Package: bslib
Title: Custom 'Bootstrap' 'Sass' Themes for 'shiny' and 'rmarkdown'
Version: 0.4.1
Authors@R: c(
    person("Carson", "Sievert", role = c("aut", "cre"), email = "carson@rstudio.com", comment = c(ORCID = "0000-0002-4958-2844")),
    person("Joe", "Cheng", role = "aut", email = "joe@rstudio.com"),
    person(family = "RStudio", role = "cph"),
    person(family = "Bootstrap contributors", role = "ctb",
    comment = "Bootstrap library"),
    person(family = "Twitter, Inc", role = "cph",
    comment = "Bootstrap library"),
    person("Javi", "Aguilar", role = c("ctb", "cph"),
    comment = "Bootstrap colorpicker library"),
    person("Thomas", "Park", role = c("ctb", "cph"),
    comment = "Bootswatch library"),
    person(family = "PayPal", role = c("ctb", "cph"),
    comment = "Bootstrap accessibility plugin")
    )
Description: Simplifies custom 'CSS' styling of both 'shiny' and 'rmarkdown' via 'Bootstrap' 'Sass'. Supports both 'Bootstrap' 3 and 4 as well as their various 'Bootswatch' themes. An interactive widget is also provided for previewing themes in real time.
Depends: R (>= 2.10)
Imports: grDevices, htmltools (>= 0.5.2), jsonlite, sass (>= 0.4.0),
        jquerylib (>= 0.1.3), rlang, cachem, memoise (>= 2.0.1)
Suggests: shiny (>= 1.6.0), rmarkdown (>= 2.7), thematic, knitr,
        testthat, withr, rappdirs, curl, magrittr
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.2.1
Collate: 'bootswatch.R' 'bs-current-theme.R' 'bs-dependencies.R'
        'bs-global.R' 'bs-remove.R' 'bs-theme-layers.R' 'utils.R'
        'bs-theme-preview.R' 'bs-theme-update.R' 'bs-theme.R'
        'deprecated.R' 'files.R' 'imports.R' 'nav-items.R'
        'nav-update.R' 'navs-legacy.R' 'navs.R' 'onLoad.R' 'page.R'
        'precompiled.R' 'print.R' 'shiny-devmode.R' 'staticimports.R'
        'utils-shiny.R' 'utils-tags.R' 'version-default.R' 'versions.R'
URL: https://rstudio.github.io/bslib/, https://github.com/rstudio/bslib
BugReports: https://github.com/rstudio/bslib/issues
Config/testthat/edition: 3
Config/Needs/routine: desc, renv
Config/Needs/website: brio, dplyr, DT, glue, purrr, rprojroot,
        rstudio/quillt, stringr, tidyr
Config/Needs/deploy: BH, DT, ggplot2, hexbin, lattice, reactable,
        rprojroot, rsconnect
NeedsCompilation: no
Packaged: 2022-11-02 20:21:31 UTC; cpsievert
Author: Carson Sievert [aut, cre] (<https://orcid.org/0000-0002-4958-2844>),
  Joe Cheng [aut],
  RStudio [cph],
  Bootstrap contributors [ctb] (Bootstrap library),
  Twitter, Inc [cph] (Bootstrap library),
  Javi Aguilar [ctb, cph] (Bootstrap colorpicker library),
  Thomas Park [ctb, cph] (Bootswatch library),
  PayPal [ctb, cph] (Bootstrap accessibility plugin)
Maintainer: Carson Sievert <carson@rstudio.com>
Repository: CRAN
Date/Publication: 2022-11-02 21:10:10 UTC

Package: bslib
Title: Custom 'Bootstrap' 'Sass' Themes for 'shiny' and 'rmarkdown'
Version: 0.10.0
Authors@R: c(
    person("Carson", "Sievert", , "carson@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-4958-2844")),
    person("Joe", "Cheng", , "joe@posit.co", role = "aut"),
    person("Garrick", "Aden-Buie", , "garrick@posit.co", role = "aut",
           comment = c(ORCID = "0000-0002-7111-0077")),
    person("Posit Software, PBC", role = c("cph", "fnd")),
    person(, "Bootstrap contributors", role = "ctb",
           comment = "Bootstrap library"),
    person(, "Twitter, Inc", role = "cph",
           comment = "Bootstrap library"),
    person("Javi", "Aguilar", role = c("ctb", "cph"),
           comment = "Bootstrap colorpicker library"),
    person("Thomas", "Park", role = c("ctb", "cph"),
           comment = "Bootswatch library"),
    person(, "PayPal", role = c("ctb", "cph"),
           comment = "Bootstrap accessibility plugin")
  )
Description: Simplifies custom 'CSS' styling of both 'shiny' and
    'rmarkdown' via 'Bootstrap' 'Sass'. Supports 'Bootstrap' 3, 4 and 5 as
    well as their various 'Bootswatch' themes. An interactive widget is
    also provided for previewing themes in real time.
License: MIT + file LICENSE
URL: https://rstudio.github.io/bslib/, https://github.com/rstudio/bslib
BugReports: https://github.com/rstudio/bslib/issues
Depends: R (>= 2.10)
Imports: base64enc, cachem, fastmap (>= 1.1.1), grDevices, htmltools
        (>= 0.5.8), jquerylib (>= 0.1.3), jsonlite, lifecycle, memoise
        (>= 2.0.1), mime, rlang, sass (>= 0.4.9)
Suggests: brand.yml, bsicons, curl, fontawesome, future, ggplot2,
        knitr, lattice, magrittr, rappdirs, rmarkdown (>= 2.7), shiny
        (>= 1.11.1), testthat, thematic, tools, utils, withr, yaml
Config/Needs/deploy: BH, chiflights22, colourpicker, commonmark, cpp11,
        cpsievert/chiflights22, cpsievert/histoslider, dplyr, DT,
        ggplot2, ggridges, gt, hexbin, histoslider, htmlwidgets,
        lattice, leaflet, lubridate, markdown, modelr, plotly,
        reactable, reshape2, rprojroot, rsconnect, rstudio/shiny,
        scales, styler, tibble
Config/Needs/routine: chromote, desc, renv
Config/Needs/website: brio, crosstalk, dplyr, DT, ggplot2, glue,
        htmlwidgets, leaflet, lorem, palmerpenguins, plotly, purrr,
        rprojroot, rstudio/htmltools, scales, stringr, tidyr, webshot2
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/testthat/start-first: zzzz-bs-sass, fonts, zzz-precompile,
        theme-*, rmd-*
Encoding: UTF-8
RoxygenNote: 7.3.3
Collate: 'accordion.R' 'breakpoints.R' 'bs-current-theme.R'
        'bs-dependencies.R' 'bs-global.R' 'bs-remove.R'
        'bs-theme-layers.R' 'bs-theme-preset-bootswatch.R'
        'bs-theme-preset-brand.R' 'bs-theme-preset-builtin.R'
        'bs-theme-preset.R' 'utils.R' 'bs-theme-preview.R'
        'bs-theme-update.R' 'bs-theme.R' 'bslib-package.R' 'buttons.R'
        'card.R' 'deprecated.R' 'files.R' 'fill.R' 'imports.R'
        'input-code-editor.R' 'input-dark-mode.R' 'input-submit.R'
        'input-switch.R' 'layout.R' 'nav-items.R' 'nav-update.R'
        'navbar_options.R' 'navs-legacy.R' 'navs.R' 'onLoad.R' 'page.R'
        'popover.R' 'precompiled.R' 'print.R' 'shiny-devmode.R'
        'sidebar.R' 'staticimports.R' 'toast.R' 'tooltip.R'
        'utils-deps.R' 'utils-shiny.R' 'utils-tags.R' 'value-box.R'
        'version-default.R' 'versions.R'
NeedsCompilation: no
Packaged: 2026-01-22 00:32:06 UTC; garrick
Author: Carson Sievert [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4958-2844>),
  Joe Cheng [aut],
  Garrick Aden-Buie [aut] (ORCID:
    <https://orcid.org/0000-0002-7111-0077>),
  Posit Software, PBC [cph, fnd],
  Bootstrap contributors [ctb] (Bootstrap library),
  Twitter, Inc [cph] (Bootstrap library),
  Javi Aguilar [ctb, cph] (Bootstrap colorpicker library),
  Thomas Park [ctb, cph] (Bootswatch library),
  PayPal [ctb, cph] (Bootstrap accessibility plugin)
Maintainer: Carson Sievert <carson@posit.co>
Repository: CRAN
Date/Publication: 2026-01-26 08:10:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.tags,bslib_sidebar)
S3method(bind_task_button,ExtendedTask)
S3method(bind_task_button,default)
S3method(is_fill_item,default)
S3method(is_fill_item,htmlwidget)
S3method(is_fillable_container,default)
S3method(is_fillable_container,htmlwidget)
S3method(print,bslib_breakpoints)
S3method(print,bslib_fragment)
S3method(print,bslib_page)
S3method(print,bslib_showcase_layout)
S3method(print,bslib_value_box_theme)
S3method(save_html,bslib_fragment)
S3method(save_html,bslib_page)
export(accordion)
export(accordion_panel)
export(accordion_panel_close)
export(accordion_panel_insert)
export(accordion_panel_open)
export(accordion_panel_remove)
export(accordion_panel_set)
export(accordion_panel_update)
export(as.card_item)
export(as_fill_carrier)
export(as_fill_item)
export(as_fillable_container)
export(bind_task_button)
export(bootstrap)
export(bootstrap_sass)
export(bootswatch_themes)
export(breakpoints)
export(bs_add_declarations)
export(bs_add_functions)
export(bs_add_mixins)
export(bs_add_rules)
export(bs_add_variables)
export(bs_bundle)
export(bs_current_theme)
export(bs_dependency)
export(bs_dependency_defer)
export(bs_get_contrast)
export(bs_get_variables)
export(bs_global_add_rules)
export(bs_global_add_variables)
export(bs_global_bundle)
export(bs_global_clear)
export(bs_global_get)
export(bs_global_set)
export(bs_global_theme)
export(bs_global_theme_update)
export(bs_remove)
export(bs_retrieve)
export(bs_theme)
export(bs_theme_accent_colors)
export(bs_theme_add)
export(bs_theme_add_variables)
export(bs_theme_base_colors)
export(bs_theme_clear)
export(bs_theme_dependencies)
export(bs_theme_fonts)
export(bs_theme_get)
export(bs_theme_get_variables)
export(bs_theme_new)
export(bs_theme_preview)
export(bs_theme_set)
export(bs_theme_update)
export(bs_themer)
export(builtin_themes)
export(card)
export(card_body)
export(card_body_fill)
export(card_footer)
export(card_header)
export(card_image)
export(card_title)
export(font_collection)
export(font_face)
export(font_google)
export(font_link)
export(input_dark_mode)
export(input_switch)
export(input_task_button)
export(is.card_item)
export(is_bs_theme)
export(is_fill_carrier)
export(is_fill_item)
export(is_fillable_container)
export(layout_column_wrap)
export(layout_columns)
export(layout_sidebar)
export(nav)
export(nav_append)
export(nav_content)
export(nav_hide)
export(nav_insert)
export(nav_item)
export(nav_menu)
export(nav_panel)
export(nav_panel_hidden)
export(nav_prepend)
export(nav_remove)
export(nav_select)
export(nav_show)
export(nav_spacer)
export(navs_bar)
export(navs_hidden)
export(navs_pill)
export(navs_pill_card)
export(navs_pill_list)
export(navs_tab)
export(navs_tab_card)
export(navset_bar)
export(navset_card_pill)
export(navset_card_tab)
export(navset_card_underline)
export(navset_hidden)
export(navset_pill)
export(navset_pill_list)
export(navset_tab)
export(navset_underline)
export(page)
export(page_fill)
export(page_fillable)
export(page_fixed)
export(page_fluid)
export(page_navbar)
export(page_sidebar)
export(popover)
export(precompiled_css_path)
export(remove_all_fill)
export(run_with_themer)
export(showcase_bottom)
export(showcase_left_center)
export(showcase_top_right)
export(sidebar)
export(sidebar_toggle)
export(theme_bootswatch)
export(theme_version)
export(toggle_dark_mode)
export(toggle_popover)
export(toggle_sidebar)
export(toggle_switch)
export(toggle_tooltip)
export(tooltip)
export(update_popover)
export(update_switch)
export(update_task_button)
export(update_tooltip)
export(value_box)
export(value_box_theme)
export(version_default)
export(versions)
import(htmltools)
import(sass)
importFrom(grDevices,col2rgb)
importFrom(htmltools,as.tags)
importFrom(jquerylib,jquery_core)
importFrom(jsonlite,fromJSON)
importFrom(lifecycle,deprecated)
importFrom(rlang,":=")
importFrom(rlang,abort)
importFrom(rlang,is_missing)
importFrom(rlang,list2)
importFrom(rlang,maybe_missing)
importFrom(rlang,missing_arg)
importFrom(sass,font_collection)
importFrom(sass,font_face)
importFrom(sass,font_google)
importFrom(sass,font_link)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(tools,file_path_sans_ext)
importFrom(utils,URLencode)
importFrom(utils,download.file)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,modifyList)
importFrom(utils,packageVersion)

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/bslib)](https://cran.r-project.org/package=bslib)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R build
status](https://github.com/rstudio/bslib/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rstudio/bslib/actions)

<!-- badges: end -->

# bslib

The `bslib` R package provides a modern UI toolkit for
[Shiny](https://shiny.posit.co/) and [R
Markdown](https://rmarkdown.rstudio.com/) based on
[Bootstrap](https://getbootstrap.com/). It facilitates:

- Creation of delightful and customizable [Shiny
  dashboards](https://rstudio.github.io/bslib/articles/dashboards.html).
  - The underlying UI components (e.g.,
    [cards](https://rstudio.github.io/bslib/articles/cards.html), [value
    boxes](https://rstudio.github.io/bslib/articles/value-boxes.html),
    [sidebars](https://rstudio.github.io/bslib/articles/sidebars.html),
    etc) are also designed to work in other contexts (e.g., in R
    Markdown).
- [Custom
  theming](https://rstudio.github.io/bslib/articles/theming.html) of
  Shiny apps and R Markdown documents.
  - Apps can even be themed interactively in real-time.
- Use of modern versions of Bootstrap and
  [Bootswatch](https://bootswatch.com/)
  - Shiny and R Markdown currently default to Bootstrap 3 and may
    continue to do so to maintain backwards compatibility.

## Installation

Install the `bslib` package from CRAN as follows:

``` r
install.packages("bslib")
```

## An example

The quickest way to see what `{bslib}` can do is to run
`bslib::bs_theme_preview()` at the console. This launches a demo app
containing both `{bslib}` and `{shiny}` UI components as well as an
theming widget (`bslib::bs_themer()`) overlaid on the app. See
[real-time
theming](https://rstudio.github.io/bslib/articles/theming.html#real-time)
to learn more.

<img src="vignettes/theming/themer.gif" alt="An animation showing bslib theming app. As the user changes the Bootswatch theme and Bootstrap settings, the app's appearance changes in real-time." width="100%" style="display: block; margin: auto;" />

## Get started

Since `{bslib}` covers a wide array of uses, it has numerous getting
started articles, each focusing on a particular interest:

- [Shiny
  dashboards](https://rstudio.github.io/bslib/articles/dashboards.html)
- [Any Bootstrap
  project](https://rstudio.github.io/bslib/articles/any-project.html)
- [Theming](https://rstudio.github.io/bslib/articles/theming.html)

## Getting help

There are two main places to get help with `bslib`:

- The [RStudio community](https://community.rstudio.com) is a friendly
  place to ask any questions (be sure to add a `bslib` tag when creating
  a topic).

- [Stack Overflow](https://stackoverflow.com/questions/tagged/bslib) is
  a great source of answers to common `bslib` questions. It is also a
  great place to get help, once you have created a reproducible example
  that illustrates your problem. Use the tags
  [`[r][bslib]`](https://stackoverflow.com/questions/tagged/bslib+r) if
  you ask a question. Add the tag `[bslib]` if you are using a Shiny
  runtime.

## Code of Conduct

Please note that the bslib project is released with a [Contributor Code
of
Conduct](https://github.com/rstudio/bslib/blob/main/CODE_OF_CONDUCT.md).
By contributing to this project, you agree to abide by its terms.

# bslib <a href='https://rstudio.github.io/bslib/'><img src='man/figures/logo.png' align="right" height="138" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/bslib)](https://cran.r-project.org/package=bslib)
[![bslib status
badge](https://rstudio.r-universe.dev/badges/bslib)](https://rstudio.r-universe.dev/bslib)
[![R build
status](https://github.com/rstudio/bslib/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rstudio/bslib/actions)

<!-- badges: end -->

The `bslib` R package provides a modern UI toolkit for
[Shiny](https://shiny.posit.co/) and [R
Markdown](https://rmarkdown.rstudio.com/) based on
[Bootstrap](https://getbootstrap.com/). It facilitates:

- Creation of delightful and customizable [Shiny
  dashboards](https://rstudio.github.io/bslib/articles/dashboards.html).
  - The underlying UI components (e.g.,
    [cards](https://rstudio.github.io/bslib/articles/cards.html), [value
    boxes](https://rstudio.github.io/bslib/articles/value-boxes.html),
    [sidebars](https://rstudio.github.io/bslib/articles/sidebars.html),
    etc) are also designed to work in other contexts (e.g., in R
    Markdown).
- [Custom
  theming](https://rstudio.github.io/bslib/articles/theming.html) of
  Shiny apps and R Markdown documents.
  - Apps can even be themed interactively in real-time.
- Use of modern versions of Bootstrap and
  [Bootswatch](https://bootswatch.com/)
  - Shiny and R Markdown currently default to Bootstrap 3 and may
    continue to do so to maintain backwards compatibility.

## Installation

Install the `bslib` package from CRAN as follows:

``` r
install.packages("bslib")
```

## An example

The quickest way to see what `{bslib}` can do is to run
`bslib::bs_theme_preview()` at the console. This launches a demo app
containing both `{bslib}` and `{shiny}` UI components as well as an
theming widget (`bslib::bs_themer()`) overlaid on the app. See
[real-time
theming](https://rstudio.github.io/bslib/articles/theming.html#real-time)
to learn more.

<img src="vignettes/theming/themer.gif" alt="An animation showing bslib theming app. As the user changes the Bootswatch theme and Bootstrap settings, the app's appearance changes in real-time." width="100%" style="display: block; margin: auto;" />

## Get started

Since `{bslib}` covers a wide array of uses, it has numerous getting
started articles, each focusing on a particular interest:

- [Shiny
  dashboards](https://rstudio.github.io/bslib/articles/dashboards.html)
- [Any Bootstrap
  project](https://rstudio.github.io/bslib/articles/any-project.html)
- [Theming](https://rstudio.github.io/bslib/articles/theming.html)

## Getting help

There are two main places to get help with `bslib`:

- The [RStudio community](https://forum.posit.co/) is a friendly place
  to ask any questions (be sure to add a `bslib` tag when creating a
  topic).

- [Stack Overflow](https://stackoverflow.com/questions/tagged/bslib) is
  a great source of answers to common `bslib` questions. It is also a
  great place to get help, once you have created a reproducible example
  that illustrates your problem. Use the tags
  [`[r][bslib]`](https://stackoverflow.com/questions/tagged/bslib+r) if
  you ask a question. Add the tag `[bslib]` if you are using a Shiny
  runtime.

## Code of Conduct

Please note that the bslib project is released with a [Contributor Code
of
Conduct](https://github.com/rstudio/bslib/blob/main/CODE_OF_CONDUCT.md).
By contributing to this project, you agree to abide by its terms.

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/bslib)](https://cran.r-project.org/package=bslib)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R build
status](https://github.com/rstudio/bslib/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rstudio/bslib/actions)

<!-- badges: end -->

<style type="text/css">
@media (min-width: 800px){.usage{display:flex}.usage *{flex:1}}
</style>

# bslib

The `bslib` R package provides tools for customizing [Bootstrap
themes](https://getbootstrap.com/docs/4.6/getting-started/theming/)
directly from R, making it much easier to customize the appearance of
[Shiny](https://shiny.rstudio.com/) apps & [R
Markdown](https://rmarkdown.rstudio.com/) documents. `bslib`’s primary
goals are:

  - Make [custom
    theming](https://rstudio.github.io/bslib/articles/bslib.html#custom)
    as easy as possible.
      - Custom themes may even be created interactively in
        [real-time](https://rstudio.github.io/bslib/articles/bslib.html#real-time).
  - Also provide easy access to pre-packaged [Bootswatch
    themes](https://rstudio.github.io/bslib/articles/bslib.html#bootswatch).
  - Make [upgrading from Bootstrap 3 to 4 (and
    beyond)](https://rstudio.github.io/bslib/articles/bslib.html#versions)
    as seamless as possible.
      - Shiny and R Markdown default to Bootstrap 3 and will continue to
        do so to avoid breaking legacy code.
  - Serve as a general foundation for Shiny and R Markdown extension
    packages.
      - Extensions such as
        [`flexdashboard`](https://flexdashboard-pkg.netlify.app/articles/theme.html),
        [`pkgdown`](https://pkgdown.r-lib.org/dev/articles/customise.html),
        and
        [`bookdown`](https://pkgs.rstudio.com/bookdown/reference/bs4_book.html)
        already fully support [`bslib`’s custom theming
        capabilities](https://rstudio.github.io/bslib/articles/bslib.html#custom).

## Installation

Install the stable release of `bslib` on CRAN:

``` r
install.packages("bslib")
```

Usage with Shiny requires version 1.6 or higher:

``` r
install.packages("shiny")
```

Usage with R Markdown requires version 2.7 or higher:

``` r
install.packages("rmarkdown")
```

## Basic usage

`bslib` is designed for use with any Shiny or R Markdown project that
uses Bootstrap. In most cases, you can identify a project that uses
Bootstrap when the relevant page constructor has a `theme` parameter.
For example, most Shiny page layout functions (e.g.,
`shiny::navbarPage()`) and some popular R Markdown formats (e.g.,
`rmarkdown::html_document`) all have a `theme` parameter.

To use `bslib` in Shiny, provide a `bs_theme()` *object* to the `theme`
parameter; and in R Markdown, provide `bs_theme()` *parameters* to
`theme`. For example, here’s a way to upgrade Shiny (left) and R
Markdown (right) from Bootstrap 3 to 5:

<div class="usage">

``` r
library(shiny)
ui <- navbarPage(
  theme = bs_theme(version = 5),
  ...
)
shinyApp(ui, function(...) {})
```

``` r
---
output:
  html_document:
    theme:
      version: 5
---
```

</div>

See the [Get
Started](https://rstudio.github.io/bslib/articles/bslib.html) article to
learn more about Bootstrap versions, pre-packaged Bootswatch themes,
(real-time) custom theming, and more.

To get started more quickly, choose a relevant R Markdown template from
inside RStudio by going to File -\> New File -\> R Markdown -\> From
Template:

<img src="man/figures/rstudio-templates.png" width="60%" style="display: block; margin: auto;" />

## Getting help

There are two main places to get help with `bslib`:

  - The [RStudio community](https://community.rstudio.com) is a friendly
    place to ask any questions (be sure to add a `bslib` tag when
    creating a topic).

  - [Stack Overflow](https://stackoverflow.com/questions/tagged/bslib)
    is a great source of answers to common `bslib` questions. It is also
    a great place to get help, once you have created a reproducible
    example that illustrates your problem. Use the tags
    [`[r][bslib]`](https://stackoverflow.com/questions/tagged/bslib+r)
    if you ask a question. Add the tag `[bslib]` if you are using a
    Shiny runtime.

## Code of Conduct

Please note that the bslib project is released with a [Contributor Code
of
Conduct](https://github.com/rstudio/bslib/blob/main/CODE_OF_CONDUCT.md).
By contributing to this project, you agree to abide by its terms.

<p align="center">
  <a href="https://getbootstrap.com/">
    <img src="https://getbootstrap.com/docs/5.3/assets/brand/bootstrap-logo-shadow.png" alt="Bootstrap logo" width="200" height="165">
  </a>
</p>

<h3 align="center">Bootstrap</h3>

<p align="center">
  Sleek, intuitive, and powerful front-end framework for faster and easier web development.
  <br>
  <a href="https://getbootstrap.com/docs/5.3/"><strong>Explore Bootstrap docs »</strong></a>
  <br>
  <br>
  <a href="https://github.com/twbs/bootstrap/issues/new?assignees=-&labels=bug&template=bug_report.yml">Report bug</a>
  ·
  <a href="https://github.com/twbs/bootstrap/issues/new?assignees=&labels=feature&template=feature_request.yml">Request feature</a>
  ·
  <a href="https://themes.getbootstrap.com/">Themes</a>
  ·
  <a href="https://blog.getbootstrap.com/">Blog</a>
</p>


## Bootstrap 5

Our default branch is for development of our Bootstrap 5 release. Head to the [`v4-dev` branch](https://github.com/twbs/bootstrap/tree/v4-dev) to view the readme, documentation, and source code for Bootstrap 4.


## Table of contents

- [Quick start](#quick-start)
- [Status](#status)
- [What's included](#whats-included)
- [Bugs and feature requests](#bugs-and-feature-requests)
- [Documentation](#documentation)
- [Contributing](#contributing)
- [Community](#community)
- [Versioning](#versioning)
- [Creators](#creators)
- [Thanks](#thanks)
- [Copyright and license](#copyright-and-license)


## Quick start

Several quick start options are available:

- [Download the latest release](https://github.com/twbs/bootstrap/archive/v5.3.1.zip)
- Clone the repo: `git clone https://github.com/twbs/bootstrap.git`
- Install with [npm](https://www.npmjs.com/): `npm install bootstrap@v5.3.1`
- Install with [yarn](https://yarnpkg.com/): `yarn add bootstrap@v5.3.1`
- Install with [Composer](https://getcomposer.org/): `composer require twbs/bootstrap:5.3.1`
- Install with [NuGet](https://www.nuget.org/): CSS: `Install-Package bootstrap` Sass: `Install-Package bootstrap.sass`

Read the [Getting started page](https://getbootstrap.com/docs/5.3/getting-started/introduction/) for information on the framework contents, templates, examples, and more.


## Status

[![Build Status](https://img.shields.io/github/actions/workflow/status/twbs/bootstrap/js.yml?branch=main&label=JS%20Tests&logo=github)](https://github.com/twbs/bootstrap/actions/workflows/js.yml?query=workflow%3AJS+branch%3Amain)
[![npm version](https://img.shields.io/npm/v/bootstrap?logo=npm&logoColor=fff)](https://www.npmjs.com/package/bootstrap)
[![Gem version](https://img.shields.io/gem/v/bootstrap?logo=rubygems&logoColor=fff)](https://rubygems.org/gems/bootstrap)
[![Meteor Atmosphere](https://img.shields.io/badge/meteor-twbs%3Abootstrap-blue?logo=meteor&logoColor=fff)](https://atmospherejs.com/twbs/bootstrap)
[![Packagist Prerelease](https://img.shields.io/packagist/vpre/twbs/bootstrap?logo=packagist&logoColor=fff)](https://packagist.org/packages/twbs/bootstrap)
[![NuGet](https://img.shields.io/nuget/vpre/bootstrap?logo=nuget&logoColor=fff)](https://www.nuget.org/packages/bootstrap/absoluteLatest)
[![Coverage Status](https://img.shields.io/coveralls/github/twbs/bootstrap/main?logo=coveralls&logoColor=fff)](https://coveralls.io/github/twbs/bootstrap?branch=main)
[![CSS gzip size](https://img.badgesize.io/twbs/bootstrap/main/dist/css/bootstrap.min.css?compression=gzip&label=CSS%20gzip%20size)](https://github.com/twbs/bootstrap/blob/main/dist/css/bootstrap.min.css)
[![CSS Brotli size](https://img.badgesize.io/twbs/bootstrap/main/dist/css/bootstrap.min.css?compression=brotli&label=CSS%20Brotli%20size)](https://github.com/twbs/bootstrap/blob/main/dist/css/bootstrap.min.css)
[![JS gzip size](https://img.badgesize.io/twbs/bootstrap/main/dist/js/bootstrap.min.js?compression=gzip&label=JS%20gzip%20size)](https://github.com/twbs/bootstrap/blob/main/dist/js/bootstrap.min.js)
[![JS Brotli size](https://img.badgesize.io/twbs/bootstrap/main/dist/js/bootstrap.min.js?compression=brotli&label=JS%20Brotli%20size)](https://github.com/twbs/bootstrap/blob/main/dist/js/bootstrap.min.js)
[![BrowserStack Status](https://www.browserstack.com/automate/badge.svg?badge_key=SkxZcStBeExEdVJqQ2hWYnlWckpkNmNEY213SFp6WHFETWk2bGFuY3pCbz0tLXhqbHJsVlZhQnRBdEpod3NLSDMzaHc9PQ==--3d0b75245708616eb93113221beece33e680b229)](https://www.browserstack.com/automate/public-build/SkxZcStBeExEdVJqQ2hWYnlWckpkNmNEY213SFp6WHFETWk2bGFuY3pCbz0tLXhqbHJsVlZhQnRBdEpod3NLSDMzaHc9PQ==--3d0b75245708616eb93113221beece33e680b229)
[![Backers on Open Collective](https://img.shields.io/opencollective/backers/bootstrap?logo=opencollective&logoColor=fff)](#backers)
[![Sponsors on Open Collective](https://img.shields.io/opencollective/sponsors/bootstrap?logo=opencollective&logoColor=fff)](#sponsors)


## What's included

Within the download you'll find the following directories and files, logically grouping common assets and providing both compiled and minified variations.

<details>
  <summary>Download contents</summary>

  ```text
  bootstrap/
  ├── css/
  │   ├── bootstrap-grid.css
  │   ├── bootstrap-grid.css.map
  │   ├── bootstrap-grid.min.css
  │   ├── bootstrap-grid.min.css.map
  │   ├── bootstrap-grid.rtl.css
  │   ├── bootstrap-grid.rtl.css.map
  │   ├── bootstrap-grid.rtl.min.css
  │   ├── bootstrap-grid.rtl.min.css.map
  │   ├── bootstrap-reboot.css
  │   ├── bootstrap-reboot.css.map
  │   ├── bootstrap-reboot.min.css
  │   ├── bootstrap-reboot.min.css.map
  │   ├── bootstrap-reboot.rtl.css
  │   ├── bootstrap-reboot.rtl.css.map
  │   ├── bootstrap-reboot.rtl.min.css
  │   ├── bootstrap-reboot.rtl.min.css.map
  │   ├── bootstrap-utilities.css
  │   ├── bootstrap-utilities.css.map
  │   ├── bootstrap-utilities.min.css
  │   ├── bootstrap-utilities.min.css.map
  │   ├── bootstrap-utilities.rtl.css
  │   ├── bootstrap-utilities.rtl.css.map
  │   ├── bootstrap-utilities.rtl.min.css
  │   ├── bootstrap-utilities.rtl.min.css.map
  │   ├── bootstrap.css
  │   ├── bootstrap.css.map
  │   ├── bootstrap.min.css
  │   ├── bootstrap.min.css.map
  │   ├── bootstrap.rtl.css
  │   ├── bootstrap.rtl.css.map
  │   ├── bootstrap.rtl.min.css
  │   └── bootstrap.rtl.min.css.map
  └── js/
      ├── bootstrap.bundle.js
      ├── bootstrap.bundle.js.map
      ├── bootstrap.bundle.min.js
      ├── bootstrap.bundle.min.js.map
      ├── bootstrap.esm.js
      ├── bootstrap.esm.js.map
      ├── bootstrap.esm.min.js
      ├── bootstrap.esm.min.js.map
      ├── bootstrap.js
      ├── bootstrap.js.map
      ├── bootstrap.min.js
      └── bootstrap.min.js.map
  ```
</details>

We provide compiled CSS and JS (`bootstrap.*`), as well as compiled and minified CSS and JS (`bootstrap.min.*`). [Source maps](https://developers.google.com/web/tools/chrome-devtools/javascript/source-maps) (`bootstrap.*.map`) are available for use with certain browsers' developer tools. Bundled JS files (`bootstrap.bundle.js` and minified `bootstrap.bundle.min.js`) include [Popper](https://popper.js.org/).


## Bugs and feature requests

Have a bug or a feature request? Please first read the [issue guidelines](https://github.com/twbs/bootstrap/blob/main/.github/CONTRIBUTING.md#using-the-issue-tracker) and search for existing and closed issues. If your problem or idea is not addressed yet, [please open a new issue](https://github.com/twbs/bootstrap/issues/new/choose).


## Documentation

Bootstrap's documentation, included in this repo in the root directory, is built with [Hugo](https://gohugo.io/) and publicly hosted on GitHub Pages at <https://getbootstrap.com/>. The docs may also be run locally.

Documentation search is powered by [Algolia's DocSearch](https://docsearch.algolia.com/).

### Running documentation locally

1. Run `npm install` to install the Node.js dependencies, including Hugo (the site builder).
2. Run `npm run test` (or a specific npm script) to rebuild distributed CSS and JavaScript files, as well as our docs assets.
3. From the root `/bootstrap` directory, run `npm run docs-serve` in the command line.
4. Open `http://localhost:9001/` in your browser, and voilà.

Learn more about using Hugo by reading its [documentation](https://gohugo.io/documentation/).

### Documentation for previous releases

You can find all our previous releases docs on <https://getbootstrap.com/docs/versions/>.

[Previous releases](https://github.com/twbs/bootstrap/releases) and their documentation are also available for download.


## Contributing

Please read through our [contributing guidelines](https://github.com/twbs/bootstrap/blob/main/.github/CONTRIBUTING.md). Included are directions for opening issues, coding standards, and notes on development.

Moreover, if your pull request contains JavaScript patches or features, you must include [relevant unit tests](https://github.com/twbs/bootstrap/tree/main/js/tests). All HTML and CSS should conform to the [Code Guide](https://github.com/mdo/code-guide), maintained by [Mark Otto](https://github.com/mdo).

Editor preferences are available in the [editor config](https://github.com/twbs/bootstrap/blob/main/.editorconfig) for easy use in common text editors. Read more and download plugins at <https://editorconfig.org/>.


## Community

Get updates on Bootstrap's development and chat with the project maintainers and community members.

- Follow [@getbootstrap on Twitter](https://twitter.com/getbootstrap).
- Read and subscribe to [The Official Bootstrap Blog](https://blog.getbootstrap.com/).
- Ask and explore [our GitHub Discussions](https://github.com/twbs/bootstrap/discussions).
- Chat with fellow Bootstrappers in IRC. On the `irc.libera.chat` server, in the `#bootstrap` channel.
- Implementation help may be found at Stack Overflow (tagged [`bootstrap-5`](https://stackoverflow.com/questions/tagged/bootstrap-5)).
- Developers should use the keyword `bootstrap` on packages which modify or add to the functionality of Bootstrap when distributing through [npm](https://www.npmjs.com/browse/keyword/bootstrap) or similar delivery mechanisms for maximum discoverability.


## Versioning

For transparency into our release cycle and in striving to maintain backward compatibility, Bootstrap is maintained under [the Semantic Versioning guidelines](https://semver.org/). Sometimes we screw up, but we adhere to those rules whenever possible.

See [the Releases section of our GitHub project](https://github.com/twbs/bootstrap/releases) for changelogs for each release version of Bootstrap. Release announcement posts on [the official Bootstrap blog](https://blog.getbootstrap.com/) contain summaries of the most noteworthy changes made in each release.


## Creators

**Mark Otto**

- <https://twitter.com/mdo>
- <https://github.com/mdo>

**Jacob Thornton**

- <https://twitter.com/fat>
- <https://github.com/fat>


## Thanks

<a href="https://www.browserstack.com/">
  <img src="https://live.browserstack.com/images/opensource/browserstack-logo.svg" alt="BrowserStack" width="192" height="42">
</a>

Thanks to [BrowserStack](https://www.browserstack.com/) for providing the infrastructure that allows us to test in real browsers!

<a href="https://www.netlify.com/">
  <img src="https://www.netlify.com/v3/img/components/full-logo-light.svg" alt="Netlify" width="147" height="40">
</a>

Thanks to [Netlify](https://www.netlify.com/) for providing us with Deploy Previews!


## Sponsors

Support this project by becoming a sponsor. Your logo will show up here with a link to your website. [[Become a sponsor](https://opencollective.com/bootstrap#sponsor)]

[![OC sponsor 0](https://opencollective.com/bootstrap/sponsor/0/avatar.svg)](https://opencollective.com/bootstrap/sponsor/0/website)
[![OC sponsor 1](https://opencollective.com/bootstrap/sponsor/1/avatar.svg)](https://opencollective.com/bootstrap/sponsor/1/website)
[![OC sponsor 2](https://opencollective.com/bootstrap/sponsor/2/avatar.svg)](https://opencollective.com/bootstrap/sponsor/2/website)
[![OC sponsor 3](https://opencollective.com/bootstrap/sponsor/3/avatar.svg)](https://opencollective.com/bootstrap/sponsor/3/website)
[![OC sponsor 4](https://opencollective.com/bootstrap/sponsor/4/avatar.svg)](https://opencollective.com/bootstrap/sponsor/4/website)
[![OC sponsor 5](https://opencollective.com/bootstrap/sponsor/5/avatar.svg)](https://opencollective.com/bootstrap/sponsor/5/website)
[![OC sponsor 6](https://opencollective.com/bootstrap/sponsor/6/avatar.svg)](https://opencollective.com/bootstrap/sponsor/6/website)
[![OC sponsor 7](https://opencollective.com/bootstrap/sponsor/7/avatar.svg)](https://opencollective.com/bootstrap/sponsor/7/website)
[![OC sponsor 8](https://opencollective.com/bootstrap/sponsor/8/avatar.svg)](https://opencollective.com/bootstrap/sponsor/8/website)
[![OC sponsor 9](https://opencollective.com/bootstrap/sponsor/9/avatar.svg)](https://opencollective.com/bootstrap/sponsor/9/website)


## Backers

Thank you to all our backers! 🙏 [[Become a backer](https://opencollective.com/bootstrap#backer)]

[![Backers](https://opencollective.com/bootstrap/backers.svg?width=890)](https://opencollective.com/bootstrap#backers)


## Copyright and license

Code and documentation copyright 2011–2023 the [Bootstrap Authors](https://github.com/twbs/bootstrap/graphs/contributors). Code released under the [MIT License](https://github.com/twbs/bootstrap/blob/main/LICENSE). Docs released under [Creative Commons](https://creativecommons.org/licenses/by/3.0/).

## Upgrade Everything

Rather than individually run each update script and worry about the correct order, use the `tools/update_all.R` script:

``` shell
Rscript tools/main.R
```


## Upgrading HTML dependencies

To upgrade or install the HTML dependencies that ship with this package, use the `tools/yarn_install.R` script:

``` shell
Rscript tools/yarn_install.R
```

**Warning:** Upgrading to a new major version of Bootstrap won't be as simple as running this R script. We'll cross that bridge when we get there.

## Local fonts

Bootswatch themes import remote google font urls by default. There is a script
to download all these fonts and ship them with the package...this script be run after the `yarn_install.R` script

``` shell
Rscript tools/fonts.R
```

<a class="readme-logo" href="https://itsjavi.com/bootstrap-colorpicker">
    <img alt="Logo by @reallinfo" src="logo.png" width="128px" />
</a>

# Bootstrap Colorpicker

<!--version-->

[Bootstrap Colorpicker](https://github.com/itsjavi/bootstrap-colorpicker/) is a modular color picker plugin for Bootstrap 4.

[![Build Status](https://img.shields.io/travis/itsjavi/bootstrap-colorpicker/master.svg?style=flat-square)](https://travis-ci.org/itsjavi/bootstrap-colorpicker)
[![npm](https://img.shields.io/npm/v/bootstrap-colorpicker.svg?style=flat-square)](https://www.npmjs.com/package/bootstrap-colorpicker)

## Install
You can get the latest version in many different ways:

- Downloading the tarball from npm the registry: https://registry.npmjs.org/bootstrap-colorpicker/-/bootstrap-colorpicker-3.2.0.tgz  (you can change the version in the url to any released tag)
- Cloning using Git: `git clone https://github.com/itsjavi/bootstrap-colorpicker.git`
- Installing via NPM: `npm install bootstrap-colorpicker`
- Installing via Yarn: `yarn add bootstrap-colorpicker`
- Installing via Composer: `composer require itsjavi/bootstrap-colorpicker`

Note that the `dist` files are only distributed via the NPM and Yarn installations.

For the rest methods, you will need to generate the files initializing the project with `yarn install`
and then building the code using `npm run build`.

## Versions

<table class="table table-bordered table-striped">
  <thead>
    <tr>
        <th>Colorpicker version</th>
        <th>Compatible Bootstrap version</th>
        <th>Dependencies</th>
    </tr>
  </thead>
  <tbody>
    <tr>
        <td>
          <a href="https://github.com/itsjavi/bootstrap-colorpicker/tree/v2.x">v2.x</a> <br>
          <a href="https://itsjavi.com/bootstrap-colorpicker/v2">Documentation</a>
        </td>
        <td>Bootstrap 3 or 4</td>
        <td>
          <ul>
            <li>jQuery >= 1.10</li>
            <li>Bootstrap CSS (input addon)</li>
          </ul>
        </td>
    </tr>
    <tr>
        <td>
          <a href="https://github.com/itsjavi/bootstrap-colorpicker">v3.x</a> <br>
          <a href="https://itsjavi.com/bootstrap-colorpicker">Documentation</a>
        </td>
        <td>Bootstrap 4 or without Bootstrap</td>
        <td>
          <ul>
            <li>jQuery >= 2.1.0</li>
            <li>Bootstrap CSS (input addon, popover)</li>
            <li>Bootstrap JS Bundle (popover)</li>
          </ul>
        </td>
    </tr>
  </thead>
</table>


Note that the plugin may work without Bootstrap if your code is not using any of the mentioned Bootstrap
dependencies.


## Examples

### With Bootstrap
The Bootstrap JS dependency is optional and it is mainly needed for the popover support.
No Bootstrap CSS is required for the plugin to work.

```html
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <link href="dist/css/bootstrap-colorpicker.css" rel="stylesheet">
</head>
<body>
  <div class="demo">
      <h1>Bootstrap Colorpicker Demo (with Bootstrap)</h1>
      <input id="demo-input" type="text" value="rgb(255, 128, 0)" />
  </div>
  <script src="//code.jquery.com/jquery-3.4.1.js"></script>
  <script src="//unpkg.com/bootstrap@4.3.1/dist/js/bootstrap.bundle.min.js"></script>
  <script src="dist/js/bootstrap-colorpicker.js"></script>
  <script>
    $(function () {
      // Basic instantiation:
      $('#demo-input').colorpicker();
      
      // Example using an event, to change the color of the #demo div background:
      $('#demo-input').on('colorpickerChange', function(event) {
        $('#demo').css('background-color', event.color.toString());
      });
    });
  </script>
</body>
```

### Without Bootstrap

To use the plugin without Bootstrap, the `popover` option should be set to `false` or `null` and, depending on your implementation,
you will usually need to set inline to `true` and a `container` selector option.

```html
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <link href="dist/css/bootstrap-colorpicker.css" rel="stylesheet">
</head>
<body>
  <div id="demo">
      <h1>Bootstrap Colorpicker Demo (without Bootstrap)</h1>
    <input type="text" value="rgb(255, 128, 0)" />
  </div>
  <script src="//code.jquery.com/jquery-3.4.1.js"></script>
  <script src="dist/js/bootstrap-colorpicker.js"></script>
  <script>
    $(function() {
      $('#demo').colorpicker({
        popover: false,
        inline: true,
        container: '#demo'
      });
    });
  </script>
</body>
```

## Contributions
* [Issues](https://github.com/itsjavi/bootstrap-colorpicker/issues)
* [Pull Requests](https://github.com/itsjavi/bootstrap-colorpicker/pulls)
* [Milestones](https://github.com/itsjavi/bootstrap-colorpicker/milestones)

This project exists thanks to all the [people who contribute](https://github.com/itsjavi/bootstrap-colorpicker/graphs/contributors).

Please read [CONTRIBUTING](https://github.com/itsjavi/bootstrap-colorpicker/blob/master/.github/CONTRIBUTING.md) 
before sending a pull request or issue.

## License
The MIT License (MIT).
Please see the [License File](https://github.com/itsjavi/bootstrap-colorpicker/blob/master/LICENSE) for more information.

## Credits

Written and maintained by [Javi Aguilar](https://itsjavi.com) and all other contributors.

*Based on Stefan Petre's color picker (2013).*

*Thanks to JetBrains for supporting this project.*

# Bootstrap 3 for Sass
[![Gem Version](https://badge.fury.io/rb/bootstrap-sass.svg)](http://badge.fury.io/rb/bootstrap-sass)
[![npm version](https://img.shields.io/npm/v/bootstrap-sass.svg?style=flat)](https://www.npmjs.com/package/bootstrap-sass)
[![Bower Version](https://badge.fury.io/bo/bootstrap-sass.svg)](http://badge.fury.io/bo/bootstrap-sass)
[![Build Status](https://img.shields.io/travis/twbs/bootstrap-sass.svg)](https://travis-ci.org/twbs/bootstrap-sass)

`bootstrap-sass` is a Sass-powered version of [Bootstrap](https://github.com/twbs/bootstrap) 3, ready to drop right into your Sass powered applications.

This is Bootstrap **3**. For Bootstrap **4** use the [Bootstrap rubygem](https://github.com/twbs/bootstrap-rubygem) if you use Ruby, and the [main repo](https://github.com/twbs/bootstrap) otherwise.

## Installation

Please see the appropriate guide for your environment of choice:

* [Ruby on Rails](#a-ruby-on-rails).
* [Bower](#b-bower).
* [npm / Node.js](#c-npm--nodejs).

### a. Ruby on Rails

`bootstrap-sass` is easy to drop into Rails with the asset pipeline.

In your Gemfile you need to add the `bootstrap-sass` gem, and ensure that the `sass-rails` gem is present - it is added to new Rails applications by default.

```ruby
gem 'bootstrap-sass', '~> 3.4.1'
gem 'sassc-rails', '>= 2.1.0'
```

`bundle install` and restart your server to make the files available through the pipeline.

Import Bootstrap styles in `app/assets/stylesheets/application.scss`:

```scss
// "bootstrap-sprockets" must be imported before "bootstrap" and "bootstrap/variables"
@import "bootstrap-sprockets";
@import "bootstrap";
```

`bootstrap-sprockets` must be imported before `bootstrap` for the icon fonts to work.

Make sure the file has `.scss` extension (or `.sass` for Sass syntax). If you have just generated a new Rails app,
it may come with a `.css` file instead. If this file exists, it will be served instead of Sass, so rename it:

```console
$ mv app/assets/stylesheets/application.css app/assets/stylesheets/application.scss
```

Then, remove all the `*= require_self` and `*= require_tree .` statements from the sass file. Instead, use `@import` to import Sass files.

Do not use `*= require` in Sass or your other stylesheets will not be [able to access][antirequire] the Bootstrap mixins or variables.

Bootstrap JavaScript depends on jQuery.
If you're using Rails 5.1+, add the `jquery-rails` gem to your Gemfile:

```ruby
gem 'jquery-rails'
```

```console
$ bundle install
```

Require Bootstrap Javascripts in `app/assets/javascripts/application.js`:

```js
//= require jquery
//= require bootstrap-sprockets
```

`bootstrap-sprockets` and `bootstrap` [should not both be included](https://github.com/twbs/bootstrap-sass/issues/829#issuecomment-75153827) in `application.js`.

`bootstrap-sprockets` provides individual Bootstrap Javascript files (`alert.js` or `dropdown.js`, for example), while
`bootstrap` provides a concatenated file containing all Bootstrap Javascripts.

#### Bower with Rails

When using [bootstrap-sass Bower package](#c-bower) instead of the gem in Rails, configure assets in `config/application.rb`:

```ruby
# Bower asset paths
root.join('vendor', 'assets', 'bower_components').to_s.tap do |bower_path|
  config.sass.load_paths << bower_path
  config.assets.paths << bower_path
end
# Precompile Bootstrap fonts
config.assets.precompile << %r(bootstrap-sass/assets/fonts/bootstrap/[\w-]+\.(?:eot|svg|ttf|woff2?)$)
# Minimum Sass number precision required by bootstrap-sass
::Sass::Script::Value::Number.precision = [8, ::Sass::Script::Value::Number.precision].max
```

Replace Bootstrap `@import` statements in `application.scss` with:

```scss
$icon-font-path: "bootstrap-sass/assets/fonts/bootstrap/";
@import "bootstrap-sass/assets/stylesheets/bootstrap-sprockets";
@import "bootstrap-sass/assets/stylesheets/bootstrap";
```

Replace Bootstrap `require` directive in `application.js` with:

```js
//= require bootstrap-sass/assets/javascripts/bootstrap-sprockets
```

#### Rails 4.x

Please make sure `sprockets-rails` is at least v2.1.4.

#### Rails 3.2.x

bootstrap-sass is no longer compatible with Rails 3. The latest version of bootstrap-sass compatible with Rails 3.2 is v3.1.1.0.

### b. Bower

bootstrap-sass Bower package is compatible with node-sass 3.2.0+. You can install it with:

```console
$ bower install bootstrap-sass
```

Sass, JS, and all other assets are located at [assets](/assets).

By default, `bower.json` main field list only the main `_bootstrap.scss` and all the static assets (fonts and JS).
This is compatible by default with asset managers such as [wiredep](https://github.com/taptapship/wiredep).

#### Node.js Mincer

If you use [mincer][mincer] with node-sass, import Bootstrap like so:

In `application.css.ejs.scss` (NB **.css.ejs.scss**):

```scss
// Import mincer asset paths helper integration
@import "bootstrap-mincer";
@import "bootstrap";
```

In `application.js`:

```js
//= require bootstrap-sprockets
```

See also this [example manifest.js](/test/dummy_node_mincer/manifest.js) for mincer.

### c. npm / Node.js
```console
$ npm install bootstrap-sass
```


## Configuration

### Sass

By default all of Bootstrap is imported.

You can also import components explicitly. To start with a full list of modules copy
[`_bootstrap.scss`](assets/stylesheets/_bootstrap.scss) file into your assets as `_bootstrap-custom.scss`.
Then comment out components you do not want from `_bootstrap-custom`.
In the application Sass file, replace `@import 'bootstrap'` with:

```scss
@import 'bootstrap-custom';
```

### Sass: Number Precision

bootstrap-sass [requires](https://github.com/twbs/bootstrap-sass/issues/409) minimum [Sass number precision][sass-precision] of 8 (default is 5).

Precision is set for Ruby automatically when using the `sassc-rails` gem.
When using the npm or Bower version with Ruby, you can set it with:

```ruby
::Sass::Script::Value::Number.precision = [8, ::Sass::Script::Value::Number.precision].max
```

### Sass: Autoprefixer

Bootstrap requires the use of [Autoprefixer][autoprefixer].
[Autoprefixer][autoprefixer] adds vendor prefixes to CSS rules using values from [Can I Use](https://caniuse.com/).

To match [upstream Bootstrap's level of browser compatibility](https://getbootstrap.com/getting-started/#support), set Autoprefixer's `browsers` option to:
```json
[
  "Android 2.3",
  "Android >= 4",
  "Chrome >= 20",
  "Firefox >= 24",
  "Explorer >= 8",
  "iOS >= 6",
  "Opera >= 12",
  "Safari >= 6"
]
```

### JavaScript

[`assets/javascripts/bootstrap.js`](/assets/javascripts/bootstrap.js) contains all of Bootstrap's JavaScript,
concatenated in the [correct order](/assets/javascripts/bootstrap-sprockets.js).


#### JavaScript with Sprockets or Mincer

If you use Sprockets or Mincer, you can require `bootstrap-sprockets` instead to load the individual modules:

```js
// Load all Bootstrap JavaScript
//= require bootstrap-sprockets
```

You can also load individual modules, provided you also require any dependencies.
You can check dependencies in the [Bootstrap JS documentation][jsdocs].

```js
//= require bootstrap/scrollspy
//= require bootstrap/modal
//= require bootstrap/dropdown
```

### Fonts

The fonts are referenced as:

```scss
"#{$icon-font-path}#{$icon-font-name}.eot"
```

`$icon-font-path` defaults to `bootstrap/` if asset path helpers are used, and `../fonts/bootstrap/` otherwise.

When using bootstrap-sass with Compass, Sprockets, or Mincer, you **must** import the relevant path helpers before Bootstrap itself, for example:

```scss
@import "bootstrap-compass";
@import "bootstrap";
```

## Usage

### Sass

Import Bootstrap into a Sass file (for example, `application.scss`) to get all of Bootstrap's styles, mixins and variables!

```scss
@import "bootstrap";
```

You can also include optional Bootstrap theme:

```scss
@import "bootstrap/theme";
```

The full list of Bootstrap variables can be found [here](https://getbootstrap.com/customize/#less-variables). You can override these by simply redefining the variable before the `@import` directive, e.g.:

```scss
$navbar-default-bg: #312312;
$light-orange: #ff8c00;
$navbar-default-color: $light-orange;

@import "bootstrap";
```

### Eyeglass

Bootstrap is available as an [Eyeglass](https://github.com/sass-eyeglass/eyeglass) module. After installing Bootstrap via NPM you can import the Bootstrap library via:

```scss
@import "bootstrap-sass/bootstrap"
```

or import only the parts of Bootstrap you need:

```scss
@import "bootstrap-sass/bootstrap/variables";
@import "bootstrap-sass/bootstrap/mixins";
@import "bootstrap-sass/bootstrap/carousel";
```

## Version

Bootstrap for Sass version may differ from the upstream version in the last number, known as
[PATCH](https://semver.org/spec/v2.0.0.html). The patch version may be ahead of the corresponding upstream minor.
This happens when we need to release Sass-specific changes.

Before v3.3.2, Bootstrap for Sass version used to reflect the upstream version, with an additional number for
Sass-specific changes. This was changed due to Bower and npm compatibility issues.

The upstream versions vs the Bootstrap for Sass versions are:

| Upstream |    Sass |
|---------:|--------:|
|    3.3.4+ |   same |
|    3.3.2 |   3.3.3 |
| <= 3.3.1 | 3.3.1.x |

Always refer to [CHANGELOG.md](/CHANGELOG.md) when upgrading.

---

## Development and Contributing

If you'd like to help with the development of bootstrap-sass itself, read this section.

### Upstream Converter

Keeping bootstrap-sass in sync with upstream changes from Bootstrap used to be an error prone and time consuming manual process. With Bootstrap 3 we have introduced a converter that automates this.

**Note: if you're just looking to *use* Bootstrap 3, see the [installation](#installation) section above.**

Upstream changes to the Bootstrap project can now be pulled in using the `convert` rake task.

Here's an example run that would pull down the master branch from the main [twbs/bootstrap](https://github.com/twbs/bootstrap) repo:

    rake convert

This will convert the latest LESS to Sass and update to the latest JS.
To convert a specific branch or version, pass the branch name or the commit hash as the first task argument:

    rake convert[e8a1df5f060bf7e6631554648e0abde150aedbe4]

The latest converter script is located [here][converter] and does the following:

* Converts upstream Bootstrap LESS files to its matching SCSS file.
* Copies all upstream JavaScript into `assets/javascripts/bootstrap`, a Sprockets manifest at `assets/javascripts/bootstrap-sprockets.js`, and a concatenation at `assets/javascripts/bootstrap.js`.
* Copies all upstream font files into `assets/fonts/bootstrap`.
* Sets `Bootstrap::BOOTSTRAP_SHA` in [version.rb][version] to the branch sha.

This converter fully converts original LESS to SCSS. Conversion is automatic but requires instructions for certain transformations (see converter output).
Please submit GitHub issues tagged with `conversion`.

## Credits

bootstrap-sass has a number of major contributors:

<!-- feel free to make these link wherever you wish -->
* [Thomas McDonald](https://twitter.com/thomasmcdonald_)
* [Tristan Harward](http://www.trisweb.com)
* Peter Gumeson
* [Gleb Mazovetskiy](https://github.com/glebm)

and a [significant number of other contributors][contrib].

## You're in good company
bootstrap-sass is used to build some awesome projects all over the web, including
[Diaspora](https://diasporafoundation.org/), [rails_admin](https://github.com/sferik/rails_admin),
Michael Hartl's [Rails Tutorial](https://www.railstutorial.org/), [gitlabhq](http://gitlabhq.com/) and
[kandan](http://getkandan.com/).

[converter]: https://github.com/twbs/bootstrap-sass/blob/master/tasks/converter/less_conversion.rb
[version]: https://github.com/twbs/bootstrap-sass/blob/master/lib/bootstrap-sass/version.rb
[contrib]: https://github.com/twbs/bootstrap-sass/graphs/contributors
[antirequire]: https://github.com/twbs/bootstrap-sass/issues/79#issuecomment-4428595
[jsdocs]: https://getbootstrap.com/javascript/#transitions
[sass-precision]: http://sass-lang.com/documentation/Sass/Script/Value/Number.html#precision%3D-class_method
[mincer]: https://github.com/nodeca/mincer
[autoprefixer]: https://github.com/postcss/autoprefixer

<p align="center">
  <img width="200" height="200" src="https://bootswatch.com/_assets/img/logo-dark.svg">
</p>

<h3 align="center">Bootswatch</h3>

<p align="center">
  A collection of open source themes for <a href="https://getbootstrap.com/">Bootstrap</a>
  <br>
  <a href="https://bootswatch.com/"><strong>View Bootswatch themes »</strong></a>
  <br>
  <br>
  <a href="https://github.com/thomaspark/bootswatch/issues/new">Report bug</a>
  ·
  <a href="https://github.com/thomaspark/bootswatch/issues/new">Request feature</a>
  ·
  <a href="https://blog.bootswatch.com/">Blog</a>
</p>

## Usage

There are a few different ways you can integrate Bootswatch into your project.

### Via Pre-compiled Asset

Download the `bootstrap.min.css` file associated with a theme and replace
Bootstrap's default stylesheet. You must still include Bootstrap's JavaScript
file to have functional dropdowns, modals, etc.

### Via CDN

You can use the themes via CDN at [jsDelivr](https://www.jsdelivr.com/package/npm/bootswatch).

### Via Sass Imports

If you're using [Sass](https://sass-lang.com/) (SCSS) in your project, you can
import the `_variables.scss` and `_bootswatch.scss` files for a given theme.
This method allows you to override theme variables.

```scss
// Your variable overrides go here, e.g.:
// $h1-font-size: 3rem;

@import "~bootswatch/dist/[theme]/variables";
@import "~bootstrap/scss/bootstrap";
@import "~bootswatch/dist/[theme]/bootswatch";
```

Make sure to import Bootstrap's `bootstrap.scss` in between `_variables.scss`
and `_bootswatch.scss`!

### Via npm

You can install as a package with the command `npm install bootswatch`.

#### React Users (`create-react-app`, or similar bundler)

Modern JavaScript bundlers (webpack, parcel, rollup) support `import`ing CSS from JS files. This can make it easier to deploy various 1st and 3rd party assets predictably. Note: _There are tradeoffs to the following method, research your tooling before deploying to production._

Before continuing, ensure you've run `npm install bootswatch` in your local project folder. (Use either `npm` or `yarn`.)

Add the following `import` to your top-level `index.js` (or `App.js`) file. Add it **before** any other `.css` imports.

```js
import "bootswatch/dist/[theme]/bootstrap.min.css";
// TODO: Note: Replace ^[theme]^ (examples: darkly, slate, cosmo, spacelab, and superhero. See https://bootswatch.com for current theme names.)
```

Here's an example of updated `imports` in `App.js` to use "slate" theme (using a `create-react-app` fresh project.)

```js
import React from 'react';
import logo from './logo.svg';
import 'bootswatch/dist/slate/bootstrap.min.css'; // Added this :boom:
import './App.css';

function App() {
...
```

### Via Ruby Gem

In your Ruby project, you can access the latest version of each theme by adding
the following to your Gemfile and running `bundle install`:

```ruby
gem "bootswatch", github: "thomaspark/bootswatch"
```

Each theme directory is then accessible via the path
`"#{Gem.loaded_specs["bootswatch"].load_paths.first}/[theme]"`.

Ruby on Rails users can add the following to an initializer (e.g.
`config/initializers/bootswatch.rb`):

```ruby
Rails.application.config.assets.paths += Gem.loaded_specs["bootswatch"].load_paths
```

And thus be able to import themes via Sass like so:

```scss
@import "[theme]/variables";
@import "~bootstrap/scss/bootstrap";
@import "[theme]/bootswatch";
```

### Via API

A simple JSON API is available for integrating your platform with Bootswatch.
More info can be found on the [Help page](https://bootswatch.com/help/#api).

## Customization

Bootswatch is open source and you’re welcome to modify the themes.

Each theme consists of two SASS files. `_variables.scss`, which is included by default in Bootstrap, allows you to customize the settings. `_bootswatch.scss` introduces more extensive structural changes.

Check out the [Help page](https://bootswatch.com/help/#customization) for more details on building your own theme.

## Contributing

It's through your contributions that Bootswatch will continue to improve. You can contribute in several ways:

* **Issues:** Provide a detailed report of any bugs you encounter and open an issue on [GitHub](https://github.com/thomaspark/bootswatch/issues).

* **Documentation:** If you'd like to fix a typo or beef up the docs, you can fork the project, make your changes, and submit a pull request.

* **Code:** Make a fix and submit it as a pull request. When making changes, it's important to keep the CSS and SASS versions in sync. To do this, be sure to edit the SASS source files for the particular theme first, then run the  tasks `grunt swatch` to build the CSS.

* **Donation:** Donations are gratefully accepted via [GitHub](https://github.com/sponsors/thomaspark) and [PayPal](https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&hosted_button_id=PU2DH4BMF9MWS&source=url).

## Author

Thomas Park

* <https://github.com/thomaspark>
* <https://thomaspark.co/>

## Thanks

* [Mark Otto](https://github.com/mdo) and [Jacob Thornton](https://github.com/fat) for [Bootstrap](https://github.com/twbs/bootstrap).
* [XhmikosR](https://github.com/xhmikosr) for ongoing maintenance support.
* [Jenil Gogari](https://jgog.in/) for his contributions to the Flatly theme.
* [James Taylor](https://github.com/jostylr) for [cors-lite](https://github.com/jostylr/cors-lite).
* [Corey Sewell](https://github.com/cjsewell) for SASS conversion.

## Copyright and License

Copyright 2014-2023 Thomas Park

Code released under the [MIT License](LICENSE).

Bootswatch
==========

[![Bootswatch Logo](./assets/img/logo-dark.png)](http://bootswatch.com)

Bootswatch is a collection of open source themes for [Bootstrap](http://getbootstrap.com/). Check it out at [bootswatch.com](http://bootswatch.com).

Usage
-----
Download the `bootstrap.min.css` file associated with a theme and replace Bootstrap's default stylesheet. You must still include Bootstrap's JavaScript file to have functional dropdowns, modals, etc.

The themes are also hosted on [BootstrapCDN](http://www.bootstrapcdn.com/bootswatch/). 

You can import a theme into your styles using either LESS or SASS.

LESS:

```
@import "bootstrap/less/bootstrap.less";
@import "bootswatch/theme/variables.less";
@import "bootswatch/theme/bootswatch.less";

```

SASS:

```
@import "bootswatch/theme/variables";
@import "bootstrap-sass-official/assets/stylesheets/bootstrap";
@import "bootswatch/theme/bootswatch";

```


Customization
------
Bootswatch is open source and you’re welcome to modify the themes.

Each theme consists of two LESS files. `variables.less`, which is included by default in Bootstrap, allows you to customize [these settings](http://getbootstrap.com/customize/#less-variables). `bootswatch.less` introduces more extensive structural changes.

These files are also available in SASS.

Check out the [Help page](http://bootswatch.com/help/) for more details on building your own theme.

API
-----

A simple API is available for integrating your platform with Bootswatch. More info at http://bootswatch.com/help/#api

Contributing
-----
It's through your contributions that Bootswatch will continue to improve. You can contribute in several ways.

**Issues:** Provide a detailed report of any bugs you encounter and open an issue on [GitHub](https://github.com/thomaspark/bootswatch/issues).

**Documentation:** If you'd like to fix a typo or beef up the docs, you can fork the project, make your changes, and submit a pull request.

**Code:** Make a fix and submit it as a pull request. When making changes, it's important to keep the CSS, LESS and SASS versions in sync. To do this, be sure to edit the LESS source files for the particular theme, then run the  tasks `grunt swatch` and `grunt:convert_less` to build the CSS and LESS.

**Donation:** Donations are gratefully accepted via [PayPal](https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&amp;hosted_button_id=F22JEM3Q78JC2) and [Gratipay](https://gratipay.com/bootswatch/).

Author
------
Thomas Park

+ http://github.com/thomaspark
+ http://thomaspark.co

Thanks
------
[Mark Otto](https://github.com/mdo) and [Jacob Thornton](https://github.com/fat) for [Bootstrap](https://github.com/twbs/bootstrap).

[Jenil Gogari](http://www.jgog.in/) for his contributions to the Flatly theme.

[James Taylor](https://github.com/jostylr) for [cors-lite](https://github.com/jostylr/cors-lite).

[Corey Sewell](https://github.com/cjsewell) for SASS conversion.


Copyright and License
----
Copyright 2014-2018 Thomas Park

Code released under the MIT License.

.onLoad <- function(libname, pkgname) {
  register_upgrade_message("shiny", "1.6.0")
  register_upgrade_message("rmarkdown", "2.7")
  .dependency_cache <<- cachem::cache_mem(max_age = 1)

  # Call setup_input_handlers() when both bslib and shiny are loaded
  if (!setup_input_handlers()) {
    # shiny is not yet loaded, if it ever gets loaded we'll register then
    setHook(packageEvent("shiny", "onLoad"), function(...) {
      setup_input_handlers()
    })
  }
}

.dependency_cache <- NULL

# Returns TRUE if input handlers were successfully loaded
setup_input_handlers <- function() {
  if (!"shiny" %in% loadedNamespaces()) {
    return(FALSE)
  }

  if (!"registerInputHandler" %in% names(asNamespace("shiny"))) {
    # In devtools::load_all(), the shiny namespace is loaded but empty at the
    # time bslib loads. It's OK not to do the registration now, it'll just
    # happen when shiny finishes loading.
    return(FALSE)
  }

  shiny::registerInputHandler(
    "bslib.taskbutton",
    input_task_button_input_handler,
    force = TRUE
  )
  TRUE
}

#' Test and/or coerce fill behavior
#'
#' Filling layouts in bslib are built on the foundation of fillable containers
#' and fill items (fill carriers are both fillable and fill). This is why most
#' bslib components (e.g., [card()], [card_body()], [layout_sidebar()]) possess
#' both `fillable` and `fill` arguments (to control their fill behavior).
#' However, sometimes it's useful to add, remove, and/or test fillable/fill
#' properties on arbitrary [htmltools::tag()], which these functions are
#' designed to do.
#'
#' @references The [Filling Layouts](https://rstudio.github.io/bslib/articles/filling.html)
#'   article on the bslib website introduces the concept of fillable containers
#'   and fill items.
#'
#' @seealso These functions provide a convenient interface to the underlying
#'   [htmltools::bindFillRole()] function.
#'
#' @details
#' Although `as_fill()`, `as_fillable()`, and `as_fill_carrier()` can work with
#' non-tag objects that have a [htmltools::as.tags] method (e.g., htmlwidgets),
#' they return the "tagified" version of that object.
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' shinyApp(
#'   page_fillable(
#'     # without `as_fill_carrier()`, the plot won't fill the page because
#'     # `uiOutput()` is neither a fillable container nor a fill item by default.
#'     as_fill_carrier(uiOutput("ui"))
#'   ),
#'   function(input, output) {
#'     output$ui <- renderUI({
#'       div(
#'         class = "bg-info text-white",
#'         as_fill_item(),
#'         "A fill item"
#'       )
#'     })
#'   }
#' )
#'
#' @return
#'   * For `as_fill()`, `as_fillable()`, and `as_fill_carrier()`: the _tagified_
#'     version `x`, with relevant tags modified to possess the relevant fill
#'     properties.
#'   * For `is_fill()`, `is_fillable()`, and `is_fill_carrier()`: a logical vector,
#'     with length matching the number of top-level tags that possess the relevant
#'     fill properties.
#'
#' @param x An [htmltools::tag()].
#' @param ... Currently ignored.
#' @param min_height,max_height Any valid [CSS unit][htmltools::validateCssUnit]
#'   (e.g., `150`).
#' @param gap Any valid [CSS unit][htmltools::validateCssUnit].
#' @param class A character vector of class names to add to the tag.
#' @param style A character vector of CSS properties to add to the tag.
#' @param css_selector A character string containing a CSS selector for
#'   targeting particular (inner) tag(s) of interest. For more details on what
#'   selector(s) are supported, see [htmltools::tagAppendAttributes()].
#'
#' @export
as_fill_carrier <- function(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
) {
  rlang::check_dots_empty()

  attrs <- fillable_attributes(
    min_height = min_height,
    max_height = max_height,
    gap = gap,
    class = class,
    style = style
  )

  if (rlang::is_missing(x)) {
    warn_css_selector_null(css_selector)
    attrs <- c(
      attrs,
      list(class = "html-fill-item html-fill-container", fill_dependency())
    )
    return(rlang::splice(attrs))
  }

  x <- bindFillRole(
    x,
    item = TRUE,
    container = TRUE,
    .cssSelector = css_selector
  )
  tagAppendAttributes(x, .cssSelector = css_selector, !!!attrs)
}

#' @rdname as_fill_carrier
#' @export
as_fillable_container <- function(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
) {
  rlang::check_dots_empty()

  attrs <- fillable_attributes(
    min_height = min_height,
    max_height = max_height,
    gap = gap,
    class = class,
    style = style
  )

  if (rlang::is_missing(x)) {
    warn_css_selector_null(css_selector)
    attrs <- c(attrs, list(class = "html-fill-container", fill_dependency()))
    return(rlang::splice(attrs))
  }

  x <- bindFillRole(x, container = TRUE, .cssSelector = css_selector)

  tagAppendAttributes(x, .cssSelector = css_selector, !!!attrs)
}

#' @rdname as_fill_carrier
#' @export
as_fill_item <- function(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
) {
  rlang::check_dots_empty()

  attrs <- fillable_attributes(
    min_height = min_height,
    max_height = max_height,
    class = class,
    style = style
  )

  if (rlang::is_missing(x)) {
    warn_css_selector_null(css_selector)
    attrs <- c(attrs, list(class = "html-fill-item", fill_dependency()))
    return(rlang::splice(attrs))
  }

  x <- bindFillRole(x, item = TRUE, .cssSelector = css_selector)

  tagAppendAttributes(x, .cssSelector = css_selector, !!!attrs)
}

fill_dependency <- local({
  fill_dep <- NULL

  function() {
    if (!is.null(fill_dep)) return(fill_dep)
    fill_dep <<- htmltools::findDependencies(
      htmltools::bindFillRole(htmltools::tags$div(), container = TRUE)
    )[[1]]
    fill_dep
  }
})

fillable_attributes <- function(
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL
) {
  list(
    style = css(
      min_height = validateCssUnit(min_height),
      max_height = validateCssUnit(max_height),
      gap = validateCssUnit(gap)
    ),
    class = class,
    style = style
  )
}

warn_css_selector_null <- function(x) {
  if (is.null(x)) return()

  rlang::warn(
    "Ignoring non-NULL `css_selector` because an input tag `x` was not provided.",
    call = rlang::caller_env()
  )
}

#' @rdname as_fill_carrier
#' @export
remove_all_fill <- function(x) {
  # NOTE: this doesn't remove the `htmltools-fill` dependency from `x`
  bindFillRole(
    x,
    item = FALSE,
    container = FALSE,
    overwrite = TRUE
  )
}

#' @rdname as_fill_carrier
#' @export
is_fill_carrier <- function(x) {
  classes <- paste0("html-fill-", c("container", "item"))
  renders_to_tag_class(x, classes)
}

#' @rdname as_fill_carrier
#' @export
is_fillable_container <- function(x) {
  UseMethod("is_fillable_container")
}

#' @export
is_fillable_container.htmlwidget <- function(x) {
  # won't actually work until (htmltools#334) gets fixed
  renders_to_tag_class(x, "html-fill-container", ".html-widget")
}

#' @export
is_fillable_container.default <- function(x) {
  renders_to_tag_class(x, "html-fill-container")
}

#' @rdname as_fill_carrier
#' @export
is_fill_item <- function(x) {
  UseMethod("is_fill_item")
}

#' @export
is_fill_item.htmlwidget <- function(x) {
  # won't actually work until (htmltools#334) gets fixed
  renders_to_tag_class(x, "html-fill-item", ".html-widget")
}

#' @export
is_fill_item.default <- function(x) {
  renders_to_tag_class(x, "html-fill-item")
}

renders_to_tag_class <- function(x, class, selector = NULL) {
  x <- try(as.tags(x), silent = TRUE)
  if (inherits(x, "try-error")) {
    return(FALSE)
  }
  xq <- tagQuery(x)
  if (length(selector)) {
    xq <- xq$find(selector)
  }
  vapply(
    xq$selectedTags(),
    function(y) tagQuery(y)$hasClass(class),
    logical(1),
    USE.NAMES = FALSE
  )
}

is_tag <- function(x) {
  inherits(x, "shiny.tag")
}

# Fill Item -----------------------------------------------------------------

test_that("as_fill_item() with a simple tag", {
  expect_false(is_fill_item(tag_simple()))
  expect_true(is_fill_item(as_fill_item(tag_simple())))

  expect_match(
    tagGetAttribute(as_fill_item(tag_simple()), "class"),
    "html-fill-item"
  )

  expect_snapshot(cat(format(as_fill_item(tag_simple()))))
  expect_equal(
    format(as_fill_item(tag_simple())),
    format(tag_simple(as_fill_item()))
  )
})

test_that("as_fill_item() with a simple tag with arguments", {
  expect_false(is_fill_item(tag_simple()))

  ts_fill <- as_fill_item(
    tag_simple(),
    min_height = "100px",
    max_height = "100%"
  )
  expect_true(is_fill_item(ts_fill))

  expect_match(tagGetAttribute(ts_fill, "class"), "html-fill-item")
  expect_match(tagGetAttribute(ts_fill, "style"), "min-height:\\s*100px")
  expect_match(tagGetAttribute(ts_fill, "style"), "max-height:\\s*100%")

  expect_snapshot(cat(format(ts_fill)))
  expect_equal(
    format(ts_fill),
    format(
      tag_simple(as_fill_item(min_height = "100px", max_height = "100%"))
    )
  )
})

test_that("as_fill_item() with a nested tag on outer tag", {
  expect_false(is_fill_item(tag_nested()))
  expect_true(is_fill_item(as_fill_item(tag_nested())))
  expect_match(
    tagGetAttribute(as_fill_item(tag_nested()), "class"),
    "html-fill-item"
  )

  expect_snapshot(cat(format(as_fill_item(tag_nested()))))
  expect_equal(
    format(as_fill_item(tag_nested())),
    format(tag_nested(as_fill_item()))
  )
})

test_that("as_fill_item() with a nested tag on inner tag", {
  # three inner tags that aren't fill items
  expect_equal(
    renders_to_tag_class(tag_nested(), "html-fill-item", selector = ".inner"),
    rep(FALSE, 3)
  )

  nested_fill <- as_fill_item(tag_nested(), css_selector = ".inner")
  expect_equal(
    renders_to_tag_class(nested_fill, "html-fill-item", selector = ".inner"),
    rep(TRUE, 3)
  )

  expect_snapshot(cat(format(nested_fill)))
})

# Fillable Container --------------------------------------------------------

test_that("as_fillable_container() with a simple tag", {
  expect_false(is_fillable_container(tag_simple()))
  expect_true(is_fillable_container(as_fillable_container(tag_simple())))

  expect_match(
    tagGetAttribute(as_fillable_container(tag_simple()), "class"),
    "html-fill-container"
  )

  expect_snapshot(cat(format(as_fillable_container(tag_simple()))))
  expect_equal(
    format(as_fillable_container(tag_simple())),
    format(tag_simple(as_fillable_container()))
  )
})

test_that("as_fillable_container() with a simple tag with arguments", {
  expect_false(is_fillable_container(tag_simple()))

  ts_fillable <-
    as_fillable_container(
      tag_simple(),
      min_height = "100px",
      max_height = "100%",
      gap = "1em"
    )
  expect_true(is_fillable_container(ts_fillable))

  expect_match(tagGetAttribute(ts_fillable, "class"), "html-fill-container")
  expect_match(tagGetAttribute(ts_fillable, "style"), "min-height:\\s*100px")
  expect_match(tagGetAttribute(ts_fillable, "style"), "max-height:\\s*100%")
  expect_match(tagGetAttribute(ts_fillable, "style"), "gap:\\s*1em")

  ts_fillable_inline <-
    tag_simple(
      as_fillable_container(
        min_height = "100px",
        max_height = "100%",
        gap = "1em"
      )
    )

  expect_snapshot(cat(format(ts_fillable)))
  expect_equal(format(ts_fillable), format(ts_fillable_inline))
})

test_that("as_fillable_container() with a nested tag on outer tag", {
  expect_false(is_fillable_container(tag_nested()))
  expect_true(is_fillable_container(as_fillable_container(tag_nested())))
  expect_match(
    tagGetAttribute(as_fillable_container(tag_nested()), "class"),
    "html-fill-container"
  )

  expect_snapshot(cat(format(as_fillable_container(tag_nested()))))
  expect_equal(
    format(as_fillable_container(tag_nested())),
    format(tag_nested(as_fillable_container()))
  )
})

test_that("as_fillable_container() with a nested tag on inner tag", {
  # three inner tags that aren't fillable containers
  expect_equal(
    renders_to_tag_class(
      tag_nested(),
      "html-fill-container",
      selector = ".inner"
    ),
    rep(FALSE, 3)
  )

  nested_fillable <- as_fillable_container(
    tag_nested(),
    css_selector = ".inner"
  )
  expect_equal(
    renders_to_tag_class(
      nested_fillable,
      "html-fill-container",
      selector = ".inner"
    ),
    rep(TRUE, 3)
  )

  expect_snapshot(cat(format(nested_fillable)))
})

# Fill Carrier --------------------------------------------------------------

expect_has_classes <- function(tag, classes) {
  tag_classes <- tagGetAttribute(tag, "class")
  tag_classes <- strsplit(tag_classes, " ")[[1]]
  expect_true(all(!!classes %in% tag_classes))
}

test_that("as_fill_carrier() with a simple tag", {
  expect_false(is_fill_carrier(tag_simple()))
  expect_true(is_fill_carrier(as_fill_carrier(tag_simple())))

  expect_has_classes(
    as_fill_carrier(tag_simple()),
    c("html-fill-container", "html-fill-item")
  )

  expect_snapshot(cat(format(as_fill_carrier(tag_simple()))))
})

test_that("as_fill_carrier() with a simple tag with arguments", {
  expect_false(is_fill_carrier(tag_simple()))

  ts_carrier <- as_fill_carrier(
    tag_simple(),
    min_height = "100px",
    max_height = "100%",
    gap = "1em"
  )
  expect_true(is_fill_carrier(ts_carrier))

  expect_match(tagGetAttribute(ts_carrier, "class"), "html-fill-container")
  expect_match(tagGetAttribute(ts_carrier, "style"), "min-height:\\s*100px")
  expect_match(tagGetAttribute(ts_carrier, "style"), "max-height:\\s*100%")
  expect_match(tagGetAttribute(ts_carrier, "style"), "gap:\\s*1em")

  ts_carrier_inline <-
    tag_simple(
      as_fill_carrier(
        min_height = "100px",
        max_height = "100%",
        gap = "1em"
      )
    )

  expect_snapshot(cat(format(ts_carrier)))
  expect_equal(format(ts_carrier), format(ts_carrier_inline))
})

test_that("as_fill_carrier() with a nested tag on outer tag", {
  expect_false(is_fill_carrier(tag_nested()))
  expect_true(is_fill_carrier(as_fill_carrier(tag_nested())))

  expect_has_classes(
    as_fill_carrier(tag_simple()),
    c("html-fill-container", "html-fill-item")
  )

  expect_snapshot(cat(format(as_fill_carrier(tag_nested()))))
})

test_that("as_fill_carrier() with a nested tag on inner tag", {
  # three inner tags that aren't fill carriers
  expect_equal(
    renders_to_tag_class(
      tag_nested(),
      "html-fill-container",
      selector = ".inner"
    ),
    rep(FALSE, 3)
  )
  expect_equal(
    renders_to_tag_class(tag_nested(), "html-fill-item", selector = ".inner"),
    rep(FALSE, 3)
  )

  nested_carrier <- as_fill_carrier(tag_nested(), css_selector = ".inner")
  expect_equal(
    renders_to_tag_class(
      nested_carrier,
      "html-fill-container",
      selector = ".inner"
    ),
    rep(TRUE, 3)
  )
  expect_equal(
    renders_to_tag_class(nested_carrier, "html-fill-item", selector = ".inner"),
    rep(TRUE, 3)
  )

  expect_snapshot(cat(format(nested_carrier)))
})

# Remove All Fill -----------------------------------------------------------

test_that("remove_all_fill() removes fill from outer element", {
  fill_removed <- remove_all_fill(as_fill_item(tag_simple()))
  expect_null(htmlDependencies(fill_removed))
  expect_equal_html(fill_removed, tag_simple())

  expect_equal_html(
    remove_all_fill(as_fillable_container(tag_simple())),
    tag_simple()
  )

  expect_equal_html(
    remove_all_fill(as_fill_carrier(tag_simple())),
    tag_simple()
  )
})

# Unexpected Uses -----------------------------------------------------------

test_that("as_fill_item() warns about usage of `css_selector` without an input tag", {
  expect_warning(
    expect_equal(
      as_fill_item(class = "test", css_selector = ".inner"),
      as_fill_item(class = "test")
    )
  )

  expect_warning(
    expect_equal(
      div(as_fill_item(class = "test", css_selector = ".inner")),
      div(as_fill_item(class = "test"))
    )
  )
})

test_that("as_fillable_container() warns about usage of `css_selector` without an input tag", {
  expect_warning(
    expect_equal(
      as_fillable_container(class = "test", css_selector = ".inner"),
      as_fillable_container(class = "test")
    )
  )

  expect_warning(
    expect_equal(
      div(as_fillable_container(class = "test", css_selector = ".inner")),
      div(as_fillable_container(class = "test"))
    )
  )
})

test_that("as_fill_carrier() warns about usage of `css_selector` without an input tag", {
  expect_warning(
    expect_equal(
      as_fill_carrier(class = "test", css_selector = ".inner"),
      as_fill_carrier(class = "test")
    )
  )

  expect_warning(
    expect_equal(
      div(as_fill_carrier(class = "test", css_selector = ".inner")),
      div(as_fill_carrier(class = "test"))
    )
  )
})

get_current_session <- function(
  require_active = TRUE,
  call = rlang::caller_env()
) {
  session <- if (isNamespaceLoaded("shiny")) {
    shiny::getDefaultReactiveDomain()
  }

  if (is.null(session) && require_active) {
    rlang::abort(
      paste(
        "This function must be called with a Shiny `server` function",
        "(i.e., it can only be used inside an active user session)."
      ),
      call = call,
      class = "shiny-no-active-session"
    )
  }

  session
}

has_valid_reactive_context <- function(session) {
  if (is.null(session)) {
    return(FALSE)
  }
  if (!"getCurrentTheme" %in% names(session)) {
    return(FALSE)
  }
  hasReactiveContext <- getFromNamespace("hasCurrentContext", "shiny") %||%
    function() FALSE
  hasReactiveContext()
}

get_current_theme <- function() {
  if (isNamespaceLoaded("shiny")) shiny::getCurrentTheme()
}

# Shiny internal funcs needed for nav_panel() (i.e., tabPanel()) logic

processDeps <- function(...) {
  getFromNamespace("processDeps", "shiny")(...)
}

p_randomInt <- function(...) {
  getFromNamespace("p_randomInt", "shiny")(...)
}

# Copy of shiny::getCurrentThemeVersion()
# (copied to avoid >1.6.0 dependency)
getCurrentThemeVersion <- function() {
  theme <- shiny::getCurrentTheme()
  if (is_bs_theme(theme)) theme_version(theme) else "3"
}

# Copy of shiny:::anyNamed()
anyNamed <- function(x) {
  if (length(x) == 0) {
    return(FALSE)
  }
  nms <- names(x)
  if (is.null(nms)) {
    return(FALSE)
  }
  any(nzchar(nms))
}


shiny_input_label <- function(id, label = NULL) {
  getFromNamespace("shinyInputLabel", "shiny")(id, label)
}

component_dependencies <- function() {
  list(
    component_dependency_js(),
    bs_dependency_defer(component_dependency_sass)
  )
}

component_dependency_js <- function() {
  minified <- get_shiny_devmode_option("shiny.minified", default = TRUE)

  htmlDependency(
    name = "bslib-component-js",
    version = get_package_version("bslib"),
    package = "bslib",
    src = "components/dist",
    script = list(
      list(src = paste0("components", if (minified) ".min", ".js")),
      list(
        src = paste0("web-components", if (minified) ".min", ".js"),
        type = "module"
      )
    )
  )
}

# Pre-compiled component styles
component_dependency_css <- function() {
  htmlDependency(
    name = "bslib-component-css",
    version = get_package_version("bslib"),
    package = "bslib",
    src = "components/dist",
    stylesheet = "components.css"
  )
}

# Run-time (Sass) component styles
component_dependency_sass <- function(theme) {
  precompiled <- isTRUE(get_precompiled_option())
  default_theme <- !is_bs_theme(theme) || identical(theme, bs_theme())
  if (precompiled && default_theme) {
    component_dependency_css()
  } else {
    component_dependency_sass_(theme)
  }
}

component_dependency_sass_files <- function() {
  scss_dir <- path_inst("components", "scss")
  scss_files <- c(
    file.path(scss_dir, "mixins", "_mixins.scss"),
    dir(scss_dir, pattern = "\\.scss$", full.names = TRUE)
  )

  lapply(scss_files, sass_file)
}

component_dependency_sass_ <- function(theme) {
  # Although rare, it's possible for bs_dependency_defer() to pass
  # along a NULL theme (e.g., renderTags(accordion())), so fallback
  # to the default theme if need be
  theme <- theme %||% bs_theme()

  if (theme_version(theme) < 5) {
    abort(
      c(
        "bslib components require Bootstrap 5 or higher.",
        "i" = "Do you need to specify a different `version` in `bs_theme()`?"
      )
    )
  }

  bs_dependency(
    input = component_dependency_sass_files(),
    theme = theme,
    name = "bslib-component-css",
    version = get_package_version("bslib"),
    cache_key_extra = get_package_version("bslib"),
    .sass_args = list(options = sass_options(output_style = "compressed"))
  )
}

web_component <- function(tagName, ...) {
  deps <- component_dependencies()
  args <- c(deps, rlang::list2(...))
  htmltools::tag(tagName, args)
}

# List which precompiled CSS files are available.
.precompiled_css_versions <- local({
  versions <- NULL
  function() {
    if (is.null(versions)) {
      versions <<- dir(system.file("css-precompiled", package = "bslib"))
    }
    versions
  }
})

.precompiled_css_themes <- new.env(parent = emptyenv())

# Precompiled themes are created for base `bs_theme()`, changing only the
# Bootstrap version number. To decide if we can use the precompiled theme, we
# hash `bs_theme(version)`, which we'll compare with a hash of the user's theme.
precompiled_bs_theme_hash <- function(version) {
  theme_hash <- get0(as.character(version), .precompiled_css_themes)

  if (is.null(theme_hash)) {
    theme_hash <- rlang::hash(bs_theme(version = version, brand = FALSE))
    assign(version, theme_hash, envir = .precompiled_css_themes)
  }

  return(theme_hash)
}

#' Get the path to a precompiled CSS file
#'
#' This function is meant for development and debugging purposes. It can be used
#' to test if a precompiled CSS file exists for a given theme object.
#'
#' @inheritParams bs_theme_update
#'
#' @return The path to the precompiled CSS file, if it exists, or `NULL`, if
#'   not.
#' @keywords internal
#' @export
precompiled_css_path <- function(theme = bs_theme()) {
  version <- theme_version(theme)
  theme_hash <- rlang::hash(theme)
  version_precompiled <- NULL

  if (is.null(version)) {
    for (v in .precompiled_css_versions()) {
      if (identical(theme_hash, precompiled_bs_theme_hash(v))) {
        version_precompiled <- v
        break
      }
    }
  } else if (version %in% .precompiled_css_versions()) {
    if (identical(theme_hash, precompiled_bs_theme_hash(version))) {
      version_precompiled <- version
    }
  }

  if (is.null(version_precompiled)) {
    return(NULL)
  }

  system_file(
    package = "bslib",
    "css-precompiled",
    version_precompiled,
    "bootstrap.min.css"
  )
}

local_disable_cache()

test_that("Can find precompiled themes", {
  # Should be present for 4, 3 without bootswatch theme
  expect_true(file.exists(precompiled_css_path(bs_theme("4"))))
  expect_true(file.exists(precompiled_css_path(bs_theme("3"))))

  # Not present for bootswatch themes
  expect_null(precompiled_css_path(bs_theme("4", "sketchy")))

  # With modifications
  theme <- bs_add_rules(bs_theme(), rules = "body { background: red; }")
  expect_null(precompiled_css_path(theme))
})


test_that("Precompiled theme output is identical to compiled themes", {
  skip_on_cran()

  get_bootstrap_path <- function(x) {
    bs_dep <- Filter(x, f = function(y) { identical(y$name, "bootstrap") })[[1]]
    bs_dep$src$file
  }

  # precompiled themes will end up in a different directory, but have identical
  # contents to non-precompiled. The dirs will be something like:
  # /tmp/RtmpVi9qRT/bslib-088b402ee16511d7
  # /tmp/RtmpVi9qRT/bslib-precompiled-4
  dir1 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4"), precompiled = FALSE))
  dir2 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4"), precompiled = TRUE))
  expect_false(identical(dir1, dir2))
  expect_true(identical_dirs(dir1, dir2))

  dir1 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("3"), precompiled = FALSE))
  dir2 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("3"), precompiled = TRUE))
  expect_false(identical(dir1, dir2))
  expect_true(identical_dirs(dir1, dir2))

  # Two calls to bs_theme_dependencies() with precompiled CSS should end up in
  # the same dir, even when caching is turned off. Something like:
  # /tmp/RtmpVi9qRT/bslib-precompiled-4
  dir1 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4"), precompiled = TRUE))
  dir2 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4"), precompiled = TRUE))
  expect_identical(dir1, dir2)

  # For default sass options, we'll end up with the precompiled dir, something
  # like:
  # /tmp/RtmpVi9qRT/bslib-precompiled-4
  default_sass_options <- eval(formals(bs_theme_dependencies)$sass_options)
  dir1 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4"), precompiled = TRUE))
  dir2 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4"), default_sass_options, precompiled = TRUE))
  expect_identical(dir1, dir2)

  # For non-default sass options, we'll end up NOT with the precompiled dir,
  # which means that it didn't use the precompiled version.
  dir1 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4"), precompiled = TRUE))
  dir2 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4"),
                             sass::sass_options(output_style = "expanded"),
                             precompiled = TRUE))
  expect_false(identical(dir1, dir2))

  # For themes where there's no precompiled version, they will be compiled to
  # the same directory whether precompiled is TRUE or FALSE. It will
  # look like:
  # /tmp/RtmpVi9qRT/bslib-5c3108f39cfd584e
  # Need to tell sass to use a cache, because otherwise it will write into a new
  # directory every time.
  cache <- sass_file_cache(tempfile())
  dir1 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4", "darkly"),
                                       cache = cache,
                                       precompiled = FALSE))
  dir2 <- get_bootstrap_path(bs_theme_dependencies(bs_theme("4", "darkly"),
                                       cache = cache,
                                       precompiled = TRUE))
  expect_true(identical(dir1, dir2))
})

local_disable_cache()

test_that("Can find precompiled themes", {
  # Should be present for 4, 3 without bootswatch theme
  expect_true(file.exists(precompiled_css_path(bs_theme("4"))))
  expect_true(file.exists(precompiled_css_path(bs_theme("3"))))

  # Not present for bootswatch themes
  expect_null(precompiled_css_path(bs_theme("4", "sketchy")))

  # With modifications
  theme <- bs_add_rules(bs_theme(), rules = "body { background: red; }")
  expect_null(precompiled_css_path(theme))
})

test_that("Precompiled theme output is identical to compiled themes", {
  skip_on_cran()
  # If this fails run: Rscript tools/main.R

  get_bootstrap_path <- function(x) {
    bs_dep <- Filter(x, f = function(y) {
      identical(y$name, "bootstrap")
    })[[1]]
    bs_dep$src$file
  }

  # precompiled themes will end up in a different directory, but have identical
  # contents to non-precompiled. The dirs will be something like:
  # /tmp/RtmpVi9qRT/bslib-088b402ee16511d7
  # /tmp/RtmpVi9qRT/bslib-precompiled-4
  dir1 <- get_bootstrap_path(
    bs_theme_dependencies(bs_theme("4"), precompiled = FALSE)
  )
  dir2 <- get_bootstrap_path(
    bs_theme_dependencies(bs_theme("4"), precompiled = TRUE)
  )
  expect_false(identical(dir1, dir2))
  expect_true(identical_dirs(dir1, dir2))

  dir1 <- get_bootstrap_path(
    bs_theme_dependencies(bs_theme("3"), precompiled = FALSE)
  )
  dir2 <- get_bootstrap_path(
    bs_theme_dependencies(bs_theme("3"), precompiled = TRUE)
  )
  expect_false(identical(dir1, dir2))
  expect_true(identical_dirs(dir1, dir2))

  # Two calls to bs_theme_dependencies() with precompiled CSS should end up in
  # the same dir, even when caching is turned off. Something like:
  # /tmp/RtmpVi9qRT/bslib-precompiled-4
  dir1 <- get_bootstrap_path(
    bs_theme_dependencies(bs_theme("4"), precompiled = TRUE)
  )
  dir2 <- get_bootstrap_path(
    bs_theme_dependencies(bs_theme("4"), precompiled = TRUE)
  )
  expect_identical(dir1, dir2)

  # For default sass options, we'll end up with the precompiled dir, something
  # like:
  # /tmp/RtmpVi9qRT/bslib-precompiled-4
  default_sass_options <- eval(formals(bs_theme_dependencies)$sass_options)
  dir1 <- get_bootstrap_path(
    bs_theme_dependencies(bs_theme("4"), precompiled = TRUE)
  )
  dir2 <- get_bootstrap_path(
    bs_theme_dependencies(
      bs_theme("4"),
      default_sass_options,
      precompiled = TRUE
    )
  )
  expect_identical(dir1, dir2)

  # For non-default sass options, we'll end up NOT with the precompiled dir,
  # which means that it didn't use the precompiled version.
  dir1 <- get_bootstrap_path(
    bs_theme_dependencies(bs_theme("4"), precompiled = TRUE)
  )
  dir2 <- get_bootstrap_path(
    bs_theme_dependencies(
      bs_theme("4"),
      sass::sass_options(output_style = "expanded"),
      precompiled = TRUE
    )
  )
  expect_false(identical(dir1, dir2))

  # For themes where there's no precompiled version, they will be compiled to
  # the same directory whether precompiled is TRUE or FALSE. It will
  # look like:
  # /tmp/RtmpVi9qRT/bslib-5c3108f39cfd584e
  # Need to tell sass to use a cache, because otherwise it will write into a new
  # directory every time.
  cache <- sass_file_cache(tempfile())
  dir1 <- get_bootstrap_path(
    bs_theme_dependencies(
      bs_theme("4", "darkly"),
      cache = cache,
      precompiled = FALSE
    )
  )
  dir2 <- get_bootstrap_path(
    bs_theme_dependencies(
      bs_theme("4", "darkly"),
      cache = cache,
      precompiled = TRUE
    )
  )
  expect_true(identical(dir1, dir2))
})

# Class added or removed from `bs_theme` when a theme includes a preset
THEME_PRESET_CLASS <- "bs_theme_with_preset"

resolve_bs_preset <- function(
  preset = NULL,
  bootswatch = NULL,
  version = version_default()
) {
  if (is.null(preset) && is.null(bootswatch)) return(NULL)

  assert_preset_scalar_string(preset)
  assert_preset_scalar_string(bootswatch)
  assert_preset_only_one_name_arg(preset, bootswatch)

  version <- switch_version(version, five = "5", four = "4", three = "3")
  preset_name <- preset %||% bootswatch

  if (preset_name %in% c("default", "bootstrap")) {
    # "bootstrap" means no preset bundle, just bare default Bootstrap
    return(new_bs_preset("bootstrap", version))
  }

  builtin_themes <- builtin_themes(version)

  if (length(builtin_themes) > 0 && preset_name %in% builtin_themes) {
    return(new_bs_preset(preset_name, version, type = "builtin"))
  }

  bootswatch_themes <- bootswatch_themes(version)

  if (length(bootswatch_themes) > 0 && preset_name %in% bootswatch_themes) {
    return(new_bs_preset(preset_name, version, type = "bootswatch"))
  }

  abort_preset_unknown_name(preset_name, version)
}

new_bs_preset <- function(name, version, type = NULL, ...) {
  preset <- list(
    version = version, # bootstrap version
    name = name, # preset name
    type = type, # preset type (e.g. "builtin", "bootswatch")
    ...
  )

  structure(dropNulls(preset), class = "bs_preset")
}

# The `bs_preset_bundle()` function is the main entry point for creating a SASS
# bundle for a theme preset. This calls out to the appropriate functions to
# create a bundle for a built-in theme or for a Bootswatch theme.
bs_preset_bundle <- function(preset) {
  if (!inherits(preset, "bs_preset")) return(NULL)
  if (is.null(preset$type)) return(NULL)

  switch(
    preset$type,
    bootswatch = bootswatch_bundle(preset$name, version = preset$version),
    builtin = builtin_bundle(preset$name, version = preset$version),
    stop("Unknown preset type: ", preset$type)
  )
}

theme_preset_info <- function(theme) {
  if (!is_bs_theme(theme)) return(NULL)

  info <- bs_get_variables(
    theme,
    c("bslib-preset-type", "bslib-preset-name", "bootstrap-version")
  )

  name <- info[["bslib-preset-name"]]
  type <- info[["bslib-preset-type"]]

  new_bs_preset(
    name = if (!is.na(name)) name else "bootstrap",
    version = info[["bootstrap-version"]],
    type = if (!is.na(type)) type
  )
}

assert_preset_scalar_string <- function(var, .frame = rlang::caller_env()) {
  var_name <- deparse(substitute(var))

  if (is.null(var) || rlang::is_string(var)) {
    return(invisible())
  }

  msg <- c(
    sprintf(
      "The preset theme `%s` must be a single character string.",
      var_name
    ),
    "x" = sprintf('Bad: `%s = c("flatly", "darkly")`', var_name),
    "v" = sprintf('Good: `%s = "flatly"`', var_name)
  )
  rlang::abort(msg, .frame = .frame)
}

assert_preset_only_one_name_arg <- function(
  preset,
  bootswatch,
  .frame = rlang::caller_env()
) {
  both_provided <- !is.null(preset) && !is.null(bootswatch)

  if (!both_provided) {
    return(invisible())
  }

  msg <- c(
    "Only one of `preset` or `bootswatch` may be provided (`preset` is the preferred choice).",
    "i" = "Did you mean one of the following options?",
    "*" = sprintf('`preset = "%s"`', preset),
    "*" = sprintf('`preset = "%s"`', bootswatch),
    "*" = sprintf('`bootswatch = "%s"`', bootswatch)
  )
  rlang::abort(msg, .frame = .frame)
}

abort_preset_unknown_name <- function(
  name,
  version,
  .frame = rlang::caller_env()
) {
  msg <- c(
    sprintf(
      "'%s' is not a known preset theme name for Bootstrap version %s.",
      name,
      version
    ),
    "i" = "You can list available preset themes:",
    "*" = sprintf("Built-in: `builtin_themes(%s)`", version),
    "*" = sprintf("Bootswatch: `bootswatch_themes(%s)`.", version)
  )
  rlang::abort(msg, .frame = .frame)
}

describe("resolve_bs_preset()", {
  it("returns NULL if both `name` and `bootswatch` are missing", {
    expect_null(resolve_bs_preset(preset = NULL, bootswatch = NULL))
  })

  it("throws an error if both `name` and `bootswatch` are provided", {
    expect_snapshot(
      error = TRUE,
      resolve_bs_preset(preset = "name", bootswatch = "bootswatch")
    )
  })

  it("throws an error if `name` or `bootswatch` are not scalar strings", {
    expect_snapshot(
      error = TRUE,
      resolve_bs_preset(preset = c("a", "b"))
    )
    expect_snapshot(
      error = TRUE,
      resolve_bs_preset(bootswatch = c("a", "b"))
    )

    expect_error(resolve_bs_preset(preset = 1))
    expect_error(resolve_bs_preset(bootswatch = 1))

    expect_error(resolve_bs_preset(preset = TRUE))
    expect_error(resolve_bs_preset(bootswatch = TRUE))

    expect_error(resolve_bs_preset(preset = NA_character_))
    expect_error(resolve_bs_preset(bootswatch = NA_character_))
  })

  it("throws an error if `name` or `bootswatch` don't match existing presets", {
    expect_snapshot(
      error = TRUE,
      resolve_bs_preset(preset = "not_a_preset", version = 4)
    )
    expect_snapshot(
      error = TRUE,
      resolve_bs_preset(bootswatch = "not_a_preset", version = 4)
    )
  })

  it("throws an error for unknown bootstrap version", {
    expect_error(resolve_bs_preset(preset = "cerulean", version = "2"))
    expect_error(resolve_bs_preset(bootswatch = "cerulean", version = "99"))

    expect_warning(resolve_bs_preset(bootswatch = "cerulean", version = "4-3"))
  })

  it("returns a 'default' preset if name or bootswatch is 'default' or 'bootstrap'", {
    expect_equal(
      resolve_bs_preset(preset = "default"),
      resolve_bs_preset(bootswatch = "default")
    )

    expect_equal(
      resolve_bs_preset(preset = "bootstrap"),
      resolve_bs_preset(bootswatch = "bootstrap")
    )

    expect_equal(
      resolve_bs_preset(preset = "default"),
      resolve_bs_preset(preset = "bootstrap")
    )

    expect_equal(
      unclass(resolve_bs_preset(preset = "default")),
      list(version = version_default(), name = "bootstrap")
    )

    expect_identical(class(resolve_bs_preset(preset = "default")), "bs_preset")
  })

  it("returns a BS5 Bootswatch theme preset", {
    bsw_darkly <- resolve_bs_preset(bootswatch = "darkly", version = 5)

    expect_s3_class(bsw_darkly, "bs_preset")
    expect_equal(bsw_darkly$name, "darkly")
    expect_equal(bsw_darkly$version, "5")
  })

  it("returns a BS4 Bootswatch theme preset", {
    bsw_cosmo <- resolve_bs_preset(bootswatch = "cosmo", version = 4)

    expect_s3_class(bsw_cosmo, "bs_preset")
    expect_equal(bsw_cosmo$name, "cosmo")
    expect_equal(bsw_cosmo$version, "4")
  })

  it("returns a BS3 Bootswatch theme preset", {
    bsw_readable <- resolve_bs_preset(bootswatch = "readable", version = 3)

    expect_s3_class(bsw_readable, "bs_preset")
    expect_equal(bsw_readable$name, "readable")
    expect_equal(bsw_readable$version, "3")
  })

  it("returns a bootswatch theme preset if `name` is used instead of `bootswatch`", {
    expect_equal(
      resolve_bs_preset(preset = "darkly", version = 5),
      resolve_bs_preset(bootswatch = "darkly", version = 5)
    )

    expect_equal(
      resolve_bs_preset(preset = "cosmo", version = 4),
      resolve_bs_preset(bootswatch = "cosmo", version = 4)
    )

    expect_equal(
      resolve_bs_preset(preset = "readable", version = 3),
      resolve_bs_preset(bootswatch = "readable", version = 3)
    )
  })

  it("returns the builtin shiny theme preset", {
    shiny <- resolve_bs_preset(preset = "shiny", version = 5)

    expect_s3_class(shiny, "bs_preset")
    expect_equal(shiny$name, "shiny")
    expect_equal(shiny$version, "5")
  })
})

test_that("bs_preset_bundle() returns `NULL` for default or empty preset", {
  expect_null(bs_preset_bundle(resolve_bs_preset(preset = "default")))
  expect_null(bs_preset_bundle(resolve_bs_preset(bootswatch = "default")))
  expect_null(bs_preset_bundle(NULL))
})

test_that("bs_preset_bundle() throws for unknown preset$type", {
  expect_error(bs_preset_bundle(new_bs_preset("invalid", 5, type = "unknown")))
})

describe("theme_preset_info()", {
  it("returns bootswatch theme information", {
    expect_equal(
      theme_preset_info(bs_theme(version = 5, bootswatch = "flatly")),
      new_bs_preset("flatly", version = "5", type = "bootswatch")
    )

    expect_equal(
      theme_preset_info(bs_theme(version = 4, bootswatch = "superhero")),
      new_bs_preset("superhero", version = "4", type = "bootswatch")
    )

    expect_equal(
      theme_preset_info(bs_theme(version = 3, bootswatch = "yeti")),
      new_bs_preset("yeti", version = "3", type = "bootswatch")
    )
  })

  it("returns builtin preset theme information", {
    expect_equal(
      theme_preset_info(bs_theme(version = 5, preset = "shiny")),
      new_bs_preset("shiny", version = "5", type = "builtin")
    )
  })

  it("returns vanilla bootstrap theme information", {
    expect_equal(
      theme_preset_info(bs_theme(preset = "bootstrap", version = 5)),
      new_bs_preset("bootstrap", version = "5")
    )

    expect_equal(
      theme_preset_info(bs_theme(version = 4)),
      new_bs_preset("bootstrap", version = "4")
    )

    expect_equal(
      theme_preset_info(bs_theme(version = 3)),
      new_bs_preset("bootstrap", version = "3")
    )
  })

  it("returns NULL if not given a bs_theme object", {
    expect_null(theme_preset_info(list()))
  })
})

tag_require <- function(tag, version = version_default(), caller = "") {
  tagAddRenderHook(
    tag,
    replace = FALSE,
    func = function(x) {
      # If we know for sure the version isn't sufficient, it's safe to throw
      current_version <- theme_version(bs_current_theme())
      if (isTRUE(current_version < version)) {
        stop(
          caller,
          " requires Bootstrap ",
          version,
          " or higher. ",
          "To specify the version of Bootstrap, see https://rstudio.github.io/bslib/#basic-usage",
          call. = FALSE
        )
      }
      # We generally don't know the theme/version if any of these conditions are true:
      # 1. We're inside an Rmd output format that doesn't run through rmarkdown::html_document_base
      #    * pkgdown is one known case where the Bootstrap version may be customized, but bslib
      #      doesn't currently have a way to know what the version is.
      # 2. shiny::bootstrapLib() is being called while shiny is not shiny::isRunning()
      #   * At one point I was hoping bootstrapLib() could set the relevant context when
      #     statically rendered, but we didn't end up merging this
      #     https://github.com/rstudio/htmltools/pull/267
      # 3. Someone else is providing the bootstrap dependency
      #   * I currently don't know of any cases where this is relevant, but it might be
      #
      # So, since there are totally legitimate cases where the version requirement
      # could be met, but we don't know for sure what's happening server-side,
      # resort to a client-side check/warning
      return(tag_require_client_side(x, version, caller))
    }
  )
}

tag_require_client_side <- function(
  tag,
  version = version_default(),
  caller = ""
) {
  tagAppendChild(
    tagAppendAttributes(
      tag,
      "data-require-bs-version" = version,
      "data-require-bs-caller" = caller
    ),
    htmlDependency(
      name = "bslib-tag-require",
      version = get_package_version("bslib"),
      package = "bslib",
      src = "components",
      script = "tag-require.js"
    )
  )
}

tag_add_outer_class <- function(x, class = NULL, ...) {
  if (is.null(class)) {
    return(x)
  }

  if (inherits(x, "shiny.tag")) {
    return(tagAppendAttributes(x, class = class, ...))
  }

  as_fill_carrier(div(x, class = class, ...))
}

#' Define breakpoint values
#'
#' @description
#' A generic constructor for responsive breakpoints.
#'
#' @param xs The default value to apply to the `xs` breakpoint. Note that this
#'   breakpoint is generally equivalent to "all sizes" and is typically treated
#'   as the base case or a value to apply by default across all breakpoints
#'   unless overridden by a larger breakpoint.
#' @param sm Values to apply at the `sm` breakpoint.
#' @param md Values to apply at the `md` breakpoint.
#' @param lg Values to apply at the `lg` breakpoint.
#' @param ... Other breakpoints (e.g., `xl`).
#'
#' @seealso `breakpoints()` is used by [layout_columns()].
#' @references Bootstrap's
#'   [Breakpoints article](https://getbootstrap.com/docs/5.3/layout/breakpoints/)
#'   provides more detail on breakpoints and how they are used and customized.
#'
#' @examples
#' breakpoints(sm = c(4, 4, 4), md = c(3, 3, 6), lg = c(-2, 8, -2))
#'
#' @export
breakpoints <- function(..., xs = NULL, sm = NULL, md = NULL, lg = NULL) {
  breaks <- dropNulls(
    rlang::list2(..., xs = xs, sm = sm, md = md, lg = lg)
  )

  if (any_unnamed(breaks)) {
    abort("All `breakpoints` values must be named")
  }

  # Ensure that breakpoints are in the increasing order
  break_nms <- sort(
    factor(
      names(breaks),
      ordered = TRUE,
      levels = unique(c(bs_breakpoints(), names(breaks)))
    )
  )

  structure(
    breaks[as.character(break_nms)],
    class = "bslib_breakpoints"
  )
}

#' @export
print.bslib_breakpoints <- function(x, ...) {
  cat("<breakpoints>\n")

  width_vals <- max(nchar(unlist(x)), na.rm = TRUE)
  width_nms <- max(nchar(names(x)), na.rm = TRUE)

  for (bp in names(x)) {
    vals <- format(x[[bp]], width = width_vals, justify = "right")
    cat(
      " ",
      format(bp, width = width_nms, justify = "right"),
      ": ",
      paste0(vals, collapse = " "),
      "\n",
      sep = ""
    )
  }

  invisible(x)
}

is_breakpoints <- function(x) {
  inherits(x, "bslib_breakpoints")
}

bs_breakpoints <- function() {
  c("xs", "sm", "md", "lg", "xl", "xxl")
}

# @staticimports pkg:staticimports
#  is_installed get_package_version system_file
#  register_upgrade_message
#  %||% is_string
#  read_utf8

switch_version <- function(
  version,
  five = default,
  four = default,
  three = default,
  default = NULL
) {
  if (is_bs_theme(version)) {
    version <- theme_version(version)
  }
  version <- as.character(version)
  if (isTRUE(version %in% c("4-3", "4+3"))) {
    warning("Version ", version, " has been renamed to 4. Please use 4 instead")
    version <- "4"
  }
  switch(
    version,
    `5` = five,
    `4` = four,
    `3` = three,
    stop("Didn't recognize Bootstrap version: ", version, call. = FALSE)
  )
}

get_exact_version <- function(version) {
  switch_version(
    version,
    five = version_bs5,
    four = version_bs4,
    three = version_bs3
  )
}

path_inst <- function(...) {
  files <- system_file(..., package = "bslib")
  files_found <- files != ""
  if (all(files_found)) return(files)

  files_not_found <- file.path(...)[!files_found]
  stop(
    "bslib file not found: '",
    files_not_found,
    "'",
    call. = FALSE
  )
}

path_lib <- function(...) path_inst("lib", ...)

is_shiny_app <- function() {
  # Make sure to not load shiny as a side-effect of calling this function.
  isNamespaceLoaded("shiny") && shiny::isRunning()
}

is_hosted_app <- function() {
  nzchar(Sys.getenv("SHINY_SERVER_VERSION")) && is_shiny_app()
}

is_shiny_runtime <- function() {
  if (!is_installed("knitr")) return(FALSE)
  isTRUE(grepl("^shiny", knitr::opts_knit$get("rmarkdown.runtime")))
}

register_runtime_package_check <- function(feature, pkg, version) {
  msg <- sprintf(
    "%s is designed to work with %s %s or higher",
    feature,
    pkg,
    version
  )

  if (isNamespaceLoaded(pkg) && !is_installed(pkg, version)) {
    warning(msg, call. = FALSE)
  }

  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      if (!is_installed(pkg, version)) warning(msg, call. = FALSE)
    }
  )
}

add_class <- function(x, y) {
  class(x) <- unique(c(y, oldClass(x)))
  x
}

dropNulls <- function(x) {
  x[!vapply(x, is.null, FUN.VALUE = logical(1))]
}

names2 <- function(x) {
  names(x) %||% rep.int("", length(x))
}

any_unnamed <- function(x) {
  if (length(x) == 0) return(FALSE)
  nms <- names(x)
  is.null(nms) || !all(nzchar(nms))
}

separate_arguments <- function(...) {
  x <- rlang::list2(...)
  x_names <- rlang::names2(x)
  is_named <- nzchar(x_names)

  if (all(is_named)) {
    return(list(attribs = x, children = list()))
  }

  if (!any(is_named)) {
    return(list(attribs = list(), children = x))
  }

  list(attribs = x[is_named], children = unname(dropNulls(x[!is_named])))
}

#' Rename a named list
#'
#' @param x a named list to be renamed
#' @param nms a named character vector defining the renaming
#' @noRd
#' @examples
#' rename2(list(a=1, b=3, c=4, a=2), b="z", f="w", a="y")
#' #> list(y = 1, z = 3, c = 4, y = 2)
#' rename2(c("a", "b", "c", "a"), b="z", f="w", a="y")
#' #> c("y", "z", "c", "y")
rename2 <- function(x, ...) {
  defs <- rlang::list2(...)
  nms <- names(x) %||% as.character(x)
  matches <- intersect(nms, names(defs))
  map <- match(nms, names(defs))
  mapidxNA <- is.na(map)
  replacement <- as.character(defs)[map[!mapidxNA]]
  if (is.null(names(x))) {
    x[!mapidxNA] <- replacement
  } else {
    names(x)[!mapidxNA] <- replacement
  }
  x
}

# Get an accessible color contrast for a specified bg_color
# (and return NULL+warn on failure)
get_color_contrast <- function(bg_color) {
  utils_layer <- sass::sass_layer(
    functions = sass::sass_file(
      path_inst("sass-utils", "color-contrast.scss")
    ),
    defaults = c(
      "$black: #000000;",
      "$white: #FFFFFF;"
    ),
    rules = sprintf(
      "._ {--RET: #{color-contrast(%s)}}",
      bg_color
    )
  )

  tryCatch(
    {
      css <- sass::sass(
        utils_layer,
        cache_key_extra = get_package_version("bslib"),
        # Don't listen to global Sass options so we can be sure
        # that stuff like source maps won't be included
        options = sass::sass_options(source_map_embed = FALSE)
      )
      # example: css <- "._ {\n  --RET: #fff;\n}"
      # we'll split to get value:     ^     ^
      ret <- strsplit(css, "--RET:")[[1]][2]
      trimws(strsplit(ret, ";")[[1]][1])
    },
    error = function(err) {
      warning(
        "Failed to compute a contrasting color for '",
        bg_color,
        "'",
        call. = FALSE
      )
      NULL
    }
  )
}

test_that("path_lib returns multiple system files", {
  # version 4 has a js file and a js map file
  js_files <- bootstrap_javascript("4")
  expect_equal(length(js_files), 1)
  js_files <- bootstrap_javascript("3")
  expect_equal(length(js_files), 1)
  js_files <- bootstrap_javascript_map("4")
  expect_equal(length(js_files), 1)
  js_files <- bootstrap_javascript_map("3")
  expect_equal(length(js_files), 0)
})

test_that("get_color_contrast() works", {
  expect_equal(get_color_contrast("white"), "#000000")
  expect_equal(get_color_contrast("black"), "#FFFFFF")
  expect_warning(expect_null(get_color_contrast("fsdnffdgdsfsd")))
})

test_that("separate_arguments() works", {
  expect_equal(
    separate_arguments(1, 2, 3)$children,
    list(1, 2, 3)
  )

  expect_equal(
    separate_arguments(a = 1, b = 2, c = 3)$attribs,
    list(a = 1, b = 2, c = 3)
  )

  expect_equal(
    separate_arguments(a = 1, 2, 3),
    list(attribs = list(a = 1), children = list(2, 3))
  )

  expect_equal(
    separate_arguments(1, a = 2, 3),
    list(attribs = list(a = 2), children = list(1, 3))
  )
})

test_that("theme_version() works for current versions", {
  bs_versions <- versions()

  for (bs_v in bs_versions) {
    expect_equal(theme_version(bs_theme(!!bs_v)), !!bs_v)
  }
})

test_that("theme_version() uses class first, then sass bundle", {
  theme <- theme_unclassed <- bs_theme(version = 5)
  class(theme_unclassed) <- setdiff(class(theme), "bs_version_5")
  theme_99 <- add_class(theme_unclassed, "bs_version_99")

  # These should use the class information
  expect_equal(theme_version(theme), "5")
  expect_equal(theme_version(theme_99), "99")

  # These should consult the sass bundle
  expect_equal(theme_version(add_class(theme, "bs_version_99")), "5")
  expect_equal(theme_version(theme_unclassed), "5")
})

#' Global theming
#'
#' `bs_global_theme()` creates and sets the global Bootstrap Sass theme. This
#' theme is typically found by [bs_theme_dependencies()] in the app or document
#' where the global theme is being used. You can obtain the current global theme
#' with [bs_global_get()] or directly set the global theme with
#' [bs_global_set()].
#'
#' @inheritParams bs_theme
#'
#' @return Functions that modify the global theme (e.g., `bs_global_set()`)
#'   invisibly return the previously set theme. `bs_global_get()` returns the
#'   current global theme.
#'
#' @family Bootstrap theme functions
#'
#' @examples
#' # Remember the global state now (so we can restore later)
#' theme <- bs_global_get()
#'
#' # Use Bootstrap 3 (globally) with some theme customization
#' bs_global_theme(3, bg = "#444", fg = "#e4e4e4", primary = "#e39777")
#' if (rlang::is_interactive()) {
#'   bs_theme_preview(with_themer = FALSE)
#' }
#'
#' # If no global theme is active, bs_global_get() returns NULL
#' bs_global_clear()
#' bs_global_get()
#'
#' # Restore the original state
#' bs_global_set(theme)
#'
#' @export
bs_global_theme <- function(
  version = version_default(),
  preset = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  ...,
  bootswatch = NULL
) {
  bs_global_set(
    bs_theme(
      version,
      preset = preset,
      bootswatch = bootswatch,
      bg = bg,
      fg = fg,
      primary = primary,
      secondary = secondary,
      success = success,
      info = info,
      warning = warning,
      danger = danger,
      base_font = base_font,
      code_font = code_font,
      heading_font = heading_font,
      ...
    )
  )
}

#' @rdname bs_global_theme
#' @inheritParams bs_theme_update
#' @export
bs_global_set <- function(theme = bs_theme()) {
  if (!is.null(theme)) {
    assert_bs_theme(theme)
  }
  # In addition to setting a bslib global option, also set shiny's
  # current theme if this code is running in an `runtime: shiny` doc
  if (is_shiny_runtime() && is_installed("shiny", "1.6")) {
    warning(
      "bs_global_set() may not work as expected inside runtime: shiny documents. ",
      "To update the document's theme, use `session$setCurrentTheme()` instead.",
      call. = FALSE
    )
  }
  old_theme <- options(bslib_theme = theme)
  invisible(old_theme[["bslib_theme"]])
}

#' @rdname bs_global_theme
#' @export
bs_global_get <- function() {
  getOption("bslib_theme")
}

#' @rdname bs_global_theme
#' @export
bs_global_clear <- function() {
  old_theme <- options(bslib_theme = NULL)
  invisible(old_theme[["bslib_theme"]])
}

#' @rdname bs_global_theme
#' @inheritParams bs_add_variables
#' @export
bs_global_add_variables <- function(
  ...,
  .where = "defaults",
  .default_flag = identical(.where, "defaults")
) {
  theme <- assert_global_theme("bs_global_add_variables()")
  theme <- bs_add_variables(
    theme,
    ...,
    .where = .where,
    .default_flag = .default_flag
  )
  bs_global_set(theme)
}

#' @rdname bs_global_theme
#' @export
bs_global_add_rules <- function(...) {
  theme <- assert_global_theme("bs_global_add_rules()")
  theme <- bs_add_rules(theme, ...)
  bs_global_set(theme)
}

#' @rdname bs_global_theme
#' @export
bs_global_bundle <- function(...) {
  theme <- assert_global_theme("bs_global_bundle()")
  theme <- bs_bundle(theme, ...)
  bs_global_set(theme)
}

assert_global_theme <- function(calling_func) {
  theme <- bs_global_get()
  if (is.null(theme)) {
    stop(
      "`",
      calling_func,
      "` requires that a global theme is first set (do you want to call `bs_global_theme()`?)"
    )
  }
  theme
}

test_that("global theme api works", {
  bs_global_theme(version = 4, bootswatch = "sketchy")
  on.exit(bs_global_clear(), add = TRUE)

  # Can retrieve theme version/bootswatch
  sketchy_theme <- bs_global_get()
  expect_equal(theme_version(sketchy_theme), "4")
  expect_equal(theme_bootswatch(sketchy_theme), "sketchy")

  # Setting a new theme overrides the old one
  bs_global_theme(version = 4, bootswatch = "cosmo")
  cosmo_theme <- bs_global_get()
  expect_equal(theme_version(cosmo_theme), "4")
  expect_equal(theme_bootswatch(cosmo_theme), "cosmo")

  # Can add new variable defaults
  bs_global_add_variables("primary" = "red !default;")
  primary <- sass_partial("body{background:$primary;}", bs_global_get())
  expect_css("body{background:red;}", primary)

  # Can clear
  bs_global_clear()
  expect_null(bs_global_get())

  # Adding without a set errors
  expect_error(bs_global_bundle())

  # Can set a theme object as the current theme
  bs_global_set(cosmo_theme)
  primary <- sass_partial("body{background:$primary;}", bs_global_get())
  expect_css("body{background:#2780e3;}", primary)
})


test_that("Theme adding works as intended", {
  bs_global_theme()
  on.exit(bs_global_clear(), add = TRUE)
  # Can override variable defaults
  bs_global_add_variables(primary = "blue !default;")
  bs_global_add_variables(primary = "#fff !default;")
  css <- sass_partial(".foo{color:$primary;}", bs_global_get())
  expect_css(".foo{color:#fff;}", css)

  # Also works without default flags and can handle numeric values
  bs_global_add_variables(primary = "blue")
  bs_global_add_variables(primary = "#fff", "font-size" = 0)
  css <- sass_partial(".foo{color:$primary;font-size:0}", bs_global_get())
  expect_css(".foo{color:#fff;font-size:0;}", css)

  # Can also override variables via declarations
  bs_global_add_variables(
    .where = "mixins",
    "primary" = "$secondary",
    "body-color" = "color-contrast($primary)"
  )
  css <- sass_partial(".foo{color:$primary;}", bs_global_get())
  expect_css(".foo{color:#6c757d;}", css)
})


test_that("rename works as intended", {
  expect_identical(
    rename2(list(a=1, b=3, c=4, a=2), b="z", f="w", a="y"),
    list(y = 1, z = 3, c = 4, y = 2)
  )
  expect_identical(
    rename2(c("a", "b", "c", "a"), b="z", f="w", a="y"),
    c("y", "z", "c", "y")
  )
})

test_that("global theme api works", {
  bs_global_theme(version = 4, bootswatch = "sketchy")
  on.exit(bs_global_clear(), add = TRUE)

  # Can retrieve theme version/bootswatch
  sketchy_theme <- bs_global_get()
  expect_equal(theme_version(sketchy_theme), "4")
  expect_equal(theme_bootswatch(sketchy_theme), "sketchy")

  # Setting a new theme overrides the old one
  bs_global_theme(version = 4, bootswatch = "cosmo")
  cosmo_theme <- bs_global_get()
  expect_equal(theme_version(cosmo_theme), "4")
  expect_equal(theme_bootswatch(cosmo_theme), "cosmo")

  # Can add new variable defaults
  bs_global_add_variables("primary" = "red !default;")
  primary <- sass_partial("body{background:$primary;}", bs_global_get())
  expect_css("body{background:red;}", primary)

  # Can clear
  bs_global_clear()
  expect_null(bs_global_get())

  # Adding without a set errors
  expect_error(bs_global_bundle())

  # Can set a theme object as the current theme
  bs_global_set(cosmo_theme)
  primary <- sass_partial("body{background:$primary;}", bs_global_get())
  expect_css("body{background:#2780e3;}", primary)
})

test_that("Theme adding works as intended", {
  bs_global_theme()
  on.exit(bs_global_clear(), add = TRUE)
  # Can override variable defaults
  bs_global_add_variables(primary = "blue !default;")
  bs_global_add_variables(primary = "#fff !default;")
  css <- sass_partial(".foo{color:$primary;}", bs_global_get())
  expect_css(".foo{color:#fff;}", css)

  # Also works without default flags and can handle numeric values
  bs_global_add_variables(primary = "blue")
  bs_global_add_variables(primary = "#fff", "font-size" = 0)
  css <- sass_partial(".foo{color:$primary;font-size:0}", bs_global_get())
  expect_css(".foo{color:#fff;font-size:0;}", css)

  # Can also override variables via declarations
  bs_global_add_variables(
    .where = "mixins",
    "primary" = "$secondary",
    "body-color" = "color-contrast($primary)"
  )
  css <- sass_partial(".foo{color:$primary;}", bs_global_get())
  expect_css(".foo{color:#404040;}", css)
})

test_that("rename works as intended", {
  expect_identical(
    rename2(list(a = 1, b = 3, c = 4, a = 2), b = "z", f = "w", a = "y"),
    list(y = 1, z = 3, c = 4, y = 2)
  )
  expect_identical(
    rename2(c("a", "b", "c", "a"), b = "z", f = "w", a = "y"),
    c("y", "z", "c", "y")
  )
})

#' Column-first uniform grid layouts
#'
#' @description
#' Wraps a 1d sequence of UI elements into a 2d grid. The number of columns (and
#' rows) in the grid dependent on the column `width` as well as the size of the
#' display. For more explanation and illustrative examples, see the _References_
#' section below.
#'
#' @param ... Unnamed arguments should be UI elements (e.g., [card()]). Named
#'   arguments become attributes on the containing [htmltools::tag] element.
#' @param width The desired width of each card, which can be any of the
#'  following:
#'   * A (unit-less) number between 0 and 1.
#'     * This should be specified as `1/num`, where `num` represents the number
#'       of desired columns.
#'   * A [CSS length unit][htmltools::validateCssUnit()]
#'     * Either the minimum (when `fixed_width=FALSE`) or fixed width
#'       (`fixed_width=TRUE`).
#'   * `NULL`
#'     * Allows power users to set the `grid-template-columns` CSS property
#'       manually, either via a `style` attribute or a CSS stylesheet.
#' @param fixed_width When `width` is greater than 1 or is a CSS length unit,
#'   e.g. `"200px"`, `fixed_width` indicates whether that `width` value
#'   represents the absolute size of each column (`fixed_width=TRUE`) or the
#'   minimum size of a column (`fixed_width=FALSE`). When `fixed_width=FALSE`,
#'   new columns are added to a row when `width` space is available and columns
#'   will never exceed the container or viewport size. When `fixed_width=TRUE`,
#'   all columns will be exactly `width` wide, which may result in columns
#'   overflowing the parent container.
#' @param heights_equal If `"all"` (the default), every card in every row of the
#'   grid will have the same height. If `"row"`, then every card in _each_ row
#'   of the grid will have the same height, but heights may vary between rows.
#' @param fill Whether or not to allow the layout to grow/shrink to fit a
#'   fillable container with an opinionated height (e.g., `page_fillable()`).
#' @param fillable Whether or not each element is wrapped in a fillable container.
#' @param height_mobile Any valid CSS unit to use for the height when on mobile
#'   devices (or narrow windows).
#' @param min_height,max_height The maximum or minimum height of the layout container.
#'   Can be any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `max_height="200px"`). Use these arguments in filling layouts to ensure that a
#'   layout container doesn't shrink below `min_height` or grow beyond `max_height`.
#' @inheritParams card
#' @inheritParams card_body
#'
#' @export
#' @family Column layouts
#'
#' @references The bslib website features `layout_column_wrap()` in two places:
#'   * [Column-based layouts](https://rstudio.github.io/bslib/articles/column-layout/index.html)
#'   * [Cards: Multiple cards](https://rstudio.github.io/bslib/articles/cards/index.html#multiple-cards)
#'
#' @examplesIf rlang::is_interactive()
#' x <- card("A simple card")
#'
#' # Always has 2 columns (on non-mobile)
#' layout_column_wrap(width = 1/2, x, x, x)
#'
#' # Automatically lays out three cards into columns
#' # such that each column is at least 200px wide:
#' layout_column_wrap(x, x, x)
#'
#' # To use larger column widths by default, set `width`.
#' # This example has 3 columns when the screen is at least 900px wide:
#' layout_column_wrap(width = "300px", x, x, x)
#'
#' # You can add a list of items, spliced with rlang's `!!!` operator
#' layout_column_wrap(!!!list(x, x, x))
#'
#' @export
layout_column_wrap <- function(
  ...,
  width = "200px",
  fixed_width = FALSE,
  heights_equal = c("all", "row"),
  fill = TRUE,
  fillable = TRUE,
  height = NULL,
  height_mobile = NULL,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL
) {
  heights_equal <- match.arg(heights_equal)

  args <- separate_arguments(...)
  attribs <- args$attribs
  children <- args$children

  if (missing(width)) {
    first_is_width <-
      is.null(children[[1]]) ||
        is_probably_a_css_unit(children[[1]])

    if (first_is_width) {
      # Assume an unnamed first argument that matches our expectations for
      # `width` is actually the width argument, with a warning
      lifecycle::deprecate_warn(
        "0.6.0",
        "layout_column_wrap(width = 'must be named')"
      )
      width <- children[[1]]
      children <- children[-1]
    }
  }

  if (length(width) > 1) {
    stop("`width` of length greater than 1 is not currently supported.")
  }

  colspec <- if (!is.null(width)) {
    if (width > 0 && width <= 1) {
      num_cols <- 1 / width
      if (num_cols != as.integer(num_cols)) {
        stop("Could not interpret width argument; see ?layout_column_wrap")
      }
      sprintf("repeat(%s, minmax(0, 1fr))", num_cols)
    } else {
      if (fixed_width) {
        paste0("repeat(auto-fit, ", validateCssUnit(width), ")")
      } else {
        sprintf(
          "repeat(auto-fit, minmax(min(%s, 100%%), 1fr))",
          validateCssUnit(width)
        )
      }
    }
  }

  # If relevant (i.e., fillable=TRUE), wrap each child in a fillable context
  # (so fill/flex items can fill the available area)
  children <- lapply(children, grid_item_container, fillable = fillable)

  tag <- div(
    class = "bslib-grid bslib-mb-spacing",
    style = css(
      grid_template_columns = colspec,
      grid_auto_rows = if (heights_equal == "all") "1fr",
      # Always provide the `height:auto` default so that the CSS variable
      # doesn't get inherited in a scenario like layout_column_wrap(height=200, ..., layout_column_wrap(...))
      "--bslib-grid-height" = validateCssUnit(height %||% "auto"),
      "--bslib-grid-height-mobile" = validateCssUnit(height_mobile %||% "auto"),
      gap = validateCssUnit(gap),
      min_height = validateCssUnit(min_height),
      max_height = validateCssUnit(max_height)
    ),
    !!!attribs,
    children,
    component_dependencies()
  )

  tag <- bindFillRole(tag, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "layout_column_wrap()")
  )
}

is_probably_a_css_unit <- function(x) {
  if (length(x) != 1) return(FALSE)
  if (is.numeric(x)) return(TRUE)
  if (!is.character(x)) return(FALSE)
  tryCatch(
    {
      validateCssUnit(x)
      TRUE
    },
    error = function(e) FALSE
  )
}

#' Responsive 12-column grid layouts
#'
#' Create responsive, column-based grid layouts, based on a 12-column grid.
#'
#' @inheritParams layout_column_wrap
#' @param col_widths One of the following:
#'   * `NA` (the default): Automatically determines a sensible number of columns
#'     based on the number of children.
#'   * A numeric vector of integers between 1 and 12, where each element
#'     represents the number of columns for the relevant UI element. Elements
#'     that happen to go beyond 12 columns wrap onto the next row. For example,
#'     `col_widths = c(4, 8, 12)` allocates 4 columns to the first element, 8
#'     columns to the second element, and 12 columns to the third element (which
#'     wraps to the next row). Negative values are also allowed, and are treated
#'     as empty columns. For example, `col_widths = c(-2, 8, -2)` would allocate
#'     8 columns to an element (with 2 empty columns on either side).
#'   * A [breakpoints()] object, where each breakpoint may be either of the
#'     above.
#' @param row_heights One of the following:
#'   * A numeric vector, where each value represents the
#'     [fractional unit](https://css-tricks.com/introduction-fr-css-unit/)
#'     (`fr`) height of the relevant row. If there are more rows than values
#'     provided, the pattern will repeat. For example, `row_heights = c(1, 2)`
#'     allows even rows to take up twice as much space as odd rows.
#'   * A list of numeric and [CSS length units][htmltools::validateCssUnit()],
#'     where each value represents the height of the relevant row. If more rows
#'     are needed than values provided, the pattern will repeat. For example,
#'     `row_heights = list("auto", 1)` allows the height of odd rows to be
#'     driven my it's contents and even rows to be
#'     [`1fr`](https://css-tricks.com/introduction-fr-css-unit/).
#'   * A character vector/string of [CSS length units][htmltools::validateCssUnit()].
#'     In this case, the value is supplied directly to `grid-auto-rows`.
#'   * A [breakpoints()] object, where each breakpoint may be either of the above.
#'
#' @export
#' @family Column layouts
#'
#' @references [Column-based layouts](https://rstudio.github.io/bslib/articles/column-layout/index.html)
#'   on the bslib website.
#'
#' @seealso [breakpoints()] for more information on specifying column widths at
#'   responsive breakpoints.
#'
#' @examplesIf rlang::is_interactive()
#' x <- card("A simple card")
#'
#' page_fillable(
#'   layout_columns(x, x, x, x)
#' )
#'
#' # Or add a list of items, spliced with rlang's `!!!` operator
#' page_fillable(
#'  layout_columns(!!!list(x, x, x))
#' )
#'
#' page_fillable(
#'   layout_columns(
#'     col_widths = c(6, 6, 12),
#'     x, x, x
#'   )
#' )
#'
#' page_fillable(
#'   layout_columns(
#'     col_widths = c(6, 6, -2, 8),
#'     row_heights = c(1, 3),
#'     x, x, x
#'   )
#' )
#'
#' page_fillable(
#'   fillable_mobile = TRUE,
#'   layout_columns(
#'     col_widths = breakpoints(
#'       sm = c(12, 12, 12),
#'       md = c(6, 6, 12),
#'       lg = c(4, 4, 4)
#'     ),
#'     x, x, x
#'   )
#' )
#'
layout_columns <- function(
  ...,
  col_widths = NA,
  row_heights = NULL,
  fill = TRUE,
  fillable = TRUE,
  gap = NULL,
  class = NULL,
  height = NULL,
  min_height = NULL,
  max_height = NULL
) {
  args <- separate_arguments(...)
  attribs <- args$attribs
  children <- args$children
  n_kids <- length(children)

  # Check for spec problems so we can stop early
  col_spec <- as_col_spec(col_widths, n_kids)

  # Wrap each child in a container (so fill/flex items can fill the available area)
  children <- Map(
    f = grid_item_container,
    children,
    fillable = fillable
  )

  tag <- web_component(
    "bslib-layout-columns",
    class = "bslib-grid grid bslib-mb-spacing",
    style = css(
      height = validateCssUnit(height),
      gap = validateCssUnit(gap),
      min_height = validateCssUnit(min_height),
      max_height = validateCssUnit(max_height)
    ),
    # We don't enable the next option by default, but users could add this
    # attribute to hide the internal elements until after the custom element
    # adds classes to all of the elements. We don't think this is needed but
    # it's here if we find there are cases where rendering the children and
    # adding classes takes longer than we've anticipated in our testing.
    # "hidden-until-init" = NA,
    !!!col_widths_attrs(col_spec),
    !!!row_heights_css_vars(row_heights),
    !!!attribs,
    !!!children
  )

  tag <- bindFillRole(tag, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "layout_columns()")
  )
}

as_col_spec <- function(col_widths, n_kids) {
  if (is.null(col_widths) || rlang::is_na(col_widths)) return(NULL)

  if (!is_breakpoints(col_widths)) {
    col_widths <- breakpoints(sm = col_widths)
  }

  for (break_name in names(col_widths)) {
    bk <- col_widths[[break_name]]

    if (rlang::is_na(bk)) {
      next
    }

    if (isTRUE(any(bk == 0))) {
      abort(
        "Column values must be greater than 0 to indicate width, or negative to indicate a column offset."
      )
    }

    if (length(bk) > 1 && anyNA(bk)) {
      abort(
        "Cannot mix widths and `NA` values. All column widths must be specified, or choose auto widths using a single `NA` value."
      )
    }

    if (!any(bk > 0)) {
      abort("Column values must include at least one positive integer width.")
    }

    if (length(bk[bk > 0]) > n_kids) {
      rlang::warn(
        sprintf(
          "More column widths than children at breakpoint '%s', extra widths will be ignored.",
          break_name
        )
      )
    }
  }

  col_widths
}

col_widths_attrs <- function(col_spec) {
  if (is.null(col_spec) || rlang::is_na(col_spec)) return(NULL)

  names(col_spec) <- paste0("col-widths-", names(col_spec))
  lapply(col_spec, function(x) {
    if (rlang::is_na(x)) NA else paste(x, collapse = ",")
  })
}

maybe_fr_unit <- function(x) {
  if (is.numeric(x)) sprintf("%0.0ffr", x) else x
}

row_heights_css_vars <- function(x) {
  if (is.null(x)) return(list())

  if (!is_breakpoints(x)) {
    # Setting the `xs` breakpoint is equivalent to setting all breaks
    x <- breakpoints(xs = x)
  }

  # creates classes that pair with CSS variables when set
  classes <- paste0("bslib-grid--row-heights--", names(x))

  css_vars <- setNames(x, paste0("--", classes))

  # Treat numeric values as fractional units
  css_vars <- rapply(css_vars, how = "replace", maybe_fr_unit)

  if (identical(names(css_vars), "--bslib-grid--row-heights--xs")) {
    names(css_vars) <- "--bslib-grid--row-heights"
    classes <- character()
  }

  list(
    style = css(!!!css_vars),
    class = classes
  )
}

grid_item_container <- function(el, ..., fillable = TRUE) {
  div(
    ...,
    class = "bslib-grid-item",
    class = if (fillable) "bslib-gap-spacing",
    if (fillable) as_fillable_container(),
    el
  )
}

test_that("layout_columns() with col_widths", {
  children <- lapply(1:2, function(...) {
    div(class = "layout-column-child-element")
  })

  expect_snapshot(
    layout_columns(col_widths = 6, !!!children)
  )

  expect_snapshot(
    layout_columns(col_widths = c(4, 8), !!!children)
  )

  expect_snapshot(
    layout_columns(
      col_widths = breakpoints(sm = 6, md = 4, lg = 3),
      !!!children
    )
  )

  expect_snapshot(
    layout_columns(col_widths = breakpoints(sm = NA, lg = c(4, 8)), !!!children)
  )

  expect_snapshot(
    layout_columns(
      col_widths = breakpoints(sm = NA, lg = c(4, -4, 4)),
      !!!children
    )
  )
})

test_that("layout_columns() without `col_widths`", {
  expect_no_match(
    format(layout_columns(div())),
    "col-widths-"
  )

  expect_no_match(
    format(layout_columns(div(), col_widths = NULL)),
    "col-widths-"
  )

  expect_no_match(
    format(layout_columns(div(), col_widths = list(NA))),
    "col-widths-"
  )
})

test_that("grid_item_container()", {
  expect_snapshot(
    grid_item_container(
      div(class = "layout-column-child-element"),
      fillable = TRUE
    )
  )

  expect_snapshot(
    grid_item_container(
      div(class = "layout-column-child-element"),
      fillable = FALSE
    )
  )
})

test_that("grid-breakpoints are known", {
  brks <- bs_get_variables(
    bs_add_variables(
      bs_theme(),
      "grid-brk-names" = "map-keys($grid-breakpoints)",
      .where = "declarations"
    ),
    "grid-brk-names"
  )

  # In bslib::breakpoints() we flag these as "known" breakpoints...
  # which mostly means we order them, but also, in the NA case,
  # we use lg/xl/xxl to fill in md
  expect_equal(
    strsplit(brks, ", ")[[1]],
    c("xs", "sm", "md", "lg", "xl", "xxl")
  )
})

test_that("breakpoints() re-orders well-known breaks and test print method", {
  bp <- breakpoints(
    xl = 5,
    sm = 2,
    xs = 1,
    giant = 7,
    lg = 4,
    huge = 8,
    md = 3,
    xxl = 6
  )

  expect_snapshot(bp)
  expect_equal(unclass(unname(unlist(bp))), 1:8)
})

test_that("breakpoints() has correct classes and structure", {
  bp <- breakpoints(
    lg = c(-2, 2, -1, -2, 3, -2),
    md = c(-1, 2, -2, 3),
    sm = c(1, -1, 1),
    xl = c(1, 2, 3, NA),
    huge = list(1, 2, 3, "auto")
  )

  expect_true(is_breakpoints(bp))

  # names are ordered, custom breakpoint names are last
  expect_named(bp, c("sm", "md", "lg", "xl", "huge"))

  expect_snapshot(bp)
})

test_that("breakpoints() must be named", {
  expect_error(breakpoints(1:3))
})

test_that("layout_columns() allows a single NA in a breakpoint", {
  expect_silent(
    layout_columns(col_widths = breakpoints(sm = NA), div())
  )
})

test_that("layout_columns() throws if NAs are mixed with other column values", {
  expect_snapshot_error(
    layout_columns(col_widths = breakpoints(sm = c(-1, NA, 1)))
  )
})

test_that("layout_columns() warns when too many column widths", {
  expect_snapshot_warning(
    layout_columns(
      col_widths = breakpoints(md = c(1, 1, 1)),
      div()
    )
  )
})

test_that("breakpoint_columns() throws if a column is 0", {
  expect_error(
    col_width_grid_classes(breakpoints(md = c(-1, 0, 1)), 3)
  )
})

test_that("layout_columns() throws if no columns are positive", {
  expect_error(
    layout_columns(col_widths = breakpoints(md = -1), div())
  )
  expect_error(
    layout_columns(
      col_widths = breakpoints(md = c(-1, -1, -1)),
      div(),
      div(),
      div()
    )
  )
})

test_that("as_column_spec() excludes offsets from oversized col_widths warning", {
  expect_silent(
    as_col_spec(col_widths = c(4, -4, 4), 2)
  )
})

test_that("row_heights_css_vars() decides fr/px for numeric, passes character", {
  expect_equal(
    row_heights_css_vars(breakpoints(md = c(2, 1)))$style,
    "--bslib-grid--row-heights--md:2fr 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list(2, 1)))$style,
    "--bslib-grid--row-heights--md:2fr 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list("auto", 1)))$style,
    "--bslib-grid--row-heights--md:auto 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list("10px", 1)))$style,
    "--bslib-grid--row-heights--md:10px 1fr;"
  )
})

test_that("row_heights_css_vars() sets row heights at all sizes if no breakpoint is provided", {
  expect_equal(
    row_heights_css_vars(c(1, 2)),
    list(
      style = "--bslib-grid--row-heights:1fr 2fr;",
      class = character(0)
    )
  )
})

test_that("row_heights_css_vars() rounds fractional numeric values (fr must be integer)", {
  expect_equal(
    row_heights_css_vars(c(1.5, 2.5)), #<< R rounds 1.5 up and 2.5 down
    list(
      style = "--bslib-grid--row-heights:2fr 2fr;",
      class = character(0)
    )
  )
})

test_that("row_heights_css_vars() doesn't include the class for xs size", {
  expect_equal(
    row_heights_css_vars(breakpoints(xs = c(1, 2))),
    list(
      style = "--bslib-grid--row-heights:1fr 2fr;",
      class = character(0)
    )
  )
})

test_that("layout_column_wrap() handles deprecated width as first arg", {
  # first arg is fractional
  lifecycle::expect_deprecated(
    lc_implicit_width_frac <- layout_column_wrap(1 / 2, "one", "two")
  )

  expect_equal(
    as.character(lc_implicit_width_frac),
    as.character(layout_column_wrap(width = 1 / 2, "one", "two"))
  )

  # first arg is explicitly px character
  lifecycle::expect_deprecated(
    lc_implicit_width_px <- layout_column_wrap("400px", "one", "two")
  )

  expect_equal(
    as.character(lc_implicit_width_px),
    as.character(layout_column_wrap(width = "400px", "one", "two"))
  )

  # first arg is px, but numeric
  lifecycle::expect_deprecated(
    lc_implicit_width_px_implied <- layout_column_wrap(365, "one", "two")
  )

  expect_equal(
    as.character(lc_implicit_width_px_implied),
    as.character(layout_column_wrap(width = 365, "one", "two"))
  )

  # first arg is NULL
  lifecycle::expect_deprecated(
    lc_implicit_width_null <- layout_column_wrap(NULL, "one", "two")
  )

  expect_equal(
    as.character(lc_implicit_width_null),
    as.character(layout_column_wrap(width = NULL, "one", "two"))
  )

  # first arg is not a CSS unit
  rlang::local_options(lifecycle_verbosity = "warning")
  testthat::expect_silent(
    as.character(layout_column_wrap("1ft", "one", "two"))
  )
})

withr::with_dir(
  rprojroot::find_package_root_file("vignettes"), {

    src4 <- knitr::knit_expand(
      file = "_variables-template.Rmd",
      version = 4
    )
    writeLines(src4, "bs4-variables.Rmd")

    src5 <- knitr::knit_expand(
      file = "_variables-template.Rmd",
      version = 5
    )
    invisible(writeLines(src5, "bs5-variables.Rmd"))
  }
)

#' Button for launching longer-running operations
#'
#' @description
#' `input_task_button` is a button that can be used in conjuction with
#' [shiny::bindEvent()] (or the older [shiny::eventReactive()] and
#' [shiny::observeEvent()] functions) to trigger actions or recomputation.
#'
#' It is similar to [shiny::actionButton()], except it prevents the user from
#' clicking when its operation is already in progress.
#'
#' Upon click, it automatically displays a customizable progress message and
#' disables itself; and after the server has dealt with whatever reactivity is
#' triggered from the click, the button automatically reverts to its original
#' appearance and re-enables itself.
#'
#' @section Manual button reset:
#' In some advanced use cases, it may be necessary to keep a task button in its
#' busy state even after the normal reactive processing has completed. Calling
#' `update_task_button(id, state = "busy")` from the server will opt out of any
#' currently pending reset for a specific task button. After doing so, the
#' button can be re-enabled by calling `update_task_button(id, state = "ready")`
#' after each click's work is complete.
#'
#' You can also pass an explicit `auto_reset = FALSE` to `input_task_button()`,
#' which means that button will _never_ be automatically re-enabled and will
#' require `update_task_button(id, state = "ready")` to be called each time.
#'
#' Note that, as a general rule, Shiny's `update` family of functions do not
#' take effect at the instant that they are called, but are held until the end
#' of the current reactive cycle. So if you have many different reactive
#' calculations and outputs, you don't have to be too careful about when you
#' call `update_task_button(id, state = "ready")`, as the button on the client
#' will not actually re-enable until the same moment that all of the updated
#' outputs simultaneously sent to the client.
#'
#' @param id The `input` slot that will be used to access the value.
#' @param label The label of the button while it is in ready (clickable) state;
#'   usually a string.
#' @param icon An optional icon to display next to the label while the button is
#'   in ready state. See [fontawesome::fa_i()].
#' @param label_busy The label of the button while it is busy.
#' @param icon_busy The icon to display while the button is busy. By default,
#'   `fontawesome::fa_i("refresh", class = "fa-spin", "aria-hidden" = "true")`
#'   is used, which displays a spinning "chasing arrows" icon. You can create
#'   spinning icons out of other Font Awesome icons by using the same
#'   expression, but replacing `"refresh"` with a different icon name. See
#'   [fontawesome::fa_i()].
#' @param type One of the Bootstrap theme colors (`"primary"`, `"default"`,
#'   `"secondary"`, `"success"`, `"danger"`, `"warning"`, `"info"`, `"light"`,
#'   `"dark"`), or `NULL` to leave off the Bootstrap-specific button CSS classes
#'   altogether.
#' @param ... Named arguments become attributes to include on the `<button>`
#'   element.
#' @param auto_reset If `TRUE` (the default), automatically put the button back
#'   in "ready" state after its click is handled by the server.
#'
#' @section Server value:
#' An integer of class `"shinyActionButtonValue"`. This class differs from
#' ordinary integers in that a value of 0 is considered "falsy".
#' This implies two things:
#'   * Event handlers (e.g., [shiny::observeEvent()], [shiny::eventReactive()]) won't execute on initial load.
#'   * Input validation (e.g., [shiny::req()], [shiny::need()]) will fail on initial load.
#'
#' @seealso [bind_task_button()]
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_sidebar(
#'   sidebar = sidebar(
#'     open = "always",
#'     input_task_button("resample", "Resample"),
#'   ),
#'   verbatimTextOutput("summary")
#' )
#'
#' server <- function(input, output, session) {
#'   sample <- eventReactive(input$resample, ignoreNULL=FALSE, {
#'     Sys.sleep(2)  # Make this artificially slow
#'     rnorm(100)
#'   })
#'
#'   output$summary <- renderPrint({
#'     summary(sample())
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
input_task_button <- function(
  id,
  label,
  ...,
  icon = NULL,
  label_busy = "Processing...",
  icon_busy = rlang::missing_arg(),
  type = "primary",
  auto_reset = TRUE
) {
  dots <- separate_arguments(...)
  attribs <- dots$attribs
  children <- dots$children

  icon_busy <- rlang::maybe_missing(
    icon_busy,
    fontawesome::fa_i("refresh", class = "fa-spin", "aria-hidden" = "true")
  )

  tags$button(
    id = id,
    class = if (!is.null(type)) paste0("btn btn-", type),
    class = "bslib-task-button",
    type = "button",
    "data-auto-reset" = if (isTRUE(auto_reset)) NA else NULL,
    !!!attribs,

    component_dependencies(),

    htmltools::tag(
      "bslib-switch-inline",
      rlang::list2(
        case = "ready",
        span(slot = "ready", icon, label),
        span(slot = "busy", icon_busy, label_busy),
        !!!children
      )
    )
  )
}

#' @param state If `"busy"`, put the button into busy/disabled state. If
#'   `"ready"`, put the button into ready/enabled state.
#' @param session The `session` object; using the default is recommended.
#' @rdname input_task_button
#' @export
update_task_button <- function(
  id,
  ...,
  state = NULL,
  session = get_current_session()
) {
  force(id)
  force(state)

  rlang::check_dots_empty()

  if (!is.null(state)) {
    if (!rlang::is_string(state)) {
      abort("`state` must be a single character value.")
    }
    set_task_button_manual_reset(
      session,
      id,
      manual = !identical(state, "ready")
    )
  }

  session$sendInputMessage(id, dropNulls(list(state = state)))
}

task_button_manual_reset_map <- function(session) {
  key <- "manual_task_button_reset"
  map <- session$userData[[key]]
  if (is.null(map)) {
    map <- fastmap::fastmap()
    session$userData[[key]] <- map
  }
  map
}

# Prevent automatic resetting of the task button when this reactive flush is
# complete
set_task_button_manual_reset <- function(session, id, manual = TRUE) {
  ns_id <- session$ns(id)
  map <- task_button_manual_reset_map(session)
  if (isTRUE(manual)) {
    map$set(ns_id, TRUE)
  } else {
    map$remove(ns_id)
  }
}

is_task_button_manual_reset <- function(session, id) {
  ns_id <- session$ns(id)
  map <- task_button_manual_reset_map(session)
  map$get(ns_id, FALSE)
}

input_task_button_input_handler <- function(val, shinysession, name) {
  value <- val[["value"]]

  if (isTRUE(val[["autoReset"]])) {
    shinysession$onFlush(once = TRUE, function() {
      # This is input_task_button's auto-reset feature: unless the button has
      # opted out using set_task_button_manual_reset(), we should reset after
      # a flush cycle where a bslib.taskbutton value is seen.
      if (!is_task_button_manual_reset(shinysession, name)) {
        update_task_button(name, state = "ready", session = shinysession)
      }
    })
  }

  # mark up the action button value with a special class so we can recognize it later
  class(value) <- c("shinyActionButtonValue", class(value))
  value
}

#' Bind `input_task_button` to `ExtendedTask`
#'
#' @description
#' Sets up a [shiny::ExtendedTask] to relay its state to an existing
#' [input_task_button()], so the task button stays in its "busy" state for as
#' long as the extended task is running.
#'
#' Note that `bind_task_button` does _not_ automatically cause button presses to
#' invoke the extended task; you still need to use [shiny::bindEvent()] (or
#' [shiny::observeEvent()]) to cause the button press to trigger an invocation,
#' as in the example below.
#'
#' `bind_task_button` cannot be used to bind one task button to multiple
#' `ExtendedTask` objects; if you attempt to do so, any bound `ExtendedTask`
#' that completes will cause the button to return to "ready" state.
#'
#' @param target The target object (i.e. `ExtendedTask`).
#' @param task_button_id A string matching the `id` argument passed to the
#'   corresponding [input_task_button()] call.
#' @param session A Shiny session object (the default should almost always be
#'   used).
#' @param ... Further arguments passed to other methods.
#'
#' @returns The `target` object that was passed in.
#'
#' @examplesIf rlang::is_interactive()
#'
#' library(shiny)
#' library(bslib)
#' library(future)
#' plan(multisession)
#'
#' ui <- page_sidebar(
#'   sidebar = sidebar(
#'     input_task_button("recalc", "Recalculate")
#'   ),
#'   textOutput("outval")
#' )
#'
#' server <- function(input, output) {
#'   rand_task <- ExtendedTask$new(function() {
#'     future({
#'       # Slow operation goes here
#'       Sys.sleep(2)
#'       runif(1)
#'     }, seed = TRUE)
#'   })
#'
#'   # Make button state reflect task.
#'   # If using R >=4.1, you can do this instead:
#'   # rand_task <- ExtendedTask$new(...) |> bind_task_button("recalc")
#'   bind_task_button(rand_task, "recalc")
#'
#'   observeEvent(input$recalc, {
#'     rand_task$invoke()
#'   })
#'
#'   output$outval <- renderText({
#'     rand_task$result()
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
bind_task_button <- function(target, task_button_id, ...) {
  UseMethod("bind_task_button")
}

#' @rdname bind_task_button
#' @export
bind_task_button.default <- function(target, task_button_id, ...) {
  abort(
    "Don't know how to bind a task button to an object of class '",
    class(target)[[0]],
    "'"
  )
}

#' @rdname bind_task_button
#' @export
bind_task_button.ExtendedTask <- function(
  target,
  task_button_id,
  ...,
  session = get_current_session()
) {
  force(target)
  force(task_button_id)
  force(session)

  was_running <- FALSE
  shiny::observe(
    {
      running <- target$status() == "running"
      if (running != was_running) {
        was_running <<- running
        if (running) {
          update_task_button(task_button_id, state = "busy", session = session)
        } else {
          update_task_button(task_button_id, state = "ready", session = session)
        }
      }
    },
    priority = 1000
  )
  return(target)
}

#' Deprecated functions
#'
#' These functions have been deprecated but remain for backwards compatibility.
#'
#' @section v0.9.1:
#'   The function `sidebar_toggle()` is now deprecated in v0.9.1. Please use
#'   [toggle_sidebar()] instead.
#'
#' @section Navigation Containers:
#'
#'   Several functions for navigation containers were renamed in version 0.5.0:
#'
#'   - `nav()` was renamed [nav_panel()]
#'   - `nav_content()` was renamed [nav_panel_hidden()]
#'   - `navs_tab()` was renamed [navset_tab()]
#'   - `navs_pill()` was renamed [navset_pill()]
#'   - `navs_pill_list()` was renamed [navset_pill_list()]
#'   - `navs_hidden()` was renamed [navset_hidden()]
#'   - `navs_bar()` was renamed [navset_bar()]
#'   - `navs_tab_card()` was renamed [navset_card_tab()]
#'   - `navs_pill_card()` was renamed [navset_card_pill()]
#'
#' @keywords internal
#' @name deprecated
#' @returns a [bs_theme()] object.
#' @export
bs_theme_new <- function(...) {
  .Deprecated("bs_global_theme")
  bs_global_theme(...)
}

#' @rdname deprecated
#' @export
bs_theme_clear <- function(...) {
  .Deprecated("bs_global_clear")
  bs_global_clear(...)
}

#' @rdname deprecated
#' @export
bs_theme_get <- function(...) {
  .Deprecated("bs_global_get")
  bs_global_get(...)
}

#' @rdname deprecated
#' @export
bs_theme_set <- function(...) {
  .Deprecated("bs_global_set")
  bs_global_set(...)
}

#' @rdname deprecated
#' @export
bs_theme_base_colors <- function(...) {
  .Deprecated("bs_global_base_colors")
  bs_global_base_colors(...)
}

bs_global_base_colors <- function(bg = "#FFFFFF", fg = "#000000") {
  theme <- assert_global_theme("bs_global_base_colors()")
  theme <- bs_base_colors(theme, bg = bg, fg = fg)
  bs_global_set(theme)
}

#' @rdname deprecated
#' @export
bs_theme_accent_colors <- function(...) {
  .Deprecated("bs_global_accent_colors")
  bs_global_accent_colors(...)
}

bs_global_accent_colors <- function(
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL
) {
  theme <- assert_global_theme("bs_global_accent_colors()")
  theme <- bs_accent_colors(
    theme,
    primary = primary,
    secondary = secondary,
    success = success,
    info = info,
    warning = warning,
    danger = danger
  )
  bs_global_set(theme)
}

#' @rdname deprecated
#' @export
bs_theme_fonts <- function(...) {
  .Deprecated("bs_global_fonts")
  bs_global_fonts(...)
}

bs_global_fonts <- function(base = NULL, code = NULL, heading = NULL) {
  theme <- assert_global_theme("bs_global_fonts()")
  theme <- bs_fonts(theme, base = base, code = code, heading = heading)
  bs_global_set(theme)
}

#' @rdname deprecated
#' @export
bs_theme_add_variables <- function(...) {
  .Deprecated("bs_global_add_variables")
  bs_global_add_variables(...)
}

#' @rdname deprecated
#' @export
bs_theme_add <- function(...) {
  .Deprecated("bs_global_bundle")
  bs_global_bundle(sass::sass_layer(...))
}

#' @rdname deprecated
#' @export
bs_theme_get_variables <- function(...) {
  .Deprecated("bs_global_get_variables")
  bs_global_get_variables(...)
}

bs_global_get_variables <- function(varnames) {
  theme <- assert_global_theme("bs_global_get_variables()")
  bs_get_variables(theme, varnames)
}

#' @rdname deprecated
#' @export
bootstrap <- function(theme = bs_theme_get(), ...) {
  .Deprecated("bs_theme_dependencies")
  bs_theme_dependencies(theme, ...)
}

#' @rdname deprecated
#' @export
bootstrap_sass <- function(rules = list(), theme = bs_theme_get(), ...) {
  .Deprecated("sass::sass_partial")
  theme <- as_bs_theme(theme)
  theme$rules <- ""
  sass_partial(rules, theme)
}

#' @rdname deprecated
#' @export
bs_add_declarations <- function(theme, declarations) {
  .Deprecated("bs_add_mixins")
  bs_bundle(theme, sass_layer(declarations = declarations))
}

#' @rdname deprecated
#' @export
card_body_fill <- function(...) {
  .Deprecated("card_body", old = "card_body_fill")
  card_body(...)
}

#' @rdname deprecated
#' @export
page_fill <- function(...) {
  .Deprecated("page_fillable", old = "page_fill")
  page_fillable(...)
}

#' @rdname deprecated
#' @export
sidebar_toggle <- function(id, open = NULL, session = get_current_session()) {
  lifecycle::deprecate_warn(
    "0.9.1",
    "sidebar_toggle()",
    "toggle_sidebar()"
  )
  toggle_sidebar(id = id, open = open, session = session)
}

#  Legacy Nav Containers --------------------------------------------------
# This is needed since some shiny functions like `tabPanel()` call
# `nav()` underneath the hood. In a future release of shiny, we
# could/should require bslib > 0.4.2, update it to use the new version
# (`nav_panel()`), then we won't need this fancy deprecation
deprecate_if_not_called_from_shiny <- function(old_name, new, version) {
  new_name <- deparse(substitute(new))

  function(...) {
    caller_fn_env <- environment(rlang::caller_fn())
    if (!is.null(caller_fn_env) && rlang::is_environment(caller_fn_env)) {
      caller_fn_env <- rlang::env_name(caller_fn_env)
    }
    if (!identical(caller_fn_env, "namespace:shiny")) {
      msg <- sprintf(
        "`%s()` was deprecated in {bslib} version %s, use `%s()` instead.",
        old_name,
        version,
        new_name
      )
      .Deprecated(msg = msg)
    }
    new(...)
  }
}

#' @rdname deprecated
#' @export
nav <- deprecate_if_not_called_from_shiny(
  old_name = "nav",
  new = nav_panel,
  version = "0.5.0"
)

#' @rdname deprecated
#' @export
nav_content <- deprecate_if_not_called_from_shiny(
  old_name = "nav_content",
  new = nav_panel_hidden,
  version = "0.5.0"
)

#' @rdname deprecated
#' @export
navs_tab <- deprecate_if_not_called_from_shiny(
  old_name = "navs_tab",
  new = navset_tab,
  version = "0.5.0"
)

#' @rdname deprecated
#' @export
navs_pill <- deprecate_if_not_called_from_shiny(
  old_name = "navs_pill",
  new = navset_pill,
  version = "0.5.0"
)

#' @rdname deprecated
#' @export
navs_pill_list <- deprecate_if_not_called_from_shiny(
  old_name = "navs_pill_list",
  new = navset_pill_list,
  version = "0.5.0"
)

#' @rdname deprecated
#' @export
navs_hidden <- deprecate_if_not_called_from_shiny(
  old_name = "navs_hidden",
  new = navset_hidden,
  version = "0.5.0"
)

#' @rdname deprecated
#' @export
navs_bar <- deprecate_if_not_called_from_shiny(
  old_name = "navs_bar",
  new = navset_bar,
  version = "0.5.0"
)

#' @rdname deprecated
#' @export
navs_tab_card <- deprecate_if_not_called_from_shiny(
  old_name = "navs_tab_card",
  new = navset_card_tab,
  version = "0.5.0"
)

#' @rdname deprecated
#' @export
navs_pill_card <- deprecate_if_not_called_from_shiny(
  old_name = "navs_pill_card",
  new = navset_card_pill,
  version = "0.5.0"
)

#' Create a vertically collapsing accordion
#'
#' @description
#' An accordion can be used to organize UI elements and content in a limited
#' space. It comprises multiple, vertically stacked panels that expand or
#' collapse when clicked, providing a compact layout that works well for
#' grouping input elements in a [sidebar()] or for organizing detailed
#' context-specific information.
#'
#' @param ... Named arguments become attributes on the `<div class="accordion">`
#'   element. Unnamed arguments should be `accordion_panel()`s.
#' @param id If provided, you can use `input$id` in your server logic to
#'   determine which of the `accordion_panel()`s are currently active. The value
#'   will correspond to the `accordion_panel()`'s `value` argument.
#' @param open A character vector of `accordion_panel()` `value`s to open
#'   (i.e., show) by default. The default value of `NULL` will open the first
#'   `accordion_panel()`. Use a value of `TRUE` to open all (or `FALSE` to
#'   open none) of the items. It's only possible to open more than one panel
#'   when `multiple=TRUE`.
#' @param multiple Whether multiple `accordion_panel()` can be `open` at once.
#' @param class Additional CSS classes to include on the accordion div.
#' @param width,height Any valid CSS unit; for example, height="100%".
#'
#' @export
#' @family Components
#'
#' @references bslib's accordion component is derived from the [Bootstrap
#'   Accordion
#'   component](https://getbootstrap.com/docs/5.3/components/accordion/).
#'   Accordions are also featured on the bslib website:
#'
#'   * [Get Started: Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html#accordions)
#'   * [Sidebars: Accordions](https://rstudio.github.io/bslib/articles/dashboards/index.html#accordions)
#'
#' @seealso [accordion_panel_set()], [accordion_panel_open()] and
#'   [accordion_panel_close()] programmatically interact with the state of an
#'   accordion panel.
#' @seealso [accordion_panel_insert()], [accordion_panel_remove()] and
#'   [accordion_panel_update()] add or remove accordion panels from an
#'   accordion.
#'
#' @examplesIf rlang::is_interactive()
#' items <- lapply(LETTERS, function(x) {
#'   accordion_panel(paste("Section", x), paste("Some narrative for section", x))
#' })
#'
#' # First shown by default
#' accordion(!!!items)
#' # Nothing shown by default
#' accordion(!!!items, open = FALSE)
#' # Everything shown by default
#' accordion(!!!items, open = TRUE)
#'
#' # Show particular sections
#' accordion(!!!items, open = "Section B")
#' accordion(!!!items, open = c("Section A", "Section B"))
#'
#' # Provide an id to create a shiny input binding
#' library(shiny)
#'
#' ui <- page_fluid(
#'   accordion(!!!items, id = "acc")
#' )
#'
#' server <- function(input, output) {
#'   observe(print(input$acc))
#' }
#'
#' shinyApp(ui, server)
#'
accordion <- function(
  ...,
  id = NULL,
  open = NULL,
  multiple = TRUE,
  class = NULL,
  width = NULL,
  height = NULL
) {
  args <- rlang::list2(...)
  argnames <- rlang::names2(args)

  attrs <- args[nzchar(argnames)]
  children <- args[!nzchar(argnames)]

  if (!is.null(id) && isNamespaceLoaded("shiny")) {
    open <- shiny::restoreInput(id = id, default = open)
  }

  is_open <- vapply(
    children,
    function(x) {
      isTRUE(open) || isTRUE(tagGetAttribute(x, "data-value") %in% open)
    },
    logical(1)
  )

  if (!any(is_open) && !identical(open, FALSE)) {
    is_open[1] <- TRUE
  }

  if (!multiple && sum(is_open) > 1) {
    abort("Can't select more than one panel when `multiple = FALSE`")
  }

  # Since multiple=FALSE requires an id, we always include one,
  # but only create a binding when it's provided
  if (is.null(id)) {
    id <- paste0("bslib-accordion-", p_randomInt(1000, 10000))
  } else {
    class <- c("bslib-accordion-input", class)
  }

  children <- Map(
    children,
    is_open,
    f = function(x, open) {
      if (!multiple) {
        x <- tagAppendAttributes(
          x,
          "data-bs-parent" = paste0("#", id),
          .cssSelector = ".accordion-collapse"
        )
      }

      # accordion_panel() defaults to a collapsed state
      if (open) {
        tq <- tagQuery(x)
        tq$find(".accordion-collapse")$addClass("show")
        tq$find(".accordion-button")$removeClass("collapsed")$removeAttrs(
          "aria-expanded"
        )$addAttrs("aria-expanded" = "true")
        x <- tq$allTags()
      }

      x
    }
  )

  tag <- div(
    id = id,
    class = "accordion",
    class = if (!multiple) "autoclose", # just for ease of identifying autoclosing client-side
    class = class,
    style = css(
      width = validateCssUnit(width),
      height = validateCssUnit(height)
    ),
    !!!attrs,
    !!!children,
    component_dependencies()
  )

  tag <- tag_require(tag, version = 5, caller = "accordion()")

  as_fragment(tag)
}

#' @rdname accordion
#' @param title A title to appear in the `accordion_panel()`'s header.
#' @param value A character string that uniquely identifies this panel.
#' @param icon A [htmltools::tag] child (e.g., [bsicons::bs_icon()]) which is positioned just before the `title`.
#' @export
accordion_panel <- function(title, ..., value = title, icon = NULL) {
  id <- paste0("bslib-accordion-panel-", p_randomInt(1000, 10000))

  btn <- tags$button(
    class = "accordion-button collapsed",
    type = "button",
    "data-bs-toggle" = "collapse",
    "data-bs-target" = paste0("#", id),
    "aria-expanded" = "false",
    "aria-controls" = id,
    # Always include an .accordion-icon container to simplify accordion_panel_update() logic
    div(class = "accordion-icon", icon),
    div(class = "accordion-title", title)
  )

  if (!rlang::is_string(value)) abort("`value` must be a character string")

  div(
    class = "accordion-item",
    "data-value" = value,
    # We don't use an <h2> because it would get included in rmd/pkgdown/qmd TOC
    # TODO: can we provide a way to put more stuff in the header? Like maybe some right-aligned controls?
    div(class = "accordion-header", btn),
    div(
      id = id,
      class = "accordion-collapse collapse",
      div(class = "accordion-body", ...)
    )
  )
}

#' Dynamically update accordions
#'
#' @description
#' Dynamically update/modify [`accordion()`]s in a Shiny app. To be updated
#' programmatically, the [accordion()] must have an `id`. These functions
#' require an active Shiny session and only work with a running Shiny app.
#'
#' @param id an character string that matches an existing [accordion()]'s `id`.
#' @param values either a character string (used to identify particular
#'   [accordion_panel()]s by their `value`) or `TRUE` (i.e., all `values`).
#' @param session a shiny session object (the default should almost always be
#'   used).
#'
#' @seealso [accordion()] and [accordion_panel()] create the accordion
#'   component.
#'
#' @describeIn accordion_panel_set same as `accordion_panel_open()`, except it
#'   also closes any currently open panels.
#'
#' @export
accordion_panel_set <- function(id, values, session = get_current_session()) {
  send_panel_message(
    id,
    session,
    method = "set",
    values = if (isTRUE(values)) values else as.list(check_character(values))
  )
}

#' @describeIn accordion_panel_set open [accordion_panel()]s.
#' @export
accordion_panel_open <- function(id, values, session = get_current_session()) {
  send_panel_message(
    id,
    session,
    method = "open",
    values = if (isTRUE(values)) values else as.list(check_character(values))
  )
}

#' @describeIn accordion_panel_set close [accordion_panel()]s.
#' @export
accordion_panel_close <- function(id, values, session = get_current_session()) {
  send_panel_message(
    id,
    session,
    method = "close",
    values = if (isTRUE(values)) values else as.list(check_character(values))
  )
}

#' @param panel an [accordion_panel()].
#' @param target The `value` of an existing panel to insert next to. If
#'   removing: the `value` of the [accordion_panel()] to remove.
#' @param position Should `panel` be added before or after the target? When
#'   `target` is `NULL` (the default), `"after"` will append after the last
#'   panel and `"before"` will prepend before the first panel.
#'
#' @describeIn accordion_panel_set insert a new [accordion_panel()]
#' @export
accordion_panel_insert <- function(
  id,
  panel,
  target = NULL,
  position = c("after", "before"),
  session = get_current_session()
) {
  position <- match.arg(position)
  send_panel_message(
    id,
    session,
    method = "insert",
    panel = processDeps(panel, session),
    target = if (!is.null(target)) check_character(target, max_length = 1),
    position = position
  )
}

#' @describeIn accordion_panel_set remove [accordion_panel()]s.
#' @export
accordion_panel_remove <- function(
  id,
  target,
  session = get_current_session()
) {
  send_panel_message(
    id,
    session,
    method = "remove",
    target = as.list(check_character(target))
  )
}

#' @describeIn accordion_panel_set update a [accordion_panel()].
#' @param ... Elements that become the new content of the panel.
#' @inheritParams accordion_panel
#' @export
accordion_panel_update <- function(
  id,
  target,
  ...,
  title = NULL,
  value = NULL,
  icon = NULL,
  session = get_current_session()
) {
  rlang::check_dots_unnamed()
  body <- rlang::list2(...)

  send_panel_message(
    id,
    session,
    method = "update",
    target = check_character(target, max_length = 1),
    value = if (!is.null(value)) check_character(value, max_length = 1),
    body = if (length(body) == 0) NULL else processDeps(body, session),
    title = if (!is.null(title)) processDeps(title, session),
    icon = if (!is.null(icon)) processDeps(icon, session)
  )
}

# Send message before the next flush since things like remove/insert may
# remove/create input/output values. Also do this for set/open/close since,
# you might want to open a panel after inserting it.
send_panel_message <- function(id, session, ...) {
  message <- dropNulls(rlang::list2(...))
  callback <- function() session$sendInputMessage(id, message)
  session$onFlush(callback, once = TRUE)
}

check_character <- function(
  x,
  max_length = Inf,
  min_length = 1,
  call = rlang::caller_env()
) {
  x_name <- deparse(substitute(x))
  if (!is.character(x)) {
    abort(
      sprintf("`%s` must be a character vector.", x_name),
      call = call
    )
  }
  if (length(x) < min_length) {
    abort(
      sprintf("`%s` must have length %s or more.", x_name, min_length),
      call = call
    )
  }
  if (length(x) > max_length) {
    abort(
      sprintf("`%s` must have length %s or less.", x_name, max_length),
      call = call
    )
  }
  x
}

#' Navigation containers
#'
#' Render a collection of [nav_panel()] items into a container.
#'
#' @section Examples:
#'
#' ```{r child="man/fragments/ex-navset_tab.Rmd"}
#' ```
#'
#' @param ... a collection of [nav_panel()] items.
#' @param id a character string used for dynamically updating the container (see
#'    [nav_select()]).
#' @param selected a character string matching the `value` of a particular
#'   [nav_panel()] item to selected by default.
#' @param header UI element(s) ([htmltools::tags]) to display _above_ the nav
#'   content. For `card`-based navsets, these elements are implicitly wrapped in
#'   a `card_body()`. To control things like `padding`, `fill`, etc., wrap the
#'   elements in an explicit [card_body()].
#' @param footer UI element(s) ([htmltools::tags]) to display _below_ the nav
#'   content. For `card`-based navsets, these elements are implicitly wrapped in
#'   a `card_body()`. To control things like `padding`, `fill`, etc., wrap the
#'   elements in an explicit [card_body()].
#'
#' @seealso [nav_panel()], [nav_panel_hidden()] create panels of content.
#' @seealso [nav_menu()], [nav_item()], [nav_spacer()] create menus, items, or
#'   space in the navset control area.
#' @seealso [nav_insert()], [nav_remove()] programmatically add or remove nav
#'   panels.
#' @seealso [nav_select()], [nav_show()], [nav_hide()] change the state of a
#'   [nav_panel()] in a navset.
#'
#' @family Panel container functions
#' @rdname navset
#' @name navset
#' @export
navset_tab <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL
) {
  tabs <- tabsetPanel_(
    ...,
    type = "tabs",
    id = id,
    selected = selected,
    header = header,
    footer = footer
  )
  as_fragment(tabs)
}

#' @export
#' @rdname navset
navset_pill <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL
) {
  pills <- tabsetPanel_(
    ...,
    type = "pills",
    id = id,
    selected = selected,
    header = header,
    footer = footer
  )
  as_fragment(pills)
}

#' @export
#' @rdname navset
navset_underline <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL
) {
  res <- tabsetPanel_(
    ...,
    type = "underline",
    id = id,
    selected = selected,
    header = header,
    footer = footer
  )
  as_fragment(res)
}

#' @export
#' @inheritParams shiny::navlistPanel
#' @rdname navset
navset_pill_list <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  well = TRUE,
  fluid = TRUE,
  widths = c(4, 8)
) {
  pill_list <- navlistPanel_(
    ...,
    id = id,
    selected = selected,
    header = header,
    footer = footer,
    well = well,
    fluid = fluid,
    widths = widths
  )
  as_fragment(pill_list)
}

#' @export
#' @rdname navset
navset_hidden <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL
) {
  hidden <- tabsetPanel_(
    ...,
    type = "hidden",
    id = id,
    selected = selected,
    header = header,
    footer = footer
  )
  as_fragment(hidden)
}

#' @inheritParams shiny::navbarPage
#' @inheritParams page_fillable
#' @param sidebar A [sidebar()] component to display on every [nav_panel()]
#'   page.
#' @param fillable Whether or not to allow `fill` items to grow/shrink to fit
#'   the browser window. If `TRUE`, all [nav_panel()] pages are `fillable`. A
#'   character vector, matching the `value` of [nav_panel()]s to be filled, may
#'   also be provided. Note that, if a `sidebar` is provided, `fillable` makes
#'   the main content portion fillable.
#' @param navbar_options Options to control the appearance and behavior of the
#'   navbar. Use [navbar_options()] to create the list of options.
#' @param position `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(position=)`][navbar_options] instead.
#' @param collapsible `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(collapsible=)`][navbar_options] instead.
#' @param bg `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(bg=)`][navbar_options] instead.
#' @param inverse `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(inverse=)`][navbar_options] instead.
#'
#' @export
#' @rdname navset
navset_bar <- function(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  gap = NULL,
  padding = NULL,
  header = NULL,
  footer = NULL,
  fluid = TRUE,
  navbar_options = NULL,
  position = deprecated(),
  bg = deprecated(),
  inverse = deprecated(),
  collapsible = deprecated()
) {
  padding <- validateCssPadding(padding)
  gap <- validateCssUnit(gap)

  .navbar_options <- navbar_options_resolve_deprecated(
    options_user = navbar_options,
    position = position,
    bg = bg,
    inverse = inverse,
    collapsible = collapsible
  )

  navs_bar_(
    ...,
    title = title,
    id = id,
    selected = selected,
    sidebar = sidebar,
    fillable = fillable,
    gap = gap,
    padding = padding,
    header = header,
    footer = footer,
    fluid = fluid,
    navbar_options = .navbar_options,
    # theme is only used to determine whether legacy style markup should be used
    # (and, at least at the moment, we don't need legacy markup for this exported function)
    theme = bs_theme()
  )
}

# This internal version of navs_bar() exists so both it and page_navbar()
# (and thus shiny::navbarPage()) can use it. And in the page_navbar() case,
# we can use addition theme information as an indication of whether we need
# to handle backwards compatibility
navs_bar_ <- function(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  gap = NULL,
  padding = NULL,
  navbar_options = NULL,
  header = NULL,
  footer = NULL,
  fluid = TRUE,
  theme = NULL
) {
  navbar_options <- navbar_options %||% navbar_options()

  navbar_theme <- navbar_options[["theme"]]
  navbar_bg <- navbar_options[["bg"]]

  if (identical(navbar_theme, "auto")) {
    if (is.null(theme) || theme_version(theme) < 5) {
      navbar_theme <- "dark"
    }
    if (!is.null(navbar_bg)) {
      navbar_bg <- htmltools::parseCssColors(navbar_bg)
      bg_contrast <- get_color_contrast(navbar_bg)
      navbar_theme <- if (bg_contrast == "#FFFFFF") "dark" else "light"
    }
  }

  navbar <- navbarPage_(
    title = title,
    ...,
    id = id,
    selected = selected,
    sidebar = sidebar,
    fillable = fillable,
    gap = gap,
    padding = padding,
    position = navbar_options[["position"]],
    header = header,
    footer = footer,
    collapsible = navbar_options[["collapsible"]],
    inverse = identical(navbar_theme, "dark"),
    underline = navbar_options[["underline"]],
    fluid = fluid,
    theme = theme
  )

  attribs <- navbar_options[["attribs"]] %||% list()

  # Use user-provided `data-bs-theme` or our internally selected `theme`. If the
  # user includes both, the attribute wins for being most technically-correct.
  attribs[["data-bs-theme"]] <- attribs[["data-bs-theme"]] %||% navbar_theme

  # navbarPage_() returns a tagList() of the nav and content
  navbar[[1]] <- tagAppendAttributes(
    navbar[[1]],
    style = if (!is.null(navbar_bg))
      css(background_color = paste(navbar_bg, "!important")),
    !!!attribs
  )

  as_fragment(navbar, page = page)
}

# -----------------------------------------------------------------------
# 'Internal' tabset logic that was pulled directly from shiny/R/bootstrap.R
#  (with minor modifications)
# -----------------------------------------------------------------------

navbarPage_ <- function(
  title,
  ...,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  gap = NULL,
  padding = NULL,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  header = NULL,
  footer = NULL,
  inverse = FALSE,
  underline = TRUE,
  collapsible = FALSE,
  fluid = TRUE,
  theme = NULL
) {
  # alias title so we can avoid conflicts w/ title in withTags
  pageTitle <- title

  # navbar class based on options
  navbarClass <- "navbar navbar-default"
  position <- match.arg(position)
  if (!is.null(position))
    navbarClass <- paste0(navbarClass, " navbar-", position)
  if (isTRUE(inverse)) navbarClass <- paste(navbarClass, "navbar-inverse")

  if (!is.null(id)) selected <- shiny::restoreInput(id = id, default = selected)

  # build the tabset
  ulClass <- "nav navbar-nav"
  if (underline) ulClass <- paste(ulClass, "nav-underline")
  tabset <- buildTabset(..., ulClass = ulClass, id = id, selected = selected)

  containerClass <- paste0("container", if (fluid) "-fluid")

  # built the container div dynamically to support optional collapsibility
  if (collapsible) {
    navId <- paste0("navbar-collapse-", p_randomInt(1000, 10000))
    containerDiv <- div(
      class = containerClass,
      div(
        class = "navbar-header",
        tags$button(
          type = "button",
          class = "navbar-toggle collapsed",
          `data-toggle` = "collapse",
          `data-target` = paste0("#", navId),
          # data-bs-* is for BS5+
          `data-bs-toggle` = "collapse",
          `data-bs-target` = paste0("#", navId),
          span(class = "sr-only visually-hidden", "Toggle navigation"),
          span(class = "icon-bar"),
          span(class = "icon-bar"),
          span(class = "icon-bar")
        ),
        span(class = "navbar-brand", pageTitle)
      ),
      div(
        class = "navbar-collapse collapse",
        id = navId,
        tabset$navList
      )
    )
  } else {
    containerDiv <- div(
      class = containerClass,
      div(
        class = "navbar-header",
        span(class = "navbar-brand", pageTitle)
      ),
      tabset$navList
    )
  }

  # Bootstrap 3 explicitly supported "dropup menus" via .navbar-fixed-bottom,
  # but BS4+ requires .dropup on menus with .navbar.fixed-bottom
  if (position == "fixed-bottom") {
    containerDiv <- tagQuery(containerDiv)$find(
      ".dropdown-menu"
    )$parent()$addClass("dropup")$allTags()
  }

  # If fillable is truthy, give the relevant .tab-content > .tab-pane containers
  # the potential to fill
  tabset$content <- makeTabsFillable(
    tabset$content,
    fillable,
    navbar = TRUE,
    gap = gap,
    padding = padding
  )

  # For backwards compatibility reasons, wrap header & footer in a .row
  # container if we're not using BS5+. I'm not entirely sure what the motivation
  # was for it in the 1st place, but now that, with BS5, .row adds
  # `display:flex` and makes children `width:100%`, which is surprising and
  # confusing from a user perspective
  isLegacy <- as.numeric(theme_version(theme) %||% 3) < 5
  if (!is.null(header) && isLegacy) {
    header <- div(class = "row", header)
  }
  if (!is.null(footer) && isLegacy) {
    footer <- div(class = "row", footer)
  }

  contents <- dropNulls(list(header, tabset$content, footer))

  if (is.null(sidebar)) {
    contentDiv <- div(class = containerClass, !!!contents)

    # If fillable is truthy, the .container also needs to be fillable
    if (!isFALSE(fillable)) {
      contentDiv <- bindFillRole(contentDiv, container = TRUE, item = TRUE)
    }
  } else {
    contentDiv <- div(
      # In the fluid case, the sidebar layout should be flush (i.e.,
      # the .container-fluid class adds padding that we don't want)
      class = if (!fluid) "container",
      layout_sidebar(
        sidebar = sidebar,
        fillable = !isFALSE(fillable),
        border_radius = FALSE,
        border = !fluid,
        page_main_container(contents)
      )
    )

    # Always have the sidebar layout fill its parent (in this case
    # fillable controls whether the _main_ content portion is fillable)
    contentDiv <- bindFillRole(contentDiv, container = TRUE, item = TRUE)
  }

  # *Don't* wrap in bootstrapPage() (shiny::navbarPage()) does that part
  tagList(
    tags$nav(class = navbarClass, role = "navigation", containerDiv),
    contentDiv
  )
}

navbarMenu_ <- function(title, ..., menuName = title, icon = NULL, align) {
  icon <- prepTabIcon(icon)
  structure(
    list(
      title = title,
      menuName = menuName,
      tabs = list2(...),
      # Here for legacy reasons
      # https://github.com/cran/miniUI/blob/74c87d3/R/layout.R#L369
      iconClass = if (inherits(icon, "shiny.tag"))
        tagGetAttribute(icon, "class"),
      icon = icon,
      align = align
    ),
    class = "shiny.navbarmenu"
  )
}

isNavbarMenu <- function(x) {
  inherits(x, "shiny.navbarmenu")
}

tabPanel_ <- function(title, ..., value = title, icon = NULL) {
  icon <- prepTabIcon(icon)
  pane <- div(
    class = "tab-pane",
    title = title,
    `data-value` = value,
    # Here for legacy reasons
    # https://github.com/cran/miniUI/blob/74c87d/R/layout.R#L395
    `data-icon-class` = if (inherits(icon, "shiny.tag"))
      tagGetAttribute(icon, "class"),
    ...
  )
  attr(pane, "_shiny_icon") <- icon
  pane
}

isTabPanel <- function(x) {
  if (!inherits(x, "shiny.tag")) return(FALSE)
  class <- tagGetAttribute(x, "class") %||% ""
  "tab-pane" %in% strsplit(class, "\\s+")[[1]]
}

tabPanelBody_ <- function(value, ..., icon = NULL) {
  if (
    !is.character(value) ||
      length(value) != 1 ||
      any(is.na(value)) ||
      nchar(value) == 0
  ) {
    stop("`value` must be a single, non-empty string value")
  }
  tabPanel_(title = NULL, ..., value = value, icon = icon)
}

tabsetPanel_ <- function(
  ...,
  id = NULL,
  selected = NULL,
  type = c("tabs", "pills", "hidden", "underline"),
  header = NULL,
  footer = NULL
) {
  if (!is.null(id)) selected <- shiny::restoreInput(id = id, default = selected)

  type <- match.arg(type)
  tabset <- buildTabset(
    ...,
    ulClass = paste0("nav nav-", type),
    id = id,
    selected = selected
  )

  tags$div(
    class = "tabbable",
    !!!dropNulls(
      list(
        tabset$navList,
        header,
        tabset$content,
        footer
      )
    )
  )
}

navlistPanel_ <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  well = TRUE,
  fluid = TRUE,
  widths = c(4, 8)
) {
  if (!is.null(id)) selected <- shiny::restoreInput(id = id, default = selected)

  tabset <- buildTabset(
    ...,
    ulClass = "nav nav-pills nav-stacked",
    textFilter = function(text) tags$li(class = "navbar-brand", text),
    id = id,
    selected = selected
  )

  row <- if (fluid) shiny::fluidRow else shiny::fixedRow

  row(
    shiny::column(widths[[1]], class = if (well) "well", tabset$navList),
    shiny::column(
      widths[[2]],
      !!!dropNulls(list(header, tabset$content, footer))
    )
  )
}

# Helpers to build tabsetPanels (& Co.) and their elements
markTabAsSelected <- function(x) {
  attr(x, "selected") <- TRUE
  x
}

isTabSelected <- function(x) {
  isTRUE(attr(x, "selected", exact = TRUE))
}

containsSelectedTab <- function(tabs) {
  any(vapply(tabs, isTabSelected, logical(1)))
}

findAndMarkSelectedTab <- function(tabs, selected, foundSelected) {
  tabs <- lapply(tabs, function(x) {
    if (foundSelected || is.character(x)) {
      # Strings are not selectable items
    } else if (isNavbarMenu(x)) {
      # Recur for navbarMenus
      res <- findAndMarkSelectedTab(x$tabs, selected, foundSelected)
      x$tabs <- res$tabs
      foundSelected <<- res$foundSelected
    } else if (isTabPanel(x)) {
      # Base case: regular tab item. If the `selected` argument is
      # provided, check for a match in the existing tabs; else,
      # mark first available item as selected
      if (is.null(selected)) {
        foundSelected <<- TRUE
        x <- markTabAsSelected(x)
      } else {
        tabValue <- x$attribs$`data-value` %||% x$attribs$title
        if (identical(selected, tabValue)) {
          foundSelected <<- TRUE
          x <- markTabAsSelected(x)
        }
      }
    }
    return(x)
  })
  return(list(tabs = tabs, foundSelected = foundSelected))
}

prepTabIcon <- function(x = NULL) {
  if (!inherits(x, "shiny.tag")) return(x)

  is_fa <- grepl("fa-", tagGetAttribute(x, "class") %||% "", fixed = TRUE)
  if (!is_fa) return(x)

  # specify fixed-width for font-awesome
  tagAppendAttributes(x, class = "fa-fw")
}

# Text filter for navbarMenu's (plain text) separators
navbarMenuTextFilter <- function(text) {
  if (grepl("^\\-+$", text)) tags$li(class = "divider") else
    tags$li(class = "dropdown-header", text)
}

# This function is called internally by navbarPage, tabsetPanel
# and navlistPanel
buildTabset <- function(
  ...,
  ulClass,
  textFilter = NULL,
  id = NULL,
  selected = NULL,
  foundSelected = FALSE
) {
  tabs <- dropNulls(list2(...))
  res <- findAndMarkSelectedTab(tabs, selected, foundSelected)
  tabs <- res$tabs
  foundSelected <- res$foundSelected

  # add input class if we have an id
  if (!is.null(id)) ulClass <- paste(ulClass, "shiny-tab-input")

  if (anyNamed(tabs)) {
    nms <- names(tabs)
    nms <- nms[nzchar(nms)]
    stop(
      "Tabs should all be unnamed arguments, but some are named: ",
      paste(nms, collapse = ", ")
    )
  }

  tabsetId <- p_randomInt(1000, 10000)
  tabs <- lapply(
    seq_len(length(tabs)),
    buildTabItem,
    tabsetId = tabsetId,
    foundSelected = foundSelected,
    tabs = tabs,
    textFilter = textFilter
  )

  tabNavList <- tags$ul(
    class = ulClass,
    id = id,
    `data-tabsetid` = tabsetId,
    !!!lapply(tabs, "[[", "liTag")
  )

  tabContent <- tags$div(
    class = "tab-content",
    `data-tabsetid` = tabsetId,
    !!!lapply(tabs, "[[", "divTag")
  )

  list(navList = tabNavList, content = tabContent)
}

# Builds tabPanel/navbarMenu items (this function used to be
# declared inside the buildTabset() function and it's been
# refactored for clarity and reusability). Called internally
# by buildTabset.
buildTabItem <- function(
  index,
  tabsetId,
  foundSelected,
  tabs = NULL,
  divTag = NULL,
  textFilter = NULL
) {
  divTag <- divTag %||% tabs[[index]]

  # Handles navlistPanel() headers and dropdown dividers
  if (is.character(divTag) && !is.null(textFilter)) {
    return(list(liTag = textFilter(divTag), divTag = NULL))
  }

  if (isNavbarMenu(divTag)) {
    # tabPanelMenu item: build the child tabset
    ulClass <- "dropdown-menu"
    if (identical(divTag$align, "right")) {
      ulClass <- paste(ulClass, "dropdown-menu-right dropdown-menu-end")
    }
    tabset <- buildTabset(
      !!!divTag$tabs,
      ulClass = ulClass,
      textFilter = navbarMenuTextFilter,
      foundSelected = foundSelected
    )
    return(buildDropdown(divTag, tabset))
  }

  if (isTabPanel(divTag)) {
    return(buildNavItem(divTag, tabsetId, index))
  }

  if (is_nav_item(divTag) || is_nav_spacer(divTag)) {
    return(
      list(liTag = divTag, divTag = NULL)
    )
  }

  # The behavior is undefined at this point, so construct a condition message
  msg <- paste0(
    "Navigation containers expect a collection of `bslib::nav_panel()`/`shiny::tabPanel()`s ",
    "and/or `bslib::nav_menu()`/`shiny::navbarMenu()`s. ",
    "Consider using `header` or `footer` if you wish to place content above ",
    "(or below) every panel's contents."
  )

  # Luckily this case has never worked, so it's safe to throw here
  # https://github.com/rstudio/shiny/issues/3313
  if (!inherits(divTag, "shiny.tag")) {
    stop(msg, call. = FALSE)
  }

  # Unfortunately, this 'off-label' use case creates an 'empty' nav and includes
  # the divTag content on every tab. There shouldn't be any reason to be relying on
  # this behavior since we now have pre/post arguments, so throw a warning, but still
  # support the use case since we don't make breaking changes
  warning(msg, call. = FALSE)

  return(buildNavItem(divTag, tabsetId, index))
}

buildNavItem <- function(divTag, tabsetId, index) {
  id <- paste("tab", tabsetId, index, sep = "-")
  # Get title attribute directory (not via tagGetAttribute()) so that contents
  # don't get passed to as.character().
  # https://github.com/rstudio/shiny/issues/3352
  title <- divTag$attribs[["title"]]
  value <- divTag$attribs[["data-value"]]
  active <- isTabSelected(divTag)
  divTag <- tagAppendAttributes(divTag, class = if (active) "active")
  divTag$attribs$id <- id
  divTag$attribs$title <- NULL
  list(
    divTag = divTag,
    liTag = tagAddRenderHook(
      liTag(id, title, value, attr(divTag, "_shiny_icon")),
      function(x) {
        if (isTRUE(getCurrentThemeVersion() >= 4)) {
          tagQuery(x)$addClass("nav-item")$find("a")$addClass(
            c("nav-link", if (active) "active")
          )$allTags()
        } else {
          tagAppendAttributes(x, class = if (active) "active")
        }
      }
    )
  )
}

liTag <- function(id, title, value, icon) {
  tags$li(
    tags$a(
      href = paste0("#", id),
      `data-toggle` = "tab",
      # data-bs-* is for BS5+
      `data-bs-toggle` = "tab",
      `data-value` = value,
      icon,
      title
    )
  )
}

buildDropdown <- function(divTag, tabset) {
  navList <- tagAddRenderHook(
    tabset$navList,
    function(x) {
      if (isTRUE(getCurrentThemeVersion() >= 4)) {
        tagQuery(x)$find(".nav-item")$removeClass("nav-item")$find(
          ".nav-link"
        )$removeClass("nav-link")$addClass("dropdown-item")$allTags()
      } else {
        x
      }
    }
  )

  active <- containsSelectedTab(divTag$tabs)

  dropdown <- tags$li(
    class = "dropdown",
    tags$a(
      href = "#",
      class = "dropdown-toggle",
      `data-toggle` = "dropdown",
      # data-bs-* is for BS5+
      `data-bs-toggle` = "dropdown",
      `data-value` = divTag$menuName,
      divTag$icon,
      divTag$title,
      tags$b(class = "caret")
    ),
    navList,
    .renderHook = function(x) {
      if (isTRUE(getCurrentThemeVersion() >= 4)) {
        tagQuery(x)$addClass("nav-item")$find(".dropdown-toggle")$addClass(
          "nav-link"
        )$addClass(if (active) "active")$allTags()
      } else {
        tagAppendAttributes(x, class = if (active) "active")
      }
    }
  )

  list(
    divTag = tabset$content$children,
    liTag = dropdown
  )
}

test_that("navset_bar() warns if using deprecated args", {
  lifecycle::expect_deprecated(
    navset_bar(position = "fixed-top")
  )
  lifecycle::expect_deprecated(
    navset_bar(bg = "red")
  )
  lifecycle::expect_deprecated(
    navset_bar(inverse = TRUE)
  )
  lifecycle::expect_deprecated(
    navset_bar(collapsible = FALSE)
  )
})

test_that("navset_bar() warns if `navbar_options()` collide with direct deprecated options", {
  rlang::local_options(lifecycle_verbosity = "quiet")

  expect_warning(
    navset_bar(
      position = "fixed-top",
      navbar_options = navbar_options(position = "static-top")
    )
  )

  expect_warning(
    navset_bar(
      bg = "red",
      navbar_options = navbar_options(bg = "blue")
    )
  )

  expect_warning(
    navset_bar(
      inverse = TRUE,
      navbar_options = navbar_options(theme = "light")
    )
  )

  expect_warning(
    navset_bar(
      collapsible = FALSE,
      navbar_options = navbar_options(collapsible = TRUE)
    )
  )
})

test_that("shiny:navbarPage() is unaffected", {
  rlang::local_options(lifecycle_verbosity = "warning")

  expect_silent(
    shiny::navbarPage(title = "test")
  )

  expect_silent(
    shiny::navbarPage(
      title = "test",
      bg = "red",
      collapsible = TRUE,
      inverse = TRUE,
      position = "fixed-top"
    )
  )
})

show_navbar_markup <- function(navbar) {
  nb <- navbar[[1]]
  nb$children <- NULL
  cat(format(nb))
}

test_that("navbar markup snapshots", {
  expect_snapshot(
    show_navbar_markup(navs_bar_(theme = bs_theme(version = 3)))
  )

  expect_snapshot(
    show_navbar_markup(navs_bar_(theme = bs_theme(version = 4)))
  )

  expect_snapshot(
    show_navbar_markup(navs_bar_(theme = bs_theme(version = 5)))
  )

  expect_snapshot(
    show_navbar_markup(
      navs_bar_(
        theme = bs_theme(version = 4),
        navbar_options = navbar_options(theme = "dark")
      )
    )
  )

  expect_snapshot(
    show_navbar_markup(
      navs_bar_(
        theme = bs_theme(version = 4),
        navbar_options = navbar_options(theme = "light")
      )
    )
  )

  expect_snapshot(
    show_navbar_markup(
      navs_bar_(
        theme = bs_theme(version = 4),
        navbar_options = navbar_options(bg = "#000")
      )
    )
  )

  expect_snapshot(
    show_navbar_markup(
      navs_bar_(
        theme = bs_theme(version = 5),
        navbar_options = navbar_options(theme = "dark")
      )
    )
  )

  expect_snapshot(
    show_navbar_markup(
      navs_bar_(
        theme = bs_theme(version = 5),
        navbar_options = navbar_options(theme = "light")
      )
    )
  )

  expect_snapshot(
    show_navbar_markup(
      navs_bar_(
        theme = bs_theme(version = 5),
        navbar_options = navbar_options(bg = "#000")
      )
    )
  )

  expect_snapshot(
    show_navbar_markup(
      navs_bar_(
        theme = bs_theme(version = 5),
        navbar_options = navbar_options(
          theme = "light",
          `data-bs-theme` = "dark"
        )
      )
    )
  )

  expect_snapshot(
    show_navbar_markup(
      navs_bar_(
        theme = bs_theme(version = 5),
        navbar_options = navbar_options(class = "bg-primary", theme = "dark")
      )
    )
  )
})

.onLoad <- function(libname, pkgname) {
  register_upgrade_message("shiny", "1.6.0")
  register_upgrade_message("rmarkdown", "2.7")
  .dependency_cache <<- cachem::cache_mem(max_age = 1)
}

.dependency_cache <- NULL

.onLoad <- function(libname, pkgname) {
  register_upgrade_message("shiny", "1.6.0")
  register_upgrade_message("rmarkdown", "2.7")
}

# Essentially verbatim from shiny:::register_upgrade_message
register_upgrade_message <- function(pkg, version) {

  msg <- sprintf(
    "This version of bslib is designed to work with %s version %s or higher. ",
    pkg, version
  )

  if (isNamespaceLoaded(pkg) && !is_available(pkg, version)) {
    packageStartupMessage(msg)
  }

  # Always register hook in case pkg is loaded at some
  # point the future (or, potentially, but less commonly,
  # unloaded & reloaded)
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      if (!is_available(pkg, version)) packageStartupMessage(msg)
    }
  )
}

#' Create a set of navbar options
#'
#' A `navbar_options()` object captures options specific to the appearance and
#' behavior of the navbar, independent from the content displayed on the page.
#' This helper should be used to create the list of options expected by
#' `navbar_options` in [page_navbar()] and [navset_bar()].
#'
#' ## Navbar style with Bootstrap 5 and Bootswatch themes
#'
#' In \pkg{bslib} v0.9.0, the default navbar colors for Bootswatch themes with
#' Bootstrap 5 changed. Prior to v0.9.0, bslib pre-selected navbar background
#' colors in light and dark mode; after v0.9.0 the default navbar colors are
#' less opinionated by default and follow light or dark mode (see
#' [input_dark_mode()]).
#'
#' You can use `navbar_options()` to adjust the colors of the navbar when using
#' a Bootswatch preset theme with Bootstrap 5. For example, the [Bootswatch
#' documentation for the Flatly theme](https://bootswatch.com/flatly/) shows
#' 4 navbar variations. Inspecting the source code for the first example reveals
#' the following markup:
#'
#' ```html
#' <nav class="navbar navbar-expand-lg bg-primary" data-bs-theme="dark">
#'   <!-- all of the navbar html -->
#' </nav>
#' ```
#'
#' Note that this navbar uses the `bg-primary` class for a dark navy background.
#' The navbar's white text is controlled by the `data-bs-theme="dark"`
#' attribute, which is used by Bootstrap for light text on a _dark_ background.
#' In \pkg{bslib}, you can achieve this look with:
#'
#' ```r
#' ui <- page_navbar(
#'   theme = bs_theme(5, "flatly"),
#'   navbar_options = navbar_options(class = "bg-primary", theme = "dark")
#' )
#' ```
#'
#' This particular combination of `class = "bg-primary"` and `theme = "dark"`
#' works well for most Bootswatch presets.
#'
#' Another variation from the Flatly documentation features a navar with dark
#' text on a light background:
#'
#' ```r
#' ui <- page_navbar(
#'   theme = bs_theme(5, "flatly"),
#'   navbar_options = navbar_options(class = "bg-light", theme = "light")
#' )
#' ```
#'
#' The above options set navbar foreground and background colors that are always
#' the same in both light and dark modes. To customize the navbar colors used in
#' light or dark mode, you can use the `$navbar-light-bg` and `$navbar-dark-bg`
#' Sass variables. When provided, bslib will automatically choose to use
#' light or dark text as the foreground color.
#'
#' ```r
#' ui <- page_navbar(
#'   theme = bs_theme(
#'     5,
#'     preset = "flatly",
#'     navbar_light_bg = "#18BC9C", # flatly's success color (teal)
#'     navbar_dark_bg = "#2C3E50"   # flatly's primary color (navy)
#'   )
#' )
#' ```
#'
#' Finally, you can also use the `$navbar-bg` Sass variable to set the navbar
#' background color for both light and dark modes:
#'
#' ```r
#' ui <- page_navbar(
#'   theme = bs_theme(
#'     5,
#'     preset = "flatly",
#'     navbar_bg = "#E74C3C" # flatly's danger color (red)
#'   )
#' )
#' ```
#'
#' @section Changelog:
#'
#' This function was introduced in \pkg{bslib} v0.9.0, replacing the `position`,
#' `bg`, `inverse`, `collapsible` and `underline` arguments of [page_navbar()]
#' and [navset_bar()]. Those arguments are deprecated with a warning and will be
#' removed in a future version of \pkg{bslib}. Note that the deprecated
#' `inverse` argument of [page_navbar()] and [navset_bar()] was replaced with
#' the `theme` argument of `navbar_options()`.
#'
#' @examples
#' navbar_options(position = "static-top", bg = "#2e9f7d", underline = FALSE)
#'
#' @inheritParams shiny::navbarPage
#' @param bg a CSS color to use for the navbar's background color.
#' @param theme Either `"dark"` for a light text color (on a **dark**
#'   background) or `"light"` for a dark text color (on a **light** background).
#'   If `"auto"` (the default) and `bg` is provided, the best contrast to `bg`
#'   is chosen.
#' @param underline Whether or not to add underline styling to page or navbar
#'   links when active or focused.
#' @param ... Additional attributes that will be passed directly to the navbar
#'   container element.
#'
#' @returns Returns a list of navbar options.
#'
#' @export
navbar_options <- function(
  ...,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  bg = NULL,
  theme = c("auto", "light", "dark"),
  collapsible = TRUE,
  underline = TRUE
) {
  # Track user-provided arguments for print method and deprecation warnings
  is_default <- list(
    position = missing(position),
    bg = missing(bg),
    theme = missing(theme),
    collapsible = missing(collapsible),
    underline = missing(underline)
  )

  opts <- list(
    position = rlang::arg_match(position),
    bg = bg,
    theme = rlang::arg_match(theme),
    collapsible = collapsible,
    underline = underline
  )

  dots <- separate_arguments(...)
  if (length(dots$children) > 0) {
    abort(
      "All arguments in `...` must be named attributes to be applied to the navbar container."
    )
  }

  if ("inverse" %in% names(dots$attribs)) {
    # Catch muscle-memory for using `inverse`. We didn't officially release
    # `navbar_options()` with an `inverse` argument, but it's reasonable people
    # might try to use it and it did exist briefly in dev versions.
    lifecycle::deprecate_soft(
      when = "0.9.0",
      what = "navbar_options(inverse=)",
      with = "navbar_options(theme=)"
    )
  }
  if (length(dots$attribs)) {
    opts$attribs <- dots$attribs
  }

  structure(
    opts,
    class = c("bslib_navbar_options", "list"),
    is_default = is_default,
    waldo_opts = list(ignore_attr = TRUE)
  )
}

navbar_options_resolve_deprecated <- function(
  options_user = list(),
  position = deprecated(),
  bg = deprecated(),
  inverse = deprecated(),
  collapsible = deprecated(),
  underline = deprecated(),
  .fn_caller = "navset_bar",
  .warn_deprecated = TRUE
) {
  options_old <- list(
    position = if (lifecycle::is_present(position)) position,
    bg = if (lifecycle::is_present(bg)) bg,
    inverse = if (lifecycle::is_present(inverse)) inverse,
    collapsible = if (lifecycle::is_present(collapsible)) collapsible,
    underline = if (lifecycle::is_present(underline)) underline
  )
  options_old <- dropNulls(options_old)

  args_deprecated <- names(options_old)

  if (.warn_deprecated && length(args_deprecated)) {
    # TODO-deprecated: (2024-12) Elevate deprecation to an error
    lifecycle::deprecate_warn(
      "0.9.0",
      I(
        sprintf(
          "The arguments of `%s()` for navbar options (including %s) have been consolidated into a single `navbar_options` argument and ",
          .fn_caller,
          paste(sprintf("`%s`", args_deprecated), collapse = ", ")
        )
      ),
      details = c(
        "i" = "See `navbar_options()` for more details.",
        "!" = if ("inverse" %in% args_deprecated)
          "Use `theme` instead of `inverse` in `navbar_options()`."
      )
    )
  }

  # Upgrade `inverse` to the new `theme` argument of `navbar_options()`
  if ("inverse" %in% names(options_old)) {
    inverse <- options_old[["inverse"]]
    options_old[["inverse"]] <- NULL

    options_old[["theme"]] <-
      if (is.character(inverse)) {
        inverse
      } else if (isTRUE(as.logical(inverse))) {
        options_old[["theme"]] <- "dark"
      } else if (isFALSE(as.logical(inverse))) {
        options_old[["theme"]] <- "light"
      } else {
        abort(paste("Invalid `inverse` value: ", inverse))
      }
  }

  # Consolidate `navbar_options` (options_user) with the deprecated direct
  # options. We take the direct option if the user option is a default value,
  # warning if otherwise ignored.
  # TODO-deprecated: Remove this and warning when direct options are hard-deprecated
  is_default <- attr(options_user, "is_default") %||% list()
  keep_user_values <- vapply(
    names(options_user),
    function(x) !isTRUE(is_default[[x]]),
    logical(1)
  )
  options_user <- options_user[keep_user_values]

  ignored <- c()
  for (opt in names(options_old)) {
    if (!opt %in% names(options_user)) {
      options_user[[opt]] <- options_old[[opt]]
    } else if (!identical(options_old[[opt]], options_user[[opt]])) {
      ignored <- c(ignored, if (opt == "theme") "inverse" else opt)
    }
  }

  if (length(ignored) > 0) {
    rlang::warn(
      c(
        sprintf(
          "`%s` %s provided twice: once directly and once in `navbar_options`.",
          paste(ignored, collapse = "`, `"),
          if (length(ignored) == 1) "was" else "were"
        ),
        "The deprecated direct option(s) will be ignored and the values from `navbar_options` will be used."
      ),
      call = rlang::caller_call()
    )
  }

  attribs <- options_user[["attribs"]] %||% list()
  options_user$attribs <- NULL

  rlang::exec(navbar_options, !!!options_user, !!!attribs)
}

#' @export
print.bslib_navbar_options <- function(x, ...) {
  cat("<bslib_navbar_options>\n")

  if (length(x) == 0) {
    return(invisible(x))
  }

  fields <- names(x)
  opt_w <- max(nchar(fields))
  is_default <- attr(x, "is_default") %||% list()
  for (opt in fields) {
    value <- x[[opt]] %||% "NULL"
    if (inherits(value, "list")) {
      value <- paste(names(value), collapse = ", ")
    }
    if (isTRUE(is_default[[opt]])) {
      if (identical(value, "NULL")) {
        # Skip printing default NULL values
        next
      }
      value <- sprintf("(%s)", value)
    }
    cat(sprintf("%*s", opt_w, opt), ": ", value, "\n", sep = "")
  }

  invisible(x)
}

test_that("navbar_options() validates position", {
  expect_equal(
    navbar_options(position = "fixed-bottom")$position,
    "fixed-bottom"
  )

  expect_error(navbar_options(position = "bad"))
})

test_that("navbar_options() print method", {
  expect_snapshot(navbar_options())
  expect_snapshot(navbar_options(theme = "dark", bg = "red"))
  expect_snapshot(
    navbar_options(position = "static-top", theme = "auto", collapsible = TRUE)
  )

  expect_output(
    print(navbar_options()),
    "<bslib_navbar_options>"
  )
})

test_that("navbar_options() adds named args from ... to `attribs`", {
  expect_equal(navbar_options(foo = "bar")$attribs, list(foo = "bar"))
})

test_that("navbar_options() throws for unnamed args in ...", {
  expect_error(navbar_options("foo", "bar"))
})

test_that("navbar_options() warns `inverse` is used instead of `type`", {
  lifecycle::expect_deprecated(
    navbar_options(inverse = TRUE)
  )
})

test_that("navbar_options_resolve_deprecated() consolidates correctly", {
  # TODO-deprecated: Remove when direction options are deprecated with an error

  # deprecation messages are handled through other tests
  rlang::local_options(lifecycle_verbosity = "quiet")

  expect_equal(
    navbar_options_resolve_deprecated(navbar_options(), bg = "red")$bg,
    "red"
  )

  expect_equal(
    navbar_options_resolve_deprecated(list(), bg = "red")$bg,
    "red"
  )

  expect_warning(
    expect_equal(
      navbar_options_resolve_deprecated(
        navbar_options(bg = "blue"),
        bg = "red"
      )$bg,
      "blue"
    )
  )

  expect_warning(
    expect_equal(
      navbar_options_resolve_deprecated(list(bg = "blue"), bg = "red")$bg,
      "blue"
    )
  )

  expect_warning(
    expect_null(
      navbar_options_resolve_deprecated(
        navbar_options(bg = NULL),
        bg = "red"
      )$bg
    )
  )

  expect_warning(
    expect_null(
      navbar_options_resolve_deprecated(list(bg = NULL), bg = "red")$bg
    )
  )

  expect_equal(
    attr(navbar_options(underline = FALSE), "is_default"),
    attr(navbar_options_resolve_deprecated(underline = FALSE), "is_default")
  )
})

test_that("navbar_options_resolve_deprecated() upgrades `inverse` to `type`", {
  # TODO-deprecated: Remove when direction options are deprecated with an error

  # deprecation messages are handled through other tests
  rlang::local_options(lifecycle_verbosity = "quiet")

  expect_equal(
    navbar_options_resolve_deprecated(navbar_options(), inverse = TRUE)$theme,
    "dark"
  )

  expect_equal(
    navbar_options_resolve_deprecated(navbar_options(), inverse = FALSE)$theme,
    "light"
  )

  expect_equal(
    navbar_options_resolve_deprecated(navbar_options(), inverse = "auto")$theme,
    "auto"
  )

  expect_warning(
    expect_equal(
      navbar_options_resolve_deprecated(
        navbar_options(theme = "light"),
        inverse = TRUE
      )$theme,
      "light"
    )
  )

  expect_warning(
    expect_equal(
      navbar_options_resolve_deprecated(
        navbar_options(theme = "dark"),
        inverse = FALSE
      )$theme,
      "dark"
    )
  )
})

test_that("navbar_options_resolve_deprecated() prefers user options over deprecated direct options", {
  rlang::local_options(lifecycle_verbosity = "quiet")

  expect_warning(
    expect_equal(
      navbar_options_resolve_deprecated(
        position = "fixed-top",
        options_user = navbar_options(position = "static-top")
      )$position,
      "static-top"
    )
  )

  expect_warning(
    expect_equal(
      navbar_options_resolve_deprecated(
        bg = "red",
        options_user = navbar_options(bg = "blue")
      )$bg,
      "blue"
    )
  )

  expect_warning(
    expect_equal(
      navbar_options_resolve_deprecated(
        inverse = TRUE,
        options_user = navbar_options(theme = "light")
      )$theme,
      "light"
    )
  )

  expect_warning(
    expect_equal(
      navbar_options_resolve_deprecated(
        collapsible = FALSE,
        options_user = navbar_options(collapsible = TRUE)
      )$collapsible,
      TRUE
    )
  )
})

#' Code editor input
#'
#' @description
#' Creates an interactive light-weight code editor input that can be used in
#' Shiny applications. The editor provides syntax highlighting, line numbers,
#' and other basic code editing features powered by Prism Code Editor. For a
#' complete example, run `shiny::runExample("code-editor", package = "bslib")`.
#'
#' The editor value is not sent to R on every keystroke. Instead, updates are
#' reflected on the server when the user moves away from the editor or when they
#' press `Ctrl/Cmd` + `Enter`.
#'
#' Note that this input is not designed for editing or rendering large files.
#' Displaying files with 1,000 lines or more may lead to performance issues.
#'
#' @section Keyboard shortcuts:
#' The editor supports the following keyboard shortcuts:
#' - `Ctrl/Cmd+Enter`: Submit the current code to R
#' - `Ctrl/Cmd+Z`: Undo
#' - `Ctrl/Cmd+Shift+Z`: Redo
#' - `Tab`: Indent selection
#' - `Shift+Tab`: Dedent selection
#'
#' @section Themes:
#' The editor automatically switches between `theme_light` and `theme_dark`
#' when used with [bslib::input_dark_mode()]. Otherwise, the editor will use
#' `theme_light` by default.
#'
#' A variety of themes are available for use. The full list of bundled themes
#' is: `r code_editor_themes(mode="docs")`.
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   input_code_editor(
#'     "sql_query",
#'     value = "SELECT * FROM table",
#'     language = "sql"
#'   ),
#'   verbatimTextOutput("code_output")
#' )
#'
#' server <- function(input, output, session) {
#'   output$code_output <- renderPrint({
#'     input$sql_query
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @param id Input ID. Access the current value with `input$<id>`.
#' @param label Display label for the input. Default is `NULL` for no label.
#' @param value Code content. Default is an empty string.
#' @param ... Named arguments, e.g. `class` and `style`, that will be added to
#'   the outer container of the input.
#' @param language Programming language for syntax highlighting. Supported
#'   languages include `"r"`, `"python"`, `"julia"`, `"sql"`, `"javascript"`,
#'   `"typescript"`, `"html"`, `"css"`, `"scss"`, `"sass"`, `"json"`,
#'   `"markdown"`, `"yaml"`, `"xml"`, `"toml"`, `"ini"`, `"bash"`, `"docker"`,
#'   `"latex"`, `"cpp"`, `"rust"`, `"diff"`, and `"plain"`. Default is
#'   `"plain"`.
#' @param height CSS height of the editor. Default is `"300px"`.
#' @param width CSS width of the editor. Default is `"100%"`.
#' @param theme_light,theme_dark Theme to use in light or dark mode. Defaults to
#'   `"github-light"` and `"github-dark"`, respectively. See the Theme section
#'   for more details.
#' @param read_only Whether the editor should be read-only. Default is `FALSE`.
#' @param line_numbers Whether to show line numbers. Default is `TRUE`, except
#'   for markdown and plain text.
#' @param word_wrap Whether to wrap long lines, by default disabled when
#'   `line_numbers` is `FALSE`.
#' @param tab_size Number of spaces per tab. Default is `2`.
#' @param indentation Type of indentation: `"space"` or `"tab"`. Default is
#'   `"space"`.
#' @inheritParams card
#'
#' @return An HTML tag object that can be included in a Shiny UI.
#'
#' @family input controls
#' @export
input_code_editor <- function(
  id,
  label = NULL,
  value = "",
  ...,
  language = "plain",
  height = "auto",
  width = "100%",
  theme_light = "github-light",
  theme_dark = "github-dark",
  read_only = FALSE,
  line_numbers = NULL,
  word_wrap = NULL,
  tab_size = 2,
  indentation = c("space", "tab"),
  fill = TRUE
) {
  dots <- separate_arguments(...)

  stopifnot("`value` must be a character" = is.character(value))
  value <- paste(value, collapse = "\n")

  # Restore input for bookmarking support
  value <- shiny::restoreInput(id, default = value)

  # Validate inputs
  check_value_line_count(value)
  language <- arg_match_language(language)
  theme_light <- arg_match_theme(theme_light, "theme_light")
  theme_dark <- arg_match_theme(theme_dark, "theme_dark")

  if (is.null(line_numbers)) {
    line_numbers <- !language %in%
      c("markdown", "md", "plain", "plaintext", "text", "txt")
  }

  if (is.null(word_wrap)) {
    word_wrap <- !line_numbers
  }

  stopifnot(
    "All arguments in `...` must be named" = length(dots$children) == 0,
    "`fill` must be boolean" = rlang::is_bool(fill),
    "`line_numbers` must be boolean" = rlang::is_bool(line_numbers),
    "`read_only` must be boolean" = rlang::is_bool(read_only),
    "`word_wrap` must be boolean" = rlang::is_bool(word_wrap),
    "`tab_size` must be a single positive integer" = {
      rlang::is_scalar_integerish(tab_size) && tab_size >= 1
    }
  )

  indentation <- rlang::arg_match(indentation)
  insert_spaces <- indentation == "space"

  # Create inner container that will hold the actual editor
  editor_inner <- htmltools::tags$div(
    class = "code-editor",
    bslib::as_fill_item(),
    style = htmltools::css(
      display = "grid"
    )
  )

  label_tag <- shiny_input_label(id, label)

  tag <- htmltools::tag(
    "bslib-code-editor",
    list(
      id = id,
      style = htmltools::css(
        height = height,
        width = width
      ),
      !!!dots$attribs,
      language = language,
      value = value,
      `theme-light` = theme_light,
      `theme-dark` = theme_dark,
      readonly = tolower(as.character(read_only)),
      `line-numbers` = tolower(as.character(line_numbers)),
      `word-wrap` = tolower(as.character(word_wrap)),
      `tab-size` = as.character(tab_size),
      `insert-spaces` = tolower(as.character(insert_spaces)),
      label_tag,
      editor_inner,
      code_editor_dependencies()
    )
  )

  tag <- as_fillable_container(tag)
  if (fill) {
    tag <- as_fill_item(tag)
  }

  tag <- tag_require(tag, version = 5, caller = "input_code_editor()")
  as_fragment(tag, page = page_fillable)
}

#' @rdname input_code_editor
#' @inheritParams nav_insert
#' @export
update_code_editor <- function(
  id,
  ...,
  value = NULL,
  label = NULL,
  language = NULL,
  theme_light = NULL,
  theme_dark = NULL,
  read_only = NULL,
  line_numbers = NULL,
  word_wrap = NULL,
  tab_size = NULL,
  indentation = NULL,
  session = get_current_session()
) {
  # Ensure no extra arguments
  rlang::check_dots_empty()

  # Validate inputs if provided
  if (!is.null(value)) {
    stopifnot("`value` must be a character" = is.character(value))
    value <- paste(value, collapse = "\n")
    check_value_line_count(value)
  }
  if (!is.null(language)) {
    language <- arg_match_language(language, "language")
  }
  if (!is.null(theme_light)) {
    theme_light <- arg_match_theme(theme_light, "theme_light")
  }
  if (!is.null(theme_dark)) {
    theme_dark <- arg_match_theme(theme_dark, "theme_dark")
  }

  # Build message with only non-NULL values
  message <- dropNulls(list(
    value = value,
    label = if (!is.null(label)) processDeps(label, session),
    language = language,
    theme_light = theme_light,
    theme_dark = theme_dark,
    read_only = read_only,
    line_numbers = line_numbers,
    word_wrap = word_wrap,
    tab_size = tab_size,
    indentation = if (!is.null(indentation)) {
      rlang::arg_match(indentation, c("space", "tab"))
    }
  ))

  # Send message to JavaScript binding
  session$sendInputMessage(id, message)

  invisible(NULL)
}

code_editor_themes <- function(mode = c("raw", "docs")) {
  mode <- match.arg(mode)
  themes_dir <- path_inst("lib", "prism-code-editor", "themes")

  if (!dir.exists(themes_dir)) {
    return(character(0))
  }

  theme_files <- list.files(themes_dir, pattern = "\\.css$")
  themes <- sub("\\.css$", "", theme_files)

  switch(
    mode,
    raw = themes,
    docs = paste(sprintf('`"%s"`', themes), collapse = ", ")
  )
}

# Internal dependency functions ------------------------------------------------

code_editor_dependencies <- function() {
  list(
    code_editor_dependency_prism(),
    code_editor_dependency()
  )
}

code_editor_dependency_prism <- function() {
  htmltools::htmlDependency(
    name = "prism-code-editor",
    version = version_prism_code_editor,
    package = "bslib",
    src = "lib/prism-code-editor",
    script = list(src = "index.js", type = "module"),
    stylesheet = c("layout.css", "copy.css", "scrollbar.css"),
    all_files = TRUE
  )
}

code_editor_dependency <- function() {
  minified <- get_shiny_devmode_option("shiny.minified", default = TRUE)
  htmltools::htmlDependency(
    name = "bslib-code-editor",
    version = get_package_version("bslib"),
    package = "bslib",
    src = "components/dist",
    script = list(
      list(
        src = paste0("code-editor", if (minified) ".min", ".js"),
        type = "module"
      )
    ),
    stylesheet = "code-editor.css",
    all_files = FALSE
  )
}

# Validation helpers -----------------------------------------------------------

arg_match_theme <- function(theme, arg_name = "theme") {
  if (is.null(theme)) {
    return(invisible(NULL))
  }

  available_themes <- code_editor_themes()

  rlang::arg_match(
    theme,
    values = available_themes,
    error_arg = arg_name,
    error_call = rlang::caller_env()
  )
}

arg_match_language <- function(language, arg_name = "language") {
  if (is.null(language)) {
    return(invisible(NULL))
  }

  # Language aliases (user-friendly names → prism grammar names)
  language_aliases <- c(
    "md" = "markdown",
    "html" = "markup",
    "plain" = "plain",
    "plaintext" = "plain",
    "text" = "plain",
    "txt" = "plain"
  )

  # All supported languages: bundled prism grammars + aliases
  supported_languages <- c(
    code_editor_bundled_languages,
    names(language_aliases)
  )

  language <- rlang::arg_match(
    language,
    values = supported_languages,
    error_arg = arg_name,
    error_call = rlang::caller_env()
  )

  # Resolve aliases to their actual grammar names
  if (language %in% names(language_aliases)) {
    language <- language_aliases[[language]]
  }

  language
}

check_value_line_count <- function(value) {
  if (is.null(value) || !is.character(value) || length(value) == 0) {
    return(invisible(NULL))
  }

  line_count <- length(strsplit(value, "\n", fixed = TRUE)[[1]])

  if (line_count >= 1000) {
    rlang::warn(c(
      sprintf("Code editor value contains %d lines.", line_count),
      "i" = "The editor may experience performance issues with 1,000 or more lines."
    ))
  }

  invisible(NULL)
}

test_that("code_editor_themes returns character vector of themes", {
  themes <- code_editor_themes()

  expect_type(themes, "character")
  expect_true(length(themes) > 0)

  # Check for expected default themes
  expect_true("github-light" %in% themes)
  expect_true("github-dark" %in% themes)
  expect_true("vs-code-light" %in% themes)
  expect_true("vs-code-dark" %in% themes)
})

test_that("input_code_editor() rejects invalid themes", {
  expect_snapshot(error = TRUE, {
    input_code_editor("test", theme_light = "invalid-theme")
  })
})

test_that("arg_match_language rejects invalid languages", {
  expect_snapshot(error = TRUE, {
    input_code_editor("test", language = "fortran")
  })
})

test_that("input_code_editor generates correct HTML structure", {
  editor <- input_code_editor(
    "test_editor",
    value = "SELECT * FROM table",
    language = "sql"
  )

  # Check that dependencies are attached
  deps <- htmltools::findDependencies(editor)
  dep_names <- sapply(deps, function(d) d$name)
  expect_true("prism-code-editor" %in% dep_names)
  expect_true("bslib-code-editor" %in% dep_names)

  html <- as.character(editor)

  # Check for bslib-code-editor custom element tag
  expect_match(html, '<bslib-code-editor')

  # Check for correct ID
  expect_match(html, 'id="test_editor"')

  # Check for attributes
  expect_match(html, 'language="sql"')
  expect_match(html, 'value="SELECT \\* FROM table"')
})

test_that("input_code_editor handles all parameters correctly", {
  editor <- input_code_editor(
    "full_editor",
    value = "print('hello')",
    language = "python",
    height = "500px",
    width = "80%",
    theme_light = "vs-code-light",
    theme_dark = "vs-code-dark",
    read_only = TRUE,
    line_numbers = FALSE,
    word_wrap = TRUE,
    tab_size = 4,
    indentation = "tab"
  )

  html <- as.character(editor)

  # Check all attributes
  expect_match(html, 'language="python"')
  expect_match(html, 'theme-light="vs-code-light"')
  expect_match(html, 'theme-dark="vs-code-dark"')
  expect_match(html, 'readonly="true"')
  expect_match(html, 'line-numbers="false"')
  expect_match(html, 'word-wrap="true"')
  expect_match(html, 'tab-size="4"')
  expect_match(html, 'insert-spaces="false"') # tab indentation

  # Check style attributes
  expect_match(html, 'height:\\s*500px')
  expect_match(html, 'width:\\s*80%')
})

test_that("input_code_editor uses correct defaults", {
  editor <- input_code_editor("default_editor", language = "r")

  html <- as.character(editor)

  expect_match(html, 'language="r"')
  expect_match(html, 'theme-light="github-light"')
  expect_match(html, 'theme-dark="github-dark"')
  expect_match(html, 'readonly="false"')
  expect_match(html, 'line-numbers="true"')
  expect_match(html, 'word-wrap="false"')
  expect_match(html, 'tab-size="2"')
  expect_match(html, 'insert-spaces="true"')
  expect_match(html, 'height:\\s*auto')
  expect_match(html, 'width:\\s*100%')
})

test_that("input_code_editor validates theme names", {
  expect_snapshot(error = TRUE, {
    input_code_editor("test", theme_light = "invalid-theme")
    input_code_editor("test", theme_dark = "invalid-theme")
  })
})

test_that("input_code_editor validates `language`", {
  expect_snapshot(error = TRUE, {
    input_code_editor("test", language = "fortran")
  })
})

test_that("input_code_editor handles empty `value`", {
  editor <- input_code_editor("empty_editor", value = "")

  html <- as.character(editor)
  expect_match(html, 'value=""')
})

test_that("input_code_editor handles character vector `value`", {
  editor <- input_code_editor("empty_editor", value = c("one", "two", "three"))

  html <- as.character(editor)
  expect_match(html, 'value="one&#10;two&#10;three"')
})

test_that("input_code_editor indentation parameter works correctly", {
  editor_spaces <- input_code_editor("test1", indentation = "space")
  editor_tabs <- input_code_editor("test2", indentation = "tab")

  html_spaces <- as.character(editor_spaces)
  html_tabs <- as.character(editor_tabs)

  expect_match(html_spaces, 'insert-spaces="true"')
  expect_match(html_tabs, 'insert-spaces="false"')
})

test_that("update_code_editor validates inputs", {
  expect_snapshot(error = TRUE, {
    update_code_editor("test", language = "fortran", session = NULL)
  })

  expect_snapshot(error = TRUE, {
    update_code_editor("test", theme_light = "invalid", session = NULL)
  })

  expect_snapshot(error = TRUE, {
    update_code_editor("test", theme_dark = "invalid", session = NULL)
  })

  expect_snapshot(error = TRUE, {
    update_code_editor("test", indentation = "invalid", session = NULL)
  })
})

test_that("input_code_editor supports label parameter", {
  editor_with_label <- input_code_editor("test", label = "SQL Query")
  editor_without_label <- input_code_editor("test2")

  html_with <- as.character(editor_with_label)
  html_without <- as.character(editor_without_label)

  # Editor with label should have a label element
  expect_match(html_with, "<label")
  expect_match(html_with, "SQL Query")

  # Editor without label should still have the custom element tag
  expect_true(grepl("<bslib-code-editor", html_without))
})

test_that("input_code_editor attaches dependencies", {
  editor <- input_code_editor("test")

  # The editor should be a tag with the dependencies attached
  expect_s3_class(editor, "shiny.tag")

  # Extract dependencies
  deps <- htmltools::findDependencies(editor)
  expect_true(length(deps) > 0)

  # Check that all expected dependencies are present
  dep_names <- sapply(deps, function(d) d$name)
  expect_true("prism-code-editor" %in% dep_names)
  expect_true("bslib-code-editor" %in% dep_names)
})

test_that("input_code_editor works with different languages", {
  languages <- c("sql", "python", "r", "javascript", "markup", "css", "json")

  for (lang in languages) {
    editor <- input_code_editor(paste0("editor_", lang), language = lang)
    html <- as.character(editor)
    expect_match(html, sprintf('language="%s"', lang))
  }
})

test_that("input_code_editor tab_size validates range", {
  # Valid tab sizes
  expect_silent(input_code_editor("test1", tab_size = 2))
  expect_silent(input_code_editor("test2", tab_size = 4))
  expect_silent(input_code_editor("test3", tab_size = 8))

  expect_error(input_code_editor("test4", tab_size = 0))
})

test_that("input_code_editor warns when value has 1,000 or more lines", {
  # No warning for small values
  expect_silent(input_code_editor("test1", value = "line1\nline2\nline3"))

  # No warning for exactly 999 lines
  value_999 <- paste(rep("line", 999), collapse = "\n")
  expect_silent(input_code_editor("test2", value = value_999))

  # Warning for exactly 1000 lines
  value_1000 <- paste(rep("line", 1000), collapse = "\n")
  expect_warning(
    input_code_editor("test3", value = value_1000),
    "Code editor value contains 1000 lines"
  )
  expect_warning(
    input_code_editor("test3", value = value_1000),
    "performance issues"
  )

  # Warning for more than 1000 lines
  value_2000 <- paste(rep("line", 2000), collapse = "\n")
  expect_warning(
    input_code_editor("test4", value = value_2000),
    "Code editor value contains 2000 lines"
  )
})

test_that("update_code_editor warns when value has 1,000 or more lines", {
  mock_session <- list(sendInputMessage = function(...) invisible())

  # No warning for small values
  expect_silent(update_code_editor(
    "test1",
    value = "line1\nline2",
    session = mock_session
  ))

  # Warning for exactly 1000 lines
  value_1000 <- paste(rep("line", 1000), collapse = "\n")
  expect_warning(
    update_code_editor("test4", value = value_1000, session = mock_session),
    "Code editor value contains 1000 lines"
  )

  # Warning for more than 1000 lines
  value_2000 <- paste(rep("line", 2000), collapse = "\n")
  expect_warning(
    update_code_editor("test4", value = value_1000, session = mock_session),
    "Code editor value contains 1000 lines"
  )
})

test_that("check_value_line_count handles edge cases", {
  # NULL value
  expect_silent(bslib:::check_value_line_count(NULL))

  # Empty string
  expect_silent(bslib:::check_value_line_count(""))

  # Empty character vector
  expect_silent(bslib:::check_value_line_count(character(0)))

  # Single line
  expect_silent(bslib:::check_value_line_count("single line"))

  # Multiple lines below threshold
  value_100 <- paste(rep("line", 100), collapse = "\n")
  expect_silent(bslib:::check_value_line_count(value_100))
})

#' A Bootstrap card component
#'
#' @description
#' A general purpose container for grouping related UI elements together with a
#' border and optional padding. To learn more about [card()]s, see [the Cards
#' article](https://rstudio.github.io/bslib/articles/cards/index.html) or the
#' other articles listed in the _References_ section below.
#'
#' @param ... Unnamed arguments can be any valid child of an [htmltools
#'   tag][htmltools::tags] (which includes card items such as [card_body()].
#'   Named arguments become HTML attributes on returned UI element.
#' @param full_screen If `TRUE`, an icon will appear when hovering over the card
#'   body. Clicking the icon expands the card to fit viewport size.
#' @param height Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `height="200px"`). Doesn't apply when a card is made `full_screen`
#'   (in this case, consider setting a `height` in [card_body()]).
#' @param max_height,min_height Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `max_height="200px"`). Doesn't apply when a card is made `full_screen`
#'   (in this case, consider setting a `max_height` in [card_body()]).
#' @param fill Whether or not to allow the card to grow/shrink to fit a
#'   fillable container with an opinionated height (e.g., `page_fillable()`).
#' @param class Additional CSS classes for the returned UI element.
#' @param wrapper A function (which returns a UI element) to call on unnamed
#'   arguments in `...` which are not already card item(s) (like
#'   [card_header()], [card_body()], etc.). Note that non-card items are grouped
#'   together into one `wrapper` call (e.g. given `card("a", "b",
#'   card_body("c"), "d")`, `wrapper` would be called twice, once with `"a"` and
#'   `"b"` and once with `"d"`).
#' @param id Provide a unique identifier for the `card()` or `value_box()` to
#'   report its full screen state to Shiny. For example, using `id = "my_card"`,
#'   you can observe the card's full screen state with
#'   `input$my_card_full_screen`.
#'
#' @return A [htmltools::div()] tag.
#'
#' @export
#' @family Components
#'
#' @seealso [Card item functions][card_body()] create the various parts of a
#'   card.
#' @seealso [navset_card_tab()], [navset_card_pill()] and
#'   [navset_card_underline()] create cards with tabbed navigation.
#' @seealso [layout_columns()] and [layout_column_wrap()] help position multiple
#'   cards into columns and rows and can also be used inside a card.
#' @seealso [layout_sidebar()] adds a sidebar to a card when nested in [card()]
#'   or [card_body()].
#' @seealso [value_box()] uses [card()] to highlight a showcase a key piece of
#'   information.
#'
#' @references Several articles on the bslib website feature the card component:
#'
#'   * [Cards](https://rstudio.github.io/bslib/articles/cards/index.html)
#'   * [Get Started: Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'   * [Get Started: Any Project](https://rstudio.github.io/bslib/articles/any-project/index.html)
#'   * [Column-based layouts](https://rstudio.github.io/bslib/articles/column-layout/index.html)
#'   * [Filling layouts: Full-screen cards](https://rstudio.github.io/bslib/articles/filling/index.html#full-screen-cards)
#'
#' @examplesIf rlang::is_interactive()
#' library(htmltools)
#'
#' card(
#'   full_screen = TRUE,
#'   card_header(
#'     "This is the header"
#'   ),
#'   card_body(
#'     p("This is the body."),
#'     p("This is still the body.")
#'   ),
#'   card_footer(
#'     "This is the footer"
#'   )
#' )
#'
card <- function(
  ...,
  full_screen = FALSE,
  height = NULL,
  max_height = NULL,
  min_height = NULL,
  fill = TRUE,
  class = NULL,
  wrapper = card_body,
  id = NULL
) {
  args <- rlang::list2(...)
  argnames <- rlang::names2(args)

  attribs <- args[nzchar(argnames)]
  children <- as_card_items(args[!nzchar(argnames)], wrapper = wrapper)
  children <- card_image_add_classes(children)

  is_shiny_input <- !is.null(id)

  if (full_screen && is.null(id)) {
    # a11y: full screen cards need an ID for aria-controls on the toggle button
    id <- paste0("bslib-card-", p_randomInt(1000, 10000))
  }

  tag <- div(
    id = id,
    class = "card bslib-card bslib-mb-spacing",
    class = if (is_shiny_input) "bslib-card-input",
    style = css(
      height = validateCssUnit(height),
      max_height = validateCssUnit(max_height),
      min_height = validateCssUnit(min_height)
    ),
    "data-bslib-card-init" = NA,
    "data-full-screen" = if (full_screen) "false",
    !!!attribs,
    !!!children,
    if (full_screen) full_screen_toggle(attribs$id),
    card_init_js(),
    component_dependencies()
  )

  tag <- bindFillRole(tag, container = TRUE, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "card()")
  )
}

as_card_items <- function(children, wrapper) {
  # We don't want NULLs creating empty card bodys
  children <- children[vapply(children, function(x) length(x) > 0, logical(1))]

  if (!is.function(wrapper)) {
    return(children)
  }

  # Any children that are `is.card_item` should be included verbatim. Any
  # children that are not, should be wrapped in card_body(). Consecutive children
  # that are not card_item, should be wrapped in a single card_body().
  needs_wrap <- !vapply(children, is.card_item, logical(1))
  needs_wrap_rle <- rle(needs_wrap)
  start_indices <- c(1, head(cumsum(needs_wrap_rle$lengths) + 1, -1))
  children <- mapply(
    start_indices,
    needs_wrap_rle$lengths,
    needs_wrap_rle$values,
    FUN = function(start, length, wrap) {
      these_children <- children[start:(start + length - 1)]
      if (wrap) {
        list(wrapper(these_children))
      } else {
        these_children
      }
    },
    SIMPLIFY = FALSE
  )
  unlist(children, recursive = FALSE)
}

#' Card items
#'
#' Components designed to be provided as direct children of a [card()]. For a
#' general overview of the [card()] API, see [the Cards
#' article](https://rstudio.github.io/bslib/articles/cards/index.html) or the
#' other articles listed in the _References_ section of the [card()]
#' documentation.
#'
#' @param ... Unnamed arguments can be any valid child of an [htmltools
#'   tag][htmltools::tags]. Named arguments become HTML attributes on returned
#'   UI element.
#' @param min_height,max_height,max_height_full_screen Any valid [CSS length
#'   unit][htmltools::validateCssUnit()].
#' @param fillable Whether or not the card item should be a fillable (i.e.
#'   flexbox) container.
#' @param fill Whether to allow this element to grow/shrink to fit its `card()`
#'   container.
#' @param gap A [CSS length unit][htmltools::validateCssUnit()] defining the
#'   `gap` (i.e., spacing) between elements provided to `...`. This argument is only applicable when `fillable = TRUE`
#' @inheritParams card
#' @inheritParams page_fillable
#'
#' @return An [htmltools::div()] tag.
#'
#' @export
#'
#' @seealso [card()] creates a card component.
#' @seealso [navset_card_tab()], [navset_card_pill()] and
#'   [navset_card_underline()] create cards with tabbed navigation.
#' @seealso [layout_columns()] and [layout_column_wrap()] help position multiple
#'   cards into columns and rows and can also be used inside a card.
#' @seealso [layout_sidebar()] adds a sidebar to a card when nested in [card()]
#'   or [card_body()].
#'
#' @describeIn card_body A general container for the "main content" of a [card()].
card_body <- function(
  ...,
  fillable = TRUE,
  min_height = NULL,
  max_height = NULL,
  max_height_full_screen = max_height,
  height = NULL,
  padding = NULL,
  gap = NULL,
  fill = TRUE,
  class = NULL
) {
  if (fillable) {
    register_runtime_package_check("`card_body()`", "shiny", "1.7.4")
    register_runtime_package_check("`card_body()`", "htmlwidgets", "1.6.0")
  }

  tag <- div(
    class = "card-body",
    class = if (fillable) "bslib-gap-spacing",
    style = css(
      min_height = validateCssUnit(min_height),
      "--bslib-card-body-max-height" = validateCssUnit(max_height),
      "--bslib-card-body-max-height-full-screen" = validateCssUnit(
        max_height_full_screen
      ),
      margin_top = "auto",
      margin_bottom = "auto",
      # .card-body already adds `flex: 1 1 auto` so make sure to override it
      flex = if (fill) "1 1 auto" else "0 0 auto",
      padding = validateCssPadding(padding),
      gap = validateCssUnit(gap),
      height = validateCssUnit(height)
    ),
    ...
  )

  tag <- bindFillRole(tag, item = fill, container = fillable)

  # Make sure user has the opportunity to override the classes added by bindFillRole()
  tag <- tagAppendAttributes(tag, class = class)

  as.card_item(tag)
}

#' @describeIn card_body Similar to `card_header()` but without the border and background color.
#' @param container a function to generate an HTML element.
#' @export
card_title <- function(..., container = htmltools::h5) {
  container(...)
}

#' @describeIn card_body A header (with border and background color) for the `card()`. Typically appears before a `card_body()`.
#' @param container a function that generates an [htmltools tag][htmltools::tags].
#' @export
card_header <- function(..., class = NULL, container = htmltools::div) {
  as.card_item(
    container(class = "card-header", class = class, ...)
  )
}

#' @describeIn card_body A header (with border and background color) for the `card()`. Typically appears after a `card_body()`.
#' @export
card_footer <- function(..., class = NULL) {
  as.card_item(
    div(class = "card-footer", class = class, ...)
  )
}

#' @describeIn card_body Include static images in a card, for example as an
#'   image cap at the top or bottom of the card.
#'
#' @param file A file path pointing an image. Local images (i.e. not a URI
#'   starting with `https://` or similar) will be base64 encoded and provided to
#'   the `src` attribute of the `<img>`. Alternatively, you may directly set
#'   the image `src`, in which case `file` is ignored.
#' @param alt Alternate text for the image, used by screen readers and assistive
#'   devices. Provide alt text with a description of the image for any images
#'   with important content. If alt text is not provided, the image will be
#'   considered to be decorative and will not be read or announced by screen
#'   readers.
#'
#'   For more information, the Web Accessibility Initiative (WAI) has a
#'   [helpful tutorial on alt text](https://www.w3.org/WAI/tutorials/images/).
#' @param src The `src` attribute of the `<img>` tag. If provided, `file` is
#'   ignored entirely. Use `src` to provide a relative path to a file that will
#'   be served by the Shiny application and should not be base64 encoded.
#' @param href An optional URL to link to when a user clicks on the image.
#' @param border_radius Which side of the image should have rounded corners,
#'   useful when `card_image()` is used as an image cap at the top or bottom of
#'   the card.
#'
#'   The value of `border_radius` determines whether the `card-img-top`
#'   (`"top"`), `card-img-bottom` (`"bottom"`), or `card-img` (`"all"`)
#'   [Bootstrap
#'   classes](https://getbootstrap.com/docs/5.3/components/card/#images) are
#'   applied to the card. The default `"auto"` value will use the image's
#'   position within a `card()` to automatically choose the appropriate class.
#' @param mime_type The mime type of the `file` when it is base64 encoded. This
#'   argument is available for advanced use cases where [mime::guess_type()] is
#'   unable to automatically determine the file type.
#' @param container A function to generate an HTML element to contain the image.
#'   Setting this value to `card_body()` places the image inside the card body
#'   area, otherwise the image will extend to the edges of the card.
#' @param width Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `width="100%"`).
#'
#' @export
card_image <- function(
  file,
  ...,
  alt = "",
  src = NULL,
  href = NULL,
  border_radius = c("auto", "top", "bottom", "all", "none"),
  mime_type = NULL,
  class = NULL,
  height = NULL,
  fill = FALSE,
  width = NULL,
  container = NULL
) {
  if (any(!nzchar(rlang::names2(list(...))))) {
    rlang::abort(
      c(
        "Unnamed arguments were included in `...`.",
        i = "All additional arguments to `card_image()` in `...` should be named attributes for the `<img>` tag."
      )
    )
  }

  border_radius <- rlang::arg_match(border_radius)

  if (is.null(src)) {
    if (grepl("^([[:alnum:]]+:)?//|data:", file)) {
      src <- file
    } else {
      if (!file.exists(file)) {
        rlang::abort(
          c(
            sprintf("`file` does not exist: %s", file),
            i = sprintf(
              "If `file` is a remote file or will be served by the Shiny app, use a URL or set `src = \"%s\"`.",
              file
            )
          )
        )
      }
      src <- base64enc::dataURI(
        file = file,
        mime = mime_type %||% mime::guess_type(file)
      )
    }
  }

  image <- tags$img(
    src = src,
    alt = alt,
    class = "img-fluid",
    class = switch(
      border_radius,
      all = "card-img",
      top = "card-img-top",
      bottom = "card-img-bottom",
      NULL
    ),
    style = css(
      height = validateCssUnit(height),
      width = validateCssUnit(width)
    ),
    ...
  )

  image <- bindFillRole(image, item = fill)
  image <- tagAppendAttributes(image, class = class)

  if (!is.null(href)) {
    image <- bindFillRole(
      tags$a(href = href, image),
      container = TRUE,
      item = TRUE
    )
  }

  if (is.function(container)) {
    image <- container(image)
  } else {
    image <- as.card_item(image)
  }

  class(image) <- c(
    if (border_radius == "auto") "card_image_auto",
    "card_image",
    class(image)
  )

  image
}

card_image_add_classes <- function(children) {
  for (idx_child in seq_along(children)) {
    if (inherits(children[[idx_child]], "card_image_auto")) {
      card_img_class <-
        if (length(children) == 1) {
          "card-img"
        } else if (idx_child == 1) {
          "card-img-top"
        } else if (idx_child == length(children)) {
          "card-img-bottom"
        }

      children[[idx_child]] <- tagAppendAttributes(
        children[[idx_child]],
        class = card_img_class
      )
    }
  }

  children
}

#' @describeIn card_body Mark an object as a card item. This will prevent the
#'   [card()] from putting the object inside a `wrapper` (i.e., a
#'   `card_body()`).
#' @param x an object to test (or coerce to) a card item.
#' @export
as.card_item <- function(x) {
  class(x) <- c("card_item", class(x))
  x
}

#' @rdname card_body
#' @export
is.card_item <- function(x) {
  inherits(x, "card_item")
}

full_screen_toggle <- function(id_controls) {
  tooltip(
    tags$button(
      class = "bslib-full-screen-enter",
      class = "badge rounded-pill",
      "aria-expanded" = "false",
      "aria-controls" = id_controls,
      "aria-label" = "Expand card",
      full_screen_toggle_icon()
    ),
    "Expand"
  )
}

card_init_js <- function() {
  tags$script(
    `data-bslib-card-init` = NA,
    HTML("bslib.Card.initializeAllCards();")
  )
}

full_screen_toggle_icon <- function() {
  # https://www.visiwig.com/icons/
  # https://www.visiwig.com/icons-license/
  HTML(
    '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24" style="height:1em;width:1em;fill:currentColor;" aria-hidden="true" role="img"><path d="M20 5C20 4.4 19.6 4 19 4H13C12.4 4 12 3.6 12 3C12 2.4 12.4 2 13 2H21C21.6 2 22 2.4 22 3V11C22 11.6 21.6 12 21 12C20.4 12 20 11.6 20 11V5ZM4 19C4 19.6 4.4 20 5 20H11C11.6 20 12 20.4 12 21C12 21.6 11.6 22 11 22H3C2.4 22 2 21.6 2 21V13C2 12.4 2.4 12 3 12C3.6 12 4 12.4 4 13V19Z"/></svg>'
  )
}

# jcheng 2022-06-06: Removing for now; list items have more features than I'm
# ready to design an API for right now
#
# #' @rdname card_body
# #' @export
# card_list <- function(...) {
#   res <- tags$ul(class = "list-group list-group-flush", ...)
#   as.card_item(res)
# }
#
# #' @export
# card_list_item <- function(...) {
#   tags$li(class = "list-group-item", ...)
# }

test_that("card_image()", {
  show_raw_html <- function(x) {
    cat(format(x))
  }

  expect_snapshot(
    show_raw_html(
      card(
        card_image("https://example.com/image.jpg"),
        card_body("image cap on top of card")
      )
    )
  )

  expect_snapshot(
    show_raw_html(
      card(
        card_body("image cap on bottom of card"),
        card_image("https://example.com/image.jpg")
      )
    )
  )

  expect_snapshot(
    show_raw_html(
      card(
        card_header("header"),
        card_image("https://example.com/image.jpg"),
        card_body("image not a cap")
      )
    )
  )

  expect_snapshot(
    show_raw_html(
      card(
        card_image("https://example.com/image.jpg", alt = "card-img")
      )
    )
  )
})

test_that("card_image() input validation", {
  expect_snapshot(
    error = TRUE,
    card_image("cat.jpg")
  )

  expect_snapshot(
    error = TRUE,
    card_image("foo", "bar")
  )

  expect_snapshot(
    error = TRUE,
    card_image("foo", border_radius = "guess")
  )
})

library(shiny)
library(bslib)
library(plotly)
library(leaflet)
library(gt)

plot_hist <- function(var) {
  config(
    plot_ly(y = diamonds[[var]], type = "histogram"),
    displayModeBar = FALSE
  )
}

lorem_ipsum_dolor_sit_amet <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Id nibh tortor id aliquet lectus proin nibh nisl. Adipiscing at in tellus integer feugiat. Arcu bibendum at varius vel pharetra vel turpis nunc eget. Cursus sit amet dictum sit amet justo. Sit amet consectetur adipiscing elit. Vestibulum mattis ullamcorper velit sed ullamcorper. Enim facilisis gravida neque convallis a. Elit duis tristique sollicitudin nibh sit amet. Magna eget est lorem ipsum. Gravida dictum fusce ut placerat orci nulla pellentesque dignissim. Mauris in aliquam sem fringilla ut morbi. Id semper risus in hendrerit gravida rutrum quisque non tellus. At erat pellentesque adipiscing commodo elit at imperdiet dui. Fames ac turpis egestas maecenas pharetra convallis posuere morbi. Duis convallis convallis tellus id interdum velit laoreet id. Aliquet lectus proin nibh nisl. Nunc vel risus commodo viverra maecenas accumsan lacus vel facilisis. Bibendum enim facilisis gravida neque convallis a."

main_grid <- layout_column_wrap(
  width = 1/3, heights_equal = "row",
  card(
    id = "card-dt",
    full_screen = TRUE,
    card_header("DT::dataTableOutput()"),
    DT::dataTableOutput("DT")
  ),
  navset_card_pill(
    title = "Shiny outputs",
    id = "card-navset-pill",
    full_screen = TRUE,
    nav_panel(
      "plotOutput",
      plotOutput("plot"),
      card_body(
        height = "100px",
        lorem_ipsum_dolor_sit_amet
      )
    ),
    nav_panel("imageOutput", plotOutput("image"))
  ),
  card(
    id = "card-leaflet",
    full_screen = TRUE,
    card_header("leaflet (via uiOutput())"),
    uiOutput("leaflet", fill = TRUE)
  ),
  card(
    id = "card-plotly",
    full_screen = TRUE,
    card_header("plotly::plotlyOutput()"),
    layout_column_wrap(
      width = 1/2,
      plotlyOutput("cut"),
      plotlyOutput("clarity")
    ),
    plotlyOutput("price")
  ),
  card(
    id = "card-plotly-static",
    full_screen = TRUE,
    card_header("Static plotly"),
    plot_hist("price"),
    layout_column_wrap(
      width = 1/2,
      plot_hist("cut"),
      plot_hist("clarity")
    )
  ),
  layout_column_wrap(
    width = 1,
    card(
      id = "card-image",
      full_screen = TRUE,
      card_header("card_image()"),
      card_image(
        file = "www/shiny-hex.svg",
        height = 200,
        href = "https://github.com/rstudio/shiny"
      )
    ),
    card(
      id = "card-gt",
      full_screen = TRUE,
      card_header("Scrollable gt()"),
      card_body(
        max_height = "400px",
        max_height_full_screen = "100%",
        gt::gt(mtcars)
      )
    )
  )
)

grid_height <- "calc(100vh - 5px)"

ui <- page_fluid(
  layout_column_wrap(
    width = NULL, heights_equal = "row",
    height = grid_height,
    style = css(grid_template_columns = "200px 1fr"),
    id = "grid_page",
    card(
      card_header("Input controls", class = "bg-primary"),
      checkboxInput(
        "fixed_height",
        "Fix height to viewport (this only applies to wide screens)",
        value = TRUE
      )
    ),
    main_grid
  )
)

server <- function(input, output, session) {

  theme_set(theme_minimal(base_size = 16))

  observeEvent(input$fixed_height, ignoreInit = TRUE, {
    css <- sprintf(
      "#grid_page { height: %s !important; }",
      if (input$fixed_height) grid_height else "auto"
    )

    insertUI(selector = "head", ui = tags$style(HTML(css)))
  })

  output$DT <- DT::renderDataTable({
    DT::datatable(mtcars, fillContainer = TRUE)
  })

  output$plot <- renderPlot({
    ggplot(mtcars) + geom_point(aes(wt, mpg))
  })

  output$image <- renderImage({
    list(src = "www/shiny-hex.svg", width = "100%", height = "100%")
  }, deleteFile = FALSE)

  output$cut <- renderPlotly(plot_hist("cut"))
  output$clarity <- renderPlotly(plot_hist("clarity"))
  output$price <- renderPlotly(plot_hist("price"))


  output$leaflet <- renderUI({
    addTiles(leaflet())
  })
  output$gt <- render_gt(gt(mtcars[1:5, ]))
}

shinyApp(ui, server)

# Generated by tools/yarn_install.R: do not edit by hand
version_bs5 <- "5.3.1"
version_bs4 <- "4.6.0"
version_bs3 <- "3.4.1"
version_accessibility <- "1.0.6"
version_prism_code_editor <- "4.2.0"
code_editor_bundled_languages <- c("r", "python", "julia", "sql", "javascript", "typescript", "markup", "css", "scss", "sass", "json", "markdown", "yaml", "xml", "toml", "ini", "bash", "docker", "latex", "cpp", "rust", "diff")

# Turn off the default cache until the calling function exits; when that
# happens, the previous default cache will be restored.
local_disable_cache <- function(env = parent.frame()) {
  withr::local_options(
    list(sass.cache = FALSE),
    .local_envir = env
  )
}

# "brand" and "preset" interact in a few ways that need to be reconciled.
# Ultimately, we want three layers:
#
# 1. The base Bootstrap styles, by version. E.g. Bootstrap 5
# 2. The `preset` styles, which are one layer above and adjust the base
#    Bootstrap with the preset defaults E.g. "shiny" (bslib) or "flatly" (Bootswatch)
# 3. The `brand` styles, which can be seen as the final layer of default styles.
#    These styles set Bootstrap variables to adjust the `preset` theme in the
#    direction of the brand's color palette and typography.
#
# The `preset` and `version` can be specified in two ways with `brand`:
#
# 1. Directly via `bs_theme()`:
#    `bs_theme(preset = "shiny", version = 5, brand = TRUE)`.
#
# 2. As part of the `brand` definition:
#    `bs_theme(brand = list(defaults = list(shiny = list(theme = list(preset = "flatly")))))`
#
# So the order of operations is:
#
# 1. Read `brand`, if necessary.
# 2. If `bs_theme()` provides guidance for `version` and `theme`, use those preferences.
# 3. If not, inspect `brand` for guidance.
# 4. Finally, use our own defaults.
#
# Importantly we need to separately read `brand` and then resolve the preset and
# then resolve the brand bundle.

bs_brand_bundle <- function(brand, version = version_default()) {
  brand <- brand_resolve(brand)

  if (is.null(brand)) {
    return()
  }

  brand_fonts <- brand_sass_fonts(brand)
  brand_color_palette <- brand_sass_color_palette(brand)
  brand_color <- brand_sass_color(brand)
  brand_defaults <- brand_sass_defaults_bootstrap(brand)
  brand_typography <- brand_sass_typography(brand)

  if (version <= 4) {
    rlang::warn(
      sprintf(
        "Branded theming works best with Bootstrap v5. Some features may not work as expected for Bootstrap v%s.",
        version
      )
    )
  }

  sass_bundle(
    "brand_base" = switch_version(
      version,
      five = sass_layer_file(
        system_file("brand", "bs5", "_brand-yml.scss", package = "bslib")
      ),
      default = list()
    ),
    "brand_defaults" = brand_defaults$layer,
    "brand" = sass_layer(
      defaults = list2(
        "//* ---- brand.color.palette ---- *//",
        !!!brand_color_palette$defaults,
        "//* ---- brand.defaults ---- *//",
        !!!brand_defaults$defaults,
        "//* ---- brand.color ---- *//",
        !!!brand_color$defaults,
        "//* ---- brand.typography ---- *//",
        !!!brand_fonts$defaults,
        !!!brand_typography$defaults,
      ),
      rules = list2(
        brand_color_palette$rules,
        !!!brand_fonts$rules
      )
    )
  )
}

brand_resolve <- function(brand, ...) {
  UseMethod("brand_resolve")
}

#' @export
brand_resolve.list <- function(brand, ...) {
  brand <- as_brand_yml(brand)
  brand_resolve(brand, ...)
}

#' @export
`brand_resolve.NULL` <- function(brand, ...) {
  brand <- tryCatch(
    read_brand_yml(NULL),
    error = function(err) {
      # Couldn't find _brand.yml but we're not going to error
      NULL
    }
  )
  if (is.null(brand)) return(NULL)
  brand_resolve(brand, ...) # future compat if we add anything to the ...
}

#' @export
brand_resolve.logical <- function(brand, ...) {
  if (identical(brand, FALSE)) {
    return(NULL)
  }
  brand <- read_brand_yml(NULL)
  brand_resolve(brand, ...) # future compat if we add anything to the ...
}

#' @export
brand_resolve.character <- function(brand, ...) {
  brand <- read_brand_yml(brand)
  brand_resolve(brand, ...)
}

#' @export
brand_resolve.brand_yml <- function(brand, ...) {
  brand
}

brand_resolve_preset <- function(brand, preset = NULL, version = NULL) {
  version_resolved <-
    version %||%
      brand_pluck(brand, "defaults", "shiny", "theme", "version") %||%
      brand_pluck(brand, "defaults", "bootstrap", "version") %||%
      version_default()

  preset_resolved <-
    preset %||%
      brand_pluck(brand, "defaults", "shiny", "theme", "preset") %||%
      switch_version(version_resolved, five = "shiny", default = "bootstrap")

  resolve_bs_preset(preset_resolved, version = version_resolved)
}

# Brand Sass -------------------------------------------------------------------
brand_sass_color_palette <- function(brand) {
  palette <- brand_pluck(brand, "color", "palette")

  if (is.null(palette)) {
    return(list(defaults = list(), rules = list()))
  }

  # Resolve internal references in colors
  palette <- lapply(
    rlang::set_names(names(palette)),
    brand_color_pluck,
    brand = brand
  )

  defaults <- palette
  defaults <- lapply(defaults, paste, "!default")
  names(defaults) <- sprintf("brand-%s", names(defaults))

  for (color in intersect(names(palette), bootstrap_colors)) {
    defaults[color] <- sprintf("$brand-%s !default", color)
  }

  css_vars <- palette
  names(css_vars) <- sprintf("--brand-%s", names(css_vars))
  rules <- sprintf(":root { %s }", css(!!!css_vars))

  list(
    defaults = defaults,
    rules = rules
  )
}

bootstrap_colors <- c(
  "white",
  "black",
  "blue",
  "indigo",
  "purple",
  "pink",
  "red",
  "orange",
  "yellow",
  "green",
  "teal",
  "cyan"
)

brand_sass_color <- function(brand) {
  # Create brand Sass variables and set related Bootstrap Sass vars
  # brand.color.primary = "#007bff"
  # ==> $brand_color_primary: #007bff !default;
  # ==> $primary: $brand_color_primary !default;

  colors <- brand_pluck(brand, "color") %||% list()
  colors$palette <- NULL

  if (length(colors) == 0) {
    return(list())
  }

  # Resolve internal references in colors
  colors <- lapply(
    rlang::set_names(names(colors)),
    brand_color_pluck,
    brand = brand
  )

  defaults <- list()
  for (thm_name in names(colors)) {
    brand_color_var <- sprintf("brand_color_%s", thm_name)
    defaults[[brand_color_var]] <- paste(colors[[thm_name]], "!default")
  }

  list(defaults = defaults)
}

brand_sass_typography <- function(brand) {
  # Creates a dictionary of Sass variables for typography settings defined in
  # the `brand` object. These are used to set brand Sass variables in the format
  # `$brand_typography_{field}_{prop}`.
  typography <- brand_pluck(brand, "typography")

  if (is.null(typography)) {
    return(list(defaults = list()))
  }

  defaults <- list()

  for (field in names(typography)) {
    if (field == "fonts") {
      next
    }

    prop <- typography[[field]]
    for (prop_key in names(prop)) {
      prop_value <- prop[[prop_key]]
      if (field == "base" && prop_key == "size") {
        prop_value <- maybe_convert_font_size_to_rem(prop_value)
      } else if (prop_key %in% c("color", "background-color")) {
        prop_value <- brand_color_pluck(brand, prop_value)
      }
      field <- gsub("-", "_", field)
      prop_key <- gsub("-", "_", prop_key)
      typo_sass_var <- paste("brand_typography", field, prop_key, sep = "_")
      defaults[[typo_sass_var]] <- paste(prop_value, "!default")
    }
  }

  list(defaults = defaults)
}

brand_sass_fonts <- function(brand) {
  fonts <- brand_pluck(brand, "typography", "fonts")

  if (is.null(fonts)) {
    return(list(defaults = list(), rules = list()))
  }

  defaults <- list()
  rules <- list()

  for (font in fonts) {
    var_name <- sprintf(
      "brand-font-%s",
      gsub("[^a-z0-9-]+", "-", tolower(font$family))
    )

    font_obj <- switch(
      font$source %||% "google",
      google = sass::font_google(
        family = font$family,
        wght = brand_remap_font_weight(font$weight) %||%
          seq(100, 900, by = 100),
        ital = c("normal" = 0, "italic" = 1)[
          font$style %||% c("normal", "italic")
        ],
        display = font$display %||% "auto"
      ),
      bunny = brand_font_bunny(
        family = font$family,
        weight = font$weight,
        style = font$style,
        display = font$display
      ),
      file = brand_font_file(
        family = font$family,
        files = font$files,
        brand_root = dirname(brand$path)
      ),
      system = NULL,
      abort(sprintf("Unknown font source '%s'.", font$source))
    )

    if (!is.null(font_obj)) {
      defaults[[var_name]] <- font_obj
      rules <- c(
        rules,
        sprintf(".%s { font-family: $%s; }", var_name, var_name)
      )
    }
  }

  list(defaults = defaults, rules = rules)
}

brand_font_bunny <- function(
  family,
  weight = NULL,
  style = NULL,
  display = NULL
) {
  rlang::check_installed("utils")

  weight <- brand_remap_font_weight(weight) %||% seq(100, 900, 100)

  style <- style %||% c("normal", "italic")
  style <- rlang::arg_match(
    style,
    values = c("normal", "italic"),
    multiple = TRUE
  )

  display <- display %||% "auto"
  display <- rlang::arg_match(
    display,
    values = c("swap", "auto", "block", "fallback", "optional"),
    error_arg = "display"
  )

  if (!is.null(weight)) {
    stopifnot(is.character(weight) || is.numeric(weight))
    weight <- sort(weight)
  }

  weight_list <- as.character(weight)
  style_map <- c(normal = "", italic = "i")
  ital <- sort(style_map[style])

  values <- character(0)
  if (length(weight_list) > 0 && length(ital) > 0) {
    # 400,700,400i,700i
    values <- as.vector(outer(weight_list, ital, paste0))
  } else if (length(weight_list) > 0) {
    values <- weight_list
  } else if (length(ital) > 0) {
    values <- ifelse(ital == "", "regular", "italic")
  }

  family_values <- ""
  if (length(values) > 0) {
    family_values <- paste0(":", paste(values, collapse = ","))
  }

  params <- list(
    family = paste0(family, family_values),
    display = display
  )

  url_base <- "https://fonts.bunny.net/css"
  url_query <- paste0(
    names(params),
    "=",
    utils::URLencode(unlist(params)),
    collapse = "&"
  )

  url <- paste0(url_base, "?", url_query)

  font_link(family, url)
}

brand_font_file <- function(family, files, brand_root = getwd()) {
  if (!(is.list(files) && length(files) > 0)) {
    abort(
      c(
        sprintf(
          "Font family '%s' must have one or more associated files.",
          family
        ),
        "i" = "Use `source: system` for fonts that are provided by the user's system."
      )
    )
  }

  font_collection_files <- lapply(files, function(file) {
    if (is.null(file$path)) {
      abort(
        sprintf(
          "All font `files` for font family '%s' must have a `path`.",
          family
        )
      )
    }

    font_data_uri <- if (grepl("^https?://", file$path)) {
      font_path <- file$path
    } else {
      font_path <- file.path(brand_root, file$path)
      base64enc::dataURI(
        file = font_path,
        mime = mime::guess_type(font_path)
      )
    }
    font_type <- switch(
      path_ext(tolower(font_path)),
      # otc = "collection",
      # ttc = "collection",
      # eot = "embedded-opentype",
      otf = "opentype",
      ttf = "truetype",
      # svg = "svg",
      # svgz = "svg",
      woff = "woff",
      woff2 = "woff2",
      abort(
        c(
          sprintf("Invalid font type: %s", font_path),
          "i" = "Font must be `.ttf`, `.otf`, `.woff` or `.woff2`."
        )
      )
    )

    sass::font_face(
      family = family,
      src = sprintf("url(%s) format(%s)", font_data_uri, font_type),
      weight = brand_remap_font_weight(file$weight),
      style = file$style,
      display = "auto"
    )
  })

  sass::font_collection(!!!font_collection_files)
}

brand_remap_font_weight <- function(x) {
  if (is.null(x)) return()

  for (i in seq_along(x)) {
    if (x[[i]] %in% names(brand_font_weight_map)) {
      x[[i]] <- brand_font_weight_map[x[[i]]]
    }
  }

  x
}

brand_font_weight_map <- c(
  "thin" = 100,
  "extra-light" = 200,
  "ultra-light" = 200,
  "light" = 300,
  "normal" = 400,
  "regular" = 400,
  "medium" = 500,
  "semi-bold" = 600,
  "demi-bold" = 600,
  "bold" = 700,
  "extra-bold" = 800,
  "ultra-bold" = 800,
  "black" = 900
)

#' Convert a font size to rem
#'
#' Some frameworks, like Bootstrap expect base font size to be in `rem`. This
#' function converts `em`, `%`, `px`, `pt` to `rem`:
#'
#' 1. `em` is directly replace with `rem`.
#' 2. `1%` is `0.01rem`, e.g. `90%` becomes `0.9rem`.
#' 3. `16px` is `1rem`, e.g. `18px` becomes `1.125rem`.
#' 4. `12pt` is `1rem`.
#' 5. `0.1666in` is `1rem`.
#' 6. `4.234cm` is `1rem`.
#' 7. `42.3mm` is `1rem`.
#'
#' @noRd
maybe_convert_font_size_to_rem <- function(x) {
  x_og <- as.character(x)
  split_result <- split_css_value_and_unit(x)
  value <- split_result$value
  unit <- split_result$unit

  if (unit %in% c("rem", "em")) {
    return(paste0(value, "rem"))
  }

  scale <- list(
    "%" = 100,
    "px" = 16,
    "pt" = 12,
    "in" = 16 / 96, # 96 px/inch
    "cm" = 16 / 96 * 2.54, # inch -> cm
    "mm" = 16 / 96 * 25.4 # cm -> mm
  )

  if (unit %in% names(scale)) {
    return(paste0(as.numeric(value) / scale[[unit]], "rem"))
  }

  if (unit == "") {
    unit <- "unknown"
  }

  abort(
    sprintf(
      "Could not convert font size '%s' from %s units to a relative unit.",
      x_og,
      unit
    )
  )
}

split_css_value_and_unit <- function(x) {
  x <- trimws(x)
  pattern <- "^(-?[0-9]*\\.?[0-9]+)\\s*([a-z%]*)$"
  match <- regexec(pattern, x)
  result <- regmatches(x, match)[[1]]

  if (length(result) != 3) {
    abort(paste0("Invalid CSS value format: ", x))
  }

  return(list(value = result[2], unit = result[3]))
}

brand_validate_bootstrap_defaults <- function(
  defaults,
  source = "brand.defaults.bootstrap.defaults"
) {
  if (is.null(defaults)) {
    return(list())
  }

  if (!is.list(defaults)) {
    abort("Invalid brand defaults in `", source, "`, must be a list.")
  }

  if (length(defaults) == 0) {
    return(list())
  }

  if (!all(nzchar(names2(defaults)))) {
    abort("Invalid brand defaults in `", source, "`, all values must be named.")
  }

  is_scalar <- function(v) {
    if (is.null(v)) return(TRUE)
    rlang::is_scalar_character(v) ||
      rlang::is_scalar_logical(v) ||
      rlang::is_scalar_double(v) ||
      rlang::is_scalar_integerish(v)
  }

  good <- vapply(defaults, is_scalar, logical(1))

  if (!all(good)) {
    abort(
      sprintf(
        "Invalid brand defaults in `%s`, all values must be scalar: %s",
        source,
        defaults[!good][1]
      )
    )
  }

  return(defaults)
}

brand_sass_defaults_bootstrap <- function(brand) {
  bootstrap <- brand_pluck(brand, "defaults", "bootstrap")
  shiny <- brand_pluck(brand, "defaults", "shiny", "theme")

  if (is.null(bootstrap) && is.null(shiny))
    return(
      list(
        defaults = list(),
        layer = list()
      )
    )

  shiny <- shiny %||% list()
  shiny_defaults <- brand_validate_bootstrap_defaults(
    shiny$defaults,
    "brand.defaults.shiny.theme"
  )

  bootstrap <- bootstrap %||% list()
  bootstrap_defaults <- brand_validate_bootstrap_defaults(bootstrap$defaults)

  defaults <- list2(!!!bootstrap_defaults, !!!shiny_defaults)
  defaults <- lapply(defaults, function(x) {
    if (is.null(x)) {
      x <- "null"
    } else if (is.logical(x)) {
      x <- tolower(as.character(x))
    }
    paste(x, "!default")
  })

  list(
    defaults = defaults,
    layer = sass_layer(
      functions = c(bootstrap$functions, shiny$functions),
      mixins = c(bootstrap$mixins, shiny$mixins),
      rules = c(bootstrap$rules, shiny$rules)
    )
  )
}

# Read Brand -------------------------------------------------------------------

read_brand_yml <- function(path = NULL) {
  path <- find_project_brand_yml(path)

  rlang::check_installed("yaml")
  brand <- yaml::read_yaml(path, readLines.warn = FALSE)

  brand <- as_brand_yml(brand)
  brand$path <- path

  brand
}

as_brand_yml <- function(brand = list()) {
  stopifnot(is.list(brand))

  # Normalize brand internals !! MINIMAL VALIDATION !!
  brand <- brand_meta_normalize(brand)
  brand <- brand_color_normalize(brand)
  brand <- brand_typography_normalize(brand)

  class(brand) <- "brand_yml"
  brand
}

brand_meta_normalize <- function(brand) {
  if (!brand_has(brand, "meta")) {
    return(brand)
  }

  if (brand_has_string(brand, "meta", "name")) {
    name <- brand[["meta"]][["name"]]
    brand[["meta"]][["name"]] <- list(short = name, full = name)
  }

  if (brand_has_string(brand, "meta", "link")) {
    brand[["meta"]][["link"]] <- list(
      home = brand[["meta"]][["link"]]
    )
  }

  brand
}

brand_color_normalize <- function(brand) {
  if (!brand_has(brand, "color")) {
    return(brand)
  }

  # Pull out colors and resolve each color from original brand
  theme <- brand_pluck(brand, "color")

  for (field in names(brand_pluck(brand, "color"))) {
    if (field == "palette") {
      theme[[field]] <- lapply(
        rlang::set_names(names(theme[[field]])),
        brand_color_pluck,
        brand = brand
      )
    } else {
      theme[[field]] <- brand_color_pluck(brand, field)
    }
  }

  # Then replace brand.color with resolved colors
  brand[["color"]] <- theme
  brand
}

brand_typography_normalize <- function(brand) {
  if (!brand_has(brand, "typography")) {
    return(brand)
  }

  expand_family <- c(
    "base",
    "headings",
    "monospace",
    "monospace-inline",
    "monospace-block"
  )

  for (field in expand_family) {
    if (brand_has_string(brand, "typography", field)) {
      brand[["typography"]][[field]] <- list(
        family = brand[["typography"]][[field]]
      )
    }
  }

  brand
}

brand_color_pluck <- function(brand, key) {
  if (!brand_has(brand, "color")) {
    return(key)
  }

  theme_colors <- brand[["color"]]
  theme_colors$palette <- NULL
  palette <- brand[["color"]][["palette"]] %||% list()

  key_og <- key
  visited <- c()

  cycle <- function(key) {
    path <- c(visited, key)
    if (length(path) > 10) {
      path <- c(path[1:2], "...", path[-(1:(length(path) - 2))])
    }
    paste(path, collapse = " -> ")
  }

  assert_no_cycles <- function(key) {
    if (key %in% visited) {
      abort(
        c(
          sprintf(
            "Cyclic references detected in `brand.color` for color '%s'.",
            key_og
          ),
          "i" = cycle(key)
        )
      )
    }
    visited <<- c(visited, key)
  }

  check_string_or_null <- function(key, value) {
    if (is.null(value)) return()
    if (rlang::is_string(value)) return(value)

    abort(sprintf("`brand.color.%s` must be a string or `NULL`.", key))
  }

  p_key <- function(key) paste0("palette.", key)
  value <- ""
  i <- 0
  while (!identical(value, key)) {
    if (is.null(key) || is.null(value)) return()

    i <- i + 1
    if (i > 100) {
      abort(
        c(
          sprintf(
            "Max recursion limit reached while trying to resolve color '%s' using `brand.color`.",
            key_og
          ),
          i = cycle(key)
        )
      )
    }

    in_theme <- key %in% names(theme_colors)
    in_theme_unseen <- in_theme && !key %in% visited
    in_pal <- key %in% names(palette)

    if (in_pal && !in_theme_unseen) {
      # Prioritize palette if theme was already visited
      assert_no_cycles(p_key(key))
      key <- check_string_or_null(p_key(key), palette[[key]])
    } else if (in_theme) {
      assert_no_cycles(key)
      key <- check_string_or_null(key, theme_colors[[key]])
    } else {
      value <- key
    }
  }

  return(value)
}

# Brand utilities --------------------------------------------------------------

brand_has <- function(brand, ...) {
  x <- brand

  for (f in c(...)) {
    if (is.null(x[[f]])) return(FALSE)
    x <- x[[f]]
  }

  TRUE
}

brand_pluck <- function(brand, ...) {
  if (!brand_has(brand, ...)) {
    return(NULL)
  }

  brand[[c(...)]]
}

brand_has_string <- function(brand, ...) {
  if (!brand_has(brand, ...)) return(FALSE)
  rlang::is_string(brand[[c(...)]])
}

brand_has_list <- function(brand, ...) {
  if (!brand_has(brand, ...)) return(FALSE)
  rlang::is_list(brand[[c(...)]])
}

# Find _brand.yml --------------------------------------------------------------

find_project_brand_yml <- function(path = NULL) {
  path <- path %||% getwd()
  path <- normalizePath(path, mustWork = FALSE)

  ext <- if (dir.exists(path)) "" else path_ext(path)
  if (ext %in% c("yml", "yaml")) {
    return(path)
  }

  if (nzchar(ext)) {
    path <- dirname(path)
  }

  find_project_file(
    filename = c("_brand.yml", "_brand.yaml"),
    dir = path,
    subdir = c("brand", "_brand")
  )
}

find_project_file <- function(filename, dir, subdir = character()) {
  dir_og <- dir
  max_parents <- 20

  while (dir != dirname(dir) && max_parents > 0) {
    for (fname in filename) {
      file_path <- file.path(dir, fname)
      if (path_is_file(file_path)) {
        return(file_path)
      }
    }

    for (sub in subdir) {
      for (fname in filename) {
        file_path <- file.path(dir, sub, fname)
        if (path_is_file(file_path)) {
          return(file_path)
        }
      }
    }

    dir <- dirname(dir)
    max_parents <- max_parents - 1
  }

  abort(
    sprintf(
      "Could not find %s in %s or its parents.",
      paste(filename, collapse = ", "),
      dir_og
    )
  )
}

path_is_file <- function(path) {
  # The file exists and is a file
  file.exists(path) && !dir.exists(path)
}

path_ext <- function(path) {
  # Same as tools::file_ext()
  pos <- regexpr("\\.([[:alnum:]]+)$", path)
  ifelse(pos > -1L, substring(path, pos + 1L), "")
}

# testthat::test_that()

describe("as_brand_yml()", {
  it("normalizes colors", {
    brand <- list(
      color = list(
        palette = list(red = "#FF1122"),
        primary = "red",
        secondary = "berry"
      )
    )

    brand <- as_brand_yml(brand)
    expect_s3_class(brand, "brand_yml")
    expect_equal(brand$color$palette$red, brand$color$primary)
    expect_equal(brand$color$secondary, "berry")
  })

  it("normalizes font family choices", {
    brand <- list(
      typography = list(
        base = "Times New Roman",
        headings = "Helvetica",
        monospace = "Courier New",
        "monospace-inline" = "Fira Code"
      )
    )

    brand <- as_brand_yml(brand)
    expect_s3_class(brand, "brand_yml")
    expect_equal(brand$typography$base$family, "Times New Roman")
    expect_equal(brand$typography$headings$family, "Helvetica")
    expect_equal(brand$typography[["monospace"]]$family, "Courier New")
    expect_equal(brand$typography[["monospace-inline"]]$family, "Fira Code")
  })
})

describe("brand_resolve()", {
  withr::local_dir(withr::local_tempdir())

  it("doesn't error if `brand = NULL` and no _brand.yml is found", {
    expect_null(brand_resolve(NULL))
  })

  it("throws if `brand = TRUE` and no `_brand.yml` is found", {
    expect_error(brand_resolve(TRUE))
  })

  writeLines(
    c("meta:", "  name: test-brand-yml"),
    "_brand.yml"
  )

  direct_is_valid <- FALSE

  it("returns NULL if `brand = FALSE`", {
    expect_null(brand_resolve(FALSE))
  })

  it("finds _brand.yml or reads from brand path", {
    path_with_parent_dir <- function(x) {
      file.path(basename(dirname(x)), basename(x))
    }

    # brand=TRUE and brand=NULL are the same in presence of _brand.yml file
    expect_equal(brand_resolve(TRUE), brand_resolve(NULL))

    brand_found <- brand_resolve(TRUE)
    brand_found$path <- path_with_parent_dir(brand_found$path)
    brand_direct <- brand_resolve("_brand.yml")
    brand_direct$path <- path_with_parent_dir(brand_direct$path)

    expect_equal(brand_found, brand_direct)
    expect_s3_class(brand_found, "brand_yml")
    expect_equal(
      brand_found$path,
      path_with_parent_dir(file.path(getwd(), "_brand.yml"))
    )
    direct_is_valid <<- TRUE
  })

  it("takes a list or a brand_yml object", {
    expect_true(direct_is_valid)

    brand_list <- brand_resolve(list(meta = list(name = "test-brand-yml")))
    brand_direct <- brand_resolve("_brand.yml")
    brand_direct$path <- NULL # brand is equal other than via path
    brand_obj <- brand_resolve(brand_direct)

    # brand_direct validated above
    expect_equal(brand_list, brand_direct)
    expect_equal(brand_obj, brand_direct)
  })

  it("uses brand.defaults.shiny.theme.preset", {
    brand <- as_brand_yml(
      list(
        meta = list(name = "test-brand-yml"),
        defaults = list(
          shiny = list(theme = list(preset = "flatly", version = "4"))
        )
      )
    )

    expected_base_preset <- resolve_bs_preset("flatly", version = "4")
    brand_preset <- brand_resolve(brand)
    preset <- brand_resolve_preset(brand)
    expect_equal(preset, expected_base_preset)
    expect_equal(preset$version, "4")
    expect_equal(preset$name, "flatly")
    expect_equal(preset$version, "4")
  })

  it("throws if `brand.defaults.shiny.theme.preset: brand`", {
    brand <- as_brand_yml(
      list(
        meta = list(name = "test-brand-yml"),
        defaults = list(
          shiny = list(theme = list(preset = "brand", version = "4"))
        )
      )
    )

    expect_error(
      brand_resolve_preset(brand),
      "is not a known preset"
    )
  })

  it("uses brand.defaults.shiny.theme.version before brand.defaults.bootstrap.version", {
    brand <- as_brand_yml(
      list(
        meta = list(name = "test-brand-yml"),
        defaults = list(
          bootstrap = list(version = 3),
          shiny = list(theme = list(preset = "flatly", version = "4"))
        )
      )
    )

    expected_base_preset <- resolve_bs_preset("flatly", version = "4")
    brand <- brand_resolve(brand)
    preset <- brand_resolve_preset(brand)
    expect_equal(preset, expected_base_preset)
    expect_equal(preset$version, "4")
    expect_equal(preset$name, "flatly")
    expect_equal(preset$version, "4")
  })

  it("uses brand.defaults.bootstrap.version", {
    brand <- as_brand_yml(
      list(
        meta = list(name = "test-brand-yml"),
        defaults = list(
          bootstrap = list(version = 4)
        )
      )
    )

    expected_base_preset <- resolve_bs_preset("bootstrap", version = "4")
    brand <- brand_resolve(brand)
    preset <- brand_resolve_preset(brand)
    expect_equal(preset, expected_base_preset)
    expect_equal(preset$version, "4")
    expect_equal(preset$name, "bootstrap")
    expect_equal(preset$version, "4")
  })
})

describe("brand_resolve_preset()", {
  withr::local_dir(withr::local_tempdir())

  write_brand_yml_preset <- function(preset = NULL, version = NULL) {
    b <- list(
      defaults = list(
        shiny = list(
          theme = list(
            preset = preset,
            version = version
          )
        )
      )
    )
    b <- dropNulls(b)
    yaml::write_yaml(b, "_brand.yml")
  }

  write_brand_yml_preset(preset = "flatly")

  it("uses `preset` and `version` from `_brand.yml`", {
    preset <- brand_resolve_preset(brand_resolve(NULL))
    expect_equal(preset$name, "flatly")
    expect_equal(preset$version, version_default())
  })

  it("uses `preset` and `version` from `bs_theme()`", {
    preset <- brand_resolve_preset(NULL, preset = "cosmo", 5)
    expect_equal(preset$name, "cosmo")
    expect_equal(preset$version, version_default())
  })

  it("uses `shiny` by default for BS 5", {
    write_brand_yml_preset(version = "5")
    preset <- brand_resolve_preset(brand_resolve(NULL))
    expect_equal(preset$name, "shiny")
    expect_equal(preset$version, "5")
  })

  it("uses `bootstrap` by default for BS 4", {
    write_brand_yml_preset(version = "4")
    preset <- brand_resolve_preset(brand_resolve(NULL))
    expect_equal(preset$name, "bootstrap")
    expect_equal(preset$version, "4")
  })
})

describe("brand_color_pluck()", {
  it("detects cyclic references in brand.color.palette", {
    brand <- list(
      color = list(
        palette = list(red = "blue", blue = "red")
      )
    )

    expect_error(
      brand_color_pluck(brand, "red"),
      "palette.red -> palette.blue -> palette.red"
    )

    expect_error(
      brand_color_pluck(brand, "blue"),
      "palette.blue -> palette.red -> palette.blue"
    )
  })

  it("detects cyclic references in brand.color", {
    brand <- list(
      color = list(
        primary = "secondary",
        secondary = "primary"
      )
    )

    expect_error(
      brand_color_pluck(brand, "primary"),
      "primary -> secondary -> primary"
    )

    expect_error(
      brand_color_pluck(brand, "secondary"),
      "secondary -> primary -> secondary"
    )
  })

  it("detects cyclic references in brand.color and brand.color.palette", {
    brand1 <- list(
      color = list(
        palette = list(
          primary = "secondary",
          secondary = "resolved" # cycles before reaches here
        ),
        primary = "primary",
        secondary = "primary" # bad
      )
    )

    expect_error(
      brand_color_pluck(brand1, "primary"),
      "primary -> palette.primary -> secondary -> palette.primary"
    )

    brand2 <- list(
      color = list(
        palette = list(red = "primary"),
        primary = "red"
      )
    )

    expect_error(
      brand_color_pluck(brand2, "red"),
      "palette.red -> primary -> palette.red"
    )

    expect_error(
      brand_color_pluck(brand2, "primary"),
      "primary -> palette.red -> primary"
    )
  })

  it("avoids high levels of recursion", {
    max_recursion <- 101
    seq_max <- 1:max_recursion
    color_ref <- function(i) sprintf("color%s", i)

    brand <- list(
      color = list(
        palette = lapply(
          rlang::set_names(seq_max, color_ref(seq_max - 1)),
          color_ref
        )
      )
    )

    expect_error(
      brand_color_pluck(brand, color_ref(0)),
      "recursion limit"
    )
  })

  it("returns `key` if `brand.color` isn't present", {
    brand <- list(meta = list(name = "no color"))
    expect_equal(brand_color_pluck(brand, "red"), "red")
  })

  it("returns `NULL` if the color is preset but `NULL`", {
    brand <- list(color = list(secondary = NULL, palette = list(black = NULL)))
    expect_null(brand_color_pluck(brand, "secondary"))
    expect_null(brand_color_pluck(brand, "black"))
  })

  it("errors if the color value is not a string", {
    brand <- list(
      color = list(
        secondary = 123456,
        palette = list(black = 123456)
      )
    )

    expect_error(brand_color_pluck(brand, "secondary"), "brand.color.secondary")
    expect_error(brand_color_pluck(brand, "black"), "brand.color.palette.black")
  })
})

describe("maybe_convert_font_size_to_rem()", {
  it("returns `rem` directly", {
    expect_equal(maybe_convert_font_size_to_rem("1rem"), "1rem")
    expect_equal(maybe_convert_font_size_to_rem("1.123rem"), "1.123rem")
    expect_equal(maybe_convert_font_size_to_rem("1.123 rem"), "1.123rem")
  })

  it("returns `em` as 1:1 with `rem`", {
    expect_equal(maybe_convert_font_size_to_rem("1em"), "1rem")
    expect_equal(maybe_convert_font_size_to_rem("1.123em"), "1.123rem")
    expect_equal(maybe_convert_font_size_to_rem("1.123 em"), "1.123rem")
  })

  it("converts `%` as 100%:1rem", {
    expect_equal(maybe_convert_font_size_to_rem("100%"), "1rem")
    expect_equal(maybe_convert_font_size_to_rem("225%"), "2.25rem")
    expect_equal(maybe_convert_font_size_to_rem("50 %"), "0.5rem")
  })

  it("converts `in`, `cm` and `mm` to `rem`", {
    expect_equal(maybe_convert_font_size_to_rem("1in"), "6rem")
    expect_equal(maybe_convert_font_size_to_rem("0.5in"), "3rem")

    expect_equal(maybe_convert_font_size_to_rem("2.54cm"), "6rem")
    expect_equal(maybe_convert_font_size_to_rem("1.27cm"), "3rem")

    expect_equal(maybe_convert_font_size_to_rem("25.4mm"), "6rem")
    expect_equal(maybe_convert_font_size_to_rem("12.7mm"), "3rem")
  })

  it("throws for unsupported units", {
    expect_error(
      maybe_convert_font_size_to_rem("1 foo")
    )
    expect_error(
      maybe_convert_font_size_to_rem("1 foo bar")
    )
    expect_error(
      maybe_convert_font_size_to_rem("1vw")
    )
    expect_error(
      maybe_convert_font_size_to_rem("123")
    )
  })
})

describe("bs_brand_bundle()", {
  it("warns with Bootstrap <= 4", {
    expect_warning(
      bs_brand_bundle(list(), "3")
    )

    expect_warning(
      bs_brand_bundle(list(), "4")
    )

    expect_silent(
      bs_brand_bundle(list(), "5")
    )
  })
})

# testthat::test_that()

describe("as_brand_yml()", {
  skip_if_not_installed("brand.yml")

  it("normalizes colors", {
    brand <- list(
      color = list(
        palette = list(red = "#FF1122"),
        primary = "red",
        secondary = "berry"
      )
    )

    brand <- brand.yml::as_brand_yml(brand)
    expect_s3_class(brand, "brand_yml")
    expect_equal(brand$color$palette$red, brand$color$primary)
    expect_equal(brand$color$secondary, "berry")
  })

  it("normalizes font family choices", {
    brand <- list(
      typography = list(
        base = "Times New Roman",
        headings = "Helvetica",
        monospace = "Courier New",
        "monospace-inline" = "Fira Code"
      )
    )

    brand <- brand.yml::as_brand_yml(brand)
    expect_s3_class(brand, "brand_yml")
    expect_equal(brand$typography$base$family, "Times New Roman")
    expect_equal(brand$typography$headings$family, "Helvetica")
    expect_equal(brand$typography[["monospace"]]$family, "Courier New")
    expect_equal(brand$typography[["monospace_inline"]]$family, "Fira Code")
  })
})

describe("brand_resolve()", {
  withr::local_dir(withr::local_tempdir())

  it("returns `NULL` if brand.yml is not installed", {
    is_installed_og <- is_installed

    local_mocked_bindings(
      is_installed = function(pkg) {
        if (pkg == "brand.yml") FALSE else is_installed_og(pkg)
      }
    )

    expect_null(brand_resolve(NULL))
  })

  it("returns NULL if `brand = FALSE`", {
    expect_null(brand_resolve(FALSE))
  })

  skip_if_not_installed("brand.yml")

  it("doesn't error if `brand = NULL` and no _brand.yml is found", {
    expect_null(brand_resolve(NULL))
  })

  it("throws if `brand = TRUE` and no `_brand.yml` is found", {
    expect_error(brand_resolve(TRUE))
  })

  writeLines(
    c("meta:", "  name: test-brand-yml"),
    "_brand.yml"
  )

  direct_is_valid <- FALSE

  it("finds _brand.yml or reads from brand path", {
    path_with_parent_dir <- function(x) {
      file.path(basename(dirname(x)), basename(x))
    }

    # brand=TRUE and brand=NULL are the same in presence of _brand.yml file
    expect_equal(brand_resolve(TRUE), brand_resolve(NULL))

    brand_found <- brand_resolve(TRUE)
    brand_found$path <- path_with_parent_dir(brand_found$path)
    brand_direct <- brand_resolve("_brand.yml")
    brand_direct$path <- path_with_parent_dir(brand_direct$path)

    expect_equal(brand_found, brand_direct)
    expect_s3_class(brand_found, "brand_yml")
    expect_equal(
      brand_found$path,
      path_with_parent_dir(file.path(getwd(), "_brand.yml"))
    )
    direct_is_valid <<- TRUE
  })

  it("takes a list or a brand_yml object", {
    expect_true(direct_is_valid)

    brand_list <- brand_resolve(list(meta = list(name = "test-brand-yml")))
    brand_direct <- brand_resolve("_brand.yml")
    brand_direct$path <- NULL # brand is equal other than via path
    brand_obj <- brand_resolve(brand_direct)

    # brand_direct validated above
    expect_equal(brand_list, brand_direct)
    expect_equal(brand_obj, brand_direct)
  })

  it("uses brand.defaults.shiny.theme.preset", {
    brand <- brand.yml::as_brand_yml(
      list(
        meta = list(name = "test-brand-yml"),
        defaults = list(
          shiny = list(theme = list(preset = "flatly", version = "4"))
        )
      )
    )

    expected_base_preset <- resolve_bs_preset("flatly", version = "4")
    brand_preset <- brand_resolve(brand)
    preset <- brand_resolve_preset(brand)
    expect_equal(preset, expected_base_preset)
    expect_equal(preset$version, "4")
    expect_equal(preset$name, "flatly")
    expect_equal(preset$version, "4")
  })

  it("throws if `brand.defaults.shiny.theme.preset: brand`", {
    brand <- brand.yml::as_brand_yml(
      list(
        meta = list(name = "test-brand-yml"),
        defaults = list(
          shiny = list(theme = list(preset = "brand", version = "4"))
        )
      )
    )

    expect_error(
      brand_resolve_preset(brand),
      "is not a known preset"
    )
  })

  it("uses brand.defaults.shiny.theme.version before brand.defaults.bootstrap.version", {
    brand <- brand.yml::as_brand_yml(
      list(
        meta = list(name = "test-brand-yml"),
        defaults = list(
          bootstrap = list(version = 3),
          shiny = list(theme = list(preset = "flatly", version = "4"))
        )
      )
    )

    expected_base_preset <- resolve_bs_preset("flatly", version = "4")
    brand <- brand_resolve(brand)
    preset <- brand_resolve_preset(brand)
    expect_equal(preset, expected_base_preset)
    expect_equal(preset$version, "4")
    expect_equal(preset$name, "flatly")
    expect_equal(preset$version, "4")
  })

  it("uses brand.defaults.bootstrap.version", {
    brand <- brand.yml::as_brand_yml(
      list(
        meta = list(name = "test-brand-yml"),
        defaults = list(
          bootstrap = list(version = 4)
        )
      )
    )

    expected_base_preset <- resolve_bs_preset("bootstrap", version = "4")
    brand <- brand_resolve(brand)
    preset <- brand_resolve_preset(brand)
    expect_equal(preset, expected_base_preset)
    expect_equal(preset$version, "4")
    expect_equal(preset$name, "bootstrap")
    expect_equal(preset$version, "4")
  })
})

describe("brand_resolve_preset()", {
  skip_if_not_installed("brand.yml")
  withr::local_dir(withr::local_tempdir())

  write_brand_yml_preset <- function(preset = NULL, version = NULL) {
    b <- list(
      defaults = list(
        shiny = list(
          theme = list(
            preset = preset,
            version = version
          )
        )
      )
    )
    b <- dropNulls(b)
    yaml::write_yaml(b, "_brand.yml")
  }

  write_brand_yml_preset(preset = "flatly")

  it("uses `preset` and `version` from `_brand.yml`", {
    preset <- brand_resolve_preset(brand_resolve(NULL))
    expect_equal(preset$name, "flatly")
    expect_equal(preset$version, version_default())
  })

  it("uses `preset` and `version` from `bs_theme()`", {
    preset <- brand_resolve_preset(NULL, preset = "cosmo", 5)
    expect_equal(preset$name, "cosmo")
    expect_equal(preset$version, version_default())
  })

  it("uses `shiny` by default for BS 5", {
    write_brand_yml_preset(version = "5")
    preset <- brand_resolve_preset(brand_resolve(NULL))
    expect_equal(preset$name, "shiny")
    expect_equal(preset$version, "5")
  })

  it("uses `bootstrap` by default for BS 4", {
    write_brand_yml_preset(version = "4")
    preset <- brand_resolve_preset(brand_resolve(NULL))
    expect_equal(preset$name, "bootstrap")
    expect_equal(preset$version, "4")
  })
})

describe("brand_color_pluck()", {
  skip_if_not_installed("brand.yml")

  it("detects cyclic references in brand.color.palette", {
    brand <- list(
      color = list(
        palette = list(red = "blue", blue = "red")
      )
    )

    expect_error(
      brand.yml::brand_color_pluck(brand, "red"),
      "palette.red -> palette.blue -> palette.red"
    )

    expect_error(
      brand.yml::brand_color_pluck(brand, "blue"),
      "palette.blue -> palette.red -> palette.blue"
    )
  })

  it("detects cyclic references in brand.color", {
    brand <- list(
      color = list(
        primary = "secondary",
        secondary = "primary"
      )
    )

    expect_error(
      brand.yml::brand_color_pluck(brand, "primary"),
      "primary -> secondary -> primary"
    )

    expect_error(
      brand.yml::brand_color_pluck(brand, "secondary"),
      "secondary -> primary -> secondary"
    )
  })

  it("detects cyclic references in brand.color and brand.color.palette", {
    brand1 <- list(
      color = list(
        palette = list(
          primary = "secondary",
          secondary = "resolved" # cycles before reaches here
        ),
        primary = "primary",
        secondary = "primary" # bad
      )
    )

    expect_error(
      brand.yml::brand_color_pluck(brand1, "primary"),
      "primary -> palette.primary -> secondary -> palette.primary"
    )

    brand2 <- list(
      color = list(
        palette = list(red = "primary"),
        primary = "red"
      )
    )

    expect_error(
      brand.yml::brand_color_pluck(brand2, "red"),
      "palette.red -> primary -> palette.red"
    )

    expect_error(
      brand.yml::brand_color_pluck(brand2, "primary"),
      "primary -> palette.red -> primary"
    )
  })

  it("avoids high levels of recursion", {
    max_recursion <- 101
    seq_max <- 1:max_recursion
    color_ref <- function(i) sprintf("color%s", i)

    brand <- list(
      color = list(
        palette = lapply(
          rlang::set_names(seq_max, color_ref(seq_max - 1)),
          color_ref
        )
      )
    )

    expect_error(
      brand.yml::brand_color_pluck(brand, color_ref(0)),
      "recursion limit"
    )
  })

  it("returns `key` if `brand.color` isn't present", {
    brand <- list(meta = list(name = "no color"))
    expect_equal(brand.yml::brand_color_pluck(brand, "red"), "red")
  })

  it("returns `NULL` if the color is preset but `NULL`", {
    brand <- list(color = list(secondary = NULL, palette = list(black = NULL)))
    expect_null(brand.yml::brand_color_pluck(brand, "secondary"))
    expect_null(brand.yml::brand_color_pluck(brand, "black"))
  })

  it("errors if the color value is not a string", {
    brand <- list(
      color = list(
        secondary = 123456,
        palette = list(black = 123456)
      )
    )

    expect_error(
      brand.yml::brand_color_pluck(brand, "secondary"),
      "brand.color.secondary"
    )
    expect_error(
      brand.yml::brand_color_pluck(brand, "black"),
      "brand.color.palette.black"
    )
  })
})

describe("bs_brand_bundle()", {
  skip_if_not_installed("brand.yml")

  it("warns with Bootstrap <= 4", {
    expect_warning(
      bs_brand_bundle(list(), "3")
    )

    expect_warning(
      bs_brand_bundle(list(), "4")
    )

    expect_silent(
      bs_brand_bundle(list(), "5")
    )
  })

  it("doesn't re-resolve brand", {
    bs_theme_base <- bs_theme()

    withr::local_dir(withr::local_tempdir())

    writeLines(
      c("color:", "  primary: '#FF1122'"),
      "_brand.yml"
    )

    expect_null(bs_brand_bundle(NULL))
    expect_true(
      bs_get_variables(bs_theme(brand = FALSE), "primary") != "#FF1122"
    )
    expect_equal(bs_theme(brand = FALSE), bs_theme_base)
  })
})

#' Create a textarea input control with explicit submission
#'
#' Creates a textarea input where users can enter multi-line text and submit
#' their input using a dedicated button or keyboard shortcut. This control is
#' ideal when you want to capture finalized input, rather than reacting to every
#' keystroke, making it useful for chat boxes, comments, or other scenarios
#' where users may compose and review their text before submitting.
#'
#' @section Server value:
#' The server receives a character string containing the user's text input.
#'
#' **Important:** The initial server value is always `""` (empty string),
#' regardless of any `value` parameter provided to `input_submit_textarea()`.
#' The server value updates only when the user explicitly submits the input by
#' either pressing the Enter key (possibly with a modifier key) or clicking the
#' submit button.
#'
#' @param id The input ID.
#' @param label The label to display above the input control. If `NULL`, no
#'   label is displayed.
#' @param ... Additional attributes to apply to the underlying `<textarea>`
#'   element (e.g., spellcheck, autocomplete, etc).
#' @param placeholder The placeholder text to display when the input is empty.
#'   This can be used to provide a hint or example of the expected input.
#' @param value The initial input text. Note that, unlike [shiny::textAreaInput()],
#'   this won't set a server-side value until the value is explicitly submitted.
#' @param width Any valid CSS unit (e.g., `width="100%"`).
#' @param rows The number of rows (i.e., height) of the textarea. This essentially
#'   sets the minimum height -- the textarea can grow taller as the user
#'   enters more text.
#' @param button A [htmltools::tags] element to use for the submit button. It's recommended
#'   that this be a [input_task_button()] since it will automatically provide a
#'   busy indicator (and disable) until the next flush occurs. Note also that if
#'   the submit button launches a [shiny::ExtendedTask], this button can also be bound
#'   to the task ([bind_task_button()]) and/or manually updated for more
#'   accurate progress reporting ([update_task_button()]).
#' @param toolbar A list of optional UI elements (e.g., links, icons) to
#'  display next to the submit button.
#' @param submit_key A character string indicating what keyboard event should
#'   trigger the submit button. The default is `enter+modifier`, which requires
#'   the user to hold down Ctrl (or Cmd on Mac) before pressing Enter to
#'   submit. This helps prevent accidental submissions. To allow submission with
#'   just the Enter key, use `enter`. In this case, the user can still insert
#'   new lines using Shift+Enter or Alt+Enter.
#'
#' @return A textarea input control that can be added to a UI definition.
#' @seealso [update_submit_textarea()], [input_task_button()]
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   input_submit_textarea("text", placeholder = "Enter some input..."),
#'   verbatimTextOutput("value")
#' )
#' server <- function(input, output) {
#'   output$value <- renderText({
#'     req(input$text)
#'     Sys.sleep(2)
#'     paste("You entered:", input$text)
#'   })
#' }
#' shinyApp(ui, server)
#'
#' @export
input_submit_textarea <- function(
  id,
  label = NULL,
  ...,
  placeholder = NULL,
  value = "",
  width = "min(680px, 100%)",
  rows = 1,
  button = NULL,
  toolbar = NULL,
  submit_key = c("enter+modifier", "enter")
) {
  rlang::check_installed("shiny", version = "1.11.1")
  args <- rlang::list2(...)
  if (any_unnamed(args)) {
    abort(c(
      "All `...` arguments must be named",
      "i" = "Did you mean to pass UI elements to `toolbar`?"
    ))
  }

  value <- shiny::restoreInput(id = id, default = value)
  if (length(value) != 1 || !is.character(value)) {
    abort("`value` must be a character string")
  }

  submit_key <- rlang::arg_match(submit_key)
  needs_modifier <- isTRUE(submit_key == "enter+modifier")

  if (is.null(button)) {
    button <- input_task_button(
      id = paste0(id, "_submit"),
      class = "btn-sm",
      label = span(class = "bslib-submit-key", "\U23CE"),
      icon = "Submit",
      label_busy = div(
        class = "spinner-border spinner-border-sm ms-2",
        role = "status",
        span(class = "visually-hidden", "Processing...")
      ),
      icon_busy = "Submit",
      title = "Press Enter to Submit",
      `aria-label` = "Press Enter to Submit"
    )
  }

  if (!is_button_tag(button)) {
    abort("`button` must be a `tags$button()`")
  }

  button <- tagAppendAttributes(button, class = "bslib-submit-textarea-btn")

  div(
    class = "bslib-input-submit-textarea shiny-input-container bslib-mb-spacing",
    style = css(
      # TODO: validateCssUnit() needs to handle more complex CSS
      width = if (is.numeric(width)) paste0(width, "px") else width,
    ),
    shiny_input_label(id, label),
    div(
      class = "bslib-submit-textarea-container",
      tags$textarea(
        id = id,
        class = "form-control",
        style = css(width = if (!is.null(width)) "100%"),
        placeholder = placeholder,
        `data-needs-modifier` = if (needs_modifier) "",
        rows = rows,
        !!!args,
        value
      ),
      tags$footer(
        div(toolbar, class = "bslib-toolbar"),
        button
      )
    )
  )
}

is_button_tag <- function(x) {
  if (!inherits(x, "shiny.tag")) {
    return(FALSE)
  }

  isTRUE(x$name == "button") ||
    isTRUE(x$attribs$type == "button")
}

#' @param value The value to set the user input to.
#' @param placeholder The placeholder text for the user input.
#' @param submit Whether to automatically submit the text for the user. Requires `value`.
#' @param focus Whether to move focus to the input element. Requires `value`.
#' @param session The `session` object; using the default is recommended.
#'
#' @rdname input_submit_textarea
#' @export
update_submit_textarea <- function(
  id,
  ...,
  value = NULL,
  placeholder = NULL,
  label = NULL,
  submit = FALSE,
  focus = FALSE,
  session = get_current_session()
) {
  rlang::check_dots_empty()

  if (is.null(value) && (submit || focus)) {
    stop(
      "An input `value` must be provided when `submit` or `focus` are `TRUE`.",
      call. = FALSE
    )
  }

  message <- dropNulls(list(
    value = value,
    placeholder = placeholder,
    label = if (!is.null(label)) processDeps(label, session),
    submit = submit,
    focus = focus
  ))

  session$sendInputMessage(id, message)
}

test_that("input_submit_textarea() markup snapshots", {
  expect_snapshot(input_submit_textarea("test"))

  expect_snapshot(
    input_submit_textarea(
      "test",
      label = "Enter text",
      placeholder = "Type here...",
      value = "Initial value",
      width = 300,
      rows = 3,
      button = tags$button(
        id = "custom_submit",
        class = "btn btn-primary",
        "Send"
      ),
      toolbar = tagList(
        tags$span("Press"),
        tags$kbd("Enter"),
        tags$span("to submit")
      ),
      submit_key = "enter",
      spellcheck = "false",
      autocomplete = "off",
    )
  )
})

test_that("input_submit_textarea() validation errors", {
  expect_snapshot(
    error = TRUE,
    input_submit_textarea("test", "With Children", div())
  )

  expect_snapshot(
    error = TRUE,
    input_submit_textarea("test", value = 123)
  )

  expect_snapshot(
    error = TRUE,
    input_submit_textarea("test", value = c("a", "b"))
  )

  expect_snapshot(
    error = TRUE,
    input_submit_textarea("test", button = tags$div("Not a button"))
  )

  expect_snapshot(
    error = TRUE,
    input_submit_textarea("test", submit_key = "invalid")
  )
})

# "brand" and "preset" interact in a few ways that need to be reconciled.
# Ultimately, we want three layers:
#
# 1. The base Bootstrap styles, by version. E.g. Bootstrap 5
# 2. The `preset` styles, which are one layer above and adjust the base
#    Bootstrap with the preset defaults E.g. "shiny" (bslib) or "flatly" (Bootswatch)
# 3. The `brand` styles, which can be seen as the final layer of default styles.
#    These styles set Bootstrap variables to adjust the `preset` theme in the
#    direction of the brand's color palette and typography.
#
# The `preset` and `version` can be specified in two ways with `brand`:
#
# 1. Directly via `bs_theme()`:
#    `bs_theme(preset = "shiny", version = 5, brand = TRUE)`.
#
# 2. As part of the `brand` definition:
#    `bs_theme(brand = list(defaults = list(shiny = list(theme = list(preset = "flatly")))))`
#
# So the order of operations is:
#
# 1. Read `brand`, if necessary.
# 2. If `bs_theme()` provides guidance for `version` and `theme`, use those preferences.
# 3. If not, inspect `brand` for guidance.
# 4. Finally, use our own defaults.
#
# Importantly we need to separately read `brand` and then resolve the preset and
# then resolve the brand bundle.

bs_brand_bundle <- function(brand, version = version_default()) {
  # `brand` is already resolved in `bs_theme()`

  if (is.null(brand)) {
    return()
  }

  rlang::check_installed("brand.yml")

  brand_fonts <- brand.yml::brand_sass_fonts(brand)
  brand_color_palette <- brand.yml::brand_sass_color_palette(brand)
  brand_color <- brand.yml::brand_sass_color(brand)
  brand_defaults <- brand.yml::brand_sass_defaults_bootstrap(brand)
  brand_typography <- brand.yml::brand_sass_typography(brand)

  if (version <= 4) {
    rlang::warn(
      sprintf(
        "Branded theming works best with Bootstrap v5. Some features may not work as expected for Bootstrap v%s.",
        version
      )
    )
  }

  sass_bundle(
    "brand_base" = switch_version(
      version,
      five = sass_layer_file(
        system_file("brand", "bs5", "_brand-yml.scss", package = "bslib")
      ),
      default = list()
    ),
    "brand_defaults" = brand_defaults$layer,
    "brand" = sass_layer(
      defaults = list2(
        "//* ---- brand.color.palette ---- *//",
        !!!brand_color_palette$defaults,
        "//* ---- brand.defaults ---- *//",
        !!!brand_defaults$defaults,
        "//* ---- brand.color ---- *//",
        !!!brand_color$defaults,
        "//* ---- brand.typography ---- *//",
        !!!brand_fonts$defaults,
        !!!brand_typography$defaults,
      ),
      rules = list2(
        brand_color_palette$rules,
        !!!brand_fonts$rules
      )
    )
  )
}

brand_resolve <- function(brand, ...) {
  UseMethod("brand_resolve")
}

#' @export
brand_resolve.list <- function(brand, ...) {
  rlang::check_installed("brand.yml")
  brand.yml::as_brand_yml(brand)
}

#' @export
`brand_resolve.NULL` <- function(brand, ...) {
  if (!is_installed("brand.yml")) {
    # If brand.yml is *not* installed, skip automatic detection
    return(NULL)
  }

  brand <- tryCatch(
    brand.yml::read_brand_yml(NULL),
    brand_yml_not_found = function(err) {
      # Couldn't find _brand.yml but we're not going to error
      NULL
    }
  )
  if (is.null(brand)) {
    return(NULL)
  }
  brand_resolve(brand, ...)
}

#' @export
brand_resolve.logical <- function(brand, ...) {
  if (identical(brand, FALSE)) {
    return(NULL)
  }
  rlang::check_installed("brand.yml")
  brand.yml::read_brand_yml(NULL)
}

#' @export
brand_resolve.character <- function(brand, ...) {
  rlang::check_installed("brand.yml")
  brand.yml::read_brand_yml(brand)
}

#' @export
brand_resolve.brand_yml <- function(brand, ...) {
  brand
}

brand_resolve_preset <- function(brand, preset = NULL, version = NULL) {
  rlang::check_installed("brand.yml")

  version_resolved <-
    version %||%
    brand.yml::brand_pluck(brand, "defaults", "shiny", "theme", "version") %||%
    brand.yml::brand_pluck(brand, "defaults", "bootstrap", "version") %||%
    version_default()

  preset_resolved <-
    preset %||%
    brand.yml::brand_pluck(brand, "defaults", "shiny", "theme", "preset") %||%
    switch_version(version_resolved, five = "shiny", default = "bootstrap")

  resolve_bs_preset(preset_resolved, version = version_resolved)
}

#' Navigation items
#'
#' Create nav item(s) for use inside nav containers (e.g., [navs_tab()],
#' [navs_bar()], etc).
#'
#' @param title A title to display. Can be a character string or UI elements
#'   (i.e., [tags]).
#' @param ... Depends on the function:
#'   * For `nav()` and `nav_content()`: UI elements (i.e., [tags]) to display
#'     when the item is active.
#'   * For `nav_menu()`: a collection of nav items (e.g., `nav()`, `nav_item()`).
#'   * For `nav_item()`: UI elements (i.e., [tags]) to place directly in
#'     the navigation panel (e.g., search forms, links to external content, etc).
#' @param value A character string to assign to the nav item. This value may be
#'   supplied to the relevant container's `selected` argument in order to show
#'   particular nav item's content immediately on page load. This value is also
#'   useful for programmatically updating the selected content via
#'   [nav_select()], [nav_hide()], etc (updating selected tabs this way is often
#'   useful for showing/hiding panels of content via other UI controls like
#'   [shiny::radioButtons()] -- in this scenario, consider using [nav_content()]
#'   with [navs_hidden()]).
#' @param icon Optional icon to appear next to the nav item's `title`.
#' @return A nav item that may be passed to a nav container (e.g. [navs_tab()]).
#' @export
#' @seealso [navs_tab()], [nav_select()].
#' @describeIn nav Content to display when the given item is selected.
nav <- function(title, ..., value = title, icon = NULL) {
  tabPanel_(title, ..., value = value, icon = icon)
}

#' @describeIn nav Create a menu of nav items.
#' @param align horizontal alignment of the dropdown menu relative to dropdown toggle.
#' @export
nav_menu <- function(title, ..., value = title, icon = NULL, align = c("left", "right")) {
  align <- match.arg(align)
  navbarMenu_(title, ..., menuName = value, icon = icon, align = align)
}

#' @describeIn nav Create nav content for use inside `navs_hidden()` (for
#'   creating custom navigation controls via `navs_select()`),
#' @export
nav_content <- function(value, ..., icon = NULL) {
  tabPanelBody_(value, ..., icon = icon)
}

#' @describeIn nav Place arbitrary content in the navigation panel (e.g., search
#'   forms, links to external content, etc.)
#' @export
nav_item <- function(...) {
  # TODO: drop form-inline since BS5 dropped it?
  # If we do that do we need navs_bar() to generate valid BS5 markup?
  tags$li(class = "bslib-nav-item nav-item form-inline", ...)
}

is_nav_item <- function(x) {
  tag_has_class(x, "bslib-nav-item")
}

#' @describeIn nav Adding spacing between nav items.
#' @export
nav_spacer <- function() {
  div(class = "bslib-nav-spacer")
}

is_nav_spacer <- function(x) {
  tag_has_class(x, "bslib-nav-spacer")
}

tag_has_class <- function(x, class) {
  if (!inherits(x, "shiny.tag")) {
    return(FALSE)
  }
  tagQuery(x)$hasClass(class)
}

bs3_sass_files <- function(x) {
  bs_sass_files(x, version = 3)
}

bs4_sass_files <- function(x) {
  bs_sass_files(x, version = 4)
}

bs5_sass_files <- function(x) {
  bs_sass_files(x, version = 5)
}

bs_sass_files <- function(files, version) {
  lapply(files, bs_sass_file, version = version)
}

# Search for one file at a time so we can throw informative errors
bs_sass_file <- function(file, version) {
  if (length(file) != 1) stop("file should be of length 1")

  file <- file.path(
    dirname(file),
    paste0("_", basename(file), ".scss")
  )

  f <- switch_version(
    version,
    five = lib_file("bs5", "scss", file),
    four = lib_file("bs4", "scss", file),
    three = lib_file("bs3", "assets", "stylesheets", "bootstrap", file)
  )
  if (f == "") stop("The bootstrap stylesheet '", file, "' doesn't exist.", call. = FALSE)
  sass_file(f)
}

bootswatch_sass_file <- function(theme, file, version = version_default()) {
  if (length(file) > 1) stop("file should be of length 1")
  theme <- match.arg(theme, bootswatch_themes(version))
  file <- paste0("_", file, ".scss")
  f <- file.path(bootswatch_dist(version), theme, file)
  if (file.exists(f)) return(sass_file(f))
  stop("Bootswatch file '", file, "' doesn't exist for theme '", theme, "'.", call. = FALSE)
}


# Given a vector of sass_file()s, create a list of sass_bundles(),
# so each rule may be removed layer (by it's files basename)
rule_bundles <- function(files) {
  paths <- vapply(files, get_sass_file_path, character(1))
  nms <- tools::file_path_sans_ext(basename(paths))
  Map(
    nms, files,
    f = function(nm, f) {
      sass_bundle(!!nm := sass_layer(rules = f))
    }
  )
}

get_sass_file_path <- function(x) {
  path <- attr(x, "sass_file_path")
  if (length(path)) return(path)

  stop("Couldn't find file path")
}

#' @export
#' @rdname navs
navs_tab_card <- function(..., id = NULL, selected = NULL,
                          header = NULL, footer = NULL) {
  tabs <- tagQuery(
    navs_tab(..., id = id, selected = selected, header = header, footer = footer)
  )

  # https://getbootstrap.com/docs/5.0/components/card/#navigation
  nav <- tabs$
    find(".nav")$
    addClass("card-header-tabs")$
    selectedTags()

  content <- tabs$find(".tab-content")$selectedTags()

  card(header = nav, content, caller = "navs_tab_card()")
}

#' @export
#' @param placement placement of the nav items relative to the content.
#' @rdname navs
navs_pill_card <- function(..., id = NULL, selected = NULL,
                           header = NULL, footer = NULL,
                           placement = c("above", "below")) {

  pills <- tagQuery(
    navs_pill(..., id = id, selected = selected, header = header, footer = footer)
  )

  above <- match.arg(placement) == "above"

  nav <- pills$
    find(".nav")$
    addClass(if (above) "card-header-pills")$
    selectedTags()

  content <- pills$find(".tab-content")$selectedTags()

  args <- list(nav, content, caller = "navs_pill_card()")

  names(args)[1] <- if (above) "header" else "footer"

  do.call(card, args)
}

card <- function(..., header = NULL, footer = NULL, caller) {
  tag <- div(
    class = "card",
    if (!is.null(header)) div(class = "card-header", header),
    div(class = "card-body", ...),
    if (!is.null(footer)) div(class = "card-footer", footer)
  )
  as_fragment(
    tag_require(tag, version = 4, caller = caller)
  )
}

withr::with_dir(
  rprojroot::find_package_root_file("vignettes"),
  {
    src4 <- knitr::knit_expand(
      file = "_variables-template.Rmd",
      version = 4
    )
    writeLines(src4, "bs4-variables/index.Rmd")

    src5 <- knitr::knit_expand(
      file = "_variables-template.Rmd",
      version = 5
    )
    invisible(writeLines(src5, "bs5-variables/index.Rmd"))
  }
)

# Returns TRUE if two files have identical contents
identical_files <- function(a, b) {
  content_a <- readBin(a, "raw", file.size(a))
  content_b <- readBin(b, "raw", file.size(b))
  identical(content_a, content_b)
}

# Returns TRUE if two directories have identical contents, recursively.
identical_dirs <- function(a, b) {
  if (!all(dir.exists(c(a, b)))) {
    return(FALSE)
  }
  files_a <- sort(dir(a, recursive = TRUE))
  files_b <- sort(dir(b, recursive = TRUE))
  if (!identical(files_a, files_b)) {
    return(FALSE)
  }

  for (file in files_a) {
    res <- identical_files(
      file.path(a, file),
      file.path(b, file)
    )
    if (!res) {
      return(FALSE)
    }
  }

  TRUE
}

#' Remove or retrieve Sass code from a theme
#'
#' @inheritParams bs_theme_update
#' @param ids a character vector of ids
#'
#' @return a modified [bs_theme()] object.
#'
#' @export
#' @examples
#'
#' # Remove CSS rules for print and carousels
#' bs4 <- bs_theme(version = 4)
#' bs_remove(bs4, c("_print", "_carousel"))
#'
#' # Remove BS3 compatibility layer
#' bs_remove(bs4, "bs3compat")
#'
bs_remove <- function(theme, ids = character(0)) {
  ids <- retain_valid_ids(theme, ids)
  sass_bundle_remove(theme, ids)
}

#' @rdname bs_remove
#' @param include_unnamed whether or not to include unnamed [sass::sass_layer()]s
#' (e.g., Bootstrap Sass variables, functions, and mixins).
#' @export
bs_retrieve <- function(theme, ids = character(0), include_unnamed = TRUE) {
  ids <- retain_valid_ids(theme, ids)
  if (isTRUE(include_unnamed)) {
    ids <- c(ids, "")
  }
  layers <- theme$layers[names(theme$layers) %in% ids]
  structure(
    sass_bundle(!!!layers),
    class = class(theme)
  )
}


retain_valid_ids <- function(theme, ids) {
  assert_bs_theme(theme)
  stopifnot(is.character(ids))

  bundle_ids <- names(theme$layers)
  if (!length(ids)) {
    stop(
      "Provide one or more of the following `ids`: ",
      paste(bundle_ids, collapse = ", ")
    )
  }

  missing_ids <- setdiff(ids, bundle_ids)
  if (length(missing_ids)) {
    warning(
      "The following `ids` weren't found in `theme`: ",
      paste(missing_ids, collapse = ", "), ".\n\n",
      "Possible `ids` include: ",
      paste(bundle_ids[nzchar(bundle_ids)], collapse = ", ")
    )
  }

  intersect(ids, bundle_ids)
}

bs_size <- function(x) {
  bs <- bs_theme_dependencies(x)[[2]]
  files <- file.path(bs$src, bs$stylesheet)
  file.info(files)["size"]
}

test_that("Can remove rules", {
  a <- bs_theme(version = "4")
  b <- bs_remove(a, "_print")
  c <- bs_remove(b, "_carousel")
  expect_true(bs_size(a) > bs_size(b))
  expect_true(bs_size(b) > bs_size(c))
})

test_that("Can retrieve rules", {
  a <- bs_theme(version = "4")
  b <- bs_retrieve(a, "_print")
  c <- bs_retrieve(a, "_carousel")
  expect_true(bs_size(a) > bs_size(b))
  expect_true(bs_size(a) > bs_size(c))
})

generate_random_walk <- function(num_steps = 90) {
  start_date <- as.POSIXct(as.integer(Sys.time()) * runif(1), origin = "1970-01-01")

  increments <- rnorm(num_steps)
  cumulative_sum <- cumsum(increments)
  time_series <- c(0, cumulative_sum) + rnorm(1, 0, 50) + 25

  dates <- seq(start_date, length.out = num_steps + 1, by = "day")

  data.frame(date = dates, value = time_series)
}

as_plotly_sparkline <- function(plot, color = "white") {
  plot |>
    layout(
      xaxis = list(visible = FALSE, showgrid = FALSE, title = ""),
      yaxis = list(visible = FALSE, showgrid = FALSE, title = ""),
      hovermode = "x",
      margin = list(t = 0, r = 0, l = 0, b = 0),
      font = list(color = color),
      paper_bgcolor = "transparent",
      plot_bgcolor = "transparent"
    ) |>
    config(displayModeBar = FALSE) |>
    htmlwidgets::onRender(
      "function(el) {
        var ro = new ResizeObserver(function() {
          var visible = el.offsetHeight > 200;
          Plotly.relayout(el, {'xaxis.visible': visible});
        });
        ro.observe(el);
      }"
    )
  }


random_plotly_plot <- function(type = NULL, color = "white") {
  if (is.null(type)) {
    type <-  sample(c("bar", "box", "line"), 1)
  }

  plot <- switch(
    type,
    bar = random_plotly_bar(color, 50),
    box = random_plotly_box(color, 50),
    line = random_plotly_line(color, 50),
    stop("Not a valid random plot type: ", type)
  )

  as_plotly_sparkline(plot, color)
}

random_plotly_bar <- function(color, n = 50) {
  plot_ly(
    x = ~ runif(n),
    type = "histogram",
    histnorm = "probability",
    nbinsx = 10,
    color = I(color),
    stroke = I(color),
    alpha_stroke = 1,
    alpha = 0.6
  )
}

random_plotly_box <- function(color, n = 50) {
  plot_ly(x = ~rnorm(n), type = "box", color = I(color))
}

random_plotly_line <- function(color, n) {
    add_lines(
      plot_ly(generate_random_walk(n)),
      x = ~ date,
      y = ~ value,
      color = I(color),
      fill = "tozeroy",
      span = I(1),
      alpha = 0.2
    )
}

#' Obtain the currently active theme at render time
#'
#' Intended for advanced use by developers to obtain the currently active theme
#' _at render time_ and primarily for implementing themable widgets that can't
#' otherwise be themed via [bs_dependency_defer()] .
#'
#' @details
#'
#' This function should generally only be called at print/render time. For
#' example:
#'
#'  * Inside the `preRenderHook` of `htmlwidgets::createWidget()`.
#'  * Inside of a custom [print] method that generates [htmltools::tags].
#'  * Inside of a [htmltools::tagFunction()]
#'
#' Calling this function at print/render time is important because it does
#' different things based on the context in which it's called:
#'
#' * If a reactive context is active, `session$getCurrentTheme()` is called
#'   (which is a reactive read).
#' * If no reactive context is active, `shiny::getCurrentTheme()` is called
#'   (which returns the current app's `theme`, if relevant).
#' * If `shiny::getCurrentTheme()` comes up empty, then `bs_global_get()`
#'   is called, which is relevant for [rmarkdown::html_document()], and
#'   possibly other static rendering contexts.
#'
#' @param session The current Shiny session (if any).
#'
#' @return a [bs_theme()] object.
#' @export
bs_current_theme <- function(session = get_reactive_domain()) {
  # If we're able to make a reactive read of the theme, then do it
  if (has_valid_reactive_context(session)) {
    return(session$getCurrentTheme())
  }
  # Otherwise, this'll be a non-reactive read of session/app/global state
  get_current_theme() %||% bs_global_get()
}

get_reactive_domain <- function() {
  if (isNamespaceLoaded("shiny")) shiny::getDefaultReactiveDomain() else NULL
}

has_valid_reactive_context <- function(session) {
  if (is.null(session)) return(FALSE)
  if (!"getCurrentTheme" %in% names(session)) return(FALSE)
  hasReactiveContext <- getFromNamespace("hasCurrentContext", "shiny") %||% function() FALSE
  hasReactiveContext()
}

get_current_theme <- function() {
  if (isNamespaceLoaded("shiny")) shiny::getCurrentTheme()
}

#' @export
#' @inheritParams card
#' @inheritParams card_body
#' @param title A (left-aligned) title to place in the card header/footer. If
#'   provided, other nav items are automatically right aligned.
#' @include navs-legacy.R
#' @rdname navset
navset_card_tab <- function(..., id = NULL, selected = NULL, title = NULL,
                          sidebar = NULL, header = NULL, footer = NULL,
                          height = NULL, full_screen = FALSE, wrapper = card_body) {

  items <- collect_nav_items(..., wrapper = wrapper)

  tabs <- navset_tab(
    !!!items, id = id, selected = selected, header = header, footer = footer
  )

  tabQ <- tagQuery(tabs)

  # https://getbootstrap.com/docs/5.0/components/card/#navigation
  nav <- tabQ$children(".nav")$addClass("card-header-tabs")$selectedTags()[[1]]
  content <- tabQ$children(".tab-content")$selectedTags()[[1]]

  card(
    height = height,
    full_screen = full_screen,
    if (!is.null(title)) {
      card_header(class = "bslib-navs-card-title", tags$span(title), nav)
    } else {
      card_header(nav)
    },
    navs_card_body(content, sidebar)
  )
}

#' @export
#' @param placement placement of the nav items relative to the content.
#' @rdname navset
navset_card_pill <- function(..., id = NULL, selected = NULL, title = NULL,
                           sidebar = NULL, header = NULL, footer = NULL,
                           height = NULL, placement = c("above", "below"),
                           full_screen = FALSE, wrapper = card_body) {

  items <- collect_nav_items(..., wrapper = wrapper)

  pills <- navset_pill(
    !!!items, id = id, selected = selected,
    header = header, footer = footer
  )

  above <- match.arg(placement) == "above"

  pillQ <- tagQuery(pills)

  # https://getbootstrap.com/docs/5.0/components/card/#navigation
  nav <- pillQ$children(".nav")$addClass(if (above) "card-header-pills")$selectedTags()[[1]]
  content <- pillQ$children(".tab-content")$selectedTags()[[1]]

  nav_args <- if (!is.null(title)) {
    list(class = "bslib-navs-card-title", tags$span(title), nav)
  } else {
    list(nav)
  }

  card(
    height = height,
    full_screen = full_screen,
    if (above) card_header(!!!nav_args),
    navs_card_body(content, sidebar),
    if (!above) card_footer(!!!nav_args)
  )
}


collect_nav_items <- function(..., wrapper) {
  items <- rlang::list2(...)

  # Wrap any nav_panel() children up into card items
  nav_to_card_item <- function(x) {
    if (isNavbarMenu(x)) {
      x$tabs <- lapply(x$tabs, nav_to_card_item)
    }
    if (isTabPanel(x)) {
      x$children <- as_card_items(x$children, wrapper = wrapper)
    }
    x
  }

  lapply(items, nav_to_card_item)
}

# Always give tab contents the potential to fill since that's akin to the
# normal card() API (i.e. the card() is a fill container) and users have
# option to make the contents fill via card_body(fill = TRUE, fillable = TRUE)
navs_card_body <- function(content, sidebar) {
  content <- makeTabsFillable(content, fillable = TRUE, gap = 0, padding = 0)
  if (!is.null(sidebar)) {
    content <- layout_sidebar(
      sidebar = sidebar,
      fillable = TRUE,
      border = FALSE,
      content
    )
  }
  as.card_item(content)
}


# Given a .tab-content container, mark each relevant .tab-pane as a
# fill container/item.
makeTabsFillable <- function(content, fillable = TRUE, navbar = FALSE, gap = NULL, padding = NULL) {
  if (!inherits(content, "shiny.tag") || !tagQuery(content)$hasClass("tab-content")) {
    abort("Expected `content` to be a tag with a tab-content class")
  }

  if (isFALSE(fillable)) {
    return(content)
  }

  # Even if only one .tab-pane wants fillable behavior, the .tab-content
  # must to be a fillable container.
  content <- bindFillRole(content, container = TRUE, item = TRUE)

  tagQuery(content)$
    find(".tab-pane")$
    each(function(x, i) {

      if (isTRUE(fillable) || isTRUE(tagGetAttribute(x, "data-value") %in% fillable)) {
        x <- tagAppendAttributes(
          class = "bslib-gap-spacing",
          style = css(
            # Remove the margin between nav and content (for page_navbar())
            "--bslib-navbar-margin" = if (navbar) 0,
            gap = gap,
            padding = padding
          ),
          bindFillRole(x, container = TRUE, item = TRUE)
        )
      }

      x
    })$
    allTags()
}

#!/usr/bin/env Rscript
#
# Build the ggsql Prism grammar file from the TextMate grammar source.
#
# Usage:
#   Rscript tools/build_ggsql_grammar.R
#
# This fetches the ggsql TextMate grammar from posit-dev/ggsql on GitHub,
# extracts token patterns, and generates a Prism-compatible grammar file
# that extends the built-in SQL grammar.
#
# The output is written to:
#   inst/lib/prism-code-editor/prism/languages/ggsql.js

TMGRAMMAR_URL <- "https://raw.githubusercontent.com/posit-dev/ggsql/main/ggsql-vscode/syntaxes/ggsql.tmLanguage.json"

# Paths relative to the bslib package root
LANGUAGES_DIR <- "inst/lib/prism-code-editor/prism/languages"
SQL_JS <- file.path(LANGUAGES_DIR, "sql.js")
OUTPUT_JS <- file.path(LANGUAGES_DIR, "ggsql.js")

# ---- Helpers ----

# Extract word alternatives from a TextMate match pattern like
# "\\b(word1|word2|word3)\\b" or "\\b(?i:word1|word2)\\b"
extract_words <- function(match_pattern) {
  # Try non-capture group with flags first: \b(?i:word1|word2)\b
  m <- regmatches(
    match_pattern,
    regexec("\\\\b\\(\\?[a-z]*:([^)]+)\\)\\\\b", match_pattern)
  )[[1]]
  if (length(m) == 2) {
    return(strsplit(m[2], "\\|")[[1]])
  }

  # Try capture group: \b(word1|word2)\b
  m <- regmatches(
    match_pattern,
    regexec("\\\\b\\(([^)]+)\\)\\\\b", match_pattern)
  )[[1]]
  if (length(m) == 2) {
    return(strsplit(m[2], "\\|")[[1]])
  }

  NULL
}

# Extract word alternatives from a function pattern like
# "(?i)\\b(func1|func2)\\b\\s*\\("
extract_function_words <- function(match_pattern) {
  m <- regmatches(
    match_pattern,
    regexec("\\\\b\\(([^)]+)\\)\\\\b\\\\s\\*\\\\\\(", match_pattern)
  )[[1]]
  if (length(m) == 2) {
    return(strsplit(m[2], "\\|")[[1]])
  }
  NULL
}

# Format a word list as a JS regex alternation: /\b(?:word1|word2)\b/flags
format_word_regex <- function(words, flags = "") {
  alts <- paste(words, collapse = "|")
  sprintf("/\\b(?:%s)\\b/%s", alts, flags)
}

# Format function words with lookahead: /\b(?:func1|func2)(?=\s*\()/flags
format_function_regex <- function(words, flags = "i") {
  alts <- paste(words, collapse = "|")
  sprintf("/\\b(?:%s)(?=\\s*\\()/%s", alts, flags)
}

# Extract the content-hashed import path from sql.js
extract_import_path <- function(sql_js_path) {
  lines <- readLines(sql_js_path, n = 1)
  m <- regmatches(lines, regexec('"([^"]+index-[^"]+\\.js)"', lines))[[1]]
  if (length(m) < 2) {
    stop("Could not extract import path from ", sql_js_path)
  }
  m[2]
}

# ---- Main ----

main <- function() {
  if (!file.exists(SQL_JS)) {
    stop("sql.js not found at ", SQL_JS, ". Run from the bslib package root.")
  }

  message("Fetching TextMate grammar from: ", TMGRAMMAR_URL)
  grammar <- jsonlite::fromJSON(TMGRAMMAR_URL, simplifyDataFrame = FALSE)
  repo <- grammar$repository

  import_path <- extract_import_path(SQL_JS)
  message("Using import path: ", import_path)

  # -- Extract ggsql clause keywords --
  # Collect from: begin captures of clause patterns + keyword.other matches
  clause_keywords <- character(0)

  # Clause names come from the `begin` regex of each *-clause pattern
  clause_names <- grep("-clause$", names(repo), value = TRUE)
  for (clause_name in clause_names) {
    clause <- repo[[clause_name]]
    if (!is.null(clause$begin)) {
      words <- extract_words(clause$begin)
      if (!is.null(words)) clause_keywords <- c(clause_keywords, words)
    }
  }

  # Also collect keyword.other.ggsql from all clause patterns and common-clause-patterns
  keyword_sources <- c("common-clause-patterns", clause_names)
  for (src in keyword_sources) {
    for (p in repo[[src]]$patterns) {
      if (identical(p$name, "keyword.other.ggsql")) {
        words <- extract_words(p$match)
        if (!is.null(words)) clause_keywords <- c(clause_keywords, words)
      }
    }
  }
  clause_keywords <- unique(clause_keywords)

  # -- Extract token groups from named patterns --
  # Searches all patterns in repo_key(s) for token_name, accumulating words
  # across multiple matching patterns.
  extract_token_group <- function(repo_keys, token_name) {
    words <- character(0)
    for (repo_key in repo_keys) {
      if (is.null(repo[[repo_key]])) {
        stop(
          "Repository key '",
          repo_key,
          "' not found in TextMate grammar. ",
          "The upstream grammar structure may have changed."
        )
      }
      for (p in repo[[repo_key]]$patterns) {
        if (identical(p$name, token_name)) {
          w <- extract_words(p$match)
          if (!is.null(w)) words <- c(words, w)
        }
      }
    }
    if (length(words) == 0) {
      stop(
        "No words extracted for '",
        token_name,
        "' from ",
        paste(repo_keys, collapse = ", "),
        ". ",
        "The upstream grammar structure may have changed."
      )
    }
    unique(words)
  }

  geoms <- extract_token_group(
    c("draw-clause", "place-clause"),
    "support.type.geom.ggsql"
  )
  scale_types <- extract_token_group(
    "scale-clause",
    "keyword.control.scale-modifier.ggsql"
  )
  scale_values <- extract_token_group(
    "scale-clause",
    "constant.language.scale-type.ggsql"
  )
  aesthetics <- extract_token_group(
    "aesthetics",
    "support.type.aesthetic.ggsql"
  )
  properties <- extract_token_group(
    c("scale-clause", "facet-clause", "project-clause", "label-clause"),
    "support.type.property.ggsql"
  )
  projects <- extract_token_group(
    "project-clause",
    "support.type.project.ggsql"
  )

  # -- Extract SQL function names --
  all_functions <- character(0)
  for (p in repo[["sql-functions"]]$patterns) {
    words <- extract_function_words(p$match)
    if (!is.null(words)) all_functions <- c(all_functions, words)
  }
  all_functions <- unique(all_functions)

  if (length(clause_keywords) == 0) {
    stop(
      "No clause keywords extracted. The upstream grammar structure may have changed."
    )
  }
  if (length(all_functions) == 0) {
    stop(
      "No SQL functions extracted. The upstream grammar structure may have changed."
    )
  }

  # -- Generate the JS file --
  js <- c(
    "// ggsql.js — ggsql language for prism-code-editor",
    "//",
    "// AUTO-GENERATED by tools/build_ggsql_grammar.R from the TextMate grammar at:",
    "// https://github.com/posit-dev/ggsql/blob/main/ggsql-vscode/syntaxes/ggsql.tmLanguage.json",
    "// Do not edit by hand.",
    "//",
    "// NOTE: The import path below is a content-hashed internal module of",
    "// prism-code-editor. If bslib updates prism-code-editor, this hash may change.",
    "// Re-run this script to regenerate.",
    "",
    'import "./sql.js";',
    sprintf('import { l as languages } from "%s";', import_path),
    "",
    "var sql = languages.sql;",
    "var ggsql = {};",
    "",
    "// Copy SQL tokens",
    'Object.keys(sql).forEach(function(k) { ggsql[k] = sql[k]; });',
    "",
    sprintf(
      r"(// ggsql clause keywords
ggsql["ggsql-keyword"] = {
  pattern: %s,
  alias: "keyword",
};)",
      format_word_regex(clause_keywords, flags = "i")
    ),
    "",
    sprintf(
      r"(// Geom types
ggsql["ggsql-geom"] = {
  pattern: %s,
  alias: "builtin",
};)",
      format_word_regex(geoms)
    ),
    "",
    sprintf(
      r"(// Scale type modifiers
ggsql["ggsql-scale-type"] = {
  pattern: %s,
  alias: "builtin",
};)",
      format_word_regex(scale_types, flags = "i")
    ),
    "",
    sprintf(
      r"(// Scale type values
ggsql["ggsql-scale-value"] = {
  pattern: %s,
  alias: "string",
};)",
      format_word_regex(scale_values)
    ),
    "",
    sprintf(
      r"(// Aesthetic names
ggsql["ggsql-aesthetic"] = {
  pattern: %s,
  alias: "attr-name",
};)",
      format_word_regex(aesthetics)
    ),
    "",
    sprintf(
      r"(// Property names
ggsql["ggsql-property"] = {
  pattern: %s,
  alias: "property",
};)",
      format_word_regex(properties)
    ),
    "",
    sprintf(
      r"(// Project types
ggsql["ggsql-project"] = {
  pattern: %s,
  alias: "class-name",
};)",
      format_word_regex(projects)
    ),
    "",
    "// Fat arrow operator",
    'ggsql["ggsql-arrow"] = {',
    "  pattern: /=>/,",
    '  alias: "operator",',
    "};",
    "",
    "// SQL functions (aggregate, window, datetime, string, math, conversion, conditional, JSON, list)",
    sprintf('ggsql["function"] = %s;', format_function_regex(all_functions)),
    "",
    "// Reorder: ggsql tokens before generic SQL keyword/boolean",
    "var ordered = {};",
    '["comment", "variable", "string", "identifier"].forEach(function(k) {',
    "  if (k in ggsql) ordered[k] = ggsql[k];",
    "});",
    '["ggsql-keyword", "ggsql-geom", "ggsql-scale-type", "ggsql-scale-value",',
    ' "ggsql-aesthetic", "ggsql-property", "ggsql-project", "ggsql-arrow"].forEach(function(k) {',
    "  if (k in ggsql) ordered[k] = ggsql[k];",
    "});",
    "Object.keys(ggsql).forEach(function(k) {",
    "  if (!(k in ordered)) ordered[k] = ggsql[k];",
    "});",
    "",
    "languages.ggsql = ordered;",
    ""
  )

  writeLines(js, OUTPUT_JS)
  message("Generated: ", OUTPUT_JS)
}

main()

local_disable_cache()

describe("Compiling against theme", {
  # Example CSS that includes one variable, one function call, one mixin
  bs4_css <- ".foo { background-color: $primary; color: color-contrast($primary); width: 120px; height: 120px; }"
  resolved_css <- ".foo { background-color: #007bff; color: #fff; width: 120px; height: 120px; }"

  # Compare bs_sass(input1) and sass(input2)
  expect_bs4_equal <- function(input1, input2, options = sass_options(), theme = bs_theme(version = 4)) {
    expect_css(
      sass_partial(input1, theme, options = options),
      sass(input2, options = options)
    )
  }

  it("automatically includes Bootstrap 4 variables/functions", {
    expect_bs4_equal(bs4_css, resolved_css)
  })

  it("respects output_style", {
    expect_bs4_equal(bs4_css, resolved_css, options = sass_options(output_style = "compressed"))
  })


  it("respects theme options", {
    # Theme options are respected
    expect_bs4_equal(
      bs4_css, ".foo { background-color: #FF0000; color: #fff; width: 120px; height: 120px; }",
      theme = bs_theme(version = 4, primary = "red")
    )
    # Unless they're not
    expect_bs4_equal(bs4_css, resolved_css)
  })

})

getDefaultReactiveDomain <- function() {
  shiny::getDefaultReactiveDomain()
}

# Shiny internal funcs needed for nav() (i.e., tabPanel()) logic

processDeps <- function(...) {
  getFromNamespace("processDeps", "shiny")(...)
}

p_randomInt <- function(...) {
  getFromNamespace("p_randomInt", "shiny")(...)
}

# Copy of shiny::getCurrentThemeVersion()
# (copied to avoid >1.6.0 dependency)
getCurrentThemeVersion <- function() {
  theme <- shiny::getCurrentTheme()
  if (is_bs_theme(theme)) theme_version(theme) else "3"
}

# Copy of shiny:::anyNamed()
anyNamed <- function(x) {
  if (length(x) == 0)
    return(FALSE)
  nms <- names(x)
  if (is.null(nms))
    return(FALSE)
  any(nzchar(nms))
}

library(stringr)

# if (Sys.getenv("RSTUDIO") == "1") {
#   stop("Please run this script from the command line: `Rscript tools/fonts.R`")
# }

# TODO: make sure this isn't drastically different from 2.1 Mb
fonts_home <- file.path("inst", "fonts")
unlink(fonts_home, recursive = TRUE)
dir.create(fonts_home)

extract_first_group <- function(x, pattern) {
  matches <- regmatches(x, regexec(pattern, x))
  na.omit(sapply(matches, "[", 2))
}

download_and_copy_fonts <-  function(theme) {
  theme_scss <- readLines(file.path(theme, "_bootswatch.scss"))
  web_font_path <- grep("$web-font-path:", theme_scss, value = TRUE, fixed = TRUE)
  if (!length(web_font_path)) {
    return()
  }
  web_font_url <- strsplit(stringr::str_trim(web_font_path), "\\s+")[[1]][2]
  if (!length(web_font_url)) {
    warning("Wasn't able to extract a url for $web-font-path from theme: ", theme)
    return()
  }
  web_font_url <- gsub('(^")|("$)', '', web_font_url)
  css_file <- file.path(theme, "font.css")
  css <- sass:::read_gfont_url(web_font_url, css_file)
  urls <- sass:::extract_group(css, "url\\(([^)]+)")
  basenames <- basename(urls)
  Map(function(url, nm) {
    sass:::download_file(url, file.path(fonts_home, nm))
    css <<- sub(url, file.path("fonts", nm), css, fixed = TRUE)
  }, urls, basenames)
  writeLines(css, css_file)
  NULL
}


themes5 <- list.dirs(
  "inst/lib/bsw5/dist",
  recursive = FALSE,
  full.names = TRUE
)
themes4 <- list.dirs(
  "inst/lib/bsw4/dist",
  recursive = FALSE,
  full.names = TRUE
)
themes3 <- list.dirs(
  "inst/lib/bsw3",
  recursive = FALSE,
  full.names = TRUE
)

lapply(themes5, download_and_copy_fonts)
lapply(themes4, download_and_copy_fonts)
lapply(themes3, download_and_copy_fonts)

# This figures out which precompiled CSS files are available.
# This is memoized - it runs once per R session.
precompiled_css <- local({
  themes <- NULL

  function() {
    if (is.null(themes)) {
      versions <- dir(system.file("css-precompiled", package = "bslib"))
      themes <<- lapply(versions, function(version) {
        list(
          version = version,
          theme = bs_theme(version)
        )
      })
    }
    themes
  }
})


precompiled_css_version <- function(theme) {
  for (precompiled in precompiled_css()) {
    if (identical(theme, precompiled$theme)) {
      return(theme_version(theme))
    }
  }

  return(NULL)
}


#' Get the path to a precompiled CSS file
#'
#' This function is meant for development and debugging purposes. It can be used
#' to test if a precompiled CSS file exists for a given theme object.
#'
#' @inheritParams bs_theme_update
#'
#' @return The path to the precompiled CSS file, if it exists, or `NULL`, if
#'   not.
#' @keywords internal
#' @export
precompiled_css_path <- function(theme = bs_theme()) {
  version <- precompiled_css_version(theme)
  if (is.null(version)) {
    return(NULL)
  }

  system_file(package = "bslib", "css-precompiled", version, "bootstrap.min.css")
}

#' Switch input control
#'
#' Create an on-off style switch control for specifying logical values.
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fixed(
#'   title = "Keyboard Settings",
#'   h2("Keyboard Settings"),
#'   input_switch("auto_capitalization", "Auto-Capitalization", TRUE),
#'   input_switch("auto_correction", "Auto-Correction", TRUE),
#'   input_switch("check_spelling", "Check Spelling", TRUE),
#'   input_switch("smart_punctuation", "Smart Punctuation"),
#'   h2("Preview"),
#'   verbatimTextOutput("preview")
#' )
#'
#' server <- function(input, output, session) {
#'   output$preview <- renderPrint({
#'     list(
#'       auto_capitalization = input$auto_capitalization,
#'       auto_correction = input$auto_correction,
#'       check_spelling = input$check_spelling,
#'       smart_punctuation = input$smart_punctuation
#'     )
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @param id An input id.
#' @param label A label for the switch.
#' @param value Whether or not the switch should be checked by default.
#' @param width Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `width="200px"`).
#'
#' @return Returns a UI element for a switch input control. The server value
#'   received for the input corresponding to `id` will be a logical
#'   (`TRUE`/`FALSE`) value.
#'
#' @family input controls
#' @export
input_switch <- function(id, label, value = FALSE, width = NULL) {
  tag <- input_checkbox(id, label, class = "form-check form-switch", value = value, width = width)
  tag <- tag_require(tag, version = 5, caller = "input_switch()")
  as_fragment(tag)
}

#' @rdname input_switch
#' @inheritParams nav_insert
#' @export
update_switch <- function(id, label = NULL, value = NULL, session = get_current_session()) {
  message <- dropNulls(list(label = label, value = value))
  session$sendInputMessage(id, message)
}

#' @rdname input_switch
#' @inheritParams toggle_sidebar
#' @export
toggle_switch <- function(id, value = NULL, session = get_current_session()) {
  if (!is.null(value) && !rlang::is_logical(value, n = 1)) {
    abort("`value` must be a `NULL` or a single logical value.")
  }

  msg <- dropNulls(list(id = id, value = value))

  callback <- function() {
    session$sendCustomMessage("bslib.toggle-input-binary", msg)
  }

  session$onFlush(callback, once = TRUE)
}

input_checkbox <- function(id, label, class = "form-check", value = FALSE, width = NULL, inline = FALSE) {
  div(
    class = "form-group shiny-input-container",
    class = if (inline) "shiny-input-container-inline",
    style = css(width = width),
    div(
      class = class,
      tags$input(
        id = id,
        class = "form-check-input",
        type = "checkbox",
        role = "switch",
        checked = if (value) NA,
      ),
      tags$label(
        # The span here is needed to adhere to shiny.js' checkbox binding logic
        # https://github.com/rstudio/shiny/blob/c21ba0b/srcts/src/bindings/input/checkbox.ts#L42-L43
        tags$span(label),
        class = "form-check-label",
        `for` = id
      )
    ),
    component_dependency_js("bslibShiny")
  )
}

library(ggplot2)
library(lattice)


ids <- factor(c("1.1", "2.1", "1.2", "2.2", "1.3", "2.3"))
values <- data.frame(
  id = ids,
  value = c(3, 3.1, 3.1, 3.2, 3.15, 3.5)
)
positions <- data.frame(
  id = rep(ids, each = 4),
  x = c(2, 1, 1.1, 2.2, 1, 0, 0.3, 1.1, 2.2, 1.1, 1.2, 2.5, 1.1, 0.3,
        0.5, 1.2, 2.5, 1.2, 1.3, 2.7, 1.2, 0.5, 0.6, 1.3),
  y = c(-0.5, 0, 1, 0.5, 0, 0.5, 1.5, 1, 0.5, 1, 2.1, 1.7, 1, 1.5,
        2.2, 2.1, 1.7, 2.1, 3.2, 2.8, 2.1, 2.2, 3.3, 3.2)
)
huron <- data.frame(year = 1875:1972, level = as.vector(LakeHuron))

ggplot2_examples <- list(
  GeomPointColor = {
    ggplot(mtcars, aes(wt, mpg, color = cyl)) + geom_point()
  },
  GeomAbline = {
    ggplot(mtcars, aes(mpg, wt)) +
      geom_point() +
      geom_hline(aes(yintercept = wt), data.frame(cyl = c(4, 6, 8), wt = c(2.28, 3.11, 4.00))) +
      facet_wrap(~ cyl)
  },
  GeomBar = ggplot(mpg, aes(class)) + geom_bar(aes(weight = displ)),
  GeomBin2d = ggplot(diamonds, aes(x, y)) + xlim(4, 10) + ylim(4, 10) + geom_bin2d(),
  GeomBox  = ggplot(mpg, aes(class, hwy)) + geom_boxplot(),
  GeomContour = ggplot(faithfuld, aes(waiting, eruptions, z = density)) + geom_contour(),
  GeomContourRaster =  ggplot(faithfuld, aes(waiting, eruptions, z = density)) + geom_raster(aes(fill = density)) + geom_contour(),
  GeomCount = ggplot(mpg, aes(cty, hwy)) + geom_count(),
  GeomDensity = ggplot(diamonds, aes(carat)) + geom_density(),
  GeomDensityColor = ggplot(diamonds, aes(depth, fill = cut, colour = cut)) + geom_density(alpha = 0.1),
  GeomDotPlot = ggplot(mtcars, aes(x = mpg)) + geom_dotplot(),
  GeomError = {
    ggplot(data.frame(
      trt = factor(c(1, 1, 2, 2)),
      resp = c(1, 5, 3, 4),
      group = factor(c(1, 2, 1, 2)),
      se = c(0.1, 0.3, 0.3, 0.2)
    ), aes(resp, trt, colour = group)) + geom_point() +
      geom_errorbarh(aes(xmax = resp + se, xmin = resp - se))
  },
  GeomHex = ggplot(diamonds, aes(carat, price)) + geom_hex(binwidth = c(.1, 500)),
  GeomHistogram = ggplot(diamonds, aes(price, fill = cut)) + geom_histogram(binwidth = 500),
  GeomJitter = ggplot(mpg, aes(cyl, hwy)) + geom_jitter(width = 0.25),
  GeomErrorbar = ggplot(data.frame(
    trt = factor(c(1, 1, 2, 2)),
    resp = c(1, 5, 3, 4),
    group = factor(c(1, 2, 1, 2)),
    upper = c(1.1, 5.3, 3.3, 4.2),
    lower = c(0.8, 4.6, 2.4, 3.6)
  ), aes(trt, resp, fill = group)) +
    geom_col(position = position_dodge(width = 0.9)) +
    geom_errorbar(aes(ymin = lower, ymax = upper), position = position_dodge(width = 0.9), width = 0.25),
  GeomPolygon = ggplot(values) +
    geom_map(aes(map_id = id), map = positions) +
    expand_limits(positions),
  GeomPolygon2 = ggplot(values, aes(fill = value)) +
    geom_map(aes(map_id = id), map = positions) +
    expand_limits(positions),
  GeomLine = ggplot(economics_long, aes(date, value01, group = variable)) + geom_line(),
  GeomLine2 = ggplot(economics_long, aes(date, value01, colour = variable)) + geom_line(),
  GeomPoint = ggplot(diamonds, aes(carat, price)) + geom_point(alpha = 1/50),
  GeomPoint2 = ggplot(mtcars, aes(wt, mpg, color = factor(cyl), shape = factor(cyl))) + geom_point(),
  GeomQuantile = ggplot(mpg, aes(displ, 1 / hwy)) + geom_point() + geom_quantile(),
  GeomRibbon = ggplot(huron, aes(year)) +
    geom_ribbon(aes(ymin = level - 1, ymax = level + 1), fill = "grey70") +
    geom_line(aes(y = level)),
  GeomRug = ggplot(mtcars, aes(wt, mpg)) + geom_point() + geom_rug(),
  GeomRug2 = ggplot(mtcars, aes(wt, mpg)) + geom_point() + geom_rug(outside = TRUE) + coord_cartesian(clip = "off"),
  GeomCurve = {
    df <- data.frame(x1 = 2.62, x2 = 3.57, y1 = 21.0, y2 = 15.0)
    ggplot(mtcars, aes(wt, mpg)) +
      geom_point() +
      geom_curve(aes(x = x1, y = y1, xend = x2, yend = y2, colour = "curve"), data = df) +
      geom_segment(aes(x = x1, y = y1, xend = x2, yend = y2, colour = "segment"), data = df) +
      geom_curve(aes(x = x1, y = y1, xend = x2, yend = y2), data = df, curvature = -0.2)
  },
  GeomSmooth = ggplot(diamonds[sample(nrow(diamonds), 1000), ], aes(carat, price)) +
    geom_point(alpha = 0.2) +
    geom_smooth() +
    facet_wrap(~cut) + ggtitle("Diamond price by carat and cut"),
  GeomSmooth2 = ggplot(mpg, aes(displ, hwy, colour = class)) +
    geom_point() +
    geom_smooth(se = FALSE, method = lm),
  GeomSpoke = {
    df <- expand.grid(x = 1:10, y=1:10)
    df$angle <- runif(100, 0, 2*pi)
    df$speed <- runif(100, 0, sqrt(0.1 * df$x))

    ggplot(df, aes(x, y)) +
      geom_point() +
      geom_spoke(aes(angle = angle), radius = 0.5)
  },
  GeomText = ggplot(mtcars, aes(wt, mpg, label = rownames(mtcars))) + geom_text(),
  GeomLabel = ggplot(mtcars, aes(wt, mpg, label = rownames(mtcars))) + geom_label(),
  GeomTile = ggplot(data.frame(
    x = rep(c(2, 5, 7, 9, 12), 2),
    y = rep(c(1, 2), each = 5),
    z = factor(rep(1:5, each = 2)),
    w = rep(diff(c(0, 4, 6, 8, 10, 14)), 2)
  ), aes(x, y, width = w)) + geom_tile(aes(fill = z)),
  GeomViolin = ggplot(mtcars, aes(factor(cyl), mpg)) + geom_violin()
)

lattice_examples <- list(
  settings = quote(show.settings()),
  densityplot = quote(densityplot(~mpg, data=mtcars,
                                  group=factor(mtcars$am, levels=c(0, 1),
                                               labels=c("Automatic", "Manual")),
                                  main="MPG Distribution by Transmission Type",
                                  xlab="Miles per Gallon",
                                  auto.key=TRUE))
)

base_examples <- list(
  scatterplot = quote(plot(iris$Sepal.Length, iris$Sepal.Width,
                           col = iris$Species,
                           main = "Sepal Length vs Width in Iris")),
  hist = quote(hist(iris$Sepal.Length))
)

#' Navigation items
#'
#' Create nav item(s) for use inside nav containers (e.g., [navset_tab()],
#' [navset_bar()], etc).
#'
#' @param title A title to display. Can be a character string or UI elements
#'   (i.e., [htmltools::tags]).
#' @param ... Depends on the function:
#'   * For `nav_panel()` and `nav_panel_hidden()`: UI elements (i.e.,
#'     [htmltools::tags]) to display when the item is active.
#'   * For `nav_menu()`: a collection of nav items (e.g., `nav_panel()`,
#'     `nav_item()`).
#'   * For `nav_item()`: UI elements (i.e., [htmltools::tags]) to place directly in
#'     the navigation panel (e.g., search forms, links to external content, etc).
#' @param value A character string to assign to the nav item. This value may be
#'   supplied to the relevant container's `selected` argument in order to show
#'   particular nav item's content immediately on page load. This value is also
#'   useful for programmatically updating the selected content via
#'   [nav_select()], [nav_hide()], etc (updating selected tabs this way is often
#'   useful for showing/hiding panels of content via other UI controls like
#'   [shiny::radioButtons()] -- in this scenario, consider using
#'   [nav_panel_hidden()] with [navset_hidden()]).
#' @param icon Optional icon to appear next to the nav item's `title`.
#'
#' @return A nav item that may be passed to a nav container (e.g. [navset_tab()]).
#'
#' @seealso [navset] create the navigation container holding the nav panels.
#' @seealso [nav_menu()], [nav_item()], [nav_spacer()] create menus, items, or
#'   space in the navset control area.
#' @seealso [nav_insert()], [nav_remove()] programmatically add or remove nav
#'   panels.
#' @seealso [nav_select()], [nav_show()], [nav_hide()] change the state of a
#'   [nav_panel()] in a navset.
#'
#' @family Panel container functions
#' @name nav-items
NULL

#' @describeIn nav-items Content to display when the given item is selected.
#' @export
nav_panel <- function(title, ..., value = title, icon = NULL) {
  tabPanel_(title, ..., value = value, icon = icon)
}

#' @describeIn nav-items Create nav content for use inside `navset_hidden()` (for
#'   creating custom navigation controls via `navs_select()`),
#' @export
nav_panel_hidden <- function(value, ..., icon = NULL) {
  tabPanelBody_(value, ..., icon = icon)
}

#' @describeIn nav-items Create a menu of nav items.
#' @param align horizontal alignment of the dropdown menu relative to dropdown toggle.
#' @export
nav_menu <- function(
  title,
  ...,
  value = title,
  icon = NULL,
  align = c("left", "right")
) {
  align <- match.arg(align)
  navbarMenu_(title, ..., menuName = value, icon = icon, align = align)
}

#' @describeIn nav-items Place arbitrary content in the navigation panel (e.g., search
#'   forms, links to external content, etc.)
#' @export
nav_item <- function(...) {
  # TODO: drop form-inline since BS5 dropped it?
  # If we do that do we need navset_bar() to generate valid BS5 markup?
  tags$li(class = "bslib-nav-item nav-item form-inline", ...)
}

is_nav_item <- function(x) {
  tag_has_class(x, "bslib-nav-item")
}

#' @describeIn nav-items Adding spacing between nav items.
#' @export
nav_spacer <- function() {
  div(class = "bslib-nav-spacer", component_dependencies())
}

is_nav_spacer <- function(x) {
  tag_has_class(x, "bslib-nav-spacer")
}

tag_has_class <- function(x, class) {
  if (!inherits(x, "shiny.tag")) {
    return(FALSE)
  }
  tagQuery(x)$hasClass(class)
}

as_fragment <- function(x, page = page_fluid) {
  stopifnot(is.function(page) && "theme" %in% names(formals(page)))
  attr(x, "bslib_page") <- page
  class(x) <- c("bslib_fragment", class(x))
  x
}

as_page <- function(x) {
  class(x) <- c("bslib_page", class(x))
  x
}

#' Make HTML browsable by default
#'
#' @param x a [tag()] object.
#' @param ... passed along to an underlying print method
#' @export
#' @keywords internal
#' @rdname html-browse
print.bslib_fragment <- function(x, ...) {
  x <- attr(x, "bslib_page")(x)
  invisible(print(x, ...))
}

#' @rdname html-browse
#' @export
print.bslib_page <- function(x, ...) {
  if (interactive()) {
    x <- htmltools::browsable(x)
  }
  invisible(NextMethod())
}

#' Add low-level theming customizations
#'
#' Compared to higher-level theme customization available in [bs_theme()], these functions
#' are a more direct interface to Bootstrap Sass, and therefore, do nothing to
#' ensure theme customizations are portable between major Bootstrap versions.
#'
#' @inheritParams bs_theme_update
#' @param ...
#'  * `bs_add_variables()`: Should be named Sass variables or values that can be passed in directly to the `defaults` argument of a [sass::sass_layer()].
#'  * `bs_bundle()`: Should be arguments that can be handled by [sass::sass_bundle()] to be appended to the `theme`
#' @param .where Whether to place the variable definitions before other Sass
#'   `"defaults"`, after other Sass `"declarations"`, or after other Sass
#'   `"rules"`.
#' @param .default_flag Whether or not to add a `!default` flag (if missing) to
#'   variable expressions. It's recommended to keep this as `TRUE` when `.where
#'   = "defaults"`.
#'
#' @return a modified [bs_theme()] object.
#'
#' @references \url{https://getbootstrap.com/docs/4.4/getting-started/theming/}
#' @references \url{https://rstudio.github.io/sass/articles/sass.html#layering}
#' @examples
#'
#' # Function to preview the styling a (primary) Bootstrap button
#' library(htmltools)
#' button <- tags$a(class = "btn btn-primary", href = "#", role = "button", "Hello")
#' preview_button <- function(theme) {
#'   if (interactive()) {
#'     browsable(tags$body(bs_theme_dependencies(theme), button))
#'   }
#' }
#'
#' # Here we start with a theme based on a Bootswatch theme,
#' # then override some variable defaults
#' theme <- bs_add_variables(
#'   bs_theme(bootswatch = "sketchy", primary = "orange"),
#'   "body-bg" = "#EEEEEE",
#'   "font-family-base" = "monospace",
#'   "font-size-base" = "1.4rem",
#'   "btn-padding-y" = ".16rem",
#'   "btn-padding-x" = "2rem"
#' )
#'
#' preview_button(theme)
#'
#' # If you need to set a variable based on another Bootstrap variable
#' theme <- bs_add_variables(theme, "body-color" = "$success", .where = "declarations")
#' preview_button(theme)
#'
#' # Start a new global theme and add some custom rules that
#' # use Bootstrap variables to define a custom styling for a
#' # 'person card'
#' person_rules <- system.file("custom", "person.scss", package = "bslib")
#' theme <- bs_add_rules(bs_theme(), sass::sass_file(person_rules))
#' # Include custom CSS that leverages bootstrap Sass variables
#' person <- function(name, title, company) {
#'   tags$div(
#'     class = "person",
#'     h3(class = "name", name),
#'     div(class = "title", title),
#'     div(class = "company", company)
#'   )
#' }
#' if (interactive()) {
#'   browsable(shiny::fluidPage(
#'     theme = theme,
#'     person("Andrew Carnegie", "Owner", "Carnegie Steel Company"),
#'     person("John D. Rockefeller", "Chairman", "Standard Oil")
#'   ))
#' }
#'
#' @export
#' @describeIn bs_bundle Add Bootstrap Sass [variable defaults](https://getbootstrap.com/docs/4.4/getting-started/theming/#variable-defaults)
bs_add_variables <- function(theme, ..., .where = "defaults", .default_flag = identical(.where, "defaults")) {
  assert_bs_theme(theme)

  vars <- rlang::list2(...)
  if (any(names2(vars) == "")) stop("Variables must be named.", call. = FALSE)

  # Workaround to the problem of 'blue' winning in the scenario of:
  # bs_add_variables("body-bg" = "blue")
  # bs_add_variables("body-bg" = "red")
  if (.default_flag) {
    vars <- ensure_default_flag(vars)
  }

  bs_bundle(
    theme, do.call(sass_layer, rlang::list2(!!.where := vars))
  )
}

# Given a named list of variable definitions,
# searches each variable's expression for a !default flag,
# and if missing, adds it.
ensure_default_flag <- function(x) {
  Map(
    x, rlang::names2(x),
    f = function(val, nm) {
      # sass::font_collection() has its own default_flag, so warn if they conflict
      if (sass::is_font_collection(val)) {
        if (identical(val$default_flag, FALSE)) {
          message(
            "Ignoring `bs_add_variables()`'s `.default_flag = TRUE` for ",
            "the ", nm, " variable (since it has it's own `default_flag`)."
          )
        }
        return(val)
      }
      val <- paste(as_sass(val), collapse = "\n")
      if (grepl("!default\\s*;*\\s*$", val)) {
        val
      } else {
        paste(sub(";+$", "", val), "!default")
      }
    }
  )
}

#' @describeIn bs_bundle Add additional [Sass rules](https://sass-lang.com/documentation/style-rules)
#' @param rules Sass rules. Anything understood by [sass::as_sass()] may be
#'   provided (e.g., a list, character vector, [sass::sass_file()], etc)
#' @export
bs_add_rules <- function(theme, rules) {
  bs_bundle(theme, sass_layer(rules = rules))
}

#' @describeIn bs_bundle Add additional [Sass
#'   functions](https://rstudio.github.io/sass/articles/sass.html#functions-1)
#' @param functions A character vector or [sass::sass_file()] containing
#'   functions definitions.
#' @export
bs_add_functions <- function(theme, functions) {
  bs_bundle(theme, sass_layer(functions = functions))
}

#' @describeIn bs_bundle Add additional [Sass
#'   mixins](https://rstudio.github.io/sass/articles/sass.html#mixins-1)
#' @param mixins A character vector or [sass::sass_file()] containing
#'   mixin definitions.
#' @export
bs_add_mixins <- function(theme, mixins) {
  bs_bundle(theme, sass_layer(mixins = mixins))
}

#' @describeIn bs_bundle Add additional [sass::sass_bundle()] objects to an existing `theme`.
#' @export
bs_bundle <- function(theme, ...) {
  assert_bs_theme(theme)
  structure(
    sass_bundle(theme, ...),
    class = class(theme)
  )
}

test_that("bs_theme_update() can update defaults", {
  vars <- c("black", "white", "primary", "danger", "font-family-base", "font-family-monospace")
  pre_theme <- bs_theme("4")
  pre_vars <- bs_get_variables(pre_theme, vars)
  post_theme <- bs_theme_update(pre_theme, bg = "black", fg = "white", primary = "#222222", base_font = "SomeFont")
  post_vars <- bs_get_variables(post_theme, vars)
  # bg/fg has been inverted, primary and base font have been updated
  expect_true(post_vars[["black"]] == "#FFFFFF")
  expect_true(post_vars[["white"]] == "#000000")
  expect_true(post_vars[["primary"]] == "#222222")
  expect_true(post_vars[["font-family-base"]] == "SomeFont")
  # Danger hasn't been set
  expect_true(post_vars[["danger"]] == pre_vars[["danger"]])
  # Neither has monospace
  expect_true(post_vars[["font-family-monospace"]] == pre_vars[["font-family-monospace"]])
})


test_that("Sass bundles work as expected with a theme", {
  theme <- bs_add_variables(bs_theme("4"), primary = "#222222")
  expect_true(bs_get_variables(theme, "primary") == "#222222")
  # declarations can be used in rules
  theme <- bs_add_mixins(theme, list(foo = "bar !default"))
  expect_identical(
    as.character(sass_partial(".foo {color: $foo}", theme)),
    ".foo {\n  color: bar;\n}\n"
  )
  # but declarations come after defaults (so this won't override the value!)
  theme <- bs_add_mixins(theme, list(primary = "#333333 !default"))
  expect_true(bs_get_variables(theme, "primary") == "#222222")
  # Can drop-down to the lower-level bs_bundle()
  theme <- bs_bundle(
    theme, sass::sass_layer(defaults = list(primary = "#333333 !default"))
  )
  expect_true(bs_get_variables(theme, "primary") == "#333333")
})

test_that("bs_theme_update() can update the bootswatch theme", {
  darkly <- bs_theme(bootswatch = "darkly")
  cosmo <- bs_theme(bootswatch = "cosmo")

  expect_false(identical(darkly, cosmo))
  darkly2 <- bs_theme_update(cosmo, bootswatch = "darkly")
  expect_identical(
    sass::sass(darkly),
    sass::sass(darkly2)
  )
  cyborg <- bs_theme(bootswatch = "cyborg")
  cyborg2 <- bs_theme_update(bs_theme(), bootswatch = "cyborg")
  expect_identical(
    sass::sass(cyborg),
    sass::sass(cyborg2)
  )
  default <- bs_theme_update(cosmo, bootswatch = "default")
  expect_identical(
    sass::sass(default),
    sass::sass(bs_theme())
  )
  expect_null(theme_bootswatch(default))
})

test_that("is_bs_theme() works", {
  expect_true(is_bs_theme(bs_theme()))
  expect_true(is_bs_theme(bs_theme(3)))
  expect_true(is_bs_theme(bs_theme(4)))
})

test_that("bs_theme_update() can update defaults", {
  vars <- c(
    "black",
    "white",
    "primary",
    "danger",
    "font-family-base",
    "font-family-monospace"
  )
  pre_theme <- bs_theme("4")
  pre_vars <- bs_get_variables(pre_theme, vars)
  post_theme <- bs_theme_update(
    pre_theme,
    bg = "black",
    fg = "white",
    primary = "#222222",
    base_font = "SomeFont"
  )
  post_vars <- bs_get_variables(post_theme, vars)
  # bg/fg has been inverted, primary and base font have been updated
  expect_true(post_vars[["black"]] == "#FFFFFF")
  expect_true(post_vars[["white"]] == "#000000")
  expect_true(post_vars[["primary"]] == "#222222")
  expect_true(post_vars[["font-family-base"]] == "SomeFont")
  # Danger hasn't been set
  expect_true(post_vars[["danger"]] == pre_vars[["danger"]])
  # Neither has monospace
  expect_true(
    post_vars[["font-family-monospace"]] == pre_vars[["font-family-monospace"]]
  )
})

test_that("Sass bundles work as expected with a theme", {
  theme <- bs_add_variables(bs_theme("4"), primary = "#222222")
  expect_true(bs_get_variables(theme, "primary") == "#222222")
  # declarations can be used in rules
  theme <- bs_add_mixins(theme, list(foo = "bar !default"))
  expect_identical(
    as.character(sass_partial(".foo {color: $foo}", theme)),
    ".foo {\n  color: bar;\n}\n"
  )
  # but declarations come after defaults (so this won't override the value!)
  theme <- bs_add_mixins(theme, list(primary = "#333333 !default"))
  expect_true(bs_get_variables(theme, "primary") == "#222222")
  # Can drop-down to the lower-level bs_bundle()
  theme <- bs_bundle(
    theme,
    sass::sass_layer(defaults = list(primary = "#333333 !default"))
  )
  expect_true(bs_get_variables(theme, "primary") == "#333333")
})

test_that("bs_theme_update() can update the bootswatch theme", {
  darkly <- bs_theme(bootswatch = "darkly")
  cosmo <- bs_theme(bootswatch = "cosmo")

  expect_false(identical(darkly, cosmo))
  darkly2 <- bs_theme_update(cosmo, bootswatch = "darkly")
  expect_identical(
    sass::sass(darkly),
    sass::sass(darkly2)
  )
  cyborg <- bs_theme(bootswatch = "cyborg")
  cyborg2 <- bs_theme_update(bs_theme(), bootswatch = "cyborg")
  expect_identical(
    sass::sass(cyborg),
    sass::sass(cyborg2)
  )
  default <- bs_theme_update(cosmo, bootswatch = "default")
  expect_identical(
    sass::sass(default),
    sass::sass(bs_theme(preset = "bootstrap"))
  )
  expect_null(theme_bootswatch(default))
})

test_that("is_bs_theme() works", {
  expect_true(is_bs_theme(bs_theme()))
  expect_true(is_bs_theme(bs_theme(3)))
  expect_true(is_bs_theme(bs_theme(4)))
})

test_that("Using bootswatch arg is equivalent to bs_bundle()", {
  cosmo_primary <- sass_partial("body{color:$primary;}", as_bs_theme("cosmo@4"))
  expect_css("body{color:#2780e3;}", cosmo_primary)
})

bs3_sass_files <- function(x) {
  bs_sass_files(x, version = 3)
}

bs4_sass_files <- function(x) {
  bs_sass_files(x, version = 4)
}

bs5_sass_files <- function(x) {
  bs_sass_files(x, version = 5)
}

bs_sass_files <- function(files, version) {
  lapply(files, bs_sass_file, version = version)
}

# Search for one file at a time so we can throw informative errors
bs_sass_file <- function(file, version) {
  if (length(file) != 1) {
    stop("file should be of length 1")
  }

  file <- file.path(
    dirname(file),
    paste0("_", basename(file), ".scss")
  )

  f <- switch_version(
    version,
    five = path_lib("bs5", "scss", file),
    four = path_lib("bs4", "scss", file),
    three = path_lib("bs3", "assets", "stylesheets", "bootstrap", file)
  )
  if (f == "") {
    stop("The bootstrap stylesheet '", file, "' doesn't exist.", call. = FALSE)
  }
  sass_file(f)
}

bootswatch_sass_file <- function(theme, file, version = version_default()) {
  if (length(file) > 1) {
    stop("file should be of length 1")
  }
  theme <- match.arg(theme, bootswatch_themes(version))
  file <- paste0("_", file, ".scss")
  f <- file.path(bootswatch_dist(version), theme, file)
  if (file.exists(f)) {
    return(sass_file(f))
  }
  stop(
    "Bootswatch file '",
    file,
    "' doesn't exist for theme '",
    theme,
    "'.",
    call. = FALSE
  )
}

# Given a vector of sass_file()s, create a list of sass_bundles(),
# so each rule may be removed layer (by its files basename)
rule_bundles <- function(files) {
  files <- lapply(files, as_sass_file)
  paths <- vapply(files, get_sass_file_path, character(1))
  nms <- tools::file_path_sans_ext(basename(paths))
  Map(
    nms,
    files,
    f = function(nm, f) {
      sass_bundle(!!nm := sass_layer(rules = f))
    }
  )
}

get_sass_file_path <- function(x) {
  path <- attr(x, "sass_file_path")
  if (length(path)) {
    return(path)
  }

  stop("Couldn't find file path")
}

as_sass_file <- function(x) {
  if (inherits(x, "sass_file")) x else sass_file(x)
}

#' Obtain a list of all available bootswatch themes.
#'
#' @param version the major version of Bootswatch.
#' @param full_path whether to return a path to the installed theme.
#' @export
#' @return a character vector of Bootswatch themes.
bootswatch_themes <- function(version = version_default(), full_path = FALSE) {
  list.dirs(bootswatch_dist(version), full.names = full_path, recursive = FALSE)
}

#' Obtain a theme's Bootswatch theme name
#'
#' @inheritParams bs_theme_update
#' @return the Bootswatch theme named used (if any) in the `theme`.
#' @export
theme_bootswatch <- function(theme) {
  if (!is_bs_theme(theme)) return(NULL)
  info <- theme_preset_info(theme)
  if (identical("bootswatch", info$type)) info$name else NULL
}

#' Obtain a theme's Bootstrap version
#'
#' @inheritParams bs_theme_update
#' @return the major version of Bootstrap used in the `theme`.
#' @export
theme_version <- function(theme) {
  if (!is_bs_theme(theme)) return(NULL)

  version <- grep("^bs_version_", class(theme), value = TRUE)
  if (length(version) == 1) {
    return(sub("^bs_version_", "", version))
  }

  # If no version class, or more than one class, we consult the sass bundle
  theme_preset_info(theme)$version
}


bootswatch_dist <- function(version) {
  switch_version(
    version,
    five = path_lib("bsw5", "dist"),
    four = path_lib("bsw4", "dist"),
    three = path_lib("bsw3")
  )
}


# -----------------------------------------------------------------
# Bootswatch bundle
# -----------------------------------------------------------------

bootswatch_bundle <- function(bootswatch, version) {
  if (!length(bootswatch) || isTRUE(bootswatch %in% c("default", "bootstrap"))) {
    return(NULL)
  }

  bootswatch <- switch_version(
    version,
    default = {
      switch(
        bootswatch,
        paper = {
          message("Bootswatch 3 theme paper has been renamed to materia in version 4 (using that theme instead)")
          "materia"
        },
        readable = {
          message("Bootswatch 3 theme readable has been renamed to litera in version 4 (using that theme instead)")
          "litera"
        },
        match.arg(bootswatch, bootswatch_themes(version))
      )
    },
    three = match.arg(bootswatch, bootswatch_themes(version))
  )

  # Attach local font files, if necessary
  font_css <- file.path(bootswatch_dist(version), bootswatch, "font.css")
  attachments <- if (file.exists(font_css)) {
    c(
      "font.css" = font_css,
      fonts = system_file("fonts", package = "bslib")
    )
  }

  sass_bundle(
    bootswatch = sass_layer(
      file_attachments = attachments,
      defaults = list(
        "bslib-preset-type" = "bootswatch",
        "bslib-preset-name" = bootswatch,
        # Use local fonts (this path is relative to the bootstrap HTML dependency dir)
        '$web-font-path: "font.css" !default;',
        bootswatch_sass_file(bootswatch, "variables", version),
        # Unless we change navbarPage()'s markup, BS4+ will likely want BS3 compatibility
        switch_version(
          version, three = "", default = bs3compat_navbar_defaults(bootswatch)
        )
      ),
      rules = list(
        bootswatch_sass_file(bootswatch, "bootswatch", version),
        # For some reason sketchy sets .dropdown-menu{overflow: hidden}
        # but this prevents .dropdown-submenu from working properly
        # https://github.com/rstudio/bootscss/blob/023d455/inst/node_modules/bootswatch/dist/sketchy/_bootswatch.scss#L204
        if (identical(bootswatch, "sketchy")) ".dropdown-menu{ overflow: inherit; }" else "",
        # Several Bootswatch themes (e.g., zephyr, simplex, etc) add custom .btn-secondary
        # rules that should also apply to .btn-default
        ".btn-default:not(.btn-primary):not(.btn-info):not(.btn-success):not(.btn-warning):not(.btn-danger):not(.btn-dark):not(.btn-light):not([class*='btn-outline-']) {
          @extend .btn-secondary !optional;
        }"
      )
    )
  )
}


# Mappings from BS3 navbar classes to BS4
bs3compat_navbar_defaults <- function(bootswatch) {
  # Do nothing if this isn't a Bootswatch 3 theme
  if (!bootswatch %in% c("materia", "litera", bootswatch_themes(3))) {
    return("")
  }

  bg_colors <- switch(
    bootswatch,
    cerulean = c("primary", "info"),
    cosmo = c("dark", "primary"),
    cyborg = c("body-bg", "secondary"),
    darkly = c("primary", "success"),
    flatly = c("primary", "success"),
    journal = c("light", "primary"),
    lumen = c("light", "white"),
    # i.e., materia
    paper = ,
    materia = c("light", "primary"),
    readable = ,
    litera = c("light", "dark"),
    sandstone = c("dark", "success"),
    simplex = c("light", "primary"),
    slate = c("primary", "light"),
    spacelab = c("light", "primary"),
    superhero = c("dark", "primary"),
    united = c("primary", "dark"),
    yeti = c("dark", "primary"),
    stop("Didn't recognize Bootswatch 3 theme: ", bootswatch, call. = FALSE)
  )


  list(
    sprintf('$navbar-light-bg: $%s !default;', bg_colors[1]),
    sprintf('$navbar-dark-bg: $%s !default;', bg_colors[2])
  )
}

test_that("Can retrieve version from theme object", {
  expect_identical(
    theme_version(bs_theme()),
    version_default()
  )
  expect_equal(
    theme_version(bs_theme(version = "3")), "3"
  )
  expect_identical(
    theme_version(bs_theme(version = "4")), "4"
  )
  deps <- bs_theme_dependencies(bs_theme(version = 4))
  # jQuery, Bootstrap, and bs3compat layer
  expect_length(deps, 3)
})

test_that("Can retrieve version from theme object", {
  expect_identical(
    theme_version(bs_theme()),
    version_default()
  )
  expect_equal(
    theme_version(bs_theme(version = "3")),
    "3"
  )
  expect_identical(
    theme_version(bs_theme(version = "4")),
    "4"
  )
  deps <- bs_theme_dependencies(bs_theme(version = 4))
  # jQuery, Bootstrap, and bs3compat layer
  expect_length(deps, 3)
})

#' @export
#' @inheritParams card
#' @inheritParams card_body
#' @param title A (left-aligned) title to place in the card header/footer. If
#'   provided, other nav items are automatically right aligned.
#' @include navs-legacy.R
#' @rdname navset
navset_card_tab <- function(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  sidebar = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  full_screen = FALSE,
  wrapper = card_body
) {
  navset_card(
    navset_tab,
    "card-header-tabs",
    ...,
    id = id,
    selected = selected,
    title = title,
    sidebar = sidebar,
    header = header,
    footer = footer,
    height = height,
    full_screen = full_screen,
    wrapper = wrapper
  )
}

#' @export
#' @param placement placement of the nav items relative to the content.
#' @rdname navset
navset_card_pill <- function(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  sidebar = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  placement = c("above", "below"),
  full_screen = FALSE,
  wrapper = card_body
) {
  items <- collect_nav_items(..., wrapper = wrapper)

  pills <- navset_pill(!!!items, id = id, selected = selected)

  above <- match.arg(placement) == "above"

  pillQ <- tagQuery(pills)

  # https://getbootstrap.com/docs/5.0/components/card/#navigation
  nav <- pillQ$children(".nav")$addClass(
    if (above) "card-header-pills"
  )$selectedTags()[[1]]
  content <- pillQ$children(".tab-content")$selectedTags()[[1]]

  nav_args <- if (!is.null(title)) {
    list(class = "bslib-navs-card-title", tags$span(title), nav)
  } else {
    list(nav)
  }

  card(
    height = height,
    full_screen = full_screen,
    if (above) card_header(!!!nav_args),
    header,
    navs_card_body(content, sidebar),
    footer,
    if (!above) card_footer(!!!nav_args)
  )
}

#' @export
#' @rdname navset
navset_card_underline <- function(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  sidebar = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  full_screen = FALSE,
  wrapper = card_body
) {
  navset_card(
    navset_underline,
    "nav-underline",
    ...,
    id = id,
    selected = selected,
    title = title,
    sidebar = sidebar,
    header = header,
    footer = footer,
    height = height,
    full_screen = full_screen,
    wrapper = wrapper
  )
}

navset_card <- function(
  navset_func,
  nav_class,
  ...,
  id,
  selected,
  title,
  sidebar,
  header,
  footer,
  height,
  full_screen,
  wrapper
) {
  items <- collect_nav_items(..., wrapper = wrapper)

  tabs <- navset_func(!!!items, id = id, selected = selected)

  tabQ <- tagQuery(tabs)

  # https://getbootstrap.com/docs/5.3/components/card/#navigation
  nav <- tabQ$children(".nav")$addClass(nav_class)$selectedTags()[[1]]
  content <- tabQ$children(".tab-content")$selectedTags()[[1]]

  card(
    height = height,
    full_screen = full_screen,
    if (!is.null(title)) {
      card_header(class = "bslib-navs-card-title", tags$span(title), nav)
    } else {
      card_header(nav)
    },
    header,
    navs_card_body(content, sidebar),
    footer
  )
}

collect_nav_items <- function(..., wrapper) {
  items <- rlang::list2(...)

  # Wrap any nav_panel() children up into card items
  nav_to_card_item <- function(x) {
    if (isNavbarMenu(x)) {
      x$tabs <- lapply(x$tabs, nav_to_card_item)
    }
    if (isTabPanel(x)) {
      x$children <- as_card_items(x$children, wrapper = wrapper)
    }
    x
  }

  lapply(items, nav_to_card_item)
}

# Always give tab contents the potential to fill since that's akin to the
# normal card() API (i.e. the card() is a fill container) and users have
# option to make the contents fill via card_body(fill = TRUE, fillable = TRUE)
navs_card_body <- function(content, sidebar) {
  content <- makeTabsFillable(content, fillable = TRUE, gap = 0, padding = 0)
  if (!is.null(sidebar)) {
    content <- layout_sidebar(
      sidebar = sidebar,
      fillable = TRUE,
      border = FALSE,
      content
    )
  }
  as.card_item(content)
}

# Given a .tab-content container, mark each relevant .tab-pane as a
# fill container/item.
makeTabsFillable <- function(
  content,
  fillable = TRUE,
  navbar = FALSE,
  gap = NULL,
  padding = NULL
) {
  if (
    !inherits(content, "shiny.tag") ||
      !tagQuery(content)$hasClass("tab-content")
  ) {
    abort("Expected `content` to be a tag with a tab-content class")
  }

  if (isFALSE(fillable)) {
    return(content)
  }

  # Even if only one .tab-pane wants fillable behavior, the .tab-content
  # must to be a fillable container.
  content <- bindFillRole(content, container = TRUE, item = TRUE)

  tagQuery(content)$find(".tab-pane")$each(function(x, i) {
    if (
      isTRUE(fillable) || isTRUE(tagGetAttribute(x, "data-value") %in% fillable)
    ) {
      x <- tagAppendAttributes(
        class = "bslib-gap-spacing",
        style = css(
          # Remove the margin between nav and content (for page_navbar())
          "--bslib-navbar-margin" = if (navbar) 0,
          gap = gap,
          padding = padding
        ),
        bindFillRole(x, container = TRUE, item = TRUE)
      )
    }

    x
  })$allTags()
}

label_with_extras <- function(label, ...) {
  div(
    class = "d-inline-block w-100",
    span(label),
    span(class = "float-right", ...),
    singleton(tags$style(HTML(".shiny-input-container .control-label { width: 100%; }")))
  )
}

ui_selextra <- function(id, label) {
  ns <- shiny::NS(id)

  selectizeInput(
    inputId = ns("selected"),
    choices = NULL,
    label_with_extras(
      label,
      actionLink(
        ns("shuffle"),
        bsicons::bs_icon("shuffle", title = paste("Random", label))
      ),
      actionLink(
        ns("prev"),
        bsicons::bs_icon("caret-left-fill", title = paste("Previous", label))
      ),
      actionLink(
        ns("next"),
        bsicons::bs_icon("caret-right-fill", title = paste("Next", label))
      )
    )
  )
}

server_selextra <- function(input, output, session, choices) {
  ns <- session$ns

  # These are server-side selectize inputs, so we update them on initialization
  updateSelectizeInput(session, "selected", choices = choices, server = TRUE)

  trigger_shuffle <- reactiveVal(0)
  trigger_next <- reactiveVal(0)
  trigger_prev <- reactiveVal(0)

  observeEvent(
    input$shuffle,
    ignoreInit = TRUE,
    trigger_shuffle(trigger_shuffle() + 1)
  )

  observeEvent(trigger_shuffle(), {
    req(trigger_shuffle() > 0)

    updateSelectizeInput(
      session,
      "selected",
      selected = sample(unlist(choices), 1),
      choices = choices,
      server = TRUE
    )
  })

  observeEvent(input[["next"]], move(1))
  observeEvent(trigger_next(), move(1), ignoreInit = TRUE)

  observeEvent(input[["prev"]], move(-1))
  observeEvent(trigger_prev(), move(-1), ignoreInit = TRUE)

  move <- reactiveVal(0)

  observeEvent(move(), {
    req(move() != 0)

    current <- input$selected
    choices_flat <- unlist(choices)

    idx <- which(choices_flat == current) + move()
    move(0)
    req(idx)

    if (idx > length(choices_flat)) idx <- 1
    if (idx <= 0) idx <- length(choices_flat)

    updateSelectizeInput(
      session,
      "selected",
      selected =  choices_flat[[idx]],
      choices = choices,
      server = TRUE
    )
  })

  list(
    "value" = reactive(input$selected %||% ""),
    "shuffle" = function() trigger_shuffle(as.integer(Sys.time())),
    "next" = function() trigger_next(as.integer(Sys.time())),
    "prev" = function() trigger_prev(as.integer(Sys.time())),
    "choices" = choices,
    "set" = function(value) {
      updateSelectizeInput(
        session,
        "selected",
        selected = value,
        choices = choices,
        server = TRUE
      )
    }
  )
}

module_selextra <- function(id, choices) {
  callModule(server_selextra, id, choices = choices)
}

#' Available Bootstrap versions
#'
#' @return A list of the Bootstrap versions available.
#'
#' @export
versions <- function() {
  rlang::set_names(
    list("5", "4", "3"),
    c(version_bs5, version_bs4, version_bs3)
  )
}

#' @export
#' @rdname versions
version_default <- function() {
  versions()[[1]]
}

# TODO: make this a getter/setter?

theme_colors <- list(
  "primary",
  "secondary",
  "success",
  "danger",
  "warning",
  "info",
  "light",
  "dark"
)

named_colors <- c(
  "blue",
  "indigo",
  "purple",
  "pink",
  "red",
  "orange",
  "yellow",
  "green",
  "teal",
  "cyan"
)

gc <- expand.grid(named_colors, named_colors)
gc <- gc[gc$Var1 != gc$Var2,]
gradient_classes <- sprintf("bg-gradient-%s-%s", gc$Var1, gc$Var2)


all_themes <- c(
  "Default" = "",
  theme_colors,
  named_colors,
  paste0("text-", theme_colors),
  paste0("text-", named_colors),
  sort(gradient_classes)
)

tag_require <- function(tag, version = 4, caller = "") {
  tagAddRenderHook(
    tag, replace = FALSE,
    func = function(x) {
      current_version <- theme_version(bs_current_theme()) %||% 3
      if (isTRUE(current_version >= version))
        return(x)

      stop(
        caller, " requires Bootstrap ", version, " or higher. ",
        "Please supply `bslib::bs_theme(version = ", version,
        ")` to the UI's page layout function.",
        call. = FALSE
      )
    }
  )
}

is_tag <- function(x) {
  inherits(x, "shiny.tag")
}

#' Switch input control
#'
#' @description
#' Create an on-off style switch control for specifying logical values.
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fixed(
#'   title = "Keyboard Settings",
#'   h2("Keyboard Settings"),
#'   input_switch("auto_capitalization", "Auto-Capitalization", TRUE),
#'   input_switch("auto_correction", "Auto-Correction", TRUE),
#'   input_switch("check_spelling", "Check Spelling", TRUE),
#'   input_switch("smart_punctuation", "Smart Punctuation"),
#'   h2("Preview"),
#'   verbatimTextOutput("preview")
#' )
#'
#' server <- function(input, output, session) {
#'   output$preview <- renderPrint({
#'     list(
#'       auto_capitalization = input$auto_capitalization,
#'       auto_correction = input$auto_correction,
#'       check_spelling = input$check_spelling,
#'       smart_punctuation = input$smart_punctuation
#'     )
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @param id An input id.
#' @param label A label for the switch.
#' @param value Whether or not the switch should be checked by default.
#' @param width Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `width="200px"`).
#'
#' @return Returns a UI element for a switch input control. The server value
#'   received for the input corresponding to `id` will be a logical
#'   (`TRUE`/`FALSE`) value.
#'
#' @family input controls
#' @export
input_switch <- function(id, label, value = FALSE, width = NULL) {
  value <- shiny::restoreInput(id, default = value)
  tag <- input_checkbox(
    id,
    label,
    class = "bslib-input-switch form-switch",
    value = value,
    width = width
  )
  tag <- tag_require(tag, version = 5, caller = "input_switch()")
  as_fragment(tag)
}

#' @rdname input_switch
#' @inheritParams nav_insert
#' @export
update_switch <- function(
  id,
  label = NULL,
  value = NULL,
  session = get_current_session()
) {
  message <- dropNulls(list(label = label, value = value))
  session$sendInputMessage(id, message)
}

#' @rdname input_switch
#' @inheritParams toggle_sidebar
#' @export
toggle_switch <- function(id, value = NULL, session = get_current_session()) {
  if (!is.null(value) && !rlang::is_logical(value, n = 1)) {
    abort("`value` must be a `NULL` or a single logical value.")
  }

  if (!rlang::is_string(id)) {
    abort("`id` must be a single string containing the `id` of a switch input.")
  }

  msg <- dropNulls(list(id = session$ns(id), value = value))

  callback <- function() {
    session$sendCustomMessage("bslib.toggle-input-binary", msg)
  }

  session$onFlush(callback, once = TRUE)
}

input_checkbox <- function(
  id,
  label,
  class = "bslib-input-checkbox",
  value = FALSE,
  width = NULL,
  inline = FALSE
) {
  div(
    class = "form-group shiny-input-container",
    class = if (inline) "shiny-input-container-inline",
    style = css(width = width),
    div(
      class = class,
      class = "form-check",
      tags$input(
        id = id,
        class = "form-check-input",
        type = "checkbox",
        role = "switch",
        checked = if (value) NA,
      ),
      tags$label(
        # The span here is needed to adhere to shiny.js' checkbox binding logic
        # https://github.com/rstudio/shiny/blob/c21ba0b/srcts/src/bindings/input/checkbox.ts#L42-L43
        tags$span(label),
        class = "form-check-label",
        `for` = id
      ),
      component_dependencies()
    )
  )
}

utils::globalVariables("!!")

#' @import htmltools
#' @import sass
#' @importFrom utils modifyList packageVersion download.file URLencode getFromNamespace
#' @importFrom stats setNames na.omit
#' @importFrom grDevices col2rgb
#' @importFrom tools file_path_sans_ext
#' @importFrom jquerylib jquery_core
#' @importFrom jsonlite fromJSON
#' @importFrom rlang := list2
NULL


#' Helpers for importing web fonts
#'
#' `font_google()`, `font_link()`, and `font_face()` are all re-exported from
#' the \pkg{sass} package (see [sass::font_face()] for details). For a quick
#' example of how to use these functions with [bslib::bs_theme()], see the
#' examples section below.
#'
#' @importFrom sass font_face
#' @export font_face
#' @name font_face
#' @examples
#'
#' # If you have an internet connection, running the following code
#' # will download, cache, and import the relevant Google Font files
#' # for local use
#' theme <- bs_theme(
#'   base_font = font_google("Fira Sans"),
#'   code_font = font_google("Fira Code"),
#'   heading_font = font_google("Fredoka One")
#' )
#' if (interactive()) {
#'   bs_theme_preview(theme)
#' }
#'
#' # Three different yet equivalent ways of importing a remotely-hosted Google Font
#' a <- font_google("Crimson Pro", wght = "200..900", local = FALSE)
#' b <- font_link(
#'   "Crimson Pro",
#'   href = "https://fonts.googleapis.com/css2?family=Crimson+Pro:wght@200..900"
#' )
#' url <- "https://fonts.gstatic.com/s/crimsonpro/v13/q5uDsoa5M_tv7IihmnkabARboYF6CsKj.woff2"
#' c <- font_face(
#'   family = "Crimson Pro",
#'   style = "normal",
#'   weight = "200 900",
#'   src = paste0("url(", url, ") format('woff2')")
#' )
#' theme <- bs_theme(base_font = c)
#' if (interactive()) {
#'   bs_theme_preview(theme)
#' }
NULL

#' @importFrom sass font_link
#' @export font_link
#' @name font_link
#' @rdname font_face
NULL

#' @importFrom sass font_google
#' @export font_google
#' @name font_google
#' @rdname font_face
NULL

#' @importFrom sass font_collection
#' @export font_collection
#' @name font_collection
#' @rdname font_face
NULL

#' Remove or retrieve Sass code from a theme
#'
#' A Bootstrap theme created with [bs_theme()] is comprised of
#' [many Sass layers](https://rstudio.github.io/sass/articles/sass.html#layering).
#' `bs_remove()` and `bs_retrieve()` allow you to remove or retrieve an
#' individual layer, either to reduce the size of the compiled CSS or to extract
#' styles from a theme.
#'
#' @inheritParams bs_theme_update
#' @param ids a character vector of ids
#'
#' @return Returns a modified [bs_theme()] object.
#'
#' @export
#' @family Bootstrap theme functions
#'
#' @examples
#' bs4 <- bs_theme(version = 4)
#'
#' # Retrieve sass bundle for print styles
#' bs_retrieve(bs4, "_print", include_unnamed = FALSE)
#'
#' # Remove CSS rules for print and carousels
#' bs4_no_print <- bs_remove(bs4, c("_print", "_carousel"))
#' suppressWarnings(
#'   bs_retrieve(bs4_no_print, "_print", include_unnamed = FALSE)
#' )
#'
#' # Remove BS3 compatibility layer
#' bs4_no_compat <- bs_remove(bs4, "bs3compat")
bs_remove <- function(theme, ids = character(0)) {
  ids <- retain_valid_ids(theme, ids)
  sass_bundle_remove(theme, ids)
}

#' @rdname bs_remove
#' @param include_unnamed whether or not to include unnamed [sass::sass_layer()]s
#' (e.g., Bootstrap Sass variables, functions, and mixins).
#' @export
bs_retrieve <- function(theme, ids = character(0), include_unnamed = TRUE) {
  ids <- retain_valid_ids(theme, ids)
  if (isTRUE(include_unnamed)) {
    ids <- c(ids, "")
  }
  layers <- theme$layers[names(theme$layers) %in% ids]
  structure(
    sass_bundle(!!!layers),
    class = class(theme)
  )
}

retain_valid_ids <- function(theme, ids) {
  assert_bs_theme(theme)
  stopifnot(is.character(ids))

  bundle_ids <- names(theme$layers)
  if (!length(ids)) {
    stop(
      "Provide one or more of the following `ids`: ",
      paste(bundle_ids, collapse = ", ")
    )
  }

  missing_ids <- setdiff(ids, bundle_ids)
  if (length(missing_ids)) {
    warning(
      "The following `ids` weren't found in `theme`: ",
      paste(missing_ids, collapse = ", "),
      ".\n\n",
      "Possible `ids` include: ",
      paste(bundle_ids[nzchar(bundle_ids)], collapse = ", ")
    )
  }

  intersect(ids, bundle_ids)
}

web_component <- function(tagName, ...) {
  js_dep <- component_dependency_js("webComponents", type = "module")
  args <- c(list(js_dep), rlang::list2(...))
  tag(tagName, args)
}

component_dependency_js <- function(name, ...) {
  minified <- get_shiny_devmode_option("shiny.minified", default = TRUE)

  htmlDependency(
    name = paste0("bslib-", name, "-js"),
    version = get_package_version("bslib"),
    package = "bslib",
    src = file.path("components", "dist", name),
    script = list(
      src = paste0(name, if (minified) ".min", ".js"),
      ...
    ),
    all_files = TRUE
  )
}

# Pre-compiled component styles
component_dependency_css <- function(name) {
  htmlDependency(
    name = paste0("bslib-", name, "-styles"),
    version = get_package_version("bslib"),
    package = "bslib",
    src = file.path("components", "dist", name),
    stylesheet = paste0(name, ".css")
  )
}

# Run-time (Sass) component styles
component_dependency_sass <- function(theme, name) {
  precompiled <- isTRUE(get_precompiled_option())
  default_theme <- !is_bs_theme(theme) || identical(theme, bs_theme())
  if (precompiled && default_theme) {
    component_dependency_css(name)
  } else {
    component_dependency_sass_(theme, name)
  }
}

component_dependency_sass_ <- function(theme, name) {
  scss_files <- list(
    path_components("scss", "mixins", "_mixins.scss"),
    path_components("scss", paste0(name, ".scss"))
  )

  bs_dependency(
    input = lapply(scss_files, sass_file),
    # At least currently, when statically rendering a component, 
    # bs_dependency_defer() passes along a NULL theme. We should 
    # eventually fix that, but for now, fall back to the default theme
    theme = theme %||% bs_theme(),
    name = paste0("bslib-", name, "-styles"),
    version = get_package_version("bslib"),
    cache_key_extra = get_package_version("bslib"),
    .sass_args = list(options = sass_options(output_style = "compressed"))
  )
}

#' Obtain the currently active theme at render time
#'
#' Intended for advanced use by developers to obtain the currently active theme
#' _at render time_ and primarily for implementing themable widgets that can't
#' otherwise be themed via [bs_dependency_defer()] .
#'
#' @details
#'
#' This function should generally only be called at print/render time. For
#' example:
#'
#'  * Inside the `preRenderHook` of `htmlwidgets::createWidget()`.
#'  * Inside of a custom [print] method that generates [htmltools::tags].
#'  * Inside of a [htmltools::tagFunction()]
#'
#' Calling this function at print/render time is important because it does
#' different things based on the context in which it's called:
#'
#' * If a reactive context is active, `session$getCurrentTheme()` is called
#'   (which is a reactive read).
#' * If no reactive context is active, `shiny::getCurrentTheme()` is called
#'   (which returns the current app's `theme`, if relevant).
#' * If `shiny::getCurrentTheme()` comes up empty, then `bs_global_get()`
#'   is called, which is relevant for [rmarkdown::html_document()], and
#'   possibly other static rendering contexts.
#'
#' @param session The current Shiny session (if any).
#'
#' @return Returns a [bs_theme()] object.
#'
#' @family Bootstrap theme functions
#' @export
bs_current_theme <- function(session = get_current_session(FALSE)) {
  # If we're able to make a reactive read of the theme, then do it
  if (has_valid_reactive_context(session)) {
    return(session$getCurrentTheme())
  }
  # Otherwise, this'll be a non-reactive read of session/app/global state
  get_current_theme() %||% bs_global_get()
}

#' Test and/or coerce fill behavior
#'
#' Filling layouts in bslib are built on the foundation of fillable containers
#' and fill items (fill carriers are both fillable and fill). This is why most
#' bslib components (e.g., [card()], [card_body()], [layout_sidebar()]) possess
#' both `fillable` and `fill` arguments (to control their fill behavior).
#' However, sometimes it's useful to add, remove, and/or test fillable/fill
#' properties on arbitrary [htmltools::tag()], which these functions are
#' designed to do.
#'
#' @references The [Filling Layouts](https://rstudio.github.io/bslib/articles/filling.html)
#'   article on the bslib website introduces the concept of fillable containers
#'   and fill items.
#'
#' @seealso These functions provide a convenient interface to the underlying
#'   [htmltools::bindFillRole()] function.
#'
#' @details
#' Although `as_fill()`, `as_fillable()`, and `as_fill_carrier()` can work with
#' non-tag objects that have a [htmltools::as.tags] method (e.g., htmlwidgets),
#' they return the "tagified" version of that object.
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' library(bslib)
#' shinyApp(
#'   page_fillable(
#'     # without `as_fill_carrier()`, the plot won't fill the page because
#'     # `uiOutput()` is neither a fillable container nor a fill item by default.
#'     as_fill_carrier(uiOutput("ui"))
#'   ),
#'   function(input, output) {
#'     output$ui <- renderUI({
#'       div(
#'         class = "bg-info text-white",
#'         as_fill_item(),
#'         "A fill item"
#'       )
#'     })
#'   }
#' )
#'
#' @return
#'   * For `as_fill()`, `as_fillable()`, and `as_fill_carrier()`: the _tagified_
#'     version `x`, with relevant tags modified to possess the relevant fill
#'     properties.
#'   * For `is_fill()`, `is_fillable()`, and `is_fill_carrier()`: a logical vector,
#'     with length matching the number of top-level tags that possess the relevant
#'     fill properties.
#'
#' @param x An [htmltools::tag()].
#' @param ... Currently ignored.
#' @param min_height,max_height Any valid [CSS unit][htmltools::validateCssUnit]
#'   (e.g., `150`).
#' @param gap Any valid [CSS unit][htmltools::validateCssUnit].
#' @param class A character vector of class names to add to the tag.
#' @param style A character vector of CSS properties to add to the tag.
#' @param css_selector A character string containing a CSS selector for
#'   targeting particular (inner) tag(s) of interest. For more details on what
#'   selector(s) are supported, see [htmltools::tagAppendAttributes()].
#'
#' @export
as_fill_carrier <- function(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
) {
  rlang::check_dots_empty()

  attrs <- fillable_attributes(
    min_height = min_height,
    max_height = max_height,
    gap = gap,
    class = class,
    style = style
  )

  if (rlang::is_missing(x)) {
    warn_css_selector_null(css_selector)
    attrs <- c(
      attrs,
      list(class = "html-fill-item html-fill-container", fill_dependency())
    )
    return(rlang::splice(attrs))
  }

  x <- bindFillRole(
    x,
    item = TRUE,
    container = TRUE,
    .cssSelector = css_selector
  )
  tagAppendAttributes(x, .cssSelector = css_selector, !!!attrs)
}

#' @rdname as_fill_carrier
#' @export
as_fillable_container <- function(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
) {
  rlang::check_dots_empty()

  attrs <- fillable_attributes(
    min_height = min_height,
    max_height = max_height,
    gap = gap,
    class = class,
    style = style
  )

  if (rlang::is_missing(x)) {
    warn_css_selector_null(css_selector)
    attrs <- c(attrs, list(class = "html-fill-container", fill_dependency()))
    return(rlang::splice(attrs))
  }

  x <- bindFillRole(x, container = TRUE, .cssSelector = css_selector)

  tagAppendAttributes(x, .cssSelector = css_selector, !!!attrs)
}

#' @rdname as_fill_carrier
#' @export
as_fill_item <- function(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
) {
  rlang::check_dots_empty()

  attrs <- fillable_attributes(
    min_height = min_height,
    max_height = max_height,
    class = class,
    style = style
  )

  if (rlang::is_missing(x)) {
    warn_css_selector_null(css_selector)
    attrs <- c(attrs, list(class = "html-fill-item", fill_dependency()))
    return(rlang::splice(attrs))
  }

  x <- bindFillRole(x, item = TRUE, .cssSelector = css_selector)

  tagAppendAttributes(x, .cssSelector = css_selector, !!!attrs)
}

fill_dependency <- local({
  fill_dep <- NULL

  function() {
    if (!is.null(fill_dep)) {
      return(fill_dep)
    }
    fill_dep <<- htmltools::findDependencies(
      htmltools::bindFillRole(htmltools::tags$div(), container = TRUE)
    )[[1]]
    fill_dep
  }
})

fillable_attributes <- function(
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL
) {
  list(
    style = css(
      min_height = validateCssUnit(min_height),
      max_height = validateCssUnit(max_height),
      gap = validateCssUnit(gap)
    ),
    class = class,
    style = style
  )
}

warn_css_selector_null <- function(x) {
  if (is.null(x)) {
    return()
  }

  rlang::warn(
    "Ignoring non-NULL `css_selector` because an input tag `x` was not provided.",
    call = rlang::caller_env()
  )
}

#' @rdname as_fill_carrier
#' @export
remove_all_fill <- function(x) {
  # NOTE: this doesn't remove the `htmltools-fill` dependency from `x`
  bindFillRole(
    x,
    item = FALSE,
    container = FALSE,
    overwrite = TRUE
  )
}

#' @rdname as_fill_carrier
#' @export
is_fill_carrier <- function(x) {
  classes <- paste0("html-fill-", c("container", "item"))
  renders_to_tag_class(x, classes)
}

#' @rdname as_fill_carrier
#' @export
is_fillable_container <- function(x) {
  UseMethod("is_fillable_container")
}

#' @export
is_fillable_container.htmlwidget <- function(x) {
  # won't actually work until (htmltools#334) gets fixed
  renders_to_tag_class(x, "html-fill-container", ".html-widget")
}

#' @export
is_fillable_container.default <- function(x) {
  renders_to_tag_class(x, "html-fill-container")
}

#' @rdname as_fill_carrier
#' @export
is_fill_item <- function(x) {
  UseMethod("is_fill_item")
}

#' @export
is_fill_item.htmlwidget <- function(x) {
  # won't actually work until (htmltools#334) gets fixed
  renders_to_tag_class(x, "html-fill-item", ".html-widget")
}

#' @export
is_fill_item.default <- function(x) {
  renders_to_tag_class(x, "html-fill-item")
}

renders_to_tag_class <- function(x, class, selector = NULL) {
  x <- try(as.tags(x), silent = TRUE)
  if (inherits(x, "try-error")) {
    return(FALSE)
  }
  xq <- tagQuery(x)
  if (length(selector)) {
    xq <- xq$find(selector)
  }
  vapply(
    xq$selectedTags(),
    function(y) tagQuery(y)$hasClass(class),
    logical(1),
    USE.NAMES = FALSE
  )
}

is_tag <- function(x) {
  inherits(x, "shiny.tag")
}

#' Obtain a list of all available bootswatch themes.
#'
#' @param version The major version of Bootswatch.
#' @param full_path Whether to return a path to the installed theme.
#' @return Returns a character vector of Bootswatch themes.
#' @family Bootstrap theme utility functions
#' @export
bootswatch_themes <- function(version = version_default(), full_path = FALSE) {
  list.dirs(bootswatch_dist(version), full.names = full_path, recursive = FALSE)
}

#' Obtain a theme's Bootswatch theme name
#'
#' @inheritParams bs_theme_update
#' @return Returns the Bootswatch theme named used (if any) in the `theme`.
#' @family Bootstrap theme utility functions
#' @export
theme_bootswatch <- function(theme) {
  if (!is_bs_theme(theme)) {
    return(NULL)
  }
  info <- theme_preset_info(theme)
  if (identical("bootswatch", info$type)) info$name else NULL
}

#' Obtain a theme's Bootstrap version
#'
#' @inheritParams bs_theme_update
#' @return Returns the major version of Bootstrap used in the `theme`.
#' @family Bootstrap theme utility functions
#' @export
theme_version <- function(theme) {
  if (!is_bs_theme(theme)) {
    return(NULL)
  }

  version <- grep("^bs_version_", class(theme), value = TRUE)
  if (length(version) == 1) {
    return(sub("^bs_version_", "", version))
  }

  # If no version class, or more than one class, we consult the sass bundle
  theme_preset_info(theme)$version
}

bootswatch_dist <- function(version) {
  switch_version(
    version,
    five = path_lib("bsw5", "dist"),
    four = path_lib("bsw4", "dist"),
    three = path_lib("bsw3")
  )
}

# -----------------------------------------------------------------
# Bootswatch bundle
# -----------------------------------------------------------------

bootswatch_bundle <- function(bootswatch, version) {
  if (
    !length(bootswatch) || isTRUE(bootswatch %in% c("default", "bootstrap"))
  ) {
    return(NULL)
  }

  bootswatch <- switch_version(
    version,
    default = {
      switch(
        bootswatch,
        paper = {
          message(
            "Bootswatch 3 theme paper has been renamed to materia in version 4 (using that theme instead)"
          )
          "materia"
        },
        readable = {
          message(
            "Bootswatch 3 theme readable has been renamed to litera in version 4 (using that theme instead)"
          )
          "litera"
        },
        match.arg(bootswatch, bootswatch_themes(version))
      )
    },
    three = match.arg(bootswatch, bootswatch_themes(version))
  )

  # Attach local font files, if necessary
  font_css <- file.path(bootswatch_dist(version), bootswatch, "font.css")
  attachments <- if (file.exists(font_css)) {
    c(
      "font.css" = font_css,
      fonts = system_file("fonts", package = "bslib")
    )
  }

  sass_bundle(
    bootswatch = sass_layer(
      file_attachments = attachments,
      defaults = list(
        "bslib-preset-type" = "bootswatch",
        "bslib-preset-name" = bootswatch,
        # Use local fonts (this path is relative to the bootstrap HTML dependency dir)
        '$web-font-path: "font.css" !default;',
        bootswatch_sass_file(bootswatch, "variables", version),
        # BS4 navbars are matched with BS3 for compatibility
        switch_version(
          version,
          three = "",
          four = bs3compat_navbar_defaults(bootswatch),
          # BS5 uses more neutral defaults (navbar that flips in light/dark mode)
          # or requires a bit more user input
          default = list()
        )
      ),
      rules = list(
        bootswatch_sass_file(bootswatch, "bootswatch", version),
        # For some reason sketchy sets .dropdown-menu{overflow: hidden}
        # but this prevents .dropdown-submenu from working properly
        # https://github.com/rstudio/bootscss/blob/023d455/inst/node_modules/bootswatch/dist/sketchy/_bootswatch.scss#L204
        if (identical(bootswatch, "sketchy")) {
          ".dropdown-menu{ overflow: inherit; }"
        } else {
          ""
        },
        # Several Bootswatch themes (e.g., zephyr, simplex, etc) add custom .btn-secondary
        # rules that should also apply to .btn-default
        ".btn-default:not(.btn-primary):not(.btn-info):not(.btn-success):not(.btn-warning):not(.btn-danger):not(.btn-dark):not(.btn-light):not([class*='btn-outline-']) {
          @extend .btn-secondary !optional;
        }"
      )
    )
  )
}

# Mappings from BS3 navbar classes to BS4
bs3compat_navbar_defaults <- function(bootswatch) {
  # Do nothing if this isn't a Bootswatch 3 theme
  if (!bootswatch %in% c("materia", "litera", bootswatch_themes(3))) {
    return("")
  }

  bg_colors <- switch(
    bootswatch,
    cerulean = c("primary", "info"),
    cosmo = c("dark", "primary"),
    cyborg = c("body-bg", "secondary"),
    darkly = c("primary", "success"),
    flatly = c("primary", "success"),
    journal = c("light", "primary"),
    lumen = c("light", "white"),
    # i.e., materia
    paper = ,
    materia = c("light", "primary"),
    readable = ,
    litera = c("light", "dark"),
    sandstone = c("dark", "success"),
    simplex = c("light", "primary"),
    slate = c("primary", "light"),
    spacelab = c("light", "primary"),
    superhero = c("dark", "primary"),
    united = c("primary", "dark"),
    yeti = c("dark", "primary"),
    stop("Didn't recognize Bootswatch 3 theme: ", bootswatch, call. = FALSE)
  )

  list(
    sprintf('$navbar-light-bg: $%s !default;', bg_colors[1]),
    sprintf('$navbar-dark-bg: $%s !default;', bg_colors[2])
  )
}

iconButton <- function(
  id,
  label,
  class = NULL,
  icon = "shuffle",
  icon_title = "Random",
  ...
) {
  tags$button(
    id = id,
    class = "btn btn-default action-button",
    class = class,
    bsicons::bs_icon("shuffle", title = "Random"),
    label
  )
}

shuffleButton <- function(id, label, class = NULL) {
  iconButton(id, label, class = c("shuffle-button", class))
}

local_disable_cache()

describe("Compiling against theme", {
  # Example CSS that includes one variable, one function call, one mixin
  bs4_css <- ".foo { background-color: $primary; color: color-contrast($primary); width: 120px; height: 120px; }"
  resolved_css <- ".foo { background-color: #007bff; color: #fff; width: 120px; height: 120px; }"

  # Compare bs_sass(input1) and sass(input2)
  expect_bs4_equal <- function(
    input1,
    input2,
    options = sass_options(),
    theme = bs_theme(version = 4)
  ) {
    expect_css(
      sass_partial(input1, theme, options = options),
      sass(input2, options = options)
    )
  }

  it("automatically includes Bootstrap 4 variables/functions", {
    expect_bs4_equal(bs4_css, resolved_css)
  })

  it("respects output_style", {
    expect_bs4_equal(
      bs4_css,
      resolved_css,
      options = sass_options(output_style = "compressed")
    )
  })

  it("respects theme options", {
    # Theme options are respected
    expect_bs4_equal(
      bs4_css,
      ".foo { background-color: #FF0000; color: #fff; width: 120px; height: 120px; }",
      theme = bs_theme(version = 4, primary = "red")
    )
    # Unless they're not
    expect_bs4_equal(bs4_css, resolved_css)
  })
})

# @staticimports pkg:staticimports
#  is_installed get_package_version system_file
#  register_upgrade_message
#  %||% is_string

switch_version <- function(version, five = default, four = default, three = default, default = NULL) {
  if (is_bs_theme(version)) {
    version <- theme_version(version)
  }
  version <- as.character(version)
  if (isTRUE(version %in% c("4-3", "4+3"))) {
    warning("Version ", version, " has been renamed to 4. Please use 4 instead")
    version <- "4"
  }
  switch(
    version, `5` = five, `4` = four, `3` = three,
    stop("Didn't recognize Bootstrap version: ", version, call. = FALSE)
  )
}

get_exact_version <- function(version) {
  switch_version(version, five = version_bs5, four = version_bs4, three = version_bs3)
}

lib_file <- function(...) {
  files <- system_file("lib", ..., package = "bslib")
  files_found <- files != ""
  if (all(files_found)) return(files)

  files_not_found <- file.path(...)[!files_found]
  stop(
    "bslib file not found: '", files_not_found, "'",
    call. = FALSE
  )
}

is_shiny_app <- function() {
  # Make sure to not load shiny as a side-effect of calling this function.
  isNamespaceLoaded("shiny") && shiny::isRunning()
}

is_hosted_app <- function() {
  nzchar(Sys.getenv("SHINY_SERVER_VERSION")) && is_shiny_app()
}

is_shiny_runtime <- function() {
  isTRUE(grepl("^shiny", knitr::opts_knit$get("rmarkdown.runtime")))
}

add_class <- function(x, y) {
  class(x) <- unique(c(y, oldClass(x)))
  x
}

dropNulls <- function(x) {
  x[!vapply(x, is.null, FUN.VALUE=logical(1))]
}

names2 <- function(x) {
  names(x) %||% rep.int("", length(x))
}

#' Rename a named list
#'
#' @param x a named list to be renamed
#' @param nms a named character vector defining the renaming
#' @noRd
#' @examples
#' rename2(list(a=1, b=3, c=4, a=2), b="z", f="w", a="y")
#' #> list(y = 1, z = 3, c = 4, y = 2)
#' rename2(c("a", "b", "c", "a"), b="z", f="w", a="y")
#' #> c("y", "z", "c", "y")
rename2 <- function(x, ...) {
  defs <- rlang::list2(...)
  nms <- names(x) %||% as.character(x)
  matches <- intersect(nms, names(defs))
  map <- match(nms, names(defs))
  mapidxNA <- is.na(map)
  replacement <- as.character(defs)[map[!mapidxNA]]
  if (is.null(names(x))) {
    x[!mapidxNA] <- replacement
  } else {
    names(x)[!mapidxNA] <- replacement
  }
  x
}

# Packages ---------------------------------------
library(shiny)
library(bslib)
library(htmltools)

pkgs_extra <- c("plotly", "colourpicker")
pkgs_yes <- vapply(pkgs_extra, rlang::is_installed, logical(1))
if (any(!pkgs_yes)) {
  rlang::abort(paste0(
    "The `build-a-box` app requires additional packages: ",
    paste(pkgs_extra[!pkgs_yes], collapse = ", ")
  ))
}

library(plotly)

# Settings ---------------------------------------
ENABLE_THEMER <- identical(Sys.getenv("ENABLE_THEMER"), "true")

# Functions ---------------------------------------
layout_value_box_options <- function(ui_opts) {
  layout_columns(
    div(h4("Content", class = "border-bottom"), ui_opts$title_value),
    div(h4("Theme", class = "border-bottom"), ui_opts$theme_opts),
    div(h4("Showcase", class = "border-bottom"), ui_opts$showcase)
  )
}

value_box_placeholder <- function(id) {
  value_box(
    id = id,
    class = "placeholder-glow",
    title = span(class = "placeholder col-7"),
    value = span(class = "placeholder col-4"),
    showcase = div(class = "placeholder bg-primary col-12", as_fill_item())
  )
}

# Theme ---------------------------------------
theme_build_a_box <- bs_add_rules(
  bs_theme(preset = "shiny"),
  sass::sass_file("www/build-a-box.scss")
)

# UI ---------------------------------------
ui <- page_fixed(
  title = "Build a Box | bslib",
  theme = theme_build_a_box,

  # Header ----
  tags$header(
    class = "mt-4 d-flex flex-row justify-content-between align-items-center",
    h2("Build a Box"),
    div(
      class = "d-flex flex-row align-items-center gap-3",
      popover(
        bsicons::bs_icon(
          "info-square-fill",
          title = "About Value Boxes",
          class = "icon-gradient"
        ),
        title = "About Value Boxes",
        HTML(commonmark::markdown_html(readLines("about-value-boxes.md")))
      ),
      actionLink(
        "show_code",
        tooltip(icon("code"), "Show code"),
        class = "nav-link text-blue",
        style = css(width = "1em")
      ),
      input_dark_mode(
        id = "color_mode",
        style = css("--text-1" = "var(--bs-blue)")
      )
    )
  ),

  # Main ----
  tags$main(
    # Value Box Previews ----
    div(
      id = "preview",
      class = "my-5",
      layout_columns(
        class = "value-box-previews",
        div(
          as_fill_carrier(),
          value_box_placeholder("one-value_box_placeholder"),
          ui_value_box_output("one")
        ),
        div(
          as_fill_carrier(),
          value_box_placeholder("two-value_box_placeholder"),
          ui_value_box_output("two")
        ),
        div(
          as_fill_carrier(),
          value_box_placeholder("three-value_box_placeholder"),
          ui_value_box_output("three")
        )
      )
    ),

    # Settings Panel ----
    navset_card_pill(
      id = "settings",
      title = span(
        "Value box settings",
        popover(
          bsicons::bs_icon("question-square-fill"),
          class = "ms-1 d-inline-block text-orange",
          title = "Getting started",
          shiny::markdown("
            The Build-a-Box app includes three value boxes. You can customize
            all three at once from the **All** tab.

            Pick an overall theme, choose whether you'd like to **showcase** a plot or icon, and decide which **showcase layout** works best for your data.

            Then, click directly on a value box or switch to the
            <a id=\"switch_to_one\" href=\"#\" class=\"action-button\">One</a>,
            <a id=\"switch_to_two\" href=\"#\" class=\"action-button\">Two</a>, or
            <a id=\"switch_to_three\" href=\"#\" class=\"action-button\">Three</a>
            tabs to customize its settings individually.
          ")
        )
      ),
      nav_panel(
        "All",
        value = "all",
        ui_global_controls("all")
      ),
      nav_panel(
        "One",
        value = "one-value_box",
        layout_value_box_options(
          ui_value_box_options("one")
        )
      ),
      nav_panel(
        "Two",
        value = "two-value_box",
        layout_value_box_options(
          ui_value_box_options("two")
        )
      ),
      nav_panel(
        "Three",
        value = "three-value_box",
        layout_value_box_options(
          ui_value_box_options("three")
        )
      )
    )
  ),

  # Footer ----
  tags$footer(
    class = "footer mt-auto py-3",
    layout_columns(
      class = "border-top pt-3 text-muted",
      div(
        class = "text-center text-sm-start",
        "Made with",
        a(href = "https://rstudio.github.io/bslib", "{bslib}"),
        "and",
        a(
          href = "https://shiny.rstudio.com",
          img(src = "shiny.png", width = "22px", alt = " "),
          "Shiny"
        )
      ),
      div(
        class = "text-center text-sm-end",
        HTML('Proudly supported by <a href="https://posit.co/"><img src="https://www.rstudio.com/assets/img/posit-logo-fullcolor-TM.svg" class="img-fluid" alt="Posit" width="65"></a>')
      )
    )
  ),
  # Extras ----
  tags$script(src = "build-a-box.js")
)

# Server ---------------------------------------
server <- function(input, output, session) {
  enable_themer <- reactive({
    query <- shiny::getQueryString()
    query_has_themer <- "themer" %in% names(query)

    if (!length(query) || !query_has_themer) return(ENABLE_THEMER)

    query$themer %in% c(1, "true", "") || ENABLE_THEMER
  })

  observeEvent(enable_themer(), {
    # TODO: This only runs on app startup right now
    req(enable_themer())

    insertUI(
      selector = "body",
      where = "beforeEnd",
      ui = tags$script("window.watchForThemer()")
    )
    bs_themer()
  })

  one <- module_value_box("one")
  two <- module_value_box("two")
  three <- module_value_box("three")

  module_global_controls("all", one, two, three)

  observeEvent(input$clicked_value_box, {
    nav_select("settings", input$clicked_value_box)
  })

  observeEvent(input$switch_to_one, nav_select("settings", "one-value_box"))
  observeEvent(input$switch_to_two, nav_select("settings", "two-value_box"))
  observeEvent(input$switch_to_three, nav_select("settings", "three-value_box"))

  observeEvent(input$settings, {
    session$sendCustomMessage("active-value-box", input$settings)
  })

  observeEvent(input$show_code, {
    layout_value_boxes <-
      paste0(
        "layout_columns(\n  ",
        rlang::expr_text(one$code()), ",\n  ",
        rlang::expr_text(two$code()), ",\n  ",
        rlang::expr_text(three$code()), "\n",
        ")"
      )

    code_modal(layout_value_boxes)
  })
}

# BUILD A BOX -----------------------------------------
shinyApp(ui, server)

as_fragment <- function(x, page = page_fluid) {
  stopifnot(is.function(page) && "theme" %in% names(formals(page)))
  attr(x, "bslib_page") <- page
  class(x) <- c("bslib_fragment", class(x))
  x
}

as_page <- function(x, theme = bs_theme()) {
  class(x) <- c("bslib_page", class(x))
  attr(x, "bs_theme") <- theme
  x
}

#' Print a bslib page/fragment
#'
#' @param x a bslib page/fragment.
#' @param ... passed along to an underlying print method.
#' @export
#' @keywords internal
#' @rdname html-browse
print.bslib_fragment <- function(x, ...) {
  x <- attr(x, "bslib_page")(x)
  invisible(print(x, ...))
}

#' Save a bslib page/fragment as HTML
#'
#' @param html a bslib page/fragment.
#' @param ... passed along to an underlying [save_html()] method.
#' @export
#' @keywords internal
#' @rdname save-html
save_html.bslib_fragment <- function(html, file, ...) {
  html <- attr(html, "bslib_page")(html)
  save_html(html, file, ...)
}

#' @rdname html-browse
#' @export
print.bslib_page <- function(x, ...) {
  old_global <- bs_global_get()
  bs_global_set(attr(x, "bs_theme", exact = TRUE))
  on.exit(bs_global_set(old_global))

  if (interactive()) {
    x <- htmltools::browsable(x)
  }

  invisible(NextMethod())
}

#' @rdname save-html
#' @export
save_html.bslib_page <- function(html, file, ...) {
  old_theme <- bs_global_get()
  bs_global_set(attr(html, "bs_theme", exact = TRUE))
  on.exit(bs_global_set(old_theme), add = TRUE)

  NextMethod()
}

# Generated by staticimports; do not edit by hand.
# ======================================================================
# Imported from pkg:staticimports
# ======================================================================

`%||%` <- function(a, b) {
  if (is.null(a)) b else a
}

# Borrowed from pkgload:::dev_meta, with some modifications.
# Returns TRUE if `pkg` was loaded with `devtools::load_all()`.
devtools_loaded <- function(pkg) {
  ns <- .getNamespace(pkg)
  if (is.null(ns) || is.null(ns$.__DEVTOOLS__)) {
    return(FALSE)
  }
  TRUE
}

get_package_version <- function(pkg) {
  # `utils::packageVersion()` can be slow, so first try the fast path of
  # checking if the package is already loaded.
  ns <- .getNamespace(pkg)
  if (is.null(ns)) {
    utils::packageVersion(pkg)
  } else {
    as.package_version(ns$.__NAMESPACE__.$spec[["version"]])
  }
}

is_installed <- function(pkg, version = NULL) {
  installed <- isNamespaceLoaded(pkg) || nzchar(system_file_cached(package = pkg))
  if (is.null(version)) {
    return(installed)
  }
  installed && isTRUE(get_package_version(pkg) >= version)
}

is_string <- function(x) {
  is.character(x) && length(x) == 1 && !is.na(x)
}

register_upgrade_message <- function(pkg, version, error = FALSE) {

  msg <- sprintf(
    "This version of '%s' is designed to work with '%s' >= %s.
    Please upgrade via install.packages('%s').",
    environmentName(environment(register_upgrade_message)),
    pkg, version, pkg
  )

  cond <- if (error) stop else packageStartupMessage

  if (pkg %in% loadedNamespaces() && !is_installed(pkg, version)) {
    cond(msg)
  }

  # Always register hook in case pkg is loaded at some
  # point the future (or, potentially, but less commonly,
  # unloaded & reloaded)
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      if (!is_installed(pkg, version)) cond(msg)
    }
  )
}

# Borrowed from pkgload::shim_system.file, with some modifications. This behaves
# like `system.file()`, except that (1) for packages loaded with
# `devtools::load_all()`, it will return the path to files in the package's
# inst/ directory, and (2) for other packages, the directory lookup is cached.
# Also, to keep the implementation simple, it doesn't support specification of
# lib.loc or mustWork.
system_file <- function(..., package = "base") {
  if (!devtools_loaded(package))  {
    return(system_file_cached(..., package = package))
  }

  if (!is.null(names(list(...)))) {
    stop("All arguments other than `package` must be unnamed.")
  }

  # If package was loaded with devtools (the package loaded with load_all),
  # also search for files under inst/, and don't cache the results (it seems
  # more likely that the package path will change during the development
  # process)
  pkg_path <- find.package(package)

  # First look in inst/
  files_inst <- file.path(pkg_path, "inst", ...)
  present_inst <- file.exists(files_inst)

  # For any files that weren't present in inst/, look in the base path
  files_top <- file.path(pkg_path, ...)
  present_top <- file.exists(files_top)

  # Merge them together. Here are the different possible conditions, and the
  # desired result. NULL means to drop that element from the result.
  #
  # files_inst:   /inst/A  /inst/B  /inst/C  /inst/D
  # present_inst:    T        T        F        F
  # files_top:      /A       /B       /C       /D
  # present_top:     T        F        T        F
  # result:       /inst/A  /inst/B    /C       NULL
  #
  files <- files_top
  files[present_inst] <- files_inst[present_inst]
  # Drop cases where not present in either location
  files <- files[present_inst | present_top]
  if (length(files) == 0) {
    return("")
  }
  # Make sure backslashes are replaced with slashes on Windows
  normalizePath(files, winslash = "/")
}

# A wrapper for `system.file()`, which caches the results, because
# `system.file()` can be slow. Note that because of caching, if
# `system_file_cached()` is called on a package that isn't installed, then the
# package is installed, and then `system_file_cached()` is called again, it will
# still return "".
system_file_cached <- local({
  pkg_dir_cache <- character()

  function(..., package = "base") {
    if (!is.null(names(list(...)))) {
      stop("All arguments other than `package` must be unnamed.")
    }

    not_cached <- is.na(match(package, names(pkg_dir_cache)))
    if (not_cached) {
      pkg_dir <- system.file(package = package)
      pkg_dir_cache[[package]] <<- pkg_dir
    } else {
      pkg_dir <- pkg_dir_cache[[package]]
    }

    file.path(pkg_dir, ...)
  }
})

#' Global theming
#'
#' `bs_global_theme()` creates a new (global) Bootstrap Sass theme which
#' [bs_theme_dependencies()] (or [sass_partial()]) can consume (their `theme`
#' argument defaults to `bs_global_get()`, which get the current global theme).
#'
#' @inheritParams bs_theme
#'
#' @return functions that modify the global theme (e.g., `bs_global_set()`)
#'   invisibly return the previously set theme. `bs_global_get()` returns the
#'   current global theme.
#'
#' @seealso [bs_theme()], [bs_theme_preview()]
#' @examples
#'
#' # Remember the global state now (so we can restore later)
#' theme <- bs_global_get()
#' # Use Bootstrap 3 (globally) with some theme customization
#' bs_global_theme(3, bg = "#444", fg = "#e4e4e4", primary = "#e39777")
#' if (interactive()) bs_theme_preview(with_themer = FALSE)
#' # If no global theme is active, bs_global_get() returns NULL
#' bs_global_clear()
#' bs_global_get()
#' # Restore the original state
#' bs_global_set(theme)
#'
#' @export
bs_global_theme <- function(version = version_default(), bootswatch = NULL, bg = NULL, fg = NULL,
                            primary = NULL, secondary = NULL, success = NULL, info = NULL, warning = NULL,
                            danger = NULL, base_font = NULL, code_font = NULL, heading_font = NULL, ...) {
  bs_global_set(bs_theme(
    version, bootswatch,
    bg = bg, fg = fg,
    primary = primary,
    secondary = secondary,
    success = success,
    info = info,
    warning = warning,
    danger = danger,
    base_font = base_font,
    code_font = code_font,
    heading_font = heading_font,
    ...
  ))
}


#' @rdname bs_global_theme
#' @inheritParams bs_theme_update
#' @export
bs_global_set <- function(theme = bs_theme()) {
  if (!is.null(theme)) {
    assert_bs_theme(theme)
  }
  # In addition to setting a bslib global option, also set shiny's
  # current theme if this code is running in an `runtime: shiny` doc
  if (is_shiny_runtime() && is_installed("shiny", "1.6")) {
    warning(
      "bs_global_set() may not work as expected inside runtime: shiny documents. ",
      "To update the document's theme, use `session$setCurrentTheme()` instead.",
      call. = FALSE
    )
  }
  old_theme <- options(bslib_theme = theme)
  invisible(old_theme[["bslib_theme"]])
}

#' @rdname bs_global_theme
#' @export
bs_global_get <- function() {
  getOption("bslib_theme")
}

#' @rdname bs_global_theme
#' @export
bs_global_clear <- function() {
  old_theme <- options(bslib_theme = NULL)
  invisible(old_theme[["bslib_theme"]])
}

#' @rdname bs_global_theme
#' @inheritParams bs_add_variables
#' @export
bs_global_add_variables <- function(..., .where = "defaults",
                                    .default_flag = identical(.where, "defaults")) {
  theme <- assert_global_theme("bs_global_add_variables()")
  theme <- bs_add_variables(theme, ..., .where = .where, .default_flag = .default_flag)
  bs_global_set(theme)
}

#' @rdname bs_global_theme
#' @export
bs_global_add_rules <- function(...) {
  theme <- assert_global_theme("bs_global_add_rules()")
  theme <- bs_add_rules(theme, ...)
  bs_global_set(theme)
}

#' @rdname bs_global_theme
#' @export
bs_global_bundle <- function(...) {
  theme <- assert_global_theme("bs_global_bundle()")
  theme <- bs_bundle(theme, ...)
  bs_global_set(theme)
}




assert_global_theme <- function(calling_func) {
  theme <- bs_global_get()
  if (is.null(theme)) {
    stop("`", calling_func, "` requires that a global theme is first set (do you want to call `bs_global_theme()`?)")
  }
  theme
}

withr::local_temp_libpaths()
install.packages(c("shiny", "ggplot2", "yaml", "thematic", "markdown"))

rsconnect::deployApp(
  rprojroot::find_package_root_file("inst/examples-shiny/brand.yml"),
  appName = "brand-yml",
  account = "bslib",
  forceUpdate = TRUE
)

#' Define breakpoint values
#'
#' A generic constructor for responsive breakpoints.
#'
#' @param sm Values to apply at the `sm` breakpoint.
#' @param md Values to apply at the `md` breakpoint.
#' @param lg Values to apply at the `lg` breakpoint.
#' @param ... Other breakpoints (e.g., `xl`).
#'
#' @references <https://getbootstrap.com/docs/5.3/layout/breakpoints/>
#'
#' @export
#' @seealso [layout_columns()]
breakpoints <- function(..., sm = NULL, md = NULL, lg = NULL) {
  breaks <- dropNulls(
    rlang::list2(..., sm = sm, md = md, lg = lg)
  )

  if (any_unnamed(breaks)) {
    abort("All `breakpoints` values must be named")
  }

  # Ensure that breakpoints are in the increasing order
  break_nms <- sort(factor(
    names(breaks), ordered = TRUE,
    levels = unique(c(bs_breakpoints(), names(breaks)))
  ))

  structure(
    breaks[as.character(break_nms)],
    class = "bslib_breakpoints"
  )
}

#' @export
print.bslib_breakpoints <- function(x, ...) {
  cat("<breakpoints>\n")

  width_vals <- max(nchar(unlist(x)), na.rm = TRUE)
  width_nms <- max(nchar(names(x)), na.rm = TRUE)

  for (bp in names(x)) {
    vals <- format(x[[bp]], width = width_vals, justify = "right")
    cat(
      " ", format(bp, width = width_nms, justify = "right"), ": ",
      paste0(vals, collapse = " "), "\n",
      sep = ""
    )
  }

  invisible(x)
}


is_breakpoints <- function(x) {
  inherits(x, "bslib_breakpoints")
}

bs_breakpoints <- function() {
  c("xs", "sm", "md", "lg", "xl", "xxl")
}

# breakpoints() passed to layout_columns(col_widths = ) is special in the sense
# that negative values are supported
as_column_breakpoints <- function(breaks) {
  if (!is_breakpoints(breaks)) {
    abort("`breaks` must be a `breakpoints()` object")
  }

  for (break_name in names(breaks)) {
    bk <- breaks[[break_name]]

    if (isTRUE(any(bk == 0))) {
      abort("Column values must be greater than 0 to indicate width, or negative to indicate a column offset.")
    }

    if (length(bk) > 1 && anyNA(bk)) {
      abort("Cannot mix widths and `NA` values. All column widths must be specified, or choose auto widths using a single `NA` value.")
    }

    if (isTRUE(is.na(bk)) || all(bk > 0)) {
      breaks[[break_name]] <- list(
        width = bk,
        before = integer(length(bk)),
        after = integer(length(bk))
      )
      next
    }

    if (!any(bk > 0)) {
      abort("Column values must include at least one positive integer width.")
    }

    idx_actual <- which(bk > 0)
    last_actual <- max(idx_actual)
    n_actual <- length(idx_actual)

    actual <- bk[idx_actual]
    before <- integer(n_actual)
    after  <- integer(n_actual)

    i <- 1L
    idx_before <- 1L
    while (i <= length(bk)) {
      if (bk[i] > 0) {
        idx_before <- idx_before + 1L
      } else if (i > last_actual) {
        # accumulate trailing offsets
        after[length(after)] <- after[length(after)] + abs(bk[i])
      } else {
        # accumulate leading offsets
        before[idx_before] <- before[idx_before] + abs(bk[i])
      }
      i <- i + 1L
    }

    breaks[[break_name]] <- list(
      width = actual,
      before = before,
      after = after
    )
  }

  breaks
}

withr::local_temp_libpaths()
install.packages("shiny")

rsconnect::deployApp(
  rprojroot::find_package_root_file("inst/examples-shiny/code-editor"),
  appName = "code-editor",
  account = "bslib",
  forceUpdate = TRUE
)

library(shiny)
library(bslib)
library(dplyr)
library(lubridate)
library(plotly)
library(chiflights22) # https://github.com/cpsievert/chiflights22
library(histoslider) # https://github.com/cpsievert/histoslider
library(rlang)

# Data prep
flights <- flights %>%
  left_join(
    airports %>%
      transmute(
        dest_name = paste0(name, " (", faa, ")"),
        dest = faa,
        end_lat = lat,
        end_lon = lon
      )
  ) %>%
  filter(!is.na(dest_name)) %>%
  left_join(
    airports %>% select(origin = faa, start_lat = lat, start_lon = lon)
  ) %>%
  left_join(airlines, by = "carrier") %>%
  rename(carrier_name = name) %>%
  select(-time_hour) %>%
  left_join(weather) %>%
  mutate(
    # The overwhelming majority of precipation is 0 so transform
    # with some help from MASS::boxcox() https://stats.stackexchange.com/a/1452/48604
    precip = scales::rescale(precip^-1.55 / -.55),
    date = lubridate::ymd(paste(year, month, day, sep = "-"))
  )

SLIDER_HEIGHT <- 150
CHOICES <- list(
  origin = c(
    "Choose origin(s)" = "",
    "O'Hare" = "ORD",
    "Midway" = "MDW",
    "Rockford" = "RFD"
  ),
  dest_name = c(
    "Choose destination(s)" = "",
    sort(unique(flights$dest_name))
  ),
  carrier_name = c(
    "Choose carrier(s)" = "",
    unique(flights$carrier_name)
  )
)

sidebar_acc <- accordion(
  open = c("Origin", "Destination"),
  accordion_panel(
    "Flight Path",
    # See https://github.com/rstudio/fontawesome/issues/114
    icon = fontawesome::fa("plane-departure"),
    uiOutput("flight_path_reset"),
    selectizeInput(
      "origin",
      "Origin",
      choices = CHOICES$origin,
      multiple = TRUE,
      options = list(plugins = "remove_button", closeAfterSelect = TRUE)
    ),
    selectizeInput(
      "dest_name",
      "Destination",
      choices = CHOICES$dest_name,
      multiple = TRUE,
      options = list(plugins = "remove_button", closeAfterSelect = TRUE)
    ),
    selectizeInput(
      "carrier_name",
      "Carrier",
      choices = CHOICES$carrier_name,
      multiple = TRUE,
      options = list(plugins = "remove_button", closeAfterSelect = TRUE)
    )
  ),
  accordion_panel(
    "Flight time",
    icon = fontawesome::fa("clock"),
    input_histoslider(
      "sched_dep_time",
      "Departure time",
      flights$sched_dep_time,
      height = SLIDER_HEIGHT,
      options = list(handleLabelFormat = "0d")
    ),
    input_histoslider(
      "sched_arr_time",
      "Arrival time",
      flights$sched_arr_time,
      height = SLIDER_HEIGHT,
      options = list(handleLabelFormat = "0d")
    ),
    input_histoslider(
      "date",
      "Date",
      flights$date,
      height = SLIDER_HEIGHT,
      breaks = "months",
      options = list(handleLabelFormat = "%b %e")
    )
  ),
  accordion_panel(
    "Weather",
    icon = fontawesome::fa("cloud-rain"),
    # TODO: problematic (many NAs)
    #input_histoslider(
    #  "precip", "Precipitation",
    #  flights$precip,
    #  height = SLIDER_HEIGHT
    #),
    input_histoslider(
      "wind_speed",
      "Wind speed",
      flights$wind_speed,
      height = SLIDER_HEIGHT
    ),
    input_histoslider(
      "wind_gust",
      "Wind gust",
      flights$wind_gust,
      height = SLIDER_HEIGHT
    )
  )
)

flights_card <- card(
  full_screen = TRUE,
  card_header(
    "Flight paths",
    tooltip(
      bsicons::bs_icon("info-circle", title = "About marker areas"),
      "Marker areas are proportional to mean arrival delay"
    ),
    class = "d-flex justify-content-between align-items-center"
  ),
  plotlyOutput("flight_paths")
)

avg_delay_by_category <- card(
  full_screen = TRUE,
  card_header(
    "Average delay by category",
    popover(
      bsicons::bs_icon("gear", title = "Settings"),
      selectInput(
        "avg_delay_category",
        "Category",
        c("Carrier", "Month", "Weekday")
      ),
      radioButtons(
        "avg_delay_type",
        "Delay type",
        c("Arrival", "Departure"),
        inline = TRUE
      )
    ),
    class = "d-flex justify-content-between align-items-center"
  ),
  plotlyOutput("scatter_delay")
)

delay_dist <- navset_card_underline(
  title = "Distribution of delay times",
  full_screen = TRUE,
  id = "delay_dist_nav",
  sidebar = sidebar(
    position = "right",
    open = FALSE,
    radioButtons(
      "delay_dist_type",
      "Delay type",
      c("Arrival" = "arr_delay", "Departure" = "dep_delay"),
      inline = TRUE
    ),
    conditionalPanel(
      "input.delay_dist_nav === 'Overall'",
      selectizeInput(
        "delay_dist_category",
        "Split by",
        c("Choose a category" = "", "Carrier", "Month", "Weekday"),
        options = list(plugins = "remove_button")
      )
    )
  ),
  nav_panel(
    "Overall",
    plotlyOutput("delay_dist")
  ),
  nav_panel(
    "Over time",
    plotlyOutput("arr_delay_series")
  )
)

PRIMARY <- "#0675DD"

ui <- page_navbar(
  theme = bs_theme(
    preset = "shiny",
    "primary" = PRIMARY
  ),
  lang = "en",
  title = tags$span(
    tags$img(
      src = "logo.png",
      width = "46px",
      height = "auto",
      class = "me-3",
      alt = "Shiny hex logo"
    ),
    "Chicago Flights"
  ),
  sidebar = sidebar(width = 275, sidebar_acc),
  nav_spacer(),
  nav_panel(
    "Delay overview",
    class = "bslib-page-dashboard",
    uiOutput("value_boxes"),
    layout_columns(
      flights_card,
      avg_delay_by_category
    ),
    delay_dist
  ),
  nav_panel(
    "Data export",
    card(
      card_header("Flight data"),
      DT::dataTableOutput("export")
    )
  ),
  nav_item(
    tags$a(
      tags$span(
        bsicons::bs_icon("code-slash"),
        "Source code"
      ),
      href = "https://github.com/rstudio/bslib/tree/main/inst/examples-shiny/flights",
      target = "_blank"
    )
  ),
  nav_item(
    input_dark_mode(id = "dark_mode", mode = "light")
  )
)

server <- function(input, output, session) {
  # ---------------------------------------------------------
  # Flights tab logic
  #
  # WARNING: this server-side filtering logic is VERY experimental
  # at this point and won't be easily adapt to different use cases.
  # If you feel tempted to use it, use with caution.
  # ---------------------------------------------------------

  # Mapping from input id name to updating input function
  input_discrete_vars <- list(
    origin = updateSelectInput,
    dest_name = updateSelectInput,
    carrier_name = updateSelectInput
  )
  input_numeric_vars <- list(
    sched_dep_time = update_histoslider,
    sched_arr_time = update_histoslider,
    date = update_histoslider
  )
  input_vars <- c(
    input_discrete_vars,
    input_numeric_vars
  )

  filter_index <- function(d) {
    idx <- rep(TRUE, nrow(d))
    for (var in names(d)) {
      idx <- idx & filter_col(d, var)
    }
    idx & !is.na(idx)
  }

  filter_col <- function(d, var) {
    vals <- d[[var]]
    input_val <- input[[var]]
    if (is.null(input_val) || identical(input_val, "")) {
      return(TRUE)
    }

    if (is.character(vals) || is.factor(vals) || is.logical(vals)) {
      return(d[[var]] %in% input_val)
    }

    # N.B. between() will remove NAs, which we probably don't
    # want to remove until the slider is considered 'active'
    rng <- range(vals, na.rm = TRUE)
    active <- isTRUE(rng[1] <= input_val[1] || input_val[2] <= rng[2])

    if (!active) {
      return(TRUE)
    }

    dplyr::between(vals, input_val[1], input_val[2])
  }

  # Set up a listener for each input that effectively says update
  # every other input when my value changes
  lapply(names(input_vars), function(var) {
    # We don't want updates to other variables to then
    # cause an update to this variable
    do_update <- reactiveVal(TRUE)

    observeEvent(input[[var]], ignoreInit = TRUE, ignoreNULL = FALSE, {
      if (!do_update()) {
        return()
      }
      do_update(FALSE)
      on.exit(do_update(TRUE), add = TRUE)

      d <- flights[filter_index(flights), ]
      if (nrow(d) == 0) {
        return()
      }

      other_vars <- setdiff(names(input_vars), var)
      lapply(other_vars, function(v) {
        input_val <- input[[v]]
        update_input_func <- input_vars[[v]]

        if (v %in% names(input_discrete_vars)) {
          choices <- CHOICES[[v]] %||% sort(unique(d[[v]]))
          selected <- input_val %||% CHOICES[[v]][CHOICES[[v]] == ""]

          update_input_func(
            inputId = v,
            choices = choices,
            selected = selected
          )
        } else {
          update_input_func(
            id = v,
            values = d[[v]],
            start = input_val[1],
            end = input_val[2]
          )
        }
      })
    })
  })

  output$flight_path_reset <- renderUI({
    req(c(input$origin, input$dest_name, input$carrier_name))

    actionLink(
      "flight_path_reset",
      "Reset",
      style = htmltools::css(
        position = "absolute",
        right = "1rem",
        text_decoration = "none",
        font_weight = 700,
        font_size = ".875rem"
      )
    )
  })

  observeEvent(input$flight_path_reset, {
    updateSelectInput(
      inputId = "origin",
      choices = CHOICES$origin
    )
    updateSelectInput(
      inputId = "dest_name",
      choices = CHOICES$dest_name
    )
    updateSelectInput(
      inputId = "carrier_name",
      choices = CHOICES$carrier_name
    )
  })

  # Flights with all filters applied (i.e., data used for value boxes/plots)
  flight_dat <- reactive({
    flights[filter_index(flights), ]
  })

  summary_vals <- reactive({
    d <- flight_dat()

    list(
      n = scales::comma(nrow(d)),
      n_dest = length(unique(d$dest_name)),
      n_carriers = length(unique(d$carrier_name)),
      dep_delay = round(mean(d$dep_delay, na.rm = T), 0),
      dep_delay_perc = round(
        100 * sum(d$dep_delay > 0, na.rm = T) / nrow(d),
        1
      ),
      arr_delay = round(mean(d$arr_delay, na.rm = T), 0),
      arr_delay_perc = round(
        100 * sum(d$arr_delay > 0, na.rm = TRUE) / nrow(d),
        1
      )
    )
  })

  output$value_boxes <- renderUI({
    vals <- summary_vals()
    n_flights <- value_box(
      "A TOTAL OF",
      paste(vals$n, "flights"),
      paste("Across", vals$n_dest, "destinations"),
      tags$p(
        paste(
          "On",
          vals$n_carriers,
          "different carriers"
        )
      ),
      showcase = bsicons::bs_icon("airplane")
    )

    late <- if (vals$dep_delay > 0) "late" else "early"
    delay_dep <- value_box(
      "AVERAGE DEPARTURE",
      paste(vals$dep_delay, "mins", late),
      paste0(vals$dep_delay_perc, "% of flights depart ", late),
      showcase = bsicons::bs_icon("hourglass-split")
    )

    late <- if (vals$arr_delay > 0) "late" else "early"
    delay_arr <- value_box(
      "AVERAGE ARRIVAL",
      paste(vals$arr_delay, "mins", late),
      paste0(vals$arr_delay_perc, "% of flights arrive ", late),
      showcase = bsicons::bs_icon("hourglass-bottom")
    )

    layout_columns(class = "mb-0", n_flights, delay_dep, delay_arr)
  }) %>%
    bindCache(flight_dat())

  plotly_base <- function(..., geo = FALSE, color = I(PRIMARY)) {
    plot_func <- if (geo) plot_geo else plot_ly
    plot_func(..., color = color) %>%
      plotly::config(displayModeBar = FALSE) %>%
      plotly::layout(
        font = list(
          family = "Open Sans",
          color = if (input$dark_mode == "dark") "white" else "#1D1F21"
        ),
        plot_bgcolor = "transparent",
        paper_bgcolor = "transparent"
      )
  }

  output$flight_paths <- renderPlotly({
    flight_dat() %>%
      group_by(start_lon, start_lat, end_lon, end_lat, origin, dest) %>%
      summarise(mean_delay = mean(arr_delay, na.rm = TRUE)) %>%
      plotly_base(geo = TRUE, showlegend = FALSE) %>%
      add_segments(
        x = ~start_lon,
        xend = ~end_lon,
        y = ~start_lat,
        yend = ~end_lat,
        alpha = 0.5,
        size = I(1),
        hoverinfo = "none"
      ) %>%
      add_markers(
        x = ~end_lon,
        y = ~end_lat,
        size = ~mean_delay,
        hoverinfo = "text",
        alpha = 0.1,
        text = ~ paste0(
          origin,
          " -> ",
          dest,
          "<br>",
          "Average delay: ",
          round(mean_delay, 1)
        )
      ) %>%
      layout(
        geo = list(
          bgcolor = "transparent",
          projection = list(
            type = 'orthographic',
            rotation = list(lon = -100, lat = 40, roll = 0)
          ),
          showland = TRUE,
          landcolor = toRGB("gray95"),
          countrycolor = toRGB("gray80")
        )
      )
  }) %>%
    bindCache(flight_dat(), input$dark_mode)

  output$scatter_delay <- renderPlotly({
    d <- flight_dat()
    req(nrow(d) > 0)

    d <- switch(
      input$avg_delay_category,
      Weekday = group_by(d, y = lubridate::wday(date, label = TRUE)),
      Month = group_by(d, y = lubridate::month(date, label = TRUE)),
      Carrier = group_by(d, y = carrier_name),
      stop("Category of ", input$avg_delay_category, "not implemented")
    )

    d <- switch(
      input$avg_delay_type,
      Arrival = summarise(d, avg = mean(arr_delay, na.rm = TRUE)),
      Departure = summarise(d, avg = mean(dep_delay, na.rm = TRUE)),
    )

    d %>%
      arrange(avg) %>%
      mutate(y = factor(y, levels = y)) %>%
      plotly_base(x = ~avg, y = ~y) %>%
      add_bars(hoverinfo = "x") %>%
      layout(
        yaxis = list(title = ""),
        xaxis = list(
          title = paste("Average", tolower(input$avg_delay_type), "delay"),
          hoverformat = ".1f",
          gridcolor = if (input$dark_mode == "dark") "#303030"
        )
      )
  }) %>%
    bindCache(
      flight_dat(),
      input$dark_mode,
      input$avg_delay_category,
      input$avg_delay_type
    )

  output$delay_dist <- renderPlotly({
    d <- flight_dat()
    x <- d[[input$delay_dist_type]]
    req(length(x) > 0)

    x_mean <- mean(x, na.rm = TRUE)
    end <- quantile(x, probs = 0.99, na.rm = TRUE)

    color <- switch(
      input$delay_dist_category,
      Carrier = d$carrier_name,
      Month = lubridate::month(d$date, label = TRUE),
      Weekday = lubridate::wday(d$date, label = TRUE),
      I(PRIMARY)
    )

    plotly_base(
      x = x,
      color = color,
      hovertemplate = "%{y} flights were<br>%{x} min late <extra></extra>"
    ) %>%
      rangeslider(start = min(x, na.rm = TRUE), end = as.numeric(end)) %>%
      add_annotations(
        text = paste(
          "Average",
          switch(
            input$delay_dist_type,
            arr_delay = "arrival",
            dep_delay = "departure"
          ),
          "<br> delay of",
          round(x_mean, 1),
          "min"
        ),
        x = x_mean,
        y = 0.5,
        yref = "paper",
        ax = 80,
        ay = -50,
        font = list(size = 14)
      ) %>%
      layout(
        barmode = "stack",
        yaxis = list(gridcolor = if (input$dark_mode == "dark") "#303030"),
        shapes = list(
          type = "line",
          x0 = x_mean,
          x1 = x_mean,
          y0 = 0,
          y1 = 1,
          yref = "paper",
          line = list(color = "lightgray", dash = "dash")
        )
      )
  }) %>%
    bindCache(
      flight_dat(),
      input$dark_mode,
      input$delay_dist_type,
      input$delay_dist_category
    )

  output$arr_delay_series <- renderPlotly({
    d <- flight_dat()
    req(nrow(d) > 0)

    d <- group_by(d, date)

    d <- switch(
      input$delay_dist_type,
      arr_delay = summarise(d, y = mean(arr_delay, na.rm = TRUE)),
      dep_delay = summarise(d, y = mean(dep_delay, na.rm = TRUE))
    )

    color <- switch(
      input$delay_dist_category,
      Carrier = d$carrier_name,
      Month = lubridate::month(d$date, label = TRUE),
      Weekday = lubridate::wday(d$date, label = TRUE),
      I(PRIMARY)
    )

    plotly_base(
      x = d$date,
      y = d$y,
      color = color,
      hovertemplate = "%{y:.1f}<extra></extra>",
    ) %>%
      add_lines() %>%
      layout(
        hovermode = "x",
        xaxis = list(title = "", tickformat = "%b %e"),
        yaxis = list(title = "Average delay", showgrid = FALSE)
      )
  }) %>%
    bindCache(
      flight_dat(),
      input$dark_mode,
      input$delay_dist_type,
      input$delay_dist_category
    )

  output$export <- DT::renderDataTable({
    DT::datatable(flight_dat(), fillContainer = TRUE)
  })
}

shinyApp(ui, server)

rlang::check_installed("reshape2")
library(reshape2)

data(tips, package = "reshape2")

tipsUI <- function(id) {
  ns <- NS(id)

  layout_sidebar(
    sidebar = sidebar(
      title = tooltip(
        span(
          "Restaurant tipping",
          bsicons::bs_icon("info-circle-fill"),
          class = "sidebar-title"
        ),
        "One waiter recorded information about each tip he received",
        "over a period of a few months working in one restaurant."
      ),
      sliderInput(
        ns("total_bill"),
        "Bill amount",
        min(tips$total_bill),
        max(tips$total_bill),
        value = range(tips$total_bill),
        pre = "$"
      ),
      checkboxGroupInput(
        ns("time"),
        "Food service",
        c("Lunch", "Dinner"),
        c("Lunch", "Dinner"),
        inline = TRUE
      ),
      actionButton(ns("reset"), "Reset filter"),
    ),
    layout_column_wrap(
      width = 1 / 3,
      fill = FALSE,
      value_box(
        "Total tippers",
        uiOutput(ns("total_tippers"), container = h2),
        showcase = bsicons::bs_icon("person")
      ),
      value_box(
        "Average tip",
        uiOutput(ns("average_tip"), container = h2),
        showcase = bsicons::bs_icon("wallet2")
      ),
      value_box(
        "Average bill",
        uiOutput(ns("average_bill"), container = h2),
        showcase = bsicons::bs_icon("currency-dollar")
      )
    ),
    layout_column_wrap(
      width = 1 / 2,
      class = "mt-3",
      card(
        full_screen = TRUE,
        card_header(
          "Total bill vs tip",
          popover(
            bsicons::bs_icon("gear"),
            radioButtons(
              ns("scatter_color"),
              NULL,
              inline = TRUE,
              c("none", "sex", "smoker", "day", "time")
            ),
            title = "Add a color variable",
            placement = "top"
          ),
          class = "d-flex justify-content-between align-items-center"
        ),
        plotOutput(ns("scatterplot"))
      ),
      card(
        full_screen = TRUE,
        class = "bslib-card-table-sm",
        card_header("Tips data"),
        DT::dataTableOutput(ns("table"))
      ),
    ),
    card(
      full_screen = TRUE,
      class = "mt-3",
      card_header(
        "Tip percentages",
        popover(
          bsicons::bs_icon("gear"),
          radioButtons(
            ns("tip_perc_y"),
            "Split by:",
            inline = TRUE,
            c("sex", "smoker", "day", "time"),
            "day"
          ),
          radioButtons(
            ns("tip_perc_facet"),
            "Facet by:",
            inline = TRUE,
            c("none", "sex", "smoker", "day", "time"),
            "none"
          ),
          title = "Add a color variable"
        ),
        class = "d-flex justify-content-between align-items-center"
      ),
      plotOutput(ns("tip_perc"))
    )
  )
}

tipsServer <- function(id) {
  moduleServer(id, function(input, output, session) {
    tips_data <- reactive({
      d <- tips
      d <- d[
        d$total_bill >= input$total_bill[1] &
          d$total_bill <= input$total_bill[2],
      ]
      d <- d[d$time %in% input$time, ]
      d
    })

    # TODO: maybe use reactable if and when this lands?
    # https://github.com/glin/reactable/pull/96
    output$table <- DT::renderDataTable({
      DT::datatable(tips_data(), fillContainer = TRUE, rownames = FALSE)
    })

    output$scatterplot <- renderPlot({
      validate(
        need(
          nrow(tips_data()) > 0,
          "No tips match the current filter. Try adjusting your filter settings."
        )
      )
      color <- if (input$scatter_color != "none") sym(input$scatter_color)
      ggplot(tips_data(), aes(x = total_bill, y = tip, color = !!color)) +
        geom_point() +
        geom_smooth() +
        labs(x = NULL, y = NULL)
    })

    output$tip_perc <- renderPlot({
      validate(
        need(
          requireNamespace("ggridges", quietly = TRUE),
          "Please install the ggridges package to see this plot."
        )
      )
      validate(
        need(
          requireNamespace("ggridges", quietly = TRUE),
          "Please install the ggridges package to see this plot."
        )
      )
      p <- ggplot(
        tips_data(),
        aes(x = tip / total_bill, y = !!sym(input$tip_perc_y))
      ) +
        ggridges::geom_density_ridges(scale = 0.9) +
        coord_cartesian(clip = "off") +
        labs(x = NULL, y = NULL)

      if (input$tip_perc_facet != "none") {
        p <- p + facet_wrap(vars(!!sym(input$tip_perc_facet)))
      }

      p
    })

    output$total_tippers <- renderUI({
      nrow(tips_data())
    })

    output$average_bill <- renderUI({
      if (nrow(tips_data()) == 0) {
        return(HTML("&ndash;"))
      }
      scales::dollar(mean(tips_data()$total_bill))
    })

    output$average_tip <- renderUI({
      if (nrow(tips_data()) == 0) {
        return(HTML("&ndash;"))
      }
      d <- tips_data()
      scales::percent(mean(d$tip / d$total_bill))
    })

    observeEvent(input$reset, {
      updateSliderInput(session, "total_bill", value = range(tips$total_bill))
      updateCheckboxGroupInput(session, "time", selected = c("Lunch", "Dinner"))
    })
  })
}

as_fragment <- function(x, page = page_fluid) {
  stopifnot(is.function(page) && "theme" %in% names(formals(page)))
  attr(x, "bslib_page") <- page
  class(x) <- c("bslib_fragment", class(x))
  x
}

as_page <- function(x, theme = bs_theme()) {
  class(x) <- c("bslib_page", class(x))
  attr(x, "bs_theme") <- theme
  x
}

#' Print a bslib page/fragment
#'
#' @param x a bslib page/fragment.
#' @param ... passed along to an underlying print method.
#' @export
#' @keywords internal
#' @rdname html-browse
print.bslib_fragment <- function(x, ...) {
  x <- attr(x, "bslib_page")(x)
  invisible(print(x, ...))
}

#' Save a bslib page/fragment as HTML
#'
#' @param html a bslib page/fragment.
#' @param ... passed along to an underlying [htmltools::save_html()] method.
#' @export
#' @keywords internal
#' @rdname save-html
save_html.bslib_fragment <- function(html, file, ...) {
  html <- attr(html, "bslib_page")(html)
  save_html(html, file, ...)
}

#' @rdname html-browse
#' @export
print.bslib_page <- function(x, ...) {
  old_global <- bs_global_get()
  bs_global_set(attr(x, "bs_theme", exact = TRUE))
  on.exit(bs_global_set(old_global))

  if (interactive()) {
    x <- htmltools::browsable(x)
  }

  invisible(NextMethod())
}

#' @rdname save-html
#' @export
save_html.bslib_page <- function(html, file, ...) {
  old_theme <- bs_global_get()
  bs_global_set(attr(html, "bs_theme", exact = TRUE))
  on.exit(bs_global_set(old_theme), add = TRUE)

  NextMethod()
}

library(shiny)
library(bslib)
library(ggplot2)
library(lattice)
library(thematic)
library(rlang)
library(curl)
library(DT)
library(knitr)
library(reactable)

theme <- bs_theme(
  bg = "#202123",
  fg = "#B8BCC2",
  primary = "#EA80FC",
  secondary = "#00DAC6",
  success = "#4F9B29",
  info = "#28B3ED",
  warning = "#FD7424",
  danger = "#F7367E",
  base_font = font_google("Open Sans"),
  heading_font = font_google("Proza Libre"),
  code_font = font_google("Fira Code")
)

old_theme <- bs_global_get()
bs_global_set(theme)
onStop(function() bs_global_set(old_theme))

bslib:::as_themer_app(
  system.file("themer-demo", package = "bslib")
)

#' @include utils.R
NULL

#' Preview the currently set theme
#'
#' Launches an example shiny app via `run_with_themer()` and
#' `bs_theme_dependencies()`. Useful for getting a quick preview of the current
#' theme setting as well as an interactive GUI for tweaking some of the main
#' theme settings.
#'
#' The app that this launches is subject to change.
#'
#' @inheritParams bs_theme_update
#' @param ... passed along to [shiny::runApp()].
#' @param with_themer whether or not to run the app with [run_with_themer()].
#' @return nothing, this function is called for its side-effects (launching an
#'   application).
#' @seealso [run_with_themer()]
#' @examples
#' theme <- bs_theme(bg = "#6c757d", fg = "white", primary = "orange")
#' if (interactive()) bs_theme_preview(theme)
#' @export
bs_theme_preview <- function(theme = bs_theme(), ..., with_themer = TRUE) {
  assert_bs_theme(theme)
  old_theme <- bs_global_get()
  on.exit(bs_global_set(old_theme), add = TRUE)
  bs_global_set(theme)
  # TODO: add more this demo and also an option for launching different demos
  app <- system_file("themer-demo", package = "bslib")
  if (with_themer) {
    run_with_themer(app, ...)
  } else {
    shiny::runApp(app, ...)
  }
}

colorpicker_deps <- function() {
  htmltools::htmlDependency(
    "bootstrap-colorpicker",
    "3.1.2",
    lib_file("bs-colorpicker"),
    stylesheet = "css/bootstrap-colorpicker.min.css",
    script = "js/bootstrap-colorpicker.js"
  )
}

opts_metadata <- function(theme) {
  opts <- jsonlite::fromJSON(
    system_file("themer/options.json", package = "bslib"),
    simplifyDataFrame = FALSE
  )
  themes <- bootswatch_themes(theme_version(theme))
  opts[[1]]$bootswatch$choices <- c("default", themes)
  opts
}

bs_themer_ui <- function(opts, vals, theme) {

  make_control <- function(id, opts) {
    value <- vals[[id]]
    lbl <- HTML(opts$label)
    desc <- HTML(opts$desc)
    text_input <- function(input_class = NULL, type = "text", ...) {
      div(
        class = "form-row form-group",
        tags$label(lbl),
        tags$input(
          type = type, value = value, "data-id" = id,
          class = "form-control form-control-sm bs-theme-value",
          class = input_class, ...
        ),
        if (!is.null(desc)) div(class = "form-text small", desc)
      )
    }
    switch(
      opts$type,
      color = text_input(input_class = "bs-theme-value-color text-monospace"),
      str = text_input(input_class = "bs-theme-value-str"),
      length = text_input(input_class = "bs-theme-value-length"),
      number = text_input(input_class = "bs-theme-value-str", type = "number", step = opts$step),
      bool = tagList(
        div(
          class = "form-check",
          tags$input(
            type = "checkbox", checked = if (value) NA else NULL,
            class = "bs-theme-value bs-theme-value-bool form-check-input",
            id = paste0(".bsthemer-", id), "data-id" = id
          ),
          tags$label("for" = paste0(".bsthemer-", id), class = "form-check-label", lbl)
        ),
        if (!is.null(desc)) div(class = "form-text small", desc)
      ),
      select = div(
        class = "form-row form-group",
        tags$label(class = "control-label", lbl),
        tags$select(
          class = "form-control", "data-id" = id,
          class = "bs-theme-value bs-theme-value-select",
          lapply(opts$choices, function(x) {
            tags$option(
              value = x, selected = if (identical(x, value)) NA else NULL,
              tools::toTitleCase(x)
            )
          })
        ),
        if (!is.null(desc)) div(class = "form-text small", desc)
      ),
      stop("unknown type")
    )
  }

  version <- theme_version(theme)
  accordion <- lapply(seq_along(opts), function(i) {
    opt_name <- names(opts)[[i]]
    elId <- paste0("bsthemerCollapse", i)
    btn <- tags$button(
      class = if (version >= 5) "accordion-button" else "btn btn-link px-3 py-2 w-100 text-left border-0",
      class = if (i != 1) "collapsed",
      "data-toggle" = "collapse",
      "data-target" = paste0("#", elId),
      # data-bs-* is for BS5+
      "data-bs-toggle" = "collapse",
      "data-bs-target" = paste0("#", elId),
      "aria-expanded" = "true", "aria-controls" = elId,
      opt_name
    )
    controls <- lapply(seq_along(opts[[i]]), function(j) {
      make_control(names(opts[[i]])[[j]], opts[[i]][[j]])
    })
    div(
      class = if (version >= 5) "accordion-item",
      div(
        class = if (version >= 5) "accordion-header" else "card-header p-0 border-0",
        btn
      ),
      div(
        id = elId, class = if (i == 1) "show" else "collapse",
        "data-parent" = "#bsthemerAccordion",
        # data-bs-* is for BS5+
        "data-bs-parent" = "#bsthemerAccordion",
        class = if (version >= 5) "accordion-collapse",
        div(
          class = if (version >= 5) "accordion-body" else "card-body",
          controls
        )
      )
    )
  })

  withTags(tagList(
    colorpicker_deps(),
    htmlDependency(
      "bs_themer", version = get_package_version("bslib"),
      src = "themer", script = c("themer.js"),
      package = "bslib", all_files = FALSE
    ),

    div(id = "bsthemerContainer",
      class = "card shadow",
      style = css(
        # The bootstrap-colorpicker plugin sets a z-index of 1060 on
        # it's inputs, so the container needs a smaller index, than that
        # https://github.com/rstudio/bslib/blob/e4da71f3/inst/lib/bs-colorpicker/css/bootstrap-colorpicker.css#L38
        #
        # It's also important that this z-index is higher than 1030 so it's
        # overlaid on-top of fixed/sticky navbars
        # https://github.com/rstudio/bslib/blob/e4da71f3/inst/lib/bs/scss/_variables.scss#L697-L701
        z_index = 1059, width = "18rem", max_height = "80vh",
        position = "fixed", top = "1rem", right = "1rem", height = "auto"
      ),

      div(id = "bsthemerHeader",
        class = "move-grabber", "data-target" = "#bsthemerContainer",
        class = "card-header font-weight-bold bg-dark text-light px-3 py-2",
        "Theme customizer",
        tags$div(id = "bsthemerToggle", class = "float-right",
          "data-toggle" = "collapse",
          "data-target" = "#bsthemerAccordion",
          # data-bs-* is for BS5+
          "data-bs-toggle" = "collapse",
          "data-bs-target" = "#bsthemerAccordion",
          style = css(cursor = "pointer"),
          tags$span(),
          bs_dependency_defer(themer_css_dependency)
        )
      ),

      div(
        id = "bsthemerAccordion", class = "collapse show",
        class = if (version >= 5) "accordion",
        style = css(overflow_y = "auto"),
        accordion
      )
    )
  ))
}

themer_css_dependency <- function(theme) {
  version <- get_package_version("bslib")
  bs_dependency(
    input = sass_file(system_file("themer/themer.scss", package = "bslib")),
    theme = theme,
    name = "bs-themer-css",
    version = version,
    cache_key_extra = version
  )
}

#' Theme customization UI
#'
#' A 'real-time' theme customization UI that you can use to easily make common
#' tweaks to Bootstrap variables and immediately see how they would affect your
#' app's appearance. There are two ways you can launch the theming UI. For most
#' Shiny apps, just use `run_with_themer()` in place of [shiny::runApp()]; they
#' should take the same arguments and work the same way. Alternatively, you can
#' call the `bs_themer()` function from inside your server function (or in an R
#' Markdown app that is using `runtime: shiny`, you can call this from any code
#' chunk). Note that this function is only intended to be used for development!
#'
#' To help you utilize the changes you see in the preview, this utility prints
#' [bs_theme()] code to the R console.
#'
#' @param appDir The application to run. This can be a file or directory path,
#'   or a [shiny::shinyApp()] object. See [shiny::runApp()] for details.
#' @param ... Additional parameters to pass through to [shiny::runApp()].
#' @param gfonts whether or not to detect Google Fonts and wrap them in
#'   [font_google()] (so that their font files are automatically imported).
#' @param gfonts_update whether or not to update the internal database of
#'   Google Fonts.
#'
#' @section Limitations:
#'
#'   * Doesn't work with Bootstrap 3.
#'   * Doesn't work with IE11.
#'   * Only works inside Shiny apps and `runtime: shiny` R Markdown documents.
#'     * Can't be used with static R Markdown documents.
#'     * Can be used to some extent with `runtime: shiny_prerendered`, but only UI
#'       rendered through a `context="server"` may update in real-time.
#'   * Doesn't work with '3rd party' custom widgets that don't make use of
#'     [bs_dependency_defer()] or [bs_current_theme()].
#'
#' @return nothing. These functions are called for their side-effects.
#'
#' @examples
#' library(shiny)
#'
#' ui <- fluidPage(
#'   theme = bs_theme(bg = "black", fg = "white"),
#'   h1("Heading 1"),
#'   h2("Heading 2"),
#'   p(
#'     "Paragraph text;",
#'     tags$a(href = "https://www.rstudio.com", "a link")
#'   ),
#'   p(
#'     actionButton("cancel", "Cancel"),
#'     actionButton("continue", "Continue", class = "btn-primary")
#'   ),
#'   tabsetPanel(
#'     tabPanel("First tab",
#'       "The contents of the first tab"
#'     ),
#'     tabPanel("Second tab",
#'       "The contents of the second tab"
#'     )
#'   )
#' )
#'
#' if (interactive()) {
#'   run_with_themer(shinyApp(ui, function(input, output) {}))
#' }
#'
#' @export
run_with_themer <- function(appDir = getwd(), ..., gfonts = TRUE, gfonts_update = FALSE) {
  shiny::runApp(
    as_themer_app(appDir, gfonts = gfonts, gfonts_update = gfonts_update),
    ...
  )
}

as_themer_app <- function(appDir, gfonts = TRUE, gfonts_update = FALSE) {
  obj <- shiny::as.shiny.appobj(appDir)
  origServerFuncSource <- obj[["serverFuncSource"]]
  obj[["serverFuncSource"]] <- function() {
    origServerFunc <- origServerFuncSource()
    function(input, output, session, ...) {
      bs_themer(gfonts, gfonts_update)
      if (!"session" %in% names(formals(origServerFunc))) {
        origServerFunc(input, output, ...)
      } else {
        origServerFunc(input, output, session, ...)
      }
    }
  }
  obj
}

#' @rdname run_with_themer
#' @export
bs_themer <- function(gfonts = TRUE, gfonts_update = FALSE) {
  session <- shiny::getDefaultReactiveDomain()
  if (is.null(session)) {
    stop(call. = FALSE, "`bslib::bs_themer()` must be called from within a ",
         "Shiny server function")
  }
  if (!identical("ok", session$ns("ok"))) {
    stop(call. = FALSE, "`bslib::bs_themer()` must be called from within a ",
         "top-level Shiny server function, not a Shiny module server function")
  }
  if (!is_installed("shiny", "1.6.0")) {
    stop(call. = FALSE, "`bslib::bs_themer()` requires shiny v1.6.0 or higher")
  }
  theme <- get_current_theme()
  if (!is_bs_theme(theme)) {
    stop(call. = FALSE, "`bslib::bs_themer()` requires `shiny::bootstrapLib()` to be present ",
         "in the app's UI. Consider providing `bslib::bs_theme()` to the theme argument of the ",
         "relevant page layout function (or, more generally, adding `bootstrapLib(bs_theme())` to the UI.")
  }
  bootswatch <- theme_bootswatch(theme)
  switch_version(
    theme, three = stop("Interactive theming for Bootstrap 3 isn't supported")
  )
  if (isTRUE(session$userData[["bs_themer_init"]])) {
    # bs_themer() was called multiple times for the same session
    return()
  } else {
    session$userData[["bs_themer_init"]] <- TRUE
  }

  gfont_info <- if (isTRUE(gfonts)) get_gfont_info(gfonts_update)

  # Insert the theming control panel with values informed by the theme settings
  themer_opts <- opts_metadata(theme)
  themer_vars <- unlist(unname(lapply(themer_opts, names)))
  sass_vars <- setdiff(themer_vars, "bootswatch")
  themer_vals <- as.list(get_themer_vals(theme, sass_vars))
  themer_vals$bootswatch <- bootswatch
  shiny::insertUI("body", where = "beforeEnd", ui = bs_themer_ui(themer_opts, themer_vals, theme))

  input <- session$input

  # We emit different 'code' for runtime:shiny in Rmd
  isRmd <- is_shiny_runtime()

  # When the bootswatch theme changes, update the themer's state to reflect
  # the new variable defaults. Note that we also update the "input theme",
  # and effectively throw out any other theming changes made (i.e., start from a new theme)
  # since it'd be messy to figure out whether changes are "real" or just a
  # consequence of a new bootswatch value
  shiny::observeEvent(input$bs_theme_bootswatch, {
    theme <<- set_current_theme(
      theme, list(bootswatch = input$bs_theme_bootswatch),
      session, rmd = isRmd
    )
    vals <- as.list(bs_get_variables(theme, sass_vars))
    session$sendCustomMessage("bs-themer-bootswatch", list(values = vals))
  })

  # Fires when anything other then the Bootswatch theme changes
  shiny::observeEvent(input$bs_theme_vars, {
    vals <- jsonlite::parse_json(input$bs_theme_vars)

    # Validate that `vals` is a simple list, containing atomic elements,
    # that are all named
    if (!identical(class(vals), "list") ||
        !all(vapply(vals, is.atomic, logical(1))) ||
        is.null(names(vals)) ||
        !isTRUE(all(nzchar(names(vals), keepNA = TRUE)))) {
      warning(call. = FALSE,
        "bs_themer() encountered malformed input; ignoring"
      )
      return()
    }

    # Makes remaining logic simpler to reason about
    if (length(vals) == 0) {
      return()
    }

    # Remember, theme at this point has been updated to reflect the current Bootswatch theme,
    # so re-query Sass values from the (possibly updated) theme, then filter down to meaningful
    # differences
    theme_vals <- get_themer_vals(theme, names(vals[sass_vars]))
    changed_vals <- as.list(diff_css_values(vals[sass_vars], theme_vals))

    if (!identical(bootswatch, input$bs_theme_bootswatch)) {
      changed_vals$bootswatch <- input$bs_theme_bootswatch
    }

    # If _either_ fg/bg has changed, bs_theme() must to be called with *both* fg and bg populated.
    if (any(c("bg", "fg") %in% names(changed_vals))) {
      changed_vals[["bg"]] <- changed_vals[["bg"]] %||% vals[["bg"]]
      changed_vals[["fg"]] <- changed_vals[["fg"]] %||% vals[["fg"]]
    }

    # Change variables names to their 'high-level' equivalents
    changed_vals <- rename2(
      changed_vals,
      "font-family-base" = "base_font", "font-family-monospace" = "code_font",
      "headings-font-family" = "heading_font",
      "font-size-base" = "font_scale"
    )

    if (length(changed_vals$font_scale)) {
      changed_vals$font_scale <- as.numeric(changed_vals$font_scale)
    }

    if (isTRUE(gfonts)) {
      for (var in c("base_font", "code_font", "heading_font")) {
        changed_vals[[var]] <- insert_font_google_call(changed_vals[[var]], gfont_info)
      }
    }

    set_current_theme(theme, changed_vals, session, rmd = isRmd)
  })
}


get_themer_vals <- function(theme, vars) {
  vals <- bs_get_variables(theme, vars)
  if (!grepl("rem$", vals[["font-size-base"]])) {
    stop("font-size-base must have a CSS unit length type of rem", call. = FALSE)
  }
  vals[["font-size-base"]] <- sub("rem$", "", vals[["font-size-base"]])
  vals
}

set_current_theme <- function(theme, changed_vals, session, rmd = FALSE) {
  shiny::insertUI("body", ui = spinner_overlay(), immediate = TRUE, session = session)
  on.exit(shiny::removeUI("body > #spinner_overlay"), add = TRUE)

  # Construct the code/yaml to display to the user
  if (isTRUE(rmd)) {
    display_vals <- lapply(changed_vals, function(x) {
      if (is.numeric(x)) {
        return(x)
      }
      if (rlang::is_call(x)) {
        str <- paste0(deparse(x, width.cutoff = 500L), collapse = "")
        return(paste("!expr", str))
      }
      # To avoid yaml parse errors with values that contain # or ",
      # first escape ", then in quote the value
      paste0('"', gsub('"', '\\"', x, fixed = TRUE), '"')
    })
    message("\n####  Update your Rmd output format's theme:  ####")
    cat(paste0(
      "    theme:\n",
      paste0(
        collapse = "\n", "      ", names(display_vals), ": ", display_vals
      ),
      "\n"
    ))
  } else {
    message("\n####  Update your bs_theme() R code with:  #####")
    print(rlang::expr(bs_theme_update(theme, !!!changed_vals)))
  }

  # Color contrast warnings are more annoying then they are useful inside the theming widget
  opts <- options(bslib.color_contrast_warnings = FALSE)
  on.exit(options(opts), add = TRUE)

  # the actual code that we evaluate should not have quoted expressions
  changed_vals[] <- lapply(changed_vals, eval_val)
  code <- rlang::expr(bs_theme_update(theme, !!!changed_vals))
  theme <- rlang::eval_tidy(code)
  # Prevent Sass compilation errors from crashing the app and relay a message to user.
  # Errors can happen if the users enters values that lead to unexpected Sass
  # expressions (e.g., "$foo: * !default")
  shiny::removeNotification("sass-compilation-error", session = session)
  tryCatch(
    session$setCurrentTheme(theme),
    error = function(e) {
      shiny::showNotification(
        "Sass -> CSS compilation failed, likely due to invalid user input.
         Other theming changes won't take effect until the invalid input is fixed.",
        duration = NULL,
        id = "sass-compilation-error",
        type = "error",
        session = session
      )
    }
  )
  invisible(theme)
}

spinner_overlay <- function() {
  tagList(
    tags$style(
      "@supports ((-webkit-backdrop-filter:blur(4px)) or (backdrop-filter:blur(4px))) {
        #spinner_overlay{ -webkit-backdrop-filter:blur(4px); backdrop-filter:blur(4px); background-color:rgba(255,255,255,.05);}
      }"
    ),
    div(
      id = "spinner_overlay",
      style = "position:absolute; top:0; left:0; min-height:100vh; width:100%; background-color:rgba(255,255,255,.8); z-index:100000",
      class = "d-flex flex-column justify-content-center align-items-center",
      div(
        class = "spinner-border",
        style = "width:5rem; height:5rem; color: rgba(0,0,0,0.8);",
        role = "status",
        span(class = "sr-only visually-hidden", "Refreshing stylesheets...")
      ),
      span(class = "lead mt-1", style = "color: rgba(0,0,0,0.8);", "Refreshing stylesheets...")
    )
  )
}

eval_val <- function(x) {
  if (is.call(x)) return(eval(x))
  if (!is.list(x)) return(x)
  lapply(x, eval_val)
}

insert_font_google_call <- function(val, gfont_info) {
  # val should be a non-empty character string
  if (!is_string(val)) return(NULL)
  if (!nzchar(val)) return(NULL)
  fams <- strsplit(as.character(val), ",")[[1]]
  fams <- vapply(
    fams, function(x) gsub("^\\s*['\"]?", "", gsub("['\"]?\\s*$", "", x)),
    character(1), USE.NAMES = FALSE
  )
  fams <- fams[nzchar(fams)]
  is_a_gfont <- tolower(fams) %in% tolower(gfont_info$family)
  if (length(fams) == 1) {
    return(if (is_a_gfont) call("font_google", fams) else fams)
  }
  fams <- as.list(fams)
  for (i in which(is_a_gfont)) {
    fams[[i]] <- call("font_google", fams[[i]])
  }
  rlang::expr(font_collection(!!!unname(fams)))
}


get_gfont_info <- function(update = FALSE) {
  if (isTRUE(update)) {
    jsonlite::fromJSON(gfont_api_url())$items
  } else {
    # See tools/update_gfont_info.R
    gfont_info
  }
}

# same as thematic:::gfont_api_url
gfont_api_url <- function() {
  paste0("https://www.googleapis.com/webfonts/v1/webfonts?key=", gfont_key())
}
# same as thematic:::gfont_key
# As mentioned in the developer API, this key is safe to be public facing
# https://developers.google.com/fonts/docs/developer_api
gfont_key <- function() {
  Sys.getenv("GFONT_KEY", paste0("AIzaSyDP", "KvElVqQ-", "26f7tjxyg", "IGpIajf", "tS_zmas"))
}

#' Retrieve Sass variable values from the current theme
#'
#' Useful for retriving a variable from the current theme and using
#' the value to inform another R function.
#'
#' @inheritParams bs_theme_update
#' @param varnames a character string referencing a Sass variable
#' in the current theme.
#' @return a character string containing a CSS/Sass value.
#' If the variable(s) are not defined, their value is `NA`.
#'
#' @export
#' @examples
#' vars <- c("body-bg", "body-color", "primary", "border-radius")
#' bs_get_variables(bs_theme(), varnames = vars)
#' bs_get_variables(bs_theme(bootswatch = "darkly"), varnames = vars)
#'
bs_get_variables <- function(theme, varnames) {
  if (length(varnames) == 0) {
    return(stats::setNames(character(0), character(0)))
  }

  # Our bg/fg are not actual Sass variables and can mean different things depending
  # on the bootswatch theme/version
  base_color_idx <- varnames %in% c("fg", "bg")
  if (any(base_color_idx)) {
    varnames[base_color_idx] <- rename2(
      varnames[base_color_idx], !!!get_base_color_map(theme)
    )
  }

  assert_bs_theme(theme)

  # Support both `bs_get_variables("$foo")` and `bs_get_variables("foo")`
  # (note that `sass::sass("$$foo:1;")` is illegal; so this seems safe)
  varnames <- sub("^\\$", "", varnames)

  # It's possible that some varnames refer to variables that aren't defined.
  # This would normally cause a crash. We define last-ditch defaults here,
  # with a magic constant that we can swap out for NA before returning to
  # the user.
  na_sentinel <- "NA_SENTINEL_CONSTANT_4902F4E"
  sassvars <- paste0(
    "$", varnames, ": ", na_sentinel, " !default;",
    collapse = "\n"
  )

  # Declare a block with a meaningless but identifiable selector (.__rstudio_bslib_get_variables)
  # and add properties for each variable that is desired.
  cssvars <- paste0(
    "--", varnames, ": #{inspect($", varnames, ")};",
    collapse = "\n"
  )
  cssvars <- sprintf(":root.__rstudio_bslib_get_variables {\n %s \n}", cssvars)

  css <- sass_partial(
    cssvars,
    # Add declarations to the current theme
    bs_bundle(theme, sass_layer(mixins = sassvars)),
  )

  # Search the output for the block of properties we just generated, using the
  # ".__rstudio_bslib_get_variables" selector. The capture group will include all of the
  # properties we care about in a single string (the propstr variable below).
  matches <- regexec("(:root)?\\.__rstudio_bslib_get_variables(:root)?\\s*\\{\\s*\\n(.*?)\\n\\s*\\}", css)
  propstr <- regmatches(css, matches)[[1]][4]
  if (is.na(propstr)) {
    stop("bs_global_get_variables failed; expected selector was not found")
  }
  # Split the propstr by newline, so we can perform vectorized regex operations
  # on all of the variables at once.
  proplines <- strsplit(propstr, "\n")[[1]]

  # Parse each line for the name and value.
  matches2 <- regmatches(proplines, regexec("\\s*--([^:]+):\\s*(.*);$", proplines))
  names <- vapply(matches2, function(x) x[2], character(1))
  values <- vapply(matches2, function(x) x[3], character(1))

  if (any(is.na(names))) {
    stop("bs_global_get_variables failed; generated output was in an unexpected format")
  }
  if (!identical(varnames, names)) {
    stop("bs_global_get_variables failed; expected properties were not found")
  }

  # Any variables that had to fall back to our defaults, we'll replace with NA
  values[values == na_sentinel] <- NA_character_


  if (any(base_color_idx)) {
    varnames[base_color_idx] <- rename2(
      varnames[base_color_idx], !!!get_base_color_map(theme, decode = FALSE)
    )
  }

  # Return as a named character vector
  stats::setNames(values, varnames)
}


diff_css_values <- function(a, b) {
  stopifnot(all(!is.na(a)))
  stopifnot(identical(names(a), names(b)))
  stopifnot(is.list(a))
  if(!is.character(b))browser()

  a_char <- vapply(a, function(x) {
    if (is.null(x) || isTRUE(is.na(x))) {
      "null"
    } else if (is.logical(x)) {
      tolower(as.character(x))
    } else if (is.character(x)) {
      x
    } else {
      as.character(x)
    }
  }, character(1))

  b <- ifelse(is.na(b), "null", b)

  # Normalize colors; ignore things that don't seem to be colors. This is
  # necessary so we don't consider "black", "#000", "#000000", "rgb(0,0,0,1)",
  # etc. to be distinct values.
  #
  # Note: This won't work with values that are colors AND other things, like
  # "solid #000 3px"; it needs the value to be solely a color to be normalized.

  a_char_colors <- htmltools::parseCssColors(a_char, mustWork = FALSE)
  a_char <- ifelse(!is.na(a_char_colors), a_char_colors, a_char)

  b_colors <- htmltools::parseCssColors(b, mustWork = FALSE)
  b <- ifelse(!is.na(b_colors), b_colors, b)

  idx <- ifelse(is.na(b), TRUE, a_char != b)
  a[idx]
}

#' @rdname bs_get_variables
#' @inheritParams bs_get_variables
#' @export
#' @examples
#'
#' bs_get_contrast(bs_theme(), c("primary", "dark", "light"))
#'
#' library(htmltools)
#' div(
#'   class = "bg-primary",
#'   style = css(
#'     color = bs_get_contrast(bs_theme(), "primary")
#'   )
#' )
#'
bs_get_contrast <- function(theme, varnames) {
  stopifnot(is.character(varnames))
  stopifnot(length(varnames) > 0)

  varnames <- sub("^\\$", "", varnames)
  prop_string <- paste0(
    paste0(varnames, ": color-contrast($", varnames, ");"),
    collapse = "\n"
  )
  css <- sass::sass_partial(
    paste0("bs_get_contrast {", prop_string, "}"),
    theme, cache_key_extra = get_package_version("bslib"),
    # Don't listen to global Sass options so we can be sure
    # that stuff like source maps won't be included
    options = sass::sass_options(source_map_embed = FALSE)
  )
  css <- gsub("\n", "", gsub("\\s*", "", css))
  css <- sub("bs_get_contrast{", "", css, fixed = TRUE)
  css <- sub("\\}$", "", css)
  props <- strsplit(strsplit(css, ";")[[1]], ":")
  setNames(
    vapply(props, function(x) htmltools::parseCssColors(sub(";$", "", x[2])), character(1)),
    vapply(props, `[[`, character(1), 1)
  )
}

render_skeleton <- function(x) {
  tmp_dir <- withr::local_tempdir()
  tmp_rmd <- file.path(tmp_dir, "tmp.Rmd")

  src <- system.file(
    "rmarkdown/templates",
    x,
    "skeleton/skeleton.Rmd",
    package = "bslib"
  )

  # Comment out bs_themer() since that needs a shiny runtime
  txt <- sub("^bslib::bs_themer()", "#bslib::bs_themer()", readLines(src))
  writeLines(txt, tmp_rmd)

  rmarkdown::render(tmp_rmd, quiet = TRUE)
}

test_that("Rmd skeletons can be render cleanly", {
  skip_on_cran()

  skip_if_not(
    rmarkdown::pandoc_available("1.12.3"),
    "Pandoc 1.12.3 or higher is required"
  )

  expect_no_error(render_skeleton("legacy"))
  expect_no_error(render_skeleton("new"))
  withr::with_namespace(
    "shiny",
    expect_no_error(render_skeleton("real-time"))
  )
})

render_skeleton <- function(x) {
  tmp <- tempfile()
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  src <- system.file("rmarkdown/templates", x, "skeleton/skeleton.Rmd", package = "bslib")
  # Comment out bs_themer() since that needs a shiny runtime
  txt <- sub("^bslib::bs_themer()", "#bslib::bs_themer()", readLines(src))
  writeLines(txt, file.path(tmp, "tmp.Rmd"))
  rmarkdown::render(file.path(tmp, "tmp.Rmd"), quiet = TRUE)
}

test_that("Rmd skeletons can be render cleanly", {
  skip_if_not(
    rmarkdown::pandoc_available("1.12.3"),
    "Pandoc 1.12.3 or higher is required"
  )
  expect_error(render_skeleton("legacy"), NA)
  expect_error(render_skeleton("new"), NA)
  withr::with_namespace(
    "shiny",
    expect_error(render_skeleton("real-time"), NA)
  )
})

render_skeleton <- function(x) {
  tmp_dir <- withr::local_tempdir()
  tmp_rmd <- file.path(tmp_dir, "tmp.Rmd")

  src <- system.file(
    "rmarkdown/templates",
    x,
    "skeleton/skeleton.Rmd",
    package = "bslib"
  )

  # Comment out bs_themer() since that needs a shiny runtime
  txt <- sub("^bslib::bs_themer()", "#bslib::bs_themer()", readLines(src))
  writeLines(txt, tmp_rmd)

  rmarkdown::render(tmp_rmd, quiet = TRUE)
}

test_that("Rmd skeletons can be render cleanly", {
  skip_on_cran()
  skip_if_not_installed("rmarkdown")

  skip_if_not(
    rmarkdown::pandoc_available("1.12.3"),
    "Pandoc 1.12.3 or higher is required"
  )

  expect_no_error(render_skeleton("legacy"))
  expect_no_error(render_skeleton("new"))
  withr::with_namespace(
    "shiny",
    expect_no_error(render_skeleton("real-time"))
  )
})

#' Define breakpoint values
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' A generic constructor for responsive breakpoints.
#'
#' @param sm Values to apply at the `sm` breakpoint.
#' @param md Values to apply at the `md` breakpoint.
#' @param lg Values to apply at the `lg` breakpoint.
#' @param ... Other breakpoints (e.g., `xl`).
#'
#' @seealso `breakpoints()` is used by [layout_columns()].
#' @references Bootstrap's
#'   [Breakpoints article](https://getbootstrap.com/docs/5.3/layout/breakpoints/)
#'   provides more detail on breakpoints and how they are used and customized.
#'
#' @examples
#' breakpoints(sm = c(4, 4, 4), md = c(3, 3, 6), lg = c(-2, 8, -2))
#'
#' @export
breakpoints <- function(..., sm = NULL, md = NULL, lg = NULL) {
  breaks <- dropNulls(
    rlang::list2(..., sm = sm, md = md, lg = lg)
  )

  if (any_unnamed(breaks)) {
    abort("All `breakpoints` values must be named")
  }

  # Ensure that breakpoints are in the increasing order
  break_nms <- sort(factor(
    names(breaks), ordered = TRUE,
    levels = unique(c(bs_breakpoints(), names(breaks)))
  ))

  structure(
    breaks[as.character(break_nms)],
    class = "bslib_breakpoints"
  )
}

#' @export
print.bslib_breakpoints <- function(x, ...) {
  cat("<breakpoints>\n")

  width_vals <- max(nchar(unlist(x)), na.rm = TRUE)
  width_nms <- max(nchar(names(x)), na.rm = TRUE)

  for (bp in names(x)) {
    vals <- format(x[[bp]], width = width_vals, justify = "right")
    cat(
      " ", format(bp, width = width_nms, justify = "right"), ": ",
      paste0(vals, collapse = " "), "\n",
      sep = ""
    )
  }

  invisible(x)
}


is_breakpoints <- function(x) {
  inherits(x, "bslib_breakpoints")
}

bs_breakpoints <- function() {
  c("xs", "sm", "md", "lg", "xl", "xxl")
}

# breakpoints() passed to layout_columns(col_widths = ) is special in the sense
# that negative values are supported
as_column_breakpoints <- function(breaks) {
  if (!is_breakpoints(breaks)) {
    abort("`breaks` must be a `breakpoints()` object")
  }

  for (break_name in names(breaks)) {
    bk <- breaks[[break_name]]

    if (isTRUE(any(bk == 0))) {
      abort("Column values must be greater than 0 to indicate width, or negative to indicate a column offset.")
    }

    if (length(bk) > 1 && anyNA(bk)) {
      abort("Cannot mix widths and `NA` values. All column widths must be specified, or choose auto widths using a single `NA` value.")
    }

    if (isTRUE(is.na(bk)) || all(bk > 0)) {
      breaks[[break_name]] <- list(
        width = bk,
        before = integer(length(bk)),
        after = integer(length(bk))
      )
      next
    }

    if (!any(bk > 0)) {
      abort("Column values must include at least one positive integer width.")
    }

    idx_actual <- which(bk > 0)
    last_actual <- max(idx_actual)
    n_actual <- length(idx_actual)

    actual <- bk[idx_actual]
    before <- integer(n_actual)
    after  <- integer(n_actual)

    i <- 1L
    idx_before <- 1L
    while (i <= length(bk)) {
      if (bk[i] > 0) {
        idx_before <- idx_before + 1L
      } else if (i > last_actual) {
        # accumulate trailing offsets
        after[length(after)] <- after[length(after)] + abs(bk[i])
      } else {
        # accumulate leading offsets
        before[idx_before] <- before[idx_before] + abs(bk[i])
      }
      i <- i + 1L
    }

    breaks[[break_name]] <- list(
      width = actual,
      before = before,
      after = after
    )
  }

  breaks
}

#' Toast notifications
#'
#' @description
#' Toast notifications are lightweight, temporary messages designed to mimic
#' push notifications from mobile and desktop operating systems. They are built
#' on [Bootstrap 5's toast
#' component](https://getbootstrap.com/docs/5.3/components/toasts/).
#'
#' bslib includes a complete example of toasts and their many configuration
#' options:
#'
#' ```r
#' shiny::runExample("toast", package = "bslib")
#' ```
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   actionButton("show_simple", "Simple Toast"),
#'   actionButton("show_header", "Toast with Header")
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$show_simple, {
#'     show_toast(
#'       toast(
#'         "Operation completed successfully!",
#'         header = "Success",
#'         type = "success"
#'       )
#'     )
#'   })
#'
#'   observeEvent(input$show_header, {
#'     show_toast(
#'       toast(
#'         "Your settings have been saved.",
#'         header = toast_header(
#'           title = "Settings Updated",
#'           status = "just now"
#'         ),
#'         type = "success"
#'       )
#'     )
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @param ... Body content of the toast. Can be a string, or any HTML elements.
#'   Named arguments will be treated as HTML attributes for the toast container.
#' @param header Optional header content. Can be a string, or the result of
#'   [toast_header()]. If provided, creates a `.toast-header` with close button
#'   (if `closable = TRUE`).
#' @param id Optional unique identifier for the toast. If `NULL`, an ID will be
#'   automatically generated when the toast is shown via [show_toast()].
#'   Providing a stable ID allows you to hide the toast later. If a toast with
#'   `id` is already visible, that toast is automatically hidden before showing
#'   the new toast with the same `id` so that only one toast with a given ID is
#'   shown at once.
#' @param type Optional semantic type. One of `NULL`, `"primary"`,
#'   `"secondary"`, `"success"`, `"info"`, `"warning"`, `"danger"`, `"light"`,
#'   or `"dark"`. Applies appropriate Bootstrap background utility classes
#'   (`text-bg-*`).
#' @param duration_s Numeric. Number of seconds after which the toast should
#'   automatically hide. Use `0`, or `NA` to disable auto-hiding (toast will
#'   remain visible until manually dismissed). Default is `5` (5 seconds).
#' @param position String or character vector specifying where to position the
#'   toast container. Can be provided in several formats:
#'   * Kebab-case: `"top-left"`, `"bottom-right"`, etc.
#'   * Space-separated: `"top left"`, `"bottom right"`, etc.
#'   * Character vector: `c("top", "left")`, `c("bottom", "right")`, etc.
#'   * Any order: `"left top"` is equivalent to `"top left"`
#'
#'   Valid vertical positions are `"top"`, `"middle"`, or `"bottom"`. Valid
#'   horizontal positions are `"left"`, `"center"`, or `"right"`. Input is
#'   case-insensitive. Default is `"bottom-right"`.
#' @param closable Logical. Whether to include a close button. Defaults to
#'   `TRUE`. When both `duration_s = NA` (or `0` or `NULL`) and `closable =
#'   FALSE`, the toast will remain visible until manually hidden via
#'   [hide_toast()]. This is useful when the toast contains interactive Shiny UI
#'   elements and you want to manage the toast display programmatically.
#'
#' @return A `bslib_toast` object that can be passed to [show_toast()].
#'
#' @seealso [show_toast()] to display a toast, [hide_toast()] to dismiss a
#'   toast, and [toast_header()] to create structured headers.
#' @describeIn toast Create a toast element.
#' @family Toast components
#' @export
toast <- function(
  ...,
  header = NULL,
  icon = NULL,
  id = NULL,
  type = NULL,
  duration_s = 5,
  position = "top-right",
  closable = TRUE
) {
  if (!is.null(type)) {
    type <- rlang::arg_match(
      type,
      # fmt: skip
      c(
        "primary", "secondary",
        "success", "info", "warning", "danger", "error",
        "light", "dark"
      )
    )
    # Support "error" as alias for "danger"
    type <- switch(type, error = "danger", type)
  }

  dots <- separate_arguments(...)

  position <- normalize_toast_position(position)

  # duration_s of 0 or NA (or NULL) disables auto-hiding
  if (is.null(duration_s) || rlang::is_na(duration_s)) {
    autohide <- FALSE
  } else {
    if (!is.numeric(duration_s) || length(duration_s) != 1 || duration_s < 0) {
      rlang::abort(
        "`duration_s` must be a single non-negative number or NA."
      )
    }
    autohide <- duration_s != 0
  }

  duration <- if (autohide) duration_s * 1000 # milliseconds

  structure(
    list(
      body = dots$children,
      header = header,
      icon = icon,
      id = id,
      type = type,
      autohide = autohide,
      duration = duration,
      position = position,
      closable = closable,
      attribs = dots$attribs
    ),
    class = "bslib_toast"
  )
}

#' @describeIn toast Create a structured toast header with optional icon and
#'   status indicator. Returns a data structure that can be passed to the
#'   `header` argument of `toast()`.
#'
#' @param title Header text (required).
#' @param icon Optional icon element, for example from [shiny::icon()],
#'   [bsicons::bs_icon()] or [fontawesome::fa()].
#' @param status Optional status text that appears as small, muted text on the
#'   right side of the header.
#'
#' @return For `toast_header()`: a toast header object that can be used with the
#'   `header` argument of `toast()`.
#'
#' @export
toast_header <- function(title, ..., icon = NULL, status = NULL) {
  dots <- separate_arguments(...)

  structure(
    list(
      title = tagList(title, !!!dots$children),
      icon = icon,
      status = status,
      attribs = dots$attribs
    ),
    class = "bslib_toast_header"
  )
}

#' @export
as.tags.bslib_toast <- function(x, ...) {
  id <- x$id %||% toast_random_id()

  toast_component(
    body = x$body,
    header = x$header,
    icon = x$icon,
    type = x$type,
    closable = x$closable,
    id = id,
    attribs = x$attribs
  )
}

# nocov start
#' @export
print.bslib_toast <- function(x, ...) {
  x_tags <- x
  # Add "show" class to make toast visible when printed
  x_tags$attribs <- c(x_tags$attribs, list(class = "show"))
  x_tags <- as.tags(x_tags)
  print(as_fragment(x_tags))
  invisible(x)
}
# nocov end

#' Show or hide a toast notification
#'
#' @description
#' Displays a toast notification in a Shiny application.
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   actionButton("show_persistent", "Show Persistent Toast"),
#'   actionButton("hide_persistent", "Hide Toast")
#' )
#'
#' server <- function(input, output, session) {
#'   toast_id <- reactiveVal(NULL)
#'
#'   observeEvent(input$show_persistent, {
#'     id <- show_toast(
#'       toast(
#'         body = "This toast won't disappear automatically.",
#'         autohide = FALSE
#'       )
#'     )
#'     toast_id(id)
#'   })
#'
#'   observeEvent(input$hide_persistent, {
#'     req(toast_id())
#'     hide_toast(toast_id())
#'     toast_id(NULL)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @param toast A [toast()], or a string that will be automatically converted to
#'   a toast with default settings.
#' @param id String with the toast ID returned by `show_toast()` or a `toast`
#'   object provided that the `id` was set when created/shown.
#' @param ... Reserved for future extensions (currently ignored).
#' @param session Shiny session object.
#'
#' @return `show_toast()` Invisibly returns the toast ID (string) that can be
#'   used with `hide_toast()`.
#'
#' @family Toast components
#' @describeIn show_toast Show a toast notification.
#' @export
show_toast <- function(
  toast,
  ...,
  session = shiny::getDefaultReactiveDomain()
) {
  rlang::check_dots_empty()

  if (!inherits(toast, "bslib_toast")) {
    toast <- toast(toast)
  }

  if (length(toast$body) == 0 && length(toast$header) == 0) {
    rlang::warn("`toast` has no content; no toast to show.")
    return(invisible(NULL))
  }

  toast$id <- toast$id %||% toast_random_id()

  toasted <- processDeps(toast, session)

  data <- list(
    html = toasted$html,
    deps = toasted$deps,
    autohide = toast$autohide,
    duration = toast$duration,
    position = toast$position,
    id = toast$id
  )

  # Use custom message to show toast immediately
  session$sendCustomMessage("bslib.show-toast", dropNulls(data))
  invisible(toast$id)
}

#' @describeIn show_toast Hide a toast notification by ID.
#' @export
hide_toast <- function(id, ..., session = shiny::getDefaultReactiveDomain()) {
  rlang::check_dots_empty()

  if (inherits(id, "bslib_toast")) {
    if (is.null(id$id)) {
      rlang::abort("Cannot hide a toast without an ID. Provide the toast ID.")
    }
    id <- id$id
  }
  if (is.null(id)) {
    rlang::warn("`id` is NULL; no toast to hide.")
    return(invisible(NULL))
  }

  # Use custom message to hide toast immediately
  session$sendCustomMessage("bslib.hide-toast", list(id = id))
  invisible(id)
}

toast_component_header <- function(x) {
  UseMethod("toast_component_header")
}

#' @export
toast_component_header.character <- function(x) {
  strong(class = "me-auto", x)
}

#' @export
toast_component_header.bslib_toast_header <- function(x) {
  # Status text (small muted text)
  status_text <- if (!is.null(x$status)) {
    tags$small(class = "text-muted text-end", x$status)
  }

  tagList(
    if (!is.null(x$icon)) span(class = "toast-header-icon", x$icon),
    strong(
      class = "me-auto",
      class = if (!is.null(x$icon)) "ms-2",
      x$title
    ),
    status_text
  )
}

#' @export
toast_component_header.list <- function(x) {
  # Treat a bare list with a `title` element as a toast_header()
  if (!rlang::has_name(x, "title")) {
    rlang::abort(
      "Invalid toast header: must be a string, toast_header(), or a list with a `title` element."
    )
  }

  toast_component_header(structure(x, class = "bslib_toast_header"))
}

#' @export
toast_component_header.default <- function(x) {
  x
}

# Internal function to build toast HTML structure
toast_component <- function(
  body,
  header = NULL,
  icon = NULL,
  type = NULL,
  closable = TRUE,
  id = NULL,
  class = NULL,
  attribs = list()
) {
  # Set ARIA attributes based on toast urgency level
  # - danger toasts use role="alert" + aria-live="assertive" to immediately
  #   interrupt screen readers (critical errors requiring immediate attention)
  # - other toasts use role="status" + aria-live="polite" to announce updates
  #   without interrupting (non-critical, announced when convenient)
  aria_role <- if (!is.null(type) && type == "danger") "alert" else "status"
  aria_live <- if (!is.null(type) && type == "danger") "assertive" else "polite"

  type_class <- if (!is.null(type)) {
    paste0("text-bg-", type)
  }

  close_button <- tags$button(
    type = "button",
    class = "btn-close",
    `data-bs-dismiss` = "toast",
    `aria-label` = "Close"
  )

  header_tag <- if (!is.null(header)) {
    div(
      class = "toast-header",
      toast_component_header(header),
      if (closable) close_button
    )
  }

  # Body with optional close button
  # * If header exists, close button goes in header
  # * If no header, close button goes in body (if closable)
  body_has_close_btn <- is.null(header) && closable
  body_tag <- if (!body_has_close_btn && is.null(icon)) {
    div(class = "toast-body", body)
  } else {
    div(
      class = "toast-body d-flex gap-2",
      if (!is.null(icon)) span(class = "toast-body-icon", icon),
      div(class = "toast-body-content flex-grow-1", body),
      if (body_has_close_btn) close_button
    )
  }

  toast <- div(
    id = id,
    class = paste(c("toast", type_class, class), collapse = " "),
    role = aria_role,
    `aria-live` = aria_live,
    `aria-atomic` = "true",
    !!!attribs,
    header_tag,
    body_tag,
    component_dependencies()
  )

  as_fragment(tag_require(toast, version = 5))
}

toast_random_id <- function() {
  paste0("bslib-toast-", p_randomInt(1000, 10000000))
}

normalize_toast_position <- function(position = "bottom-right") {
  if (is.null(position) || length(position) == 0) {
    return("bottom-right")
  }

  original_position <- position # for error messages

  # If position is a vector, collapse it to a single string with spaces
  if (length(position) > 1) {
    position <- paste(position, collapse = " ")
  }

  position <- tolower(trimws(position))
  components <- strsplit(position, "[-\\s]+", perl = TRUE)[[1]]
  components <- components[nzchar(components)] # Remove empty strings

  vertical <- intersect(components, c("top", "middle", "bottom"))
  horizontal <- intersect(components, c("left", "center", "right"))
  invalid <- setdiff(components, c(vertical, horizontal))

  if (length(vertical) != 1 || length(horizontal) != 1 || length(invalid) > 0) {
    rlang::abort(
      paste0(
        "Invalid toast position: '",
        paste(original_position, collapse = " "),
        "'. ",
        "Must specify one vertical position (top, middle, bottom) and ",
        "one horizontal position (left, center, right)."
      )
    )
  }

  result <- paste0(vertical, "-", horizontal)

  rlang::arg_match(
    result,
    c(
      "top-left",
      "top-center",
      "top-right",
      "middle-left",
      "middle-center",
      "middle-right",
      "bottom-left",
      "bottom-center",
      "bottom-right"
    )
  )
}

# toast() constructor tests ----

test_that("toast() creates bslib_toast object with defaults", {
  t <- toast("Test message")

  expect_s3_class(t, "bslib_toast")
  expect_null(t$id)
  expect_true(t$autohide)
  expect_equal(t$duration, 5000) # Default 5 seconds in milliseconds
  expect_true(t$closable)
  expect_null(t$header)
  expect_null(t$icon)
  expect_equal(t$position, "top-right")
})

test_that("toast() validates position argument", {
  expect_no_error(toast("Test", position = "bottom-left"))
  expect_no_error(toast("Test", position = "top-center"))
  expect_no_error(toast("Test", position = "middle-center"))
  expect_snapshot(error = TRUE, toast("Test", position = "invalid"))
})

test_that("toast() validates type argument", {
  expect_no_error(toast("Test", type = "success"))
  expect_no_error(toast("Test", type = "danger"))
  expect_no_error(toast("Test", type = "info"))
  expect_snapshot(error = TRUE, toast("Test", type = "invalid"))
})

test_that("toast() type 'error' is aliased to 'danger'", {
  t <- toast("Test", type = "error")
  expect_equal(t$type, "danger")
})

test_that("toast() autohide disabled (0, NA, NULL)", {
  # When autohide is disabled, closable can be set to FALSE
  # This allows app authors to manage toast display manually
  t1 <- toast("Test", duration_s = 0, closable = FALSE)
  expect_false(t1$autohide)
  expect_false(t1$closable)

  t2 <- toast("Test", duration_s = NA, closable = FALSE)
  expect_false(t2$autohide)
  expect_false(t2$closable)

  t3 <- toast("Test", duration_s = NULL, closable = FALSE)
  expect_false(t3$autohide)
  expect_false(t3$closable)

  # closable can also be TRUE when autohide is disabled
  t4 <- toast("Test", duration_s = NA, closable = TRUE)
  expect_false(t4$autohide)
  expect_true(t4$closable)
})

test_that("toast() `closable` when autohide enabled", {
  # When autohide is enabled, user can control closable
  t_closable <- toast("Test", duration_s = 10, closable = TRUE)
  expect_true(t_closable$autohide)
  expect_equal(t_closable$duration, 10000) # Converted to milliseconds
  expect_true(t_closable$closable)

  t_not_closable <- toast("Test", duration_s = 5, closable = FALSE)
  expect_true(t_not_closable$autohide)
  expect_equal(t_not_closable$duration, 5000)
  expect_false(t_not_closable$closable)
})

test_that("toast() duration_s throws for invalid values", {
  expect_snapshot(error = TRUE, {
    toast("Test", duration_s = -5)
    toast("Test", duration_s = "invalid")
    toast("Test", duration_s = c(5, 10))
  })
})

test_that("toast() stores icon argument", {
  icon_elem <- span(class = "test-icon", HTML("&#9733;"))

  t <- toast("Test message", icon = icon_elem)

  expect_s3_class(t, "bslib_toast")
  expect_s3_class(t$icon, "shiny.tag")
  expect_equal(t$icon$attribs$class, "test-icon")
  expect_equal(t$icon$children[[1]], HTML("&#9733;"))
})

test_that("toast() icon is NULL by default", {
  t <- toast("Test message")
  expect_null(t$icon)
})


# toast() rendering tests ----

test_that("as.tags.bslib_toast creates proper HTML structure", {
  t <- toast(
    body = "Test message",
    header = "Test",
    type = "success",
    id = "test-toast"
  )

  tag <- as.tags(t)
  expect_s3_class(tag, "shiny.tag")
  expect_snapshot(cat(format(tag)))
})

test_that("as.tags.bslib_toast generates ID if not provided", {
  t <- toast("Test message")
  tag <- as.tags(t)

  html_str <- as.character(tag)
  # Verify an auto-generated ID is present
  expect_match(html_str, 'id="bslib-toast-[0-9]+"')
})

test_that("as.tags.bslib_toast respects accessibility attributes", {
  # Danger type gets assertive role
  t_danger <- toast("Error message", type = "danger", id = "danger-toast")
  html_danger <- as.character(as.tags(t_danger))
  expect_match(html_danger, 'role="alert"')
  expect_match(html_danger, 'aria-live="assertive"')
  expect_snapshot(cat(format(as.tags(t_danger))))

  # Info type gets polite role
  t_info <- toast("Info message", type = "info", id = "info-toast")
  html_info <- as.character(as.tags(t_info))
  expect_match(html_info, 'role="status"')
  expect_match(html_info, 'aria-live="polite"')
  expect_snapshot(cat(format(as.tags(t_info))))

  # NULL type (default) gets polite role
  t_default <- toast("Default message", id = "default-toast")
  html_default <- as.character(as.tags(t_default))
  expect_match(html_default, 'role="status"')
  expect_match(html_default, 'aria-live="polite"')
  expect_snapshot(cat(format(as.tags(t_default))))
})

test_that("as.tags.bslib_toast includes close button appropriately", {
  # With header, closable
  t_header <- toast(
    "Message",
    header = "Title",
    closable = TRUE,
    id = "header-toast"
  )
  expect_snapshot(cat(format(as.tags(t_header))))

  # Without header, closable
  t_no_header <- toast("Message", closable = TRUE, id = "no-header-toast")
  expect_snapshot(cat(format(as.tags(t_no_header))))

  # Non-closable with autohide
  t_non_closable <- toast(
    "Message",
    closable = FALSE,
    duration_s = 5,
    id = "non-closable-toast"
  )
  expect_snapshot(cat(format(as.tags(t_non_closable))))

  # Non-closable with autohide disabled (for manual management)
  t_manual <- toast(
    "Message",
    closable = FALSE,
    duration_s = NA,
    id = "manual-toast"
  )
  expect_snapshot(cat(format(as.tags(t_manual))))
})

test_that("toast() icon renders in body without header", {
  icon_elem <- span(class = "my-icon", HTML("&#9733;"))
  t <- toast("You have new messages", icon = icon_elem, id = "icon-toast")

  tag <- as.tags(t)
  html <- as.character(tag)

  # Icon should be in toast-body with special wrapper
  expect_match(html, 'class="toast-body d-flex gap-2"')
  expect_match(html, 'class="toast-body-icon"')
  expect_match(html, 'class="my-icon"')
  expect_match(html, "&#9733;")
  expect_match(html, 'class="toast-body-content flex-grow-1"')
  expect_snapshot(cat(format(tag)))
})

test_that("toast() icon renders in body with header", {
  icon_elem <- span(class = "header-icon", HTML("&#9733;"))
  t <- toast(
    "Message content",
    header = "New Mail",
    icon = icon_elem,
    id = "icon-header-toast"
  )

  tag <- as.tags(t)
  html <- as.character(tag)

  # Icon should still be in body when header is present
  expect_match(html, 'class="toast-body d-flex gap-2"')
  expect_match(html, 'class="toast-body-icon"')
  expect_match(html, 'class="header-icon"')
  expect_match(html, "&#9733;")
  expect_snapshot(cat(format(tag)))
})

test_that("toast() icon works with closable button in body", {
  icon_elem <- span(class = "alert-icon", HTML("&#9733;"))
  t <- toast(
    "Warning message",
    icon = icon_elem,
    closable = TRUE,
    id = "icon-closable-toast"
  )

  tag <- as.tags(t)
  html <- as.character(tag)

  # Should have both icon and close button in body
  expect_match(html, 'class="toast-body d-flex gap-2"')
  expect_match(html, 'class="toast-body-icon"')
  expect_match(html, 'class="alert-icon"')
  expect_match(html, "&#9733;")
  expect_match(html, 'class="btn-close"')
  expect_snapshot(cat(format(tag)))
})

test_that("toast() without icon or close button has simple body", {
  t <- toast(
    "Simple message",
    header = "Header",
    closable = FALSE,
    id = "simple-body-toast"
  )

  tag <- as.tags(t)
  html <- as.character(tag)

  # Should have simple toast-body (no d-flex gap-2)
  expect_match(html, 'class="toast-body"')
  expect_false(grepl('class="toast-body d-flex gap-2"', html))
  expect_false(grepl('toast-body-icon', html))
  expect_false(grepl('toast-body-content', html))
})


# toast_header() tests ----

test_that("toast_header() creates structured header data", {
  # Simple header with just title
  h1 <- toast_header("My Title")
  expect_s3_class(h1, "bslib_toast_header")
  expect_equal(as.character(h1$title), "My Title")
  expect_null(h1$icon)
  expect_null(h1$status)

  # Header with status text
  h2 <- toast_header("Success", status = "11 mins ago")
  expect_s3_class(h2, "bslib_toast_header")
  expect_equal(as.character(h2$title), "Success")
  expect_equal(h2$status, "11 mins ago")
})

test_that("toast_header() works with icons", {
  icon <- span(class = "test-icon")

  h <- toast_header("Title", icon = icon)
  expect_s3_class(h, "bslib_toast_header")
  expect_equal(as.character(h$title), "Title")
  expect_s3_class(h$icon, "shiny.tag")
  expect_equal(h$icon$attribs$class, "test-icon")
})

test_that("toast_header() icon renders in header", {
  icon_elem <- span(class = "header-test-icon", HTML("&#9733;"))
  h <- toast_header("Notification", icon = icon_elem, status = "now")

  t <- toast("Body content", header = h, id = "header-icon-toast")
  tag <- as.tags(t)
  html <- as.character(tag)

  # Icon should be in toast-header with wrapper
  expect_match(html, 'class="toast-header"')
  expect_match(html, 'class="toast-header-icon"')
  expect_match(html, 'class="header-test-icon"')
  expect_match(html, "&#9733;")
  expect_snapshot(cat(format(tag)))
})

test_that("toast_header() icon with status and title", {
  icon_elem <- span(class = "success-icon", "✓")
  h <- toast_header("Success", icon = icon_elem, status = "just now")

  t <- toast("Operation completed", header = h, id = "full-header-toast")
  tag <- as.tags(t)
  html <- as.character(tag)

  # Should have all three elements: icon, title, status
  expect_match(html, 'class="toast-header-icon"')
  expect_match(html, 'class="success-icon"')
  expect_match(html, "✓")
  expect_match(html, "Success")
  expect_match(html, "just now")
  expect_match(html, 'class="text-muted text-end"')
  expect_snapshot(cat(format(tag)))
})

test_that("toast() stores additional attributes", {
  t <- toast("Test", `data-test` = "value", class = "extra-class")

  expect_equal(t$attribs$`data-test`, "value")
  expect_equal(t$attribs$class, "extra-class")

  tag <- as.tags(t)
  html <- as.character(tag)
  expect_true(grepl('data-test="value"', html))
  expect_true(grepl('class="toast[^"]+extra-class"', html))
})


test_that("toast() type is reflected in rendered HTML", {
  t_success <- toast("Test", type = "success")
  expect_equal(t_success$type, "success")

  tag_success <- as.tags(t_success)
  html_success <- as.character(tag_success)
  expect_true(grepl("text-bg-success", html_success))

  t_danger <- toast("Test", type = "danger")
  expect_equal(t_danger$type, "danger")

  tag_danger <- as.tags(t_danger)
  html_danger <- as.character(tag_danger)
  expect_true(grepl("text-bg-danger", html_danger))
})

test_that("toast() position is stored correctly", {
  t1 <- toast("Test", position = "top-left")
  expect_equal(t1$position, "top-left")

  t2 <- toast("Test", position = "middle-center")
  expect_equal(t2$position, "middle-center")

  t3 <- toast("Test", position = "bottom-right")
  expect_equal(t3$position, "bottom-right")
})

# toast() header integration tests ----

test_that("toast header with character", {
  t <- toast("Body", header = "Simple Header")
  tag <- as.tags(t)
  html <- as.character(tag)

  expect_true(grepl("toast-header", html))
  expect_true(grepl("Simple Header", html))
  expect_true(grepl("me-auto", html))
})

test_that("toast header with toast_header()", {
  t <- toast(
    "Body",
    header = toast_header("Title", status = "just now")
  )
  tag <- as.tags(t)
  html <- as.character(tag)

  expect_true(grepl("toast-header", html))
  expect_true(grepl("Title", html))
  expect_true(grepl("just now", html))
  expect_true(grepl("text-muted", html))
})

test_that("toast header with list(title = ...) pattern", {
  # Bare list with title should work like toast_header()
  t <- toast(
    "Body",
    header = list(title = "Title", status = "just now")
  )
  tag <- as.tags(t)
  html <- as.character(tag)

  expect_true(grepl("toast-header", html))
  expect_true(grepl("Title", html))
  expect_true(grepl("just now", html))
  expect_true(grepl("text-muted", html))
})

test_that("toast header with custom tag", {
  t <- toast(
    "Body",
    header = div(class = "custom-header", "My Header")
  )

  tag <- as.tags(t)
  expect_equal(tag$children[[1]]$attribs$class, "toast-header")
  expect_equal(
    format(tag$children[[1]]$children[[1]]),
    '<div class="custom-header">My Header</div>'
  )
})

test_that("toast header can be modified after creation", {
  # Create toast with toast_header()
  t <- toast(
    "Body",
    header = toast_header("Original Title", status = "1 min ago")
  )

  # Modify the header
  t$header$title <- "Updated Title"
  t$header$status <- "just now"

  tag <- as.tags(t)
  html <- as.character(tag)

  expect_true(grepl("Updated Title", html))
  expect_true(grepl("just now", html))
  expect_false(grepl("Original Title", html))
  expect_false(grepl("1 min ago", html))
})

test_that("toast header with icon can be modified after creation", {
  # Create toast with toast_header() including icon
  icon1 <- span(class = "icon-1", "A")
  t <- toast(
    "Body",
    header = toast_header("Title", icon = icon1)
  )

  # Modify the header icon
  icon2 <- span(class = "icon-2", "B")
  t$header$icon <- icon2

  tag <- as.tags(t)
  html <- as.character(tag)

  expect_true(grepl("icon-2", html))
  expect_true(grepl("B", html, fixed = TRUE))
  expect_false(grepl("icon-1", html))
  expect_false(grepl("A", html, fixed = TRUE))
})

test_that("toast header with list pattern and icon", {
  # Bare list with title and icon
  icon_elem <- span(class = "list-icon", HTML("&#9733;"))
  t <- toast(
    "Body",
    header = list(
      title = "Notes",
      icon = icon_elem,
      status = "updated"
    )
  )

  tag <- as.tags(t)
  html <- as.character(tag)

  expect_true(grepl("toast-header", html))
  expect_true(grepl("Notes", html))
  expect_true(grepl("updated", html))
  expect_true(grepl("list-icon", html))
  expect_true(grepl("&#9733;", html))
})

test_that("toast header can be replaced with list pattern", {
  # Create toast with simple character header
  t <- toast("Body", header = "Simple")

  # Replace with list pattern
  t$header <- list(
    title = "New Title",
    status = "now",
    icon = span(class = "icon")
  )

  tag <- as.tags(t)
  html <- as.character(tag)

  expect_true(grepl("New Title", html))
  expect_true(grepl("now", html))
  expect_true(grepl("icon", html))
  expect_false(grepl("Simple", html))
})

test_that("toast with both header icon and body icon", {
  # Both header and body can have their own icons
  header_icon <- span(class = "h-icon", "H")
  body_icon <- span(class = "b-icon", "B")

  t <- toast(
    "Message content",
    header = toast_header("Title", icon = header_icon),
    icon = body_icon,
    id = "dual-icon-toast"
  )

  tag <- as.tags(t)
  html <- as.character(tag)

  # Both icons should be present in different locations
  expect_match(html, 'class="toast-header-icon"')
  expect_match(html, 'class="h-icon"')
  expect_match(html, 'class="toast-body-icon"')
  expect_match(html, 'class="b-icon"')
  expect_snapshot(cat(format(tag)))
})

# normalize_toast_position() helper tests ----

test_that("normalize_toast_position() handles standard kebab-case format", {
  expect_equal(normalize_toast_position("top-left"), "top-left")
  expect_equal(normalize_toast_position("bottom-right"), "bottom-right")
  expect_equal(
    normalize_toast_position("middle-center"),
    "middle-center"
  )
})

test_that("normalize_toast_position() handles space-separated format", {
  expect_equal(normalize_toast_position("top left"), "top-left")
  expect_equal(normalize_toast_position("bottom right"), "bottom-right")
  expect_equal(
    normalize_toast_position("middle center"),
    "middle-center"
  )
})

test_that("normalize_toast_position() handles reversed order", {
  expect_equal(normalize_toast_position("left top"), "top-left")
  expect_equal(normalize_toast_position("right bottom"), "bottom-right")
  expect_equal(
    normalize_toast_position("center middle"),
    "middle-center"
  )
})

test_that("normalize_toast_position() handles vector input", {
  expect_equal(normalize_toast_position(c("top", "left")), "top-left")
  expect_equal(
    normalize_toast_position(c("bottom", "right")),
    "bottom-right"
  )
  expect_equal(normalize_toast_position(c("left", "top")), "top-left")
  expect_equal(
    normalize_toast_position(c("right", "bottom")),
    "bottom-right"
  )
})

test_that("normalize_toast_position() is case-insensitive", {
  expect_equal(normalize_toast_position("TOP LEFT"), "top-left")
  expect_equal(normalize_toast_position("Bottom Right"), "bottom-right")
  expect_equal(
    normalize_toast_position("MIDDLE center"),
    "middle-center"
  )
})

test_that("normalize_toast_position() handles default NULL or empty", {
  expect_equal(normalize_toast_position(NULL), "bottom-right")
  expect_equal(normalize_toast_position(character(0)), "bottom-right")
})

test_that("normalize_toast_position() defaults to bottom-right when unspecified", {
  expect_equal(normalize_toast_position(), "bottom-right")
})

test_that("normalize_toast_position() handles all valid combinations", {
  # Space-separated
  expect_equal(normalize_toast_position("top left"), "top-left")
  expect_equal(normalize_toast_position("middle center"), "middle-center")
  expect_equal(normalize_toast_position("bottom right"), "bottom-right")

  # Reversed order
  expect_equal(normalize_toast_position("left top"), "top-left")
  expect_equal(normalize_toast_position("center middle"), "middle-center")
  expect_equal(normalize_toast_position("right bottom"), "bottom-right")

  # Vector input
  expect_equal(normalize_toast_position(c("top", "left")), "top-left")
  expect_equal(normalize_toast_position(c("center", "middle")), "middle-center")
  expect_equal(normalize_toast_position(c("bottom", "right")), "bottom-right")
})

test_that("normalize_toast_position() errors on invalid input", {
  expect_snapshot(error = TRUE, {
    # Missing components
    normalize_toast_position("top")
    normalize_toast_position("left")

    # Duplicate components
    normalize_toast_position("top bottom left")
    normalize_toast_position(c("top", "bottom", "left"))

    # Invalid components
    normalize_toast_position("top invalid")
    normalize_toast_position("foo bar")
  })
})

test_that("normalize_toast_position() handles extra whitespace", {
  expect_equal(normalize_toast_position("  top   left  "), "top-left")
  expect_equal(
    normalize_toast_position("bottom  right"),
    "bottom-right"
  )
})

# show_toast() and hide_toast() tests ----

test_that("toast() works with flexible position formats", {
  # Space-separated
  t1 <- toast("Test", position = "top left")
  expect_equal(t1$position, "top-left")

  # Reversed order
  t2 <- toast("Test", position = "right bottom")
  expect_equal(t2$position, "bottom-right")

  # Vector
  t3 <- toast("Test", position = c("middle", "center"))
  expect_equal(t3$position, "middle-center")
})

test_that("show_toast() returns the toast id", {
  local_mocked_bindings(
    toast_random_id = function() "bslib-toast-1234"
  )

  session <- list(sendCustomMessage = function(type, message) {
    expect_equal(type, "bslib.show-toast")
    expect_equal(message$id, !!exp_toast_id)
  })

  t <- toast("Test message")
  exp_toast_id <- "bslib-toast-1234"
  toast_id <- show_toast(t, session = session)
  expect_equal(toast_id, exp_toast_id)

  exp_toast_id <- "custom-id"
  t2 <- toast("Another message", id = exp_toast_id)
  toast_id2 <- show_toast(t2, session = session)
  expect_equal(toast_id2, exp_toast_id)
})

test_that("show_toast() converts string to toast automatically", {
  local_mocked_bindings(
    toast_random_id = function() "bslib-toast-auto"
  )

  message_sent <- FALSE
  session <- list(sendCustomMessage = function(type, message) {
    expect_equal(type, "bslib.show-toast")
    expect_equal(message$id, "bslib-toast-auto")
    message_sent <<- TRUE
  })

  # Pass a plain string instead of a toast object
  toast_id <- show_toast("Simple message", session = session)
  expect_true(message_sent)
  expect_equal(toast_id, "bslib-toast-auto")
})

test_that("show_toast() and hide_toast() warn if nothing to show/hide", {
  session <- list(sendCustomMessage = function(type, message) {
    stop("sendCustomMessage should not be called")
  })

  expect_snapshot({
    hide_toast(show_toast(toast(), session = session), session = session)
  })
})

test_that("hide_toast() works", {
  session <- list(sendCustomMessage = function(type, message) {
    expect_equal(type, "bslib.hide-toast")
    expect_equal(message$id, !!exp_toast_id)
  })

  exp_toast_id <- "bslib-toast-1234"
  t_id <- hide_toast(exp_toast_id, session = session)
  expect_equal(t_id, exp_toast_id)

  exp_toast_id <- "custom-id"
  t_id2 <- hide_toast(toast("Test", id = exp_toast_id), session = session)
  expect_equal(t_id2, exp_toast_id)

  expect_snapshot(error = TRUE, {
    hide_toast(toast())
  })
})

#' Add low-level theming customizations
#'
#' These functions provide direct access to the layers of a bslib theme created
#' with [bs_theme()]. Learn more about [composable Sass
#' layers](https://rstudio.github.io/sass/articles/sass.html#layering) on the
#' \pkg{sass} website.
#'
#' @details
#' Compared to higher-level theme customization available in [bs_theme()], these
#' functions are a more direct interface to Bootstrap Sass, and therefore, do
#' nothing to ensure theme customizations are portable between major Bootstrap
#' versions.
#'
#' @inheritParams bs_theme_update
#' @param ...
#'  * `bs_add_variables()`: Should be named Sass variables or values that can be
#'    passed in directly to the `defaults` argument of a [sass::sass_layer()].
#'  * `bs_bundle()`: Should be arguments that can be handled by
#'    [sass::sass_bundle()] to be appended to the `theme`
#' @param .where Whether to place the variable definitions before other Sass
#'   `"defaults"`, after other Sass `"declarations"`, or after other Sass
#'   `"rules"`.
#' @param .default_flag Whether or not to add a `!default` flag (if missing) to
#'   variable expressions. It's recommended to keep this as `TRUE` when `.where
#'   = "defaults"`.
#'
#' @return Returns a modified [bs_theme()] object.
#'
#' @family Bootstrap theme functions
#'
#' @seealso [bs_theme()] creates a Bootstrap theme object, and is the best place
#'   to start learning about bslib's theming capabilities.
#'
#' @references
#'   * bslib's theming capabilities are powered by
#'     [the {sass} package](https://rstudio.github.io/sass/).
#'   * Learn more about
#'     [composable Sass layers](https://rstudio.github.io/sass/articles/sass.html#layering)
#'     on the \pkg{sass} website.
#'
#' @examplesIf rlang::is_interactive()
#'
#' # Function to preview the styling a (primary) Bootstrap button
#' library(htmltools)
#' button <- tags$a(class = "btn btn-primary", href = "#", role = "button", "Hello")
#' preview_button <- function(theme) {
#'   browsable(tags$body(bs_theme_dependencies(theme), button))
#' }
#'
#' # Here we start with a theme based on a Bootswatch theme,
#' # then override some variable defaults
#' theme <- bs_add_variables(
#'   bs_theme(bootswatch = "sketchy", primary = "orange"),
#'   "body-bg" = "#EEEEEE",
#'   "font-family-base" = "monospace",
#'   "font-size-base" = "1.4rem",
#'   "btn-padding-y" = ".16rem",
#'   "btn-padding-x" = "2rem"
#' )
#'
#' preview_button(theme)
#'
#' # If you need to set a variable based on another Bootstrap variable
#' theme <- bs_add_variables(theme, "body-color" = "$success", .where = "declarations")
#' preview_button(theme)
#'
#' # Start a new global theme and add some custom rules that
#' # use Bootstrap variables to define a custom styling for a
#' # 'person card'
#' person_rules <- system.file("custom", "person.scss", package = "bslib")
#' theme <- bs_add_rules(bs_theme(), sass::sass_file(person_rules))
#'
#' # Include custom CSS that leverages bootstrap Sass variables
#' person <- function(name, title, company) {
#'   tags$div(
#'     class = "person",
#'     h3(class = "name", name),
#'     div(class = "title", title),
#'     div(class = "company", company)
#'   )
#' }
#'
#' page_fluid(
#'   theme = theme,
#'   person("Andrew Carnegie", "Owner", "Carnegie Steel Company"),
#'   person("John D. Rockefeller", "Chairman", "Standard Oil")
#' )
#'
#' @describeIn bs_bundle Add Bootstrap Sass
#'   [variable defaults](https://rstudio.github.io/bslib/articles/bs5-variables/index.html).
#' @export
bs_add_variables <- function(
  theme,
  ...,
  .where = "defaults",
  .default_flag = identical(.where, "defaults")
) {
  assert_bs_theme(theme)

  vars <- rlang::list2(...)
  if (any(names2(vars) == "")) {
    stop("Variables must be named.", call. = FALSE)
  }

  # Workaround to the problem of 'blue' winning in the scenario of:
  # bs_add_variables("body-bg" = "blue")
  # bs_add_variables("body-bg" = "red")
  if (.default_flag) {
    vars <- ensure_default_flag(vars)
  }

  bs_bundle(
    theme,
    do.call(sass_layer, rlang::list2(!!.where := vars))
  )
}

# Given a named list of variable definitions,
# searches each variable's expression for a !default flag,
# and if missing, adds it.
ensure_default_flag <- function(x) {
  Map(
    x,
    rlang::names2(x),
    f = function(val, nm) {
      # sass::font_collection() has its own default_flag, so warn if they conflict
      if (sass::is_font_collection(val)) {
        if (identical(val$default_flag, FALSE)) {
          message(
            "Ignoring `bs_add_variables()`'s `.default_flag = TRUE` for ",
            "the ",
            nm,
            " variable (since it has its own `default_flag`)."
          )
        }
        return(val)
      }
      val <- paste(as_sass(val), collapse = "\n")
      if (grepl("!default\\s*;*\\s*$", val)) {
        val
      } else {
        paste(sub(";+$", "", val), "!default")
      }
    }
  )
}

#' @describeIn bs_bundle Add additional
#'   [Sass rules](https://sass-lang.com/documentation/style-rules).
#' @param rules Sass rules. Anything understood by [sass::as_sass()] may be
#'   provided (e.g., a list, character vector, [sass::sass_file()], etc)
#' @export
bs_add_rules <- function(theme, rules) {
  bs_bundle(theme, sass_layer(rules = rules))
}

#' @describeIn bs_bundle Add additional [Sass
#'   functions](https://rstudio.github.io/sass/articles/sass.html#functions).
#' @param functions A character vector or [sass::sass_file()] containing
#'   functions definitions.
#' @export
bs_add_functions <- function(theme, functions) {
  bs_bundle(theme, sass_layer(functions = functions))
}

#' @describeIn bs_bundle Add additional [Sass
#'   mixins](https://rstudio.github.io/sass/articles/sass.html#mixins).
#' @param mixins A character vector or [sass::sass_file()] containing
#'   mixin definitions.
#' @export
bs_add_mixins <- function(theme, mixins) {
  bs_bundle(theme, sass_layer(mixins = mixins))
}

#' @describeIn bs_bundle Add additional [sass::sass_bundle()] objects to an
#'   existing `theme`.
#' @export
bs_bundle <- function(theme, ...) {
  assert_bs_theme(theme)
  structure(
    sass_bundle(theme, ...),
    class = class(theme)
  )
}

#' Deprecated functions
#'
#' These functions have been deprecated but remain for backwards compatibility.
#'
#' @keywords internal
#' @rdname deprecated
#' @returns a [bs_theme()] object.
#' @export
bs_theme_new <- function(...) {
  .Deprecated("bs_global_theme")
  bs_global_theme(...)
}

#' @rdname deprecated
#' @export
bs_theme_clear <- function(...) {
  .Deprecated("bs_global_clear")
  bs_global_clear(...)
}

#' @rdname deprecated
#' @export
bs_theme_get <- function(...) {
  .Deprecated("bs_global_get")
  bs_global_get(...)
}

#' @rdname deprecated
#' @export
bs_theme_set <- function(...) {
  .Deprecated("bs_global_set")
  bs_global_set(...)
}

#' @rdname deprecated
#' @export
bs_theme_base_colors <- function(...) {
  .Deprecated("bs_global_base_colors")
  bs_global_base_colors(...)
}

bs_global_base_colors <- function(bg = "#FFFFFF", fg = "#000000") {
  theme <- assert_global_theme("bs_global_base_colors()")
  theme <- bs_base_colors(theme, bg = bg, fg = fg)
  bs_global_set(theme)
}

#' @rdname deprecated
#' @export
bs_theme_accent_colors <- function(...) {
  .Deprecated("bs_global_accent_colors")
  bs_global_accent_colors(...)
}

bs_global_accent_colors <- function(primary = NULL, secondary = NULL, success = NULL,
                                    info = NULL, warning = NULL, danger = NULL) {
  theme <- assert_global_theme("bs_global_accent_colors()")
  theme <- bs_accent_colors(
    theme, primary = primary, secondary = secondary, success = success,
    info = info, warning = warning, danger = danger
  )
  bs_global_set(theme)
}

#' @rdname deprecated
#' @export
bs_theme_fonts <- function(...) {
  .Deprecated("bs_global_fonts")
  bs_global_fonts(...)
}

bs_global_fonts <- function(base = NULL, code = NULL, heading = NULL) {
  theme <- assert_global_theme("bs_global_fonts()")
  theme <- bs_fonts(theme, base = base, code = code, heading = heading)
  bs_global_set(theme)
}

#' @rdname deprecated
#' @export
bs_theme_add_variables <- function(...) {
  .Deprecated("bs_global_add_variables")
  bs_global_add_variables(...)
}

#' @rdname deprecated
#' @export
bs_theme_add <- function(...) {
  .Deprecated("bs_global_bundle")
  bs_global_bundle(sass::sass_layer(...))
}


#' @rdname deprecated
#' @export
bs_theme_get_variables <- function(...) {
  .Deprecated("bs_global_get_variables")
  bs_global_get_variables(...)
}

bs_global_get_variables <- function(varnames) {
  theme <- assert_global_theme("bs_global_get_variables()")
  bs_get_variables(theme, varnames)
}

#' @rdname deprecated
#' @export
bootstrap <- function(theme = bs_theme_get(), ...) {
  .Deprecated("bs_theme_dependencies")
  bs_theme_dependencies(theme, ...)
}

#' @rdname deprecated
#' @export
bootstrap_sass <- function(rules = list(), theme = bs_theme_get(), ...) {
  .Deprecated("sass::sass_partial")
  theme <- as_bs_theme(theme)
  theme$rules <- ""
  sass_partial(rules, theme)
}

#' @rdname deprecated
#' @export
bs_add_declarations <- function(theme, declarations) {
  .Deprecated("bs_add_mixins")
  bs_bundle(theme, sass_layer(declarations = declarations))
}

library(shiny)
library(bslib)
library(ggplot2)
library(lattice)
library(thematic)
library(rlang)
library(curl)
library(DT)
library(knitr)
library(reactable)
library(reshape2)
library(bsicons)
library(ggridges)

theme <- bs_theme(preset = "shiny")

old_theme <- bs_global_get()
bs_global_set(theme)
onStop(function() bs_global_set(old_theme))

bslib:::as_themer_app(
  system.file("themer-demo", package = "bslib")
)

#' Available Bootstrap versions
#'
#' @return Returns a list of the Bootstrap versions available.
#'
#' @family Bootstrap theme utility functions
#' @export
versions <- function() {
  rlang::set_names(
    list("5", "4", "3"),
    c(version_bs5, version_bs4, version_bs3)
  )
}

#' @export
#' @rdname versions
version_default <- function() {
  versions()[[1]]
}

# TODO: make this a getter/setter?

library(shiny)
library(bslib)
library(bsicons)
library(plotly)

ui <- page_fluid(
  theme = bslib::bs_theme(
    base_font = font_google("Atkinson Hyperlegible")
  ),
  br(),
  layout_column_wrap(
    width = "200px",
    class = "mt-3",
    value_box(
      title = "Unemployment Rate",
      value = "2.7%",
      p("Started at 1.5%"),
      p("Averaging 3%"),
      p("Peaked at 5.2% in Dec 1982"),
      showcase = plotlyOutput("unemploy"),
      full_screen = TRUE
    ),
    value_box(
      title = "Personal Savings Rate",
      value = "7.6%",
      p("Started at 12.6%"),
      p("Averaging 8.6%"),
      p("Peaked at 17.3% in May 1975"),
      showcase = plotlyOutput("psavert"),
      showcase_layout = showcase_top_right(),
      full_screen = TRUE,
      theme = "success"
    ),
    value_box(
      title = "Personal Consumption",
      value = "$3.8B",
      p("Started at $0.25B"),
      p("Averaging $1.7B"),
      showcase = bsicons::bs_icon("piggy-bank", size = "100%"),
      full_screen = TRUE,
      theme = "danger"
    )
  )
)

server <- function(input, output) {
  output$unemploy <- renderPlotly({
    plotly_time_series(
      economics,
      x = ~date,
      y = ~ 100 * unemploy / pop
    )
  })

  output$psavert <- renderPlotly({
    plotly_time_series(
      economics,
      x = ~date,
      y = ~psavert
    )
  })

  output$pce <- renderPlotly({
    plotly_time_series(
      economics,
      x = ~date,
      y = ~ 100 * pce / pop
    )
  })

  plotly_time_series <- function(d, x, y) {
    info <- getCurrentOutputInfo()
    large <- isTRUE(info$height() > 200)

    plot_ly(d, x = x, y = y) %>%
      add_lines(
        color = I(info$fg()),
        span = I(1),
        #hoverinfo = if (!large) "none",
        fill = 'tozeroy',
        alpha = 0.2
      ) %>%
      layout(
        hovermode = "x",
        margin = list(t = 0, r = 0, l = 0, b = 0),
        font = list(color = info$fg()),
        paper_bgcolor = "transparent",
        plot_bgcolor = "transparent",
        xaxis = list(
          title = "",
          visible = large,
          showgrid = FALSE
        ),
        yaxis = list(
          title = "",
          visible = large,
          showgrid = FALSE
        )
      ) %>%
      config(displayModeBar = FALSE)
  }
}

shinyApp(ui, server)

rlang::check_installed("reshape2")
library(reshape2)

data(tips, package = "reshape2")


tipsUI <- function(id) {
  ns <- NS(id)

  layout_sidebar(
    sidebar = sidebar(
      title = tooltip(
        span(
          "Restaurant tipping",
          bsicons::bs_icon("info-circle-fill"),
          class = "sidebar-title"
        ),
        "One waiter recorded information about each tip he received",
        "over a period of a few months working in one restaurant."
      ),
      sliderInput(
        ns("total_bill"),
        "Bill amount",
        min(tips$total_bill),
        max(tips$total_bill),
        value = range(tips$total_bill),
        pre = "$"
      ),
      checkboxGroupInput(ns("time"), "Food service", c("Lunch", "Dinner"), c("Lunch", "Dinner"), inline = TRUE),
      actionButton(ns("reset"), "Reset filter"),
    ),
    layout_column_wrap(
      width = 1/3,
      fill = FALSE,
      value_box(
        "Total tippers",
        uiOutput(ns("total_tippers"), container = h2),
        showcase = bsicons::bs_icon("person")
      ),
      value_box(
        "Average tip",
        uiOutput(ns("average_tip"), container = h2),
        showcase = bsicons::bs_icon("wallet2")
      ),
      value_box(
        "Average bill",
        uiOutput(ns("average_bill"), container = h2),
        showcase = bsicons::bs_icon("currency-dollar")
      )
    ),
    layout_column_wrap(
      width = 1/2,
      class = "mt-3",
      card(
        full_screen = TRUE,
        card_header(
          "Total bill vs tip",
          popover(
            bsicons::bs_icon("gear"),
            radioButtons(
              ns("scatter_color"), NULL, inline = TRUE,
              c("none", "sex", "smoker", "day", "time")
            ),
            title = "Add a color variable",
            placement = "top"
          ),
          class = "d-flex justify-content-between align-items-center"
        ),
        plotOutput(ns("scatterplot"))
      ),
      card(
        full_screen = TRUE,
        class = "bslib-card-table-sm",
        card_header("Tips data"),
        DT::dataTableOutput(ns("table"))
      ),
    ),
    card(
      full_screen = TRUE,
      class = "mt-3",
      card_header(
        "Tip percentages",
        popover(
          bsicons::bs_icon("gear"),
          radioButtons(
            ns("tip_perc_y"), "Split by:", inline = TRUE,
            c("sex", "smoker", "day", "time"), "day"
          ),
          radioButtons(
            ns("tip_perc_facet"), "Facet by:", inline = TRUE,
            c("none", "sex", "smoker", "day", "time"), "none"
          ),
          title = "Add a color variable"
        ),
        class = "d-flex justify-content-between align-items-center"
      ),
      plotOutput(ns("tip_perc"))
    )
  )
}



tipsServer <- function(id) {
  moduleServer(id, function(input, output, session) {

    tips_data <- reactive({
      d <- tips
      d <- d[d$total_bill >= input$total_bill[1] & d$total_bill <= input$total_bill[2], ]
      d <- d[d$time %in% input$time, ]
      d
    })

    # TODO: maybe use reactable if and when this lands?
    # https://github.com/glin/reactable/pull/96
    output$table <- DT::renderDataTable({
      DT::datatable(tips_data(), fillContainer = TRUE, rownames = FALSE)
    })

    output$scatterplot <- renderPlot({
      validate(need(
        nrow(tips_data()) > 0,
        "No tips match the current filter. Try adjusting your filter settings."
      ))
      color <-  if (input$scatter_color != "none") sym(input$scatter_color)
      ggplot(tips_data(), aes(x = total_bill, y = tip, color = !!color)) +
        geom_point() +
        geom_smooth() +
        labs(x = NULL, y = NULL)
    })

    output$tip_perc <- renderPlot({
      validate(need(
        requireNamespace("ggridges", quietly = TRUE),
        "Please install the ggridges package to see this plot."
      ))
      validate(need(
        requireNamespace("ggridges", quietly = TRUE),
        "Please install the ggridges package to see this plot."
      ))
      p <- ggplot(tips_data(), aes(x = tip / total_bill, y = !!sym(input$tip_perc_y))) +
        ggridges::geom_density_ridges(scale = 0.9) +
        coord_cartesian(clip = "off") +
        labs(x = NULL, y = NULL)

      if (input$tip_perc_facet != "none") {
        p <- p + facet_wrap(vars(!!sym(input$tip_perc_facet)))
      }

      p
    })

    output$total_tippers <- renderUI({
      nrow(tips_data())
    })

    output$average_bill <- renderUI({
      if (nrow(tips_data()) == 0) return(HTML("&ndash;"))
      scales::dollar(mean(tips_data()$total_bill))
    })

    output$average_tip <- renderUI({
      if (nrow(tips_data()) == 0) return(HTML("&ndash;"))
      d <- tips_data()
      scales::percent(mean(d$tip / d$total_bill))
    })

    observeEvent(input$reset, {
      updateSliderInput(session, "total_bill", value = range(tips$total_bill))
      updateCheckboxGroupInput(session, "time", selected = c("Lunch", "Dinner"))
    })
  })
}

# Class added or removed from `bs_theme` when a theme includes a preset
THEME_PRESET_CLASS <- "bs_theme_with_preset"

resolve_bs_preset <- function(
  preset = NULL,
  bootswatch = NULL,
  version = version_default()
) {
  if (is.null(preset) && is.null(bootswatch)) {
    return(NULL)
  }

  assert_preset_scalar_string(preset)
  assert_preset_scalar_string(bootswatch)
  assert_preset_only_one_name_arg(preset, bootswatch)

  version <- switch_version(version, five = "5", four = "4", three = "3")
  preset_name <- preset %||% bootswatch

  if (preset_name %in% c("default", "bootstrap")) {
    # "bootstrap" means no preset bundle, just bare default Bootstrap
    return(new_bs_preset("bootstrap", version))
  }

  builtin_themes <- builtin_themes(version)

  if (length(builtin_themes) > 0 && preset_name %in% builtin_themes) {
    return(new_bs_preset(preset_name, version, type = "builtin"))
  }

  bootswatch_themes <- bootswatch_themes(version)

  if (length(bootswatch_themes) > 0 && preset_name %in% bootswatch_themes) {
    return(new_bs_preset(preset_name, version, type = "bootswatch"))
  }

  abort_preset_unknown_name(preset_name, version)
}

new_bs_preset <- function(name, version, type = NULL, ...) {
  preset <- list(
    version = version, # bootstrap version
    name = name, # preset name
    type = type, # preset type (e.g. "builtin", "bootswatch")
    ...
  )

  structure(dropNulls(preset), class = "bs_preset")
}

# The `bs_preset_bundle()` function is the main entry point for creating a SASS
# bundle for a theme preset. This calls out to the appropriate functions to
# create a bundle for a built-in theme or for a Bootswatch theme.
bs_preset_bundle <- function(preset) {
  if (!inherits(preset, "bs_preset")) {
    return(NULL)
  }
  if (is.null(preset$type)) {
    return(NULL)
  }

  switch(
    preset$type,
    bootswatch = bootswatch_bundle(preset$name, version = preset$version),
    builtin = builtin_bundle(preset$name, version = preset$version),
    stop("Unknown preset type: ", preset$type)
  )
}

theme_preset_info <- function(theme) {
  if (!is_bs_theme(theme)) {
    return(NULL)
  }

  info <- bs_get_variables(
    theme,
    c("bslib-preset-type", "bslib-preset-name", "bootstrap-version")
  )

  name <- info[["bslib-preset-name"]]
  type <- info[["bslib-preset-type"]]

  new_bs_preset(
    name = if (!is.na(name)) name else "bootstrap",
    version = info[["bootstrap-version"]],
    type = if (!is.na(type)) type
  )
}

assert_preset_scalar_string <- function(var, .frame = rlang::caller_env()) {
  var_name <- deparse(substitute(var))

  if (is.null(var) || rlang::is_string(var)) {
    return(invisible())
  }

  msg <- c(
    sprintf(
      "The preset theme `%s` must be a single character string.",
      var_name
    ),
    "x" = sprintf('Bad: `%s = c("flatly", "darkly")`', var_name),
    "v" = sprintf('Good: `%s = "flatly"`', var_name)
  )
  rlang::abort(msg, .frame = .frame)
}

assert_preset_only_one_name_arg <- function(
  preset,
  bootswatch,
  .frame = rlang::caller_env()
) {
  both_provided <- !is.null(preset) && !is.null(bootswatch)

  if (!both_provided) {
    return(invisible())
  }

  msg <- c(
    "Only one of `preset` or `bootswatch` may be provided (`preset` is the preferred choice).",
    "i" = "Did you mean one of the following options?",
    "*" = sprintf('`preset = "%s"`', preset),
    "*" = sprintf('`preset = "%s"`', bootswatch),
    "*" = sprintf('`bootswatch = "%s"`', bootswatch)
  )
  rlang::abort(msg, .frame = .frame)
}

abort_preset_unknown_name <- function(
  name,
  version,
  .frame = rlang::caller_env()
) {
  msg <- c(
    sprintf(
      "'%s' is not a known preset theme name for Bootstrap version %s.",
      name,
      version
    ),
    "i" = "You can list available preset themes:",
    "*" = sprintf("Built-in: `builtin_themes(%s)`", version),
    "*" = sprintf("Bootswatch: `bootswatch_themes(%s)`.", version)
  )
  rlang::abort(msg, .frame = .frame)
}

#' Navigation containers
#'
#' Render a collection of [nav()] items into a container.
#'
#' @param ... a collection of [nav()] items.
#' @param id a character string used for dynamically updating the container (see [nav_select()]).
#' @param selected a character string matching the `value` of a particular [nav()] item to selected by default.
#' @param header UI element(s) ([tags]) to display _above_ the nav content.
#' @param footer UI element(s) ([tags]) to display _below_ the nav content.
#' @export
#' @seealso [nav()], [nav_select()].
#' @rdname navs
#' @examples
#'
#' library(shiny)
#'
#' nav_items <- function(prefix) {
#'   list(
#'     nav("a", paste(prefix, ": tab a content")),
#'     nav("b", paste(prefix, ": tab b content")),
#'     nav_item(
#'       tags$a(icon("github"), "Shiny", href = "https://github.com/rstudio/shiny", target = "_blank")
#'     ),
#'     nav_spacer(),
#'     nav_menu(
#'       "Other links", align = "right",
#'       nav("c", paste(prefix, ": tab c content")),
#'       nav_item(
#'         tags$a(icon("r-project"), "RStudio", href = "https://rstudio.com", target = "_blank")
#'       )
#'     )
#'   )
#' }
#'
#' if (interactive()) {
#'   shinyApp(
#'     page_navbar(
#'       title = "page_navbar()",
#'       bg = "#0062cc",
#'       !!!nav_items("page_navbar()"),
#'       footer = div(
#'         style = "width:80%; margin: 0 auto",
#'         h4("navs_tab()"),
#'         navs_tab(!!!nav_items("navs_tab()")),
#'         h4("navs_pill()"),
#'         navs_pill(!!!nav_items("navs_pill()")),
#'         h4("navs_tab_card()"),
#'         navs_tab_card(!!!nav_items("navs_tab_card()")),
#'         h4("navs_pill_card()"),
#'         navs_pill_card(!!!nav_items("navs_pill_card()")),
#'         h4("navs_pill_list()"),
#'         navs_pill_list(!!!nav_items("navs_pill_list()"))
#'       )
#'     ),
#'     function(...) { }
#'   )
#' }
navs_tab <- function(..., id = NULL, selected = NULL,
                     header = NULL, footer = NULL) {
  tabs <- tabsetPanel_(
    ..., type = "tabs", id = id, selected = selected,
    header = header, footer = footer
  )
  as_fragment(tabs)
}

#' @export
#' @rdname navs
navs_pill <- function(..., id = NULL, selected = NULL,
                      header = NULL, footer = NULL) {
  pills <- tabsetPanel_(
    ..., type = "pills", id = id, selected = selected,
    header = header, footer = footer
  )
  as_fragment(pills)
}

#' @export
#' @inheritParams shiny::navlistPanel
#' @rdname navs
navs_pill_list <- function(..., id = NULL, selected = NULL,
                           header = NULL, footer = NULL,
                           well = TRUE, fluid = TRUE,
                           widths = c(4, 8)) {
  pill_list <- navlistPanel_(
    ..., id = id, selected = selected,
    header = header, footer = footer,
    well = well, fluid = fluid,
    widths = widths
  )
  as_fragment(pill_list)
}

#' @export
#' @rdname navs
navs_hidden <- function(..., id = NULL, selected = NULL,
                        header = NULL, footer = NULL) {
  hidden <- tabsetPanel_(
    ..., type = "hidden", id = id, selected = selected,
    header = header, footer = footer
  )
  as_fragment(hidden)
}


#' @inheritParams shiny::navbarPage
#' @param bg a CSS color to use for the navbar's background color.
#' @param inverse Either `TRUE` for a light text color or `FALSE` for a dark
#'   text color. If `"auto"` (the default), the best contrast to `bg` is chosen.
#' @export
#' @rdname navs
navs_bar <- function(..., title = NULL, id = NULL, selected = NULL,
                     # TODO: add sticky-top as well?
                     position = c("static-top", "fixed-top", "fixed-bottom"),
                     header = NULL, footer = NULL,
                     bg = NULL, inverse = "auto",
                     collapsible = TRUE, fluid = TRUE) {

  if (identical(inverse, "auto")) {
    inverse <- TRUE
    if (!is.null(bg)) {
      bg <- htmltools::parseCssColors(bg)
      bg_contrast <- bs_get_contrast(bs_theme("navbar-bg" = bg), "navbar-bg")
      inverse <- col2rgb(bg_contrast)[1,] > 127.5
    }
  }

  navbar <- navbarPage_(
    title = title, ..., id = id, selected = selected,
    position = match.arg(position),
    header = header, footer = footer, collapsible = collapsible,
    inverse = inverse, fluid = fluid
  )

  if (!is.null(bg)) {
    # navbarPage_() returns a tagList() of the nav and content
    navbar[[1]] <- tagAppendAttributes(
      navbar[[1]], style = css(background_color = paste(bg, "!important"))
    )
  }

  as_fragment(navbar, page = page)
}



# -----------------------------------------------------------------------
# 'Internal' tabset logic that was pulled directly from shiny/R/bootstrap.R
#  (with minor modifications)
# -----------------------------------------------------------------------

navbarPage_ <- function(title,
                       ...,
                       id = NULL,
                       selected = NULL,
                       position = c("static-top", "fixed-top", "fixed-bottom"),
                       header = NULL,
                       footer = NULL,
                       inverse = FALSE,
                       collapsible = FALSE,
                       fluid = TRUE,
                       theme = NULL,
                       windowTitle = title,
                       lang = NULL) {

  # alias title so we can avoid conflicts w/ title in withTags
  pageTitle <- title

  # navbar class based on options
  navbarClass <- "navbar navbar-default"
  position <- match.arg(position)
  if (!is.null(position))
    navbarClass <- paste0(navbarClass, " navbar-", position)
  if (inverse)
    navbarClass <- paste(navbarClass, "navbar-inverse")

  if (!is.null(id))
    selected <- shiny::restoreInput(id = id, default = selected)

  # build the tabset
  tabset <- buildTabset(..., ulClass = "nav navbar-nav", id = id, selected = selected)

  containerClass <- paste0("container", if (fluid) "-fluid")

  # built the container div dynamically to support optional collapsibility
  if (collapsible) {
    navId <- paste0("navbar-collapse-", p_randomInt(1000, 10000))
    containerDiv <- div(
      class = containerClass,
      div(
        class = "navbar-header",
        tags$button(
          type = "button",
          class = "navbar-toggle collapsed",
          `data-toggle` = "collapse",
          `data-target` = paste0("#", navId),
          # data-bs-* is for BS5+
          `data-bs-toggle` = "collapse",
          `data-bs-target` = paste0("#", navId),
          span(class="sr-only visually-hidden", "Toggle navigation"),
          span(class = "icon-bar"),
          span(class = "icon-bar"),
          span(class = "icon-bar")
        ),
        span(class = "navbar-brand", pageTitle)
      ),
      div(
        class = "navbar-collapse collapse",
        id = navId, tabset$navList
      )
    )
  } else {
    containerDiv <- div(
      class = containerClass,
      div(
        class = "navbar-header",
        span(class = "navbar-brand", pageTitle)
      ),
      tabset$navList
    )
  }

  # Bootstrap 3 explicitly supported "dropup menus" via .navbar-fixed-bottom,
  # but BS4+ requires .dropup on menus with .navbar.fixed-bottom
  if (position == "fixed-bottom") {
    containerDiv <- tagQuery(containerDiv)$
      find(".dropdown-menu")$
      parent()$
      addClass("dropup")$
      allTags()
  }

  # build the main tab content div
  contentDiv <- div(class = containerClass)
  if (!is.null(header))
    contentDiv <- tagAppendChild(contentDiv, div(class = "row", header))
  contentDiv <- tagAppendChild(contentDiv, tabset$content)
  if (!is.null(footer))
    contentDiv <- tagAppendChild(contentDiv, div(class = "row", footer))

  # *Don't* wrap in bootstrapPage() (shiny::navbarPage()) does that part
  tagList(
    tags$nav(class = navbarClass, role = "navigation", containerDiv),
    contentDiv
  )
}

navbarMenu_ <- function(title, ..., menuName = title, icon = NULL, align) {
  icon <- prepTabIcon(icon)
  structure(
    list(
      title = title,
      menuName = menuName,
      tabs = list2(...),
      # Here for legacy reasons
      # https://github.com/cran/miniUI/blob/74c87d3/R/layout.R#L369
      iconClass = tagGetAttribute(icon, "class"),
      icon = icon,
      align = align
    ),
    class = "shiny.navbarmenu"
  )
}

isNavbarMenu <- function(x) {
  inherits(x, "shiny.navbarmenu")
}

tabPanel_ <- function(title, ..., value = title, icon = NULL) {
  icon <- prepTabIcon(icon)
  pane <- div(
    class = "tab-pane",
    title = title,
    `data-value` = value,
    # Here for legacy reasons
    # https://github.com/cran/miniUI/blob/74c87d/R/layout.R#L395
    `data-icon-class` = tagGetAttribute(icon, "class"),
    ...
  )
  attr(pane, "_shiny_icon") <- icon
  pane
}

isTabPanel <- function(x) {
  if (!inherits(x, "shiny.tag")) return(FALSE)
  class <- tagGetAttribute(x, "class") %||% ""
  "tab-pane" %in% strsplit(class, "\\s+")[[1]]
}

tabPanelBody_ <- function(value, ..., icon = NULL) {
  if (
    !is.character(value) ||
    length(value) != 1 ||
    any(is.na(value)) ||
    nchar(value) == 0
  ) {
    stop("`value` must be a single, non-empty string value")
  }
  tabPanel_(title = NULL, ..., value = value, icon = icon)
}

tabsetPanel_ <- function(...,
                        id = NULL,
                        selected = NULL,
                        type = c("tabs", "pills", "hidden"),
                        header = NULL,
                        footer = NULL) {

  if (!is.null(id))
    selected <- shiny::restoreInput(id = id, default = selected)

  type <- match.arg(type)
  tabset <- buildTabset(..., ulClass = paste0("nav nav-", type), id = id, selected = selected)

  tags$div(
    class = "tabbable",
    !!!dropNulls(list(
      tabset$navList,
      header,
      tabset$content,
      footer
    ))
  )
}

navlistPanel_ <- function(...,
                         id = NULL,
                         selected = NULL,
                         header = NULL,
                         footer = NULL,
                         well = TRUE,
                         fluid = TRUE,
                         widths = c(4, 8)) {

  if (!is.null(id))
    selected <- shiny::restoreInput(id = id, default = selected)

  tabset <- buildTabset(
    ..., ulClass = "nav nav-pills nav-stacked",
    textFilter = function(text) tags$li(class = "navbar-brand", text),
    id = id, selected = selected
  )

  row <- if (fluid) shiny::fluidRow else shiny::fixedRow

  row(
    shiny::column(widths[[1]], class = if (well) "well", tabset$navList),
    shiny::column(
      widths[[2]],
      !!!dropNulls(list(header, tabset$content, footer))
    )
  )
}


# Helpers to build tabsetPanels (& Co.) and their elements
markTabAsSelected <- function(x) {
  attr(x, "selected") <- TRUE
  x
}

isTabSelected <- function(x) {
  isTRUE(attr(x, "selected", exact = TRUE))
}

containsSelectedTab <- function(tabs) {
  any(vapply(tabs, isTabSelected, logical(1)))
}

findAndMarkSelectedTab <- function(tabs, selected, foundSelected) {
  tabs <- lapply(tabs, function(x) {
    if (foundSelected || is.character(x)) {
      # Strings are not selectable items

    } else if (isNavbarMenu(x)) {
      # Recur for navbarMenus
      res <- findAndMarkSelectedTab(x$tabs, selected, foundSelected)
      x$tabs <- res$tabs
      foundSelected <<- res$foundSelected

    } else if (isTabPanel(x)) {
      # Base case: regular tab item. If the `selected` argument is
      # provided, check for a match in the existing tabs; else,
      # mark first available item as selected
      if (is.null(selected)) {
        foundSelected <<- TRUE
        x <- markTabAsSelected(x)
      } else {
        tabValue <- x$attribs$`data-value` %||% x$attribs$title
        if (identical(selected, tabValue)) {
          foundSelected <<- TRUE
          x <- markTabAsSelected(x)
        }
      }
    }
    return(x)
  })
  return(list(tabs = tabs, foundSelected = foundSelected))
}

prepTabIcon <- function(x = NULL) {
  if (is.null(x)) return(NULL)
  if (!inherits(x, "shiny.tag")) {
    stop(
      "`icon` must be a `shiny.tag` object. ",
      "Try passing `icon()` (or `tags$i()`) to the `icon` parameter.",
      call. = FALSE
    )
  }

  is_fa <- grepl("fa-", tagGetAttribute(x, "class") %||% "", fixed = TRUE)
  if (!is_fa) {
    return(x)
  }

  # for font-awesome we specify fixed-width
  tagAppendAttributes(x, class = "fa-fw")
}

# Text filter for navbarMenu's (plain text) separators
navbarMenuTextFilter <- function(text) {
  if (grepl("^\\-+$", text)) tags$li(class = "divider")
  else tags$li(class = "dropdown-header", text)
}

# This function is called internally by navbarPage, tabsetPanel
# and navlistPanel
buildTabset <- function(..., ulClass, textFilter = NULL, id = NULL,
                        selected = NULL, foundSelected = FALSE) {

  tabs <- dropNulls(list2(...))
  res <- findAndMarkSelectedTab(tabs, selected, foundSelected)
  tabs <- res$tabs
  foundSelected <- res$foundSelected

  # add input class if we have an id
  if (!is.null(id)) ulClass <- paste(ulClass, "shiny-tab-input")

  if (anyNamed(tabs)) {
    nms <- names(tabs)
    nms <- nms[nzchar(nms)]
    stop("Tabs should all be unnamed arguments, but some are named: ",
         paste(nms, collapse = ", "))
  }

  tabsetId <- p_randomInt(1000, 10000)
  tabs <- lapply(seq_len(length(tabs)), buildTabItem,
                 tabsetId = tabsetId, foundSelected = foundSelected,
                 tabs = tabs, textFilter = textFilter)

  tabNavList <- tags$ul(class = ulClass, id = id,
                        `data-tabsetid` = tabsetId, !!!lapply(tabs, "[[", "liTag"))

  tabContent <- tags$div(class = "tab-content",
                         `data-tabsetid` = tabsetId, !!!lapply(tabs, "[[", "divTag"))

  list(navList = tabNavList, content = tabContent)
}

# Builds tabPanel/navbarMenu items (this function used to be
# declared inside the buildTabset() function and it's been
# refactored for clarity and reusability). Called internally
# by buildTabset.
buildTabItem <- function(index, tabsetId, foundSelected, tabs = NULL,
                         divTag = NULL, textFilter = NULL) {

  divTag <- divTag %||% tabs[[index]]

  # Handles navlistPanel() headers and dropdown dividers
  if (is.character(divTag) && !is.null(textFilter)) {
    return(list(liTag = textFilter(divTag), divTag = NULL))
  }

  if (isNavbarMenu(divTag)) {
    # tabPanelMenu item: build the child tabset
    ulClass <- "dropdown-menu"
    if (identical(divTag$align, "right")) {
      ulClass <- paste(ulClass, "dropdown-menu-right dropdown-menu-end")
    }
    tabset <- buildTabset(
      !!!divTag$tabs, ulClass = ulClass,
      textFilter = navbarMenuTextFilter,
      foundSelected = foundSelected
    )
    return(buildDropdown(divTag, tabset))
  }

  if (isTabPanel(divTag)) {
    return(buildNavItem(divTag, tabsetId, index))
  }

  if (is_nav_item(divTag) || is_nav_spacer(divTag)) {
    return(
      list(liTag = divTag, divTag = NULL)
    )
  }

  # The behavior is undefined at this point, so construct a condition message
  msg <- paste0(
    "Navigation containers expect a collection of `bslib::nav()`/`shiny::tabPanel()`s ",
    "and/or `bslib::nav_menu()`/`shiny::navbarMenu()`s. ",
    "Consider using `header` or `footer` if you wish to place content above ",
    "(or below) every panel's contents."
  )

  # Luckily this case has never worked, so it's safe to throw here
  # https://github.com/rstudio/shiny/issues/3313
  if (!inherits(divTag, "shiny.tag"))  {
    stop(msg, call. = FALSE)
  }

  # Unfortunately, this 'off-label' use case creates an 'empty' nav and includes
  # the divTag content on every tab. There shouldn't be any reason to be relying on
  # this behavior since we now have pre/post arguments, so throw a warning, but still
  # support the use case since we don't make breaking changes
  warning(msg, call. = FALSE)

  return(buildNavItem(divTag, tabsetId, index))
}

buildNavItem <- function(divTag, tabsetId, index) {
  id <- paste("tab", tabsetId, index, sep = "-")
  # Get title attribute directory (not via tagGetAttribute()) so that contents
  # don't get passed to as.character().
  # https://github.com/rstudio/shiny/issues/3352
  title <- divTag$attribs[["title"]]
  value <- divTag$attribs[["data-value"]]
  active <- isTabSelected(divTag)
  divTag <- tagAppendAttributes(divTag, class = if (active) "active")
  divTag$attribs$id <- id
  divTag$attribs$title <- NULL
  list(
    divTag = divTag,
    liTag = tagAddRenderHook(
      liTag(id, title, value, attr(divTag, "_shiny_icon")),
      function(x) {
        if (isTRUE(getCurrentThemeVersion() >= 4)) {
          tagQuery(x)$
            addClass("nav-item")$
            find("a")$
            addClass(c("nav-link", if (active) "active"))$
            allTags()
        } else {
          tagAppendAttributes(x, class = if (active) "active")
        }
      }
    )
  )
}

liTag <- function(id, title, value, icon) {
  tags$li(
    tags$a(
      href = paste0("#", id),
      `data-toggle` = "tab",
      # data-bs-* is for BS5+
      `data-bs-toggle` = "tab",
      `data-value` = value,
      icon, title
    )
  )
}

buildDropdown <- function(divTag, tabset) {

  navList <- tagAddRenderHook(
    tabset$navList,
    function(x) {
      if (isTRUE(getCurrentThemeVersion() >= 4)) {
        tagQuery(x)$
          find(".nav-item")$
          removeClass("nav-item")$
          find(".nav-link")$
          removeClass("nav-link")$
          addClass("dropdown-item")$
          allTags()
      } else {
        x
      }
    }
  )

  active <- containsSelectedTab(divTag$tabs)

  dropdown <- tags$li(
    class = "dropdown",
    tags$a(
      href = "#",
      class = "dropdown-toggle",
      `data-toggle` = "dropdown",
      # data-bs-* is for BS5+
      `data-bs-toggle` = "dropdown",
      `data-value` = divTag$menuName,
      divTag$icon,
      divTag$title,
      tags$b(class = "caret")
    ),
    navList,
    .renderHook = function(x) {
      if (isTRUE(getCurrentThemeVersion() >= 4)) {
        tagQuery(x)$
          addClass("nav-item")$
          find(".dropdown-toggle")$
          addClass("nav-link")$
          addClass(if (active) "active")$
          allTags()
      } else {
        tagAppendAttributes(x, class = if (active) "active")
      }
    }
  )

  list(
    divTag = tabset$content$children,
    liTag = dropdown
  )
}

#' Add a popover to a UI element
#'
#' Display additional information when clicking on a UI element (typically a
#' button).
#'
#' @param trigger The UI element to serve as the popover trigger (typically a
#'   [shiny::actionButton()] or similar). If `trigger` renders as multiple HTML
#'   elements (e.g., it's a `tagList()`), the last HTML element is used for the
#'   trigger. If the `trigger` should contain all of those elements, wrap the
#'   object in a [div()] or [span()].
#' @param ... UI elements for the popover's body. Character strings are
#'   [automatically escaped][htmlEscape()] unless marked as [HTML()].
#' @param title A title (header) for the popover.
#' @param id A character string. Required to re-actively respond to the
#'   visibility of the popover (via the `input[[id]]` value) and/or update the
#'   visibility/contents of the popover.
#' @param placement The placement of the popover relative to its trigger.
#' @param options A list of additional
#'   [options](https://getbootstrap.com/docs/5.3/components/popovers/#options).
#'
#' @section Closing popovers:
#'
#'   In addition to clicking the `close_button`, popovers can be closed by
#'   pressing the Esc/Space key when the popover (and/or its trigger) is
#'   focused.
#'
#' @section Theming/Styling:
#'
#'   Like other bslib components, popovers can be themed by supplying [relevant
#'   theming
#'   variables](https://rstudio.github.io/bslib/articles/bs5-variables/index.html#popover-bg)
#'   to [bs_theme()], which effects styling of every popover on the page. To
#'   style a _specific_ popover differently from other popovers, utilize the
#'   `customClass` option:
#'
#'   ```
#'   popover(
#'     "Trigger", "Popover message",
#'     options = list(customClass = "my-pop")
#'   )
#'   ```
#'
#'   And then add relevant rules to [bs_theme()] via [bs_add_rules()]:
#'
#'   ```
#'   bs_theme() |> bs_add_rules(".my-pop { max-width: none; }")
#'   ```
#'
#' @describeIn popover Add a popover to a UI element
#' @references <https://getbootstrap.com/docs/5.3/components/popovers/>
#' @export
#' @seealso [tooltip()]
#' @examplesIf interactive()
#'
#' popover(
#'   shiny::actionButton("btn", "A button"),
#'   "Popover body content...",
#'   title = "Popover title"
#' )
#'
#' library(shiny)
#'
#' ui <- page_fixed(
#'   card(class = "mt-5",
#'     card_header(
#'       popover(
#'         uiOutput("card_title", inline = TRUE),
#'         title = "Provide a new title",
#'         textInput("card_title", NULL, "An editable title")
#'       )
#'     ),
#'     "The card body..."
#'   )
#' )
#'
#' server <- function(input, output) {
#'   output$card_title <- renderUI({
#'     list(input$card_title, bsicons::bs_icon("pencil-square"))
#'   })
#' }
#'
#' shinyApp(ui, server)
popover <- function(
  trigger,
  ...,
  title = NULL,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
) {

  args <- separate_arguments(...)
  children <- args$children
  attribs <- args$attribs

  if (length(children) == 0) {
    abort("At least one value must be provided to `...`.")
  }

  bad_opts <- intersect(c("content", "title", "placement"), names(options))
  if (length(bad_opts) > 0) {
    rlang::abort(
      sprintf("The `%s` option cannot be specified directly.", bad_opts[1])
    )
  }

  res <- web_component(
    "bslib-popover",
    id = id,
    placement = rlang::arg_match(placement),
    bsOptions = to_json(options),
    !!!attribs,
    # Use <template> as a way to protect these children from potentially being
    # pulled outside this element (the browser's parser does this to, for
    # example, block elements inside a <p> tag)
    tags$template(
      div(!!!children, style = "display:contents;"),
      div(title, style = "display:contents;")
    ),
    trigger
  )

  res <- tag_require(res, version = 5, caller = "popover()")
  as_fragment(res)
}

#' @describeIn popover Programmatically show/hide a popover.
#'
#' @param show Whether to show (`TRUE`) or hide (`FALSE`) the popover. The
#'   default (`NULL`) will show if currently hidden and hide if currently shown.
#'   Note that a popover will not be shown if the trigger is not visible (e.g.,
#'   it's hidden behind a tab).
#' @param session A Shiny session object (the default should almost always be
#'   used).
#'
#' @export
toggle_popover <- function(id, show = NULL, session = get_current_session()) {
  show <- normalize_show_value(show)

  msg <- list(method = "toggle", value = show)
  force(id)
  callback <- function() {
    session$sendInputMessage(id, msg)
  }
  session$onFlush(callback, once = TRUE)
}

#' @describeIn popover Update the contents of a popover.
#' @export
update_popover <- function(id, ..., title = NULL, session = get_current_session()) {

  body <- tagList(...)

  msg <- dropNulls(list(
    method = "update",
    content = if (length(body) > 0) processDeps(body, session),
    header = if (length(title) > 0) processDeps(title, session)
  ))

  force(id)
  callback <- function() {
    session$sendInputMessage(id, msg)
  }
  session$onFlush(callback, once = TRUE)
}

#' Helpers for importing web fonts
#'
#' `font_google()`, `font_link()`, and `font_face()` are all re-exported from
#' the \pkg{sass} package (see [sass::font_face()] for details). For a quick
#' example of how to use these functions with [bslib::bs_theme()], see the
#' examples section below.
#'
#' @importFrom sass font_face
#' @export font_face
#' @name font_face
#' @examples
#'
#' # If you have an internet connection, running the following code
#' # will download, cache, and import the relevant Google Font files
#' # for local use
#' theme <- bs_theme(
#'   base_font = font_google("Fira Sans"),
#'   code_font = font_google("Fira Code"),
#'   heading_font = font_google("Fredoka One")
#' )
#' if (interactive()) {
#'   bs_theme_preview(theme)
#' }
#'
#' # Three different yet equivalent ways of importing a remotely-hosted Google Font
#' a <- font_google("Crimson Pro", wght = "200..900", local = FALSE)
#' b <- font_link(
#'   "Crimson Pro",
#'   href = "https://fonts.googleapis.com/css2?family=Crimson+Pro:wght@200..900"
#' )
#' url <- "https://fonts.gstatic.com/s/crimsonpro/v13/q5uDsoa5M_tv7IihmnkabARboYF6CsKj.woff2"
#' c <- font_face(
#'   family = "Crimson Pro",
#'   style = "normal",
#'   weight = "200 900",
#'   src = paste0("url(", url, ") format('woff2')")
#' )
#' theme <- bs_theme(base_font = c)
#' if (interactive()) {
#'   bs_theme_preview(theme)
#' }
NULL

#' @importFrom sass font_link
#' @export font_link
#' @name font_link
#' @rdname font_face
NULL

#' @importFrom sass font_google
#' @export font_google
#' @name font_google
#' @rdname font_face
NULL

#' @importFrom sass font_collection
#' @export font_collection
#' @name font_collection
#' @rdname font_face
NULL

#' Column-first uniform grid layouts
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Wraps a 1d sequence of UI elements into a 2d grid. The number of columns (and
#' rows) in the grid dependent on the column `width` as well as the size of the
#' display. For more explanation and illustrative examples, see the _References_
#' section below.
#'
#' @param ... Unnamed arguments should be UI elements (e.g., [card()]). Named
#'   arguments become attributes on the containing [htmltools::tag] element.
#' @param width The desired width of each card, which can be any of the
#'  following:
#'   * A (unit-less) number between 0 and 1.
#'     * This should be specified as `1/num`, where `num` represents the number
#'       of desired columns.
#'   * A [CSS length unit][htmltools::validateCssUnit()]
#'     * Either the minimum (when `fixed_width=FALSE`) or fixed width
#'       (`fixed_width=TRUE`).
#'   * `NULL`
#'     * Allows power users to set the `grid-template-columns` CSS property
#'       manually, either via a `style` attribute or a CSS stylesheet.
#' @param fixed_width When `width` is greater than 1 or is a CSS length unit,
#'   e.g. `"200px"`, `fixed_width` indicates whether that `width` value
#'   represents the absolute size of each column (`fixed_width=TRUE`) or the
#'   minimum size of a column (`fixed_width=FALSE`). When `fixed_width=FALSE`,
#'   new columns are added to a row when `width` space is available and columns
#'   will never exceed the container or viewport size. When `fixed_width=TRUE`,
#'   all columns will be exactly `width` wide, which may result in columns
#'   overflowing the parent container.
#' @param heights_equal If `"all"` (the default), every card in every row of the
#'   grid will have the same height. If `"row"`, then every card in _each_ row
#'   of the grid will have the same height, but heights may vary between rows.
#' @param fill Whether or not to allow the layout to grow/shrink to fit a
#'   fillable container with an opinionated height (e.g., `page_fillable()`).
#' @param fillable Whether or not each element is wrapped in a fillable container.
#' @param height_mobile Any valid CSS unit to use for the height when on mobile
#'   devices (or narrow windows).
#' @inheritParams card
#' @inheritParams card_body
#'
#' @export
#' @family Column layouts
#'
#' @references The bslib website features `layout_column_wrap()` in two places:
#'   * [Column-based layouts](https://rstudio.github.io/bslib/articles/column-layout/index.html)
#'   * [Cards: Multiple cards](https://rstudio.github.io/bslib/articles/cards/index.html#multiple-cards)
#'
#' @examplesIf rlang::is_interactive()
#' x <- card("A simple card")
#'
#' # Always has 2 columns (on non-mobile)
#' layout_column_wrap(width = 1/2, x, x, x)
#'
#' # Automatically lays out three cards into columns
#' # such that each column is at least 200px wide:
#' layout_column_wrap(x, x, x)
#'
#' # To use larger column widths by default, set `width`.
#' # This example has 3 columns when the screen is at least 900px wide:
#' layout_column_wrap(width = "300px", x, x, x)
#'
#' @export
layout_column_wrap <- function(
  ...,
  width = "200px",
  fixed_width = FALSE,
  heights_equal = c("all", "row"),
  fill = TRUE,
  fillable = TRUE,
  height = NULL,
  height_mobile = NULL,
  gap = NULL,
  class = NULL
) {

  heights_equal <- match.arg(heights_equal)

  args <- separate_arguments(...)
  attribs <- args$attribs
  children <- args$children

  if (missing(width)) {
    first_is_width <-
      is.null(children[[1]]) ||
      is_probably_a_css_unit(children[[1]])

    if (first_is_width) {
      # Assume an unnamed first argument that matches our expectations for
      # `width` is actually the width argument, with a warning
      lifecycle::deprecate_warn(
        "0.6.0",
        "layout_column_wrap(width = 'must be named')"
      )
      width <- children[[1]]
      children <- children[-1]
    }
  }

  if (length(width) > 1) {
    stop("`width` of length greater than 1 is not currently supported.")
  }

  colspec <- if (!is.null(width)) {
    if (width > 0 && width <= 1) {
      num_cols <- 1 / width
      if (num_cols != as.integer(num_cols)) {
        stop("Could not interpret width argument; see ?layout_column_wrap")
      }
      sprintf("repeat(%s, minmax(0, 1fr))", num_cols)
    } else {
      if (fixed_width) {
        paste0("repeat(auto-fit, ", validateCssUnit(width), ")")
      } else {
        sprintf(
          "repeat(auto-fit, minmax(min(%s, 100%%), 1fr))",
          validateCssUnit(width)
        )
      }
    }
  }

  # If relevant (i.e., fillable=TRUE), wrap each child in a fillable context
  # (so fill/flex items can fill the available area)
  children <- lapply(children, grid_item_container, fillable = fillable)

  tag <- div(
    class = "bslib-grid bslib-mb-spacing",
    style = css(
      grid_template_columns = colspec,
      grid_auto_rows = if (heights_equal == "all") "1fr",
      # Always provide the `height:auto` default so that the CSS variable
      # doesn't get inherited in a scenario like layout_column_wrap(height=200, ..., layout_column_wrap(...))
      "--bslib-grid-height" = validateCssUnit(height %||% "auto"),
      "--bslib-grid-height-mobile" = validateCssUnit(height_mobile %||% "auto"),
      gap = validateCssUnit(gap)
    ),
    !!!attribs,
    children,
    component_dependencies()
  )

  tag <- bindFillRole(tag, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "layout_column_wrap()")
  )
}

is_probably_a_css_unit <- function(x) {
  if (length(x) != 1) return(FALSE)
  if (is.numeric(x)) return(TRUE)
  if (!is.character(x)) return(FALSE)
  tryCatch(
    { validateCssUnit(x); TRUE },
    error = function(e) FALSE
  )
}

#' Responsive 12-column grid layouts
#'
#' Create responsive, column-based grid layouts, based on a 12-column grid.
#'
#' @inheritParams layout_column_wrap
#' @param col_widths One of the following:
#'   * `NA` (the default): Automatically determines a sensible number of columns
#'     based on the number of children.
#'   * A numeric vector of integers between 1 and 12, where each element
#'     represents the number of columns for the relevant UI element. Elements
#'     that happen to go beyond 12 columns wrap onto the next row. For example,
#'     `col_widths = c(4, 8, 12)` allocates 4 columns to the first element, 8
#'     columns to the second element, and 12 columns to the third element (which
#'     wraps to the next row). Negative values are also allowed, and are treated
#'     as empty columns. For example, `col_widths = c(-2, 8, -2)` would allocate
#'     8 columns to an element (with 2 empty columns on either side).
#'   * A [breakpoints()] object, where each breakpoint may be either of the
#'     above.
#' @param row_heights One of the following:
#'   * A numeric vector, where each value represents the
#'     [fractional unit](https://css-tricks.com/introduction-fr-css-unit/)
#'     (`fr`) height of the relevant row. If there are more rows than values
#'     provided, the pattern will repeat. For example, `row_heights = c(1, 2)`
#'     allows even rows to take up twice as much space as odd rows.
#'   * A list of numeric and [CSS length units][htmltools::validateCssUnit()],
#'     where each value represents the height of the relevant row. If more rows
#'     are needed than values provided, the pattern will repeat. For example,
#'     `row_heights = list("auto", 1)` allows the height of odd rows to be
#'     driven my it's contents and even rows to be
#'     [`1fr`](https://css-tricks.com/introduction-fr-css-unit/).
#'   * A character vector/string of [CSS length units][htmltools::validateCssUnit()].
#'     In this case, the value is supplied directly to `grid-auto-rows`.
#'   * A [breakpoints()] object, where each breakpoint may be either of the above.
#'
#' @export
#' @family Column layouts
#'
#' @references [Column-based layouts](https://rstudio.github.io/bslib/articles/column-layout/index.html)
#'   on the bslib website.
#'
#' @seealso [breakpoints()] for more information on specifying column widths at
#'   responsive breakpoints.
#'
#' @examplesIf rlang::is_interactive()
#' x <- card("A simple card")
#'
#' page_fillable(
#'   layout_columns(x, x, x, x)
#' )
#'
#' page_fillable(
#'   layout_columns(
#'     col_widths = c(6, 6, 12),
#'     x, x, x
#'   )
#' )
#'
#' page_fillable(
#'   layout_columns(
#'     col_widths = c(6, 6, -2, 8),
#'     row_heights = c(1, 3),
#'     x, x, x
#'   )
#' )
#'
#' page_fillable(
#'   fillable_mobile = TRUE,
#'   layout_columns(
#'     col_widths = breakpoints(
#'       sm = c(12, 12, 12),
#'       md = c(6, 6, 12),
#'       lg = c(4, 4, 4)
#'     ),
#'     x, x, x
#'   )
#' )
#'
layout_columns <- function(
  ...,
  col_widths = NA,
  row_heights = NULL,
  fill = TRUE,
  fillable = TRUE,
  gap = NULL,
  class = NULL,
  height = NULL
) {
  args <- separate_arguments(...)
  attribs <- args$attribs
  children <- args$children
  n_kids <- length(children)

  # Resolve missing value(s) for col_widths, etc.
  spec <- impute_col_spec(col_widths, n_kids)

  # Translate col_widths into Bootstrap's .g-col-* classes
  width_classes <- col_width_grid_classes(
    spec$col_widths, n_kids, spec$n_cols
  )

  # Add a class to each item that helps provide "fallback" rules
  width_classes <- paste0(width_classes, " bslib-grid-item")

  # Wrap each child in a container (so fill/flex items can fill the available area)
  children <- Map(
    f = grid_item_container, children,
    class = width_classes,
    fillable = fillable
  )

  tag <- div(
    class = "bslib-grid grid bslib-mb-spacing",
    style = css(
      height = validateCssUnit(height),
      gap = validateCssUnit(gap),
      "--bs-columns" = spec$n_cols
    ),
    !!!row_heights_css_vars(row_heights),
    !!!attribs,
    !!!children,
    component_dependencies()
  )

  tag <- bindFillRole(tag, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "layout_columns()")
  )
}



impute_col_spec <- function(x, n_kids) {
  if (isTRUE(is.na(x))) {
    x <- breakpoints(sm = NA, lg = NA)
  }

  if (!is_breakpoints(x)) {
    x <- breakpoints(md = x)
  }

  # if smallest defined breakpoint is large, fill in gap with 'md'
  if (names(x)[1] %in% c("lg", "xl", "xxl")) {
    x <- breakpoints(md = x[[1]], !!!x)
  }

  has_auto_spec <- vapply(x, function(y) isTRUE(is.na(y)), logical(1))

  if (!any(has_auto_spec)) {
    return(list(n_cols = 12, col_widths = x))
  }

  n_cols <- if (n_kids > 7) 12 else if (n_kids > 3) n_kids * 2 else n_kids

  for (break_name in names(x)[has_auto_spec]) {
    prefer_wider <- break_name %in% c("sm", "md")
    x[[break_name]] <- col_width_best_fit(n_kids, prefer_wider)
  }

  list(n_cols = n_cols, col_widths = x)
}



row_heights_css_vars <- function(x) {
  if (is.null(x)) return(list())

  if (!is_breakpoints(x)) {
    x <- breakpoints(sm = x)
  }

  # creates classes that pair with CSS variables when set
  classes <- paste0("bslib-grid--row-heights--", names(x))

  css_vars <- setNames(x, paste0("--", classes))

  # mobile row height is derived from xs or defaults to auto in the CSS,
  # so we don't need the class to activate it
  classes <- setdiff(classes, "bslib-grid--row-heights--xs")

  # Treat numeric values as fractional units
  css_vars <- rapply(
    css_vars, how = "replace",
    function(x) {
      if (is.numeric(x)) paste0(x, "fr") else x
    }
  )

  list(
    style = css(!!!css_vars),
    class = classes
  )
}



col_width_grid_classes <- function(breaks, n_kids, n_cols = 12) {
  if (!is_breakpoints(breaks)) {
    abort("`breaks` must be a `breakpoints()` object")
  }

  classes <- vector("list", n_kids)
  add_class <- function(idx, new) {
    classes[[idx]] <<- c(classes[[idx]], new)
  }

  breaks <- as_column_breakpoints(breaks)

  for (break_name in names(breaks)) {
    bk <- breaks[[break_name]]

    if (length(bk$width) > n_kids) {
      msg <- sprintf(
        "Truncating number of widths at '%s' breakpoint to match number of elements.",
        break_name
      )
      rlang::warn(c(
        msg,
        "*" = paste("widths:", length(bk$width)),
        "*" = paste("elements:", n_kids)
      ))
    }

    widths <- rep_len(bk$width, n_kids)
    before <- rep_len(bk$before, n_kids)
    after <- rep_len(bk$after, n_kids)

    # This next section implements a content-layout algorithm, motivated by
    # supporting empty columns. In particular, we need to know two things about
    # the content item:
    #
    # 1. How wide is the content item?
    # 2. What is its starting column position?
    #
    # The following example illustrates a few layout cases (. = empty column):
    # > breakpoint_columns(md = c(-1, 4, 5, -4, 3, 9, -3, 2))
    #
    # | . | 4 | 4 | 4 | 4 | 5 | 5 | 5 | 5 | 5 | . |
    # | . | . | . | 3 | 3 | 3 |   |   |   |   |   |
    # | 9 | 9 | 9 | 9 | 9 | 9 | 9 | 9 | . | . | . |
    # | 2 | 2 | . | 4 | 4 | 4 | 4 |   |   |   |   | ...
    #
    # Because we recycle column widths to match the number of kids, we can't
    # guarantee that the pattern repeats by row. To quickly summarize:
    #
    # * Each content item has a width (`widths[idx]`) and empty space before
    #   the item: `before[idx]` + `after[idx - 1]` (the space before this item
    #   plus the space _after_ the previous item).
    # * We maintain a cursor that knows the 0-indexed column position. At each
    #   step we:
    #   * Move the cursor forward by the empty space before the item
    #   * Decide if we require a starting class (`g-start-{break}-{cursor + 1}`)
    #   * Add starting class and content width class (`g-start-{break}-{width}`)
    #     for the item
    #   * Move the cursor forward by the width of the item.
    #
    # We take into account a few edge cases:
    #
    # * We *don't need* a starting class if the item would naturally reflow to the
    #   next row.
    # * We *do need* a starting class if the item would fit into the empty space
    #   of the current row, but there isn't enough room for the item _after_
    #   accounting for blocked empty space.
    # * If adding empty space causes a new row, but adding the content item
    #   would cause _another row break_, we skip the empty row.

    cursor <- 0L
    update_cursor <- function(incr, is_empty = FALSE) {
      cursor <<- abs(cursor)
      new <- cursor + incr
      if (new == n_cols) {
        # we reached the final column, allow for a natural break
        new <- 0L
      }
      if (new > n_cols) {
        # this row is full, empty columns can break (with <0 cursor to signal)
        # and content columns fit on the next row
        new <- if (is_empty) -1 * new %% n_cols else incr
      }
      # message("cursor: ", cursor, " -> ", new, " (+", incr, if (is_empty) " empty", ")")
      cursor <<- new
    }

    add_start_class <- FALSE
    for (idx in seq_len(n_kids)) {
      move_ahead <- before[idx] + if (idx > 1) after[idx - 1] else 0L
      this_width <- min(widths[idx], n_cols)

      # when we move ahead, we need a start class unless the current item
      # wouldn't fit on the row anyway (ignoring empty cols)
      row_remaining <- n_cols - cursor

      if (move_ahead > 0) {
        update_cursor(move_ahead, is_empty = TRUE)
        if (cursor < 0) {
          cursor <- abs(cursor)
          # adding empty cols caused a row wrap, so we need a start class if
          # 1. we're not at the beginning of the row
          # 2. But: if the current item is wider than the remaining space after
          #    accounting for empty columns, reset cursor to start of the row
          #    rather than causing an empty row.
          if (widths[idx] > (n_cols - cursor)) {
            cursor <- 0L
          }
          row_remaining <- 0L
        }
        add_start_class <- row_remaining >= widths[idx] || cursor > 0
      }

      if (add_start_class) {
        add_class(idx, sprintf("g-start-%s-%s", break_name, cursor + 1L))
        add_start_class <- FALSE
      }

      add_class(idx, sprintf("g-col-%s-%s", break_name, this_width))
      update_cursor(this_width, is_empty = FALSE)
    }
  }

  vapply(classes, paste, character(1), collapse = " ")
}


col_width_best_fit <- function(kids, prefer_wider = FALSE) {
  if (kids < 4) return(1)
  if (kids <= 7) {
    # sizes 4-7 are special cased to use (2 * kids) columns
    return(if (prefer_wider) kids else 2)
  }

  fctrs <- c(if (!prefer_wider) 2, 3, 4, if (prefer_wider) 6)

  col_units <- kids * fctrs
  rows <- ceiling(col_units / 12)
  total_units <- rows * 12
  empty_units <- total_units - col_units

  if (prefer_wider) {
    fctrs <- rev(fctrs)
    empty_units <- rev(empty_units)
  }

  fctrs[which.min(empty_units)]
}



grid_item_container <- function(el, ..., fillable = TRUE) {
  div(
    ...,
    class = "bslib-gap-spacing",
    if (fillable) as_fillable_container(),
    el
  )
}

test_that("grid_item_container()", {
  expect_snapshot(
    grid_item_container(
      div(class = "layout-column-child-element"),
      class = "g-col-md-6 g-col-lg-4",
      fillable = TRUE
    )
  )

  expect_snapshot(
    grid_item_container(
      div(class = "layout-column-child-element"),
      class = "g-col-md-6 g-col-lg-4",
      fillable = FALSE
    )
  )
})

test_that("grid-breakpoints are known", {
  brks <- bs_get_variables(
    bs_add_variables(
      bs_theme(),
      "grid-brk-names" = "map-keys($grid-breakpoints)",
      .where = "declarations"
    ),
    "grid-brk-names"
  )

  # In bslib::breakpoints() we flag these as "known" breakpoints...
  # which mostly means we order them, but also, in the NA case,
  # we use lg/xl/xxl to fill in md
  expect_equal(
    strsplit(brks, ", ")[[1]],
    c("xs", "sm", "md", "lg", "xl", "xxl")
  )
})

test_that("breakpoints() re-orders well-known breaks and test print method", {
  bp <- breakpoints(
    xl = 5,
    sm = 2,
    xs = 1,
    giant = 7,
    lg = 4,
    huge = 8,
    md = 3,
    xxl = 6
  )

  expect_snapshot(bp)
  expect_equal(unclass(unname(unlist(bp))), 1:8)
})

test_that("breakpoints() has correct classes and structure", {
  bp <- breakpoints(
    lg = c(-2, 2, -1, -2, 3, -2),
    md = c(-1, 2, -2, 3),
    sm = c(1, -1, 1),
    xl = c(1, 2, 3, NA),
    huge = list(1, 2, 3, "auto")
  )

  expect_true(is_breakpoints(bp))

  # names are ordered, custom breakpoint names are last
  expect_named(bp, c("sm", "md", "lg", "xl", "huge"))

  expect_snapshot(bp)
})

test_that("breakpoints() with negative widths to indicate empty columns", {
  bp <- breakpoints(
    sm = c(1, -1, 1),
    md = c(-1, 2, -2, 3),
    lg = c(-2, 2, -1, -2, 3, -2),
    xl = c(1, 2, 3)
  )

  bp <- as_column_breakpoints(bp)

  expect_equal(bp$sm$width, c(1, 1), ignore_attr = TRUE)
  expect_equal(bp$md$width, c(2, 3), ignore_attr = TRUE)
  expect_equal(bp$lg$width, c(2, 3), ignore_attr = TRUE)
  expect_equal(bp$xl$width, c(1, 2, 3), ignore_attr = TRUE)

  expect_equal(bp$sm[["before"]], c(0, 1))
  expect_equal(bp$md[["before"]], c(1, 2))
  expect_equal(bp$lg[["before"]], c(2, 3))
  expect_equal(bp$xl[["before"]], c(0, 0, 0))

  expect_equal(bp$sm[["after"]], c(0, 0))
  expect_equal(bp$md[["after"]], c(0, 0))
  expect_equal(bp$lg[["after"]], c(0, 2))
  expect_equal(bp$xl[["after"]], c(0, 0, 0))
})

test_that("breakpoints() must be named", {
  expect_error(breakpoints(1:3))
})

test_that("breakpoints() can be a single NA", {
  bp <- breakpoints(
    sm = NA,
    md = c(-1, 2, -2, 3),
    lg = c(-2, 2, -1, -2, 3, -2),
    xl = NA
  )

  bp <- as_column_breakpoints(bp)

  expect_equal(bp$sm$width, NA, ignore_attr = TRUE)
  expect_equal(bp$md$width, c(2, 3), ignore_attr = TRUE)
  expect_equal(bp$lg$width, c(2, 3), ignore_attr = TRUE)
  expect_equal(bp$xl$width, NA, ignore_attr = TRUE)

  expect_equal(bp$sm[["before"]], c(0))
  expect_equal(bp$md[["before"]], c(1, 2))
  expect_equal(bp$lg[["before"]], c(2, 3))
  expect_equal(bp$xl[["before"]], c(0))

  expect_equal(bp$sm[["after"]], c(0))
  expect_equal(bp$md[["after"]], c(0, 0))
  expect_equal(bp$lg[["after"]], c(0, 2))
  expect_equal(bp$xl[["after"]], c(0))
})

test_that("breakpoints() throws if NAs are mixed with other column values", {
  expect_snapshot_error(
    layout_columns(col_widths = breakpoints(sm = c(-1, NA, 1)))
  )
})

test_that("bs_css_grid_width_classes() warns when too many column widths", {
  expect_snapshot_warning(
    expect_equal(
      col_width_grid_classes(breakpoints(md = 1:4), 3),
      c("g-col-md-1", "g-col-md-2", "g-col-md-3")
    )
  )
})

test_that("bs_css_grid_width_classes()", {
  expect_equal(
    col_width_grid_classes(breakpoints(md = 6), 3),
    c("g-col-md-6", "g-col-md-6", "g-col-md-6")
  )

  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(4, 8)), 3),
    c("g-col-lg-4", "g-col-lg-8", "g-col-lg-4")
  )

  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(12)), n_kids = 3),
    c("g-col-lg-12", "g-col-lg-12", "g-col-lg-12")
  )

  # accounts for leading and trailing empty columns
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(-1, 4, 6, -1)), n_kids = 3),
    c("g-start-lg-2 g-col-lg-4", "g-col-lg-6", "g-start-lg-2 g-col-lg-4")
  )

  # accounts for trailing empty columns (restarts on first column with class if needed)
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(1, 3, -1, 6, -1)), n_kids = 4),
    c("g-col-lg-1", "g-col-lg-3", "g-start-lg-6 g-col-lg-6", "g-start-lg-1 g-col-lg-1")
  )

  # accounts for trailing empty columns (restarts on first column without class if possible)
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(4, -1, 6, -1)), n_kids = 3),
    c("g-col-lg-4", "g-start-lg-6 g-col-lg-6", "g-col-lg-4")
  )

  # doesn't allow empty columns to cause an empty row due to row break
  # skipping 8 columns would start the second row on column 5, but the next 9
  # column-item would cause a row break, leaving an empty row (which we avoid)
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(8, -8, 9)), n_kids = 4),
    c("g-col-lg-8", "g-col-lg-9", "g-col-lg-8", "g-col-lg-9")
  )

  # Same as above, except that 8 columns *will* fit on next row with an offset
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(8, -8, 8)), n_kids = 4),
    c("g-col-lg-8", "g-start-lg-5 g-col-lg-8", "g-col-lg-8", "g-start-lg-5 g-col-lg-8")
  )

  # recycles the pattern to match number of kids
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(5, -2, 5, 12)), n_kids = 7),
    c(
      "g-col-lg-5",
      "g-start-lg-8 g-col-lg-5",
      "g-col-lg-12",
      # repeats
      "g-col-lg-5",
      "g-start-lg-8 g-col-lg-5",
      "g-col-lg-12",
      # repeats (partially)
      "g-col-lg-5"
    )
  )

  # variant of previous test where first/last columns are empty
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(-1, 5, 5, -1, 12)), n_kids = 7),
    c(
      "g-start-lg-2 g-col-lg-5",
      "g-col-lg-5",
      "g-col-lg-12",
      # repeats
      "g-start-lg-2 g-col-lg-5",
      "g-col-lg-5",
      "g-col-lg-12",
      # repeats (partially)
      "g-start-lg-2 g-col-lg-5"
    )
  )
})

test_that("breakpoint_columns() throws if a column is 0", {
  expect_error(
    col_width_grid_classes(breakpoints(md = c(-1, 0, 1)), 3)
  )
})

test_that("breakpoint_columns() throws if no columns are positive", {
  expect_error(
    col_width_grid_classes(breakpoints(md = -1), 1)
  )
  expect_error(
    col_width_grid_classes(breakpoints(md = c(-1, -1, -1)), 3)
  )
})

test_that("impute_col_spec() auto layout with col_widths = NA", {

  get_col_spec <- function(x, n_kids) {
    spec <- impute_col_spec(x, n_kids)
    list(
      n_cols = spec$n_cols,
      col_widths = as_column_breakpoints(spec$col_widths)
    )
  }

  kids_1 <- get_col_spec(NA, 1)

  # bare NA is the same as NA breakpoints at sm and lg
  expect_equal(
    get_col_spec(breakpoints(sm = NA, lg = NA), 1),
    kids_1
  )

  expect_equal(kids_1$n_cols, 1)
  expect_equal(kids_1$col_widths$sm, list(width = 1, before = 0, after = 0))

  kids_2 <- get_col_spec(NA, 2)
  expect_equal(kids_2$n_cols, 2)
  expect_equal(kids_2$col_widths$sm, list(width = 1, before = 0, after = 0))

  kids_3 <- get_col_spec(NA, 3)
  expect_equal(kids_3$n_cols, 3)
  expect_equal(kids_3$col_widths$sm, list(width = 1, before = 0, after = 0))

  kids_4 <- get_col_spec(NA, 4)
  expect_equal(kids_4$n_cols, 8)
  expect_equal(kids_4$col_widths$sm, list(width = 4, before = 0, after = 0))
  expect_equal(kids_4$col_widths$lg, list(width = 2, before = 0, after = 0))

  kids_5 <- get_col_spec(NA, 5)
  expect_equal(kids_5$n_cols, 5 * 2)
  expect_equal(kids_5$col_widths$sm, list(width = 5, before = 0, after = 0))
  expect_equal(kids_5$col_widths$lg, list(width = 2, before = 0, after = 0))

  kids_6 <- get_col_spec(NA, 6)
  expect_equal(kids_6$n_cols, 6 * 2)
  expect_equal(kids_6$col_widths$sm, list(width = 6, before = 0, after = 0))
  expect_equal(kids_6$col_widths$lg, list(width = 2, before = 0, after = 0))

  kids_7 <- get_col_spec(NA, 7)
  expect_equal(kids_7$n_cols, 7 * 2)
  expect_equal(kids_7$col_widths$sm, list(width = 7, before = 0, after = 0))
  expect_equal(kids_7$col_widths$lg, list(width = 2, before = 0, after = 0))

  kids_8 <- get_col_spec(NA, 8)
  expect_equal(kids_8$n_cols, 12)
  expect_equal(kids_8$col_widths$sm, list(width = 6, before = 0, after = 0))
  expect_equal(kids_8$col_widths$lg, list(width = 3, before = 0, after = 0))
})

test_that("impute_col_spec() missing smaller breakpoints inherit from lg+", {
  expect_equal(
    impute_col_spec(breakpoints(lg = c(4, -4, 4)), 2)$col_widths,
    breakpoints(
      md = c(4, -4, 4),
      lg = c(4, -4, 4)
    )
  )

  expect_equal(
    impute_col_spec(breakpoints(xl = c(4, -4, 4)), 2)$col_widths,
    breakpoints(
      md = c(4, -4, 4),
      xl = c(4, -4, 4)
    )
  )

  expect_equal(
    impute_col_spec(breakpoints(xxl = c(4, -4, 4)), 2)$col_widths,
    breakpoints(
      md = c(4, -4, 4),
      xxl = c(4, -4, 4)
    )
  )
})

test_that("impute_col_spec() base case", {
  expect_equal(
    impute_col_spec(1:3, 6),
    list(n_cols = 12, col_widths = breakpoints(md = 1:3))
  )
})

test_that("row_heights_css_vars() decides fr/px for numeric, passes character", {
  expect_equal(
    row_heights_css_vars(breakpoints(md = c(2, 1)))$style,
    "--bslib-grid--row-heights--md:2fr 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list(2, 1)))$style,
    "--bslib-grid--row-heights--md:2fr 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list("auto", 1)))$style,
    "--bslib-grid--row-heights--md:auto 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list("10px", 1)))$style,
    "--bslib-grid--row-heights--md:10px 1fr;"
  )
})

test_that("grid_item_container()", {
  expect_snapshot(
    grid_item_container(
      div(class = "layout-column-child-element"),
      class = "g-col-md-6 g-col-lg-4",
      fillable = TRUE
    )
  )

  expect_snapshot(
    grid_item_container(
      div(class = "layout-column-child-element"),
      class = "g-col-md-6 g-col-lg-4",
      fillable = FALSE
    )
  )
})

test_that("grid-breakpoints are known", {
  brks <- bs_get_variables(
    bs_add_variables(
      bs_theme(),
      "grid-brk-names" = "map-keys($grid-breakpoints)",
      .where = "declarations"
    ),
    "grid-brk-names"
  )

  # In bslib::breakpoints() we flag these as "known" breakpoints...
  # which mostly means we order them, but also, in the NA case,
  # we use lg/xl/xxl to fill in md
  expect_equal(
    strsplit(brks, ", ")[[1]],
    c("xs", "sm", "md", "lg", "xl", "xxl")
  )
})

test_that("breakpoints() re-orders well-known breaks and test print method", {
  bp <- breakpoints(
    xl = 5,
    sm = 2,
    xs = 1,
    giant = 7,
    lg = 4,
    huge = 8,
    md = 3,
    xxl = 6
  )

  expect_snapshot(bp)
  expect_equal(unclass(unname(unlist(bp))), 1:8)
})

test_that("breakpoints() has correct classes and structure", {
  bp <- breakpoints(
    lg = c(-2, 2, -1, -2, 3, -2),
    md = c(-1, 2, -2, 3),
    sm = c(1, -1, 1),
    xl = c(1, 2, 3, NA),
    huge = list(1, 2, 3, "auto")
  )

  expect_true(is_breakpoints(bp))

  # names are ordered, custom breakpoint names are last
  expect_named(bp, c("sm", "md", "lg", "xl", "huge"))

  expect_snapshot(bp)
})

test_that("breakpoints() with negative widths to indicate empty columns", {
  bp <- breakpoints(
    sm = c(1, -1, 1),
    md = c(-1, 2, -2, 3),
    lg = c(-2, 2, -1, -2, 3, -2),
    xl = c(1, 2, 3)
  )

  bp <- as_column_breakpoints(bp)

  expect_equal(bp$sm$width, c(1, 1), ignore_attr = TRUE)
  expect_equal(bp$md$width, c(2, 3), ignore_attr = TRUE)
  expect_equal(bp$lg$width, c(2, 3), ignore_attr = TRUE)
  expect_equal(bp$xl$width, c(1, 2, 3), ignore_attr = TRUE)

  expect_equal(bp$sm[["before"]], c(0, 1))
  expect_equal(bp$md[["before"]], c(1, 2))
  expect_equal(bp$lg[["before"]], c(2, 3))
  expect_equal(bp$xl[["before"]], c(0, 0, 0))

  expect_equal(bp$sm[["after"]], c(0, 0))
  expect_equal(bp$md[["after"]], c(0, 0))
  expect_equal(bp$lg[["after"]], c(0, 2))
  expect_equal(bp$xl[["after"]], c(0, 0, 0))
})

test_that("breakpoints() must be named", {
  expect_error(breakpoints(1:3))
})

test_that("breakpoints() can be a single NA", {
  bp <- breakpoints(
    sm = NA,
    md = c(-1, 2, -2, 3),
    lg = c(-2, 2, -1, -2, 3, -2),
    xl = NA
  )

  bp <- as_column_breakpoints(bp)

  expect_equal(bp$sm$width, NA, ignore_attr = TRUE)
  expect_equal(bp$md$width, c(2, 3), ignore_attr = TRUE)
  expect_equal(bp$lg$width, c(2, 3), ignore_attr = TRUE)
  expect_equal(bp$xl$width, NA, ignore_attr = TRUE)

  expect_equal(bp$sm[["before"]], c(0))
  expect_equal(bp$md[["before"]], c(1, 2))
  expect_equal(bp$lg[["before"]], c(2, 3))
  expect_equal(bp$xl[["before"]], c(0))

  expect_equal(bp$sm[["after"]], c(0))
  expect_equal(bp$md[["after"]], c(0, 0))
  expect_equal(bp$lg[["after"]], c(0, 2))
  expect_equal(bp$xl[["after"]], c(0))
})

test_that("breakpoints() throws if NAs are mixed with other column values", {
  expect_snapshot_error(
    layout_columns(col_widths = breakpoints(sm = c(-1, NA, 1)))
  )
})

test_that("bs_css_grid_width_classes() warns when too many column widths", {
  expect_snapshot_warning(
    expect_equal(
      col_width_grid_classes(breakpoints(md = 1:4), 3),
      c("g-col-md-1", "g-col-md-2", "g-col-md-3")
    )
  )
})

test_that("bs_css_grid_width_classes()", {
  expect_equal(
    col_width_grid_classes(breakpoints(md = 6), 3),
    c("g-col-md-6", "g-col-md-6", "g-col-md-6")
  )

  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(4, 8)), 3),
    c("g-col-lg-4", "g-col-lg-8", "g-col-lg-4")
  )

  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(12)), n_kids = 3),
    c("g-col-lg-12", "g-col-lg-12", "g-col-lg-12")
  )

  # accounts for leading and trailing empty columns
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(-1, 4, 6, -1)), n_kids = 3),
    c("g-start-lg-2 g-col-lg-4", "g-col-lg-6", "g-start-lg-2 g-col-lg-4")
  )

  # accounts for trailing empty columns (restarts on first column with class if needed)
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(1, 3, -1, 6, -1)), n_kids = 4),
    c("g-col-lg-1", "g-col-lg-3", "g-start-lg-6 g-col-lg-6", "g-start-lg-1 g-col-lg-1")
  )

  # accounts for trailing empty columns (restarts on first column without class if possible)
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(4, -1, 6, -1)), n_kids = 3),
    c("g-col-lg-4", "g-start-lg-6 g-col-lg-6", "g-col-lg-4")
  )

  # doesn't allow empty columns to cause an empty row due to row break
  # skipping 8 columns would start the second row on column 5, but the next 9
  # column-item would cause a row break, leaving an empty row (which we avoid)
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(8, -8, 9)), n_kids = 4),
    c("g-col-lg-8", "g-col-lg-9", "g-col-lg-8", "g-col-lg-9")
  )

  # Same as above, except that 8 columns *will* fit on next row with an offset
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(8, -8, 8)), n_kids = 4),
    c("g-col-lg-8", "g-start-lg-5 g-col-lg-8", "g-col-lg-8", "g-start-lg-5 g-col-lg-8")
  )

  # recycles the pattern to match number of kids
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(5, -2, 5, 12)), n_kids = 7),
    c(
      "g-col-lg-5",
      "g-start-lg-8 g-col-lg-5",
      "g-col-lg-12",
      # repeats
      "g-col-lg-5",
      "g-start-lg-8 g-col-lg-5",
      "g-col-lg-12",
      # repeats (partially)
      "g-col-lg-5"
    )
  )

  # variant of previous test where first/last columns are empty
  expect_equal(
    col_width_grid_classes(breakpoints(lg = c(-1, 5, 5, -1, 12)), n_kids = 7),
    c(
      "g-start-lg-2 g-col-lg-5",
      "g-col-lg-5",
      "g-col-lg-12",
      # repeats
      "g-start-lg-2 g-col-lg-5",
      "g-col-lg-5",
      "g-col-lg-12",
      # repeats (partially)
      "g-start-lg-2 g-col-lg-5"
    )
  )
})

test_that("breakpoint_columns() throws if a column is 0", {
  expect_error(
    col_width_grid_classes(breakpoints(md = c(-1, 0, 1)), 3)
  )
})

test_that("breakpoint_columns() throws if no columns are positive", {
  expect_error(
    col_width_grid_classes(breakpoints(md = -1), 1)
  )
  expect_error(
    col_width_grid_classes(breakpoints(md = c(-1, -1, -1)), 3)
  )
})

test_that("impute_col_spec() auto layout with col_widths = NA", {

  get_col_spec <- function(x, n_kids) {
    spec <- impute_col_spec(x, n_kids)
    list(
      n_cols = spec$n_cols,
      col_widths = as_column_breakpoints(spec$col_widths)
    )
  }

  kids_1 <- get_col_spec(NA, 1)

  # bare NA is the same as NA breakpoints at sm and lg
  expect_equal(
    get_col_spec(breakpoints(sm = NA, lg = NA), 1),
    kids_1
  )

  expect_equal(kids_1$n_cols, 1)
  expect_equal(kids_1$col_widths$sm, list(width = 1, before = 0, after = 0))

  kids_2 <- get_col_spec(NA, 2)
  expect_equal(kids_2$n_cols, 2)
  expect_equal(kids_2$col_widths$sm, list(width = 1, before = 0, after = 0))

  kids_3 <- get_col_spec(NA, 3)
  expect_equal(kids_3$n_cols, 3)
  expect_equal(kids_3$col_widths$sm, list(width = 1, before = 0, after = 0))

  kids_4 <- get_col_spec(NA, 4)
  expect_equal(kids_4$n_cols, 8)
  expect_equal(kids_4$col_widths$sm, list(width = 4, before = 0, after = 0))
  expect_equal(kids_4$col_widths$lg, list(width = 2, before = 0, after = 0))

  kids_5 <- get_col_spec(NA, 5)
  expect_equal(kids_5$n_cols, 5 * 2)
  expect_equal(kids_5$col_widths$sm, list(width = 5, before = 0, after = 0))
  expect_equal(kids_5$col_widths$lg, list(width = 2, before = 0, after = 0))

  kids_6 <- get_col_spec(NA, 6)
  expect_equal(kids_6$n_cols, 6 * 2)
  expect_equal(kids_6$col_widths$sm, list(width = 6, before = 0, after = 0))
  expect_equal(kids_6$col_widths$lg, list(width = 2, before = 0, after = 0))

  kids_7 <- get_col_spec(NA, 7)
  expect_equal(kids_7$n_cols, 7 * 2)
  expect_equal(kids_7$col_widths$sm, list(width = 7, before = 0, after = 0))
  expect_equal(kids_7$col_widths$lg, list(width = 2, before = 0, after = 0))

  kids_8 <- get_col_spec(NA, 8)
  expect_equal(kids_8$n_cols, 12)
  expect_equal(kids_8$col_widths$sm, list(width = 6, before = 0, after = 0))
  expect_equal(kids_8$col_widths$lg, list(width = 3, before = 0, after = 0))
})

test_that("impute_col_spec() missing smaller breakpoints inherit from lg+", {
  expect_equal(
    impute_col_spec(breakpoints(lg = c(4, -4, 4)), 2)$col_widths,
    breakpoints(
      md = c(4, -4, 4),
      lg = c(4, -4, 4)
    )
  )

  expect_equal(
    impute_col_spec(breakpoints(xl = c(4, -4, 4)), 2)$col_widths,
    breakpoints(
      md = c(4, -4, 4),
      xl = c(4, -4, 4)
    )
  )

  expect_equal(
    impute_col_spec(breakpoints(xxl = c(4, -4, 4)), 2)$col_widths,
    breakpoints(
      md = c(4, -4, 4),
      xxl = c(4, -4, 4)
    )
  )
})

test_that("impute_col_spec() base case", {
  expect_equal(
    impute_col_spec(1:3, 6),
    list(n_cols = 12, col_widths = breakpoints(md = 1:3))
  )
})

test_that("row_heights_css_vars() decides fr/px for numeric, passes character", {
  expect_equal(
    row_heights_css_vars(breakpoints(md = c(2, 1)))$style,
    "--bslib-grid--row-heights--md:2fr 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list(2, 1)))$style,
    "--bslib-grid--row-heights--md:2fr 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list("auto", 1)))$style,
    "--bslib-grid--row-heights--md:auto 1fr;"
  )

  expect_equal(
    row_heights_css_vars(breakpoints(md = list("10px", 1)))$style,
    "--bslib-grid--row-heights--md:10px 1fr;"
  )
})

test_that("layout_column_wrap() handles deprecated width as first arg", {
  # first arg is fractional
  lifecycle::expect_deprecated(
    lc_implicit_width_frac <- layout_column_wrap(1/2, "one", "two")
  )

  expect_equal(
    as.character(lc_implicit_width_frac),
    as.character(layout_column_wrap(width = 1/2, "one", "two"))
  )

  # first arg is explicitly px character
  lifecycle::expect_deprecated(
    lc_implicit_width_px <- layout_column_wrap("400px", "one", "two")
  )

  expect_equal(
    as.character(lc_implicit_width_px),
    as.character(layout_column_wrap(width = "400px", "one", "two"))
  )

  # first arg is px, but numeric
  lifecycle::expect_deprecated(
    lc_implicit_width_px_implied <- layout_column_wrap(365, "one", "two")
  )

  expect_equal(
    as.character(lc_implicit_width_px_implied),
    as.character(layout_column_wrap(width = 365, "one", "two"))
  )

  # first arg is NULL
  lifecycle::expect_deprecated(
    lc_implicit_width_null <- layout_column_wrap(NULL, "one", "two")
  )

  expect_equal(
    as.character(lc_implicit_width_null),
    as.character(layout_column_wrap(width = NULL, "one", "two"))
  )

  # first arg is not a CSS unit
  rlang::local_options(lifecycle_verbosity = "warning")
  testthat::expect_silent(
    as.character(layout_column_wrap("1ft", "one", "two"))
  )
})

# Load required libraries for data manipulation and visualization
library(dplyr)
library(ggplot2)

# Create a sample dataset with random values
data <- data.frame(
  group = rep(c("A", "B", "C"), each = 10),
  value = rnorm(30, mean = 50, sd = 15),
  count = sample(1:100, 30)
)

# Process data using tidyverse pipes and functions
result <- data %>%
  filter(count > 20) %>%
  group_by(group) %>%
  summarise(
    mean_value = mean(value),
    median_count = median(count),
    n = n(),
    .groups = "drop"
  )

# Create visualization with ggplot2
plot_output = ggplot(result, aes(x = group, y = mean_value, fill = group)) +
  geom_col(alpha = 0.8, show.legend = FALSE) +
  labs(title = "Summary Statistics by Group", x = "Group", y = "Mean Value") +
  theme_minimal() +
  theme(plot.title = element_text(hjust = 0.5, size = 14, face = "bold"))

print(plot_output)

library(stringr)

# if (Sys.getenv("RSTUDIO") == "1") {
#   stop("Please run this script from the command line: `Rscript tools/download_preset_fonts.R`")
# }

# TODO: make sure this isn't drastically different from 2.1 Mb
fonts_home <- file.path("inst", "fonts")
unlink(fonts_home, recursive = TRUE)
dir.create(fonts_home)

extract_first_group <- function(x, pattern) {
  matches <- regmatches(x, regexec(pattern, x))
  na.omit(sapply(matches, "[", 2))
}

# Google Fonts does browser sniffing to determine which font file type is appropriate.
# At the time of writing, woff2 is recommended for Chrome, and actually provides pretty
# good browser support (and compression), so we'll adopt it universally
# https://caniuse.com/woff2
request_headers <- c(
  "User-Agent" = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36"
)

download_and_copy_fonts <- function(theme, rule_file = "_bootswatch.scss") {
  theme_file <- file.path(theme, rule_file)

  if (!file.exists(theme_file)) {
    stop("Couldn't find '", rule_file, "' in ", theme)
  }

  theme_scss <- readLines(theme_file)

  web_font_path <- grep(
    "$web-font-path:",
    theme_scss,
    value = TRUE,
    fixed = TRUE
  )
  if (!length(web_font_path)) {
    return()
  }
  web_font_url <- strsplit(stringr::str_trim(web_font_path), "\\s+")[[1]][2]
  if (!length(web_font_url)) {
    warning(
      "Wasn't able to extract a url for $web-font-path from theme: ",
      theme
    )
    return()
  }
  web_font_url <- gsub('(^")|("$)', '', web_font_url)
  css_file <- file.path(theme, "font.css")
  download.file(
    utils::URLencode(web_font_url),
    css_file,
    method = "libcurl",
    headers = request_headers
  )
  css <- readLines(css_file)
  urls <- sass:::extract_group(css, "url\\(([^)]+)")
  basenames <- basename(urls)
  Map(
    function(url, nm) {
      if (!grepl("[.]woff2$", nm)) {
        stop("Expected a woff2 font file")
      }

      target <- file.path(fonts_home, nm)
      # The basename can sometimes be very long, and R CMD check
      # will complain if the target file is over 100 characters long,
      # so shorten it if necessary
      if (nchar(file.path("bslib", target)) > 100) {
        nm <- paste0(
          rlang::hash(tools::file_path_sans_ext(nm)),
          ".",
          tools::file_ext(nm)
        )
        target <- file.path(fonts_home, nm)
      }
      download.file(url, target, method = "libcurl", headers = request_headers)
      css <<- sub(url, file.path("fonts", nm), css, fixed = TRUE)
    },
    urls,
    basenames
  )
  writeLines(css, css_file)
  NULL
}

themes5 <- list.dirs(
  "inst/lib/bsw5/dist",
  recursive = FALSE,
  full.names = TRUE
)
themes4 <- list.dirs(
  "inst/lib/bsw4/dist",
  recursive = FALSE,
  full.names = TRUE
)
themes3 <- list.dirs(
  "inst/lib/bsw3",
  recursive = FALSE,
  full.names = TRUE
)
builtin5 <- list.dirs(
  "inst/builtin/bs5",
  recursive = FALSE,
  full.names = TRUE
)

invisible({
  lapply(themes5, download_and_copy_fonts)
  lapply(themes4, download_and_copy_fonts)
  lapply(themes3, download_and_copy_fonts)
})

lapply(builtin5, download_and_copy_fonts, rule_file = "_rules.scss")

###
# These methods could be removed if shiny were able to be imported.
# * `shiny::in_devmode()`
# * `shiny::get_devmode_option()`
###

get_precompiled_option <- function(
  name = "bslib.precompiled",
  default = TRUE,
  devmode_default = FALSE,
  devmode_message = "Disabling the use of bslib precompiled themes. To be able to use precompiled themes, call `options(bslib.precompiled = TRUE)`"
) {
  get_shiny_devmode_option(
    name,
    default = default,
    devmode_default = devmode_default,
    devmode_message = devmode_message
  )
}


in_shiny_devmode <- function() {
  isTRUE(getOption("shiny.devmode", FALSE)) &&
    !identical(Sys.getenv("TESTTHAT"), "true")
}

#' @importFrom rlang missing_arg is_missing maybe_missing
get_shiny_devmode_option <- function(
  name,
  default = NULL,
  devmode_default = missing_arg(),
  devmode_message = missing_arg()
) {

  if (!in_shiny_devmode()) {
    # Dev Mode disabled, act like `getOption()`
    return(getOption(name, default = default))
  }

  # Dev Mode enabled, update the default value for `getOption()`
  getOption(name, default = {
    # Notify developer
    if (
      !is_missing(devmode_message) &&
      !is.null(devmode_message) &&
      getOption("shiny.devmode.verbose", TRUE)
    ) {
      devmode_message <- paste0("shiny devmode - ", devmode_message)
      rlang::inform(
        message = devmode_message,
        .frequency = "regularly",
        .frequency_id = devmode_message,
        .file = stderr()
      )
    }

    # Return Dev Mode default value `devmode_default`
    maybe_missing(devmode_default, default)
  })
}

test_that("shiny devmode shuts off default caching", {
  expect_equal(
    get_precompiled_option(),
    TRUE
  )

  withr::local_options(
    list(
      shiny.devmode = TRUE,
      shiny.devmode.verbose = FALSE
    )
  )
  withr::local_envvar(
    list(
      TESTTHAT = "false"
    )
  )

  expect_equal(
    get_precompiled_option(),
    FALSE
  )
})

test_that("shiny devmode shuts off default caching", {

  expect_equal(
    get_precompiled_option(),
    TRUE
  )


  withr::local_options(list(
    shiny.devmode = TRUE,
    shiny.devmode.verbose = FALSE
  ))
  withr::local_envvar(list(
    TESTTHAT = "false"
  ))

  expect_equal(
    get_precompiled_option(),
    FALSE
  )

})

#' Compile Bootstrap Sass with (optional) theming
#'
#' `bs_theme_dependencies()` compiles Bootstrap Sass into CSS and returns it,
#' along with other HTML dependencies, as a list of
#' [htmltools::htmlDependency()]s. Most users won't need to call this function
#' directly as Shiny & R Markdown will perform this compilation automatically
#' when handed a [bs_theme()]. If you're here looking to create a themeable
#' component, see [bs_dependency()].
#'
#' @section Sass caching and precompilation:
#'
#' If Shiny Developer Mode is enabled (by setting `options(shiny.devmode =
#' TRUE)` or calling `shiny::devmode(TRUE)`), both \pkg{sass} caching and \pkg{bslib}
#' precompilation are disabled by default; that is, a call to
#' `bs_theme_dependencies(theme)` expands to `bs_theme_dependencies(theme, cache
#' = F, precompiled = F)`). This is useful for local development as
#' enabling caching/precompilation may produce incorrect results if local
#' changes are made to bslib's source files.
#'
#' @inheritParams bs_theme_update
#' @param sass_options a [sass::sass_options()] object.
#' @param jquery a [jquerylib::jquery_core()] object.
#' @param precompiled Before compiling the theme object, first look for a
#'   precompiled CSS file for the [theme_version()].  If `precompiled = TRUE`
#'   and a precompiled CSS file exists for the theme object, it will be fetched
#'   immediately and not compiled. At the moment, we only provide precompiled
#'   CSS for "stock" builds of Bootstrap (i.e., no theming additions, bootswatch
#'   themes, or non-default `sass_options`).
#' @inheritParams sass::sass
#'
#' @return a list of HTML dependencies containing Bootstrap CSS, Bootstrap
#'   JavaScript, and `jquery`. This list may contain additional HTML
#'   dependencies if bundled with the `theme`.
#'
#' @export
#' @seealso [bs_theme()], [bs_dependency()]
#' @examples
#'
#' # Function to preview the styling a (primary) Bootstrap button
#' library(htmltools)
#' button <- tags$a(class = "btn btn-primary", href = "#", role = "button", "Hello")
#' preview_button <- function(theme) {
#'   if (interactive()) {
#'     browsable(tags$body(bs_theme_dependencies(theme), button))
#'   }
#' }
#'
#' # Latest Bootstrap
#' preview_button(bs_theme())
#' # Bootstrap 3
#' preview_button(bs_theme(3))
#' # Bootswatch 4 minty theme
#' preview_button(bs_theme(4, bootswatch = "minty"))
#' # Bootswatch 4 sketchy theme
#' preview_button(bs_theme(4, bootswatch = "sketchy"))
#'
bs_theme_dependencies <- function(
  theme,
  sass_options = sass::sass_options_get(output_style = "compressed"),
  cache = sass::sass_cache_get(),
  jquery = jquerylib::jquery_core(3),
  precompiled = get_precompiled_option("bslib.precompiled", default = TRUE)
) {

  theme <- as_bs_theme(theme)
  version <- theme_version(theme)

  if (isTRUE(version >= 5)) {
    shiny_version <- "1.6.0.9001"
    msg <- sprintf("`bs_theme(version = 5)` is designed to work with shiny %s or higher", shiny_version)
    if (isNamespaceLoaded("shiny") && !is_installed("shiny", shiny_version)) warning(msg, call. = FALSE)
    setHook(packageEvent("shiny", "onLoad"), function(...) {
      if (!is_installed("shiny", shiny_version)) warning(msg, call. = FALSE)
    })
  }

  if (is.character(cache)) {
    cache <- sass_cache_get_dir(cache)
  }

  out_file <- NULL
  # Look for a precompiled css file if user asks for it AND the default options
  # are used.
  if (precompiled &&
      identical(sass_options, sass_options(output_style = "compressed")))
  {
    precompiled_css <- precompiled_css_path(theme)
    if (!is.null(precompiled_css)) {
      out_dir <- file.path(tempdir(), paste0("bslib-precompiled-", version))
      if (!dir.exists(out_dir)) {
        dir.create(out_dir)
      }
      out_file <- file.path(out_dir, basename(precompiled_css))
      file.copy(precompiled_css, out_file)

      # Usually sass() would handle file_attachments and dependencies,
      # but we need to do this manually
      out_file <- attachDependencies(out_file, htmlDependencies(as_sass(theme)))
      write_file_attachments(
        as_sass_layer(theme)$file_attachments,
        out_dir
      )
    }
  }

  # If precompiled css not found, compile normally.
  if (is.null(out_file)) {

    contrast_warn <- get_shiny_devmode_option(
      "bslib.color_contrast_warnings",
      default = FALSE,
      devmode_default = TRUE,
      devmode_message = paste(
        "Enabling warnings about low color contrasts found inside `bslib::bs_theme()`.",
        "To suppress these warnings, set `options(bslib.color_contrast_warnings = FALSE)`"
      )
    )
    theme <- bs_add_variables(theme, "color-contrast-warnings" = contrast_warn)

    out_file <- sass(
      input = theme,
      options = sass_options,
      output = output_template(basename = "bootstrap", dirname = "bslib-"),
      cache = cache,
      write_attachments = TRUE,
      cache_key_extra = list(
        get_exact_version(version),
        get_package_version("bslib")
      )
    )
  }

  out_file_dir <- dirname(out_file)

  js_files <- bootstrap_javascript(version)
  js_map_files <- bootstrap_javascript_map(version)
  success_js_files <- file.copy(c(js_files, js_map_files), out_file_dir, overwrite = TRUE)
  if (any(!success_js_files)) {
    warning("Failed to copy over bootstrap's javascript files into the htmlDependency() directory.")
  }

  htmltools::resolveDependencies(c(
    if (inherits(jquery, "html_dependency")) list(jquery) else jquery,
    list(
      htmlDependency(
        name = "bootstrap",
        version = get_exact_version(version),
        src = out_file_dir,
        stylesheet = basename(out_file),
        script = basename(js_files),
        all_files = TRUE, # include font and map files
        meta = list(viewport = "width=device-width, initial-scale=1, shrink-to-fit=no")
      )
    ),
    htmlDependencies(out_file)
  ))
}


#' Themeable HTML components
#'
#' @description
#'
#' Themeable HTML components use Sass to generate CSS rules from Bootstrap Sass
#' variables, functions, and/or mixins (i.e., stuff inside of `theme`).
#' `bs_dependencies()` makes it a bit easier to create themeable components by
#' compiling [sass::sass()] (`input`) together with Bootstrap Sass inside of a
#' `theme`, and packaging up the result into an [htmlDependency()].
#'
#' Themable components can also be  _dynamically_ themed inside of Shiny (i.e.,
#' they may be themed in 'real-time' via [bs_themer()], and more generally,
#' update their styles in response to [shiny::session]'s `setCurrentTheme()`
#' method). Dynamically themeable components provide a "recipe" (i.e., a
#' function) to `bs_dependency_defer()`, describing how to generate new CSS
#' stylesheet(s) from a new `theme`. This function is called when the HTML page
#' is first rendered, and may be invoked again with a new `theme` whenever
#' [shiny::session]'s `setCurrentTheme()` is called.
#'
#' @param input Sass rules to compile, using `theme`.
#' @param theme A [bs_theme()] object.
#' @param cache_key_extra Extra information to add to the sass cache key. It is
#'   useful to add the version of your package.
#' @param .dep_args A list of additional arguments to pass to
#'   [htmltools::htmlDependency()]. Note that `package` has no effect and
#'   `script` must be absolute path(s).
#' @param .sass_args A list of additional arguments to pass to
#'   [sass::sass_partial()].
#' @inheritParams htmltools::htmlDependency
#' @references
#' <https://rstudio.github.io/bslib/articles/custom-components.html>
#'
#'
#' @return `bs_dependency()` returns an [htmltools::htmlDependency()] and
#'   `bs_dependency_defer()` returns an [htmltools::tagFunction()]
#' @export
bs_dependency <- function(input = list(), theme, name, version,
  cache_key_extra = NULL, .dep_args = list(), .sass_args = list())
{
  sass_args <- c(
    list(
      rules = input,
      bundle = theme,
      output = output_template(basename = name, dirname = name),
      write_attachments = TRUE,
      cache_key_extra = cache_key_extra
    ),
    .sass_args
  )
  outfile <- do.call(sass_partial, sass_args)

  dep_args <- list(
    name = name,
    version = version,
    src = dirname(outfile),
    stylesheet = basename(outfile)
  )

  bad_args <- intersect(names(.dep_args), names(dep_args))
  if (length(bad_args)) {
    stop(
      "The following `.dep_args` must be provided as top-level args to `bs_dependency()`: ",
      paste(bad_args, collapse = ", ")
    )
  }

  if ("package" %in% names(.dep_args)) {
    warning("`package` won't have any effect since `src` must be an absolute path")
  }

  script <- .dep_args[["script"]]
  if (length(script)) {
    if (basename(outfile) %in% basename(script)) {
      stop("`script` file basename(s) must all be something other than ", basename(outfile))
    }
    success <- file.copy(script, dirname(outfile), overwrite = TRUE)
    if (!all(success)) {
      stop(
        "Failed to copy the following script(s): ",
        paste(script[!success], collapse = ", "), ".\n\n",
        "Make sure script are absolute path(s)."
      )
    }
    .dep_args[["script"]] <- basename(script)
  }

  do.call(htmlDependency, c(dep_args, .dep_args))
}

#' @param func a _non-anonymous_ function, with a _single_ argument.
#'   This function should accept a [bs_theme()] object and return a single
#'   [htmlDependency()], a list of them, or `NULL`.
#' @param memoise whether or not to memoise (i.e., cache) `func` results for a
#'   short period of time. The default, `TRUE`, can have large performance
#'   benefits when many instances of the same themable widget are rendered. Note
#'   that you may want to avoid memoisation if `func` relies on side-effects
#'   (e.g., files on-disk) that need to change for each themable widget
#'   instance.
#' @export
#'
#' @examples
#'
#' \dontrun{
#'
#' myWidgetVersion <- "1.2.3"
#'
#' myWidgetDependency <- function() {
#'   list(
#'     bs_dependency_defer(myWidgetCss),
#'     htmlDependency(
#'       name = "mywidget-js",
#'       version = myWidgetVersion,
#'       src = system.file(package = "mypackage", "js"),
#'       script = "mywidget.js"
#'     )
#'   )
#' }
#'
#' myWidgetCSS <- function(theme) {
#'   if (!is_bs_theme(theme)) {
#'     return(
#'       htmlDependency(
#'         name = "mywidget-css",
#'         version = myWidgetVersion,
#'         src = system.file(package = "mypackage", "css"),
#'         stylesheet = "mywidget.css"
#'       )
#'     )
#'   }
#'
#'   # Compile mywidget.scss using the variables and defaults from the theme
#'   # object.
#'   sass_input <- sass::sass_file(system.file(package = "mypackage", "scss/mywidget.scss"))
#'
#'   bs_dependency(
#'     input = sass_input,
#'     theme = theme,
#'     name = "mywidget",
#'     version = myWidgetVersion,
#'     cache_key_extra = utils::packageVersion("mypackage")
#'   )
#' }
#'
#' # Note that myWidgetDependency is not defined inside of myWidget. This is so
#' # that, if `myWidget()` is called multiple times, Shiny can tell that the
#' # function objects are identical and deduplicate them.
#' myWidget <- function(id) {
#'   div(
#'     id = id,
#'     span("myWidget"),
#'     myWidgetDependency()
#'   )
#' }
#' }
#'
#' @rdname bs_dependency
bs_dependency_defer <- function(func, memoise = TRUE) {
  # func() most likely calls stuff like sass_file() and bs_dependency() ->
  # sass_partial() -> sass() (e.g., see example section above) Even though
  # sass() calls can be cached, there is still considerable overhead involved
  # in computing the cache key itself because it involves file-level operations
  # (e.g. getting the mtimes of file attachments).
  # So, to help with performance (in the case where we repeatedly call the same
  # func with the same theme), memoise func with a short-lived cache
  if (memoise) {
    # The memoized function is created each time bs_dependency_defer is called,
    # and then it is used once. This is not how memoized functions are normally
    # used, but in this case it works because the caching object is re-used, and
    # it still provides very significant improvement in performance.
    mfunc <- memoise::memoise(func, cache = .dependency_cache)
  } else {
    mfunc <- func
  }

  tagFunction(function() {
    if (is_shiny_app()) {
      # If we're in a Shiny app, do two things:
      # (1) Register this function as a dependency so that Shiny will know to
      # update it later if the theme dynamically changes (i.e.,
      # session$setCurrentTheme() is called). Note that we *don't* provide
      # the memoised version since Shiny will de-duplicate identical()
      # functions anyway.
      # (2) Call the user's `func()` with the current theme, and return the
      # resulting htmlDependency so that it can be embedded in the static page.
      shiny::registerThemeDependency(func)

      return(mfunc(get_current_theme()))
    }

    # Outside of a Shiny context, we'll just get the global theme.
    mfunc(bs_global_get())
  })
}

as_bs_theme <- function(theme) {
  if (is_bs_theme(theme)) return(theme)

  # This is a historical artifact that should happen
  if (is_sass_bundle(theme) || inherits(theme, "sass_layer")) {
    stop("`theme` cannot be a `sass_bundle()` or `sass_layer()` (use `bs_bundle()` to add a bundle)")
  }

  # NULL means default Bootstrap
  if (is.null(theme)) return(bs_theme())

  # For example, `bs_theme_dependencies(theme = 4)`
  if (is.numeric(theme)) return(bs_theme(version = theme))

  # For example, `bs_theme_dependencies(theme = 'bootswatch@version')`
  if (is_string(theme)) {
    theme <- strsplit(theme, "@", fixed = TRUE)[[1]]
    if (length(theme) == 2) {
      return(bs_theme(version = theme[2], bootswatch = theme[1]))
    }
    # Also support `bs_theme_dependencies(version = '4')` and
    # `bs_theme_dependencies(theme = 'bootswatch')`
    if (length(theme) == 1) {
      if (theme %in% c(versions(), "4-3", "4+3")) {
        return(bs_theme(version = theme))
      } else {
        return(bs_theme(bootswatch = theme))
      }
    }
    stop("If `theme` is a string, it can't contain more than one @")
  }

  stop(
    "`theme` must be one of the following: (1) `NULL`, ",
    "(2) a `'bootswatch@version'` string, ",
    "(3) the result of `bs_global_get()`."
  )
}

local_disable_cache()

test_that("Can access the sass behind all versions and Bootswatch themes", {
  skip_on_cran()

  for (version in versions()) {
    # Can compile CSS against variables (in each version)
    css <- sass_partial("body{background-color:$body-bg}", as_bs_theme(version))
    css <- sub("#ffffff", "#fff", css)
    expect_css(css, "body{background-color:#fff;}")
    themes <- bootswatch_themes(version)
    for (theme in themes) {
      # Can compile each bootswatch theme/version combination
      bs <- bs_theme_dependencies(
        theme = as_bs_theme(paste0(theme, "@", version))
      )
      expect_true(length(bs) > 1)
    }
  }
})

# The internal bootstrap_bundle() splits up bootstrap.scss into defaults/declarations/rules,
# and so, has to make assumptions about what's in that file. Thus, everytime this file
# changes, we should check to make sure we've made the appropriate changes in bootstrap_bundle()
# (and once we have, then this hash should be updated as well).
test_that("bootstrap.scss hasn't changed", {
  expect_snapshot_file(
    path_lib("bs4", "scss", "bootstrap.scss"),
    name = "bs4-imports"
  )
  expect_snapshot_file(
    path_lib("bs5", "scss", "bootstrap.scss"),
    name = "bs5-imports"
  )
})

# If Bootstrap changes their %heading placeholder, then we'll likely
# need to update our bootstrap-heading() mixin to replicate those changes
# https://github.com/twbs/bootstrap/blob/2c7f88/scss/_reboot.scss#L83-L96
# https://github.com/rstudio/bslib/tree/main/inst/components/scss/mixins/_mixins.scss
test_that("bootstrap-heading() mixin approximates %heading placeholder", {
  reboot <- readLines(path_lib("bs5", "scss", "_reboot.scss"))
  idx_start <- which(reboot == "%heading {")
  idx_end <- NA
  for (idx in which(reboot == "}")) {
    if (idx > idx_start) {
      idx_end <- idx
      break
    }
  }
  heading_placeholder <- reboot[seq.int(idx_start, idx_end)]
  expect_snapshot(heading_placeholder)
})

local_disable_cache()

test_that("Can access the sass behind all versions and Bootswatch themes", {
  skip_on_cran()

  for (version in versions()) {
    # Can compile CSS against variables (in each version)
    css <- sass_partial("body{background-color:$body-bg}", as_bs_theme(version))
    expect_css(css, "body{background-color:#fff;}")
    themes <- bootswatch_themes(version)
    for (theme in themes) {
      # Can compile each bootswatch theme/version combination
      bs <- bs_theme_dependencies(theme = as_bs_theme(paste0(theme, "@", version)))
      expect_true(length(bs) > 1)
    }
  }
})


# The internal bootstrap_bundle() splits up bootstrap.scss into defaults/declarations/rules,
# and so, has to make assumptions about what's in that file. Thus, everytime this file
# changes, we should check to make sure we've made the appropriate changes in bootstrap_bundle()
# (and once we have, then this hash should be updated as well).
test_that("Make sure bootstrap.scss hasn't changed", {
  expect_snapshot_file(
    lib_file("bs4", "scss", "bootstrap.scss"),
    name = "bs4-imports"
  )
  expect_snapshot_file(
    lib_file("bs5", "scss", "bootstrap.scss"),
    name = "bs5-imports"
  )
})

label_with_extras <- function(label, ...) {
  div(
    class = "d-inline-block w-100",
    span(label),
    span(class = "float-right", ...),
    singleton(
      tags$style(HTML(".shiny-input-container .control-label { width: 100%; }"))
    )
  )
}

ui_selextra <- function(id, label) {
  ns <- shiny::NS(id)

  selectizeInput(
    inputId = ns("selected"),
    choices = NULL,
    label_with_extras(
      label,
      actionLink(
        ns("shuffle"),
        bsicons::bs_icon("shuffle", title = paste("Random", label))
      ),
      actionLink(
        ns("prev"),
        bsicons::bs_icon("caret-left-fill", title = paste("Previous", label))
      ),
      actionLink(
        ns("next"),
        bsicons::bs_icon("caret-right-fill", title = paste("Next", label))
      )
    )
  )
}

server_selextra <- function(input, output, session, choices) {
  ns <- session$ns

  # These are server-side selectize inputs, so we update them on initialization
  updateSelectizeInput(session, "selected", choices = choices, server = TRUE)

  trigger_shuffle <- reactiveVal(0)
  trigger_next <- reactiveVal(0)
  trigger_prev <- reactiveVal(0)

  observeEvent(
    input$shuffle,
    ignoreInit = TRUE,
    trigger_shuffle(trigger_shuffle() + 1)
  )

  observeEvent(trigger_shuffle(), {
    req(trigger_shuffle() > 0)

    updateSelectizeInput(
      session,
      "selected",
      selected = sample(unlist(choices), 1),
      choices = choices,
      server = TRUE
    )
  })

  observeEvent(input[["next"]], move(1))
  observeEvent(trigger_next(), move(1), ignoreInit = TRUE)

  observeEvent(input[["prev"]], move(-1))
  observeEvent(trigger_prev(), move(-1), ignoreInit = TRUE)

  move <- reactiveVal(0)

  observeEvent(move(), {
    req(move() != 0)

    current <- input$selected
    choices_flat <- unlist(choices)

    idx <- which(choices_flat == current) + move()
    move(0)
    req(idx)

    if (idx > length(choices_flat)) {
      idx <- 1
    }
    if (idx <= 0) {
      idx <- length(choices_flat)
    }

    updateSelectizeInput(
      session,
      "selected",
      selected = choices_flat[[idx]],
      choices = choices,
      server = TRUE
    )
  })

  list(
    "value" = reactive(input$selected %||% ""),
    "shuffle" = function() trigger_shuffle(as.integer(Sys.time())),
    "next" = function() trigger_next(as.integer(Sys.time())),
    "prev" = function() trigger_prev(as.integer(Sys.time())),
    "choices" = choices,
    "set" = function(value) {
      updateSelectizeInput(
        session,
        "selected",
        selected = value,
        choices = choices,
        server = TRUE
      )
    }
  )
}

module_selextra <- function(id, choices) {
  callModule(server_selextra, id, choices = choices)
}

#!/usr/bin/env Rscript
library(sass)
library(rprojroot)
library(brio)
library(withr)


if (!identical(getwd(), find_package_root_file())) {
  stop("This script must be run from the top directory of the bslib package")
}

if (Sys.which("yarn") == "") {
  stop("The yarn CLI must be installed and in your PATH")
}

# only install the direct deps
with_dir("inst", system("yarn install --production"))

unlink("inst/lib", recursive = TRUE)
file.rename("inst/node_modules/", "inst/lib/")

# not used
unlink("inst/lib/.yarn-integrity")

# jquery comes in via jquerylib (R package)
unlink("inst/lib/jquery", recursive = TRUE)

# bootstrap is a peer dependency of bs-colorpicker?
unlink("inst/lib/bootstrap", recursive = TRUE)

# bootstrap.bundle.min.js includes popper (but not jQuery)
# https://getbootstrap.com/docs/4.4/getting-started/introduction/#js
unlink("inst/lib/popper.js", recursive = TRUE)


# ----------------------------------------------------------------------
# Add known vendor prefixes to all scss files since we don't want
# to rely on a node run-time to prefix after compilation
# https://github.com/twbs/bootstrap/blob/d438f3/package.json#L30
# ----------------------------------------------------------------------
scss_files <- dir("inst", pattern = "\\.scss$", recursive = TRUE, full.names = TRUE)
# These libs should already have prefixes in their source
# TODO: add test(s) that we aren't missing vendor prefixes
scss_files <- scss_files[!grepl("(^inst/lib/bs3)|(^inst/bs3compat)|(^inst/themer)|(^inst/nav-spacer)", scss_files)]

scss_src <- lapply(scss_files, readLines)

add_property_prefixes <- function(src, property, ok_values = NULL, vendors = c("-webkit-", "-moz-", "-ms-", "-o-")) {
  pattern <- paste0("^\\s*", property, ":\\s*(.+);")
  idx <- grep(pattern, src)
  for (i in idx) {
    prop <- src[[i]]
    if (length(ok_values)) {
      value <- regmatches(prop, regexec(pattern, prop))[[1]][2]
      vals <- strsplit(value, "\\s+")[[1]]
      if (all(vals %in% c(ok_values, "!important"))) next
    }
    leading_ws <- regmatches(prop, regexpr("^\\s+", prop))
    src[[i]] <- paste0(
      leading_ws,
      c("", vendors),
      sub("^\\s+", "", prop),
      collapse = "\n"
    )
  }
  src
}

# Unconditionally prefix the following CSS properties for all vendors
needs_prefix <- c(
  "appearance", "user-select", "backdrop-filter",
  "backface-visibility", "touch-action",
  "animation-duration"
)
for (prop in needs_prefix) {
  scss_src <- lapply(scss_src, add_property_prefixes, prop)
}

# Only add webkit prefix for BS5+ since other vendors aren't really relevant anymore
for (prop in c('mask-image', 'mask-size', 'mask-position')) {
  scss_src <- lapply(scss_src, add_property_prefixes, prop, vendors = "-webkit-")
}

# Print specific vendor prefixes
scss_src <- lapply(scss_src, add_property_prefixes, "color-adjust", vendors = "-webkit-print-")

# phantomjs 2.1.1 needs webkit vendor prefix on flex properties to work correctly
flex_props <- c(
  "flex-direction", "flex-wrap", "flex-flow",  "justify-content",
  "align-items", "align-content", "order", "flex-grow", "flex-shrink",
  "flex-basis", "flex", "align-self"
)
for (prop in flex_props) {
  scss_src <- lapply(scss_src, add_property_prefixes, prop, vendors = "-webkit-")
}

# Conditionally prefix text-decoration if its not CSS2 compliant
# https://www.w3.org/TR/CSS2/text.html#lining-striking-props
# https://developer.mozilla.org/en-US/docs/Web/CSS/text-decoration
# https://caniuse.com/#search=text-decoration
scss_src <- lapply(
  scss_src, add_property_prefixes, "text-decoration",
  ok_values = c("none", "underline", "overline", "line-through", "blink", "inherit")
)


add_value_prefixes <- function(src, value, vendors = c("-webkit-", "-moz-", "-ms-", "-o-")) {
  pattern <- paste0("^\\s*[^/]+:\\s*", value, "\\s*(!important)?\\s*;")
  idx <- grep(pattern, src)
  for (i in idx) {
    prop_val <- strsplit(src[[i]], ":\\s*")[[1]]
    src[[i]] <- paste0(
      prop_val[1], ": ", c("", vendors), prop_val[2],
      collapse = "\n"
    )
  }
  src
}

# https://caniuse.com/?search=min-content
scss_src <- lapply(scss_src, add_value_prefixes, "min-content")
scss_src <- lapply(scss_src, add_value_prefixes, "max-content")

# phantomjs 2.1.1 needs `display: -webkit-flex` to work properly
scss_src <- lapply(scss_src, add_value_prefixes, "flex", vendors = "-webkit-")

# Write modified source to disk
invisible(Map(writeLines, scss_src, scss_files))

# ----------------------------------------------------------------------
# Check to make sure we're not missing any vendor prefixes
# that we don't already know about in the distributed CSS
# ----------------------------------------------------------------------

find_prefixed_css <- function(css) {
  vendors <- c("webkit", "moz", "ms")
  prefixes <- lapply(vendors, function(vendor) {
    pattern <- sprintf("-%s-([^:|;| |\\)]+)", vendor)
    prefixes <- regmatches(css, regexec(pattern, css))
    lapply(prefixes, function(x) if (length(x) > 1) x[2] else NULL)
  })
  unique(unlist(prefixes, recursive = TRUE))
}

# TODO: do for each bootstrap?
src_prefixes <- find_prefixed_css(unlist(scss_src))
dist_prefixes <- find_prefixed_css(
  c(
    readLines("inst/lib/bs4/dist/css/bootstrap.css"),
    readLines("inst/lib/bs5/dist/css/bootstrap.css")
  )
)
auto_prefixes <- setdiff(dist_prefixes, src_prefixes)

whitelist <- c(
  # https://caniuse.com/#feat=flexbox
  "flex", "inline-flex", "inline-flexbox",
  # https://caniuse.com/#feat=mdn-api_csskeyframesrule
  "keyframes",
  # https://caniuse.com/#feat=mdn-css_properties_transition
  "transition",
  # https://caniuse.com/#feat=css-animation
  "animation",
  # https://caniuse.com/#feat=transforms2d
  "transform",
  # https://caniuse.com/#feat=mdn-css_properties_column-count
  "column-count",
  # https://caniuse.com/#feat=mdn-css_properties_column-gap_multicol_context
  "column-gap",
  # https://caniuse.com/#feat=css-placeholder
  "placeholder", "input-placeholder",
  # https://developer.mozilla.org/en-US/docs/Web/CSS/text-decoration-skip-ink
  "text-decoration-skip-ink",
  # webkit prefix doesn't seem necessary? And Bootstrap docs warn about IE...
  # https://caniuse.com/#search=sticky
  "sticky",
  # Already applied conditional prefixing
  "text-decoration",
  # IE11 technically needs a prefix, but this is just for floating labels
  # and I'm lazy https://caniuse.com/?search=placeholder-shown
  "placeholder-shown",
  # False positive?
  "margin-end"
)

unknown_prefixes <- setdiff(
  # whitelist flexbox props (there are many)
  auto_prefixes[!grepl("^flex-*", auto_prefixes)],
  whitelist
)
if (length(unknown_prefixes)) {
  stop(
    "Unknown vendor prefixes introduced by Bootstrap's autoprefixer. ",
    "Use either add_property_prefixes() to add prefixes or whitelist them ",
    "(if they're not needed for modern browsers): ",
    "'", paste(collapse = "', '", unknown_prefixes), "'.",
    call. = FALSE
  )
}


# ----------------------------------------------------------------------
# Now, get rid of files that we don't need to bundle with the package
# ----------------------------------------------------------------------

with_dir(
  "inst/lib", {
    # Downsize Bootstrap
    for (bs in c("bs5", "bs4")) {
      unlink(file.path(bs, "dist/css"), recursive = TRUE)
      unlink(file.path(bs, "js"), recursive = TRUE)
      js_dist <- file.path(bs, "dist/js")
      non_bundle <- setdiff(
        dir(js_dist), c("bootstrap.bundle.min.js", "bootstrap.bundle.min.js.map")
      )
      file.remove(file.path(js_dist, non_bundle))
    }

    # Downsize bootswatch
    for (bsw in c("bsw5", "bsw4")) {
      file.remove(Sys.glob(file.path(bsw, "dist/*/bootstrap.css")))
      file.remove(Sys.glob(file.path(bsw, "dist/*/bootstrap.min.css")))
      unlink("bsw5/docs", recursive = TRUE)
      # Remove hidden files
      unlink(
        dir(bsw, pattern = "^\\.[a-z]+", all.files = TRUE, full.names = TRUE), recursive = TRUE
      )
    }
    file.remove(c(
      Sys.glob("bsw3/*/bootstrap.css"),
      Sys.glob("bsw3/*/bootstrap.min.css"),
      Sys.glob("bsw3/*/thumbnail.png"),
      Sys.glob("bsw3/*/*.less")
    ))
    unlink("bsw3/docs", recursive = TRUE)
    unlink("bsw3/.github", recursive = TRUE)
    unlink("bsw3/fonts", recursive = TRUE) # have fonts via tools/download_fonts.R

    # Downsize bootstrap-accessibility
    with_dir("bs-a11y-p", {
      discard <- setdiff(dir(), c("src", "plugins", "LICENSE.md", "package.json"))
      unlink(discard, recursive = TRUE)
    })

    # Downsize bootstrap-colorpicker
    with_dir(
      "bs-colorpicker", {
        file.rename("dist/css", "css")
        file.rename("dist/js", "js")
        unlink("node_modules", recursive = TRUE)
        unlink("dist", recursive = TRUE)
        unlink("src", recursive = TRUE)
        unlink("logo.png")
      }
    )

    # GitHub reports security issues of devDependencies, but that's irrelevant to us
    remove_dev_dependencies <- function(pkg_file) {
      if (!file.exists(pkg_file)) return()
      json <- jsonlite::fromJSON(pkg_file)
      json <- json[setdiff(names(json), "devDependencies")]
      jsonlite::write_json(json, pkg_file, pretty = TRUE, auto_unbox = TRUE)
    }
    invisible(lapply(Sys.glob("*/package.json"), remove_dev_dependencies))

    # Get BS4/BS3 versions (for bslib::bs_dependencies() version-ing)
    version_bs5 <- sub("-beta[0-9]+", "", jsonlite::fromJSON("bs5/package.json")$version)
    version_bs4 <- jsonlite::fromJSON("bs4/package.json")$version
    version_bs3 <- jsonlite::fromJSON("bs3/package.json")$version
    version_accessibility <- jsonlite::fromJSON("bs-a11y-p/package.json")$version
  }
)

writeLines(
  c(
    '# DO NOT EDIT',
    '# This file is auto-generated by tools/yarn_install.R',
    paste0('version_bs5 <- ', deparse( version_bs5)),
    paste0('version_bs4 <- ', deparse(version_bs4)),
    paste0('version_bs3 <- ',  deparse(version_bs3)),
    paste0('version_accessibility <- ',  deparse(version_accessibility))
  ),
  "R/versions.R"
)


# ----------------------------------------------------------------------
# Apply any patches to source
# ----------------------------------------------------------------------

patch_files <- list.files(
  find_package_root_file("tools/patches"),
  full.names = TRUE
)

rej_pre <- dir(pattern = "\\.rej$", recursive = TRUE)
for (patch in patch_files) {
  message(sprintf("Applying %s", basename(patch)))
  res <- system(sprintf("git apply --reject --whitespace=fix '%s'", patch))
  if (res > 0) stop("Couldn't successfully apply patch: ", patch, call. = FALSE)
}
rej_post <- dir(pattern = "\\.rej$", recursive = TRUE)
if (length(rej_post) > length(rej_pre)) {
  warning(
    "Running `git apply --reject` generated `.rej` files. \n",
    "An 'easy' way to do this is to first `git add` the new source changes, ",
    "then manually make the relevant changes from the patch file,",
    "then `git diff` to get the relevant diff output and update the patch diff with the new diff."
  )
}

# Tracking changes in solar isn't so important since we've basically
# re-implemented it using a proper color system
# TODO: do the same for BS5?
writeLines(
  sass::as_sass(list(
    "white" = "#092B36 !default",
    "gray-100" = "#173741 !default",
    "gray-200" = "#25434B !default",
    "gray-300" = "#324E56 !default",
    "gray-400" = "#405A61 !default",
    "gray-500" = "#4E666B !default",
    "gray-600" = "#5C7276 !default",
    "gray-700" = "#6A7E81 !default",
    "gray-800" = "#77898C !default",
    "gray-900" = "#859596 !default",
    "black" = "#93A1A1 !default",
    "blue" = "#b58900 !default",
    "indigo" = "#6610f2 !default",
    "purple" = "#6f42c1 !default",
    "pink" = "#e83e8c !default",
    "red" = "#d33682 !default",
    "orange" = "#fd7e14 !default",
    "yellow" = "#cb4b16 !default",
    "green" = "#2aa198 !default",
    "teal" = "#20c997 !default",
    "cyan" = "#268bd2 !default",
    "enable-gradients" = "true !default",
    "secondary" = "$gray-800 !default",
    # Darker green and lighter green than `$black` and `$white`
    "color-contrast-dark" = "#031014 !default",
    "color-contrast-light" = "#BBD0D0 !default"
  )),
  "inst/lib/bsw4/dist/solar/_variables.scss"
)
writeLines(
  c(
    '$web-font-path: "https://fonts.googleapis.com/css2?family=Source+Sans+Pro:wght@400;600;700&display=swap" !default;',
    '@import url($web-font-path);'
  ),
  "inst/lib/bsw4/dist/solar/_bootswatch.scss"
)

# ----------------------------------------------------------------------
# Apply minification to patched files
# ----------------------------------------------------------------------

with_dir("inst", {
  local({
    # install all deps
    system("yarn install")
    # remove node modules
    on.exit({
      unlink("node_modules", recursive = TRUE)
    }, add = TRUE)

    bslib_plugin_paths <- file.path(
      "lib", "bs-a11y-p", "plugins", "js", c(
        "bootstrap-accessibility.js"
      ))

    for (unminified_file in c(
      bslib_plugin_paths
    )) {
      message("Minifying ", basename(unminified_file))
      cmd <- paste0(
        "yarn parcel",
        " build ", unminified_file,
        " --no-source-maps",
        " --no-cache",
        " --out-dir ", dirname(unminified_file),
        " --out-file ", sub(".js", ".min.js", fixed = TRUE, basename(unminified_file))
      )

      system(cmd)
    }

  })

})

# Clean up
unlink("inst/yarn.lock")
unlink("inst/node_modules", recursive = TRUE)

# ----------------------------------------------------------------------
# Precompile Bootstrap CSS
# ----------------------------------------------------------------------

# This generates precompiled builds of Bootstrap's css. It would be nice to do
# it at binary package build time, but I couldn't get that to work, using either
# src/install.libs.R (because the bslib functions used in this script
# aren't available yet), or by putting this code directly in the R/ directory
# (because the R/ files are evaluated only after the inst directory is copied
# over).
library(bslib)

precompiled_dir <- find_package_root_file("inst/css-precompiled")
unlink(precompiled_dir, recursive = TRUE)
dir.create(precompiled_dir, recursive = TRUE)

lapply(versions(), function(version) {
  res <- bs_theme_dependencies(
    bs_theme(version), precompiled = FALSE,
    sass_options = sass_options(output_style = "compressed"),
    cache = NULL
  )
  # Extract the Bootstrap dependency object (as opposed to, say, jQuery)
  bs_dep <- Filter(res, f = function(x) { identical(x$name, "bootstrap") })[[1]]

  tmp_css <- file.path(bs_dep$src$file, bs_dep$stylesheet)
  dest_dir <- file.path(precompiled_dir, version)
  if (!dir.exists(dest_dir)) {
    dir.create(dest_dir)
  }
  file.copy(tmp_css, dest_dir)
})

#!/usr/bin/env Rscript
library(sass)
library(rprojroot)
library(brio)
library(withr)

if (!identical(getwd(), find_package_root_file())) {
  stop("This script must be run from the top directory of the bslib package")
}

if (Sys.which("yarn") == "") {
  stop("The yarn CLI must be installed and in your PATH")
}

# Build _our_ JS assets (e.g., accordion component)
system("yarn install")
system("yarn build")

# only install the direct deps
with_dir("inst", system("yarn install --production"))

unlink("inst/lib", recursive = TRUE)
file.rename("inst/node_modules/", "inst/lib/")

# not used
unlink("inst/lib/.yarn-integrity")

# jquery comes in via jquerylib (R package)
unlink("inst/lib/jquery", recursive = TRUE)

# bootstrap is a peer dependency of bs-colorpicker?
unlink("inst/lib/bootstrap", recursive = TRUE)

# bootstrap.bundle.min.js includes popper (but not jQuery)
# https://getbootstrap.com/docs/4.4/getting-started/introduction/#js
unlink("inst/lib/popper.js", recursive = TRUE)

# ----------------------------------------------------------------------
# Add known vendor prefixes to all scss files since we don't want
# to rely on a node run-time to prefix after compilation
# https://github.com/twbs/bootstrap/blob/d438f3/package.json#L30
# ----------------------------------------------------------------------
scss_files <- dir(
  "inst",
  pattern = "\\.scss$",
  recursive = TRUE,
  full.names = TRUE
)
# These libs should already have prefixes in their source
# TODO: add test(s) that we aren't missing vendor prefixes
scss_files <- scss_files[
  !grepl(
    "^inst/(lib/bs3|bs3compat|themer|components|bslib-scss|builtin|examples)",
    scss_files
  )
]

scss_src <- lapply(scss_files, readLines)

add_property_prefixes <- function(
  src,
  property,
  ok_values = NULL,
  vendors = c("-webkit-", "-moz-", "-ms-", "-o-")
) {
  pattern <- paste0("^\\s*", property, ":\\s*(.+);")
  idx <- grep(pattern, src)
  for (i in idx) {
    prop <- src[[i]]
    if (length(ok_values)) {
      value <- regmatches(prop, regexec(pattern, prop))[[1]][2]
      vals <- strsplit(value, "\\s+")[[1]]
      if (all(vals %in% c(ok_values, "!important"))) next
    }
    leading_ws <- regmatches(prop, regexpr("^\\s+", prop))
    src[[i]] <- paste0(
      leading_ws,
      c("", vendors),
      sub("^\\s+", "", prop),
      collapse = "\n"
    )
  }
  src
}

# Unconditionally prefix the following CSS properties for all vendors
needs_prefix <- c(
  "appearance",
  "user-select",
  "backdrop-filter",
  "backface-visibility",
  "touch-action",
  "animation-duration"
)
for (prop in needs_prefix) {
  scss_src <- lapply(scss_src, add_property_prefixes, prop)
}

# Only add webkit prefix for BS5+ since other vendors aren't really relevant anymore
for (prop in c('mask-image', 'mask-size', 'mask-position')) {
  scss_src <- lapply(
    scss_src,
    add_property_prefixes,
    prop,
    vendors = "-webkit-"
  )
}

# Print specific vendor prefixes
scss_src <- lapply(
  scss_src,
  add_property_prefixes,
  "color-adjust",
  vendors = "-webkit-print-"
)

# phantomjs 2.1.1 needs webkit vendor prefix on flex properties to work correctly
flex_props <- c(
  "flex-direction",
  "flex-wrap",
  "flex-flow",
  "justify-content",
  "align-items",
  "align-content",
  "order",
  "flex-grow",
  "flex-shrink",
  "flex-basis",
  "flex",
  "align-self"
)
for (prop in flex_props) {
  scss_src <- lapply(
    scss_src,
    add_property_prefixes,
    prop,
    vendors = "-webkit-"
  )
}

# Conditionally prefix text-decoration if its not CSS2 compliant
# https://www.w3.org/TR/CSS2/text.html#lining-striking-props
# https://developer.mozilla.org/en-US/docs/Web/CSS/text-decoration
# https://caniuse.com/#search=text-decoration
scss_src <- lapply(
  scss_src,
  add_property_prefixes,
  "text-decoration",
  ok_values = c(
    "none",
    "underline",
    "overline",
    "line-through",
    "blink",
    "inherit"
  )
)

add_value_prefixes <- function(
  src,
  value,
  vendors = c("-webkit-", "-moz-", "-ms-", "-o-")
) {
  pattern <- paste0("^\\s*[^/]+:\\s*", value, "\\s*(!important)?\\s*;")
  idx <- grep(pattern, src)
  for (i in idx) {
    prop_val <- strsplit(src[[i]], ":\\s*")[[1]]
    src[[i]] <- paste0(
      prop_val[1],
      ": ",
      c("", vendors),
      prop_val[2],
      collapse = "\n"
    )
  }
  src
}

# https://caniuse.com/?search=min-content
scss_src <- lapply(scss_src, add_value_prefixes, "min-content")
scss_src <- lapply(scss_src, add_value_prefixes, "max-content")

# phantomjs 2.1.1 needs `display: -webkit-flex` to work properly
scss_src <- lapply(scss_src, add_value_prefixes, "flex", vendors = "-webkit-")

# Write modified source to disk
invisible(Map(writeLines, scss_src, scss_files))

# ----------------------------------------------------------------------
# Check to make sure we're not missing any vendor prefixes
# that we don't already know about in the distributed CSS
# ----------------------------------------------------------------------

find_prefixed_css <- function(css) {
  vendors <- c("webkit", "moz", "ms")
  prefixes <- lapply(vendors, function(vendor) {
    pattern <- sprintf("-%s-([^:|;| |\\|,)]+)", vendor)
    prefixes <- regmatches(css, regexec(pattern, css))
    lapply(prefixes, function(x) if (length(x) > 1) x[2] else NULL)
  })
  unique(unlist(prefixes, recursive = TRUE))
}

# TODO: do for each bootstrap?
src_prefixes <- find_prefixed_css(unlist(scss_src))
dist_prefixes <- find_prefixed_css(
  c(
    readLines("inst/lib/bs4/dist/css/bootstrap.css"),
    readLines("inst/lib/bs5/dist/css/bootstrap.css")
  )
)
auto_prefixes <- setdiff(dist_prefixes, src_prefixes)

whitelist <- c(
  # https://caniuse.com/#feat=flexbox
  "flex",
  "inline-flex",
  "inline-flexbox",
  # https://caniuse.com/#feat=mdn-api_csskeyframesrule
  "keyframes",
  # https://caniuse.com/#feat=mdn-css_properties_transition
  "transition",
  # https://caniuse.com/#feat=css-animation
  "animation",
  # https://caniuse.com/#feat=transforms2d
  "transform",
  # https://caniuse.com/#feat=mdn-css_properties_column-count
  "column-count",
  # https://caniuse.com/#feat=mdn-css_properties_column-gap_multicol_context
  "column-gap",
  # https://caniuse.com/#feat=css-placeholder
  "placeholder",
  "input-placeholder",
  # https://developer.mozilla.org/en-US/docs/Web/CSS/text-decoration-skip-ink
  "text-decoration-skip-ink",
  # webkit prefix doesn't seem necessary? And Bootstrap docs warn about IE...
  # https://caniuse.com/#search=sticky
  "sticky",
  # Already applied conditional prefixing
  "text-decoration",
  "text-decoration-color",
  # IE11 technically needs a prefix, but this is just for floating labels
  # and I'm lazy https://caniuse.com/?search=placeholder-shown
  "placeholder-shown",
  # False positive?
  "margin-end"
)

unknown_prefixes <- setdiff(
  # whitelist flexbox props (there are many)
  auto_prefixes[!grepl("^flex-*", auto_prefixes)],
  whitelist
)
if (length(unknown_prefixes)) {
  stop(
    "Unknown vendor prefixes introduced by Bootstrap's autoprefixer. ",
    "Use either add_property_prefixes() to add prefixes or whitelist them ",
    "(if they're not needed for modern browsers): ",
    "'",
    paste(collapse = "', '", unknown_prefixes),
    "'.",
    call. = FALSE
  )
}

# ----------------------------------------------------------------------
# Now, get rid of files that we don't need to bundle with the package
# ----------------------------------------------------------------------

with_dir("inst/lib", {
  # Downsize Bootstrap
  for (bs in c("bs5", "bs4")) {
    unlink(file.path(bs, "dist/css"), recursive = TRUE)
    unlink(file.path(bs, "js"), recursive = TRUE)
    js_dist <- file.path(bs, "dist/js")
    non_bundle <- setdiff(
      dir(js_dist),
      c("bootstrap.bundle.min.js", "bootstrap.bundle.min.js.map")
    )
    file.remove(file.path(js_dist, non_bundle))
  }

  # Only keep Bootswatch's Sass source files
  for (bsw in c("bsw5", "bsw4")) {
    file.remove(Sys.glob(file.path(bsw, "dist/*/*.css")))
    file.remove(Sys.glob(file.path(bsw, "dist/*/*.map")))
    unlink("bsw5/docs", recursive = TRUE)
    # Remove hidden files
    unlink(
      dir(bsw, pattern = "^\\.[a-z]+", all.files = TRUE, full.names = TRUE),
      recursive = TRUE
    )
  }
  file.remove(
    c(
      Sys.glob("bsw3/*/bootstrap.css"),
      Sys.glob("bsw3/*/bootstrap.min.css"),
      Sys.glob("bsw3/*/thumbnail.png"),
      Sys.glob("bsw3/*/*.less")
    )
  )
  unlink("bsw3/docs", recursive = TRUE)
  unlink("bsw3/.github", recursive = TRUE)
  unlink("bsw3/fonts", recursive = TRUE) # have fonts via tools/download_fonts.R

  # Downsize bootstrap-accessibility
  with_dir("bs-a11y-p", {
    discard <- setdiff(dir(), c("src", "plugins", "LICENSE.md", "package.json"))
    unlink(discard, recursive = TRUE)
  })

  # Downsize bootstrap-colorpicker
  with_dir("bs-colorpicker", {
    file.rename("dist/css", "css")
    file.rename("dist/js", "js")
    # For the sake of simplicity, a patch is applied to just the
    # non-minified version (and the minified version isn't used)
    unlink(Sys.glob("js/*.min.js"))
    unlink("node_modules", recursive = TRUE)
    unlink("dist", recursive = TRUE)
    unlink("src", recursive = TRUE)
    unlink("logo.png")
  })

  # ----------------------------------------------------------------------
  # Copy prism-code-editor files (selective copy from dist/)
  # The full package is installed as prism-code-editor-full, and we copy
  # only the needed files to prism-code-editor, then remove the full package
  # ----------------------------------------------------------------------
  cat("\n\n==== Processing prism-code-editor ====\n")
  unlink("prism-code-editor", recursive = TRUE)

  src <- "prism-code-editor-full/dist"
  dest <- "prism-code-editor"

  # Languages to bundle locally for syntax highlighting
  code_editor_bundled_languages <<- c(
    # Data Science
    "r",
    "python",
    "julia",
    "sql",
    # Web/Frontend
    "javascript",
    "typescript",
    "markup",
    "css",
    "scss",
    "sass",
    "json",
    # Markup/Config
    "markdown",
    "yaml",
    "xml",
    "toml",
    "ini",
    # Shell/Infrastructure
    "bash",
    "docker",
    # Extras
    "latex",
    "cpp",
    "rust",
    "diff"
  )

  # Create destination directories
  dir.create(dest, recursive = TRUE)
  dir.create(file.path(dest, "utils"), recursive = TRUE)

  dir.create(file.path(dest, "languages"), recursive = TRUE)
  dir.create(file.path(dest, "prism", "languages"), recursive = TRUE)
  dir.create(file.path(dest, "extensions", "copyButton"), recursive = TRUE)
  dir.create(file.path(dest, "themes"), recursive = TRUE)

  # Copy core files (*.js and *.css from dist root)
  # Exclude theme JS files - we use the CSS files in themes/ instead.
  # Theme JS files match pattern: <theme-name>.js or <theme-name>-<hash>.js
  # fmt: skip
  theme_names <- sub("\\.css$", "", basename(Sys.glob(file.path(src, "themes", "*.css"))))
  # fmt: skip
  theme_js_pattern <- sprintf("^(%s)(-[^.]+)?\\.js$", paste(theme_names, collapse = "|"))
  core_files <- c(
    Sys.glob(file.path(src, "*.js")),
    Sys.glob(file.path(src, "*.css"))
  )
  core_files <- core_files[!grepl(theme_js_pattern, basename(core_files))]
  file.copy(core_files, dest)

  # Copy utils
  file.copy(
    Sys.glob(file.path(src, "utils", "*.js")),
    file.path(dest, "utils")
  )

  # Copy languages (only bundled)
  bundled_lang_files <- paste0(code_editor_bundled_languages, ".js")
  bundled_languages <- file.path(src, "languages", bundled_lang_files)
  bundled_languages <- bundled_languages[file.exists(bundled_languages)]
  file.copy(
    bundled_languages,
    file.path(dest, "languages")
  )

  # Copy prism grammars (only bundled)
  file.copy(
    file.path(src, "prism", "languages", bundled_lang_files),
    file.path(dest, "prism", "languages")
  )

  # Copy extensions
  ext_files <- list.files(
    file.path(src, "extensions"),
    full.names = TRUE,
    recursive = FALSE
  )
  # Copy top-level extension files
  ext_js <- ext_files[grepl("\\.js$", ext_files)]
  file.copy(ext_js, file.path(dest, "extensions"))
  # Copy copyButton extension folder
  file.copy(
    Sys.glob(file.path(src, "extensions", "copyButton", "*")),
    file.path(dest, "extensions", "copyButton")
  )

  # ============================================================================
  # Remove unused CSS files
  # ============================================================================
  unused_css <- c(
    "autocomplete.css",
    "autocomplete-icons.css",
    "invisibles.css",
    "code-block.css",
    "folding.css",
    "guides.css",
    "search.css"
  )
  for (css in unused_css) {
    css_path <- file.path(dest, css)
    if (file.exists(css_path)) {
      file.remove(css_path)
      cat("Removed unused CSS:", css, "\n")
    }
  }

  # Keep scrollbar.css and rtl-layout.css (don't remove)

  # ============================================================================
  # Remove unused extensions
  # ============================================================================
  unused_extensions <- c(
    "guides.js",
    "matchTags.js"
  )
  for (ext in unused_extensions) {
    ext_path <- file.path(dest, "extensions", ext)
    if (file.exists(ext_path)) {
      file.remove(ext_path)
      cat("Removed unused extension:", ext, "\n")
    }
  }

  # Keep cursor.js (required by tooltips)

  # ============================================================================
  # Remove unused chunks
  # ============================================================================
  # Remove chunks matching these patterns
  chunk_patterns <- c(
    "basic-*.js",
    "readonly-*.js",
    "search-*.js",
    "selection-*.js",
    "styles-*.js"
  )

  for (pattern in chunk_patterns) {
    chunks <- Sys.glob(file.path(dest, pattern))
    if (length(chunks) > 0) {
      file.remove(chunks)
      cat("Removed", length(chunks), "chunk(s) matching", pattern, "\n")
    }
  }

  # Remove webComponent.js specifically
  webcomp <- file.path(dest, "webComponent.js")
  if (file.exists(webcomp)) {
    file.remove(webcomp)
    cat("Removed webComponent.js\n")
  }

  # ============================================================================
  # Remove development files (TypeScript definitions and source maps)
  # ============================================================================
  ts_defs <- Sys.glob(file.path(dest, "**", "*.d.ts"))
  js_maps <- Sys.glob(file.path(dest, "**", "*.js.map"))
  if (length(ts_defs) > 0) {
    file.remove(ts_defs)
  }
  if (length(js_maps) > 0) {
    file.remove(js_maps)
  }
  cat("Removed", length(ts_defs), "TypeScript definition files\n")
  cat("Removed", length(js_maps), "source map files\n")

  # Copy themes
  file.copy(
    Sys.glob(file.path(src, "themes", "*.css")),
    file.path(dest, "themes")
  )

  # Scope theme CSS files to support multiple editors with different themes.
  # Each theme is wrapped with attribute selectors that match the editor's
  # theme-light/theme-dark attributes, combined with the page's data-bs-theme
  # attribute, using CSS nesting (supported in modern browsers).
  theme_files <- Sys.glob(file.path(dest, "themes", "*.css"))
  for (theme_file in theme_files) {
    theme_name <- sub("\\.css$", "", basename(theme_file))
    css_content <- readLines(theme_file)

    # Wrap with scoped selectors using CSS nesting
    scoped_css <- c(
      sprintf(
        "html:not([data-bs-theme]) [theme-light=\"%s\"],[data-bs-theme=\"light\"] [theme-light=\"%s\"], [data-bs-theme=\"light\"][theme-light=\"%s\"],",
        theme_name,
        theme_name,
        theme_name
      ),
      sprintf(
        "[data-bs-theme=\"dark\"] [theme-dark=\"%s\"], [data-bs-theme=\"dark\"][theme-dark=\"%s\"] {",
        theme_name,
        theme_name
      ),
      css_content,
      "}"
    )

    writeLines(scoped_css, theme_file)
  }

  # Get version for tracking
  version_prism_code_editor <<- jsonlite::fromJSON(
    "prism-code-editor-full/package.json"
  )$version

  # Remove the full package now that we've copied what we need

  unlink("prism-code-editor-full", recursive = TRUE)
  cat("\n\n==== Finished prism-code-editor ====\n")

  # GitHub reports security issues of devDependencies, but that's irrelevant to us
  remove_dev_dependencies <- function(pkg_file) {
    if (!file.exists(pkg_file)) {
      return()
    }
    json <- jsonlite::fromJSON(pkg_file)
    json <- json[setdiff(names(json), "devDependencies")]
    jsonlite::write_json(json, pkg_file, pretty = TRUE, auto_unbox = TRUE)
  }
  invisible(lapply(Sys.glob("*/package.json"), remove_dev_dependencies))

  # Get BS4/BS3 versions (for bslib::bs_dependencies() version-ing)
  version_bs5 <- sub(
    "-beta[0-9]+",
    "",
    jsonlite::fromJSON("bs5/package.json")$version
  )
  version_bs4 <- jsonlite::fromJSON("bs4/package.json")$version
  version_bs3 <- jsonlite::fromJSON("bs3/package.json")$version
  version_accessibility <- jsonlite::fromJSON("bs-a11y-p/package.json")$version
})

writeLines(
  c(
    '# Generated by tools/yarn_install.R: do not edit by hand',
    paste0('version_bs5 <- ', deparse(version_bs5)),
    paste0('version_bs4 <- ', deparse(version_bs4)),
    paste0('version_bs3 <- ', deparse(version_bs3)),
    paste0('version_accessibility <- ', deparse(version_accessibility)),
    paste0('version_prism_code_editor <- ', deparse(version_prism_code_editor)),
    paste0(
      'code_editor_bundled_languages <- ',
      deparse1(code_editor_bundled_languages)
    )
  ),
  "R/versions.R"
)

# ----------------------------------------------------------------------
# Apply any patches to source
# ----------------------------------------------------------------------

patch_files <- list.files(
  find_package_root_file("tools/patches"),
  full.names = TRUE
)

rej_pre <- dir(pattern = "\\.rej$", recursive = TRUE)
for (patch in patch_files) {
  message(sprintf("Applying %s", basename(patch)))
  res <- system(sprintf("git apply --reject --whitespace=fix '%s'", patch))
  if (res > 0) stop("Couldn't successfully apply patch: ", patch, call. = FALSE)
}
rej_post <- dir(pattern = "\\.rej$", recursive = TRUE)
if (length(rej_post) > length(rej_pre)) {
  warning(
    "Running `git apply --reject` generated `.rej` files. \n",
    "An 'easy' way to do this is to first `git add` the new source changes, ",
    "then manually make the relevant changes from the patch file,",
    "then `git diff` to get the relevant diff output and update the patch diff with the new diff."
  )
}

# Tracking changes in solar isn't so important since we've basically
# re-implemented it using a proper color system
# TODO: do the same for BS5?
writeLines(
  sass::as_sass(
    list(
      "white" = "#092B36 !default",
      "gray-100" = "#173741 !default",
      "gray-200" = "#25434B !default",
      "gray-300" = "#324E56 !default",
      "gray-400" = "#405A61 !default",
      "gray-500" = "#4E666B !default",
      "gray-600" = "#5C7276 !default",
      "gray-700" = "#6A7E81 !default",
      "gray-800" = "#77898C !default",
      "gray-900" = "#859596 !default",
      "black" = "#93A1A1 !default",
      "blue" = "#b58900 !default",
      "indigo" = "#6610f2 !default",
      "purple" = "#6f42c1 !default",
      "pink" = "#e83e8c !default",
      "red" = "#d33682 !default",
      "orange" = "#fd7e14 !default",
      "yellow" = "#cb4b16 !default",
      "green" = "#2aa198 !default",
      "teal" = "#20c997 !default",
      "cyan" = "#268bd2 !default",
      "enable-gradients" = "true !default",
      "secondary" = "$gray-800 !default",
      # Darker green and lighter green than `$black` and `$white`
      "color-contrast-dark" = "#031014 !default",
      "color-contrast-light" = "#BBD0D0 !default"
    )
  ),
  "inst/lib/bsw4/dist/solar/_variables.scss"
)
writeLines(
  c(
    '$web-font-path: "https://fonts.googleapis.com/css2?family=Source+Sans+Pro:wght@400;600;700&display=swap" !default;',
    '@import url($web-font-path);'
  ),
  "inst/lib/bsw4/dist/solar/_bootswatch.scss"
)

# ----------------------------------------------------------------------
# Apply minification to patched files
# ----------------------------------------------------------------------

with_dir("inst", {
  local({
    # install all deps
    system("yarn install")
    # remove node modules
    on.exit(
      {
        unlink("node_modules", recursive = TRUE)
      },
      add = TRUE
    )

    bslib_plugin_paths <- file.path(
      "lib",
      "bs-a11y-p",
      "plugins",
      "js",
      c(
        "bootstrap-accessibility.js"
      )
    )

    for (unminified_file in c(
      bslib_plugin_paths
    )) {
      message("Minifying ", basename(unminified_file))
      cmd <- paste0(
        "yarn parcel",
        " build ",
        unminified_file,
        " --no-source-maps",
        " --no-cache",
        " --out-dir ",
        dirname(unminified_file),
        " --out-file ",
        sub(".js", ".min.js", fixed = TRUE, basename(unminified_file))
      )

      system(cmd)
    }
  })
})

# Clean up
unlink("inst/yarn.lock")
unlink("inst/node_modules", recursive = TRUE)

# ----------------------------------------------------------------------
# Precompile Bootstrap CSS
# ----------------------------------------------------------------------

# This generates precompiled builds of Bootstrap's css. It would be nice to do
# it at binary package build time, but I couldn't get that to work, using either
# src/install.libs.R (because the bslib functions used in this script
# aren't available yet), or by putting this code directly in the R/ directory
# (because the R/ files are evaluated only after the inst directory is copied
# over).
library(bslib)

precompiled_dir <- find_package_root_file("inst/css-precompiled")
unlink(precompiled_dir, recursive = TRUE)
dir.create(precompiled_dir, recursive = TRUE)

invisible(
  lapply(versions(), function(version) {
    res <- bs_theme_dependencies(
      bs_theme(version, brand = FALSE),
      precompiled = FALSE,
      sass_options = sass_options(output_style = "compressed"),
      cache = NULL
    )
    # Extract the Bootstrap dependency object (as opposed to, say, jQuery)
    bs_dep <- Filter(res, f = function(x) {
      identical(x$name, "bootstrap")
    })[[1]]

    tmp_css <- file.path(bs_dep$src$file, bs_dep$stylesheet)
    dest_dir <- file.path(precompiled_dir, version)
    if (!dir.exists(dest_dir)) {
      dir.create(dest_dir)
    }
    file.copy(tmp_css, dest_dir)

    # Also save the BS5+ Sass code used to generate the pre-compiled CSS.
    # This is primarily here to help Quarto more easily replicate bs_theme()'s Sass.
    if (version >= 5) {
      theme_sass <- gsub(
        paste0("@import \"", getwd(), "/"),
        "@import \"",
        as_sass(bs_theme(version))
      )
      writeLines(theme_sass, file.path(dest_dir, "bootstrap.scss"))
      # Sanity check that we we can compile by moving file to home dir
      file.copy(file.path(dest_dir, "bootstrap.scss"), "bootstrap.scss")
      on.exit(unlink("bootstrap.scss"), add = TRUE)
      testthat::expect_error(sass(sass_file("bootstrap.scss")), NA)
    }
  })
)

random_values <- tibble::tribble(
  ~title,
  ~value,
  "Sales revenue",
  "$22,456.78",
  "Customer satisfaction",
  "94.5%",
  "Inventory turnover",
  "8/month",
  "Number of employees",
  "75",
  "Website traffic",
  "12,345 visits/day",
  "Customer retention",
  "87%",
  "Average transaction",
  "$56.73",
  "Followers",
  "10,234",
  "Open rate",
  "22.6%",
  "Production efficiency",
  "92%",
  "Project completion",
  "17 days",
  "Employee turnover",
  "10%",
  "Market share",
  "15.2%",
  "ROI",
  "8.3%",
  "Average response time",
  "3.2 seconds",
  "Energy consumption",
  "345 kwh/month",
  "Customer churn",
  "5%",
  "Production defect rate",
  "0.6%",
  "Engagement rate",
  "3.8%",
  "Average wait",
  "4 minutes",
  "Project budget variance",
  "$2,345.67",
  "Employee productivity",
  "95%",
  "Conversion rate",
  "2.5%",
  "Website bounce rate",
  "42%",
  "Revenue growth",
  "+15.2%",
  "Customer loyalty",
  "4.5 out of 5",
  "Product quality",
  "8.9/10",
  "Social media followers",
  "50,000+",
  "Market reach",
  "10.5 million",
  "Employee satisfaction",
  "92.3%",
  "Average order value",
  "$123.45",
  "Website conversion rate",
  "3.6%",
  "Customer support",
  "9.8/10",
  "Productivity index",
  "120%",
  "Customer lifetime value",
  "$5,000",
  "Brand awareness",
  "85%",
  "Time to market",
  "4 weeks",
  "Return on investment",
  "18.5%",
  "Net promoter score",
  "8.2 out of 10",
  "Email click-through rate",
  "12.7%",
  "Supply chain efficiency",
  "92%",
  "Website loading time",
  "2.3 seconds",
  "Cost per acquisition",
  "$25.60",
  "Employee engagement",
  "83%",
  "Innovation index",
  "9.5/10",
  "Customer complaints resolved",
  "98.6%",
  "Market share growth",
  "+2.3%",
  "Production capacity utilization",
  "87%",
  "Website traffic source diversity",
  "5 channels",
  "Brand equity",
  "$1.2 billion",
  "Code Coverage",
  "85%",
  "Lines of Code",
  "46,157",
  "Commits",
  "339",
  "Review Turnaround",
  "2 days",
  "Community Contributors",
  "34",
  "Project Forks",
  "36",
  "Open Issues",
  "87",
  "Open Pull Requests",
  "12",
  "Closed Issues",
  "19",
  "AWS Cloud Spending",
  "$3,463",
  "Diversity Index",
  "0.75",
  "API Response Time",
  "150 ms",
  "Daily Active Users",
  "8,507",
  "Project Stars",
  "5,100",
  "Package Downloads",
  "409,446",
  "Average Response Time",
  "50 ms",
  "Uptime",
  "99.97%",
  "Developer Satisfaction",
  "4.8/5",
  "Project Funding",
  "$2,805"
)

random_title_value <- function() {
  as.list(random_values[sample(nrow(random_values), 1), ])
}

random_values <- tibble::tribble(
  ~title, ~value,
  "Sales revenue", "$22,456.78",
  "Customer satisfaction", "94.5%",
  "Inventory turnover", "8/month",
  "Number of employees", "75",
  "Website traffic", "12,345 visits/day",
  "Customer retention", "87%",
  "Average transaction", "$56.73",
  "Followers", "10,234",
  "Open rate", "22.6%",
  "Production efficiency", "92%",
  "Project completion", "17 days",
  "Employee turnover", "10%",
  "Market share", "15.2%",
  "ROI", "8.3%",
  "Average response time", "3.2 seconds",
  "Energy consumption", "345 kwh/month",
  "Customer churn", "5%",
  "Production defect rate", "0.6%",
  "Engagement rate", "3.8%",
  "Average wait", "4 minutes",
  "Project budget variance", "$2,345.67",
  "Employee productivity", "95%",
  "Conversion rate", "2.5%",
  "Website bounce rate", "42%",
  "Revenue growth", "+15.2%",
  "Customer loyalty", "4.5 out of 5",
  "Product quality", "8.9/10",
  "Social media followers", "50,000+",
  "Market reach", "10.5 million",
  "Employee satisfaction", "92.3%",
  "Average order value", "$123.45",
  "Website conversion rate", "3.6%",
  "Customer support", "9.8/10",
  "Productivity index", "120%",
  "Customer lifetime value", "$5,000",
  "Brand awareness", "85%",
  "Time to market", "4 weeks",
  "Return on investment", "18.5%",
  "Net promoter score", "8.2 out of 10",
  "Email click-through rate", "12.7%",
  "Supply chain efficiency", "92%",
  "Website loading time", "2.3 seconds",
  "Cost per acquisition", "$25.60",
  "Employee engagement", "83%",
  "Innovation index", "9.5/10",
  "Customer complaints resolved", "98.6%",
  "Market share growth", "+2.3%",
  "Production capacity utilization", "87%",
  "Website traffic source diversity", "5 channels",
  "Brand equity", "$1.2 billion",
  "Code Coverage", "85%",
  "Lines of Code", "46,157",
  "Commits", "339",
  "Review Turnaround", "2 days",
  "Community Contributors", "34",
  "Project Forks", "36",
  "Open Issues", "87",
  "Open Pull Requests", "12",
  "Closed Issues", "19",
  "AWS Cloud Spending", "$3,463",
  "Diversity Index", "0.75",
  "API Response Time", "150 ms",
  "Daily Active Users", "8,507",
  "Project Stars", "5,100",
  "Package Downloads", "409,446",
  "Average Response Time", "50 ms",
  "Uptime", "99.97%",
  "Developer Satisfaction", "4.8/5",
  "Project Funding", "$2,805"
)

random_title_value <- function() {
  as.list(random_values[sample(nrow(random_values), 1), ])
}

# @staticimports pkg:staticimports
#  is_installed get_package_version system_file
#  register_upgrade_message
#  %||% is_string
#  read_utf8

switch_version <- function(
  version,
  five = default,
  four = default,
  three = default,
  default = NULL
) {
  if (is_bs_theme(version)) {
    version <- theme_version(version)
  }
  version <- as.character(version)
  if (isTRUE(version %in% c("4-3", "4+3"))) {
    warning("Version ", version, " has been renamed to 4. Please use 4 instead")
    version <- "4"
  }
  switch(
    version,
    `5` = five,
    `4` = four,
    `3` = three,
    stop("Didn't recognize Bootstrap version: ", version, call. = FALSE)
  )
}

get_exact_version <- function(version) {
  switch_version(
    version,
    five = version_bs5,
    four = version_bs4,
    three = version_bs3
  )
}

path_inst <- function(...) {
  files <- system_file(..., package = "bslib")
  files_found <- files != ""
  if (all(files_found)) {
    return(files)
  }

  files_not_found <- file.path(...)[!files_found]
  stop(
    "bslib file not found: '",
    files_not_found,
    "'",
    call. = FALSE
  )
}

path_lib <- function(...) path_inst("lib", ...)

is_shiny_app <- function() {
  # Make sure to not load shiny as a side-effect of calling this function.
  isNamespaceLoaded("shiny") && shiny::isRunning()
}

is_hosted_app <- function() {
  nzchar(Sys.getenv("SHINY_SERVER_VERSION")) && is_shiny_app()
}

is_shiny_runtime <- function() {
  if (!is_installed("knitr")) {
    return(FALSE)
  }
  isTRUE(grepl("^shiny", knitr::opts_knit$get("rmarkdown.runtime")))
}

register_runtime_package_check <- function(feature, pkg, version) {
  msg <- sprintf(
    "%s is designed to work with %s %s or higher",
    feature,
    pkg,
    version
  )

  if (isNamespaceLoaded(pkg) && !is_installed(pkg, version)) {
    warning(msg, call. = FALSE)
  }

  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      if (!is_installed(pkg, version)) warning(msg, call. = FALSE)
    }
  )
}

add_class <- function(x, y) {
  class(x) <- unique(c(y, oldClass(x)))
  x
}

dropNulls <- function(x) {
  x[!vapply(x, is.null, FUN.VALUE = logical(1))]
}

names2 <- function(x) {
  names(x) %||% rep.int("", length(x))
}

any_unnamed <- function(x) {
  if (length(x) == 0) {
    return(FALSE)
  }
  nms <- names(x)
  is.null(nms) || !all(nzchar(nms))
}

separate_arguments <- function(...) {
  x <- rlang::list2(...)
  x_names <- rlang::names2(x)
  is_named <- nzchar(x_names)

  if (all(is_named)) {
    return(list(attribs = x, children = list()))
  }

  if (!any(is_named)) {
    return(list(attribs = list(), children = x))
  }

  list(attribs = x[is_named], children = unname(dropNulls(x[!is_named])))
}

#' Rename a named list
#'
#' @param x a named list to be renamed
#' @param nms a named character vector defining the renaming
#' @noRd
#' @examples
#' rename2(list(a=1, b=3, c=4, a=2), b="z", f="w", a="y")
#' #> list(y = 1, z = 3, c = 4, y = 2)
#' rename2(c("a", "b", "c", "a"), b="z", f="w", a="y")
#' #> c("y", "z", "c", "y")
rename2 <- function(x, ...) {
  defs <- rlang::list2(...)
  nms <- names(x) %||% as.character(x)
  matches <- intersect(nms, names(defs))
  map <- match(nms, names(defs))
  mapidxNA <- is.na(map)
  replacement <- as.character(defs)[map[!mapidxNA]]
  if (is.null(names(x))) {
    x[!mapidxNA] <- replacement
  } else {
    names(x)[!mapidxNA] <- replacement
  }
  x
}

# Get an accessible color contrast for a specified bg_color
# (and return NULL+warn on failure)
get_color_contrast <- function(bg_color) {
  utils_layer <- sass::sass_layer(
    functions = sass::sass_file(
      path_inst("sass-utils", "color-contrast.scss")
    ),
    defaults = c(
      "$black: #000000;",
      "$white: #FFFFFF;"
    ),
    rules = sprintf(
      "._ {--RET: #{color-contrast(%s)}}",
      bg_color
    )
  )

  tryCatch(
    {
      css <- sass::sass(
        utils_layer,
        cache_key_extra = get_package_version("bslib"),
        # Don't listen to global Sass options so we can be sure
        # that stuff like source maps won't be included
        options = sass::sass_options(source_map_embed = FALSE)
      )
      # example: css <- "._ {\n  --RET: #fff;\n}"
      # we'll split to get value:     ^     ^
      ret <- strsplit(css, "--RET:")[[1]][2]
      trimws(strsplit(ret, ";")[[1]][1])
    },
    error = function(err) {
      warning(
        "Failed to compute a contrasting color for '",
        bg_color,
        "'",
        call. = FALSE
      )
      NULL
    }
  )
}

#' Sidebar layouts
#'
#' @description Create a collapsing sidebar layout by providing a `sidebar()`
#'   object to the `sidebar` argument of:
#'
#' * [page_sidebar()]
#'     * Creates a "page-level" sidebar.
#' * [page_navbar()]
#'     * Creates a multi-page app with a "page-level" sidebar.
#'   * Creates a multi page/tab UI with a singular `sidebar()` (which is
#'     shown on every page/tab).
#' * `layout_sidebar()`
#'   * Creates a "floating" sidebar layout component which can be dropped
#'     inside any [page()] and/or [card()] context.
#' * [navset_card_tab()] and [navset_card_pill()]
#'   * Creates a multi-tab card with a sidebar inside of it.
#'
#' See [this article](https://rstudio.github.io/bslib/articles/sidebars.html)
#'   to learn more.
#'
#' @param ... Unnamed arguments can be any valid child of an [htmltools
#'   tag][htmltools::tags] and named arguments become HTML attributes on
#'   returned UI element. In the case of `layout_sidebar()`, these arguments are
#'   passed to the main content tag (not the sidebar+main content container).
#' @param width A valid [CSS unit][htmltools::validateCssUnit] used for the
#'   width of the sidebar.
#' @param position Where the sidebar should appear relative to the main content.
#' @param open The initial state of the sidebar, choosing from the following
#'   options:
#'
#'   * `"desktop"`: The sidebar starts open on desktop screen, closed on mobile.
#'     This is default sidebar behavior.
#'   * `"open"` or `TRUE`: The sidebar starts open.
#'   * `"closed"` or `FALSE`: The sidebar starts closed.
#'   * `"always"` or `NA`: The sidebar is always open and cannot be closed.
#'
#'   In `sidebar_toggle()`, `open` indicates the desired state of the sidebar,
#'   where the default of `open = NULL` will cause the sidebar to be toggled
#'   open if closed or vice versa. Note that `sidebar_toggle()` can only open or
#'   close the sidebar, so it does not support the `"desktop"` and `"always"`
#'   options.
#' @param id A character string. Required if wanting to re-actively read (or
#'   update) the `collapsible` state in a Shiny app.
#' @param title A character title to be used as the sidebar title, which will be
#'   wrapped in a `<div>` element with class `sidebar-title`. You can also
#'   provide a custom [htmltools::tag()] for the title element, in which case
#'   you'll likely want to give this element `class = "sidebar-title"`.
#' @param bg,fg A background or foreground color. If only one of either is
#'   provided, an accessible contrasting color is provided for the opposite
#'   color, e.g. setting `bg` chooses an appropriate `fg` color.
#' @param class CSS classes for the sidebar container element, in addition to
#'   the fixed `.sidebar` class.
#' @param max_height_mobile The maximum height of the horizontal sidebar when
#'   viewed on mobile devices. The default is `250px` unless the sidebar is
#'   included in a [layout_sidebar()] with a specified height, in which case
#'   the default is to take up no more than 50% of the layout container.
#' @param gap A [CSS length unit][htmltools::validateCssUnit()] defining the
#'   vertical `gap` (i.e., spacing) between adjacent elements provided to `...`.
#' @param padding Padding within the sidebar itself. This can be a numeric
#'   vector (which will be interpreted as pixels) or a character vector with
#'   valid CSS lengths. `padding` may be one to four values. If one, then
#'   that value will be used for all four sides. If two, then the first value
#'   will be used for the top and bottom, while the second value will be used
#'   for left and right. If three, then the first will be used for top, the
#'   second will be left and right, and the third will be bottom. If four, then
#'   the values will be interpreted as top, right, bottom, and left
#'   respectively.
#'
#' @export
sidebar <- function(
  ...,
  width = 250,
  position = c("left", "right"),
  open = c("desktop", "open", "closed", "always"),
  id = NULL,
  title = NULL,
  bg = NULL,
  fg = NULL,
  class = NULL,
  max_height_mobile = NULL,
  gap = NULL,
  padding = NULL
) {
  if (isTRUE(open)) {
    open <- "open"
  } else if (identical(open, FALSE)) {
    open <- "closed"
  } else if (isTRUE(is.na(open))) {
    open <- "always"
  }

  open <- rlang::arg_match(open)

  if (!is.null(id)) {
    if (length(id) != 1 || is.na(id) || !nzchar(id)) {
      rlang::abort("`id` must be a non-empty, length-1 character string or `NULL`.")
    }

    # create input binding when id is provided by adding input class
    class <- c("bslib-sidebar-input", class)
  }

  if (is.null(id) && open != "always") {
    # always provide id when collapsible for accessibility reasons
    id <- paste0("bslib-sidebar-", p_randomInt(1000, 10000))
  }

  if (is.null(fg) && !is.null(bg)) {
    fg <- get_color_contrast(bg)
  }
  if (is.null(bg) && !is.null(fg)) {
    bg <- get_color_contrast(fg)
  }

  if (rlang::is_bare_character(title) || rlang::is_bare_numeric(title)) {
    title <- div(title, class = "sidebar-title")
  }

  collapse_tag <-
    if (open != "always") {
      tags$button(
        class = "collapse-toggle",
        type = "button",
        title = "Toggle sidebar",
        "aria-expanded" = if (open %in% c("open", "desktop")) "true" else "false",
        "aria-controls" = id,
        collapse_icon()
      )
    }

  res <- list2(
    tag = tags$div(
      id = id,
      role = "complementary",
      class = c("sidebar", class),
      hidden = if (open == "closed") NA,
      tags$div(
        class = "sidebar-content bslib-gap-spacing",
        title,
        style = css(
          gap = validateCssUnit(gap),
          padding = validateCssPadding(padding)
        ),
        ...
      )
    ),
    collapse_tag = collapse_tag,
    position = match.arg(position),
    open = open,
    width = validateCssUnit(width),
    max_height_mobile = validateCssUnit(max_height_mobile),
    color = list(bg = bg, fg = fg)
  )

  class(res) <- c("sidebar", class(res))
  res
}

#' @rdname sidebar
#' @param sidebar A [sidebar()] object.
#' @param fillable Whether or not the `main` content area should be considered a
#'   fillable (i.e., flexbox) container.
#' @param fill Whether or not to allow the layout container to grow/shrink to fit a
#'   fillable container with an opinionated height (e.g., `page_fillable()`).
#' @param border Whether or not to add a border.
#' @param border_radius Whether or not to add a border radius.
#' @param border_color The border color that is applied to the entire layout (if
#'   `border = TRUE`) and the color of the border between the sidebar and the
#'   main content area.
#' @inheritParams card
#' @inheritParams page_fillable
#'
#' @export
layout_sidebar <- function(
  ...,
  sidebar = NULL,
  fillable = TRUE,
  fill = TRUE,
  bg = NULL,
  fg = NULL,
  border = NULL,
  border_radius = NULL,
  border_color = NULL,
  padding = NULL,
  gap = NULL,
  height = NULL
) {

  if (!inherits(sidebar, "sidebar")) {
    sidebar <- sidebar(sidebar)
  }

  if (!(is.null(border) || isTRUE(border) || isFALSE(border))) {
    abort("`border` must be `NULL`, `TRUE`, or `FALSE`")
  }
  if (!(is.null(border_radius) || isTRUE(border_radius) || isFALSE(border_radius))) {
    abort("`border`_radius must be `NULL`, `TRUE`, or `FALSE`")
  }

  # main content area colors, if not provided ----
  if (is.null(fg) && !is.null(bg)) {
    fg <- get_color_contrast(bg)
  }
  if (is.null(bg) && !is.null(fg)) {
    bg <- get_color_contrast(fg)
  }

  main <- div(
    role = "main",
    class = "main",
    class = if (fillable) "bslib-gap-spacing",
    style = css(
      background_color = bg,
      color = fg,
      padding = validateCssPadding(padding),
      gap = validateCssUnit(gap)
    ),
    ...
  )

  main <- bindFillRole(main, container = fillable)

  contents <- list(main, sidebar$tag, sidebar$collapse_tag)

  right <- identical(sidebar$position, "right")

  max_height_mobile <- sidebar$max_height_mobile %||%
    if (is.null(height)) "250px" else "50%"

  sidebar_init <- if (!identical(sidebar$open, "always")) TRUE

  res <- div(
    class = "bslib-sidebar-layout bslib-mb-spacing",
    class = if (right) "sidebar-right",
    class = if (identical(sidebar$open, "closed")) "sidebar-collapsed",
    `data-bslib-sidebar-init` = sidebar_init,
    `data-bslib-sidebar-open` = sidebar$open,
    `data-bslib-sidebar-border` = if (!is.null(border)) tolower(border),
    `data-bslib-sidebar-border-radius` = if (!is.null(border_radius)) tolower(border_radius),
    style = css(
      "--bslib-sidebar-width" = sidebar$width,
      "--bslib-sidebar-bg" = if (!is.null(sidebar$color$bg)) sidebar$color$bg,
      "--bslib-sidebar-fg" = if (!is.null(sidebar$color$fg)) sidebar$color$fg,
      "--bs-card-border-color" = border_color,
      height = validateCssUnit(height),
      "--bslib-sidebar-max-height-mobile" = max_height_mobile
    ),
    !!!contents,
    sidebar_dependency(),
    sidebar_init_js()
  )

  res <- bindFillRole(res, item = fill)

  res <- as.card_item(res)

  as_fragment(
    tag_require(res, version = 5, caller = "layout_sidebar()")
  )
}

#' @describeIn sidebar Toggle a `sidebar()` state during an active Shiny user
#'   session.
#' @param session A Shiny session object (the default should almost always be
#'   used).
#' @export
toggle_sidebar <- function(id, open = NULL, session = get_current_session()) {
  method <-
    if (is.null(open) || identical(open, "toggle")) {
      "toggle"
    } else if (isTRUE(open) || identical(open, "open")) {
      "open"
    } else if (isFALSE(open) || identical(open, "closed")) {
      "close"
    } else if (isTRUE(is.na(open)) || identical(open, "always")) {
      abort('`open = "always"` is not supported by `sidebar_toggle()`.')
    } else if (identical(open, "desktop")) {
      abort('`open = "desktop"` is not supported by `sidebar_toggle()`.')
    } else {
      abort('`open` must be `NULL`, `TRUE` (or "open"), or `FALSE` (or "closed").')
    }

  force(id)
  callback <- function() {
    session$sendInputMessage(id, list(method = method))
  }
  session$onFlush(callback, once = TRUE)
}

#' @describeIn sidebar An alias for [toggle_sidebar()].
#' @export
sidebar_toggle <- toggle_sidebar

collapse_icon <- function() {
  if (!is_installed("bsicons")) {
    icon <- "<svg xmlns=\"http://www.w3.org/2000/svg\" viewBox=\"0 0 16 16\" class=\"bi bi-chevron-down collapse-icon\" style=\"fill:currentColor;\" aria-hidden=\"true\" role=\"img\" ><path fill-rule=\"evenodd\" d=\"M1.646 4.646a.5.5 0 0 1 .708 0L8 10.293l5.646-5.647a.5.5 0 0 1 .708.708l-6 6a.5.5 0 0 1-.708 0l-6-6a.5.5 0 0 1 0-.708z\"></path></svg>"
    return(HTML(icon))
  }
  bsicons::bs_icon("chevron-down", class = "collapse-icon", size = NULL)
}

sidebar_init_js <- function() {
  # Note: if we want to avoid inline `<script>` tags in the future for
  # initialization code, we might be able to do so by turning the sidebar layout
  # container into a web component
  tags$script(
    `data-bslib-sidebar-init` = NA,
    HTML("bslib.Sidebar.initCollapsibleAll()")
  )
}


sidebar_dependency <- function() {
  list(
    component_dependency_js("sidebar"),
    bs_dependency_defer(sidebar_dependency_sass)
  )
}

sidebar_dependency_sass <- function(theme) {
  component_dependency_sass(theme, "sidebar")
}

test_that("sidebar(open =) ensures backwards compatibility of defaults", {
  expect_equal(
    sidebar(open = TRUE, id = "open"),
    sidebar(open = "open", id = "open")
  )

  expect_equal(
    sidebar(open = FALSE, id = "closed"),
    sidebar(open = "closed", id = "closed")
  )

  expect_equal(
    sidebar(open = NA, id = "always"),
    sidebar(open = "always", id = "always")
  )
})

test_that("sidebar(open = 'always') doesn't include the collapse toggle", {
  expect_null(
    sidebar(open = "always")$collapse_tag
  )
})

test_that("sidebar() - assigns input binding class if `id` is provided", {
  my_sidebar <- sidebar(id = "my-id")

  expect_match(
    htmltools::tagGetAttribute(my_sidebar$tag, "class"),
    "bslib-sidebar-input"
  )

  expect_equal(
    htmltools::tagGetAttribute(my_sidebar$tag, "id"),
    "my-id"
  )
})

test_that("sidebar() - assigns a random `id` if collapsible and `id` not provided", {
  sidebar_open <- sidebar(open = "open")
  expect_match(
    htmltools::tagGetAttribute(sidebar_open$tag, "id"),
    "bslib-sidebar-\\d+"
  )

  expect_equal(
    htmltools::tagGetAttribute(sidebar_open$collapse_tag, "aria-controls"),
    htmltools::tagGetAttribute(sidebar_open$tag, "id")
  )

  sidebar_closed <- sidebar(open = "closed")
  expect_match(
    htmltools::tagGetAttribute(sidebar_closed$tag, "id"),
    "bslib-sidebar-\\d+"
  )

  expect_equal(
    htmltools::tagGetAttribute(sidebar_closed$collapse_tag, "aria-controls"),
    htmltools::tagGetAttribute(sidebar_closed$tag, "id")
  )

  sidebar_desktop <- sidebar(open = "desktop")
  expect_match(
    htmltools::tagGetAttribute(sidebar_desktop$tag, "id"),
    "bslib-sidebar-\\d+"
  )

  expect_equal(
    htmltools::tagGetAttribute(sidebar_desktop$collapse_tag, "aria-controls"),
    htmltools::tagGetAttribute(sidebar_desktop$tag, "id")
  )

  expect_null(
    htmltools::tagGetAttribute(sidebar(open = "always")$tag, "id")
  )
})

test_that("sidebar() - sets `aria-expanded` correctly on collapse toggle", {
  expect_equal(
    htmltools::tagGetAttribute(sidebar(open = "open")$collapse_tag, "aria-expanded"),
    "true"
  )

  expect_equal(
    htmltools::tagGetAttribute(sidebar(open = "closed")$collapse_tag, "aria-expanded"),
    "false"
  )

  expect_equal(
    htmltools::tagGetAttribute(sidebar(open = "desktop")$collapse_tag, "aria-expanded"),
    "true"
  )
})

test_that("sidebar() - throws an error for invalid `open` values", {
  expect_error(
    sidebar(open = "bad")
  )

  expect_error(
    sidebar(open = c("open", "closed"))
  )

  expect_error(
    sidebar(open = 1)
  )

  expect_error(
    sidebar(open = "OPEN")
  )

  expect_error(
    sidebar(open = character())
  )
})
test_that("layout_sidebar() - errors with unexpected border, border_radius input values", {
  expect_error(
    layout_sidebar(sidebar(), border = "1px solid blue")
  )
  expect_error(
    layout_sidebar(sidebar(), border = c(TRUE, FALSE))
  )
  expect_error(
    layout_sidebar(sidebar(), border_radius = c(TRUE, FALSE))
  )
  expect_error(
    layout_sidebar(sidebar(), border = c(TRUE, NA))
  )
  expect_error(
    layout_sidebar(sidebar(), border_radius = c(TRUE, NA))
  )
  expect_error(
    layout_sidebar(sidebar(), border = NA)
  )
  expect_error(
    layout_sidebar(sidebar(), border_radius = NA)
  )
})

#' Dynamically update nav containers
#'
#' Functions for dynamically updating nav containers (e.g., select, insert, and
#' remove nav items). These functions require an `id` on the nav container to be
#' specified.
#'
#' @param id a character string used to identify the nav container.
#' @param selected a character string used to identify a particular [nav()] item.
#' @param session a shiny session object (the default should almost always be used).
#' @export
#' @seealso [nav()], [navs_tab()].
#' @examples
#'
#' can_browse <- function() interactive() && require("shiny")
#'
#' # Selecting a tab
#' if (can_browse()) {
#'   shinyApp(
#'     page_fluid(
#'       radioButtons("item", "Choose", c("A", "B")),
#'       navs_hidden(
#'         id = "container",
#'         nav_content("A", "a"),
#'         nav_content("B", "b")
#'       )
#'     ),
#'     function(input, output) {
#'       observe(nav_select("container", input$item))
#'     }
#'   )
#' }
#'
#' # Inserting and removing
#' if (can_browse()) {
#'   ui <- page_fluid(
#'     actionButton("add", "Add 'Dynamic' tab"),
#'     actionButton("remove", "Remove 'Foo' tab"),
#'     navs_tab(
#'       id = "tabs",
#'       nav("Hello", "hello"),
#'       nav("Foo", "foo"),
#'       nav("Bar", "bar tab")
#'     )
#'   )
#'   server <- function(input, output) {
#'     observeEvent(input$add, {
#'       nav_insert(
#'         "tabs", target = "Bar", select = TRUE,
#'         nav("Dynamic", "Dynamically added content")
#'       )
#'     })
#'     observeEvent(input$remove, {
#'       nav_remove("tabs", target = "Foo")
#'     })
#'   }
#'   shinyApp(ui, server)
#' }
#'
nav_select <- function(id, selected = NULL,
                       session = getDefaultReactiveDomain()) {
  shiny::updateTabsetPanel(session, id, selected)
}


#' @param nav a [nav()] item.
#' @param target The `value` of an existing `nav()` item, next to which tab will be added. If removing: the `value` of the `nav()` item that you want to remove.
#' @param position Should `nav` be added before or after the target?
#' @param select Should `nav` be selected upon being inserted?
#' @rdname nav_select
#' @export
nav_insert <- function(id, nav, target = NULL, position = c("after", "before"),
                       select = FALSE, session = getDefaultReactiveDomain()) {

  force(target)
  force(select)
  position <- match.arg(position)
  inputId <- session$ns(id)

  # Barbara -- August 2017
  # Note: until now, the number of tabs in a tabsetPanel (or navbarPage
  # or navlistPanel) was always fixed. So, an easy way to give an id to
  # a tab was simply incrementing a counter. (Just like it was easy to
  # give a random 4-digit number to identify the tabsetPanel). Since we
  # can only know this in the client side, we'll just pass `id` and
  # `tsid` (TabSetID) as dummy values that will be fixed in the JS code.
  item <- buildTabItem("id", "tsid", TRUE, divTag = nav,
                       textFilter = if (is.character(nav)) navbarMenuTextFilter else NULL)

  callback <- function() {
    session$sendInsertTab(
      inputId = inputId,
      liTag = processDeps(item$liTag, session),
      divTag = processDeps(item$divTag, session),
      menuName = NULL,
      target = target,
      position = position,
      select = select)
  }
  session$onFlush(callback, once = TRUE)
}

#' @export
#' @rdname nav_select
nav_remove <- function(id, target, session = getDefaultReactiveDomain()) {
  force(target)
  inputId <- session$ns(id)

  callback <- function() {
    session$sendRemoveTab(
      inputId = inputId,
      target = target
    )
  }
  session$onFlush(callback, once = TRUE)
}

#' @export
#' @rdname nav_select
nav_show <- function(id, target, select = FALSE,
                     session = getDefaultReactiveDomain()) {
  shiny::showTab(id, target, select, session)
}

#' @export
#' @rdname nav_select
nav_hide <- function(id, target,
                     session = getDefaultReactiveDomain()) {
  shiny::hideTab(id, target, session)
}

#' Append or prepend nav item to a dropdown menu
#'
#' Exported for use by [shiny::prependTab()]/[shiny::appendTab()]. These
#' functions have been superseded by [nav_insert()] (i.e.,
#' [shiny::insertTab()]), since it can do everything these functions do (i.e.,
#' add a [nav()] to the start or end of a [nav_menu()]) and more (i.e., insert a
#' [nav()] anywhere inside a nav container).
#'
#' @inheritParams nav_insert
#' @param menu_title The title of a [nav_menu()].
#' @keywords internal
#' @export
nav_prepend <- function(id, nav, menu_title, select = FALSE, session = getDefaultReactiveDomain()) {

  force(select)
  force(menu_title)
  inputId <- session$ns(id)

  item <- buildTabItem("id", "tsid", TRUE, divTag = nav,
                       textFilter = if (is.character(nav)) navbarMenuTextFilter else NULL)

  callback <- function() {
    session$sendInsertTab(
      inputId = inputId,
      liTag = processDeps(item$liTag, session),
      divTag = processDeps(item$divTag, session),
      menuName = menu_title,
      target = NULL,
      position = "before",
      select = select)
  }
  session$onFlush(callback, once = TRUE)
}

#' @rdname nav_prepend
#' @keywords internal
#' @export
nav_append <- function(id, nav, menu_title, select = FALSE, session = getDefaultReactiveDomain()) {

  force(select)
  force(menu_title)
  inputId <- session$ns(id)

  item <- buildTabItem("id", "tsid", TRUE, divTag = nav,
                       textFilter = if (is.character(nav)) navbarMenuTextFilter else NULL)

  callback <- function() {
    session$sendInsertTab(
      inputId = inputId,
      liTag = processDeps(item$liTag, session),
      divTag = processDeps(item$divTag, session),
      menuName = menu_title,
      target = NULL,
      position = "after",
      select = select)
  }
  session$onFlush(callback, once = TRUE)
}

# This script generates code for replacing hardcoded colors for Bootstrap 3.
# Unlike Bootstrap 4, the Bootstrap 3 _variables.scss has a lot of repeated
# color literals. We want to replace all of the ones that are monochrome with
# dynamically calculated replacements.

# Load bootstrap-sass (bs3) _variables.scss
lines <- readLines(system.file(package = "bslib",
  "lib/bs3/assets/stylesheets/bootstrap/_variables.scss"))
# Retain lines that seem like they contain a hex color.
# (Note that colors specified by keyword would break this logic, but, AFAICT,
# there aren't any such colors in _variables.scss.)
lines <- lines[grepl("#[0-9a-f-A-F]", lines, perl = TRUE)]

# Parse each line into variable name and hex color
df <- data.frame(stringsAsFactors = FALSE,
  name = sub("^\\$(.+?):.*", "\\1", lines, perl = TRUE),
  color = sub(".*(#[0-9A-Za-z]+).*", "\\1", lines, perl = TRUE)
)
# Sort by color value
df <- df[order(df$color),]

# All of the color values that we want to replace, are assigned a variable name.
# Any variables in _variables.scss that point to one of these color values, will
# now point to the variable name by default. For example, since _variables.scss
# contains this:
#
#   $navbar-inverse-bg: #222 !default;
#
# Our goal is to generate the rule:
#
#   $navbar-inverse-bg: $gray-darker !default;
#
# So that all such variables (i.e. ones whose default value is #222) can be
# changed just by modifying $gray-darker.
named_colors <- c(
  # These variables already exist in _variables.scss
  "gray-base" = "#000",
  "gray-darker" = "#222",
  "gray-dark" = "#333",
  "gray" = "#555",
  "gray-light" = "#777",
  "gray-lighter" = "#eee",

  # These are variables we're introducing
  "gray-44" = "#444",
  "gray-88" = "#888",
  "gray-99" = "#999",
  "gray-cc" = "#ccc",
  "gray-dd" = "#ddd",
  "gray-e5" = "#e5e5e5",
  "gray-f5" = "#f5f5f5",
  "gray-f8" = "#f8f8f8",
  "gray-f9" = "#f9f9f9",
  "white" = "#fff"
)

# Color variables in _variables.scss that we don't have names for.
# Consider this a potential TODO list of colors we could dynamically set.
df[!df$color %in% named_colors,]

# Generate the code for color_mapping in bs3_theme_base_colors
library(dplyr)
df %>%
  mutate(varname = names(named_colors[match(color, named_colors)])) %>%
  filter(!is.na(varname)) %>%
  filter(name != varname) %>%
  mutate(varname = paste0("$", varname)) %>%
  select(name, varname) %>%
  { setNames(.$varname, .$name) } %>%
  as.list() %>%
  dput()

#' Column-first uniform grid layouts
#'
#' @description
#' Wraps a 1d sequence of UI elements into a 2d grid. The number of columns (and
#' rows) in the grid dependent on the column `width` as well as the size of the
#' display. For more explanation and illustrative examples, see the _References_
#' section below.
#'
#' @param ... Unnamed arguments should be UI elements (e.g., [card()]). Named
#'   arguments become attributes on the containing [htmltools::tag] element.
#' @param width The desired width of each card, which can be any of the
#'  following:
#'   * A (unit-less) number between 0 and 1.
#'     * This should be specified as `1/num`, where `num` represents the number
#'       of desired columns.
#'   * A [CSS length unit][htmltools::validateCssUnit()]
#'     * Either the minimum (when `fixed_width=FALSE`) or fixed width
#'       (`fixed_width=TRUE`).
#'   * `NULL`
#'     * Allows power users to set the `grid-template-columns` CSS property
#'       manually, either via a `style` attribute or a CSS stylesheet.
#' @param fixed_width When `width` is greater than 1 or is a CSS length unit,
#'   e.g. `"200px"`, `fixed_width` indicates whether that `width` value
#'   represents the absolute size of each column (`fixed_width=TRUE`) or the
#'   minimum size of a column (`fixed_width=FALSE`). When `fixed_width=FALSE`,
#'   new columns are added to a row when `width` space is available and columns
#'   will never exceed the container or viewport size. When `fixed_width=TRUE`,
#'   all columns will be exactly `width` wide, which may result in columns
#'   overflowing the parent container.
#' @param heights_equal If `"all"` (the default), every card in every row of the
#'   grid will have the same height. If `"row"`, then every card in _each_ row
#'   of the grid will have the same height, but heights may vary between rows.
#' @param fill Whether or not to allow the layout to grow/shrink to fit a
#'   fillable container with an opinionated height (e.g., `page_fillable()`).
#' @param fillable Whether or not each element is wrapped in a fillable container.
#' @param height_mobile Any valid CSS unit to use for the height when on mobile
#'   devices (or narrow windows).
#' @param min_height,max_height The maximum or minimum height of the layout container.
#'   Can be any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `max_height="200px"`). Use these arguments in filling layouts to ensure that a
#'   layout container doesn't shrink below `min_height` or grow beyond `max_height`.
#' @inheritParams card
#' @inheritParams card_body
#'
#' @export
#' @family Column layouts
#'
#' @references The bslib website features `layout_column_wrap()` in two places:
#'   * [Column-based layouts](https://rstudio.github.io/bslib/articles/column-layout/index.html)
#'   * [Cards: Multiple cards](https://rstudio.github.io/bslib/articles/cards/index.html#multiple-cards)
#'
#' @examplesIf rlang::is_interactive()
#' x <- card("A simple card")
#'
#' # Always has 2 columns (on non-mobile)
#' layout_column_wrap(width = 1/2, x, x, x)
#'
#' # Automatically lays out three cards into columns
#' # such that each column is at least 200px wide:
#' layout_column_wrap(x, x, x)
#'
#' # To use larger column widths by default, set `width`.
#' # This example has 3 columns when the screen is at least 900px wide:
#' layout_column_wrap(width = "300px", x, x, x)
#'
#' # You can add a list of items, spliced with rlang's `!!!` operator
#' layout_column_wrap(!!!list(x, x, x))
#'
#' @export
layout_column_wrap <- function(
  ...,
  width = "200px",
  fixed_width = FALSE,
  heights_equal = c("all", "row"),
  fill = TRUE,
  fillable = TRUE,
  height = NULL,
  height_mobile = NULL,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL
) {
  heights_equal <- match.arg(heights_equal)

  args <- separate_arguments(...)
  attribs <- args$attribs
  children <- args$children

  if (missing(width)) {
    first_is_width <-
      is.null(children[[1]]) ||
      is_probably_a_css_unit(children[[1]])

    if (first_is_width) {
      # Assume an unnamed first argument that matches our expectations for
      # `width` is actually the width argument, with a warning
      lifecycle::deprecate_warn(
        "0.6.0",
        "layout_column_wrap(width = 'must be named')"
      )
      width <- children[[1]]
      children <- children[-1]
    }
  }

  if (length(width) > 1) {
    stop("`width` of length greater than 1 is not currently supported.")
  }

  colspec <- if (!is.null(width)) {
    if (width > 0 && width <= 1) {
      num_cols <- 1 / width
      if (num_cols != as.integer(num_cols)) {
        stop("Could not interpret width argument; see ?layout_column_wrap")
      }
      sprintf("repeat(%s, minmax(0, 1fr))", num_cols)
    } else {
      if (fixed_width) {
        paste0("repeat(auto-fit, ", validateCssUnit(width), ")")
      } else {
        sprintf(
          "repeat(auto-fit, minmax(min(%s, 100%%), 1fr))",
          validateCssUnit(width)
        )
      }
    }
  }

  # If relevant (i.e., fillable=TRUE), wrap each child in a fillable context
  # (so fill/flex items can fill the available area)
  children <- lapply(children, grid_item_container, fillable = fillable)

  tag <- div(
    class = "bslib-grid bslib-mb-spacing",
    style = css(
      grid_template_columns = colspec,
      grid_auto_rows = if (heights_equal == "all") "1fr",
      # Always provide the `height:auto` default so that the CSS variable
      # doesn't get inherited in a scenario like layout_column_wrap(height=200, ..., layout_column_wrap(...))
      "--bslib-grid-height" = validateCssUnit(height %||% "auto"),
      "--bslib-grid-height-mobile" = validateCssUnit(height_mobile %||% "auto"),
      gap = validateCssUnit(gap),
      min_height = validateCssUnit(min_height),
      max_height = validateCssUnit(max_height)
    ),
    !!!attribs,
    children,
    component_dependencies()
  )

  tag <- bindFillRole(tag, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "layout_column_wrap()")
  )
}

is_probably_a_css_unit <- function(x) {
  if (length(x) != 1) {
    return(FALSE)
  }
  if (is.numeric(x)) {
    return(TRUE)
  }
  if (!is.character(x)) {
    return(FALSE)
  }
  tryCatch(
    {
      validateCssUnit(x)
      TRUE
    },
    error = function(e) FALSE
  )
}

#' Responsive 12-column grid layouts
#'
#' Create responsive, column-based grid layouts, based on a 12-column grid.
#'
#' @inheritParams layout_column_wrap
#' @param col_widths One of the following:
#'   * `NA` (the default): Automatically determines a sensible number of columns
#'     based on the number of children.
#'   * A numeric vector of integers between 1 and 12, where each element
#'     represents the number of columns for the relevant UI element. Elements
#'     that happen to go beyond 12 columns wrap onto the next row. For example,
#'     `col_widths = c(4, 8, 12)` allocates 4 columns to the first element, 8
#'     columns to the second element, and 12 columns to the third element (which
#'     wraps to the next row). Negative values are also allowed, and are treated
#'     as empty columns. For example, `col_widths = c(-2, 8, -2)` would allocate
#'     8 columns to an element (with 2 empty columns on either side).
#'   * A [breakpoints()] object, where each breakpoint may be either of the
#'     above.
#' @param row_heights One of the following:
#'   * A numeric vector, where each value represents the
#'     [fractional unit](https://css-tricks.com/introduction-fr-css-unit/)
#'     (`fr`) height of the relevant row. If there are more rows than values
#'     provided, the pattern will repeat. For example, `row_heights = c(1, 2)`
#'     allows even rows to take up twice as much space as odd rows.
#'   * A list of numeric and [CSS length units][htmltools::validateCssUnit()],
#'     where each value represents the height of the relevant row. If more rows
#'     are needed than values provided, the pattern will repeat. For example,
#'     `row_heights = list("auto", 1)` allows the height of odd rows to be
#'     driven by its contents and even rows to be
#'     [`1fr`](https://css-tricks.com/introduction-fr-css-unit/).
#'   * A character vector/string of [CSS length units][htmltools::validateCssUnit()].
#'     In this case, the value is supplied directly to `grid-auto-rows`.
#'   * A [breakpoints()] object, where each breakpoint may be either of the above.
#'
#' @export
#' @family Column layouts
#'
#' @references [Column-based layouts](https://rstudio.github.io/bslib/articles/column-layout/index.html)
#'   on the bslib website.
#'
#' @seealso [breakpoints()] for more information on specifying column widths at
#'   responsive breakpoints.
#'
#' @examplesIf rlang::is_interactive()
#' x <- card("A simple card")
#'
#' page_fillable(
#'   layout_columns(x, x, x, x)
#' )
#'
#' # Or add a list of items, spliced with rlang's `!!!` operator
#' page_fillable(
#'  layout_columns(!!!list(x, x, x))
#' )
#'
#' page_fillable(
#'   layout_columns(
#'     col_widths = c(6, 6, 12),
#'     x, x, x
#'   )
#' )
#'
#' page_fillable(
#'   layout_columns(
#'     col_widths = c(6, 6, -2, 8),
#'     row_heights = c(1, 3),
#'     x, x, x
#'   )
#' )
#'
#' page_fillable(
#'   fillable_mobile = TRUE,
#'   layout_columns(
#'     col_widths = breakpoints(
#'       sm = c(12, 12, 12),
#'       md = c(6, 6, 12),
#'       lg = c(4, 4, 4)
#'     ),
#'     x, x, x
#'   )
#' )
#'
layout_columns <- function(
  ...,
  col_widths = NA,
  row_heights = NULL,
  fill = TRUE,
  fillable = TRUE,
  gap = NULL,
  class = NULL,
  height = NULL,
  min_height = NULL,
  max_height = NULL
) {
  args <- separate_arguments(...)
  attribs <- args$attribs
  children <- args$children
  n_kids <- length(children)

  # Check for spec problems so we can stop early
  col_spec <- as_col_spec(col_widths, n_kids)

  # Wrap each child in a container (so fill/flex items can fill the available area)
  children <- Map(
    f = grid_item_container,
    children,
    fillable = fillable
  )

  tag <- web_component(
    "bslib-layout-columns",
    class = "bslib-grid grid bslib-mb-spacing",
    style = css(
      height = validateCssUnit(height),
      gap = validateCssUnit(gap),
      min_height = validateCssUnit(min_height),
      max_height = validateCssUnit(max_height)
    ),
    # We don't enable the next option by default, but users could add this
    # attribute to hide the internal elements until after the custom element
    # adds classes to all of the elements. We don't think this is needed but
    # it's here if we find there are cases where rendering the children and
    # adding classes takes longer than we've anticipated in our testing.
    # "hidden-until-init" = NA,
    !!!col_widths_attrs(col_spec),
    !!!row_heights_css_vars(row_heights),
    !!!attribs,
    !!!children
  )

  tag <- bindFillRole(tag, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "layout_columns()")
  )
}

as_col_spec <- function(col_widths, n_kids) {
  if (is.null(col_widths) || rlang::is_na(col_widths)) {
    return(NULL)
  }

  if (!is_breakpoints(col_widths)) {
    col_widths <- breakpoints(sm = col_widths)
  }

  for (break_name in names(col_widths)) {
    bk <- col_widths[[break_name]]

    if (rlang::is_na(bk)) {
      next
    }

    if (isTRUE(any(bk == 0))) {
      abort(
        "Column values must be greater than 0 to indicate width, or negative to indicate a column offset."
      )
    }

    if (length(bk) > 1 && anyNA(bk)) {
      abort(
        "Cannot mix widths and `NA` values. All column widths must be specified, or choose auto widths using a single `NA` value."
      )
    }

    if (!any(bk > 0)) {
      abort("Column values must include at least one positive integer width.")
    }

    if (length(bk[bk > 0]) > n_kids) {
      rlang::warn(
        sprintf(
          "More column widths than children at breakpoint '%s', extra widths will be ignored.",
          break_name
        )
      )
    }
  }

  col_widths
}

col_widths_attrs <- function(col_spec) {
  if (is.null(col_spec) || rlang::is_na(col_spec)) {
    return(NULL)
  }

  names(col_spec) <- paste0("col-widths-", names(col_spec))
  lapply(col_spec, function(x) {
    if (rlang::is_na(x)) NA else paste(x, collapse = ",")
  })
}

maybe_fr_unit <- function(x) {
  if (is.numeric(x)) sprintf("%0.0ffr", x) else x
}

row_heights_css_vars <- function(x) {
  if (is.null(x)) {
    return(list())
  }

  if (!is_breakpoints(x)) {
    # Setting the `xs` breakpoint is equivalent to setting all breaks
    x <- breakpoints(xs = x)
  }

  # creates classes that pair with CSS variables when set
  classes <- paste0("bslib-grid--row-heights--", names(x))

  css_vars <- setNames(x, paste0("--", classes))

  # Treat numeric values as fractional units
  css_vars <- rapply(css_vars, how = "replace", maybe_fr_unit)

  if (identical(names(css_vars), "--bslib-grid--row-heights--xs")) {
    names(css_vars) <- "--bslib-grid--row-heights"
    classes <- character()
  }

  list(
    style = css(!!!css_vars),
    class = classes
  )
}

grid_item_container <- function(el, ..., fillable = TRUE) {
  div(
    ...,
    class = "bslib-grid-item",
    class = if (fillable) "bslib-gap-spacing",
    if (fillable) as_fillable_container(),
    el
  )
}

rlang::check_installed("shiny", version = "1.8.1")
rlang::check_installed("bslib", version = "0.8.0.9000")
rlang::check_installed("future")
rlang::check_installed("ggplot2")
rlang::check_installed("markdown")

library(shiny)
library(bslib)
library(ggplot2)

library(future)
plan(multisession)

options(
  bslib.color_contrast_warnings = FALSE
  # shiny.autoreload.pattern = "_brand[.]yml|app[.]R|[.]s?css" ## TODO: Enable after fixing autoreload
)

theme_brand <- bs_theme()

brand <- attr(theme_brand, "brand")
BRAND_PATH <- brand$path %||% "_brand.yml"

theme_set(theme_minimal())

if (requireNamespace("thematic", quietly = TRUE)) {
  base_font <- bslib:::brand_pluck(brand, "typography", "base", "family")
  if (!is.null(base_font)) {
    # TODO: Update plot fonts dynamically
    thematic::thematic_shiny(font = base_font)
  } else {
    thematic::thematic_shiny()
  }
}

use_download_button <- FALSE

# ---- For hosted demo, delete if local ----
is_app_hosted <-
  Sys.getenv("R_CONFIG_ACTIVE") %in%
  c("shinylive", "shinyapps", "rsconnect", "rstudio_cloud")
is_app_packaged <-
  getwd() == system.file("examples-shiny/brand.yml", package = "bslib")
use_download_button <- is_app_hosted || is_app_packaged

if (bslib:::brand_has(brand, "typography", "fonts")) {
  tryCatch(
    {
      if (brand$typography$fonts[[2]]$files[[1]]$path == "Monda.ttf") {
        if (!file.exists("Monda.ttf")) {
          download.file(
            "https://github.com/google/fonts/raw/48db77e32954f6f5e65a7122ecbe8a2093c4f5d7/ofl/monda/Monda%5Bwght%5D.ttf",
            "Monda.ttf"
          )
          download.file(
            "https://github.com/google/fonts/raw/48db77e32954f6f5e65a7122ecbe8a2093c4f5d7/ofl/monda/OFL.txt",
            "Monda-OFL.txt"
          )
        }
      }
    },
    error = function(err) {}
  )
}
# ---- For hosted demo, delete if local ----

ui <- page_navbar(
  theme = bs_add_rules(theme_brand, sass::sass_file("_colors.scss")),
  title = tagList(
    uiOutput("brand_icon", inline = TRUE),
    uiOutput("brand_name", inline = TRUE)
  ),
  window_title = "brand.yml Demo",
  fillable = TRUE,

  sidebar = sidebar(
    id = "sidebar_editor",
    position = "right",
    open = "closed",
    width = "40%",
    bg = "var(--bs-dark)",
    fg = "var(--bs-light)",
    `data-bs-theme` = "dark",

    card(
      card_header(
        class = "text-bg-secondary hstack",
        div("Edit", code("brand.yml")),
        div(
          class = "ms-auto",
          tooltip(
            tags$a(
              class = "btn btn-link p-0",
              href = "https://posit-dev.github.io/brand-yml/brand/",
              target = "_blank",
              bsicons::bs_icon(
                "question-square-fill",
                title = "About brand.yml",
                size = "1.25rem"
              )
            ),
            "About brand.yml"
          )
        )
      ),
      htmltools::tagAppendAttributes(
        textAreaInput(
          "txt_brand_yml",
          label = NULL,
          value = if (!is.null(brand)) {
            paste(readLines(brand$path, warn = FALSE), collapse = "\n")
          },
          width = "100%",
          height = "80%",
          rows = 20
        ),
        class = "font-monospace",
        .cssSelector = "textarea"
      ),
      card_body(
        padding = 0,
        div(
          id = "editor_brand_yml",
          style = "overflow: auto;",
          as_fill_item()
        )
      )
    ),

    tags$script(
      type = "module",
      HTML(
        '
import { basicEditor } from "https://esm.sh/prism-code-editor@3.4.0/setups";
import "https://esm.sh/prism-code-editor@3.4.0/prism/languages/yaml";

const shinyInput = document.getElementById("txt_brand_yml");

function initBrandEditor() {
  if (typeof Shiny.setInputValue !== "function") {
    setTimeout(initBrandEditor, 100);
    return;
  }
  window.brandEditor = basicEditor(
    "#editor_brand_yml",
    {
      language: "yml",
      theme: "github-dark",
      value: shinyInput.value,
      onUpdate: (value) => {
        Shiny.setInputValue("txt_brand_yml", value);
      },
    },
    () => shinyInput.parentElement.parentElement.remove()
  );
}

initBrandEditor();
'
      )
    ),

    tags$style(
      HTML(
        '
.bslib-sidebar-layout .sidebar-title { margin-bottom: 0 }
#sidebar_editor .sidebar-content { height: max(600px, 100%) }'
      )
    ),

    if (use_download_button) {
      shiny::downloadButton(
        "download",
        label = span(
          "Download",
          code("_brand.yml", style = "color: currentColor"),
          "file"
        ),
        class = "btn-outline-light"
      )
    } else {
      actionButton(
        "save",
        label = span(
          "Save",
          code("_brand.yml", style = "color: currentColor"),
          "file"
        ),
        class = "btn-outline-light"
      )
    }
  ),

  nav_panel(
    "Input Output Demo",
    value = "dashboard",
    layout_sidebar(
      sidebar = sidebar(
        sliderInput("slider1", "Numeric Slider Input", 0, 11, 11),
        numericInput("numeric1", "Numeric Input Widget", 30),
        dateInput("date1", "Date Input Component", value = "2024-01-01"),
        input_switch("switch1", "Binary Switch Input", value = TRUE),
        radioButtons(
          "radio1",
          "Radio Button Group",
          choices = c("Option A", "Option B", "Option C", "Option D")
        ),
        actionButton("action1", "Action Button")
      ),
      shiny::useBusyIndicators(),
      layout_column_wrap(
        value_box(
          title = "Metric 1",
          value = "100",
          theme = "primary",
          id = "value_box_one"
        ),
        value_box(
          title = "Metric 2",
          value = "200",
          theme = "secondary",
          id = "value_box_two"
        ),
        value_box(
          title = "Metric 3",
          value = "300",
          theme = "info",
          id = "value_box_three"
        )
      ),
      card(
        card_header("Plot Output"),
        plotOutput("out_plot")
      ),
      card(
        card_header("Text Output"),
        verbatimTextOutput("out_text")
      )
    )
  ),

  nav_panel(
    "Widget Gallery",
    layout_column_wrap(
      width = 300,
      heights_equal = "row",
      card(
        card_header("Button Variants"),
        actionButton("btn_default", "Default"),
        actionButton("btn_primary", "Primary", class = "btn-primary"),
        actionButton("btn_secondary", "Secondary", class = "btn-secondary"),
        actionButton("btn_success", "Success", class = "btn-success"),
        actionButton("btn_danger", "Danger", class = "btn-danger"),
        actionButton("btn_warning", "Warning", class = "btn-warning"),
        actionButton("btn_info", "Info", class = "btn-info")
      ),
      card(
        card_header("Radio Button Examples"),
        radioButtons(
          "radio2",
          "Standard Radio Group",
          choices = c("Selection 1", "Selection 2", "Selection 3")
        ),
        radioButtons(
          "radio3",
          "Inline Radio Group",
          choices = c("Option 1", "Option 2", "Option 3"),
          inline = TRUE
        )
      ),
      card(
        card_header("Checkbox Examples"),
        checkboxGroupInput(
          "check1",
          "Standard Checkbox Group",
          choices = c("Item 1", "Item 2", "Item 3")
        ),
        checkboxGroupInput(
          "check2",
          "Inline Checkbox Group",
          choices = c("Choice A", "Choice B", "Choice C"),
          inline = TRUE
        )
      ),
      card(
        card_header("Select Input Widgets"),
        selectizeInput(
          "select1",
          "Selectize Input",
          choices = c("Selection A", "Selection B", "Selection C")
        ),
        selectInput(
          "select2",
          "Multiple Select Input",
          choices = c("Item X", "Item Y", "Item Z"),
          multiple = TRUE
        )
      ),
      card(
        card_header("Text Input Widgets"),
        textInput("text1", "Text Input"),
        textAreaInput(
          "textarea1",
          "Text Area Input",
          value = "Default text content for the text area widget"
        ),
        passwordInput("password1", "Password Input")
      )
    )
  ),

  nav_panel(
    "Colors",
    div(
      class = "container-sm overflow-y-auto",
      uiOutput("ui_colors")
    )
  ),

  nav_panel(
    "Documentation",
    div(
      class = "container-sm overflow-y-auto",
      if (FALSE) library(markdown), # for shinyapps.io
      includeMarkdown("documentation.md")
    )
  ),

  nav_spacer(),
  nav_item(input_dark_mode(id = "color_mode")),
  nav_item(
    actionLink(
      "show_editor",
      bsicons::bs_icon(
        "pencil-fill",
        size = "1rem",
        title = "Show/hide editor"
      ),
      class = "nav-link"
    )
  ),
)

errors <- rlang::new_environment()

error_notification <- function(context) {
  function(err) {
    time <- as.character(Sys.time())

    msg <- conditionMessage(err)
    # Strip ANSI color sequences from error messages
    msg <- gsub(
      pattern = "\u001b\\[.*?m",
      replacement = "",
      msg
    )
    # Wrap at 40 characters
    msg <- paste(strwrap(msg, width = 60), collapse = "\n")

    err_id <- rlang::hash(list(time, msg))
    assign(err_id, list(message = msg, context = context), envir = errors)

    showNotification(
      markdown(context),
      action = tags$button(
        class = "btn btn-outline-danger pull-right",
        onclick = sprintf(
          "event.preventDefault(); Shiny.setInputValue('show_error', '%s')",
          err_id
        ),
        "Show details"
      ),
      duration = 10,
      type = "error",
      id = err_id
    )
  }
}

server <- function(input, output, session) {
  brand_yml_text <- debounce(reactive(input$txt_brand_yml), 1000)
  brand_yml <- reactiveVal()

  observeEvent(input$show_editor, sidebar_toggle("sidebar_editor"))

  observeEvent(input$show_error, {
    req(input$show_error)
    err <- get0(input$show_error, errors)

    if (is.null(err)) {
      message("Could not find error with id ", input$show_error)
      return()
    }

    removeNotification(input$show_error)
    rm(list = input$show_error, envir = errors)

    showModal(
      modalDialog(
        size = "l",
        easyClose = TRUE,
        markdown(err$context),
        pre(err$message)
      )
    )
  })

  observeEvent(brand_yml_text(), {
    req(brand_yml_text())

    tryCatch(
      {
        b <- yaml::yaml.load(brand_yml_text())
        b$path <- normalizePath(BRAND_PATH, mustWork = FALSE)
        brand_yml(b)
      },
      error = error_notification(
        "Could not parse `_brand.yml` file. Check for syntax errors."
      )
    )
  })

  observeEvent(brand_yml(), {
    req(brand_yml())

    tryCatch(
      {
        theme <- bs_theme(brand = brand_yml())
        theme <- bs_add_rules(theme, sass::sass_file("_colors.scss"))
        session$setCurrentTheme(theme)
      },
      error = error_notification(
        "Could not compile branded theme. Please check your `_brand.yml` file."
      )
    )
  })

  output$brand_name <- renderUI({
    brand_name <-
      bslib:::brand_pluck(brand_yml(), "meta", "name", "short") %||%
      bslib:::brand_pluck(brand_yml(), "meta", "name")

    if (rlang::is_string(brand_name)) brand_name else "brand.yml Demo"
  })

  output$brand_icon <- renderUI({
    brand <- brand_yml()

    logo <-
      bslib:::brand_pluck(brand, "logo", "small") %||%
      bslib:::brand_pluck(brand, "logo")

    req(rlang::is_string(logo))

    logo_path <- file.path(dirname(BRAND_PATH), logo)
    if (file.exists(logo_path)) {
      img(
        src = base64enc::dataURI(
          file = logo_path,
          mime = mime::guess_type(logo_path)
        ),
        height = 30
      )
    }
  })

  observeEvent(input$save, {
    validate(
      need(input$txt_brand_yml, "_brand.yml file contents cannot be empty.")
    )

    tryCatch(
      {
        writeLines(input$txt_brand_yml, BRAND_PATH)
        showNotification(markdown("Saved `_brand.yml`!"))
      },
      error = error_notification("Could not save `_brand.yml`.")
    )
  })

  output$download <- downloadHandler(
    filename = "_brand.yml",
    content = function(file) {
      validate(
        need(input$txt_brand_yml, "_brand.yml file contents cannot be empty.")
      )
      writeLines(input$txt_brand_yml, file)
    }
  )

  PlotTask <- ExtendedTask$new(function(x_max, y_factor) {
    x <- seq(0, x_max, length.out = 100)
    y <- sin(x) * y_factor

    future({
      Sys.sleep(3)

      df <- data.frame(x = x, y = y)

      ggplot(df, aes(x = x, y = y)) +
        geom_col(width = 1, position = "identity") +
        labs(title = "Sine Wave Output", x = "", y = "")
    })
  })

  observe({
    x_max <- debounce(reactive(input$numeric1), 500)()
    y_factor <- debounce(reactive(input$slider1), 500)()

    PlotTask$invoke(x_max = x_max, y_factor = y_factor)
  })

  output$out_plot <- renderPlot({
    PlotTask$result()
  })

  output$out_text <- renderText({
    "example_function <- function() {\n  return(\"Function output text\")\n}"
  })

  output$ui_colors <- renderUI({
    bootstrap_colors <- c(
      "blue",
      "indigo",
      "purple",
      "pink",
      "red",
      "orange",
      "yellow",
      "green",
      "teal",
      "cyan"
    )
    colors <- c("gray", bootstrap_colors)

    tagList(
      layout_columns(
        col_widths = 3,
        class = "font-monospace",
        !!!lapply(
          c(
            "primary",
            "secondary",
            "dark",
            "light",
            "info",
            "success",
            "warning",
            "danger"
          ),
          function(color) {
            div(
              color,
              class = paste0("p-3 mb-2 position-relative text-bg-", color)
            )
          }
        )
      ),
      layout_columns(
        col_widths = 3,
        class = "font-monospace",
        !!!lapply(
          c("black", "white", "foreground", "background"),
          function(color) {
            div(
              color,
              class = paste0("p-3 mb-2 position-relative bd-", color)
            )
          }
        )
      ),
      layout_column_wrap(
        width = 200,
        !!!lapply(colors, function(color) {
          if (!color %in% c("white", "black")) {
            div(
              class = "mb-3",
              div(
                color,
                class = paste0("p-3 mb-2 position-relative bd-", color, "-500")
              ),
              lapply(seq(100, 900, 100), function(r) {
                div(
                  paste0(color, "-", r),
                  class = paste0("p-3 bd-", color, "-", r)
                )
              })
            )
          }
        })
      )
    )
  })
}

shinyApp(ui, server)

library(shiny)
library(bslib)
library(dplyr)
library(lubridate)
library(plotly)
library(nycflights13)
library(histoslider)
library(rlang)

# Data prep
flights <- flights %>%
  left_join(
    airports %>%
      transmute(dest_name = paste0(name, " (", faa, ")"), dest = faa, end_lat = lat, end_lon = lon)
  ) %>%
  filter(!is.na(dest_name)) %>%
  left_join(
    airports %>% select(origin = faa, start_lat = lat, start_lon = lon)
  ) %>%
  left_join(airlines, by = "carrier") %>%
  rename(carrier_name = name) %>%
  left_join(weather) %>%
  mutate(
    # The overwhelming majority of precipation is 0 so transform
    # with some help from MASS::boxcox() https://stats.stackexchange.com/a/1452/48604
    precip = scales::rescale(precip^-1.55/-.55),
    date = lubridate::ymd(paste(year, month, day, sep = "-"))
  )

PRIMARY <- "#0675DD"

sidebar <- card(
  card_header("Filter controls"),
  uiOutput("origin_reset"),
  checkboxGroupInput(
    "origin",
    span("Origin", icon("plane-departure")),
    choices = sort(unique(flights$origin)),
    inline = TRUE
  ),
  selectInput(
    "dest_name",
    span("Destination", icon("plane-arrival")),
    choices = sort(unique(flights$dest_name)),
    multiple = TRUE,
    width = "100%"
  ),
  selectInput(
    "carrier_name",
    span("Carrier", icon("user-tie")),
    unique(flights$carrier_name),
    multiple = TRUE,
    width = "100%"
  ),
  input_histoslider(
    "sched_dep_time", "Departure time",
    flights$sched_dep_time,
    options = list(
      handleLabelFormat = "0d",
      selectedColor = PRIMARY
    )
  ),
  input_histoslider(
    "sched_arr_time", "Arrival time",
    flights$sched_arr_time,
    options = list(
      handleLabelFormat = "0d",
      selectedColor = PRIMARY
    )
  ),
  input_histoslider(
    "date", "Date",
    flights$date, breaks = "months",
    options = list(
      handleLabelFormat = "%b %e",
      selectedColor = PRIMARY
    )
  )
)


flights_card <- card(
  full_screen = TRUE,
  card_header("Flight paths"),
  card_body_fill(plotlyOutput("flight_paths")),
  card_footer("Marker areas are proportional to mean arrival delay")
)

delay_corr_card <- card(
  full_screen = TRUE,
  card_header(
    "Arrival vs departure delay",
    checkboxInput("scatter_summarize", "Summarize", TRUE, width = "fit-content"),
    class = "d-flex justify-content-between"
  ),
  card_body_fill(plotlyOutput("scatter_delay"))
)

delay_card <- navs_pill_card(
  title = "Arrival delay",
  full_screen = TRUE,
  wrapper = card_body_fill,
  nav(
    "Overall",
    plotlyOutput("arr_delay")
  ),
  nav(
    "Over time",
    plotlyOutput("arr_delay_series")
  )
)

main <- layout_column_wrap(
  width = 1, heights_equal = "row",
  uiOutput("value_boxes", fill = TRUE),
  layout_column_wrap(
    width = "200px", class = "my-3",
    flights_card, delay_corr_card
  ),
  layout_column_wrap(
    width = 1,
    delay_card
  )
)


flights_ui <- layout_column_wrap(
  height = "calc(100vh - 90px)",
  width = NULL, style = "grid-template-columns: 280px 1fr",
  sidebar, main
)


ui <- page_navbar(
  fluid = TRUE,
  theme = bs_theme(
    base_font = font_google(
      "Open Sans", wght = c(300, 400, 500, 600, 700, 800),
      ital = c(0, 1)
    ),
    "primary" = PRIMARY,
    "navbar-bg" = PRIMARY
  ),
  title = tags$span(
    tags$img(
      src = "logo.png",
      style = "width:46px;height:auto;margin-right:24px;"
    ),
    "Demo"
  ),
  nav("NYC Flights", flights_ui),
  nav_spacer(),
  nav_item(
    tags$a(
      tags$span(
        bsicons::bs_icon("code-slash"), "Source code"
      ),
      href = "https://github.com/rstudio/bslib/tree/main/inst/examples/demo",
      target = "_blank"
    )
  )
)

server <- function(input, output, session) {
  # ---------------------------------------------------------
  # Flights tab logic
  #
  # WARNING: this server-side filtering logic is VERY experimental
  # at this point and won't be easily adapt to different use cases.
  # If you feel tempted to use it, use with caution.
  # ---------------------------------------------------------

  # Mapping from input id name to updating input function
  input_discrete_vars <- list(
    origin = function(...) updateCheckboxGroupInput(..., inline = TRUE),
    dest_name = updateSelectInput,
    carrier_name = updateSelectInput
  )
  input_numeric_vars <- list(
    sched_dep_time = update_histoslider,
    sched_arr_time = update_histoslider,
    date = update_histoslider
  )
  input_vars <- c(
    input_discrete_vars,
    input_numeric_vars
  )

  filter_index <- function(d) {
    idx <- rep(TRUE, nrow(d))
    for (var in names(d)) {
      idx <- idx & filter_col(d, var)
    }
    idx & !is.na(idx)
  }

  filter_col <- function(d, var) {
    vals <- d[[var]]
    input_val <- input[[var]]
    if (is.null(input_val)) {
      return(TRUE)
    }

    if (is.character(vals) || is.factor(vals) || is.logical(vals)) {
      return(d[[var]] %in% input_val)
    }

    # N.B. between() will remove NAs, which we probably don't
    # want to remove until the slider is considered 'active'
    rng <- range(vals, na.rm = TRUE)
    active <- isTRUE(rng[1] <= input_val[1] || input_val[2] <= rng[2])

    if (!active) {
      return(TRUE)
    }

    dplyr::between(vals, input_val[1], input_val[2])
  }

  # When a particular input filter changes, we need to update every other filter
  # with data
  lapply(names(input_vars), function(var) {

    index <- reactive({
      filter_index(flights[, setdiff(names(input_vars), var)])
    })

    observeEvent(index(), ignoreInit = TRUE, ignoreNULL = FALSE, {

      d <- flights[index(), ]
      if (nrow(d) == 0) return()
      input_val <- input[[var]]
      update_input_func <- input_vars[[var]]

      if (var %in% names(input_discrete_vars)) {
        choices <- sort(unique(d[[var]]))
        update_input_func(
          inputId = var, choices = choices, selected = input_val %||% character(0)
        )
      } else {
        update_input_func(
          id = var,
          values = d[[var]],
          start = input_val[1],
          end = input_val[2]
        )
      }

    })
  })

  output$origin_reset <- renderUI({
    actionLink(
      "origin_reset", "Reset",
      style = htmltools::css(
        text_decoration = "none",
        font_weight = 700,
        font_size = ".875rem",
        float = "right",
        visibility = if (length(input$origin) > 0) "visible" else "hidden",
        margin_top = "-7px"
      )
    )
  })

  observeEvent(input$origin_reset, {
    updateCheckboxGroupInput(
      inputId = "origin",
      choices = sort(unique(flights$origin)),
      selected = character(0)
    )
  })

  # Flights with all filters applied (i.e., data used for value boxes/plots)
  flight_dat <- reactive({
    flights[filter_index(flights), ]
  })

  output$value_boxes <- renderUI({
    d <- flight_dat()
    n_flights <- value_box(
      "A TOTAL OF",
      paste(nrow(d), "flights"),
      paste(
        "Across", length(unique(d$dest_name)),
        "destinations"
      ),
      tags$p(paste(
        "On", length(unique(d$carrier_name)), "different carriers"
      )),
      showcase = bsicons::bs_icon("airplane")
    )

    delay_dep <- value_box(
      "AVERAGE DEPARTURE",
      paste(
        round(mean(d$dep_delay, na.rm = T), 0),
        "mins late"
      ),
      paste0(
        round(100 * sum(d$dep_delay > 0, na.rm = T) / nrow(d), 1),
        "% of flights depart late"
      ),
      showcase = bsicons::bs_icon("hourglass-split")
    )

    delay_arr <- value_box(
      "AVERAGE ARRIVAL",
      paste(
        round(mean(d$arr_delay, na.rm = T), 0),
        "mins late"
      ),
      paste0(
        round(100 * sum(d$arr_delay > 0, na.rm = TRUE) / nrow(d), 1),
        "% of flights arrive late"
      ),
      showcase = bsicons::bs_icon("hourglass-bottom")
    )

    layout_column_wrap(width = 1/3, n_flights, delay_dep, delay_arr)
  })

  output$flight_paths <- renderPlotly({
    flight_dat() %>%
      group_by(start_lon, start_lat, end_lon, end_lat, origin, dest) %>%
      summarise(mean_delay = mean(arr_delay, na.rm = TRUE)) %>%
      plot_geo(color = I("#0D6EFD"), showlegend = FALSE) %>%
      add_segments(
        x = ~start_lon, xend = ~end_lon,
        y = ~start_lat, yend = ~end_lat,
        alpha = 0.5, size = I(1), hoverinfo = "none"
      ) %>%
      add_markers(
        x = ~end_lon, y = ~end_lat, size = ~mean_delay,
        hoverinfo = "text", alpha = 0.1,
        text = ~paste0(
          origin, " -> ", dest, "<br>",
          "Average delay: ", round(mean_delay, 1)
        )
      ) %>%
      config(displayModeBar = FALSE) %>%
      layout(
        geo = list(
          projection = list(
            type = 'orthographic',
            rotation = list(lon = -100, lat = 40, roll = 0)
          ),
          showland = TRUE,
          landcolor = toRGB("gray95"),
          countrycolor = toRGB("gray80")
        )
      )
  })

  output$scatter_delay <- renderPlotly({
    d <- flight_dat()
    req(nrow(d) > 0)

    if (isTRUE(input$scatter_summarize)) {
      x <- d$dep_delay
      y <- d$arr_delay
      p <- plot_ly() %>%
        add_histogram2dcontour(
          x = x, y = y, hovertemplate = "Number of flights: %{z}<extra></extra>"
        ) %>%
        layout(
          xaxis = list(
            title = "Departure delay",
            range = c(-30, 100)
          ),
          yaxis = list(
            title = "Arrival delay",
            range = c(-50, 100)
          )
        ) %>%
        colorbar(title = "Number of flights", y = 0.5, yanchor = "middle")

      return(p)
    }

    m <- loess(arr_delay~dep_delay, data = d)
    rng <- range(d$dep_delay, na.rm = TRUE)
    grid <- data.frame(
      dep_delay = seq(rng[1], rng[2], length = 100)
    )
    grid <- modelr::add_predictions(grid, m)

    plot_ly() %>%
      add_markers(
        data = d,
        x = ~dep_delay,
        y = ~arr_delay,
        stroke = I(PRIMARY),
        color = I("transparent"),
        span = I(1),
        text = ~paste0(
          origin, " -> ",
          substr(dest_name, nchar(dest_name) - 3, nchar(dest_name) - 1), "\n",
          carrier_name, "\n",
          "Arrival: ", arr_delay, "\n",
          "Departure: ", dep_delay, "\n"
        ),
        hoverinfo = "text",
        name = "Flights"
      ) %>%
      add_lines(
        data = grid,
        x = ~dep_delay,
        y = ~pred,
        text = ~paste0("Given a departure delay of ", round(dep_delay), "\n expect an arrival delay of ", round(pred)),
        hoverinfo = "text",
        name = "Predicted"
      ) %>%
      layout(
        legend = list(
          orientation = "h",
          x = 0.1, xanchor = "left",
          y = 0.95, yanchor = "top",
          bgcolor = "transparent"
        ),
        xaxis = list(title = "Departure delay"),
        yaxis = list(title = "Arrival delay")
      ) %>%
      toWebGL()
  })


  output$arr_delay <- renderPlotly({
    x <- flight_dat()$arr_delay
    req(length(x) > 0)

    x_mean <- mean(x, na.rm = TRUE)
    end <- quantile(x, probs = 0.99, na.rm = TRUE)
    plot_ly(x = x, hovertemplate = "%{y} flights were %{x} min late <extra></extra>", color = I(PRIMARY)) %>%
      config(displayModeBar = FALSE) %>%
      rangeslider(start = min(x, na.rm = TRUE), end = as.numeric(end)) %>%
      # TODO: add annotations for each line?
      layout(
        shapes = list(
          type = "line",
          x0 = x_mean, x1 = x_mean,
          y0 = 0, y1 = 1,
          yref = "paper",
          line = list(color = "orange", dash = "solid")
        )
      )
  })

  output$arr_delay_series <- renderPlotly({
    d <- flight_dat()
    req(nrow(d) > 0)

    d %>%
      group_by(date) %>%
      summarise(arr_delay_mean = mean(arr_delay, na.rm = TRUE)) %>%
      plot_ly(
        x = ~date, y = ~arr_delay_mean,
        hovertemplate = "%{y:.1f}<extra></extra>",
        color = I(PRIMARY)
      ) %>%
      add_lines() %>%
      layout(
        hovermode = "x",
        xaxis = list(title = "", tickformat = "%b %e"),
        yaxis = list(title = "Average delay")
      )
  })

}

shinyApp(ui, server)

# TODO: Once we have more UI stuff in bslib, copy over the
# page constructors from shiny so folks can create static
# pages without a shiny dependency.

#' Create a Bootstrap page
#'
#' These functions are small wrappers around shiny's page constructors (i.e., [shiny::fluidPage()], [shiny::navbarPage()], etc) that differ in two ways:
#'  * The `theme` parameter defaults bslib's recommended version of Bootstrap (for new projects).
#'  * The return value is rendered as an static HTML page when printed interactively at the console.
#'
#' @inheritParams shiny::bootstrapPage
#' @seealso [shiny::bootstrapPage()]
#' @export
page <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  as_page(
    shiny::bootstrapPage(..., title = title, theme = theme, lang = lang)
  )
}

#' @rdname page
#' @inheritParams shiny::fluidPage
#' @seealso [shiny::fluidPage()]
#' @export
page_fluid <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  as_page(
    shiny::fluidPage(..., title = title, theme = theme, lang = lang)
  )
}

#' @rdname page
#' @inheritParams shiny::fixedPage
#' @seealso [shiny::fixedPage()]
#' @export
page_fixed <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  as_page(
    shiny::fixedPage(..., title = title, theme = theme, lang = lang)
  )
}

#' @rdname page
#' @inheritParams shiny::fillPage
#' @seealso [shiny::fillPage()]
#' @export
page_fill <- function(..., padding = 0, title = NULL,
                      theme = bs_theme(), lang = NULL) {
  as_page(
    shiny::fillPage(..., padding = padding, title = title, theme = theme, lang = lang)
  )
}

#' @rdname page
#' @inheritParams navs_bar
#' @inheritParams bs_page
#' @seealso [shiny::navbarPage()]
#' @param window_title the browser window title. The default value, `NA`, means
#'   to use any character strings that appear in `title` (if none are found, the
#'   host URL of the page is displayed by default).
#' @export
page_navbar <- function(..., title = NULL, id = NULL, selected = NULL,
                        position = c("static-top", "fixed-top", "fixed-bottom"),
                        header = NULL, footer = NULL,
                        bg = NULL, inverse = "auto",
                        collapsible = TRUE, fluid = TRUE,
                        theme =  bs_theme(),
                        window_title = NA,
                        lang = NULL) {

  # https://github.com/rstudio/shiny/issues/2310
  if (!is.null(title) && isTRUE(is.na(window_title))) {
    window_title <- unlist(find_characters(title))
    if (is.null(window_title)) {
      warning("Unable to infer a `window_title` default from `title`. Consider providing a character string to `window_title`.")
    } else {
      window_title <- paste(window_title, collapse = " ")
    }
  }

  page(
    title = window_title,
    theme = theme,
    lang = lang,
    navs_bar(
      ..., title = title, id = id, selected = selected,
      position = match.arg(position), header = header,
      footer = footer, bg = bg, inverse = inverse,
      collapsible = collapsible, fluid = fluid
    )
  )
}

#> unlist(find_characters(div(h1("foo"), h2("bar"))))
#> [1] "foo" "bar"
find_characters <- function(x) {
  if (is.character(x)) {
    return(x)
  }
  if (inherits(x, "shiny.tag")) {
    return(lapply(x$children, find_characters))
  }
  if (is.list(x)) {
    return(lapply(x, find_characters))
  }
  NULL
}

library(htmltools)

# All page_*() functions are very thin wrappers around
# shiny::*Page() at the moment (except for page_navbar(), which
# which why we only have tests for page_navbar())

test_that("page_navbar()", {

  expect_snapshot(
    renderTags(
      page_navbar(
        title = div(h1("foo"), h2("bar"))
      )
    )$head,
    cran = TRUE
  )

  expect_snapshot(
    renderTags(
      page_navbar(
        title = "foo",
        window_title = "bar"
      )
    )$head,
    cran = TRUE
  )

})

#' Obtain a list of all available bootswatch themes.
#'
#' @param version the major version of Bootswatch.
#' @param full_path whether to return a path to the installed theme.
#' @export
#' @return a character vector of Bootswatch themes.
bootswatch_themes <- function(version = version_default(), full_path = FALSE) {
  list.dirs(bootswatch_dist(version), full.names = full_path, recursive = FALSE)
}

#' Obtain a theme's Bootswatch theme name
#'
#' @inheritParams bs_theme_update
#' @return the Bootswatch theme named used (if any) in the `theme`.
#' @export
theme_bootswatch <- function(theme) {
  if (!is_bs_theme(theme)) return(NULL)

  swatch <- grep("^bs_bootswatch_", class(theme), value = TRUE)
  if (!length(swatch)) return(NULL)

  sub("^bs_bootswatch_", "", swatch)
}

#' Obtain a theme's Bootstrap version
#'
#' @inheritParams bs_theme_update
#' @return the major version of Bootstrap used in the `theme`.
#' @export
theme_version <- function(theme) {
  if (!is_bs_theme(theme)) return(NULL)

  version <- grep("^bs_version_", class(theme), value = TRUE)
  sub("^bs_version_", "", version)
}


bootswatch_dist <- function(version) {
  switch_version(
    version,
    five = lib_file("bsw5", "dist"),
    four = lib_file("bsw4", "dist"),
    three = lib_file("bsw3")
  )
}

#' Add a popover to a UI element
#'
#' @description
#' Display additional information when clicking on a UI element (typically a
#' button).
#'
#' @param trigger The UI element to serve as the popover trigger (typically a
#'   [shiny::actionButton()] or similar). If `trigger` renders as multiple HTML
#'   elements (e.g., it's a `tagList()`), the last HTML element is used for the
#'   trigger. If the `trigger` should contain all of those elements, wrap the
#'   object in a [htmltools::div()] or [htmltools::span()].
#' @param ... UI elements for the popover's body. Character strings are
#'   [automatically escaped][htmltools::htmlEscape()] unless marked as
#'   [htmltools::HTML()].
#' @param title A title (header) for the popover. To remove a header
#'   with `update_popover()`, provide a either an empty string or `character(0)`.
#' @param id A character string. Required to re-actively respond to the
#'   visibility of the popover (via the `input[[id]]` value) and/or update the
#'   visibility/contents of the popover.
#' @param placement The placement of the popover relative to its trigger.
#' @param options A list of additional
#'   [options](https://getbootstrap.com/docs/5.3/components/popovers/#options).
#'
#' @section Closing popovers:
#'
#'   In addition to clicking the `close_button`, popovers can be closed by
#'   pressing the Esc/Space key when the popover (and/or its trigger) is
#'   focused.
#'
#' @section Theming/Styling:
#'
#'   ```{r child="man/fragments/tooltip-popover_theming.Rmd", el = "popover"}
#'   ```
#'
#'   ```
#'   popover(
#'     "Trigger", "Popover message",
#'     options = list(customClass = "my-pop")
#'   )
#'   ```
#'
#'   And then add relevant rules to [bs_theme()] via [bs_add_rules()]:
#'
#'   ```
#'   bs_theme() |> bs_add_rules(".my-pop { max-width: none; }")
#'   ```
#'
#' @section Accessibility of Popover Triggers:
#'
#'   ```{r child="man/fragments/tooltip-popover_a11y-trigger.Rmd", el = "popover"}
#'   ```
#'
#'   ```r
#'   popover(
#'     bsicons::bs_icon("gear", title = "Settings"),
#'     title = "Settings",
#'     sliderInput("n", "Number of points", 1, 100, 50)
#'   )
#'   ```
#'
#'   ```r
#'   popover(
#'     fontawesome::fa("gear", a11y = "sem", title = "Settings"),
#'     title = "Settings",
#'     sliderInput("n", "Number of points", 1, 100, 50)
#'   )
#'   ```
#'
#' @describeIn popover Add a popover to a UI element
#'
#' @export
#' @family Components
#'
#' @references Popovers are based on [Bootstrap's Popover
#'   component](https://getbootstrap.com/docs/5.3/components/popovers/). See the
#'   bslib website for an [interactive introduction to tooltips and
#'   popovers](https://rstudio.github.io/bslib/articles/tooltips-popovers/index.html).
#'
#' @seealso [tooltip()] provides an alternative way to display informational
#'   text on demand, typically when focusing or hovering over a trigger element.
#'
#' @examplesIf rlang::is_interactive()
#'
#' popover(
#'   shiny::actionButton("btn", "A button"),
#'   "Popover body content...",
#'   title = "Popover title"
#' )
#'
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fixed(
#'   card(class = "mt-5",
#'     card_header(
#'       popover(
#'         uiOutput("card_title", inline = TRUE),
#'         title = "Provide a new title",
#'         textInput("card_title", NULL, "An editable title")
#'       )
#'     ),
#'     "The card body..."
#'   )
#' )
#'
#' server <- function(input, output) {
#'   output$card_title <- renderUI({
#'     list(input$card_title, bsicons::bs_icon("pencil-square"))
#'   })
#' }
#'
#' shinyApp(ui, server)
popover <- function(
  trigger,
  ...,
  title = NULL,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
) {
  args <- separate_arguments(...)
  children <- args$children
  attribs <- args$attribs

  if (length(children) == 0) {
    abort("At least one value must be provided to `...`.")
  }

  bad_opts <- intersect(c("content", "title", "placement"), names(options))
  if (length(bad_opts) > 0) {
    rlang::abort(
      sprintf("The `%s` option cannot be specified directly.", bad_opts[1])
    )
  }

  res <- web_component(
    "bslib-popover",
    id = id,
    placement = rlang::arg_match(placement),
    bsOptions = to_json(options),
    !!!attribs,
    # Use <template> as a way to protect these children from potentially being
    # pulled outside this element (the browser's parser does this to, for
    # example, block elements inside a <p> tag)
    tags$template(
      div(!!!children, style = "display:contents;"),
      div(title, style = "display:contents;")
    ),
    trigger
  )

  res <- tag_require(res, version = 5, caller = "popover()")
  as_fragment(res)
}

#' @describeIn popover Programmatically show/hide a popover.
#'
#' @param show Whether to show (`TRUE`) or hide (`FALSE`) the popover. The
#'   default (`NULL`) will show if currently hidden and hide if currently shown.
#'   Note that a popover will not be shown if the trigger is not visible (e.g.,
#'   it's hidden behind a tab).
#' @param session A Shiny session object (the default should almost always be
#'   used).
#'
#' @export
toggle_popover <- function(id, show = NULL, session = get_current_session()) {
  show <- normalize_show_value(show)

  msg <- list(method = "toggle", value = show)
  force(id)
  callback <- function() {
    session$sendInputMessage(id, msg)
  }
  session$onFlush(callback, once = TRUE)
}

#' @describeIn popover Update the contents of a popover.
#' @export
update_popover <- function(
  id,
  ...,
  title = NULL,
  session = get_current_session()
) {
  body <- tagList(...)

  msg <- dropNulls(
    list(
      method = "update",
      content = if (length(body) > 0) processDeps(body, session),
      header = if (!is.null(title)) processDeps(title, session)
    )
  )

  force(id)
  callback <- function() {
    session$sendInputMessage(id, msg)
  }
  session$onFlush(callback, once = TRUE)
}

#' Create a Bootstrap theme
#'
#' @description
#'
#' Creates a Bootstrap theme object, where you can:
#'
#' * Choose a (major) Bootstrap `version`.
#' * Choose a [Bootswatch theme](https://bootswatch.com) (optional).
#' * Customize main colors and fonts via explicitly named arguments (e.g.,
#'   `bg`, `fg`, `primary`, etc).
#' * Customize other, lower-level, Bootstrap Sass variable defaults via `...`.
#'
#' To learn more about how to implement custom themes, as well as how to use them inside Shiny and R Markdown, [see here](https://rstudio.github.io/bslib/articles/bslib.html).
#'
#' @section Colors:
#'
#'  Colors may be provided in any format that [htmltools::parseCssColors()] can
#'  understand. To control the vast majority of the ('grayscale') color
#'  defaults, specify both the `fg` (foreground) and `bg` (background) colors.
#'  The `primary` and `secondary` theme colors are also useful for accenting the
#'  main grayscale colors in things like hyperlinks, tabset panels, and buttons.
#'
#' @section Fonts:
#'
#'  Use `base_font`, `code_font`, and `heading_font` to control the main
#'  typefaces. These arguments set new defaults for the relevant `font-family`
#'  CSS properties, but don't necessarily import the relevant font files. To
#'  both set CSS properties _and_ import font files, consider using the various
#'  [font_face()] helpers.
#'
#'  Each `*_font` argument may be collection of character vectors,
#'  [font_google()]s, [font_link()]s and/or [font_face()]s. Note that a
#'  character vector can have:
#'    * A single unquoted name (e.g., `"Source Sans Pro"`).
#'    * A single quoted name (e.g., `"'Source Sans Pro'"`).
#'    * A comma-separated list of names w/ individual names quoted as necessary.
#'      (e.g. `c("Open Sans", "'Source Sans Pro'", "'Helvetica Neue', Helvetica, sans-serif")`)
#'
#'  Since `font_google(..., local = TRUE)` guarantees that the client has access to
#'  the font family, meaning it's relatively safe to specify just one font
#'  family, for instance:
#'
#'  ```
#'  bs_theme(base_font = font_google("Pacifico", local = TRUE))
#'  ```
#'
#'  However, specifying multiple "fallback" font families is recommended,
#'  especially when relying on remote and/or system fonts being available, for
#'  instance. Fallback fonts are useful not only for handling missing fonts, but
#'  also for handling a Flash of Invisible Text (FOIT) which can be quite
#'  noticeable with remote web fonts on a slow internet connection.
#'
#'  ```
#'  bs_theme(base_font = font_collection(font_google("Pacifico", local = FALSE), "Roboto", "sans-serif"))
#'  ````
#'
#' @param version The major version of Bootstrap to use (see [versions()]
#'   for possible values). Defaults to the currently recommended version
#'   for new projects (currently Bootstrap 5).
#' @param bootswatch The name of a bootswatch theme (see [bootswatch_themes()]
#'   for possible values). When provided to `bs_theme_update()`, any previous
#'   Bootswatch theme is first removed before the new one is applied (use
#'   `bootswatch = "default"` to effectively remove the Bootswatch theme).
#' @param ... arguments passed along to [bs_add_variables()].
#' @param bg A color string for the background.
#' @param fg A color string for the foreground.
#' @param primary A color to be used for hyperlinks, to indicate primary/default
#'   actions, and to show active selection state in some Bootstrap components.
#'   Generally a bold, saturated color that contrasts with the theme's base
#'   colors.
#' @param secondary A color for components and messages that don't need to stand
#'   out. (Not supported in Bootstrap 3.)
#' @param success A color for messages that indicate an operation has succeeded.
#'   Typically green.
#' @param info A color for messages that are informative but not critical.
#'   Typically a shade of blue-green.
#' @param warning A color for warning messages. Typically yellow.
#' @param danger A color for errors. Typically red.
#' @param base_font The default typeface.
#' @param code_font The typeface to be used for code. Be sure this is monospace!
#' @param heading_font The typeface to be used for heading elements.
#' @param font_scale A scalar multiplier to apply to the base font size. For
#'   example, a value of `1.5` scales font sizes to 150% and a value of `0.8`
#'   scales to 80%. Must be a positive number.
#'
#' @return a [sass::sass_bundle()] (list-like) object.
#'
#' @references \url{https://rstudio.github.io/bslib/articles/bslib.html}
#' @references \url{https://rstudio.github.io/sass/}
#' @seealso [bs_add_variables()], [bs_theme_preview()]
#' @examples
#'
#' theme <- bs_theme(
#'   # Controls the default grayscale palette
#'   bg = "#202123", fg = "#B8BCC2",
#'   # Controls the accent (e.g., hyperlink, button, etc) colors
#'   primary = "#EA80FC", secondary = "#48DAC6",
#'   base_font = c("Grandstander", "sans-serif"),
#'   code_font = c("Courier", "monospace"),
#'   heading_font = "'Helvetica Neue', Helvetica, sans-serif",
#'   # Can also add lower-level customization
#'   "input-border-color" = "#EA80FC"
#' )
#' if (interactive()) {
#'   bs_theme_preview(theme)
#' }
#'
#' # Lower-level bs_add_*() functions allow you to work more
#' # directly with the underlying Sass code
#' theme <- bs_add_variables(theme, "my-class-color" = "red")
#' theme <- bs_add_rules(theme, ".my-class { color: $my-class-color }")
#'
#' @export
bs_theme <- function(version = version_default(), bootswatch = NULL, ...,
                     bg = NULL, fg = NULL, primary = NULL, secondary = NULL,
                     success = NULL, info = NULL, warning = NULL, danger = NULL,
                     base_font = NULL, code_font = NULL, heading_font = NULL,
                     font_scale = NULL) {

  theme <- bs_bundle(
    bs_theme_init(version, bootswatch),
    bootstrap_bundle(version),
    bootswatch_bundle(bootswatch, version)
  )
  bs_theme_update(
    theme, ...,
    bg = bg, fg = fg,
    primary = primary,
    secondary = secondary,
    success = success,
    info = info,
    warning = warning,
    danger = danger,
    base_font = base_font,
    code_font = code_font,
    heading_font = heading_font,
    font_scale = font_scale
  )
}

#' @rdname bs_theme
#' @param theme a [bs_theme()] object.
#' @export
bs_theme_update <- function(theme, ..., bootswatch = NULL, bg = NULL, fg = NULL,
                            primary = NULL, secondary = NULL, success = NULL,
                            info = NULL, warning = NULL, danger = NULL,
                            base_font = NULL, code_font = NULL, heading_font = NULL,
                            font_scale = NULL) {
  assert_bs_theme(theme)

  if (!is.null(bootswatch)) {
    old_swatch <- theme_bootswatch(theme)
    # You're only allowed one Bootswatch theme!
    if (length(old_swatch)) {
      theme <- bs_remove(theme, "bootswatch")
      class(theme) <- setdiff(class(theme), bootswatch_class(old_swatch))
    }
    if (!identical(bootswatch, "default")) {
      theme <- add_class(theme, bootswatch_class(bootswatch))
      theme <- bs_bundle(theme, bootswatch_bundle(bootswatch, theme_version(theme)))
    }
  }
  # See R/bs-theme-update.R for the implementation of these
  theme <- bs_base_colors(theme, bg = bg, fg = fg)
  theme <- bs_accent_colors(
    theme, primary = primary, secondary = secondary, success = success,
    info = info, warning = warning, danger = danger
  )
  theme <- bs_fonts(theme, base = base_font, code = code_font, heading = heading_font)
  if (!is.null(font_scale)) {
    stopifnot(is.numeric(font_scale) && length(font_scale) == 1)
    theme <- bs_add_variables(
      theme, "font-size-base" = paste(
        font_scale, "*", bs_get_variables(theme, "font-size-base")
      )
    )
  }
  bs_add_variables(theme, ...)
}

#' @rdname bs_global_theme
#' @export
bs_global_theme_update <- function(..., bootswatch = NULL, bg = NULL, fg = NULL,
                                   primary = NULL,  secondary = NULL, success = NULL,
                                   info = NULL, warning = NULL, danger = NULL,
                                   base_font = NULL, code_font = NULL, heading_font = NULL) {
  theme <- assert_global_theme("bs_theme_global_update()")
  bs_global_set(bs_theme_update(
    theme, ...,
    bg = bg, fg = fg,
    primary = primary,
    secondary = secondary,
    success = success,
    info = info,
    warning = warning,
    danger = danger,
    base_font = base_font,
    code_font = code_font,
    heading_font = heading_font
  ))
}

#' @rdname bs_theme
#' @param x an object.
#' @export
is_bs_theme <- function(x) {
  inherits(x, "bs_theme")
}

# Start an empty bundle with special classes that
# theme_version() & theme_bootswatch() search for
bs_theme_init <- function(version, bootswatch = NULL) {
  add_class(
    sass_layer(defaults = list("bootstrap-version" = version)),
    c(
      bootswatch_class(bootswatch),
      paste0("bs_version_", version),
      "bs_theme"
    )
  )
}

bootswatch_class <- function(bootswatch = NULL) {
  if (is.null(bootswatch)) NULL else paste0("bs_bootswatch_", bootswatch)
}

assert_bs_theme <- function(theme) {
  if (!is_bs_theme(theme)) {
    stop("`theme` must be a `bs_theme()` object")
  }
  invisible(theme)
}

# -----------------------------------------------------------------
# Core Bootstrap bundle
# -----------------------------------------------------------------

bootstrap_bundle <- function(version) {
  pandoc_tables <- list(
    # Pandoc uses align attribute to align content but BS4 styles take precedence...
    # we may want to consider adopting this more generally in "strict" BS4 mode as well
    ".table th[align=left] { text-align: left; }",
    ".table th[align=right] { text-align: right; }",
    ".table th[align=center] { text-align: center; }"
  )

  main_bundle <- switch_version(
    version,
    five = sass_bundle(
      # Don't name this "core" bundle so it can't easily be removed
      sass_layer(
        functions = bs5_sass_files("functions"),
        defaults = bs5_sass_files("variables"),
        mixins = bs5_sass_files("mixins")
      ),
      # Returns a _named_ list of bundles (i.e., these should be easily removed)
      !!!rule_bundles(
        # Names here should match https://github.com/twbs/bs5/blob/master/scss/bootstrap.scss
        bs5_sass_files(c(
          "utilities",
          "root", "reboot", "type", "images", "containers", "grid",
          "tables", "forms", "buttons", "transitions", "dropdown",
          "button-group", "nav", "navbar", "card", "accordion", "breadcrumb",
          "pagination", "badge", "alert", "progress", "list-group", "close",
          "toasts", "modal", "tooltip", "popover", "carousel", "spinners",
          "offcanvas", "placeholders", "helpers", "utilities/api"
        ))
      ),
      # Additions to BS5 that are always included (i.e., not a part of compatibility)
      sass_layer(rules = pandoc_tables),
      bs3compat = bs3compat_bundle()
    ),
    four = sass_bundle(
      sass_layer(
        functions = bs4_sass_files(c("deprecated", "functions")),
        defaults = bs4_sass_files("variables"),
        mixins = bs4_sass_files("mixins")
      ),
      # Returns a _named_ list of bundles (i.e., these should be easily removed)
      !!!rule_bundles(
        # Names here should match https://github.com/twbs/bs4/blob/master/scss/bootstrap.scss
        bs4_sass_files(c(
          "root", "reboot", "type", "images", "code", "grid", "tables",
          "forms", "buttons", "transitions", "dropdown", "button-group",
          "input-group", "custom-forms", "nav", "navbar", "card",
          "breadcrumb", "pagination", "badge", "jumbotron", "alert",
          "progress", "media", "list-group", "close", "toasts", "modal",
          "tooltip", "popover", "carousel", "spinners", "utilities", "print"
        ))
      ),
      # Additions to BS4 that are always included (i.e., not a part of compatibility)
      sass_layer(rules = pandoc_tables),
      bs3compat = bs3compat_bundle()
    ),
    three = sass_bundle(
      sass_layer(
        defaults = bs3_sass_files("variables"),
        mixins = bs3_sass_files("mixins")
      ),
      # Should match https://github.com/twbs/bootstrap-sass/blob/master/assets/stylesheets/_bootstrap.scss
      !!!rule_bundles(
        bs3_sass_files(c(
          "normalize", "print", "glyphicons", "scaffolding", "type", "code", "grid",
          "tables", "forms", "buttons", "component-animations", "dropdowns", "button-groups",
          "input-groups", "navs", "navbar", "breadcrumbs", "pagination", "pager", "labels",
          "badges", "jumbotron", "thumbnails", "alerts", "progress-bars", "media",
          "list-group", "panels", "responsive-embed", "wells", "close", "modals",
          "tooltip", "popovers", "carousel", "utilities", "responsive-utilities"
        ))
      ),
      accessibility = bs3_accessibility_bundle(),
      glyphicon_font_files = sass_layer(
        defaults = list("icon-font-path" = "'glyphicon-fonts/'"),
        file_attachments = c(
          "glyphicon-fonts" = lib_file("bs3", "assets", "fonts", "bootstrap")
        )
      )
    )
  )

  sass_bundle(
    main_bundle,
    # color-contrast() was introduced in Bootstrap 5.
    # We include our own version for a few reasons:
    # 1. Easily turn off warnings options(bslib.color_contrast_warnings=F)
    # 2. Allow Bootstrap 3 & 4 to use color-contrast() in variable definitions
    # 3. Allow Bootstrap 3 & 4 to use bs_get_contrast()
    sass_layer(
      functions = sass_file(system_file("sass-utils/color-contrast.scss", package = "bslib"))
    ),
    # nav_spacer() CSS (can be removed)
    nav_spacer = sass_layer(
      rules = sass_file(system_file("nav-spacer/nav-spacer.scss", package = "bslib"))
    )
  )
}


bootstrap_javascript_map <- function(version) {
  switch_version(
    version,
    five = lib_file("bs5", "dist", "js", "bootstrap.bundle.min.js.map"),
    four = lib_file("bs4", "dist", "js", "bootstrap.bundle.min.js.map")
  )
}
bootstrap_javascript <- function(version) {
  switch_version(
    version,
    five = lib_file("bs5", "dist", "js", "bootstrap.bundle.min.js"),
    four = lib_file("bs4", "dist", "js", "bootstrap.bundle.min.js"),
    three = lib_file("bs3", "assets", "javascripts", "bootstrap.min.js")
  )
}


# -----------------------------------------------------------------
# BS3 compatibility bundle
# -----------------------------------------------------------------

bs3compat_bundle <- function() {
  sass_layer(
    defaults = sass_file(system_file("bs3compat", "_defaults.scss", package = "bslib")),
    mixins = sass_file(system_file("bs3compat", "_declarations.scss", package = "bslib")),
    rules = sass_file(system_file("bs3compat", "_rules.scss", package = "bslib")),
    # Gyliphicon font files
    file_attachments = c(
      fonts = lib_file("bs3", "assets", "fonts")
    ),
    html_deps = htmltools::htmlDependency(
      "bs3compat", packageVersion("bslib"),
      package = "bslib",
      src = "bs3compat/js",
      script = c("transition.js", "tabs.js", "bs3compat.js")
    )
  )
}

# -----------------------------------------------------------------
# BS3 accessibility bundle
# -----------------------------------------------------------------

bs3_accessibility_bundle <- function() {
  sass_layer(
    rules = sass_file(
      system_file(
        "lib", "bs-a11y-p",
        "src", "sass", "bootstrap-accessibility.scss",
        package = "bslib"
      )
    ),
    html_deps = htmltools::htmlDependency(
      "bootstrap-accessibility", version_accessibility,
      package = "bslib", src = "lib/bs-a11y-p",
      script = "plugins/js/bootstrap-accessibility.min.js",
      all_files = FALSE
    )
  )
}

# -----------------------------------------------------------------
# Bootswatch bundle
# -----------------------------------------------------------------

bootswatch_bundle <- function(bootswatch, version) {
  if (!length(bootswatch) || isTRUE(bootswatch %in% c("default", "bootstrap"))) {
    return(NULL)
  }

  bootswatch <- switch_version(
    version,
    default = {
      switch(
        bootswatch,
        paper = {
          message("Bootswatch 3 theme paper has been renamed to materia in version 4 (using that theme instead)")
          "materia"
        },
        readable = {
          message("Bootswatch 3 theme readable has been renamed to litera in version 4 (using that theme instead)")
          "litera"
        },
        match.arg(bootswatch, bootswatch_themes(version))
      )
    },
    three = match.arg(bootswatch, bootswatch_themes(version))
  )

  # Attach local font files, if necessary
  font_css <- file.path(bootswatch_dist(version), bootswatch, "font.css")
  attachments <- if (file.exists(font_css)) {
    c(
      "font.css" = font_css,
      fonts = system_file("fonts", package = "bslib")
    )
  }

  sass_bundle(
    bootswatch = sass_layer(
      file_attachments = attachments,
      defaults = list(
        # Use local fonts (this path is relative to the bootstrap HTML dependency dir)
        '$web-font-path: "font.css" !default;',
        bootswatch_sass_file(bootswatch, "variables", version),
        # Unless we change navbarPage()'s markup, BS4+ will likely want BS3 compatibility
        switch_version(
          version, three = "", default = bs3compat_navbar_defaults(bootswatch)
        )
      ),
      rules = list(
        bootswatch_sass_file(bootswatch, "bootswatch", version),
        # For some reason sketchy sets .dropdown-menu{overflow: hidden}
        # but this prevents .dropdown-submenu from working properly
        # https://github.com/rstudio/bootscss/blob/023d455/inst/node_modules/bootswatch/dist/sketchy/_bootswatch.scss#L204
        if (identical(bootswatch, "sketchy")) ".dropdown-menu{ overflow: inherit; }" else "",
        # Several Bootswatch themes (e.g., zephyr, simplex, etc) add custom .btn-secondary
        # rules that should also apply to .btn-default
        ".btn-default:not(.btn-primary):not(.btn-info):not(.btn-success):not(.btn-warning):not(.btn-danger):not(.btn-dark):not(.btn-outline-primary):not(.btn-outline-info):not(.btn-outline-success):not(.btn-outline-warning):not(.btn-outline-danger):not(.btn-outline-dark) {
          @extend .btn-secondary !optional;
        }"
      )
    )
  )
}


# Mappings from BS3 navbar classes to BS4
bs3compat_navbar_defaults <- function(bootswatch) {
  # Do nothing if this isn't a Bootswatch 3 theme
  if (!bootswatch %in% c("materia", "litera", bootswatch_themes(3))) {
    return("")
  }

  bg_colors <- switch(
    bootswatch,
    cerulean = c("primary", "info"),
    cosmo = c("dark", "primary"),
    cyborg = c("body-bg", "secondary"),
    darkly = c("primary", "success"),
    flatly = c("primary", "success"),
    journal = c("light", "primary"),
    lumen = c("light", "white"),
    # i.e., materia
    paper = ,
    materia = c("light", "primary"),
    readable = ,
    litera = c("light", "dark"),
    sandstone = c("dark", "success"),
    simplex = c("light", "primary"),
    slate = c("primary", "light"),
    spacelab = c("light", "primary"),
    superhero = c("dark", "primary"),
    united = c("primary", "dark"),
    yeti = c("dark", "primary"),
    stop("Didn't recognize Bootswatch 3 theme: ", bootswatch, call. = FALSE)
  )


  list(
    sprintf('$navbar-light-bg: $%s !default;', bg_colors[1]),
    sprintf('$navbar-dark-bg: $%s !default;', bg_colors[2])
  )
}

test_that("bs_theme() and bs_theme_update() add and remove preset class", {
  theme <- bs_theme(5, preset = "bootstrap")
  expect_s3_class(theme, "bs_theme")
  expect_no_match(class(theme), THEME_PRESET_CLASS)

  # Adds preset class when adding a new preset
  theme_bsw_cyborg <- bs_theme_update(theme, preset = "cyborg")
  expect_s3_class(theme_bsw_cyborg, "bs_theme")
  expect_s3_class(theme_bsw_cyborg, THEME_PRESET_CLASS)

  # Keeps preset theme class when preset changes
  expect_s3_class(
    bs_theme_update(theme_bsw_cyborg, preset = "lumen"),
    THEME_PRESET_CLASS
  )

  # Drops preset theme class when preset is removed
  expect_no_match(
    class(bs_theme_update(theme_bsw_cyborg, preset = "bootstrap")),
    THEME_PRESET_CLASS
  )
})

test_that("lib_file returns multiple system files", {

  # version 4 has a js file and a js map file
  js_files <- bootstrap_javascript("4")
  expect_equal(length(js_files), 1)
  js_files <- bootstrap_javascript("3")
  expect_equal(length(js_files), 1)
  js_files <- bootstrap_javascript_map("4")
  expect_equal(length(js_files), 1)
  js_files <- bootstrap_javascript_map("3")
  expect_equal(length(js_files), 0)
})

test_that("Can retrieve version from theme object", {
  expect_null(theme_bootswatch(bs_theme()))

  theme <- bs_theme(version = "3", bootswatch = "paper")
  expect_equal(
    theme_bootswatch(theme), "paper"
  )

  theme <- bs_theme(version = "4", bootswatch = "materia")
  expect_identical(
    theme_bootswatch(theme), "materia"
  )

  theme <- bs_theme_update(theme, bootswatch = "darkly")
  expect_identical(
    theme_bootswatch(theme), "darkly"
  )

  # Can use default to effectively remove bootswatch
  expect_identical(
    sass::sass(bs_theme_update(theme, bootswatch = "default")),
    sass::sass(bs_theme(version = "4"))
  )

  # Can use default as a way to "explicitly don't use" a Bootswatch theme
  default <- bs_theme(version = "4", bootswatch = "default")
  expect_equal(default, bs_theme(version = "4"), ignore_attr = TRUE)
  expect_equal(theme_bootswatch(default), "default")
})

test_that("Can retrieve version from theme object", {
  expect_null(theme_bootswatch(bs_theme()))

  theme <- bs_theme(version = "3", bootswatch = "paper")
  expect_equal(
    theme_bootswatch(theme),
    "paper"
  )

  theme <- bs_theme(version = "4", bootswatch = "materia")
  expect_identical(
    theme_bootswatch(theme),
    "materia"
  )

  theme <- bs_theme_update(theme, bootswatch = "darkly")
  expect_identical(
    theme_bootswatch(theme),
    "darkly"
  )

  # Can use default to effectively remove bootswatch
  expect_identical(
    sass::sass(bs_theme_update(theme, bootswatch = "default")),
    sass::sass(bs_theme(version = "4"))
  )

  # Can use default as a way to "explicitly don't use" a Bootswatch theme
  default <- bs_theme(version = "4", bootswatch = "default")
  expect_equal(default, bs_theme(version = "4"), ignore_attr = TRUE)
  expect_no_match(class(default), "bs_bootswatch_")
})

# Expect that a font input overrides the relevant Sass default
expect_font_defaults <- function(input, output = input) {
  base <- bs_theme(version = "4", base_font = input)
  expect_equal(unname(bs_get_variables(base, "font-family-base")), output)
  code <- bs_theme(version = "4", code_font = input)
  expect_equal(unname(bs_get_variables(code, "font-family-monospace")), output)
  heading <- bs_theme(version = "4", heading_font = input)
  expect_equal(unname(bs_get_variables(heading, "headings-font-family")), output)
}
# Expect that a font input adds some number (n) of HTML dependencies
expect_new_dependencies <- function(input, n = 1) {
  theme <- bs_theme()
  n_default <- length(bs_theme_dependencies(theme))
  base <- bs_theme_dependencies(
    bs_theme_update(theme, base_font = input)
  )
  code <- bs_theme_dependencies(
    bs_theme_update(theme, code_font = input)
  )
  heading <- bs_theme_dependencies(
    bs_theme_update(theme, heading_font = input)
  )
  if (n_default + n != length(base)) browser()
  expect_equal(n_default + n, length(base))
  expect_equal(n_default + n, length(code))
  expect_equal(n_default + n, length(heading))
}

test_that("Strings are quoted, if needed", {
  expect_font_defaults('"foo-bar"')
  expect_font_defaults("foo bar", '"foo bar"')
  expect_font_defaults('"foo bar"')
  expect_font_defaults('"foo bar", "baz"')
})

test_that("Lists are collapsed into quoted strings", {
  expect_font_defaults(list("foo-bar"), "foo-bar")
  expect_font_defaults(list("foo-bar", "foo bar"), 'foo-bar, "foo bar"')
  expect_font_defaults(list("foo-bar", '"foo bar", baz'), 'foo-bar, "foo bar", baz')
})

test_that("Single font objects set defaults and add dependencies", {
  cg <- font_google("Crimson Pro", local = FALSE)
  cl <- font_link("Crimson Pro", href = "foo")
  cf <- font_face("Crimson Pro", src = "foo")
  expect_font_defaults(cg, '"Crimson Pro"')
  expect_font_defaults(cl, '"Crimson Pro"')
  expect_font_defaults(cf, '"Crimson Pro"')
  expect_new_dependencies(cg)
  expect_new_dependencies(cl)
  expect_new_dependencies(cf)
})


test_that("Mix of font objects and character strings", {
  font <- font_collection(font_google("Pacifico", local = FALSE), "Sans Serif")
  expect_font_defaults(font, 'Pacifico, "Sans Serif"')
  expect_new_dependencies(font)

  font <- font_collection(font, font_link("foo", "bar"))
  expect_font_defaults(font, 'Pacifico, "Sans Serif", foo')
  expect_new_dependencies(font, n = 2)

  font <- font_collection(font, font_face("foo bar", "baz"))
  expect_font_defaults(font, 'Pacifico, "Sans Serif", foo, "foo bar"')
  expect_new_dependencies(font, n = 3)
})


# Note, we don't really test local google fonts here, but that code path
# will be tested in shinycoreci-apps

# Expect that a font input overrides the relevant Sass default
expect_font_defaults <- function(input, output = input) {
  base <- bs_theme(version = "4", base_font = input)
  expect_equal(unname(bs_get_variables(base, "font-family-base")), output)
  code <- bs_theme(version = "4", code_font = input)
  expect_equal(unname(bs_get_variables(code, "font-family-monospace")), output)
  heading <- bs_theme(version = "4", heading_font = input)
  expect_equal(
    unname(bs_get_variables(heading, "headings-font-family")),
    output
  )
}
# Expect that a font input adds some number (n) of HTML dependencies
expect_new_dependencies <- function(input, n = 1) {
  theme <- bs_theme()
  n_default <- length(bs_theme_dependencies(theme))
  base <- bs_theme_dependencies(
    bs_theme_update(theme, base_font = input)
  )
  code <- bs_theme_dependencies(
    bs_theme_update(theme, code_font = input)
  )
  heading <- bs_theme_dependencies(
    bs_theme_update(theme, heading_font = input)
  )
  if (n_default + n != length(base)) {
    browser()
  }
  expect_equal(n_default + n, length(base))
  expect_equal(n_default + n, length(code))
  expect_equal(n_default + n, length(heading))
}

test_that("Strings are quoted, if needed", {
  expect_font_defaults('"foo-bar"')
  expect_font_defaults("foo bar", '"foo bar"')
  expect_font_defaults('"foo bar"')
  expect_font_defaults('"foo bar", "baz"')
})

test_that("Lists are collapsed into quoted strings", {
  expect_font_defaults(list("foo-bar"), "foo-bar")
  expect_font_defaults(list("foo-bar", "foo bar"), 'foo-bar, "foo bar"')
  expect_font_defaults(
    list("foo-bar", '"foo bar", baz'),
    'foo-bar, "foo bar", baz'
  )
})

test_that("Single font objects set defaults and add dependencies", {
  cg <- font_google("Crimson Pro", local = FALSE)
  cl <- font_link("Crimson Pro", href = "foo")
  cf <- font_face("Crimson Pro", src = "foo")
  expect_font_defaults(cg, '"Crimson Pro"')
  expect_font_defaults(cl, '"Crimson Pro"')
  expect_font_defaults(cf, '"Crimson Pro"')
  expect_new_dependencies(cg)
  expect_new_dependencies(cl)
  expect_new_dependencies(cf)
})

test_that("Mix of font objects and character strings", {
  font <- font_collection(font_google("Pacifico", local = FALSE), "Sans Serif")
  expect_font_defaults(font, 'Pacifico, "Sans Serif"')
  expect_new_dependencies(font)

  font <- font_collection(font, font_link("foo", "bar"))
  expect_font_defaults(font, 'Pacifico, "Sans Serif", foo')
  expect_new_dependencies(font, n = 2)

  font <- font_collection(font, font_face("foo bar", "baz"))
  expect_font_defaults(font, 'Pacifico, "Sans Serif", foo, "foo bar"')
  expect_new_dependencies(font, n = 3)
})

# Note, we don't really test local google fonts here, but that code path
# will be tested in shinycoreci-apps

rlang::check_installed("shiny", version = "1.8.1")
rlang::check_installed("bslib", version = "0.8.0.9000")
rlang::check_installed("future")
rlang::check_installed("ggplot2")
rlang::check_installed("markdown")

library(shiny)
library(bslib)
library(ggplot2)

library(future)
plan(multisession)

options(
  bslib.color_contrast_warnings = FALSE
  # shiny.autoreload.pattern = "_brand[.]yml|app[.]R|[.]s?css" ## TODO: Enable after fixing autoreload
)

if (!file.exists("Monda.ttf")) {
  download.file(
    "https://github.com/google/fonts/raw/48db77e32954f6f5e65a7122ecbe8a2093c4f5d7/ofl/monda/Monda%5Bwght%5D.ttf",
    "Monda.ttf"
  )
  download.file(
    "https://github.com/google/fonts/raw/48db77e32954f6f5e65a7122ecbe8a2093c4f5d7/ofl/monda/OFL.txt",
    "Monda-OFL.txt"
  )
}

theme_brand <- bs_theme(brand = TRUE)

brand <- attr(theme_brand, "brand")

theme_set(theme_minimal())

if (requireNamespace("thematic", quietly = TRUE)) {
  if (!is.null(brand)) {
    # TODO: Update plot fonts dynamically
    thematic::thematic_shiny(
      font = bslib:::brand_pluck(brand, "typography", "base", "family")
    )
  } else {
    thematic::thematic_shiny()
  }
}

is_app_hosted <-
  Sys.getenv("R_CONFIG_ACTIVE") %in%
    c("shinylive", "shinyapps", "rsconnect", "rstudio_cloud")
is_app_packaged <-
  getwd() != system.file("examples-shiny/brand.yml", package = "bslib")

ui <- page_navbar(
  theme = bs_add_rules(theme_brand, sass::sass_file("_colors.scss")),
  title = "brand.yml Demo",
  fillable = TRUE,

  sidebar = sidebar(
    id = "sidebar_editor",
    position = "right",
    open = "closed",
    width = "40%",
    bg = "var(--bs-dark)",
    fg = "var(--bs-light)",

    card(
      card_header(
        class = "text-bg-secondary hstack",
        div("Edit", code("brand.yml")),
        div(
          class = "ms-auto",
          tooltip(
            tags$a(
              class = "btn btn-link p-0",
              href = "https://posit-dev.github.io/brand-yml/brand/",
              target = "_blank",
              bsicons::bs_icon(
                "question-square-fill",
                title = "About brand.yml",
                size = "1.25rem"
              )
            ),
            "About brand.yml"
          )
        )
      ),
      htmltools::tagAppendAttributes(
        textAreaInput(
          "txt_brand_yml",
          label = NULL,
          value = paste(readLines("_brand.yml", warn = FALSE), collapse = "\n"),
          width = "100%",
          height = "80%",
          rows = 20
        ),
        class = "font-monospace",
        .cssSelector = "textarea"
      ),
      card_body(
        padding = 0,
        div(
          id = "editor_brand_yml",
          style = "overflow: auto;",
          as_fill_item()
        )
      )
    ),

    tags$script(
      type = "module",
      HTML(
        '
import { basicEditor } from "https://esm.sh/prism-code-editor@3.4.0/setups";
import "https://esm.sh/prism-code-editor@3.4.0/prism/languages/yaml";

const shinyInput = document.getElementById("txt_brand_yml");

function initBrandEditor() {
  if (typeof Shiny.setInputValue !== "function") {
    setTimeout(initBrandEditor, 100);
    return;
  }
  window.brandEditor = basicEditor(
    "#editor_brand_yml",
    {
      language: "yml",
      theme: "github-dark",
      value: shinyInput.value,
      onUpdate: (value) => {
        Shiny.setInputValue("txt_brand_yml", value);
      },
    },
    () => shinyInput.parentElement.parentElement.remove()
  );
}

initBrandEditor();
'
      )
    ),

    tags$style(
      HTML(
        '
.bslib-sidebar-layout .sidebar-title { margin-bottom: 0 }
#sidebar_editor .sidebar-content { height: max(600px, 100%) }'
      )
    ),

    if (is_app_hosted || is_app_packaged) {
      shiny::downloadButton(
        "download",
        label = span("Download", code("_brand.yml"), "file"),
        class = "btn-outline-light"
      )
    } else {
      actionButton(
        "save",
        label = span("Save", code("_brand.yml"), "file"),
        class = "btn-outline-light"
      )
    }
  ),

  nav_panel(
    "Input Output Demo",
    value = "dashboard",
    layout_sidebar(
      sidebar = sidebar(
        sliderInput("slider1", "Numeric Slider Input", 0, 11, 11),
        numericInput("numeric1", "Numeric Input Widget", 30),
        dateInput("date1", "Date Input Component", value = "2024-01-01"),
        input_switch("switch1", "Binary Switch Input", value = TRUE),
        radioButtons(
          "radio1",
          "Radio Button Group",
          choices = c("Option A", "Option B", "Option C", "Option D")
        ),
        actionButton("action1", "Action Button")
      ),
      shiny::useBusyIndicators(),
      layout_column_wrap(
        value_box(
          title = "Metric 1",
          value = "100",
          theme = "primary",
          id = "value_box_one"
        ),
        value_box(
          title = "Metric 2",
          value = "200",
          theme = "secondary",
          id = "value_box_two"
        ),
        value_box(
          title = "Metric 3",
          value = "300",
          theme = "info",
          id = "value_box_three"
        )
      ),
      card(
        card_header("Plot Output"),
        plotOutput("out_plot")
      ),
      card(
        card_header("Text Output"),
        verbatimTextOutput("out_text")
      )
    )
  ),

  nav_panel(
    "Widget Gallery",
    layout_column_wrap(
      width = 300,
      heights_equal = "row",
      card(
        card_header("Button Variants"),
        actionButton("btn_default", "Default"),
        actionButton("btn_primary", "Primary", class = "btn-primary"),
        actionButton("btn_secondary", "Secondary", class = "btn-secondary"),
        actionButton("btn_success", "Success", class = "btn-success"),
        actionButton("btn_danger", "Danger", class = "btn-danger"),
        actionButton("btn_warning", "Warning", class = "btn-warning"),
        actionButton("btn_info", "Info", class = "btn-info")
      ),
      card(
        card_header("Radio Button Examples"),
        radioButtons(
          "radio2",
          "Standard Radio Group",
          choices = c("Selection 1", "Selection 2", "Selection 3")
        ),
        radioButtons(
          "radio3",
          "Inline Radio Group",
          choices = c("Option 1", "Option 2", "Option 3"),
          inline = TRUE
        )
      ),
      card(
        card_header("Checkbox Examples"),
        checkboxGroupInput(
          "check1",
          "Standard Checkbox Group",
          choices = c("Item 1", "Item 2", "Item 3")
        ),
        checkboxGroupInput(
          "check2",
          "Inline Checkbox Group",
          choices = c("Choice A", "Choice B", "Choice C"),
          inline = TRUE
        )
      ),
      card(
        card_header("Select Input Widgets"),
        selectizeInput(
          "select1",
          "Selectize Input",
          choices = c("Selection A", "Selection B", "Selection C")
        ),
        selectInput(
          "select2",
          "Multiple Select Input",
          choices = c("Item X", "Item Y", "Item Z"),
          multiple = TRUE
        )
      ),
      card(
        card_header("Text Input Widgets"),
        textInput("text1", "Text Input"),
        textAreaInput(
          "textarea1",
          "Text Area Input",
          value = "Default text content for the text area widget"
        ),
        passwordInput("password1", "Password Input")
      )
    )
  ),

  nav_panel(
    "Colors",
    div(
      class = "container-sm overflow-y-auto",
      uiOutput("ui_colors")
    )
  ),

  nav_panel(
    "Documentation",
    div(
      class = "container-sm overflow-y-auto",
      if (FALSE) library(markdown), # for shinyapps.io
      includeMarkdown("documentation.md")
    )
  ),

  nav_spacer(),
  nav_item(input_dark_mode(id = "color_mode")),
  nav_item(
    actionLink(
      "show_editor",
      bsicons::bs_icon(
        "pencil-fill",
        size = "1rem",
        title = "Show/hide editor"
      ),
      class = "nav-link"
    )
  ),
)

errors <- rlang::new_environment()

error_notification <- function(context) {
  function(err) {
    time <- as.character(Sys.time())

    msg <- conditionMessage(err)
    # Strip ANSI color sequences from error messages
    msg <- gsub(
      pattern = "\u001b\\[.*?m",
      replacement = "",
      msg
    )
    # Wrap at 40 characters
    msg <- paste(strwrap(msg, width = 60), collapse = "\n")

    err_id <- rlang::hash(list(time, msg))
    assign(err_id, list(message = msg, context = context), envir = errors)

    showNotification(
      markdown(context),
      action = tags$button(
        class = "btn btn-outline-danger pull-right",
        onclick = sprintf(
          "event.preventDefault(); Shiny.setInputValue('show_error', '%s')",
          err_id
        ),
        "Show details"
      ),
      duration = 10,
      type = "error",
      id = err_id
    )
  }
}

server <- function(input, output, session) {
  brand_yml_text <- debounce(reactive(input$txt_brand_yml), 1000)
  brand_yml <- reactiveVal()

  observeEvent(input$show_editor, sidebar_toggle("sidebar_editor"))

  observeEvent(input$show_error, {
    req(input$show_error)
    err <- get0(input$show_error, errors)

    if (is.null(err)) {
      message("Could not find error with id ", input$show_error)
      return()
    }

    removeNotification(input$show_error)
    rm(list = input$show_error, envir = errors)

    showModal(
      modalDialog(
        size = "l",
        easyClose = TRUE,
        markdown(err$context),
        pre(err$message)
      )
    )
  })

  observeEvent(brand_yml_text(), {
    req(brand_yml_text())

    tryCatch(
      {
        b <- yaml::yaml.load(brand_yml_text())
        b$path <- normalizePath("_brand.yml")
        brand_yml(b)
      },
      error = error_notification(
        "Could not parse `_brand.yml` file. Check for syntax errors."
      )
    )
  })

  observeEvent(brand_yml(), {
    req(brand_yml())

    tryCatch(
      {
        theme <- bs_theme(brand = brand_yml())
        theme <- bs_add_rules(theme, sass::sass_file("_colors.scss"))
        session$setCurrentTheme(theme)
      },
      error = error_notification(
        "Could not compile branded theme. Please check your `_brand.yml` file."
      )
    )
  })

  observeEvent(input$save, {
    validate(
      need(input$txt_brand_yml, "_brand.yml file contents cannot be empty.")
    )

    tryCatch(
      {
        writeLines(input$txt_brand_yml, "_brand.yml")
        showNotification(markdown("Saved `_brand.yml`!"))
      },
      error = error_notification("Could not save `_brand.yml`.")
    )
  })

  output$download <- downloadHandler(
    filename = "_brand.yml",
    content = function(file) {
      validate(
        need(input$txt_brand_yml, "_brand.yml file contents cannot be empty.")
      )
      writeLines(input$txt_brand_yml, file)
    }
  )

  PlotTask <- ExtendedTask$new(function(x_max, y_factor) {
    x <- seq(0, x_max, length.out = 100)
    y <- sin(x) * y_factor

    future({
      Sys.sleep(3)

      df <- data.frame(x = x, y = y)

      ggplot(df, aes(x = x, y = y)) +
        geom_col(width = 1, position = "identity") +
        labs(title = "Sine Wave Output", x = "", y = "")
    })
  })

  observe({
    x_max <- debounce(reactive(input$numeric1), 500)()
    y_factor <- debounce(reactive(input$slider1), 500)()

    PlotTask$invoke(x_max = x_max, y_factor = y_factor)
  })

  output$out_plot <- renderPlot({
    PlotTask$result()
  })

  output$out_text <- renderText({
    "example_function <- function() {\n  return(\"Function output text\")\n}"
  })

  output$ui_colors <- renderUI({
    bootstrap_colors <- c(
      "blue",
      "indigo",
      "purple",
      "pink",
      "red",
      "orange",
      "yellow",
      "green",
      "teal",
      "cyan"
    )
    colors <- c("gray", bootstrap_colors)

    tagList(
      layout_columns(
        col_widths = 3,
        class = "font-monospace",
        !!!lapply(
          c(
            "primary",
            "secondary",
            "dark",
            "light",
            "info",
            "success",
            "warning",
            "danger"
          ),
          function(color) {
            div(
              color,
              class = paste0("p-3 mb-2 position-relative text-bg-", color)
            )
          }
        )
      ),
      layout_columns(
        col_widths = 3,
        class = "font-monospace",
        !!!lapply(
          c("black", "white", "foreground", "background"),
          function(color) {
            div(
              color,
              class = paste0("p-3 mb-2 position-relative bd-", color)
            )
          }
        )
      ),
      layout_column_wrap(
        width = 200,
        !!!lapply(colors, function(color) {
          if (!color %in% c("white", "black")) {
            div(
              class = "mb-3",
              div(
                color,
                class = paste0("p-3 mb-2 position-relative bd-", color, "-500")
              ),
              lapply(seq(100, 900, 100), function(r) {
                div(
                  paste0(color, "-", r),
                  class = paste0("p-3 bd-", color, "-", r)
                )
              })
            )
          }
        })
      )
    )
  })
}

shinyApp(ui, server)

library(shiny)
library(ggplot2)
library(bslib)
library(rlang)
library(curl)

# enlarged auto fonts
if (is_installed("thematic")) {
  thematic::thematic_shiny(
    font = thematic::font_spec("auto", scale = 2, update = TRUE)
  )
}

# Source in ggplot2 examples
source("global.R")

theme <- bs_global_get() %||% bs_theme()
if ("3" %in% theme_version(theme)) {
  warning("This example app requires Bootstrap 4 or higher", call. = FALSE)
}

rounded <- isTRUE(as.logical(bs_get_variables(theme %||% bslib::bs_theme(), "enable-rounded")))
pill <- function(...) {
  shiny::tabPanel(..., class = "p-3 border", class = if (rounded) "rounded")
}
tab <- function(...) {
  shiny::tabPanel(..., class = "p-3 border border-top-0", class = if (rounded) "rounded-bottom")
}
gradient <- function(theme_color = "primary") {
  bg_color <- paste0("bg-", theme_color)
  bgg_color <- if ("4" %in% theme_version(theme)) {
    paste0("bg-gradient-", theme_color)
  } else {
    paste(bg_color, "bg-gradient")
  }
  bg_div <- function(color_class, ...) {
    display_classes <- paste(
      paste0(".", strsplit(color_class, "\\s+")[[1]]),
      collapse = " "
    )
    div(
      class = "p-3", class = color_class,
      display_classes, ...
    )
  }
  fluidRow(
    column(6, bg_div(bg_color)),
    column(6, bg_div(bgg_color))
  )
}

theme_colors <- c("primary", "secondary", "default", "success", "info", "warning", "danger", "dark")
gradients <- lapply(theme_colors, gradient)

progressBar <- div(
  class="progress",
  div(
    class="progress-bar w-25",
    role="progressbar",
    "aria-valuenow"="25",
    "aria-valuemin"="0",
    "aria-valuemax"="100"
  )
)

# This is here for shinycoreci to take advantage of (so we don't need to update a bunch of screenshots)
IS_LEGACY <- as.logical(Sys.getenv("BSLIB_LEGACY_THEMER_APP", FALSE))
is_legacy_version <- theme_version(theme) %in% c("3", "4")
if (isTRUE(IS_LEGACY || is_legacy_version)) {
  dashboardTab <- NULL
} else {
  dashboardTab <- nav_panel("Dashboard", tipsUI("tips"))
  theme_set(theme_minimal())
}

layout_buttons <- if (is_legacy_version) tags$div else layout_columns

bs_table <- function(x, class = NULL, ...) {
  class <- paste(c("table", class), collapse = " ")
  class <- sprintf('class="%s"', class)
  HTML(knitr::kable(x, format = "html", table.attr = class))
}

shinyApp(
  page_navbar(
    theme = theme,
    title = "Theme demo",
    collapsible = TRUE,
    inverse = !IS_LEGACY,
    id = "navbar",
    fillable = "Dashboard",
    dashboardTab,
    nav_panel(
      "Inputs",
      tabsetPanel(
        type = "pills", id = "inputs",
        pill(
          "inputPanel()",
          inputPanel(
            sliderInput("slider", "sliderInput()", min = 0, max = 100, value = c(30, 70), step = 20),
            selectInput("selectize", "selectizeInput()", choices = state.abb),
            selectInput("selectizeMulti", "selectizeInput(multiple=T)", choices = state.abb, multiple = TRUE),
            dateInput("date", "dateInput()", value = "2020-12-24"),
            dateRangeInput("dateRange", "dateRangeInput()", start = "2020-12-24", end = "2020-12-31")
          ),
          br(),
          textOutput("inputPanelOutputHeader"),
          verbatimTextOutput("inputPanelOutput"),
          br(),
          tags$p("Here are some", tags$code("actionButton()"), "s demonstrating different theme (i.e., accent) colors"),
          layout_buttons( # Regular <div> if IS_LEGACY, otherwise layout_columns
            actionButton("primary", "Primary", icon("product-hunt"), class = "btn-primary m-2"),
            actionButton("secondary", "Secondary (default)", class = "m-2"),
            actionButton("success", "Success", icon("check"), class = "btn-success m-2"),
            actionButton("info", "Info", icon("info"),  class = "btn-info m-2"),
            actionButton("warning", "warning", icon("exclamation"), class = "btn-warning m-2"),
            actionButton(
              "danger", "Danger",
              icon(if (packageVersion("fontawesome") >= "0.3") "triangle-exclamation" else "exclamation-triangle"),
              class = "btn-danger m-2"
            ),
            actionButton("dark", "Dark", icon("moon"), class = "btn-dark m-2"),
            actionButton("light", "Light", icon("sun"), class = "btn-light m-2")
          )
        ),
        pill(
          "wellPanel()",
          wellPanel(
            fluidRow(
              column(
                6,
                selectInput("select", "selectInput()", choices = state.abb, selectize = FALSE),
                selectInput("selectMulti", "selectInput(multiple=T)", choices = state.abb, multiple = TRUE, selectize = FALSE),
                textInput("text", "textInput()", placeholder = "Enter some text"),
                numericInput("numeric", "numericInput()", value = 0)
              ),
              column(
                6,
                passwordInput("password", "passwordInput()", "secret"),
                textAreaInput("textArea", "textAreaInput()", placeholder = "A text area"),
                checkboxInput("check", "checkboxInput()", value = TRUE),
                checkboxGroupInput("checkGroup", "checkboxGroupInput()", choices = c("A", "B")),
                radioButtons("radioButtons", "radioButtons()", choices = c("A", "B"))
              )
            )
          ),
          br(),
          textOutput("wellPanelOutputHeader"),
          br(),
          verbatimTextOutput("wellPanelOutput")
        )
      )
    ),
    nav_panel(
      "Plots",
      uiOutput("thematic_needed"),
      plotOutput("plot"),
      selectizeInput(
        "plot_example", "Choose an example",
        selected = "GeomSmooth",
        choices = list(
          ggplot2 = names(ggplot2_examples),
          lattice = names(lattice_examples),
          base = names(base_examples)
        )
      )
    ),
    nav_panel(
      "Tables",
      tabsetPanel(
        id = "tables",
        tab(
          "DataTables",
          DT::dataTableOutput("DT")
        ),
        tab(
          "Bootstrap",
          h3("Plain tables", class = "mt-2 mb-1"),
          bs_table(mtcars[1:5, 1:5]),
          h3("Striped tables", class = "mt-4 mb-1"),
          bs_table(mtcars[1:5, 1:5], class = "table-striped"),
          h3("Striped tables with hover", class = "mt-4 mb-1"),
          bs_table(mtcars[1:5, 1:5], class = "table-striped table-hover")
        )
      )
    ),
    nav_panel(
      "Notifications",
      tabsetPanel(
        id = "otherNav",
        tab(
          "Messages",
          br(),
          actionButton("showProgress", "Progress", style = "margin: 1rem"),
          actionButton("showModal", "Modals", style = "margin: 1rem"),
          lapply(c("default", "message", "warning", "error"), function(x) {
            X <- tools::toTitleCase(x)
            class <- switch(x, message = "btn-info", warning = "btn-warning", error = "btn-danger")
            actionButton(
              paste0("show", X), paste(X, "notification"),
              class = class, style = "margin: 1rem"
            )
          })
        ),
        tab(
          "Uploads & Downloads",
          br(),
          fileInput("file", "fileInput()"),
          downloadButton("downloadButton", "downloadButton()", style = "margin: 1rem"),
          downloadLink("downloadLink", "downloadLink()", style = "margin: 1rem")
        )
      )
    ),
    nav_panel(
      "Fonts",
      h1("Heading font:", class = "text-primary"),
      hr(class = "bg-primary", style = "height: 5px"),
      h1("Heading 1"),
      h2("Heading 2"),
      h3("Heading 3"),
      h4("Heading 4"),
      h5("Heading 5"),
      h1("Base font:", class = "text-primary"),
      hr(class = "bg-primary", style = "height: 5px"),
      tags$p("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."),
      h1("Code font:", class = "text-primary"),
      hr(class = "bg-primary", style = "height: 5px"),
      tags$pre(
        class = "shiny-text-output",
        style = "white-space: pre-wrap",
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
      )
    ),
    nav_panel(
      "Options",
      if ("4" %in% theme_version(theme)) {
        p(
          "Background color gradients are disabled by default.",
          "Enable them to see the difference here.",
          "If enabled, gradients automatically apply buttons and progress bars, ",
          "but you may also add to a .bg-gradient-* modified class to arbitrary elements."
        )
      } else {
        markdown(
          "`{bslib}` 'extends' Bootstrap's [`background-color` utility classes](https://getbootstrap.com/docs/5.0/utilities/background/) to also sets the foreground `color` to a sensible contrasting color (i.e., no additional `text-*` classes are needed)."
        )
      },
      !!!gradients,
      br(),
      tags$p(
        "With the default settings, enabling of box shadows adds a very subtle and
              barely noticable inner box-shadow to most input widgets. The difference is
              a little more obvious in a progress bar:"
      ),
      progressBar,
      br(),
      tags$p(
        "Rounded corners are enabled by default and apply to numerous components (e.g., ",
        code("tabsetPanel()"), ",", code("wellPanel()"), ", and ", code("actionButton()"), "):",
        actionButton("showProgress2", "Show progress")
      ),
      br(),
      tags$p(
        "Smooth transitions are enabled by default, but if you disable them, progress updating",
        "in ", code("fileInput()"), "and ", code("Progress"), " will appear more staggered",
        "(e.g., click button above)."
      )
    )
  ),
  function(input, output, session) {

    output$DT <- DT::renderDataTable({
      cars <- mtcars
      for (var in c("cyl", "vs", "am", "gear", "carb")) {
        cars[[var]] <- factor(cars[[var]])
      }

      DT::datatable(
        cars,
        style = "auto",
        caption = "A generic table made with the DT package and the DataTables library.",
        filter = "top"
      )
    })

    output$inputPanelOutputHeader <- renderText({
      "Below are the values bound to each input widget above"
    })

    output$inputPanelOutput <- renderPrint({
      str(list(
        sliderInput = input$slider,
        selectizeInput = input$selectize,
        selectizeMultiInput = input$selectizeMulti,
        dateInput = input$date,
        dateRangeInput = input$dateRange
      ))
    })

    output$wellPanelOutputHeader <- renderText({
      "Below are the values bound to each input widget above"
    })

    output$wellPanelOutput <- renderPrint({
      str(list(
        selectInput = input$select,
        selectMultiInput = input$selectMulti,
        textInput = input$text,
        numericInput = input$numeric,
        passwordInput = input$password,
        textAreaInput = input$textArea,
        checkInput = input$check,
        checkGroupInput = input$checkGroup,
        radioButtonsInput = input$radioButtons
      ))
    })

    observeEvent(input$showModal, {
      showModal(modalDialog(
        title = "Somewhat important message",
        "This is a somewhat important message.",
        easyClose = TRUE,
        footer = modalButton("Close")
      ))
    })

    fake_progress <- function(style = "notification") {
      withProgress(
        message = 'Calculation in progress',
        detail = 'This may take a while...',
        value = 0,
        style = style,
        {
          for (i in 1:15) {
            incProgress(
              amount = 1/15,
              detail = paste("Step", i, "of", 15)
            )
            Sys.sleep(0.25)
          }
        })
    }

    observeEvent(input$showProgress, {
      fake_progress()
      # TODO: old progress styling could be improved
      #fake_progress("old")
    })

    observeEvent(input$showProgress2, {
      p <- Progress$new()
      p$set(
        message = 'Calculation in progress',
        detail = 'This may take a while...',
        value = 0.5
      )
    })

    lapply(c("default", "message", "warning", "error"), function(x) {
      X <- tools::toTitleCase(x)
      observeEvent(input[[paste0("show", X)]], {
        q <- parseQueryString(session$clientData$url_search)
        if ("notification-duration" %in% names(q)) {
          duration <- as.numeric(q[["notification-duration"]])
          if (duration <= 0) duration <- NULL
        } else {
          duration <- 5
        }
        showNotification(paste(X, "notification styling"), type = x, duration = duration)
      })
    })

    output$plot <- renderPlot({
      ggplot2_examples[[input$plot_example]] %||%
        eval(lattice_examples[[input$plot_example]]) %||%
        eval(base_examples[[input$plot_example]])
    })

    output$thematic_needed <- renderUI({
      if (bslib:::is_installed("thematic")) return(NULL)

      htmltools::HTML(
        "<span class=\"bg-warning\">&nbsp;!! Install the <a href='https://rstudio.github.io/thematic/'><code>thematic</code></a> package to enable auto-theming of static R plots !!&nbsp;</span>"
      )
    })

    if (!IS_LEGACY) tipsServer("tips")

  }
)

random_values <- tibble::tribble(
  ~title                             , ~value              ,
  "Sales revenue"                    , "$22,456.78"        ,
  "Customer satisfaction"            , "94.5%"             ,
  "Inventory turnover"               , "8/month"           ,
  "Number of employees"              , "75"                ,
  "Website traffic"                  , "12,345 visits/day" ,
  "Customer retention"               , "87%"               ,
  "Average transaction"              , "$56.73"            ,
  "Followers"                        , "10,234"            ,
  "Open rate"                        , "22.6%"             ,
  "Production efficiency"            , "92%"               ,
  "Project completion"               , "17 days"           ,
  "Employee turnover"                , "10%"               ,
  "Market share"                     , "15.2%"             ,
  "ROI"                              , "8.3%"              ,
  "Average response time"            , "3.2 seconds"       ,
  "Energy consumption"               , "345 kwh/month"     ,
  "Customer churn"                   , "5%"                ,
  "Production defect rate"           , "0.6%"              ,
  "Engagement rate"                  , "3.8%"              ,
  "Average wait"                     , "4 minutes"         ,
  "Project budget variance"          , "$2,345.67"         ,
  "Employee productivity"            , "95%"               ,
  "Conversion rate"                  , "2.5%"              ,
  "Website bounce rate"              , "42%"               ,
  "Revenue growth"                   , "+15.2%"            ,
  "Customer loyalty"                 , "4.5 out of 5"      ,
  "Product quality"                  , "8.9/10"            ,
  "Social media followers"           , "50,000+"           ,
  "Market reach"                     , "10.5 million"      ,
  "Employee satisfaction"            , "92.3%"             ,
  "Average order value"              , "$123.45"           ,
  "Website conversion rate"          , "3.6%"              ,
  "Customer support"                 , "9.8/10"            ,
  "Productivity index"               , "120%"              ,
  "Customer lifetime value"          , "$5,000"            ,
  "Brand awareness"                  , "85%"               ,
  "Time to market"                   , "4 weeks"           ,
  "Return on investment"             , "18.5%"             ,
  "Net promoter score"               , "8.2 out of 10"     ,
  "Email click-through rate"         , "12.7%"             ,
  "Supply chain efficiency"          , "92%"               ,
  "Website loading time"             , "2.3 seconds"       ,
  "Cost per acquisition"             , "$25.60"            ,
  "Employee engagement"              , "83%"               ,
  "Innovation index"                 , "9.5/10"            ,
  "Customer complaints resolved"     , "98.6%"             ,
  "Market share growth"              , "+2.3%"             ,
  "Production capacity utilization"  , "87%"               ,
  "Website traffic source diversity" , "5 channels"        ,
  "Brand equity"                     , "$1.2 billion"      ,
  "Code Coverage"                    , "85%"               ,
  "Lines of Code"                    , "46,157"            ,
  "Commits"                          , "339"               ,
  "Review Turnaround"                , "2 days"            ,
  "Community Contributors"           , "34"                ,
  "Project Forks"                    , "36"                ,
  "Open Issues"                      , "87"                ,
  "Open Pull Requests"               , "12"                ,
  "Closed Issues"                    , "19"                ,
  "AWS Cloud Spending"               , "$3,463"            ,
  "Diversity Index"                  , "0.75"              ,
  "API Response Time"                , "150 ms"            ,
  "Daily Active Users"               , "8,507"             ,
  "Project Stars"                    , "5,100"             ,
  "Package Downloads"                , "409,446"           ,
  "Average Response Time"            , "50 ms"             ,
  "Uptime"                           , "99.97%"            ,
  "Developer Satisfaction"           , "4.8/5"             ,
  "Project Funding"                  , "$2,805"
)

random_title_value <- function() {
  as.list(random_values[sample(nrow(random_values), 1), ])
}

#' Obtain a list of all available built-in \pkg{bslib} themes.
#'
#' @param version the major version of Bootstrap.
#' @param full_path whether to return a path to the installed theme.
#' @export
#' @return a character vector of built-in themes provided by \pkg{bslib}.
builtin_themes <- function(version = version_default(), full_path = FALSE) {
  path_builtins <- path_builtin_theme(version = version)
  if (is.null(path_builtins)) return(NULL)

  list.dirs(path_builtins, full.names = full_path, recursive = FALSE)
}

builtin_bundle <- function(name = "shiny", version = version_default()) {
  name <- validate_builtin_preset_name(name, version = version)

  font_css <- path_builtin_theme(name, "font.css", version = version)
  attachments <- if (file.exists(font_css)) {
    c(
      "font.css" = font_css,
      fonts = system_file("fonts", package = "bslib")
    )
  }

  sass_parts <- rlang::set_names(c("variables", "mixins", "rules"))

  sass_files <- lapply(sass_parts, function(part) {
    path <- path_builtin_theme(name, paste0("_", part, ".scss"), version = version)
    if (file.exists(path)) sass_file(path)
  })

  sass_bundle(
    "builtin" = sass_layer(
      file_attachments = attachments,
      defaults = list(
        "bslib-preset-type" = "builtin",
        "bslib-preset-name" = name,
        '$web-font-path: "font.css" !default;',
        sass_files$variables
      ),
      mixins = list(sass_files$mixins),
      rules = list(sass_files$rules)
    )
  )
}

validate_builtin_preset_name <- function(name, version = version_default()) {
  builtin_names <- builtin_themes(version)

  if (name %in% builtin_names) {
    return(name)
  }

  msg <- "'%s' is not a valid built-in theme preset provided by {bslib}."

  info <-
    if (length(builtin_names)) {
      sprintf(
        "Available Bootstrap %s themes are: '%s'",
        version,
        paste0(builtin_themes(version), collapse = "', '"))
    } else {
      "No built-in theme presets are available for this version of Bootstrap."
    }

  rlang::abort(c(sprintf(msg, name), "i" = info))
}

path_builtin_theme <- function(..., version = version_default()) {
  switch_version(
    version,
    five = system.file("builtin", "bs5", ..., package = "bslib")
  )
}

# test_that()

describe("builtin_bundle()", {
  it("errors for unknown preset names", {
    expect_snapshot(
      error = TRUE,
      builtin_bundle("not-a-preset", version = "5")
    )

    expect_snapshot(
      error = TRUE,
      builtin_bundle("not-a-preset", version = "4")
    )
  })

  it("errors for unknown Bootstrap version", {
    expect_error(builtin_bundle("shiny", version = 99))
  })

  it("returns a bundle for a known preset name", {
    bundle <- builtin_bundle("shiny", version = "5")
    expect_s3_class(bundle, "sass_bundle")
    expect_named(bundle$layers, "builtin")
  })
})

if (!identical(getwd(), rprojroot::find_package_root_file())) {
  stop("This script must be run from the top directory of the bslib package")
}

pkgload::load_all()

lapply(c(
  "yarn_install.R",
  "download_bootswatch_fonts.R",
  "update_gfont_info.R",
  "expand_variables_article_template.R"
), function(file) {
  message("Updating: ", file)
  source(file.path("tools", file), local = TRUE)
})

# ----------------------------------------------------------------------------------------
# Base colors
# ----------------------------------------------------------------------------------------

bs_base_colors <- function(theme, bg = NULL, fg = NULL) {
  assert_bs_theme(theme)
  if (is.null(bg) && is.null(fg)) return(theme)
  if (is.null(bg)) stop("Cannot specify bg without fg.")
  if (is.null(fg)) stop("Cannot specify fg without bg.")

  args <- validate_and_normalize_colors(list(bg = bg, fg = fg))
  # In some cases, bg/fg really means $body-bg/$body-color, not $white/$black
  use_body <- has_body_base_colors(theme)
  if (use_body) {
    args <- rename2(args, !!!get_base_color_map(theme))
  }

  switch_add_variables(
    theme, args,
    default = if (use_body) identity else bs4_base_colors,
    three = if (use_body) identity else bs3_base_colors
  )
}

switch_add_variables <- function(theme, args, ...) {
  func <- switch_version(theme, ...)
  vars <- do.call(func, list(args))
  bs_add_variables(theme, !!!vars)
}

# Obtain a mapping from (to) fg/bg to (from) relevant Sass vars
get_base_color_map <- function(theme, decode = TRUE) {
  use_body <- has_body_base_colors(theme)
  vars <- switch_version(
    theme,
    three = list("body-bg", if (use_body) "text-color" else "gray-base"),
    default = if (use_body) list("body-bg", "body-color") else list("white", "black")
  )
  if (decode) {
    setNames(vars, c("bg", "fg"))
  } else {
    setNames(c("bg", "fg"), vars)
  }
}

# We've modified these "dark mode" themes to be more themable by cascading
# defaults from $body-bg/$body-color instead of $white/$black
has_body_base_colors <- function(theme) {
  body_themes <- c("darkly", "cyborg", "superhero", "slate")
  # TODO: do the same for BS3? I guess?
  isTRUE(theme_bootswatch(theme) %in% body_themes)
}

bs4_base_colors <- function(args) {
  white <- args$bg
  black <- args$fg

  grays <- grDevices::colorRamp(c(white, black), alpha = TRUE)(0:10/10)

  if (any(grays[,4] != 255)) {
    warning(call. = FALSE,
      "bs_base_colors does not respect alpha in `white` and `black` arguments"
    )
  }

  grays <- sprintf("#%02X%02X%02X",
    round(grays[,1]),
    round(grays[,2]),
    round(grays[,3])
  )

  names(grays) <- c(
    "white",
    paste0("gray-", 1:9 * 100),
    "black"
  )

  as.list(grays)
}

bs3_base_colors <- function(args) {
  white <- args$bg
  black <- args$fg

  ramp <- grDevices::colorRamp(c(black, white))
  gray <- function(level = 255) {
    val <- round(ramp(level / 255))
    sprintf("#%02X%02X%02X", val[1,1], val[1,2], val[1,3])
  }

  result_colors <- list(
    "gray-base" = gray(0x00),
    "gray-darker" = gray(0x22),
    "gray-dark" = gray(0x33),
    "gray" = gray(0x55),
    "gray-light" = gray(0x77),
    "gray-lighter" = gray(0xEE),

    "gray-44" = gray(0x44),
    "gray-88" = gray(0x88),
    "gray-99" = gray(0x99),
    "gray-cc" = gray(0xcc),
    "gray-dd" = gray(0xdd),
    "gray-e5" = gray(0xe5),
    "gray-f5" = gray(0xf5),
    "gray-f8" = gray(0xf8),
    "gray-f9" = gray(0xf9),
    "white" = gray(0xff),
    "black" = gray(0x00)
  )

  # There's code in tools/bs3_theme_base_colors.R for generating this list.
  color_mapping <- list(
    `dropdown-caret-color` = "$gray-base", `tooltip-bg` = "$gray-base",
    `modal-backdrop-bg` = "$gray-base", `close-color` = "$gray-base",
    `navbar-inverse-bg` = "$gray-darker", `btn-default-color` = "$gray-dark",
    `navbar-default-link-hover-color` = "$gray-dark", `navbar-inverse-toggle-hover-bg` = "$gray-dark",
    `navbar-inverse-toggle-border-color` = "$gray-dark", `list-group-link-heading-color` = "$gray-dark",
    `kbd-bg` = "$gray-dark", `navbar-inverse-link-disabled-color` = "$gray-44",
    `navbar-default-link-active-color` = "$gray", `list-group-link-color` = "$gray",
    `navbar-default-color` = "$gray-light", `navbar-default-link-color` = "$gray-light",
    `navbar-default-toggle-icon-bar-bg` = "$gray-88", `input-color-placeholder` = "$gray-99",
    `modal-content-fallback-border-color` = "$gray-99", `btn-default-border` = "$gray-cc",
    `input-border` = "$gray-cc", `dropdown-fallback-border` = "$gray-cc",
    `navbar-default-link-disabled-color` = "$gray-cc", `popover-fallback-border-color` = "$gray-cc",
    `breadcrumb-color` = "$gray-cc", `pre-border-color` = "$gray-cc",
    `table-border-color` = "$gray-dd", `navbar-default-toggle-hover-bg` = "$gray-dd",
    `navbar-default-toggle-border-color` = "$gray-dd", `nav-tabs-border-color` = "$gray-dd",
    `nav-tabs-active-link-hover-border-color` = "$gray-dd", `nav-tabs-justified-link-border-color` = "$gray-dd",
    `pagination-border` = "$gray-dd", `pagination-hover-border` = "$gray-dd",
    `pagination-disabled-border` = "$gray-dd", `list-group-border` = "$gray-dd",
    `panel-inner-border` = "$gray-dd", `panel-default-border` = "$gray-dd",
    `thumbnail-border` = "$gray-dd", `legend-border-color` = "$gray-e5",
    `dropdown-divider-bg` = "$gray-e5", `modal-header-border-color` = "$gray-e5",
    `table-bg-hover` = "$gray-f5", `dropdown-link-hover-bg` = "$gray-f5",
    `progress-bg` = "$gray-f5", `list-group-hover-bg` = "$gray-f5",
    `panel-footer-bg` = "$gray-f5", `panel-default-heading-bg` = "$gray-f5",
    `well-bg` = "$gray-f5", `breadcrumb-bg` = "$gray-f5", `code-bg` = "$gray-f5",
    `pre-bg` = "$gray-f5", `navbar-default-bg` = "$gray-f8",
    `table-bg-accent` = "$gray-f9", `body-bg` = "$white", `btn-default-bg` = "$white",
    `btn-primary-color` = "$white", `btn-success-color` = "$white",
    `btn-info-color` = "$white", `btn-warning-color` = "$white",
    `btn-danger-color` = "$white", `input-bg` = "$white", `dropdown-bg` = "$white",
    `navbar-inverse-link-hover-color` = "$white", `navbar-inverse-brand-hover-color` = "$white",
    `navbar-inverse-toggle-icon-bar-bg` = "$white", `pagination-bg` = "$white",
    `pagination-active-color` = "$white", `pagination-disabled-bg` = "$white",
    `tooltip-color` = "$white", `popover-bg` = "$white", `label-color` = "$white",
    `label-link-hover-color` = "$white", `modal-content-bg` = "$white",
    `progress-bar-color` = "$white", `list-group-bg` = "$white",
    `panel-bg` = "$white", `panel-primary-text` = "$white", `badge-color` = "$white",
    `badge-link-hover-color` = "$white", `badge-active-bg` = "$white",
    `carousel-control-color` = "$white", `carousel-indicator-active-bg` = "$white",
    `carousel-indicator-border-color` = "$white", `carousel-caption-color` = "$white",
    `close-text-shadow` = "$white", `kbd-color` = "$white"
  )

  results <- c(result_colors, color_mapping)

  results
}

# ----------------------------------------------------------------------------------------
# Accent colors
# ----------------------------------------------------------------------------------------

bs_accent_colors <- function(theme, primary = NULL, secondary = NULL,
                             success = NULL, info = NULL, warning = NULL, danger = NULL) {
  assert_bs_theme(theme)

  args <- validate_and_normalize_colors(
    list(
      primary = primary, secondary = secondary, success = success,
      info = info, warning = warning, danger = danger
    )
  )

  switch_add_variables(
    theme, args, three = bs3_accent_colors, default = identity
  )
}


bs3_accent_colors <- function(args) {
  # Warns and filters out unsupported arguments
  supported <- c("primary", "success", "info", "warning", "danger")

  args <- retain_known_vars("Bootstrap 3", "accent color", supported, args)

  # Bootstrap 3 uses brand-primary, brand-danger, etc. as var names
  if (length(args) > 0) {
    names(args) <- paste0("brand-", names(args))
  }

  args
}

# ----------------------------------------------------------------------------------------
# Fonts
# ----------------------------------------------------------------------------------------

bs_fonts <- function(theme, base = NULL, code = NULL, heading = NULL) {
  assert_bs_theme(theme)
  args <- dropNulls(list(
    base = base,
    code = code,
    heading = heading
  ))
  args <- lapply(args, as_font_collection)
  switch_add_variables(theme, args, three = bs3_fonts, default = bs4_fonts)
}

as_font_collection <- function(x) {
  if (sass::is_font_collection(x)) {
    return(x)
  }
  do.call(sass::font_collection, as.list(x))
}


bs4_fonts <- function(args) {
  name_map <- c(
    base = "font-family-base",
    code = "font-family-monospace",
    heading = "headings-font-family",
    # TODO: we don't have a input_font...should we?
    input = "input-btn-font-family"
  )

  names(args) <- name_map[names(args)]
  args
}

bs3_fonts <- function(args) {
  name_map <- c(
    base = "font-family-base",
    code = "font-family-monospace",
    heading = "headings-font-family"
  )

  args <- retain_known_vars("Bootstrap 3", "font", names(name_map), args)

  names(args) <- name_map[names(args)]
  args
}

#' Ensures all arguments are either NULL, or length 1 character vectors with
#' valid CSS color strings; returned is a list with no NULLs and normalized
#' color strings
#' @param args A named list
#' @noRd
validate_and_normalize_colors <- function(args) {
  args <- dropNulls(args)
  if (length(args) == 0) return(args)

  is_char <- vapply(args, is.character, logical(1))
  vec_len <- vapply(args, length, integer(1))
  bad <- !is_char | vec_len != 1
  if (any(bad)) {
    stop(call. = FALSE,
      "Invalid HTML color strings for argument(s) ",
      format_varnames(names(args)[bad]),
      "; single-element character vectors are required")
  }
  args <- lapply(args, htmltools::parseCssColors, mustWork = FALSE)
  bad <- vapply(args, rlang::is_na, logical(1))
  if (any(bad)) {
    stop(
      "Invalid HTML color strings for argument(s) ",
      format_varnames(names(args)[bad]), call. = FALSE
    )
  }
  args
}


#' Remove unsupported arguments, with a nicely formatted warning
#'
#' @param caller_name String naming the calling function; used for error
#'   messages
#' @param arg_name A name for the group of variables (e.g., accent, font); used for error messages
#' @param supported_vars Character vector of known names
#' @param args List of args
#' @return List with unsupported vars removed, possibly warning in the process
#' @noRd
retain_known_vars <- function(caller_name, arg_name = "", supported_vars, args) {
  argnames <- names(args)
  unknown_idx <- !argnames %in% supported_vars

  if (any(unknown_idx)) {
    warning(call. = FALSE,
      caller_name, " doesn't support the following ",
      arg_name, if (nzchar(arg_name)) " ",
      "argument(s), they will be ignored: ",
      format_varnames(argnames[unknown_idx])
    )
  }

  args[!unknown_idx]
}

# Format a vector of variable names
format_varnames <- function(varnames, quot = "`", delim = ", ") {
  between <- paste0(quot, delim, quot)

  paste0(
    quot,
    paste(collapse = between, varnames),
    quot
  )
}

test_that("bs4 base colors", {
  varnames <- c(
    "black", "white", paste0("gray-", 1:9 * 100),
    "body-bg", "body-color", "primary", "secondary"
  )

  theme <- bs_theme(4, bg = "white", fg = "black", primary = "blue", secondary = "silver")
  colors <- bs_get_variables(theme, varnames)

  expect_identical(colors,
    c(
      black = "#000000", white = "#FFFFFF", `gray-100` = "#E6E6E6",
      `gray-200` = "#CCCCCC", `gray-300` = "#B2B2B2", `gray-400` = "#999999",
      `gray-500` = "#808080", `gray-600` = "#666666", `gray-700` = "#4D4D4D",
      `gray-800` = "#333333", `gray-900` = "#191919", `body-bg` = "#FFFFFF",
      `body-color` = "#191919", primary = "#0000FF", secondary = "#C0C0C0")
  )

  theme <- bs_theme("4", bg = "#112233", fg = "#FFEEDD", primary = "orange", secondary = "brown")
  colors <- bs_get_variables(theme, varnames)

  expect_identical(colors,
    c(
      black = "#FFEEDD", white = "#112233", `gray-100` = "#293644",
      `gray-200` = "#414B55", `gray-300` = "#585F66", `gray-400` = "#707477",
      `gray-500` = "#888888", `gray-600` = "#A09C99", `gray-700` = "#B8B1AA",
      `gray-800` = "#CFC5BB", `gray-900` = "#E7DACC", `body-bg` = "#112233",
      `body-color` = "#E7DACC", primary = "#FFA500", secondary = "#A52A2A")
  )

  # Can individual colors still be overridden?
  theme <- bs_add_variables(theme, "body-bg" = "white", black = "red")
  expect_identical(bs_get_variables(theme, c("body-bg", "black")),
    c("body-bg" = "white", black = "red"))
})

test_that("bs3 base colors", {

  varnames <- c(
    "gray-base", "white",
    paste0("gray", c("-darker", "-dark", "", "-light", "-lighter")),
    "body-bg", "text-color", "brand-primary")

  theme <- bs_theme("3", bg = "white", fg = "black", primary = "blue")
  colors <- bs_get_variables(theme, varnames)

  expect_identical(colors,
    c(`gray-base` = "#000000", white = "#FFFFFF", `gray-darker` = "#222222",
      `gray-dark` = "#333333", gray = "#555555", `gray-light` = "#777777",
      `gray-lighter` = "#EEEEEE", `body-bg` = "#FFFFFF", `text-color` = "#333333",
      `brand-primary` = "#0000FF")
  )

  expect_warning(
    theme <- bs_theme("3", bg = "#112233", fg = "#FFEEDD", primary = "orange", secondary = "brown"),
    "doesn't support.*ignored"
  )
  colors <- bs_get_variables(theme, varnames)

  expect_identical(colors,
    c(`gray-base` = "#FFEEDD", white = "#112233", `gray-darker` = "#DFD3C6",
      `gray-dark` = "#CFC5BB", gray = "#B0AAA4", `gray-light` = "#908F8E",
      `gray-lighter` = "#21303E", `body-bg` = "#112233", `text-color` = "#CFC5BB",
      `brand-primary` = "#FFA500")
  )

  # Can individual colors still be overridden?
  theme <- bs_add_variables(theme, "body-bg" = "black")
  expect_identical(bs_get_variables(theme, c("body-bg", "gray-darker")),
    c("body-bg" = "black", "gray-darker" = "#DFD3C6"))
})

test_that("bs4 accent colors", {
  varnames <- c("primary", "secondary", "success", "info", "warning",
    "danger")

  # Baseline
  theme <- bs_theme(4)
  expect_identical(bs_get_variables(theme, varnames),
    c(primary = "#007bff", secondary = "#6c757d",
      success = "#28a745", info = "#17a2b8", warning = "#ffc107", danger = "#dc3545")
  )

  theme <- bs_theme(4, primary = "#123", secondary = "#234",
                    success = "#345", info = "#456", warning = "#567", danger = "#678")
  expect_identical(bs_get_variables(theme, varnames),
    c(primary = "#112233", secondary = "#223344", success = "#334455",
      info = "#445566", warning = "#556677", danger = "#667788")
  )

  theme <- bs_theme(4, primary = "#123", secondary = "#234",
                    success = "#345", info = "#456", warning = "#567", danger = "#678")
  expect_identical(bs_get_variables(theme, varnames),
    c(primary = "#112233", secondary = "#223344",
      success = "#334455", info = "#445566", warning = "#556677", danger = "#667788")
  )
})

test_that("bs3 accent colors", {
  varnames <- c("primary", "secondary", "default", "success", "info", "warning",
    "danger")
  varnames <- paste0("brand-", varnames)

  # Baseline
  theme <- bs_theme("3")
  expect_identical(bs_get_variables(theme, varnames),
    c(`brand-primary` = "#337ab7", `brand-secondary` = NA, `brand-default` = NA,
      `brand-success` = "#5cb85c", `brand-info` = "#5bc0de", `brand-warning` = "#f0ad4e",
      `brand-danger` = "#d9534f")
  )

  expect_warning(
    theme <- bs_theme("3", primary = "#123", secondary = "#234",
                      success = "#345", info = "#456", warning = "#567", danger = "#678"),
    "doesn't support"
  )
  expect_identical(bs_get_variables(theme, varnames),
    c(`brand-primary` = "#112233", `brand-secondary` = NA, `brand-default` = NA,
      `brand-success` = "#334455", `brand-info` = "#445566", `brand-warning` = "#556677",
      `brand-danger` = "#667788")
  )
})

test_that("bs_theme_fonts", {
  expected_result <- "\"Source Sans Pro\", \"Open Sans\", Helvetica, sans-serif"
  candidates <- list(
    # All separate
    c("Source Sans Pro", "Open Sans", "Helvetica", "sans-serif"),
    # All together
    c("'Source Sans Pro'", "Open Sans", "Helvetica, sans-serif"),
    # Mixed
    c("'Source Sans Pro', \"Open Sans\", Helvetica, sans-serif")
  )

  lapply(versions(), function(version) {

    for (candidate in candidates) {
      theme <- bs_theme(version, base_font = candidate)
      expect_identical(
        bs_get_variables(theme, "font-family-base"),
        c(`font-family-base` = expected_result)
      )
    }
    expect_error(bs_fonts(theme, ""))
    expect_error(bs_fonts(theme, code = NA_character_))
    expect_error(bs_fonts(theme, heading = TRUE))
    expect_error(bs_fonts(theme, 10L))

    # NOT an error
    expect_error(bs_fonts(theme, NULL), NA)

    theme <- bs_fonts(bs_theme(), base = "a", code = "b", heading = "c")
    expect_identical(
      bs_get_variables(theme, c("font-family-base", "font-family-monospace", "headings-font-family")),
      c(`font-family-base` = "a", `font-family-monospace` = "b", `headings-font-family` = "c")
    )
  })
})

test_that("format_varnames", {
  expect_identical(format_varnames("a"), "`a`")
  expect_identical(format_varnames(c("a", "b")), "`a`, `b`")
  expect_identical(format_varnames(c("foo bar", "baz qux"), quot = "'", delim = "/"), "'foo bar'/'baz qux'")
})

test_that("validate_and_normalize_colors", {
  expect_null(validate_and_normalize_colors(NULL))

  # Bad
  expect_error(validate_and_normalize_colors(list(mycolor = "")), "Invalid.*mycolor")
  expect_error(validate_and_normalize_colors(list(foo = "bar")), "Invalid")
  expect_error(
    validate_and_normalize_colors(list(valid = "black", notvalid1 = "#00", notvalid2 = "#11111")),
    "Invalid.*notvalid.*notvalid2"
  )
  expect_error(validate_and_normalize_colors(list(toomany = c("red", "blue"))), "Invalid.*toomany")

  expect_identical(
    validate_and_normalize_colors(list(a = "#000", b = "white", c = "rgb(16, 32, 64)")),
    list(a = "#000000", b = "#FFFFFF", c = "#102040")
  )
})


test_that("bs_get_variables is fg/bg aware", {
  expect_equal(
    bs_get_variables(bs_theme(), c("bg", "fg")),
    c(bg = "#fff", fg = "#000")
  )
  expect_equal(
    bs_get_variables(bs_theme(version = 3), c("bg", "fg")),
    c(bg = "#fff", fg = "#000")
  )
  expect_equal(
    bs_get_variables(bs_theme(version = 4, bootswatch = "darkly"), c("bg", "fg")),
    c(bg = "#222", fg = "#fff")
  )
  expect_equal(
    bs_get_variables(bs_theme(version = 3, bootswatch = "darkly"), c("bg", "fg")),
    c(bg = "#222222", fg = "#fff")
  )
})


test_that("theme-color('default') works as expected", {
  # default theme color derives defaults from $gray-300
  css <- sass::sass_partial(
    ".foo{color:theme-color('default')}", bs_theme(version = 4)
  )
  expect_css(css, ".foo{color:#dee2e6;}")
  # but can be customized via secondary
  css <- sass::sass_partial(
    ".foo{color:theme-color('default')}", bs_theme(version = 4, secondary = "red")
  )
  expect_css(css, ".foo{color:#FF0000;}")
  expect_error(
    sass::sass(
      list(bs_theme(), ".foo { @extend .bg-default; }")
    ),
    NA
  )
})

test_that("bs_get_contrast() works as expected", {
  for (version in versions()) {
    base <- bs_theme(version)
    expect_equal(
      bs_get_contrast(base, "input-bg"), c("input-bg" = "#000000")
    )
    primary <- if ("3" %in% version) "brand-primary" else "primary"
    expect_equal(
      bs_get_contrast(base, c("input-bg", primary)),
      setNames(c("#000000", "#FFFFFF"), c("input-bg", primary))
    )
    inverted <- bs_theme_update(base, bg = "black", fg = "white")
    expect_equal(
      bs_get_contrast(inverted, "input-bg"),
      c("input-bg" = "#FFFFFF")
    )
    yb <- bs_theme(4, bg = "yellow", fg = "#0000FF")
    expect_equal(
      bs_get_contrast(yb, "input-bg"),
      c("input-bg" = "#0000FF")
    )
    yb <- bs_theme_update(yb, bg = "black", fg = "white", "color-contrast-light" = "#0000FF", "color-contrast-dark" = "yellow")
    expect_equal(
      bs_get_contrast(yb, "input-bg"),
      c("input-bg" = "#FFFF00")
    )
    expect_error(
      bs_get_contrast(base, "font"),
      "Undefined variable"
    )
    expect_error(
      bs_get_contrast(base, "font-family-base"),
      "must be a color"
    )
    if ("3" %in% version) {
      next
    }
    light_success <- bs_theme_update(inverted, success = "#E5FFE5")
    expect_equal(
      bs_get_contrast(light_success, "success"),
      c("success" = "#000000")
    )
    expect_equal(
      bs_get_contrast(bs_theme_update(light_success, "color-contrast-light" = "#222", "min-contrast-ratio" = 1), "success"),
      c("success" = "#222222")
    )
    dark_success <- bs_theme_update(inverted, success = "#193319")
    expect_equal(
      bs_get_contrast(dark_success, "success"),
      c("success" = "#FFFFFF")
    )
    expect_equal(
      bs_get_contrast(bs_theme_update(dark_success, "color-contrast-light" = "#F8F9FA"), "success"),
      c("success" = "#F8F9FA")
    )
  }
})

test_that("bs4 base colors", {
  varnames <- c(
    "black",
    "white",
    paste0("gray-", 1:9 * 100),
    "body-bg",
    "body-color",
    "primary",
    "secondary"
  )

  theme <- bs_theme(
    4,
    bg = "white",
    fg = "black",
    primary = "blue",
    secondary = "silver"
  )
  colors <- bs_get_variables(theme, varnames)

  expect_identical(
    colors,
    c(
      black = "#000000",
      white = "#FFFFFF",
      `gray-100` = "#E6E6E6",
      `gray-200` = "#CCCCCC",
      `gray-300` = "#B2B2B2",
      `gray-400` = "#999999",
      `gray-500` = "#808080",
      `gray-600` = "#666666",
      `gray-700` = "#4D4D4D",
      `gray-800` = "#333333",
      `gray-900` = "#191919",
      `body-bg` = "#FFFFFF",
      `body-color` = "#191919",
      primary = "#0000FF",
      secondary = "#C0C0C0"
    )
  )

  theme <- bs_theme(
    "4",
    bg = "#112233",
    fg = "#FFEEDD",
    primary = "orange",
    secondary = "brown"
  )
  colors <- bs_get_variables(theme, varnames)

  expect_identical(
    colors,
    c(
      black = "#FFEEDD",
      white = "#112233",
      `gray-100` = "#293644",
      `gray-200` = "#414B55",
      `gray-300` = "#585F66",
      `gray-400` = "#707477",
      `gray-500` = "#888888",
      `gray-600` = "#A09C99",
      `gray-700` = "#B8B1AA",
      `gray-800` = "#CFC5BB",
      `gray-900` = "#E7DACC",
      `body-bg` = "#112233",
      `body-color` = "#E7DACC",
      primary = "#FFA500",
      secondary = "#A52A2A"
    )
  )

  # Can individual colors still be overridden?
  theme <- bs_add_variables(theme, "body-bg" = "white", black = "red")
  expect_identical(
    bs_get_variables(theme, c("body-bg", "black")),
    c("body-bg" = "white", black = "red")
  )
})

test_that("bs3 base colors", {
  varnames <- c(
    "gray-base",
    "white",
    paste0("gray", c("-darker", "-dark", "", "-light", "-lighter")),
    "body-bg",
    "text-color",
    "brand-primary"
  )

  theme <- bs_theme("3", bg = "white", fg = "black", primary = "blue")
  colors <- bs_get_variables(theme, varnames)

  expect_identical(
    colors,
    c(
      `gray-base` = "#000000",
      white = "#FFFFFF",
      `gray-darker` = "#222222",
      `gray-dark` = "#333333",
      gray = "#555555",
      `gray-light` = "#777777",
      `gray-lighter` = "#EEEEEE",
      `body-bg` = "#FFFFFF",
      `text-color` = "#333333",
      `brand-primary` = "#0000FF"
    )
  )

  expect_warning(
    theme <- bs_theme(
      "3",
      bg = "#112233",
      fg = "#FFEEDD",
      primary = "orange",
      secondary = "brown"
    ),
    "doesn't support.*ignored"
  )
  colors <- bs_get_variables(theme, varnames)

  expect_identical(
    colors,
    c(
      `gray-base` = "#FFEEDD",
      white = "#112233",
      `gray-darker` = "#DFD3C6",
      `gray-dark` = "#CFC5BB",
      gray = "#B0AAA4",
      `gray-light` = "#908F8E",
      `gray-lighter` = "#21303E",
      `body-bg` = "#112233",
      `text-color` = "#CFC5BB",
      `brand-primary` = "#FFA500"
    )
  )

  # Can individual colors still be overridden?
  theme <- bs_add_variables(theme, "body-bg" = "black")
  expect_identical(
    bs_get_variables(theme, c("body-bg", "gray-darker")),
    c("body-bg" = "black", "gray-darker" = "#DFD3C6")
  )
})

test_that("bs4 accent colors", {
  varnames <- c("primary", "secondary", "success", "info", "warning", "danger")

  # Baseline
  theme <- bs_theme(4)
  expect_identical(
    bs_get_variables(theme, varnames),
    c(
      primary = "#007bff",
      secondary = "#6c757d",
      success = "#28a745",
      info = "#17a2b8",
      warning = "#ffc107",
      danger = "#dc3545"
    )
  )

  theme <- bs_theme(
    4,
    primary = "#123",
    secondary = "#234",
    success = "#345",
    info = "#456",
    warning = "#567",
    danger = "#678"
  )
  expect_identical(
    bs_get_variables(theme, varnames),
    c(
      primary = "#112233",
      secondary = "#223344",
      success = "#334455",
      info = "#445566",
      warning = "#556677",
      danger = "#667788"
    )
  )

  theme <- bs_theme(
    4,
    primary = "#123",
    secondary = "#234",
    success = "#345",
    info = "#456",
    warning = "#567",
    danger = "#678"
  )
  expect_identical(
    bs_get_variables(theme, varnames),
    c(
      primary = "#112233",
      secondary = "#223344",
      success = "#334455",
      info = "#445566",
      warning = "#556677",
      danger = "#667788"
    )
  )
})

test_that("bs3 accent colors", {
  varnames <- c(
    "primary",
    "secondary",
    "default",
    "success",
    "info",
    "warning",
    "danger"
  )
  varnames <- paste0("brand-", varnames)

  # Baseline
  theme <- bs_theme("3")
  expect_identical(
    bs_get_variables(theme, varnames),
    c(
      `brand-primary` = "#337ab7",
      `brand-secondary` = NA,
      `brand-default` = NA,
      `brand-success` = "#5cb85c",
      `brand-info` = "#5bc0de",
      `brand-warning` = "#f0ad4e",
      `brand-danger` = "#d9534f"
    )
  )

  expect_warning(
    theme <- bs_theme(
      "3",
      primary = "#123",
      secondary = "#234",
      success = "#345",
      info = "#456",
      warning = "#567",
      danger = "#678"
    ),
    "doesn't support"
  )
  expect_identical(
    bs_get_variables(theme, varnames),
    c(
      `brand-primary` = "#112233",
      `brand-secondary` = NA,
      `brand-default` = NA,
      `brand-success` = "#334455",
      `brand-info` = "#445566",
      `brand-warning` = "#556677",
      `brand-danger` = "#667788"
    )
  )
})

test_that("bs_theme_fonts", {
  expected_result <- "\"Source Sans Pro\", \"Open Sans\", Helvetica, sans-serif"
  candidates <- list(
    # All separate
    c("Source Sans Pro", "Open Sans", "Helvetica", "sans-serif"),
    # All together
    c("'Source Sans Pro'", "Open Sans", "Helvetica, sans-serif"),
    # Mixed
    c("'Source Sans Pro', \"Open Sans\", Helvetica, sans-serif")
  )

  lapply(versions(), function(version) {
    for (candidate in candidates) {
      theme <- bs_theme(version, base_font = candidate)
      expect_identical(
        bs_get_variables(theme, "font-family-base"),
        c(`font-family-base` = expected_result)
      )
    }
    expect_error(bs_fonts(theme, ""))
    expect_error(bs_fonts(theme, code = NA_character_))
    expect_error(bs_fonts(theme, heading = TRUE))
    expect_error(bs_fonts(theme, 10L))

    # NOT an error
    expect_error(bs_fonts(theme, NULL), NA)

    theme <- bs_fonts(bs_theme(), base = "a", code = "b", heading = "c")
    expect_identical(
      bs_get_variables(
        theme,
        c("font-family-base", "font-family-monospace", "headings-font-family")
      ),
      c(
        `font-family-base` = "a",
        `font-family-monospace` = "b",
        `headings-font-family` = "c"
      )
    )
  })
})

test_that("format_varnames", {
  expect_identical(format_varnames("a"), "`a`")
  expect_identical(format_varnames(c("a", "b")), "`a`, `b`")
  expect_identical(
    format_varnames(c("foo bar", "baz qux"), quot = "'", delim = "/"),
    "'foo bar'/'baz qux'"
  )
})

test_that("validate_and_normalize_colors", {
  expect_null(validate_and_normalize_colors(NULL))

  # Bad
  expect_error(
    validate_and_normalize_colors(list(mycolor = "")),
    "Invalid.*mycolor"
  )
  expect_error(validate_and_normalize_colors(list(foo = "bar")), "Invalid")
  expect_error(
    validate_and_normalize_colors(
      list(valid = "black", notvalid1 = "#00", notvalid2 = "#11111")
    ),
    "Invalid.*notvalid.*notvalid2"
  )
  expect_error(
    validate_and_normalize_colors(list(toomany = c("red", "blue"))),
    "Invalid.*toomany"
  )

  expect_identical(
    validate_and_normalize_colors(
      list(a = "#000", b = "white", c = "rgb(16, 32, 64)")
    ),
    list(a = "#000000", b = "#FFFFFF", c = "#102040")
  )
})

test_that("bs_get_variables is fg/bg aware", {
  expect_equal(
    bs_get_variables(bs_theme(), c("bg", "fg")),
    c(bg = "#ffffff", fg = "#000")
  )
  expect_equal(
    bs_get_variables(bs_theme(version = 3), c("bg", "fg")),
    c(bg = "#fff", fg = "#000")
  )
  expect_equal(
    bs_get_variables(
      bs_theme(version = 4, bootswatch = "darkly"),
      c("bg", "fg")
    ),
    c(bg = "#222", fg = "#fff")
  )
  expect_equal(
    bs_get_variables(
      bs_theme(version = 3, bootswatch = "darkly"),
      c("bg", "fg")
    ),
    c(bg = "#222222", fg = "#fff")
  )
})

test_that("theme-color('default') works as expected", {
  # default theme color derives defaults from $gray-300
  css <- sass::sass_partial(
    ".foo{color:theme-color('default')}",
    bs_theme(version = 4)
  )
  expect_css(css, ".foo{color:#dee2e6;}")
  # but can be customized via secondary
  css <- sass::sass_partial(
    ".foo{color:theme-color('default')}",
    bs_theme(version = 4, secondary = "red")
  )
  expect_css(css, ".foo{color:#FF0000;}")
  expect_error(
    sass::sass(
      list(bs_theme(), ".foo { @extend .bg-default; }")
    ),
    NA
  )
})

test_that("bs_get_contrast() works as expected", {
  for (version in versions()) {
    base <- bs_theme(version)
    expect_equal(
      bs_get_contrast(base, "body-bg"),
      c("body-bg" = "#000000")
    )
    primary <- if ("3" %in% version) "brand-primary" else "primary"
    expect_equal(
      bs_get_contrast(base, c("body-bg", primary)),
      setNames(c("#000000", "#FFFFFF"), c("body-bg", primary))
    )
    inverted <- bs_theme_update(base, bg = "black", fg = "white")
    expect_equal(
      bs_get_contrast(inverted, "body-bg"),
      c("body-bg" = "#FFFFFF")
    )
    yb <- bs_theme(4, bg = "yellow", fg = "#0000FF")
    expect_equal(
      bs_get_contrast(yb, "input-bg"),
      c("input-bg" = "#0000FF")
    )
    yb <- bs_theme_update(
      yb,
      bg = "black",
      fg = "white",
      "color-contrast-light" = "#0000FF",
      "color-contrast-dark" = "yellow"
    )
    expect_equal(
      bs_get_contrast(yb, "input-bg"),
      c("input-bg" = "#FFFF00")
    )
    expect_error(
      bs_get_contrast(base, "font"),
      "Undefined variable"
    )
    expect_error(
      bs_get_contrast(base, "font-family-base"),
      "must be a color"
    )
    if ("3" %in% version) {
      next
    }
    light_success <- bs_theme_update(inverted, success = "#E5FFE5")
    expect_equal(
      bs_get_contrast(light_success, "success"),
      c("success" = "#000000")
    )
    expect_equal(
      bs_get_contrast(
        bs_theme_update(
          light_success,
          "color-contrast-light" = "#222",
          "min-contrast-ratio" = 1
        ),
        "success"
      ),
      c("success" = "#222222")
    )
    dark_success <- bs_theme_update(inverted, success = "#193319")
    expect_equal(
      bs_get_contrast(dark_success, "success"),
      c("success" = "#FFFFFF")
    )
    expect_equal(
      bs_get_contrast(
        bs_theme_update(dark_success, "color-contrast-light" = "#F8F9FA"),
        "success"
      ),
      c("success" = "#F8F9FA")
    )
  }
})

#' Button for launching longer-running operations
#'
#' @description
#' `input_task_button` is a button that can be used in conjuction with
#' [shiny::bindEvent()] (or the older [shiny::eventReactive()] and
#' [shiny::observeEvent()] functions) to trigger actions or recomputation.
#'
#' It is similar to [shiny::actionButton()], except it prevents the user from
#' clicking when its operation is already in progress.
#'
#' Upon click, it automatically displays a customizable progress message and
#' disables itself; and after the server has dealt with whatever reactivity is
#' triggered from the click, the button automatically reverts to its original
#' appearance and re-enables itself.
#'
#' @section Manual button reset:
#' In some advanced use cases, it may be necessary to keep a task button in its
#' busy state even after the normal reactive processing has completed. Calling
#' `update_task_button(id, state = "busy")` from the server will opt out of any
#' currently pending reset for a specific task button. After doing so, the
#' button can be re-enabled by calling `update_task_button(id, state = "ready")`
#' after each click's work is complete.
#'
#' You can also pass an explicit `auto_reset = FALSE` to `input_task_button()`,
#' which means that button will _never_ be automatically re-enabled and will
#' require `update_task_button(id, state = "ready")` to be called each time.
#'
#' Note that, as a general rule, Shiny's `update` family of functions do not
#' take effect at the instant that they are called, but are held until the end
#' of the current reactive cycle. So if you have many different reactive
#' calculations and outputs, you don't have to be too careful about when you
#' call `update_task_button(id, state = "ready")`, as the button on the client
#' will not actually re-enable until the same moment that all of the updated
#' outputs simultaneously sent to the client.
#'
#' @section Custom states:
#' The task button is designed to automatically switch between two states: the
#' "ready" state, where the button is clickable and displays the `label` and
#' `icon`; and the "busy" state, where the button is disabled and displays
#' `label_busy` and `icon_busy`.
#'
#' In advanced use cases, you can include additional states by adding an
#' [htmltools::div()] with a `slot` attribute naming the state and the icon and
#' label as the first and second children, respectively.
#'
#' ```r
#' input_task_button(
#'   label = "Ring home",
#'   icon = fontawesome::fa_i("phone"),
#'   div(slot = "ringing", fontawesome::fa_i("bell"), "Ringing..."),
#'   div(
#'     slot = "voice-mail",
#'     fontawesome::fa_i("voicemail"),
#'     "Leaving a message..."
#'   )
#' )
#' ```
#'
#' You can move between these states by calling `update_task_button()` and
#' passing the slot name to the `state` argument, e.g. `state="ringing"`. See
#' the section above on manual button resetting, which you will likely need to
#' use in conjunction with custom states.
#'
#' @section Server value:
#' An integer of class `"shinyActionButtonValue"`. This class differs from
#' ordinary integers in that a value of 0 is considered "falsy". This implies
#' two things:
#'   * Event handlers (e.g., [shiny::observeEvent()], [shiny::eventReactive()])
#'     won't execute on initial load.
#'   * Input validation (e.g., [shiny::req()], [shiny::need()]) will fail on
#'     initial load.
#'
#' @param id The `input` slot that will be used to access the value.
#' @param label The label of the button while it is in ready (clickable) state;
#'   usually a string.
#' @param icon An optional icon to display next to the label while the button is
#'   in ready state. See [fontawesome::fa_i()].
#' @param label_busy The label of the button while it is busy.
#' @param icon_busy The icon to display while the button is busy. By default,
#'   `fontawesome::fa_i("refresh", class = "fa-spin", "aria-hidden" = "true")`
#'   is used, which displays a spinning "chasing arrows" icon. You can create
#'   spinning icons out of other Font Awesome icons by using the same
#'   expression, but replacing `"refresh"` with a different icon name. See
#'   [fontawesome::fa_i()].
#' @param type One of the Bootstrap theme colors (`"primary"`, `"default"`,
#'   `"secondary"`, `"success"`, `"danger"`, `"warning"`, `"info"`, `"light"`,
#'   `"dark"`), or `NULL` to leave off the Bootstrap-specific button CSS classes
#'   altogether.
#' @param ... In `input_task_button()`, named arguments become attributes to
#'   include on the `<button>` element, e.g. `class` or data attributes.
#'   Unnamed arguments can provide additional states for the button, see the
#'   "Custom states" section.
#'
#'   In `update_task_button()`, `...` are ignored and must be empty. The task
#'   button only supports changing between pre-defined states.
#' @param auto_reset If `TRUE` (the default), automatically put the button back
#'   in "ready" state after its click is handled by the server.
#'
#' @seealso [bind_task_button()]
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_sidebar(
#'   sidebar = sidebar(
#'     open = "always",
#'     input_task_button("resample", "Resample"),
#'   ),
#'   verbatimTextOutput("summary")
#' )
#'
#' server <- function(input, output, session) {
#'   sample <- eventReactive(input$resample, ignoreNULL=FALSE, {
#'     Sys.sleep(2)  # Make this artificially slow
#'     rnorm(100)
#'   })
#'
#'   output$summary <- renderPrint({
#'     summary(sample())
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
input_task_button <- function(
  id,
  label,
  ...,
  icon = NULL,
  label_busy = "Processing...",
  icon_busy = rlang::missing_arg(),
  type = "primary",
  auto_reset = TRUE
) {
  dots <- separate_arguments(...)
  attribs <- dots$attribs
  children <- dots$children

  icon_busy <- rlang::maybe_missing(
    icon_busy,
    fontawesome::fa_i("refresh", class = "fa-spin", "aria-hidden" = "true")
  )

  tags$button(
    id = id,
    class = if (!is.null(type)) paste0("btn btn-", type),
    class = "bslib-task-button",
    type = "button",
    "data-auto-reset" = if (isTRUE(auto_reset)) NA else NULL,
    !!!attribs,

    component_dependencies(),

    htmltools::tag(
      "bslib-switch-inline",
      rlang::list2(
        case = "ready",
        span(slot = "ready", icon, label),
        span(slot = "busy", icon_busy, label_busy),
        !!!children
      )
    )
  )
}

#' @param state If `"busy"`, put the button into busy/disabled state. If
#'   `"ready"`, put the button into ready/enabled state.
#' @param session The `session` object; using the default is recommended.
#' @rdname input_task_button
#' @export
update_task_button <- function(
  id,
  ...,
  state = NULL,
  session = get_current_session()
) {
  force(id)
  force(state)

  rlang::check_dots_empty()

  if (!is.null(state)) {
    if (!rlang::is_string(state)) {
      abort("`state` must be a single character value.")
    }
    set_task_button_manual_reset(
      session,
      id,
      manual = !identical(state, "ready")
    )
  }

  session$sendInputMessage(id, dropNulls(list(state = state)))
}

task_button_manual_reset_map <- function(session) {
  key <- "manual_task_button_reset"
  map <- session$userData[[key]]
  if (is.null(map)) {
    map <- fastmap::fastmap()
    session$userData[[key]] <- map
  }
  map
}

# Prevent automatic resetting of the task button when this reactive flush is
# complete
set_task_button_manual_reset <- function(session, id, manual = TRUE) {
  ns_id <- session$ns(id)
  map <- task_button_manual_reset_map(session)
  if (isTRUE(manual)) {
    map$set(ns_id, TRUE)
  } else {
    map$remove(ns_id)
  }
}

is_task_button_manual_reset <- function(session, id) {
  ns_id <- session$ns(id)
  map <- task_button_manual_reset_map(session)
  map$get(ns_id, FALSE)
}

input_task_button_input_handler <- function(val, shinysession, name) {
  value <- val[["value"]]

  if (isTRUE(val[["autoReset"]])) {
    shinysession$onFlush(once = TRUE, function() {
      # This is input_task_button's auto-reset feature: unless the button has
      # opted out using set_task_button_manual_reset(), we should reset after
      # a flush cycle where a bslib.taskbutton value is seen.
      if (!is_task_button_manual_reset(shinysession, name)) {
        update_task_button(name, state = "ready", session = shinysession)
      }
    })
  }

  # mark up the action button value with a special class so we can recognize it later
  class(value) <- c("shinyActionButtonValue", class(value))
  value
}

#' Bind `input_task_button` to `ExtendedTask`
#'
#' @description
#' Sets up a [shiny::ExtendedTask] to relay its state to an existing
#' [input_task_button()], so the task button stays in its "busy" state for as
#' long as the extended task is running.
#'
#' Note that `bind_task_button` does _not_ automatically cause button presses to
#' invoke the extended task; you still need to use [shiny::bindEvent()] (or
#' [shiny::observeEvent()]) to cause the button press to trigger an invocation,
#' as in the example below.
#'
#' `bind_task_button` cannot be used to bind one task button to multiple
#' `ExtendedTask` objects; if you attempt to do so, any bound `ExtendedTask`
#' that completes will cause the button to return to "ready" state.
#'
#' @param target The target object (i.e. `ExtendedTask`).
#' @param task_button_id A string matching the `id` argument passed to the
#'   corresponding [input_task_button()] call.
#' @param session A Shiny session object (the default should almost always be
#'   used).
#' @param ... Further arguments passed to other methods.
#'
#' @returns The `target` object that was passed in.
#'
#' @examplesIf rlang::is_interactive()
#'
#' library(shiny)
#' library(bslib)
#' library(future)
#' plan(multisession)
#'
#' ui <- page_sidebar(
#'   sidebar = sidebar(
#'     input_task_button("recalc", "Recalculate")
#'   ),
#'   textOutput("outval")
#' )
#'
#' server <- function(input, output) {
#'   rand_task <- ExtendedTask$new(function() {
#'     future({
#'       # Slow operation goes here
#'       Sys.sleep(2)
#'       runif(1)
#'     }, seed = TRUE)
#'   })
#'
#'   # Make button state reflect task.
#'   # If using R >=4.1, you can do this instead:
#'   # rand_task <- ExtendedTask$new(...) |> bind_task_button("recalc")
#'   bind_task_button(rand_task, "recalc")
#'
#'   observeEvent(input$recalc, {
#'     rand_task$invoke()
#'   })
#'
#'   output$outval <- renderText({
#'     rand_task$result()
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
bind_task_button <- function(target, task_button_id, ...) {
  UseMethod("bind_task_button")
}

#' @rdname bind_task_button
#' @export
bind_task_button.default <- function(target, task_button_id, ...) {
  abort(
    "Don't know how to bind a task button to an object of class '",
    class(target)[[0]],
    "'"
  )
}

#' @rdname bind_task_button
#' @export
bind_task_button.ExtendedTask <- function(
  target,
  task_button_id,
  ...,
  session = get_current_session()
) {
  force(target)
  force(task_button_id)
  force(session)

  was_running <- FALSE
  shiny::observe(
    label = "update_task_button_state",
    {
      running <- target$status() == "running"
      if (running != was_running) {
        was_running <<- running
        if (running) {
          update_task_button(task_button_id, state = "busy", session = session)
        } else {
          update_task_button(task_button_id, state = "ready", session = session)
        }
      }
    },
    priority = 1000
  )
  return(target)
}

# Turn off the default cache until the calling function exits; when that
# happens, the previous default cache will be restored.
local_disable_cache <- function(env = parent.frame()) {
  old_opts <- options(sass.cache = FALSE)
  withr::defer(
    options(old_opts),
    envir = env
  )
}

#' @include utils.R
NULL

#' Preview a Bootstrap theme
#'
#' Launches an example shiny app that can be used to get a quick preview of a
#' [bs_theme()], as well as an interactive GUI for tweaking some of the
#' main theme settings. Calling `bs_theme_preview()` with no arguments starts
#' the theme preview app with the default theme, which is a great way to see
#' the available theme presets or to start creating your own theme.
#'
#' The app that this launches is subject to change as new features are
#' developed in \pkg{bslib} and \pkg{shiny}.
#'
#' @inheritParams bs_theme_update
#' @param ... passed along to [shiny::runApp()].
#' @param with_themer whether or not to run the app with [run_with_themer()].
#'
#' @return nothing, this function is called for its side-effects (launching an
#'   application).
#'
#' @seealso Use [run_with_themer()] or [bs_themer()] to add the theming UI to
#'   an existing shiny app.
#'
#' @examplesIf rlang::is_interactive()
#' theme <- bs_theme(bg = "#6c757d", fg = "white", primary = "orange")
#' bs_theme_preview(theme)
#'
#' @family Bootstrap theme functions
#' @export
bs_theme_preview <- function(theme = bs_theme(), ..., with_themer = TRUE) {
  assert_bs_theme(theme)
  old_theme <- bs_global_get()
  on.exit(bs_global_set(old_theme), add = TRUE)
  bs_global_set(theme)
  # TODO: add more this demo and also an option for launching different demos
  app <- system_file("themer-demo", package = "bslib")
  if (with_themer) {
    run_with_themer(app, ...)
  } else {
    shiny::runApp(app, ...)
  }
}

colorpicker_deps <- function() {
  htmltools::htmlDependency(
    "bootstrap-colorpicker",
    "3.1.2",
    path_lib("bs-colorpicker"),
    stylesheet = "css/bootstrap-colorpicker.min.css",
    script = "js/bootstrap-colorpicker.js"
  )
}

opts_metadata <- function(theme) {
  opts <- jsonlite::fromJSON(
    system_file("themer/options.json", package = "bslib"),
    simplifyDataFrame = FALSE
  )
  version <- theme_version(theme)
  themes <- list(
    "Bootstrap" = "bootstrap",
    "bslib" = builtin_themes(version),
    "Bootswatch" = bootswatch_themes(version)
  )
  opts[[1]]$preset$choices <- dropNulls(themes)
  opts
}

bs_themer_ui <- function(opts, vals, theme) {
  theme_v <- theme_version(theme)

  make_control <- function(id, opts) {
    if (!is.null(opts$version)) {
      if (!is.null(opts$version$min) && theme_v < opts$version$min) {
        return(NULL)
      }
      if (!is.null(opts$version$max) && theme_v > opts$version$max) {
        return(NULL)
      }
    }

    value <- vals[[id]]
    lbl <- HTML(opts$label)
    desc <- HTML(opts$desc)

    text_input <- function(input_class = NULL, type = "text", ...) {
      div(
        class = "form-row form-group",
        tags$label(lbl),
        tags$input(
          type = type,
          value = value,
          "data-id" = id,
          class = "form-control form-control-sm bs-theme-value",
          class = input_class,
          ...
        ),
        if (!is.null(desc)) div(class = "form-text small", desc)
      )
    }

    switch_input <- function(input_class = NULL, ...) {
      # https://getbootstrap.com/docs/5.3/forms/checks-radios/#switches
      div(
        class = "form-row form-group form-check form-switch",
        tags$label(
          class = "form-check-label",
          tags$input(
            class = "form-check-input",
            class = input_class,
            type = "checkbox",
            role = "switch",
            id = paste0("bsthemer-", id)
          ),
          lbl
        )
      )
    }

    switch(
      opts$type,
      color = text_input(input_class = "bs-theme-value-color text-monospace"),
      str = text_input(input_class = "bs-theme-value-str"),
      length = text_input(input_class = "bs-theme-value-length"),
      number = text_input(
        input_class = "bs-theme-value-str",
        type = "number",
        step = opts$step
      ),
      bool = tagList(
        div(
          class = "form-check",
          tags$input(
            type = "checkbox",
            checked = if (value) NA else NULL,
            class = "bs-theme-value bs-theme-value-bool form-check-input",
            id = paste0(".bsthemer-", id),
            "data-id" = id
          ),
          tags$label(
            "for" = paste0(".bsthemer-", id),
            class = "form-check-label",
            lbl
          )
        ),
        if (!is.null(desc)) div(class = "form-text small", desc)
      ),
      switch = switch_input(),
      select = div(
        class = "form-row form-group",
        tags$label(class = "control-label", lbl),
        tags$select(
          class = "form-control",
          "data-id" = id,
          class = "bs-theme-value bs-theme-value-select",
          # This select is designed for the 'bootswatch' input and assumes that
          # the choices options are a list of lists, each list is an optgroup.
          lapply(seq_along(opts$choices), function(idx) {
            group_name <- names(opts$choices)[[idx]]
            choice_group <- opts$choices[[idx]]
            opts <- lapply(choice_group, function(x) {
              tags$option(
                value = x,
                selected = if (identical(x, value)) NA else NULL,
                tools::toTitleCase(x)
              )
            })
            if (!nzchar(group_name)) {
              return(opts)
            }
            tags$optgroup(label = group_name, opts)
          })
        ),
        if (!is.null(desc)) div(class = "form-text small", desc)
      ),
      stop("unknown type")
    )
  }

  version <- theme_version(theme)
  accordion <- lapply(seq_along(opts), function(i) {
    opt_name <- names(opts)[[i]]
    elId <- paste0("bsthemerCollapse", i)
    btn <- tags$button(
      class = if (version >= 5) {
        "accordion-button"
      } else {
        "btn btn-link px-3 py-2 w-100 text-left border-0"
      },
      class = if (i != 1) "collapsed",
      "data-toggle" = "collapse",
      "data-target" = paste0("#", elId),
      # data-bs-* is for BS5+
      "data-bs-toggle" = "collapse",
      "data-bs-target" = paste0("#", elId),
      "aria-expanded" = "true",
      "aria-controls" = elId,
      opt_name
    )
    controls <- lapply(seq_along(opts[[i]]), function(j) {
      make_control(names(opts[[i]])[[j]], opts[[i]][[j]])
    })
    div(
      class = if (version >= 5) "accordion-item",
      div(
        class = if (version >= 5) {
          "accordion-header"
        } else {
          "card-header p-0 border-0"
        },
        btn
      ),
      div(
        id = elId,
        class = if (i == 1) "show" else "collapse",
        "data-parent" = "#bsthemerAccordion",
        # data-bs-* is for BS5+
        "data-bs-parent" = "#bsthemerAccordion",
        class = if (version >= 5) "accordion-collapse",
        div(
          class = if (version >= 5) "accordion-body" else "card-body",
          controls
        )
      )
    )
  })

  withTags(
    tagList(
      colorpicker_deps(),
      htmlDependency(
        "bs_themer",
        version = get_package_version("bslib"),
        src = "themer",
        script = c("themer.js"),
        package = "bslib",
        all_files = FALSE
      ),

      div(
        id = "bsthemerContainer",
        class = "card shadow",
        style = css(
          # The bootstrap-colorpicker plugin sets a z-index of 1060 on
          # its inputs, so the container needs a smaller index, than that
          # https://github.com/rstudio/bslib/blob/e4da71f3/inst/lib/bs-colorpicker/css/bootstrap-colorpicker.css#L38
          #
          # It's also important that this z-index is higher than 1030 so it's
          # overlaid on-top of fixed/sticky navbars
          # https://github.com/rstudio/bslib/blob/e4da71f3/inst/lib/bs/scss/_variables.scss#L697-L701
          z_index = 1059,
          width = "18rem",
          max_height = "80vh",
          position = "fixed",
          top = "1rem",
          right = "1rem",
          height = "auto"
        ),

        div(
          id = "bsthemerHeader",
          class = "move-grabber",
          "data-target" = "#bsthemerContainer",
          class = "card-header font-weight-bold bg-dark text-light px-3 py-2",
          "Theme customizer",
          tags$div(
            id = "bsthemerToggle",
            class = "float-right",
            "data-toggle" = "collapse",
            "data-target" = "#bsthemerAccordion",
            # data-bs-* is for BS5+
            "data-bs-toggle" = "collapse",
            "data-bs-target" = "#bsthemerAccordion",
            style = css(cursor = "pointer"),
            tags$span(),
            bs_dependency_defer(themer_css_dependency)
          )
        ),

        div(
          id = "bsthemerAccordion",
          class = "collapse show",
          class = if (version >= 5) "accordion",
          style = css(overflow_y = "auto"),
          accordion
        )
      )
    )
  )
}

themer_css_dependency <- function(theme) {
  version <- get_package_version("bslib")
  bs_dependency(
    input = sass_file(system_file("themer/themer.scss", package = "bslib")),
    theme = theme,
    name = "bs-themer-css",
    version = version,
    cache_key_extra = version
  )
}

#' Theme customization UI
#'
#' A 'real-time' theme customization UI that you can use to easily make common
#' tweaks to Bootstrap variables and immediately see how they would affect your
#' app's appearance. There are two ways you can launch the theming UI. For most
#' Shiny apps, just use `run_with_themer()` in place of [shiny::runApp()]; they
#' should take the same arguments and work the same way. Alternatively, you can
#' call the `bs_themer()` function from inside your server function (or in an R
#' Markdown app that is using `runtime: shiny`, you can call this from any code
#' chunk). Note that this function is only intended to be used for development!
#'
#' To help you utilize the changes you see in the preview, this utility prints
#' [bs_theme()] code to the R console.
#'
#' @param appDir The application to run. This can be a file or directory path,
#'   or a [shiny::shinyApp()] object. See [shiny::runApp()] for details.
#' @param ... Additional parameters to pass through to [shiny::runApp()].
#' @param gfonts whether or not to detect Google Fonts and wrap them in
#'   [font_google()] (so that their font files are automatically imported).
#' @param gfonts_update whether or not to update the internal database of
#'   Google Fonts.
#'
#' @section Limitations:
#'
#'   * Doesn't work with Bootstrap 3.
#'   * Doesn't work with IE11.
#'   * Only works inside Shiny apps and `runtime: shiny` R Markdown documents.
#'     * Can't be used with static R Markdown documents.
#'     * Can be used to some extent with `runtime: shiny_prerendered`, but only UI
#'       rendered through a `context="server"` may update in real-time.
#'   * Doesn't work with '3rd party' custom widgets that don't make use of
#'     [bs_dependency_defer()] or [bs_current_theme()].
#'
#' @return nothing. These functions are called for their side-effects.
#'
#' @examples
#' library(shiny)
#' library(bslib)
#'
#' ui <- fluidPage(
#'   theme = bs_theme(bg = "black", fg = "white"),
#'   h1("Heading 1"),
#'   h2("Heading 2"),
#'   p(
#'     "Paragraph text;",
#'     tags$a(href = "https://www.rstudio.com", "a link")
#'   ),
#'   p(
#'     actionButton("cancel", "Cancel"),
#'     actionButton("continue", "Continue", class = "btn-primary")
#'   ),
#'   tabsetPanel(
#'     tabPanel("First tab",
#'       "The contents of the first tab"
#'     ),
#'     tabPanel("Second tab",
#'       "The contents of the second tab"
#'     )
#'   )
#' )
#'
#' if (interactive()) {
#'   run_with_themer(shinyApp(ui, function(input, output) {}))
#' }
#'
#' @export
run_with_themer <- function(
  appDir = getwd(),
  ...,
  gfonts = TRUE,
  gfonts_update = FALSE
) {
  shiny::runApp(
    as_themer_app(appDir, gfonts = gfonts, gfonts_update = gfonts_update),
    ...
  )
}

as_themer_app <- function(appDir, gfonts = TRUE, gfonts_update = FALSE) {
  obj <- shiny::as.shiny.appobj(appDir)
  origServerFuncSource <- obj[["serverFuncSource"]]
  obj[["serverFuncSource"]] <- function() {
    origServerFunc <- origServerFuncSource()
    function(input, output, session, ...) {
      bs_themer(gfonts, gfonts_update)
      if (!"session" %in% names(formals(origServerFunc))) {
        origServerFunc(input, output, ...)
      } else {
        origServerFunc(input, output, session, ...)
      }
    }
  }
  obj
}

#' @rdname run_with_themer
#' @export
bs_themer <- function(gfonts = TRUE, gfonts_update = FALSE) {
  session <- get_current_session()
  if (!identical("ok", session$ns("ok"))) {
    stop(
      call. = FALSE,
      "`bslib::bs_themer()` must be called from within a ",
      "top-level Shiny server function, not a Shiny module server function"
    )
  }
  if (!is_installed("shiny", "1.6.0")) {
    stop(call. = FALSE, "`bslib::bs_themer()` requires shiny v1.6.0 or higher")
  }
  theme <- get_current_theme()
  if (!is_bs_theme(theme)) {
    stop(
      call. = FALSE,
      "`bslib::bs_themer()` requires `shiny::bootstrapLib()` to be present ",
      "in the app's UI. Consider providing `bslib::bs_theme()` to the theme argument of the ",
      "relevant page layout function (or, more generally, adding `bootstrapLib(bs_theme())` to the UI."
    )
  }
  preset_initial <- theme_preset_info(theme)$name
  switch_version(
    theme,
    three = stop("Interactive theming for Bootstrap 3 isn't supported")
  )
  if (isTRUE(session$userData[["bs_themer_init"]])) {
    # bs_themer() was called multiple times for the same session
    return()
  } else {
    session$userData[["bs_themer_init"]] <- TRUE
  }

  gfont_info <- if (isTRUE(gfonts)) get_gfont_info(gfonts_update)

  # Insert the theming control panel with values informed by the theme settings
  themer_opts <- opts_metadata(theme)
  themer_vars <- unlist(unname(lapply(themer_opts, names)))
  sass_vars <- setdiff(themer_vars, c("preset", "dark-mode"))
  themer_vals <- as.list(get_themer_vals(theme, sass_vars))
  themer_vals$preset <- preset_initial
  shiny::insertUI(
    "body",
    where = "beforeEnd",
    ui = bs_themer_ui(themer_opts, themer_vals, theme)
  )

  input <- session$input

  # We emit different 'code' for runtime:shiny in Rmd
  isRmd <- is_shiny_runtime()

  # When the bootswatch theme changes, update the themer's state to reflect
  # the new variable defaults. Note that we also update the "input theme",
  # and effectively throw out any other theming changes made (i.e., start from a new theme)
  # since it'd be messy to figure out whether changes are "real" or just a
  # consequence of a new bootswatch value
  shiny::observeEvent(input$bs_theme_preset, {
    theme <<- set_current_theme(
      theme,
      list(preset = input$bs_theme_preset),
      session,
      rmd = isRmd
    )
    vals <- as.list(bs_get_variables(theme, sass_vars))
    session$sendCustomMessage("bs-themer-preset", list(values = vals))
  })

  # Fires when anything other then the Bootswatch theme changes
  shiny::observeEvent(input$bs_theme_vars, {
    vals <- jsonlite::parse_json(input$bs_theme_vars)

    # Validate that `vals` is a simple list, containing atomic elements,
    # that are all named
    if (
      !identical(class(vals), "list") ||
        !all(
          vapply(
            vals,
            function(x) {
              is.atomic(x) || is.null(x)
            },
            logical(1)
          )
        ) ||
        is.null(names(vals)) ||
        !isTRUE(all(nzchar(names(vals), keepNA = TRUE)))
    ) {
      warning(
        call. = FALSE,
        "bs_themer() encountered malformed input; ignoring"
      )
      return()
    }

    # Makes remaining logic simpler to reason about
    if (length(vals) == 0) {
      return()
    }

    # Remember, theme at this point has been updated to reflect the current Bootswatch theme,
    # so re-query Sass values from the (possibly updated) theme, then filter down to meaningful
    # differences
    theme_vals <- get_themer_vals(theme, names(vals[sass_vars]))
    changed_vals <- as.list(diff_css_values(vals[sass_vars], theme_vals))

    if (!identical(preset_initial, input$bs_theme_preset)) {
      changed_vals$preset <- input$bs_theme_preset
    }

    # If _either_ fg/bg has changed, bs_theme() must to be called with *both* fg and bg populated.
    if (any(c("bg", "fg") %in% names(changed_vals))) {
      changed_vals[["bg"]] <- changed_vals[["bg"]] %||% vals[["bg"]]
      changed_vals[["fg"]] <- changed_vals[["fg"]] %||% vals[["fg"]]
    }

    # Change variables names to their 'high-level' equivalents
    changed_vals <- rename2(
      changed_vals,
      "font-family-base" = "base_font",
      "font-family-monospace" = "code_font",
      "headings-font-family" = "heading_font",
      "font-size-base" = "font_scale"
    )

    if (length(changed_vals$font_scale)) {
      changed_vals$font_scale <- as.numeric(changed_vals$font_scale)
    }

    if (isTRUE(gfonts)) {
      for (var in c("base_font", "code_font", "heading_font")) {
        changed_vals[[var]] <- insert_font_google_call(
          changed_vals[[var]],
          gfont_info
        )
      }
    }

    set_current_theme(theme, changed_vals, session, rmd = isRmd)
  })
}

get_themer_vals <- function(theme, vars) {
  vals <- bs_get_variables(theme, vars)
  if (!grepl("rem$", vals[["font-size-base"]])) {
    stop(
      "font-size-base must have a CSS unit length type of rem",
      call. = FALSE
    )
  }
  vals[["font-size-base"]] <- sub("rem$", "", vals[["font-size-base"]])
  vals
}

set_current_theme <- function(theme, changed_vals, session, rmd = FALSE) {
  shiny::insertUI(
    "body",
    ui = spinner_overlay(),
    immediate = TRUE,
    session = session
  )
  on.exit(shiny::removeUI("body > #spinner_overlay"), add = TRUE)

  # Construct the code/yaml to display to the user
  if (isTRUE(rmd)) {
    display_vals <- lapply(changed_vals, function(x) {
      if (is.numeric(x)) {
        return(x)
      }
      if (rlang::is_call(x)) {
        str <- paste0(deparse(x, width.cutoff = 500L), collapse = "")
        return(paste("!expr", str))
      }
      # To avoid yaml parse errors with values that contain # or ",
      # first escape ", then in quote the value
      paste0('"', gsub('"', '\\"', x, fixed = TRUE), '"')
    })
    message("\n####  Update your Rmd output format's theme:  ####")
    cat(
      paste0(
        "    theme:\n",
        paste0(
          collapse = "\n",
          "      ",
          names(display_vals),
          ": ",
          display_vals
        ),
        "\n"
      )
    )
  } else {
    message("\n####  Update your bs_theme() R code with:  #####")
    print(rlang::expr(bs_theme_update(theme, !!!changed_vals)))
  }

  # Color contrast warnings are more annoying then they are useful inside the theming widget
  opts <- options(bslib.color_contrast_warnings = FALSE)
  on.exit(options(opts), add = TRUE)

  # the actual code that we evaluate should not have quoted expressions
  changed_vals[] <- lapply(changed_vals, eval_val)
  code <- rlang::expr(bs_theme_update(theme, !!!changed_vals))
  theme <- rlang::eval_tidy(code)
  # Prevent Sass compilation errors from crashing the app and relay a message to user.
  # Errors can happen if the users enters values that lead to unexpected Sass
  # expressions (e.g., "$foo: * !default")
  shiny::removeNotification("sass-compilation-error", session = session)
  tryCatch(
    session$setCurrentTheme(theme),
    error = function(e) {
      shiny::showNotification(
        "Sass -> CSS compilation failed, likely due to invalid user input.
         Other theming changes won't take effect until the invalid input is fixed.",
        duration = NULL,
        id = "sass-compilation-error",
        type = "error",
        session = session
      )
    }
  )
  invisible(theme)
}

spinner_overlay <- function() {
  tagList(
    tags$style(
      "@supports ((-webkit-backdrop-filter:blur(4px)) or (backdrop-filter:blur(4px))) {
        #spinner_overlay{ -webkit-backdrop-filter:blur(4px); backdrop-filter:blur(4px); background-color:rgba(255,255,255,.05);}
      }"
    ),
    div(
      id = "spinner_overlay",
      style = "position:absolute; top:0; left:0; min-height:100vh; width:100%; background-color:rgba(255,255,255,.8); z-index:100000",
      class = "d-flex flex-column justify-content-center align-items-center",
      div(
        class = "spinner-border",
        style = "width:5rem; height:5rem; color: rgba(0,0,0,0.8);",
        role = "status",
        span(class = "sr-only visually-hidden", "Refreshing stylesheets...")
      ),
      span(
        class = "lead mt-1",
        style = "color: rgba(0,0,0,0.8);",
        "Refreshing stylesheets..."
      )
    )
  )
}

eval_val <- function(x) {
  if (is.call(x)) {
    return(eval(x))
  }
  if (!is.list(x)) {
    return(x)
  }
  lapply(x, eval_val)
}

insert_font_google_call <- function(val, gfont_info) {
  # val should be a non-empty character string
  if (!is_string(val)) {
    return(NULL)
  }
  if (!nzchar(val)) {
    return(NULL)
  }
  fams <- strsplit(as.character(val), ",")[[1]]
  fams <- vapply(
    fams,
    function(x) gsub("^\\s*['\"]?", "", gsub("['\"]?\\s*$", "", x)),
    character(1),
    USE.NAMES = FALSE
  )
  fams <- fams[nzchar(fams)]
  is_a_gfont <- tolower(fams) %in% tolower(gfont_info$family)
  if (length(fams) == 1) {
    return(if (is_a_gfont) call("font_google", fams) else fams)
  }
  fams <- as.list(fams)
  for (i in which(is_a_gfont)) {
    fams[[i]] <- call("font_google", fams[[i]])
  }
  rlang::expr(font_collection(!!!unname(fams)))
}

get_gfont_info <- function(update = FALSE) {
  if (isTRUE(update)) {
    jsonlite::fromJSON(gfont_api_url())$items
  } else {
    # See tools/update_gfont_info.R
    gfont_info
  }
}

# same as thematic:::gfont_api_url
gfont_api_url <- function() {
  paste0("https://www.googleapis.com/webfonts/v1/webfonts?key=", gfont_key())
}
# same as thematic:::gfont_key
# As mentioned in the developer API, this key is safe to be public facing
# https://developers.google.com/fonts/docs/developer_api
gfont_key <- function() {
  Sys.getenv(
    "GFONT_KEY",
    paste0("AIzaSyDP", "KvElVqQ-", "26f7tjxyg", "IGpIajf", "tS_zmas")
  )
}

#' Retrieve Sass variable values from the current theme
#'
#' Useful for retrieving a variable from the current theme and using
#' the value to inform another R function.
#'
#' @inheritParams bs_theme_update
#' @param varnames A character string referencing a Sass variable in the current
#'   theme.
#'
#' @return Returns a character string containing a CSS/Sass value. If the
#'   variable(s) are not defined, their value is `NA`.
#'
#' @references [Theming: Bootstrap 5 variables](https://rstudio.github.io/bslib/articles/bs5-variables/index.html)
#'   provides a searchable reference of all theming variables available in
#'   Bootstrap 5.
#'
#' @export
#' @family Bootstrap theme utility functions
#'
#' @examples
#' vars <- c("body-bg", "body-color", "primary", "border-radius")
#' bs_get_variables(bs_theme(), varnames = vars)
#' bs_get_variables(bs_theme(bootswatch = "darkly"), varnames = vars)
bs_get_variables <- function(theme, varnames) {
  if (length(varnames) == 0) {
    return(stats::setNames(character(0), character(0)))
  }

  # Our bg/fg are not actual Sass variables and can mean different things depending
  # on the bootswatch theme/version
  base_color_idx <- varnames %in% c("fg", "bg")
  if (any(base_color_idx)) {
    varnames[base_color_idx] <- rename2(
      varnames[base_color_idx],
      !!!get_base_color_map(theme)
    )
  }

  assert_bs_theme(theme)

  # Support both `bs_get_variables("$foo")` and `bs_get_variables("foo")`
  # (note that `sass::sass("$$foo:1;")` is illegal; so this seems safe)
  varnames <- sub("^\\$", "", varnames)

  # It's possible that some varnames refer to variables that aren't defined.
  # This would normally cause a crash. We define last-ditch defaults here,
  # with a magic constant that we can swap out for NA before returning to
  # the user.
  na_sentinel <- "NA_SENTINEL_CONSTANT_4902F4E"
  sassvars <- paste0(
    "$",
    varnames,
    ": ",
    na_sentinel,
    " !default;",
    collapse = "\n"
  )

  # Declare a block with a meaningless but identifiable selector (.__rstudio_bslib_get_variables)
  # and add properties for each variable that is desired.
  cssvars <- paste0(
    "--",
    varnames,
    ": #{inspect($",
    varnames,
    ")};",
    collapse = "\n"
  )
  cssvars <- sprintf(":root.__rstudio_bslib_get_variables {\n %s \n}", cssvars)

  css <- sass_partial(
    cssvars,
    # Add declarations to the current theme
    bs_bundle(theme, sass_layer(mixins = sassvars)),
  )

  # Search the output for the block of properties we just generated, using the
  # ".__rstudio_bslib_get_variables" selector. The capture group will include all of the
  # properties we care about in a single string (the propstr variable below).
  matches <- regexec(
    "(:root)?\\.__rstudio_bslib_get_variables(:root)?\\s*\\{\\s*\\n(.*?)\\n\\s*\\}",
    css
  )
  propstr <- regmatches(css, matches)[[1]][4]
  if (is.na(propstr)) {
    stop("bs_global_get_variables failed; expected selector was not found")
  }
  # Split the propstr by newline, so we can perform vectorized regex operations
  # on all of the variables at once.
  proplines <- strsplit(propstr, "\n")[[1]]

  # Parse each line for the name and value.
  matches2 <- regmatches(
    proplines,
    regexec("\\s*--([^:]+):\\s*(.*);$", proplines)
  )
  names <- vapply(matches2, function(x) x[2], character(1))
  values <- vapply(matches2, function(x) x[3], character(1))

  if (any(is.na(names))) {
    stop(
      "bs_global_get_variables failed; generated output was in an unexpected format"
    )
  }
  if (!identical(varnames, names)) {
    stop("bs_global_get_variables failed; expected properties were not found")
  }

  # Any variables that had to fall back to our defaults, we'll replace with NA
  values[values == na_sentinel] <- NA_character_

  if (any(base_color_idx)) {
    varnames[base_color_idx] <- rename2(
      varnames[base_color_idx],
      !!!get_base_color_map(theme, decode = FALSE)
    )
  }

  # Return as a named character vector
  stats::setNames(values, varnames)
}

diff_css_values <- function(a, b) {
  stopifnot(all(!is.na(a)))
  stopifnot(identical(names(a), names(b)))
  stopifnot(is.list(a))
  if (!is.character(b)) {
    browser()
  }

  a_char <- vapply(
    a,
    function(x) {
      if (is.null(x) || isTRUE(is.na(x))) {
        "null"
      } else if (is.logical(x)) {
        tolower(as.character(x))
      } else if (is.character(x)) {
        x
      } else {
        as.character(x)
      }
    },
    character(1)
  )

  b <- ifelse(is.na(b), "null", b)

  # Normalize colors; ignore things that don't seem to be colors. This is
  # necessary so we don't consider "black", "#000", "#000000", "rgb(0,0,0,1)",
  # etc. to be distinct values.
  #
  # Note: This won't work with values that are colors AND other things, like
  # "solid #000 3px"; it needs the value to be solely a color to be normalized.

  a_char_colors <- htmltools::parseCssColors(a_char, mustWork = FALSE)
  a_char <- ifelse(!is.na(a_char_colors), a_char_colors, a_char)

  b_colors <- htmltools::parseCssColors(b, mustWork = FALSE)
  b <- ifelse(!is.na(b_colors), b_colors, b)

  idx <- ifelse(is.na(b), TRUE, a_char != b)
  a[idx]
}

#' @rdname bs_get_variables
#' @inheritParams bs_get_variables
#' @export
#' @examples
#'
#' bs_get_contrast(bs_theme(), c("primary", "dark", "light"))
#'
#' library(htmltools)
#' div(
#'   class = "bg-primary",
#'   style = css(
#'     color = bs_get_contrast(bs_theme(), "primary")
#'   )
#' )
#'
bs_get_contrast <- function(theme, varnames) {
  stopifnot(is.character(varnames))
  stopifnot(length(varnames) > 0)

  varnames <- sub("^\\$", "", varnames)
  prop_string <- paste0(
    paste0(varnames, ": color-contrast($", varnames, ");"),
    collapse = "\n"
  )
  css <- sass::sass_partial(
    paste0("bs_get_contrast {", prop_string, "}"),
    theme,
    cache_key_extra = get_package_version("bslib"),
    # Don't listen to global Sass options so we can be sure
    # that stuff like source maps won't be included
    options = sass::sass_options(source_map_embed = FALSE)
  )
  css <- gsub("\n", "", gsub("\\s*", "", css))
  css <- sub("bs_get_contrast{", "", css, fixed = TRUE)
  css <- sub("\\}$", "", css)
  props <- strsplit(strsplit(css, ";")[[1]], ":")
  setNames(
    vapply(
      props,
      function(x) htmltools::parseCssColors(sub(";$", "", x[2])),
      character(1)
    ),
    vapply(props, `[[`, character(1), 1)
  )
}

#' A Bootstrap card component
#'
#' A general purpose container for grouping related UI elements together with a
#' border and optional padding. To learn more about [card()]s, see [this
#' article](https://rstudio.github.io/bslib/articles/cards.html).
#'
#' @param ... Unnamed arguments can be any valid child of an [htmltools
#'   tag][htmltools::tags] (which includes card items such as [card_body()].
#'   Named arguments become HTML attributes on returned UI element.
#' @param full_screen If `TRUE`, an icon will appear when hovering over the card
#'   body. Clicking the icon expands the card to fit viewport size.
#' @param height Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `height="200px"`). Doesn't apply when a card is made `full_screen`
#'   (in this case, consider setting a `height` in [card_body()]).
#' @param max_height,min_height Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `max_height="200px"`). Doesn't apply when a card is made `full_screen`
#'   (in this case, consider setting a `max_height` in [card_body()]).
#' @param fill Whether or not to allow the card to grow/shrink to fit a
#'   fillable container with an opinionated height (e.g., `page_fillable()`).
#' @param class Additional CSS classes for the returned UI element.
#' @param wrapper A function (which returns a UI element) to call on unnamed
#'   arguments in `...` which are not already card item(s) (like
#'   [card_header()], [card_body()], etc.). Note that non-card items are grouped
#'   together into one `wrapper` call (e.g. given `card("a", "b",
#'   card_body("c"), "d")`, `wrapper` would be called twice, once with `"a"` and
#'   `"b"` and once with `"d"`).
#'
#' @return A [htmltools::div()] tag.
#'
#' @export
#' @seealso [card_body()] for putting stuff inside the card.
#' @seealso [navset_card_tab()] for cards with multiple tabs.
#' @seealso [layout_column_wrap()] for laying out multiple cards (or multiple
#'   columns inside a card).
#' @examples
#'
#' library(htmltools)
#'
#' if (interactive()) {
#'   card(
#'     full_screen = TRUE,
#'     card_header(
#'       "This is the header"
#'     ),
#'     card_body(
#'       p("This is the body."),
#'       p("This is still the body.")
#'     ),
#'     card_footer(
#'       "This is the footer"
#'     )
#'   )
#' }
#'
card <- function(..., full_screen = FALSE, height = NULL, max_height = NULL, min_height = NULL, fill = TRUE, class = NULL, wrapper = card_body) {

  args <- rlang::list2(...)
  argnames <- rlang::names2(args)

  attribs <- args[nzchar(argnames)]
  children <- as_card_items(args[!nzchar(argnames)], wrapper = wrapper)

  tag <- div(
    class = "card bslib-card bslib-mb-spacing",
    style = css(
      height = validateCssUnit(height),
      max_height = validateCssUnit(max_height),
      min_height = validateCssUnit(min_height)
    ),
    "data-bslib-card-init" = NA,
    "data-full-screen" = if (full_screen) "false",
    !!!attribs,
    !!!children,
    if (full_screen) full_screen_toggle(),
    card_dependency(),
    card_init_js()
  )

  tag <- bindFillRole(tag, container = TRUE, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "card()")
  )
}


as_card_items <- function(children, wrapper) {
  # We don't want NULLs creating empty card bodys
  children <- children[vapply(children, function(x) length(x) > 0, logical(1))]

  if (!is.function(wrapper)) {
    return(children)
  }

  # Any children that are `is.card_item` should be included verbatim. Any
  # children that are not, should be wrapped in card_body(). Consecutive children
  # that are not card_item, should be wrapped in a single card_body().
  needs_wrap <- !vapply(children, is.card_item, logical(1))
  needs_wrap_rle <- rle(needs_wrap)
  start_indices <- c(1, head(cumsum(needs_wrap_rle$lengths) + 1, -1))
  children <- mapply(
    start_indices, needs_wrap_rle$lengths, needs_wrap_rle$values,
    FUN = function(start, length, wrap) {
      these_children <- children[start:(start + length - 1)]
      if (wrap) {
        list(wrapper(these_children))
      } else {
        these_children
      }
    },
    SIMPLIFY = FALSE
  )
  unlist(children, recursive = FALSE)
}

#' Card items
#'
#' Components designed to be provided as direct children of a [card()]. For a
#' general overview of the [card()] API, see [this
#' article](https://rstudio.github.io/bslib/articles/cards.html).
#'
#' @param ... Unnamed arguments can be any valid child of an [htmltools
#'   tag][htmltools::tags]. Named arguments become HTML attributes on returned
#'   UI element.
#' @param min_height,max_height,max_height_full_screen Any valid [CSS length
#'   unit][htmltools::validateCssUnit()].
#' @param fillable Whether or not the card item should be a fillable (i.e.
#'   flexbox) container.
#' @param fill Whether to allow this element to grow/shrink to fit its `card()`
#'   container.
#' @param gap A [CSS length unit][htmltools::validateCssUnit()] defining the
#'   `gap` (i.e., spacing) between elements provided to `...`. This argument is only applicable when `fillable = TRUE`
#' @inheritParams card
#' @inheritParams page_fillable
#'
#' @return An [htmltools::div()] tag.
#'
#' @export
#' @seealso [card()] for creating a card component.
#' @seealso [navset_card_tab()] for cards with multiple tabs.
#' @seealso [layout_column_wrap()] for laying out multiple cards (or multiple
#'   columns inside a card).
#'
#' @describeIn card_body A general container for the "main content" of a [card()].
card_body <- function(..., fillable = TRUE, min_height = NULL, max_height = NULL, max_height_full_screen = max_height, height = NULL, padding = NULL, gap = NULL, fill = TRUE, class = NULL) {

  if (fillable) {
    register_runtime_package_check("`card_body()`", "shiny", "1.7.4")
    register_runtime_package_check("`card_body()`", "htmlwidgets", "1.6.0")
  }

  tag <- div(
    class = "card-body bslib-gap-spacing",
    style = css(
      min_height = validateCssUnit(min_height),
      "--bslib-card-body-max-height" = validateCssUnit(max_height),
      "--bslib-card-body-max-height-full-screen" = validateCssUnit(max_height_full_screen),
      margin_top = "auto",
      margin_bottom = "auto",
      # .card-body already adds `flex: 1 1 auto` so make sure to override it
      flex = if (fill) "1 1 auto" else "0 0 auto",
      padding = validateCssPadding(padding),
      gap = validateCssUnit(gap),
      height = validateCssUnit(height)
    ),
    ...
  )

  tag <- bindFillRole(tag, item = fill, container = fillable)

  # Make sure user has the opportunity to override the classes added by bindFillRole()
  tag <- tagAppendAttributes(tag, class = class)

  as.card_item(tag)
}


#' @describeIn card_body Similar to `card_header()` but without the border and background color.
#' @param container a function to generate an HTML element.
#' @export
card_title <- function(..., container = htmltools::h5) {
  container(...)
}


#' @describeIn card_body A header (with border and background color) for the `card()`. Typically appears before a `card_body()`.
#' @param container a function that generates an [htmltools tag][htmltools::tags].
#' @export
card_header <- function(..., class = NULL, container = htmltools::div) {
  as.card_item(
    container(class = "card-header", class = class, ...)
  )
}

#' @describeIn card_body A header (with border and background color) for the `card()`. Typically appears after a `card_body()`.
#' @export
card_footer <- function(..., class = NULL) {
  as.card_item(
    div(class = "card-footer", class = class, ...)
  )
}

#' @describeIn card_body Include static (i.e., pre-generated) images.
#' @param file a file path pointing an image. The image will be base64 encoded
#' and provided to the `src` attribute of the `<img>`. Alternatively, you may
#' set this value to `NULL` and provide the `src` yourself.
#' @param href an optional URL to link to.
#' @param border_radius where to apply `border-radius` on the image.
#' @param mime_type the mime type of the `file`.
#' @param container a function to generate an HTML element to contain the image.
#' @param width Any valid [CSS unit][htmltools::validateCssUnit] (e.g., `width="100%"`).
#' @export
card_image <- function(
  file, ..., href = NULL, border_radius = c("top", "bottom", "all", "none"),
  mime_type = NULL, class = NULL, height = NULL, fill = TRUE, width = NULL, container = card_body) {

  src <- NULL
  if (length(file) > 0) {
    src <- base64enc::dataURI(
      file = file, mime = mime_type %||% mime::guess_type(file)
    )
  }

  image <- tags$img(
    src = src,
    class = "img-fluid",
    class = switch(
      match.arg(border_radius),
      all = "card-img",
      top = "card-img-top",
      bottom = "card-img-bottom",
      NULL
    ),
    style = css(
      height = validateCssUnit(height),
      width = validateCssUnit(width)
    ),
    ...
  )

  image <- bindFillRole(image, item = fill)
  image <- tagAppendAttributes(image, class = class)

  if (!is.null(href)) {
    image <- bindFillRole(tags$a(href = href, image), container = TRUE, item = TRUE)
  }

  if (is.function(container)) {
    image <- container(image)
  }

  image
}

#' @describeIn card_body Mark an object as a card item. This will prevent the
#'   [card()] from putting the object inside a `wrapper` (i.e., a
#'   `card_body()`).
#' @param x an object to test (or coerce to) a card item.
#' @export
as.card_item <- function(x) {
  class(x) <- c("card_item", class(x))
  x
}

#' @rdname card_body
#' @export
is.card_item <- function(x) {
  inherits(x, "card_item")
}


full_screen_toggle <- function() {
  tooltip(
    tags$span(
      class = "bslib-full-screen-enter",
      class = "badge rounded-pill bg-dark",
      full_screen_toggle_icon()
    ),
    "Expand"
  )
}

card_dependency <- function() {
  list(
    component_dependency_js("card"),
    bs_dependency_defer(card_dependency_sass)
  )
}

card_dependency_sass <- function(theme) {
  component_dependency_sass(theme, "card")
}

card_init_js <- function() {
  tags$script(
    `data-bslib-card-init` = NA,
    HTML("bslib.Card.initializeAllCards();")
  )
}

full_screen_toggle_icon <- function() {
  if (is_installed("bsicons")) {
    return(bsicons::bs_icon("arrows-fullscreen"))
  }
  HTML('<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16" class="bi bi-arrows-fullscreen " style="height:1em;width:1em;fill:currentColor;" aria-hidden="true" role="img" ><path fill-rule="evenodd" d="M5.828 10.172a.5.5 0 0 0-.707 0l-4.096 4.096V11.5a.5.5 0 0 0-1 0v3.975a.5.5 0 0 0 .5.5H4.5a.5.5 0 0 0 0-1H1.732l4.096-4.096a.5.5 0 0 0 0-.707zm4.344 0a.5.5 0 0 1 .707 0l4.096 4.096V11.5a.5.5 0 1 1 1 0v3.975a.5.5 0 0 1-.5.5H11.5a.5.5 0 0 1 0-1h2.768l-4.096-4.096a.5.5 0 0 1 0-.707zm0-4.344a.5.5 0 0 0 .707 0l4.096-4.096V4.5a.5.5 0 1 0 1 0V.525a.5.5 0 0 0-.5-.5H11.5a.5.5 0 0 0 0 1h2.768l-4.096 4.096a.5.5 0 0 0 0 .707zm-4.344 0a.5.5 0 0 1-.707 0L1.025 1.732V4.5a.5.5 0 0 1-1 0V.525a.5.5 0 0 1 .5-.5H4.5a.5.5 0 0 1 0 1H1.732l4.096 4.096a.5.5 0 0 1 0 .707z"></path></svg>')
}


# jcheng 2022-06-06: Removing for now; list items have more features than I'm
# ready to design an API for right now
#
# #' @rdname card_body
# #' @export
# card_list <- function(...) {
#   res <- tags$ul(class = "list-group list-group-flush", ...)
#   as.card_item(res)
# }
#
# #' @export
# card_list_item <- function(...) {
#   tags$li(class = "list-group-item", ...)
# }

library(shiny)
library(bslib)
library(dplyr)
library(lubridate)
library(plotly)
library(nycflights13)
library(histoslider)
library(rlang)

# Data prep
flights <- flights %>%
  left_join(
    airports %>%
      transmute(dest_name = paste0(name, " (", faa, ")"), dest = faa, end_lat = lat, end_lon = lon)
  ) %>%
  filter(!is.na(dest_name)) %>%
  left_join(
    airports %>% select(origin = faa, start_lat = lat, start_lon = lon)
  ) %>%
  left_join(airlines, by = "carrier") %>%
  rename(carrier_name = name) %>%
  left_join(weather) %>%
  mutate(
    # The overwhelming majority of precipation is 0 so transform
    # with some help from MASS::boxcox() https://stats.stackexchange.com/a/1452/48604
    precip = scales::rescale(precip^-1.55/-.55),
    date = lubridate::ymd(paste(year, month, day, sep = "-"))
  )

PRIMARY <- "#0675DD"

sidebar_acc <- accordion(
  open = c("Origin", "Destination"),
  accordion_panel(
    "Origin", icon = icon("plane-departure"),
    uiOutput("origin_reset"),
    checkboxGroupInput(
      "origin", NULL,
      choices = sort(unique(flights$origin)),
      inline = TRUE
    )
  ),
  accordion_panel(
    "Destination", icon = icon("plane-arrival"),
    selectInput(
      "dest_name", NULL,
      sort(unique(flights$dest_name)),
      multiple = TRUE,
      width = "100%"
    )
  ),
  accordion_panel(
    "Carrier", icon = icon("user-tie"),
    selectInput(
      "carrier_name", NULL,
      unique(flights$carrier_name),
      multiple = TRUE,
      width = "100%"
    )
  ),
  accordion_panel(
    "Flight time", icon = icon("clock"),
    input_histoslider(
      "sched_dep_time", "Departure time",
      flights$sched_dep_time, height = 125,
      options = list(
        handleLabelFormat = "0d",
        selectedColor = PRIMARY
      )
    ),
    input_histoslider(
      "sched_arr_time", "Arrival time",
      flights$sched_arr_time, height = 125,
      options = list(
        handleLabelFormat = "0d",
        selectedColor = PRIMARY
      )
    ),
    input_histoslider(
      "date", "Date",
      flights$date, height = 125, breaks = "months",
      options = list(
        handleLabelFormat = "%b %e",
        selectedColor = PRIMARY
      )
    )
  ),
  accordion_panel(
    "Weather", icon = icon("cloud-rain"),
    input_histoslider(
      "precip", "Precipitation",
      flights$precip, height = 125,
      options = list(selectedColor = PRIMARY)
    ),
    input_histoslider(
      "wind_speed", "Wind speed",
      flights$wind_speed, height = 125,
      options = list(selectedColor = PRIMARY)
    ),
    input_histoslider(
      "wind_gust", "Wind gust",
      flights$wind_gust, height = 125,
      options = list(selectedColor = PRIMARY)
    )
  )
)


flights_card <- card(
  full_screen = TRUE,
  card_header("Flight paths"),
  plotlyOutput("flight_paths"),
  card_footer("Marker areas are proportional to mean arrival delay")
)

delay_corr_card <- card(
  full_screen = TRUE,
  card_header(
    "Arrival vs departure delay",
    checkboxInput("scatter_summarize", "Summarize", TRUE, width = "fit-content"),
    class = "d-flex justify-content-between"
  ),
  plotlyOutput("scatter_delay")
)

delay_card <- navset_card_pill(
  title = "Arrival delay",
  full_screen = TRUE,
  nav_panel(
    "Overall",
    plotlyOutput("arr_delay")
  ),
  nav_panel(
    "Over time",
    plotlyOutput("arr_delay_series")
  )
)


ui <- page_navbar(
  theme = bs_theme(
    base_font = font_google(
      "Open Sans", wght = c(300, 400, 500, 600, 700, 800),
      ital = c(0, 1)
    ),
    "primary" = PRIMARY,
    "navbar-bg" = PRIMARY
  ),
  title = tags$span(
    tags$img(src = "logo.png", width = "46px", height = "auto", class = "me-3"),
    "NYC Flights"
  ),
  fillable = TRUE,
  sidebar = sidebar(sidebar_acc),
  nav_panel(
    "Delays",
    uiOutput("value_boxes"),
    layout_column_wrap(
      width = "200px", class = "my-3",
      flights_card, delay_corr_card
    ),
    delay_card
  ),
  nav_panel(
    "Durations",
    "Coming soon"
  ),
  nav_spacer(),
  nav_item(
    tags$a(
      tags$span(
        bsicons::bs_icon("code-slash"), "Source code"
      ),
      href = "https://github.com/rstudio/bslib/tree/main/inst/examples/flights",
      target = "_blank"
    )
  )
)

server <- function(input, output, session) {
  # ---------------------------------------------------------
  # Flights tab logic
  #
  # WARNING: this server-side filtering logic is VERY experimental
  # at this point and won't be easily adapt to different use cases.
  # If you feel tempted to use it, use with caution.
  # ---------------------------------------------------------

  # Mapping from input id name to updating input function
  input_discrete_vars <- list(
    origin = function(...) updateCheckboxGroupInput(..., inline = TRUE),
    dest_name = updateSelectInput,
    carrier_name = updateSelectInput
  )
  input_numeric_vars <- list(
    sched_dep_time = update_histoslider,
    sched_arr_time = update_histoslider,
    date = update_histoslider
  )
  input_vars <- c(
    input_discrete_vars,
    input_numeric_vars
  )

  filter_index <- function(d) {
    idx <- rep(TRUE, nrow(d))
    for (var in names(d)) {
      idx <- idx & filter_col(d, var)
    }
    idx & !is.na(idx)
  }

  filter_col <- function(d, var) {
    vals <- d[[var]]
    input_val <- input[[var]]
    if (is.null(input_val)) {
      return(TRUE)
    }

    if (is.character(vals) || is.factor(vals) || is.logical(vals)) {
      return(d[[var]] %in% input_val)
    }

    # N.B. between() will remove NAs, which we probably don't
    # want to remove until the slider is considered 'active'
    rng <- range(vals, na.rm = TRUE)
    active <- isTRUE(rng[1] <= input_val[1] || input_val[2] <= rng[2])

    if (!active) {
      return(TRUE)
    }

    dplyr::between(vals, input_val[1], input_val[2])
  }

  # When a particular input filter changes, we need to update every other filter
  # with data
  lapply(names(input_vars), function(var) {

    index <- reactive({
      filter_index(flights[, setdiff(names(input_vars), var)])
    })

    observeEvent(index(), ignoreInit = TRUE, ignoreNULL = FALSE, {

      d <- flights[index(), ]
      if (nrow(d) == 0) return()
      input_val <- input[[var]]
      update_input_func <- input_vars[[var]]

      if (var %in% names(input_discrete_vars)) {
        choices <- sort(unique(d[[var]]))
        update_input_func(
          inputId = var, choices = choices, selected = input_val %||% character(0)
        )
      } else {
        update_input_func(
          id = var,
          values = d[[var]],
          start = input_val[1],
          end = input_val[2]
        )
      }

    })
  })

  output$origin_reset <- renderUI({
    actionLink(
      "origin_reset", "Reset",
      style = htmltools::css(
        text_decoration = "none",
        font_weight = 700,
        font_size = ".875rem",
        float = "right",
        visibility = if (length(input$origin) > 0) "visible" else "hidden",
        margin_top = "-7px"
      )
    )
  })

  observeEvent(input$origin_reset, {
    updateCheckboxGroupInput(
      inputId = "origin",
      choices = sort(unique(flights$origin)),
      selected = character(0)
    )
  })

  # Flights with all filters applied (i.e., data used for value boxes/plots)
  flight_dat <- reactive({
    flights[filter_index(flights), ]
  })

  output$value_boxes <- renderUI({
    d <- flight_dat()
    n_flights <- value_box(
      "A TOTAL OF",
      paste(nrow(d), "flights"),
      paste(
        "Across", length(unique(d$dest_name)),
        "destinations"
      ),
      tags$p(paste(
        "On", length(unique(d$carrier_name)), "different carriers"
      )),
      showcase = bsicons::bs_icon("airplane")
    )

    delay_dep <- value_box(
      "AVERAGE DEPARTURE",
      paste(
        round(mean(d$dep_delay, na.rm = T), 0),
        "mins late"
      ),
      paste0(
        round(100 * sum(d$dep_delay > 0, na.rm = T) / nrow(d), 1),
        "% of flights depart late"
      ),
      showcase = bsicons::bs_icon("hourglass-split")
    )

    delay_arr <- value_box(
      "AVERAGE ARRIVAL",
      paste(
        round(mean(d$arr_delay, na.rm = T), 0),
        "mins late"
      ),
      paste0(
        round(100 * sum(d$arr_delay > 0, na.rm = TRUE) / nrow(d), 1),
        "% of flights arrive late"
      ),
      showcase = bsicons::bs_icon("hourglass-bottom")
    )

    layout_column_wrap(width = 1/3, n_flights, delay_dep, delay_arr)
  }) %>%
    bindCache(flight_dat())

  output$flight_paths <- renderPlotly({
    flight_dat() %>%
      group_by(start_lon, start_lat, end_lon, end_lat, origin, dest) %>%
      summarise(mean_delay = mean(arr_delay, na.rm = TRUE)) %>%
      plot_geo(color = I("#0D6EFD"), showlegend = FALSE) %>%
      add_segments(
        x = ~start_lon, xend = ~end_lon,
        y = ~start_lat, yend = ~end_lat,
        alpha = 0.5, size = I(1), hoverinfo = "none"
      ) %>%
      add_markers(
        x = ~end_lon, y = ~end_lat, size = ~mean_delay,
        hoverinfo = "text", alpha = 0.1,
        text = ~paste0(
          origin, " -> ", dest, "<br>",
          "Average delay: ", round(mean_delay, 1)
        )
      ) %>%
      config(displayModeBar = FALSE) %>%
      layout(
        geo = list(
          projection = list(
            type = 'orthographic',
            rotation = list(lon = -100, lat = 40, roll = 0)
          ),
          showland = TRUE,
          landcolor = toRGB("gray95"),
          countrycolor = toRGB("gray80")
        )
      )
  }) %>%
    bindCache(flight_dat())

  output$scatter_delay <- renderPlotly({
    d <- flight_dat()
    req(nrow(d) > 0)

    if (isTRUE(input$scatter_summarize)) {
      x <- d$dep_delay
      y <- d$arr_delay
      p <- plot_ly() %>%
        add_histogram2dcontour(
          x = x, y = y, hovertemplate = "Number of flights: %{z}<extra></extra>"
        ) %>%
        layout(
          xaxis = list(
            title = "Departure delay",
            range = c(-30, 100)
          ),
          yaxis = list(
            title = "Arrival delay",
            range = c(-50, 100)
          )
        ) %>%
        colorbar(title = "Number of flights", y = 0.5, yanchor = "middle")

      return(p)
    }

    m <- loess(arr_delay~dep_delay, data = d)
    rng <- range(d$dep_delay, na.rm = TRUE)
    grid <- data.frame(
      dep_delay = seq(rng[1], rng[2], length = 100)
    )
    grid <- modelr::add_predictions(grid, m)

    plot_ly() %>%
      add_markers(
        data = d,
        x = ~dep_delay,
        y = ~arr_delay,
        stroke = I(PRIMARY),
        color = I("transparent"),
        span = I(1),
        text = ~paste0(
          origin, " -> ",
          substr(dest_name, nchar(dest_name) - 3, nchar(dest_name) - 1), "\n",
          carrier_name, "\n",
          "Arrival: ", arr_delay, "\n",
          "Departure: ", dep_delay, "\n"
        ),
        hoverinfo = "text",
        name = "Flights"
      ) %>%
      add_lines(
        data = grid,
        x = ~dep_delay,
        y = ~pred,
        text = ~paste0("Given a departure delay of ", round(dep_delay), "\n expect an arrival delay of ", round(pred)),
        hoverinfo = "text",
        name = "Predicted"
      ) %>%
      layout(
        legend = list(
          orientation = "h",
          x = 0.1, xanchor = "left",
          y = 0.95, yanchor = "top",
          bgcolor = "transparent"
        ),
        xaxis = list(title = "Departure delay"),
        yaxis = list(title = "Arrival delay")
      ) %>%
      toWebGL()
  }) %>%
    bindCache(flight_dat(), input$scatter_summarize)


  output$arr_delay <- renderPlotly({
    x <- flight_dat()$arr_delay
    req(length(x) > 0)

    x_mean <- mean(x, na.rm = TRUE)
    end <- quantile(x, probs = 0.99, na.rm = TRUE)
    plot_ly(x = x, hovertemplate = "%{y} flights were %{x} min late <extra></extra>", color = I(PRIMARY)) %>%
      config(displayModeBar = FALSE) %>%
      rangeslider(start = min(x, na.rm = TRUE), end = as.numeric(end)) %>%
      # TODO: add annotations for each line?
      layout(
        shapes = list(
          type = "line",
          x0 = x_mean, x1 = x_mean,
          y0 = 0, y1 = 1,
          yref = "paper",
          line = list(color = "orange", dash = "solid")
        )
      )
  }) %>%
    bindCache(flight_dat())

  output$arr_delay_series <- renderPlotly({
    d <- flight_dat()
    req(nrow(d) > 0)

    d %>%
      group_by(date) %>%
      summarise(arr_delay_mean = mean(arr_delay, na.rm = TRUE)) %>%
      plot_ly(
        x = ~date, y = ~arr_delay_mean,
        hovertemplate = "%{y:.1f}<extra></extra>",
        color = I(PRIMARY)
      ) %>%
      add_lines() %>%
      layout(
        hovermode = "x",
        xaxis = list(title = "", tickformat = "%b %e"),
        yaxis = list(title = "Average delay")
      )
  }) %>%
    bindCache(flight_dat())

}

shinyApp(ui, server)

library(shiny)
library(ggplot2)
library(bslib)
library(rlang)
library(curl)

# enlarged auto fonts
if (is_installed("thematic")) {
  thematic::thematic_shiny(
    font = thematic::font_spec("auto", scale = 2, update = TRUE)
  )
}

# Source in ggplot2 examples
source("global.R")

theme <- bs_global_get() %||% bs_theme()
if ("3" %in% theme_version(theme)) {
  warning("This example app requires Bootstrap 4 or higher", call. = FALSE)
}

rounded <- isTRUE(
  as.logical(bs_get_variables(theme %||% bslib::bs_theme(), "enable-rounded"))
)
pill <- function(...) {
  shiny::tabPanel(..., class = "p-3 border", class = if (rounded) "rounded")
}
tab <- function(...) {
  shiny::tabPanel(
    ...,
    class = "p-3 border border-top-0",
    class = if (rounded) "rounded-bottom"
  )
}
gradient <- function(theme_color = "primary") {
  bg_color <- paste0("bg-", theme_color)
  bgg_color <- if ("4" %in% theme_version(theme)) {
    paste0("bg-gradient-", theme_color)
  } else {
    paste(bg_color, "bg-gradient")
  }
  bg_div <- function(color_class, ...) {
    display_classes <- paste(
      paste0(".", strsplit(color_class, "\\s+")[[1]]),
      collapse = " "
    )
    div(
      class = "p-3",
      class = color_class,
      display_classes,
      ...
    )
  }
  fluidRow(
    column(6, bg_div(bg_color)),
    column(6, bg_div(bgg_color))
  )
}

theme_colors <- c(
  "primary",
  "secondary",
  "default",
  "success",
  "info",
  "warning",
  "danger",
  "dark"
)
gradients <- lapply(theme_colors, gradient)

progressBar <- div(
  class = "progress",
  div(
    class = "progress-bar w-25",
    role = "progressbar",
    "aria-valuenow" = "25",
    "aria-valuemin" = "0",
    "aria-valuemax" = "100"
  )
)

# This is here for shinycoreci to take advantage of (so we don't need to update a bunch of screenshots)
IS_LEGACY <- as.logical(Sys.getenv("BSLIB_LEGACY_THEMER_APP", FALSE))
is_legacy_version <- theme_version(theme) %in% c("3", "4")
if (isTRUE(IS_LEGACY || is_legacy_version)) {
  dashboardTab <- NULL
} else {
  dashboardTab <- nav_panel("Dashboard", tipsUI("tips"))
  theme_set(theme_minimal())
}

layout_buttons <- if (is_legacy_version) tags$div else layout_columns

bs_table <- function(x, class = NULL, ...) {
  class <- paste(c("table", class), collapse = " ")
  class <- sprintf('class="%s"', class)
  HTML(knitr::kable(x, format = "html", table.attr = class))
}

shinyApp(
  page_navbar(
    theme = theme,
    title = "Theme demo",
    collapsible = TRUE,
    inverse = !IS_LEGACY,
    id = "navbar",
    fillable = "Dashboard",
    dashboardTab,
    nav_panel(
      "Inputs",
      tabsetPanel(
        type = "pills",
        id = "inputs",
        pill(
          "inputPanel()",
          inputPanel(
            sliderInput(
              "slider",
              "sliderInput()",
              min = 0,
              max = 100,
              value = c(30, 70),
              step = 20
            ),
            selectInput("selectize", "selectizeInput()", choices = state.abb),
            selectInput(
              "selectizeMulti",
              "selectizeInput(multiple=T)",
              choices = state.abb,
              multiple = TRUE
            ),
            dateInput("date", "dateInput()", value = "2020-12-24"),
            dateRangeInput(
              "dateRange",
              "dateRangeInput()",
              start = "2020-12-24",
              end = "2020-12-31"
            )
          ),
          br(),
          textOutput("inputPanelOutputHeader"),
          verbatimTextOutput("inputPanelOutput"),
          br(),
          tags$p(
            "Here are some",
            tags$code("actionButton()"),
            "s demonstrating different theme (i.e., accent) colors"
          ),
          layout_buttons(
            # Regular <div> if IS_LEGACY, otherwise layout_columns
            actionButton(
              "primary",
              "Primary",
              icon("product-hunt"),
              class = "btn-primary m-2"
            ),
            actionButton("secondary", "Secondary (default)", class = "m-2"),
            actionButton(
              "success",
              "Success",
              icon("check"),
              class = "btn-success m-2"
            ),
            actionButton("info", "Info", icon("info"), class = "btn-info m-2"),
            actionButton(
              "warning",
              "warning",
              icon("exclamation"),
              class = "btn-warning m-2"
            ),
            actionButton(
              "danger",
              "Danger",
              icon(
                if (packageVersion("fontawesome") >= "0.3") {
                  "triangle-exclamation"
                } else {
                  "exclamation-triangle"
                }
              ),
              class = "btn-danger m-2"
            ),
            actionButton("dark", "Dark", icon("moon"), class = "btn-dark m-2"),
            actionButton("light", "Light", icon("sun"), class = "btn-light m-2")
          )
        ),
        pill(
          "wellPanel()",
          wellPanel(
            fluidRow(
              column(
                6,
                selectInput(
                  "select",
                  "selectInput()",
                  choices = state.abb,
                  selectize = FALSE
                ),
                selectInput(
                  "selectMulti",
                  "selectInput(multiple=T)",
                  choices = state.abb,
                  multiple = TRUE,
                  selectize = FALSE
                ),
                textInput(
                  "text",
                  "textInput()",
                  placeholder = "Enter some text"
                ),
                numericInput("numeric", "numericInput()", value = 0)
              ),
              column(
                6,
                passwordInput("password", "passwordInput()", "secret"),
                textAreaInput(
                  "textArea",
                  "textAreaInput()",
                  placeholder = "A text area"
                ),
                checkboxInput("check", "checkboxInput()", value = TRUE),
                checkboxGroupInput(
                  "checkGroup",
                  "checkboxGroupInput()",
                  choices = c("A", "B")
                ),
                radioButtons(
                  "radioButtons",
                  "radioButtons()",
                  choices = c("A", "B")
                )
              )
            )
          ),
          br(),
          textOutput("wellPanelOutputHeader"),
          br(),
          verbatimTextOutput("wellPanelOutput")
        )
      )
    ),
    nav_panel(
      "Plots",
      uiOutput("thematic_needed"),
      plotOutput("plot"),
      selectizeInput(
        "plot_example",
        "Choose an example",
        selected = "GeomSmooth",
        choices = list(
          ggplot2 = names(ggplot2_examples),
          lattice = names(lattice_examples),
          base = names(base_examples)
        )
      )
    ),
    nav_panel(
      "Tables",
      tabsetPanel(
        id = "tables",
        tab(
          "DataTables",
          DT::dataTableOutput("DT")
        ),
        tab(
          "Bootstrap",
          h3("Plain tables", class = "mt-2 mb-1"),
          bs_table(mtcars[1:5, 1:5]),
          h3("Striped tables", class = "mt-4 mb-1"),
          bs_table(mtcars[1:5, 1:5], class = "table-striped"),
          h3("Striped tables with hover", class = "mt-4 mb-1"),
          bs_table(mtcars[1:5, 1:5], class = "table-striped table-hover")
        )
      )
    ),
    nav_panel(
      "Notifications",
      tabsetPanel(
        id = "otherNav",
        tab(
          "Messages",
          br(),
          actionButton("showProgress", "Progress", style = "margin: 1rem"),
          actionButton("showModal", "Modals", style = "margin: 1rem"),
          lapply(c("default", "message", "warning", "error"), function(x) {
            X <- tools::toTitleCase(x)
            class <- switch(
              x,
              message = "btn-info",
              warning = "btn-warning",
              error = "btn-danger"
            )
            actionButton(
              paste0("show", X),
              paste(X, "notification"),
              class = class,
              style = "margin: 1rem"
            )
          })
        ),
        tab(
          "Uploads & Downloads",
          br(),
          fileInput("file", "fileInput()"),
          downloadButton(
            "downloadButton",
            "downloadButton()",
            style = "margin: 1rem"
          ),
          downloadLink("downloadLink", "downloadLink()", style = "margin: 1rem")
        )
      )
    ),
    nav_panel(
      "Fonts",
      h1("Heading font:", class = "text-primary"),
      hr(class = "bg-primary", style = "height: 5px"),
      h1("Heading 1"),
      h2("Heading 2"),
      h3("Heading 3"),
      h4("Heading 4"),
      h5("Heading 5"),
      h1("Base font:", class = "text-primary"),
      hr(class = "bg-primary", style = "height: 5px"),
      tags$p(
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
      ),
      h1("Code font:", class = "text-primary"),
      hr(class = "bg-primary", style = "height: 5px"),
      tags$pre(
        class = "shiny-text-output",
        style = "white-space: pre-wrap",
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
      )
    ),
    nav_panel(
      "Options",
      if ("4" %in% theme_version(theme)) {
        p(
          "Background color gradients are disabled by default.",
          "Enable them to see the difference here.",
          "If enabled, gradients automatically apply buttons and progress bars, ",
          "but you may also add to a .bg-gradient-* modified class to arbitrary elements."
        )
      } else {
        markdown(
          "`{bslib}` 'extends' Bootstrap's [`background-color` utility classes](https://getbootstrap.com/docs/5.0/utilities/background/) to also sets the foreground `color` to a sensible contrasting color (i.e., no additional `text-*` classes are needed)."
        )
      },
      !!!gradients,
      br(),
      tags$p(
        "With the default settings, enabling of box shadows adds a very subtle and
              barely noticable inner box-shadow to most input widgets. The difference is
              a little more obvious in a progress bar:"
      ),
      progressBar,
      br(),
      tags$p(
        "Rounded corners are enabled by default and apply to numerous components (e.g., ",
        code("tabsetPanel()"),
        ",",
        code("wellPanel()"),
        ", and ",
        code("actionButton()"),
        "):",
        actionButton("showProgress2", "Show progress")
      ),
      br(),
      tags$p(
        "Smooth transitions are enabled by default, but if you disable them, progress updating",
        "in ",
        code("fileInput()"),
        "and ",
        code("Progress"),
        " will appear more staggered",
        "(e.g., click button above)."
      )
    )
  ),
  function(input, output, session) {
    output$DT <- DT::renderDataTable({
      cars <- mtcars
      for (var in c("cyl", "vs", "am", "gear", "carb")) {
        cars[[var]] <- factor(cars[[var]])
      }

      DT::datatable(
        cars,
        style = "auto",
        caption = "A generic table made with the DT package and the DataTables library.",
        filter = "top"
      )
    })

    output$inputPanelOutputHeader <- renderText({
      "Below are the values bound to each input widget above"
    })

    output$inputPanelOutput <- renderPrint({
      str(
        list(
          sliderInput = input$slider,
          selectizeInput = input$selectize,
          selectizeMultiInput = input$selectizeMulti,
          dateInput = input$date,
          dateRangeInput = input$dateRange
        )
      )
    })

    output$wellPanelOutputHeader <- renderText({
      "Below are the values bound to each input widget above"
    })

    output$wellPanelOutput <- renderPrint({
      str(
        list(
          selectInput = input$select,
          selectMultiInput = input$selectMulti,
          textInput = input$text,
          numericInput = input$numeric,
          passwordInput = input$password,
          textAreaInput = input$textArea,
          checkInput = input$check,
          checkGroupInput = input$checkGroup,
          radioButtonsInput = input$radioButtons
        )
      )
    })

    observeEvent(input$showModal, {
      showModal(
        modalDialog(
          title = "Somewhat important message",
          "This is a somewhat important message.",
          easyClose = TRUE,
          footer = modalButton("Close")
        )
      )
    })

    fake_progress <- function(style = "notification") {
      withProgress(
        message = 'Calculation in progress',
        detail = 'This may take a while...',
        value = 0,
        style = style,
        {
          for (i in 1:15) {
            incProgress(
              amount = 1 / 15,
              detail = paste("Step", i, "of", 15)
            )
            Sys.sleep(0.25)
          }
        }
      )
    }

    observeEvent(input$showProgress, {
      fake_progress()
      # TODO: old progress styling could be improved
      #fake_progress("old")
    })

    observeEvent(input$showProgress2, {
      p <- Progress$new()
      p$set(
        message = 'Calculation in progress',
        detail = 'This may take a while...',
        value = 0.5
      )
    })

    lapply(c("default", "message", "warning", "error"), function(x) {
      X <- tools::toTitleCase(x)
      observeEvent(input[[paste0("show", X)]], {
        q <- parseQueryString(session$clientData$url_search)
        if ("notification-duration" %in% names(q)) {
          duration <- as.numeric(q[["notification-duration"]])
          if (duration <= 0) duration <- NULL
        } else {
          duration <- 5
        }
        showNotification(
          paste(X, "notification styling"),
          type = x,
          duration = duration
        )
      })
    })

    output$plot <- renderPlot({
      ggplot2_examples[[input$plot_example]] %||%
        eval(lattice_examples[[input$plot_example]]) %||%
        eval(base_examples[[input$plot_example]])
    })

    output$thematic_needed <- renderUI({
      if (bslib:::is_installed("thematic")) {
        return(NULL)
      }

      htmltools::HTML(
        "<span class=\"bg-warning\">&nbsp;!! Install the <a href='https://rstudio.github.io/thematic/'><code>thematic</code></a> package to enable auto-theming of static R plots !!&nbsp;</span>"
      )
    })

    if (!IS_LEGACY) tipsServer("tips")
  }
)

#' Sidebar layouts
#'
#' @description
#' Sidebar layouts place UI elements, like input controls or additional context,
#' next to the main content area which often holds output elements like plots or
#' tables.
#'
#' There are several page, navigation, and layout functions that allow you to
#' create a sidebar layout. In each case, you can create a collapsing sidebar
#' layout by providing a `sidebar()` object to the `sidebar` argument the
#' following functions.
#'
#' * [page_sidebar()] creates a "page-level" sidebar.
#' * [page_navbar()] creates a multi-panel app with an (optional, page-level)
#'   sidebar that is shown on every panel.
#' * `layout_sidebar()` creates a "floating" sidebar layout component which can
#'   be used inside any [page()] and/or [card()] context.
#' * [navset_card_tab()] and [navset_card_pill()] create multi-tab cards with a
#'   shared sidebar that is accessible from every panel.
#'
#' See [the Sidebars article](https://rstudio.github.io/bslib/articles/sidebars.html)
#' on the bslib website to learn more.
#'
#' @references Sidebar layouts are featured in a number of pages on the bslib
#'   website:
#'
#'   * [Sidebars](https://rstudio.github.io/bslib/articles/sidebars.html)
#'   * [Cards: Sidebars](https://rstudio.github.io/bslib/articles/cards/index.html#sidebars)
#'   * [Getting Started: Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'
#' @param ... Unnamed arguments can be any valid child of an [htmltools
#'   tag][htmltools::tags] and named arguments become HTML attributes on
#'   returned UI element. In the case of `layout_sidebar()`, these arguments are
#'   passed to the main content tag (not the sidebar+main content container).
#' @param width A valid [CSS unit][htmltools::validateCssUnit] used for the
#'   width of the sidebar.
#' @param position Where the sidebar should appear relative to the main content.
#' @param open The initial state of the sidebar, choosing from the following
#'   options:
#'
#'   * `"desktop"`: The sidebar starts open on desktop screen, closed on mobile.
#'     This is default sidebar behavior.
#'   * `"open"` or `TRUE`: The sidebar starts open.
#'   * `"closed"` or `FALSE`: The sidebar starts closed.
#'   * `"always"` or `NA`: The sidebar is always open and cannot be closed.
#'
#'   Alternatively, you can use a list with `desktop` or `mobile` items to set
#'   the initial sidebar state independently for `desktop` and `mobile` screen
#'   sizes. In this case, `desktop` or `mobile` can use any of the above options
#'   except `"desktop"`, which is equivalent to
#'   `list(desktop = "open", mobile = "closed")`. You can also choose to place
#'   an always open sidebar above the main content on mobile devices by setting
#'   `mobile = "always-above"`.
#'
#'   In `sidebar_toggle()`, `open` indicates the desired state of the sidebar,
#'   where the default of `open = NULL` will cause the sidebar to be toggled
#'   open if closed or vice versa. Note that `sidebar_toggle()` can only open or
#'   close the sidebar, so it does not support the `"desktop"` and `"always"`
#'   options.
#' @param id A character string. Required if wanting to re-actively read (or
#'   update) the `collapsible` state in a Shiny app.
#' @param title A character title to be used as the sidebar title, which will be
#'   wrapped in a `<header>` element with class `sidebar-title`. You can also
#'   provide a custom [htmltools::tag()] for the title element, in which case
#'   you'll likely want to give this element `class = "sidebar-title"`.
#' @param bg,fg A background or foreground color. If only one of either is
#'   provided, an accessible contrasting color is provided for the opposite
#'   color, e.g. setting `bg` chooses an appropriate `fg` color.
#' @param class CSS classes for the sidebar container element, in addition to
#'   the fixed `.sidebar` class.
#' @param max_height_mobile A [CSS length unit][htmltools::validateCssUnit()]
#'   defining the maximum height of the horizontal sidebar when viewed on mobile
#'   devices. Only applies to always-open sidebars that use `open = "always"`,
#'   where by default the sidebar container is placed below the main content
#'   container on mobile devices.
#' @param gap A [CSS length unit][htmltools::validateCssUnit()] defining the
#'   vertical `gap` (i.e., spacing) between adjacent elements provided to `...`.
#' @param padding Padding within the sidebar itself. This can be a numeric
#'   vector (which will be interpreted as pixels) or a character vector with
#'   valid CSS lengths. `padding` may be one to four values. If one, then
#'   that value will be used for all four sides. If two, then the first value
#'   will be used for the top and bottom, while the second value will be used
#'   for left and right. If three, then the first will be used for top, the
#'   second will be left and right, and the third will be bottom. If four, then
#'   the values will be interpreted as top, right, bottom, and left
#'   respectively.
#' @param fillable Whether or not the sidebar should be considered a fillable
#'   container. When `TRUE`, the sidebar and its content can use `fill` to
#'   consume available vertical space.
#'
#' @export
sidebar <- function(
  ...,
  width = 250,
  position = c("left", "right"),
  open = NULL,
  id = NULL,
  title = NULL,
  bg = NULL,
  fg = NULL,
  class = NULL,
  max_height_mobile = NULL,
  gap = NULL,
  padding = NULL,
  fillable = FALSE
) {
  position <- rlang::arg_match(position)
  gap <- validateCssUnit(gap)
  padding <- validateCssPadding(padding)
  width <- validateCssUnit(width)
  max_height_mobile <- validateCssUnit(max_height_mobile)

  if (!is.null(open)) {
    open <- as_sidebar_open_on(open)
  }

  if (!is.null(id)) {
    if (length(id) != 1 || is.na(id) || !nzchar(id)) {
      rlang::abort(
        "`id` must be a non-empty, length-1 character string or `NULL`."
      )
    }

    # when user provides id, make the sidebar a shiny input to report its state
    class <- c("bslib-sidebar-input", class)
  }

  if (is.null(fg) && !is.null(bg)) {
    fg <- get_color_contrast(bg)
  }
  if (is.null(bg) && !is.null(fg)) {
    bg <- get_color_contrast(fg)
  }

  if (rlang::is_bare_character(title) || rlang::is_bare_numeric(title)) {
    title <- tags$header(title, class = "sidebar-title")
  }

  dots <- separate_arguments(...)

  res <- list2(
    id = id,
    title = title,
    class = class,
    gap = gap,
    padding = padding,
    open = open,
    position = position,
    width = width,
    max_height_mobile = max_height_mobile,
    color = list(bg = bg, fg = fg),
    fillable = fillable,
    attributes = dots$attribs,
    children = dots$children
  )

  class(res) <- c("bslib_sidebar", "sidebar", class(res))
  res
}

#' Render a sidebar as HTML tags
#'
#' Renders the sidebar element and collapse toggle elements for a [sidebar()] in
#' a [layout_sidebar()] context.
#'
#' @param x A [sidebar()] object.
#' @param ... Additional arguments passed to [htmltools::as.tags()].
#'
#' @importFrom htmltools as.tags
#' @keywords internal
#' @export
as.tags.bslib_sidebar <- function(x, ...) {
  if (is.null(open)) {
    open <- sidebar_open_on()
  }

  if (x$open$mobile == "always-above") {
    x$open$mobile <- "always"
  }

  is_always_open <- all(vapply(x$open, identical, logical(1), "always"))

  if (!is_always_open && is.null(x$id)) {
    # always provide an id for accessibility reasons when collapsible
    x$id <- paste0("bslib-sidebar-", p_randomInt(1000, 10000))
  }

  if (!is.null(x$max_height_mobile) && x$open$mobile != "always") {
    rlang::warn(
      'The `max_height_mobile` argument only applies to the sidebar when `open = "always"` on mobile.'
    )
    x$max_height_mobile <- NULL
  }

  is_starts_closed <- all(vapply(x$open, identical, logical(1), "closed"))
  is_starts_open <- all(vapply(x$open, identical, logical(1), "open"))

  hidden_initially <-
    if (is_always_open) {
      FALSE
    } else if (is_starts_open) {
      FALSE
    } else {
      TRUE
    }

  collapse_tag <-
    tags$button(
      class = "collapse-toggle",
      type = "button",
      title = "Toggle sidebar",
      "aria-expanded" = tolower(
        is_starts_open || identical(x$open$desktop, "open")
      ),
      "aria-controls" = x$id,
      collapse_icon()
    )

  sidebar_tag <- tags$aside(
    id = x$id,
    class = c("sidebar", x$class),
    hidden = if (hidden_initially) NA,
    if (isTRUE(x$fillable)) as_fillable_container(),
    tags$div(
      class = "sidebar-content bslib-gap-spacing",
      if (isTRUE(x$fillable)) as_fill_carrier(),
      x$title,
      style = css(
        gap = x$gap,
        padding = x$padding
      ),
      !!!x$attributes,
      !!!x$children
    )
  )

  htmltools::tagList(sidebar_tag, collapse_tag)
}

as_sidebar_open_on <- function(open) {
  if (is.null(open)) {
    return(NULL)
  }

  if (rlang::is_list(open)) {
    unknown <- setdiff(names(open), c("desktop", "mobile"))
    if (length(unknown)) {
      rlang::warn(
        paste(
          "Ignoring unknown items in `open` options list:",
          paste0('"', unknown, '"', collapse = ", ")
        ),
      )
    }
    return(sidebar_open_on(open[["desktop"]], open[["mobile"]]))
  }

  if (length(open) == 1) {
    open <- sidebar_open_as_string(open, extra = "desktop", rlang::caller_env())
    if (identical(open, "desktop")) {
      return(sidebar_open_on("open", "closed"))
    }
    return(sidebar_open_on(open, open))
  }

  msg <- sprintf(
    "`open` must be a character string, a boolean, or a list with `desktop` and `mobile` items, not `%s`.",
    rlang::expr_text(open)
  )

  rlang::abort(
    c(
      msg,
      "i" = 'Character options include "open", "closed", "always" or "desktop".'
    ),
    call = rlang::caller_call()
  )
}

#' @param desktop,mobile The initial state of the sidebar on desktop or mobile
#'   screen sizes. Can be one of `"open"` (or `TRUE`), `"closed"` (or `FALSE`),
#'   or `"always"` (or `NA`). `mobile` also accepts `"always-above"`.
#' @noRd
sidebar_open_on <- function(
  desktop = "open",
  mobile = "closed"
) {
  desktop <- desktop %||% "open"
  mobile <- mobile %||% "closed"

  desktop <- sidebar_open_as_string(desktop)
  mobile <- sidebar_open_as_string(mobile, extra = "always-above")

  structure(
    list(desktop = desktop, mobile = mobile),
    class = "bslib_sidebar_open_options"
  )
}

sidebar_open_as_string <- function(
  open,
  extra = NULL,
  error_call = rlang::caller_env()
) {
  error_arg <- deparse(substitute(open))

  if (is.null(open)) {
    return(NULL)
  }
  if (identical(open, NA)) {
    return("always")
  }
  if (isTRUE(open)) {
    return("open")
  }
  if (isFALSE(open)) {
    return("closed")
  }

  rlang::arg_match(
    open,
    values = c("open", "closed", "always", extra),
    error_arg = error_arg,
    error_call = error_call
  )
}

#' @rdname sidebar
#' @param sidebar A [sidebar()] object.
#' @param fillable Whether or not the `main` content area should be considered a
#'   fillable (i.e., flexbox) container.
#' @param fill Whether or not to allow the layout container to grow/shrink to fit a
#'   fillable container with an opinionated height (e.g., `page_fillable()`).
#' @param border Whether or not to add a border.
#' @param border_radius Whether or not to add a border radius.
#' @param border_color The border color that is applied to the entire layout (if
#'   `border = TRUE`) and the color of the border between the sidebar and the
#'   main content area.
#' @inheritParams card
#' @inheritParams page_fillable
#'
#' @export
layout_sidebar <- function(
  ...,
  sidebar = NULL,
  fillable = TRUE,
  fill = TRUE,
  bg = NULL,
  fg = NULL,
  border = NULL,
  border_radius = NULL,
  border_color = NULL,
  padding = NULL,
  gap = NULL,
  height = NULL
) {
  if (!inherits(sidebar, "sidebar")) {
    sidebar <- sidebar(sidebar)
  }

  if (is.null(sidebar$open)) {
    sidebar$open <- sidebar_open_on()
  }

  if (!(is.null(border) || isTRUE(border) || isFALSE(border))) {
    abort("`border` must be `NULL`, `TRUE`, or `FALSE`")
  }
  if (
    !(is.null(border_radius) || isTRUE(border_radius) || isFALSE(border_radius))
  ) {
    abort("`border_radius` must be `NULL`, `TRUE`, or `FALSE`")
  }

  # main content area colors, if not provided ----
  if (is.null(fg) && !is.null(bg)) {
    fg <- get_color_contrast(bg)
  }
  if (is.null(bg) && !is.null(fg)) {
    bg <- get_color_contrast(fg)
  }

  main <- div(
    class = "main",
    class = if (fillable) "bslib-gap-spacing",
    style = css(
      padding = validateCssPadding(padding),
      gap = validateCssUnit(gap)
    ),
    ...
  )

  main <- bindFillRole(main, container = fillable)
  contents <- if (sidebar$open$mobile == "always-above") {
    list(as.tags(sidebar), main)
  } else {
    list(main, as.tags(sidebar))
  }

  right <- identical(sidebar$position, "right")

  res <- div(
    class = "bslib-sidebar-layout bslib-mb-spacing",
    class = if (right) "sidebar-right",
    class = if (identical(sidebar$open$desktop, "closed")) "sidebar-collapsed",
    `data-bslib-sidebar-init` = TRUE,
    `data-open-desktop` = sidebar$open$desktop,
    `data-open-mobile` = sidebar$open$mobile,
    `data-collapsible-mobile` = tolower(
      !sidebar$open$mobile %in% c("always", "always-above")
    ),
    `data-collapsible-desktop` = tolower(
      !identical(sidebar$open$desktop, "always")
    ),
    `data-bslib-sidebar-border` = if (!is.null(border)) tolower(border),
    `data-bslib-sidebar-border-radius` = if (!is.null(border_radius)) {
      tolower(border_radius)
    },
    style = css(
      "--_sidebar-width" = sidebar$width,
      "--_sidebar-bg" = sidebar$color$bg,
      "--_sidebar-fg" = sidebar$color$fg,
      "--_main-fg" = fg,
      "--_main-bg" = bg,
      "--bs-card-border-color" = border_color,
      height = validateCssUnit(height),
      "--_mobile-max-height" = sidebar$max_height_mobile
    ),
    !!!contents,
    sidebar_init_js(),
    component_dependencies()
  )

  res <- bindFillRole(res, item = fill)

  res <- as.card_item(res)

  as_fragment(
    tag_require(res, version = 5, caller = "layout_sidebar()")
  )
}

#' @describeIn sidebar Toggle a `sidebar()` state during an active Shiny user
#'   session. To use this function, the `sidebar()` you want to open or close
#'   must have an `id` value.
#' @param session A Shiny session object (the default should almost always be
#'   used).
#' @export
toggle_sidebar <- function(id, open = NULL, session = get_current_session()) {
  method <- sidebar_open_as_string(open %||% "toggle", "toggle")

  if (identical(method, "always")) {
    abort('`open = "always"` is not supported by `sidebar_toggle()`.')
  }

  force(id)
  callback <- function() {
    session$sendInputMessage(id, list(method = method))
  }
  session$onFlush(callback, once = TRUE)
}

collapse_icon <- function() {
  if (!is_installed("bsicons")) {
    icon <- "<svg xmlns=\"http://www.w3.org/2000/svg\" viewBox=\"0 0 16 16\" class=\"bi bi-chevron-left collapse-icon\" style=\"fill:currentColor;\" aria-hidden=\"true\" role=\"img\" ><path fill-rule=\"evenodd\" d=\"M11.354 1.646a.5.5 0 0 1 0 .708L5.707 8l5.647 5.646a.5.5 0 0 1-.708.708l-6-6a.5.5 0 0 1 0-.708l6-6a.5.5 0 0 1 .708 0z\"></path></svg>"
    return(HTML(icon))
  }
  bsicons::bs_icon("chevron-left", class = "collapse-icon", size = NULL)
}

sidebar_init_js <- function() {
  # Note: if we want to avoid inline `<script>` tags in the future for
  # initialization code, we might be able to do so by turning the sidebar layout
  # container into a web component
  tags$script(
    `data-bslib-sidebar-init` = NA,
    HTML("bslib.Sidebar.initCollapsibleAll()")
  )
}

test_that("sidebar(open =) ensures backwards compatibility of defaults", {
  expect_equal(
    sidebar(open = TRUE, id = "open"),
    sidebar(open = "open", id = "open")
  )

  expect_equal(
    sidebar(open = FALSE, id = "closed"),
    sidebar(open = "closed", id = "closed")
  )

  expect_equal(
    sidebar(open = NA, id = "always"),
    sidebar(open = "always", id = "always")
  )
})

test_that("sidebar() - assigns input binding class if `id` is provided", {
  my_sidebar <- as.tags(sidebar(id = "my-id", open = sidebar_open_on()))

  expect_match(
    htmltools::tagGetAttribute(my_sidebar[[1]], "class"),
    "bslib-sidebar-input"
  )

  expect_equal(
    htmltools::tagGetAttribute(my_sidebar[[1]], "id"),
    "my-id"
  )
})

test_that("sidebar() - assigns a random `id` if collapsible and `id` not provided", {
  sidebar_open <- as.tags(sidebar(open = "open"))

  expect_match(
    htmltools::tagGetAttribute(sidebar_open[[1]], "id"),
    "bslib-sidebar-\\d+"
  )

  expect_equal(
    htmltools::tagGetAttribute(sidebar_open[[2]], "aria-controls"),
    htmltools::tagGetAttribute(sidebar_open[[1]], "id")
  )

  sidebar_closed <- as.tags(sidebar(open = "closed"))
  expect_match(
    htmltools::tagGetAttribute(sidebar_closed[[1]], "id"),
    "bslib-sidebar-\\d+"
  )

  expect_equal(
    htmltools::tagGetAttribute(sidebar_closed[[2]], "aria-controls"),
    htmltools::tagGetAttribute(sidebar_closed[[1]], "id")
  )

  sidebar_desktop <- as.tags(sidebar(open = "desktop"))
  expect_match(
    htmltools::tagGetAttribute(sidebar_desktop[[1]], "id"),
    "bslib-sidebar-\\d+"
  )

  expect_equal(
    htmltools::tagGetAttribute(sidebar_desktop[[2]], "aria-controls"),
    htmltools::tagGetAttribute(sidebar_desktop[[1]], "id")
  )

  expect_null(
    htmltools::tagGetAttribute(sidebar(open = "always")[[1]], "id")
  )
})

test_that("sidebar() - sets `aria-expanded` correctly on collapse toggle", {
  get_sidebar_collapse_tag <- function(...) {
    as.tags(sidebar(...))[[2]]
  }

  expect_equal(
    htmltools::tagGetAttribute(
      get_sidebar_collapse_tag(open = "open"),
      "aria-expanded"
    ),
    "true"
  )

  expect_equal(
    htmltools::tagGetAttribute(
      get_sidebar_collapse_tag(open = "closed"),
      "aria-expanded"
    ),
    "false"
  )

  expect_equal(
    htmltools::tagGetAttribute(
      get_sidebar_collapse_tag(open = "desktop"),
      "aria-expanded"
    ),
    "true"
  )
})

test_that("sidebar() - throws an error for invalid `open` values", {
  expect_error(
    sidebar(open = "bad")
  )

  expect_error(
    sidebar(open = c("open", "closed"))
  )

  expect_error(
    sidebar(open = 1)
  )

  expect_error(
    sidebar(open = "OPEN")
  )

  expect_error(
    sidebar(open = character())
  )
})

test_that("sidebar() - warns if `max_height_mobile` used with `open != 'always'", {
  expect_warning(
    as.tags(sidebar(open = "open", max_height_mobile = "100px"))
  )

  expect_silent(
    as.tags(sidebar(open = "always", max_height_mobile = "100px"))
  )
})

test_that("layout_sidebar() - errors with unexpected border, border_radius input values", {
  expect_error(
    layout_sidebar(sidebar(), border = "1px solid blue")
  )
  expect_error(
    layout_sidebar(sidebar(), border = c(TRUE, FALSE))
  )
  expect_error(
    layout_sidebar(sidebar(), border_radius = c(TRUE, FALSE))
  )
  expect_error(
    layout_sidebar(sidebar(), border = c(TRUE, NA))
  )
  expect_error(
    layout_sidebar(sidebar(), border_radius = c(TRUE, NA))
  )
  expect_error(
    layout_sidebar(sidebar(), border = NA)
  )
  expect_error(
    layout_sidebar(sidebar(), border_radius = NA)
  )
})

withr::local_temp_libpaths()
install.packages("shiny")

rsconnect::deployApp(
  rprojroot::find_package_root_file("inst/examples/build-a-box"),
  appName = "build-a-box",
  account = "bslib",
  forceUpdate = TRUE
)

# Packages ---------------------------------------
library(shiny)
library(bslib)
library(htmltools)

pkgs_extra <- c("plotly", "colourpicker")
pkgs_yes <- vapply(pkgs_extra, rlang::is_installed, logical(1))
if (any(!pkgs_yes)) {
  rlang::abort(
    paste0(
      "The `build-a-box` app requires additional packages: ",
      paste(pkgs_extra[!pkgs_yes], collapse = ", ")
    )
  )
}

library(plotly)

# Settings ---------------------------------------
ENABLE_THEMER <- identical(Sys.getenv("ENABLE_THEMER"), "true")

# Functions ---------------------------------------
layout_value_box_options <- function(ui_opts) {
  layout_columns(
    div(h4("Content", class = "border-bottom"), ui_opts$title_value),
    div(h4("Theme", class = "border-bottom"), ui_opts$theme_opts),
    div(h4("Showcase", class = "border-bottom"), ui_opts$showcase)
  )
}

value_box_placeholder <- function(id) {
  value_box(
    id = id,
    class = "placeholder-glow",
    title = span(class = "placeholder col-7"),
    value = span(class = "placeholder col-4"),
    showcase = div(class = "placeholder bg-primary col-12", as_fill_item())
  )
}

# Theme ---------------------------------------
theme_build_a_box <- bs_add_rules(
  bs_theme(preset = "shiny"),
  sass::sass_file("www/build-a-box.scss")
)

# UI ---------------------------------------
ui <- page_fixed(
  title = "Build a Box | bslib",
  theme = theme_build_a_box,

  # Header ----
  tags$header(
    class = "mt-4 d-flex flex-row justify-content-between align-items-center",
    h2("Build a Box"),
    div(
      class = "d-flex flex-row align-items-center gap-3",
      popover(
        bsicons::bs_icon(
          "info-square-fill",
          title = "About Value Boxes",
          class = "icon-gradient"
        ),
        title = "About Value Boxes",
        HTML(commonmark::markdown_html(readLines("about-value-boxes.md")))
      ),
      actionLink(
        "show_code",
        tooltip(icon("code"), "Show code"),
        class = "nav-link text-blue",
        style = css(width = "1em")
      ),
      input_dark_mode(
        id = "color_mode",
        style = css("--text-1" = "var(--bs-blue)")
      )
    )
  ),

  # Main ----
  tags$main(
    # Value Box Previews ----
    div(
      id = "preview",
      class = "my-5",
      layout_columns(
        class = "value-box-previews",
        div(
          as_fill_carrier(),
          value_box_placeholder("one-value_box_placeholder"),
          ui_value_box_output("one")
        ),
        div(
          as_fill_carrier(),
          value_box_placeholder("two-value_box_placeholder"),
          ui_value_box_output("two")
        ),
        div(
          as_fill_carrier(),
          value_box_placeholder("three-value_box_placeholder"),
          ui_value_box_output("three")
        )
      )
    ),

    # Settings Panel ----
    navset_card_pill(
      id = "settings",
      title = span(
        "Value box settings",
        popover(
          bsicons::bs_icon("question-square-fill"),
          class = "ms-1 d-inline-block text-orange",
          title = "Getting started",
          shiny::markdown(
            "
            The Build-a-Box app includes three value boxes. You can customize
            all three at once from the **All** tab.

            Pick an overall theme, choose whether you'd like to **showcase** a plot or icon, and decide which **showcase layout** works best for your data.

            Then, click directly on a value box or switch to the
            <a id=\"switch_to_one\" href=\"#\" class=\"action-button\">One</a>,
            <a id=\"switch_to_two\" href=\"#\" class=\"action-button\">Two</a>, or
            <a id=\"switch_to_three\" href=\"#\" class=\"action-button\">Three</a>
            tabs to customize its settings individually.
          "
          )
        )
      ),
      nav_panel(
        "All",
        value = "all",
        ui_global_controls("all")
      ),
      nav_panel(
        "One",
        value = "one-value_box",
        layout_value_box_options(
          ui_value_box_options("one")
        )
      ),
      nav_panel(
        "Two",
        value = "two-value_box",
        layout_value_box_options(
          ui_value_box_options("two")
        )
      ),
      nav_panel(
        "Three",
        value = "three-value_box",
        layout_value_box_options(
          ui_value_box_options("three")
        )
      )
    )
  ),

  # Footer ----
  tags$footer(
    class = "footer mt-auto py-3",
    layout_columns(
      class = "border-top pt-3 text-muted",
      div(
        class = "text-center text-sm-start",
        "Made with",
        a(href = "https://rstudio.github.io/bslib", "{bslib}"),
        "and",
        a(
          href = "https://shiny.rstudio.com",
          img(src = "shiny.png", width = "22px", alt = " "),
          "Shiny"
        )
      ),
      div(
        class = "text-center text-sm-end",
        HTML(
          'Proudly supported by <a href="https://posit.co/"><img src="https://www.rstudio.com/assets/img/posit-logo-fullcolor-TM.svg" class="img-fluid" alt="Posit" width="65"></a>'
        )
      )
    )
  ),
  # Extras ----
  tags$script(src = "build-a-box.js")
)

# Server ---------------------------------------
server <- function(input, output, session) {
  enable_themer <- reactive({
    query <- shiny::getQueryString()
    query_has_themer <- "themer" %in% names(query)

    if (!length(query) || !query_has_themer) {
      return(ENABLE_THEMER)
    }

    query$themer %in% c(1, "true", "") || ENABLE_THEMER
  })

  observeEvent(enable_themer(), {
    # TODO: This only runs on app startup right now
    req(enable_themer())

    insertUI(
      selector = "body",
      where = "beforeEnd",
      ui = tags$script("window.watchForThemer()")
    )
    bs_themer()
  })

  one <- module_value_box("one")
  two <- module_value_box("two")
  three <- module_value_box("three")

  module_global_controls("all", one, two, three)

  observeEvent(input$clicked_value_box, {
    nav_select("settings", input$clicked_value_box)
  })

  observeEvent(input$switch_to_one, nav_select("settings", "one-value_box"))
  observeEvent(input$switch_to_two, nav_select("settings", "two-value_box"))
  observeEvent(input$switch_to_three, nav_select("settings", "three-value_box"))

  observeEvent(input$settings, {
    session$sendCustomMessage("active-value-box", input$settings)
  })

  observeEvent(input$show_code, {
    layout_value_boxes <-
      paste0(
        "layout_columns(\n  ",
        rlang::expr_text(one$code()),
        ",\n  ",
        rlang::expr_text(two$code()),
        ",\n  ",
        rlang::expr_text(three$code()),
        "\n",
        ")"
      )

    code_modal(layout_value_boxes)
  })
}

# BUILD A BOX -----------------------------------------
shinyApp(ui, server)

#' Create a vertically collapsing accordion
#'
#' @description
#' An accordion can be used to organize UI elements and content in a limited
#' space. It comprises multiple, vertically stacked panels that expand or
#' collapse when clicked, providing a compact layout that works well for
#' grouping input elements in a [sidebar()] or for organizing detailed
#' context-specific information.
#'
#' @param ... Named arguments become attributes on the `<div class="accordion">`
#'   element. Unnamed arguments should be `accordion_panel()`s.
#' @param id If provided, you can use `input$id` in your server logic to
#'   determine which of the `accordion_panel()`s are currently active. The value
#'   will correspond to the `accordion_panel()`'s `value` argument.
#' @param open A character vector of `accordion_panel()` `value`s to open
#'   (i.e., show) by default. The default value of `NULL` will open the first
#'   `accordion_panel()`. Use a value of `TRUE` to open all (or `FALSE` to
#'   open none) of the items. It's only possible to open more than one panel
#'   when `multiple=TRUE`.
#' @param multiple Whether multiple `accordion_panel()` can be `open` at once.
#' @param class Additional CSS classes to include on the accordion div.
#' @param width,height Any valid CSS unit; for example, height="100%".
#'
#' @export
#' @family Components
#'
#' @references bslib's accordion component is derived from the [Bootstrap
#'   Accordion
#'   component](https://getbootstrap.com/docs/5.3/components/accordion/).
#'   Accordions are also featured on the bslib website:
#'
#'   * [Get Started: Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html#accordions)
#'   * [Sidebars: Accordions](https://rstudio.github.io/bslib/articles/dashboards/index.html#accordions)
#'
#' @seealso [accordion_panel_set()], [accordion_panel_open()] and
#'   [accordion_panel_close()] programmatically interact with the state of an
#'   accordion panel.
#' @seealso [accordion_panel_insert()], [accordion_panel_remove()] and
#'   [accordion_panel_update()] add or remove accordion panels from an
#'   accordion.
#'
#' @examplesIf rlang::is_interactive()
#' items <- lapply(LETTERS, function(x) {
#'   accordion_panel(paste("Section", x), paste("Some narrative for section", x))
#' })
#'
#' # First shown by default
#' accordion(!!!items)
#' # Nothing shown by default
#' accordion(!!!items, open = FALSE)
#' # Everything shown by default
#' accordion(!!!items, open = TRUE)
#'
#' # Show particular sections
#' accordion(!!!items, open = "Section B")
#' accordion(!!!items, open = c("Section A", "Section B"))
#'
#' # Provide an id to create a shiny input binding
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   accordion(!!!items, id = "acc")
#' )
#'
#' server <- function(input, output) {
#'   observe(print(input$acc))
#' }
#'
#' shinyApp(ui, server)
#'
accordion <- function(
  ...,
  id = NULL,
  open = NULL,
  multiple = TRUE,
  class = NULL,
  width = NULL,
  height = NULL
) {
  args <- rlang::list2(...)
  argnames <- rlang::names2(args)

  attrs <- args[nzchar(argnames)]
  children <- args[!nzchar(argnames)]

  if (!is.null(id) && isNamespaceLoaded("shiny")) {
    open <- shiny::restoreInput(id = id, default = open)
  }

  is_open <- vapply(
    children,
    function(x) {
      isTRUE(open) || isTRUE(tagGetAttribute(x, "data-value") %in% open)
    },
    logical(1)
  )

  if (!any(is_open) && !identical(open, FALSE)) {
    is_open[1] <- TRUE
  }

  if (!multiple && sum(is_open) > 1) {
    abort("Can't select more than one panel when `multiple = FALSE`")
  }

  # Since multiple=FALSE requires an id, we always include one,
  # but only create a binding when it's provided
  if (is.null(id)) {
    id <- paste0("bslib-accordion-", p_randomInt(1000, 10000))
  } else {
    class <- c("bslib-accordion-input", class)
  }

  children <- Map(
    children,
    is_open,
    f = function(x, open) {
      if (!multiple) {
        x <- tagAppendAttributes(
          x,
          "data-bs-parent" = paste0("#", id),
          .cssSelector = ".accordion-collapse"
        )
      }

      # accordion_panel() defaults to a collapsed state
      if (open) {
        tq <- tagQuery(x)
        tq$find(".accordion-collapse")$addClass("show")
        tq$find(".accordion-button")$removeClass("collapsed")$removeAttrs(
          "aria-expanded"
        )$addAttrs("aria-expanded" = "true")
        x <- tq$allTags()
      }

      x
    }
  )

  tag <- div(
    id = id,
    class = "accordion",
    class = if (!multiple) "autoclose", # just for ease of identifying autoclosing client-side
    class = class,
    style = css(
      width = validateCssUnit(width),
      height = validateCssUnit(height)
    ),
    !!!attrs,
    !!!children,
    component_dependencies()
  )

  tag <- tag_require(tag, version = 5, caller = "accordion()")

  as_fragment(tag)
}

#' @rdname accordion
#' @param title A title to appear in the `accordion_panel()`'s header.
#' @param value A character string that uniquely identifies this panel.
#' @param icon A [htmltools::tag] child (e.g., [bsicons::bs_icon()]) which is positioned just before the `title`.
#' @export
accordion_panel <- function(title, ..., value = title, icon = NULL) {
  id <- paste0("bslib-accordion-panel-", p_randomInt(1000, 10000))

  btn <- tags$button(
    class = "accordion-button collapsed",
    type = "button",
    "data-bs-toggle" = "collapse",
    "data-bs-target" = paste0("#", id),
    "aria-expanded" = "false",
    "aria-controls" = id,
    # Always include an .accordion-icon container to simplify accordion_panel_update() logic
    div(class = "accordion-icon", icon),
    div(class = "accordion-title", title)
  )

  if (!rlang::is_string(value)) {
    abort("`value` must be a character string")
  }

  div(
    class = "accordion-item",
    "data-value" = value,
    # We don't use an <h2> because it would get included in rmd/pkgdown/qmd TOC
    # TODO: can we provide a way to put more stuff in the header? Like maybe some right-aligned controls?
    div(class = "accordion-header", btn),
    div(
      id = id,
      class = "accordion-collapse collapse",
      div(class = "accordion-body", ...)
    )
  )
}

#' Dynamically update accordions
#'
#' @description
#' Dynamically update/modify [`accordion()`]s in a Shiny app. To be updated
#' programmatically, the [accordion()] must have an `id`. These functions
#' require an active Shiny session and only work with a running Shiny app.
#'
#' @param id an character string that matches an existing [accordion()]'s `id`.
#' @param values either a character string (used to identify particular
#'   [accordion_panel()]s by their `value`) or `TRUE` (i.e., all `values`).
#' @param session a shiny session object (the default should almost always be
#'   used).
#'
#' @seealso [accordion()] and [accordion_panel()] create the accordion
#'   component.
#'
#' @describeIn accordion_panel_set same as `accordion_panel_open()`, except it
#'   also closes any currently open panels.
#'
#' @export
accordion_panel_set <- function(id, values, session = get_current_session()) {
  send_panel_message(
    id,
    session,
    method = "set",
    values = if (isTRUE(values)) values else as.list(check_character(values))
  )
}

#' @describeIn accordion_panel_set open [accordion_panel()]s.
#' @export
accordion_panel_open <- function(id, values, session = get_current_session()) {
  send_panel_message(
    id,
    session,
    method = "open",
    values = if (isTRUE(values)) values else as.list(check_character(values))
  )
}

#' @describeIn accordion_panel_set close [accordion_panel()]s.
#' @export
accordion_panel_close <- function(id, values, session = get_current_session()) {
  send_panel_message(
    id,
    session,
    method = "close",
    values = if (isTRUE(values)) values else as.list(check_character(values))
  )
}

#' @param panel an [accordion_panel()].
#' @param target The `value` of an existing panel to insert next to. If
#'   removing: the `value` of the [accordion_panel()] to remove.
#' @param position Should `panel` be added before or after the target? When
#'   `target` is `NULL` (the default), `"after"` will append after the last
#'   panel and `"before"` will prepend before the first panel.
#'
#' @describeIn accordion_panel_set insert a new [accordion_panel()]
#' @export
accordion_panel_insert <- function(
  id,
  panel,
  target = NULL,
  position = c("after", "before"),
  session = get_current_session()
) {
  position <- match.arg(position)
  send_panel_message(
    id,
    session,
    method = "insert",
    panel = processDeps(panel, session),
    target = if (!is.null(target)) check_character(target, max_length = 1),
    position = position
  )
}

#' @describeIn accordion_panel_set remove [accordion_panel()]s.
#' @export
accordion_panel_remove <- function(
  id,
  target,
  session = get_current_session()
) {
  send_panel_message(
    id,
    session,
    method = "remove",
    target = as.list(check_character(target))
  )
}

#' @describeIn accordion_panel_set update a [accordion_panel()].
#' @param ... Elements that become the new content of the panel.
#' @inheritParams accordion_panel
#' @export
accordion_panel_update <- function(
  id,
  target,
  ...,
  title = NULL,
  value = NULL,
  icon = NULL,
  session = get_current_session()
) {
  rlang::check_dots_unnamed()
  body <- rlang::list2(...)

  send_panel_message(
    id,
    session,
    method = "update",
    target = check_character(target, max_length = 1),
    value = if (!is.null(value)) check_character(value, max_length = 1),
    body = if (length(body) == 0) NULL else processDeps(body, session),
    title = if (!is.null(title)) processDeps(title, session),
    icon = if (!is.null(icon)) processDeps(icon, session)
  )
}

# Send message before the next flush since things like remove/insert may
# remove/create input/output values. Also do this for set/open/close since,
# you might want to open a panel after inserting it.
send_panel_message <- function(id, session, ...) {
  message <- dropNulls(rlang::list2(...))
  callback <- function() session$sendInputMessage(id, message)
  session$onFlush(callback, once = TRUE)
}

check_character <- function(
  x,
  max_length = Inf,
  min_length = 1,
  call = rlang::caller_env()
) {
  x_name <- deparse(substitute(x))
  if (!is.character(x)) {
    abort(
      sprintf("`%s` must be a character vector.", x_name),
      call = call
    )
  }
  if (length(x) < min_length) {
    abort(
      sprintf("`%s` must have length %s or more.", x_name, min_length),
      call = call
    )
  }
  if (length(x) > max_length) {
    abort(
      sprintf("`%s` must have length %s or less.", x_name, max_length),
      call = call
    )
  }
  x
}

show_raw_html <- function(x) {
  cat(format(x))
}

expect_snapshot_html <- function(x, .envir = parent.frame()) {
  with_private_seed()
  x_str <- deparse1(substitute(x))
  code <- parse(text = sprintf("expect_snapshot(show_raw_html(%s))", x_str))
  eval(code, envir = .envir)
}

`%||%` <- function(x, y) if (is.null(x)) y else x

icon_list <- list(
  "bsicons" = setNames(
    paste0("bsicons::", bsicons:::icon_info$name),
    bsicons:::icon_info$name
  ),
  "fontawesome" = setNames(
    paste0("fontawesome::", fontawesome::fa_metadata()$icon_names),
    fontawesome::fa_metadata()$icon_names
  )
)

resolve_icon <- function(name) {
  icon <- strsplit(name, "::", fixed = TRUE)[[1]]
  icon_fn <- switch(
    icon[1],
    bsicons = {
      ns <- "bsicons"
      fn <- "bs_icon"
    },
    {
      ns <- "fontawesome"
      fn <- "fa_i"
    }
  )
  rlang::call2(.ns = ns, fn, icon[2])
}

ui_value_box_options <- function(id) {
  ns <- NS(id)

  conditional_panel <- function(condition, ...) {
    conditionalPanel(condition, ..., ns = ns)
  }

  init <- random_title_value()

  list(
    "title_value" = list(
      textInput(ns("title"), "Title", init$title),
      textInput(ns("value"), "Value", init$value),
      tagAppendAttributes(
        textAreaInput(ns("extra"), "Extra text (markdown allowed)", ""),
        class = "input-text-code"
      ),
      actionButton(ns("random_stat"), "Random stat")
    ),
    "theme_opts" = list(
      input_switch(ns("use_custom_colors"), "Use custom colors", FALSE),
      conditional_panel(
        "!input.use_custom_colors",
        ui_selextra(ns("theme"), "Theme")
      ),
      conditional_panel(
        "input.use_custom_colors",
        colourpicker::colourInput(ns("foreground"), "Foreground", value = "#000000"),
        colourpicker::colourInput(ns("background"), "Background", value = "#FFFFFF")
      ),
      input_switch(ns("full_screen"), "Allow full screen", value = FALSE),
      input_switch(ns("fill"), "Fill vertical space", value = TRUE),
      input_switch(ns("fixed_height"), "Fixed height", value = FALSE),
      conditional_panel(
        "input.fixed_height",
        sliderInput(
          inputId = ns("height"),
          label = "Height",
          min = 100,
          max = 500,
          value = 150,
          post = "px",
          step = 10
        )
      )
    ),
    showcase = list(
      input_switch(ns("showcase"), "Include showcase", value = TRUE),
      conditional_panel(
        "input.showcase",
        radioButtons(
          ns("showcase_layout"),
          "Showcase Layout",
          c("left center", "top right", "bottom"),
          inline = TRUE
        ),
        radioButtons(
          ns("showcase_item"),
          "Showcase Item",
          c("icon", "plot"),
          selected = "plot",
          inline = TRUE
        ),
        conditional_panel(
          "input.showcase_item == 'plot'",
          selectInput(ns("showcase_plot_type"), "Plot Type", c("line", "bar", "box")),
          radioButtons(ns("showcase_plot_color"), "Plot Color", c("auto", "black", "white"), inline = TRUE)
        ),
        conditional_panel(
          "input.showcase_item == 'icon'",
          ui_selextra(ns("showcase_icon"), "Icon")
        )
      )
    )
  )
}

ui_value_box_output <- function(id) {
  uiOutput(NS(id)("value_box"), class = "shiny-report-theme", fill = TRUE)
}

server_value_box <- function(input, output, session, ...) {
  ns <- session$ns

  theme <- module_selextra("theme", all_themes)
  showcase_icon <- module_selextra("showcase_icon", icon_list)

  random_plot <- reactive({
    req(isolate(input$showcase), input$showcase_plot_type, plot_color())
    random_plotly_plot(input$showcase_plot_type, plot_color())
  })

  plot_color <- reactiveVal("#FFFFFF")

  observe({
    if (input$showcase_plot_color != "auto") {
      plot_color(input$showcase_plot_color)
      return()
    }

    if (input$use_custom_colors) {
      plot_color(input$foreground)
      return()
    }

    if (!is.null(input$value_box_fg_color)) {
      fg <- input$value_box_fg_color
      fg <- htmltools::parseCssColors(fg)
      plot_color(fg)
    } else {
      plot_color("#808080")
    }
  })

  observe({
    req(input$showcase, input$showcase_item == "icon")
    showcase_icon$shuffle()
  })

  # ┌─ {bslib} ──────────────────────┐
  # │                                │
  # │           value_box()          │
  # │                                │
  # └────────────────────────────────┘

  value_box_args_impl <- reactive({
    req(input$title, input$value)

    theme <-
      if (input$use_custom_colors) {
        rlang::call2(
          "value_box_theme",
          bg = input$background,
          fg = input$foreground
        )
      } else {
        if (nzchar(theme$value())) theme$value()
      }

    extra <-
      if (nzchar(input$extra)) {
        rlang::call2(.ns = "shiny", "markdown", input$extra)
      }

    showcase <-
      if (input$showcase) {
        req(showcase_icon$value())

        switch(
          input$showcase_item,
          icon = resolve_icon(showcase_icon$value()),
          plot = "Your Plot"
        )
      }

    rlang::list2(
      title = input$title,
      value = input$value,
      if (!is.null(extra)) extra else rlang::missing_arg(),
      theme = theme,
      showcase = showcase,
      showcase_layout = input$showcase_layout,
      full_screen = input$full_screen,
      fill = input$fill,
      height = if (input$fixed_height) input$height
    )
  })

  value_box_args <- debounce(value_box_args_impl, 250)

  value_box_call <- reactive({
    req(input$title, input$value)

    rlang::call2("value_box", !!!value_box_args())
  })

  output$value_box <- renderUI({
    req(value_box_call())

    call <- value_box_call()

    is_showcase_plot <- isolate(input$showcase && input$showcase_item == "plot")

    if (is_showcase_plot) {
      call <- rlang::call_modify(call, showcase = random_plot())
    }

    rlang::eval_bare(call)
  })

  observeEvent(input$showcase_item, {
    updateCheckboxInput(session, "full_screen", value = input$showcase_item == "plot")
  })

  observeEvent(input$random_stat, {
    random <- random_title_value()
    updateTextInput(session, "title", value = random$title)
    updateTextInput(session, "value", value = random$value)
  })

  observeEvent(input$shuffle_showcase_icon, {
    new <- sample(bsicons:::icon_info$name, 1)
    updateSelectInput(session, "showcase_icon", selected = paste0("bsicons::", new))
  })

  list(
    code = value_box_call,
    theme = theme,
    showcase_icon = showcase_icon,
    random_stat = function() {
      random <- random_title_value()
      updateTextInput(session, "title", value = random$title)
      updateTextInput(session, "value", value = random$value)
    },
    set_showcase_layout = function(layout) {
      updateRadioButtons(session, "showcase_layout", selected = layout)
    },
    set_showcase_item = function(item) {
      updateRadioButtons(session, "showcase_item", selected = item)
    }
  )
}

module_value_box <- function(id) {
  moduleServer(id, server_value_box)
}

#' Navigation containers
#'
#' Render a collection of [nav_panel()] items into a container.
#'
#' @section Examples:
#'
#' ```{r child="man/fragments/ex-navset_tab.Rmd"}
#' ```
#'
#' @param ... a collection of [nav_panel()] items.
#' @param id a character string used for dynamically updating the container (see
#'    [nav_select()]).
#' @param selected a character string matching the `value` of a particular
#'   [nav_panel()] item to selected by default.
#' @param header UI element(s) ([htmltools::tags]) to display _above_ the nav
#'   content. For `card`-based navsets, these elements are implicitly wrapped in
#'   a `card_body()`. To control things like `padding`, `fill`, etc., wrap the
#'   elements in an explicit [card_body()].
#' @param footer UI element(s) ([htmltools::tags]) to display _below_ the nav
#'   content. For `card`-based navsets, these elements are implicitly wrapped in
#'   a `card_body()`. To control things like `padding`, `fill`, etc., wrap the
#'   elements in an explicit [card_body()].
#'
#' @seealso [nav_panel()], [nav_panel_hidden()] create panels of content.
#' @seealso [nav_menu()], [nav_item()], [nav_spacer()] create menus, items, or
#'   space in the navset control area.
#' @seealso [nav_insert()], [nav_remove()] programmatically add or remove nav
#'   panels.
#' @seealso [nav_select()], [nav_show()], [nav_hide()] change the state of a
#'   [nav_panel()] in a navset.
#'
#' @family Panel container functions
#' @rdname navset
#' @name navset
#' @export
navset_tab <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL
) {
  tabs <- tabsetPanel_(
    ...,
    type = "tabs",
    id = id,
    selected = selected,
    header = header,
    footer = footer
  )
  as_fragment(tabs)
}

#' @export
#' @rdname navset
navset_pill <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL
) {
  pills <- tabsetPanel_(
    ...,
    type = "pills",
    id = id,
    selected = selected,
    header = header,
    footer = footer
  )
  as_fragment(pills)
}

#' @export
#' @rdname navset
navset_underline <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL
) {
  res <- tabsetPanel_(
    ...,
    type = "underline",
    id = id,
    selected = selected,
    header = header,
    footer = footer
  )
  as_fragment(res)
}

#' @export
#' @inheritParams shiny::navlistPanel
#' @rdname navset
navset_pill_list <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  well = TRUE,
  fluid = TRUE,
  widths = c(4, 8)
) {
  pill_list <- navlistPanel_(
    ...,
    id = id,
    selected = selected,
    header = header,
    footer = footer,
    well = well,
    fluid = fluid,
    widths = widths
  )
  as_fragment(pill_list)
}

#' @export
#' @rdname navset
navset_hidden <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL
) {
  hidden <- tabsetPanel_(
    ...,
    type = "hidden",
    id = id,
    selected = selected,
    header = header,
    footer = footer
  )
  as_fragment(hidden)
}

#' @inheritParams shiny::navbarPage
#' @inheritParams page_fillable
#' @param sidebar A [sidebar()] component to display on every [nav_panel()]
#'   page.
#' @param fillable Whether or not to allow `fill` items to grow/shrink to fit
#'   the browser window. If `TRUE`, all [nav_panel()] pages are `fillable`. A
#'   character vector, matching the `value` of [nav_panel()]s to be filled, may
#'   also be provided. Note that, if a `sidebar` is provided, `fillable` makes
#'   the main content portion fillable.
#' @param navbar_options Options to control the appearance and behavior of the
#'   navbar. Use [navbar_options()] to create the list of options.
#' @param position `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(position=)`][navbar_options] instead.
#' @param collapsible `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(collapsible=)`][navbar_options] instead.
#' @param bg `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(bg=)`][navbar_options] instead.
#' @param inverse `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(inverse=)`][navbar_options] instead.
#'
#' @export
#' @rdname navset
navset_bar <- function(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  gap = NULL,
  padding = NULL,
  header = NULL,
  footer = NULL,
  fluid = TRUE,
  navbar_options = NULL,
  position = deprecated(),
  bg = deprecated(),
  inverse = deprecated(),
  collapsible = deprecated()
) {
  padding <- validateCssPadding(padding)
  gap <- validateCssUnit(gap)

  .navbar_options <- navbar_options_resolve_deprecated(
    options_user = navbar_options,
    position = position,
    bg = bg,
    inverse = inverse,
    collapsible = collapsible
  )

  navs_bar_(
    ...,
    title = title,
    id = id,
    selected = selected,
    sidebar = sidebar,
    fillable = fillable,
    gap = gap,
    padding = padding,
    header = header,
    footer = footer,
    fluid = fluid,
    navbar_options = .navbar_options,
    # theme is only used to determine whether legacy style markup should be used
    # (and, at least at the moment, we don't need legacy markup for this exported function)
    theme = bs_theme()
  )
}

# This internal version of navs_bar() exists so both it and page_navbar()
# (and thus shiny::navbarPage()) can use it. And in the page_navbar() case,
# we can use addition theme information as an indication of whether we need
# to handle backwards compatibility
navs_bar_ <- function(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  gap = NULL,
  padding = NULL,
  navbar_options = NULL,
  header = NULL,
  footer = NULL,
  fluid = TRUE,
  theme = NULL
) {
  navbar_options <- navbar_options %||% navbar_options()

  navbar_theme <- navbar_options[["theme"]]
  navbar_bg <- navbar_options[["bg"]]

  if (identical(navbar_theme, "auto")) {
    if (is.null(theme) || theme_version(theme) < 5) {
      navbar_theme <- "dark"
    }
    if (!is.null(navbar_bg)) {
      navbar_bg <- htmltools::parseCssColors(navbar_bg)
      bg_contrast <- get_color_contrast(navbar_bg)
      navbar_theme <- if (bg_contrast == "#FFFFFF") "dark" else "light"
    }
  }

  navbar <- navbarPage_(
    title = title,
    ...,
    id = id,
    selected = selected,
    sidebar = sidebar,
    fillable = fillable,
    gap = gap,
    padding = padding,
    position = navbar_options[["position"]],
    header = header,
    footer = footer,
    collapsible = navbar_options[["collapsible"]],
    inverse = identical(navbar_theme, "dark"),
    underline = navbar_options[["underline"]],
    fluid = fluid,
    theme = theme
  )

  attribs <- navbar_options[["attribs"]] %||% list()

  # Use user-provided `data-bs-theme` or our internally selected `theme`. If the
  # user includes both, the attribute wins for being most technically-correct.
  attribs[["data-bs-theme"]] <- attribs[["data-bs-theme"]] %||% navbar_theme

  # navbarPage_() returns a tagList() of the nav and content
  navbar[[1]] <- tagAppendAttributes(
    navbar[[1]],
    style = if (!is.null(navbar_bg)) {
      css(background_color = paste(navbar_bg, "!important"))
    },
    !!!attribs
  )

  as_fragment(navbar, page = page)
}

# -----------------------------------------------------------------------
# 'Internal' tabset logic that was pulled directly from shiny/R/bootstrap.R
#  (with minor modifications)
# -----------------------------------------------------------------------

navbarPage_ <- function(
  title,
  ...,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  gap = NULL,
  padding = NULL,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  header = NULL,
  footer = NULL,
  inverse = FALSE,
  underline = TRUE,
  collapsible = FALSE,
  fluid = TRUE,
  theme = NULL
) {
  # alias title so we can avoid conflicts w/ title in withTags
  pageTitle <- title

  # navbar class based on options
  navbarClass <- "navbar navbar-default"
  position <- match.arg(position)
  if (!is.null(position)) {
    navbarClass <- paste0(navbarClass, " navbar-", position)
  }
  if (isTRUE(inverse)) {
    navbarClass <- paste(navbarClass, "navbar-inverse")
  }

  if (!is.null(id)) {
    selected <- shiny::restoreInput(id = id, default = selected)
  }

  # build the tabset
  ulClass <- "nav navbar-nav"
  if (underline) {
    ulClass <- paste(ulClass, "nav-underline")
  }
  tabset <- buildTabset(..., ulClass = ulClass, id = id, selected = selected)

  containerClass <- paste0("container", if (fluid) "-fluid")

  # built the container div dynamically to support optional collapsibility
  if (collapsible) {
    navId <- paste0("navbar-collapse-", p_randomInt(1000, 10000))
    containerDiv <- div(
      class = containerClass,
      div(
        class = "navbar-header",
        tags$button(
          type = "button",
          class = "navbar-toggle collapsed",
          `data-toggle` = "collapse",
          `data-target` = paste0("#", navId),
          # data-bs-* is for BS5+
          `data-bs-toggle` = "collapse",
          `data-bs-target` = paste0("#", navId),
          span(class = "sr-only visually-hidden", "Toggle navigation"),
          span(class = "icon-bar"),
          span(class = "icon-bar"),
          span(class = "icon-bar")
        ),
        span(class = "navbar-brand", pageTitle)
      ),
      div(
        class = "navbar-collapse collapse",
        id = navId,
        tabset$navList
      )
    )
  } else {
    containerDiv <- div(
      class = containerClass,
      div(
        class = "navbar-header",
        span(class = "navbar-brand", pageTitle)
      ),
      tabset$navList
    )
  }

  # Bootstrap 3 explicitly supported "dropup menus" via .navbar-fixed-bottom,
  # but BS4+ requires .dropup on menus with .navbar.fixed-bottom
  if (position == "fixed-bottom") {
    containerDiv <- tagQuery(containerDiv)$find(
      ".dropdown-menu"
    )$parent()$addClass("dropup")$allTags()
  }

  # If fillable is truthy, give the relevant .tab-content > .tab-pane containers
  # the potential to fill
  tabset$content <- makeTabsFillable(
    tabset$content,
    fillable,
    navbar = TRUE,
    gap = gap,
    padding = padding
  )

  # For backwards compatibility reasons, wrap header & footer in a .row
  # container if we're not using BS5+. I'm not entirely sure what the motivation
  # was for it in the 1st place, but now that, with BS5, .row adds
  # `display:flex` and makes children `width:100%`, which is surprising and
  # confusing from a user perspective
  isLegacy <- as.numeric(theme_version(theme) %||% 3) < 5
  if (!is.null(header) && isLegacy) {
    header <- div(class = "row", header)
  }
  if (!is.null(footer) && isLegacy) {
    footer <- div(class = "row", footer)
  }

  contents <- dropNulls(list(header, tabset$content, footer))

  if (is.null(sidebar)) {
    contentDiv <- div(class = containerClass, !!!contents)

    # If fillable is truthy, the .container also needs to be fillable
    if (!isFALSE(fillable)) {
      contentDiv <- bindFillRole(contentDiv, container = TRUE, item = TRUE)
    }
  } else {
    contentDiv <- div(
      # In the fluid case, the sidebar layout should be flush (i.e.,
      # the .container-fluid class adds padding that we don't want)
      class = if (!fluid) "container",
      layout_sidebar(
        sidebar = sidebar,
        fillable = !isFALSE(fillable),
        border_radius = FALSE,
        border = !fluid,
        page_main_container(contents, fillable = !isFALSE(fillable))
      )
    )

    # Always have the sidebar layout fill its parent (in this case
    # fillable controls whether the _main_ content portion is fillable)
    contentDiv <- bindFillRole(contentDiv, container = TRUE, item = TRUE)
  }

  # *Don't* wrap in bootstrapPage() (shiny::navbarPage()) does that part
  tagList(
    tags$nav(class = navbarClass, role = "navigation", containerDiv),
    contentDiv
  )
}

navbarMenu_ <- function(title, ..., menuName = title, icon = NULL, align) {
  icon <- prepTabIcon(icon)
  structure(
    list(
      title = title,
      menuName = menuName,
      tabs = list2(...),
      # Here for legacy reasons
      # https://github.com/cran/miniUI/blob/74c87d3/R/layout.R#L369
      iconClass = if (inherits(icon, "shiny.tag")) {
        tagGetAttribute(icon, "class")
      },
      icon = icon,
      align = align
    ),
    class = "shiny.navbarmenu"
  )
}

isNavbarMenu <- function(x) {
  inherits(x, "shiny.navbarmenu")
}

tabPanel_ <- function(title, ..., value = title, icon = NULL) {
  icon <- prepTabIcon(icon)
  pane <- div(
    class = "tab-pane",
    title = title,
    `data-value` = value,
    # Here for legacy reasons
    # https://github.com/cran/miniUI/blob/74c87d/R/layout.R#L395
    `data-icon-class` = if (inherits(icon, "shiny.tag")) {
      tagGetAttribute(icon, "class")
    },
    ...
  )
  attr(pane, "_shiny_icon") <- icon
  pane
}

isTabPanel <- function(x) {
  if (!inherits(x, "shiny.tag")) {
    return(FALSE)
  }
  class <- tagGetAttribute(x, "class") %||% ""
  "tab-pane" %in% strsplit(class, "\\s+")[[1]]
}

tabPanelBody_ <- function(value, ..., icon = NULL) {
  if (
    !is.character(value) ||
      length(value) != 1 ||
      any(is.na(value)) ||
      nchar(value) == 0
  ) {
    stop("`value` must be a single, non-empty string value")
  }
  tabPanel_(title = NULL, ..., value = value, icon = icon)
}

tabsetPanel_ <- function(
  ...,
  id = NULL,
  selected = NULL,
  type = c("tabs", "pills", "hidden", "underline"),
  header = NULL,
  footer = NULL
) {
  if (!is.null(id)) {
    selected <- shiny::restoreInput(id = id, default = selected)
  }

  type <- match.arg(type)
  tabset <- buildTabset(
    ...,
    ulClass = paste0("nav nav-", type),
    id = id,
    selected = selected
  )

  tags$div(
    class = "tabbable",
    !!!dropNulls(
      list(
        tabset$navList,
        header,
        tabset$content,
        footer
      )
    )
  )
}

navlistPanel_ <- function(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  well = TRUE,
  fluid = TRUE,
  widths = c(4, 8)
) {
  if (!is.null(id)) {
    selected <- shiny::restoreInput(id = id, default = selected)
  }

  tabset <- buildTabset(
    ...,
    ulClass = "nav nav-pills nav-stacked",
    textFilter = function(text) tags$li(class = "navbar-brand", text),
    id = id,
    selected = selected
  )

  row <- if (fluid) shiny::fluidRow else shiny::fixedRow

  row(
    shiny::column(widths[[1]], class = if (well) "well", tabset$navList),
    shiny::column(
      widths[[2]],
      !!!dropNulls(list(header, tabset$content, footer))
    )
  )
}

# Helpers to build tabsetPanels (& Co.) and their elements
markTabAsSelected <- function(x) {
  attr(x, "selected") <- TRUE
  x
}

isTabSelected <- function(x) {
  isTRUE(attr(x, "selected", exact = TRUE))
}

containsSelectedTab <- function(tabs) {
  any(vapply(tabs, isTabSelected, logical(1)))
}

findAndMarkSelectedTab <- function(tabs, selected, foundSelected) {
  tabs <- lapply(tabs, function(x) {
    if (foundSelected || is.character(x)) {
      # Strings are not selectable items
    } else if (isNavbarMenu(x)) {
      # Recur for navbarMenus
      res <- findAndMarkSelectedTab(x$tabs, selected, foundSelected)
      x$tabs <- res$tabs
      foundSelected <<- res$foundSelected
    } else if (isTabPanel(x)) {
      # Base case: regular tab item. If the `selected` argument is
      # provided, check for a match in the existing tabs; else,
      # mark first available item as selected
      if (is.null(selected)) {
        foundSelected <<- TRUE
        x <- markTabAsSelected(x)
      } else {
        tabValue <- x$attribs$`data-value` %||% x$attribs$title
        if (identical(selected, tabValue)) {
          foundSelected <<- TRUE
          x <- markTabAsSelected(x)
        }
      }
    }
    return(x)
  })
  return(list(tabs = tabs, foundSelected = foundSelected))
}

prepTabIcon <- function(x = NULL) {
  if (!inherits(x, "shiny.tag")) {
    return(x)
  }

  is_fa <- grepl("fa-", tagGetAttribute(x, "class") %||% "", fixed = TRUE)
  if (!is_fa) {
    return(x)
  }

  # specify fixed-width for font-awesome
  tagAppendAttributes(x, class = "fa-fw")
}

# Text filter for navbarMenu's (plain text) separators
navbarMenuTextFilter <- function(text) {
  if (grepl("^\\-+$", text)) {
    tags$li(class = "divider")
  } else {
    tags$li(class = "dropdown-header", text)
  }
}

# This function is called internally by navbarPage, tabsetPanel
# and navlistPanel
buildTabset <- function(
  ...,
  ulClass,
  textFilter = NULL,
  id = NULL,
  selected = NULL,
  foundSelected = FALSE
) {
  tabs <- dropNulls(list2(...))
  res <- findAndMarkSelectedTab(tabs, selected, foundSelected)
  tabs <- res$tabs
  foundSelected <- res$foundSelected

  # add input class if we have an id
  if (!is.null(id)) {
    ulClass <- paste(ulClass, "shiny-tab-input")
  }

  if (anyNamed(tabs)) {
    nms <- names(tabs)
    nms <- nms[nzchar(nms)]
    stop(
      "Tabs should all be unnamed arguments, but some are named: ",
      paste(nms, collapse = ", ")
    )
  }

  tabsetId <- p_randomInt(1000, 10000)
  tabs <- lapply(
    seq_len(length(tabs)),
    buildTabItem,
    tabsetId = tabsetId,
    foundSelected = foundSelected,
    tabs = tabs,
    textFilter = textFilter
  )

  tabNavList <- tags$ul(
    class = ulClass,
    id = id,
    `data-tabsetid` = tabsetId,
    !!!lapply(tabs, "[[", "liTag")
  )

  tabContent <- tags$div(
    class = "tab-content",
    `data-tabsetid` = tabsetId,
    !!!lapply(tabs, "[[", "divTag")
  )

  list(navList = tabNavList, content = tabContent)
}

# Builds tabPanel/navbarMenu items (this function used to be
# declared inside the buildTabset() function and it's been
# refactored for clarity and reusability). Called internally
# by buildTabset.
buildTabItem <- function(
  index,
  tabsetId,
  foundSelected,
  tabs = NULL,
  divTag = NULL,
  textFilter = NULL
) {
  divTag <- divTag %||% tabs[[index]]

  # Handles navlistPanel() headers and dropdown dividers
  if (is.character(divTag) && !is.null(textFilter)) {
    return(list(liTag = textFilter(divTag), divTag = NULL))
  }

  if (isNavbarMenu(divTag)) {
    # tabPanelMenu item: build the child tabset
    ulClass <- "dropdown-menu"
    if (identical(divTag$align, "right")) {
      ulClass <- paste(ulClass, "dropdown-menu-right dropdown-menu-end")
    }
    tabset <- buildTabset(
      !!!divTag$tabs,
      ulClass = ulClass,
      textFilter = navbarMenuTextFilter,
      foundSelected = foundSelected
    )
    return(buildDropdown(divTag, tabset))
  }

  if (isTabPanel(divTag)) {
    return(buildNavItem(divTag, tabsetId, index))
  }

  if (is_nav_item(divTag) || is_nav_spacer(divTag)) {
    return(
      list(liTag = divTag, divTag = NULL)
    )
  }

  # The behavior is undefined at this point, so construct a condition message
  msg <- paste0(
    "Navigation containers expect a collection of `bslib::nav_panel()`/`shiny::tabPanel()`s ",
    "and/or `bslib::nav_menu()`/`shiny::navbarMenu()`s. ",
    "Consider using `header` or `footer` if you wish to place content above ",
    "(or below) every panel's contents."
  )

  # Luckily this case has never worked, so it's safe to throw here
  # https://github.com/rstudio/shiny/issues/3313
  if (!inherits(divTag, "shiny.tag")) {
    stop(msg, call. = FALSE)
  }

  # Unfortunately, this 'off-label' use case creates an 'empty' nav and includes
  # the divTag content on every tab. There shouldn't be any reason to be relying on
  # this behavior since we now have pre/post arguments, so throw a warning, but still
  # support the use case since we don't make breaking changes
  warning(msg, call. = FALSE)

  return(buildNavItem(divTag, tabsetId, index))
}

buildNavItem <- function(divTag, tabsetId, index) {
  id <- paste("tab", tabsetId, index, sep = "-")
  # Get title attribute directory (not via tagGetAttribute()) so that contents
  # don't get passed to as.character().
  # https://github.com/rstudio/shiny/issues/3352
  title <- divTag$attribs[["title"]]
  value <- divTag$attribs[["data-value"]]
  active <- isTabSelected(divTag)
  divTag <- tagAppendAttributes(divTag, class = if (active) "active")
  divTag$attribs$id <- id
  divTag$attribs$title <- NULL
  list(
    divTag = divTag,
    liTag = tagAddRenderHook(
      liTag(id, title, value, attr(divTag, "_shiny_icon")),
      function(x) {
        if (isTRUE(getCurrentThemeVersion() >= 4)) {
          tagQuery(x)$addClass("nav-item")$find("a")$addClass(
            c("nav-link", if (active) "active")
          )$allTags()
        } else {
          tagAppendAttributes(x, class = if (active) "active")
        }
      }
    )
  )
}

liTag <- function(id, title, value, icon) {
  # When title is NULL (i.e., nav_panel_hidden()), add disabled attribute to
  # prevent keyboard navigation
  is_hidden <- is.null(title)

  tags$li(
    tags$a(
      href = paste0("#", id),
      `data-toggle` = "tab",
      # data-bs-* is for BS5+
      `data-bs-toggle` = "tab",
      `data-value` = value,
      disabled = if (is_hidden) NA,
      icon,
      title
    )
  )
}

buildDropdown <- function(divTag, tabset) {
  navList <- tagAddRenderHook(
    tabset$navList,
    function(x) {
      if (isTRUE(getCurrentThemeVersion() >= 4)) {
        tagQuery(x)$find(".nav-item")$removeClass("nav-item")$find(
          ".nav-link"
        )$removeClass("nav-link")$addClass("dropdown-item")$allTags()
      } else {
        x
      }
    }
  )

  active <- containsSelectedTab(divTag$tabs)

  dropdown <- tags$li(
    class = "dropdown",
    tags$a(
      href = "#",
      class = "dropdown-toggle",
      `data-toggle` = "dropdown",
      # data-bs-* is for BS5+
      `data-bs-toggle` = "dropdown",
      `data-value` = divTag$menuName,
      divTag$icon,
      divTag$title,
      tags$b(class = "caret")
    ),
    navList,
    .renderHook = function(x) {
      if (isTRUE(getCurrentThemeVersion() >= 4)) {
        tagQuery(x)$addClass("nav-item")$find(".dropdown-toggle")$addClass(
          "nav-link"
        )$addClass(if (active) "active")$allTags()
      } else {
        tagAppendAttributes(x, class = if (active) "active")
      }
    }
  )

  list(
    divTag = tabset$content$children,
    liTag = dropdown
  )
}

utils::globalVariables("!!")

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import htmltools
#' @import sass
#' @importFrom utils modifyList packageVersion download.file URLencode getFromNamespace head
#' @importFrom stats setNames na.omit
#' @importFrom grDevices col2rgb
#' @importFrom tools file_path_sans_ext
#' @importFrom jquerylib jquery_core
#' @importFrom jsonlite fromJSON
#' @importFrom rlang := list2 abort
## usethis namespace: end
NULL

library(shiny)
library(bslib)
library(bsicons)
library(plotly)

ui <- page_fluid(
  theme = bslib::bs_theme(
    base_font = font_google("Atkinson Hyperlegible")
  ),
  br(),
  layout_column_wrap(
    width = "200px",
    class = "mt-3",
    value_box(
      title = "Unemployment Rate",
      value = "2.7%",
      p("Started at 1.5%"),
      p("Averaging 3%"),
      p("Peaked at 5.2% in Dec 1982"),
      showcase = plotlyOutput("unemploy"),
      full_screen = TRUE
    ),
    value_box(
      title = "Personal Savings Rate",
      value = "7.6%",
      p("Started at 12.6%"),
      p("Averaging 8.6%"),
      p("Peaked at 17.3% in May 1975"),
      showcase = plotlyOutput("psavert"),
      showcase_layout = showcase_top_right(),
      full_screen = TRUE,
      theme = "success"
    ),
    value_box(
      title = "Personal Consumption",
      value = "$3.8B",
      p("Started at $0.25B"),
      p("Averaging $1.7B"),
      showcase = bsicons::bs_icon("piggy-bank", size = "100%"),
      full_screen = TRUE,
      theme = "danger"
    )
  )
)

server <- function(input, output) {

  output$unemploy <- renderPlotly({
    plotly_time_series(
      economics, x = ~date, y = ~100 * unemploy / pop
    )
  })

  output$psavert <- renderPlotly({
    plotly_time_series(
      economics, x = ~date, y = ~psavert
    )
  })

  output$pce <- renderPlotly({
    plotly_time_series(
      economics, x = ~date, y = ~ 100 * pce / pop
    )
  })

  plotly_time_series <- function(d, x, y) {
    info <- getCurrentOutputInfo()
    large <- isTRUE(info$height() > 200)

    plot_ly(d, x = x, y = y) %>%
      add_lines(
        color = I(info$fg()),
        span = I(1),
        #hoverinfo = if (!large) "none",
        fill = 'tozeroy',
        alpha = 0.2
      ) %>%
      layout(
        hovermode = "x",
        margin = list(t = 0, r = 0, l = 0, b = 0),
        font = list(color = info$fg()),
        paper_bgcolor = "transparent",
        plot_bgcolor = "transparent",
        xaxis = list(
          title = "",
          visible = large,
          showgrid = FALSE
        ),
        yaxis = list(
          title = "",
          visible = large,
          showgrid = FALSE
        )
      ) %>%
      config(displayModeBar = FALSE)
  }
}

shinyApp(ui, server)

# TODO: Once we have more UI stuff in bslib, copy over the
# page constructors from shiny so folks can create static
# pages without a shiny dependency.

#' Modern Bootstrap page layouts
#'
#' These functions are small wrappers around shiny's page constructors (i.e.,
#' [shiny::fluidPage()], [shiny::navbarPage()], etc) that differ in two ways:
#'  * The `theme` parameter defaults bslib's recommended version of Bootstrap (for new projects).
#'  * The return value is rendered as an static HTML page when printed interactively at the console.
#'
#' @inheritParams shiny::bootstrapPage
#' @param ... UI elements for the page. Named arguments become HTML attributes.
#' @param theme A [bslib::bs_theme()] object.
#'
#' @seealso Dashboard-style pages: [page_sidebar()], [page_navbar()],
#'   [page_fillable()].
#'
#' @export
page <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  # Wrap ... in <body> since bootstrapPage() passes ... to tagList(),
  # which means named arguments aren't handled sensibly
  as_page(
    shiny::bootstrapPage(
      tags$body(...),
      title = title,
      theme = theme,
      lang = lang,
      # Components require Bootstrap 5+
      if (isTRUE(theme_version(theme) >= 5)) component_dependencies()
    ),
    theme = theme
  )
}

#' @describeIn page A \pkg{bslib} wrapper for [shiny::fluidPage()], a fluid
#'   Bootstrap-based page layout that extends to the full viewport width.
#' @inheritParams shiny::fluidPage
#' @export
page_fluid <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  as_page(
    shiny::fluidPage(
      ...,
      title = title,
      theme = theme,
      lang = lang,
      # Components require Bootstrap 5+
      if (isTRUE(theme_version(theme) >= 5)) component_dependencies()
    ),
    theme = theme
  )
}

#' @describeIn page A \pkg{bslib} wrapper for [shiny::fixedPage()], a fixed
#'   Bootstrap-based page layout where the page content container is centered
#'   horizontally and its width is constrained.
#' @inheritParams shiny::fixedPage
#' @export
page_fixed <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  as_page(
    shiny::fixedPage(
      ...,
      title = title,
      theme = theme,
      lang = lang,
      # Components require Bootstrap 5+
      if (isTRUE(theme_version(theme) >= 5)) component_dependencies()
    ),
    theme = theme
  )
}

#' A screen-filling page layout
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' A Bootstrap-based page layout whose contents fill the full height and width
#' of the browser window.
#'
#' @param padding Padding to use for the body. This can be a numeric vector
#'   (which will be interpreted as pixels) or a character vector with valid CSS
#'   lengths. The length can be between one and four. If one, then that value
#'   will be used for all four sides. If two, then the first value will be used
#'   for the top and bottom, while the second value will be used for left and
#'   right. If three, then the first will be used for top, the second will be
#'   left and right, and the third will be bottom. If four, then the values will
#'   be interpreted as top, right, bottom, and left respectively.
#' @param fillable_mobile Whether or not the page should fill the viewport's
#'   height on mobile devices (i.e., narrow windows).
#' @param gap A [CSS length unit][htmltools::validateCssUnit()] defining the
#'   `gap` (i.e., spacing) between elements provided to `...`.
#' @inheritParams page
#'
#' @export
#' @family Dashboard page layouts
#'
#' @seealso [layout_columns()] and [layout_column_wrap()] for laying out content
#'   into rows and columns.
#' @seealso [layout_sidebar()] for 'floating' sidebar layouts.
#' @seealso [accordion()] for grouping related input controls in the `sidebar`.
#' @seealso [card()] for wrapping outputs in the 'main' content area.
#' @seealso [value_box()] for highlighting values.
#'
#' @references
#'   * [Filling Layouts](https://rstudio.github.io/bslib/articles/filling/index.html)
#'     on the bslib website.
#'   * [Getting Started with Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'     on the bslib website.
#'
#' @examplesIf rlang::is_interactive()
#'
#' library(shiny)
#' library(ggplot2)
#'
#' ui <- page_fillable(
#'   h1("Example", code("mtcars"), "dashboard"),
#'   layout_columns(
#'     card(
#'       full_screen = TRUE,
#'       card_header("Number of forward gears"),
#'       plotOutput("gear")
#'     ),
#'     card(
#'       full_screen = TRUE,
#'       card_header("Number of carburetors"),
#'       plotOutput("carb")
#'     )
#'   ),
#'   card(
#'     full_screen = TRUE,
#'     card_header("Weight vs. Quarter Mile Time"),
#'     layout_sidebar(
#'       sidebar = sidebar(
#'         varSelectInput("var_x", "Compare to qsec:", mtcars[-7], "wt"),
#'         varSelectInput("color", "Color by:", mtcars[-7], "cyl"),
#'         position = "right"
#'       ),
#'       plotOutput("var_vs_qsec")
#'     )
#'   )
#' )
#'
#' server <- function(input, output) {
#'   for (var in c("cyl", "vs", "am", "gear", "carb")) {
#'     mtcars[[var]] <- as.factor(mtcars[[var]])
#'   }
#'
#'   output$gear <- renderPlot({
#'     ggplot(mtcars, aes(gear)) + geom_bar()
#'   })
#'
#'   output$carb <- renderPlot({
#'     ggplot(mtcars, aes(carb)) + geom_bar()
#'   })
#'
#'   output$var_vs_qsec <- renderPlot({
#'     req(input$var_x, input$color)
#'
#'     ggplot(mtcars) +
#'       aes(y = qsec, x = !!input$var_x, color = !!input$color) +
#'       geom_point()
#'   })
#' }
#'
#' shinyApp(ui, server)
page_fillable <- function(
  ...,
  padding = NULL,
  gap = NULL,
  fillable_mobile = FALSE,
  title = NULL,
  theme = bs_theme(),
  lang = NULL
) {
  page(
    title = title,
    theme = theme,
    lang = lang,
    class = "bslib-page-fill bslib-gap-spacing",
    class = if (!fillable_mobile) "bslib-flow-mobile",
    style = css(
      padding = validateCssPadding(padding),
      gap = validateCssUnit(gap)
    ),
    ...,
    tags$head(tags$style("html { height: 100%; }")),
    as_fillable_container()
  )
}

validateCssPadding <- function(padding = NULL) {
  paste(
    vapply(padding, validateCssUnit, character(1)),
    collapse = " "
  )
}

#' A sidebar page (i.e., dashboard)
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Create a dashboard layout with a full-bleed header (`title`) and [sidebar()].
#'
#' @param ... UI elements to display in the 'main' content area (i.e., next to
#'   the `sidebar`). These arguments are passed to `layout_sidebar()`, which has
#'   more details.
#' @param title A string, number, or [htmltools::tag()] child to display as the
#'   title (just above the `sidebar`).
#' @inheritParams layout_sidebar
#' @inheritParams page_fillable
#' @inheritParams page_navbar
#'
#' @export
#' @family Dashboard page layouts
#'
#' @seealso [layout_columns()] and [layout_column_wrap()] for laying out content
#'   into rows and columns.
#' @seealso [accordion()] for grouping related input controls in the `sidebar`.
#' @seealso [card()] for wrapping outputs in the 'main' content area.
#' @seealso [value_box()] for highlighting values.
#'
#' @references [Getting Started with Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'   on the bslib website.
#'
#' @examplesIf rlang::is_interactive()
#'
#' library(shiny)
#' library(ggplot2)
#'
#' ui <- page_sidebar(
#'   title = "Example dashboard",
#'   sidebar = sidebar(
#'     varSelectInput("var", "Select variable", mtcars)
#'   ),
#'   card(
#'     full_screen = TRUE,
#'     card_header("My plot"),
#'     plotOutput("p")
#'   )
#' )
#'
#' server <- function(input, output) {
#'   output$p <- renderPlot({
#'     ggplot(mtcars) + geom_histogram(aes(!!input$var))
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
page_sidebar <- function(
  ...,
  sidebar = NULL,
  title = NULL,
  fillable = TRUE,
  fillable_mobile = FALSE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL
) {
  if (rlang::is_bare_character(title) || rlang::is_bare_numeric(title)) {
    title <- h1(title, class = "bslib-page-title")
  }

  page_fillable(
    padding = 0,
    gap = 0,
    title = infer_window_title(title, window_title),
    theme = theme,
    lang = lang,
    fillable_mobile = fillable_mobile,
    class = "bslib-page-sidebar",
    title,
    layout_sidebar(
      sidebar = sidebar,
      fillable = fillable,
      border = FALSE,
      border_radius = FALSE,
      ...
    )
  )
}


#' Multi-page app with a top navigation bar
#'
#' @description
#' Create a page that contains a top level navigation bar that can be used to
#' toggle a set of [nav_panel()] elements. Use this page layout to create the
#' effect of a multi-page app, where your app's content is broken up into
#' multiple "pages" that can be navigated to via the top navigation bar.
#'
#' @param fillable_mobile Whether or not `fillable` pages should fill the viewport's
#'   height on mobile devices (i.e., narrow windows).
#' @param underline Whether or not to add underline styling to page links when
#'   active or focused.
#' @param window_title the browser window title. The default value, `NA`, means
#'   to use any character strings that appear in `title` (if none are found, the
#'   host URL of the page is displayed by default).
#' @inheritParams navset_bar
#' @inheritParams page
#'
#' @export
#' @family Dashboard page layouts
#'
#' @seealso [nav_panel()], [nav_menu()], and [nav_item()] for adding content
#'   sections and organizing or creating items in the navigation bar.
#' @seealso [layout_columns()] and [layout_column_wrap()] for laying out content
#'   into rows and columns.
#' @seealso [card()] for wrapping outputs in the 'main' content area.
#' @seealso [value_box()] for highlighting values.
#' @seealso [accordion()] for grouping related input controls in the `sidebar`.
#'
#' @references [Getting Started with Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'   on the bslib website.
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' library(bslib)
#'
#' link_shiny <- tags$a(
#'   shiny::icon("github"), "Shiny",
#'   href = "https://github.com/rstudio/shiny",
#'   target = "_blank"
#' )
#' link_posit <- tags$a(
#'   shiny::icon("r-project"), "Posit",
#'   href = "https://posit.co",
#'   target = "_blank"
#' )
#'
#' ui <- page_navbar(
#'   title = "My App",
#'   nav_panel(title = "One", p("First page content.")),
#'   nav_panel(title = "Two", p("Second page content.")),
#'   nav_panel("Three", p("Third page content.")),
#'   nav_spacer(),
#'   nav_menu(
#'     title = "Links",
#'     align = "right",
#'     nav_item(link_shiny),
#'     nav_item(link_posit)
#'   )
#' )
#'
#' server <- function(...) { } # not used in this example
#'
#' shinyApp(ui, server)
page_navbar <- function(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  fillable_mobile = FALSE,
  gap = NULL,
  padding = NULL,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  header = NULL,
  footer = NULL,
  bg = NULL,
  inverse = "auto",
  underline = TRUE,
  collapsible = TRUE,
  fluid = TRUE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL
) {


  # Default to fillable = F when this is called from shiny::navbarPage()
  # TODO: update shiny::navbarPage() to set fillable = FALSE and get rid of this hack
  if (missing(fillable)) {
    isNavbarPage <- isNamespaceLoaded("shiny") && identical(rlang::caller_fn(), shiny::navbarPage)
    if (isNavbarPage) {
      fillable <- FALSE
    }
  }

  # If a sidebar is provided, we want the layout_sidebar(fill = TRUE) component
  # (which is a sibling of the <nav>) to always fill the page
  page_func <- if (isFALSE(fillable) && is.null(sidebar)) {
    page
  } else {
    function(...) page_fillable(..., fillable_mobile = fillable_mobile, padding = 0, gap = 0)
  }

  page_func(
    title = infer_window_title(title, window_title),
    theme = theme,
    lang = lang,
    class = "bslib-page-navbar",
    class = if (!is.null(sidebar)) "has-page-sidebar",
    navs_bar_(
      ..., title = title, id = id, selected = selected,
      sidebar = sidebar, fillable = fillable,
      gap = gap, padding = padding,
      position = match.arg(position), header = header,
      footer = footer, bg = bg, inverse = inverse,
      underline = underline, collapsible = collapsible,
      fluid = fluid, theme = theme
    )
  )
}

# https://github.com/rstudio/shiny/issues/2310
infer_window_title <- function(title = NULL, window_title = NA) {
  if (!isTRUE(is.na(window_title))) {
    return(window_title)
  }

  if (!is.null(title)) {
    window_title <- unlist(find_characters(title))
    if (is.null(window_title)) {
      warning("Unable to infer a `window_title` default from `title`. Consider providing a character string to `window_title`.")
    } else {
      window_title <- paste(window_title, collapse = " ")
    }
  }

  window_title
}


# CPS (2023-02-09): Joe is currently working on a potentially
# more compelling contain_width() interface, so we'll punt on this for now
#
# #' Contain, pad, and align content
# #'
# #' @param ... A collection of [htmltools::tag()] children.
# #' @param size A size (i.e., max-width policy) for the container.
# #' @param bg A background color.
# #' @param class Additional CSS classes for the container.
# #'
# #' @references <https://getbootstrap.com/docs/5.3/layout/containers/>
# #'
# #' @export
# contain_width <- function(..., size = c("sm", "md", "lg", "xl", "xxl", # "fluid"), bg = NULL, class = NULL) {
#
#   size <- match.arg(size)
#
#   res <- div(
#     class = paste0("container-", size),
#     class = class,
#     # TODO: parseCssColors(), once it supports var() and !important
#     style = css(background_color = bg),
#     ...
#   )
#
#   as_fragment(
#     tag_require(res, version = 5, caller = "contain_width()")
#   )
# }

#> unlist(find_characters(div(h1("foo"), h2("bar"))))
#> [1] "foo" "bar"
find_characters <- function(x) {
  if (is.character(x)) {
    return(x)
  }
  if (inherits(x, "shiny.tag")) {
    return(lapply(x$children, find_characters))
  }
  if (is.list(x)) {
    return(lapply(x, find_characters))
  }
  NULL
}

library(htmltools)

# All page_*() functions are very thin wrappers around
# shiny::*Page() at the moment (except for page_navbar(), which
# which why we only have tests for page_navbar())

test_that("page_navbar()", {
  skip_if_not_installed("shiny")
  expect_snapshot(
    renderTags(
      page_navbar(
        title = div(h1("foo"), h2("bar"))
      )
    )$head,
    cran = TRUE
  )

  expect_snapshot(
    renderTags(
      page_navbar(
        title = "foo",
        window_title = "bar"
      )
    )$head,
    cran = TRUE
  )

  expect_snapshot(
    renderTags(
      page_sidebar(title = "foo")
    )$head,
    cran = TRUE
  )
})

test_that("page_sidebar()", {
  with_private_seed()

  expect_snapshot(
    renderTags(
      page_sidebar(
        title = "foo",
        window_title = "bar"
      )
    )$head,
    cran = TRUE
  )

  with_private_seed()

  expect_snapshot(
    renderTags(
      page_sidebar(
        "main",
        title = "Title",
        # Removes the {bsicons} icon
        sidebar = sidebar(open = "always"),
        "data-attr" = "here"
      )
    )$html,
    cran = TRUE
  )

  with_private_seed()

  skip_if_not_installed("bsicons", "0.1.0.9000")

  expect_snapshot(
    renderTags(
      page_sidebar(
        "main",
        title = "Title",
        sidebar = "side"
      )
    )$html,
    # Don't run on CRAN since the output depends on {bsicons}
    cran = FALSE
  )
})

test_that("save_html() works on components and pages with a custom theme", {
  withr::with_tempdir({
    withr::local_options(list(htmltools.dir.version = FALSE))
    save_html(
      card("A simple card"),
      "card.html"
    )
    expect_snapshot_file("card.html")
  })

  withr::with_tempdir({
    withr::local_options(list(htmltools.dir.version = FALSE))
    save_html(
      page(
        theme = bs_remove(bs_theme(), "bs3compat"),
        "A simple page without bs3compat dependencies"
      ),
      "modern-page.html"
    )
    expect_snapshot_file("modern-page.html")
  })
})

test_that("page_*() functions can handle trailing commas", {
  expect_no_error(
    page("foo", )
  )
  expect_no_error(
    page_fluid("foo", )
  )
  expect_no_error(
    page_fixed("foo", )
  )
  expect_no_error(
    page_fillable("foo", )
  )
  expect_no_error(
    page_sidebar("foo", )
  )
  expect_no_error(
    page_navbar(nav_panel("foo", "bar"), )
  )
})

library(htmltools)

# All page_*() functions are very thin wrappers around
# shiny::*Page() at the moment (except for page_navbar(), which
# which why we only have tests for page_navbar())

test_that("page_navbar()", {
  skip_if_not_installed("shiny")
  expect_snapshot(
    renderTags(
      page_navbar(
        title = div(h1("foo"), h2("bar"))
      )
    )$head,
    cran = TRUE
  )

  expect_snapshot(
    renderTags(
      page_navbar(
        title = "foo",
        window_title = "bar"
      )
    )$head,
    cran = TRUE
  )

  expect_snapshot(
    renderTags(
      page_sidebar(title = "foo")
    )$head,
    cran = TRUE
  )
})

test_that("page_sidebar()", {

  with_private_seed()

  expect_snapshot(
    renderTags(
      page_sidebar(
        title = "foo",
        window_title = "bar"
      )
    )$head,
    cran = TRUE
  )

  with_private_seed()

  expect_snapshot(
    renderTags(
      page_sidebar(
        "main",
        title = "Title",
        # Removes the {bsicons} icon
        sidebar = sidebar(open = "always")
      )
    )$html,
    cran = TRUE
  )

  with_private_seed()

  skip_if_not_installed("bsicons", "0.1.0.9000")

  expect_snapshot(
    renderTags(
      page_sidebar(
        "main",
        title = "Title",
        sidebar = "side"
      )
    )$html,
    # Don't run on CRAN since the output depends on {bsicons}
    cran = FALSE
  )

})

library(shiny)
library(bslib)

# Load sample code snippets into a named list
sample_code_files <- dir("examples", full.names = TRUE)
names(sample_code_files) <- gsub(
  "\\.[[:alpha:]]+$",
  "",
  basename(sample_code_files)
)
sample_code <- lapply(sample_code_files, readLines)
sample_code <- lapply(sample_code, paste, collapse = "\n")

ui <- page_sidebar(
  title = "Code Editor Demo",
  theme = bs_theme(version = 5),
  class = "bslib-page-dashboard",

  sidebar = sidebar(
    width = 300,
    title = "Editor Controls",
    gap = "0.5rem",

    textInput("label", "Label:", value = "Code Editor"),

    selectInput(
      "language",
      "Language:",
      choices = c("plain", "html", bslib:::code_editor_bundled_languages),
      selected = "r"
    ),
    actionButton(
      "load_sample",
      "Load Sample Code",
      class = "btn-secondary btn-sm w-100 mb-2"
    ),
    actionButton(
      "clear_code",
      "Clear Editor",
      class = "btn-warning btn-sm w-100 mb-2"
    ),

    selectInput(
      "theme_light",
      "Light Theme:",
      choices = bslib:::code_editor_themes(),
      selected = "github-light"
    ),

    selectInput(
      "theme_dark",
      "Dark Theme:",
      choices = bslib:::code_editor_themes(),
      selected = "github-dark"
    ),

    p(
      actionLink(
        inputId = "toggle_dark_mode",
        style = css(text_decoration = "none"),
        label = tagList(
          input_dark_mode(
            id = "dark_mode",
            mode = "light",
            style = css("--vertical-correction" = "5px"),
            style = "color: currentColor"
          ),
          "Toggle Theme"
        )
      ),
      class = "text-end small"
    ),

    checkboxInput("read_only", "Read Only", value = FALSE),
    checkboxInput("line_numbers", "Line Numbers", value = TRUE),
    checkboxInput("word_wrap", "Word Wrap", value = FALSE),

    sliderInput("tab_size", "Tab Size:", min = 2, max = 8, value = 2, step = 1),

    radioButtons(
      "indentation",
      "Indentation:",
      choices = c("Spaces" = "space", "Tabs" = "tab"),
      selected = "space",
      inline = TRUE
    )
  ),

  layout_columns(
    card(
      card_header("Code Editor"),
      card_body(
        p(
          "This editor supports syntax highlighting, line numbers, word wrap, and more. ",
          "Try pressing ",
          tags$kbd("Ctrl/Cmd+Enter"),
          " to submit the code."
        ),
        input_code_editor(
          "code",
          value = "library(tidyverse)\n\n# Sample R code\nmtcars |>\n  group_by(cyl) |>\n  summarise(avg_mpg = mean(mpg)) |>\n  arrange(desc(avg_mpg))",
          label = "Code Editor",
          language = "r",
          line_numbers = TRUE,
          word_wrap = FALSE,
          # height = "400px",
          fill = TRUE
        )
      )
    ),
    layout_columns(
      col_widths = 12,
      navset_card_underline(
        title = "Editor Info",
        nav_panel("Value", verbatimTextOutput("code_output"), ),
        nav_panel("Settings", verbatimTextOutput("editor_info"))
      ),

      card(
        card_header("Features & Keyboard Shortcuts"),
        card_body(
          tags$ul(
            tags$li(
              tags$kbd("Ctrl/Cmd+Enter"),
              " - Submit code to R (triggers reactive update)"
            ),
            tags$li(tags$kbd("Ctrl/Cmd+Z"), " - Undo"),
            tags$li(tags$kbd("Ctrl/Cmd+Shift+Z"), " - Redo"),
            tags$li(tags$kbd("Tab"), " - Indent selection"),
            tags$li(tags$kbd("Shift+Tab"), " - Dedent selection"),
            tags$li("Copy button in top-right corner"),
            tags$li("Automatic light/dark theme switching"),
            tags$li("Update on blur (when editor loses focus)")
          )
        )
      )
    )
  )
)

server <- function(input, output, session) {
  observeEvent(input$label, {
    update_code_editor("code", label = input$label)
  })

  observeEvent(input$language, {
    language <- switch(
      input$language,
      plain = "markdown",
      html = "markup",
      input$language
    )
    updateActionButton(
      session,
      "load_sample",
      disabled = !language %in% names(sample_code)
    )
    update_code_editor("code", language = language)
  })
  observeEvent(input$theme_light, {
    update_code_editor("code", theme_light = input$theme_light)
  })
  observeEvent(input$theme_dark, {
    update_code_editor("code", theme_dark = input$theme_dark)
  })
  observeEvent(input$read_only, {
    update_code_editor("code", read_only = input$read_only)
  })
  observeEvent(input$line_numbers, {
    update_code_editor("code", line_numbers = input$line_numbers)
  })
  observeEvent(input$word_wrap, {
    update_code_editor("code", word_wrap = input$word_wrap)
  })
  observeEvent(input$tab_size, {
    update_code_editor("code", tab_size = input$tab_size)
  })
  observeEvent(input$indentation, {
    update_code_editor("code", indentation = input$indentation)
  })

  # Load sample code for selected language
  observeEvent(input$load_sample, {
    lang <- switch(
      input$language,
      plain = "markdown",
      html = "markup",
      input$language
    )
    sample <- sample_code[[lang]]
    if (!is.null(sample)) {
      update_code_editor(
        "code",
        value = sample,
        language = lang
      )
    }
  })

  # Clear editor
  observeEvent(input$clear_code, {
    update_code_editor("code", value = "")
  })

  observeEvent(input$toggle_dark_mode, {
    bslib::toggle_dark_mode()
  })

  # Display current code
  output$code_output <- renderText({
    code <- input$code
    if (is.null(code) || code == "") {
      "[Editor is empty]"
    } else {
      code
    }
  })

  # Display editor information
  output$editor_info <- renderText({
    code <- input$code
    if (is.null(code)) {
      code <- ""
    }

    lines <- length(strsplit(code, "\n")[[1]])
    chars <- nchar(code)

    paste(
      sprintf("Language: %s", input$language),
      sprintf("Lines: %d", lines),
      sprintf("Characters: %d", chars),
      sprintf("Read Only: %s", input$read_only),
      sprintf("Line Numbers: %s", input$line_numbers),
      sprintf("Word Wrap: %s", input$word_wrap),
      sprintf("Tab Size: %d", input$tab_size),
      sprintf("Indentation: %s", input$indentation),
      sep = "\n"
    )
  })
}

shinyApp(ui = ui, server = server)

#' Add a tooltip to a UI element
#'
#' Display additional information when focusing (or hovering over) a UI element.
#'
#' @param trigger A UI element (i.e., [htmltools tag][htmltools::tags]) to serve
#'   as the tooltip trigger. If `trigger` renders as multiple HTML
#'   elements (e.g., it's a `tagList()`), the last HTML element is used for the
#'   trigger. If the `trigger` should contain all of those elements, wrap the
#'   object in a [div()] or [span()].
#' @param ... UI elements for the tooltip. Character strings are [automatically
#'   escaped][htmlEscape()] unless marked as [HTML()].
#' @param id A character string. Required to re-actively respond to the
#'   visibility of the tooltip (via the `input[[id]]` value) and/or update the
#'   visibility/contents of the tooltip.
#' @param placement The placement of the tooltip relative to its trigger.
#' @param options A list of additional [options](https://getbootstrap.com/docs/5.3/components/tooltips/#options).
#'
#' @section Theming/Styling:
#'
#'   Like other bslib components, tooltips can be themed by supplying [relevant
#'   theming
#'   variables](https://rstudio.github.io/bslib/articles/bs5-variables/index.html#tooltip-bg)
#'   to [bs_theme()], which effects styling of every popover on the page. To
#'   style a _specific_ popover differently from other popovers, utilize the
#'   `customClass` option:
#'
#'   ```
#'   tooltip(
#'     "Trigger", "Tooltip message",
#'     options = list(customClass = "my-tip")
#'   )
#'   ```
#'
#'   And then add relevant rules to [bs_theme()] via [bs_add_rules()]:
#'
#'   ```
#'   bs_theme() |> bs_add_rules(".my-tip { max-width: none; }")
#'   ```
#'
#' @describeIn tooltip Add a tooltip to a UI element
#' @references <https://getbootstrap.com/docs/5.3/components/tooltips/>
#' @export
#' @seealso [popover()]
#' @examplesIf interactive()
#'
#' tooltip(
#'   shiny::actionButton("btn", "A button"),
#'   "A message"
#' )
#'
#' card(
#'   card_header(
#'     tooltip(
#'       span("Card title ", bsicons::bs_icon("question-circle-fill")),
#'       "Additional info",
#'       placement = "right"
#'     )
#'   ),
#'   "Card body content..."
#' )
tooltip <- function(
  trigger,
  ...,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
) {

  args <- separate_arguments(...)
  children <- args$children
  attribs <- args$attribs

  if (length(children) == 0) {
    abort("At least one value must be provided to `...`.")
  }

  bad_opts <- intersect(c("title", "placement"), names(options))
  if (length(bad_opts) > 0) {
    rlang::abort(
      sprintf("The `%s` option cannot be specified directly.", bad_opts[1])
    )
  }

  res <- web_component(
    "bslib-tooltip",
    id = id,
    placement = rlang::arg_match(placement),
    bsOptions = to_json(options),
    !!!attribs,
    # Use <template> as a way to protect these children from potentially being
    # pulled outside this element (the browser's parser does this to, for
    # example, block elements inside a <p> tag)
    tags$template(!!!children),
    trigger
  )

  res <- tag_require(res, version = 5, caller = "tooltip()")
  as_fragment(res)
}

#' @describeIn tooltip Programmatically show/hide a tooltip.
#'
#' @param id a character string that matches an existing tooltip id.
#' @param show Whether to show (`TRUE`) or hide (`FALSE`) the tooltip. The
#'   default (`NULL`) will show if currently hidden and hide if currently shown.
#'   Note that a tooltip will not be shown if the trigger is not visible (e.g.,
#'   it's hidden behind a tab).
#' @param session A Shiny session object (the default should almost always be
#'   used).
#'
#' @export
toggle_tooltip <- function(id, show = NULL, session = get_current_session()) {
  show <- normalize_show_value(show)

  msg <- list(method = "toggle", value = show)
  force(id)
  callback <- function() {
    session$sendInputMessage(id, msg)
  }
  session$onFlush(callback, once = TRUE)
}


#' @describeIn tooltip Update the contents of a tooltip.
#' @export
update_tooltip <- function(id, ..., session = get_current_session()) {

  title <- tagList(...)

  msg <- dropNulls(list(
    method = "update",
    title = if (length(title) > 0) processDeps(title, session)
  ))

  force(id)
  callback <- function() {
    session$sendInputMessage(id, msg)
  }
  session$onFlush(callback, once = TRUE)
}

normalize_show_value <- function(show) {
  if (is.null(show)) return("toggle")

  if (length(show) != 1 || !is.logical(show)) {
    abort("`show` must be `TRUE`, `FALSE`, or `NULL`.")
  }

  if (show) "show" else "hide"
}


to_json <- function(..., auto_unbox = TRUE, null = "null") {
  jsonlite::toJSON(..., auto_unbox = auto_unbox, null = null)
}

if (!identical(getwd(), rprojroot::find_package_root_file())) {
  stop("This script must be run from the top directory of the bslib package")
}

pkgload::load_all()

components <- fs::dir_ls(path_components("scss"), type = "file")
components <- basename(tools::file_path_sans_ext(components))

# TODO: save a CSS file for every version >= 5?
lapply(components, function(x) {
  dep <- component_dependency_sass_(bs_theme(), x)
  css <- file.path(dep$src, dep$stylesheet)
  target_dir <- file.path(path_components("dist"), x)
  if (!dir.exists(target_dir)) dir.create(target_dir)
  target_css <- file.path(target_dir, paste0(x, ".css"))
  file.copy(css, target_css, overwrite = TRUE)
})

#' Obtain a list of all available built-in \pkg{bslib} themes.
#'
#' @param version the major version of Bootstrap.
#' @param full_path whether to return a path to the installed theme.
#'
#' @return Returns a character vector of built-in themes provided by
#'   \pkg{bslib}.
#'
#' @family Bootstrap theme utility functions
#' @export
builtin_themes <- function(version = version_default(), full_path = FALSE) {
  path_builtins <- path_builtin_theme(version = version)
  if (is.null(path_builtins)) {
    return(NULL)
  }

  list.dirs(path_builtins, full.names = full_path, recursive = FALSE)
}

builtin_bundle <- function(name = "shiny", version = version_default()) {
  name <- validate_builtin_preset_name(name, version = version)

  font_css <- path_builtin_theme(name, "font.css", version = version)
  attachments <- if (file.exists(font_css)) {
    c(
      "font.css" = font_css,
      fonts = system_file("fonts", package = "bslib")
    )
  }

  sass_parts <- rlang::set_names(c("variables", "mixins", "rules"))

  sass_files <- lapply(sass_parts, function(part) {
    path <- path_builtin_theme(
      name,
      paste0("_", part, ".scss"),
      version = version
    )
    if (file.exists(path)) sass_file(path)
  })

  sass_bundle(
    "builtin" = sass_layer(
      file_attachments = attachments,
      defaults = list(
        "bslib-preset-type" = "builtin",
        "bslib-preset-name" = name,
        '$web-font-path: "font.css" !default;',
        sass_files$variables
      ),
      mixins = list(sass_files$mixins),
      rules = list(sass_files$rules)
    )
  )
}

validate_builtin_preset_name <- function(name, version = version_default()) {
  builtin_names <- builtin_themes(version)

  if (name %in% builtin_names) {
    return(name)
  }

  msg <- "'%s' is not a valid built-in theme preset provided by {bslib}."

  info <-
    if (length(builtin_names)) {
      sprintf(
        "Available Bootstrap %s themes are: '%s'",
        version,
        paste0(builtin_themes(version), collapse = "', '")
      )
    } else {
      "No built-in theme presets are available for this version of Bootstrap."
    }

  rlang::abort(c(sprintf(msg, name), "i" = info))
}

path_builtin_theme <- function(..., version = version_default()) {
  switch_version(
    version,
    five = system.file("builtin", "bs5", ..., package = "bslib")
  )
}

utils::globalVariables("!!")

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import htmltools
#' @import sass
#' @importFrom grDevices col2rgb
#' @importFrom jquerylib jquery_core
#' @importFrom jsonlite fromJSON
#' @importFrom lifecycle deprecated
#' @importFrom rlang := list2 abort
#' @importFrom stats setNames na.omit
#' @importFrom tools file_path_sans_ext
#' @importFrom utils modifyList packageVersion download.file URLencode getFromNamespace head
## usethis namespace: end
NULL


# For usethis::use_release_issue()
release_bullets <- function() {
  c(
    "Update static imports: `staticimports::import()`"
  )
}

#' Compile Bootstrap Sass with (optional) theming
#'
#' `bs_theme_dependencies()` compiles Bootstrap Sass into CSS and returns it,
#' along with other HTML dependencies, as a list of
#' [htmltools::htmlDependency()]s. Most users won't need to call this function
#' directly as Shiny & R Markdown will perform this compilation automatically
#' when handed a [bs_theme()]. If you're here looking to create a themeable
#' component, see [bs_dependency()].
#'
#' @section Sass caching and precompilation:
#'
#' If Shiny Developer Mode is enabled (by setting `options(shiny.devmode =
#' TRUE)` or calling `shiny::devmode(TRUE)`), both \pkg{sass} caching and
#' \pkg{bslib} precompilation are disabled by default; that is, a call to
#' `bs_theme_dependencies(theme)` expands to `bs_theme_dependencies(theme, cache
#' = F, precompiled = F)`). This is useful for local development as enabling
#' caching/precompilation may produce incorrect results if local changes are
#' made to bslib's source files.
#'
#' @inheritParams bs_theme_update
#' @param sass_options a [sass::sass_options()] object.
#' @param jquery a [jquerylib::jquery_core()] object.
#' @param precompiled Before compiling the theme object, first look for a
#'   precompiled CSS file for the [theme_version()].  If `precompiled = TRUE`
#'   and a precompiled CSS file exists for the theme object, it will be fetched
#'   immediately and not compiled. At the moment, we only provide precompiled
#'   CSS for "stock" builds of Bootstrap (i.e., no theming additions, Bootswatch
#'   themes, or non-default `sass_options`).
#' @inheritParams sass::sass
#'
#' @return Returns a list of HTML dependencies containing Bootstrap CSS,
#'   Bootstrap JavaScript, and `jquery`. This list may contain additional HTML
#'   dependencies if bundled with the `theme`.
#'
#' @export
#' @family Bootstrap theme functions
#'
#' @examplesIf rlang::is_interactive()
#'
#' # Function to preview the styling a (primary) Bootstrap button
#' library(htmltools)
#' button <- tags$a(class = "btn btn-primary", href = "#", role = "button", "Hello")
#' preview_button <- function(theme) {
#'   browsable(tags$body(bs_theme_dependencies(theme), button))
#' }
#'
#' # Latest Bootstrap
#' preview_button(bs_theme())
#' # Bootstrap 3
#' preview_button(bs_theme(3))
#' # Bootswatch 4 minty theme
#' preview_button(bs_theme(4, bootswatch = "minty"))
#' # Bootswatch 4 sketchy theme
#' preview_button(bs_theme(4, bootswatch = "sketchy"))
#'
bs_theme_dependencies <- function(
  theme,
  sass_options = sass::sass_options_get(output_style = "compressed"),
  cache = sass::sass_cache_get(),
  jquery = jquerylib::jquery_core(3),
  precompiled = get_precompiled_option("bslib.precompiled", default = TRUE)
) {
  theme <- as_bs_theme(theme)
  version <- theme_version(theme)

  if (isTRUE(version >= 5)) {
    register_runtime_package_check("`bs_theme(version = 5)`", "shiny", "1.7.0")
  }

  if (is.character(cache)) {
    cache <- sass_cache_get_dir(cache)
  }

  out_file <- maybe_precompiled_css(theme, sass_options, precompiled)

  if (is.null(out_file)) {
    out_file <- sass_compile_theme(theme, sass_options, cache)
  }

  htmltools::resolveDependencies(
    c(
      if (inherits(jquery, "html_dependency")) list(jquery) else jquery,
      list(
        htmlDependency(
          name = "bootstrap",
          version = get_exact_version(version),
          src = dirname(out_file),
          stylesheet = basename(out_file),
          script = basename(bootstrap_javascript(version)),
          all_files = TRUE, # include font and map files
          meta = list(
            viewport = "width=device-width, initial-scale=1, shrink-to-fit=no"
          )
        )
      ),
      htmlDependencies(out_file)
    )
  )
}

maybe_precompiled_css <- function(theme, sass_options, precompiled) {
  if (!precompiled) {
    return()
  }
  if (!identical(sass_options, sass_options(output_style = "compressed"))) {
    return()
  }

  precompiled_css <- precompiled_css_path(theme)
  if (is.null(precompiled_css)) {
    return()
  }

  version <- theme_version(theme)
  bs_js_hash <- rlang::hash_file(bootstrap_javascript(version))
  out_dir <- file.path(
    tempdir(),
    sprintf("bslib-precompiled-%s-%s", version, bs_js_hash)
  )

  out_file <- file.path(out_dir, basename(precompiled_css))
  out_file <- attachDependencies(out_file, htmlDependencies(as_sass(theme)))

  if (dir.exists(out_dir)) {
    # We've already copied all the files for this precompiled theme, there's no
    # need to copy them again.
    return(out_file)
  }

  dir.create(out_dir)
  file.copy(precompiled_css, out_file)

  # Usually sass() would handle file_attachments and dependencies,
  # but we need to do this manually
  write_file_attachments(
    as_sass_layer(theme)$file_attachments,
    out_dir
  )

  bootstrap_javascript_copy_assets(version, dirname(out_file))

  out_file
}

sass_compile_theme <- function(theme, sass_options, sass_cache) {
  version <- theme_version(theme)

  contrast_warn <- get_shiny_devmode_option(
    "bslib.color_contrast_warnings",
    default = FALSE,
    devmode_default = TRUE,
    devmode_message = paste(
      "Enabling warnings about low color contrasts found inside `bslib::bs_theme()`.",
      "To suppress these warnings, set `options(bslib.color_contrast_warnings = FALSE)`"
    )
  )
  theme <- bs_add_variables(theme, "color-contrast-warnings" = contrast_warn)

  out_file <- sass(
    input = theme,
    options = sass_options,
    output = output_template(basename = "bootstrap", dirname = "bslib-"),
    cache = sass_cache,
    write_attachments = TRUE,
    cache_key_extra = list(
      get_exact_version(version),
      get_package_version("bslib")
    )
  )

  bootstrap_javascript_copy_assets(version, dirname(out_file))

  out_file
}

bootstrap_javascript_copy_assets <- function(version, to) {
  js_files <- bootstrap_javascript(version)
  js_map_files <- bootstrap_javascript_map(version)
  success_js_files <- file.copy(c(js_files, js_map_files), to, overwrite = TRUE)
  if (any(!success_js_files)) {
    warning(
      "Failed to copy over bootstrap's javascript files into the htmlDependency() directory."
    )
  }
}

#' Themeable HTML components
#'
#' @description
#'
#' Themeable HTML components use Sass to generate CSS rules from Bootstrap Sass
#' variables, functions, and/or mixins (i.e., stuff inside of `theme`).
#' `bs_dependencies()` makes it a bit easier to create themeable components by
#' compiling [sass::sass()] (`input`) together with Bootstrap Sass inside of a
#' `theme`, and packaging up the result into an [htmltools::htmlDependency()].
#'
#' Themable components can also be  _dynamically_ themed inside of Shiny (i.e.,
#' they may be themed in 'real-time' via [bs_themer()], and more generally,
#' update their styles in response to [shiny::session]'s `setCurrentTheme()`
#' method). Dynamically themeable components provide a "recipe" (i.e., a
#' function) to `bs_dependency_defer()`, describing how to generate new CSS
#' stylesheet(s) from a new `theme`. This function is called when the HTML page
#' is first rendered, and may be invoked again with a new `theme` whenever
#' [shiny::session]'s `setCurrentTheme()` is called.
#'
#' @param input Sass rules to compile, using `theme`.
#' @param theme A [bs_theme()] object.
#' @param cache_key_extra Extra information to add to the sass cache key. It is
#'   useful to add the version of your package.
#' @param .dep_args A list of additional arguments to pass to
#'   [htmltools::htmlDependency()]. Note that `package` has no effect and
#'   `script` must be absolute path(s).
#' @param .sass_args A list of additional arguments to pass to
#'   [sass::sass_partial()].
#' @inheritParams htmltools::htmlDependency
#'
#' @references
#'   * [Theming: Custom components](https://rstudio.github.io/bslib/articles/custom-components/index.html)
#'     gives a tutorial on creating a dynamically themable custom component.
#'
#' @return `bs_dependency()` returns an [htmltools::htmlDependency()] and
#'   `bs_dependency_defer()` returns an [htmltools::tagFunction()]
#'
#' @family Bootstrap theme functions
#' @export
bs_dependency <- function(
  input = list(),
  theme,
  name,
  version,
  cache_key_extra = NULL,
  .dep_args = list(),
  .sass_args = list()
) {
  sass_args <- c(
    list(
      rules = input,
      bundle = theme,
      output = output_template(basename = name, dirname = name),
      write_attachments = TRUE,
      cache_key_extra = cache_key_extra
    ),
    .sass_args
  )
  outfile <- do.call(sass_partial, sass_args)

  dep_args <- list(
    name = name,
    version = version,
    src = dirname(outfile),
    stylesheet = basename(outfile)
  )

  bad_args <- intersect(names(.dep_args), names(dep_args))
  if (length(bad_args)) {
    stop(
      "The following `.dep_args` must be provided as top-level args to `bs_dependency()`: ",
      paste(bad_args, collapse = ", ")
    )
  }

  if ("package" %in% names(.dep_args)) {
    warning(
      "`package` won't have any effect since `src` must be an absolute path"
    )
  }

  script <- .dep_args[["script"]]
  if (length(script)) {
    if (basename(outfile) %in% basename(script)) {
      stop(
        "`script` file basename(s) must all be something other than ",
        basename(outfile)
      )
    }
    success <- file.copy(script, dirname(outfile), overwrite = TRUE)
    if (!all(success)) {
      stop(
        "Failed to copy the following script(s): ",
        paste(script[!success], collapse = ", "),
        ".\n\n",
        "Make sure script are absolute path(s)."
      )
    }
    .dep_args[["script"]] <- basename(script)
  }

  do.call(htmlDependency, c(dep_args, .dep_args))
}

#' @rdname bs_dependency
#' @param func a _non-anonymous_ function, with a _single_ argument.
#'   This function should accept a [bs_theme()] object and return a single
#'   [htmltools::htmlDependency()], a list of them, or `NULL`.
#' @param memoise whether or not to memoise (i.e., cache) `func` results for a
#'   short period of time. The default, `TRUE`, can have large performance
#'   benefits when many instances of the same themable widget are rendered. Note
#'   that you may want to avoid memoisation if `func` relies on side-effects
#'   (e.g., files on-disk) that need to change for each themable widget
#'   instance.
#'
#' @export
#'
#' @examplesIf rlang::is_interactive()
#' myWidgetVersion <- "1.2.3"
#'
#' myWidgetDependency <- function() {
#'   list(
#'     bs_dependency_defer(myWidgetCss),
#'     htmlDependency(
#'       name = "mywidget-js",
#'       version = myWidgetVersion,
#'       src = system.file(package = "mypackage", "js"),
#'       script = "mywidget.js"
#'     )
#'   )
#' }
#'
#' myWidgetCSS <- function(theme) {
#'   if (!is_bs_theme(theme)) {
#'     return(
#'       htmlDependency(
#'         name = "mywidget-css",
#'         version = myWidgetVersion,
#'         src = system.file(package = "mypackage", "css"),
#'         stylesheet = "mywidget.css"
#'       )
#'     )
#'   }
#'
#'   # Compile mywidget.scss using the variables and defaults from the theme
#'   # object.
#'   sass_input <- sass::sass_file(system.file(package = "mypackage", "scss/mywidget.scss"))
#'
#'   bs_dependency(
#'     input = sass_input,
#'     theme = theme,
#'     name = "mywidget",
#'     version = myWidgetVersion,
#'     cache_key_extra = utils::packageVersion("mypackage")
#'   )
#' }
#'
#' # Note that myWidgetDependency is not defined inside of myWidget. This is so
#' # that, if `myWidget()` is called multiple times, Shiny can tell that the
#' # function objects are identical and deduplicate them.
#' myWidget <- function(id) {
#'   div(
#'     id = id,
#'     span("myWidget"),
#'     myWidgetDependency()
#'   )
#' }
bs_dependency_defer <- function(func, memoise = TRUE) {
  # func() most likely calls stuff like sass_file() and bs_dependency() ->
  # sass_partial() -> sass() (e.g., see example section above) Even though
  # sass() calls can be cached, there is still considerable overhead involved
  # in computing the cache key itself because it involves file-level operations
  # (e.g. getting the mtimes of file attachments).
  # So, to help with performance (in the case where we repeatedly call the same
  # func with the same theme), memoise func with a short-lived cache
  if (memoise) {
    # The memoized function is created each time bs_dependency_defer is called,
    # and then it is used once. This is not how memoized functions are normally
    # used, but in this case it works because the caching object is re-used, and
    # it still provides very significant improvement in performance.
    mfunc <- memoise::memoise(func, cache = .dependency_cache)
  } else {
    mfunc <- func
  }

  tagFunction(function() {
    if (is_shiny_app()) {
      # If we're in a Shiny app, do two things:
      # (1) Register this function as a dependency so that Shiny will know to
      # update it later if the theme dynamically changes (i.e.,
      # session$setCurrentTheme() is called). Note that we *don't* provide
      # the memoised version since Shiny will de-duplicate identical()
      # functions anyway.
      # (2) Call the user's `func()` with the current theme, and return the
      # resulting htmlDependency so that it can be embedded in the static page.
      shiny::registerThemeDependency(func)

      return(mfunc(get_current_theme()))
    }

    # Outside of a Shiny context, we'll just get the global theme.
    mfunc(bs_global_get())
  })
}

as_bs_theme <- function(theme) {
  if (is_bs_theme(theme)) {
    return(theme)
  }

  # This is a historical artifact that should happen
  if (is_sass_bundle(theme) || inherits(theme, "sass_layer")) {
    stop(
      "`theme` cannot be a `sass_bundle()` or `sass_layer()` (use `bs_bundle()` to add a bundle)"
    )
  }

  # NULL means default Bootstrap
  if (is.null(theme)) {
    return(bs_theme())
  }

  # For example, `bs_theme_dependencies(theme = 4)`
  if (is.numeric(theme)) {
    return(bs_theme(version = theme))
  }

  # For example, `bs_theme_dependencies(theme = 'bootswatch@version')`
  if (is_string(theme)) {
    theme <- strsplit(theme, "@", fixed = TRUE)[[1]]
    if (length(theme) == 2) {
      return(bs_theme(version = theme[2], bootswatch = theme[1]))
    }
    # Also support `bs_theme_dependencies(version = '4')` and
    # `bs_theme_dependencies(theme = 'bootswatch')`
    if (length(theme) == 1) {
      if (theme %in% c(versions(), "4-3", "4+3")) {
        return(bs_theme(version = theme))
      } else {
        return(bs_theme(bootswatch = theme))
      }
    }
    stop("If `theme` is a string, it can't contain more than one @")
  }

  stop(
    "`theme` must be one of the following: (1) `NULL`, ",
    "(2) a `'bootswatch@version'` string, ",
    "(3) the result of `bs_global_get()`."
  )
}

library(shiny)
library(bslib)
library(dplyr)
library(lubridate)
library(plotly)
library(chiflights22)
library(histoslider)
library(rlang)

# Data prep
flights <- flights %>%
  left_join(
    airports %>%
      transmute(dest_name = paste0(name, " (", faa, ")"), dest = faa, end_lat = lat, end_lon = lon)
  ) %>%
  filter(!is.na(dest_name)) %>%
  left_join(
    airports %>% select(origin = faa, start_lat = lat, start_lon = lon)
  ) %>%
  left_join(airlines, by = "carrier") %>%
  rename(carrier_name = name) %>%
  select(-time_hour) %>%
  left_join(weather) %>%
  mutate(
    # The overwhelming majority of precipation is 0 so transform
    # with some help from MASS::boxcox() https://stats.stackexchange.com/a/1452/48604
    precip = scales::rescale(precip^-1.55/-.55),
    date = lubridate::ymd(paste(year, month, day, sep = "-"))
  )

SLIDER_HEIGHT <- 150
CHOICES <- list(
  origin = c(
    "Choose origin(s)" = "",
    "O'Hare" = "ORD",
    "Midway" = "MDW",
    "Rockford" = "RFD"
  ),
  dest_name = c(
    "Choose destination(s)" = "",
    sort(unique(flights$dest_name))
  ),
  carrier_name = c(
    "Choose carrier(s)" = "",
    unique(flights$carrier_name)
  )
)


sidebar_acc <- accordion(
  open = c("Origin", "Destination"),
  accordion_panel(
    "Flight Path",
    # See https://github.com/rstudio/fontawesome/issues/114
    icon = fontawesome::fa("plane-departure"),
    uiOutput("flight_path_reset"),
    selectizeInput(
      "origin", "Origin",
      choices = CHOICES$origin,
      multiple = TRUE,
      options = list(plugins = "remove_button", closeAfterSelect = TRUE)
    ),
    selectizeInput(
      "dest_name", "Destination",
      choices = CHOICES$dest_name,
      multiple = TRUE,
      options = list(plugins = "remove_button", closeAfterSelect = TRUE)
    ),
    selectizeInput(
      "carrier_name", "Carrier",
      choices = CHOICES$carrier_name,
      multiple = TRUE,
      options = list(plugins = "remove_button", closeAfterSelect = TRUE)
    )
  ),
  accordion_panel(
    "Flight time",
    icon = fontawesome::fa("clock"),
    input_histoslider(
      "sched_dep_time", "Departure time",
      flights$sched_dep_time,
      height = SLIDER_HEIGHT,
      options = list(handleLabelFormat = "0d")
    ),
    input_histoslider(
      "sched_arr_time", "Arrival time",
      flights$sched_arr_time,
      height = SLIDER_HEIGHT,
      options = list(handleLabelFormat = "0d")
    ),
    input_histoslider(
      "date", "Date",
      flights$date,
      height = SLIDER_HEIGHT,
      breaks = "months",
      options = list(handleLabelFormat = "%b %e")
    )
  ),
  accordion_panel(
    "Weather",
    icon = fontawesome::fa("cloud-rain"),
    # TODO: problematic (many NAs)
    #input_histoslider(
    #  "precip", "Precipitation",
    #  flights$precip,
    #  height = SLIDER_HEIGHT
    #),
    input_histoslider(
      "wind_speed", "Wind speed",
      flights$wind_speed,
      height = SLIDER_HEIGHT
    ),
    input_histoslider(
      "wind_gust", "Wind gust",
      flights$wind_gust,
      height = SLIDER_HEIGHT
    )
  )
)


flights_card <- card(
  full_screen = TRUE,
  card_header(
    "Flight paths",
    tooltip(
      bsicons::bs_icon("info-circle", title = "About marker areas"),
      "Marker areas are proportional to mean arrival delay"
    ),
    class = "d-flex justify-content-between align-items-center"
  ),
  plotlyOutput("flight_paths")
)

avg_delay_by_category <- card(
  full_screen = TRUE,
  card_header(
    "Average delay by category",
    popover(
      bsicons::bs_icon("gear", title = "Settings"),
      selectInput(
        "avg_delay_category", "Category",
        c("Carrier", "Month", "Weekday")
      ),
      radioButtons(
        "avg_delay_type", "Delay type",
        c("Arrival", "Departure"),
        inline = TRUE
      )
    ),
    class = "d-flex justify-content-between align-items-center"
  ),
  plotlyOutput("scatter_delay")
)

delay_dist <- navset_card_underline(
  title = "Distribution of delay times",
  full_screen = TRUE,
  id = "delay_dist_nav",
  sidebar = sidebar(
    position = "right",
    open = FALSE,
    radioButtons(
      "delay_dist_type", "Delay type",
      c("Arrival" = "arr_delay", "Departure" = "dep_delay"),
      inline = TRUE
    ),
    conditionalPanel(
      "input.delay_dist_nav === 'Overall'",
      selectizeInput(
        "delay_dist_category", "Split by",
        c("Choose a category" = "", "Carrier", "Month", "Weekday"),
        options = list(plugins = "remove_button")
      )
    )
  ),
  nav_panel(
    "Overall",
    plotlyOutput("delay_dist")
  ),
  nav_panel(
    "Over time",
    plotlyOutput("arr_delay_series")
  )
)

PRIMARY <- "#0675DD"

ui <- page_navbar(
  theme = bs_theme(
    preset = "shiny",
    "primary" = PRIMARY
  ),
  lang = "en",
  title = tags$span(
    tags$img(
      src = "logo.png",
      width = "46px",
      height = "auto",
      class = "me-3",
      alt = "Shiny hex logo"
    ),
    "Chicago Flights"
  ),
  sidebar = sidebar(width = 275, sidebar_acc),
  nav_spacer(),
  nav_panel(
    "Delay overview",
    class = "bslib-page-dashboard",
    uiOutput("value_boxes"),
    layout_columns(
      flights_card, avg_delay_by_category
    ),
    delay_dist
  ),
  nav_panel(
    "Data export",
    card(
      card_header("Flight data"),
      DT::dataTableOutput("export")
    )
  ),
  nav_item(
    tags$a(
      tags$span(
        bsicons::bs_icon("code-slash"), "Source code"
      ),
      href = "https://github.com/rstudio/bslib/tree/main/inst/examples/flights",
      target = "_blank"
    )
  ),
  nav_item(
    input_dark_mode(id = "dark_mode", mode = "light")
  )
)

server <- function(input, output, session) {
  # ---------------------------------------------------------
  # Flights tab logic
  #
  # WARNING: this server-side filtering logic is VERY experimental
  # at this point and won't be easily adapt to different use cases.
  # If you feel tempted to use it, use with caution.
  # ---------------------------------------------------------

  # Mapping from input id name to updating input function
  input_discrete_vars <- list(
    origin = updateSelectInput,
    dest_name = updateSelectInput,
    carrier_name = updateSelectInput
  )
  input_numeric_vars <- list(
    sched_dep_time = update_histoslider,
    sched_arr_time = update_histoslider,
    date = update_histoslider
  )
  input_vars <- c(
    input_discrete_vars,
    input_numeric_vars
  )

  filter_index <- function(d) {
    idx <- rep(TRUE, nrow(d))
    for (var in names(d)) {
      idx <- idx & filter_col(d, var)
    }
    idx & !is.na(idx)
  }

  filter_col <- function(d, var) {
    vals <- d[[var]]
    input_val <- input[[var]]
    if (is.null(input_val) || identical(input_val, "")) {
      return(TRUE)
    }

    if (is.character(vals) || is.factor(vals) || is.logical(vals)) {
      return(d[[var]] %in% input_val)
    }

    # N.B. between() will remove NAs, which we probably don't
    # want to remove until the slider is considered 'active'
    rng <- range(vals, na.rm = TRUE)
    active <- isTRUE(rng[1] <= input_val[1] || input_val[2] <= rng[2])

    if (!active) {
      return(TRUE)
    }

    dplyr::between(vals, input_val[1], input_val[2])
  }

  # Set up a listener for each input that effectively says update
  # every other input when my value changes
  lapply(names(input_vars), function(var) {
    # We don't want updates to other variables to then
    # cause an update to this variable
    do_update <- reactiveVal(TRUE)

    observeEvent(input[[var]], ignoreInit = TRUE, ignoreNULL = FALSE, {

      if (!do_update()) return()
      do_update(FALSE)
      on.exit(do_update(TRUE), add = TRUE)

      d <- flights[filter_index(flights), ]
      if (nrow(d) == 0) return()

      other_vars <- setdiff(names(input_vars), var)
      lapply(other_vars, function(v) {

        input_val <- input[[v]]
        update_input_func <- input_vars[[v]]

        if (v %in% names(input_discrete_vars)) {
          choices <- CHOICES[[v]] %||% sort(unique(d[[v]]))
          selected <- input_val %||% CHOICES[[v]][CHOICES[[v]] == ""]

          update_input_func(
            inputId = v, choices = choices,
            selected = selected
          )
        } else {
          update_input_func(
            id = v,
            values = d[[v]],
            start = input_val[1],
            end = input_val[2]
          )
        }
      })
    })
  })


  output$flight_path_reset <- renderUI({
    req(c(input$origin, input$dest_name, input$carrier_name))

    actionLink(
      "flight_path_reset", "Reset",
      style = htmltools::css(
        position = "absolute",
        right = "1rem",
        text_decoration = "none",
        font_weight = 700,
        font_size = ".875rem"
      )
    )
  })

  observeEvent(input$flight_path_reset, {
    updateSelectInput(
      inputId = "origin",
      choices = CHOICES$origin
    )
    updateSelectInput(
      inputId = "dest_name",
      choices = CHOICES$dest_name
    )
    updateSelectInput(
      inputId = "carrier_name",
      choices = CHOICES$carrier_name
    )
  })

  # Flights with all filters applied (i.e., data used for value boxes/plots)
  flight_dat <- reactive({
    flights[filter_index(flights), ]
  })

  summary_vals <- reactive({
    d <- flight_dat()

    list(
      n = scales::comma(nrow(d)),
      n_dest = length(unique(d$dest_name)),
      n_carriers = length(unique(d$carrier_name)),
      dep_delay = round(mean(d$dep_delay, na.rm = T), 0),
      dep_delay_perc = round(100 * sum(d$dep_delay > 0, na.rm = T) / nrow(d), 1),
      arr_delay = round(mean(d$arr_delay, na.rm = T), 0),
      arr_delay_perc = round(100 * sum(d$arr_delay > 0, na.rm = TRUE) / nrow(d), 1)
    )
  })

  output$value_boxes <- renderUI({
    vals <- summary_vals()
    n_flights <- value_box(
      "A TOTAL OF",
      paste(vals$n, "flights"),
      paste("Across", vals$n_dest, "destinations"),
      tags$p(paste(
        "On", vals$n_carriers, "different carriers"
      )),
      showcase = bsicons::bs_icon("airplane")
    )

    late <- if (vals$dep_delay > 0) "late" else "early"
    delay_dep <- value_box(
      "AVERAGE DEPARTURE",
      paste(vals$dep_delay, "mins", late),
      paste0(vals$dep_delay_perc, "% of flights depart ", late),
      showcase = bsicons::bs_icon("hourglass-split")
    )

    late <- if (vals$arr_delay > 0) "late" else "early"
    delay_arr <- value_box(
      "AVERAGE ARRIVAL",
      paste(vals$arr_delay, "mins", late),
      paste0(vals$arr_delay_perc, "% of flights arrive ", late),
      showcase = bsicons::bs_icon("hourglass-bottom")
    )

    layout_columns(class = "mb-0", n_flights, delay_dep, delay_arr)
  }) %>%
    bindCache(flight_dat())

  plotly_base <- function(..., geo = FALSE, color = I(PRIMARY)) {
    plot_func <- if (geo) plot_geo else plot_ly
    plot_func(..., color = color) %>%
      plotly::config(displayModeBar = FALSE) %>%
      plotly::layout(
        font = list(
          family = "Open Sans",
          color = if (input$dark_mode == "dark") "white" else "#1D1F21"
        ),
        plot_bgcolor = "transparent",
        paper_bgcolor = "transparent"
      )
  }

  output$flight_paths <- renderPlotly({
    flight_dat() %>%
      group_by(start_lon, start_lat, end_lon, end_lat, origin, dest) %>%
      summarise(mean_delay = mean(arr_delay, na.rm = TRUE)) %>%
      plotly_base(geo = TRUE, showlegend = FALSE) %>%
      add_segments(
        x = ~start_lon, xend = ~end_lon,
        y = ~start_lat, yend = ~end_lat,
        alpha = 0.5, size = I(1), hoverinfo = "none"
      ) %>%
      add_markers(
        x = ~end_lon, y = ~end_lat, size = ~mean_delay,
        hoverinfo = "text", alpha = 0.1,
        text = ~paste0(
          origin, " -> ", dest, "<br>",
          "Average delay: ", round(mean_delay, 1)
        )
      ) %>%
      layout(
        geo = list(
          bgcolor = "transparent",
          projection = list(
            type = 'orthographic',
            rotation = list(lon = -100, lat = 40, roll = 0)
          ),
          showland = TRUE,
          landcolor = toRGB("gray95"),
          countrycolor = toRGB("gray80")
        )
      )
  }) %>%
    bindCache(flight_dat(), input$dark_mode)

  output$scatter_delay <- renderPlotly({
    d <- flight_dat()
    req(nrow(d) > 0)

    d <- switch(
      input$avg_delay_category,
      Weekday = group_by(d, y = lubridate::wday(date, label = TRUE)),
      Month = group_by(d, y = lubridate::month(date, label = TRUE)),
      Carrier = group_by(d, y = carrier_name),
      stop("Category of ", input$avg_delay_category, "not implemented")
    )

    d <- switch(
      input$avg_delay_type,
      Arrival = summarise(d, avg = mean(arr_delay, na.rm = TRUE)),
      Departure = summarise(d, avg = mean(dep_delay, na.rm = TRUE)),
    )

    d %>%
      arrange(avg) %>%
      mutate(y = factor(y, levels = y)) %>%
      plotly_base(x = ~avg, y = ~y) %>%
      add_bars(hoverinfo = "x") %>%
      layout(
        yaxis = list(title = ""),
        xaxis = list(
          title = paste("Average", tolower(input$avg_delay_type), "delay"),
          hoverformat = ".1f",
          gridcolor = if (input$dark_mode == "dark") "#303030"
        )
      )
  }) %>%
    bindCache(flight_dat(), input$dark_mode, input$avg_delay_category, input$avg_delay_type)


  output$delay_dist <- renderPlotly({
    d <- flight_dat()
    x <- d[[input$delay_dist_type]]
    req(length(x) > 0)

    x_mean <- mean(x, na.rm = TRUE)
    end <- quantile(x, probs = 0.99, na.rm = TRUE)

    color <- switch(
      input$delay_dist_category,
      Carrier = d$carrier_name,
      Month = lubridate::month(d$date, label = TRUE),
      Weekday = lubridate::wday(d$date, label = TRUE),
      I(PRIMARY)
    )

    plotly_base(
      x = x, color = color,
      hovertemplate = "%{y} flights were<br>%{x} min late <extra></extra>"
    ) %>%
      rangeslider(start = min(x, na.rm = TRUE), end = as.numeric(end)) %>%
      add_annotations(
        text = paste(
          "Average",
          switch(
            input$delay_dist_type,
            arr_delay = "arrival",
            dep_delay = "departure"
          ),
          "<br> delay of",
          round(x_mean, 1), "min"
        ),
        x = x_mean, y = 0.5, yref = "paper",
        ax = 80, ay = -50,
        font = list(size = 14)
      ) %>%
      layout(
        barmode = "stack",
        yaxis = list(gridcolor = if (input$dark_mode == "dark") "#303030"),
        shapes = list(
          type = "line",
          x0 = x_mean, x1 = x_mean,
          y0 = 0, y1 = 1,
          yref = "paper",
          line = list(color = "lightgray", dash = "dash")
        )
      )
  }) %>%
    bindCache(flight_dat(), input$dark_mode, input$delay_dist_type, input$delay_dist_category)

  output$arr_delay_series <- renderPlotly({
    d <- flight_dat()
    req(nrow(d) > 0)

    d <- group_by(d, date)

    d <- switch(
      input$delay_dist_type,
      arr_delay = summarise(d, y = mean(arr_delay, na.rm = TRUE)),
      dep_delay = summarise(d, y = mean(dep_delay, na.rm = TRUE))
    )

    color <- switch(
      input$delay_dist_category,
      Carrier = d$carrier_name,
      Month = lubridate::month(d$date, label = TRUE),
      Weekday = lubridate::wday(d$date, label = TRUE),
      I(PRIMARY)
    )

    plotly_base(
      x = d$date, y = d$y, color = color,
      hovertemplate = "%{y:.1f}<extra></extra>",
    ) %>%
      add_lines() %>%
      layout(
        hovermode = "x",
        xaxis = list(title = "", tickformat = "%b %e"),
        yaxis = list(title = "Average delay", showgrid = FALSE)
      )
  }) %>%
    bindCache(flight_dat(), input$dark_mode, input$delay_dist_type, input$delay_dist_category)

  output$export <- DT::renderDataTable({
    DT::datatable(flight_dat(), fillContainer = TRUE)
  })

}

shinyApp(ui, server)

#' Dark mode input control
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Creates a button that toggles between dark and light modes, specifically for
#' toggling between light and dark [Bootstrap color
#' modes](https://getbootstrap.com/docs/5.3/customize/color-modes/) -- a new
#' feature added in [Bootstrap
#' 5.3](https://getbootstrap.com/docs/5.3/migration/#color-modes).
#'
#' @param ... Additional attributes to be passed to the input control UI, such
#'   as `class`, `style`, etc.
#'
#'   In `toggle_dark_mode()`, the `...` are included for future extensibility
#'   and are currently ignored.
#' @param id An optional input id, required to reactively read the current color
#'   mode.
#' @param mode The initial mode of the dark mode switch. By default or when set
#'   to `NULL`, the user's system settings for preferred color scheme will be
#'   used. Otherwise, set to `"light"` or `"dark"` to force a particular initial
#'   mode.
#'
#' @return Returns a UI element for a dark mode switch input control. The server
#'   value received for the input corresponding to `id` will be a string value
#'   with the current color mode (`"light"` or `"dark"`).
#'
#' @describeIn input_dark_mode Create a dark mode switch input control
#' @family input controls
#' @export
input_dark_mode <- function(..., id = NULL, mode = NULL) {
  if (!is.null(mode)) {
    mode <- rlang::arg_match(mode, c("light", "dark"))
  }

  if (any(!nzchar(rlang::names2(rlang::list2(...))))) {
    abort("All arguments in `...` must be named.")
  }

  res <- web_component(
    "bslib-input-dark-mode",
    id = id,
    attribute = "data-bs-theme",
    mode = mode,
    style = css(
      "--text-1" = "var(--bs-emphasis-color)",
      "--text-2" = "var(--bs-tertiary-color)",
      # The vertical correction used in the dark mode component isn't quite
      # right on Bootstrap pages. This next line overrides it and removes the
      # vertical correction. But users can still set the CSS property manually
      # in their a `style` argument passed in via `...` if they want.
      "--vertical-correction" = " "
    ),
    ...
  )

  res <- tag_require(res, version = 5, caller = "input_dark_mode()")
  as_fragment(res)
}

#' @describeIn input_dark_mode Programmatically toggle or set the current light
#'   or dark color mode.
#' @param session A Shiny session object (the default should almost always be
#'   used).
#' @export
toggle_dark_mode <- function(mode = NULL, ..., session = get_current_session()) {
  rlang::check_dots_empty(
    error = function(err) rlang::warn(rlang::cnd_message(err))
  )

  if (!is.null(mode)) {
    mode <- tryCatch(
      rlang::arg_match(mode, c("light", "dark")),
      error = function(err) {
        rlang::warn(rlang::cnd_message(err))
        mode
      }
    )
  }

  data <- dropNulls(list(method = "toggle", value = mode))

  # We're using sendCustomMessage() here because dark mode is set globally and
  # `id` is not required for `toggle_dark_mode()`. `$sendInputMessage()` would
  # require us to know the `id` of at least one input control.
  callback <- function() {
    session$sendCustomMessage("bslib.toggle-dark-mode", data)
  }

  session$onFlush(callback, once = TRUE)
}

if (!identical(getwd(), rprojroot::find_package_root_file())) {
  stop("This script must be run from the top directory of the bslib package")
}

pkgload::load_all()

# TODO: save a CSS file for every version >= 5?
dep <- component_dependency_sass_(bs_theme())
css <- file.path(dep$src, dep$stylesheet)
target_dir <- path_inst("components", "dist")
target_css <- file.path(target_dir, "components.css")
file.copy(css, target_css, overwrite = TRUE)

#' Create a set of navbar options
#'
#' A `navbar_options()` object captures options specific to the appearance and
#' behavior of the navbar, independent from the content displayed on the page.
#' This helper should be used to create the list of options expected by
#' `navbar_options` in [page_navbar()] and [navset_bar()].
#'
#' ## Navbar style with Bootstrap 5 and Bootswatch themes
#'
#' In \pkg{bslib} v0.9.0, the default navbar colors for Bootswatch themes with
#' Bootstrap 5 changed. Prior to v0.9.0, bslib pre-selected navbar background
#' colors in light and dark mode; after v0.9.0 the default navbar colors are
#' less opinionated by default and follow light or dark mode (see
#' [input_dark_mode()]).
#'
#' You can use `navbar_options()` to adjust the colors of the navbar when using
#' a Bootswatch preset theme with Bootstrap 5. For example, the [Bootswatch
#' documentation for the Flatly theme](https://bootswatch.com/flatly/) shows
#' 4 navbar variations. Inspecting the source code for the first example reveals
#' the following markup:
#'
#' ```html
#' <nav class="navbar navbar-expand-lg bg-primary" data-bs-theme="dark">
#'   <!-- all of the navbar html -->
#' </nav>
#' ```
#'
#' Note that this navbar uses the `bg-primary` class for a dark navy background.
#' The navbar's white text is controlled by the `data-bs-theme="dark"`
#' attribute, which is used by Bootstrap for light text on a _dark_ background.
#' In \pkg{bslib}, you can achieve this look with:
#'
#' ```r
#' ui <- page_navbar(
#'   theme = bs_theme(5, "flatly"),
#'   navbar_options = navbar_options(class = "bg-primary", theme = "dark")
#' )
#' ```
#'
#' This particular combination of `class = "bg-primary"` and `theme = "dark"`
#' works well for most Bootswatch presets.
#'
#' Another variation from the Flatly documentation features a navar with dark
#' text on a light background:
#'
#' ```r
#' ui <- page_navbar(
#'   theme = bs_theme(5, "flatly"),
#'   navbar_options = navbar_options(class = "bg-light", theme = "light")
#' )
#' ```
#'
#' The above options set navbar foreground and background colors that are always
#' the same in both light and dark modes. To customize the navbar colors used in
#' light or dark mode, you can use the `$navbar-light-bg` and `$navbar-dark-bg`
#' Sass variables. When provided, bslib will automatically choose to use
#' light or dark text as the foreground color.
#'
#' ```r
#' ui <- page_navbar(
#'   theme = bs_theme(
#'     5,
#'     preset = "flatly",
#'     navbar_light_bg = "#18BC9C", # flatly's success color (teal)
#'     navbar_dark_bg = "#2C3E50"   # flatly's primary color (navy)
#'   )
#' )
#' ```
#'
#' Finally, you can also use the `$navbar-bg` Sass variable to set the navbar
#' background color for both light and dark modes:
#'
#' ```r
#' ui <- page_navbar(
#'   theme = bs_theme(
#'     5,
#'     preset = "flatly",
#'     navbar_bg = "#E74C3C" # flatly's danger color (red)
#'   )
#' )
#' ```
#'
#' @section Changelog:
#'
#' This function was introduced in \pkg{bslib} v0.9.0, replacing the `position`,
#' `bg`, `inverse`, `collapsible` and `underline` arguments of [page_navbar()]
#' and [navset_bar()]. Those arguments are deprecated with a warning and will be
#' removed in a future version of \pkg{bslib}. Note that the deprecated
#' `inverse` argument of [page_navbar()] and [navset_bar()] was replaced with
#' the `theme` argument of `navbar_options()`.
#'
#' @examples
#' navbar_options(position = "static-top", bg = "#2e9f7d", underline = FALSE)
#'
#' @inheritParams shiny::navbarPage
#' @param bg a CSS color to use for the navbar's background color.
#' @param theme Either `"dark"` for a light text color (on a **dark**
#'   background) or `"light"` for a dark text color (on a **light** background).
#'   If `"auto"` (the default) and `bg` is provided, the best contrast to `bg`
#'   is chosen.
#' @param underline Whether or not to add underline styling to page or navbar
#'   links when active or focused.
#' @param ... Additional attributes that will be passed directly to the navbar
#'   container element.
#'
#' @returns Returns a list of navbar options.
#'
#' @export
navbar_options <- function(
  ...,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  bg = NULL,
  theme = c("auto", "light", "dark"),
  collapsible = TRUE,
  underline = TRUE
) {
  # Track user-provided arguments for print method and deprecation warnings
  is_default <- list(
    position = missing(position),
    bg = missing(bg),
    theme = missing(theme),
    collapsible = missing(collapsible),
    underline = missing(underline)
  )

  opts <- list(
    position = rlang::arg_match(position),
    bg = bg,
    theme = rlang::arg_match(theme),
    collapsible = collapsible,
    underline = underline
  )

  dots <- separate_arguments(...)
  if (length(dots$children) > 0) {
    abort(
      "All arguments in `...` must be named attributes to be applied to the navbar container."
    )
  }

  if ("inverse" %in% names(dots$attribs)) {
    # Catch muscle-memory for using `inverse`. We didn't officially release
    # `navbar_options()` with an `inverse` argument, but it's reasonable people
    # might try to use it and it did exist briefly in dev versions.
    lifecycle::deprecate_soft(
      when = "0.9.0",
      what = "navbar_options(inverse=)",
      with = "navbar_options(theme=)"
    )
  }
  if (length(dots$attribs)) {
    opts$attribs <- dots$attribs
  }

  structure(
    opts,
    class = c("bslib_navbar_options", "list"),
    is_default = is_default,
    waldo_opts = list(ignore_attr = TRUE)
  )
}

navbar_options_resolve_deprecated <- function(
  options_user = list(),
  position = deprecated(),
  bg = deprecated(),
  inverse = deprecated(),
  collapsible = deprecated(),
  underline = deprecated(),
  .fn_caller = "navset_bar",
  .warn_deprecated = TRUE
) {
  options_old <- list(
    position = if (lifecycle::is_present(position)) position,
    bg = if (lifecycle::is_present(bg)) bg,
    inverse = if (lifecycle::is_present(inverse)) inverse,
    collapsible = if (lifecycle::is_present(collapsible)) collapsible,
    underline = if (lifecycle::is_present(underline)) underline
  )
  options_old <- dropNulls(options_old)

  args_deprecated <- names(options_old)

  if (.warn_deprecated && length(args_deprecated)) {
    # TODO-deprecated: (2024-12) Elevate deprecation to an error
    lifecycle::deprecate_warn(
      "0.9.0",
      I(
        sprintf(
          "The arguments of `%s()` for navbar options (including %s) have been consolidated into a single `navbar_options` argument and ",
          .fn_caller,
          paste(sprintf("`%s`", args_deprecated), collapse = ", ")
        )
      ),
      details = c(
        "i" = "See `navbar_options()` for more details.",
        "!" = if ("inverse" %in% args_deprecated) {
          "Use `theme` instead of `inverse` in `navbar_options()`."
        }
      )
    )
  }

  # Upgrade `inverse` to the new `theme` argument of `navbar_options()`
  if ("inverse" %in% names(options_old)) {
    inverse <- options_old[["inverse"]]
    options_old[["inverse"]] <- NULL

    options_old[["theme"]] <-
      if (is.character(inverse)) {
        inverse
      } else if (isTRUE(as.logical(inverse))) {
        options_old[["theme"]] <- "dark"
      } else if (isFALSE(as.logical(inverse))) {
        options_old[["theme"]] <- "light"
      } else {
        abort(paste("Invalid `inverse` value: ", inverse))
      }
  }

  # Consolidate `navbar_options` (options_user) with the deprecated direct
  # options. We take the direct option if the user option is a default value,
  # warning if otherwise ignored.
  # TODO-deprecated: Remove this and warning when direct options are hard-deprecated
  is_default <- attr(options_user, "is_default") %||% list()
  keep_user_values <- vapply(
    names(options_user),
    function(x) !isTRUE(is_default[[x]]),
    logical(1)
  )
  options_user <- options_user[keep_user_values]

  ignored <- c()
  for (opt in names(options_old)) {
    if (!opt %in% names(options_user)) {
      options_user[[opt]] <- options_old[[opt]]
    } else if (!identical(options_old[[opt]], options_user[[opt]])) {
      ignored <- c(ignored, if (opt == "theme") "inverse" else opt)
    }
  }

  if (length(ignored) > 0) {
    rlang::warn(
      c(
        sprintf(
          "`%s` %s provided twice: once directly and once in `navbar_options`.",
          paste(ignored, collapse = "`, `"),
          if (length(ignored) == 1) "was" else "were"
        ),
        "The deprecated direct option(s) will be ignored and the values from `navbar_options` will be used."
      ),
      call = rlang::caller_call()
    )
  }

  attribs <- options_user[["attribs"]] %||% list()
  options_user$attribs <- NULL

  rlang::exec(navbar_options, !!!options_user, !!!attribs)
}

#' @export
print.bslib_navbar_options <- function(x, ...) {
  cat("<bslib_navbar_options>\n")

  if (length(x) == 0) {
    return(invisible(x))
  }

  fields <- names(x)
  opt_w <- max(nchar(fields))
  is_default <- attr(x, "is_default") %||% list()
  for (opt in fields) {
    value <- x[[opt]] %||% "NULL"
    if (inherits(value, "list")) {
      value <- paste(names(value), collapse = ", ")
    }
    if (isTRUE(is_default[[opt]])) {
      if (identical(value, "NULL")) {
        # Skip printing default NULL values
        next
      }
      value <- sprintf("(%s)", value)
    }
    cat(sprintf("%*s", opt_w, opt), ": ", value, "\n", sep = "")
  }

  invisible(x)
}

#' Value box
#'
#' An opinionated ([card()]-powered) box, designed for displaying a `value` and
#' `title`. Optionally, a `showcase` can provide for context for what the
#' `value` represents (for example, it could hold a [bsicons::bs_icon()], or
#' even a [shiny::plotOutput()]).
#'
#' @param title,value A string, number, or [htmltools::tag()] child to display as
#'   the title or value of the value box. The `title` appears above the `value`.
#' @param ... Unnamed arguments may be any [htmltools::tag()] children to
#'   display below `value`. Named arguments become attributes on the containing
#'   element.
#' @param showcase A [htmltools::tag()] child to showcase (e.g., a
#'   [bsicons::bs_icon()], a `plotly::plotlyOutput()`, etc).
#' @param showcase_layout either `showcase_left_center()` or
#'   `showcase_top_right()`.
#' @param theme_color A theme color to use for the background color. Should
#'   match a name in the Bootstrap Sass variable `$theme-colors` (e.g.,
#'   `"secondary"`, `"success"`, `"danger"`, etc).
#' @param class Utility classes for customizing the appearance of the summary
#'   card. Use `bg-*` and `text-*` classes (e.g, `"bg-danger"` and
#'   `"text-light"`) to customize the background/foreground colors.
#' @param fill Whether to allow the value box to grow/shrink to fit a fillable
#'   container with an opinionated height (e.g., `page_fillable()`).
#' @inheritParams card
#' @export
#' @seealso [card()]
#' @examples
#'
#' library(htmltools)
#'
#' if (interactive()) {
#'   value_box(
#'     "KPI Title",
#'     h1(HTML("$1 <i>Billion</i> Dollars")),
#'     span(
#'       bsicons::bs_icon("arrow-up"),
#'       " 30% VS PREVIOUS 30 DAYS"
#'     ),
#'     showcase = bsicons::bs_icon("piggy-bank"),
#'     class = "bg-success"
#'   )
#' }
value_box <- function(title, value, ..., showcase = NULL, showcase_layout = showcase_left_center(), full_screen = FALSE, theme_color = "primary", height = NULL, max_height = NULL, fill = TRUE, class = NULL) {

  args <- rlang::list2(...)
  argnames <- rlang::names2(args)

  attribs <- args[nzchar(argnames)]
  children <- args[!nzchar(argnames)]

  if (rlang::is_bare_character(title) || rlang::is_bare_numeric(title)) {
    title <- tags$p(title)
  }
  if (rlang::is_bare_character(value) || rlang::is_bare_numeric(value)) {
    value <- tags$p(value)
  }

  contents <- div(class = "value-box-area", title, value, !!!children)
  contents <- bindFillRole(contents, container = TRUE, item = TRUE)

  if (!is.null(showcase)) {
    contents <- showcase_layout(showcase, contents)
  }

  res <- card(
    class = c("bslib-value-box border-0", paste0("bg-", theme_color), class),
    full_screen = full_screen,
    height = height,
    max_height = max_height,
    fill = fill,
    !!!attribs,
    contents,
    as.card_item(value_box_dependency())
  )

  as_fragment(tag_require(res, version = 5, caller = "value_box()"))
}

value_box_dependency <- function() {
  bs_dependency_defer(value_box_dependency_sass)
}

value_box_dependency_sass <- function(theme) {
  component_dependency_sass(theme, "value_box")
}

#' @param width one of the following:
#'   * A proportion (i.e., a number between 0 and 1) of available width to
#'     allocate to the showcase.
#'   * A vector of length 2 valid [CSS unit][htmltools::validateCssUnit] defining
#'     the width of each column (for `showcase_left_center()` the 1st unit defines
#'     the showcase width and for `showcase_top_right` the 2nd unit defines the
#'     showcase width). Note that any units supported by the CSS grid
#'     `grid-template-columns` property may be used (e.g., `fr` units).
#' @param max_height,max_height_full_screen A proportion (i.e., a number between
#'   0 and 1) or any valid [CSS unit][htmltools::validateCssUnit] defining the
#'   showcase max_height.
#'
#' @export
#' @rdname value_box
showcase_left_center <- function(width = 0.3, max_height = "100px", max_height_full_screen = 0.67) {
  showcase_layout_(width, max_height, max_height_full_screen, top_right = FALSE)
}

#' @export
#' @rdname value_box
showcase_top_right <- function(width = 0.3, max_height = "75px", max_height_full_screen = 0.67) {
  if (is_01_scalar(width)) {
    width <- 1 - width
  }
  showcase_layout_(width, max_height, max_height_full_screen, top_right = TRUE)
}


showcase_layout_ <- function(width, max_height, max_height_full_screen, top_right) {

  width <- validate_width_unit(width)
  max_height <- validate_height_unit(max_height)
  max_height_full_screen <- validate_height_unit(max_height_full_screen)

  function(showcase, contents) {

    showcase_container <- div(
      class = "value-box-showcase",
      class = "overflow-hidden",
      class = if (top_right) "showcase-top-right",
      style = css(
        "--bslib-value-box-max-height" = max_height,
        "--bslib-value-box-max-height-full-screen" = max_height_full_screen
      ),
      showcase
    )

    showcase_container <- bindFillRole(showcase_container, container = TRUE, item = TRUE)

    if (!top_right) {
      contents <- tagAppendAttributes(contents, class = "border-start")
    }

    items <- list(showcase_container, contents)
    width_fs <- c("1fr", "auto")
    if (top_right) {
      items <- rev(items)
      width_fs <- rev(width_fs)
    }

    card_body(
      style = css(padding = 0),
      layout_column_wrap(
        width = NULL, gap = 0,
        heights_equal = "row",
        class = "value-box-grid",
        style = css(
          "--bslib-value-box-widths" = width,
          "--bslib-value-box-widths-full-screen" = width_fs
        ),
        !!!items
      )
    )
  }
}


# It seems to be to use % over fr here since there is no gap on the grid
validate_width_unit <- function(x) {
  if (!is_01_scalar(x)) {
    # TODO: validateCssUnit() should maybe support fr units?
    return(paste(x, collapse = " "))
  }
  paste(paste0(c(100 * x, 100 * (1 - x)), "%"), collapse = " ")
}

validate_height_unit <- function(x) {
  if (!is_01_scalar(x)) {
    return(validateCssUnit(x))
  }
  paste0(100 * x, "%")
}

is_01_scalar <- function(x) {
  rlang::is_scalar_double(x) && x >= 0 && x <= 1
}

library(shiny)
library(bslib)

ui <- page_fillable(
  theme = bs_theme(version = 5),
  title = "Toast Notifications Demo",
  padding = 0,
  gap = 0,

  h2("Toast Notifications Demo", class = "p-3 border-bottom mb-0"),
  input_dark_mode(class = "position-absolute top-0 end-0 p-3"),

  layout_column_wrap(
    width = 1 / 2,
    class = "bslib-page-dashboard",
    style = css(
      background = "var(--bslib-dashboard-main-bg)",
      padding = "15px",
      gap = "15px"
    ),

    # Toast Builder Card
    card(
      card_header("Toast Builder"),
      card_body(
        # Body content
        textAreaInput(
          "body",
          "Body Content",
          value = "This is a toast notification!",
          rows = 3,
          width = "100%"
        ),

        layout_columns(
          div(
            # Type
            selectInput(
              "type",
              "Type (Background Color)",
              choices = c(
                "None (default)" = "",
                "Primary" = "primary",
                "Secondary" = "secondary",
                "Success" = "success",
                "Info" = "info",
                "Warning" = "warning",
                "Danger" = "danger",
                "Light" = "light",
                "Dark" = "dark"
              ),
              selected = ""
            ),

            # Position
            selectInput(
              "position",
              "Position",
              choices = c(
                "Top Left" = "top-left",
                "Top Center" = "top-center",
                "Top Right" = "top-right",
                "Middle Left" = "middle-left",
                "Middle Center" = "middle-center",
                "Middle Right" = "middle-right",
                "Bottom Left" = "bottom-left",
                "Bottom Center" = "bottom-center",
                "Bottom Right" = "bottom-right"
              ),
              selected = "top-right"
            ),

            # Auto-hide options
            sliderInput(
              "duration_s",
              "Duration (seconds, 0 = disabled)",
              min = 0,
              max = 25,
              value = 5,
              step = 1,
              ticks = FALSE
            ),

            # Close button
            input_switch("closable", "Show Close Button", value = TRUE),

            textInput(
              "custom_id",
              "Toast ID",
              placeholder = "Automatically generated"
            )
          ),
          div(
            # Header options
            input_switch("use_header", "Include Header", value = FALSE),
            conditionalPanel(
              "input.use_header",
              textInput("header_title", "Header Title", value = "Notification"),
              selectInput(
                "header_icon",
                "Icon",
                choices = c(
                  "None" = "",
                  "Check" = "check-circle-fill",
                  "Info" = "info-circle-fill",
                  "Warning" = "exclamation-triangle-fill",
                  "Error" = "x-circle-fill",
                  "Star" = "star-fill",
                  "Heart" = "heart-fill",
                  "Bell" = "bell-fill",
                  "User" = "person-fill",
                  "Cog" = "gear-fill"
                ),
                selected = ""
              ),
              textInput(
                "header_status",
                "Custom Status Text",
                placeholder = "'Just now', '2 mins ago'"
              )
            ),
            conditionalPanel(
              "!input.use_header",
              selectInput(
                "icon_body",
                "Icon",
                # A set of bootstrap icon names for the notification body
                choices = c(
                  "None" = "",
                  "Check" = "check-circle-fill",
                  "Info" = "info-circle-fill",
                  "Warning" = "exclamation-triangle-fill",
                  "Error" = "x-circle-fill",
                  "Star" = "star-fill",
                  "Heart" = "heart-fill",
                  "Bell" = "bell-fill",
                  "User" = "person-fill",
                  "Settings" = "gear-fill",
                  "Chat" = "chat-dots-fill",
                  "Envelope" = "envelope-fill",
                  "Lightbulb" = "lightbulb-fill",
                  "Rocket" = "rocket-takeoff-fill",
                  "Shield" = "shield-fill-check",
                  "Thumbs Up" = "hand-thumbs-up-fill",
                  "Download" = "download",
                  "Upload" = "upload",
                  "Calendar" = "calendar-fill",
                  "Clock" = "clock-fill",
                  "Fire" = "fire",
                  "Gift" = "gift-fill",
                  "Trophy" = "trophy-fill",
                  "Flag" = "flag-fill",
                  "Pin" = "pin-fill"
                )
              )
            )
          )
        ),
      ),
      card_footer(
        class = "hstack gap-2 justify-content-end",
        actionButton("show_toast", "Show Toast", class = "btn-primary"),
        actionButton("hide_toast", "Hide Last Toast", class = "btn-secondary")
      )
    ),

    layout_column_wrap(
      width = 1,

      # Advanced Features and Examples
      card(
        card_header("Advanced Features"),
        card_body(
          actionButton(
            "show_persistent",
            "Show Persistent Toast",
            class = "mb-2 w-100"
          ),
          actionButton(
            "hide_persistent",
            "Hide Persistent Toast",
            class = "mb-2 w-100"
          ),
          actionButton(
            "show_long_duration",
            "Long Duration (10s)",
            class = "mb-2 w-100"
          ),
          actionButton(
            "show_no_close",
            "No Close Button",
            class = "mb-2 w-100"
          ),
          actionButton(
            "show_custom_header",
            "Custom Header with Icon & Status",
            class = "mb-2 w-100"
          )
        )
      ),

      card(
        card_header("Interactive Toasts"),
        card_body(
          actionButton(
            "show_action_buttons",
            "Toast with Action Buttons",
            class = "mb-2 w-100"
          ),
          actionButton(
            "show_multiple",
            "Show Multiple Toasts",
            class = "mb-2 w-100"
          ),
          actionButton(
            "show_all_positions",
            "Test All Positions",
            class = "mb-2 w-100"
          ),
          actionButton(
            "show_dynamic_content",
            "Toast with Dynamic Content",
            class = "mb-2 w-100"
          )
        )
      )
    )
  )
)

server <- function(input, output, session) {
  # Store last toast ID
  last_toast_id <- reactiveVal(NULL)
  persistent_toast_id <- reactiveVal(NULL)

  # Show toast from builder
  observeEvent(input$show_toast, {
    # Build header if needed
    header <- NULL
    if (input$use_header) {
      header <- toast_header(
        title = input$header_title,
        icon = if (nzchar(input$header_icon)) {
          bsicons::bs_icon(input$header_icon)
        },
        status = if (nzchar(input$header_status)) input$header_status
      )
    }

    body_icon <- if (!input$use_header && nzchar(input$icon_body)) {
      bsicons::bs_icon(input$icon_body)
    }

    # Build toast
    toast_obj <- toast(
      input$body,
      header = header,
      icon = body_icon,
      id = if (nzchar(input$custom_id)) input$custom_id,
      type = if (nzchar(input$type)) input$type,
      duration_s = input$duration_s,
      position = input$position,
      closable = input$closable
    )

    # Show and store ID
    id <- show_toast(toast_obj)
    last_toast_id(id)
  })

  # Hide last toast
  observeEvent(input$hide_toast, {
    req(last_toast_id())
    hide_toast(last_toast_id())
    last_toast_id(NULL)
  })

  # Advanced features
  observeEvent(input$show_persistent, {
    id <- show_toast(
      toast(
        "This toast won't disappear automatically. Use the 'Hide' button to dismiss it.",
        header = "Persistent Toast",
        type = "info",
        duration_s = 0
      )
    )
    persistent_toast_id(id)
  })

  observeEvent(input$hide_persistent, {
    req(persistent_toast_id())
    hide_toast(persistent_toast_id())
    persistent_toast_id(NULL)
  })

  observeEvent(input$show_long_duration, {
    show_toast(
      toast(
        "This toast will stay visible for 10 seconds.",
        header = "Long Duration",
        type = "primary",
        duration_s = 10
      )
    )
  })

  observeEvent(input$show_no_close, {
    show_toast(
      toast(
        "This toast has no close button but will auto-hide in 3 seconds.",
        type = "secondary",
        closable = FALSE,
        duration_s = 3
      )
    )
  })

  observeEvent(input$show_custom_header, {
    show_toast(
      toast(
        "Your profile has been updated successfully.",
        header = toast_header(
          title = "Profile Updated",
          icon = icon("check"),
          status = "just now"
        ),
        type = "success"
      )
    )
  })

  # Interactive toasts
  observeEvent(input$show_action_buttons, {
    show_toast(
      toast(
        id = "unsaved_changes_toast",
        p("Would you like to save your changes?"),
        div(
          class = "mt-2",
          actionButton("save_yes", "Save", class = "btn-sm btn-primary me-2"),
          actionButton(
            "save_no",
            "Don't Save",
            class = "btn-sm btn-secondary"
          )
        ),
        header = "Unsaved Changes",
        type = "warning",
        duration_s = 0,
        closable = FALSE
      )
    )
  })

  observeEvent(input$save_yes, {
    hide_toast("unsaved_changes_toast")
    show_toast(toast("Saved changes", type = "success"))
  })

  observeEvent(input$save_no, {
    hide_toast("unsaved_changes_toast")
    show_toast(toast("Changes were not saved", type = "danger"))
  })

  observeEvent(input$show_multiple, {
    show_toast(toast("First notification", type = "primary"))
    Sys.sleep(0.2)
    show_toast(toast("Second notification", type = "success"))
    Sys.sleep(0.2)
    show_toast(toast("Third notification", type = "info"))
  })

  observeEvent(input$show_all_positions, {
    positions <- c(
      "top-left",
      "top-center",
      "top-right",
      "middle-left",
      "middle-center",
      "middle-right",
      "bottom-left",
      "bottom-center",
      "bottom-right"
    )

    types <- c(
      "primary",
      "success",
      "info",
      "warning",
      "danger",
      "secondary",
      "light",
      "dark",
      "primary"
    )

    for (i in seq_along(positions)) {
      pos <- positions[i]
      show_toast(
        toast(
          paste("Toast at", pos),
          type = types[i],
          duration_s = 4,
          position = pos
        )
      )
    }
  })

  # Dynamic content toast
  inserted_time <- reactiveVal(NULL)

  observeEvent(input$show_dynamic_content, {
    show_toast(
      toast(
        id = "dynamic_content_toast",
        div(
          p("Current time:", strong(textOutput("toast_time", inline = TRUE))),
          plotOutput("toast_plot", height = "200px"),
          sliderInput(
            "toast_bins",
            "Number of bins:",
            min = 5,
            max = 50,
            value = 30,
            width = "100%"
          ),
        ),
        header = toast_header(
          title = "Dynamic Toast",
          status = textOutput("toast_status", inline = TRUE)
        ),
        type = "light",
        duration_s = 0
      )
    )
    inserted_time(Sys.time())
  })

  output$toast_time <- renderText({
    invalidateLater(1000)
    format(Sys.time(), "%H:%M:%S")
  })

  output$toast_status <- renderText({
    req(inserted_time())

    invalidateLater(1000)
    elapsed <- as.numeric(difftime(Sys.time(), inserted_time(), units = "secs"))
    if (elapsed < 60) {
      paste0(floor(elapsed), "s ago")
    } else {
      paste0(floor(elapsed / 60), "m ago")
    }
  })

  output$toast_plot <- renderPlot(
    {
      req(input$toast_bins)
      par(
        mar = c(0, 0, 2, 0), # Margins
        mgp = c(1, 0, 0)
      )
      hist(
        faithful$eruptions,
        breaks = input$toast_bins,
        col = "#444",
        border = "transparent",
        main = "Eruption Times",
        xlab = NULL,
        ylab = NULL,
        axes = FALSE
      )
    },
    bg = "transparent"
  )
}

shinyApp(ui, server)

#' Dynamically update nav containers
#'
#' Functions for dynamically updating nav containers (e.g., select, insert, and
#' remove nav items). These functions require an `id` on the nav container to be
#' specified and must be called within an active Shiny session.
#'
#' @param id a character string used to identify the nav container.
#' @param selected a character string used to identify a particular
#'   [nav_panel()] item.
#' @param session a shiny session object (the default should almost always be
#'   used).
#'
#' @seealso [Navset functions][navset] create the navigation container holding
#'   the nav panels.
#' @seealso [nav_panel()], [nav_panel_hidden()] create panels of content.
#' @seealso [nav_menu()], [nav_item()], [nav_spacer()] create menus, items, or
#'   space in the navset control area.
#'
#' @export
#' @family Panel container functions
#'
#' @examples
#'
#' can_browse <- function() rlang::is_interactive() && require("shiny")
#'
#' # Selecting a tab
#' if (can_browse()) {
#'   shinyApp(
#'     page_fluid(
#'       radioButtons("item", "Choose", c("A", "B")),
#'       navset_hidden(
#'         id = "container",
#'         nav_panel_hidden("A", "a"),
#'         nav_panel_hidden("B", "b")
#'       )
#'     ),
#'     function(input, output) {
#'       observe(nav_select("container", input$item))
#'     }
#'   )
#' }
#'
#' # Inserting and removing
#' if (can_browse()) {
#'   ui <- page_fluid(
#'     actionButton("add", "Add 'Dynamic' tab"),
#'     actionButton("remove", "Remove 'Foo' tab"),
#'     navset_tab(
#'       id = "tabs",
#'       nav_panel("Hello", "hello"),
#'       nav_panel("Foo", "foo"),
#'       nav_panel("Bar", "bar tab")
#'     )
#'   )
#'   server <- function(input, output) {
#'     observeEvent(input$add, {
#'       nav_insert(
#'         "tabs", target = "Bar", select = TRUE,
#'         nav_panel("Dynamic", "Dynamically added content")
#'       )
#'     })
#'     observeEvent(input$remove, {
#'       nav_remove("tabs", target = "Foo")
#'     })
#'   }
#'   shinyApp(ui, server)
#' }
#'
nav_select <- function(id, selected = NULL, session = get_current_session()) {
  shiny::updateTabsetPanel(session, id, selected)
}

#' @param nav a [nav_panel()] item.
#' @param target The `value` of an existing `nav_panel()` item, next to which tab will be added. If removing: the `value` of the `nav_panel()` item that you want to remove.
#' @param position Should `nav` be added before or after the target?
#' @param select Should `nav` be selected upon being inserted?
#' @rdname nav_select
#' @export
nav_insert <- function(
  id,
  nav,
  target = NULL,
  position = c("after", "before"),
  select = FALSE,
  session = get_current_session()
) {
  force(target)
  force(select)
  position <- match.arg(position)
  inputId <- session$ns(id)

  # Barbara -- August 2017
  # Note: until now, the number of tabs in a tabsetPanel (or navbarPage
  # or navlistPanel) was always fixed. So, an easy way to give an id to
  # a tab was simply incrementing a counter. (Just like it was easy to
  # give a random 4-digit number to identify the tabsetPanel). Since we
  # can only know this in the client side, we'll just pass `id` and
  # `tsid` (TabSetID) as dummy values that will be fixed in the JS code.
  item <- buildTabItem(
    "id",
    "tsid",
    TRUE,
    divTag = nav,
    textFilter = if (is.character(nav)) navbarMenuTextFilter else NULL
  )

  callback <- function() {
    session$sendInsertTab(
      inputId = inputId,
      liTag = processDeps(item$liTag, session),
      divTag = processDeps(item$divTag, session),
      menuName = NULL,
      target = target,
      position = position,
      select = select
    )
  }
  session$onFlush(callback, once = TRUE)
}

#' @export
#' @rdname nav_select
nav_remove <- function(id, target, session = get_current_session()) {
  force(target)
  inputId <- session$ns(id)

  callback <- function() {
    session$sendRemoveTab(
      inputId = inputId,
      target = target
    )
  }
  session$onFlush(callback, once = TRUE)
}

#' @export
#' @rdname nav_select
nav_show <- function(
  id,
  target,
  select = FALSE,
  session = get_current_session()
) {
  shiny::showTab(id, target, select, session)
}

#' @export
#' @rdname nav_select
nav_hide <- function(id, target, session = get_current_session()) {
  shiny::hideTab(id, target, session)
}

#' Append or prepend nav item to a dropdown menu
#'
#' Exported for use by [shiny::prependTab()]/[shiny::appendTab()]. These
#' functions have been superseded by [nav_insert()] (i.e.,
#' [shiny::insertTab()]), since it can do everything these functions do (i.e.,
#' add a [nav_panel()] to the start or end of a [nav_menu()]) and more (i.e., insert a
#' [nav_panel()] anywhere inside a nav container).
#'
#' @inheritParams nav_insert
#' @param menu_title The title of a [nav_menu()].
#' @keywords internal
#' @export
nav_prepend <- function(
  id,
  nav,
  menu_title,
  select = FALSE,
  session = get_current_session()
) {
  force(select)
  force(menu_title)
  inputId <- session$ns(id)

  item <- buildTabItem(
    "id",
    "tsid",
    TRUE,
    divTag = nav,
    textFilter = if (is.character(nav)) navbarMenuTextFilter else NULL
  )

  callback <- function() {
    session$sendInsertTab(
      inputId = inputId,
      liTag = processDeps(item$liTag, session),
      divTag = processDeps(item$divTag, session),
      menuName = menu_title,
      target = NULL,
      position = "before",
      select = select
    )
  }
  session$onFlush(callback, once = TRUE)
}

#' @rdname nav_prepend
#' @keywords internal
#' @export
nav_append <- function(
  id,
  nav,
  menu_title,
  select = FALSE,
  session = get_current_session()
) {
  force(select)
  force(menu_title)
  inputId <- session$ns(id)

  item <- buildTabItem(
    "id",
    "tsid",
    TRUE,
    divTag = nav,
    textFilter = if (is.character(nav)) navbarMenuTextFilter else NULL
  )

  callback <- function() {
    session$sendInsertTab(
      inputId = inputId,
      liTag = processDeps(item$liTag, session),
      divTag = processDeps(item$divTag, session),
      menuName = menu_title,
      target = NULL,
      position = "after",
      select = select
    )
  }
  session$onFlush(callback, once = TRUE)
}

#' A Bootstrap card component
#'
#' @description
#' A general purpose container for grouping related UI elements together with a
#' border and optional padding. To learn more about [card()]s, see [the Cards
#' article](https://rstudio.github.io/bslib/articles/cards/index.html) or the
#' other articles listed in the _References_ section below.
#'
#' @param ... Unnamed arguments can be any valid child of an [htmltools
#'   tag][htmltools::tags] (which includes card items such as [card_body()].
#'   Named arguments become HTML attributes on returned UI element.
#' @param full_screen If `TRUE`, an icon will appear when hovering over the card
#'   body. Clicking the icon expands the card to fit viewport size.
#' @param height Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `height="200px"`). Doesn't apply when a card is made `full_screen`
#'   (in this case, consider setting a `height` in [card_body()]).
#' @param max_height,min_height Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `max_height="200px"`). Doesn't apply when a card is made `full_screen`
#'   (in this case, consider setting a `max_height` in [card_body()]).
#' @param fill Whether or not to allow the card to grow/shrink to fit a
#'   fillable container with an opinionated height (e.g., `page_fillable()`).
#' @param class Additional CSS classes for the returned UI element.
#' @param wrapper A function (which returns a UI element) to call on unnamed
#'   arguments in `...` which are not already card item(s) (like
#'   [card_header()], [card_body()], etc.). Note that non-card items are grouped
#'   together into one `wrapper` call (e.g. given `card("a", "b",
#'   card_body("c"), "d")`, `wrapper` would be called twice, once with `"a"` and
#'   `"b"` and once with `"d"`).
#' @param id Provide a unique identifier for the `card()` or `value_box()` to
#'   report its full screen state to Shiny. For example, using `id = "my_card"`,
#'   you can observe the card's full screen state with
#'   `input$my_card_full_screen`.
#'
#' @return A [htmltools::div()] tag.
#'
#' @export
#' @family Components
#'
#' @seealso [Card item functions][card_body()] create the various parts of a
#'   card.
#' @seealso [navset_card_tab()], [navset_card_pill()] and
#'   [navset_card_underline()] create cards with tabbed navigation.
#' @seealso [layout_columns()] and [layout_column_wrap()] help position multiple
#'   cards into columns and rows and can also be used inside a card.
#' @seealso [layout_sidebar()] adds a sidebar to a card when nested in [card()]
#'   or [card_body()].
#' @seealso [value_box()] uses [card()] to highlight a showcase a key piece of
#'   information.
#'
#' @references Several articles on the bslib website feature the card component:
#'
#'   * [Cards](https://rstudio.github.io/bslib/articles/cards/index.html)
#'   * [Get Started: Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'   * [Get Started: Any Project](https://rstudio.github.io/bslib/articles/any-project/index.html)
#'   * [Column-based layouts](https://rstudio.github.io/bslib/articles/column-layout/index.html)
#'   * [Filling layouts: Full-screen cards](https://rstudio.github.io/bslib/articles/filling/index.html#full-screen-cards)
#'
#' @examplesIf rlang::is_interactive()
#' library(htmltools)
#'
#' card(
#'   full_screen = TRUE,
#'   card_header(
#'     "This is the header"
#'   ),
#'   card_body(
#'     p("This is the body."),
#'     p("This is still the body.")
#'   ),
#'   card_footer(
#'     "This is the footer"
#'   )
#' )
#'
card <- function(
  ...,
  full_screen = FALSE,
  height = NULL,
  max_height = NULL,
  min_height = NULL,
  fill = TRUE,
  class = NULL,
  wrapper = card_body,
  id = NULL
) {
  args <- rlang::list2(...)
  argnames <- rlang::names2(args)

  attribs <- args[nzchar(argnames)]
  children <- as_card_items(args[!nzchar(argnames)], wrapper = wrapper)
  children <- card_image_add_classes(children)

  is_shiny_input <- !is.null(id)

  if (full_screen && is.null(id)) {
    # a11y: full screen cards need an ID for aria-controls on the toggle button
    id <- paste0("bslib-card-", p_randomInt(1000, 10000))
  }

  tag <- div(
    id = id,
    class = "card bslib-card bslib-mb-spacing",
    class = if (is_shiny_input) "bslib-card-input",
    style = css(
      height = validateCssUnit(height),
      max_height = validateCssUnit(max_height),
      min_height = validateCssUnit(min_height)
    ),
    "data-bslib-card-init" = NA,
    "data-full-screen" = if (full_screen) "false",
    !!!attribs,
    !!!children,
    if (full_screen) full_screen_toggle(attribs$id),
    card_init_js(),
    component_dependencies()
  )

  tag <- bindFillRole(tag, container = TRUE, item = fill)
  tag <- tagAppendAttributes(tag, class = class)

  as_fragment(
    tag_require(tag, version = 5, caller = "card()")
  )
}

as_card_items <- function(children, wrapper) {
  # We don't want NULLs creating empty card bodys
  children <- children[vapply(children, function(x) length(x) > 0, logical(1))]

  if (!is.function(wrapper)) {
    return(children)
  }

  # Any children that are `is.card_item` should be included verbatim. Any
  # children that are not, should be wrapped in card_body(). Consecutive children
  # that are not card_item, should be wrapped in a single card_body().
  needs_wrap <- !vapply(children, is.card_item, logical(1))
  needs_wrap_rle <- rle(needs_wrap)
  start_indices <- c(1, head(cumsum(needs_wrap_rle$lengths) + 1, -1))
  children <- mapply(
    start_indices,
    needs_wrap_rle$lengths,
    needs_wrap_rle$values,
    FUN = function(start, length, wrap) {
      these_children <- children[start:(start + length - 1)]
      if (wrap) {
        list(wrapper(these_children))
      } else {
        these_children
      }
    },
    SIMPLIFY = FALSE
  )
  unlist(children, recursive = FALSE)
}

#' Card items
#'
#' Components designed to be provided as direct children of a [card()]. For a
#' general overview of the [card()] API, see [the Cards
#' article](https://rstudio.github.io/bslib/articles/cards/index.html) or the
#' other articles listed in the _References_ section of the [card()]
#' documentation.
#'
#' @param ... Unnamed arguments can be any valid child of an [htmltools
#'   tag][htmltools::tags]. Named arguments become HTML attributes on returned
#'   UI element.
#' @param min_height,max_height,max_height_full_screen Any valid [CSS length
#'   unit][htmltools::validateCssUnit()].
#' @param fillable Whether or not the card item should be a fillable (i.e.
#'   flexbox) container.
#' @param fill Whether to allow this element to grow/shrink to fit its `card()`
#'   container.
#' @param gap A [CSS length unit][htmltools::validateCssUnit()] defining the
#'   `gap` (i.e., spacing) between elements provided to `...`. This argument is only applicable when `fillable = TRUE`
#' @inheritParams card
#' @inheritParams page_fillable
#'
#' @return An [htmltools::div()] tag.
#'
#' @export
#'
#' @seealso [card()] creates a card component.
#' @seealso [navset_card_tab()], [navset_card_pill()] and
#'   [navset_card_underline()] create cards with tabbed navigation.
#' @seealso [layout_columns()] and [layout_column_wrap()] help position multiple
#'   cards into columns and rows and can also be used inside a card.
#' @seealso [layout_sidebar()] adds a sidebar to a card when nested in [card()]
#'   or [card_body()].
#'
#' @describeIn card_body A general container for the "main content" of a [card()].
card_body <- function(
  ...,
  fillable = TRUE,
  min_height = NULL,
  max_height = NULL,
  max_height_full_screen = max_height,
  height = NULL,
  padding = NULL,
  gap = NULL,
  fill = TRUE,
  class = NULL
) {
  if (fillable) {
    register_runtime_package_check("`card_body()`", "shiny", "1.7.4")
    register_runtime_package_check("`card_body()`", "htmlwidgets", "1.6.0")
  }

  tag <- div(
    class = "card-body",
    class = if (fillable) "bslib-gap-spacing",
    style = css(
      min_height = validateCssUnit(min_height),
      "--bslib-card-body-max-height" = validateCssUnit(max_height),
      "--bslib-card-body-max-height-full-screen" = validateCssUnit(
        max_height_full_screen
      ),
      margin_top = "auto",
      margin_bottom = "auto",
      # .card-body already adds `flex: 1 1 auto` so make sure to override it
      flex = if (fill) "1 1 auto" else "0 0 auto",
      padding = validateCssPadding(padding),
      gap = validateCssUnit(gap),
      height = validateCssUnit(height)
    ),
    ...
  )

  tag <- bindFillRole(tag, item = fill, container = fillable)

  # Make sure user has the opportunity to override the classes added by bindFillRole()
  tag <- tagAppendAttributes(tag, class = class)

  as.card_item(tag)
}

#' @describeIn card_body Similar to `card_header()` but without the border and background color.
#' @param container a function to generate an HTML element.
#' @export
card_title <- function(..., container = htmltools::h5) {
  container(...)
}

#' @describeIn card_body A header (with border and background color) for the `card()`. Typically appears before a `card_body()`.
#' @param container a function that generates an [htmltools tag][htmltools::tags].
#' @export
card_header <- function(
  ...,
  gap = NULL,
  class = NULL,
  container = htmltools::div
) {
  as.card_item(
    container(
      class = "card-header bslib-gap-spacing",
      class = class,
      style = css(gap = validateCssUnit(gap)),
      ...
    )
  )
}

#' @describeIn card_body A header (with border and background color) for the `card()`. Typically appears after a `card_body()`.
#' @export
card_footer <- function(..., class = NULL) {
  as.card_item(
    div(class = "card-footer", class = class, ...)
  )
}

#' @describeIn card_body Include static images in a card, for example as an
#'   image cap at the top or bottom of the card.
#'
#' @param file A file path pointing an image. Local images (i.e. not a URI
#'   starting with `https://` or similar) will be base64 encoded and provided to
#'   the `src` attribute of the `<img>`. Alternatively, you may directly set
#'   the image `src`, in which case `file` is ignored.
#' @param alt Alternate text for the image, used by screen readers and assistive
#'   devices. Provide alt text with a description of the image for any images
#'   with important content. If alt text is not provided, the image will be
#'   considered to be decorative and will not be read or announced by screen
#'   readers.
#'
#'   For more information, the Web Accessibility Initiative (WAI) has a
#'   [helpful tutorial on alt text](https://www.w3.org/WAI/tutorials/images/).
#' @param src The `src` attribute of the `<img>` tag. If provided, `file` is
#'   ignored entirely. Use `src` to provide a relative path to a file that will
#'   be served by the Shiny application and should not be base64 encoded.
#' @param href An optional URL to link to when a user clicks on the image.
#' @param border_radius Which side of the image should have rounded corners,
#'   useful when `card_image()` is used as an image cap at the top or bottom of
#'   the card.
#'
#'   The value of `border_radius` determines whether the `card-img-top`
#'   (`"top"`), `card-img-bottom` (`"bottom"`), or `card-img` (`"all"`)
#'   [Bootstrap
#'   classes](https://getbootstrap.com/docs/5.3/components/card/#images) are
#'   applied to the card. The default `"auto"` value will use the image's
#'   position within a `card()` to automatically choose the appropriate class.
#' @param mime_type The mime type of the `file` when it is base64 encoded. This
#'   argument is available for advanced use cases where [mime::guess_type()] is
#'   unable to automatically determine the file type.
#' @param container A function to generate an HTML element to contain the image.
#'   Setting this value to `card_body()` places the image inside the card body
#'   area, otherwise the image will extend to the edges of the card.
#' @param width Any valid [CSS unit][htmltools::validateCssUnit] (e.g.,
#'   `width="100%"`).
#'
#' @export
card_image <- function(
  file,
  ...,
  alt = "",
  src = NULL,
  href = NULL,
  border_radius = c("auto", "top", "bottom", "all", "none"),
  mime_type = NULL,
  class = NULL,
  height = NULL,
  fill = FALSE,
  width = NULL,
  container = NULL
) {
  if (any(!nzchar(rlang::names2(list(...))))) {
    rlang::abort(
      c(
        "Unnamed arguments were included in `...`.",
        i = "All additional arguments to `card_image()` in `...` should be named attributes for the `<img>` tag."
      )
    )
  }

  border_radius <- rlang::arg_match(border_radius)

  if (is.null(src)) {
    if (grepl("^([[:alnum:]]+:)?//|data:", file)) {
      src <- file
    } else {
      if (!file.exists(file)) {
        rlang::abort(
          c(
            sprintf("`file` does not exist: %s", file),
            i = sprintf(
              "If `file` is a remote file or will be served by the Shiny app, use a URL or set `src = \"%s\"`.",
              file
            )
          )
        )
      }
      src <- base64enc::dataURI(
        file = file,
        mime = mime_type %||% mime::guess_type(file)
      )
    }
  }

  image <- tags$img(
    src = src,
    alt = alt,
    class = "img-fluid",
    class = switch(
      border_radius,
      all = "card-img",
      top = "card-img-top",
      bottom = "card-img-bottom",
      NULL
    ),
    style = css(
      height = validateCssUnit(height),
      width = validateCssUnit(width)
    ),
    ...
  )

  image <- bindFillRole(image, item = fill)
  image <- tagAppendAttributes(image, class = class)

  if (!is.null(href)) {
    image <- bindFillRole(
      tags$a(href = href, image),
      container = TRUE,
      item = TRUE
    )
  }

  if (is.function(container)) {
    image <- container(image)
  } else {
    image <- as.card_item(image)
  }

  class(image) <- c(
    if (border_radius == "auto") "card_image_auto",
    "card_image",
    class(image)
  )

  image
}

card_image_add_classes <- function(children) {
  for (idx_child in seq_along(children)) {
    if (inherits(children[[idx_child]], "card_image_auto")) {
      card_img_class <-
        if (length(children) == 1) {
          "card-img"
        } else if (idx_child == 1) {
          "card-img-top"
        } else if (idx_child == length(children)) {
          "card-img-bottom"
        }

      children[[idx_child]] <- tagAppendAttributes(
        children[[idx_child]],
        class = card_img_class
      )
    }
  }

  children
}

#' @describeIn card_body Mark an object as a card item. This will prevent the
#'   [card()] from putting the object inside a `wrapper` (i.e., a
#'   `card_body()`).
#' @param x an object to test (or coerce to) a card item.
#' @export
as.card_item <- function(x) {
  class(x) <- c("card_item", class(x))
  x
}

#' @rdname card_body
#' @export
is.card_item <- function(x) {
  inherits(x, "card_item")
}

full_screen_toggle <- function(id_controls) {
  tooltip(
    tags$button(
      class = "bslib-full-screen-enter",
      class = "badge rounded-pill",
      "aria-expanded" = "false",
      "aria-controls" = id_controls,
      "aria-label" = "Expand card",
      full_screen_toggle_icon()
    ),
    "Expand"
  )
}

card_init_js <- function() {
  tags$script(
    `data-bslib-card-init` = NA,
    HTML("bslib.Card.initializeAllCards();")
  )
}

full_screen_toggle_icon <- function() {
  # https://www.visiwig.com/icons/
  # https://www.visiwig.com/icons-license/
  HTML(
    '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24" style="height:1em;width:1em;fill:currentColor;" aria-hidden="true" role="img"><path d="M20 5C20 4.4 19.6 4 19 4H13C12.4 4 12 3.6 12 3C12 2.4 12.4 2 13 2H21C21.6 2 22 2.4 22 3V11C22 11.6 21.6 12 21 12C20.4 12 20 11.6 20 11V5ZM4 19C4 19.6 4.4 20 5 20H11C11.6 20 12 20.4 12 21C12 21.6 11.6 22 11 22H3C2.4 22 2 21.6 2 21V13C2 12.4 2.4 12 3 12C3.6 12 4 12.4 4 13V19Z"/></svg>'
  )
}

# jcheng 2022-06-06: Removing for now; list items have more features than I'm
# ready to design an API for right now
#
# #' @rdname card_body
# #' @export
# card_list <- function(...) {
#   res <- tags$ul(class = "list-group list-group-flush", ...)
#   as.card_item(res)
# }
#
# #' @export
# card_list_item <- function(...) {
#   tags$li(class = "list-group-item", ...)
# }

code_modal <- function(code) {
  if (rlang::is_call(code)) {
    code <- rlang::expr_text(code)
  }

  if (requireNamespace("styler", quietly = TRUE)) {
    code <- styler::style_text(code)
  }

  code <- paste(code, collapse = "\n")

  showModal(
    modalDialog(
      HTML(sprintf(
        '<pre><code id="value-box-code">%s</code></pre>',
        code
      )),
      p(
        id = "copy-clipboard-not-supported",
        class = "text-muted d-none",
        HTML("Press <kbd>Ctrl</kbd>/<kbd>Cmd</kbd> + <kbd>C</kbd> to copy the value box example code.")
      ),
      tags$button(
        id = "copy-code-to-clipboard",
        class = "btn btn-outline-primary",
        onclick = "copyValueBoxCode()",
        "Copy to clipboard"
      ),
      singleton(tags$script(src = "code-modal.js")),
      footer = modalButton("Done"),
      easyClose = TRUE
    )
  )
}

#' Value box
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' An opinionated ([card()]-powered) box, designed for displaying a `value` and
#' `title`. Optionally, a `showcase` can provide for context for what the
#' `value` represents (for example, it could hold a [bsicons::bs_icon()], or
#' even a [shiny::plotOutput()]). Find examples and template code you can use to
#' create engaging [value boxes on the bslib
#' website](https://rstudio.github.io/bslib/articles/value-boxes/index.html).
#'
#' @section Build-a-Box App:
#'
#'   Explore all of the `value_box()` options and layouts interactively with the
#'   [Build-a-Box app](https://bslib.shinyapps.io/build-a-box/), available
#'   online thanks to [shinyapps.io](https://www.shinyapps.io/). Or, you can
#'   run the app locally with:
#'
#'   ```r
#'   shiny::runApp(system.file("examples", "build-a-box", package = "bslib"))
#'   ```
#'
#' @section Themes:
#'
#'   ```{r child="man/fragments/value-box-themes.Rmd"}
#'   ```
#'
#' @section Showcase Layouts:
#'
#'   ```{r child="man/fragments/value-box-showcase.Rmd"}
#'   ```
#'
#' @param title,value A string, number, or [htmltools::tag()] child to display as
#'   the title or value of the value box. The `title` appears above the `value`.
#' @param ... Unnamed arguments may be any [htmltools::tag()] children to
#'   display below `value`. Named arguments become attributes on the containing
#'   element.
#' @param showcase A [htmltools::tag()] child to showcase (e.g., a
#'   [bsicons::bs_icon()], a `plotly::plotlyOutput()`, etc).
#' @param showcase_layout One of `"left center"` (default), `"top right"` or
#'   `"bottom"`. Alternatively, you can customize the showcase layout options
#'   with the [showcase_left_center()], [showcase_top_right()], or
#'   [showcase_bottom()] functions. Use the options functions when you want to
#'   control the height or width of the showcase area.
#' @param theme The name of a theme for the value box, or a theme constructed
#'   with `value_box_theme()`. The theme names provide a convenient way to use
#'   your app's Bootstrap theme colors as the foreground or background colors of
#'   the value box. See below for more details on the provided themes. For more
#'   control, you can create your own theme with `value_box_theme()` where you
#'   can pass foreground and background colors directly. See the **Themes**
#'   section for more details.
#' @param class Utility classes for customizing the appearance of the summary
#'   card. Use `bg-*` and `text-*` classes (e.g, `"bg-danger"` and
#'   `"text-light"`) to customize the background/foreground colors.
#' @param fill Whether to allow the value box to grow/shrink to fit a fillable
#'   container with an opinionated height (e.g., `page_fillable()`).
#' @inheritParams card
#' @param theme_color `r lifecycle::badge("deprecated")` Use `theme` instead.
#'
#' @seealso Value boxes are a specialized form of a [card()] component.
#' @seealso [layout_columns()] and [layout_column_wrap()] help position multiple
#'   value boxes into columns and rows.
#'
#' @references Value boxes are featured on the bslib website in a few articles:
#'
#'   * [Value boxes](https://rstudio.github.io/bslib/articles/value-boxes/index.html)
#'   * [Build-a-Box App](https://bslib.shinyapps.io/build-a-box)
#'   * [Get Started: Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html#value-boxes)
#'
#' @examplesIf rlang::is_interactive()
#' library(htmltools)
#'
#' value_box(
#'   "KPI Title",
#'   h1(HTML("$1 <i>Billion</i> Dollars")),
#'   span(
#'     bsicons::bs_icon("arrow-up"),
#'     " 30% VS PREVIOUS 30 DAYS"
#'   ),
#'   showcase = bsicons::bs_icon("piggy-bank"),
#'   theme = "success"
#' )
#'
#' @family Components
#' @export
value_box <- function(
  title,
  value,
  ...,
  showcase = NULL,
  showcase_layout = c("left center", "top right", "bottom"),
  full_screen = FALSE,
  theme = NULL,
  height = NULL,
  max_height = NULL,
  fill = TRUE,
  class = NULL,
  theme_color = deprecated()
) {
  dots <- separate_arguments(...)
  attribs <- dots$attribs
  children <- dots$children

  # ---- Title and value ----
  if (rlang::is_bare_character(title) || rlang::is_bare_numeric(title)) {
    title <- tags$p(title)
  }
  if (rlang::is_bare_character(value) || rlang::is_bare_numeric(value)) {
    value <- tags$p(value)
  }

  if (!is.null(title)) {
    title <- tag_add_outer_class(title, class = "value-box-title")
  }
  if (!is.null(value)) {
    value <- tag_add_outer_class(value, class = "value-box-value")
  }

  # ---- Contents ----
  contents <- div(class = "value-box-area", title, value, !!!children)
  contents <- as_fill_carrier(contents)

  # ---- Showcase ----
  if (!is.null(showcase)) {
    showcase_layout <- resolve_showcase_layout(showcase_layout)
    contents <- render_showcase_layout(showcase_layout, showcase, contents)
  }

  # ---- Theme ----
  if (lifecycle::is_present(theme_color)) {
    lifecycle::deprecate_soft(
      "0.6.0",
      "value_box(theme_color =)",
      "value_box(theme =)",
      details = if (!missing(theme)) {
        "Both `theme` and `theme_color` were provided, but only the `theme` argument will be used."
      }
    )

    if (missing(theme)) {
      theme <- theme_color
    }
  }

  if (!inherits(theme, "bslib_value_box_theme")) {
    theme <- value_box_theme(theme)
  }

  # ---- Layout ----
  res <- card(
    class = c(
      "bslib-value-box",
      theme$class,
      class,
      if (!is.null(showcase)) showcase_layout$class
    ),
    full_screen = full_screen,
    height = height,
    max_height = max_height,
    fill = fill,
    style = css(
      color = theme$fg,
      background_color = theme$bg,
      # These variables are used by the full screen card button
      "--bslib-color-fg" = theme$fg,
      "--bslib-color-bg" = theme$bg
    ),
    !!!attribs,
    contents
  )

  as_fragment(tag_require(res, version = 5, caller = "value_box()"))
}

#' @param name The name of the theme, e.g. `"primary"`, `"danger"`, `"purple"`.
#' @param bg,fg The background and foreground colors for the theme. If only `bg`
#'   is provided, then the foreground color is automatically chosen from
#'   `$black` or `$white` to provide the best contrast with the background
#'   color.
#'
#' @rdname value_box
#' @export
value_box_theme <- function(name = NULL, bg = NULL, fg = NULL) {
  if (is.null(name)) {
    if (is.null(bg)) {
      name <- "default"
    } else {
      # don't warn if we can't get a contrast color, `bg` might be valid
      # CSS but not something sass can compute on
      fg <- fg %||% suppressWarnings(get_color_contrast(bg))
    }

    return(new_value_box_theme(name, bg, fg))
  }

  if (!rlang::is_string(name)) {
    rlang::abort('`theme` must be a single value, e.g. "primary", "danger", "purple", etc.')
  }

  if (!grepl("^(text|bg)-", name)) {
    name <- paste0("bg-", name)
  }

  new_value_box_theme(name, bg, fg)
}

new_value_box_theme <- function(class = NULL, bg = NULL, fg = NULL) {
  structure(
    list(
      class = class,
      fg = fg,
      bg = bg
    ),
    class = "bslib_value_box_theme"
  )
}

#' @param width,width_full_screen,height,height_full_screen one of the
#'   following:
#'   * A proportion (i.e., a number between 0 and 1) of available width or
#'     height to allocate to the showcase.
#'   * A valid [CSS unit][htmltools::validateCssUnit] defining the width or
#'     height of the showcase column, or a valid value accepted by the
#'     `grid-template-columns` (width) or `grid-template-rows` (height) CSS
#'     property to define the width or height of the showcase column or row.
#'     Accepted values in the second category are `"auto"`, `"min-content"`,
#'     `"max-content"`, a fractional unit (e.g. `2fr`), or a `minmax()` function
#'     (e.g., `minmax(100px, 1fr)`).
#' @param max_height,max_height_full_screen A proportion (i.e., a number between
#'   0 and 1) or any valid [CSS unit][htmltools::validateCssUnit] defining the
#'   showcase max_height.
#'
#' @export
#' @rdname value_box
showcase_left_center <- function(
  width = 0.3,
  width_full_screen = "1fr",
  max_height = "100px",
  max_height_full_screen = 0.67
) {
  new_showcase_layout(
    class = "showcase-left-center",
    width = width,
    width_full_screen = width_full_screen,
    max_height = max_height,
    max_height_full_screen = max_height_full_screen
  )
}

#' @export
#' @rdname value_box
showcase_top_right <- function(
  width = 0.4,
  width_full_screen = "1fr",
  max_height = "75px",
  max_height_full_screen = 0.67
) {
  new_showcase_layout(
    class = "showcase-top-right",
    width = width,
    width_full_screen = width_full_screen,
    max_height = max_height,
    max_height_full_screen = max_height_full_screen
  )
}

#' @export
#' @rdname value_box
showcase_bottom <- function(
  width = "100%",
  width_full_screen = NULL,
  height = "auto",
  height_full_screen = "2fr",
  max_height = "100px",
  max_height_full_screen = NULL
) {
  new_showcase_layout(
    class = "showcase-bottom",
    width = width,
    width_full_screen = width_full_screen,
    height = height,
    height_full_screen = height_full_screen,
    max_height = max_height,
    max_height_full_screen = max_height_full_screen
  )
}

new_showcase_layout <- function(
  class,
  width = 0.3,
  width_full_screen = "1fr",
  height = NULL,
  height_full_screen = NULL,
  max_height = "100px",
  max_height_full_screen = 0.67
) {
  width <- validate_grid_unit(width)
  width_full_screen <- validate_grid_unit(width_full_screen)

  height <- validate_grid_unit(height)
  height_full_screen <- validate_grid_unit(height_full_screen)

  max_height <- validate_height_unit(max_height)
  max_height_full_screen <- validate_height_unit(max_height_full_screen)

  structure(
    list(
      class = class,
      width = width,
      width_full_screen = width_full_screen,
      height = height,
      height_full_screen = height_full_screen,
      max_height = max_height,
      max_height_full_screen = max_height_full_screen
    ),
    class = "bslib_showcase_layout"
  )
}

resolve_showcase_layout <- function(showcase_layout) {
  if (inherits(showcase_layout, "bslib_showcase_layout")) {
    return(showcase_layout)
  }

  if (is.character(showcase_layout)) {
    layout_choices <- c("left center", "top right", "bottom")
    showcase_layout <- rlang::arg_match(showcase_layout, layout_choices)
    showcase_layout <- switch(
      showcase_layout,
      "left center" = showcase_left_center(),
      "top right" = showcase_top_right(),
      bottom = showcase_bottom()
    )
    return(showcase_layout)
  }

  rlang::abort("`showcase_layout` must be a string or a `showcase_*()` layout")
}

render_showcase_layout <- function(showcase_layout, showcase, contents) {
  stopifnot(inherits(showcase_layout, "bslib_showcase_layout"))

  showcase <- div(class = "value-box-showcase", showcase)
  showcase <- as_fill_carrier(showcase)

  grid_props <- css(
    "--bslib-grid-height" = "auto",
    "--bslib-grid-height-mobile" = "auto",
    # These are private variables but css() can't handle _ prefix yet
    # FIXME: rstudio/htmltools#397
    "---bslib-value-box-showcase-w" = showcase_layout$width,
    "---bslib-value-box-showcase-w-fs" = showcase_layout$width_full_screen,
    "---bslib-value-box-showcase-h" = showcase_layout$height,
    "---bslib-value-box-showcase-h-fs" = showcase_layout$height_full_screen,
    "---bslib-value-box-showcase-max-h" = showcase_layout$max_height,
    "---bslib-value-box-showcase-max-h-fs" = showcase_layout$max_height_full_screen
  )

  value_box_grid <- div(
    class = "value-box-grid",
    style = grid_props,
    showcase, # .value-box-showcase
    contents  # .value-box-area (prepared in value_box())
  )

  card_body(
    style = css(padding = 0),
    fillable = TRUE,
    as_fill_item(value_box_grid)
  )
}


# It seems to be to use % over fr here since there is no gap on the grid
validate_grid_unit <- function(x) {
  if (is.null(x)) return(x)

  if (!is_01_scalar(x)) {
    if (tolower(x) %in% c("auto", "min-content", "max-content")) {
      return(tolower(x))
    }
    if (grepl("^minmax\\(", x)) {
      return(x)
    }
    if (!grepl("\\d+\\s*fr", x)) {
      x <- validateCssUnit(x)
    }
    return(x)
  }
  paste0(100 * x, "%")
}

validate_height_unit <- function(x) {
  if (is.null(x)) return(x)

  if (!is_01_scalar(x)) {
    return(validateCssUnit(x))
  }
  paste0(100 * x, "%")
}

is_01_scalar <- function(x) {
  rlang::is_scalar_double(x) && x >= 0 && x <= 1
}

# Print methods ---------------------------------------

#' @export
print.bslib_value_box_theme <- function(x, ...) {
  cat("<bslib_value_box_theme>\n")
  if (!is.null(x$class)) {
    cat("theme: ", x$class, "\n", sep = "")
  }
  if (!is.null(x$bg)) {
    cat("background-color: ", x$bg, "\n", sep = "")
  }
  if (!is.null(x$fg)) {
    cat("color: ", x$fg, "\n", sep = "")
  }
  invisible(x)
}

#' @export
print.bslib_showcase_layout <- function(x, ...) {
  cat("<", x$class, ">\n", sep = "")

  fields <- c("width", "height", "max_height")
  unset <- "---"
  nchar_fields <- max(nchar(fields))
  nchar_value <- max(vapply(dropNulls(x[fields]), nchar, integer(1)))

  if (any(vapply(x[fields], is.null, logical(1)))) {
    nchar_value <- max(nchar_value, nchar(unset))
  }

  for (field in fields) {
    t_field <- format(paste0(field, ": "), width = nchar_fields + 2, align = "left")
    t_value <- format(x[[field]] %||% unset, width = nchar_value, align = "right")
    t_value_fs <- x[[paste0(field, "_full_screen")]] %||% unset

    cat(t_field, t_value, " [fs: ", t_value_fs, "]\n", sep = "")
  }

  invisible(x)
}

test_that("value_box(theme_color=) is deprecated", {
  lifecycle::expect_deprecated(
    value_box("title", "value", theme_color = "primary")
  )

  lifecycle::expect_deprecated(
    expect_equal(
      value_box("title", "value", theme_color = "primary"),
      value_box("title", "value", theme = "primary")
    )
  )

  lifecycle::expect_deprecated(
    expect_equal(
      value_box("title", "value", theme = "primary", theme_color = "secondary"),
      value_box("title", "value", theme = "primary")
    )
  )
})

test_that("resolve_showcase_layout()", {
  expect_equal(
    resolve_showcase_layout("left center"),
    showcase_left_center()
  )

  expect_equal(
    resolve_showcase_layout("top right"),
    showcase_top_right()
  )

  expect_equal(
    resolve_showcase_layout("bottom"),
    showcase_bottom()
  )

  expect_equal(
    resolve_showcase_layout(showcase_left_center(width = 0.5)),
    showcase_left_center(width = 0.5)
  )

  expect_error(
    resolve_showcase_layout("left"),
  )
})

test_that("resolve_showcase_layout() can resolve all layouts in value_box()", {
  layouts <- eval(rlang::fn_fmls(value_box)$showcase_layout)

  for (layout in layouts) {
    expect_silent(resolve_showcase_layout(layout))
  }
})

test_that("print.bslib_showcase_layout()", {
  expect_snapshot(showcase_left_center())
  expect_snapshot(showcase_top_right())
  expect_snapshot(showcase_bottom())
})

test_that("value_box_theme() basic usage", {
  expect_equal(
    value_box_theme(),
    new_value_box_theme("default")
  )

  expect_equal(
    value_box_theme("primary"),
    value_box_theme("bg-primary")
  )

  expect_null(value_box_theme("orange")$fg)
  expect_null(value_box_theme("orange")$bg)
  expect_null(value_box_theme(bg = "orange")$class)

  expect_error(value_box_theme(c("text-primary", "bg-dark")))
})

test_that("value_box_theme() fills in `fg` when only `bg` is provided", {
  # fills in foreground
  expect_equal(
    value_box_theme(bg = "black"),
    value_box_theme(bg = "black", fg = "#FFFFFF")
  )

  # doesn't fill in foreground because `name` was used
  theme_no_fg <- value_box_theme("text-yellow", bg = "black")
  expect_equal(theme_no_fg$class, "text-yellow")
  expect_equal(theme_no_fg$bg, "black")
  expect_null(theme_no_fg$fg)
})

test_that("value_box_theme() allows `bg` and `fg` to be CSS variables", {
  expect_equal(
    value_box_theme(bg = "var(--bs-primary)")$bg,
    "var(--bs-primary)"
  )

  expect_equal(
    value_box_theme(fg = "var(--bs-secondary)")$fg,
    "var(--bs-secondary)"
  )
})

test_that("value_box_theme() print method", {
  expect_snapshot(value_box_theme("bg-gradient-blue-purple"))
  expect_snapshot(value_box_theme("red", fg = "white"))
  expect_snapshot(value_box_theme(bg = "black"))
  # Can set `name` and `bg` but not `fg` without automatic contrast calculation
  expect_snapshot(value_box_theme("text-red", bg = "#FFE8E8"))
})

# This script generates code for replacing hardcoded colors for Bootstrap 3.
# Unlike Bootstrap 4, the Bootstrap 3 _variables.scss has a lot of repeated
# color literals. We want to replace all of the ones that are monochrome with
# dynamically calculated replacements.

# Load bootstrap-sass (bs3) _variables.scss
lines <- readLines(
  system.file(
    package = "bslib",
    "lib/bs3/assets/stylesheets/bootstrap/_variables.scss"
  )
)
# Retain lines that seem like they contain a hex color.
# (Note that colors specified by keyword would break this logic, but, AFAICT,
# there aren't any such colors in _variables.scss.)
lines <- lines[grepl("#[0-9a-f-A-F]", lines, perl = TRUE)]

# Parse each line into variable name and hex color
df <- data.frame(
  stringsAsFactors = FALSE,
  name = sub("^\\$(.+?):.*", "\\1", lines, perl = TRUE),
  color = sub(".*(#[0-9A-Za-z]+).*", "\\1", lines, perl = TRUE)
)
# Sort by color value
df <- df[order(df$color), ]

# All of the color values that we want to replace, are assigned a variable name.
# Any variables in _variables.scss that point to one of these color values, will
# now point to the variable name by default. For example, since _variables.scss
# contains this:
#
#   $navbar-inverse-bg: #222 !default;
#
# Our goal is to generate the rule:
#
#   $navbar-inverse-bg: $gray-darker !default;
#
# So that all such variables (i.e. ones whose default value is #222) can be
# changed just by modifying $gray-darker.
named_colors <- c(
  # These variables already exist in _variables.scss
  "gray-base" = "#000",
  "gray-darker" = "#222",
  "gray-dark" = "#333",
  "gray" = "#555",
  "gray-light" = "#777",
  "gray-lighter" = "#eee",

  # These are variables we're introducing
  "gray-44" = "#444",
  "gray-88" = "#888",
  "gray-99" = "#999",
  "gray-cc" = "#ccc",
  "gray-dd" = "#ddd",
  "gray-e5" = "#e5e5e5",
  "gray-f5" = "#f5f5f5",
  "gray-f8" = "#f8f8f8",
  "gray-f9" = "#f9f9f9",
  "white" = "#fff"
)

# Color variables in _variables.scss that we don't have names for.
# Consider this a potential TODO list of colors we could dynamically set.
df[!df$color %in% named_colors, ]

# Generate the code for color_mapping in bs3_theme_base_colors
library(dplyr)
df %>%
  mutate(varname = names(named_colors[match(color, named_colors)])) %>%
  filter(!is.na(varname)) %>%
  filter(name != varname) %>%
  mutate(varname = paste0("$", varname)) %>%
  select(name, varname) %>%
  {
    setNames(.$varname, .$name)
  } %>%
  as.list() %>%
  dput()

library(shiny)
library(bslib)

ui <- page_navbar(
  title = "Analytics Dashboard",
  theme = bs_theme(preset = "shiny"),
  fillable = TRUE,

  nav_panel(
    "Overview",
    layout_columns(
      col_widths = c(8, 4),

      # Main data table card
      card(
        full_screen = TRUE,
        card_header(
          "Sales Data",
          toolbar(
            align = "right",
            toolbar_input_select(
              id = "filter",
              label = "Filter",
              choices = c("All", "Active", "Inactive"),
              icon = icon("filter")
            ),
            toolbar_input_select(
              id = "sort",
              label = "Sort by",
              choices = c("Date", "Amount", "Customer"),
              icon = icon("sort")
            ),
            toolbar_divider(),
            toolbar_input_button(
              id = "refresh",
              label = "Refresh",
              icon = icon("arrows-rotate")
            ),
            toolbar_input_button(
              id = "export",
              label = "Export",
              icon = icon("download"),
              class = "btn-primary",
              show_label = TRUE
            )
          )
        ),
        card_body(
          tableOutput("sales_table")
        ),
        card_footer(
          toolbar(
            align = "left",
            "Showing 10 of 247 records"
          ),
          toolbar(
            align = "right",
            toolbar_input_button(
              id = "update_filter_btn",
              label = "Update Filter Choices",
              icon = icon("wand-magic-sparkles"),
              show_label = TRUE
            ),
            toolbar_input_button(
              id = "update_export_btn",
              label = "Update Export Button",
              icon = icon("wand-magic-sparkles"),
              show_label = TRUE
            )
          )
        )
      ),

      # Stats card
      card(
        card_header(
          "Quick Stats",
          toolbar(
            align = "right",
            toolbar_input_select(
              id = "period",
              label = "Period",
              choices = c("Today", "Week", "Month", "Year"),
              selected = "Week",
              show_label = TRUE
            ),
            toolbar_divider(),
            toolbar_input_button(
              id = "stats_refresh",
              label = "Refresh",
              icon = icon("arrows-rotate")
            )
          )
        ),
        card_body(
          numericInput(
            "stats_threshold",
            label = toolbar(
              "Sales Threshold",
              toolbar_spacer(),
              toolbar_input_button(
                "threshold_decrease",
                label = "Decrease",
                icon = icon("minus")
              ),
              toolbar_input_button(
                "threshold_increase",
                label = "Increase",
                icon = icon("plus")
              )
            ),
            value = 1000,
            min = 0,
            max = 10000,
            step = 100
          ),
          verbatimTextOutput("stats")
        ),
        card_footer(
          toolbar(
            align = "left",
            toolbar_input_select(
              id = "stats_comparison",
              label = "Compare to",
              choices = c("Previous Period", "Last Year", "Budget"),
              show_label = TRUE
            )
          )
        )
      )
    ),

    layout_columns(
      col_widths = c(6, 6),

      # Chart card with view controls
      card(
        full_screen = TRUE,
        card_header(
          "Revenue Trend",
          toolbar(
            align = "right",
            toolbar_input_button(
              id = "chart_type",
              label = "Bar",
              icon = icon("chart-bar"),
              tooltip = "Switch to Bar Chart"
            ),
            toolbar_divider(),
            popover(
              toolbar_input_button(
                id = "chart_settings",
                label = "Settings",
                icon = icon("sliders"),
                tooltip = FALSE,
                show_label = FALSE,
                title = "Settings"
              ),
              radioButtons(
                "color_scheme",
                "Color Scheme",
                choices = c(
                  "Blue" = "#0d6efd",
                  "Green" = "#198754",
                  "Purple" = "#6f42c1",
                  "Orange" = "#fd7e14",
                  "Red" = "#dc3545"
                ),
                selected = "#0d6efd",
                inline = TRUE
              )
            )
          )
        ),
        card_body(
          plotOutput("revenue_chart", height = "300px")
        )
      ),

      # Activity feed card
      card(
        full_screen = TRUE,
        card_header(
          "Recent Activity",
          toolbar(
            align = "right",
            toolbar_input_button(
              id = "activity_refresh",
              label = "Refresh",
              icon = icon("rotate")
            ),
            toolbar_input_button(
              id = "activity_filter",
              label = "Filter",
              icon = icon("filter")
            )
          )
        ),
        card_body(
          uiOutput("activity_feed")
        ),
        card_footer(
          toolbar(
            align = "right",
            toolbar_input_button(
              id = "view_all",
              label = "View All",
              show_label = TRUE
            )
          )
        )
      )
    )
  ),

  nav_panel(
    "Chat",
    layout_columns(
      col_widths = 12,
      card(
        full_screen = TRUE,
        card_header("Customer Support Chat"),
        card_body(
          min_height = "400px",
          uiOutput("chat_messages")
        ),
        card_footer(
          input_submit_textarea(
            "chat_input",
            placeholder = "Type your message here... (Cmd/Ctrl + Enter to send)",
            toolbar = toolbar(
              toolbar_input_button(
                "attach_file",
                icon = icon("paperclip"),
                label = "Attach"
              ),
              toolbar_input_button(
                "emoji",
                icon = icon("face-smile"),
                label = "Emoji"
              ),
              toolbar_divider(),
              toolbar_input_select(
                "message_format",
                label = "Format",
                choices = c("Plain", "Markdown", "HTML"),
                icon = icon("code")
              )
            )
          )
        )
      )
    )
  ),

  nav_panel(
    "Reports",
    layout_columns(
      col_widths = 12,

      # Monthly Sales Report
      card(
        full_screen = TRUE,
        card_header(
          "Q4 2024 Sales Report",
          toolbar(
            align = "right",
            toolbar_input_button(
              id = "new_report",
              label = "New",
              icon = icon("file"),
              class = "btn-outline-success",
              show_label = TRUE
            ),
            toolbar_input_button(
              id = "open_report",
              label = "Open",
              icon = icon("folder-open")
            ),
            toolbar_divider(),
            toolbar_input_button(
              id = "save_report",
              label = "Save",
              icon = icon("save")
            ),
            toolbar_input_button(
              id = "print_report",
              label = "Print",
              icon = icon("print")
            ),
            toolbar_divider(),
            toolbar_input_button(
              id = "share_report",
              label = "Share",
              icon = icon("share-nodes"),
              show_label = TRUE,
              class = "btn-outline-primary",
              border = TRUE
            )
          )
        ),
        card_body(
          min_height = "500px",
          tags$div(
            style = "padding: 20px;",
            tags$h3("Executive Summary"),
            tags$p(
              "This report provides a comprehensive analysis of sales performance for Q4 2024, highlighting key trends, top-performing products, and regional performance metrics."
            ),

            tags$h4(style = "margin-top: 30px;", "Key Metrics"),
            tags$ul(
              tags$li(tags$strong("Total Revenue:"), " $2,847,392"),
              tags$li(tags$strong("Total Orders:"), " 1,247"),
              tags$li(tags$strong("Average Order Value:"), " $2,283"),
              tags$li(tags$strong("Year-over-Year Growth:"), " +18.5%")
            ),

            tags$h4(style = "margin-top: 30px;", "Top Products"),
            tableOutput("report_products"),

            tags$h4(style = "margin-top: 30px;", "Regional Performance"),
            plotOutput("report_regional_chart", height = "300px"),

            tags$h4(style = "margin-top: 30px;", "Recommendations"),
            tags$ol(
              tags$li(
                "Increase inventory for top-performing products to meet Q1 demand"
              ),
              tags$li("Focus marketing efforts on underperforming regions"),
              tags$li(
                "Implement customer retention programs to maintain growth momentum"
              )
            )
          )
        ),
        card_footer(
          toolbar(
            align = "left",
            tags$span(
              style = "color: #6c757d;",
              "Generated: ",
              format(Sys.time(), "%B %d, %Y at %I:%M %p")
            )
          ),
          toolbar(
            align = "right",
            toolbar_input_button(
              id = "add_chart",
              label = "Add Chart",
              icon = icon("chart-column"),
              class = "btn-info",
              show_label = TRUE
            ),
            toolbar_input_button(
              id = "add_table",
              label = "Add Table",
              icon = icon("table"),
              class = "btn-warning",
              show_label = TRUE
            ),
            toolbar_input_button(
              id = "add_text",
              label = "Add Text",
              icon = icon("paragraph"),
              class = "btn-danger",
              show_label = TRUE
            )
          )
        )
      )
    )
  )
)

server <- function(input, output, session) {
  # Chat messages storage
  chat_messages <- reactiveVal(list(
    list(
      text = "Welcome to customer support! How can we help you today?",
      from = "Agent",
      time = Sys.time()
    )
  ))

  # Sample data - larger dataset for filtering/sorting
  set.seed(123)
  sales_data <- data.frame(
    Date = rep(seq.Date(Sys.Date() - 29, Sys.Date(), by = "day"), each = 3),
    Customer = paste("Customer", rep(1:30, 3)),
    Amount = round(runif(90, 100, 1000), 2),
    Status = sample(
      c("Active", "Inactive"),
      90,
      replace = TRUE,
      prob = c(0.7, 0.3)
    )
  )

  # Chart type and color reactive values
  chart_type <- reactiveVal("line")
  chart_color <- reactiveVal("#0d6efd")

  # Chat message rendering
  output$chat_messages <- renderUI({
    messages <- chat_messages()

    tags$div(
      style = "display: flex; flex-direction: column; gap: 12px; padding: 15px;",
      lapply(messages, function(msg) {
        align <- if (msg$from == "You") "flex-end" else "flex-start"
        bg_color <- if (msg$from == "You") "#0d6efd" else "#6c757d"
        text_color <- "white"

        tags$div(
          style = sprintf("display: flex; justify-content: %s;", align),
          tags$div(
            style = sprintf(
              "max-width: 70%%; padding: 10px 15px; border-radius: 12px; background-color: %s; color: %s;",
              bg_color,
              text_color
            ),
            tags$div(
              style = "font-weight: 500; font-size: 0.85em; margin-bottom: 4px;",
              msg$from
            ),
            tags$div(msg$text),
            tags$div(
              style = "font-size: 0.75em; margin-top: 4px; opacity: 0.8;",
              format(msg$time, "%I:%M %p")
            )
          )
        )
      })
    )
  })

  # Handle chat message submission
  observeEvent(input$chat_input, {
    req(nchar(input$chat_input) > 0)

    # Add user message
    current_messages <- chat_messages()
    current_messages <- c(
      current_messages,
      list(list(text = input$chat_input, from = "You", time = Sys.time()))
    )

    # Simulate agent response
    agent_responses <- c(
      "I understand your concern. Let me help you with that.",
      "That's a great question! Here's what I found...",
      "Thank you for reaching out. I'll look into this right away.",
      "I appreciate your patience. Let me check our records.",
      "Got it! I'll process that for you now."
    )

    current_messages <- c(
      current_messages,
      list(list(
        text = sample(agent_responses, 1),
        from = "Agent",
        time = Sys.time() + 2
      ))
    )

    chat_messages(current_messages)

    # Clear the input
    update_submit_textarea("chat_input", value = "", focus = TRUE)
  })

  # Chat toolbar buttons
  observeEvent(input$attach_file, {
    show_toast(
      toast(
        "File attachment feature would open here",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$emoji, {
    show_toast(
      toast(
        "Emoji picker would open here",
        type = "info",
        duration_s = 2
      )
    )
  })

  # Threshold increment/decrement buttons
  observeEvent(input$threshold_increase, {
    updateNumericInput(
      session,
      "stats_threshold",
      value = input$stats_threshold + 100
    )
  })

  observeEvent(input$threshold_decrease, {
    updateNumericInput(
      session,
      "stats_threshold",
      value = input$stats_threshold - 100
    )
  })

  # Filter and sort sales data
  filtered_sales <- reactive({
    data <- sales_data

    # Apply status filter
    if (!is.null(input$filter) && input$filter != "All") {
      data <- data[data$Status == input$filter, ]
    }

    # Apply sorting
    if (!is.null(input$sort)) {
      data <- switch(
        input$sort,
        "Date" = data[order(data$Date, decreasing = TRUE), ],
        "Amount" = data[order(data$Amount, decreasing = TRUE), ],
        "Customer" = data[order(data$Customer), ],
        data
      )
    }

    data
  })

  output$sales_table <- renderTable({
    head(filtered_sales(), 10)
  })

  # Stats that update based on filter
  output$stats <- renderText({
    data <- filtered_sales()
    total_sales <- sum(data$Amount)
    avg_order <- mean(data$Amount)
    num_customers <- length(unique(data$Customer))

    paste(
      sprintf(
        "Total Sales: $%s",
        format(round(total_sales, 2), big.mark = ",")
      ),
      sprintf("Customers: %d", num_customers),
      sprintf("Avg Order: $%.2f", avg_order),
      sprintf("Records: %d", nrow(data)),
      sep = "\n"
    )
  })

  # Revenue chart with dynamic type and color
  output$revenue_chart <- renderPlot({
    dates <- seq.Date(Sys.Date() - 29, Sys.Date(), by = "day")
    revenue <- cumsum(runif(30, 500, 2000))

    if (chart_type() == "line") {
      plot(
        dates,
        revenue,
        type = "l",
        lwd = 3,
        col = chart_color(),
        xlab = "Date",
        ylab = "Revenue ($)",
        main = "30-Day Revenue Trend (Line)"
      )
      grid()
    } else {
      barplot(
        revenue,
        names.arg = format(dates, "%m/%d"),
        col = chart_color(),
        border = NA,
        xlab = "Date",
        ylab = "Revenue ($)",
        main = "30-Day Revenue Trend (Bar)",
        las = 2,
        cex.names = 0.7
      )
      grid(nx = NA, ny = NULL)
    }
  })

  output$activity_feed <- renderUI({
    activities <- c(
      "New order #1234 received",
      "Customer John Doe registered",
      "Payment processed: $450.00",
      "Report generated: Monthly Summary",
      "User Jane Smith updated profile"
    )

    tags$div(
      lapply(activities, function(activity) {
        tags$div(
          style = "padding: 8px; border-bottom: 1px solid #dee2e6;",
          tags$small(class = "text-muted", format(Sys.time(), "%H:%M")),
          tags$span(style = "margin-left: 10px;", activity)
        )
      })
    )
  })

  # Chart type toggle buttons
  observeEvent(input$chart_type, {
    new_type <- if (input$chart_type %% 2 == 0) "line" else "bar"
    chart_type(new_type)

    update_toolbar_input_button(
      "chart_type",
      label = if (new_type == "line") "Bar Chart" else "Line Chart",
      icon = if (new_type == "line") icon("chart-bar") else icon("chart-line")
    )
    update_tooltip(
      "chart_type-tooltip",
      paste("Switch to", if (new_type == "line") "Bar Chart" else "Line Chart")
    )
  })

  # Data refresh button
  observeEvent(input$refresh, {
    show_toast(
      toast(
        "Data refreshed!",
        type = "info",
        duration_s = 2
      )
    )
  })

  # Update filter select button - changes icon, label, and choices
  observeEvent(input$update_filter_btn, {
    update_toolbar_input_select(
      "filter",
      label = "Status",
      icon = icon("circle-check"),
      choices = c("All", "Approved", "Pending", "Rejected"),
      selected = "All"
    )
    show_toast(
      toast(
        "Filter updated with new icon, label, and choices!",
        type = "success",
        duration_s = 3
      )
    )
  })

  # Update export button - changes icon and label
  observeEvent(input$update_export_btn, {
    update_toolbar_input_button(
      "export",
      label = "Download CSV",
      icon = icon("file-csv")
    )
    show_toast(
      toast(
        "Export button updated with new icon and label!",
        type = "success",
        duration_s = 3
      )
    )
  })

  # Data refresh button
  observeEvent(input$share_data, {
    show_toast(
      toast(
        "Magical sharing button clicked!",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$export, {
    show_toast(
      toast(
        "Exporting data to CSV...",
        type = "info",
        duration_s = 2
      )
    )
  })

  # Pagination
  observeEvent(input$prev_page, {
    show_toast(
      toast(
        "Previous page",
        type = "info",
        duration_s = 1
      )
    )
  })

  observeEvent(input$next_page, {
    show_toast(
      toast(
        "Next page",
        type = "info",
        duration_s = 1
      )
    )
  })

  # Apply settings from popover
  observe({
    chart_color(input$color_scheme)
  })

  # Activity feed controls
  observeEvent(input$activity_refresh, {
    show_toast(
      toast(
        "Activity refreshed!",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$activity_filter, {
    show_toast(
      toast(
        "Opening activity filter...",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$view_all, {
    show_toast(
      toast(
        "Opening full activity log...",
        type = "info",
        duration_s = 2
      )
    )
  })

  # Report builder
  observeEvent(input$new_report, {
    show_toast(
      toast(
        "Creating new report...",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$open_report, {
    show_toast(
      toast(
        "Opening report...",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$save_report, {
    show_toast(
      toast(
        "Report saved!",
        type = "success",
        duration_s = 2
      )
    )
  })

  observeEvent(input$print_report, {
    show_toast(
      toast(
        "Printing report...",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$share_report, {
    show_toast(
      toast(
        "This is where you'd add logic to share the report.",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$add_chart, {
    show_toast(
      toast(
        "Adding chart to report...",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$add_table, {
    show_toast(
      toast(
        "Adding table to report...",
        type = "info",
        duration_s = 2
      )
    )
  })

  observeEvent(input$add_text, {
    show_toast(
      toast(
        "Adding text block to report...",
        type = "info",
        duration_s = 2
      )
    )
  })

  # Report data
  output$report_products <- renderTable(
    {
      data.frame(
        Product = c(
          "Enterprise License",
          "Pro Subscription",
          "Basic Package",
          "Consulting Services",
          "Training Program"
        ),
        Revenue = c("$847,200", "$623,400", "$412,800", "$589,392", "$374,600"),
        Units = c(354, 1247, 2064, 89, 156),
        Growth = c("+23%", "+18%", "+12%", "+31%", "+8%")
      )
    },
    striped = TRUE
  )

  output$report_regional_chart <- renderPlot({
    regions <- c(
      "North America",
      "Europe",
      "Asia Pacific",
      "Latin America",
      "Middle East"
    )
    revenue <- c(1200000, 850000, 620000, 120000, 57392)

    barplot(
      revenue / 1000,
      names.arg = regions,
      col = "#0d6efd",
      border = NA,
      ylab = "Revenue ($1000s)",
      main = "Q4 2024 Revenue by Region",
      las = 2,
      cex.names = 0.8
    )
    grid(nx = NA, ny = NULL)
  })
}

shinyApp(ui = ui, server = server)

ui_global_controls <- function(id) {
  ns <- shiny::NS(id)

  tagList(
    layout_columns(
      class = "align-items-end",
      selectizeInput(
        ns("theme_style"),
        "Theme style",
        list(
          "Default" = "default",
          "All" = "all",
          "Semantic Colors" = list(
            "Semantic Background" = "semantic-bg",
            "Semantic Text" = "semantic-fg"
          ),
          "Theme Colors" = list(
            "Colored Background" = "colors-bg",
            "Colored Text" = "colors-fg"
          ),
          "Vibrant" = list(
            "Gradient Background" = "gradient"
          )
        )
      ),
      shuffleButton(ns("random_theme"), "Theme"),
      shuffleButton(ns("random_stat"), "Stats")
    ),
    layout_columns(
      class = "align-items-start",
      div(
        radioButtons(
          ns("showcase_item"),
          "Showcase Item",
          choices = c("Plot", "Icon"),
          inline = TRUE
        ),
        conditionalPanel(
          "input.showcase_item == 'Plot'",
          ns = ns,
          p(
            class = "text-muted",
            "See",
            tags$a(href = "https://rstudio.github.io/bslib/articles/value-boxes/index.html#expandable-sparklines", "Expandable Sparklines"),
            "for example plot code."
          )
        ),
        conditionalPanel(
          "input.showcase_item == 'Icon'",
          ns = ns,
          shuffleButton(ns("random_icon"), "Icons")
        )
      ),
      radioButtons(
        ns("showcase_layout"),
        "Showcase Layout",
        choices = c("Left center", "Top right", "Bottom"),
        inline = TRUE
      )
    )
  )
}

server_global_controls <- function(input, output, sessions, one, two, three) {
  observeEvent(c(input$random_theme, input$theme_style), {
    new_values <- switch(
      input$theme_style,
      all = {
        one$theme$shuffle()
        two$theme$shuffle()
        three$theme$shuffle()
        NULL
      },
      default = {
        one$theme$set("Default")
        two$theme$set("Default")
        three$theme$set("Default")
        NULL
      },
      "semantic-bg" = sample(setdiff(theme_colors, c("light", "dark")), 3),
      "semantic-fg" = paste0("text-", sample(setdiff(theme_colors, c("light", "dark")), 3)),
      "colors-bg" = sample(named_colors, 3, replace = TRUE),
      "colors-fg" = paste0("text-", sample(named_colors, 3, replace = TRUE)),
      gradient = sample(gradient_classes, 3)
    )

    if (is.null(new_values)) return()

    one$theme$set(new_values[[1]])
    two$theme$set(new_values[[2]])
    three$theme$set(new_values[[3]])
  }, ignoreInit = TRUE)

  observeEvent(input$random_stat, {
    one$random_stat()
    two$random_stat()
    three$random_stat()
  })

  observeEvent(input$random_icon, {
    one$showcase_icon$shuffle()
    two$showcase_icon$shuffle()
    three$showcase_icon$shuffle()
  })

  observeEvent(input$showcase_item, {
    item <- tolower(input$showcase_item)
    one$set_showcase_item(item)
    two$set_showcase_item(item)
    three$set_showcase_item(item)
  }, ignoreInit = TRUE)

  observeEvent(input$showcase_layout, {
    layout <- tolower(input$showcase_layout)
    one$set_showcase_layout(layout)
    two$set_showcase_layout(layout)
    three$set_showcase_layout(layout)
  }, ignoreInit = TRUE)
}

module_global_controls <- function(id, one, two, three) {
  callModule(server_global_controls, id, one = one, two = two, three = three)
}

# TODO: Once we have more UI stuff in bslib, copy over the
# page constructors from shiny so folks can create static
# pages without a shiny dependency.

#' Modern Bootstrap page layouts
#'
#' These functions are small wrappers around shiny's page constructors (i.e.,
#' [shiny::fluidPage()], [shiny::navbarPage()], etc) that differ in two ways:
#'  * The `theme` parameter defaults bslib's recommended version of Bootstrap (for new projects).
#'  * The return value is rendered as an static HTML page when printed interactively at the console.
#'
#' @inheritParams shiny::bootstrapPage
#' @param ... UI elements for the page. Named arguments become HTML attributes.
#' @param theme A [bslib::bs_theme()] object.
#'
#' @seealso Dashboard-style pages: [page_sidebar()], [page_navbar()],
#'   [page_fillable()].
#'
#' @describeIn page A \pkg{bslib} wrapper for [shiny::bootstrapPage()], a basic
#'   Boostrap page where the content is added directly to the page body.
#'
#' @export
page <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  # Wrap ... in <body> since bootstrapPage() passes ... to tagList(),
  # which means named arguments aren't handled sensibly
  as_page(
    shiny::bootstrapPage(
      tags$body(...),
      title = title,
      theme = theme,
      lang = lang,
      # Components require Bootstrap 5+
      if (isTRUE(theme_version(theme) >= 5)) component_dependencies()
    ),
    theme = theme
  )
}

#' @describeIn page A \pkg{bslib} wrapper for [shiny::fluidPage()], a fluid
#'   Bootstrap-based page layout that extends to the full viewport width.
#' @inheritParams shiny::fluidPage
#' @export
page_fluid <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  as_page(
    shiny::fluidPage(
      # Components require Bootstrap 5+
      if (isTRUE(theme_version(theme) >= 5)) component_dependencies(),
      ...,
      title = title,
      theme = theme,
      lang = lang
    ),
    theme = theme
  )
}

#' @describeIn page A \pkg{bslib} wrapper for [shiny::fixedPage()], a fixed
#'   Bootstrap-based page layout where the page content container is centered
#'   horizontally and its width is constrained.
#' @inheritParams shiny::fixedPage
#' @export
page_fixed <- function(..., title = NULL, theme = bs_theme(), lang = NULL) {
  as_page(
    shiny::fixedPage(
      # Components require Bootstrap 5+
      if (isTRUE(theme_version(theme) >= 5)) component_dependencies(),
      ...,
      title = title,
      theme = theme,
      lang = lang
    ),
    theme = theme
  )
}

#' A screen-filling page layout
#'
#' @description
#' A Bootstrap-based page layout whose contents fill the full height and width
#' of the browser window.
#'
#' @param padding Padding to use for the body. This can be a numeric vector
#'   (which will be interpreted as pixels) or a character vector with valid CSS
#'   lengths. The length can be between one and four. If one, then that value
#'   will be used for all four sides. If two, then the first value will be used
#'   for the top and bottom, while the second value will be used for left and
#'   right. If three, then the first will be used for top, the second will be
#'   left and right, and the third will be bottom. If four, then the values will
#'   be interpreted as top, right, bottom, and left respectively.
#' @param fillable_mobile Whether or not the page should fill the viewport's
#'   height on mobile devices (i.e., narrow windows).
#' @param gap A [CSS length unit][htmltools::validateCssUnit()] defining the
#'   `gap` (i.e., spacing) between elements provided to `...`.
#' @inheritParams page
#'
#' @export
#' @family Dashboard page layouts
#'
#' @seealso [layout_columns()] and [layout_column_wrap()] for laying out content
#'   into rows and columns.
#' @seealso [layout_sidebar()] for 'floating' sidebar layouts.
#' @seealso [accordion()] for grouping related input controls in the `sidebar`.
#' @seealso [card()] for wrapping outputs in the 'main' content area.
#' @seealso [value_box()] for highlighting values.
#'
#' @references
#'   * [Filling Layouts](https://rstudio.github.io/bslib/articles/filling/index.html)
#'     on the bslib website.
#'   * [Getting Started with Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'     on the bslib website.
#'
#' @examplesIf rlang::is_interactive()
#'
#' library(shiny)
#' library(ggplot2)
#' library(bslib)
#'
#' ui <- page_fillable(
#'   h1("Example", code("mtcars"), "dashboard"),
#'   layout_columns(
#'     card(
#'       full_screen = TRUE,
#'       card_header("Number of forward gears"),
#'       plotOutput("gear")
#'     ),
#'     card(
#'       full_screen = TRUE,
#'       card_header("Number of carburetors"),
#'       plotOutput("carb")
#'     )
#'   ),
#'   card(
#'     full_screen = TRUE,
#'     card_header("Weight vs. Quarter Mile Time"),
#'     layout_sidebar(
#'       sidebar = sidebar(
#'         varSelectInput("var_x", "Compare to qsec:", mtcars[-7], "wt"),
#'         varSelectInput("color", "Color by:", mtcars[-7], "cyl"),
#'         position = "right"
#'       ),
#'       plotOutput("var_vs_qsec")
#'     )
#'   )
#' )
#'
#' server <- function(input, output) {
#'   for (var in c("cyl", "vs", "am", "gear", "carb")) {
#'     mtcars[[var]] <- as.factor(mtcars[[var]])
#'   }
#'
#'   output$gear <- renderPlot({
#'     ggplot(mtcars, aes(gear)) + geom_bar()
#'   })
#'
#'   output$carb <- renderPlot({
#'     ggplot(mtcars, aes(carb)) + geom_bar()
#'   })
#'
#'   output$var_vs_qsec <- renderPlot({
#'     req(input$var_x, input$color)
#'
#'     ggplot(mtcars) +
#'       aes(y = qsec, x = !!input$var_x, color = !!input$color) +
#'       geom_point()
#'   })
#' }
#'
#' shinyApp(ui, server)
page_fillable <- function(
  ...,
  padding = NULL,
  gap = NULL,
  fillable_mobile = FALSE,
  title = NULL,
  theme = bs_theme(),
  lang = NULL
) {
  page(
    title = title,
    theme = theme,
    lang = lang,
    class = "bslib-page-fill bslib-gap-spacing",
    class = if (!fillable_mobile) "bslib-flow-mobile",
    style = css(
      padding = validateCssPadding(padding),
      gap = validateCssUnit(gap)
    ),
    as_fillable_container(),
    tags$head(tags$style("html { height: 100%; }")),
    ...
  )
}

validateCssPadding <- function(padding = NULL) {
  if (is.null(padding)) {
    return(NULL)
  }
  paste(
    vapply(padding, validateCssUnit, character(1)),
    collapse = " "
  )
}

#' A sidebar page (i.e., dashboard)
#'
#' @description
#' Create a dashboard layout with a full-width header (`title`) and [sidebar()].
#'
#' @param ... UI elements to display in the 'main' content area (i.e., next to
#'   the `sidebar`). These arguments are passed to [layout_sidebar()], which has
#'   more details.
#' @param title A string, number, or [htmltools::tag()] child to display as the
#'   title (just above the `sidebar`).
#' @inheritParams layout_sidebar
#' @inheritParams page_fillable
#' @inheritParams page_navbar
#'
#' @export
#' @family Dashboard page layouts
#'
#' @seealso [layout_columns()] and [layout_column_wrap()] for laying out content
#'   into rows and columns.
#' @seealso [accordion()] for grouping related input controls in the `sidebar`.
#' @seealso [card()] for wrapping outputs in the 'main' content area.
#' @seealso [value_box()] for highlighting values.
#'
#' @references [Getting Started with Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'   on the bslib website.
#'
#' @examplesIf rlang::is_interactive()
#'
#' library(shiny)
#' library(ggplot2)
#' library(bslib)
#'
#' ui <- page_sidebar(
#'   title = "Example dashboard",
#'   sidebar = sidebar(
#'     varSelectInput("var", "Select variable", mtcars)
#'   ),
#'   card(
#'     full_screen = TRUE,
#'     card_header("My plot"),
#'     plotOutput("p")
#'   )
#' )
#'
#' server <- function(input, output) {
#'   output$p <- renderPlot({
#'     ggplot(mtcars) + geom_histogram(aes(!!input$var))
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
page_sidebar <- function(
  ...,
  sidebar = NULL,
  title = NULL,
  fillable = TRUE,
  fillable_mobile = FALSE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL
) {
  if (rlang::is_bare_character(title) || rlang::is_bare_numeric(title)) {
    title <- h1(title, class = "bslib-page-title navbar-brand")
  }

  navbar_title <-
    if (!is.null(title)) {
      div(
        class = "navbar navbar-static-top",
        div(title, class = "container-fluid")
      )
    }

  sidebar <- maybe_page_sidebar(sidebar)

  dots <- separate_arguments(...)

  layout_sidebar_args <- rlang::list2(
    sidebar = sidebar,
    fillable = fillable,
    border = FALSE,
    border_radius = FALSE,
    !!!dots$attribs,
    page_main_container(dots$children, fillable = fillable)
  )

  page_fillable(
    padding = 0,
    gap = 0,
    title = infer_window_title(title, window_title),
    theme = theme,
    lang = lang,
    fillable_mobile = fillable_mobile,
    class = "bslib-page-sidebar",
    navbar_title,
    rlang::exec(layout_sidebar, !!!layout_sidebar_args)
  )
}

page_main_container <- function(..., fillable = TRUE) {
  main <- tags$main(
    class = "bslib-page-main bslib-gap-spacing",
    ...
  )
  if (fillable) as_fill_carrier(main) else main
}

maybe_page_sidebar <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }
  if (!inherits(x, "sidebar")) {
    x <- sidebar(x)
  }

  # If `open` is not provided, choose a page-level default
  x$open <- x$open %||% sidebar_open_on(desktop = "open", mobile = "always")

  x
}

#' Multi-page app with a top navigation bar
#'
#' @description
#' Create a page that contains a top level navigation bar that can be used to
#' toggle a set of [nav_panel()] elements. Use this page layout to create the
#' effect of a multi-page app, where your app's content is broken up into
#' multiple "pages" that can be navigated to via the top navigation bar.
#'
#' @param fillable_mobile Whether or not `fillable` pages should fill the viewport's
#'   height on mobile devices (i.e., narrow windows).
#' @param underline `r lifecycle::badge("deprecated")` Please use
#'   [`navbar_options = navbar_options(underline=)`][navbar_options] instead.
#' @param window_title the browser window title. The default value, `NA`, means
#'   to use any character strings that appear in `title` (if none are found, the
#'   host URL of the page is displayed by default).
#' @inheritParams navset_bar
#' @inheritParams page
#'
#' @export
#' @family Dashboard page layouts
#'
#' @seealso [nav_panel()], [nav_menu()], and [nav_item()] for adding content
#'   sections and organizing or creating items in the navigation bar.
#' @seealso [layout_columns()] and [layout_column_wrap()] for laying out content
#'   into rows and columns.
#' @seealso [card()] for wrapping outputs in the 'main' content area.
#' @seealso [value_box()] for highlighting values.
#' @seealso [accordion()] for grouping related input controls in the `sidebar`.
#'
#' @references [Getting Started with Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html)
#'   on the bslib website.
#'
#' @examplesIf rlang::is_interactive()
#' library(shiny)
#' library(bslib)
#'
#' link_shiny <- tags$a(
#'   shiny::icon("github"), "Shiny",
#'   href = "https://github.com/rstudio/shiny",
#'   target = "_blank"
#' )
#' link_posit <- tags$a(
#'   shiny::icon("r-project"), "Posit",
#'   href = "https://posit.co",
#'   target = "_blank"
#' )
#'
#' ui <- page_navbar(
#'   title = "My App",
#'   nav_panel(title = "One", p("First page content.")),
#'   nav_panel(title = "Two", p("Second page content.")),
#'   nav_panel("Three", p("Third page content.")),
#'   nav_spacer(),
#'   nav_menu(
#'     title = "Links",
#'     align = "right",
#'     nav_item(link_shiny),
#'     nav_item(link_posit)
#'   )
#' )
#'
#' server <- function(...) { } # not used in this example
#'
#' shinyApp(ui, server)
page_navbar <- function(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  fillable_mobile = FALSE,
  gap = NULL,
  padding = NULL,
  header = NULL,
  footer = NULL,
  navbar_options = NULL,
  fluid = TRUE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL,
  position = deprecated(),
  bg = deprecated(),
  inverse = deprecated(),
  underline = deprecated(),
  collapsible = deprecated()
) {
  sidebar <- maybe_page_sidebar(sidebar)

  padding <- validateCssPadding(padding)
  gap <- validateCssUnit(gap)

  # Change behavior when called by Shiny
  # TODO: Coordinate with next bslib version bump in Shiny to use the new interface
  was_called_by_shiny <-
    isNamespaceLoaded("shiny") &&
    identical(rlang::caller_fn(), shiny::navbarPage)

  .navbar_options <- navbar_options_resolve_deprecated(
    options_user = navbar_options,
    position = position,
    bg = bg,
    inverse = inverse,
    collapsible = collapsible,
    underline = underline,
    .fn_caller = "page_navbar",
    .warn_deprecated = !was_called_by_shiny
  )

  # Default to fillable = F when this is called from shiny::navbarPage()
  # TODO: update shiny::navbarPage() to set fillable = FALSE and get rid of this hack
  if (missing(fillable) && was_called_by_shiny) {
    fillable <- FALSE
  }

  # If a sidebar is provided, we want the layout_sidebar(fill = TRUE) component
  # (which is a sibling of the <nav>) to always fill the page
  page_func <- if (isFALSE(fillable) && is.null(sidebar)) {
    page
  } else {
    function(...) {
      page_fillable(
        ...,
        fillable_mobile = fillable_mobile,
        padding = 0,
        gap = 0
      )
    }
  }

  navbar <- navs_bar_(
    ...,
    title = title,
    id = id,
    selected = selected,
    sidebar = sidebar,
    fillable = fillable,
    gap = gap,
    padding = padding,
    header = header,
    footer = footer,
    navbar_options = .navbar_options,
    fluid = fluid,
    theme = theme
  )

  page_func(
    title = infer_window_title(title, window_title),
    theme = theme,
    lang = lang,
    class = "bslib-page-navbar",
    class = if (!is.null(sidebar)) "has-page-sidebar",
    navbar
  )
}

# https://github.com/rstudio/shiny/issues/2310
infer_window_title <- function(title = NULL, window_title = NA) {
  if (!isTRUE(is.na(window_title))) {
    return(window_title)
  }

  if (!is.null(title)) {
    window_title <- unlist(find_characters(title))
    if (is.null(window_title)) {
      warning(
        "Unable to infer a `window_title` default from `title`. Consider providing a character string to `window_title`."
      )
    } else {
      window_title <- paste(window_title, collapse = " ")
    }
  }

  if (isTRUE(is.na(window_title))) NULL else window_title
}

# CPS (2023-02-09): Joe is currently working on a potentially
# more compelling contain_width() interface, so we'll punt on this for now
#
# #' Contain, pad, and align content
# #'
# #' @param ... A collection of [htmltools::tag()] children.
# #' @param size A size (i.e., max-width policy) for the container.
# #' @param bg A background color.
# #' @param class Additional CSS classes for the container.
# #'
# #' @references <https://getbootstrap.com/docs/5.3/layout/containers/>
# #'
# #' @export
# contain_width <- function(..., size = c("sm", "md", "lg", "xl", "xxl", # "fluid"), bg = NULL, class = NULL) {
#
#   size <- match.arg(size)
#
#   res <- div(
#     class = paste0("container-", size),
#     class = class,
#     # TODO: parseCssColors(), once it supports var() and !important
#     style = css(background_color = bg),
#     ...
#   )
#
#   as_fragment(
#     tag_require(res, version = 5, caller = "contain_width()")
#   )
# }

#> unlist(find_characters(div(h1("foo"), h2("bar"))))
#> [1] "foo" "bar"
find_characters <- function(x) {
  if (is.character(x)) {
    return(x)
  }
  if (inherits(x, "shiny.tag")) {
    return(lapply(x$children, find_characters))
  }
  if (is.list(x)) {
    return(lapply(x, find_characters))
  }
  NULL
}

utils::globalVariables("!!")

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import htmltools
#' @import sass
#' @importFrom grDevices col2rgb
#' @importFrom jquerylib jquery_core
#' @importFrom jsonlite fromJSON
#' @importFrom lifecycle deprecated
#' @importFrom rlang := list2 abort
#' @importFrom stats setNames na.omit
#' @importFrom tools file_path_sans_ext
#' @importFrom utils modifyList packageVersion download.file URLencode getFromNamespace head
## usethis namespace: end
NULL

#' @export
#' @importFrom htmltools css
htmltools::css

# For usethis::use_release_issue()
release_bullets <- function() {
  c(
    "Update static imports: `staticimports::import()`"
  )
}

#' Toolbar component
#'
#' @description
#' A toolbar which can contain buttons, inputs, and other UI elements in a small
#' form suitable for inclusion in card headers, footers, and other small places.
#'
#' bslib includes a complete example of toolbars and the many ways they can be used:
#'
#' ```r
#' shiny::runExample("toolbar", package = "bslib")
#' ```
#'
#' @section Cookbook:
#'
#' ## Toolbars in Card Headers and Footers
#'
#' Card headers and footers are a common places you might want to use toolbars. Toolbars
#' allow you to clearly show that a selection of inputs pertain to that
#' particular card. For example, this card uses multiple
#' [toolbar_input_select()] for filtering and sorting, along with toolbar
#' buttons and a "share" button in the toolbar in the card footer.
#'
#' ```r
#' card(
#'   full_screen = TRUE,
#'   card_header(
#'     "Sales Data",
#'     toolbar(
#'       align = "right",
#'       toolbar_input_select(
#'         id = "filter",
#'         label = "Filter",
#'         choices = c("All", "Active", "Inactive"),
#'         icon = icon("filter")
#'       ),
#'       toolbar_input_select(
#'         id = "sort",
#'         label = "Sort by",
#'         choices = c("Date", "Amount", "Customer"),
#'         icon = icon("sort")
#'       ),
#'       toolbar_divider(),
#'       toolbar_input_button(
#'         id = "refresh",
#'         label = "Refresh",
#'         icon = icon("arrows-rotate")
#'       ),
#'       toolbar_input_button(
#'         id = "export",
#'         label = "Export",
#'         icon = icon("download"),
#'         show_label = TRUE
#'       )
#'     )
#'   ),
#'   card_body(
#'     h3("Card Body Here"),
#'   ),
#'   card_footer(
#'     toolbar(
#'       align = "right",
#'       toolbar_input_button(
#'         id = "share_data",
#'         label = "Share",
#'         icon = icon("share-nodes"),
#'         show_label = TRUE,
#'         border = TRUE
#'       )
#'     )
#'   )
#' )
#' ```
#'
#' ## Toolbars in labels
#'
#' You can use toolbars in the labels of other Shiny inputs to add a composite
#' input with additional buttons or controls. The following example uses a
#' [toolbar()] in the label of a [shiny::numericInput()] to add increment and
#' decrement buttons next to the label:
#'
#' ```r
#' shiny::numericInput(
#'   "quantity",
#'   label = toolbar(
#'     "Quantity",
#'     toolbar_spacer(), # push buttons to the right
#'     toolbar_input_button("decrement", "Less", icon("minus")),
#'     toolbar_input_button("increment", "More", icon("plus"))
#'   ),
#'   value = 1,
#'   min = 0,
#'   max = 100
#' )
#' ```
#'
#' You can also use toolbars in the labels of text area inputs. For example,
#' this text editor uses a toolbar with formatting buttons:
#'
#' ```r
#' textAreaInput(
#'   "editor",
#'   label = toolbar(
#'     "Comment",
#'     toolbar_spacer(),
#'     toolbar_input_button("bold", label = "Bold", icon = icon("bold")),
#'     toolbar_input_button("italic", label = "Italic", icon = icon("italic")),
#'     toolbar_input_button("link", label = "Link", icon = icon("link"))
#'   ),
#'   value = "",
#'   rows = 5,
#'   placeholder = "Type your comment here..."
#' )
#' ```
#'
#' The [input_submit_textarea()] function from bslib allows you to create a
#' text area input with a submit button and an optional toolbar. Here is an
#' example of using a toolbar with formatting buttons and options:
#'
#' ```r
#' input_submit_textarea(
#'   "message",
#'   placeholder = "Type a message...",
#'   toolbar = toolbar(
#'     toolbar_input_button("attach", icon = icon("paperclip"), label = "Attach"),
#'     toolbar_input_button("emoji", icon = icon("face-smile"), label = "Emoji"),
#'     toolbar_divider(),
#'     toolbar_input_select(
#'       "format",
#'       label = "Format",
#'       choices = c("Plain", "Markdown", "HTML"),
#'       icon = icon("code")
#'     )
#'   )
#' )
#' ```
#'
#'
#' @examplesIf rlang::is_interactive()
#' # Minimal toolbar example
#' toolbar(
#'   toolbar_input_button(id = "view", icon = icon("eye"), label = "View"),
#'   toolbar_input_button(id = "save", icon = icon("save"), label = "Save"),
#'   toolbar_divider(),
#'   toolbar_input_select(
#'     id = "filter",
#'     label = "Filter",
#'     choices = c("All", "Active", "Inactive")
#'   )
#' )
#'
#' # Toolbar with text input
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   numericInput(
#'     "quantity",
#'     label = toolbar(
#'       "Quantity",
#'       toolbar_spacer(),
#'       toolbar_input_button("decrement", "Less", icon("minus")),
#'       toolbar_input_button("increment", "More", icon("plus"))
#'     ),
#'     value = 5,
#'     min = 0,
#'     max = 100
#'   ),
#'   verbatimTextOutput("value")
#' )
#'
#' server <- function(input, output, session) {
#'   output$value <- renderText({
#'     paste("Current value:", input$quantity)
#'   })
#'
#'   observeEvent(input$increment, {
#'     updateNumericInput(session, "quantity", value = input$quantity + 1)
#'   })
#'
#'   observeEvent(input$decrement, {
#'     updateNumericInput(session, "quantity", value = input$quantity - 1)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' # Toolbar with input_submit_textarea()
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   input_submit_textarea(
#'     "message",
#'     placeholder = "Type a message...",
#'     toolbar = toolbar(
#'       toolbar_input_button("attach", icon = icon("paperclip"), label = "Attach"),
#'       toolbar_input_button("emoji", icon = icon("face-smile"), label = "Emoji"),
#'       toolbar_divider(),
#'       toolbar_input_select(
#'         "format",
#'         label = "Format",
#'         choices = c("Plain", "Markdown", "HTML"),
#'         icon = icon("code")
#'       )
#'     )
#'   ),
#'   verbatimTextOutput("output")
#' )
#'
#' server <- function(input, output, session) {
#'   output$output <- renderText({
#'     req(input$message)
#'     paste("You said:", input$message)
#'   })
#'
#'   observeEvent(input$attach, {
#'     showNotification("Attach clicked!", duration = 2)
#'   })
#'
#'   observeEvent(input$emoji, {
#'     showNotification("Emoji clicked!", duration = 2)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @param ... UI elements for the toolbar.
#' @param align Determines if toolbar should be aligned to the `"right"` or
#'   `"left"`.
#' @param gap A CSS length unit defining the gap (i.e., spacing) between
#'   elements in the toolbar. Defaults to `0` (no gap).
#' @param width CSS width of the toolbar. Defaults to `NULL`, which will
#'   automatically set `width: 100%` when the toolbar is a direct child of a
#'   label element (e.g., when used in input labels). For [toolbar_spacer()] to
#'   push elements effectively, the toolbar needs `width: 100%` to expand and
#'   create space. Set this explicitly if you need to control the width in other
#'   contexts.
#'
#' @return Returns a toolbar element.
#'
#' @seealso [card_header()] for using toolbars in card headers/footers
#' @family toolbar components
#' @export
toolbar <- function(
  ...,
  align = c("right", "left"),
  gap = NULL,
  width = NULL
) {
  align <- rlang::arg_match(align)
  gap <- validateCssUnit(gap)
  width <- validateCssUnit(width)

  tag <- div(
    class = "bslib-toolbar bslib-gap-spacing",
    "data-align" = align,
    style = css(gap = gap, width = width),
    ...,
    component_dependencies()
  )

  tag_require(tag, version = 5, caller = "toolbar()")
  as_fragment(tag)
}

#' Add a toolbar input button
#'
#' @description
#' A button designed to fit well in small places such as in a [toolbar()].
#'
#' @section Updating toolbar buttons:
#'
#' You can dynamically update the appearance and enabled/disabled state
#' of a toolbar button on the client side using [update_toolbar_input_button()].
#' This function works similarly to [shiny::updateActionButton()].
#'
#' Note that you cannot change the `tooltip` or `border` parameters after the
#' button has been created, as these affect the button's structure and ARIA
#' attributes.
#' Please use [update_tooltip()] to update the text of the tooltip if one is
#' present.
#'
#' For example:
#'
#' ```r
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   card(
#'     card_header(
#'       "Toolbar Demo",
#'       toolbar(
#'         align = "right",
#'         toolbar_input_button("btn", label = "Click me", icon = icon("play")),
#'         toolbar_input_button(
#'           "task",
#'           label = "Do a Task",
#'           icon = icon("play-circle"),
#'           show_label = TRUE
#'         )
#'       )
#'     ),
#'     card_body(
#'       verbatimTextOutput("count")
#'     )
#'   )
#' )
#'
#' server <- function(input, output, session) {
#'   output$count <- renderPrint({
#'     input$btn
#'   })
#'
#'   observeEvent(input$btn, {
#'     update_toolbar_input_button(
#'       "btn",
#'       label = "Clicked!",
#'       icon = icon("check")
#'     )
#'     # Update the tooltip text
#'     update_tooltip("btn_tooltip", "Button was clicked!")
#'   })
#'
#'   # Handle task button - toggle between states
#'   observeEvent(input$task, {
#'     if (input$task %% 2 == 1) {
#'       # Show task is in progress
#'       update_toolbar_input_button(
#'         "task",
#'         label = "Task Running...",
#'         icon = icon("spinner")
#'       )
#'     } else {
#'       # Reset to original
#'       update_toolbar_input_button(
#'         "task",
#'         label = "Do a Task",
#'         icon = icon("play-circle"),
#'         session = session
#'       )
#'     }
#'   })
#' }
#'
#' shinyApp(ui, server)
#' ```
#'
#' @examplesIf rlang::is_interactive()
#' # Basic toolbar button
#' toolbar(
#'   toolbar_input_button(id = "view", icon = icon("eye"), label = "View"),
#'   toolbar_input_button(id = "save", icon = icon("save"), label = "Save")
#' )

#' @param id The input ID.
#' @param icon An icon. If provided without `show_label = TRUE`, only the icon
#'   will be visible.
#' @param label The input label. By default, `label` is not shown but is used by
#'   `tooltip`. Set `show_label = TRUE` to show the label (see `tooltip` for
#'   details on how this affects the tooltip behavior).
#' @param show_label Whether to show the label text. If `FALSE` (the default),
#'   only the icon is shown (if provided). If `TRUE`, the label text is shown
#'   alongside the icon. Note that `show_label` can be dynamically updated using
#'   [update_toolbar_input_button()].
#' @param tooltip Tooltip text to display when hovering over the input. Can be:
#'   * `TRUE` (default when `show_label = FALSE`) - shows a tooltip with the
#'     `label` text
#'   * `FALSE` (default when `show_label = TRUE`) - no tooltip
#'   * A character string - shows a tooltip with custom text
#'
#'   Defaults to `!show_label`. When a tooltip is created, it will have an ID of
#'   `"{id}_tooltip"` which can be used to update the tooltip text dynamically
#'   via [update_tooltip()].
#' @param ... Additional attributes to pass to the button.
#' @param disabled If `TRUE`, the button will not be clickable. Use
#'   [update_toolbar_input_button()] to dynamically enable/disable the button.
#' @param border Whether to show a border around the button.
#'
#' @return Returns a button suitable for use in a toolbar.
#'
#' @describeIn toolbar_input_button Create a toolbar button.
#' @family toolbar components
#' @export
toolbar_input_button <- function(
  id,
  label,
  icon = NULL,
  show_label = is.null(icon),
  tooltip = !show_label,
  ...,
  disabled = FALSE,
  border = FALSE
) {
  btn_type <-
    if (is.null(icon)) {
      if (!show_label) {
        rlang::abort(
          "If `show_label` is FALSE, `icon` must be provided."
        )
      }
      "label"
    } else {
      if (show_label) "both" else "icon"
    }

  # Validate that label has text for accessibility
  label_text <- paste(unlist(find_characters(label)), collapse = " ")
  # Verifies the label contains non-empty text
  if (!nzchar(trimws(label_text))) {
    warning(
      "Consider providing a non-empty string label for accessibility."
    )
  }

  label_id <- paste0("btn-label-", p_randomInt(1000, 10000))

  # We hide the label visually if `!show_label` but keep the label field for
  # use with `aria-labelledby`. This ensures that ARIA will always use the
  # label text. We found that screen readers will read out the icon's `aria-
  # label` even if it is a descendent of an element with `aria-hidden=true`.
  label_elem <- span(
    id = label_id,
    class = "bslib-toolbar-label",
    hidden = if (!show_label) NA else NULL,
    label
  )

  # And we wrap the icon to ensure that it is always treated as decorative
  icon_elem <- span(
    class = "bslib-toolbar-icon",
    `aria-hidden` = "true",
    style = "pointer-events: none",
    icon,
  )

  button <- shiny::actionButton(
    id,
    label = label_elem,
    icon = icon_elem,
    disabled = disabled,
    class = "bslib-toolbar-input-button btn-sm",
    class = if (!border) "border-0" else "border-1",
    "data-type" = btn_type,
    "aria-labelledby" = label_id,
    ...
  )

  # If tooltip is literally TRUE, use the label as the tooltip text.
  if (isTRUE(tooltip)) {
    tooltip <- label
  }
  if (isFALSE(tooltip)) {
    tooltip <- NULL
  }
  if (!is.null(tooltip)) {
    # Default placement is "bottom" for the toolbar case because otherwise the
    # tooltip ends up covering the neighboring buttons in the header/footer.
    button <- tooltip(
      button,
      tooltip,
      id = sprintf("%s_tooltip", id),
      placement = "bottom"
    )
  }

  button
}

#' @param session A Shiny session object (the default should almost always be
#'   used).
#'
#' @describeIn toolbar_input_button Update a toolbar button.
#' @export
update_toolbar_input_button <- function(
  id,
  label = NULL,
  show_label = NULL,
  icon = NULL,
  disabled = NULL,
  session = get_current_session()
) {
  label_text <- paste(unlist(find_characters(label)), collapse = " ")
  if (!nzchar(trimws(label_text))) {
    rlang::warn(
      "Consider providing a non-empty string label for accessibility."
    )
  }

  icon <- validateIcon(icon)
  icon_processed <- if (!is.null(icon)) processDeps(icon, session)
  label_processed <- if (!is.null(label)) processDeps(label, session)

  message <- dropNulls(list(
    label = label_processed,
    showLabel = show_label,
    icon = icon_processed,
    disabled = disabled
  ))

  session$sendInputMessage(id, message)
}

toolbar_input_button_input_handler <- function(value, shinysession, name) {
  # Treat like a standard action button for event handlers and input validation
  class(value) <- c("shinyActionButtonValue", class(value))
  value
}

#' Toolbar Input Select
#'
#' @description
#' Create a select input control that can be used to choose a single item from
#' a list of values, suitable for use within a [toolbar()].
#'
#' @section Updating toolbar select inputs:
#'
#' You can update the appearance and choices of a toolbar select input. This
#' function works similarly to [shiny::updateSelectInput()], but is specifically
#' designed for [toolbar_input_select()]. It allows you to update the select's
#' label, icon, choices, selected value, and label visibility from the server.
#'
#' Note that you cannot enable or disable the `tooltip` parameter after the
#' select has been created, only update the text of the tooltip.
#' When a tooltip is created for the select input, it will have an ID of
#' `"{id}_tooltip"` which can be used to update the tooltip text dynamically
#' via [update_tooltip()].
#'
#' For example:
#'
#' ```r
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fluid(
#'   toolbar(
#'     align = "right",
#'     toolbar_input_select(
#'       "select",
#'       label = "Choose",
#'       icon = bsicons::bs_icon("filter"),
#'       choices = c("A", "B", "C")
#'     ),
#'     toolbar_input_button(
#'       "change_choices",
#'       label = "Change Choices",
#'       show_label = TRUE,
#'       icon = bsicons::bs_icon("arrow-repeat")
#'     )
#'   ),
#'   verbatimTextOutput("value")
#' )
#'
#' server <- function(input, output, session) {
#'   output$value <- renderPrint({
#'     input$select
#'   })
#'
#'   observeEvent(input$change_choices, {
#'     update_toolbar_input_select(
#'       "select",
#'       label = "Pick one",
#'       choices = c("hi", "hello", "hey"),
#'       selected = "hello"
#'     )
#'     # Update the tooltip text
#'     update_tooltip("select_tooltip", "Choose your NEW option")
#'   })
#' }
#'
#' shinyApp(ui, server)
#' ```
#'
#' @examplesIf rlang::is_interactive()
#' toolbar(
#'   align = "right",
#'   toolbar_input_select(
#'     id = "select",
#'     label = "Choose option",
#'     choices = c("Option 1", "Option 2", "Option 3"),
#'     selected = "Option 2"
#'   )
#' )
#'
#' @param selected The initially selected value. If not provided on input creation, the
#'   first choice will be selected by default.
#'   If provided in `update_toolbar_input_select()` with a new set of
#'   `choices`, it will replace the currently selected value.
#' @param ... Additional named arguments passed as attributes to the outer
#'   container div.
#' @inheritParams toolbar_input_button
#' @inheritParams shiny::selectInput
#'
#' @details
#' When a tooltip is created for the select input, it will have an ID of
#' `"{id}_tooltip"` which can be used to update the tooltip text dynamically
#' via [update_tooltip()].
#'
#' @return Returns a select input control suitable for use in a toolbar.
#'
#' @describeIn toolbar_input_select Create a toolbar select input.
#' @family toolbar components
#' @export
toolbar_input_select <- function(
  id,
  label,
  choices,
  ...,
  selected = NULL,
  icon = NULL,
  show_label = FALSE,
  tooltip = !show_label
) {
  # Validate that ... contains only named arguments
  dots <- separate_arguments(...)
  if (length(dots$children) > 0) {
    rlang::abort("All arguments in `...` must be named.")
  }

  # Validate that label is a non-empty string
  # TODO: Use `check_string()`
  if (!is.character(label) || length(label) != 1 || !nzchar(trimws(label))) {
    rlang::abort("`label` must be a non-empty string.")
  }

  # Restore input for bookmarking
  selected <- shiny::restoreInput(id = id, default = selected)

  # use_first_choice = TRUE for initial creation so that if the user does not
  # provide an initial selected value, we select the first choice
  processed <- process_choices_selected(
    choices,
    selected,
    id,
    use_first_choice = TRUE
  )

  if (!is.null(processed$error)) {
    rlang::abort(processed$error)
  }

  select_tag <- tags$select(
    id = id,
    class = "form-select form-select-sm",
    `data-shiny-no-bind-input` = NA,
    processed$options
  )

  # Add optional icon before the select
  icon_elem <- span(
    icon,
    class = "bslib-toolbar-icon",
    `aria-hidden` = "true",
    style = "pointer-events: none",
    `role` = "none",
    tabindex = "-1"
  )

  label_elem <- tags$label(
    # shiny::selectInput() append `-label` to id for the label `for` attribute
    id = sprintf("%s-label", id),
    class = "control-label",
    `for` = id,
    icon_elem,
    tags$span(
      class = "bslib-toolbar-label",
      class = if (!show_label) "visually-hidden",
      label
    )
  )

  if (isTRUE(tooltip)) {
    # If tooltip is literally TRUE, use the label as the tooltip text, but hide
    # it from screen readers since it repeats the label content.
    tooltip <- tags$span(label, `aria-hidden` = "true")
  }
  if (isFALSE(tooltip)) {
    tooltip <- NULL
  }
  if (!is.null(tooltip)) {
    select_tag <- bslib::tooltip(
      select_tag,
      tooltip,
      id = sprintf("%s_tooltip", id),
      placement = "bottom"
    )
  }

  div(
    id = id,
    class = "bslib-toolbar-input-select shiny-input-container",
    !!!dots$attribs,
    label_elem,
    select_tag
  )
}

#' @param session A Shiny session object (the default should almost always be
#'   used).
#'
#' @describeIn toolbar_input_select Update a toolbar select input.
#' @export
update_toolbar_input_select <- function(
  id,
  label = NULL,
  show_label = NULL,
  choices = NULL,
  selected = NULL,
  icon = NULL,
  session = get_current_session()
) {
  # Label can be null if there is no update, but if supplied it must be valid
  if (!(is.null(label) || (rlang::is_string(label) && nzchar(trimws(label))))) {
    rlang::abort("`label` must be a non-empty string.")
  }
  icon <- validateIcon(icon)
  icon_processed <- if (!is.null(icon)) processDeps(icon, session)
  label_processed <- if (!is.null(label)) processDeps(label, session)

  # Process and validate choices and selected
  processed <- process_choices_selected(
    choices,
    selected,
    id,
    use_first_choice = FALSE
  )

  if (!is.null(processed$error)) {
    rlang::warn(processed$error)
  }

  message <- dropNulls(list(
    label = label_processed,
    showLabel = show_label,
    icon = icon_processed,
    options = processed$options,
    value = processed$value
  ))

  session$sendInputMessage(id, message)
}

# Helper function to normalize choices using Shiny's choicesWithNames
normalize_choices <- function(choices) {
  choicesWithNames <- asNamespace("shiny")[["choicesWithNames"]]
  choicesWithNames(choices)
}

# Helper function to extract all choice values from normalized choices
# Note: expects choices to already be normalized via normalize_choices()
extract_choice_values <- function(choices_normalized) {
  # If it's a list of lists (grouped choices), flatten it and extract the values
  if (
    is.list(choices_normalized) &&
      any(vapply(choices_normalized, is.list, logical(1)))
  ) {
    values <- unlist(
      lapply(choices_normalized, function(group) {
        if (is.list(group)) {
          unname(group)
        } else {
          group
        }
      }),
      use.names = FALSE
    )
  } else {
    # Flat choices - extract values
    values <- unname(choices_normalized)
  }

  as.character(values)
}

# Helper function to process and validate choices and selected
# Returns a list with:
#   - options: HTML string for <option> elements (or NULL)
#   - value: selected value as character (or NULL)
#   - error: error message if validation failed (NULL if no error)
# @param use_first_choice If TRUE and selected is NULL, use firstChoice().
#   If FALSE, leave selected as NULL (for updates where we want to keep current value)
process_choices_selected <- function(
  choices,
  selected,
  inputId,
  use_first_choice = TRUE
) {
  if (is.null(choices) && is.null(selected)) {
    return(list(options = NULL, value = NULL, error = NULL))
  }

  # Normalize choices once if provided
  choices_normalized <- NULL
  if (!is.null(choices)) {
    choices_normalized <- normalize_choices(choices)
  }

  # If selected is NULL and choices are provided, optionally select the first choice
  if (is.null(selected) && !is.null(choices_normalized) && use_first_choice) {
    firstChoice <- asNamespace("shiny")[["firstChoice"]]
    selected <- firstChoice(choices_normalized)
  }

  # Validate selected if provided
  error_msg <- NULL
  selected_validated <- selected

  if (!is.null(selected)) {
    if (length(selected) != 1) {
      error_msg <- "`selected` must be a single value, not a vector."
      selected_validated <- NULL
    } else if (is.null(choices)) {
      error_msg <- "`selected` cannot be set without `choices`."
      selected_validated <- NULL
    } else {
      # Extract all valid choice values from normalized choices
      choice_values <- extract_choice_values(choices_normalized)
      if (!as.character(selected) %in% choice_values) {
        error_msg <- sprintf(
          "`selected` value '%s' is not in `choices`.",
          as.character(selected)
        )
        selected_validated <- NULL
      }
    }
  }

  options_html <- NULL
  value <- NULL

  # Process choices into HTML options
  if (!is.null(choices_normalized)) {
    options_html <- HTML(as.character(selectOptions(
      choices_normalized,
      selected_validated,
      inputId = inputId
    )))
  }

  # Process selected value
  if (!is.null(selected_validated)) {
    value <- as.character(selected_validated)
  }

  list(
    options = options_html,
    value = value,
    error = error_msg
  )
}

# This function was copied from shiny's `input-select.R` with a small change
selectOptions <- function(
  choices,
  selected = NULL,
  inputId,
  perfWarning = FALSE
) {
  if (length(choices) >= 1000) {
    # CHANGED: This warning differs to remove mention of sever-side options
    rlang::warn(
      sprintf(
        paste0(
          "Select input `%s` contains a large number of options; ",
          "this may cause performance issues."
        ),
        inputId
      )
    )
  }

  html <- mapply(
    choices,
    names(choices),
    FUN = function(choice, label) {
      if (is.list(choice)) {
        # If sub-list, create an optgroup and recurse into the sublist
        sprintf(
          '<optgroup label="%s">\n%s\n</optgroup>',
          htmlEscape(label, TRUE),
          selectOptions(choice, selected, inputId, perfWarning)
        )
      } else {
        # If single item, just return option string
        sprintf(
          '<option value="%s"%s>%s</option>',
          htmlEscape(choice, TRUE),
          if (choice %in% selected) " selected" else "",
          htmlEscape(label)
        )
      }
    }
  )

  HTML(paste(html, collapse = "\n"))
}

#' Toolbar: Add a divider or spacer to a toolbar
#'
#' @description
#' [toolbar_divider()] creates a visual divider line with customizable and fixed
#' width and spacing between toolbar elements. [toolbar_spacer()] creates empty
#' space that expands to push adjacent toolbar elements apart as much as
#' possible.
#'
#' @param width A CSS length unit specifying the width of the divider line.
#'   Defaults to `"2px"` for a sensible dividing line. Pass `0px` for no
#'   divider line.
#' @param gap A CSS length unit defining the spacing around the divider.
#'   Defaults to `"1rem"` for sensible fixed spacing.
#' @param ... Ignored, reserved for future use and to require named arguments
#'   in [toolbar_divider()].
#'
#' @examplesIf rlang::is_interactive()
#' toolbar(
#'   toolbar_input_button(id = "left1", label = "Left"),
#'   toolbar_divider(),
#'   toolbar_input_button(id = "right1", label = "Right")
#' )
#'
#' toolbar(
#'   toolbar_input_button(id = "a", label = "A"),
#'   toolbar_divider(width = "5px", gap = "20px"),
#'   toolbar_input_button(id = "b", label = "B")
#' )
#'
#' toolbar(
#'   toolbar_input_button(id = "previous", label = "Previous"),
#'   toolbar_spacer(),
#'   toolbar_input_button(id = "next", label = "Next")
#' )
#'
#' @family toolbar components
#' @describeIn toolbar_divider Create a dividing line and fixed space between
#'   toolbar elements.
#' @export
toolbar_divider <- function(..., width = NULL, gap = NULL) {
  rlang::check_dots_empty()
  width <- validateCssUnit(width)
  gap <- validateCssUnit(gap)

  div(
    class = "bslib-toolbar-divider",
    style = css(
      # Sets the overall width of divider space
      `--_divider-gap` = gap,
      # Sets the width of the pseudo-element divider line, defaults to 2px
      `--_divider-width` = width
    ),
    `aria-hidden` = "true"
  )
}

#' @describeIn toolbar_divider Create empty, expanding space between toolbar
#'   elements. Note that for the spacer to push elements effectively, the parent
#'   toolbar needs `width: 100%` (which is automatically applied when the toolbar
#'   is a direct child of a label element). If the spacer doesn't appear to work,
#'   you may need to set `width = "100%"` explicitly on the [toolbar()].
#' @export
toolbar_spacer <- function() {
  div(class = "bslib-toolbar-spacer")
}

# Tests for toolbar container #
test_that("toolbar() basic attributes and defaults", {
  tb <- as.tags(toolbar(htmltools::span("Test")))
  expect_match(htmltools::tagGetAttribute(tb, "class"), "bslib-toolbar")
  expect_match(htmltools::tagGetAttribute(tb, "data-align"), "right")
  expect_snapshot_html(
    toolbar("Item 1", "Item 2")
  )
  expect_snapshot_html(
    toolbar("Item 1", "Item 2", gap = "10px")
  )
})

test_that("toolbar() aligns correctly", {
  tb <- as.tags(toolbar(align = "left"))
  expect_equal(htmltools::tagGetAttribute(tb, "data-align"), "left")
  expect_snapshot_html(
    toolbar("Item 1", "Item 2", align = "left")
  )
  expect_snapshot_html(
    toolbar("Item 1", "Item 2", align = "right")
  )
  expect_error(toolbar("x", align = "center"))
})


# Tests for toolbar_input_button() #
test_that("toolbar_input_button() tests", {
  # Label-only Button
  btn_label <- toolbar_input_button(
    id = "test_btn",
    label = "Click me",
    show_label = TRUE
  )
  expect_match(
    htmltools::tagGetAttribute(btn_label, "class"),
    "bslib-toolbar-input-button"
  )
  expect_match(htmltools::tagGetAttribute(btn_label, "class"), "btn-sm")
  expect_match(htmltools::tagGetAttribute(btn_label, "data-type"), "label")

  expect_snapshot_html(
    toolbar_input_button(
      id = "label_only",
      label = "Click me",
      show_label = TRUE
    )
  )

  btn_icon <- toolbar_input_button(
    id = "test_btn",
    label = "Click me",
    icon = shiny::icon("star"),
  )
  # Button is wrapped in tooltip by default, use tagQuery to extract it
  btn_icon_tag <- tagQuery(as.tags(btn_icon))$find("button")$selectedTags()[[1]]

  expect_match(
    htmltools::tagGetAttribute(btn_icon_tag, "class"),
    "bslib-toolbar-input-button"
  )
  expect_match(htmltools::tagGetAttribute(btn_icon_tag, "class"), "btn-sm")
  expect_match(htmltools::tagGetAttribute(btn_icon_tag, "data-type"), "icon")
  expect_snapshot_html(
    toolbar_input_button(
      id = "test_btn",
      label = "Click me",
      icon = shiny::icon("star"),
    )
  )

  btn_both <- toolbar_input_button(
    id = "test_btn",
    label = "Click me",
    icon = shiny::icon("star"),
    show_label = TRUE
  )
  expect_match(
    htmltools::tagGetAttribute(btn_both, "class"),
    "bslib-toolbar-input-button"
  )
  expect_match(htmltools::tagGetAttribute(btn_both, "class"), "btn-sm")
  expect_match(htmltools::tagGetAttribute(btn_both, "data-type"), "both")
  expect_snapshot_html(
    toolbar_input_button(
      id = "test_btn",
      label = "Click me",
      icon = shiny::icon("star"),
      show_label = TRUE
    )
  )
})


test_that("toolbar_input_button() disabled parameter", {
  expect_snapshot_html(
    toolbar_input_button(
      id = "disabled_btn",
      label = "Disabled",
      disabled = TRUE,
      show_label = TRUE
    )
  )
  expect_snapshot_html(
    toolbar_input_button(
      id = "enabled_btn",
      label = "Enabled",
      disabled = FALSE,
      show_label = TRUE
    )
  )
})

test_that("toolbar_input_button() border parameter", {
  expect_snapshot_html(
    toolbar_input_button(
      id = "no_border",
      label = "No Border",
      border = FALSE,
      show_label = TRUE
    )
  )

  expect_snapshot_html(
    toolbar_input_button(
      id = "with_border",
      label = "With Border",
      border = TRUE,
      show_label = TRUE
    )
  )
})


test_that("toolbar_input_button() tooltip parameter", {
  # Default: show_label = FALSE means tooltip = TRUE (shows label in tooltip)
  expect_snapshot_html(
    toolbar_input_button(
      id = "tooltip_default",
      label = "Help",
      icon = shiny::icon("question")
    )
  )

  # Explicit tooltip = FALSE disables tooltip
  expect_snapshot_html(
    toolbar_input_button(
      id = "tooltip_false",
      label = "No Tooltip",
      icon = shiny::icon("question"),
      tooltip = FALSE
    )
  )

  # Custom tooltip text
  expect_snapshot_html(
    toolbar_input_button(
      id = "tooltip_custom",
      label = "Help",
      icon = shiny::icon("question"),
      tooltip = "Click for assistance"
    )
  )

  # show_label = TRUE means tooltip = FALSE by default
  btn_no_tooltip <- toolbar_input_button(
    id = "label_visible",
    label = "Visible Label",
    show_label = TRUE
  )
  expect_false(inherits(btn_no_tooltip, "bslib_tooltip"))

  # But you can explicitly add tooltip when show_label = TRUE
  expect_snapshot_html(
    toolbar_input_button(
      id = "both_label_tooltip",
      label = "Save",
      icon = shiny::icon("save"),
      show_label = TRUE,
      tooltip = "Save your work"
    )
  )
})

test_that("toolbar_input_button() validates label for accessibility", {
  # Empty label should warn when show_label = FALSE
  expect_warning(
    toolbar_input_button(
      id = "btn",
      label = "",
      icon = shiny::icon("star")
    ),
    "non-empty string label"
  )

  # Whitespace-only label should warn
  expect_warning(
    toolbar_input_button(
      id = "btn",
      label = "   ",
      icon = shiny::icon("star")
    ),
    "non-empty string label"
  )

  # Empty tag label should warn
  expect_warning(
    toolbar_input_button(
      id = "btn",
      label = span(""),
      icon = shiny::icon("star")
    ),
    "non-empty string label"
  )

  # Valid label should not warn
  expect_no_warning(
    toolbar_input_button(
      id = "btn",
      label = "Click me",
      icon = shiny::icon("star")
    )
  )

  # Label with tag containing text should not warn
  expect_no_warning(
    toolbar_input_button(
      id = "btn",
      label = span("Click me"),
      icon = shiny::icon("star")
    )
  )
})

# Tests for toolbar_divider() #
test_that("toolbar_divider() creates divider element", {
  expect_snapshot_html(
    toolbar_divider()
  )
  expect_snapshot_html(
    toolbar_divider(gap = "20px")
  )
  expect_snapshot_html(
    toolbar_divider(width = "5px", gap = "2rem")
  )
})

test_that("toolbar_divider() validates dots are empty", {
  expect_error(
    toolbar_divider("fake"),
    "must be empty"
  )
})

# Additional Toolbar Input Select Tests #

test_that("toolbar_input_select() accepts named attributes in ...", {
  tis <- toolbar_input_select(
    id = "select",
    label = "Choose option",
    choices = c("Option 1", "Option 2", "Option 3"),
    class = "bg-success-subtle",
    `data-test` = "custom",
    tooltip = FALSE
  )

  # Check that the outer div (with bslib-toolbar-input-select class) has the custom attributes
  expect_match(htmltools::tagGetAttribute(tis, "class"), "bg-success-subtle")
  expect_equal(htmltools::tagGetAttribute(tis, "data-test"), "custom")
})

test_that("toolbar_input_select() rejects unnamed arguments in ...", {
  expect_error(
    toolbar_input_select(
      id = "select",
      label = "Choose option",
      choices = c("Option 1", "Option 2", "Option 3"),
      "bad"
    ),
    "All arguments in `...` must be named"
  )
})

test_that("toolbar_input_select() has proper label structure", {
  tis <- as.tags(
    toolbar_input_select(
      id = "select",
      label = "Choose option",
      choices = c("Option 1", "Option 2", "Option 3"),
      tooltip = FALSE
    )
  )

  # Check that a <label> element exists with proper attributes
  label_elem <- tagQuery(tis)$find("label")$selectedTags()[[1]]
  expect_true(!is.null(label_elem))

  # Label should have id matching pattern "{id}-label"
  label_id <- htmltools::tagGetAttribute(label_elem, "id")
  expect_equal(label_id, "select-label")

  # Label should have for attribute pointing to select
  label_for <- htmltools::tagGetAttribute(label_elem, "for")
  expect_equal(label_for, "select")

  # Find the label text span
  label_spans <- tagQuery(label_elem)$find(
    "span.bslib-toolbar-label"
  )$selectedTags()
  expect_true(length(label_spans) > 0)

  label_text_span <- label_spans[[1]]
  # With show_label=FALSE (default), label should be visually hidden
  expect_match(
    htmltools::tagGetAttribute(label_text_span, "class"),
    "visually-hidden"
  )

  # Check label text content
  label_text <- as.character(label_text_span$children[[1]])
  expect_equal(label_text, "Choose option")
})

test_that("toolbar_input_select() markup snapshots", {
  expect_snapshot_html(
    toolbar_input_select(
      id = "select1",
      label = "Basic select",
      choices = c("A", "B", "C"),
      tooltip = FALSE
    )
  )

  expect_snapshot_html(
    toolbar_input_select(
      id = "select2",
      label = "Select with selected",
      choices = c("Option 1", "Option 2", "Option 3"),
      selected = "Option 2",
      tooltip = FALSE
    )
  )

  expect_snapshot_html(
    toolbar_input_select(
      id = "select3",
      label = "Select with custom class",
      choices = c("X", "Y", "Z"),
      class = "bg-success-subtle",
      "style" = "width: 400px",
      tooltip = FALSE
    )
  )
})

test_that("toolbar_input_select() handles grouped choices", {
  grouped_select <- toolbar_input_select(
    id = "grouped",
    label = "Grouped select",
    choices = list(
      "Group A" = c("A1", "A2"),
      "Group B" = c("B1", "B2")
    ),
    tooltip = FALSE
  )

  expect_snapshot_html(grouped_select)
})

test_that("toolbar_input_select() handles named choices", {
  named_select <- toolbar_input_select(
    id = "named",
    label = "Named choices",
    choices = c("Label 1" = "val1", "Label 2" = "val2"),
    tooltip = FALSE
  )

  html_output <- as.character(as.tags(named_select))
  expect_match(html_output, "Label 1")
  expect_match(html_output, "val1")
  expect_match(html_output, "Label 2")
  expect_match(html_output, "val2")
})

test_that("toolbar_input_select() respects selected parameter", {
  select_with_default <- as.tags(
    toolbar_input_select(
      id = "default",
      label = "With default",
      choices = c("A", "B", "C"),
      selected = "B",
      tooltip = FALSE
    )
  )

  html_output <- as.character(select_with_default)
  expect_match(html_output, '<option value="B" selected>B</option>')
})

test_that("toolbar_input_select() selects first choice by default", {
  select_no_default <- as.tags(
    toolbar_input_select(
      id = "no_default",
      label = "No default",
      choices = c("X", "Y", "Z"),
      tooltip = FALSE
    )
  )

  html_output <- as.character(select_no_default)
  expect_match(html_output, '<option value="X" selected>X</option>')
})

test_that("toolbar_input_select() validates label parameter", {
  expect_error(
    toolbar_input_select(
      id = "test",
      label = "",
      choices = c("A", "B"),
      tooltip = FALSE
    ),
    "`label` must be a non-empty string"
  )

  expect_error(
    toolbar_input_select(
      id = "test",
      label = c("A", "B"),
      choices = c("A", "B"),
      tooltip = FALSE
    ),
    "`label` must be a non-empty string"
  )

  expect_error(
    toolbar_input_select(
      id = "test",
      label = 123,
      choices = c("A", "B"),
      tooltip = FALSE
    ),
    "`label` must be a non-empty string"
  )
})

test_that("toolbar_input_select() has correct classes", {
  select <- as.tags(
    toolbar_input_select(
      id = "test",
      label = "Test",
      choices = c("A", "B"),
      tooltip = FALSE
    )
  )

  # Check outer div has correct classes
  expect_match(
    htmltools::tagGetAttribute(select, "class"),
    "bslib-toolbar-input-select"
  )
  expect_match(
    htmltools::tagGetAttribute(select, "class"),
    "shiny-input-container"
  )

  # Check select element has Bootstrap classes
  select_elem <- tagQuery(select)$find("select")$selectedTags()[[1]]
  expect_match(htmltools::tagGetAttribute(select_elem, "class"), "form-select")
  expect_match(
    htmltools::tagGetAttribute(select_elem, "class"),
    "form-select-sm"
  )
})

test_that("toolbar_input_select() tooltip parameter", {
  # Default has tooltip=TRUE, uses label text for tooltip text
  select_default <- as.tags(
    toolbar_input_select(
      id = "default",
      label = "Has label as tooltip",
      choices = c("A", "B")
    )
  )
  html_output <- as.character(select_default)
  expect_true(grepl("bslib-tooltip", html_output))

  # With tooltip = FALSE explicitly - no bslib-tooltip wrapper
  select_tooltip_false <- as.tags(
    toolbar_input_select(
      id = "tooltip_false",
      label = "Explicitly no tooltip",
      choices = c("A", "B"),
      tooltip = FALSE
    )
  )
  html_output_false <- as.character(select_tooltip_false)
  expect_false(grepl("bslib-tooltip", html_output_false))

  # With tooltip = TRUE - uses label as tooltip text
  expect_snapshot_html(
    toolbar_input_select(
      id = "tooltip_true",
      label = "My Select Label",
      choices = c("A", "B"),
      tooltip = TRUE
    )
  )

  # With tooltip - wrapped in bslib-tooltip
  expect_snapshot_html(
    toolbar_input_select(
      id = "with_tooltip",
      label = "With tooltip",
      choices = c("A", "B"),
      tooltip = "This is helpful information"
    )
  )
})

test_that("toolbar_input_select() icon parameter", {
  # Without icon - no icon element
  select_no_icon <- as.tags(
    toolbar_input_select(
      id = "no_icon",
      label = "No icon",
      choices = c("A", "B"),
      tooltip = FALSE
    )
  )
  html_output <- as.character(select_no_icon)
  expect_false(grepl("bslib-toolbar-input-select-icon", html_output))

  # With icon
  expect_snapshot_html(
    toolbar_input_select(
      id = "with_icon",
      label = "With icon",
      choices = c("A", "B"),
      icon = shiny::icon("filter"),
      tooltip = FALSE
    )
  )

  # With both icon and tooltip
  expect_snapshot_html(
    toolbar_input_select(
      id = "icon_tooltip",
      label = "Icon and tooltip",
      choices = c("A", "B"),
      icon = shiny::icon("star"),
      tooltip = "Select an option"
    )
  )
})

# Tests to detect if the functions we import from Shiny have changed #
test_that("Shiny's firstChoice() function maintains expected behavior", {
  # These tests verify that Shiny's internal firstChoice() function
  # continues to work as expected for toolbar_input_select()
  # Note we don't test on vectors here because choicesWithNames() ensures we
  # only have lists when passed to firstChoice()

  firstChoice <- asNamespace("shiny")[["firstChoice"]]

  # Simple vector - should return first element
  expect_equal(firstChoice(c("A", "B", "C")), "A")

  # Named vector - should return first value (not name)
  expect_equal(firstChoice(c("Label 1" = "val1", "Label 2" = "val2")), "val1")

  # Nested list - should recursively find first non-list element
  nested <- list(
    "Group A" = list("A1", "A2"),
    "Group B" = list("B1", "B2")
  )
  expect_equal(firstChoice(nested), "A1")

  # Nested list with named choices
  nested_named <- list(
    "Group A" = list("Label A1" = "valA1", "Label A2" = "valA2"),
    "Group B" = list("Label B1" = "valB1")
  )
  expect_equal(firstChoice(nested_named), "valA1")

  # Empty choices should return NULL or empty
  expect_true(
    is.null(firstChoice(character(0))) ||
      identical(firstChoice(character(0)), character(0))
  )
})

test_that("Shiny's choicesWithNames() function maintains expected behavior", {
  # These tests verify that Shiny's internal choicesWithNames() function
  # continues to work as expected for toolbar_input_select()

  choicesWithNames <- asNamespace("shiny")[["choicesWithNames"]]

  # Unnamed list - names should equal values
  result1 <- choicesWithNames(list("A", "B", "C"))
  expect_equal(names(result1), c("A", "B", "C"))
  expect_equal(as.character(result1), c("A", "B", "C"))

  # Named list - preserve names and values
  result2 <- choicesWithNames(list("Label 1" = "val1", "Label 2" = "val2"))
  expect_equal(names(result2), c("Label 1", "Label 2"))
  expect_equal(as.character(result2), c("val1", "val2"))

  # Partially named list - use value as name where missing
  result3 <- choicesWithNames(list("Label 1" = "val1", "val2"))
  expect_equal(names(result3), c("Label 1", "val2"))
  expect_equal(as.character(result3), c("val1", "val2"))

  # Grouped choices (list) - should preserve structure
  grouped <- list(
    "Group A" = c("A1", "A2"),
    "Group B" = c("B1", "B2")
  )
  result4 <- choicesWithNames(grouped)
  expect_true(is.list(result4))
  expect_equal(names(result4), c("Group A", "Group B"))
  expect_equal(names(result4[["Group A"]]), c("A1", "A2"))
  expect_equal(as.character(result4[["Group A"]]), c("A1", "A2"))

  # Grouped with named choices
  grouped_named <- list(
    "Group A" = c("Label A1" = "valA1", "Label A2" = "valA2")
  )
  result5 <- choicesWithNames(grouped_named)
  expect_equal(names(result5[["Group A"]]), c("Label A1", "Label A2"))
  expect_equal(as.character(result5[["Group A"]]), c("valA1", "valA2"))
})

test_that("bslib::selectOptions() matches shiny::selectOptions() output", {
  # These tests verify that bslib's selectOptions() function produces
  # the same HTML output as Shiny's selectOptions() function
  # NOTE: All choices are preprocessed (as if by choicesWithNames())

  bslib_selectOptions <- asNamespace("bslib")[["selectOptions"]]
  shiny_selectOptions <- asNamespace("shiny")[["selectOptions"]]

  # Simple unnamed choices (preprocessed)
  choices1 <- list(A = "A", B = "B", C = "C")
  bslib_out1 <- as.character(bslib_selectOptions(choices1, inputId = "test1"))
  shiny_out1 <- as.character(shiny_selectOptions(choices1, inputId = "test1"))
  expect_equal(bslib_out1, shiny_out1)

  # Named choices (preprocessed)
  choices2 <- list(`Label A` = "valA", `Label B` = "valB", `Label C` = "valC")
  bslib_out2 <- as.character(bslib_selectOptions(choices2, inputId = "test2"))
  shiny_out2 <- as.character(shiny_selectOptions(choices2, inputId = "test2"))
  expect_equal(bslib_out2, shiny_out2)

  # With selected value
  bslib_out3 <- as.character(bslib_selectOptions(
    choices1,
    selected = "B",
    inputId = "test3"
  ))
  shiny_out3 <- as.character(shiny_selectOptions(
    choices1,
    selected = "B",
    inputId = "test3"
  ))
  expect_equal(bslib_out3, shiny_out3)

  # Grouped choices (preprocessed)
  grouped <- list(
    `Group 1` = list(A1 = "A1", A2 = "A2", A3 = "A3"),
    `Group 2` = list(B1 = "B1", B2 = "B2")
  )
  bslib_out5 <- as.character(bslib_selectOptions(grouped, inputId = "test5"))
  shiny_out5 <- as.character(shiny_selectOptions(grouped, inputId = "test5"))
  expect_equal(bslib_out5, shiny_out5)

  # Grouped with named choices (preprocessed)
  grouped_named <- list(
    `Group A` = list(`Label A1` = "valA1", `Label A2` = "valA2"),
    `Group B` = list(`Label B1` = "valB1", `Label B2` = "valB2")
  )
  bslib_out6 <- as.character(bslib_selectOptions(
    grouped_named,
    inputId = "test6"
  ))
  shiny_out6 <- as.character(shiny_selectOptions(
    grouped_named,
    inputId = "test6"
  ))
  expect_equal(bslib_out6, shiny_out6)

  # Grouped with selected value
  bslib_out7 <- as.character(bslib_selectOptions(
    grouped,
    selected = "A2",
    inputId = "test7"
  ))
  shiny_out7 <- as.character(shiny_selectOptions(
    grouped,
    selected = "A2",
    inputId = "test7"
  ))
  expect_equal(bslib_out7, shiny_out7)

  # Special characters that need escaping (preprocessed)
  choices_special <- list(
    `Label <with> HTML` = "val1",
    `Label & ampersand` = "val2"
  )
  bslib_out8 <- as.character(bslib_selectOptions(
    choices_special,
    inputId = "test8"
  ))
  shiny_out8 <- as.character(shiny_selectOptions(
    choices_special,
    inputId = "test8"
  ))
  expect_equal(bslib_out8, shiny_out8)
})

# Tests for update functions #

test_that("update_toolbar_input_select() validates label parameter", {
  session <- list(sendInputMessage = function(id, message) {
    stop("sendInputMessage should not be called")
  })

  # Empty string label should error (validation happens before session is used)
  expect_snapshot(error = TRUE, {
    update_toolbar_input_select("test_id", label = "", session = session)
  })

  # Whitespace-only label should error
  expect_snapshot(error = TRUE, {
    update_toolbar_input_select("test_id", label = "   ", session = session)
  })

  # Non-character label should error
  expect_snapshot(error = TRUE, {
    update_toolbar_input_select("test_id", label = 123, session = session)
  })

  # Multiple strings should error
  expect_snapshot(error = TRUE, {
    update_toolbar_input_select(
      "test_id",
      label = c("A", "B"),
      session = session
    )
  })
})

test_that("toolbar_input_select() validates selected is in choices", {
  # Invalid selected value should error
  expect_snapshot(error = TRUE, {
    toolbar_input_select(
      id = "test",
      label = "Test",
      choices = c("A", "B", "C"),
      selected = "D",
      tooltip = FALSE
    )
  })

  # Valid selected value should not error
  expect_no_error(
    toolbar_input_select(
      id = "test",
      label = "Test",
      choices = c("A", "B", "C"),
      selected = "B",
      tooltip = FALSE
    )
  )

  # Works with named choices
  expect_snapshot(error = TRUE, {
    toolbar_input_select(
      id = "test",
      label = "Test",
      choices = c("Label A" = "val_a", "Label B" = "val_b"),
      selected = "Label A", # Should use value, not label
      tooltip = FALSE
    )
  })

  # Works with grouped choices
  expect_snapshot(error = TRUE, {
    toolbar_input_select(
      id = "test",
      label = "Test",
      choices = list(
        "Group 1" = c("A", "B"),
        "Group 2" = c("C", "D")
      ),
      selected = "E",
      tooltip = FALSE
    )
  })
})

test_that("update_toolbar_input_select() validates selected is in choices", {
  session <- list(
    sendInputMessage = function(id, message) {
      session$last_message <<- message
    },
    input = list()
  )

  # Invalid selected value should warn and not update
  expect_snapshot({
    update_toolbar_input_select(
      "test_id",
      choices = c("A", "B", "C"),
      selected = "D",
      session = session
    )
  })
  expect_null(session$last_message$value)

  # Valid selected value should not warn
  expect_no_warning(
    update_toolbar_input_select(
      "test_id",
      choices = c("A", "B", "C"),
      selected = "B",
      session = session
    )
  )
  expect_equal(session$last_message$value, "B") # Should set to B
})

test_that("update_toolbar_input_select() keeps current value when choices change", {
  # Mock session with a current input value
  session <- list(
    sendInputMessage = function(id, message) {
      # Capture the message to verify behavior
      session$last_message <<- message
    },
    input = list(test_select = "B")
  )

  # Update choices - current value "B" is still valid, should be kept
  update_toolbar_input_select(
    "test_select",
    choices = c("A", "B", "C"),
    session = session
  )
  expect_null(session$last_message$value)

  # Update choices - current value "B" is no longer valid, should keep current (don't update)
  session$input$test_select <- "B"
  update_toolbar_input_select(
    "test_select",
    choices = c("X", "Y", "Z"),
    session = session
  )
  expect_null(session$last_message$value)
})

test_that("process_choices_selected() handles all cases correctly", {
  # Case 1: Valid selected value with choices
  result <- process_choices_selected(c("A", "B", "C"), "B", "test_id")
  expect_equal(result$value, "B")
  expect_null(result$error)

  # Case 2: Invalid selected value with choices
  result <- process_choices_selected(c("A", "B", "C"), "D", "test_id")
  expect_null(result$value)
  expect_match(result$error, "not in `choices`")

  # Case 3: selected is a vector (invalid)
  result <- process_choices_selected(c("A", "B", "C"), c("A", "B"), "test_id")
  expect_null(result$value)
  expect_match(result$error, "single value")

  # Case 4: No selected, no choices
  result <- process_choices_selected(NULL, NULL, "test_id")
  expect_null(result$value)
  expect_null(result$error)

  # Case 5: No selected, has choices - selects first choice (use_first_choice = TRUE)
  result <- process_choices_selected(
    c("A", "B", "C"),
    NULL,
    "test_id",
    use_first_choice = TRUE
  )
  expect_equal(result$value, "A")
  expect_null(result$error)

  # Case 5b: No selected, has choices - keeps NULL (use_first_choice = FALSE, for updates)
  result <- process_choices_selected(
    c("A", "B", "C"),
    NULL,
    "test_id",
    use_first_choice = FALSE
  )
  expect_null(result$value)
  expect_null(result$error)

  # Case 6: Valid selected with named choices
  result <- process_choices_selected(
    c("Label A" = "val_a", "Label B" = "val_b"),
    "val_a",
    "test_id"
  )
  expect_equal(result$value, "val_a")
  expect_null(result$error)

  # Case 7: Invalid selected (using label instead of value)
  result <- process_choices_selected(
    c("Label A" = "val_a", "Label B" = "val_b"),
    "Label A",
    "test_id"
  )
  expect_null(result$value)
  expect_match(result$error, "not in `choices`")

  # Case 8: Valid selected with grouped choices
  result <- process_choices_selected(
    list("Group 1" = c("A", "B"), "Group 2" = c("C", "D")),
    "B",
    "test_id"
  )
  expect_equal(result$value, "B")
  expect_null(result$error)

  # Case 9: Invalid selected with grouped choices
  result <- process_choices_selected(
    list("Group 1" = c("A", "B"), "Group 2" = c("C", "D")),
    "E",
    "test_id"
  )
  expect_null(result$value)
  expect_match(result$error, "not in `choices`")

  # Case 10: Selected without choices (invalid)
  result <- process_choices_selected(NULL, "B", "test_id")
  expect_null(result$value)
  expect_match(result$error, "cannot be set without `choices`")
})

test_that("update_toolbar_input_button() warns for blank label", {
  # Note: We can't fully test these functions without a Shiny session,
  # but we can test that the warning is issued before the session error occurs.

  # Empty string label should warn
  expect_warning(
    expect_error(
      update_toolbar_input_button(
        "test_id",
        label = ""
      )
    ),
    "non-empty string label"
  )

  # Whitespace-only label should warn
  expect_warning(
    expect_error(
      update_toolbar_input_button(
        "test_id",
        label = "   "
      )
    ),
    "non-empty string label"
  )

  # Empty tag label should warn
  expect_warning(
    expect_error(
      update_toolbar_input_button(
        "test_id",
        label = span("")
      )
    ),
    "non-empty string label"
  )
})

test_that("update_toolbar_input_button() can disable and reenable button", {
  # Mock session that captures sendInputMessage calls
  session <- list(
    sendInputMessage = function(id, message) {
      session$last_id <<- id
      session$last_message <<- message
    }
  )

  # Test disabling a button
  update_toolbar_input_button(
    "test_btn",
    label = "Test Button",
    disabled = TRUE,
    session = session
  )

  expect_equal(session$last_id, "test_btn")
  expect_equal(session$last_message$disabled, TRUE)

  # Test re-enabling a button
  update_toolbar_input_button(
    "test_btn",
    label = "Test Button",
    disabled = FALSE,
    session = session
  )

  expect_equal(session$last_id, "test_btn")
  expect_equal(session$last_message$disabled, FALSE)

  # Test updating multiple properties including disabled state
  update_toolbar_input_button(
    "test_btn",
    label = "Updated",
    icon = shiny::icon("check"),
    disabled = TRUE,
    session = session
  )

  expect_equal(session$last_id, "test_btn")
  expect_equal(session$last_message$disabled, TRUE)
  expect_true(!is.null(session$last_message$label))
  expect_true(!is.null(session$last_message$icon))
})

generate_random_walk <- function(num_steps = 90) {
  start_date <- as.POSIXct(
    as.integer(Sys.time()) * runif(1),
    origin = "1970-01-01"
  )

  increments <- rnorm(num_steps)
  cumulative_sum <- cumsum(increments)
  time_series <- c(0, cumulative_sum) + rnorm(1, 0, 50) + 25

  dates <- seq(start_date, length.out = num_steps + 1, by = "day")

  data.frame(date = dates, value = time_series)
}

as_plotly_sparkline <- function(plot, color = "white") {
  plot |>
    layout(
      xaxis = list(visible = FALSE, showgrid = FALSE, title = ""),
      yaxis = list(visible = FALSE, showgrid = FALSE, title = ""),
      hovermode = "x",
      margin = list(t = 0, r = 0, l = 0, b = 0),
      font = list(color = color),
      paper_bgcolor = "transparent",
      plot_bgcolor = "transparent"
    ) |>
    config(displayModeBar = FALSE) |>
    htmlwidgets::onRender(
      "function(el) {
        var ro = new ResizeObserver(function() {
          var visible = el.offsetHeight > 200;
          Plotly.relayout(el, {'xaxis.visible': visible});
        });
        ro.observe(el);
      }"
    )
}

random_plotly_plot <- function(type = NULL, color = "white") {
  if (is.null(type)) {
    type <- sample(c("bar", "box", "line"), 1)
  }

  plot <- switch(
    type,
    bar = random_plotly_bar(color, 50),
    box = random_plotly_box(color, 50),
    line = random_plotly_line(color, 50),
    stop("Not a valid random plot type: ", type)
  )

  as_plotly_sparkline(plot, color)
}

random_plotly_bar <- function(color, n = 50) {
  plot_ly(
    x = ~ runif(n),
    type = "histogram",
    histnorm = "probability",
    nbinsx = 10,
    color = I(color),
    stroke = I(color),
    alpha_stroke = 1,
    alpha = 0.6
  )
}

random_plotly_box <- function(color, n = 50) {
  plot_ly(x = ~ rnorm(n), type = "box", color = I(color))
}

random_plotly_line <- function(color, n) {
  add_lines(
    plot_ly(generate_random_walk(n)),
    x = ~date,
    y = ~value,
    color = I(color),
    fill = "tozeroy",
    span = I(1),
    alpha = 0.2
  )
}

#' Create a Bootstrap theme
#'
#' @description
#'
#' Creates a Bootstrap theme object, where you can:
#'
#' * Choose a (major) Bootstrap `version`.
#' * Choose a [Bootswatch theme](https://bootswatch.com) (optional).
#' * Customize main colors and fonts via explicitly named arguments (e.g.,
#'   `bg`, `fg`, `primary`, etc).
#' * Customize other, lower-level, Bootstrap Sass variable defaults via `...`.
#'
#' To learn more about how to implement custom themes, as well as how to use
#' them inside Shiny and R Markdown, [see
#' here](https://rstudio.github.io/bslib/articles/theming.html).
#'
#' @section Colors:
#'
#'  Colors may be provided in any format that [htmltools::parseCssColors()] can
#'  understand. To control the vast majority of the ('grayscale') color
#'  defaults, specify both the `fg` (foreground) and `bg` (background) colors.
#'  The `primary` and `secondary` theme colors are also useful for accenting the
#'  main grayscale colors in things like hyperlinks, tabset panels, and buttons.
#'
#' @section Fonts:
#'
#'  Use `base_font`, `code_font`, and `heading_font` to control the main
#'  typefaces. These arguments set new defaults for the relevant `font-family`
#'  CSS properties, but don't necessarily import the relevant font files. To
#'  both set CSS properties _and_ import font files, consider using the various
#'  [font_face()] helpers.
#'
#'  Each `*_font` argument may be a single font or a `font_collection()`.
#'  A font can be created with [font_google()], [font_link()], or [font_face()],
#'  or it can be a character vector of font names in the following format:
#'    * A single unquoted name (e.g., `"Source Sans Pro"`).
#'    * A single quoted name (e.g., `"'Source Sans Pro'"`).
#'    * A comma-separated list of names w/ individual names quoted as necessary.
#'      (e.g. `c("Open Sans", "'Source Sans Pro'", "'Helvetica Neue', Helvetica, sans-serif")`)
#'
#'  `font_google()` sets `local = TRUE` by default, which ensures that the font
#'  files are downloaded from [Google Fonts](https://fonts.google.com/) when
#'  your document or app is rendered. This guarantees that the client has access
#'  to the font family, making it relatively safe to specify just one font
#'  family:
#'
#'  ```r
#'  bs_theme(base_font = font_google("Pacifico", local = TRUE))
#'  ```
#'
#'  That said, we recommend you specify multiple "fallback" font families,
#'  especially when relying on remote and/or system fonts being available.
#'  Fallback fonts are useful not only for handling missing fonts, but also
#'  ensure that your users don't experience a Flash of Invisible Text (FOIT)
#'  which can be quite noticeable with remote web fonts on a slow internet
#'  connection.
#'
#'  ```r
#'  bs_theme(base_font = font_collection(font_google("Pacifico", local = FALSE), "Roboto", "sans-serif"))
#'  ````
#'
#' @references
#'   * [Get Started: Theming](https://rstudio.github.io/bslib/articles/theming/index.html)
#'     introduces theming with bslib in Shiny apps and R Markdown documents.
#'   * [Theming: Bootstrap 5 variables](https://rstudio.github.io/bslib/articles/bs5-variables/index.html)
#'     provides a searchable reference of all theming variables available in
#'     Bootstrap 5.
#'   * [Theming: Custom components](https://rstudio.github.io/bslib/articles/custom-components/index.html)
#'     gives a tutorial on creating a dynamically themable custom component.
#'   * bslib's theming capabilities are powered by
#'     [the {sass} package](https://rstudio.github.io/sass/).
#'   * [Bootstrap's utility classes](https://rstudio.github.io/bslib/articles/utility-classes/index.html)
#'     can be helpful when you want to change the appearance of an element
#'     without writing CSS or customizing your `bs_theme()`.
#'
#' @family Bootstrap theme functions
#'
#' @param version The major version of Bootstrap to use (see [versions()]
#'   for possible values). Defaults to the currently recommended version
#'   for new projects (currently Bootstrap 5).
#' @param preset The name of a theme preset, either a built-in theme provided by
#'   bslib or a Bootswatch theme (see [builtin_themes()] and
#'   [bootswatch_themes()] for possible values). This argument takes precedence
#'   over the `bootswatch` argument and only one `preset` or `bootswatch` can be
#'   provided. When no `bootswatch` theme is specified, and `version` is 5 or
#'   higher, `preset` defaults to `"shiny"`. To remove the `"shiny"` preset,
#'   provide a value of `"bootstrap"` (this value will also work in
#'   `bs_theme_update()` to remove a `preset` or `bootswatch` theme).
#' @param brand Specifies how to apply branding to your theme using
#'   [brand.yml](https://posit-dev.github.io/brand-yml/), a simple YAML file that
#'   defines key brand elements like colors, fonts, and logos. Valid options:
#'
#'   - `NULL` (default): Automatically looks for a `_brand.yml` file in the
#'     current directory or in `_brand/` or `brand/` in the current directory.
#'     If not found, it searches parent project directories for a `_brand.yml`
#'     file (also possibly in `_brand/` or `brand/`). If a `_brand.yml` file is
#'     found, it is applied to the Bootstrap theme.
#'   - `TRUE` (default): Automatically looks for a `_brand.yml` file in the
#'     current or app directory as described above. If a `_brand.yml` file
#'     *is not found*, `bs_theme()` will throw an error.
#'   - `FALSE`: Disables any brand.yml usage, even if a `_brand.yml` file is
#'     present.
#'   - A file path that directly points to a specific brand.yml file (with any
#'     file name) that you want to use.
#'   - Use a list to directly provide brand settings directly in R, following
#'     the brand.yml structure.
#'
#'   Learn more about creating and using brand.yml files at the
#'   [brand.yml homepage](https://posit-dev.github.io/brand-yml/) or run
#'   `shiny::runExample("brand.yml", package = "bslib")` to try brand.yml in a
#'   demo app.
#' @param bootswatch The name of a bootswatch theme (see [bootswatch_themes()]
#'   for possible values). When provided to `bs_theme_update()`, any previous
#'   Bootswatch theme is first removed before the new one is applied (use
#'   `bootswatch = "bootstrap"` to effectively remove the Bootswatch theme).
#' @param ... arguments passed along to [bs_add_variables()].
#' @param bg A color string for the background.
#' @param fg A color string for the foreground.
#' @param primary A color to be used for hyperlinks, to indicate primary/default
#'   actions, and to show active selection state in some Bootstrap components.
#'   Generally a bold, saturated color that contrasts with the theme's base
#'   colors.
#' @param secondary A color for components and messages that don't need to stand
#'   out. (Not supported in Bootstrap 3.)
#' @param success A color for messages that indicate an operation has succeeded.
#'   Typically green.
#' @param info A color for messages that are informative but not critical.
#'   Typically a shade of blue-green.
#' @param warning A color for warning messages. Typically yellow.
#' @param danger A color for errors. Typically red.
#' @param base_font The default typeface.
#' @param code_font The typeface to be used for code. Be sure this is monospace!
#' @param heading_font The typeface to be used for heading elements.
#' @param font_scale A scalar multiplier to apply to the base font size. For
#'   example, a value of `1.5` scales font sizes to 150% and a value of `0.8`
#'   scales to 80%. Must be a positive number.
#'
#' @return Returns a [sass::sass_bundle()] (list-like) object.
#'
#' @examplesIf rlang::is_interactive()
#'
#' theme <- bs_theme(
#'   # Controls the default grayscale palette
#'   bg = "#202123", fg = "#B8BCC2",
#'   # Controls the accent (e.g., hyperlink, button, etc) colors
#'   primary = "#EA80FC", secondary = "#48DAC6",
#'   base_font = c("Grandstander", "sans-serif"),
#'   code_font = c("Courier", "monospace"),
#'   heading_font = "'Helvetica Neue', Helvetica, sans-serif",
#'   # Can also add lower-level customization
#'   "input-border-color" = "#EA80FC"
#' )
#'
#' bs_theme_preview(theme)
#'
#' # Lower-level bs_add_*() functions allow you to work more
#' # directly with the underlying Sass code
#' theme <- bs_add_variables(theme, "my-class-color" = "red")
#' theme <- bs_add_rules(theme, ".my-class { color: $my-class-color }")
#'
#' @export
bs_theme <- function(
  version = version_default(),
  preset = NULL,
  ...,
  brand = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  font_scale = NULL,
  bootswatch = NULL
) {
  is_version_from_user <- !missing(version)

  brand <- brand_resolve(brand)

  preset <-
    if (!is.null(brand)) {
      brand_resolve_preset(
        brand = brand,
        preset = preset,
        version = if (is_version_from_user) version
      )
    } else {
      if (is.null(preset) && is.null(bootswatch) && version >= 5) {
        preset <- "shiny"
      }
      resolve_bs_preset(preset, bootswatch, version = version)
    }

  version <- preset$version %||% version %||% version_default()

  bundle <- bs_bundle(
    bs_theme_init(version),
    bootstrap_bundle(version),
    bs_preset_bundle(preset),
    bs_brand_bundle(brand, version = version)
  )

  if (!is.null(preset$type)) {
    bundle <- add_class(bundle, THEME_PRESET_CLASS)
  }

  bundle <- bs_theme_update(
    bundle,
    ...,
    bg = bg,
    fg = fg,
    primary = primary,
    secondary = secondary,
    success = success,
    info = info,
    warning = warning,
    danger = danger,
    base_font = base_font,
    code_font = code_font,
    heading_font = heading_font,
    font_scale = font_scale
  )

  if (!is.null(brand)) {
    bundle <- add_class(bundle, "bs_theme_brand")
    attr(bundle, "brand") <- brand
  }

  bundle
}

#' @rdname bs_theme
#' @param theme A [bs_theme()] object.
#' @export
bs_theme_update <- function(
  theme,
  ...,
  preset = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  font_scale = NULL,
  bootswatch = NULL
) {
  assert_bs_theme(theme)

  preset <- resolve_bs_preset(
    preset,
    bootswatch,
    version = theme_version(theme)
  )

  if (!is.null(preset)) {
    theme_has_preset <- inherits(theme, THEME_PRESET_CLASS)

    if (theme_has_preset) {
      # remove the old preset
      theme <- bs_remove(theme, theme_preset_info(theme)$type)
      class(theme) <- setdiff(class(theme), THEME_PRESET_CLASS)
    }

    # Add in the new preset unless vanilla bootstrap was requested
    if (!identical(preset$name, "bootstrap")) {
      theme <- add_class(theme, THEME_PRESET_CLASS)
      theme <- bs_bundle(theme, bs_preset_bundle(preset))
    }
  }

  # See R/bs-theme-update.R for the implementation of these
  theme <- bs_base_colors(theme, bg = bg, fg = fg)
  theme <- bs_accent_colors(
    theme,
    primary = primary,
    secondary = secondary,
    success = success,
    info = info,
    warning = warning,
    danger = danger
  )
  theme <- bs_fonts(
    theme,
    base = base_font,
    code = code_font,
    heading = heading_font
  )
  if (!is.null(font_scale)) {
    stopifnot(is.numeric(font_scale) && length(font_scale) == 1)
    theme <- bs_add_variables(
      theme,
      "font-size-base" = paste(
        font_scale,
        "*",
        bs_get_variables(theme, "font-size-base")
      )
    )
  }
  bs_add_variables(theme, ...)
}

#' @rdname bs_global_theme
#' @export
bs_global_theme_update <- function(
  ...,
  preset = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  bootswatch = NULL
) {
  theme <- assert_global_theme("bs_theme_global_update()")
  bs_global_set(
    bs_theme_update(
      theme,
      ...,
      preset = preset,
      bootswatch = bootswatch,
      bg = bg,
      fg = fg,
      primary = primary,
      secondary = secondary,
      success = success,
      info = info,
      warning = warning,
      danger = danger,
      base_font = base_font,
      code_font = code_font,
      heading_font = heading_font
    )
  )
}

#' @rdname bs_theme
#' @param x an object.
#' @export
is_bs_theme <- function(x) {
  inherits(x, "bs_theme")
}

# Start an empty bundle with special classes that
# theme_version() & theme_bootswatch() search for
bs_theme_init <- function(version) {
  init_layer <- sass_layer(
    defaults = list(
      "bootstrap-version" = version,
      "bslib-preset-name" = "null !default",
      "bslib-preset-type" = "null !default"
    ),
    rules = c(
      ":root {",
      "--bslib-bootstrap-version: #{$bootstrap-version};",
      "--bslib-preset-name: #{$bslib-preset-name};",
      "--bslib-preset-type: #{$bslib-preset-type};",
      "}"
    )
  )

  add_class(init_layer, c(paste0("bs_version_", version), "bs_theme"))
}

assert_bs_theme <- function(theme) {
  if (!is_bs_theme(theme)) {
    stop("`theme` must be a `bs_theme()` object")
  }
  invisible(theme)
}

# -----------------------------------------------------------------
# Core Bootstrap bundle
# -----------------------------------------------------------------

bootstrap_bundle <- function(version) {
  pandoc_tables <- list(
    # Pandoc uses align attribute to align content but BS4 styles take precedence...
    # we may want to consider adopting this more generally in "strict" BS4 mode as well
    ".table th[align=left] { text-align: left; }",
    ".table th[align=right] { text-align: right; }",
    ".table th[align=center] { text-align: center; }"
  )

  main_bundle <- switch_version(
    version,
    five = sass_bundle(
      # Don't name this "core" bundle so it can't easily be removed
      sass_layer(
        functions = bs5_sass_files("functions"),
        defaults = list(
          bs5_sass_files("variables"),
          bs5_sass_files("variables-dark")
        ),
        mixins = list(bs5_sass_files("maps"), bs5_sass_files("mixins")),
        rules = list(bs5_sass_files("mixins/banner"), "@include bsBanner('')")
      ),
      # Returns a _named_ list of bundles (i.e., these should be easily removed)
      !!!rule_bundles(
        # Names here should match https://github.com/twbs/bs5/blob/master/scss/bootstrap.scss
        bs5_sass_files(
          c(
            "utilities",
            "root",
            "reboot",
            "type",
            "images",
            "containers",
            "grid",
            "tables",
            "forms",
            "buttons",
            "transitions",
            "dropdown",
            "button-group",
            "nav",
            "navbar",
            "card",
            "accordion",
            "breadcrumb",
            "pagination",
            "badge",
            "alert",
            "progress",
            "list-group",
            "close",
            "toasts",
            "modal",
            "tooltip",
            "popover",
            "carousel",
            "spinners",
            "offcanvas",
            "placeholders",
            "helpers",
            "utilities/api"
          )
        )
      ),
      # Additions to BS5 that are always included (i.e., not a part of compatibility)
      sass_layer(rules = pandoc_tables),
      bs3compat = bs3compat_bundle(),
      # Enable CSS Grid powered Bootstrap grid
      sass_layer(
        defaults = list("enable-cssgrid" = "true !default")
      )
    ),
    four = sass_bundle(
      sass_layer(
        functions = bs4_sass_files(c("deprecated", "functions")),
        defaults = bs4_sass_files("variables"),
        mixins = bs4_sass_files("mixins")
      ),
      # Returns a _named_ list of bundles (i.e., these should be easily removed)
      !!!rule_bundles(
        # Names here should match https://github.com/twbs/bs4/blob/master/scss/bootstrap.scss
        bs4_sass_files(
          c(
            "root",
            "reboot",
            "type",
            "images",
            "code",
            "grid",
            "tables",
            "forms",
            "buttons",
            "transitions",
            "dropdown",
            "button-group",
            "input-group",
            "custom-forms",
            "nav",
            "navbar",
            "card",
            "breadcrumb",
            "pagination",
            "badge",
            "jumbotron",
            "alert",
            "progress",
            "media",
            "list-group",
            "close",
            "toasts",
            "modal",
            "tooltip",
            "popover",
            "carousel",
            "spinners",
            "utilities",
            "print"
          )
        )
      ),
      # Additions to BS4 that are always included (i.e., not a part of compatibility)
      sass_layer(rules = pandoc_tables),
      bs3compat = bs3compat_bundle()
    ),
    three = sass_bundle(
      sass_layer(
        defaults = bs3_sass_files("variables"),
        mixins = bs3_sass_files("mixins")
      ),
      # Should match https://github.com/twbs/bootstrap-sass/blob/master/assets/stylesheets/_bootstrap.scss
      !!!rule_bundles(
        bs3_sass_files(
          c(
            "normalize",
            "print",
            "glyphicons",
            "scaffolding",
            "type",
            "code",
            "grid",
            "tables",
            "forms",
            "buttons",
            "component-animations",
            "dropdowns",
            "button-groups",
            "input-groups",
            "navs",
            "navbar",
            "breadcrumbs",
            "pagination",
            "pager",
            "labels",
            "badges",
            "jumbotron",
            "thumbnails",
            "alerts",
            "progress-bars",
            "media",
            "list-group",
            "panels",
            "responsive-embed",
            "wells",
            "close",
            "modals",
            "tooltip",
            "popovers",
            "carousel",
            "utilities",
            "responsive-utilities"
          )
        )
      ),
      accessibility = bs3_accessibility_bundle(),
      glyphicon_font_files = sass_layer(
        defaults = list("icon-font-path" = "'glyphicon-fonts/'"),
        file_attachments = c(
          "glyphicon-fonts" = path_lib("bs3", "assets", "fonts", "bootstrap")
        )
      )
    )
  )

  sass_bundle(
    main_bundle,
    bslib_bundle()
  )
}

bootstrap_javascript_map <- function(version) {
  switch_version(
    version,
    five = path_lib("bs5", "dist", "js", "bootstrap.bundle.min.js.map"),
    four = path_lib("bs4", "dist", "js", "bootstrap.bundle.min.js.map")
  )
}
bootstrap_javascript <- function(version) {
  switch_version(
    version,
    five = path_lib("bs5", "dist", "js", "bootstrap.bundle.min.js"),
    four = path_lib("bs4", "dist", "js", "bootstrap.bundle.min.js"),
    three = path_lib("bs3", "assets", "javascripts", "bootstrap.min.js")
  )
}

# -----------------------------------------------------------------
# bslib specific Sass that gets bundled with Bootstrap
# -----------------------------------------------------------------

bslib_bundle <- function() {
  sass_layer(
    functions = sass_file(path_inst("bslib-scss", "functions.scss")),
    rules = sass_file(path_inst("bslib-scss", "bslib.scss"))
  )
}

# -----------------------------------------------------------------
# BS3 compatibility bundle
# -----------------------------------------------------------------

bs3compat_bundle <- function() {
  sass_layer(
    defaults = sass_file(
      system_file("bs3compat", "_defaults.scss", package = "bslib")
    ),
    mixins = sass_file(
      system_file("bs3compat", "_declarations.scss", package = "bslib")
    ),
    rules = sass_file(
      system_file("bs3compat", "_rules.scss", package = "bslib")
    ),
    # Gyliphicon font files
    file_attachments = c(
      fonts = path_lib("bs3", "assets", "fonts")
    ),
    html_deps = htmltools::htmlDependency(
      "bs3compat",
      packageVersion("bslib"),
      package = "bslib",
      src = "bs3compat/js",
      script = c("transition.js", "tabs.js", "bs3compat.js")
    )
  )
}

# -----------------------------------------------------------------
# BS3 accessibility bundle
# -----------------------------------------------------------------

bs3_accessibility_bundle <- function() {
  sass_layer(
    rules = sass_file(
      system_file(
        "lib",
        "bs-a11y-p",
        "src",
        "sass",
        "bootstrap-accessibility.scss",
        package = "bslib"
      )
    ),
    html_deps = htmltools::htmlDependency(
      "bootstrap-accessibility",
      version_accessibility,
      package = "bslib",
      src = "lib/bs-a11y-p",
      script = "plugins/js/bootstrap-accessibility.min.js",
      all_files = FALSE
    )
  )
}

if (!identical(getwd(), rprojroot::find_package_root_file())) {
  stop("This script must be run from the top directory of the bslib package")
}

pkgload::load_all()

lapply(
  c(
    "yarn_install.R",
    "download_preset_fonts.R",
    "update_gfont_info.R",
    "expand_variables_article_template.R",
    "compile_component_sass.R"
  ),
  function(file) {
    message("Updating: ", file)
    source(file.path("tools", file), local = TRUE)
  }
)

#' Dark mode input control
#'
#' @description
#' Creates a button that toggles between dark and light modes, specifically for
#' toggling between light and dark [Bootstrap color
#' modes](https://getbootstrap.com/docs/5.3/customize/color-modes/) -- a new
#' feature added in [Bootstrap
#' 5.3](https://getbootstrap.com/docs/5.3/migration/#color-modes).
#'
#' @param ... Additional attributes to be passed to the input control UI, such
#'   as `class`, `style`, etc.
#'
#'   In `toggle_dark_mode()`, the `...` are included for future extensibility
#'   and are currently ignored.
#' @param id An optional input id, required to reactively read the current color
#'   mode.
#' @param mode The initial mode of the dark mode switch. By default or when set
#'   to `NULL`, the user's system settings for preferred color scheme will be
#'   used. Otherwise, set to `"light"` or `"dark"` to force a particular initial
#'   mode.
#'
#' @return Returns a UI element for a dark mode switch input control. The server
#'   value received for the input corresponding to `id` will be a string value
#'   with the current color mode (`"light"` or `"dark"`).
#'
#' @describeIn input_dark_mode Create a dark mode switch input control
#' @family input controls
#' @export
input_dark_mode <- function(..., id = NULL, mode = NULL) {
  if (!is.null(mode)) {
    mode <- rlang::arg_match(mode, c("light", "dark"))
  }

  if (!is.null(id)) {
    mode <- shiny::restoreInput(id, default = mode)
  }

  if (any(!nzchar(rlang::names2(rlang::list2(...))))) {
    abort("All arguments in `...` must be named.")
  }

  res <- web_component(
    "bslib-input-dark-mode",
    id = id,
    attribute = "data-bs-theme",
    mode = mode,
    style = css(
      "--text-1" = "var(--bs-emphasis-color)",
      "--text-2" = "var(--bs-tertiary-color)",
      # The vertical correction used in the dark mode component isn't quite
      # right on Bootstrap pages. This next line overrides it and removes the
      # vertical correction. But users can still set the CSS property manually
      # in their a `style` argument passed in via `...` if they want.
      "--vertical-correction" = " "
    ),
    ...
  )

  res <- tag_require(res, version = 5, caller = "input_dark_mode()")
  as_fragment(res)
}

#' @describeIn input_dark_mode Programmatically toggle or set the current light
#'   or dark color mode.
#' @param session A Shiny session object (the default should almost always be
#'   used).
#' @export
toggle_dark_mode <- function(
  mode = NULL,
  ...,
  session = get_current_session()
) {
  rlang::check_dots_empty(
    error = function(err) rlang::warn(rlang::cnd_message(err))
  )

  if (!is.null(mode)) {
    mode <- tryCatch(
      rlang::arg_match(mode, c("light", "dark")),
      error = function(err) {
        rlang::warn(rlang::cnd_message(err))
        mode
      }
    )
  }

  data <- dropNulls(list(method = "toggle", value = mode))

  # We're using sendCustomMessage() here because dark mode is set globally and
  # `id` is not required for `toggle_dark_mode()`. `$sendInputMessage()` would
  # require us to know the `id` of at least one input control.
  callback <- function() {
    session$sendCustomMessage("bslib.toggle-dark-mode", data)
  }

  session$onFlush(callback, once = TRUE)
}

# ----------------------------------------------------------------------------------------
# Base colors
# ----------------------------------------------------------------------------------------

bs_base_colors <- function(theme, bg = NULL, fg = NULL) {
  assert_bs_theme(theme)
  if (is.null(bg) && is.null(fg)) {
    return(theme)
  }
  if (is.null(bg)) {
    stop("Cannot specify bg without fg.")
  }
  if (is.null(fg)) {
    stop("Cannot specify fg without bg.")
  }

  args <- validate_and_normalize_colors(list(bg = bg, fg = fg))
  # In some cases, bg/fg really means $body-bg/$body-color, not $white/$black
  use_body <- has_body_base_colors(theme)
  if (use_body) {
    args <- rename2(args, !!!get_base_color_map(theme))
  }

  switch_add_variables(
    theme,
    args,
    default = if (use_body) identity else bs4_base_colors,
    three = if (use_body) identity else bs3_base_colors
  )
}

switch_add_variables <- function(theme, args, ...) {
  func <- switch_version(theme, ...)
  vars <- do.call(func, list(args))
  bs_add_variables(theme, !!!vars)
}

# Obtain a mapping from (to) fg/bg to (from) relevant Sass vars
get_base_color_map <- function(theme, decode = TRUE) {
  use_body <- has_body_base_colors(theme)
  vars <- switch_version(
    theme,
    three = list("body-bg", if (use_body) "text-color" else "gray-base"),
    default = if (use_body) {
      list("body-bg", "body-color")
    } else {
      list("white", "black")
    }
  )
  if (decode) {
    setNames(vars, c("bg", "fg"))
  } else {
    setNames(c("bg", "fg"), vars)
  }
}

# We've modified these "dark mode" themes to be more themable by cascading
# defaults from $body-bg/$body-color instead of $white/$black
has_body_base_colors <- function(theme) {
  body_themes <- c("darkly", "cyborg", "superhero", "slate")
  # TODO: do the same for BS3? I guess?
  isTRUE(theme_bootswatch(theme) %in% body_themes)
}

bs4_base_colors <- function(args) {
  white <- args$bg
  black <- args$fg

  grays <- grDevices::colorRamp(c(white, black), alpha = TRUE)(0:10 / 10)

  if (any(grays[, 4] != 255)) {
    warning(
      call. = FALSE,
      "bs_base_colors does not respect alpha in `white` and `black` arguments"
    )
  }

  grays <- sprintf(
    "#%02X%02X%02X",
    round(grays[, 1]),
    round(grays[, 2]),
    round(grays[, 3])
  )

  names(grays) <- c(
    "white",
    paste0("gray-", 1:9 * 100),
    "black"
  )

  as.list(grays)
}

bs3_base_colors <- function(args) {
  white <- args$bg
  black <- args$fg

  ramp <- grDevices::colorRamp(c(black, white))
  gray <- function(level = 255) {
    val <- round(ramp(level / 255))
    sprintf("#%02X%02X%02X", val[1, 1], val[1, 2], val[1, 3])
  }

  result_colors <- list(
    "gray-base" = gray(0x00),
    "gray-darker" = gray(0x22),
    "gray-dark" = gray(0x33),
    "gray" = gray(0x55),
    "gray-light" = gray(0x77),
    "gray-lighter" = gray(0xEE),

    "gray-44" = gray(0x44),
    "gray-88" = gray(0x88),
    "gray-99" = gray(0x99),
    "gray-cc" = gray(0xcc),
    "gray-dd" = gray(0xdd),
    "gray-e5" = gray(0xe5),
    "gray-f5" = gray(0xf5),
    "gray-f8" = gray(0xf8),
    "gray-f9" = gray(0xf9),
    "white" = gray(0xff),
    "black" = gray(0x00)
  )

  # There's code in tools/bs3_theme_base_colors.R for generating this list.
  color_mapping <- list(
    `dropdown-caret-color` = "$gray-base",
    `tooltip-bg` = "$gray-base",
    `modal-backdrop-bg` = "$gray-base",
    `close-color` = "$gray-base",
    `navbar-inverse-bg` = "$gray-darker",
    `btn-default-color` = "$gray-dark",
    `navbar-default-link-hover-color` = "$gray-dark",
    `navbar-inverse-toggle-hover-bg` = "$gray-dark",
    `navbar-inverse-toggle-border-color` = "$gray-dark",
    `list-group-link-heading-color` = "$gray-dark",
    `kbd-bg` = "$gray-dark",
    `navbar-inverse-link-disabled-color` = "$gray-44",
    `navbar-default-link-active-color` = "$gray",
    `list-group-link-color` = "$gray",
    `navbar-default-color` = "$gray-light",
    `navbar-default-link-color` = "$gray-light",
    `navbar-default-toggle-icon-bar-bg` = "$gray-88",
    `input-color-placeholder` = "$gray-99",
    `modal-content-fallback-border-color` = "$gray-99",
    `btn-default-border` = "$gray-cc",
    `input-border` = "$gray-cc",
    `dropdown-fallback-border` = "$gray-cc",
    `navbar-default-link-disabled-color` = "$gray-cc",
    `popover-fallback-border-color` = "$gray-cc",
    `breadcrumb-color` = "$gray-cc",
    `pre-border-color` = "$gray-cc",
    `table-border-color` = "$gray-dd",
    `navbar-default-toggle-hover-bg` = "$gray-dd",
    `navbar-default-toggle-border-color` = "$gray-dd",
    `nav-tabs-border-color` = "$gray-dd",
    `nav-tabs-active-link-hover-border-color` = "$gray-dd",
    `nav-tabs-justified-link-border-color` = "$gray-dd",
    `pagination-border` = "$gray-dd",
    `pagination-hover-border` = "$gray-dd",
    `pagination-disabled-border` = "$gray-dd",
    `list-group-border` = "$gray-dd",
    `panel-inner-border` = "$gray-dd",
    `panel-default-border` = "$gray-dd",
    `thumbnail-border` = "$gray-dd",
    `legend-border-color` = "$gray-e5",
    `dropdown-divider-bg` = "$gray-e5",
    `modal-header-border-color` = "$gray-e5",
    `table-bg-hover` = "$gray-f5",
    `dropdown-link-hover-bg` = "$gray-f5",
    `progress-bg` = "$gray-f5",
    `list-group-hover-bg` = "$gray-f5",
    `panel-footer-bg` = "$gray-f5",
    `panel-default-heading-bg` = "$gray-f5",
    `well-bg` = "$gray-f5",
    `breadcrumb-bg` = "$gray-f5",
    `code-bg` = "$gray-f5",
    `pre-bg` = "$gray-f5",
    `navbar-default-bg` = "$gray-f8",
    `table-bg-accent` = "$gray-f9",
    `body-bg` = "$white",
    `btn-default-bg` = "$white",
    `btn-primary-color` = "$white",
    `btn-success-color` = "$white",
    `btn-info-color` = "$white",
    `btn-warning-color` = "$white",
    `btn-danger-color` = "$white",
    `input-bg` = "$white",
    `dropdown-bg` = "$white",
    `navbar-inverse-link-hover-color` = "$white",
    `navbar-inverse-brand-hover-color` = "$white",
    `navbar-inverse-toggle-icon-bar-bg` = "$white",
    `pagination-bg` = "$white",
    `pagination-active-color` = "$white",
    `pagination-disabled-bg` = "$white",
    `tooltip-color` = "$white",
    `popover-bg` = "$white",
    `label-color` = "$white",
    `label-link-hover-color` = "$white",
    `modal-content-bg` = "$white",
    `progress-bar-color` = "$white",
    `list-group-bg` = "$white",
    `panel-bg` = "$white",
    `panel-primary-text` = "$white",
    `badge-color` = "$white",
    `badge-link-hover-color` = "$white",
    `badge-active-bg` = "$white",
    `carousel-control-color` = "$white",
    `carousel-indicator-active-bg` = "$white",
    `carousel-indicator-border-color` = "$white",
    `carousel-caption-color` = "$white",
    `close-text-shadow` = "$white",
    `kbd-color` = "$white"
  )

  results <- c(result_colors, color_mapping)

  results
}

# ----------------------------------------------------------------------------------------
# Accent colors
# ----------------------------------------------------------------------------------------

bs_accent_colors <- function(
  theme,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL
) {
  assert_bs_theme(theme)

  args <- validate_and_normalize_colors(
    list(
      primary = primary,
      secondary = secondary,
      success = success,
      info = info,
      warning = warning,
      danger = danger
    )
  )

  switch_add_variables(
    theme,
    args,
    three = bs3_accent_colors,
    default = identity
  )
}

bs3_accent_colors <- function(args) {
  # Warns and filters out unsupported arguments
  supported <- c("primary", "success", "info", "warning", "danger")

  args <- retain_known_vars("Bootstrap 3", "accent color", supported, args)

  # Bootstrap 3 uses brand-primary, brand-danger, etc. as var names
  if (length(args) > 0) {
    names(args) <- paste0("brand-", names(args))
  }

  args
}

# ----------------------------------------------------------------------------------------
# Fonts
# ----------------------------------------------------------------------------------------

bs_fonts <- function(theme, base = NULL, code = NULL, heading = NULL) {
  assert_bs_theme(theme)
  args <- dropNulls(
    list(
      base = base,
      code = code,
      heading = heading
    )
  )
  args <- lapply(args, as_font_collection)
  switch_add_variables(theme, args, three = bs3_fonts, default = bs4_fonts)
}

as_font_collection <- function(x) {
  if (sass::is_font_collection(x)) {
    return(x)
  }
  do.call(sass::font_collection, as.list(x))
}

bs4_fonts <- function(args) {
  name_map <- c(
    base = "font-family-base",
    code = "font-family-monospace",
    heading = "headings-font-family",
    # TODO: we don't have a input_font...should we?
    input = "input-btn-font-family"
  )

  names(args) <- name_map[names(args)]
  args
}

bs3_fonts <- function(args) {
  name_map <- c(
    base = "font-family-base",
    code = "font-family-monospace",
    heading = "headings-font-family"
  )

  args <- retain_known_vars("Bootstrap 3", "font", names(name_map), args)

  names(args) <- name_map[names(args)]
  args
}

#' Ensures all arguments are either NULL, or length 1 character vectors with
#' valid CSS color strings; returned is a list with no NULLs and normalized
#' color strings
#' @param args A named list
#' @noRd
validate_and_normalize_colors <- function(args) {
  args <- dropNulls(args)
  if (length(args) == 0) {
    return(args)
  }

  is_char <- vapply(args, is.character, logical(1))
  vec_len <- vapply(args, length, integer(1))
  bad <- !is_char | vec_len != 1
  if (any(bad)) {
    stop(
      call. = FALSE,
      "Invalid HTML color strings for argument(s) ",
      format_varnames(names(args)[bad]),
      "; single-element character vectors are required"
    )
  }
  args <- lapply(args, htmltools::parseCssColors, mustWork = FALSE)
  bad <- vapply(args, rlang::is_na, logical(1))
  if (any(bad)) {
    stop(
      "Invalid HTML color strings for argument(s) ",
      format_varnames(names(args)[bad]),
      call. = FALSE
    )
  }
  args
}

#' Remove unsupported arguments, with a nicely formatted warning
#'
#' @param caller_name String naming the calling function; used for error
#'   messages
#' @param arg_name A name for the group of variables (e.g., accent, font); used for error messages
#' @param supported_vars Character vector of known names
#' @param args List of args
#' @return List with unsupported vars removed, possibly warning in the process
#' @noRd
retain_known_vars <- function(
  caller_name,
  arg_name = "",
  supported_vars,
  args
) {
  argnames <- names(args)
  unknown_idx <- !argnames %in% supported_vars

  if (any(unknown_idx)) {
    warning(
      call. = FALSE,
      caller_name,
      " doesn't support the following ",
      arg_name,
      if (nzchar(arg_name)) " ",
      "argument(s), they will be ignored: ",
      format_varnames(argnames[unknown_idx])
    )
  }

  args[!unknown_idx]
}

# Format a vector of variable names
format_varnames <- function(varnames, quot = "`", delim = ", ") {
  between <- paste0(quot, delim, quot)

  paste0(
    quot,
    paste(collapse = between, varnames),
    quot
  )
}

#' Add a tooltip to a UI element
#'
#' @description
#' Display additional information when focusing (or hovering over) a UI element.
#'
#' @param trigger A UI element (i.e., [htmltools tag][htmltools::tags]) to serve
#'   as the tooltip trigger. If `trigger` renders as multiple HTML
#'   elements (e.g., it's a `tagList()`), the last HTML element is used for the
#'   trigger. If the `trigger` should contain all of those elements, wrap the
#'   object in a [htmltools::div()] or [htmltools::span()].
#' @param ... UI elements for the tooltip. Character strings are [automatically
#'   escaped][htmltools::htmlEscape()] unless marked as [htmltools::HTML()].
#' @param id A character string. Required to re-actively respond to the
#'   visibility of the tooltip (via the `input[[id]]` value) and/or update the
#'   visibility/contents of the tooltip.
#' @param placement The placement of the tooltip relative to its trigger.
#' @param options A list of additional [options](https://getbootstrap.com/docs/5.3/components/tooltips/#options).
#'
#' @section Theming/Styling:
#'
#'   ```{r child="man/fragments/tooltip-popover_theming.Rmd", el="tooltip"}
#'   ```
#'
#'   ```
#'   tooltip(
#'     "Trigger", "Tooltip message",
#'     options = list(customClass = "my-tip")
#'   )
#'   ```
#'
#'   And then add relevant rules to [bs_theme()] via [bs_add_rules()]:
#'
#'   ```
#'   bs_theme() |> bs_add_rules(".my-tip { max-width: none; }")
#'   ```
#'
#' @section Accessibility of Tooltip Triggers:
#'
#'   ```{r child="man/fragments/tooltip-popover_a11y-trigger.Rmd", el = "tooltip"}
#'   ```
#'
#'   ```r
#'   tooltip(
#'     bsicons::bs_icon("info-circle", title = "About tooltips"),
#'     "Text shown in the tooltip."
#'   )
#'   ```
#'
#'   ```r
#'   tooltip(
#'     fontawesome::fa("info-circle", a11y = "sem", title = "About tooltips"),
#'     "Text shown in the tooltip."
#'   )
#'   ```
#'
#' @describeIn tooltip Add a tooltip to a UI element
#'
#' @export
#' @family Components
#'
#' @references Tooltips are based on [Bootstrap's Tooltip
#'   component](https://getbootstrap.com/docs/5.3/components/tooltips/). See the
#'   bslib website for an [interactive introduction to tooltips and
#'   popovers](https://rstudio.github.io/bslib/articles/tooltips-popovers/index.html).
#'
#' @seealso [popover()] provides a an alternative and more persistent container
#'   for additional elements, typically revealed by clicking on a target
#'   element.
#'
#' @examplesIf rlang::is_interactive()
#'
#' tooltip(
#'   shiny::actionButton("btn", "A button"),
#'   "A message"
#' )
#'
#' card(
#'   card_header(
#'     tooltip(
#'       span("Card title ", bsicons::bs_icon("question-circle-fill")),
#'       "Additional info",
#'       placement = "right"
#'     )
#'   ),
#'   "Card body content..."
#' )
tooltip <- function(
  trigger,
  ...,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
) {
  args <- separate_arguments(...)
  children <- args$children
  attribs <- args$attribs

  if (length(children) == 0) {
    abort("At least one value must be provided to `...`.")
  }

  bad_opts <- intersect(c("title", "placement"), names(options))
  if (length(bad_opts) > 0) {
    rlang::abort(
      sprintf("The `%s` option cannot be specified directly.", bad_opts[1])
    )
  }

  res <- web_component(
    "bslib-tooltip",
    id = id,
    placement = rlang::arg_match(placement),
    bsOptions = to_json(options),
    !!!attribs,
    # Use <template> as a way to protect these children from potentially being
    # pulled outside this element (the browser's parser does this to, for
    # example, block elements inside a <p> tag)
    tags$template(!!!children),
    trigger
  )

  res <- tag_require(res, version = 5, caller = "tooltip()")
  as_fragment(res)
}

#' @describeIn tooltip Programmatically show/hide a tooltip.
#'
#' @param id a character string that matches an existing tooltip id.
#' @param show Whether to show (`TRUE`) or hide (`FALSE`) the tooltip. The
#'   default (`NULL`) will show if currently hidden and hide if currently shown.
#'   Note that a tooltip will not be shown if the trigger is not visible (e.g.,
#'   it's hidden behind a tab).
#' @param session A Shiny session object (the default should almost always be
#'   used).
#'
#' @export
toggle_tooltip <- function(id, show = NULL, session = get_current_session()) {
  show <- normalize_show_value(show)

  msg <- list(method = "toggle", value = show)
  force(id)
  callback <- function() {
    session$sendInputMessage(id, msg)
  }
  session$onFlush(callback, once = TRUE)
}

#' @describeIn tooltip Update the contents of a tooltip.
#' @export
update_tooltip <- function(id, ..., session = get_current_session()) {
  title <- tagList(...)

  msg <- dropNulls(
    list(
      method = "update",
      title = if (length(title) > 0) processDeps(title, session)
    )
  )

  force(id)
  callback <- function() {
    session$sendInputMessage(id, msg)
  }
  session$onFlush(callback, once = TRUE)
}

normalize_show_value <- function(show) {
  if (is.null(show)) {
    return("toggle")
  }

  if (length(show) != 1 || !is.logical(show)) {
    abort("`show` must be `TRUE`, `FALSE`, or `NULL`.")
  }

  if (show) "show" else "hide"
}

to_json <- function(..., auto_unbox = TRUE, null = "null") {
  jsonlite::toJSON(..., auto_unbox = auto_unbox, null = null)
}

library(ggplot2)
library(lattice)

ids <- factor(c("1.1", "2.1", "1.2", "2.2", "1.3", "2.3"))
values <- data.frame(
  id = ids,
  value = c(3, 3.1, 3.1, 3.2, 3.15, 3.5)
)
positions <- data.frame(
  id = rep(ids, each = 4),
  x = c(
    2,
    1,
    1.1,
    2.2,
    1,
    0,
    0.3,
    1.1,
    2.2,
    1.1,
    1.2,
    2.5,
    1.1,
    0.3,
    0.5,
    1.2,
    2.5,
    1.2,
    1.3,
    2.7,
    1.2,
    0.5,
    0.6,
    1.3
  ),
  y = c(
    -0.5,
    0,
    1,
    0.5,
    0,
    0.5,
    1.5,
    1,
    0.5,
    1,
    2.1,
    1.7,
    1,
    1.5,
    2.2,
    2.1,
    1.7,
    2.1,
    3.2,
    2.8,
    2.1,
    2.2,
    3.3,
    3.2
  )
)
huron <- data.frame(year = 1875:1972, level = as.vector(LakeHuron))

ggplot2_examples <- list(
  GeomPointColor = {
    ggplot(mtcars, aes(wt, mpg, color = cyl)) + geom_point()
  },
  GeomAbline = {
    ggplot(mtcars, aes(mpg, wt)) +
      geom_point() +
      geom_hline(
        aes(yintercept = wt),
        data.frame(cyl = c(4, 6, 8), wt = c(2.28, 3.11, 4.00))
      ) +
      facet_wrap(~cyl)
  },
  GeomBar = ggplot(mpg, aes(class)) + geom_bar(aes(weight = displ)),
  GeomBin2d = ggplot(diamonds, aes(x, y)) +
    xlim(4, 10) +
    ylim(4, 10) +
    geom_bin2d(),
  GeomBox = ggplot(mpg, aes(class, hwy)) + geom_boxplot(),
  GeomContour = ggplot(faithfuld, aes(waiting, eruptions, z = density)) +
    geom_contour(),
  GeomContourRaster = ggplot(faithfuld, aes(waiting, eruptions, z = density)) +
    geom_raster(aes(fill = density)) +
    geom_contour(),
  GeomCount = ggplot(mpg, aes(cty, hwy)) + geom_count(),
  GeomDensity = ggplot(diamonds, aes(carat)) + geom_density(),
  GeomDensityColor = ggplot(diamonds, aes(depth, fill = cut, colour = cut)) +
    geom_density(alpha = 0.1),
  GeomDotPlot = ggplot(mtcars, aes(x = mpg)) + geom_dotplot(),
  GeomError = {
    ggplot(
      data.frame(
        trt = factor(c(1, 1, 2, 2)),
        resp = c(1, 5, 3, 4),
        group = factor(c(1, 2, 1, 2)),
        se = c(0.1, 0.3, 0.3, 0.2)
      ),
      aes(resp, trt, colour = group)
    ) +
      geom_point() +
      geom_errorbarh(aes(xmax = resp + se, xmin = resp - se))
  },
  GeomHex = ggplot(diamonds, aes(carat, price)) +
    geom_hex(binwidth = c(.1, 500)),
  GeomHistogram = ggplot(diamonds, aes(price, fill = cut)) +
    geom_histogram(binwidth = 500),
  GeomJitter = ggplot(mpg, aes(cyl, hwy)) + geom_jitter(width = 0.25),
  GeomErrorbar = ggplot(
    data.frame(
      trt = factor(c(1, 1, 2, 2)),
      resp = c(1, 5, 3, 4),
      group = factor(c(1, 2, 1, 2)),
      upper = c(1.1, 5.3, 3.3, 4.2),
      lower = c(0.8, 4.6, 2.4, 3.6)
    ),
    aes(trt, resp, fill = group)
  ) +
    geom_col(position = position_dodge(width = 0.9)) +
    geom_errorbar(
      aes(ymin = lower, ymax = upper),
      position = position_dodge(width = 0.9),
      width = 0.25
    ),
  GeomPolygon = ggplot(values) +
    geom_map(aes(map_id = id), map = positions) +
    expand_limits(positions),
  GeomPolygon2 = ggplot(values, aes(fill = value)) +
    geom_map(aes(map_id = id), map = positions) +
    expand_limits(positions),
  GeomLine = ggplot(economics_long, aes(date, value01, group = variable)) +
    geom_line(),
  GeomLine2 = ggplot(economics_long, aes(date, value01, colour = variable)) +
    geom_line(),
  GeomPoint = ggplot(diamonds, aes(carat, price)) + geom_point(alpha = 1 / 50),
  GeomPoint2 = ggplot(
    mtcars,
    aes(wt, mpg, color = factor(cyl), shape = factor(cyl))
  ) +
    geom_point(),
  GeomQuantile = ggplot(mpg, aes(displ, 1 / hwy)) +
    geom_point() +
    geom_quantile(),
  GeomRibbon = ggplot(huron, aes(year)) +
    geom_ribbon(aes(ymin = level - 1, ymax = level + 1), fill = "grey70") +
    geom_line(aes(y = level)),
  GeomRug = ggplot(mtcars, aes(wt, mpg)) + geom_point() + geom_rug(),
  GeomRug2 = ggplot(mtcars, aes(wt, mpg)) +
    geom_point() +
    geom_rug(outside = TRUE) +
    coord_cartesian(clip = "off"),
  GeomCurve = {
    df <- data.frame(x1 = 2.62, x2 = 3.57, y1 = 21.0, y2 = 15.0)
    ggplot(mtcars, aes(wt, mpg)) +
      geom_point() +
      geom_curve(
        aes(x = x1, y = y1, xend = x2, yend = y2, colour = "curve"),
        data = df
      ) +
      geom_segment(
        aes(x = x1, y = y1, xend = x2, yend = y2, colour = "segment"),
        data = df
      ) +
      geom_curve(
        aes(x = x1, y = y1, xend = x2, yend = y2),
        data = df,
        curvature = -0.2
      )
  },
  GeomSmooth = ggplot(
    diamonds[sample(nrow(diamonds), 1000), ],
    aes(carat, price)
  ) +
    geom_point(alpha = 0.2) +
    geom_smooth() +
    facet_wrap(~cut) +
    ggtitle("Diamond price by carat and cut"),
  GeomSmooth2 = ggplot(mpg, aes(displ, hwy, colour = class)) +
    geom_point() +
    geom_smooth(se = FALSE, method = lm),
  GeomSpoke = {
    df <- expand.grid(x = 1:10, y = 1:10)
    df$angle <- runif(100, 0, 2 * pi)
    df$speed <- runif(100, 0, sqrt(0.1 * df$x))

    ggplot(df, aes(x, y)) +
      geom_point() +
      geom_spoke(aes(angle = angle), radius = 0.5)
  },
  GeomText = ggplot(mtcars, aes(wt, mpg, label = rownames(mtcars))) +
    geom_text(),
  GeomLabel = ggplot(mtcars, aes(wt, mpg, label = rownames(mtcars))) +
    geom_label(),
  GeomTile = ggplot(
    data.frame(
      x = rep(c(2, 5, 7, 9, 12), 2),
      y = rep(c(1, 2), each = 5),
      z = factor(rep(1:5, each = 2)),
      w = rep(diff(c(0, 4, 6, 8, 10, 14)), 2)
    ),
    aes(x, y, width = w)
  ) +
    geom_tile(aes(fill = z)),
  GeomViolin = ggplot(mtcars, aes(factor(cyl), mpg)) + geom_violin()
)

lattice_examples <- list(
  settings = quote(show.settings()),
  densityplot = quote(
    densityplot(
      ~mpg,
      data = mtcars,
      group = factor(
        mtcars$am,
        levels = c(0, 1),
        labels = c("Automatic", "Manual")
      ),
      main = "MPG Distribution by Transmission Type",
      xlab = "Miles per Gallon",
      auto.key = TRUE
    )
  )
)

base_examples <- list(
  scatterplot = quote(
    plot(
      iris$Sepal.Length,
      iris$Sepal.Width,
      col = iris$Species,
      main = "Sepal Length vs Width in Iris"
    )
  ),
  hist = quote(hist(iris$Sepal.Length))
)

using Statistics
using Plots

struct DataPoint
    x::Float64
    y::Float64
    label::String
end

function analyze_data(points::Vector{DataPoint})::Dict{String, Any}
    values = [p.y for p in points]

    stats = Dict(
        "mean" => mean(values),
        "std" => std(values),
        "min" => minimum(values),
        "max" => maximum(values)
    )

    @info "Analyzed $(length(points)) data points"

    return stats
end

data = [DataPoint(i, sin(i * π / 10), "point_$i") for i in 1:10]
results = analyze_data(data)
println("Statistics: $(results["mean"])")

#' Value box
#'
#' @description
#' An opinionated ([card()]-powered) box, designed for displaying a `value` and
#' `title`. Optionally, a `showcase` can provide for context for what the
#' `value` represents (for example, it could hold a [bsicons::bs_icon()], or
#' even a [shiny::plotOutput()]). Find examples and template code you can use to
#' create engaging [value boxes on the bslib
#' website](https://rstudio.github.io/bslib/articles/value-boxes/index.html).
#'
#' @section Build-a-Box App:
#'
#'   Explore all of the `value_box()` options and layouts interactively with the
#'   [Build-a-Box app](https://bslib.shinyapps.io/build-a-box/), available
#'   online thanks to [shinyapps.io](https://www.shinyapps.io/). Or, you can
#'   run the app locally with:
#'
#'   ```r
#'   # shiny >= 1.8.1
#'   shiny::runExample("build-a-box", package = "bslib")
#'
#'   # shiny < 1.8.1
#'   shiny::runApp(system.file("examples-shiny", "build-a-box", package = "bslib"))
#'   ```
#'
#' @section Themes:
#'
#'   ```{r child="man/fragments/value-box-themes.Rmd"}
#'   ```
#'
#' @section Showcase Layouts:
#'
#'   ```{r child="man/fragments/value-box-showcase.Rmd"}
#'   ```
#'
#' @param title,value A string, number, or [htmltools::tag()] child to display as
#'   the title or value of the value box. The `title` appears above the `value`.
#' @param ... Unnamed arguments may be any [htmltools::tag()] children to
#'   display below `value`. Named arguments become attributes on the containing
#'   element.
#' @param showcase A [htmltools::tag()] child to showcase (e.g., a
#'   [bsicons::bs_icon()], a `plotly::plotlyOutput()`, etc).
#' @param showcase_layout One of `"left center"` (default), `"top right"` or
#'   `"bottom"`. Alternatively, you can customize the showcase layout options
#'   with the [showcase_left_center()], [showcase_top_right()], or
#'   [showcase_bottom()] functions. Use the options functions when you want to
#'   control the height or width of the showcase area.
#' @param theme The name of a theme for the value box, or a theme constructed
#'   with `value_box_theme()`. The theme names provide a convenient way to use
#'   your app's Bootstrap theme colors as the foreground or background colors of
#'   the value box. See below for more details on the provided themes. For more
#'   control, you can create your own theme with `value_box_theme()` where you
#'   can pass foreground and background colors directly. See the **Themes**
#'   section for more details.
#' @param class Utility classes for customizing the appearance of the summary
#'   card. Use `bg-*` and `text-*` classes (e.g, `"bg-danger"` and
#'   `"text-light"`) to customize the background/foreground colors.
#' @param fill Whether to allow the value box to grow/shrink to fit a fillable
#'   container with an opinionated height (e.g., `page_fillable()`).
#' @param max_height The maximum height of the `value_box()` or the showcase
#'   area when used in a `showcase_layout_*()` function. Can be any valid [CSS
#'   unit][htmltools::validateCssUnit] (e.g., `max_height="200px"`).
#' @param min_height The minimum height of the values box. Can be any valid [CSS
#'   unit][htmltools::validateCssUnit] (e.g., `min_height="200px"`).
#' @inheritParams card
#' @param theme_color `r lifecycle::badge("deprecated")` Use `theme` instead.
#'
#' @seealso Value boxes are a specialized form of a [card()] component.
#' @seealso [layout_columns()] and [layout_column_wrap()] help position multiple
#'   value boxes into columns and rows.
#'
#' @references Value boxes are featured on the bslib website in a few articles:
#'
#'   * [Value boxes](https://rstudio.github.io/bslib/articles/value-boxes/index.html)
#'   * [Build-a-Box App](https://bslib.shinyapps.io/build-a-box)
#'   * [Get Started: Dashboards](https://rstudio.github.io/bslib/articles/dashboards/index.html#value-boxes)
#'
#' @examplesIf rlang::is_interactive()
#' library(htmltools)
#'
#' value_box(
#'   "KPI Title",
#'   h1(HTML("$1 <i>Billion</i> Dollars")),
#'   span(
#'     bsicons::bs_icon("arrow-up"),
#'     " 30% VS PREVIOUS 30 DAYS"
#'   ),
#'   showcase = bsicons::bs_icon("piggy-bank"),
#'   theme = "success"
#' )
#'
#' @family Components
#' @export
value_box <- function(
  title,
  value,
  ...,
  showcase = NULL,
  showcase_layout = c("left center", "top right", "bottom"),
  full_screen = FALSE,
  theme = NULL,
  height = NULL,
  max_height = NULL,
  min_height = NULL,
  fill = TRUE,
  class = NULL,
  id = NULL,
  theme_color = deprecated()
) {
  dots <- separate_arguments(...)
  attribs <- dots$attribs
  children <- dots$children

  # ---- Title and value ----
  if (rlang::is_bare_character(title) || rlang::is_bare_numeric(title)) {
    title <- tags$p(title)
  }
  if (rlang::is_bare_character(value) || rlang::is_bare_numeric(value)) {
    value <- tags$p(value)
  }

  if (!is.null(title)) {
    title <- tag_add_outer_class(title, class = "value-box-title")
  }
  if (!is.null(value)) {
    value <- tag_add_outer_class(value, class = "value-box-value")
  }

  # ---- Contents ----
  contents <- div(class = "value-box-area", title, value, !!!children)
  contents <- as_fill_carrier(contents)

  # ---- Showcase ----
  if (!is.null(showcase)) {
    showcase_layout <- resolve_showcase_layout(showcase_layout)
    contents <- render_showcase_layout(showcase_layout, showcase, contents)
  }

  # ---- Theme ----
  if (lifecycle::is_present(theme_color)) {
    lifecycle::deprecate_soft(
      "0.6.0",
      "value_box(theme_color =)",
      "value_box(theme =)",
      details = if (!missing(theme)) {
        "Both `theme` and `theme_color` were provided, but only the `theme` argument will be used."
      }
    )

    if (missing(theme)) {
      theme <- theme_color
    }
  }

  if (!inherits(theme, "bslib_value_box_theme")) {
    theme <- value_box_theme(theme)
  }

  # ---- Layout ----
  res <- card(
    class = c(
      "bslib-value-box",
      theme$class,
      class,
      if (!is.null(showcase)) showcase_layout$class
    ),
    id = id,
    full_screen = full_screen,
    height = height,
    max_height = max_height,
    min_height = min_height,
    fill = fill,
    style = css(
      color = theme$fg,
      background_color = theme$bg,
      # These variables are used by the full screen card button
      "--bslib-color-fg" = theme$fg,
      "--bslib-color-bg" = theme$bg
    ),
    !!!attribs,
    contents
  )

  as_fragment(tag_require(res, version = 5, caller = "value_box()"))
}

#' @param name The name of the theme, e.g. `"primary"`, `"danger"`, `"purple"`.
#' @param bg,fg The background and foreground colors for the theme. If only `bg`
#'   is provided, then the foreground color is automatically chosen from
#'   `$black` or `$white` to provide the best contrast with the background
#'   color.
#'
#' @rdname value_box
#' @export
value_box_theme <- function(name = NULL, bg = NULL, fg = NULL) {
  if (is.null(name)) {
    if (is.null(bg)) {
      name <- "default"
    } else {
      # don't warn if we can't get a contrast color, `bg` might be valid
      # CSS but not something sass can compute on
      fg <- fg %||% suppressWarnings(get_color_contrast(bg))
    }

    return(new_value_box_theme(name, bg, fg))
  }

  if (!rlang::is_string(name)) {
    rlang::abort(
      '`theme` must be a single value, e.g. "primary", "danger", "purple", etc.'
    )
  }

  if (!grepl("^(text|bg)-", name)) {
    name <- paste0("bg-", name)
  }

  new_value_box_theme(name, bg, fg)
}

new_value_box_theme <- function(class = NULL, bg = NULL, fg = NULL) {
  structure(
    list(
      class = class,
      fg = fg,
      bg = bg
    ),
    class = "bslib_value_box_theme"
  )
}

#' @param width,width_full_screen,height,height_full_screen one of the
#'   following:
#'   * A proportion (i.e., a number between 0 and 1) of available width or
#'     height to allocate to the showcase.
#'   * A valid [CSS unit][htmltools::validateCssUnit] defining the width or
#'     height of the showcase column, or a valid value accepted by the
#'     `grid-template-columns` (width) or `grid-template-rows` (height) CSS
#'     property to define the width or height of the showcase column or row.
#'     Accepted values in the second category are `"auto"`, `"min-content"`,
#'     `"max-content"`, a fractional unit (e.g. `2fr`), or a `minmax()` function
#'     (e.g., `minmax(100px, 1fr)`).
#' @param max_height_full_screen A proportion (i.e., a number between
#'   0 and 1) or any valid [CSS unit][htmltools::validateCssUnit] defining the
#'   showcase `max_height` in a full screen card.
#'
#' @export
#' @rdname value_box
showcase_left_center <- function(
  width = 0.3,
  width_full_screen = "1fr",
  max_height = "100px",
  max_height_full_screen = 0.67
) {
  new_showcase_layout(
    class = "showcase-left-center",
    width = width,
    width_full_screen = width_full_screen,
    max_height = max_height,
    max_height_full_screen = max_height_full_screen
  )
}

#' @export
#' @rdname value_box
showcase_top_right <- function(
  width = 0.4,
  width_full_screen = "1fr",
  max_height = "75px",
  max_height_full_screen = 0.67
) {
  new_showcase_layout(
    class = "showcase-top-right",
    width = width,
    width_full_screen = width_full_screen,
    max_height = max_height,
    max_height_full_screen = max_height_full_screen
  )
}

#' @export
#' @rdname value_box
showcase_bottom <- function(
  width = "100%",
  width_full_screen = NULL,
  height = "auto",
  height_full_screen = "2fr",
  max_height = "100px",
  max_height_full_screen = NULL
) {
  new_showcase_layout(
    class = "showcase-bottom",
    width = width,
    width_full_screen = width_full_screen,
    height = height,
    height_full_screen = height_full_screen,
    max_height = max_height,
    max_height_full_screen = max_height_full_screen
  )
}

new_showcase_layout <- function(
  class,
  width = 0.3,
  width_full_screen = "1fr",
  height = NULL,
  height_full_screen = NULL,
  max_height = "100px",
  max_height_full_screen = 0.67
) {
  width <- validate_grid_unit(width)
  width_full_screen <- validate_grid_unit(width_full_screen)

  height <- validate_grid_unit(height)
  height_full_screen <- validate_grid_unit(height_full_screen)

  max_height <- validate_height_unit(max_height)
  max_height_full_screen <- validate_height_unit(max_height_full_screen)

  structure(
    list(
      class = class,
      width = width,
      width_full_screen = width_full_screen,
      height = height,
      height_full_screen = height_full_screen,
      max_height = max_height,
      max_height_full_screen = max_height_full_screen
    ),
    class = "bslib_showcase_layout"
  )
}

resolve_showcase_layout <- function(showcase_layout) {
  if (inherits(showcase_layout, "bslib_showcase_layout")) {
    return(showcase_layout)
  }

  if (is.character(showcase_layout)) {
    layout_choices <- c("left center", "top right", "bottom")
    showcase_layout <- rlang::arg_match(showcase_layout, layout_choices)
    showcase_layout <- switch(
      showcase_layout,
      "left center" = showcase_left_center(),
      "top right" = showcase_top_right(),
      bottom = showcase_bottom()
    )
    return(showcase_layout)
  }

  rlang::abort("`showcase_layout` must be a string or a `showcase_*()` layout")
}

render_showcase_layout <- function(showcase_layout, showcase, contents) {
  stopifnot(inherits(showcase_layout, "bslib_showcase_layout"))

  showcase <- div(class = "value-box-showcase", showcase)
  showcase <- as_fill_carrier(showcase)

  grid_props <- css(
    "--bslib-grid-height" = "auto",
    "--bslib-grid-height-mobile" = "auto",
    # These are private variables but css() can't handle _ prefix yet
    # FIXME: rstudio/htmltools#397
    "---bslib-value-box-showcase-w" = showcase_layout$width,
    "---bslib-value-box-showcase-w-fs" = showcase_layout$width_full_screen,
    "---bslib-value-box-showcase-h" = showcase_layout$height,
    "---bslib-value-box-showcase-h-fs" = showcase_layout$height_full_screen,
    "---bslib-value-box-showcase-max-h" = showcase_layout$max_height,
    "---bslib-value-box-showcase-max-h-fs" = showcase_layout$max_height_full_screen
  )

  value_box_grid <- div(
    class = "value-box-grid",
    style = grid_props,
    showcase, # .value-box-showcase
    contents # .value-box-area (prepared in value_box())
  )

  card_body(
    style = css(padding = 0),
    fillable = TRUE,
    as_fill_item(value_box_grid)
  )
}

# It seems to be to use % over fr here since there is no gap on the grid
validate_grid_unit <- function(x) {
  if (is.null(x)) {
    return(x)
  }

  if (!is_01_scalar(x)) {
    if (tolower(x) %in% c("auto", "min-content", "max-content")) {
      return(tolower(x))
    }
    if (grepl("^minmax\\(", x)) {
      return(x)
    }
    if (!grepl("\\d+\\s*fr", x)) {
      x <- validateCssUnit(x)
    }
    return(x)
  }
  paste0(100 * x, "%")
}

validate_height_unit <- function(x) {
  if (is.null(x)) {
    return(x)
  }

  if (!is_01_scalar(x)) {
    return(validateCssUnit(x))
  }
  paste0(100 * x, "%")
}

is_01_scalar <- function(x) {
  rlang::is_scalar_double(x) && x >= 0 && x <= 1
}

# Print methods ---------------------------------------

#' @export
print.bslib_value_box_theme <- function(x, ...) {
  cat("<bslib_value_box_theme>\n")
  if (!is.null(x$class)) {
    cat("theme: ", x$class, "\n", sep = "")
  }
  if (!is.null(x$bg)) {
    cat("background-color: ", x$bg, "\n", sep = "")
  }
  if (!is.null(x$fg)) {
    cat("color: ", x$fg, "\n", sep = "")
  }
  invisible(x)
}

#' @export
print.bslib_showcase_layout <- function(x, ...) {
  cat("<", x$class, ">\n", sep = "")

  fields <- c("width", "height", "max_height")
  unset <- "---"
  nchar_fields <- max(nchar(fields))
  nchar_value <- max(vapply(dropNulls(x[fields]), nchar, integer(1)))

  if (any(vapply(x[fields], is.null, logical(1)))) {
    nchar_value <- max(nchar_value, nchar(unset))
  }

  for (field in fields) {
    t_field <- format(
      paste0(field, ": "),
      width = nchar_fields + 2,
      align = "left"
    )
    t_value <- format(
      x[[field]] %||% unset,
      width = nchar_value,
      align = "right"
    )
    t_value_fs <- x[[paste0(field, "_full_screen")]] %||% unset

    cat(t_field, t_value, " [fs: ", t_value_fs, "]\n", sep = "")
  }

  invisible(x)
}

ui_global_controls <- function(id) {
  ns <- shiny::NS(id)

  tagList(
    layout_columns(
      class = "align-items-end",
      selectizeInput(
        ns("theme_style"),
        "Theme style",
        list(
          "Default" = "default",
          "All" = "all",
          "Semantic Colors" = list(
            "Semantic Background" = "semantic-bg",
            "Semantic Text" = "semantic-fg"
          ),
          "Theme Colors" = list(
            "Colored Background" = "colors-bg",
            "Colored Text" = "colors-fg"
          ),
          "Vibrant" = list(
            "Gradient Background" = "gradient"
          )
        )
      ),
      shuffleButton(ns("random_theme"), "Theme"),
      shuffleButton(ns("random_stat"), "Stats")
    ),
    layout_columns(
      class = "align-items-start",
      div(
        radioButtons(
          ns("showcase_item"),
          "Showcase Item",
          choices = c("Plot", "Icon"),
          inline = TRUE
        ),
        conditionalPanel(
          "input.showcase_item == 'Plot'",
          ns = ns,
          p(
            class = "text-muted",
            "See",
            tags$a(
              href = "https://rstudio.github.io/bslib/articles/value-boxes/index.html#expandable-sparklines",
              "Expandable Sparklines"
            ),
            "for example plot code."
          )
        ),
        conditionalPanel(
          "input.showcase_item == 'Icon'",
          ns = ns,
          shuffleButton(ns("random_icon"), "Icons")
        )
      ),
      radioButtons(
        ns("showcase_layout"),
        "Showcase Layout",
        choices = c("Left center", "Top right", "Bottom"),
        inline = TRUE
      )
    )
  )
}

server_global_controls <- function(input, output, sessions, one, two, three) {
  observeEvent(
    c(input$random_theme, input$theme_style),
    {
      new_values <- switch(
        input$theme_style,
        all = {
          one$theme$shuffle()
          two$theme$shuffle()
          three$theme$shuffle()
          NULL
        },
        default = {
          one$theme$set("Default")
          two$theme$set("Default")
          three$theme$set("Default")
          NULL
        },
        "semantic-bg" = sample(setdiff(theme_colors, c("light", "dark")), 3),
        "semantic-fg" = paste0(
          "text-",
          sample(setdiff(theme_colors, c("light", "dark")), 3)
        ),
        "colors-bg" = sample(named_colors, 3, replace = TRUE),
        "colors-fg" = paste0("text-", sample(named_colors, 3, replace = TRUE)),
        gradient = sample(gradient_classes, 3)
      )

      if (is.null(new_values)) {
        return()
      }

      one$theme$set(new_values[[1]])
      two$theme$set(new_values[[2]])
      three$theme$set(new_values[[3]])
    },
    ignoreInit = TRUE
  )

  observeEvent(input$random_stat, {
    one$random_stat()
    two$random_stat()
    three$random_stat()
  })

  observeEvent(input$random_icon, {
    one$showcase_icon$shuffle()
    two$showcase_icon$shuffle()
    three$showcase_icon$shuffle()
  })

  observeEvent(
    input$showcase_item,
    {
      item <- tolower(input$showcase_item)
      one$set_showcase_item(item)
      two$set_showcase_item(item)
      three$set_showcase_item(item)
    },
    ignoreInit = TRUE
  )

  observeEvent(
    input$showcase_layout,
    {
      layout <- tolower(input$showcase_layout)
      one$set_showcase_layout(layout)
      two$set_showcase_layout(layout)
      three$set_showcase_layout(layout)
    },
    ignoreInit = TRUE
  )
}

module_global_controls <- function(id, one, two, three) {
  callModule(server_global_controls, id, one = one, two = two, three = three)
}

tag_simple <- function(...) {
  htmltools::div(class = "test", id = "test", ...)
}

tag_nested <- function(...) {
  htmltools::withTags(
    div(
      div(class = "inner"),
      div(
        class = "nested",
        div(class = "inner deep")
      ),
      div(class = "inner"),
      ...
    )
  )
}

expect_equal_html <- function(actual, expected, ...) {
  expect_equal(
    format(actual),
    format(expected),
    ...
  )
}

# expect should be a css string with no spaces or linebreaks
# actual should be the output of sass compilation
expect_css <- function(expect, actual) {
  expect_identical(
    gsub("\\s+|\\n", "", as.character(expect)),
    gsub("\\s+|\\n", "", as.character(actual))
  )
}

#!/bin/bash

# Script to process log files and generate a summary report
# Author: Data Team

OUTPUT_DIR="${1:-.}"
LOG_FILES=( /var/log/*.log )

# Function to count error occurrences
count_errors() {
    local file="$1"
    grep -c "ERROR" "$file" 2>/dev/null || echo 0
}

# Function to extract timestamp
get_timestamp() {
    date "+%Y-%m-%d %H:%M:%S"
}

# Main processing loop
echo "Report generated on $(get_timestamp)" > "$OUTPUT_DIR/report.txt"

for logfile in "${LOG_FILES[@]}"; do
    if [[ -f "$logfile" ]]; then
        error_count=$(count_errors "$logfile")
        if (( error_count > 0 )); then
            echo "Processing $logfile: $error_count errors found"
            tail -n 10 "$logfile" | grep "ERROR" >> "$OUTPUT_DIR/errors.log"
        fi
    fi
done

echo "Report complete. Results saved to $OUTPUT_DIR"

# input_submit_textarea() markup snapshots Code input_submit_textarea("test") Output
    ---

        Code
          input_submit_textarea("test", label = "Enter text", placeholder = "Type here...",
            value = "Initial value", width = 300, rows = 3, button = tags$button(id = "custom_submit",
              class = "btn btn-primary", "Send"), toolbar = tagList(tags$span("Press"),
            tags$kbd("Enter"), tags$span("to submit")), submit_key = "enter", spellcheck = "false",
            autocomplete = "off", )
        Output
Enter text
    Initial value
    # input_submit_textarea() validation errors

        Code
          input_submit_textarea("test", "With Children", div())
        Condition
          Error in `input_submit_textarea()`:
          ! All `...` arguments must be named
          i Did you mean to pass UI elements to `toolbar`?

    ---

        Code
          input_submit_textarea("test", value = 123)
        Condition
          Error in `input_submit_textarea()`:
          ! `value` must be a character string

    ---

        Code
          input_submit_textarea("test", value = c("a", "b"))
        Condition
          Error in `input_submit_textarea()`:
          ! `value` must be a character string

    ---

        Code
          input_submit_textarea("test", button = tags$div("Not a button"))
        Condition
          Error in `input_submit_textarea()`:
          ! `button` must be a `tags$button()`

    ---

        Code
          input_submit_textarea("test", submit_key = "invalid")
        Condition
          Error in `input_submit_textarea()`:
          ! `submit_key` must be one of "enter+modifier" or "enter", not "invalid".

# navbar_options() print method

    Code
      navbar_options()
    Output
      <bslib_navbar_options>
         position: (static-top)
            theme: (auto)
      collapsible: (TRUE)
        underline: (TRUE)

---

    Code
      navbar_options(theme = "dark", bg = "red")
    Output
      <bslib_navbar_options>
         position: (static-top)
               bg: red
            theme: dark
      collapsible: (TRUE)
        underline: (TRUE)

---

    Code
      navbar_options(position = "static-top", theme = "auto", collapsible = TRUE)
    Output
      <bslib_navbar_options>
         position: static-top
            theme: auto
      collapsible: TRUE
        underline: (TRUE)

# navbar markup snapshots Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 3))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 4))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 5))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 4), navbar_options = navbar_options( theme = "dark"))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 4), navbar_options = navbar_options( theme = "light"))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 4), navbar_options = navbar_options( bg = "#000"))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 5), navbar_options = navbar_options( theme = "dark"))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 5), navbar_options = navbar_options( theme = "light"))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 5), navbar_options = navbar_options( bg = "#000"))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 5), navbar_options = navbar_options( theme = "light", `data-bs-theme` = "dark"))) Output
--- Code show_navbar_markup(navs_bar_(theme = bs_theme(version = 5), navbar_options = navbar_options( class = "bg-primary", theme = "dark"))) Output

# print.bslib_showcase_layout()

    Code
      showcase_left_center()
    Output
      <showcase-left-center>
      width:      30%   [fs: 1fr]
      height:     ---   [fs: ---]
      max_height: 100px [fs: 67%]

---

    Code
      showcase_top_right()
    Output
      <showcase-top-right>
      width:      40%  [fs: 1fr]
      height:     ---  [fs: ---]
      max_height: 75px [fs: 67%]

---

    Code
      showcase_bottom()
    Output
      <showcase-bottom>
      width:      100%  [fs: ---]
      height:     auto  [fs: 2fr]
      max_height: 100px [fs: ---]

# value_box_theme() print method

    Code
      value_box_theme("bg-gradient-blue-purple")
    Output
      <bslib_value_box_theme>
      theme: bg-gradient-blue-purple

---

    Code
      value_box_theme("red", fg = "white")
    Output
      <bslib_value_box_theme>
      theme: bg-red
      color: white

---

    Code
      value_box_theme(bg = "black")
    Output
      <bslib_value_box_theme>
      background-color: black
      color: #FFFFFF

---

    Code
      value_box_theme("text-red", bg = "#FFE8E8")
    Output
      <bslib_value_box_theme>
      theme: text-red
      background-color: #FFE8E8

# input_code_editor() rejects invalid themes

    Code
      input_code_editor("test", theme_light = "invalid-theme")
    Condition
      Error in `input_code_editor()`:
      ! `theme_light` must be one of "atom-one-dark", "dracula", "github-dark-dimmed", "github-dark", "github-light", "night-owl-light", "night-owl", "prism-okaidia", "prism-solarized-light", "prism-tomorrow", "prism-twilight", "prism", "vs-code-dark", or "vs-code-light", not "invalid-theme".

# arg_match_language rejects invalid languages

    Code
      input_code_editor("test", language = "fortran")
    Condition
      Error in `input_code_editor()`:
      ! `language` must be one of "r", "python", "julia", "sql", "javascript", "typescript", "markup", "css", "scss", "sass", "json", "markdown", "yaml", "xml", "toml", "ini", "bash", "docker", "latex", "cpp", "rust", "diff", "md", "html", "plain", "plaintext", "text", or "txt", not "fortran".

# input_code_editor validates theme names

    Code
      input_code_editor("test", theme_light = "invalid-theme")
    Condition
      Error in `input_code_editor()`:
      ! `theme_light` must be one of "atom-one-dark", "dracula", "github-dark-dimmed", "github-dark", "github-light", "night-owl-light", "night-owl", "prism-okaidia", "prism-solarized-light", "prism-tomorrow", "prism-twilight", "prism", "vs-code-dark", or "vs-code-light", not "invalid-theme".
    Code
      input_code_editor("test", theme_dark = "invalid-theme")
    Condition
      Error in `input_code_editor()`:
      ! `theme_dark` must be one of "atom-one-dark", "dracula", "github-dark-dimmed", "github-dark", "github-light", "night-owl-light", "night-owl", "prism-okaidia", "prism-solarized-light", "prism-tomorrow", "prism-twilight", "prism", "vs-code-dark", or "vs-code-light", not "invalid-theme".

# input_code_editor validates `language`

    Code
      input_code_editor("test", language = "fortran")
    Condition
      Error in `input_code_editor()`:
      ! `language` must be one of "r", "python", "julia", "sql", "javascript", "typescript", "markup", "css", "scss", "sass", "json", "markdown", "yaml", "xml", "toml", "ini", "bash", "docker", "latex", "cpp", "rust", "diff", "md", "html", "plain", "plaintext", "text", or "txt", not "fortran".

# update_code_editor validates inputs

    Code
      update_code_editor("test", language = "fortran", session = NULL)
    Condition
      Error in `update_code_editor()`:
      ! `language` must be one of "r", "python", "julia", "sql", "javascript", "typescript", "markup", "css", "scss", "sass", "json", "markdown", "yaml", "xml", "toml", "ini", "bash", "docker", "latex", "cpp", "rust", "diff", "md", "html", "plain", "plaintext", "text", or "txt", not "fortran".

---

    Code
      update_code_editor("test", theme_light = "invalid", session = NULL)
    Condition
      Error in `update_code_editor()`:
      ! `theme_light` must be one of "atom-one-dark", "dracula", "github-dark-dimmed", "github-dark", "github-light", "night-owl-light", "night-owl", "prism-okaidia", "prism-solarized-light", "prism-tomorrow", "prism-twilight", "prism", "vs-code-dark", or "vs-code-light", not "invalid".

---

    Code
      update_code_editor("test", theme_dark = "invalid", session = NULL)
    Condition
      Error in `update_code_editor()`:
      ! `theme_dark` must be one of "atom-one-dark", "dracula", "github-dark-dimmed", "github-dark", "github-light", "night-owl-light", "night-owl", "prism-okaidia", "prism-solarized-light", "prism-tomorrow", "prism-twilight", "prism", "vs-code-dark", or "vs-code-light", not "invalid".

---

    Code
      update_code_editor("test", indentation = "invalid", session = NULL)
    Condition
      Error in `update_code_editor()`:
      ! `indentation` must be one of "space" or "tab", not "invalid".

    # layout_columns() with col_widths

        Code
          layout_columns(col_widths = 6, !!!children)
        Output
--- Code layout_columns(col_widths = c(4, 8), !!!children) Output
    ---

        Code
          layout_columns(col_widths = breakpoints(sm = 6, md = 4, lg = 3), !!!children)
        Output
    ---

        Code
          layout_columns(col_widths = breakpoints(sm = NA, lg = c(4, 8)), !!!children)
        Output
    ---

        Code
          layout_columns(col_widths = breakpoints(sm = NA, lg = c(4, -4, 4)), !!!children)
        Output
    # grid_item_container()

        Code
          grid_item_container(div(class = "layout-column-child-element"), fillable = TRUE)
        Output
    ---

        Code
          grid_item_container(div(class = "layout-column-child-element"), fillable = FALSE)
        Output
# breakpoints() re-orders well-known breaks and test print method Code bp Output xs: 1 sm: 2 md: 3 lg: 4 xl: 5 xxl: 6 giant: 7 huge: 8 # breakpoints() has correct classes and structure Code bp Output sm: 1 -1 1 md: -1 2 -2 3 lg: -2 2 -1 -2 3 -2 xl: 1 2 3 NA huge: 1 2 3 auto # layout_columns() throws if NAs are mixed with other column values Cannot mix widths and `NA` values. All column widths must be specified, or choose auto widths using a single `NA` value. # layout_columns() warns when too many column widths More column widths than children at breakpoint 'md', extra widths will be ignored.

    # card_image()

        Code
          show_raw_html(card(card_image("https://example.com/image.jpg"), card_body(
            "image cap on top of card")))
        Output
image cap on top of card
    ---

        Code
          show_raw_html(card(card_body("image cap on bottom of card"), card_image(
            "https://example.com/image.jpg")))
        Output
image cap on bottom of card
    ---

        Code
          show_raw_html(card(card_header("header"), card_image(
            "https://example.com/image.jpg"), card_body("image not a cap")))
        Output
header
image not a cap
    ---

        Code
          show_raw_html(card(card_image("https://example.com/image.jpg", alt = "card-img")))
        Output
    # card_image() input validation

        Code
          card_image("cat.jpg")
        Condition
          Error in `card_image()`:
          ! `file` does not exist: cat.jpg
          i If `file` is a remote file or will be served by the Shiny app, use a URL or set `src = "cat.jpg"`.

    ---

        Code
          card_image("foo", "bar")
        Condition
          Error in `card_image()`:
          ! Unnamed arguments were included in `...`.
          i All additional arguments to `card_image()` in `...` should be named attributes for the `
` tag. --- Code card_image("foo", border_radius = "guess") Condition Error in `card_image()`: ! `border_radius` must be one of "auto", "top", "bottom", "all", or "none", not "guess".

    # toast() validates position argument

        Code
          toast("Test", position = "invalid")
        Condition
          Error in `normalize_toast_position()`:
          ! Invalid toast position: 'invalid'. Must specify one vertical position (top, middle, bottom) and one horizontal position (left, center, right).

    # toast() validates type argument

        Code
          toast("Test", type = "invalid")
        Condition
          Error in `toast()`:
          ! `type` must be one of "primary", "secondary", "success", "info", "warning", "danger", "error", "light", or "dark", not "invalid".

    # toast() duration_s throws for invalid values

        Code
          toast("Test", duration_s = -5)
        Condition
          Error in `toast()`:
          ! `duration_s` must be a single non-negative number or NA.
        Code
          toast("Test", duration_s = "invalid")
        Condition
          Error in `toast()`:
          ! `duration_s` must be a single non-negative number or NA.
        Code
          toast("Test", duration_s = c(5, 10))
        Condition
          Error in `toast()`:
          ! `duration_s` must be a single non-negative number or NA.

    # as.tags.bslib_toast creates proper HTML structure

        Code
          cat(format(tag))
        Output
Test
    # as.tags.bslib_toast respects accessibility attributes

        Code
          cat(format(as.tags(t_danger)))
        Output
Error message
    ---

        Code
          cat(format(as.tags(t_info)))
        Output
Info message
    ---

        Code
          cat(format(as.tags(t_default)))
        Output
Default message
    # as.tags.bslib_toast includes close button appropriately

        Code
          cat(format(as.tags(t_header)))
        Output
Title
Message
    ---

        Code
          cat(format(as.tags(t_no_header)))
        Output
Message
    ---

        Code
          cat(format(as.tags(t_non_closable)))
        Output
Message
    ---

        Code
          cat(format(as.tags(t_manual)))
        Output
Message
# toast() icon renders in body without header Code cat(format(tag)) Output
★
You have new messages
# toast() icon renders in body with header Code cat(format(tag)) Output
New Mail
★
Message content
# toast() icon works with closable button in body Code cat(format(tag)) Output
★
Warning message
    # toast_header() icon renders in header

        Code
          cat(format(tag))
        Output
★ Notification now
Body content
    # toast_header() icon with status and title

        Code
          cat(format(tag))
        Output
✓ Success just now
Operation completed
# toast with both header icon and body icon Code cat(format(tag)) Output
H Title
B
Message content
# normalize_toast_position() errors on invalid input Code normalize_toast_position("top") Condition Error in `normalize_toast_position()`: ! Invalid toast position: 'top'. Must specify one vertical position (top, middle, bottom) and one horizontal position (left, center, right). Code normalize_toast_position("left") Condition Error in `normalize_toast_position()`: ! Invalid toast position: 'left'. Must specify one vertical position (top, middle, bottom) and one horizontal position (left, center, right). Code normalize_toast_position("top bottom left") Condition Error in `normalize_toast_position()`: ! Invalid toast position: 'top bottom left'. Must specify one vertical position (top, middle, bottom) and one horizontal position (left, center, right). Code normalize_toast_position(c("top", "bottom", "left")) Condition Error in `normalize_toast_position()`: ! Invalid toast position: 'top bottom left'. Must specify one vertical position (top, middle, bottom) and one horizontal position (left, center, right). Code normalize_toast_position("top invalid") Condition Error in `normalize_toast_position()`: ! Invalid toast position: 'top invalid'. Must specify one vertical position (top, middle, bottom) and one horizontal position (left, center, right). Code normalize_toast_position("foo bar") Condition Error in `normalize_toast_position()`: ! Invalid toast position: 'foo bar'. Must specify one vertical position (top, middle, bottom) and one horizontal position (left, center, right). # show_toast() and hide_toast() warn if nothing to show/hide Code hide_toast(show_toast(toast(), session = session), session = session) Condition Warning: `toast` has no content; no toast to show. Warning: `id` is NULL; no toast to hide. # hide_toast() works Code hide_toast(toast()) Condition Error in `hide_toast()`: ! Cannot hide a toast without an ID. Provide the toast ID.

```{r 0-screenshot-examples, include = FALSE}
.UPDATE_SCREENSHOT <-
  isTRUE(knitr::opts_chunk$get(".update_screenshot")) ||
  nzchar(Sys.getenv("CI"))

if (.UPDATE_SCREENSHOT) {
  library(chromote)
  tryCatch(
    {
      b <- ChromoteSession$new(width = 300, height = 800, wait_ = TRUE)
      knitr::opts_knit$set(document = function(x) {
        b$close()
        x
      })
    },
    error = function(err) {
      message("Failed to start chromote session. Screenshots will not be updated.")
      .UPDATE_SCREENSHOT <<- FALSE
    }
  )
}

render_screenshot <- function(x, options) {
  ss_width <- options$fig.width
  ss_height <- options$fig.height
  dpi <- options$dpi

  ss_width <- ss_width * dpi
  ss_height <- ss_height * dpi

  b$Emulation$setDeviceMetricsOverride(
    width = ss_width,
    height = ss_height,
    deviceScaleFactor = 0,
    mobile = FALSE,
    wait_ = TRUE
  )

  if (inherits(x, "bslib_fragment")) {
    x <- attr(x, "bslib_page")(x)
  }

  # Without this hack, bslib won't generate BS4+ compliant nav markup (when
  # rendered statically). We should really fix this situation...
  oldTheme <- shiny::getCurrentTheme()
  shiny:::setCurrentTheme(bs_theme())
  on.exit(shiny:::setCurrentTheme(oldTheme), add = TRUE)

  page_style <- sprintf(
    "body { min-height: %spx; padding: 1rem; overflow: hidden; }
    .bslib-card .card-body { min-height: 200px; }",
    ss_height
  )

  x <- tagList(
    x,
    tags$head(tags$style(HTML(page_style)))
  )
  tmpdir <- tempfile()
  dir.create(tmpdir)
  on.exit(unlink(tmpdir, recursive = TRUE))

  tmpfile <- file.path(tmpdir, basename(tempfile(fileext = ".html")))
  save_html(x, tmpfile)

  file_out <- paste0(
    gsub("_", "-", options$label), ".png"
  )

  {
    p <- b$Page$loadEventFired(wait_ = FALSE)
    b$Page$navigate(paste0("file://", tmpfile), wait_ = FALSE)
    b$wait_for(p)
  }

  b$Runtime$evaluate(
    "if ($('.dropdown-toggle').length > 0) $('.dropdown-toggle').dropdown('toggle')"
  )
  b$screenshot(filename = file.path(
    rprojroot::find_package_root_file(),
    "man/figures/", file_out
  ))

  invisible(file_out)
}

include_screenshot <- function(x, options) {
  file_name <- paste0(
    gsub("_", "-", options$label), ".png"
  )
  alt <- options$fig.alt
  if (is.null(alt)) {
    alt <- sprintf("Screenshot of a %s() example.", options$label)
  }
  img <- sprintf('![](%s "%s")', file_name, alt)
  knitr::asis_output(img)
}

og_render <- knitr::opts_chunk$get("render")

knitr::opts_chunk$set(
  render = function(x, options) {
    no_screenshot <- isTRUE(options$no_screenshot)
    if (!(inherits(x, "shiny.tag") || inherits(x, "shiny.tag.list"))) {
      no_screenshot <- TRUE
    }

    if (no_screenshot) {
      if (is.null(og_render)) return(x)
      else return(og_render(x, options))
    }

    if (isTRUE(.UPDATE_SCREENSHOT)) {
      render_screenshot(x, options)
    }
    include_screenshot(x, options)
  }
)
```

---
title: "Theming with bslib and thematic"
output: 
  html_document:
    code_folding: show
    theme:
      bg: "#202123"
      fg: "#B8BCC2"
      primary: "#EA80FC"
      secondary: "#00DAC6"
      base_font:
        google: Prompt
      heading_font:
        google: Proza Libre
---

```{r setup, include=FALSE}
if (requireNamespace("thematic")) 
  thematic::thematic_rmd(font = "auto")
```

## R Markdown

This is an R Markdown document themed with [`{bslib}` package](https://rstudio.github.io/bslib/). `{bslib}` makes it easy to customize the main colors and fonts of a `html_document`, [`flexdashboard::flex_dashboard`](https://flexdashboard-pkg.netlify.app/articles/articles/theme.html), [shiny::fluidPage()](https://shiny.rstudio.com/reference/shiny/latest/fluidPage.html), or more generally any website that uses [Bootstrap](https://getbootstrap.com/) for styling. The `theme` parameter in the yaml front-matter of this Rmd document describes a [`bslib::bs_theme()`](https://rstudio.github.io/bslib/reference/bs_theme.html) object, which provides access to 100s of [theming options](https://rstudio.github.io/bslib/articles/bs5-variables.html) (via its `...` argument) in addition to the main options demonstrated here (e.g., `bg`, `fg`, `primary`, etc).

This particular example uses `{bslib}`'s default [Bootstrap version (which, at the time of writing, is Bootstrap 5)](https://rstudio.github.io/bslib/articles/bslib.html#versions). However, if reproducibility is important, it's recommended that you "lock-in" the version by adding `version: 5` to the `theme` definition.

## Themed Plots {.tabset .tabset-pills}

When running this document with [`{thematic}`](https://rstudio.github.io/thematic/) installed, the `thematic::thematic_rmd(font = "auto")` effectively translates `theme` (CSS) settings to new global theming defaults for `{ggplot2}`, `{lattice}`, and `{base}` R graphics:

### ggplot2

```{r}
library(ggplot2)

ggplot(mpg, aes(displ, hwy)) +
  geom_point() + geom_smooth()
```

### lattice

```{r}
lattice::show.settings()
```

### base

```{r}
plot(pressure, col = thematic::thematic_get_option("accent"))
```

_Just in case it isn't obvious, this text was written by an LLM._

# Component Documentation

The Shiny for Python framework, available at [shiny.posit.co/py](https://shiny.posit.co/py/),
provides a comprehensive set of UI components for building interactive web applications. These
components are designed with **consistency and usability** in mind, making it easier for
developers to create professional-grade applications.

Our framework implements the `ui.page_navbar()` component as the primary navigation structure,
allowing for intuitive organization of content across multiple views. Each view can contain
various input and output elements, managed through the `ui.card()` container system.

## Component Architecture

*The architecture of our application framework* emphasizes modularity and reusability. Key
components like `ui.value_box()` and `ui.layout_column_wrap()` work together to create
structured, responsive layouts that adapt to different screen sizes.

<table class="table table-striped">
<thead>
<tr>
<th>Component</th>
<th>Implementation</th>
<th>Use Case</th>
<th>Status</th>
</tr>
</thead>
<tbody>
<tr>
<td>Value Box</td>
<td><code>ui.value_box()</code></td>
<td>Metric Display</td>
<td>Production Ready</td>
</tr>
<tr>
<td>Card</td>
<td><code>ui.card()</code></td>
<td>Content Container</td>
<td>Production Ready</td>
</tr>
<tr>
<td>Layout</td>
<td><code>ui.layout_column_wrap()</code></td>
<td>Component Organization</td>
<td>Production Ready</td>
</tr>
<tr>
<td>Navigation</td>
<td><code>ui.page_navbar()</code></td>
<td>Page Structure</td>
<td>Production Ready</td>
</tr>
</tbody>
</table>

## Implementation Best Practices

When implementing components, maintain consistent patterns in your code. Use the
`@render` decorators for output functions and follow the reactive programming model
with `@reactive.effect` for side effects.

Error handling should be implemented at both the UI and server levels. For input
validation, use the `req()` function to ensure all required values are present
before processing.

## Corporate Brand Guidelines

Effective corporate brand guidelines should accomplish several key objectives:

1. **Visual Consistency**: Establish a clear color palette using our theming system.
Primary colors should be defined using `class_="btn-primary"` and similar Bootstrap
classes.

2. *Typography Standards*: Maintain consistent font usage across all text elements.
Headers should use the built-in styling provided by the `ui.card_header()` component.

3. `Component Styling`: Apply consistent styling to UI elements such as buttons,
cards, and value boxes. Use the theme parameter in components like
`ui.value_box(theme="primary")`.

4. **Layout Principles**: Follow a grid-based layout system using
`ui.layout_column_wrap()` with appropriate width parameters to ensure consistent
spacing and alignment.

5. *Responsive Design*: Implement layouts that adapt gracefully to different screen
sizes using the `fillable` parameter in page components.

Remember that brand guidelines should serve as a framework for consistency while
remaining flexible enough to accommodate future updates and modifications to the
application interface.

---
title: "Legacy theming with bslib and thematic"
output: 
  html_document:
    code_folding: show
    theme:
      bg: "#202123"
      fg: "#B8BCC2"
      primary: "#EA80FC"
      base_font:
        google: Prompt
      heading_font:
        google: Proza Libre
      version: 3
---

```{r setup, include=FALSE}
if (requireNamespace("thematic")) 
  thematic::thematic_rmd(font = "auto")
```

## R Markdown

This is an R Markdown document themed with [`{bslib}` package](https://rstudio.github.io/bslib/). `{bslib}` makes it easy to customize the main colors and fonts of a `html_document`, [`flexdashboard::flex_dashboard`](https://flexdashboard-pkg.netlify.app/articles/articles/theme.html), [shiny::fluidPage()](https://shiny.rstudio.com/reference/shiny/latest/fluidPage.html), or more generally any website that uses [Bootstrap](https://getbootstrap.com/) for styling. The `theme` parameter in the yaml front-matter of this Rmd document describes a [`bslib::bs_theme()`](https://rstudio.github.io/bslib/reference/bs_theme.html) object. This particular example uses Bootstrap 3 (`version: 3`), which is primarily for 'legacy' documents that would break with Bootstrap 4 or above.

## Themed Plots {.tabset .tabset-pills}

When running this document with [`{thematic}`](https://rstudio.github.io/thematic/) installed, the `thematic::thematic_rmd(font = "auto")` effectively translates `theme` (CSS) settings to new global theming defaults for `{ggplot2}`, `{lattice}`, and `{base}` R graphics:

### ggplot2

```{r}
library(ggplot2)

ggplot(mpg, aes(displ, hwy)) +
  geom_point() + geom_smooth()
```

### lattice

```{r}
lattice::show.settings()
```

### base

```{r}
plot(pressure, col = thematic::thematic_get_option("accent"))
```

Use `value_box()`
from [bslib](https://rstudio.github.io/bslib/reference/value_box.html)
to showcase values in your dashboards and apps!

The **Build A Box** app is designed to help you quickly choose
the right styles for your value boxes.

When you're ready to start coding,
click the
<i class="fas fa-code" role="img" aria-hidden="true"></i>
**Code** icon
to copy the code you need to get started
in your own apps!

# Getting Started with R Shiny

**R Shiny** is a powerful framework for building *interactive web applications* with R. Learn more at [shiny.posit.co](https://shiny.posit.co).

## Key Features

1. Reactive programming model
2. Built-in UI components
3. No JavaScript knowledge required
4. Easy deployment options

---

### Installation

Install Shiny from CRAN using the code below:

```r
install.packages("shiny")
library(shiny)
```

### Creating Your First App

A basic Shiny app requires:

- **UI function** to define the layout
- **Server function** to handle logic
- `shinyApp()` to combine both

> Tip: Use `shinyApp()` for simple apps or create separate `ui.R` and `server.R` files for larger projects.

For inline code, use backticks like `runApp()` to execute your application.

---

**Ready to build?** Start with `shinyAppTemplate()` and customize from there!

```{r include = FALSE}
el <- function() knitr::opts_current$get("el")
```

Like other bslib components, `r el()`s can be themed by supplying
[relevant theming variables](https://rstudio.github.io/bslib/articles/bs5-variables/index.html#`r el()`-bg)
to [bs_theme()],
which effects styling of every `r el()` on the page.
To style a _specific_ `r el()` differently from other `r el()`, utilize the `customClass` option:

---
title: "Real-time theming with bslib and thematic"
runtime: shiny
output: 
  html_document:
    code_folding: show
    theme:
      color-contrast-warnings: false
      bg: "#202123"
      fg: "#B8BCC2"
      primary: "#EA80FC"
      secondary: "#00DAC6"
      base_font:
        google: Prompt
      heading_font:
        google: Proza Libre
---

```{r setup, include=FALSE}
bslib::bs_themer()
if (requireNamespace("thematic")) 
  thematic::thematic_rmd(font = "auto")
```

## R Markdown

This is an R Markdown document themed with [`{bslib}` package](https://rstudio.github.io/bslib/). `{bslib}` makes it easy to customize the main colors and fonts of a `html_document`, [`flexdashboard::flex_dashboard`](https://flexdashboard-pkg.netlify.app/articles/articles/theme.html), [shiny::fluidPage()](https://shiny.rstudio.com/reference/shiny/latest/fluidPage.html), or more generally any website that uses [Bootstrap](https://getbootstrap.com/) for styling. The `theme` parameter in the yaml front-matter of this Rmd document describes a [`bslib::bs_theme()`](https://rstudio.github.io/bslib/reference/bs_theme.html) object, which provides access to 100s of [theming options](https://rstudio.github.io/bslib/articles/bs4-variables.html) (via its `...` argument) in addition to the main options demonstrated here (e.g., `bg`, `fg`, `primary`, etc).

This document also uses [`runtime: shiny`](https://shiny.rstudio.com/articles/interactive-docs.html) to call [`bslib::bs_themer()`](https://rstudio.github.io/bslib/reference/run_with_themer.html) which overlays a "real-time" theming widget for interactively customizing main theme settings.

## Themed Plots {.tabset .tabset-pills}

When running this document with [`{thematic}`](https://rstudio.github.io/thematic/) installed, the `thematic::thematic_rmd(font = "auto")` effectively translates `theme` (CSS) settings to new global theming defaults for `{ggplot2}`, `{lattice}`, and `{base}` R graphics:

### ggplot2

```{r, message = FALSE}
library(ggplot2)

renderPlot({
  ggplot(mpg, aes(displ, hwy)) +
    geom_point() + geom_smooth()
})
```

### lattice

```{r}
renderPlot({
  lattice::show.settings()
})
```

### base

```{r}
renderPlot({
  plot(pressure, col = thematic::thematic_get_option("accent"))
})
```

# page_navbar()

    Code
      renderTags(page_navbar(title = div(h1("foo"), h2("bar"))))$head
    Output
        <title>foo bar</title>

---

    Code
      renderTags(page_navbar(title = "foo", window_title = "bar"))$head
    Output
        <title>bar</title>

```{r setup, include = FALSE}
library(bslib)

knitr::opts_chunk$set(
  ## Screenshots are created during `devtools::document()` in CI
  ## or set to `TRUE` below to test or update locally
  # .update_screenshot = TRUE,
  "fig.width" = 5,
  "fig.height" = 2.5,
  "dpi" = 100
)
```

```{r child="0-screenshot-examples.Rmd"}
```

The appearance of a `value_box()`
can be controlled via the `theme` argument
in one of two ways:
1.  a character value describing the theme,
    such as `theme = "primary"` or `theme = "blue"`; or
3.  `theme = value_box_theme()` to create a custom theme.

We recommend using named themes for most value boxes (the first approach),
because these themes will automatically match your Bootstrap theme.

### Named themes

Bootstrap provides a list of
[theme colors](https://getbootstrap.com/docs/5.3/customize/color/#theme-colors),
with semantic names like `"primary"`, `"secondary"`, `"success"`, `"danger"`, etc.
You can set `theme` to one of these names to use the corresponding theme color as the background color of your value box.

```{r value-box-named-theme, fig.alt = "A value box using 'primary', a named Bootstrap theme color."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "primary"
)
```

Bootstrap's theme colors are drawn from
[a second color list](https://getbootstrap.com/docs/5.3/customize/color/#all-colors)
that includes variations on several main colors,
named literally.
These colors include `"blue"`, `"purple"`, `"pink"`, `"red"`, `"orange"`,
`"yellow"`, `"green"`, `"teal"`, and `"cyan"`.

```{r value-box-named-color, fig.alt = "A value box using 'purple', a named Bootstrap main color."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "teal"
)
```

### Background colors

If the theme or color name is provided without any prefix, the color will be used for the background of the value box.
You can also explicitly prefix the theme or color name with `bg-` to indicate that it should apply to the value box background.
When the theme sets the background color, either black or white is chosen automatically for the text color using Bootstrap's color contrast algorithm.

As before, you can reference semantic theme color names or literal color names.

```{r value-box-background-theme, fig.alt = "A value box using 'bg-success', a named Bootstrap theme color."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-success"
)
```

```{r value-box-background-color, fig.alt = "A value box using 'bg-orange', a named Bootstrap main color."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-purple"
)
```

### Foreground colors

To set only the foreground colors of the value box, you can prefix the theme or color name with `text-`.
This changes the text color without affecting the background color.

```{r value-box-text-theme, fig.alt = "A value box using 'text-success', a named Bootstrap theme color."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "text-success"
)
```

```{r value-box-text-color, fig.alt = "A value box using 'text-orange', a named Bootstrap main color."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "text-purple"
)
```

Occasionally you may want to adjust use both background and foreground themes on your value box.
To achieve this, set `theme` to one of the theme names and use `class` for the complementary style.
The example below uses `theme = "purple"` (which could also be `"bg-purple"`) for a purple background, and `class = "text-light"` for light-colored text.

```{r value-box-theme-class, fig.alt = "A value box using 'purple' for the background and 'text-light' for the text color."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "purple",
  class = "text-light"
)
```

### Gradient backgrounds

For a vibrant and attention-grabbing effect, bslib provides an array of gradient
background options.
Provide `theme` with a theme name in the form `bg-gradient-{from}-{to}`, where `{from}` and `{to}` are named main colors, e.g. `bg-gradient-indigo-blue`.

```{r value-box-gradient-1, fig.alt = "A value box using 'bg-gradient-blue-purple', a gradient background between two named Bootstrap colors."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-gradient-indigo-blue"
)
```

### Custom colors

Finally, for complete customization, you can use `value_box_theme()` to create a custom theme.
This function takes arguments `bg` and `fg` to set the background and foreground colors, respectively.
Like with the `bg-` theme names, if only `bg` is provided, `value_box_theme()` will choose an appropriate light or dark color for the text color.

```{r value-box-custom, fig.alt = "A value box using 'value_box_theme()' for complete customization."}
value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = value_box_theme(bg = "#e6f2fd", fg = "#0B538E"),
  class = "border"
)
```

Note that `value_box_theme()` optionally takes a `theme` name, which can be helpful if you want to use a named theme and modify the default `bg` or `fg` colors of that theme.

```{r}
value_box_theme(name = "orange", bg = "#FFFFFF")
value_box_theme(name = "text-danger", fg = "#FFB6C1")
```

Also note that `bg`/`fg` _must_ be CSS colors, not Bootstrap theme or color names.
This means that `theme = "purple"` will use your Bootstrap theme's purple color, and `bg = "purple"` will use the CSS color for _purple_, i.e. `"`r htmltools::parseCssColors("purple")`"`.

---
output:
  html_document:
    theme:
      version: 5
---

```{r setup, include = FALSE}
library(bslib)

knitr::opts_chunk$set(
  ## Screenshots are created during `devtools::document()` in CI
  ## or set to `TRUE` below to test or update locally
  # .update_screenshot = TRUE,
  "fig.width" = 5,
  "fig.height" = 2.5,
  "dpi" = 100
)
```

```{r child="0-screenshot-examples.Rmd"}
```

Use the `showcase` argument to add a plot or icon to your `value_box()`.
There are three layouts available: `"left center"`, `"top right"`, and `"bottom"`.
Set `showcase` to the name of the layout you'd like, or use the `showcase_left_center()`, `showcase_top_right()`, or `showcase_bottom()` helper functions to customize the showcase area's size.

If you're using a plot as your showcase, you may also want to set `fullscreen = TRUE` so that your users can expand the value box into a full screen card.
See the [value box article](https://rstudio.github.io/bslib/articles/value-boxes/index.html) for more details.

### Left-center showcase

The `"left center"` showcase layout is the default, and is perfect for an icon or a small plot.
This layout works best for short value boxes.

```{r value-box-showcase-left-center, fig.alt = "A value box using the left-center showcase option."}
value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = bsicons::bs_icon("ev-station-fill")
)
```

### Top-right showcase

The `"top right"` showcase layout places the icon or plot in the upper right corner of the value box.
This layout works best for medium-height to square value boxes.

```{r value-box-showcase-top-right, fig.alt = "A value box using the top-right showcase option."}
value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = bsicons::bs_icon("ev-station-fill"),
  showcase_layout = "top right"
)
```

### Bottom showcase

Finally, the `"bottom"` showcase layout is perfect for full-bleed plots.
This layout places the plot below the title and value, with the plot taking up the full width of the bottom half.

Try this layout with sparkline-style plots.
These can be a little tricky to set up, so be sure to check out the [Expandable sparklines](https://rstudio.github.io/bslib/articles/value-boxes/index.html#expandable-sparklines) section of the [value boxes article](https://rstudio.github.io/bslib/articles/value-boxes/index.html) on the bslib website.
In this example, we've created a sparkline plot using base R graphics, which isn't generally recommended.
\if{latex}{View the bslib documentation online to see the source of \verb{sparkline_plot()}.}

```{r showcase-plot, include = FALSE}
#| fig.width: 10
#| fig.show: hide

set.seed(4242)

random_sparkline_plot <- function() {
  timeseries <- cumsum(runif(100, -2, 2))
  x_axis <- seq_along(timeseries)
  x_lim <- c(1, length(timeseries))
  y_lim <- range(timeseries) + c(-2, 0)

  par(mar = c(0, 0, 0, 0))

  # Set up the plot area
  plot(
    timeseries, type = "n",
    axes = FALSE, frame.plot = FALSE,
    ylim = y_lim, xlim = x_lim,
    ylab = "",    xlab = "",
    yaxs = "i",   xaxs = "i",
  )

  # Add the sparkline line
  lines(timeseries, type = "l", pch = NA, col = "#0B538E", lwd = 3)

  # Create polygon coordinates for shading
  polygon_x <- c(1, x_axis, length(timeseries))
  polygon_y <- c(min(y_lim), timeseries, min(y_lim))

  # Add shading under the line
  polygon(polygon_x, polygon_y, col = "#e6f2fd", border = NA)
}
```

```{r sparkline_plot, include = FALSE}
sparkline_plot <- function() {
  as_fill_item(
    htmltools::plotTag(
      random_sparkline_plot(),
      width = 500,
      height = 125,
      suppressSize = "xy",
      alt = paste(
        "A sparkline plot with a randomly-generated timeseries.",
        "The timeseries starts high and ends low, with lots of variation."
      )
    )
  )
}
```

```{r show-source, echo=FALSE, comment = "", results = "asis"}
x <- htmltools::withTags(
  .noWS = c("after-begin", "before-end"),
  details(
    summary("Code for a sparkline plot with base R"),
    pre(code(paste0(
      paste(knitr::knit_code$get("showcase-plot"), collapse = "\n"),
      "\n\n",
      trimws(paste(knitr::knit_code$get("sparkline_plot"), collapse = "\n"))
    )))
  )
)
x <- gsub("\n", "<br>", x)
cat(sprintf("\\if{html}{\\out{%s}}", x))
```

```{r value-box-showcase-bottom, fig.alt = "A value box using the top-right showcase option."}
value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = sparkline_plot(),
  showcase_layout = "bottom"
)
```

```{r setup-ex-navs-tab, include = FALSE}
library(bslib)
## Screenshots are created during `devtools::document()` in CI
## or set to `TRUE` below to test or update locally
SCREENSHOT <- nzchar(Sys.getenv("CI", "")) || FALSE
```

```{r include = FALSE}
if (isTRUE(SCREENSHOT)) {
  library(chromote)
  b <- ChromoteSession$new(width = 800, height = 300, wait_ = TRUE)
}

render_screenshot <- function(x, options) {
  if (inherits(x, "bslib_fragment")) {
    x <- attr(x, "bslib_page")(x)
  }
  x <- tagList(
    x,
    tags$head(tags$style("
      body { min-height: 300px; padding: 1rem; overflow: hidden; }
      .bslib-card .card-body { min-height: 200px; }
    "))
  )
  tmpdir <- tempfile()
  dir.create(tmpdir)
  on.exit(unlink(tmpdir, recursive = TRUE))

  tmpfile <- file.path(tmpdir, basename(tempfile(fileext = ".html")))
  save_html(x, tmpfile)

  file_out <- paste0(
    gsub("_", "-", options$label), ".png"
  )

  {
    p <- b$Page$loadEventFired(wait_ = FALSE)
    b$Page$navigate(paste0("file://", tmpfile), wait_ = FALSE)
    b$wait_for(p)
  }

  b$Runtime$evaluate(
    "if ($('.dropdown-toggle').length > 0) $('.dropdown-toggle').dropdown('toggle')"
  )
  b$screenshot(filename = file.path(
    rprojroot::find_package_root_file(),
    "man/figures/", file_out
  ))

  invisible(file_out)
}

include_screenshot <- function(x, options) {
  file_name <- paste0(
    gsub("_", "-", options$label), ".png"
  )
  alt <- options$fig.alt
  if (is.null(alt)) {
    alt <- sprintf("Screenshot of a %s() example.", options$label)
  }
  img <- sprintf('![](%s "%s")', file_name, alt)
  knitr::asis_output(img)
}

knitr::opts_chunk$set(
  render = function(x, options) {
    if (isTRUE(SCREENSHOT)) {
      render_screenshot(x, options)
    }
    include_screenshot(x, options)
  }
)
```

## A basic example

This first example creates a simple tabbed navigation container with two tabs.
The tab name and the content of each tab are specified in the `nav_panel()` calls
and `navset_tab()` creates the tabbed navigation around these two tabs.

```{r navset_tab-basic, fig.alt = "Screenshot of a basic navset_tab() example."}
library(htmltools)

navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content."))
)
```

In the rest of the examples, we'll include links among the tabs (or pills) in the navigation controls.

```{r echo = TRUE, eval = TRUE, results="hide"}
link_shiny <- tags$a(shiny::icon("github"), "Shiny", href = "https://github.com/rstudio/shiny", target = "_blank")
link_posit <- tags$a(shiny::icon("r-project"), "Posit", href = "https://posit.co", target = "_blank")
```

## `navset_tab()`

You can fully customize the controls in the navigation component.
In this example, we've added a direct link to the Shiny repository using `nav_item()`.
We've also included a dropdown menu using `nav_menu()`
containing an option to select a third tab panel
and another direct link to Posit's website.
Finally, we've separated the primary tabs on the left
from the direct link and dropdown menu on the right
using `nav_spacer()`.

```{r navset_tab}
navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
```

## `navset_pill()`

`navset_pill()` creates a navigation container that behaves exactly like `navset_tab()`,
but the tab toggles are _pills_ or button-shaped.

```{r navset_pill}
navset_pill(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
```

## `navset_card_tab()`

The tabbed navigation container can also be used in a `card()` component
thanks to `navset_card_tab()`.
Learn more about this approach in the
[article about Cards](https://pkgs.rstudio.com/bslib/articles/cards.html),
including how to add [a shared sidebar](https://pkgs.rstudio.com/bslib/articles/sidebars.html#multi-page-layout)
to all tabs in the card
using the `sidebar` argument of `navset_card_tab()`.

```{r navset_tab_card}
navset_card_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
```

## `navset_card_pill()`

Similar to `navset_pill()`,
`navset_card_pill()` provides a pill-shaped variant to `navset_card_tab()`.
You can use the `placement` argument to position the navbar
`"above"` or `"below"` the card body.

```{r navset_card_pill}
navset_card_pill(
  placement = "above",
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
```

## `navset_pill_list()`

Furthermore, `navset_pill_list()` creates a vertical list of navigation controls
adjacent to, rather than on top of, the tab content panels.

```{r navset_pill_list}
navset_pill_list(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
```

## `page_navbar()`

Finally, `page_navbar()` provides full-page navigation container
similar to `navset_tab()` but where each `nav_panel()` is treated as a full page of content
and the navigation controls appear in a top-level navigation bar.

```{r page_navbar}
page_navbar(
  title = "My App",
  bg = "#0062cc",
  nav_panel(title = "One", p("First page content.")),
  nav_panel(title = "Two", p("Second page content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third page content.")),
    nav_item(link_posit)
  )
)
```

```{r setup, include = FALSE}
library(bslib)

knitr::opts_chunk$set(
  ## Screenshots are created during `devtools::document()` in CI
  ## or set to `TRUE` below to test or update locally
  # .update_screenshot = TRUE,
  "fig.width" = 8,
  "fig.height" = 3,
  "dpi" = 100
)
```

```{r child="0-screenshot-examples.Rmd"}
```

### A basic example

This first example creates a simple tabbed navigation container with two tabs.
The tab name and the content of each tab are specified in the `nav_panel()` calls
and `navset_tab()` creates the tabbed navigation around these two tabs.

```{r navset_tab-basic, fig.alt = "Screenshot of a basic navset_tab() example."}
library(htmltools)

navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content."))
)
```

In the rest of the examples, we'll include links among the tabs (or pills) in the navigation controls.

```{r echo = TRUE, eval = TRUE, results="hide"}
link_shiny <- tags$a(shiny::icon("github"), "Shiny", href = "https://github.com/rstudio/shiny", target = "_blank")
link_posit <- tags$a(shiny::icon("r-project"), "Posit", href = "https://posit.co", target = "_blank")
```

### `navset_tab()`

You can fully customize the controls in the navigation component.
In this example, we've added a direct link to the Shiny repository using `nav_item()`.
We've also included a dropdown menu using `nav_menu()`
containing an option to select a third tab panel
and another direct link to Posit's website.
Finally, we've separated the primary tabs on the left
from the direct link and dropdown menu on the right
using `nav_spacer()`.

```{r navset_tab}
navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_pill()`

`navset_pill()` creates a navigation container that behaves exactly like `navset_tab()`,
but the tab toggles are _pills_ or button-shaped.

```{r navset_pill}
navset_pill(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_underline()`

`navset_underline()` creates a navigation container that behaves exactly like `navset_tab()` and `navset_pill()`, but the active/focused navigation links are styled with an underline.

```{r navset_underline}
navset_underline(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_card_tab()`

The tabbed navigation container can also be used in a `card()` component
thanks to `navset_card_tab()`.
Learn more about this approach in the
[article about Cards](https://pkgs.rstudio.com/bslib/articles/cards.html),
including how to add [a shared sidebar](https://pkgs.rstudio.com/bslib/articles/sidebars.html#multi-page-layout)
to all tabs in the card
using the `sidebar` argument of `navset_card_tab()`.

```{r navset_tab_card}
navset_card_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_card_pill()`

Similar to `navset_pill()`,
`navset_card_pill()` provides a pill-shaped variant to `navset_card_tab()`.
You can use the `placement` argument to position the navbar
`"above"` or `"below"` the card body.

```{r navset_card_pill}
navset_card_pill(
  placement = "above",
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_card_underline()`

`navset_card_underline()` provides a card-based variant of `navset_underline()`.

```{r navset_card_underline}
navset_card_underline(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_pill_list()`

Furthermore, `navset_pill_list()` creates a vertical list of navigation controls
adjacent to, rather than on top of, the tab content panels.

```{r navset_pill_list}
navset_pill_list(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `page_navbar()`

Finally, `page_navbar()` provides full-page navigation container
similar to `navset_underline()` but where each `nav_panel()` is treated as a full page of content
and the navigation controls appear in a top-level navigation bar.
Note also that the underline styling can be removed via the `underline` argument.

```{r page_navbar, fig.width=10}
page_navbar(
  title = "My App",
  bg = "#0062cc",
  underline = TRUE,
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    align = "right",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

# bootstrap-heading() mixin approximates %heading placeholder

    Code
      heading_placeholder
    Output
      [1] "%heading {"                               
      [2] "  margin-top: 0; // 1"                    
      [3] "  margin-bottom: $headings-margin-bottom;"
      [4] "  font-family: $headings-font-family;"    
      [5] "  font-style: $headings-font-style;"      
      [6] "  font-weight: $headings-font-weight;"    
      [7] "  line-height: $headings-line-height;"    
      [8] "  color: $headings-color;"                
      [9] "}"                                        

# grid_item_container()

    Code
      grid_item_container(div(class = "layout-column-child-element"), class = "g-col-md-6 g-col-lg-4",
      fillable = TRUE)
    Output
      <div class="g-col-md-6 g-col-lg-4 bslib-gap-spacing html-fill-container">
        <div class="layout-column-child-element"></div>
      </div>

---

    Code
      grid_item_container(div(class = "layout-column-child-element"), class = "g-col-md-6 g-col-lg-4",
      fillable = FALSE)
    Output
      <div class="g-col-md-6 g-col-lg-4 bslib-gap-spacing">
        <div class="layout-column-child-element"></div>
      </div>

# breakpoints() re-orders well-known breaks and test print method

    Code
      bp
    Output
      <breakpoints>
          xs: 1
          sm: 2
          md: 3
          lg: 4
          xl: 5
         xxl: 6
       giant: 7
        huge: 8

# breakpoints() has correct classes and structure

    Code
      bp
    Output
      <breakpoints>
         sm:    1   -1    1
         md:   -1    2   -2    3
         lg:   -2    2   -1   -2    3   -2
         xl:    1    2    3   NA
       huge:    1    2    3 auto

# breakpoints() throws if NAs are mixed with other column values

    Cannot mix widths and `NA` values. All column widths must be specified, or choose auto widths using a single `NA` value.

# bs_css_grid_width_classes() warns when too many column widths

    Truncating number of widths at 'md' breakpoint to match number of elements.
    * widths: 4
    * elements: 3

```{r include = FALSE}
el <- function() knitr::opts_current$get("el")
```

Because the user needs to interact with the `trigger` element to see the `r el()`, it's best practice to use an element that is typically accessible via keyboard interactions, like a button or a link.
If you use a non-interactive element, like a `<span>` or text, bslib will automatically add the `tabindex="0"` attribute to the trigger element to make sure that users can reach the element with the keyboard.
This means that in most cases you can use any element you want as the trigger.

One place where it's important to consider the accessibility of the trigger is when using an icon without any accompanying text.
In these cases, many R packages that provide icons will create an icon element with the assumption that the icon is decorative, which will make it inaccessible to users of assistive technologies.

When using an icon as the primary trigger, ensure that the icon does not have `aria-hidden="true"` or `role="presentation"` attributes.
Icon packages typically provide a way to specify a title for the icon, as well as a way to specify that the icon is not decorative.
The title should be a short description of the purpose of the trigger, rather than a description of the icon itself.

* If you're using [bsicons::bs_icon()], provide a `title`.
* If you're using [fontawesome::fa()], set `a11y = "sem"` and provide a `title`.

For example:

    # page_navbar()

        Code
          renderTags(page_navbar(title = div(h1("foo"), h2("bar"))))$head
        Output
    ---

        Code
          renderTags(page_navbar(title = "foo", window_title = "bar"))$head
        Output
bar
    ---

        Code
          renderTags(page_sidebar(title = "foo"))$head
        Output
foo
    # page_sidebar()

        Code
          renderTags(page_sidebar(title = "foo", window_title = "bar"))$head
        Output
bar
    ---

        Code
          renderTags(page_sidebar("main", title = "Title", sidebar = sidebar(open = "always")))$
            html
        Output

# Title

main
    ---

        Code
          renderTags(page_sidebar("main", title = "Title", sidebar = "side"))$html
        Output

# Title

main

# resolve_bs_preset(): throws an error if both `name` and `bootswatch` are provided

    Code
      resolve_bs_preset(preset = "name", bootswatch = "bootswatch")
    Error <rlang_error>
      Only one of `preset` or `bootswatch` may be provided, and `name` is preferred.
      i Did you mean one of the following options?
      * `preset = "name"`
      * `preset = "bootswatch"`
      * `bootswatch = "bootswatch"`

# resolve_bs_preset(): throws an error if `name` or `bootswatch` are not scalar strings

    Code
      resolve_bs_preset(preset = c("a", "b"))
    Error <rlang_error>
      The preset theme `preset` must be a single character string.
      x Bad: `preset = c("flatly", "darkly")`
      v Good: `preset = "flatly"`

---

    Code
      resolve_bs_preset(bootswatch = c("a", "b"))
    Error <rlang_error>
      The preset theme `bootswatch` must be a single character string.
      x Bad: `bootswatch = c("flatly", "darkly")`
      v Good: `bootswatch = "flatly"`

# resolve_bs_preset(): throws an error if `name` or `bootswatch` don't match existing presets

    Code
      resolve_bs_preset(preset = "not_a_preset", version = 4)
    Error <rlang_error>
      'not_a_preset' is not a known preset theme name for Bootstrap version 4.
      i You can list available preset themes:
      * Built-in: `builtin_themes(4)`
      * Bootswatch: `bootswatch_themes(4)`.

---

    Code
      resolve_bs_preset(bootswatch = "not_a_preset", version = 4)
    Error <rlang_error>
      'not_a_preset' is not a known preset theme name for Bootstrap version 4.
      i You can list available preset themes:
      * Built-in: `builtin_themes(4)`
      * Bootswatch: `bootswatch_themes(4)`.

```{r setup, include = FALSE}
library(bslib)

knitr::opts_chunk$set(
  ## Screenshots are created during `devtools::document()` in CI
  ## or set to `TRUE` below to test or update locally
  # .update_screenshot = TRUE,
  "fig.width" = 8,
  "fig.height" = 3,
  "dpi" = 100
)
```

```{r child="0-screenshot-examples.Rmd"}
```

### A basic example

This first example creates a simple tabbed navigation container with two tabs.
The tab name and the content of each tab are specified in the `nav_panel()` calls
and `navset_tab()` creates the tabbed navigation around these two tabs.

```{r navset_tab-basic, fig.alt = "Screenshot of a basic navset_tab() example."}
library(htmltools)

navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content."))
)
```

In the rest of the examples, we'll include links among the tabs (or pills) in the navigation controls.

```{r echo = TRUE, eval = TRUE, results="hide"}
link_shiny <- tags$a(shiny::icon("github"), "Shiny", href = "https://github.com/rstudio/shiny", target = "_blank")
link_posit <- tags$a(shiny::icon("r-project"), "Posit", href = "https://posit.co", target = "_blank")
```

### `navset_tab()`

You can fully customize the controls in the navigation component.
In this example, we've added a direct link to the Shiny repository using `nav_item()`.
We've also included a dropdown menu using `nav_menu()`
containing an option to select a third tab panel
and another direct link to Posit's website.
Finally, we've separated the primary tabs on the left
from the direct link and dropdown menu on the right
using `nav_spacer()`.

```{r navset_tab}
navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_pill()`

`navset_pill()` creates a navigation container that behaves exactly like `navset_tab()`,
but the tab toggles are _pills_ or button-shaped.

```{r navset_pill}
navset_pill(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_underline()`

`navset_underline()` creates a navigation container that behaves exactly like `navset_tab()` and `navset_pill()`, but the active/focused navigation links are styled with an underline.

```{r navset_underline}
navset_underline(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_card_tab()`

The tabbed navigation container can also be used in a `card()` component
thanks to `navset_card_tab()`.
Learn more about this approach in the
[article about Cards](https://pkgs.rstudio.com/bslib/articles/cards.html),
including how to add [a shared sidebar](https://pkgs.rstudio.com/bslib/articles/sidebars.html#multi-page-layout)
to all tabs in the card
using the `sidebar` argument of `navset_card_tab()`.

```{r navset_tab_card}
navset_card_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_card_pill()`

Similar to `navset_pill()`,
`navset_card_pill()` provides a pill-shaped variant to `navset_card_tab()`.
You can use the `placement` argument to position the navbar
`"above"` or `"below"` the card body.

```{r navset_card_pill}
navset_card_pill(
  placement = "above",
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_card_underline()`

`navset_card_underline()` provides a card-based variant of `navset_underline()`.

```{r navset_card_underline}
navset_card_underline(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `navset_pill_list()`

Furthermore, `navset_pill_list()` creates a vertical list of navigation controls
adjacent to, rather than on top of, the tab content panels.

```{r navset_pill_list}
navset_pill_list(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

### `page_navbar()`

Finally, `page_navbar()` provides full-page navigation container
similar to `navset_underline()` but where each `nav_panel()` is treated as a full page of content
and the navigation controls appear in a top-level navigation bar.
Note that the navbar background and underline styling can be controlled via `navbar_options`.

`page_navbar()` is complimented by `navset_bar()` which produces a similar layout intended to be used within an app.

```{r page_navbar, fig.width=10}
page_navbar(
  title = "My App",
  navbar_options = navbar_options(
    bg = "#0062cc",
    underline = TRUE
  ),
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    align = "right",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
```

# builtin_bundle(): errors for unknown preset names

    Code
      builtin_bundle("not-a-preset", version = "5")
    Error <rlang_error>
      'not-a-preset' is not a valid built-in theme preset provided by {bslib}.
      i Available Bootstrap 5 themes are: 'shiny'

---

    Code
      builtin_bundle("not-a-preset", version = "4")
    Error <rlang_error>
      'not-a-preset' is not a valid built-in theme preset provided by {bslib}.
      i No built-in theme presets are available for this version of Bootstrap.

# as_fill_item() with a simple tag Code cat(format(as_fill_item(tag_simple()))) Output
# as_fill_item() with a simple tag with arguments Code cat(format(ts_fill)) Output
# as_fill_item() with a nested tag on outer tag Code cat(format(as_fill_item(tag_nested()))) Output
# as_fill_item() with a nested tag on inner tag Code cat(format(nested_fill)) Output
# as_fillable_container() with a simple tag Code cat(format(as_fillable_container(tag_simple()))) Output
# as_fillable_container() with a simple tag with arguments Code cat(format(ts_fillable)) Output
# as_fillable_container() with a nested tag on outer tag Code cat(format(as_fillable_container(tag_nested()))) Output
# as_fillable_container() with a nested tag on inner tag Code cat(format(nested_fillable)) Output
# as_fill_carrier() with a simple tag Code cat(format(as_fill_carrier(tag_simple()))) Output
# as_fill_carrier() with a simple tag with arguments Code cat(format(ts_carrier)) Output
# as_fill_carrier() with a nested tag on outer tag Code cat(format(as_fill_carrier(tag_nested()))) Output
# as_fill_carrier() with a nested tag on inner tag Code cat(format(nested_carrier)) Output

# bootstrap-heading() mixin approximates %heading placeholder

    Code
      heading_placeholder
    Output
      [1] "%heading {"                               
      [2] "  margin-top: 0; // 1"                    
      [3] "  margin-bottom: $headings-margin-bottom;"
      [4] "  font-family: $headings-font-family;"    
      [5] "  font-style: $headings-font-style;"      
      [6] "  font-weight: $headings-font-weight;"    
      [7] "  line-height: $headings-line-height;"    
      [8] "  color: var(--#{$prefix}heading-color);" 
      [9] "}"                                        

# page_navbar() Code renderTags(page_navbar(title = div(h1("foo"), h2("bar"))))$head Output
--- Code renderTags(page_navbar(title = "foo", window_title = "bar"))$head Output bar
--- Code renderTags(page_sidebar(title = "foo"))$head Output foo
# page_sidebar() Code renderTags(page_sidebar(title = "foo", window_title = "bar"))$head Output bar
--- Code renderTags(page_sidebar("main", title = "Title", sidebar = sidebar(open = "always"), `data-attr` = "here"))$html Output
main
--- Code renderTags(page_sidebar("main", title = "Title", sidebar = "side"))$html Output
main

# toolbar() basic attributes and defaults Code show_raw_html(toolbar("Item 1", "Item 2")) Output
Item 1 Item 2
--- Code show_raw_html(toolbar("Item 1", "Item 2", gap = "10px")) Output
Item 1 Item 2
# toolbar() aligns correctly Code show_raw_html(toolbar("Item 1", "Item 2", align = "left")) Output
Item 1 Item 2
--- Code show_raw_html(toolbar("Item 1", "Item 2", align = "right")) Output
Item 1 Item 2
    # toolbar_input_button() tests

        Code
          show_raw_html(toolbar_input_button(id = "label_only", label = "Click me",
            show_label = TRUE))
        Output
    ---

        Code
          show_raw_html(toolbar_input_button(id = "test_btn", label = "Click me", icon = shiny::icon(
            "star"), ))
        Output
    ---

        Code
          show_raw_html(toolbar_input_button(id = "test_btn", label = "Click me", icon = shiny::icon(
            "star"), show_label = TRUE))
        Output
    # toolbar_input_button() disabled parameter

        Code
          show_raw_html(toolbar_input_button(id = "disabled_btn", label = "Disabled",
            disabled = TRUE, show_label = TRUE))
        Output
    ---

        Code
          show_raw_html(toolbar_input_button(id = "enabled_btn", label = "Enabled",
            disabled = FALSE, show_label = TRUE))
        Output
    # toolbar_input_button() border parameter

        Code
          show_raw_html(toolbar_input_button(id = "no_border", label = "No Border",
            border = FALSE, show_label = TRUE))
        Output
    ---

        Code
          show_raw_html(toolbar_input_button(id = "with_border", label = "With Border",
            border = TRUE, show_label = TRUE))
        Output
    # toolbar_input_button() tooltip parameter

        Code
          show_raw_html(toolbar_input_button(id = "tooltip_default", label = "Help",
            icon = shiny::icon("question")))
        Output
    ---

        Code
          show_raw_html(toolbar_input_button(id = "tooltip_false", label = "No Tooltip",
            icon = shiny::icon("question"), tooltip = FALSE))
        Output
    ---

        Code
          show_raw_html(toolbar_input_button(id = "tooltip_custom", label = "Help", icon = shiny::icon(
            "question"), tooltip = "Click for assistance"))
        Output
    ---

        Code
          show_raw_html(toolbar_input_button(id = "both_label_tooltip", label = "Save",
            icon = shiny::icon("save"), show_label = TRUE, tooltip = "Save your work"))
        Output
    # toolbar_divider() creates divider element

        Code
          show_raw_html(toolbar_divider())
        Output
--- Code show_raw_html(toolbar_divider(gap = "20px")) Output
    ---

        Code
          show_raw_html(toolbar_divider(width = "5px", gap = "2rem"))
        Output
    # toolbar_input_select() markup snapshots

        Code
          show_raw_html(toolbar_input_select(id = "select1", label = "Basic select",
            choices = c("A", "B", "C"), tooltip = FALSE))
        Output
Basic select
    ---

        Code
          show_raw_html(toolbar_input_select(id = "select2", label = "Select with selected",
            choices = c("Option 1", "Option 2", "Option 3"), selected = "Option 2",
            tooltip = FALSE))
        Output
Select with selected
    ---

        Code
          show_raw_html(toolbar_input_select(id = "select3", label = "Select with custom class",
            choices = c("X", "Y", "Z"), class = "bg-success-subtle", style = "width: 400px",
            tooltip = FALSE))
        Output
Select with custom class
# toolbar_input_select() handles grouped choices Code show_raw_html(grouped_select) Output
Grouped select
    # toolbar_input_select() tooltip parameter

        Code
          show_raw_html(toolbar_input_select(id = "tooltip_true", label = "My Select Label",
            choices = c("A", "B"), tooltip = TRUE))
        Output
My Select Label
    ---

        Code
          show_raw_html(toolbar_input_select(id = "with_tooltip", label = "With tooltip",
            choices = c("A", "B"), tooltip = "This is helpful information"))
        Output
With tooltip
    # toolbar_input_select() icon parameter

        Code
          show_raw_html(toolbar_input_select(id = "with_icon", label = "With icon",
            choices = c("A", "B"), icon = shiny::icon("filter"), tooltip = FALSE))
        Output
With icon
    ---

        Code
          show_raw_html(toolbar_input_select(id = "icon_tooltip", label = "Icon and tooltip",
            choices = c("A", "B"), icon = shiny::icon("star"), tooltip = "Select an option"))
        Output
Icon and tooltip
    # update_toolbar_input_select() validates label parameter

        Code
          update_toolbar_input_select("test_id", label = "", session = session)
        Condition
          Error in `update_toolbar_input_select()`:
          ! `label` must be a non-empty string.

    ---

        Code
          update_toolbar_input_select("test_id", label = "   ", session = session)
        Condition
          Error in `update_toolbar_input_select()`:
          ! `label` must be a non-empty string.

    ---

        Code
          update_toolbar_input_select("test_id", label = 123, session = session)
        Condition
          Error in `update_toolbar_input_select()`:
          ! `label` must be a non-empty string.

    ---

        Code
          update_toolbar_input_select("test_id", label = c("A", "B"), session = session)
        Condition
          Error in `update_toolbar_input_select()`:
          ! `label` must be a non-empty string.

    # toolbar_input_select() validates selected is in choices

        Code
          toolbar_input_select(id = "test", label = "Test", choices = c("A", "B", "C"),
          selected = "D", tooltip = FALSE)
        Condition
          Error in `toolbar_input_select()`:
          ! `selected` value 'D' is not in `choices`.

    ---

        Code
          toolbar_input_select(id = "test", label = "Test", choices = c(`Label A` = "val_a",
            `Label B` = "val_b"), selected = "Label A", tooltip = FALSE)
        Condition
          Error in `toolbar_input_select()`:
          ! `selected` value 'Label A' is not in `choices`.

    ---

        Code
          toolbar_input_select(id = "test", label = "Test", choices = list(`Group 1` = c(
            "A", "B"), `Group 2` = c("C", "D")), selected = "E", tooltip = FALSE)
        Condition
          Error in `toolbar_input_select()`:
          ! `selected` value 'E' is not in `choices`.

    # update_toolbar_input_select() validates selected is in choices

        Code
          update_toolbar_input_select("test_id", choices = c("A", "B", "C"), selected = "D",
          session = session)
        Condition
          Warning:
          `selected` value 'D' is not in `choices`.

# resolve_bs_preset() / throws an error if both `name` and `bootswatch` are provided

    Code
      resolve_bs_preset(preset = "name", bootswatch = "bootswatch")
    Condition
      Error in `resolve_bs_preset()`:
      ! Only one of `preset` or `bootswatch` may be provided (`preset` is the preferred choice).
      i Did you mean one of the following options?
      * `preset = "name"`
      * `preset = "bootswatch"`
      * `bootswatch = "bootswatch"`

# resolve_bs_preset() / throws an error if `name` or `bootswatch` are not scalar strings

    Code
      resolve_bs_preset(preset = c("a", "b"))
    Condition
      Error in `resolve_bs_preset()`:
      ! The preset theme `preset` must be a single character string.
      x Bad: `preset = c("flatly", "darkly")`
      v Good: `preset = "flatly"`

---

    Code
      resolve_bs_preset(bootswatch = c("a", "b"))
    Condition
      Error in `resolve_bs_preset()`:
      ! The preset theme `bootswatch` must be a single character string.
      x Bad: `bootswatch = c("flatly", "darkly")`
      v Good: `bootswatch = "flatly"`

# resolve_bs_preset() / throws an error if `name` or `bootswatch` don't match existing presets

    Code
      resolve_bs_preset(preset = "not_a_preset", version = 4)
    Condition
      Error in `resolve_bs_preset()`:
      ! 'not_a_preset' is not a known preset theme name for Bootstrap version 4.
      i You can list available preset themes:
      * Built-in: `builtin_themes(4)`
      * Bootswatch: `bootswatch_themes(4)`.

---

    Code
      resolve_bs_preset(bootswatch = "not_a_preset", version = 4)
    Condition
      Error in `resolve_bs_preset()`:
      ! 'not_a_preset' is not a known preset theme name for Bootstrap version 4.
      i You can list available preset themes:
      * Built-in: `builtin_themes(4)`
      * Bootswatch: `bootswatch_themes(4)`.

# builtin_bundle() / errors for unknown preset names

    Code
      builtin_bundle("not-a-preset", version = "5")
    Condition
      Error in `validate_builtin_preset_name()`:
      ! 'not-a-preset' is not a valid built-in theme preset provided by {bslib}.
      i Available Bootstrap 5 themes are: 'shiny'

---

    Code
      builtin_bundle("not-a-preset", version = "4")
    Condition
      Error in `validate_builtin_preset_name()`:
      ! 'not-a-preset' is not a valid built-in theme preset provided by {bslib}.
      i No built-in theme presets are available for this version of Bootstrap.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card.R
\name{card_body}
\alias{card_body}
\alias{card_body_fill}
\alias{card_title}
\alias{card_header}
\alias{card_footer}
\alias{card_image}
\alias{as.card_item}
\alias{is.card_item}
\title{Card items}
\usage{
card_body(..., fill = FALSE, height = NULL, class = NULL)

card_body_fill(
  ...,
  gap = NULL,
  max_height = NULL,
  max_height_full_screen = max_height,
  min_height = NULL,
  class = NULL
)

card_title(..., container = htmltools::h5)

card_header(..., class = NULL, container = htmltools::div)

card_footer(..., class = NULL)

card_image(
  file,
  ...,
  href = NULL,
  border_radius = c("top", "bottom", "all", "none"),
  mime_type = NULL,
  class = NULL,
  height = NULL,
  width = NULL,
  container = card_body_fill
)

as.card_item(x)

is.card_item(x)
}
\arguments{
\item{...}{Unnamed arguments can be any valid child of an \link[htmltools:builder]{htmltools tag}. Named arguments become HTML attributes on returned
UI element.}

\item{fill}{whether to allow the \code{card_body()} to grow and shrink to fit its
\code{card()}.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}).}

\item{class}{Additional CSS classes for the returned UI element.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}.}

\item{max_height, max_height_full_screen, min_height}{Any valid \link[htmltools:validateCssUnit]{CSS length unit}.}

\item{container}{a function to generate an HTML element to contain the image.}

\item{file}{a file path pointing an image. The image will be base64 encoded
and provided to the \code{src} attribute of the \verb{<img>}. Alternatively, you may
set this value to \code{NULL} and provide the \code{src} yourself.}

\item{href}{an optional URL to link to.}

\item{border_radius}{where to apply \code{border-radius} on the image.}

\item{mime_type}{the mime type of the \code{file}.}

\item{width}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g., \code{width="100\%"}).}

\item{x}{an object to test (or coerce to) a card item.}
}
\value{
An \code{\link[htmltools:builder]{htmltools::div()}} tag.
}
\description{
Components designed to be provided as direct children of a \code{\link[=card]{card()}}. For a
general overview of the \code{\link[=card]{card()}} API, see \href{https://rstudio.github.io/bslib/articles/cards.html}{this article}.
}
\section{Functions}{
\itemize{
\item \code{card_body()}: A general container for the "main content" of a \code{\link[=card]{card()}}.

\item \code{card_body_fill()}: Similar to \code{card_body(fill = TRUE)}, but also marks the
return element as a "fill container" (via \code{\link[htmltools:bindFillRole]{htmltools::bindFillRole()}}) so
that its immediate children are allowed to grow and shrink to fit.

\item \code{card_title()}: Similar to \code{card_header()} but without the border and background color.

\item \code{card_header()}: A header (with border and background color) for the \code{card()}. Typically appears before a \code{card_body()}.

\item \code{card_footer()}: A header (with border and background color) for the \code{card()}. Typically appears after a \code{card_body()}.

\item \code{card_image()}: Include static (i.e., pre-generated) images.

\item \code{as.card_item()}: Mark an object as a card item. This will prevent the
\code{\link[=card]{card()}} from putting the object inside a \code{wrapper} (i.e., a
\code{card_body()}).

}}
\seealso{
\code{\link[=card]{card()}} for creating a card component.

\code{\link[=navs_tab_card]{navs_tab_card()}} for cards with multiple tabs.

\code{\link[=layout_column_wrap]{layout_column_wrap()}} for laying out multiple cards (or multiple
columns inside a card).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-current-theme.R
\name{bs_current_theme}
\alias{bs_current_theme}
\title{Obtain the currently active theme at render time}
\usage{
bs_current_theme(session = get_reactive_domain())
}
\arguments{
\item{session}{The current Shiny session (if any).}
}
\value{
a \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
Intended for advanced use by developers to obtain the currently active theme
\emph{at render time} and primarily for implementing themable widgets that can't
otherwise be themed via \code{\link[=bs_dependency_defer]{bs_dependency_defer()}} .
}
\details{
This function should generally only be called at print/render time. For
example:
\itemize{
\item Inside the \code{preRenderHook} of \code{htmlwidgets::createWidget()}.
\item Inside of a custom \link{print} method that generates \link[htmltools:builder]{htmltools::tags}.
\item Inside of a \code{\link[htmltools:tagFunction]{htmltools::tagFunction()}}
}

Calling this function at print/render time is important because it does
different things based on the context in which it's called:
\itemize{
\item If a reactive context is active, \code{session$getCurrentTheme()} is called
(which is a reactive read).
\item If no reactive context is active, \code{shiny::getCurrentTheme()} is called
(which returns the current app's \code{theme}, if relevant).
\item If \code{shiny::getCurrentTheme()} comes up empty, then \code{bs_global_get()}
is called, which is relevant for \code{\link[rmarkdown:html_document]{rmarkdown::html_document()}}, and
possibly other static rendering contexts.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sidebar.R
\name{as.tags.bslib_sidebar}
\alias{as.tags.bslib_sidebar}
\title{Render a sidebar as HTML tags}
\usage{
\method{as.tags}{bslib_sidebar}(x, ...)
}
\arguments{
\item{x}{A \code{\link[=sidebar]{sidebar()}} object.}

\item{...}{Additional arguments passed to \code{\link[htmltools:as.tags]{htmltools::as.tags()}}.}
}
\description{
Renders the sidebar element and collapse toggle elements for a \code{\link[=sidebar]{sidebar()}} in
a \code{\link[=layout_sidebar]{layout_sidebar()}} context.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accordion.R
\name{accordion}
\alias{accordion}
\alias{accordion_panel}
\title{Create a vertically collapsing accordion}
\usage{
accordion(
  ...,
  id = NULL,
  open = NULL,
  multiple = TRUE,
  class = NULL,
  width = NULL,
  height = NULL
)

accordion_panel(title, ..., value = title, icon = NULL)
}
\arguments{
\item{...}{Named arguments become attributes on the \verb{<div class="accordion">}
element. Unnamed arguments should be \code{accordion_panel()}s.}

\item{id}{If provided, you can use \code{input$id} in your server logic to
determine which of the \code{accordion_panel()}s are currently active. The value
will correspond to the \code{accordion_panel()}'s \code{value} argument.}

\item{open}{A character vector of \code{accordion_panel()} \code{value}s to open
(i.e., show) by default. The default value of \code{NULL} will open the first
\code{accordion_panel()}. Use a value of \code{TRUE} to open all (or \code{FALSE} to
open none) of the items. It's only possible to open more than one panel
when \code{multiple=TRUE}.}

\item{multiple}{Whether multiple \code{accordion_panel()} can be \code{open} at once.}

\item{class}{Additional CSS classes to include on the accordion div.}

\item{width, height}{Any valid CSS unit; for example, height="100\%".}

\item{title}{A title to appear in the \code{accordion_panel()}'s header.}

\item{value}{A character string that uniquely identifies this panel.}

\item{icon}{A \link[htmltools:builder]{htmltools::tag} child (e.g., \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}) which is positioned just before the \code{title}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

An accordion can be used to organize UI elements and content in a limited
space. It comprises multiple, vertically stacked panels that expand or
collapse when clicked, providing a compact layout that works well for
grouping input elements in a \code{\link[=sidebar]{sidebar()}} or for organizing detailed
context-specific information.
}
\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
items <- lapply(LETTERS, function(x) {
  accordion_panel(paste("Section", x), paste("Some narrative for section", x))
})

# First shown by default
accordion(!!!items)
# Nothing shown by default
accordion(!!!items, open = FALSE)
# Everything shown by default
accordion(!!!items, open = TRUE)

# Show particular sections
accordion(!!!items, open = "Section B")
accordion(!!!items, open = c("Section A", "Section B"))

# Provide an id to create a shiny input binding
library(shiny)

ui <- page_fluid(
  accordion(!!!items, id = "acc")
)

server <- function(input, output) {
  observe(print(input$acc))
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
bslib's accordion component is derived from the \href{https://getbootstrap.com/docs/5.3/components/accordion/}{Bootstrap Accordion component}.
Accordions are also featured on the bslib website:
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html#accordions}{Get Started: Dashboards}
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html#accordions}{Sidebars: Accordions}
}
}
\seealso{
\code{\link[=accordion_panel_set]{accordion_panel_set()}}, \code{\link[=accordion_panel_open]{accordion_panel_open()}} and
\code{\link[=accordion_panel_close]{accordion_panel_close()}} programmatically interact with the state of an
accordion panel.

\code{\link[=accordion_panel_insert]{accordion_panel_insert()}}, \code{\link[=accordion_panel_remove]{accordion_panel_remove()}} and
\code{\link[=accordion_panel_update]{accordion_panel_update()}} add or remove accordion panels from an
accordion.

Other Components: 
\code{\link{card}()},
\code{\link{popover}()},
\code{\link{tooltip}()},
\code{\link{value_box}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accordion.R
\name{accordion}
\alias{accordion}
\alias{accordion_panel}
\title{Create a vertically collapsing accordion}
\usage{
accordion(
  ...,
  id = NULL,
  open = NULL,
  multiple = TRUE,
  class = NULL,
  width = NULL,
  height = NULL
)

accordion_panel(title, ..., value = title, icon = NULL)
}
\arguments{
\item{...}{Named arguments become attributes on the \verb{<div class="accordion">}
element. Unnamed arguments should be \code{accordion_panel()}s.}

\item{id}{If provided, you can use \code{input$id} in your server logic to
determine which of the \code{accordion_panel()}s are currently active. The value
will correspond to the \code{accordion_panel()}'s \code{value} argument.}

\item{open}{A character vector of \code{accordion_panel()} \code{value}s to open
(i.e., show) by default. The default value of \code{NULL} will open the first
\code{accordion_panel()}. Use a value of \code{TRUE} to open all (or \code{FALSE} to
open none) of the items. It's only possible to open more than one panel
when \code{multiple=TRUE}.}

\item{multiple}{Whether multiple \code{accordion_panel()} can be \code{open} at once.}

\item{class}{Additional CSS classes to include on the accordion div.}

\item{width, height}{Any valid CSS unit; for example, height="100\%".}

\item{title}{A title to appear in the \code{accordion_panel()}'s header.}

\item{value}{A character string that uniquely identifies this panel.}

\item{icon}{A \link[htmltools:builder]{htmltools::tag} child (e.g., \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}) which is positioned just before the \code{title}.}
}
\description{
An accordion can be used to organize UI elements and content in a limited
space. It comprises multiple, vertically stacked panels that expand or
collapse when clicked, providing a compact layout that works well for
grouping input elements in a \code{\link[=sidebar]{sidebar()}} or for organizing detailed
context-specific information.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
items <- lapply(LETTERS, function(x) {
  accordion_panel(paste("Section", x), paste("Some narrative for section", x))
})

# First shown by default
accordion(!!!items)
# Nothing shown by default
accordion(!!!items, open = FALSE)
# Everything shown by default
accordion(!!!items, open = TRUE)

# Show particular sections
accordion(!!!items, open = "Section B")
accordion(!!!items, open = c("Section A", "Section B"))

# Provide an id to create a shiny input binding
library(shiny)
library(bslib)

ui <- page_fluid(
  accordion(!!!items, id = "acc")
)

server <- function(input, output) {
  observe(print(input$acc))
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
bslib's accordion component is derived from the \href{https://getbootstrap.com/docs/5.3/components/accordion/}{Bootstrap Accordion component}.
Accordions are also featured on the bslib website:
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html#accordions}{Get Started: Dashboards}
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html#accordions}{Sidebars: Accordions}
}
}
\seealso{
\code{\link[=accordion_panel_set]{accordion_panel_set()}}, \code{\link[=accordion_panel_open]{accordion_panel_open()}} and
\code{\link[=accordion_panel_close]{accordion_panel_close()}} programmatically interact with the state of an
accordion panel.

\code{\link[=accordion_panel_insert]{accordion_panel_insert()}}, \code{\link[=accordion_panel_remove]{accordion_panel_remove()}} and
\code{\link[=accordion_panel_update]{accordion_panel_update()}} add or remove accordion panels from an
accordion.

Other Components: 
\code{\link{card}()},
\code{\link{popover}()},
\code{\link{tooltip}()},
\code{\link{value_box}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accordion.R
\name{accordion_panel_set}
\alias{accordion_panel_set}
\alias{accordion_panel_open}
\alias{accordion_panel_close}
\alias{accordion_panel_insert}
\alias{accordion_panel_remove}
\alias{accordion_panel_update}
\title{Dynamically update accordions}
\usage{
accordion_panel_set(id, values, session = get_current_session())

accordion_panel_open(id, values, session = get_current_session())

accordion_panel_close(id, values, session = get_current_session())

accordion_panel_insert(
  id,
  panel,
  target = NULL,
  position = c("after", "before"),
  session = get_current_session()
)

accordion_panel_remove(id, target, session = get_current_session())

accordion_panel_update(
  id,
  target,
  ...,
  title = NULL,
  value = NULL,
  icon = NULL,
  session = get_current_session()
)
}
\arguments{
\item{id}{an character string that matches an existing \code{\link[=accordion]{accordion()}}'s \code{id}.}

\item{values}{either a character string (used to identify particular
\code{\link[=accordion_panel]{accordion_panel()}}s by their \code{value}) or \code{TRUE} (i.e., all \code{values}).}

\item{session}{a shiny session object (the default should almost always be
used).}

\item{panel}{an \code{\link[=accordion_panel]{accordion_panel()}}.}

\item{target}{The \code{value} of an existing panel to insert next to. If
removing: the \code{value} of the \code{\link[=accordion_panel]{accordion_panel()}} to remove.}

\item{position}{Should \code{panel} be added before or after the target? When
\code{target} is \code{NULL} (the default), \code{"after"} will append after the last
panel and \code{"before"} will prepend before the first panel.}

\item{...}{Named arguments become attributes on the \verb{<div class="accordion">}
element. Unnamed arguments should be \code{accordion_panel()}s.}

\item{title}{A title to appear in the \code{accordion_panel()}'s header.}

\item{value}{A character string that uniquely identifies this panel.}

\item{icon}{A \link[htmltools:builder]{htmltools::tag} child (e.g., \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}) which is positioned just before the \code{title}.}
}
\description{
Dynamically (i.e., programmatically) update/modify \code{\link[=accordion]{accordion()}}s in a
Shiny app. These functions require an \code{id} to be provided to the
\code{accordion()} and must also be called within an active Shiny session.
}
\section{Functions}{
\itemize{
\item \code{accordion_panel_set()}: same as \code{accordion_panel_open()}, except it
also closes any currently open panels.

\item \code{accordion_panel_open()}: open \code{\link[=accordion_panel]{accordion_panel()}}s.

\item \code{accordion_panel_close()}: close \code{\link[=accordion_panel]{accordion_panel()}}s.

\item \code{accordion_panel_insert()}: insert a new \code{\link[=accordion_panel]{accordion_panel()}}

\item \code{accordion_panel_remove()}: remove \code{\link[=accordion_panel]{accordion_panel()}}s.

\item \code{accordion_panel_update()}: update a \code{\link[=accordion_panel]{accordion_panel()}}.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precompiled.R
\name{precompiled_css_path}
\alias{precompiled_css_path}
\title{Get the path to a precompiled CSS file}
\usage{
precompiled_css_path(theme = bs_theme())
}
\arguments{
\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}
}
\value{
The path to the precompiled CSS file, if it exists, or \code{NULL}, if
not.
}
\description{
This function is meant for development and debugging purposes. It can be used
to test if a precompiled CSS file exists for a given theme object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-current-theme.R
\name{bs_current_theme}
\alias{bs_current_theme}
\title{Obtain the currently active theme at render time}
\usage{
bs_current_theme(session = get_current_session(FALSE))
}
\arguments{
\item{session}{The current Shiny session (if any).}
}
\value{
Returns a \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
Intended for advanced use by developers to obtain the currently active theme
\emph{at render time} and primarily for implementing themable widgets that can't
otherwise be themed via \code{\link[=bs_dependency_defer]{bs_dependency_defer()}} .
}
\details{
This function should generally only be called at print/render time. For
example:
\itemize{
\item Inside the \code{preRenderHook} of \code{htmlwidgets::createWidget()}.
\item Inside of a custom \link{print} method that generates \link[htmltools:builder]{htmltools::tags}.
\item Inside of a \code{\link[htmltools:tagFunction]{htmltools::tagFunction()}}
}

Calling this function at print/render time is important because it does
different things based on the context in which it's called:
\itemize{
\item If a reactive context is active, \code{session$getCurrentTheme()} is called
(which is a reactive read).
\item If no reactive context is active, \code{shiny::getCurrentTheme()} is called
(which returns the current app's \code{theme}, if relevant).
\item If \code{shiny::getCurrentTheme()} comes up empty, then \code{bs_global_get()}
is called, which is relevant for \code{\link[rmarkdown:html_document]{rmarkdown::html_document()}}, and
possibly other static rendering contexts.
}
}
\seealso{
Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme_preview}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-switch.R
\name{input_switch}
\alias{input_switch}
\alias{update_switch}
\alias{toggle_switch}
\title{Switch input control}
\usage{
input_switch(id, label, value = FALSE, width = NULL)

update_switch(id, label = NULL, value = NULL, session = get_current_session())

toggle_switch(id, value = NULL, session = get_current_session())
}
\arguments{
\item{id}{An input id.}

\item{label}{A label for the switch.}

\item{value}{Whether or not the switch should be checked by default.}

\item{width}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{width="200px"}).}

\item{session}{a shiny session object (the default should almost always be used).}
}
\value{
Returns a UI element for a switch input control. The server value
received for the input corresponding to \code{id} will be a logical
(\code{TRUE}/\code{FALSE}) value.
}
\description{
Create an on-off style switch control for specifying logical values.
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
library(shiny)
library(bslib)

ui <- page_fixed(
  title = "Keyboard Settings",
  h2("Keyboard Settings"),
  input_switch("auto_capitalization", "Auto-Capitalization", TRUE),
  input_switch("auto_correction", "Auto-Correction", TRUE),
  input_switch("check_spelling", "Check Spelling", TRUE),
  input_switch("smart_punctuation", "Smart Punctuation"),
  h2("Preview"),
  verbatimTextOutput("preview")
)

server <- function(input, output, session) {
  output$preview <- renderPrint({
    list(
      auto_capitalization = input$auto_capitalization,
      auto_correction = input$auto_correction,
      check_spelling = input$check_spelling,
      smart_punctuation = input$smart_punctuation
    )
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\concept{input controls}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/navs-legacy.R, R/navs.R
\name{navs_tab}
\alias{navs_tab}
\alias{navs_pill}
\alias{navs_pill_list}
\alias{navs_hidden}
\alias{navs_bar}
\alias{navs_tab_card}
\alias{navs_pill_card}
\title{Navigation containers}
\usage{
navs_tab(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navs_pill(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navs_pill_list(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  well = TRUE,
  fluid = TRUE,
  widths = c(4, 8)
)

navs_hidden(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navs_bar(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  header = NULL,
  footer = NULL,
  bg = NULL,
  inverse = "auto",
  collapsible = TRUE,
  fluid = TRUE
)

navs_tab_card(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navs_pill_card(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  placement = c("above", "below")
)
}
\arguments{
\item{...}{a collection of \code{\link[=nav]{nav()}} items.}

\item{id}{a character string used for dynamically updating the container (see \code{\link[=nav_select]{nav_select()}}).}

\item{selected}{a character string matching the \code{value} of a particular \code{\link[=nav]{nav()}} item to selected by default.}

\item{header}{UI element(s) (\link{tags}) to display \emph{above} the nav content.}

\item{footer}{UI element(s) (\link{tags}) to display \emph{below} the nav content.}

\item{well}{\code{TRUE} to place a well (gray rounded rectangle) around the
navigation list.}

\item{fluid}{\code{TRUE} to use fluid layout; \code{FALSE} to use fixed
layout.}

\item{widths}{Column widths of the navigation list and tabset content areas
respectively.}

\item{title}{The title to display in the navbar}

\item{position}{Determines whether the navbar should be displayed at the top
of the page with normal scrolling behavior (\code{"static-top"}), pinned at
the top (\code{"fixed-top"}), or pinned at the bottom
(\code{"fixed-bottom"}). Note that using \code{"fixed-top"} or
\code{"fixed-bottom"} will cause the navbar to overlay your body content,
unless you add padding, e.g.: \code{tags$style(type="text/css", "body
  {padding-top: 70px;}")}}

\item{bg}{a CSS color to use for the navbar's background color.}

\item{inverse}{Either \code{TRUE} for a light text color or \code{FALSE} for a dark
text color. If \code{"auto"} (the default), the best contrast to \code{bg} is chosen.}

\item{collapsible}{\code{TRUE} to automatically collapse the navigation
elements into a menu when the width of the browser is less than 940 pixels
(useful for viewing on smaller touchscreen device)}

\item{placement}{placement of the nav items relative to the content.}
}
\description{
Render a collection of \code{\link[=nav]{nav()}} items into a container.
}
\examples{

library(shiny)

nav_items <- function(prefix) {
  list(
    nav("a", paste(prefix, ": tab a content")),
    nav("b", paste(prefix, ": tab b content")),
    nav_item(
      tags$a(icon("github"), "Shiny", href = "https://github.com/rstudio/shiny", target = "_blank")
    ),
    nav_spacer(),
    nav_menu(
      "Other links", align = "right",
      nav("c", paste(prefix, ": tab c content")),
      nav_item(
        tags$a(icon("r-project"), "RStudio", href = "https://rstudio.com", target = "_blank")
      )
    )
  )
}

if (interactive()) {
  shinyApp(
    page_navbar(
      title = "page_navbar()",
      bg = "#0062cc",
      !!!nav_items("page_navbar()"),
      footer = div(
        style = "width:80\%; margin: 0 auto",
        h4("navs_tab()"),
        navs_tab(!!!nav_items("navs_tab()")),
        h4("navs_pill()"),
        navs_pill(!!!nav_items("navs_pill()")),
        h4("navs_tab_card()"),
        navs_tab_card(!!!nav_items("navs_tab_card()")),
        h4("navs_pill_card()"),
        navs_pill_card(!!!nav_items("navs_pill_card()")),
        h4("navs_pill_list()"),
        navs_pill_list(!!!nav_items("navs_pill_list()"))
      )
    ),
    function(...) { }
  )
}
}
\seealso{
\code{\link[=nav]{nav()}}, \code{\link[=nav_select]{nav_select()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page_fillable}
\alias{page_fillable}
\title{A screen-filling page layout}
\usage{
page_fillable(
  ...,
  padding = NULL,
  gap = NULL,
  fillable_mobile = FALSE,
  title = NULL,
  theme = bs_theme(),
  lang = NULL
)
}
\arguments{
\item{...}{UI elements for the page. Named arguments become HTML attributes.}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}.}

\item{fillable_mobile}{Whether or not the page should fill the viewport's
height on mobile devices (i.e., narrow windows).}

\item{title}{The browser window title (defaults to the host URL of the page)}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

A Bootstrap-based page layout whose contents fill the full height and width
of the browser window.
}
\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}

library(shiny)
library(ggplot2)

ui <- page_fillable(
  h1("Example", code("mtcars"), "dashboard"),
  layout_columns(
    card(
      full_screen = TRUE,
      card_header("Number of forward gears"),
      plotOutput("gear")
    ),
    card(
      full_screen = TRUE,
      card_header("Number of carburetors"),
      plotOutput("carb")
    )
  ),
  card(
    full_screen = TRUE,
    card_header("Weight vs. Quarter Mile Time"),
    layout_sidebar(
      sidebar = sidebar(
        varSelectInput("var_x", "Compare to qsec:", mtcars[-7], "wt"),
        varSelectInput("color", "Color by:", mtcars[-7], "cyl"),
        position = "right"
      ),
      plotOutput("var_vs_qsec")
    )
  )
)

server <- function(input, output) {
  for (var in c("cyl", "vs", "am", "gear", "carb")) {
    mtcars[[var]] <- as.factor(mtcars[[var]])
  }

  output$gear <- renderPlot({
    ggplot(mtcars, aes(gear)) + geom_bar()
  })

  output$carb <- renderPlot({
    ggplot(mtcars, aes(carb)) + geom_bar()
  })

  output$var_vs_qsec <- renderPlot({
    req(input$var_x, input$color)

    ggplot(mtcars) +
      aes(y = qsec, x = !!input$var_x, color = !!input$color) +
      geom_point()
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/filling/index.html}{Filling Layouts}
on the bslib website.
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html}{Getting Started with Dashboards}
on the bslib website.
}
}
\seealso{
\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} for laying out content
into rows and columns.

\code{\link[=layout_sidebar]{layout_sidebar()}} for 'floating' sidebar layouts.

\code{\link[=accordion]{accordion()}} for grouping related input controls in the \code{sidebar}.

\code{\link[=card]{card()}} for wrapping outputs in the 'main' content area.

\code{\link[=value_box]{value_box()}} for highlighting values.

Other Dashboard page layouts: 
\code{\link{page_navbar}()},
\code{\link{page_sidebar}()}
}
\concept{Dashboard page layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/navs-legacy.R, R/navs.R
\name{navs_tab}
\alias{navs_tab}
\alias{navs_pill}
\alias{navs_pill_list}
\alias{navs_hidden}
\alias{navs_bar}
\alias{navs_tab_card}
\alias{navs_pill_card}
\title{Navigation containers}
\usage{
navs_tab(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navs_pill(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navs_pill_list(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  well = TRUE,
  fluid = TRUE,
  widths = c(4, 8)
)

navs_hidden(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navs_bar(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  header = NULL,
  footer = NULL,
  bg = NULL,
  inverse = "auto",
  collapsible = TRUE,
  fluid = TRUE
)

navs_tab_card(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  full_screen = FALSE,
  wrapper = card_body
)

navs_pill_card(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  placement = c("above", "below"),
  full_screen = FALSE,
  wrapper = card_body
)
}
\arguments{
\item{...}{a collection of \code{\link[=nav]{nav()}} items.}

\item{id}{a character string used for dynamically updating the container (see \code{\link[=nav_select]{nav_select()}}).}

\item{selected}{a character string matching the \code{value} of a particular \code{\link[=nav]{nav()}} item to selected by default.}

\item{header}{UI element(s) (\link{tags}) to display \emph{above} the nav content.}

\item{footer}{UI element(s) (\link{tags}) to display \emph{below} the nav content.}

\item{well}{\code{TRUE} to place a well (gray rounded rectangle) around the
navigation list.}

\item{fluid}{\code{TRUE} to use fluid layout; \code{FALSE} to use fixed
layout.}

\item{widths}{Column widths of the navigation list and tabset content areas
respectively.}

\item{title}{A (left-aligned) title to place in the card header/footer. If
provided, other nav items are automatically right aligned.}

\item{position}{Determines whether the navbar should be displayed at the top
of the page with normal scrolling behavior (\code{"static-top"}), pinned at
the top (\code{"fixed-top"}), or pinned at the bottom
(\code{"fixed-bottom"}). Note that using \code{"fixed-top"} or
\code{"fixed-bottom"} will cause the navbar to overlay your body content,
unless you add padding, e.g.: \code{tags$style(type="text/css", "body
  {padding-top: 70px;}")}}

\item{bg}{a CSS color to use for the navbar's background color.}

\item{inverse}{Either \code{TRUE} for a light text color or \code{FALSE} for a dark
text color. If \code{"auto"} (the default), the best contrast to \code{bg} is chosen.}

\item{collapsible}{\code{TRUE} to automatically collapse the navigation
elements into a menu when the width of the browser is less than 940 pixels
(useful for viewing on smaller touchscreen device)}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}).}

\item{full_screen}{If \code{TRUE}, an icon will appear when hovering over the card
body. Clicking the icon expands the card to fit viewport size. Consider
pairing this feature with \code{\link[=card_body_fill]{card_body_fill()}} to get output that responds to
changes in the size of the card.}

\item{wrapper}{A function (which returns a UI element) to call on unnamed
arguments in \code{...} which are not already card item(s) (like
\code{\link[=card_header]{card_header()}}, \code{\link[=card_body]{card_body()}}, etc.). Note that non-card items are grouped
together into one \code{wrapper} call (e.g. given \code{card("a", "b", card_body("c"), "d")}, \code{wrapper} would be called twice, once with \code{"a"} and
\code{"b"} and once with \code{"d"}). Consider setting \code{wrapper} to \link{card_body_fill}
if the entire card wants responsive sizing or \code{NULL} to avoid wrapping
altogether}

\item{placement}{placement of the nav items relative to the content.}
}
\description{
Render a collection of \code{\link[=nav]{nav()}} items into a container.
}
\examples{

library(shiny)

nav_items <- function(prefix) {
  list(
    nav("a", paste(prefix, ": tab a content")),
    nav("b", paste(prefix, ": tab b content")),
    nav_item(
      tags$a(icon("github"), "Shiny", href = "https://github.com/rstudio/shiny", target = "_blank")
    ),
    nav_spacer(),
    nav_menu(
      "Other links", align = "right",
      nav("c", paste(prefix, ": tab c content")),
      nav_item(
        tags$a(icon("r-project"), "RStudio", href = "https://rstudio.com", target = "_blank")
      )
    )
  )
}

if (interactive()) {
  shinyApp(
    page_navbar(
      title = "page_navbar()",
      bg = "#0062cc",
      !!!nav_items("page_navbar()"),
      footer = div(
        style = "width:80\%; margin: 0 auto",
        h4("navs_tab()"),
        navs_tab(!!!nav_items("navs_tab()")),
        h4("navs_pill()"),
        navs_pill(!!!nav_items("navs_pill()")),
        h4("navs_tab_card()"),
        navs_tab_card(!!!nav_items("navs_tab_card()")),
        h4("navs_pill_card()"),
        navs_pill_card(!!!nav_items("navs_pill_card()")),
        h4("navs_pill_list()"),
        navs_pill_list(!!!nav_items("navs_pill_list()"))
      )
    ),
    function(...) { }
  )
}
}
\seealso{
\code{\link[=nav]{nav()}}, \code{\link[=nav_select]{nav_select()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page_fillable}
\alias{page_fillable}
\title{A screen-filling page layout}
\usage{
page_fillable(
  ...,
  padding = NULL,
  gap = NULL,
  fillable_mobile = FALSE,
  title = NULL,
  theme = bs_theme(),
  lang = NULL
)
}
\arguments{
\item{...}{UI elements for the page. Named arguments become HTML attributes.}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}.}

\item{fillable_mobile}{Whether or not the page should fill the viewport's
height on mobile devices (i.e., narrow windows).}

\item{title}{The browser window title (defaults to the host URL of the page)}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}
}
\description{
A Bootstrap-based page layout whose contents fill the full height and width
of the browser window.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}

library(shiny)
library(ggplot2)
library(bslib)

ui <- page_fillable(
  h1("Example", code("mtcars"), "dashboard"),
  layout_columns(
    card(
      full_screen = TRUE,
      card_header("Number of forward gears"),
      plotOutput("gear")
    ),
    card(
      full_screen = TRUE,
      card_header("Number of carburetors"),
      plotOutput("carb")
    )
  ),
  card(
    full_screen = TRUE,
    card_header("Weight vs. Quarter Mile Time"),
    layout_sidebar(
      sidebar = sidebar(
        varSelectInput("var_x", "Compare to qsec:", mtcars[-7], "wt"),
        varSelectInput("color", "Color by:", mtcars[-7], "cyl"),
        position = "right"
      ),
      plotOutput("var_vs_qsec")
    )
  )
)

server <- function(input, output) {
  for (var in c("cyl", "vs", "am", "gear", "carb")) {
    mtcars[[var]] <- as.factor(mtcars[[var]])
  }

  output$gear <- renderPlot({
    ggplot(mtcars, aes(gear)) + geom_bar()
  })

  output$carb <- renderPlot({
    ggplot(mtcars, aes(carb)) + geom_bar()
  })

  output$var_vs_qsec <- renderPlot({
    req(input$var_x, input$color)

    ggplot(mtcars) +
      aes(y = qsec, x = !!input$var_x, color = !!input$color) +
      geom_point()
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/filling/index.html}{Filling Layouts}
on the bslib website.
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html}{Getting Started with Dashboards}
on the bslib website.
}
}
\seealso{
\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} for laying out content
into rows and columns.

\code{\link[=layout_sidebar]{layout_sidebar()}} for 'floating' sidebar layouts.

\code{\link[=accordion]{accordion()}} for grouping related input controls in the \code{sidebar}.

\code{\link[=card]{card()}} for wrapping outputs in the 'main' content area.

\code{\link[=value_box]{value_box()}} for highlighting values.

Other Dashboard page layouts: 
\code{\link{page_navbar}()},
\code{\link{page_sidebar}()}
}
\concept{Dashboard page layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-remove.R
\name{bs_remove}
\alias{bs_remove}
\alias{bs_retrieve}
\title{Remove or retrieve Sass code from a theme}
\usage{
bs_remove(theme, ids = character(0))

bs_retrieve(theme, ids = character(0), include_unnamed = TRUE)
}
\arguments{
\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}

\item{ids}{a character vector of ids}

\item{include_unnamed}{whether or not to include unnamed \code{\link[sass:sass_layer]{sass::sass_layer()}}s
(e.g., Bootstrap Sass variables, functions, and mixins).}
}
\value{
a modified \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
Remove or retrieve Sass code from a theme
}
\examples{

# Remove CSS rules for print and carousels
bs4 <- bs_theme(version = 4)
bs_remove(bs4, c("_print", "_carousel"))

# Remove BS3 compatibility layer
bs_remove(bs4, "bs3compat")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accordion.R
\name{accordion_panel_set}
\alias{accordion_panel_set}
\alias{accordion_panel_open}
\alias{accordion_panel_close}
\alias{accordion_panel_insert}
\alias{accordion_panel_remove}
\alias{accordion_panel_update}
\title{Dynamically update accordions}
\usage{
accordion_panel_set(id, values, session = get_current_session())

accordion_panel_open(id, values, session = get_current_session())

accordion_panel_close(id, values, session = get_current_session())

accordion_panel_insert(
  id,
  panel,
  target = NULL,
  position = c("after", "before"),
  session = get_current_session()
)

accordion_panel_remove(id, target, session = get_current_session())

accordion_panel_update(
  id,
  target,
  ...,
  title = NULL,
  value = NULL,
  icon = NULL,
  session = get_current_session()
)
}
\arguments{
\item{id}{an character string that matches an existing \code{\link[=accordion]{accordion()}}'s \code{id}.}

\item{values}{either a character string (used to identify particular
\code{\link[=accordion_panel]{accordion_panel()}}s by their \code{value}) or \code{TRUE} (i.e., all \code{values}).}

\item{session}{a shiny session object (the default should almost always be
used).}

\item{panel}{an \code{\link[=accordion_panel]{accordion_panel()}}.}

\item{target}{The \code{value} of an existing panel to insert next to. If
removing: the \code{value} of the \code{\link[=accordion_panel]{accordion_panel()}} to remove.}

\item{position}{Should \code{panel} be added before or after the target? When
\code{target} is \code{NULL} (the default), \code{"after"} will append after the last
panel and \code{"before"} will prepend before the first panel.}

\item{...}{Elements that become the new content of the panel.}

\item{title}{A title to appear in the \code{accordion_panel()}'s header.}

\item{value}{A character string that uniquely identifies this panel.}

\item{icon}{A \link[htmltools:builder]{htmltools::tag} child (e.g., \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}) which is positioned just before the \code{title}.}
}
\description{
Dynamically update/modify \code{\link[=accordion]{accordion()}}s in a Shiny app. To be updated
programmatically, the \code{\link[=accordion]{accordion()}} must have an \code{id}. These functions
require an active Shiny session and only work with a running Shiny app.
}
\section{Functions}{
\itemize{
\item \code{accordion_panel_set()}: same as \code{accordion_panel_open()}, except it
also closes any currently open panels.

\item \code{accordion_panel_open()}: open \code{\link[=accordion_panel]{accordion_panel()}}s.

\item \code{accordion_panel_close()}: close \code{\link[=accordion_panel]{accordion_panel()}}s.

\item \code{accordion_panel_insert()}: insert a new \code{\link[=accordion_panel]{accordion_panel()}}

\item \code{accordion_panel_remove()}: remove \code{\link[=accordion_panel]{accordion_panel()}}s.

\item \code{accordion_panel_update()}: update a \code{\link[=accordion_panel]{accordion_panel()}}.

}}
\seealso{
\code{\link[=accordion]{accordion()}} and \code{\link[=accordion_panel]{accordion_panel()}} create the accordion
component.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-dependencies.R
\name{bs_dependency}
\alias{bs_dependency}
\alias{bs_dependency_defer}
\title{Themeable HTML components}
\usage{
bs_dependency(
  input = list(),
  theme,
  name,
  version,
  cache_key_extra = NULL,
  .dep_args = list(),
  .sass_args = list()
)

bs_dependency_defer(func, memoise = TRUE)
}
\arguments{
\item{input}{Sass rules to compile, using \code{theme}.}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{name}{Library name}

\item{version}{Library version}

\item{cache_key_extra}{Extra information to add to the sass cache key. It is
useful to add the version of your package.}

\item{.dep_args}{A list of additional arguments to pass to
\code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}}. Note that \code{package} has no effect and
\code{script} must be absolute path(s).}

\item{.sass_args}{A list of additional arguments to pass to
\code{\link[sass:sass_partial]{sass::sass_partial()}}.}

\item{func}{a \emph{non-anonymous} function, with a \emph{single} argument.
This function should accept a \code{\link[=bs_theme]{bs_theme()}} object and return a single
\code{\link[=htmlDependency]{htmlDependency()}}, a list of them, or \code{NULL}.}

\item{memoise}{whether or not to memoise (i.e., cache) \code{func} results for a
short period of time. The default, \code{TRUE}, can have large performance
benefits when many instances of the same themable widget are rendered. Note
that you may want to avoid memoisation if \code{func} relies on side-effects
(e.g., files on-disk) that need to change for each themable widget
instance.}
}
\value{
\code{bs_dependency()} returns an \code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}} and
\code{bs_dependency_defer()} returns an \code{\link[htmltools:tagFunction]{htmltools::tagFunction()}}
}
\description{
Themeable HTML components use Sass to generate CSS rules from Bootstrap Sass
variables, functions, and/or mixins (i.e., stuff inside of \code{theme}).
\code{bs_dependencies()} makes it a bit easier to create themeable components by
compiling \code{\link[sass:sass]{sass::sass()}} (\code{input}) together with Bootstrap Sass inside of a
\code{theme}, and packaging up the result into an \code{\link[=htmlDependency]{htmlDependency()}}.

Themable components can also be  \emph{dynamically} themed inside of Shiny (i.e.,
they may be themed in 'real-time' via \code{\link[=bs_themer]{bs_themer()}}, and more generally,
update their styles in response to \link[shiny:session]{shiny::session}'s \code{setCurrentTheme()}
method). Dynamically themeable components provide a "recipe" (i.e., a
function) to \code{bs_dependency_defer()}, describing how to generate new CSS
stylesheet(s) from a new \code{theme}. This function is called when the HTML page
is first rendered, and may be invoked again with a new \code{theme} whenever
\link[shiny:session]{shiny::session}'s \code{setCurrentTheme()} is called.
}
\examples{

\dontrun{

myWidgetVersion <- "1.2.3"

myWidgetDependency <- function() {
  list(
    bs_dependency_defer(myWidgetCss),
    htmlDependency(
      name = "mywidget-js",
      version = myWidgetVersion,
      src = system.file(package = "mypackage", "js"),
      script = "mywidget.js"
    )
  )
}

myWidgetCSS <- function(theme) {
  if (!is_bs_theme(theme)) {
    return(
      htmlDependency(
        name = "mywidget-css",
        version = myWidgetVersion,
        src = system.file(package = "mypackage", "css"),
        stylesheet = "mywidget.css"
      )
    )
  }

  # Compile mywidget.scss using the variables and defaults from the theme
  # object.
  sass_input <- sass::sass_file(system.file(package = "mypackage", "scss/mywidget.scss"))

  bs_dependency(
    input = sass_input,
    theme = theme,
    name = "mywidget",
    version = myWidgetVersion,
    cache_key_extra = utils::packageVersion("mypackage")
  )
}

# Note that myWidgetDependency is not defined inside of myWidget. This is so
# that, if `myWidget()` is called multiple times, Shiny can tell that the
# function objects are identical and deduplicate them.
myWidget <- function(id) {
  div(
    id = id,
    span("myWidget"),
    myWidgetDependency()
  )
}
}

}
\references{
\url{https://rstudio.github.io/bslib/articles/custom-components.html}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-switch.R
\name{input_switch}
\alias{input_switch}
\alias{update_switch}
\alias{toggle_switch}
\title{Switch input control}
\usage{
input_switch(id, label, value = FALSE, width = NULL)

update_switch(id, label = NULL, value = NULL, session = get_current_session())

toggle_switch(id, value = NULL, session = get_current_session())
}
\arguments{
\item{id}{An input id.}

\item{label}{A label for the switch.}

\item{value}{Whether or not the switch should be checked by default.}

\item{width}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{width="200px"}).}

\item{session}{a shiny session object (the default should almost always be
used).}
}
\value{
Returns a UI element for a switch input control. The server value
received for the input corresponding to \code{id} will be a logical
(\code{TRUE}/\code{FALSE}) value.
}
\description{
Create an on-off style switch control for specifying logical values.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)

ui <- page_fixed(
  title = "Keyboard Settings",
  h2("Keyboard Settings"),
  input_switch("auto_capitalization", "Auto-Capitalization", TRUE),
  input_switch("auto_correction", "Auto-Correction", TRUE),
  input_switch("check_spelling", "Check Spelling", TRUE),
  input_switch("smart_punctuation", "Smart Punctuation"),
  h2("Preview"),
  verbatimTextOutput("preview")
)

server <- function(input, output, session) {
  output$preview <- renderPrint({
    list(
      auto_capitalization = input$auto_capitalization,
      auto_correction = input$auto_correction,
      check_spelling = input$check_spelling,
      smart_punctuation = input$smart_punctuation
    )
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\seealso{
Other input controls: 
\code{\link{input_code_editor}()},
\code{\link{input_dark_mode}()}
}
\concept{input controls}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page_navbar}
\alias{page_navbar}
\title{Multi-page app with a top navigation bar}
\usage{
page_navbar(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  fillable_mobile = FALSE,
  gap = NULL,
  padding = NULL,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  header = NULL,
  footer = NULL,
  bg = NULL,
  inverse = "auto",
  underline = TRUE,
  collapsible = TRUE,
  fluid = TRUE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL
)
}
\arguments{
\item{...}{a collection of \code{\link[=nav_panel]{nav_panel()}} items.}

\item{title}{A (left-aligned) title to place in the card header/footer. If
provided, other nav items are automatically right aligned.}

\item{id}{a character string used for dynamically updating the container (see
\code{\link[=nav_select]{nav_select()}}).}

\item{selected}{a character string matching the \code{value} of a particular
\code{\link[=nav_panel]{nav_panel()}} item to selected by default.}

\item{sidebar}{A \code{\link[=sidebar]{sidebar()}} component to display on every \code{\link[=nav_panel]{nav_panel()}}
page.}

\item{fillable}{Whether or not to allow \code{fill} items to grow/shrink to fit
the browser window. If \code{TRUE}, all \code{\link[=nav_panel]{nav_panel()}} pages are \code{fillable}. A
character vector, matching the \code{value} of \code{\link[=nav_panel]{nav_panel()}}s to be filled, may
also be provided. Note that, if a \code{sidebar} is provided, \code{fillable} makes
the main content portion fillable.}

\item{fillable_mobile}{Whether or not \code{fillable} pages should fill the viewport's
height on mobile devices (i.e., narrow windows).}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}.}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{position}{Determines whether the navbar should be displayed at the top
of the page with normal scrolling behavior (\code{"static-top"}), pinned at
the top (\code{"fixed-top"}), or pinned at the bottom
(\code{"fixed-bottom"}). Note that using \code{"fixed-top"} or
\code{"fixed-bottom"} will cause the navbar to overlay your body content,
unless you add padding, e.g.: \code{tags$style(type="text/css", "body
  {padding-top: 70px;}")}}

\item{header}{UI element(s) (\link{tags}) to display \emph{above} the nav content.}

\item{footer}{UI element(s) (\link{tags}) to display \emph{below} the nav content.}

\item{bg}{a CSS color to use for the navbar's background color.}

\item{inverse}{Either \code{TRUE} for a light text color or \code{FALSE} for a dark
text color. If \code{"auto"} (the default), the best contrast to \code{bg} is chosen.}

\item{underline}{Whether or not to add underline styling to page links when
active or focused.}

\item{collapsible}{\code{TRUE} to automatically collapse the navigation
elements into an expandable menu on mobile devices or narrow window widths.}

\item{fluid}{\code{TRUE} to use fluid layout; \code{FALSE} to use fixed
layout.}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{window_title}{the browser window title. The default value, \code{NA}, means
to use any character strings that appear in \code{title} (if none are found, the
host URL of the page is displayed by default).}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}
}
\description{
Create a page that contains a top level navigation bar that can be used to
toggle a set of \code{\link[=nav_panel]{nav_panel()}} elements. Use this page layout to create the
effect of a multi-page app, where your app's content is broken up into
multiple "pages" that can be navigated to via the top navigation bar.
}
\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
library(shiny)
library(bslib)

link_shiny <- tags$a(
  shiny::icon("github"), "Shiny",
  href = "https://github.com/rstudio/shiny",
  target = "_blank"
)
link_posit <- tags$a(
  shiny::icon("r-project"), "Posit",
  href = "https://posit.co",
  target = "_blank"
)

ui <- page_navbar(
  title = "My App",
  nav_panel(title = "One", p("First page content.")),
  nav_panel(title = "Two", p("Second page content.")),
  nav_panel("Three", p("Third page content.")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    align = "right",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)

server <- function(...) { } # not used in this example

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
\href{https://rstudio.github.io/bslib/articles/dashboards/index.html}{Getting Started with Dashboards}
on the bslib website.
}
\seealso{
\code{\link[=nav_panel]{nav_panel()}}, \code{\link[=nav_menu]{nav_menu()}}, and \code{\link[=nav_item]{nav_item()}} for adding content
sections and organizing or creating items in the navigation bar.

\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} for laying out content
into rows and columns.

\code{\link[=card]{card()}} for wrapping outputs in the 'main' content area.

\code{\link[=value_box]{value_box()}} for highlighting values.

\code{\link[=accordion]{accordion()}} for grouping related input controls in the \code{sidebar}.

Other Dashboard page layouts: 
\code{\link{page_fillable}()},
\code{\link{page_sidebar}()}
}
\concept{Dashboard page layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/popover.R
\name{popover}
\alias{popover}
\alias{toggle_popover}
\alias{update_popover}
\title{Add a popover to a UI element}
\usage{
popover(
  trigger,
  ...,
  title = NULL,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
)

toggle_popover(id, show = NULL, session = get_current_session())

update_popover(id, ..., title = NULL, session = get_current_session())
}
\arguments{
\item{trigger}{The UI element to serve as the popover trigger (typically a
\code{\link[shiny:actionButton]{shiny::actionButton()}} or similar). If \code{trigger} renders as multiple HTML
elements (e.g., it's a \code{tagList()}), the last HTML element is used for the
trigger. If the \code{trigger} should contain all of those elements, wrap the
object in a \code{\link[=div]{div()}} or \code{\link[=span]{span()}}.}

\item{...}{UI elements for the popover's body. Character strings are
\link[=htmlEscape]{automatically escaped} unless marked as \code{\link[=HTML]{HTML()}}.}

\item{title}{A title (header) for the popover.}

\item{id}{A character string. Required to re-actively respond to the
visibility of the popover (via the \code{input[[id]]} value) and/or update the
visibility/contents of the popover.}

\item{placement}{The placement of the popover relative to its trigger.}

\item{options}{A list of additional
\href{https://getbootstrap.com/docs/5.3/components/popovers/#options}{options}.}

\item{show}{Whether to show (\code{TRUE}) or hide (\code{FALSE}) the popover. The
default (\code{NULL}) will show if currently hidden and hide if currently shown.
Note that a popover will not be shown if the trigger is not visible (e.g.,
it's hidden behind a tab).}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\description{
Display additional information when clicking on a UI element (typically a
button).
}
\section{Functions}{
\itemize{
\item \code{popover()}: Add a popover to a UI element

\item \code{toggle_popover()}: Programmatically show/hide a popover.

\item \code{update_popover()}: Update the contents of a popover.

}}
\section{Closing popovers}{


In addition to clicking the \code{close_button}, popovers can be closed by
pressing the Esc/Space key when the popover (and/or its trigger) is
focused.
}

\section{Theming/Styling}{


Like other bslib components, popovers can be themed by supplying \href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html#popover-bg}{relevant theming variables}
to \code{\link[=bs_theme]{bs_theme()}}, which effects styling of every popover on the page. To
style a \emph{specific} popover differently from other popovers, utilize the
\code{customClass} option:

\if{html}{\out{<div class="sourceCode">}}\preformatted{popover(
  "Trigger", "Popover message",
  options = list(customClass = "my-pop")
)
}\if{html}{\out{</div>}}

And then add relevant rules to \code{\link[=bs_theme]{bs_theme()}} via \code{\link[=bs_add_rules]{bs_add_rules()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{bs_theme() |> bs_add_rules(".my-pop \{ max-width: none; \}")
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}

popover(
  shiny::actionButton("btn", "A button"),
  "Popover body content...",
  title = "Popover title"
)

library(shiny)

ui <- page_fixed(
  card(class = "mt-5",
    card_header(
      popover(
        uiOutput("card_title", inline = TRUE),
        title = "Provide a new title",
        textInput("card_title", NULL, "An editable title")
      )
    ),
    "The card body..."
  )
)

server <- function(input, output) {
  output$card_title <- renderUI({
    list(input$card_title, bsicons::bs_icon("pencil-square"))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
\url{https://getbootstrap.com/docs/5.3/components/popovers/}
}
\seealso{
\code{\link[=tooltip]{tooltip()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card.R
\name{card_body}
\alias{card_body}
\alias{card_title}
\alias{card_header}
\alias{card_footer}
\alias{card_image}
\alias{as.card_item}
\alias{is.card_item}
\title{Card items}
\usage{
card_body(
  ...,
  fillable = TRUE,
  min_height = NULL,
  max_height = NULL,
  max_height_full_screen = max_height,
  height = NULL,
  padding = NULL,
  gap = NULL,
  fill = TRUE,
  class = NULL
)

card_title(..., container = htmltools::h5)

card_header(..., class = NULL, container = htmltools::div)

card_footer(..., class = NULL)

card_image(
  file,
  ...,
  href = NULL,
  border_radius = c("top", "bottom", "all", "none"),
  mime_type = NULL,
  class = NULL,
  height = NULL,
  fill = TRUE,
  width = NULL,
  container = card_body
)

as.card_item(x)

is.card_item(x)
}
\arguments{
\item{...}{Unnamed arguments can be any valid child of an \link[htmltools:builder]{htmltools tag}. Named arguments become HTML attributes on returned
UI element.}

\item{fillable}{Whether or not the card item should be a fillable (i.e.
flexbox) container.}

\item{min_height, max_height, max_height_full_screen}{Any valid \link[htmltools:validateCssUnit]{CSS length unit}.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}. This argument is only applicable when \code{fillable = TRUE}}

\item{fill}{Whether to allow this element to grow/shrink to fit its \code{card()}
container.}

\item{class}{Additional CSS classes for the returned UI element.}

\item{container}{a function to generate an HTML element to contain the image.}

\item{file}{a file path pointing an image. The image will be base64 encoded
and provided to the \code{src} attribute of the \verb{<img>}. Alternatively, you may
set this value to \code{NULL} and provide the \code{src} yourself.}

\item{href}{an optional URL to link to.}

\item{border_radius}{where to apply \code{border-radius} on the image.}

\item{mime_type}{the mime type of the \code{file}.}

\item{width}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g., \code{width="100\%"}).}

\item{x}{an object to test (or coerce to) a card item.}
}
\value{
An \code{\link[htmltools:builder]{htmltools::div()}} tag.
}
\description{
Components designed to be provided as direct children of a \code{\link[=card]{card()}}. For a
general overview of the \code{\link[=card]{card()}} API, see \href{https://rstudio.github.io/bslib/articles/cards/index.html}{the Cards article} or the
other articles listed in the \emph{References} section of the \code{\link[=card]{card()}}
documentation.
}
\section{Functions}{
\itemize{
\item \code{card_body()}: A general container for the "main content" of a \code{\link[=card]{card()}}.

\item \code{card_title()}: Similar to \code{card_header()} but without the border and background color.

\item \code{card_header()}: A header (with border and background color) for the \code{card()}. Typically appears before a \code{card_body()}.

\item \code{card_footer()}: A header (with border and background color) for the \code{card()}. Typically appears after a \code{card_body()}.

\item \code{card_image()}: Include static (i.e., pre-generated) images.

\item \code{as.card_item()}: Mark an object as a card item. This will prevent the
\code{\link[=card]{card()}} from putting the object inside a \code{wrapper} (i.e., a
\code{card_body()}).

}}
\seealso{
\code{\link[=card]{card()}} creates a card component.

\code{\link[=navset_card_tab]{navset_card_tab()}}, \code{\link[=navset_card_pill]{navset_card_pill()}} and
\code{\link[=navset_card_underline]{navset_card_underline()}} create cards with tabbed navigation.

\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} help position multiple
cards into columns and rows and can also be used inside a card.

\code{\link[=layout_sidebar]{layout_sidebar()}} adds a sidebar to a card when nested in \code{\link[=card]{card()}}
or \code{\link[=card_body]{card_body()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootswatch.R
\name{bootswatch_themes}
\alias{bootswatch_themes}
\title{Obtain a list of all available bootswatch themes.}
\usage{
bootswatch_themes(version = version_default(), full_path = FALSE)
}
\arguments{
\item{version}{the major version of Bootswatch.}

\item{full_path}{whether to return a path to the installed theme.}
}
\value{
a character vector of Bootswatch themes.
}
\description{
Obtain a list of all available bootswatch themes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tooltip.R
\name{tooltip}
\alias{tooltip}
\alias{toggle_tooltip}
\alias{update_tooltip}
\title{Add a tooltip to a UI element}
\usage{
tooltip(
  trigger,
  ...,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
)

toggle_tooltip(id, show = NULL, session = get_current_session())

update_tooltip(id, ..., session = get_current_session())
}
\arguments{
\item{trigger}{A UI element (i.e., \link[htmltools:builder]{htmltools tag}) to serve
as the tooltip trigger. If \code{trigger} renders as multiple HTML
elements (e.g., it's a \code{tagList()}), the last HTML element is used for the
trigger. If the \code{trigger} should contain all of those elements, wrap the
object in a \code{\link[=div]{div()}} or \code{\link[=span]{span()}}.}

\item{...}{UI elements for the tooltip. Character strings are \link[=htmlEscape]{automatically escaped} unless marked as \code{\link[=HTML]{HTML()}}.}

\item{id}{a character string that matches an existing tooltip id.}

\item{placement}{The placement of the tooltip relative to its trigger.}

\item{options}{A list of additional \href{https://getbootstrap.com/docs/5.3/components/tooltips/#options}{options}.}

\item{show}{Whether to show (\code{TRUE}) or hide (\code{FALSE}) the tooltip. The
default (\code{NULL}) will show if currently hidden and hide if currently shown.
Note that a tooltip will not be shown if the trigger is not visible (e.g.,
it's hidden behind a tab).}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\description{
Display additional information when focusing (or hovering over) a UI element.
}
\section{Functions}{
\itemize{
\item \code{tooltip()}: Add a tooltip to a UI element

\item \code{toggle_tooltip()}: Programmatically show/hide a tooltip.

\item \code{update_tooltip()}: Update the contents of a tooltip.

}}
\section{Theming/Styling}{


Like other bslib components, tooltips can be themed by supplying \href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html#tooltip-bg}{relevant theming variables}
to \code{\link[=bs_theme]{bs_theme()}}, which effects styling of every popover on the page. To
style a \emph{specific} popover differently from other popovers, utilize the
\code{customClass} option:

\if{html}{\out{<div class="sourceCode">}}\preformatted{tooltip(
  "Trigger", "Tooltip message",
  options = list(customClass = "my-tip")
)
}\if{html}{\out{</div>}}

And then add relevant rules to \code{\link[=bs_theme]{bs_theme()}} via \code{\link[=bs_add_rules]{bs_add_rules()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{bs_theme() |> bs_add_rules(".my-tip \{ max-width: none; \}")
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}

tooltip(
  shiny::actionButton("btn", "A button"),
  "A message"
)

card(
  card_header(
    tooltip(
      span("Card title ", bsicons::bs_icon("question-circle-fill")),
      "Additional info",
      placement = "right"
    )
  ),
  "Card body content..."
)
\dontshow{\}) # examplesIf}
}
\references{
\url{https://getbootstrap.com/docs/5.3/components/tooltips/}
}
\seealso{
\code{\link[=popover]{popover()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page_sidebar}
\alias{page_sidebar}
\title{A sidebar page (i.e., dashboard)}
\usage{
page_sidebar(
  ...,
  sidebar = NULL,
  title = NULL,
  fillable = TRUE,
  fillable_mobile = FALSE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL
)
}
\arguments{
\item{...}{UI elements to display in the 'main' content area (i.e., next to
the \code{sidebar}). These arguments are passed to \code{layout_sidebar()}, which has
more details.}

\item{sidebar}{A \code{\link[=sidebar]{sidebar()}} object.}

\item{title}{A string, number, or \code{\link[htmltools:builder]{htmltools::tag()}} child to display as the
title (just above the \code{sidebar}).}

\item{fillable}{Whether or not the \code{main} content area should be considered a
fillable (i.e., flexbox) container.}

\item{fillable_mobile}{Whether or not \code{fillable} pages should fill the viewport's
height on mobile devices (i.e., narrow windows).}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{window_title}{the browser window title. The default value, \code{NA}, means
to use any character strings that appear in \code{title} (if none are found, the
host URL of the page is displayed by default).}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}
}
\description{
Create a dashboard layout with a full-bleed header (\code{title}) and \code{\link[=sidebar]{sidebar()}}.
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}

library(shiny)
library(ggplot2)

ui <- page_sidebar(
  title = "Example dashboard",
  sidebar = sidebar(
    varSelectInput("var", "Select variable", mtcars)
  ),
  card(
    full_screen = TRUE,
    card_header("My plot"),
    plotOutput("p")
  )
)

server <- function(input, output) {
  output$p <- renderPlot({
    ggplot(mtcars) + geom_histogram(aes(!!input$var))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=layout_sidebar]{layout_sidebar()}} for 'floating' sidebar layouts.

\code{\link[=accordion]{accordion()}} for grouping related input controls in the \code{sidebar}.

\code{\link[=card]{card()}} for wrapping outputs in the 'main' content area.

\code{\link[=value_box]{value_box()}} for highlighting values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page_navbar}
\alias{page_navbar}
\title{Multi-page app with a top navigation bar}
\usage{
page_navbar(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  fillable_mobile = FALSE,
  gap = NULL,
  padding = NULL,
  header = NULL,
  footer = NULL,
  navbar_options = NULL,
  fluid = TRUE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL,
  position = deprecated(),
  bg = deprecated(),
  inverse = deprecated(),
  underline = deprecated(),
  collapsible = deprecated()
)
}
\arguments{
\item{...}{a collection of \code{\link[=nav_panel]{nav_panel()}} items.}

\item{title}{A (left-aligned) title to place in the card header/footer. If
provided, other nav items are automatically right aligned.}

\item{id}{a character string used for dynamically updating the container (see
\code{\link[=nav_select]{nav_select()}}).}

\item{selected}{a character string matching the \code{value} of a particular
\code{\link[=nav_panel]{nav_panel()}} item to selected by default.}

\item{sidebar}{A \code{\link[=sidebar]{sidebar()}} component to display on every \code{\link[=nav_panel]{nav_panel()}}
page.}

\item{fillable}{Whether or not to allow \code{fill} items to grow/shrink to fit
the browser window. If \code{TRUE}, all \code{\link[=nav_panel]{nav_panel()}} pages are \code{fillable}. A
character vector, matching the \code{value} of \code{\link[=nav_panel]{nav_panel()}}s to be filled, may
also be provided. Note that, if a \code{sidebar} is provided, \code{fillable} makes
the main content portion fillable.}

\item{fillable_mobile}{Whether or not \code{fillable} pages should fill the viewport's
height on mobile devices (i.e., narrow windows).}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}.}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{header}{UI element(s) (\link[htmltools:builder]{htmltools::tags}) to display \emph{above} the nav
content. For \code{card}-based navsets, these elements are implicitly wrapped in
a \code{card_body()}. To control things like \code{padding}, \code{fill}, etc., wrap the
elements in an explicit \code{\link[=card_body]{card_body()}}.}

\item{footer}{UI element(s) (\link[htmltools:builder]{htmltools::tags}) to display \emph{below} the nav
content. For \code{card}-based navsets, these elements are implicitly wrapped in
a \code{card_body()}. To control things like \code{padding}, \code{fill}, etc., wrap the
elements in an explicit \code{\link[=card_body]{card_body()}}.}

\item{navbar_options}{Options to control the appearance and behavior of the
navbar. Use \code{\link[=navbar_options]{navbar_options()}} to create the list of options.}

\item{fluid}{\code{TRUE} to use fluid layout; \code{FALSE} to use fixed
layout.}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{window_title}{the browser window title. The default value, \code{NA}, means
to use any character strings that appear in \code{title} (if none are found, the
host URL of the page is displayed by default).}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}

\item{position}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(position=)}} instead.}

\item{bg}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(bg=)}} instead.}

\item{inverse}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(inverse=)}} instead.}

\item{underline}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(underline=)}} instead.}

\item{collapsible}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(collapsible=)}} instead.}
}
\description{
Create a page that contains a top level navigation bar that can be used to
toggle a set of \code{\link[=nav_panel]{nav_panel()}} elements. Use this page layout to create the
effect of a multi-page app, where your app's content is broken up into
multiple "pages" that can be navigated to via the top navigation bar.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)

link_shiny <- tags$a(
  shiny::icon("github"), "Shiny",
  href = "https://github.com/rstudio/shiny",
  target = "_blank"
)
link_posit <- tags$a(
  shiny::icon("r-project"), "Posit",
  href = "https://posit.co",
  target = "_blank"
)

ui <- page_navbar(
  title = "My App",
  nav_panel(title = "One", p("First page content.")),
  nav_panel(title = "Two", p("Second page content.")),
  nav_panel("Three", p("Third page content.")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    align = "right",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)

server <- function(...) { } # not used in this example

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
\href{https://rstudio.github.io/bslib/articles/dashboards/index.html}{Getting Started with Dashboards}
on the bslib website.
}
\seealso{
\code{\link[=nav_panel]{nav_panel()}}, \code{\link[=nav_menu]{nav_menu()}}, and \code{\link[=nav_item]{nav_item()}} for adding content
sections and organizing or creating items in the navigation bar.

\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} for laying out content
into rows and columns.

\code{\link[=card]{card()}} for wrapping outputs in the 'main' content area.

\code{\link[=value_box]{value_box()}} for highlighting values.

\code{\link[=accordion]{accordion()}} for grouping related input controls in the \code{sidebar}.

Other Dashboard page layouts: 
\code{\link{page_fillable}()},
\code{\link{page_sidebar}()}
}
\concept{Dashboard page layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toolbar.R
\name{toolbar_input_button}
\alias{toolbar_input_button}
\alias{update_toolbar_input_button}
\title{Add a toolbar input button}
\usage{
toolbar_input_button(
  id,
  label,
  icon = NULL,
  show_label = is.null(icon),
  tooltip = !show_label,
  ...,
  disabled = FALSE,
  border = FALSE
)

update_toolbar_input_button(
  id,
  label = NULL,
  show_label = NULL,
  icon = NULL,
  disabled = NULL,
  session = get_current_session()
)
}
\arguments{
\item{id}{The input ID.}

\item{label}{The input label. By default, \code{label} is not shown but is used by
\code{tooltip}. Set \code{show_label = TRUE} to show the label (see \code{tooltip} for
details on how this affects the tooltip behavior).}

\item{icon}{An icon. If provided without \code{show_label = TRUE}, only the icon
will be visible.}

\item{show_label}{Whether to show the label text. If \code{FALSE} (the default),
only the icon is shown (if provided). If \code{TRUE}, the label text is shown
alongside the icon. Note that \code{show_label} can be dynamically updated using
\code{\link[=update_toolbar_input_button]{update_toolbar_input_button()}}.}

\item{tooltip}{Tooltip text to display when hovering over the input. Can be:
\itemize{
\item \code{TRUE} (default when \code{show_label = FALSE}) - shows a tooltip with the
\code{label} text
\item \code{FALSE} (default when \code{show_label = TRUE}) - no tooltip
\item A character string - shows a tooltip with custom text
}

Defaults to \code{!show_label}. When a tooltip is created, it will have an ID of
\code{"{id}_tooltip"} which can be used to update the tooltip text dynamically
via \code{\link[=update_tooltip]{update_tooltip()}}.}

\item{...}{Additional attributes to pass to the button.}

\item{disabled}{If \code{TRUE}, the button will not be clickable. Use
\code{\link[=update_toolbar_input_button]{update_toolbar_input_button()}} to dynamically enable/disable the button.}

\item{border}{Whether to show a border around the button.}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\value{
Returns a button suitable for use in a toolbar.
}
\description{
A button designed to fit well in small places such as in a \code{\link[=toolbar]{toolbar()}}.
}
\section{Functions}{
\itemize{
\item \code{toolbar_input_button()}: Create a toolbar button.

\item \code{update_toolbar_input_button()}: Update a toolbar button.

}}
\section{Updating toolbar buttons}{


You can dynamically update the appearance and enabled/disabled state
of a toolbar button on the client side using \code{\link[=update_toolbar_input_button]{update_toolbar_input_button()}}.
This function works similarly to \code{\link[shiny:updateActionButton]{shiny::updateActionButton()}}.

Note that you cannot change the \code{tooltip} or \code{border} parameters after the
button has been created, as these affect the button's structure and ARIA
attributes.
Please use \code{\link[=update_tooltip]{update_tooltip()}} to update the text of the tooltip if one is
present.

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(shiny)
library(bslib)

ui <- page_fluid(
  card(
    card_header(
      "Toolbar Demo",
      toolbar(
        align = "right",
        toolbar_input_button("btn", label = "Click me", icon = icon("play")),
        toolbar_input_button(
          "task",
          label = "Do a Task",
          icon = icon("play-circle"),
          show_label = TRUE
        )
      )
    ),
    card_body(
      verbatimTextOutput("count")
    )
  )
)

server <- function(input, output, session) \{
  output$count <- renderPrint(\{
    input$btn
  \})

  observeEvent(input$btn, \{
    update_toolbar_input_button(
      "btn",
      label = "Clicked!",
      icon = icon("check")
    )
    # Update the tooltip text
    update_tooltip("btn_tooltip", "Button was clicked!")
  \})

  # Handle task button - toggle between states
  observeEvent(input$task, \{
    if (input$task \%\% 2 == 1) \{
      # Show task is in progress
      update_toolbar_input_button(
        "task",
        label = "Task Running...",
        icon = icon("spinner")
      )
    \} else \{
      # Reset to original
      update_toolbar_input_button(
        "task",
        label = "Do a Task",
        icon = icon("play-circle"),
        session = session
      )
    \}
  \})
\}

shinyApp(ui, server)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
# Basic toolbar button
toolbar(
  toolbar_input_button(id = "view", icon = icon("eye"), label = "View"),
  toolbar_input_button(id = "save", icon = icon("save"), label = "Save")
)
\dontshow{\}) # examplesIf}
}
\seealso{
Other toolbar components:
\code{\link[=toolbar]{toolbar()}},
\code{\link[=toolbar_divider]{toolbar_divider()}},
\code{\link[=toolbar_input_select]{toolbar_input_select()}}
}
\concept{toolbar components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imports.R
\name{font_face}
\alias{font_face}
\alias{font_link}
\alias{font_google}
\alias{font_collection}
\title{Helpers for importing web fonts}
\description{
\code{font_google()}, \code{font_link()}, and \code{font_face()} are all re-exported from
the \pkg{sass} package (see \code{\link[sass:font_face]{sass::font_face()}} for details). For a quick
example of how to use these functions with \code{\link[=bs_theme]{bs_theme()}}, see the
examples section below.
}
\examples{

# If you have an internet connection, running the following code
# will download, cache, and import the relevant Google Font files
# for local use
theme <- bs_theme(
  base_font = font_google("Fira Sans"),
  code_font = font_google("Fira Code"),
  heading_font = font_google("Fredoka One")
)
if (interactive()) {
  bs_theme_preview(theme)
}

# Three different yet equivalent ways of importing a remotely-hosted Google Font
a <- font_google("Crimson Pro", wght = "200..900", local = FALSE)
b <- font_link(
  "Crimson Pro",
  href = "https://fonts.googleapis.com/css2?family=Crimson+Pro:wght@200..900"
)
url <- "https://fonts.gstatic.com/s/crimsonpro/v13/q5uDsoa5M_tv7IihmnkabARboYF6CsKj.woff2"
c <- font_face(
  family = "Crimson Pro",
  style = "normal",
  weight = "200 900",
  src = paste0("url(", url, ") format('woff2')")
)
theme <- bs_theme(base_font = c)
if (interactive()) {
  bs_theme_preview(theme)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preview.R
\name{run_with_themer}
\alias{run_with_themer}
\alias{bs_themer}
\title{Theme customization UI}
\usage{
run_with_themer(appDir = getwd(), ..., gfonts = TRUE, gfonts_update = FALSE)

bs_themer(gfonts = TRUE, gfonts_update = FALSE)
}
\arguments{
\item{appDir}{The application to run. This can be a file or directory path,
or a \code{\link[shiny:shinyApp]{shiny::shinyApp()}} object. See \code{\link[shiny:runApp]{shiny::runApp()}} for details.}

\item{...}{Additional parameters to pass through to \code{\link[shiny:runApp]{shiny::runApp()}}.}

\item{gfonts}{whether or not to detect Google Fonts and wrap them in
\code{\link[=font_google]{font_google()}} (so that their font files are automatically imported).}

\item{gfonts_update}{whether or not to update the internal database of
Google Fonts.}
}
\value{
nothing. These functions are called for their side-effects.
}
\description{
A 'real-time' theme customization UI that you can use to easily make common
tweaks to Bootstrap variables and immediately see how they would affect your
app's appearance. There are two ways you can launch the theming UI. For most
Shiny apps, just use \code{run_with_themer()} in place of \code{\link[shiny:runApp]{shiny::runApp()}}; they
should take the same arguments and work the same way. Alternatively, you can
call the \code{bs_themer()} function from inside your server function (or in an R
Markdown app that is using \code{runtime: shiny}, you can call this from any code
chunk). Note that this function is only intended to be used for development!
}
\details{
To help you utilize the changes you see in the preview, this utility prints
\code{\link[=bs_theme]{bs_theme()}} code to the R console.
}
\section{Limitations}{

\itemize{
\item Doesn't work with Bootstrap 3.
\item Doesn't work with IE11.
\item Only works inside Shiny apps and \code{runtime: shiny} R Markdown documents.
\itemize{
\item Can't be used with static R Markdown documents.
\item Can be used to some extent with \code{runtime: shiny_prerendered}, but only UI
rendered through a \code{context="server"} may update in real-time.
}
\item Doesn't work with '3rd party' custom widgets that don't make use of
\code{\link[=bs_dependency_defer]{bs_dependency_defer()}} or \code{\link[=bs_current_theme]{bs_current_theme()}}.
}
}

\examples{
library(shiny)

ui <- fluidPage(
  theme = bs_theme(bg = "black", fg = "white"),
  h1("Heading 1"),
  h2("Heading 2"),
  p(
    "Paragraph text;",
    tags$a(href = "https://www.rstudio.com", "a link")
  ),
  p(
    actionButton("cancel", "Cancel"),
    actionButton("continue", "Continue", class = "btn-primary")
  ),
  tabsetPanel(
    tabPanel("First tab",
      "The contents of the first tab"
    ),
    tabPanel("Second tab",
      "The contents of the second tab"
    )
  )
)

if (interactive()) {
  run_with_themer(shinyApp(ui, function(input, output) {}))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/breakpoints.R
\name{breakpoints}
\alias{breakpoints}
\title{Define breakpoint values}
\usage{
breakpoints(..., sm = NULL, md = NULL, lg = NULL)
}
\arguments{
\item{...}{Other breakpoints (e.g., \code{xl}).}

\item{sm}{Values to apply at the \code{sm} breakpoint.}

\item{md}{Values to apply at the \code{md} breakpoint.}

\item{lg}{Values to apply at the \code{lg} breakpoint.}
}
\description{
A generic constructor for responsive breakpoints.
}
\references{
\url{https://getbootstrap.com/docs/5.3/layout/breakpoints/}
}
\seealso{
\code{\link[=layout_columns]{layout_columns()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card.R
\name{card_body}
\alias{card_body}
\alias{card_title}
\alias{card_header}
\alias{card_footer}
\alias{card_image}
\alias{as.card_item}
\alias{is.card_item}
\title{Card items}
\usage{
card_body(
  ...,
  fillable = TRUE,
  min_height = NULL,
  max_height = NULL,
  max_height_full_screen = max_height,
  height = NULL,
  padding = NULL,
  gap = NULL,
  fill = TRUE,
  class = NULL
)

card_title(..., container = htmltools::h5)

card_header(..., gap = NULL, class = NULL, container = htmltools::div)

card_footer(..., class = NULL)

card_image(
  file,
  ...,
  alt = "",
  src = NULL,
  href = NULL,
  border_radius = c("auto", "top", "bottom", "all", "none"),
  mime_type = NULL,
  class = NULL,
  height = NULL,
  fill = FALSE,
  width = NULL,
  container = NULL
)

as.card_item(x)

is.card_item(x)
}
\arguments{
\item{...}{Unnamed arguments can be any valid child of an \link[htmltools:builder]{htmltools tag}. Named arguments become HTML attributes on returned
UI element.}

\item{fillable}{Whether or not the card item should be a fillable (i.e.
flexbox) container.}

\item{min_height, max_height, max_height_full_screen}{Any valid \link[htmltools:validateCssUnit]{CSS length unit}.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}. This argument is only applicable when \code{fillable = TRUE}}

\item{fill}{Whether to allow this element to grow/shrink to fit its \code{card()}
container.}

\item{class}{Additional CSS classes for the returned UI element.}

\item{container}{A function to generate an HTML element to contain the image.
Setting this value to \code{card_body()} places the image inside the card body
area, otherwise the image will extend to the edges of the card.}

\item{file}{A file path pointing an image. Local images (i.e. not a URI
starting with \verb{https://} or similar) will be base64 encoded and provided to
the \code{src} attribute of the \verb{<img>}. Alternatively, you may directly set
the image \code{src}, in which case \code{file} is ignored.}

\item{alt}{Alternate text for the image, used by screen readers and assistive
devices. Provide alt text with a description of the image for any images
with important content. If alt text is not provided, the image will be
considered to be decorative and will not be read or announced by screen
readers.

For more information, the Web Accessibility Initiative (WAI) has a
\href{https://www.w3.org/WAI/tutorials/images/}{helpful tutorial on alt text}.}

\item{src}{The \code{src} attribute of the \verb{<img>} tag. If provided, \code{file} is
ignored entirely. Use \code{src} to provide a relative path to a file that will
be served by the Shiny application and should not be base64 encoded.}

\item{href}{An optional URL to link to when a user clicks on the image.}

\item{border_radius}{Which side of the image should have rounded corners,
useful when \code{card_image()} is used as an image cap at the top or bottom of
the card.

The value of \code{border_radius} determines whether the \code{card-img-top}
(\code{"top"}), \code{card-img-bottom} (\code{"bottom"}), or \code{card-img} (\code{"all"})
\href{https://getbootstrap.com/docs/5.3/components/card/#images}{Bootstrap classes} are
applied to the card. The default \code{"auto"} value will use the image's
position within a \code{card()} to automatically choose the appropriate class.}

\item{mime_type}{The mime type of the \code{file} when it is base64 encoded. This
argument is available for advanced use cases where \code{\link[mime:guess_type]{mime::guess_type()}} is
unable to automatically determine the file type.}

\item{width}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{width="100\%"}).}

\item{x}{an object to test (or coerce to) a card item.}
}
\value{
An \code{\link[htmltools:builder]{htmltools::div()}} tag.
}
\description{
Components designed to be provided as direct children of a \code{\link[=card]{card()}}. For a
general overview of the \code{\link[=card]{card()}} API, see \href{https://rstudio.github.io/bslib/articles/cards/index.html}{the Cards article} or the
other articles listed in the \emph{References} section of the \code{\link[=card]{card()}}
documentation.
}
\section{Functions}{
\itemize{
\item \code{card_body()}: A general container for the "main content" of a \code{\link[=card]{card()}}.

\item \code{card_title()}: Similar to \code{card_header()} but without the border and background color.

\item \code{card_header()}: A header (with border and background color) for the \code{card()}. Typically appears before a \code{card_body()}.

\item \code{card_footer()}: A header (with border and background color) for the \code{card()}. Typically appears after a \code{card_body()}.

\item \code{card_image()}: Include static images in a card, for example as an
image cap at the top or bottom of the card.

\item \code{as.card_item()}: Mark an object as a card item. This will prevent the
\code{\link[=card]{card()}} from putting the object inside a \code{wrapper} (i.e., a
\code{card_body()}).

}}
\seealso{
\code{\link[=card]{card()}} creates a card component.

\code{\link[=navset_card_tab]{navset_card_tab()}}, \code{\link[=navset_card_pill]{navset_card_pill()}} and
\code{\link[=navset_card_underline]{navset_card_underline()}} create cards with tabbed navigation.

\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} help position multiple
cards into columns and rows and can also be used inside a card.

\code{\link[=layout_sidebar]{layout_sidebar()}} adds a sidebar to a card when nested in \code{\link[=card]{card()}}
or \code{\link[=card_body]{card_body()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-dark-mode.R
\name{input_dark_mode}
\alias{input_dark_mode}
\alias{toggle_dark_mode}
\title{Dark mode input control}
\usage{
input_dark_mode(..., id = NULL, mode = NULL)

toggle_dark_mode(mode = NULL, ..., session = get_current_session())
}
\arguments{
\item{...}{Additional attributes to be passed to the input control UI, such
as \code{class}, \code{style}, etc.

In \code{toggle_dark_mode()}, the \code{...} are included for future extensibility
and are currently ignored.}

\item{id}{An optional input id, required to reactively read the current color
mode.}

\item{mode}{The initial mode of the dark mode switch. By default or when set
to \code{NULL}, the user's system settings for preferred color scheme will be
used. Otherwise, set to \code{"light"} or \code{"dark"} to force a particular initial
mode.}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\value{
Returns a UI element for a dark mode switch input control. The server
value received for the input corresponding to \code{id} will be a string value
with the current color mode (\code{"light"} or \code{"dark"}).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Creates a button that toggles between dark and light modes, specifically for
toggling between light and dark \href{https://getbootstrap.com/docs/5.3/customize/color-modes/}{Bootstrap color modes} -- a new
feature added in \href{https://getbootstrap.com/docs/5.3/migration/#color-modes}{Bootstrap 5.3}.
}
\section{Functions}{
\itemize{
\item \code{input_dark_mode()}: Create a dark mode switch input control

\item \code{toggle_dark_mode()}: Programmatically toggle or set the current light
or dark color mode.

}}
\seealso{
Other input controls: 
\code{\link{input_switch}()}
}
\concept{input controls}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precompiled.R
\name{precompiled_css_path}
\alias{precompiled_css_path}
\title{Get the path to a precompiled CSS file}
\usage{
precompiled_css_path(theme = bs_theme())
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}
}
\value{
The path to the precompiled CSS file, if it exists, or \code{NULL}, if
not.
}
\description{
This function is meant for development and debugging purposes. It can be used
to test if a precompiled CSS file exists for a given theme object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/popover.R
\name{popover}
\alias{popover}
\alias{toggle_popover}
\alias{update_popover}
\title{Add a popover to a UI element}
\usage{
popover(
  trigger,
  ...,
  title = NULL,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
)

toggle_popover(id, show = NULL, session = get_current_session())

update_popover(id, ..., title = NULL, session = get_current_session())
}
\arguments{
\item{trigger}{The UI element to serve as the popover trigger (typically a
\code{\link[shiny:actionButton]{shiny::actionButton()}} or similar). If \code{trigger} renders as multiple HTML
elements (e.g., it's a \code{tagList()}), the last HTML element is used for the
trigger. If the \code{trigger} should contain all of those elements, wrap the
object in a \code{\link[=div]{div()}} or \code{\link[=span]{span()}}.}

\item{...}{UI elements for the popover's body. Character strings are
\link[=htmlEscape]{automatically escaped} unless marked as \code{\link[=HTML]{HTML()}}.}

\item{title}{A title (header) for the popover.}

\item{id}{A character string. Required to re-actively respond to the
visibility of the popover (via the \code{input[[id]]} value) and/or update the
visibility/contents of the popover.}

\item{placement}{The placement of the popover relative to its trigger.}

\item{options}{A list of additional
\href{https://getbootstrap.com/docs/5.3/components/popovers/#options}{options}.}

\item{show}{Whether to show (\code{TRUE}) or hide (\code{FALSE}) the popover. The
default (\code{NULL}) will show if currently hidden and hide if currently shown.
Note that a popover will not be shown if the trigger is not visible (e.g.,
it's hidden behind a tab).}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Display additional information when clicking on a UI element (typically a
button).
}
\section{Functions}{
\itemize{
\item \code{popover()}: Add a popover to a UI element

\item \code{toggle_popover()}: Programmatically show/hide a popover.

\item \code{update_popover()}: Update the contents of a popover.

}}
\section{Closing popovers}{


In addition to clicking the \code{close_button}, popovers can be closed by
pressing the Esc/Space key when the popover (and/or its trigger) is
focused.
}

\section{Theming/Styling}{


Like other bslib components, popovers can be themed by supplying
\href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html#popover-bg}{relevant theming variables}
to \code{\link[=bs_theme]{bs_theme()}},
which effects styling of every popover on the page.
To style a \emph{specific} popover differently from other popover, utilize the \code{customClass} option:

\if{html}{\out{<div class="sourceCode">}}\preformatted{popover(
  "Trigger", "Popover message",
  options = list(customClass = "my-pop")
)
}\if{html}{\out{</div>}}

And then add relevant rules to \code{\link[=bs_theme]{bs_theme()}} via \code{\link[=bs_add_rules]{bs_add_rules()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{bs_theme() |> bs_add_rules(".my-pop \{ max-width: none; \}")
}\if{html}{\out{</div>}}
}

\section{Accessibility of Popover Triggers}{


Because the user needs to interact with the \code{trigger} element to see the popover, it's best practice to use an element that is typically accessible via keyboard interactions, like a button or a link.
If you use a non-interactive element, like a \verb{<span>} or text, bslib will automatically add the \code{tabindex="0"} attribute to the trigger element to make sure that users can reach the element with the keyboard.
This means that in most cases you can use any element you want as the trigger.

One place where it's important to consider the accessibility of the trigger is when using an icon without any accompanying text.
In these cases, many R packages that provide icons will create an icon element with the assumption that the icon is decorative, which will make it inaccessible to users of assistive technologies.

When using an icon as the primary trigger, ensure that the icon does not have \code{aria-hidden="true"} or \code{role="presentation"} attributes.
Icon packages typically provide a way to specify a title for the icon, as well as a way to specify that the icon is not decorative.
The title should be a short description of the purpose of the trigger, rather than a description of the icon itself.
\itemize{
\item If you're using \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, provide a \code{title}.
\item If you're using \code{\link[fontawesome:fa]{fontawesome::fa()}}, set \code{a11y = "sem"} and provide a \code{title}.
}

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{popover(
  bsicons::bs_icon("gear", title = "Settings"),
  title = "Settings",
  sliderInput("n", "Number of points", 1, 100, 50)
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{popover(
  fontawesome::fa("gear", a11y = "sem", title = "Settings"),
  title = "Settings",
  sliderInput("n", "Number of points", 1, 100, 50)
)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}

popover(
  shiny::actionButton("btn", "A button"),
  "Popover body content...",
  title = "Popover title"
)

library(shiny)

ui <- page_fixed(
  card(class = "mt-5",
    card_header(
      popover(
        uiOutput("card_title", inline = TRUE),
        title = "Provide a new title",
        textInput("card_title", NULL, "An editable title")
      )
    ),
    "The card body..."
  )
)

server <- function(input, output) {
  output$card_title <- renderUI({
    list(input$card_title, bsicons::bs_icon("pencil-square"))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
Popovers are based on \href{https://getbootstrap.com/docs/5.3/components/popovers/}{Bootstrap's Popover component}. See the
bslib website for an \href{https://rstudio.github.io/bslib/articles/tooltips-popovers/index.html}{interactive introduction to tooltips and popovers}.
}
\seealso{
\code{\link[=tooltip]{tooltip()}} provides an alternative way to display informational
text on demand, typically when focusing or hovering over a trigger element.

Other Components: 
\code{\link{accordion}()},
\code{\link{card}()},
\code{\link{tooltip}()},
\code{\link{value_box}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version-default.R
\name{versions}
\alias{versions}
\alias{version_default}
\title{Available Bootstrap versions}
\usage{
versions()

version_default()
}
\value{
A list of the Bootstrap versions available.
}
\description{
Available Bootstrap versions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-dark-mode.R
\name{input_dark_mode}
\alias{input_dark_mode}
\alias{toggle_dark_mode}
\title{Dark mode input control}
\usage{
input_dark_mode(..., id = NULL, mode = NULL)

toggle_dark_mode(mode = NULL, ..., session = get_current_session())
}
\arguments{
\item{...}{Additional attributes to be passed to the input control UI, such
as \code{class}, \code{style}, etc.

In \code{toggle_dark_mode()}, the \code{...} are included for future extensibility
and are currently ignored.}

\item{id}{An optional input id, required to reactively read the current color
mode.}

\item{mode}{The initial mode of the dark mode switch. By default or when set
to \code{NULL}, the user's system settings for preferred color scheme will be
used. Otherwise, set to \code{"light"} or \code{"dark"} to force a particular initial
mode.}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\value{
Returns a UI element for a dark mode switch input control. The server
value received for the input corresponding to \code{id} will be a string value
with the current color mode (\code{"light"} or \code{"dark"}).
}
\description{
Creates a button that toggles between dark and light modes, specifically for
toggling between light and dark \href{https://getbootstrap.com/docs/5.3/customize/color-modes/}{Bootstrap color modes} -- a new
feature added in \href{https://getbootstrap.com/docs/5.3/migration/#color-modes}{Bootstrap 5.3}.
}
\section{Functions}{
\itemize{
\item \code{input_dark_mode()}: Create a dark mode switch input control

\item \code{toggle_dark_mode()}: Programmatically toggle or set the current light
or dark color mode.

}}
\seealso{
Other input controls: 
\code{\link{input_code_editor}()},
\code{\link{input_switch}()}
}
\concept{input controls}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toolbar.R
\name{toolbar_divider}
\alias{toolbar_divider}
\alias{toolbar_spacer}
\title{Toolbar: Add a divider or spacer to a toolbar}
\usage{
toolbar_divider(..., width = NULL, gap = NULL)

toolbar_spacer()
}
\arguments{
\item{...}{Ignored, reserved for future use and to require named arguments
in \code{\link[=toolbar_divider]{toolbar_divider()}}.}

\item{width}{A CSS length unit specifying the width of the divider line.
Defaults to \code{"2px"} for a sensible dividing line. Pass \verb{0px} for no
divider line.}

\item{gap}{A CSS length unit defining the spacing around the divider.
Defaults to \code{"1rem"} for sensible fixed spacing.}
}
\description{
\code{\link[=toolbar_divider]{toolbar_divider()}} creates a visual divider line with customizable and fixed
width and spacing between toolbar elements. \code{\link[=toolbar_spacer]{toolbar_spacer()}} creates empty
space that expands to push adjacent toolbar elements apart as much as
possible.
}
\section{Functions}{
\itemize{
\item \code{toolbar_divider()}: Create a dividing line and fixed space between
toolbar elements.

\item \code{toolbar_spacer()}: Create empty, expanding space between toolbar
elements. Note that for the spacer to push elements effectively, the parent
toolbar needs \verb{width: 100\%} (which is automatically applied when the toolbar
is a direct child of a label element). If the spacer doesn't appear to work,
you may need to set \code{width = "100\%"} explicitly on the \code{\link[=toolbar]{toolbar()}}.

}}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
toolbar(
  toolbar_input_button(id = "left1", label = "Left"),
  toolbar_divider(),
  toolbar_input_button(id = "right1", label = "Right")
)

toolbar(
  toolbar_input_button(id = "a", label = "A"),
  toolbar_divider(width = "5px", gap = "20px"),
  toolbar_input_button(id = "b", label = "B")
)

toolbar(
  toolbar_input_button(id = "previous", label = "Previous"),
  toolbar_spacer(),
  toolbar_input_button(id = "next", label = "Next")
)
\dontshow{\}) # examplesIf}
}
\seealso{
Other toolbar components:
\code{\link[=toolbar]{toolbar()}},
\code{\link[=toolbar_input_button]{toolbar_input_button()}},
\code{\link[=toolbar_input_select]{toolbar_input_select()}}
}
\concept{toolbar components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tooltip.R
\name{tooltip}
\alias{tooltip}
\alias{toggle_tooltip}
\alias{update_tooltip}
\title{Add a tooltip to a UI element}
\usage{
tooltip(
  trigger,
  ...,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
)

toggle_tooltip(id, show = NULL, session = get_current_session())

update_tooltip(id, ..., session = get_current_session())
}
\arguments{
\item{trigger}{A UI element (i.e., \link[htmltools:builder]{htmltools tag}) to serve
as the tooltip trigger. If \code{trigger} renders as multiple HTML
elements (e.g., it's a \code{tagList()}), the last HTML element is used for the
trigger. If the \code{trigger} should contain all of those elements, wrap the
object in a \code{\link[=div]{div()}} or \code{\link[=span]{span()}}.}

\item{...}{UI elements for the tooltip. Character strings are \link[=htmlEscape]{automatically escaped} unless marked as \code{\link[=HTML]{HTML()}}.}

\item{id}{a character string that matches an existing tooltip id.}

\item{placement}{The placement of the tooltip relative to its trigger.}

\item{options}{A list of additional \href{https://getbootstrap.com/docs/5.3/components/tooltips/#options}{options}.}

\item{show}{Whether to show (\code{TRUE}) or hide (\code{FALSE}) the tooltip. The
default (\code{NULL}) will show if currently hidden and hide if currently shown.
Note that a tooltip will not be shown if the trigger is not visible (e.g.,
it's hidden behind a tab).}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Display additional information when focusing (or hovering over) a UI element.
}
\section{Functions}{
\itemize{
\item \code{tooltip()}: Add a tooltip to a UI element

\item \code{toggle_tooltip()}: Programmatically show/hide a tooltip.

\item \code{update_tooltip()}: Update the contents of a tooltip.

}}
\section{Theming/Styling}{


Like other bslib components, tooltips can be themed by supplying
\href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html#tooltip-bg}{relevant theming variables}
to \code{\link[=bs_theme]{bs_theme()}},
which effects styling of every tooltip on the page.
To style a \emph{specific} tooltip differently from other tooltip, utilize the \code{customClass} option:

\if{html}{\out{<div class="sourceCode">}}\preformatted{tooltip(
  "Trigger", "Tooltip message",
  options = list(customClass = "my-tip")
)
}\if{html}{\out{</div>}}

And then add relevant rules to \code{\link[=bs_theme]{bs_theme()}} via \code{\link[=bs_add_rules]{bs_add_rules()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{bs_theme() |> bs_add_rules(".my-tip \{ max-width: none; \}")
}\if{html}{\out{</div>}}
}

\section{Accessibility of Tooltip Triggers}{


Because the user needs to interact with the \code{trigger} element to see the tooltip, it's best practice to use an element that is typically accessible via keyboard interactions, like a button or a link.
If you use a non-interactive element, like a \verb{<span>} or text, bslib will automatically add the \code{tabindex="0"} attribute to the trigger element to make sure that users can reach the element with the keyboard.
This means that in most cases you can use any element you want as the trigger.

One place where it's important to consider the accessibility of the trigger is when using an icon without any accompanying text.
In these cases, many R packages that provide icons will create an icon element with the assumption that the icon is decorative, which will make it inaccessible to users of assistive technologies.

When using an icon as the primary trigger, ensure that the icon does not have \code{aria-hidden="true"} or \code{role="presentation"} attributes.
Icon packages typically provide a way to specify a title for the icon, as well as a way to specify that the icon is not decorative.
The title should be a short description of the purpose of the trigger, rather than a description of the icon itself.
\itemize{
\item If you're using \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, provide a \code{title}.
\item If you're using \code{\link[fontawesome:fa]{fontawesome::fa()}}, set \code{a11y = "sem"} and provide a \code{title}.
}

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tooltip(
  bsicons::bs_icon("info-circle", title = "About tooltips"),
  "Text shown in the tooltip."
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tooltip(
  fontawesome::fa("info-circle", a11y = "sem", title = "About tooltips"),
  "Text shown in the tooltip."
)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}

tooltip(
  shiny::actionButton("btn", "A button"),
  "A message"
)

card(
  card_header(
    tooltip(
      span("Card title ", bsicons::bs_icon("question-circle-fill")),
      "Additional info",
      placement = "right"
    )
  ),
  "Card body content..."
)
\dontshow{\}) # examplesIf}
}
\references{
Tooltips are based on \href{https://getbootstrap.com/docs/5.3/components/tooltips/}{Bootstrap's Tooltip component}. See the
bslib website for an \href{https://rstudio.github.io/bslib/articles/tooltips-popovers/index.html}{interactive introduction to tooltips and popovers}.
}
\seealso{
\code{\link[=popover]{popover()}} provides a an alternative and more persistent container
for additional elements, typically revealed by clicking on a target
element.

Other Components: 
\code{\link{accordion}()},
\code{\link{card}()},
\code{\link{popover}()},
\code{\link{value_box}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.bslib_fragment}
\alias{print.bslib_fragment}
\alias{print.bslib_page}
\title{Make HTML browsable by default}
\usage{
\method{print}{bslib_fragment}(x, ...)

\method{print}{bslib_page}(x, ...)
}
\arguments{
\item{x}{a \code{\link[=tag]{tag()}} object.}

\item{...}{passed along to an underlying print method}
}
\description{
Make HTML browsable by default
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/popover.R
\name{popover}
\alias{popover}
\alias{toggle_popover}
\alias{update_popover}
\title{Add a popover to a UI element}
\usage{
popover(
  trigger,
  ...,
  title = NULL,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
)

toggle_popover(id, show = NULL, session = get_current_session())

update_popover(id, ..., title = NULL, session = get_current_session())
}
\arguments{
\item{trigger}{The UI element to serve as the popover trigger (typically a
\code{\link[shiny:actionButton]{shiny::actionButton()}} or similar). If \code{trigger} renders as multiple HTML
elements (e.g., it's a \code{tagList()}), the last HTML element is used for the
trigger. If the \code{trigger} should contain all of those elements, wrap the
object in a \code{\link[htmltools:builder]{htmltools::div()}} or \code{\link[htmltools:builder]{htmltools::span()}}.}

\item{...}{UI elements for the popover's body. Character strings are
\link[htmltools:htmlEscape]{automatically escaped} unless marked as
\code{\link[htmltools:HTML]{htmltools::HTML()}}.}

\item{title}{A title (header) for the popover. To remove a header
with \code{update_popover()}, provide a either an empty string or \code{character(0)}.}

\item{id}{A character string. Required to re-actively respond to the
visibility of the popover (via the \code{input[[id]]} value) and/or update the
visibility/contents of the popover.}

\item{placement}{The placement of the popover relative to its trigger.}

\item{options}{A list of additional
\href{https://getbootstrap.com/docs/5.3/components/popovers/#options}{options}.}

\item{show}{Whether to show (\code{TRUE}) or hide (\code{FALSE}) the popover. The
default (\code{NULL}) will show if currently hidden and hide if currently shown.
Note that a popover will not be shown if the trigger is not visible (e.g.,
it's hidden behind a tab).}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\description{
Display additional information when clicking on a UI element (typically a
button).
}
\section{Functions}{
\itemize{
\item \code{popover()}: Add a popover to a UI element

\item \code{toggle_popover()}: Programmatically show/hide a popover.

\item \code{update_popover()}: Update the contents of a popover.

}}
\section{Closing popovers}{


In addition to clicking the \code{close_button}, popovers can be closed by
pressing the Esc/Space key when the popover (and/or its trigger) is
focused.
}

\section{Theming/Styling}{


Like other bslib components, popovers can be themed by supplying
\href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html#popover-bg}{relevant theming variables}
to \code{\link[=bs_theme]{bs_theme()}},
which effects styling of every popover on the page.
To style a \emph{specific} popover differently from other popover, utilize the \code{customClass} option:

\if{html}{\out{<div class="sourceCode">}}\preformatted{popover(
  "Trigger", "Popover message",
  options = list(customClass = "my-pop")
)
}\if{html}{\out{</div>}}

And then add relevant rules to \code{\link[=bs_theme]{bs_theme()}} via \code{\link[=bs_add_rules]{bs_add_rules()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{bs_theme() |> bs_add_rules(".my-pop \{ max-width: none; \}")
}\if{html}{\out{</div>}}
}

\section{Accessibility of Popover Triggers}{


Because the user needs to interact with the \code{trigger} element to see the popover, it's best practice to use an element that is typically accessible via keyboard interactions, like a button or a link.
If you use a non-interactive element, like a \verb{<span>} or text, bslib will automatically add the \code{tabindex="0"} attribute to the trigger element to make sure that users can reach the element with the keyboard.
This means that in most cases you can use any element you want as the trigger.

One place where it's important to consider the accessibility of the trigger is when using an icon without any accompanying text.
In these cases, many R packages that provide icons will create an icon element with the assumption that the icon is decorative, which will make it inaccessible to users of assistive technologies.

When using an icon as the primary trigger, ensure that the icon does not have \code{aria-hidden="true"} or \code{role="presentation"} attributes.
Icon packages typically provide a way to specify a title for the icon, as well as a way to specify that the icon is not decorative.
The title should be a short description of the purpose of the trigger, rather than a description of the icon itself.
\itemize{
\item If you're using \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, provide a \code{title}.
\item If you're using \code{\link[fontawesome:fa]{fontawesome::fa()}}, set \code{a11y = "sem"} and provide a \code{title}.
}

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{popover(
  bsicons::bs_icon("gear", title = "Settings"),
  title = "Settings",
  sliderInput("n", "Number of points", 1, 100, 50)
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{popover(
  fontawesome::fa("gear", a11y = "sem", title = "Settings"),
  title = "Settings",
  sliderInput("n", "Number of points", 1, 100, 50)
)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}

popover(
  shiny::actionButton("btn", "A button"),
  "Popover body content...",
  title = "Popover title"
)

library(shiny)
library(bslib)

ui <- page_fixed(
  card(class = "mt-5",
    card_header(
      popover(
        uiOutput("card_title", inline = TRUE),
        title = "Provide a new title",
        textInput("card_title", NULL, "An editable title")
      )
    ),
    "The card body..."
  )
)

server <- function(input, output) {
  output$card_title <- renderUI({
    list(input$card_title, bsicons::bs_icon("pencil-square"))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
Popovers are based on \href{https://getbootstrap.com/docs/5.3/components/popovers/}{Bootstrap's Popover component}. See the
bslib website for an \href{https://rstudio.github.io/bslib/articles/tooltips-popovers/index.html}{interactive introduction to tooltips and popovers}.
}
\seealso{
\code{\link[=tooltip]{tooltip()}} provides an alternative way to display informational
text on demand, typically when focusing or hovering over a trigger element.

Other Components: 
\code{\link{accordion}()},
\code{\link{card}()},
\code{\link{tooltip}()},
\code{\link{value_box}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-layers.R
\name{bs_add_variables}
\alias{bs_add_variables}
\alias{bs_add_rules}
\alias{bs_add_functions}
\alias{bs_add_mixins}
\alias{bs_bundle}
\title{Add low-level theming customizations}
\usage{
bs_add_variables(
  theme,
  ...,
  .where = "defaults",
  .default_flag = identical(.where, "defaults")
)

bs_add_rules(theme, rules)

bs_add_functions(theme, functions)

bs_add_mixins(theme, mixins)

bs_bundle(theme, ...)
}
\arguments{
\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}

\item{...}{\itemize{
\item \code{bs_add_variables()}: Should be named Sass variables or values that can be passed in directly to the \code{defaults} argument of a \code{\link[sass:sass_layer]{sass::sass_layer()}}.
\item \code{bs_bundle()}: Should be arguments that can be handled by \code{\link[sass:sass_layer]{sass::sass_bundle()}} to be appended to the \code{theme}
}}

\item{.where}{Whether to place the variable definitions before other Sass
\code{"defaults"}, after other Sass \code{"declarations"}, or after other Sass
\code{"rules"}.}

\item{.default_flag}{Whether or not to add a \code{!default} flag (if missing) to
variable expressions. It's recommended to keep this as \code{TRUE} when \code{.where = "defaults"}.}

\item{rules}{Sass rules. Anything understood by \code{\link[sass:as_sass]{sass::as_sass()}} may be
provided (e.g., a list, character vector, \code{\link[sass:sass_import]{sass::sass_file()}}, etc)}

\item{functions}{A character vector or \code{\link[sass:sass_import]{sass::sass_file()}} containing
functions definitions.}

\item{mixins}{A character vector or \code{\link[sass:sass_import]{sass::sass_file()}} containing
mixin definitions.}
}
\value{
a modified \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
Compared to higher-level theme customization available in \code{\link[=bs_theme]{bs_theme()}}, these functions
are a more direct interface to Bootstrap Sass, and therefore, do nothing to
ensure theme customizations are portable between major Bootstrap versions.
}
\section{Functions}{
\itemize{
\item \code{bs_add_variables()}: Add Bootstrap Sass \href{https://getbootstrap.com/docs/4.4/getting-started/theming/#variable-defaults}{variable defaults}

\item \code{bs_add_rules()}: Add additional \href{https://sass-lang.com/documentation/style-rules}{Sass rules}

\item \code{bs_add_functions()}: Add additional \href{https://rstudio.github.io/sass/articles/sass.html#functions-1}{Sass functions}

\item \code{bs_add_mixins()}: Add additional \href{https://rstudio.github.io/sass/articles/sass.html#mixins-1}{Sass mixins}

\item \code{bs_bundle()}: Add additional \code{\link[sass:sass_layer]{sass::sass_bundle()}} objects to an existing \code{theme}.

}}
\examples{

# Function to preview the styling a (primary) Bootstrap button
library(htmltools)
button <- tags$a(class = "btn btn-primary", href = "#", role = "button", "Hello")
preview_button <- function(theme) {
  if (interactive()) {
    browsable(tags$body(bs_theme_dependencies(theme), button))
  }
}

# Here we start with a theme based on a Bootswatch theme,
# then override some variable defaults
theme <- bs_add_variables(
  bs_theme(bootswatch = "sketchy", primary = "orange"),
  "body-bg" = "#EEEEEE",
  "font-family-base" = "monospace",
  "font-size-base" = "1.4rem",
  "btn-padding-y" = ".16rem",
  "btn-padding-x" = "2rem"
)

preview_button(theme)

# If you need to set a variable based on another Bootstrap variable
theme <- bs_add_variables(theme, "body-color" = "$success", .where = "declarations")
preview_button(theme)

# Start a new global theme and add some custom rules that
# use Bootstrap variables to define a custom styling for a
# 'person card'
person_rules <- system.file("custom", "person.scss", package = "bslib")
theme <- bs_add_rules(bs_theme(), sass::sass_file(person_rules))
# Include custom CSS that leverages bootstrap Sass variables
person <- function(name, title, company) {
  tags$div(
    class = "person",
    h3(class = "name", name),
    div(class = "title", title),
    div(class = "company", company)
  )
}
if (interactive()) {
  browsable(shiny::fluidPage(
    theme = theme,
    person("Andrew Carnegie", "Owner", "Carnegie Steel Company"),
    person("John D. Rockefeller", "Chairman", "Standard Oil")
  ))
}

}
\references{
\url{https://getbootstrap.com/docs/4.4/getting-started/theming/}

\url{https://rstudio.github.io/sass/articles/sass.html#layering}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tooltip.R
\name{tooltip}
\alias{tooltip}
\alias{toggle_tooltip}
\alias{update_tooltip}
\title{Add a tooltip to a UI element}
\usage{
tooltip(
  trigger,
  ...,
  id = NULL,
  placement = c("auto", "top", "right", "bottom", "left"),
  options = list()
)

toggle_tooltip(id, show = NULL, session = get_current_session())

update_tooltip(id, ..., session = get_current_session())
}
\arguments{
\item{trigger}{A UI element (i.e., \link[htmltools:builder]{htmltools tag}) to serve
as the tooltip trigger. If \code{trigger} renders as multiple HTML
elements (e.g., it's a \code{tagList()}), the last HTML element is used for the
trigger. If the \code{trigger} should contain all of those elements, wrap the
object in a \code{\link[htmltools:builder]{htmltools::div()}} or \code{\link[htmltools:builder]{htmltools::span()}}.}

\item{...}{UI elements for the tooltip. Character strings are \link[htmltools:htmlEscape]{automatically escaped} unless marked as \code{\link[htmltools:HTML]{htmltools::HTML()}}.}

\item{id}{a character string that matches an existing tooltip id.}

\item{placement}{The placement of the tooltip relative to its trigger.}

\item{options}{A list of additional \href{https://getbootstrap.com/docs/5.3/components/tooltips/#options}{options}.}

\item{show}{Whether to show (\code{TRUE}) or hide (\code{FALSE}) the tooltip. The
default (\code{NULL}) will show if currently hidden and hide if currently shown.
Note that a tooltip will not be shown if the trigger is not visible (e.g.,
it's hidden behind a tab).}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\description{
Display additional information when focusing (or hovering over) a UI element.
}
\section{Functions}{
\itemize{
\item \code{tooltip()}: Add a tooltip to a UI element

\item \code{toggle_tooltip()}: Programmatically show/hide a tooltip.

\item \code{update_tooltip()}: Update the contents of a tooltip.

}}
\section{Theming/Styling}{


Like other bslib components, tooltips can be themed by supplying
\href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html#tooltip-bg}{relevant theming variables}
to \code{\link[=bs_theme]{bs_theme()}},
which effects styling of every tooltip on the page.
To style a \emph{specific} tooltip differently from other tooltip, utilize the \code{customClass} option:

\if{html}{\out{<div class="sourceCode">}}\preformatted{tooltip(
  "Trigger", "Tooltip message",
  options = list(customClass = "my-tip")
)
}\if{html}{\out{</div>}}

And then add relevant rules to \code{\link[=bs_theme]{bs_theme()}} via \code{\link[=bs_add_rules]{bs_add_rules()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{bs_theme() |> bs_add_rules(".my-tip \{ max-width: none; \}")
}\if{html}{\out{</div>}}
}

\section{Accessibility of Tooltip Triggers}{


Because the user needs to interact with the \code{trigger} element to see the tooltip, it's best practice to use an element that is typically accessible via keyboard interactions, like a button or a link.
If you use a non-interactive element, like a \verb{<span>} or text, bslib will automatically add the \code{tabindex="0"} attribute to the trigger element to make sure that users can reach the element with the keyboard.
This means that in most cases you can use any element you want as the trigger.

One place where it's important to consider the accessibility of the trigger is when using an icon without any accompanying text.
In these cases, many R packages that provide icons will create an icon element with the assumption that the icon is decorative, which will make it inaccessible to users of assistive technologies.

When using an icon as the primary trigger, ensure that the icon does not have \code{aria-hidden="true"} or \code{role="presentation"} attributes.
Icon packages typically provide a way to specify a title for the icon, as well as a way to specify that the icon is not decorative.
The title should be a short description of the purpose of the trigger, rather than a description of the icon itself.
\itemize{
\item If you're using \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, provide a \code{title}.
\item If you're using \code{\link[fontawesome:fa]{fontawesome::fa()}}, set \code{a11y = "sem"} and provide a \code{title}.
}

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tooltip(
  bsicons::bs_icon("info-circle", title = "About tooltips"),
  "Text shown in the tooltip."
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tooltip(
  fontawesome::fa("info-circle", a11y = "sem", title = "About tooltips"),
  "Text shown in the tooltip."
)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}

tooltip(
  shiny::actionButton("btn", "A button"),
  "A message"
)

card(
  card_header(
    tooltip(
      span("Card title ", bsicons::bs_icon("question-circle-fill")),
      "Additional info",
      placement = "right"
    )
  ),
  "Card body content..."
)
\dontshow{\}) # examplesIf}
}
\references{
Tooltips are based on \href{https://getbootstrap.com/docs/5.3/components/tooltips/}{Bootstrap's Tooltip component}. See the
bslib website for an \href{https://rstudio.github.io/bslib/articles/tooltips-popovers/index.html}{interactive introduction to tooltips and popovers}.
}
\seealso{
\code{\link[=popover]{popover()}} provides a an alternative and more persistent container
for additional elements, typically revealed by clicking on a target
element.

Other Components: 
\code{\link{accordion}()},
\code{\link{card}()},
\code{\link{popover}()},
\code{\link{value_box}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page_sidebar}
\alias{page_sidebar}
\title{A sidebar page (i.e., dashboard)}
\usage{
page_sidebar(
  ...,
  sidebar = NULL,
  title = NULL,
  fillable = TRUE,
  fillable_mobile = FALSE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL
)
}
\arguments{
\item{...}{UI elements to display in the 'main' content area (i.e., next to
the \code{sidebar}). These arguments are passed to \code{\link[=layout_sidebar]{layout_sidebar()}}, which has
more details.}

\item{sidebar}{A \code{\link[=sidebar]{sidebar()}} object.}

\item{title}{A string, number, or \code{\link[htmltools:builder]{htmltools::tag()}} child to display as the
title (just above the \code{sidebar}).}

\item{fillable}{Whether or not the \code{main} content area should be considered a
fillable (i.e., flexbox) container.}

\item{fillable_mobile}{Whether or not the page should fill the viewport's
height on mobile devices (i.e., narrow windows).}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{window_title}{the browser window title. The default value, \code{NA}, means
to use any character strings that appear in \code{title} (if none are found, the
host URL of the page is displayed by default).}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}
}
\description{
Create a dashboard layout with a full-width header (\code{title}) and \code{\link[=sidebar]{sidebar()}}.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}

library(shiny)
library(ggplot2)
library(bslib)

ui <- page_sidebar(
  title = "Example dashboard",
  sidebar = sidebar(
    varSelectInput("var", "Select variable", mtcars)
  ),
  card(
    full_screen = TRUE,
    card_header("My plot"),
    plotOutput("p")
  )
)

server <- function(input, output) {
  output$p <- renderPlot({
    ggplot(mtcars) + geom_histogram(aes(!!input$var))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\references{
\href{https://rstudio.github.io/bslib/articles/dashboards/index.html}{Getting Started with Dashboards}
on the bslib website.
}
\seealso{
\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} for laying out content
into rows and columns.

\code{\link[=accordion]{accordion()}} for grouping related input controls in the \code{sidebar}.

\code{\link[=card]{card()}} for wrapping outputs in the 'main' content area.

\code{\link[=value_box]{value_box()}} for highlighting values.

Other Dashboard page layouts: 
\code{\link{page_fillable}()},
\code{\link{page_navbar}()}
}
\concept{Dashboard page layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme.R
\name{bs_theme}
\alias{bs_theme}
\alias{bs_theme_update}
\alias{is_bs_theme}
\title{Create a Bootstrap theme}
\usage{
bs_theme(
  version = version_default(),
  bootswatch = NULL,
  ...,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  font_scale = NULL
)

bs_theme_update(
  theme,
  ...,
  bootswatch = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  font_scale = NULL
)

is_bs_theme(x)
}
\arguments{
\item{version}{The major version of Bootstrap to use (see \code{\link[=versions]{versions()}}
for possible values). Defaults to the currently recommended version
for new projects (currently Bootstrap 5).}

\item{bootswatch}{The name of a bootswatch theme (see \code{\link[=bootswatch_themes]{bootswatch_themes()}}
for possible values). When provided to \code{bs_theme_update()}, any previous
Bootswatch theme is first removed before the new one is applied (use
\code{bootswatch = "default"} to effectively remove the Bootswatch theme).}

\item{...}{arguments passed along to \code{\link[=bs_add_variables]{bs_add_variables()}}.}

\item{bg}{A color string for the background.}

\item{fg}{A color string for the foreground.}

\item{primary}{A color to be used for hyperlinks, to indicate primary/default
actions, and to show active selection state in some Bootstrap components.
Generally a bold, saturated color that contrasts with the theme's base
colors.}

\item{secondary}{A color for components and messages that don't need to stand
out. (Not supported in Bootstrap 3.)}

\item{success}{A color for messages that indicate an operation has succeeded.
Typically green.}

\item{info}{A color for messages that are informative but not critical.
Typically a shade of blue-green.}

\item{warning}{A color for warning messages. Typically yellow.}

\item{danger}{A color for errors. Typically red.}

\item{base_font}{The default typeface.}

\item{code_font}{The typeface to be used for code. Be sure this is monospace!}

\item{heading_font}{The typeface to be used for heading elements.}

\item{font_scale}{A scalar multiplier to apply to the base font size. For
example, a value of \code{1.5} scales font sizes to 150\% and a value of \code{0.8}
scales to 80\%. Must be a positive number.}

\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}

\item{x}{an object.}
}
\value{
a \code{\link[sass:sass_layer]{sass::sass_bundle()}} (list-like) object.
}
\description{
Creates a Bootstrap theme object, where you can:
\itemize{
\item Choose a (major) Bootstrap \code{version}.
\item Choose a \href{https://bootswatch.com}{Bootswatch theme} (optional).
\item Customize main colors and fonts via explicitly named arguments (e.g.,
\code{bg}, \code{fg}, \code{primary}, etc).
\item Customize other, lower-level, Bootstrap Sass variable defaults via \code{...}.
}

To learn more about how to implement custom themes, as well as how to use them inside Shiny and R Markdown, \href{https://rstudio.github.io/bslib/articles/bslib.html}{see here}.
}
\section{Colors}{


Colors may be provided in any format that \code{\link[htmltools:parseCssColors]{htmltools::parseCssColors()}} can
understand. To control the vast majority of the ('grayscale') color
defaults, specify both the \code{fg} (foreground) and \code{bg} (background) colors.
The \code{primary} and \code{secondary} theme colors are also useful for accenting the
main grayscale colors in things like hyperlinks, tabset panels, and buttons.
}

\section{Fonts}{


Use \code{base_font}, \code{code_font}, and \code{heading_font} to control the main
typefaces. These arguments set new defaults for the relevant \code{font-family}
CSS properties, but don't necessarily import the relevant font files. To
both set CSS properties \emph{and} import font files, consider using the various
\code{\link[=font_face]{font_face()}} helpers.

Each \verb{*_font} argument may be collection of character vectors,
\code{\link[=font_google]{font_google()}}s, \code{\link[=font_link]{font_link()}}s and/or \code{\link[=font_face]{font_face()}}s. Note that a
character vector can have:
\itemize{
\item A single unquoted name (e.g., \code{"Source Sans Pro"}).
\item A single quoted name (e.g., \code{"'Source Sans Pro'"}).
\item A comma-separated list of names w/ individual names quoted as necessary.
(e.g. \code{c("Open Sans", "'Source Sans Pro'", "'Helvetica Neue', Helvetica, sans-serif")})
}

Since \code{font_google(..., local = TRUE)} guarantees that the client has access to
the font family, meaning it's relatively safe to specify just one font
family, for instance:

\if{html}{\out{<div class="sourceCode">}}\preformatted{bs_theme(base_font = font_google("Pacifico", local = TRUE))
}\if{html}{\out{</div>}}

However, specifying multiple "fallback" font families is recommended,
especially when relying on remote and/or system fonts being available, for
instance. Fallback fonts are useful not only for handling missing fonts, but
also for handling a Flash of Invisible Text (FOIT) which can be quite
noticeable with remote web fonts on a slow internet connection.

\if{html}{\out{<div class="sourceCode">}}\preformatted{bs_theme(base_font = font_collection(font_google("Pacifico", local = FALSE), "Roboto", "sans-serif"))
}\if{html}{\out{</div>}}
}

\examples{

theme <- bs_theme(
  # Controls the default grayscale palette
  bg = "#202123", fg = "#B8BCC2",
  # Controls the accent (e.g., hyperlink, button, etc) colors
  primary = "#EA80FC", secondary = "#48DAC6",
  base_font = c("Grandstander", "sans-serif"),
  code_font = c("Courier", "monospace"),
  heading_font = "'Helvetica Neue', Helvetica, sans-serif",
  # Can also add lower-level customization
  "input-border-color" = "#EA80FC"
)
if (interactive()) {
  bs_theme_preview(theme)
}

# Lower-level bs_add_*() functions allow you to work more
# directly with the underlying Sass code
theme <- bs_add_variables(theme, "my-class-color" = "red")
theme <- bs_add_rules(theme, ".my-class { color: $my-class-color }")

}
\references{
\url{https://rstudio.github.io/bslib/articles/bslib.html}

\url{https://rstudio.github.io/sass/}
}
\seealso{
\code{\link[=bs_add_variables]{bs_add_variables()}}, \code{\link[=bs_theme_preview]{bs_theme_preview()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/breakpoints.R
\name{breakpoints}
\alias{breakpoints}
\title{Define breakpoint values}
\usage{
breakpoints(..., sm = NULL, md = NULL, lg = NULL)
}
\arguments{
\item{...}{Other breakpoints (e.g., \code{xl}).}

\item{sm}{Values to apply at the \code{sm} breakpoint.}

\item{md}{Values to apply at the \code{md} breakpoint.}

\item{lg}{Values to apply at the \code{lg} breakpoint.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

A generic constructor for responsive breakpoints.
}
\examples{
breakpoints(sm = c(4, 4, 4), md = c(3, 3, 6), lg = c(-2, 8, -2))

}
\references{
Bootstrap's
\href{https://getbootstrap.com/docs/5.3/layout/breakpoints/}{Breakpoints article}
provides more detail on breakpoints and how they are used and customized.
}
\seealso{
\code{breakpoints()} is used by \code{\link[=layout_columns]{layout_columns()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nav-items.R
\name{nav}
\alias{nav}
\alias{nav_menu}
\alias{nav_content}
\alias{nav_item}
\alias{nav_spacer}
\title{Navigation items}
\usage{
nav(title, ..., value = title, icon = NULL)

nav_menu(title, ..., value = title, icon = NULL, align = c("left", "right"))

nav_content(value, ..., icon = NULL)

nav_item(...)

nav_spacer()
}
\arguments{
\item{title}{A title to display. Can be a character string or UI elements
(i.e., \link{tags}).}

\item{...}{Depends on the function:
\itemize{
\item For \code{nav()} and \code{nav_content()}: UI elements (i.e., \link{tags}) to display
when the item is active.
\item For \code{nav_menu()}: a collection of nav items (e.g., \code{nav()}, \code{nav_item()}).
\item For \code{nav_item()}: UI elements (i.e., \link{tags}) to place directly in
the navigation panel (e.g., search forms, links to external content, etc).
}}

\item{value}{A character string to assign to the nav item. This value may be
supplied to the relevant container's \code{selected} argument in order to show
particular nav item's content immediately on page load. This value is also
useful for programmatically updating the selected content via
\code{\link[=nav_select]{nav_select()}}, \code{\link[=nav_hide]{nav_hide()}}, etc (updating selected tabs this way is often
useful for showing/hiding panels of content via other UI controls like
\code{\link[shiny:radioButtons]{shiny::radioButtons()}} -- in this scenario, consider using \code{\link[=nav_content]{nav_content()}}
with \code{\link[=navs_hidden]{navs_hidden()}}).}

\item{icon}{Optional icon to appear next to the nav item's \code{title}.}

\item{align}{horizontal alignment of the dropdown menu relative to dropdown toggle.}
}
\value{
A nav item that may be passed to a nav container (e.g. \code{\link[=navs_tab]{navs_tab()}}).
}
\description{
Create nav item(s) for use inside nav containers (e.g., \code{\link[=navs_tab]{navs_tab()}},
\code{\link[=navs_bar]{navs_bar()}}, etc).
}
\section{Functions}{
\itemize{
\item \code{nav()}: Content to display when the given item is selected.

\item \code{nav_menu()}: Create a menu of nav items.

\item \code{nav_content()}: Create nav content for use inside \code{navs_hidden()} (for
creating custom navigation controls via \code{navs_select()}),

\item \code{nav_item()}: Place arbitrary content in the navigation panel (e.g., search
forms, links to external content, etc.)

\item \code{nav_spacer()}: Adding spacing between nav items.

}}
\seealso{
\code{\link[=navs_tab]{navs_tab()}}, \code{\link[=nav_select]{nav_select()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-remove.R
\name{bs_remove}
\alias{bs_remove}
\alias{bs_retrieve}
\title{Remove or retrieve Sass code from a theme}
\usage{
bs_remove(theme, ids = character(0))

bs_retrieve(theme, ids = character(0), include_unnamed = TRUE)
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{ids}{a character vector of ids}

\item{include_unnamed}{whether or not to include unnamed \code{\link[sass:sass_layer]{sass::sass_layer()}}s
(e.g., Bootstrap Sass variables, functions, and mixins).}
}
\value{
Returns a modified \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
A Bootstrap theme created with \code{\link[=bs_theme]{bs_theme()}} is comprised of
\href{https://rstudio.github.io/sass/articles/sass.html#layering}{many Sass layers}.
\code{bs_remove()} and \code{bs_retrieve()} allow you to remove or retrieve an
individual layer, either to reduce the size of the compiled CSS or to extract
styles from a theme.
}
\examples{
bs4 <- bs_theme(version = 4)

# Retrieve sass bundle for print styles
bs_retrieve(bs4, "_print", include_unnamed = FALSE)

# Remove CSS rules for print and carousels
bs4_no_print <- bs_remove(bs4, c("_print", "_carousel"))
suppressWarnings(
  bs_retrieve(bs4_no_print, "_print", include_unnamed = FALSE)
)

# Remove BS3 compatibility layer
bs4_no_compat <- bs_remove(bs4, "bs3compat")
}
\seealso{
Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme_preview}()},
\code{\link{bs_theme}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/breakpoints.R
\name{breakpoints}
\alias{breakpoints}
\title{Define breakpoint values}
\usage{
breakpoints(..., xs = NULL, sm = NULL, md = NULL, lg = NULL)
}
\arguments{
\item{...}{Other breakpoints (e.g., \code{xl}).}

\item{xs}{The default value to apply to the \code{xs} breakpoint. Note that this
breakpoint is generally equivalent to "all sizes" and is typically treated
as the base case or a value to apply by default across all breakpoints
unless overridden by a larger breakpoint.}

\item{sm}{Values to apply at the \code{sm} breakpoint.}

\item{md}{Values to apply at the \code{md} breakpoint.}

\item{lg}{Values to apply at the \code{lg} breakpoint.}
}
\description{
A generic constructor for responsive breakpoints.
}
\examples{
breakpoints(sm = c(4, 4, 4), md = c(3, 3, 6), lg = c(-2, 8, -2))

}
\references{
Bootstrap's
\href{https://getbootstrap.com/docs/5.3/layout/breakpoints/}{Breakpoints article}
provides more detail on breakpoints and how they are used and customized.
}
\seealso{
\code{breakpoints()} is used by \code{\link[=layout_columns]{layout_columns()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-global.R, R/bs-theme.R
\name{bs_global_theme}
\alias{bs_global_theme}
\alias{bs_global_set}
\alias{bs_global_get}
\alias{bs_global_clear}
\alias{bs_global_add_variables}
\alias{bs_global_add_rules}
\alias{bs_global_bundle}
\alias{bs_global_theme_update}
\title{Global theming}
\usage{
bs_global_theme(
  version = version_default(),
  bootswatch = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  ...
)

bs_global_set(theme = bs_theme())

bs_global_get()

bs_global_clear()

bs_global_add_variables(
  ...,
  .where = "defaults",
  .default_flag = identical(.where, "defaults")
)

bs_global_add_rules(...)

bs_global_bundle(...)

bs_global_theme_update(
  ...,
  bootswatch = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL
)
}
\arguments{
\item{version}{The major version of Bootstrap to use (see \code{\link[=versions]{versions()}}
for possible values). Defaults to the currently recommended version
for new projects (currently Bootstrap 5).}

\item{bootswatch}{The name of a bootswatch theme (see \code{\link[=bootswatch_themes]{bootswatch_themes()}}
for possible values). When provided to \code{bs_theme_update()}, any previous
Bootswatch theme is first removed before the new one is applied (use
\code{bootswatch = "default"} to effectively remove the Bootswatch theme).}

\item{bg}{A color string for the background.}

\item{fg}{A color string for the foreground.}

\item{primary}{A color to be used for hyperlinks, to indicate primary/default
actions, and to show active selection state in some Bootstrap components.
Generally a bold, saturated color that contrasts with the theme's base
colors.}

\item{secondary}{A color for components and messages that don't need to stand
out. (Not supported in Bootstrap 3.)}

\item{success}{A color for messages that indicate an operation has succeeded.
Typically green.}

\item{info}{A color for messages that are informative but not critical.
Typically a shade of blue-green.}

\item{warning}{A color for warning messages. Typically yellow.}

\item{danger}{A color for errors. Typically red.}

\item{base_font}{The default typeface.}

\item{code_font}{The typeface to be used for code. Be sure this is monospace!}

\item{heading_font}{The typeface to be used for heading elements.}

\item{...}{arguments passed along to \code{\link[=bs_add_variables]{bs_add_variables()}}.}

\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}

\item{.where}{Whether to place the variable definitions before other Sass
\code{"defaults"}, after other Sass \code{"declarations"}, or after other Sass
\code{"rules"}.}

\item{.default_flag}{Whether or not to add a \code{!default} flag (if missing) to
variable expressions. It's recommended to keep this as \code{TRUE} when \code{.where = "defaults"}.}
}
\value{
functions that modify the global theme (e.g., \code{bs_global_set()})
invisibly return the previously set theme. \code{bs_global_get()} returns the
current global theme.
}
\description{
\code{bs_global_theme()} creates a new (global) Bootstrap Sass theme which
\code{\link[=bs_theme_dependencies]{bs_theme_dependencies()}} (or \code{\link[=sass_partial]{sass_partial()}}) can consume (their \code{theme}
argument defaults to \code{bs_global_get()}, which get the current global theme).
}
\examples{

# Remember the global state now (so we can restore later)
theme <- bs_global_get()
# Use Bootstrap 3 (globally) with some theme customization
bs_global_theme(3, bg = "#444", fg = "#e4e4e4", primary = "#e39777")
if (interactive()) bs_theme_preview(with_themer = FALSE)
# If no global theme is active, bs_global_get() returns NULL
bs_global_clear()
bs_global_get()
# Restore the original state
bs_global_set(theme)

}
\seealso{
\code{\link[=bs_theme]{bs_theme()}}, \code{\link[=bs_theme_preview]{bs_theme_preview()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootswatch.R
\name{theme_version}
\alias{theme_version}
\title{Obtain a theme's Bootstrap version}
\usage{
theme_version(theme)
}
\arguments{
\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}
}
\value{
the major version of Bootstrap used in the \code{theme}.
}
\description{
Obtain a theme's Bootstrap version
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nav-update.R
\name{nav_select}
\alias{nav_select}
\alias{nav_insert}
\alias{nav_remove}
\alias{nav_show}
\alias{nav_hide}
\title{Dynamically update nav containers}
\usage{
nav_select(id, selected = NULL, session = getDefaultReactiveDomain())

nav_insert(
  id,
  nav,
  target = NULL,
  position = c("after", "before"),
  select = FALSE,
  session = getDefaultReactiveDomain()
)

nav_remove(id, target, session = getDefaultReactiveDomain())

nav_show(id, target, select = FALSE, session = getDefaultReactiveDomain())

nav_hide(id, target, session = getDefaultReactiveDomain())
}
\arguments{
\item{id}{a character string used to identify the nav container.}

\item{selected}{a character string used to identify a particular \code{\link[=nav]{nav()}} item.}

\item{session}{a shiny session object (the default should almost always be used).}

\item{nav}{a \code{\link[=nav]{nav()}} item.}

\item{target}{The \code{value} of an existing \code{nav()} item, next to which tab will be added. If removing: the \code{value} of the \code{nav()} item that you want to remove.}

\item{position}{Should \code{nav} be added before or after the target?}

\item{select}{Should \code{nav} be selected upon being inserted?}
}
\description{
Functions for dynamically updating nav containers (e.g., select, insert, and
remove nav items). These functions require an \code{id} on the nav container to be
specified.
}
\examples{

can_browse <- function() interactive() && require("shiny")

# Selecting a tab
if (can_browse()) {
  shinyApp(
    page_fluid(
      radioButtons("item", "Choose", c("A", "B")),
      navs_hidden(
        id = "container",
        nav_content("A", "a"),
        nav_content("B", "b")
      )
    ),
    function(input, output) {
      observe(nav_select("container", input$item))
    }
  )
}

# Inserting and removing
if (can_browse()) {
  ui <- page_fluid(
    actionButton("add", "Add 'Dynamic' tab"),
    actionButton("remove", "Remove 'Foo' tab"),
    navs_tab(
      id = "tabs",
      nav("Hello", "hello"),
      nav("Foo", "foo"),
      nav("Bar", "bar tab")
    )
  )
  server <- function(input, output) {
    observeEvent(input$add, {
      nav_insert(
        "tabs", target = "Bar", select = TRUE,
        nav("Dynamic", "Dynamically added content")
      )
    })
    observeEvent(input$remove, {
      nav_remove("tabs", target = "Foo")
    })
  }
  shinyApp(ui, server)
}

}
\seealso{
\code{\link[=nav]{nav()}}, \code{\link[=navs_tab]{navs_tab()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-remove.R
\name{bs_remove}
\alias{bs_remove}
\alias{bs_retrieve}
\title{Remove or retrieve Sass code from a theme}
\usage{
bs_remove(theme, ids = character(0))

bs_retrieve(theme, ids = character(0), include_unnamed = TRUE)
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{ids}{a character vector of ids}

\item{include_unnamed}{whether or not to include unnamed \code{\link[sass:sass_layer]{sass::sass_layer()}}s
(e.g., Bootstrap Sass variables, functions, and mixins).}
}
\value{
Returns a modified \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
A Bootstrap theme created with \code{\link[=bs_theme]{bs_theme()}} is comprised of
\href{https://rstudio.github.io/sass/articles/sass.html#layering}{many Sass layers}.
\code{bs_remove()} and \code{bs_retrieve()} allow you to remove or retrieve an
individual layer, either to reduce the size of the compiled CSS or to extract
styles from a theme.
}
\examples{
bs4 <- bs_theme(version = 4)

# Retrieve sass bundle for print styles
bs_retrieve(bs4, "_print", include_unnamed = FALSE)

# Remove CSS rules for print and carousels
bs4_no_print <- bs_remove(bs4, c("_print", "_carousel"))
suppressWarnings(
  bs_retrieve(bs4_no_print, "_print", include_unnamed = FALSE)
)

# Remove BS3 compatibility layer
bs4_no_compat <- bs_remove(bs4, "bs3compat")
}
\seealso{
Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_theme}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme_preview}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-dependencies.R
\name{bs_dependency}
\alias{bs_dependency}
\alias{bs_dependency_defer}
\title{Themeable HTML components}
\usage{
bs_dependency(
  input = list(),
  theme,
  name,
  version,
  cache_key_extra = NULL,
  .dep_args = list(),
  .sass_args = list()
)

bs_dependency_defer(func, memoise = TRUE)
}
\arguments{
\item{input}{Sass rules to compile, using \code{theme}.}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{name}{Library name}

\item{version}{Library version}

\item{cache_key_extra}{Extra information to add to the sass cache key. It is
useful to add the version of your package.}

\item{.dep_args}{A list of additional arguments to pass to
\code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}}. Note that \code{package} has no effect and
\code{script} must be absolute path(s).}

\item{.sass_args}{A list of additional arguments to pass to
\code{\link[sass:sass_partial]{sass::sass_partial()}}.}

\item{func}{a \emph{non-anonymous} function, with a \emph{single} argument.
This function should accept a \code{\link[=bs_theme]{bs_theme()}} object and return a single
\code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}}, a list of them, or \code{NULL}.}

\item{memoise}{whether or not to memoise (i.e., cache) \code{func} results for a
short period of time. The default, \code{TRUE}, can have large performance
benefits when many instances of the same themable widget are rendered. Note
that you may want to avoid memoisation if \code{func} relies on side-effects
(e.g., files on-disk) that need to change for each themable widget
instance.}
}
\value{
\code{bs_dependency()} returns an \code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}} and
\code{bs_dependency_defer()} returns an \code{\link[htmltools:tagFunction]{htmltools::tagFunction()}}
}
\description{
Themeable HTML components use Sass to generate CSS rules from Bootstrap Sass
variables, functions, and/or mixins (i.e., stuff inside of \code{theme}).
\code{bs_dependencies()} makes it a bit easier to create themeable components by
compiling \code{\link[sass:sass]{sass::sass()}} (\code{input}) together with Bootstrap Sass inside of a
\code{theme}, and packaging up the result into an \code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}}.

Themable components can also be  \emph{dynamically} themed inside of Shiny (i.e.,
they may be themed in 'real-time' via \code{\link[=bs_themer]{bs_themer()}}, and more generally,
update their styles in response to \link[shiny:session]{shiny::session}'s \code{setCurrentTheme()}
method). Dynamically themeable components provide a "recipe" (i.e., a
function) to \code{bs_dependency_defer()}, describing how to generate new CSS
stylesheet(s) from a new \code{theme}. This function is called when the HTML page
is first rendered, and may be invoked again with a new \code{theme} whenever
\link[shiny:session]{shiny::session}'s \code{setCurrentTheme()} is called.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
myWidgetVersion <- "1.2.3"

myWidgetDependency <- function() {
  list(
    bs_dependency_defer(myWidgetCss),
    htmlDependency(
      name = "mywidget-js",
      version = myWidgetVersion,
      src = system.file(package = "mypackage", "js"),
      script = "mywidget.js"
    )
  )
}

myWidgetCSS <- function(theme) {
  if (!is_bs_theme(theme)) {
    return(
      htmlDependency(
        name = "mywidget-css",
        version = myWidgetVersion,
        src = system.file(package = "mypackage", "css"),
        stylesheet = "mywidget.css"
      )
    )
  }

  # Compile mywidget.scss using the variables and defaults from the theme
  # object.
  sass_input <- sass::sass_file(system.file(package = "mypackage", "scss/mywidget.scss"))

  bs_dependency(
    input = sass_input,
    theme = theme,
    name = "mywidget",
    version = myWidgetVersion,
    cache_key_extra = utils::packageVersion("mypackage")
  )
}

# Note that myWidgetDependency is not defined inside of myWidget. This is so
# that, if `myWidget()` is called multiple times, Shiny can tell that the
# function objects are identical and deduplicate them.
myWidget <- function(id) {
  div(
    id = id,
    span("myWidget"),
    myWidgetDependency()
  )
}
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/custom-components/index.html}{Theming: Custom components}
gives a tutorial on creating a dynamically themable custom component.
}
}
\seealso{
Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme_preview}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/navs-legacy.R, R/navs.R
\name{navset}
\alias{navset}
\alias{navset_tab}
\alias{navset_pill}
\alias{navset_pill_list}
\alias{navset_hidden}
\alias{navset_bar}
\alias{navset_card_tab}
\alias{navset_card_pill}
\title{Navigation containers}
\usage{
navset_tab(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navset_pill(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navset_pill_list(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  well = TRUE,
  fluid = TRUE,
  widths = c(4, 8)
)

navset_hidden(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navset_bar(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  gap = NULL,
  padding = NULL,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  header = NULL,
  footer = NULL,
  bg = NULL,
  inverse = "auto",
  collapsible = TRUE,
  fluid = TRUE
)

navset_card_tab(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  sidebar = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  full_screen = FALSE,
  wrapper = card_body
)

navset_card_pill(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  sidebar = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  placement = c("above", "below"),
  full_screen = FALSE,
  wrapper = card_body
)
}
\arguments{
\item{...}{a collection of \code{\link[=nav_panel]{nav_panel()}} items.}

\item{id}{a character string used for dynamically updating the container (see
\code{\link[=nav_select]{nav_select()}}).}

\item{selected}{a character string matching the \code{value} of a particular
\code{\link[=nav_panel]{nav_panel()}} item to selected by default.}

\item{header}{UI element(s) (\link{tags}) to display \emph{above} the nav content.}

\item{footer}{UI element(s) (\link{tags}) to display \emph{below} the nav content.}

\item{well}{\code{TRUE} to place a well (gray rounded rectangle) around the
navigation list.}

\item{fluid}{\code{TRUE} to use fluid layout; \code{FALSE} to use fixed
layout.}

\item{widths}{Column widths of the navigation list and tabset content areas
respectively.}

\item{title}{A (left-aligned) title to place in the card header/footer. If
provided, other nav items are automatically right aligned.}

\item{sidebar}{A \code{\link[=sidebar]{sidebar()}} component to display on every \code{\link[=nav_panel]{nav_panel()}}
page.}

\item{fillable}{Whether or not to allow \code{fill} items to grow/shrink to fit
the browser window. If \code{TRUE}, all \code{\link[=nav_panel]{nav_panel()}} pages are \code{fillable}. A
character vector, matching the \code{value} of \code{\link[=nav_panel]{nav_panel()}}s to be filled, may
also be provided. Note that, if a \code{sidebar} is provided, \code{fillable} makes
the main content portion fillable.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}.}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{position}{Determines whether the navbar should be displayed at the top
of the page with normal scrolling behavior (\code{"static-top"}), pinned at
the top (\code{"fixed-top"}), or pinned at the bottom
(\code{"fixed-bottom"}). Note that using \code{"fixed-top"} or
\code{"fixed-bottom"} will cause the navbar to overlay your body content,
unless you add padding, e.g.: \code{tags$style(type="text/css", "body
  {padding-top: 70px;}")}}

\item{bg}{a CSS color to use for the navbar's background color.}

\item{inverse}{Either \code{TRUE} for a light text color or \code{FALSE} for a dark
text color. If \code{"auto"} (the default), the best contrast to \code{bg} is chosen.}

\item{collapsible}{\code{TRUE} to automatically collapse the navigation
elements into a menu when the width of the browser is less than 940 pixels
(useful for viewing on smaller touchscreen device)}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{full_screen}{If \code{TRUE}, an icon will appear when hovering over the card
body. Clicking the icon expands the card to fit viewport size.}

\item{wrapper}{A function (which returns a UI element) to call on unnamed
arguments in \code{...} which are not already card item(s) (like
\code{\link[=card_header]{card_header()}}, \code{\link[=card_body]{card_body()}}, etc.). Note that non-card items are grouped
together into one \code{wrapper} call (e.g. given \code{card("a", "b", card_body("c"), "d")}, \code{wrapper} would be called twice, once with \code{"a"} and
\code{"b"} and once with \code{"d"}).}

\item{placement}{placement of the nav items relative to the content.}
}
\description{
Render a collection of \code{\link[=nav_panel]{nav_panel()}} items into a container.
}
\details{
\subsection{A basic example}{

This first example creates a simple tabbed navigation container with two
tabs. The tab name and the content of each tab are specified in the
\code{nav_panel()} calls and \code{navset_tab()} creates the tabbed navigation
around these two tabs.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(htmltools)

navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content."))
)
}\if{html}{\out{</div>}}

\figure{navset-tab-basic.png}{Screenshot of a basic navset_tab() example.}

In the rest of the examples, we’ll include links among the tabs (or
pills) in the navigation controls.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{link_shiny <- tags$a(shiny::icon("github"), "Shiny", href = "https://github.com/rstudio/shiny", target = "_blank")
link_posit <- tags$a(shiny::icon("r-project"), "Posit", href = "https://posit.co", target = "_blank")
}\if{html}{\out{</div>}}
}

\subsection{\code{navset_tab()}}{

You can fully customize the controls in the navigation component. In
this example, we’ve added a direct link to the Shiny repository using
\code{nav_item()}. We’ve also included a dropdown menu using \code{nav_menu()}
containing an option to select a third tab panel and another direct link
to Posit’s website. Finally, we’ve separated the primary tabs on the
left from the direct link and dropdown menu on the right using
\code{nav_spacer()}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-tab.png}{Screenshot of a navset_tab() example.}
}

\subsection{\code{navset_pill()}}{

\code{navset_pill()} creates a navigation container that behaves exactly like
\code{navset_tab()}, but the tab toggles are \emph{pills} or button-shaped.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_pill(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-pill.png}{Screenshot of a navset_pill() example.}
}

\subsection{\code{navset_card_tab()}}{

The tabbed navigation container can also be used in a \code{card()} component
thanks to \code{navset_card_tab()}. Learn more about this approach in the
\href{https://pkgs.rstudio.com/bslib/articles/cards.html}{article about Cards}, including
how to add \href{https://pkgs.rstudio.com/bslib/articles/sidebars.html#multi-page-layout}{a shared sidebar}
to all tabs in the card using the \code{sidebar} argument of
\code{navset_card_tab()}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_card_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-tab-card.png}{Screenshot of a navset_tab_card() example.}
}

\subsection{\code{navset_card_pill()}}{

Similar to \code{navset_pill()}, \code{navset_card_pill()} provides a pill-shaped
variant to \code{navset_card_tab()}. You can use the \code{placement} argument to
position the navbar \code{"above"} or \code{"below"} the card body.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_card_pill(
  placement = "above",
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-card-pill.png}{Screenshot of a navset_card_pill() example.}
}

\subsection{\code{navset_pill_list()}}{

Furthermore, \code{navset_pill_list()} creates a vertical list of navigation
controls adjacent to, rather than on top of, the tab content panels.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_pill_list(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third tab content")),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-pill-list.png}{Screenshot of a navset_pill_list() example.}
}

\subsection{\code{page_navbar()}}{

Finally, \code{page_navbar()} provides full-page navigation container similar
to \code{navset_tab()} but where each \code{nav_panel()} is treated as a full page
of content and the navigation controls appear in a top-level navigation
bar.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{page_navbar(
  title = "My App",
  bg = "#0062cc",
  nav_panel(title = "One", p("First page content.")),
  nav_panel(title = "Two", p("Second page content.")),
  nav_spacer(),
  nav_item(link_shiny),
  nav_menu(
    title = "Other links",
    align = "right",
    nav_panel("Three", p("Third page content.")),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{page-navbar.png}{Screenshot of a page_navbar() example.}
}
}
\seealso{
\code{\link[=nav_panel]{nav_panel()}}, \code{\link[=nav_select]{nav_select()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page}
\alias{page}
\alias{page_fluid}
\alias{page_fixed}
\alias{page_fill}
\alias{page_navbar}
\title{Create a Bootstrap page}
\usage{
page(..., title = NULL, theme = bs_theme(), lang = NULL)

page_fluid(..., title = NULL, theme = bs_theme(), lang = NULL)

page_fixed(..., title = NULL, theme = bs_theme(), lang = NULL)

page_fill(..., padding = 0, title = NULL, theme = bs_theme(), lang = NULL)

page_navbar(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  header = NULL,
  footer = NULL,
  bg = NULL,
  inverse = "auto",
  collapsible = TRUE,
  fluid = TRUE,
  theme = bs_theme(),
  window_title = NA,
  lang = NULL
)
}
\arguments{
\item{...}{The contents of the document body.}

\item{title}{The browser window title (defaults to the host URL of the page)}

\item{theme}{One of the following:
\itemize{
\item \code{NULL} (the default), which implies a "stock" build of Bootstrap 3.
\item A \code{\link[bslib:bs_theme]{bslib::bs_theme()}} object. This can be used to replace a stock
build of Bootstrap 3 with a customized version of Bootstrap 3 or higher.
\item A character string pointing to an alternative Bootstrap stylesheet
(normally a css file within the www directory, e.g. \code{www/bootstrap.css}).
}}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{id}{a character string used for dynamically updating the container (see \code{\link[=nav_select]{nav_select()}}).}

\item{selected}{a character string matching the \code{value} of a particular \code{\link[=nav]{nav()}} item to selected by default.}

\item{position}{Determines whether the navbar should be displayed at the top
of the page with normal scrolling behavior (\code{"static-top"}), pinned at
the top (\code{"fixed-top"}), or pinned at the bottom
(\code{"fixed-bottom"}). Note that using \code{"fixed-top"} or
\code{"fixed-bottom"} will cause the navbar to overlay your body content,
unless you add padding, e.g.: \code{tags$style(type="text/css", "body
  {padding-top: 70px;}")}}

\item{header}{UI element(s) (\link{tags}) to display \emph{above} the nav content.}

\item{footer}{UI element(s) (\link{tags}) to display \emph{below} the nav content.}

\item{bg}{a CSS color to use for the navbar's background color.}

\item{inverse}{Either \code{TRUE} for a light text color or \code{FALSE} for a dark
text color. If \code{"auto"} (the default), the best contrast to \code{bg} is chosen.}

\item{collapsible}{\code{TRUE} to automatically collapse the navigation
elements into a menu when the width of the browser is less than 940 pixels
(useful for viewing on smaller touchscreen device)}

\item{fluid}{\code{TRUE} to use fluid layout; \code{FALSE} to use fixed
layout.}

\item{window_title}{the browser window title. The default value, \code{NA}, means
to use any character strings that appear in \code{title} (if none are found, the
host URL of the page is displayed by default).}
}
\description{
These functions are small wrappers around shiny's page constructors (i.e., \code{\link[shiny:fluidPage]{shiny::fluidPage()}}, \code{\link[shiny:navbarPage]{shiny::navbarPage()}}, etc) that differ in two ways:
\itemize{
\item The \code{theme} parameter defaults bslib's recommended version of Bootstrap (for new projects).
\item The return value is rendered as an static HTML page when printed interactively at the console.
}
}
\seealso{
\code{\link[shiny:bootstrapPage]{shiny::bootstrapPage()}}

\code{\link[shiny:fluidPage]{shiny::fluidPage()}}

\code{\link[shiny:fixedPage]{shiny::fixedPage()}}

\code{\link[shiny:fillPage]{shiny::fillPage()}}

\code{\link[shiny:navbarPage]{shiny::navbarPage()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-dependencies.R
\name{bs_theme_dependencies}
\alias{bs_theme_dependencies}
\title{Compile Bootstrap Sass with (optional) theming}
\usage{
bs_theme_dependencies(
  theme,
  sass_options = sass::sass_options_get(output_style = "compressed"),
  cache = sass::sass_cache_get(),
  jquery = jquerylib::jquery_core(3),
  precompiled = get_precompiled_option("bslib.precompiled", default = TRUE)
)
}
\arguments{
\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}

\item{sass_options}{a \code{\link[sass:sass_options]{sass::sass_options()}} object.}

\item{cache}{This can be a directory to use for the cache, a \link[sass]{FileCache}
object created by \code{\link[sass:sass_file_cache]{sass_file_cache()}}, or \code{FALSE} or \code{NULL} for no caching.}

\item{jquery}{a \code{\link[jquerylib:jquery_core]{jquerylib::jquery_core()}} object.}

\item{precompiled}{Before compiling the theme object, first look for a
precompiled CSS file for the \code{\link[=theme_version]{theme_version()}}.  If \code{precompiled = TRUE}
and a precompiled CSS file exists for the theme object, it will be fetched
immediately and not compiled. At the moment, we only provide precompiled
CSS for "stock" builds of Bootstrap (i.e., no theming additions, bootswatch
themes, or non-default \code{sass_options}).}
}
\value{
a list of HTML dependencies containing Bootstrap CSS, Bootstrap
JavaScript, and \code{jquery}. This list may contain additional HTML
dependencies if bundled with the \code{theme}.
}
\description{
\code{bs_theme_dependencies()} compiles Bootstrap Sass into CSS and returns it,
along with other HTML dependencies, as a list of
\code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}}s. Most users won't need to call this function
directly as Shiny & R Markdown will perform this compilation automatically
when handed a \code{\link[=bs_theme]{bs_theme()}}. If you're here looking to create a themeable
component, see \code{\link[=bs_dependency]{bs_dependency()}}.
}
\section{Sass caching and precompilation}{


If Shiny Developer Mode is enabled (by setting \code{options(shiny.devmode = TRUE)} or calling \code{shiny::devmode(TRUE)}), both \pkg{sass} caching and \pkg{bslib}
precompilation are disabled by default; that is, a call to
\code{bs_theme_dependencies(theme)} expands to \code{bs_theme_dependencies(theme, cache = F, precompiled = F)}). This is useful for local development as
enabling caching/precompilation may produce incorrect results if local
changes are made to bslib's source files.
}

\examples{

# Function to preview the styling a (primary) Bootstrap button
library(htmltools)
button <- tags$a(class = "btn btn-primary", href = "#", role = "button", "Hello")
preview_button <- function(theme) {
  if (interactive()) {
    browsable(tags$body(bs_theme_dependencies(theme), button))
  }
}

# Latest Bootstrap
preview_button(bs_theme())
# Bootstrap 3
preview_button(bs_theme(3))
# Bootswatch 4 minty theme
preview_button(bs_theme(4, bootswatch = "minty"))
# Bootswatch 4 sketchy theme
preview_button(bs_theme(4, bootswatch = "sketchy"))

}
\seealso{
\code{\link[=bs_theme]{bs_theme()}}, \code{\link[=bs_dependency]{bs_dependency()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toolbar.R
\name{toolbar}
\alias{toolbar}
\title{Toolbar component}
\usage{
toolbar(..., align = c("right", "left"), gap = NULL, width = NULL)
}
\arguments{
\item{...}{UI elements for the toolbar.}

\item{align}{Determines if toolbar should be aligned to the \code{"right"} or
\code{"left"}.}

\item{gap}{A CSS length unit defining the gap (i.e., spacing) between
elements in the toolbar. Defaults to \code{0} (no gap).}

\item{width}{CSS width of the toolbar. Defaults to \code{NULL}, which will
automatically set \verb{width: 100\%} when the toolbar is a direct child of a
label element (e.g., when used in input labels). For \code{\link[=toolbar_spacer]{toolbar_spacer()}} to
push elements effectively, the toolbar needs \verb{width: 100\%} to expand and
create space. Set this explicitly if you need to control the width in other
contexts.}
}
\value{
Returns a toolbar element.
}
\description{
A toolbar which can contain buttons, inputs, and other UI elements in a small
form suitable for inclusion in card headers, footers, and other small places.

bslib includes a complete example of toolbars and the many ways they can be used:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{shiny::runExample("toolbar", package = "bslib")
}\if{html}{\out{</div>}}
}
\section{Cookbook}{

\subsection{Toolbars in Card Headers and Footers}{

Card headers and footers are a common places you might want to use toolbars. Toolbars
allow you to clearly show that a selection of inputs pertain to that
particular card. For example, this card uses multiple
\code{\link[=toolbar_input_select]{toolbar_input_select()}} for filtering and sorting, along with toolbar
buttons and a "share" button in the toolbar in the card footer.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{card(
  full_screen = TRUE,
  card_header(
    "Sales Data",
    toolbar(
      align = "right",
      toolbar_input_select(
        id = "filter",
        label = "Filter",
        choices = c("All", "Active", "Inactive"),
        icon = icon("filter")
      ),
      toolbar_input_select(
        id = "sort",
        label = "Sort by",
        choices = c("Date", "Amount", "Customer"),
        icon = icon("sort")
      ),
      toolbar_divider(),
      toolbar_input_button(
        id = "refresh",
        label = "Refresh",
        icon = icon("arrows-rotate")
      ),
      toolbar_input_button(
        id = "export",
        label = "Export",
        icon = icon("download"),
        show_label = TRUE
      )
    )
  ),
  card_body(
    h3("Card Body Here"),
  ),
  card_footer(
    toolbar(
      align = "right",
      toolbar_input_button(
        id = "share_data",
        label = "Share",
        icon = icon("share-nodes"),
        show_label = TRUE,
        border = TRUE
      )
    )
  )
)
}\if{html}{\out{</div>}}
}

\subsection{Toolbars in labels}{

You can use toolbars in the labels of other Shiny inputs to add a composite
input with additional buttons or controls. The following example uses a
\code{\link[=toolbar]{toolbar()}} in the label of a \code{\link[shiny:numericInput]{shiny::numericInput()}} to add increment and
decrement buttons next to the label:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{shiny::numericInput(
  "quantity",
  label = toolbar(
    "Quantity",
    toolbar_spacer(), # push buttons to the right
    toolbar_input_button("decrement", "Less", icon("minus")),
    toolbar_input_button("increment", "More", icon("plus"))
  ),
  value = 1,
  min = 0,
  max = 100
)
}\if{html}{\out{</div>}}

You can also use toolbars in the labels of text area inputs. For example,
this text editor uses a toolbar with formatting buttons:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{textAreaInput(
  "editor",
  label = toolbar(
    "Comment",
    toolbar_spacer(),
    toolbar_input_button("bold", label = "Bold", icon = icon("bold")),
    toolbar_input_button("italic", label = "Italic", icon = icon("italic")),
    toolbar_input_button("link", label = "Link", icon = icon("link"))
  ),
  value = "",
  rows = 5,
  placeholder = "Type your comment here..."
)
}\if{html}{\out{</div>}}

The \code{\link[=input_submit_textarea]{input_submit_textarea()}} function from bslib allows you to create a
text area input with a submit button and an optional toolbar. Here is an
example of using a toolbar with formatting buttons and options:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{input_submit_textarea(
  "message",
  placeholder = "Type a message...",
  toolbar = toolbar(
    toolbar_input_button("attach", icon = icon("paperclip"), label = "Attach"),
    toolbar_input_button("emoji", icon = icon("face-smile"), label = "Emoji"),
    toolbar_divider(),
    toolbar_input_select(
      "format",
      label = "Format",
      choices = c("Plain", "Markdown", "HTML"),
      icon = icon("code")
    )
  )
)
}\if{html}{\out{</div>}}
}
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
# Minimal toolbar example
toolbar(
  toolbar_input_button(id = "view", icon = icon("eye"), label = "View"),
  toolbar_input_button(id = "save", icon = icon("save"), label = "Save"),
  toolbar_divider(),
  toolbar_input_select(
    id = "filter",
    label = "Filter",
    choices = c("All", "Active", "Inactive")
  )
)

# Toolbar with text input
library(shiny)
library(bslib)

ui <- page_fluid(
  numericInput(
    "quantity",
    label = toolbar(
      "Quantity",
      toolbar_spacer(),
      toolbar_input_button("decrement", "Less", icon("minus")),
      toolbar_input_button("increment", "More", icon("plus"))
    ),
    value = 5,
    min = 0,
    max = 100
  ),
  verbatimTextOutput("value")
)

server <- function(input, output, session) {
  output$value <- renderText({
    paste("Current value:", input$quantity)
  })

  observeEvent(input$increment, {
    updateNumericInput(session, "quantity", value = input$quantity + 1)
  })

  observeEvent(input$decrement, {
    updateNumericInput(session, "quantity", value = input$quantity - 1)
  })
}

shinyApp(ui, server)

# Toolbar with input_submit_textarea()
library(shiny)
library(bslib)

ui <- page_fluid(
  input_submit_textarea(
    "message",
    placeholder = "Type a message...",
    toolbar = toolbar(
      toolbar_input_button("attach", icon = icon("paperclip"), label = "Attach"),
      toolbar_input_button("emoji", icon = icon("face-smile"), label = "Emoji"),
      toolbar_divider(),
      toolbar_input_select(
        "format",
        label = "Format",
        choices = c("Plain", "Markdown", "HTML"),
        icon = icon("code")
      )
    )
  ),
  verbatimTextOutput("output")
)

server <- function(input, output, session) {
  output$output <- renderText({
    req(input$message)
    paste("You said:", input$message)
  })

  observeEvent(input$attach, {
    showNotification("Attach clicked!", duration = 2)
  })

  observeEvent(input$emoji, {
    showNotification("Emoji clicked!", duration = 2)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=card_header]{card_header()}} for using toolbars in card headers/footers

Other toolbar components:
\code{\link[=toolbar_divider]{toolbar_divider()}},
\code{\link[=toolbar_input_button]{toolbar_input_button()}},
\code{\link[=toolbar_input_select]{toolbar_input_select()}}
}
\concept{toolbar components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{bs_theme_new}
\alias{bs_theme_new}
\alias{bs_theme_clear}
\alias{bs_theme_get}
\alias{bs_theme_set}
\alias{bs_theme_base_colors}
\alias{bs_theme_accent_colors}
\alias{bs_theme_fonts}
\alias{bs_theme_add_variables}
\alias{bs_theme_add}
\alias{bs_theme_get_variables}
\alias{bootstrap}
\alias{bootstrap_sass}
\alias{bs_add_declarations}
\title{Deprecated functions}
\usage{
bs_theme_new(...)

bs_theme_clear(...)

bs_theme_get(...)

bs_theme_set(...)

bs_theme_base_colors(...)

bs_theme_accent_colors(...)

bs_theme_fonts(...)

bs_theme_add_variables(...)

bs_theme_add(...)

bs_theme_get_variables(...)

bootstrap(theme = bs_theme_get(), ...)

bootstrap_sass(rules = list(), theme = bs_theme_get(), ...)

bs_add_declarations(theme, declarations)
}
\value{
a \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
These functions have been deprecated but remain for backwards compatibility.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preview.R
\name{bs_get_variables}
\alias{bs_get_variables}
\alias{bs_get_contrast}
\title{Retrieve Sass variable values from the current theme}
\usage{
bs_get_variables(theme, varnames)

bs_get_contrast(theme, varnames)
}
\arguments{
\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}

\item{varnames}{a character string referencing a Sass variable
in the current theme.}
}
\value{
a character string containing a CSS/Sass value.
If the variable(s) are not defined, their value is \code{NA}.
}
\description{
Useful for retriving a variable from the current theme and using
the value to inform another R function.
}
\examples{
vars <- c("body-bg", "body-color", "primary", "border-radius")
bs_get_variables(bs_theme(), varnames = vars)
bs_get_variables(bs_theme(bootswatch = "darkly"), varnames = vars)


bs_get_contrast(bs_theme(), c("primary", "dark", "light"))

library(htmltools)
div(
  class = "bg-primary",
  style = css(
    color = bs_get_contrast(bs_theme(), "primary")
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sidebar.R
\name{sidebar}
\alias{sidebar}
\alias{layout_sidebar}
\alias{toggle_sidebar}
\alias{sidebar_toggle}
\title{Sidebar layouts}
\usage{
sidebar(
  ...,
  width = 250,
  position = c("left", "right"),
  open = c("desktop", "open", "closed", "always"),
  id = NULL,
  title = NULL,
  bg = NULL,
  fg = NULL,
  class = NULL,
  max_height_mobile = NULL,
  gap = NULL,
  padding = NULL
)

layout_sidebar(
  ...,
  sidebar = NULL,
  fillable = TRUE,
  fill = TRUE,
  bg = NULL,
  fg = NULL,
  border = NULL,
  border_radius = NULL,
  border_color = NULL,
  padding = NULL,
  gap = NULL,
  height = NULL
)

toggle_sidebar(id, open = NULL, session = get_current_session())

sidebar_toggle(id, open = NULL, session = get_current_session())
}
\arguments{
\item{...}{Unnamed arguments can be any valid child of an \link[htmltools:builder]{htmltools tag} and named arguments become HTML attributes on
returned UI element. In the case of \code{layout_sidebar()}, these arguments are
passed to the main content tag (not the sidebar+main content container).}

\item{width}{A valid \link[htmltools:validateCssUnit]{CSS unit} used for the
width of the sidebar.}

\item{position}{Where the sidebar should appear relative to the main content.}

\item{open}{The initial state of the sidebar, choosing from the following
options:
\itemize{
\item \code{"desktop"}: The sidebar starts open on desktop screen, closed on mobile.
This is default sidebar behavior.
\item \code{"open"} or \code{TRUE}: The sidebar starts open.
\item \code{"closed"} or \code{FALSE}: The sidebar starts closed.
\item \code{"always"} or \code{NA}: The sidebar is always open and cannot be closed.
}

In \code{sidebar_toggle()}, \code{open} indicates the desired state of the sidebar,
where the default of \code{open = NULL} will cause the sidebar to be toggled
open if closed or vice versa. Note that \code{sidebar_toggle()} can only open or
close the sidebar, so it does not support the \code{"desktop"} and \code{"always"}
options.}

\item{id}{A character string. Required if wanting to re-actively read (or
update) the \code{collapsible} state in a Shiny app.}

\item{title}{A character title to be used as the sidebar title, which will be
wrapped in a \verb{<div>} element with class \code{sidebar-title}. You can also
provide a custom \code{\link[htmltools:builder]{htmltools::tag()}} for the title element, in which case
you'll likely want to give this element \code{class = "sidebar-title"}.}

\item{bg, fg}{A background or foreground color. If only one of either is
provided, an accessible contrasting color is provided for the opposite
color, e.g. setting \code{bg} chooses an appropriate \code{fg} color.}

\item{class}{CSS classes for the sidebar container element, in addition to
the fixed \code{.sidebar} class.}

\item{max_height_mobile}{The maximum height of the horizontal sidebar when
viewed on mobile devices. The default is \verb{250px} unless the sidebar is
included in a \code{\link[=layout_sidebar]{layout_sidebar()}} with a specified height, in which case
the default is to take up no more than 50\% of the layout container.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
vertical \code{gap} (i.e., spacing) between adjacent elements provided to \code{...}.}

\item{padding}{Padding within the sidebar itself. This can be a numeric
vector (which will be interpreted as pixels) or a character vector with
valid CSS lengths. \code{padding} may be one to four values. If one, then
that value will be used for all four sides. If two, then the first value
will be used for the top and bottom, while the second value will be used
for left and right. If three, then the first will be used for top, the
second will be left and right, and the third will be bottom. If four, then
the values will be interpreted as top, right, bottom, and left
respectively.}

\item{sidebar}{A \code{\link[=sidebar]{sidebar()}} object.}

\item{fillable}{Whether or not the \code{main} content area should be considered a
fillable (i.e., flexbox) container.}

\item{fill}{Whether or not to allow the layout container to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{border}{Whether or not to add a border.}

\item{border_radius}{Whether or not to add a border radius.}

\item{border_color}{The border color that is applied to the entire layout (if
\code{border = TRUE}) and the color of the border between the sidebar and the
main content area.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\description{
Create a collapsing sidebar layout by providing a \code{sidebar()}
object to the \code{sidebar} argument of:
\itemize{
\item \code{\link[=page_sidebar]{page_sidebar()}}
\itemize{
\item Creates a "page-level" sidebar.
}
\item \code{\link[=page_navbar]{page_navbar()}}
\itemize{
\item Creates a multi-page app with a "page-level" sidebar.
\item Creates a multi page/tab UI with a singular \code{sidebar()} (which is
shown on every page/tab).
}
\item \code{layout_sidebar()}
\itemize{
\item Creates a "floating" sidebar layout component which can be dropped
inside any \code{\link[=page]{page()}} and/or \code{\link[=card]{card()}} context.
}
\item \code{\link[=navset_card_tab]{navset_card_tab()}} and \code{\link[=navset_card_pill]{navset_card_pill()}}
\itemize{
\item Creates a multi-tab card with a sidebar inside of it.
}
}

See \href{https://rstudio.github.io/bslib/articles/sidebars.html}{this article}
to learn more.
}
\section{Functions}{
\itemize{
\item \code{toggle_sidebar()}: Toggle a \code{sidebar()} state during an active Shiny user
session.

\item \code{sidebar_toggle()}: An alias for \code{\link[=toggle_sidebar]{toggle_sidebar()}}.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-code-editor.R
\name{input_code_editor}
\alias{input_code_editor}
\alias{update_code_editor}
\title{Code editor input}
\usage{
input_code_editor(
  id,
  label = NULL,
  value = "",
  ...,
  language = "plain",
  height = "auto",
  width = "100\%",
  theme_light = "github-light",
  theme_dark = "github-dark",
  read_only = FALSE,
  line_numbers = NULL,
  word_wrap = NULL,
  tab_size = 2,
  indentation = c("space", "tab"),
  fill = TRUE
)

update_code_editor(
  id,
  ...,
  value = NULL,
  label = NULL,
  language = NULL,
  theme_light = NULL,
  theme_dark = NULL,
  read_only = NULL,
  line_numbers = NULL,
  word_wrap = NULL,
  tab_size = NULL,
  indentation = NULL,
  session = get_current_session()
)
}
\arguments{
\item{id}{Input ID. Access the current value with \verb{input$<id>}.}

\item{label}{Display label for the input. Default is \code{NULL} for no label.}

\item{value}{Code content. Default is an empty string.}

\item{...}{Named arguments, e.g. \code{class} and \code{style}, that will be added to
the outer container of the input.}

\item{language}{Programming language for syntax highlighting. Supported
languages include \code{"r"}, \code{"python"}, \code{"julia"}, \code{"sql"}, \code{"javascript"},
\code{"typescript"}, \code{"html"}, \code{"css"}, \code{"scss"}, \code{"sass"}, \code{"json"},
\code{"markdown"}, \code{"yaml"}, \code{"xml"}, \code{"toml"}, \code{"ini"}, \code{"bash"}, \code{"docker"},
\code{"latex"}, \code{"cpp"}, \code{"rust"}, \code{"diff"}, and \code{"plain"}. Default is
\code{"plain"}.}

\item{height}{CSS height of the editor. Default is \code{"300px"}.}

\item{width}{CSS width of the editor. Default is \code{"100\%"}.}

\item{theme_light, theme_dark}{Theme to use in light or dark mode. Defaults to
\code{"github-light"} and \code{"github-dark"}, respectively. See the Theme section
for more details.}

\item{read_only}{Whether the editor should be read-only. Default is \code{FALSE}.}

\item{line_numbers}{Whether to show line numbers. Default is \code{TRUE}, except
for markdown and plain text.}

\item{word_wrap}{Whether to wrap long lines, by default disabled when
\code{line_numbers} is \code{FALSE}.}

\item{tab_size}{Number of spaces per tab. Default is \code{2}.}

\item{indentation}{Type of indentation: \code{"space"} or \code{"tab"}. Default is
\code{"space"}.}

\item{fill}{Whether or not to allow the card to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{session}{a shiny session object (the default should almost always be
used).}
}
\value{
An HTML tag object that can be included in a Shiny UI.
}
\description{
Creates an interactive light-weight code editor input that can be used in
Shiny applications. The editor provides syntax highlighting, line numbers,
and other basic code editing features powered by Prism Code Editor. For a
complete example, run \code{shiny::runExample("code-editor", package = "bslib")}.

The editor value is not sent to R on every keystroke. Instead, updates are
reflected on the server when the user moves away from the editor or when they
press \code{Ctrl/Cmd} + \code{Enter}.

Note that this input is not designed for editing or rendering large files.
Displaying files with 1,000 lines or more may lead to performance issues.
}
\section{Keyboard shortcuts}{

The editor supports the following keyboard shortcuts:
\itemize{
\item \code{Ctrl/Cmd+Enter}: Submit the current code to R
\item \code{Ctrl/Cmd+Z}: Undo
\item \code{Ctrl/Cmd+Shift+Z}: Redo
\item \code{Tab}: Indent selection
\item \code{Shift+Tab}: Dedent selection
}
}

\section{Themes}{

The editor automatically switches between \code{theme_light} and \code{theme_dark}
when used with \code{\link[=input_dark_mode]{input_dark_mode()}}. Otherwise, the editor will use
\code{theme_light} by default.

A variety of themes are available for use. The full list of bundled themes
is: \code{"atom-one-dark"}, \code{"dracula"}, \code{"github-dark-dimmed"}, \code{"github-dark"}, \code{"github-light"}, \code{"night-owl-light"}, \code{"night-owl"}, \code{"prism-okaidia"}, \code{"prism-solarized-light"}, \code{"prism-tomorrow"}, \code{"prism-twilight"}, \code{"prism"}, \code{"vs-code-dark"}, \code{"vs-code-light"}.
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)

ui <- page_fluid(
  input_code_editor(
    "sql_query",
    value = "SELECT * FROM table",
    language = "sql"
  ),
  verbatimTextOutput("code_output")
)

server <- function(input, output, session) {
  output$code_output <- renderPrint({
    input$sql_query
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\seealso{
Other input controls: 
\code{\link{input_dark_mode}()},
\code{\link{input_switch}()}
}
\concept{input controls}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preview.R
\name{bs_theme_preview}
\alias{bs_theme_preview}
\title{Preview the currently set theme}
\usage{
bs_theme_preview(theme = bs_theme(), ..., with_themer = TRUE)
}
\arguments{
\item{theme}{a \code{\link[=bs_theme]{bs_theme()}} object.}

\item{...}{passed along to \code{\link[shiny:runApp]{shiny::runApp()}}.}

\item{with_themer}{whether or not to run the app with \code{\link[=run_with_themer]{run_with_themer()}}.}
}
\value{
nothing, this function is called for its side-effects (launching an
application).
}
\description{
Launches an example shiny app via \code{run_with_themer()} and
\code{bs_theme_dependencies()}. Useful for getting a quick preview of the current
theme setting as well as an interactive GUI for tweaking some of the main
theme settings.
}
\details{
The app that this launches is subject to change.
}
\examples{
theme <- bs_theme(bg = "#6c757d", fg = "white", primary = "orange")
if (interactive()) bs_theme_preview(theme)
}
\seealso{
\code{\link[=run_with_themer]{run_with_themer()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nav-update.R
\name{nav_prepend}
\alias{nav_prepend}
\alias{nav_append}
\title{Append or prepend nav item to a dropdown menu}
\usage{
nav_prepend(
  id,
  nav,
  menu_title,
  select = FALSE,
  session = getDefaultReactiveDomain()
)

nav_append(
  id,
  nav,
  menu_title,
  select = FALSE,
  session = getDefaultReactiveDomain()
)
}
\arguments{
\item{id}{a character string used to identify the nav container.}

\item{nav}{a \code{\link[=nav]{nav()}} item.}

\item{menu_title}{The title of a \code{\link[=nav_menu]{nav_menu()}}.}

\item{select}{Should \code{nav} be selected upon being inserted?}

\item{session}{a shiny session object (the default should almost always be used).}
}
\description{
Exported for use by \code{\link[shiny:insertTab]{shiny::prependTab()}}/\code{\link[shiny:insertTab]{shiny::appendTab()}}. These
functions have been superseded by \code{\link[=nav_insert]{nav_insert()}} (i.e.,
\code{\link[shiny:insertTab]{shiny::insertTab()}}), since it can do everything these functions do (i.e.,
add a \code{\link[=nav]{nav()}} to the start or end of a \code{\link[=nav_menu]{nav_menu()}}) and more (i.e., insert a
\code{\link[=nav]{nav()}} anywhere inside a nav container).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fill.R
\name{as_fill_carrier}
\alias{as_fill_carrier}
\alias{as_fillable_container}
\alias{as_fill_item}
\alias{remove_all_fill}
\alias{is_fill_carrier}
\alias{is_fillable_container}
\alias{is_fill_item}
\title{Test and/or coerce fill behavior}
\usage{
as_fill_carrier(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
)

as_fillable_container(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
)

as_fill_item(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
)

remove_all_fill(x)

is_fill_carrier(x)

is_fillable_container(x)

is_fill_item(x)
}
\arguments{
\item{x}{a \code{\link[htmltools:builder]{htmltools::tag()}}.}

\item{...}{currently ignored.}

\item{min_height, max_height}{Any valid \link[htmltools:validateCssUnit]{CSS unit}
(e.g., \code{150}).}

\item{gap}{Any valid \link[htmltools:validateCssUnit]{CSS unit}.}

\item{class}{A character vector of class names to add to the tag.}

\item{style}{A character vector of CSS properties to add to the tag.}

\item{css_selector}{A character string containing a CSS selector for
targeting particular (inner) tag(s) of interest. For more details on what
selector(s) are supported, see \code{\link[=tagAppendAttributes]{tagAppendAttributes()}}.}
}
\value{
\itemize{
\item For \code{as_fill()}, \code{as_fillable()}, and \code{as_fill_carrier()}: the \emph{tagified}
version \code{x}, with relevant tags modified to possess the relevant fill
properties.
\item For \code{is_fill()}, \code{is_fillable()}, and \code{is_fill_carrier()}: a logical vector,
with length matching the number of top-level tags that possess the relevant
fill properties.
}
}
\description{
Filling layouts in bslib are built on the foundation of fillable containers
and fill items (fill carriers are both fillable and fill). This is why most
bslib components (e.g., \code{\link[=card]{card()}}, \code{\link[=card_body]{card_body()}}, \code{\link[=layout_sidebar]{layout_sidebar()}}) possess
both \code{fillable} and \code{fill} arguments (to control their fill behavior).
However, sometimes it's useful to add, remove, and/or test fillable/fill
properties on arbitrary \code{\link[htmltools:builder]{htmltools::tag()}}, which these functions are
designed to do.
}
\details{
Although \code{as_fill()}, \code{as_fillable()}, and \code{as_fill_carrier()} can work with
non-tag objects that have a \link{as.tags} method (e.g., htmlwidgets), they return
the "tagified" version of that object.
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
library(shiny)
shinyApp(
  page_fillable(
    # without `as_fill_carrier()`, the plot won't fill the page because
    # `uiOutput()` is neither a fillable container nor a fill item by default.
    as_fill_carrier(uiOutput("ui"))
  ),
  function(input, output) {
    output$ui <- renderUI({
      div(
        class = "bg-info text-white",
        as_fill_item(),
        "A fill item"
      )
    })
  }
)
\dontshow{\}) # examplesIf}
}
\references{
\url{https://rstudio.github.io/bslib/articles/filling.html}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preview.R
\name{run_with_themer}
\alias{run_with_themer}
\alias{bs_themer}
\title{Theme customization UI}
\usage{
run_with_themer(appDir = getwd(), ..., gfonts = TRUE, gfonts_update = FALSE)

bs_themer(gfonts = TRUE, gfonts_update = FALSE)
}
\arguments{
\item{appDir}{The application to run. This can be a file or directory path,
or a \code{\link[shiny:shinyApp]{shiny::shinyApp()}} object. See \code{\link[shiny:runApp]{shiny::runApp()}} for details.}

\item{...}{Additional parameters to pass through to \code{\link[shiny:runApp]{shiny::runApp()}}.}

\item{gfonts}{whether or not to detect Google Fonts and wrap them in
\code{\link[=font_google]{font_google()}} (so that their font files are automatically imported).}

\item{gfonts_update}{whether or not to update the internal database of
Google Fonts.}
}
\value{
nothing. These functions are called for their side-effects.
}
\description{
A 'real-time' theme customization UI that you can use to easily make common
tweaks to Bootstrap variables and immediately see how they would affect your
app's appearance. There are two ways you can launch the theming UI. For most
Shiny apps, just use \code{run_with_themer()} in place of \code{\link[shiny:runApp]{shiny::runApp()}}; they
should take the same arguments and work the same way. Alternatively, you can
call the \code{bs_themer()} function from inside your server function (or in an R
Markdown app that is using \code{runtime: shiny}, you can call this from any code
chunk). Note that this function is only intended to be used for development!
}
\details{
To help you utilize the changes you see in the preview, this utility prints
\code{\link[=bs_theme]{bs_theme()}} code to the R console.
}
\section{Limitations}{

\itemize{
\item Doesn't work with Bootstrap 3.
\item Doesn't work with IE11.
\item Only works inside Shiny apps and \code{runtime: shiny} R Markdown documents.
\itemize{
\item Can't be used with static R Markdown documents.
\item Can be used to some extent with \code{runtime: shiny_prerendered}, but only UI
rendered through a \code{context="server"} may update in real-time.
}
\item Doesn't work with '3rd party' custom widgets that don't make use of
\code{\link[=bs_dependency_defer]{bs_dependency_defer()}} or \code{\link[=bs_current_theme]{bs_current_theme()}}.
}
}

\examples{
library(shiny)
library(bslib)

ui <- fluidPage(
  theme = bs_theme(bg = "black", fg = "white"),
  h1("Heading 1"),
  h2("Heading 2"),
  p(
    "Paragraph text;",
    tags$a(href = "https://www.rstudio.com", "a link")
  ),
  p(
    actionButton("cancel", "Cancel"),
    actionButton("continue", "Continue", class = "btn-primary")
  ),
  tabsetPanel(
    tabPanel("First tab",
      "The contents of the first tab"
    ),
    tabPanel("Second tab",
      "The contents of the second tab"
    )
  )
)

if (interactive()) {
  run_with_themer(shinyApp(ui, function(input, output) {}))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/navbar_options.R
\name{navbar_options}
\alias{navbar_options}
\title{Create a set of navbar options}
\usage{
navbar_options(
  ...,
  position = c("static-top", "fixed-top", "fixed-bottom"),
  bg = NULL,
  theme = c("auto", "light", "dark"),
  collapsible = TRUE,
  underline = TRUE
)
}
\arguments{
\item{...}{Additional attributes that will be passed directly to the navbar
container element.}

\item{position}{Determines whether the navbar should be displayed at the top
of the page with normal scrolling behavior (\code{"static-top"}), pinned at
the top (\code{"fixed-top"}), or pinned at the bottom
(\code{"fixed-bottom"}). Note that using \code{"fixed-top"} or
\code{"fixed-bottom"} will cause the navbar to overlay your body content,
unless you add padding, e.g.: \code{tags$style(type="text/css", "body
  {padding-top: 70px;}")}}

\item{bg}{a CSS color to use for the navbar's background color.}

\item{theme}{Either \code{"dark"} for a light text color (on a \strong{dark}
background) or \code{"light"} for a dark text color (on a \strong{light} background).
If \code{"auto"} (the default) and \code{bg} is provided, the best contrast to \code{bg}
is chosen.}

\item{collapsible}{\code{TRUE} to automatically collapse the navigation
elements into an expandable menu on mobile devices or narrow window widths.}

\item{underline}{Whether or not to add underline styling to page or navbar
links when active or focused.}
}
\value{
Returns a list of navbar options.
}
\description{
A \code{navbar_options()} object captures options specific to the appearance and
behavior of the navbar, independent from the content displayed on the page.
This helper should be used to create the list of options expected by
\code{navbar_options} in \code{\link[=page_navbar]{page_navbar()}} and \code{\link[=navset_bar]{navset_bar()}}.
}
\details{
\subsection{Navbar style with Bootstrap 5 and Bootswatch themes}{

In \pkg{bslib} v0.9.0, the default navbar colors for Bootswatch themes with
Bootstrap 5 changed. Prior to v0.9.0, bslib pre-selected navbar background
colors in light and dark mode; after v0.9.0 the default navbar colors are
less opinionated by default and follow light or dark mode (see
\code{\link[=input_dark_mode]{input_dark_mode()}}).

You can use \code{navbar_options()} to adjust the colors of the navbar when using
a Bootswatch preset theme with Bootstrap 5. For example, the \href{https://bootswatch.com/flatly/}{Bootswatch documentation for the Flatly theme} shows
4 navbar variations. Inspecting the source code for the first example reveals
the following markup:

\if{html}{\out{<div class="sourceCode html">}}\preformatted{<nav class="navbar navbar-expand-lg bg-primary" data-bs-theme="dark">
  <!-- all of the navbar html -->
</nav>
}\if{html}{\out{</div>}}

Note that this navbar uses the \code{bg-primary} class for a dark navy background.
The navbar's white text is controlled by the \code{data-bs-theme="dark"}
attribute, which is used by Bootstrap for light text on a \emph{dark} background.
In \pkg{bslib}, you can achieve this look with:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ui <- page_navbar(
  theme = bs_theme(5, "flatly"),
  navbar_options = navbar_options(class = "bg-primary", theme = "dark")
)
}\if{html}{\out{</div>}}

This particular combination of \code{class = "bg-primary"} and \code{theme = "dark"}
works well for most Bootswatch presets.

Another variation from the Flatly documentation features a navar with dark
text on a light background:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ui <- page_navbar(
  theme = bs_theme(5, "flatly"),
  navbar_options = navbar_options(class = "bg-light", theme = "light")
)
}\if{html}{\out{</div>}}

The above options set navbar foreground and background colors that are always
the same in both light and dark modes. To customize the navbar colors used in
light or dark mode, you can use the \verb{$navbar-light-bg} and \verb{$navbar-dark-bg}
Sass variables. When provided, bslib will automatically choose to use
light or dark text as the foreground color.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ui <- page_navbar(
  theme = bs_theme(
    5,
    preset = "flatly",
    navbar_light_bg = "#18BC9C", # flatly's success color (teal)
    navbar_dark_bg = "#2C3E50"   # flatly's primary color (navy)
  )
)
}\if{html}{\out{</div>}}

Finally, you can also use the \verb{$navbar-bg} Sass variable to set the navbar
background color for both light and dark modes:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ui <- page_navbar(
  theme = bs_theme(
    5,
    preset = "flatly",
    navbar_bg = "#E74C3C" # flatly's danger color (red)
  )
)
}\if{html}{\out{</div>}}
}
}
\section{Changelog}{


This function was introduced in \pkg{bslib} v0.9.0, replacing the \code{position},
\code{bg}, \code{inverse}, \code{collapsible} and \code{underline} arguments of \code{\link[=page_navbar]{page_navbar()}}
and \code{\link[=navset_bar]{navset_bar()}}. Those arguments are deprecated with a warning and will be
removed in a future version of \pkg{bslib}. Note that the deprecated
\code{inverse} argument of \code{\link[=page_navbar]{page_navbar()}} and \code{\link[=navset_bar]{navset_bar()}} was replaced with
the \code{theme} argument of \code{navbar_options()}.
}

\examples{
navbar_options(position = "static-top", bg = "#2e9f7d", underline = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version-default.R
\name{versions}
\alias{versions}
\alias{version_default}
\title{Available Bootstrap versions}
\usage{
versions()

version_default()
}
\value{
Returns a list of the Bootstrap versions available.
}
\description{
Available Bootstrap versions
}
\seealso{
Other Bootstrap theme utility functions: 
\code{\link{bootswatch_themes}()},
\code{\link{bs_get_variables}()},
\code{\link{builtin_themes}()},
\code{\link{theme_bootswatch}()},
\code{\link{theme_version}()}
}
\concept{Bootstrap theme utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nav-items.R
\name{nav-items}
\alias{nav-items}
\alias{nav_panel}
\alias{nav_panel_hidden}
\alias{nav_menu}
\alias{nav_item}
\alias{nav_spacer}
\title{Navigation items}
\usage{
nav_panel(title, ..., value = title, icon = NULL)

nav_panel_hidden(value, ..., icon = NULL)

nav_menu(title, ..., value = title, icon = NULL, align = c("left", "right"))

nav_item(...)

nav_spacer()
}
\arguments{
\item{title}{A title to display. Can be a character string or UI elements
(i.e., \link[htmltools:builder]{htmltools::tags}).}

\item{...}{Depends on the function:
\itemize{
\item For \code{nav_panel()} and \code{nav_panel_hidden()}: UI elements (i.e.,
\link[htmltools:builder]{htmltools::tags}) to display when the item is active.
\item For \code{nav_menu()}: a collection of nav items (e.g., \code{nav_panel()},
\code{nav_item()}).
\item For \code{nav_item()}: UI elements (i.e., \link[htmltools:builder]{htmltools::tags}) to place directly in
the navigation panel (e.g., search forms, links to external content, etc).
}}

\item{value}{A character string to assign to the nav item. This value may be
supplied to the relevant container's \code{selected} argument in order to show
particular nav item's content immediately on page load. This value is also
useful for programmatically updating the selected content via
\code{\link[=nav_select]{nav_select()}}, \code{\link[=nav_hide]{nav_hide()}}, etc (updating selected tabs this way is often
useful for showing/hiding panels of content via other UI controls like
\code{\link[shiny:radioButtons]{shiny::radioButtons()}} -- in this scenario, consider using
\code{\link[=nav_panel_hidden]{nav_panel_hidden()}} with \code{\link[=navset_hidden]{navset_hidden()}}).}

\item{icon}{Optional icon to appear next to the nav item's \code{title}.}

\item{align}{horizontal alignment of the dropdown menu relative to dropdown toggle.}
}
\value{
A nav item that may be passed to a nav container (e.g. \code{\link[=navset_tab]{navset_tab()}}).
}
\description{
Create nav item(s) for use inside nav containers (e.g., \code{\link[=navset_tab]{navset_tab()}},
\code{\link[=navset_bar]{navset_bar()}}, etc).
}
\section{Functions}{
\itemize{
\item \code{nav_panel()}: Content to display when the given item is selected.

\item \code{nav_panel_hidden()}: Create nav content for use inside \code{navset_hidden()} (for
creating custom navigation controls via \code{navs_select()}),

\item \code{nav_menu()}: Create a menu of nav items.

\item \code{nav_item()}: Place arbitrary content in the navigation panel (e.g., search
forms, links to external content, etc.)

\item \code{nav_spacer()}: Adding spacing between nav items.

}}
\seealso{
\link{navset} create the navigation container holding the nav panels.

\code{\link[=nav_menu]{nav_menu()}}, \code{\link[=nav_item]{nav_item()}}, \code{\link[=nav_spacer]{nav_spacer()}} create menus, items, or
space in the navset control area.

\code{\link[=nav_insert]{nav_insert()}}, \code{\link[=nav_remove]{nav_remove()}} programmatically add or remove nav
panels.

\code{\link[=nav_select]{nav_select()}}, \code{\link[=nav_show]{nav_show()}}, \code{\link[=nav_hide]{nav_hide()}} change the state of a
\code{\link[=nav_panel]{nav_panel()}} in a navset.

Other Panel container functions: 
\code{\link{nav_select}()},
\code{\link{navset}}
}
\concept{Panel container functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/navs-legacy.R, R/navs.R
\name{navset}
\alias{navset}
\alias{navset_tab}
\alias{navset_pill}
\alias{navset_underline}
\alias{navset_pill_list}
\alias{navset_hidden}
\alias{navset_bar}
\alias{navset_card_tab}
\alias{navset_card_pill}
\alias{navset_card_underline}
\title{Navigation containers}
\usage{
navset_tab(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navset_pill(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navset_underline(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navset_pill_list(
  ...,
  id = NULL,
  selected = NULL,
  header = NULL,
  footer = NULL,
  well = TRUE,
  fluid = TRUE,
  widths = c(4, 8)
)

navset_hidden(..., id = NULL, selected = NULL, header = NULL, footer = NULL)

navset_bar(
  ...,
  title = NULL,
  id = NULL,
  selected = NULL,
  sidebar = NULL,
  fillable = TRUE,
  gap = NULL,
  padding = NULL,
  header = NULL,
  footer = NULL,
  fluid = TRUE,
  navbar_options = NULL,
  position = deprecated(),
  bg = deprecated(),
  inverse = deprecated(),
  collapsible = deprecated()
)

navset_card_tab(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  sidebar = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  full_screen = FALSE,
  wrapper = card_body
)

navset_card_pill(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  sidebar = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  placement = c("above", "below"),
  full_screen = FALSE,
  wrapper = card_body
)

navset_card_underline(
  ...,
  id = NULL,
  selected = NULL,
  title = NULL,
  sidebar = NULL,
  header = NULL,
  footer = NULL,
  height = NULL,
  full_screen = FALSE,
  wrapper = card_body
)
}
\arguments{
\item{...}{a collection of \code{\link[=nav_panel]{nav_panel()}} items.}

\item{id}{a character string used for dynamically updating the container (see
\code{\link[=nav_select]{nav_select()}}).}

\item{selected}{a character string matching the \code{value} of a particular
\code{\link[=nav_panel]{nav_panel()}} item to selected by default.}

\item{header}{UI element(s) (\link[htmltools:builder]{htmltools::tags}) to display \emph{above} the nav
content. For \code{card}-based navsets, these elements are implicitly wrapped in
a \code{card_body()}. To control things like \code{padding}, \code{fill}, etc., wrap the
elements in an explicit \code{\link[=card_body]{card_body()}}.}

\item{footer}{UI element(s) (\link[htmltools:builder]{htmltools::tags}) to display \emph{below} the nav
content. For \code{card}-based navsets, these elements are implicitly wrapped in
a \code{card_body()}. To control things like \code{padding}, \code{fill}, etc., wrap the
elements in an explicit \code{\link[=card_body]{card_body()}}.}

\item{well}{\code{TRUE} to place a well (gray rounded rectangle) around the
navigation list.}

\item{fluid}{\code{TRUE} to use fluid layout; \code{FALSE} to use fixed
layout.}

\item{widths}{Column widths of the navigation list and tabset content areas
respectively.}

\item{title}{A (left-aligned) title to place in the card header/footer. If
provided, other nav items are automatically right aligned.}

\item{sidebar}{A \code{\link[=sidebar]{sidebar()}} component to display on every \code{\link[=nav_panel]{nav_panel()}}
page.}

\item{fillable}{Whether or not to allow \code{fill} items to grow/shrink to fit
the browser window. If \code{TRUE}, all \code{\link[=nav_panel]{nav_panel()}} pages are \code{fillable}. A
character vector, matching the \code{value} of \code{\link[=nav_panel]{nav_panel()}}s to be filled, may
also be provided. Note that, if a \code{sidebar} is provided, \code{fillable} makes
the main content portion fillable.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}.}

\item{padding}{Padding to use for the body. This can be a numeric vector
(which will be interpreted as pixels) or a character vector with valid CSS
lengths. The length can be between one and four. If one, then that value
will be used for all four sides. If two, then the first value will be used
for the top and bottom, while the second value will be used for left and
right. If three, then the first will be used for top, the second will be
left and right, and the third will be bottom. If four, then the values will
be interpreted as top, right, bottom, and left respectively.}

\item{navbar_options}{Options to control the appearance and behavior of the
navbar. Use \code{\link[=navbar_options]{navbar_options()}} to create the list of options.}

\item{position}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(position=)}} instead.}

\item{bg}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(bg=)}} instead.}

\item{inverse}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(inverse=)}} instead.}

\item{collapsible}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use
\code{\link[=navbar_options]{navbar_options = navbar_options(collapsible=)}} instead.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{full_screen}{If \code{TRUE}, an icon will appear when hovering over the card
body. Clicking the icon expands the card to fit viewport size.}

\item{wrapper}{A function (which returns a UI element) to call on unnamed
arguments in \code{...} which are not already card item(s) (like
\code{\link[=card_header]{card_header()}}, \code{\link[=card_body]{card_body()}}, etc.). Note that non-card items are grouped
together into one \code{wrapper} call (e.g. given \code{card("a", "b", card_body("c"), "d")}, \code{wrapper} would be called twice, once with \code{"a"} and
\code{"b"} and once with \code{"d"}).}

\item{placement}{placement of the nav items relative to the content.}
}
\description{
Render a collection of \code{\link[=nav_panel]{nav_panel()}} items into a container.
}
\section{Examples}{

\subsection{A basic example}{

This first example creates a simple tabbed navigation container with two tabs.
The tab name and the content of each tab are specified in the \code{nav_panel()} calls
and \code{navset_tab()} creates the tabbed navigation around these two tabs.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(htmltools)

navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content."))
)
}\if{html}{\out{</div>}}

\figure{navset-tab-basic.png}{Screenshot of a basic navset_tab() example.}

In the rest of the examples, we'll include links among the tabs (or pills) in the navigation controls.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{link_shiny <- tags$a(shiny::icon("github"), "Shiny", href = "https://github.com/rstudio/shiny", target = "_blank")
link_posit <- tags$a(shiny::icon("r-project"), "Posit", href = "https://posit.co", target = "_blank")
}\if{html}{\out{</div>}}
}

\subsection{\code{navset_tab()}}{

You can fully customize the controls in the navigation component.
In this example, we've added a direct link to the Shiny repository using \code{nav_item()}.
We've also included a dropdown menu using \code{nav_menu()}
containing an option to select a third tab panel
and another direct link to Posit's website.
Finally, we've separated the primary tabs on the left
from the direct link and dropdown menu on the right
using \code{nav_spacer()}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-tab.png}{Screenshot of a navset_tab() example.}
}

\subsection{\code{navset_pill()}}{

\code{navset_pill()} creates a navigation container that behaves exactly like \code{navset_tab()},
but the tab toggles are \emph{pills} or button-shaped.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_pill(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-pill.png}{Screenshot of a navset_pill() example.}
}

\subsection{\code{navset_underline()}}{

\code{navset_underline()} creates a navigation container that behaves exactly like \code{navset_tab()} and \code{navset_pill()}, but the active/focused navigation links are styled with an underline.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_underline(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-underline.png}{Screenshot of a navset_underline() example.}
}

\subsection{\code{navset_card_tab()}}{

The tabbed navigation container can also be used in a \code{card()} component
thanks to \code{navset_card_tab()}.
Learn more about this approach in the
\href{https://pkgs.rstudio.com/bslib/articles/cards.html}{article about Cards},
including how to add \href{https://pkgs.rstudio.com/bslib/articles/sidebars.html#multi-page-layout}{a shared sidebar}
to all tabs in the card
using the \code{sidebar} argument of \code{navset_card_tab()}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_card_tab(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-tab-card.png}{Screenshot of a navset_tab_card() example.}
}

\subsection{\code{navset_card_pill()}}{

Similar to \code{navset_pill()},
\code{navset_card_pill()} provides a pill-shaped variant to \code{navset_card_tab()}.
You can use the \code{placement} argument to position the navbar
\code{"above"} or \code{"below"} the card body.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_card_pill(
  placement = "above",
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-card-pill.png}{Screenshot of a navset_card_pill() example.}
}

\subsection{\code{navset_card_underline()}}{

\code{navset_card_underline()} provides a card-based variant of \code{navset_underline()}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_card_underline(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-card-underline.png}{Screenshot of a navset_card_underline() example.}
}

\subsection{\code{navset_pill_list()}}{

Furthermore, \code{navset_pill_list()} creates a vertical list of navigation controls
adjacent to, rather than on top of, the tab content panels.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{navset_pill_list(
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
}\if{html}{\out{</div>}}

\figure{navset-pill-list.png}{Screenshot of a navset_pill_list() example.}
}

\subsection{\code{page_navbar()}}{

Finally, \code{page_navbar()} provides full-page navigation container
similar to \code{navset_underline()} but where each \code{nav_panel()} is treated as a full page of content
and the navigation controls appear in a top-level navigation bar.
Note that the navbar background and underline styling can be controlled via \code{navbar_options}.

\code{page_navbar()} is complimented by \code{navset_bar()} which produces a similar layout intended to be used within an app.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{page_navbar(
  title = "My App",
  navbar_options = navbar_options(
    bg = "#0062cc",
    underline = TRUE
  ),
  nav_panel(title = "One", p("First tab content.")),
  nav_panel(title = "Two", p("Second tab content.")),
  nav_panel(title = "Three", p("Third tab content")),
  nav_spacer(),
  nav_menu(
    title = "Links",
    align = "right",
    nav_item(link_shiny),
    nav_item(link_posit)
  )
)
#> shiny devmode - Turning off caching of Sass -> CSS compilation. To turn caching on, call `options(sass.cache = TRUE)`
#> This message is displayed once every 8 hours.
}\if{html}{\out{</div>}}

\figure{page-navbar.png}{Screenshot of a page_navbar() example.}
}
}

\seealso{
\code{\link[=nav_panel]{nav_panel()}}, \code{\link[=nav_panel_hidden]{nav_panel_hidden()}} create panels of content.

\code{\link[=nav_menu]{nav_menu()}}, \code{\link[=nav_item]{nav_item()}}, \code{\link[=nav_spacer]{nav_spacer()}} create menus, items, or
space in the navset control area.

\code{\link[=nav_insert]{nav_insert()}}, \code{\link[=nav_remove]{nav_remove()}} programmatically add or remove nav
panels.

\code{\link[=nav_select]{nav_select()}}, \code{\link[=nav_show]{nav_show()}}, \code{\link[=nav_hide]{nav_hide()}} change the state of a
\code{\link[=nav_panel]{nav_panel()}} in a navset.

Other Panel container functions: 
\code{\link{nav-items}},
\code{\link{nav_select}()}
}
\concept{Panel container functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{save_html.bslib_fragment}
\alias{save_html.bslib_fragment}
\alias{save_html.bslib_page}
\title{Save a bslib page/fragment as HTML}
\usage{
\method{save_html}{bslib_fragment}(html, file, ...)

\method{save_html}{bslib_page}(html, file, ...)
}
\arguments{
\item{html}{a bslib page/fragment.}

\item{...}{passed along to an underlying \code{\link[=save_html]{save_html()}} method.}
}
\description{
Save a bslib page/fragment as HTML
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preset-bootswatch.R
\name{bootswatch_themes}
\alias{bootswatch_themes}
\title{Obtain a list of all available bootswatch themes.}
\usage{
bootswatch_themes(version = version_default(), full_path = FALSE)
}
\arguments{
\item{version}{The major version of Bootswatch.}

\item{full_path}{Whether to return a path to the installed theme.}
}
\value{
Returns a character vector of Bootswatch themes.
}
\description{
Obtain a list of all available bootswatch themes.
}
\seealso{
Other Bootstrap theme utility functions:
\code{\link[=bs_get_variables]{bs_get_variables()}},
\code{\link[=builtin_themes]{builtin_themes()}},
\code{\link[=theme_bootswatch]{theme_bootswatch()}},
\code{\link[=theme_version]{theme_version()}},
\code{\link[=versions]{versions()}}
}
\concept{Bootstrap theme utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toast.R
\name{show_toast}
\alias{show_toast}
\alias{hide_toast}
\title{Show or hide a toast notification}
\usage{
show_toast(toast, ..., session = shiny::getDefaultReactiveDomain())

hide_toast(id, ..., session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{toast}{A \code{\link[=toast]{toast()}}, or a string that will be automatically converted to
a toast with default settings.}

\item{...}{Reserved for future extensions (currently ignored).}

\item{session}{Shiny session object.}

\item{id}{String with the toast ID returned by \code{show_toast()} or a \code{toast}
object provided that the \code{id} was set when created/shown.}
}
\value{
\code{show_toast()} Invisibly returns the toast ID (string) that can be
used with \code{hide_toast()}.
}
\description{
Displays a toast notification in a Shiny application.
}
\section{Functions}{
\itemize{
\item \code{show_toast()}: Show a toast notification.

\item \code{hide_toast()}: Hide a toast notification by ID.

}}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)

ui <- page_fluid(
  actionButton("show_persistent", "Show Persistent Toast"),
  actionButton("hide_persistent", "Hide Toast")
)

server <- function(input, output, session) {
  toast_id <- reactiveVal(NULL)

  observeEvent(input$show_persistent, {
    id <- show_toast(
      toast(
        body = "This toast won't disappear automatically.",
        autohide = FALSE
      )
    )
    toast_id(id)
  })

  observeEvent(input$hide_persistent, {
    req(toast_id())
    hide_toast(toast_id())
    toast_id(NULL)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\seealso{
Other Toast components: 
\code{\link{toast}()}
}
\concept{Toast components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buttons.R
\name{input_task_button}
\alias{input_task_button}
\alias{update_task_button}
\title{Button for launching longer-running operations}
\usage{
input_task_button(
  id,
  label,
  ...,
  icon = NULL,
  label_busy = "Processing...",
  icon_busy = rlang::missing_arg(),
  type = "primary",
  auto_reset = TRUE
)

update_task_button(id, ..., state = NULL, session = get_current_session())
}
\arguments{
\item{id}{The \code{input} slot that will be used to access the value.}

\item{label}{The label of the button while it is in ready (clickable) state;
usually a string.}

\item{...}{In \code{input_task_button()}, named arguments become attributes to
include on the \verb{<button>} element, e.g. \code{class} or data attributes.
Unnamed arguments can provide additional states for the button, see the
"Custom states" section.

In \code{update_task_button()}, \code{...} are ignored and must be empty. The task
button only supports changing between pre-defined states.}

\item{icon}{An optional icon to display next to the label while the button is
in ready state. See \code{\link[fontawesome:fa_i]{fontawesome::fa_i()}}.}

\item{label_busy}{The label of the button while it is busy.}

\item{icon_busy}{The icon to display while the button is busy. By default,
\code{fontawesome::fa_i("refresh", class = "fa-spin", "aria-hidden" = "true")}
is used, which displays a spinning "chasing arrows" icon. You can create
spinning icons out of other Font Awesome icons by using the same
expression, but replacing \code{"refresh"} with a different icon name. See
\code{\link[fontawesome:fa_i]{fontawesome::fa_i()}}.}

\item{type}{One of the Bootstrap theme colors (\code{"primary"}, \code{"default"},
\code{"secondary"}, \code{"success"}, \code{"danger"}, \code{"warning"}, \code{"info"}, \code{"light"},
\code{"dark"}), or \code{NULL} to leave off the Bootstrap-specific button CSS classes
altogether.}

\item{auto_reset}{If \code{TRUE} (the default), automatically put the button back
in "ready" state after its click is handled by the server.}

\item{state}{If \code{"busy"}, put the button into busy/disabled state. If
\code{"ready"}, put the button into ready/enabled state.}

\item{session}{The \code{session} object; using the default is recommended.}
}
\description{
\code{input_task_button} is a button that can be used in conjuction with
\code{\link[shiny:bindEvent]{shiny::bindEvent()}} (or the older \code{\link[shiny:observeEvent]{shiny::eventReactive()}} and
\code{\link[shiny:observeEvent]{shiny::observeEvent()}} functions) to trigger actions or recomputation.

It is similar to \code{\link[shiny:actionButton]{shiny::actionButton()}}, except it prevents the user from
clicking when its operation is already in progress.

Upon click, it automatically displays a customizable progress message and
disables itself; and after the server has dealt with whatever reactivity is
triggered from the click, the button automatically reverts to its original
appearance and re-enables itself.
}
\section{Manual button reset}{

In some advanced use cases, it may be necessary to keep a task button in its
busy state even after the normal reactive processing has completed. Calling
\code{update_task_button(id, state = "busy")} from the server will opt out of any
currently pending reset for a specific task button. After doing so, the
button can be re-enabled by calling \code{update_task_button(id, state = "ready")}
after each click's work is complete.

You can also pass an explicit \code{auto_reset = FALSE} to \code{input_task_button()},
which means that button will \emph{never} be automatically re-enabled and will
require \code{update_task_button(id, state = "ready")} to be called each time.

Note that, as a general rule, Shiny's \code{update} family of functions do not
take effect at the instant that they are called, but are held until the end
of the current reactive cycle. So if you have many different reactive
calculations and outputs, you don't have to be too careful about when you
call \code{update_task_button(id, state = "ready")}, as the button on the client
will not actually re-enable until the same moment that all of the updated
outputs simultaneously sent to the client.
}

\section{Custom states}{

The task button is designed to automatically switch between two states: the
"ready" state, where the button is clickable and displays the \code{label} and
\code{icon}; and the "busy" state, where the button is disabled and displays
\code{label_busy} and \code{icon_busy}.

In advanced use cases, you can include additional states by adding an
\code{\link[htmltools:builder]{htmltools::div()}} with a \code{slot} attribute naming the state and the icon and
label as the first and second children, respectively.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{input_task_button(
  label = "Ring home",
  icon = fontawesome::fa_i("phone"),
  div(slot = "ringing", fontawesome::fa_i("bell"), "Ringing..."),
  div(
    slot = "voice-mail",
    fontawesome::fa_i("voicemail"),
    "Leaving a message..."
  )
)
}\if{html}{\out{</div>}}

You can move between these states by calling \code{update_task_button()} and
passing the slot name to the \code{state} argument, e.g. \code{state="ringing"}. See
the section above on manual button resetting, which you will likely need to
use in conjunction with custom states.
}

\section{Server value}{

An integer of class \code{"shinyActionButtonValue"}. This class differs from
ordinary integers in that a value of 0 is considered "falsy". This implies
two things:
\itemize{
\item Event handlers (e.g., \code{\link[shiny:observeEvent]{shiny::observeEvent()}}, \code{\link[shiny:observeEvent]{shiny::eventReactive()}})
won't execute on initial load.
\item Input validation (e.g., \code{\link[shiny:req]{shiny::req()}}, \code{\link[shiny:validate]{shiny::need()}}) will fail on
initial load.
}
}

\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)

ui <- page_sidebar(
  sidebar = sidebar(
    open = "always",
    input_task_button("resample", "Resample"),
  ),
  verbatimTextOutput("summary")
)

server <- function(input, output, session) {
  sample <- eventReactive(input$resample, ignoreNULL=FALSE, {
    Sys.sleep(2)  # Make this artificially slow
    rnorm(100)
  })

  output$summary <- renderPrint({
    summary(sample())
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=bind_task_button]{bind_task_button()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card.R
\name{card}
\alias{card}
\title{A Bootstrap card component}
\usage{
card(
  ...,
  full_screen = FALSE,
  height = NULL,
  max_height = NULL,
  min_height = NULL,
  fill = TRUE,
  class = NULL,
  wrapper = card_body
)
}
\arguments{
\item{...}{Unnamed arguments can be any valid child of an \link[htmltools:builder]{htmltools tag} (which includes card items such as \code{\link[=card_body]{card_body()}}.
Named arguments become HTML attributes on returned UI element.}

\item{full_screen}{If \code{TRUE}, an icon will appear when hovering over the card
body. Clicking the icon expands the card to fit viewport size.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{max_height, min_height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{max_height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{max_height} in \code{\link[=card_body]{card_body()}}).}

\item{fill}{Whether or not to allow the card to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{class}{Additional CSS classes for the returned UI element.}

\item{wrapper}{A function (which returns a UI element) to call on unnamed
arguments in \code{...} which are not already card item(s) (like
\code{\link[=card_header]{card_header()}}, \code{\link[=card_body]{card_body()}}, etc.). Note that non-card items are grouped
together into one \code{wrapper} call (e.g. given \code{card("a", "b", card_body("c"), "d")}, \code{wrapper} would be called twice, once with \code{"a"} and
\code{"b"} and once with \code{"d"}).}
}
\value{
A \code{\link[htmltools:builder]{htmltools::div()}} tag.
}
\description{
A general purpose container for grouping related UI elements together with a
border and optional padding. To learn more about \code{\link[=card]{card()}}s, see \href{https://rstudio.github.io/bslib/articles/cards.html}{this article}.
}
\examples{

library(htmltools)

if (interactive()) {
  card(
    full_screen = TRUE,
    card_header(
      "This is the header"
    ),
    card_body(
      p("This is the body."),
      p("This is still the body.")
    ),
    card_footer(
      "This is the footer"
    )
  )
}

}
\seealso{
\code{\link[=card_body]{card_body()}} for putting stuff inside the card.

\code{\link[=navset_card_tab]{navset_card_tab()}} for cards with multiple tabs.

\code{\link[=layout_column_wrap]{layout_column_wrap()}} for laying out multiple cards (or multiple
columns inside a card).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme.R
\name{bs_theme}
\alias{bs_theme}
\alias{bs_theme_update}
\alias{is_bs_theme}
\title{Create a Bootstrap theme}
\usage{
bs_theme(
  version = version_default(),
  preset = NULL,
  ...,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  font_scale = NULL,
  bootswatch = NULL
)

bs_theme_update(
  theme,
  ...,
  preset = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  font_scale = NULL,
  bootswatch = NULL
)

is_bs_theme(x)
}
\arguments{
\item{version}{The major version of Bootstrap to use (see \code{\link[=versions]{versions()}}
for possible values). Defaults to the currently recommended version
for new projects (currently Bootstrap 5).}

\item{preset}{The name of a theme preset, either a built-in theme provided by
bslib or a Bootswatch theme (see \code{\link[=builtin_themes]{builtin_themes()}} and
\code{\link[=bootswatch_themes]{bootswatch_themes()}} for possible values). This argument takes precedence
over the \code{bootswatch} argument and only one \code{preset} or \code{bootswatch} can be
provided. When no \code{bootswatch} theme is specified, and \code{version} is 5 or
higher, \code{preset} defaults to \code{"shiny"}. To remove the \code{"shiny"} preset,
provide a value of \code{"bootstrap"} (this value will also work in
\code{bs_theme_update()} to remove a \code{preset} or \code{bootswatch} theme).}

\item{...}{arguments passed along to \code{\link[=bs_add_variables]{bs_add_variables()}}.}

\item{bg}{A color string for the background.}

\item{fg}{A color string for the foreground.}

\item{primary}{A color to be used for hyperlinks, to indicate primary/default
actions, and to show active selection state in some Bootstrap components.
Generally a bold, saturated color that contrasts with the theme's base
colors.}

\item{secondary}{A color for components and messages that don't need to stand
out. (Not supported in Bootstrap 3.)}

\item{success}{A color for messages that indicate an operation has succeeded.
Typically green.}

\item{info}{A color for messages that are informative but not critical.
Typically a shade of blue-green.}

\item{warning}{A color for warning messages. Typically yellow.}

\item{danger}{A color for errors. Typically red.}

\item{base_font}{The default typeface.}

\item{code_font}{The typeface to be used for code. Be sure this is monospace!}

\item{heading_font}{The typeface to be used for heading elements.}

\item{font_scale}{A scalar multiplier to apply to the base font size. For
example, a value of \code{1.5} scales font sizes to 150\% and a value of \code{0.8}
scales to 80\%. Must be a positive number.}

\item{bootswatch}{The name of a bootswatch theme (see \code{\link[=bootswatch_themes]{bootswatch_themes()}}
for possible values). When provided to \code{bs_theme_update()}, any previous
Bootswatch theme is first removed before the new one is applied (use
\code{bootswatch = "bootstrap"} to effectively remove the Bootswatch theme).}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{x}{an object.}
}
\value{
Returns a \code{\link[sass:sass_layer]{sass::sass_bundle()}} (list-like) object.
}
\description{
Creates a Bootstrap theme object, where you can:
\itemize{
\item Choose a (major) Bootstrap \code{version}.
\item Choose a \href{https://bootswatch.com}{Bootswatch theme} (optional).
\item Customize main colors and fonts via explicitly named arguments (e.g.,
\code{bg}, \code{fg}, \code{primary}, etc).
\item Customize other, lower-level, Bootstrap Sass variable defaults via \code{...}.
}

To learn more about how to implement custom themes, as well as how to use
them inside Shiny and R Markdown, \href{https://rstudio.github.io/bslib/articles/theming.html}{see here}.
}
\section{Colors}{


Colors may be provided in any format that \code{\link[htmltools:parseCssColors]{htmltools::parseCssColors()}} can
understand. To control the vast majority of the ('grayscale') color
defaults, specify both the \code{fg} (foreground) and \code{bg} (background) colors.
The \code{primary} and \code{secondary} theme colors are also useful for accenting the
main grayscale colors in things like hyperlinks, tabset panels, and buttons.
}

\section{Fonts}{


Use \code{base_font}, \code{code_font}, and \code{heading_font} to control the main
typefaces. These arguments set new defaults for the relevant \code{font-family}
CSS properties, but don't necessarily import the relevant font files. To
both set CSS properties \emph{and} import font files, consider using the various
\code{\link[=font_face]{font_face()}} helpers.

Each \verb{*_font} argument may be a single font or a \code{font_collection()}.
A font can be created with \code{\link[=font_google]{font_google()}}, \code{\link[=font_link]{font_link()}}, or \code{\link[=font_face]{font_face()}},
or it can be a character vector of font names in the following format:
\itemize{
\item A single unquoted name (e.g., \code{"Source Sans Pro"}).
\item A single quoted name (e.g., \code{"'Source Sans Pro'"}).
\item A comma-separated list of names w/ individual names quoted as necessary.
(e.g. \code{c("Open Sans", "'Source Sans Pro'", "'Helvetica Neue', Helvetica, sans-serif")})
}

\code{font_google()} sets \code{local = TRUE} by default, which ensures that the font
files are downloaded from \href{https://fonts.google.com/}{Google Fonts} when
your document or app is rendered. This guarantees that the client has access
to the font family, making it relatively safe to specify just one font
family:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{bs_theme(base_font = font_google("Pacifico", local = TRUE))
}\if{html}{\out{</div>}}

That said, we recommend you specify multiple "fallback" font families,
especially when relying on remote and/or system fonts being available.
Fallback fonts are useful not only for handling missing fonts, but also
ensure that your users don't experience a Flash of Invisible Text (FOIT)
which can be quite noticeable with remote web fonts on a slow internet
connection.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{bs_theme(base_font = font_collection(font_google("Pacifico", local = FALSE), "Roboto", "sans-serif"))
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}

theme <- bs_theme(
  # Controls the default grayscale palette
  bg = "#202123", fg = "#B8BCC2",
  # Controls the accent (e.g., hyperlink, button, etc) colors
  primary = "#EA80FC", secondary = "#48DAC6",
  base_font = c("Grandstander", "sans-serif"),
  code_font = c("Courier", "monospace"),
  heading_font = "'Helvetica Neue', Helvetica, sans-serif",
  # Can also add lower-level customization
  "input-border-color" = "#EA80FC"
)

bs_theme_preview(theme)

# Lower-level bs_add_*() functions allow you to work more
# directly with the underlying Sass code
theme <- bs_add_variables(theme, "my-class-color" = "red")
theme <- bs_add_rules(theme, ".my-class { color: $my-class-color }")
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/theming/index.html}{Get Started: Theming}
introduces theming with bslib in Shiny apps and R Markdown documents.
\item \href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html}{Theming: Bootstrap 5 variables}
provides a searchable reference of all theming variables available in
Bootstrap 5.
\item \href{https://rstudio.github.io/bslib/articles/custom-components/index.html}{Theming: Custom components}
gives a tutorial on creating a dynamically themable custom component.
\item bslib's theming capabilities are powered by
\href{https://rstudio.github.io/sass/}{the {sass} package}.
\item \href{https://rstudio.github.io/bslib/articles/utility-classes/index.html}{Bootstrap's utility classes}
can be helpful when you want to change the appearance of an element
without writing CSS or customizing your \code{bs_theme()}.
}
}
\seealso{
Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme_preview}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toast.R
\name{toast}
\alias{toast}
\alias{toast_header}
\title{Toast notifications}
\usage{
toast(
  ...,
  header = NULL,
  icon = NULL,
  id = NULL,
  type = NULL,
  duration_s = 5,
  position = "top-right",
  closable = TRUE
)

toast_header(title, ..., icon = NULL, status = NULL)
}
\arguments{
\item{...}{Body content of the toast. Can be a string, or any HTML elements.
Named arguments will be treated as HTML attributes for the toast container.}

\item{header}{Optional header content. Can be a string, or the result of
\code{\link[=toast_header]{toast_header()}}. If provided, creates a \code{.toast-header} with close button
(if \code{closable = TRUE}).}

\item{icon}{Optional icon element, for example from \code{\link[shiny:icon]{shiny::icon()}},
\code{\link[bsicons:bs_icon]{bsicons::bs_icon()}} or \code{\link[fontawesome:fa]{fontawesome::fa()}}.}

\item{id}{Optional unique identifier for the toast. If \code{NULL}, an ID will be
automatically generated when the toast is shown via \code{\link[=show_toast]{show_toast()}}.
Providing a stable ID allows you to hide the toast later. If a toast with
\code{id} is already visible, that toast is automatically hidden before showing
the new toast with the same \code{id} so that only one toast with a given ID is
shown at once.}

\item{type}{Optional semantic type. One of \code{NULL}, \code{"primary"},
\code{"secondary"}, \code{"success"}, \code{"info"}, \code{"warning"}, \code{"danger"}, \code{"light"},
or \code{"dark"}. Applies appropriate Bootstrap background utility classes
(\verb{text-bg-*}).}

\item{duration_s}{Numeric. Number of seconds after which the toast should
automatically hide. Use \code{0}, or \code{NA} to disable auto-hiding (toast will
remain visible until manually dismissed). Default is \code{5} (5 seconds).}

\item{position}{String or character vector specifying where to position the
toast container. Can be provided in several formats:
\itemize{
\item Kebab-case: \code{"top-left"}, \code{"bottom-right"}, etc.
\item Space-separated: \code{"top left"}, \code{"bottom right"}, etc.
\item Character vector: \code{c("top", "left")}, \code{c("bottom", "right")}, etc.
\item Any order: \code{"left top"} is equivalent to \code{"top left"}
}

Valid vertical positions are \code{"top"}, \code{"middle"}, or \code{"bottom"}. Valid
horizontal positions are \code{"left"}, \code{"center"}, or \code{"right"}. Input is
case-insensitive. Default is \code{"bottom-right"}.}

\item{closable}{Logical. Whether to include a close button. Defaults to
\code{TRUE}. When both \code{duration_s = NA} (or \code{0} or \code{NULL}) and \code{closable = FALSE}, the toast will remain visible until manually hidden via
\code{\link[=hide_toast]{hide_toast()}}. This is useful when the toast contains interactive Shiny UI
elements and you want to manage the toast display programmatically.}

\item{title}{Header text (required).}

\item{status}{Optional status text that appears as small, muted text on the
right side of the header.}
}
\value{
A \code{bslib_toast} object that can be passed to \code{\link[=show_toast]{show_toast()}}.

For \code{toast_header()}: a toast header object that can be used with the
\code{header} argument of \code{toast()}.
}
\description{
Toast notifications are lightweight, temporary messages designed to mimic
push notifications from mobile and desktop operating systems. They are built
on \href{https://getbootstrap.com/docs/5.3/components/toasts/}{Bootstrap 5's toast component}.

bslib includes a complete example of toasts and their many configuration
options:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{shiny::runExample("toast", package = "bslib")
}\if{html}{\out{</div>}}
}
\section{Functions}{
\itemize{
\item \code{toast()}: Create a toast element.

\item \code{toast_header()}: Create a structured toast header with optional icon and
status indicator. Returns a data structure that can be passed to the
\code{header} argument of \code{toast()}.

}}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)

ui <- page_fluid(
  actionButton("show_simple", "Simple Toast"),
  actionButton("show_header", "Toast with Header")
)

server <- function(input, output, session) {
  observeEvent(input$show_simple, {
    show_toast(
      toast(
        "Operation completed successfully!",
        header = "Success",
        type = "success"
      )
    )
  })

  observeEvent(input$show_header, {
    show_toast(
      toast(
        "Your settings have been saved.",
        header = toast_header(
          title = "Settings Updated",
          status = "just now"
        ),
        type = "success"
      )
    )
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=show_toast]{show_toast()}} to display a toast, \code{\link[=hide_toast]{hide_toast()}} to dismiss a
toast, and \code{\link[=toast_header]{toast_header()}} to create structured headers.

Other Toast components: 
\code{\link{show_toast}()}
}
\concept{Toast components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layout.R
\name{layout_columns}
\alias{layout_columns}
\title{Responsive column-based grid layouts}
\usage{
layout_columns(
  ...,
  col_widths = NA,
  row_heights = NULL,
  fill = TRUE,
  fillable = TRUE,
  gap = NULL,
  class = NULL,
  height = NULL
)
}
\arguments{
\item{...}{Unnamed arguments should be UI elements (e.g., \code{\link[=card]{card()}})
Named arguments become attributes on the containing \link[htmltools:builder]{htmltools::tag} element.}

\item{col_widths}{One of the following:
\itemize{
\item \code{NA} (the default): Automatically determines a sensible number of columns
based on the number of children.
\item A numeric vector of integers between 1 and 12, where each element
represents the number of columns for the relevant UI element. Elements
that happen to go beyond 12 columns wrap onto the next row. For example,
\code{col_widths = c(4, 8, 12)} allocates 4 columns to the first element, 8
columns to the second element, and 12 columns to the third element (which
wraps to the next row). Negative values are also allowed, and are treated
as empty columns. For example, \code{col_widths = c(-2, 8, -2)} would allocate
8 columns to an element (with 2 empty columns on either side).
\item A \code{\link[=breakpoints]{breakpoints()}} object, where each breakpoint may be either of the
above.
}}

\item{row_heights}{One of the following:
\itemize{
\item A numeric vector, where each value represents the
\href{https://css-tricks.com/introduction-fr-css-unit/}{fractional unit} (\code{fr})
height of the relevant row. If there are more rows than values provided,
the pattern will repeat. For example, \code{row_heights = c(1, 2)} allows even
rows to take up twice as much space as odd rows.
\item A list of numeric and \link[htmltools:validateCssUnit]{CSS length units},
where each value represents the height of the relevant row. If more rows
are needed than values provided, the pattern will repeat. For example,
\code{row_heights = list("auto", 1)} allows the height of odd rows to be
driven my it's contents and even rows to be
\href{https://css-tricks.com/introduction-fr-css-unit/}{\verb{1fr}}.
\item A character vector/string of \link[htmltools:validateCssUnit]{CSS length units}.
In this case, the value is supplied directly to \code{grid-auto-rows}.
\item A \code{\link[=breakpoints]{breakpoints()}} object, where each breakpoint may be either of the above.
}}

\item{fill}{Whether or not to allow the layout to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{fillable}{Whether or not each element is wrapped in a fillable container.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}. This argument is only applicable when \code{fillable = TRUE}}

\item{class}{Additional CSS classes for the returned UI element.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}
}
\description{
Create responsive, column-based grid layouts, based on a 12-column grid.
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}


x <- card("A simple card")

page_fillable(
  layout_columns(x, x, x, x)
)

page_fillable(
  layout_columns(
    col_widths = c(6, 6, 12),
    x, x, x
  )
)

page_fillable(
  layout_columns(
    col_widths = c(6, 6, -2, 8),
    row_heights = c(1, 3),
    x, x, x
  )
)

page_fillable(
  fillable_mobile = TRUE,
  layout_columns(
    col_widths = breakpoints(
      sm = c(12, 12, 12),
      md = c(6, 6, 12),
      lg = c(4, 4, 4)
    ),
    x, x, x
  )
)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=breakpoints]{breakpoints()}} for more information on breakpoints.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preset-builtin.R
\name{builtin_themes}
\alias{builtin_themes}
\title{Obtain a list of all available built-in \pkg{bslib} themes.}
\usage{
builtin_themes(version = version_default(), full_path = FALSE)
}
\arguments{
\item{version}{the major version of Bootstrap.}

\item{full_path}{whether to return a path to the installed theme.}
}
\value{
Returns a character vector of built-in themes provided by
\pkg{bslib}.
}
\description{
Obtain a list of all available built-in \pkg{bslib} themes.
}
\seealso{
Other Bootstrap theme utility functions: 
\code{\link{bootswatch_themes}()},
\code{\link{bs_get_variables}()},
\code{\link{theme_bootswatch}()},
\code{\link{theme_version}()},
\code{\link{versions}()}
}
\concept{Bootstrap theme utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.bslib_fragment}
\alias{print.bslib_fragment}
\alias{print.bslib_page}
\title{Print a bslib page/fragment}
\usage{
\method{print}{bslib_fragment}(x, ...)

\method{print}{bslib_page}(x, ...)
}
\arguments{
\item{x}{a bslib page/fragment.}

\item{...}{passed along to an underlying print method.}
}
\description{
Print a bslib page/fragment
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{deprecated}
\alias{deprecated}
\alias{bs_theme_new}
\alias{bs_theme_clear}
\alias{bs_theme_get}
\alias{bs_theme_set}
\alias{bs_theme_base_colors}
\alias{bs_theme_accent_colors}
\alias{bs_theme_fonts}
\alias{bs_theme_add_variables}
\alias{bs_theme_add}
\alias{bs_theme_get_variables}
\alias{bootstrap}
\alias{bootstrap_sass}
\alias{bs_add_declarations}
\alias{card_body_fill}
\alias{page_fill}
\alias{nav}
\alias{nav_content}
\alias{navs_tab}
\alias{navs_pill}
\alias{navs_pill_list}
\alias{navs_hidden}
\alias{navs_bar}
\alias{navs_tab_card}
\alias{navs_pill_card}
\title{Deprecated functions}
\usage{
bs_theme_new(...)

bs_theme_clear(...)

bs_theme_get(...)

bs_theme_set(...)

bs_theme_base_colors(...)

bs_theme_accent_colors(...)

bs_theme_fonts(...)

bs_theme_add_variables(...)

bs_theme_add(...)

bs_theme_get_variables(...)

bootstrap(theme = bs_theme_get(), ...)

bootstrap_sass(rules = list(), theme = bs_theme_get(), ...)

bs_add_declarations(theme, declarations)

card_body_fill(...)

page_fill(...)

nav(...)

nav_content(...)

navs_tab(...)

navs_pill(...)

navs_pill_list(...)

navs_hidden(...)

navs_bar(...)

navs_tab_card(...)

navs_pill_card(...)
}
\value{
a \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
These functions have been deprecated but remain for backwards compatibility.
}
\section{Navigation Containers}{


Several functions for navigation containers were renamed in version 0.5.0:
\itemize{
\item \code{nav()} was renamed \code{\link[=nav_panel]{nav_panel()}}
\item \code{nav_content()} was renamed \code{\link[=nav_panel_hidden]{nav_panel_hidden()}}
\item \code{navs_tab()} was renamed \code{\link[=navset_tab]{navset_tab()}}
\item \code{navs_pill()} was renamed \code{\link[=navset_pill]{navset_pill()}}
\item \code{navs_pill_list()} was renamed \code{\link[=navset_pill_list]{navset_pill_list()}}
\item \code{navs_hidden()} was renamed \code{\link[=navset_hidden]{navset_hidden()}}
\item \code{navs_bar()} was renamed \code{\link[=navset_bar]{navset_bar()}}
\item \code{navs_tab_card()} was renamed \code{\link[=navset_card_tab]{navset_card_tab()}}
\item \code{navs_pill_card()} was renamed \code{\link[=navset_card_pill]{navset_card_pill()}}
}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buttons.R
\name{bind_task_button}
\alias{bind_task_button}
\alias{bind_task_button.default}
\alias{bind_task_button.ExtendedTask}
\title{Bind \code{input_task_button} to \code{ExtendedTask}}
\usage{
bind_task_button(target, task_button_id, ...)

\method{bind_task_button}{default}(target, task_button_id, ...)

\method{bind_task_button}{ExtendedTask}(target, task_button_id, ..., session = get_current_session())
}
\arguments{
\item{target}{The target object (i.e. \code{ExtendedTask}).}

\item{task_button_id}{A string matching the \code{id} argument passed to the
corresponding \code{\link[=input_task_button]{input_task_button()}} call.}

\item{...}{Further arguments passed to other methods.}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\value{
The \code{target} object that was passed in.
}
\description{
Sets up a \link[shiny:ExtendedTask]{shiny::ExtendedTask} to relay its state to an existing
\code{\link[=input_task_button]{input_task_button()}}, so the task button stays in its "busy" state for as
long as the extended task is running.

Note that \code{bind_task_button} does \emph{not} automatically cause button presses to
invoke the extended task; you still need to use \code{\link[shiny:bindEvent]{shiny::bindEvent()}} (or
\code{\link[shiny:observeEvent]{shiny::observeEvent()}}) to cause the button press to trigger an invocation,
as in the example below.

\code{bind_task_button} cannot be used to bind one task button to multiple
\code{ExtendedTask} objects; if you attempt to do so, any bound \code{ExtendedTask}
that completes will cause the button to return to "ready" state.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}

library(shiny)
library(bslib)
library(future)
plan(multisession)

ui <- page_sidebar(
  sidebar = sidebar(
    input_task_button("recalc", "Recalculate")
  ),
  textOutput("outval")
)

server <- function(input, output) {
  rand_task <- ExtendedTask$new(function() {
    future({
      # Slow operation goes here
      Sys.sleep(2)
      runif(1)
    }, seed = TRUE)
  })

  # Make button state reflect task.
  # If using R >=4.1, you can do this instead:
  # rand_task <- ExtendedTask$new(...) |> bind_task_button("recalc")
  bind_task_button(rand_task, "recalc")

  observeEvent(input$recalc, {
    rand_task$invoke()
  })

  output$outval <- renderText({
    rand_task$result()
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layout.R
\name{layout_column_wrap}
\alias{layout_column_wrap}
\title{Column-first uniform grid layouts}
\usage{
layout_column_wrap(
  ...,
  width = "200px",
  fixed_width = FALSE,
  heights_equal = c("all", "row"),
  fill = TRUE,
  fillable = TRUE,
  height = NULL,
  height_mobile = NULL,
  gap = NULL,
  class = NULL
)
}
\arguments{
\item{...}{Unnamed arguments should be UI elements (e.g., \code{\link[=card]{card()}}). Named
arguments become attributes on the containing \link[htmltools:builder]{htmltools::tag} element.}

\item{width}{The desired width of each card, which can be any of the
following:
\itemize{
\item A (unit-less) number between 0 and 1.
\itemize{
\item This should be specified as \code{1/num}, where \code{num} represents the number
of desired columns.
}
\item A \link[htmltools:validateCssUnit]{CSS length unit}
\itemize{
\item Either the minimum (when \code{fixed_width=FALSE}) or fixed width
(\code{fixed_width=TRUE}).
}
\item \code{NULL}
\itemize{
\item Allows power users to set the \code{grid-template-columns} CSS property
manually, either via a \code{style} attribute or a CSS stylesheet.
}
}}

\item{fixed_width}{When \code{width} is greater than 1 or is a CSS length unit,
e.g. \code{"200px"}, \code{fixed_width} indicates whether that \code{width} value
represents the absolute size of each column (\code{fixed_width=TRUE}) or the
minimum size of a column (\code{fixed_width=FALSE}). When \code{fixed_width=FALSE},
new columns are added to a row when \code{width} space is available and columns
will never exceed the container or viewport size. When \code{fixed_width=TRUE},
all columns will be exactly \code{width} wide, which may result in columns
overflowing the parent container.}

\item{heights_equal}{If \code{"all"} (the default), every card in every row of the
grid will have the same height. If \code{"row"}, then every card in \emph{each} row
of the grid will have the same height, but heights may vary between rows.}

\item{fill}{Whether or not to allow the layout to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{fillable}{Whether or not each element is wrapped in a fillable container.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{height_mobile}{Any valid CSS unit to use for the height when on mobile
devices (or narrow windows).}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}. This argument is only applicable when \code{fillable = TRUE}}

\item{class}{Additional CSS classes for the returned UI element.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Wraps a 1d sequence of UI elements into a 2d grid. The number of columns (and
rows) in the grid dependent on the column \code{width} as well as the size of the
display. For more explanation and illustrative examples, see the \emph{References}
section below.
}
\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
x <- card("A simple card")

# Always has 2 columns (on non-mobile)
layout_column_wrap(width = 1/2, x, x, x)

# Automatically lays out three cards into columns
# such that each column is at least 200px wide:
layout_column_wrap(x, x, x)

# To use larger column widths by default, set `width`.
# This example has 3 columns when the screen is at least 900px wide:
layout_column_wrap(width = "300px", x, x, x)
\dontshow{\}) # examplesIf}
}
\references{
The bslib website features \code{layout_column_wrap()} in two places:
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/column-layout/index.html}{Column-based layouts}
\item \href{https://rstudio.github.io/bslib/articles/cards/index.html#multiple-cards}{Cards: Multiple cards}
}
}
\seealso{
Other Column layouts: 
\code{\link{layout_columns}()}
}
\concept{Column layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/value-box.R
\name{value_box}
\alias{value_box}
\alias{showcase_left_center}
\alias{showcase_top_right}
\title{Value box}
\usage{
value_box(
  title,
  value,
  ...,
  showcase = NULL,
  showcase_layout = showcase_left_center(),
  full_screen = FALSE,
  theme_color = "primary",
  height = NULL,
  max_height = NULL,
  fill = TRUE,
  class = NULL
)

showcase_left_center(
  width = 0.3,
  max_height = "100px",
  max_height_full_screen = 0.67
)

showcase_top_right(
  width = 0.3,
  max_height = "75px",
  max_height_full_screen = 0.67
)
}
\arguments{
\item{title, value}{A string, number, or \code{\link[htmltools:builder]{htmltools::tag()}} child to display as
the title or value of the value box. The \code{title} appears above the \code{value}.}

\item{...}{Unnamed arguments may be any \code{\link[htmltools:builder]{htmltools::tag()}} children to
display below \code{value}. Named arguments become attributes on the containing
element.}

\item{showcase}{A \code{\link[htmltools:builder]{htmltools::tag()}} child to showcase (e.g., a
\code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, a \code{plotly::plotlyOutput()}, etc).}

\item{showcase_layout}{either \code{showcase_left_center()} or
\code{showcase_top_right()}.}

\item{full_screen}{If \code{TRUE}, an icon will appear when hovering over the card
body. Clicking the icon expands the card to fit viewport size.}

\item{theme_color}{A theme color to use for the background color. Should
match a name in the Bootstrap Sass variable \verb{$theme-colors} (e.g.,
\code{"secondary"}, \code{"success"}, \code{"danger"}, etc).}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{max_height, max_height_full_screen}{A proportion (i.e., a number between
0 and 1) or any valid \link[htmltools:validateCssUnit]{CSS unit} defining the
showcase max_height.}

\item{fill}{Whether to allow the value box to grow/shrink to fit a fillable
container with an opinionated height (e.g., \code{page_fillable()}).}

\item{class}{Utility classes for customizing the appearance of the summary
card. Use \verb{bg-*} and \verb{text-*} classes (e.g, \code{"bg-danger"} and
\code{"text-light"}) to customize the background/foreground colors.}

\item{width}{one of the following:
\itemize{
\item A proportion (i.e., a number between 0 and 1) of available width to
allocate to the showcase.
\item A vector of length 2 valid \link[htmltools:validateCssUnit]{CSS unit} defining
the width of each column (for \code{showcase_left_center()} the 1st unit defines
the showcase width and for \code{showcase_top_right} the 2nd unit defines the
showcase width). Note that any units supported by the CSS grid
\code{grid-template-columns} property may be used (e.g., \code{fr} units).
}}
}
\description{
An opinionated (\code{\link[=card]{card()}}-powered) box, designed for displaying a \code{value} and
\code{title}. Optionally, a \code{showcase} can provide for context for what the
\code{value} represents (for example, it could hold a \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, or
even a \code{\link[shiny:plotOutput]{shiny::plotOutput()}}).
}
\examples{

library(htmltools)

if (interactive()) {
  value_box(
    "KPI Title",
    h1(HTML("$1 <i>Billion</i> Dollars")),
    span(
      bsicons::bs_icon("arrow-up"),
      " 30\% VS PREVIOUS 30 DAYS"
    ),
    showcase = bsicons::bs_icon("piggy-bank"),
    class = "bg-success"
  )
}
}
\seealso{
\code{\link[=card]{card()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version-default.R
\name{versions}
\alias{versions}
\alias{version_default}
\title{Available Bootstrap versions}
\usage{
versions()

version_default()
}
\value{
Returns a list of the Bootstrap versions available.
}
\description{
Available Bootstrap versions
}
\seealso{
Other Bootstrap theme utility functions:
\code{\link[=bootswatch_themes]{bootswatch_themes()}},
\code{\link[=bs_get_variables]{bs_get_variables()}},
\code{\link[=builtin_themes]{builtin_themes()}},
\code{\link[=theme_bootswatch]{theme_bootswatch()}},
\code{\link[=theme_version]{theme_version()}}
}
\concept{Bootstrap theme utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{save_html.bslib_fragment}
\alias{save_html.bslib_fragment}
\alias{save_html.bslib_page}
\title{Save a bslib page/fragment as HTML}
\usage{
\method{save_html}{bslib_fragment}(html, file, ...)

\method{save_html}{bslib_page}(html, file, ...)
}
\arguments{
\item{html}{a bslib page/fragment.}

\item{...}{passed along to an underlying \code{\link[htmltools:save_html]{htmltools::save_html()}} method.}
}
\description{
Save a bslib page/fragment as HTML
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sidebar.R
\name{sidebar}
\alias{sidebar}
\alias{layout_sidebar}
\alias{toggle_sidebar}
\title{Sidebar layouts}
\usage{
sidebar(
  ...,
  width = 250,
  position = c("left", "right"),
  open = NULL,
  id = NULL,
  title = NULL,
  bg = NULL,
  fg = NULL,
  class = NULL,
  max_height_mobile = NULL,
  gap = NULL,
  padding = NULL,
  fillable = FALSE
)

layout_sidebar(
  ...,
  sidebar = NULL,
  fillable = TRUE,
  fill = TRUE,
  bg = NULL,
  fg = NULL,
  border = NULL,
  border_radius = NULL,
  border_color = NULL,
  padding = NULL,
  gap = NULL,
  height = NULL
)

toggle_sidebar(id, open = NULL, session = get_current_session())
}
\arguments{
\item{...}{Unnamed arguments can be any valid child of an \link[htmltools:builder]{htmltools tag} and named arguments become HTML attributes on
returned UI element. In the case of \code{layout_sidebar()}, these arguments are
passed to the main content tag (not the sidebar+main content container).}

\item{width}{A valid \link[htmltools:validateCssUnit]{CSS unit} used for the
width of the sidebar.}

\item{position}{Where the sidebar should appear relative to the main content.}

\item{open}{The initial state of the sidebar, choosing from the following
options:
\itemize{
\item \code{"desktop"}: The sidebar starts open on desktop screen, closed on mobile.
This is default sidebar behavior.
\item \code{"open"} or \code{TRUE}: The sidebar starts open.
\item \code{"closed"} or \code{FALSE}: The sidebar starts closed.
\item \code{"always"} or \code{NA}: The sidebar is always open and cannot be closed.
}

Alternatively, you can use a list with \code{desktop} or \code{mobile} items to set
the initial sidebar state independently for \code{desktop} and \code{mobile} screen
sizes. In this case, \code{desktop} or \code{mobile} can use any of the above options
except \code{"desktop"}, which is equivalent to
\code{list(desktop = "open", mobile = "closed")}. You can also choose to place
an always open sidebar above the main content on mobile devices by setting
\code{mobile = "always-above"}.

In \code{sidebar_toggle()}, \code{open} indicates the desired state of the sidebar,
where the default of \code{open = NULL} will cause the sidebar to be toggled
open if closed or vice versa. Note that \code{sidebar_toggle()} can only open or
close the sidebar, so it does not support the \code{"desktop"} and \code{"always"}
options.}

\item{id}{A character string. Required if wanting to re-actively read (or
update) the \code{collapsible} state in a Shiny app.}

\item{title}{A character title to be used as the sidebar title, which will be
wrapped in a \verb{<header>} element with class \code{sidebar-title}. You can also
provide a custom \code{\link[htmltools:builder]{htmltools::tag()}} for the title element, in which case
you'll likely want to give this element \code{class = "sidebar-title"}.}

\item{bg, fg}{A background or foreground color. If only one of either is
provided, an accessible contrasting color is provided for the opposite
color, e.g. setting \code{bg} chooses an appropriate \code{fg} color.}

\item{class}{CSS classes for the sidebar container element, in addition to
the fixed \code{.sidebar} class.}

\item{max_height_mobile}{A \link[htmltools:validateCssUnit]{CSS length unit}
defining the maximum height of the horizontal sidebar when viewed on mobile
devices. Only applies to always-open sidebars that use \code{open = "always"},
where by default the sidebar container is placed below the main content
container on mobile devices.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
vertical \code{gap} (i.e., spacing) between adjacent elements provided to \code{...}.}

\item{padding}{Padding within the sidebar itself. This can be a numeric
vector (which will be interpreted as pixels) or a character vector with
valid CSS lengths. \code{padding} may be one to four values. If one, then
that value will be used for all four sides. If two, then the first value
will be used for the top and bottom, while the second value will be used
for left and right. If three, then the first will be used for top, the
second will be left and right, and the third will be bottom. If four, then
the values will be interpreted as top, right, bottom, and left
respectively.}

\item{fillable}{Whether or not the \code{main} content area should be considered a
fillable (i.e., flexbox) container.}

\item{sidebar}{A \code{\link[=sidebar]{sidebar()}} object.}

\item{fill}{Whether or not to allow the layout container to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{border}{Whether or not to add a border.}

\item{border_radius}{Whether or not to add a border radius.}

\item{border_color}{The border color that is applied to the entire layout (if
\code{border = TRUE}) and the color of the border between the sidebar and the
main content area.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\description{
Sidebar layouts place UI elements, like input controls or additional context,
next to the main content area which often holds output elements like plots or
tables.

There are several page, navigation, and layout functions that allow you to
create a sidebar layout. In each case, you can create a collapsing sidebar
layout by providing a \code{sidebar()} object to the \code{sidebar} argument the
following functions.
\itemize{
\item \code{\link[=page_sidebar]{page_sidebar()}} creates a "page-level" sidebar.
\item \code{\link[=page_navbar]{page_navbar()}} creates a multi-panel app with an (optional, page-level)
sidebar that is shown on every panel.
\item \code{layout_sidebar()} creates a "floating" sidebar layout component which can
be used inside any \code{\link[=page]{page()}} and/or \code{\link[=card]{card()}} context.
\item \code{\link[=navset_card_tab]{navset_card_tab()}} and \code{\link[=navset_card_pill]{navset_card_pill()}} create multi-tab cards with a
shared sidebar that is accessible from every panel.
}

See \href{https://rstudio.github.io/bslib/articles/sidebars.html}{the Sidebars article}
on the bslib website to learn more.
}
\section{Functions}{
\itemize{
\item \code{toggle_sidebar()}: Toggle a \code{sidebar()} state during an active Shiny user
session. To use this function, the \code{sidebar()} you want to open or close
must have an \code{id} value.

}}
\references{
Sidebar layouts are featured in a number of pages on the bslib
website:
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/sidebars.html}{Sidebars}
\item \href{https://rstudio.github.io/bslib/articles/cards/index.html#sidebars}{Cards: Sidebars}
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html}{Getting Started: Dashboards}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page}
\alias{page}
\alias{page_fluid}
\alias{page_fixed}
\title{Modern Bootstrap page layouts}
\usage{
page(..., title = NULL, theme = bs_theme(), lang = NULL)

page_fluid(..., title = NULL, theme = bs_theme(), lang = NULL)

page_fixed(..., title = NULL, theme = bs_theme(), lang = NULL)
}
\arguments{
\item{...}{UI elements for the page. Named arguments become HTML attributes.}

\item{title}{The browser window title (defaults to the host URL of the page)}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}
}
\description{
These functions are small wrappers around shiny's page constructors (i.e.,
\code{\link[shiny:fluidPage]{shiny::fluidPage()}}, \code{\link[shiny:navbarPage]{shiny::navbarPage()}}, etc) that differ in two ways:
\itemize{
\item The \code{theme} parameter defaults bslib's recommended version of Bootstrap (for new projects).
\item The return value is rendered as an static HTML page when printed interactively at the console.
}
}
\section{Functions}{
\itemize{
\item \code{page_fluid()}: A \pkg{bslib} wrapper for \code{\link[shiny:fluidPage]{shiny::fluidPage()}}, a fluid
Bootstrap-based page layout that extends to the full viewport width.

\item \code{page_fixed()}: A \pkg{bslib} wrapper for \code{\link[shiny:fixedPage]{shiny::fixedPage()}}, a fixed
Bootstrap-based page layout where the page content container is centered
horizontally and its width is constrained.

}}
\seealso{
Dashboard-style pages: \code{\link[=page_sidebar]{page_sidebar()}}, \code{\link[=page_navbar]{page_navbar()}},
\code{\link[=page_fillable]{page_fillable()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-layers.R
\name{bs_add_variables}
\alias{bs_add_variables}
\alias{bs_add_rules}
\alias{bs_add_functions}
\alias{bs_add_mixins}
\alias{bs_bundle}
\title{Add low-level theming customizations}
\usage{
bs_add_variables(
  theme,
  ...,
  .where = "defaults",
  .default_flag = identical(.where, "defaults")
)

bs_add_rules(theme, rules)

bs_add_functions(theme, functions)

bs_add_mixins(theme, mixins)

bs_bundle(theme, ...)
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{...}{\itemize{
\item \code{bs_add_variables()}: Should be named Sass variables or values that can be
passed in directly to the \code{defaults} argument of a \code{\link[sass:sass_layer]{sass::sass_layer()}}.
\item \code{bs_bundle()}: Should be arguments that can be handled by
\code{\link[sass:sass_layer]{sass::sass_bundle()}} to be appended to the \code{theme}
}}

\item{.where}{Whether to place the variable definitions before other Sass
\code{"defaults"}, after other Sass \code{"declarations"}, or after other Sass
\code{"rules"}.}

\item{.default_flag}{Whether or not to add a \code{!default} flag (if missing) to
variable expressions. It's recommended to keep this as \code{TRUE} when \code{.where = "defaults"}.}

\item{rules}{Sass rules. Anything understood by \code{\link[sass:as_sass]{sass::as_sass()}} may be
provided (e.g., a list, character vector, \code{\link[sass:sass_import]{sass::sass_file()}}, etc)}

\item{functions}{A character vector or \code{\link[sass:sass_import]{sass::sass_file()}} containing
functions definitions.}

\item{mixins}{A character vector or \code{\link[sass:sass_import]{sass::sass_file()}} containing
mixin definitions.}
}
\value{
Returns a modified \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
These functions provide direct access to the layers of a bslib theme created
with \code{\link[=bs_theme]{bs_theme()}}. Learn more about \href{https://rstudio.github.io/sass/articles/sass.html#layering}{composable Sass layers} on the
\pkg{sass} website.
}
\details{
Compared to higher-level theme customization available in \code{\link[=bs_theme]{bs_theme()}}, these
functions are a more direct interface to Bootstrap Sass, and therefore, do
nothing to ensure theme customizations are portable between major Bootstrap
versions.
}
\section{Functions}{
\itemize{
\item \code{bs_add_variables()}: Add Bootstrap Sass
\href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html}{variable defaults}.

\item \code{bs_add_rules()}: Add additional
\href{https://sass-lang.com/documentation/style-rules}{Sass rules}.

\item \code{bs_add_functions()}: Add additional \href{https://rstudio.github.io/sass/articles/sass.html#functions}{Sass functions}.

\item \code{bs_add_mixins()}: Add additional \href{https://rstudio.github.io/sass/articles/sass.html#mixins}{Sass mixins}.

\item \code{bs_bundle()}: Add additional \code{\link[sass:sass_layer]{sass::sass_bundle()}} objects to an
existing \code{theme}.

}}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}

# Function to preview the styling a (primary) Bootstrap button
library(htmltools)
button <- tags$a(class = "btn btn-primary", href = "#", role = "button", "Hello")
preview_button <- function(theme) {
  browsable(tags$body(bs_theme_dependencies(theme), button))
}

# Here we start with a theme based on a Bootswatch theme,
# then override some variable defaults
theme <- bs_add_variables(
  bs_theme(bootswatch = "sketchy", primary = "orange"),
  "body-bg" = "#EEEEEE",
  "font-family-base" = "monospace",
  "font-size-base" = "1.4rem",
  "btn-padding-y" = ".16rem",
  "btn-padding-x" = "2rem"
)

preview_button(theme)

# If you need to set a variable based on another Bootstrap variable
theme <- bs_add_variables(theme, "body-color" = "$success", .where = "declarations")
preview_button(theme)

# Start a new global theme and add some custom rules that
# use Bootstrap variables to define a custom styling for a
# 'person card'
person_rules <- system.file("custom", "person.scss", package = "bslib")
theme <- bs_add_rules(bs_theme(), sass::sass_file(person_rules))

# Include custom CSS that leverages bootstrap Sass variables
person <- function(name, title, company) {
  tags$div(
    class = "person",
    h3(class = "name", name),
    div(class = "title", title),
    div(class = "company", company)
  )
}

page_fluid(
  theme = theme,
  person("Andrew Carnegie", "Owner", "Carnegie Steel Company"),
  person("John D. Rockefeller", "Chairman", "Standard Oil")
)
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item bslib's theming capabilities are powered by
\href{https://rstudio.github.io/sass/}{the {sass} package}.
\item Learn more about
\href{https://rstudio.github.io/sass/articles/sass.html#layering}{composable Sass layers}
on the \pkg{sass} website.
}
}
\seealso{
\code{\link[=bs_theme]{bs_theme()}} creates a Bootstrap theme object, and is the best place
to start learning about bslib's theming capabilities.

Other Bootstrap theme functions: 
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme_preview}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fill.R
\name{as_fill_carrier}
\alias{as_fill_carrier}
\alias{as_fillable_container}
\alias{as_fill_item}
\alias{remove_all_fill}
\alias{is_fill_carrier}
\alias{is_fillable_container}
\alias{is_fill_item}
\title{Test and/or coerce fill behavior}
\usage{
as_fill_carrier(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
)

as_fillable_container(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
)

as_fill_item(
  x,
  ...,
  min_height = NULL,
  max_height = NULL,
  class = NULL,
  style = NULL,
  css_selector = NULL
)

remove_all_fill(x)

is_fill_carrier(x)

is_fillable_container(x)

is_fill_item(x)
}
\arguments{
\item{x}{An \code{\link[htmltools:builder]{htmltools::tag()}}.}

\item{...}{Currently ignored.}

\item{min_height, max_height}{Any valid \link[htmltools:validateCssUnit]{CSS unit}
(e.g., \code{150}).}

\item{gap}{Any valid \link[htmltools:validateCssUnit]{CSS unit}.}

\item{class}{A character vector of class names to add to the tag.}

\item{style}{A character vector of CSS properties to add to the tag.}

\item{css_selector}{A character string containing a CSS selector for
targeting particular (inner) tag(s) of interest. For more details on what
selector(s) are supported, see \code{\link[htmltools:tagAppendAttributes]{htmltools::tagAppendAttributes()}}.}
}
\value{
\itemize{
\item For \code{as_fill()}, \code{as_fillable()}, and \code{as_fill_carrier()}: the \emph{tagified}
version \code{x}, with relevant tags modified to possess the relevant fill
properties.
\item For \code{is_fill()}, \code{is_fillable()}, and \code{is_fill_carrier()}: a logical vector,
with length matching the number of top-level tags that possess the relevant
fill properties.
}
}
\description{
Filling layouts in bslib are built on the foundation of fillable containers
and fill items (fill carriers are both fillable and fill). This is why most
bslib components (e.g., \code{\link[=card]{card()}}, \code{\link[=card_body]{card_body()}}, \code{\link[=layout_sidebar]{layout_sidebar()}}) possess
both \code{fillable} and \code{fill} arguments (to control their fill behavior).
However, sometimes it's useful to add, remove, and/or test fillable/fill
properties on arbitrary \code{\link[htmltools:builder]{htmltools::tag()}}, which these functions are
designed to do.
}
\details{
Although \code{as_fill()}, \code{as_fillable()}, and \code{as_fill_carrier()} can work with
non-tag objects that have a \link[htmltools:as.tags]{htmltools::as.tags} method (e.g., htmlwidgets),
they return the "tagified" version of that object.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
shinyApp(
  page_fillable(
    # without `as_fill_carrier()`, the plot won't fill the page because
    # `uiOutput()` is neither a fillable container nor a fill item by default.
    as_fill_carrier(uiOutput("ui"))
  ),
  function(input, output) {
    output$ui <- renderUI({
      div(
        class = "bg-info text-white",
        as_fill_item(),
        "A fill item"
      )
    })
  }
)
\dontshow{\}) # examplesIf}
}
\references{
The \href{https://rstudio.github.io/bslib/articles/filling.html}{Filling Layouts}
article on the bslib website introduces the concept of fillable containers
and fill items.
}
\seealso{
These functions provide a convenient interface to the underlying
\code{\link[htmltools:bindFillRole]{htmltools::bindFillRole()}} function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toolbar.R
\name{toolbar_input_select}
\alias{toolbar_input_select}
\alias{update_toolbar_input_select}
\title{Toolbar Input Select}
\usage{
toolbar_input_select(
  id,
  label,
  choices,
  ...,
  selected = NULL,
  icon = NULL,
  show_label = FALSE,
  tooltip = !show_label
)

update_toolbar_input_select(
  id,
  label = NULL,
  show_label = NULL,
  choices = NULL,
  selected = NULL,
  icon = NULL,
  session = get_current_session()
)
}
\arguments{
\item{id}{The input ID.}

\item{label}{The input label. By default, \code{label} is not shown but is used by
\code{tooltip}. Set \code{show_label = TRUE} to show the label (see \code{tooltip} for
details on how this affects the tooltip behavior).}

\item{choices}{List of values to select from. If elements of the list are
named, then that name --- rather than the value --- is displayed to the
user. It's also possible to group related inputs by providing a named list
whose elements are (either named or unnamed) lists, vectors, or factors. In
this case, the outermost names will be used as the group labels (leveraging
the \verb{<optgroup>} HTML tag) for the elements in the respective sublist. See
the example section for a small demo of this feature.}

\item{...}{Additional named arguments passed as attributes to the outer
container div.}

\item{selected}{The initially selected value. If not provided on input creation, the
first choice will be selected by default.
If provided in \code{update_toolbar_input_select()} with a new set of
\code{choices}, it will replace the currently selected value.}

\item{icon}{An icon. If provided without \code{show_label = TRUE}, only the icon
will be visible.}

\item{show_label}{Whether to show the label text. If \code{FALSE} (the default),
only the icon is shown (if provided). If \code{TRUE}, the label text is shown
alongside the icon. Note that \code{show_label} can be dynamically updated using
\code{\link[=update_toolbar_input_button]{update_toolbar_input_button()}}.}

\item{tooltip}{Tooltip text to display when hovering over the input. Can be:
\itemize{
\item \code{TRUE} (default when \code{show_label = FALSE}) - shows a tooltip with the
\code{label} text
\item \code{FALSE} (default when \code{show_label = TRUE}) - no tooltip
\item A character string - shows a tooltip with custom text
}

Defaults to \code{!show_label}. When a tooltip is created, it will have an ID of
\code{"{id}_tooltip"} which can be used to update the tooltip text dynamically
via \code{\link[=update_tooltip]{update_tooltip()}}.}

\item{session}{A Shiny session object (the default should almost always be
used).}
}
\value{
Returns a select input control suitable for use in a toolbar.
}
\description{
Create a select input control that can be used to choose a single item from
a list of values, suitable for use within a \code{\link[=toolbar]{toolbar()}}.
}
\details{
When a tooltip is created for the select input, it will have an ID of
\code{"{id}_tooltip"} which can be used to update the tooltip text dynamically
via \code{\link[=update_tooltip]{update_tooltip()}}.
}
\section{Functions}{
\itemize{
\item \code{toolbar_input_select()}: Create a toolbar select input.

\item \code{update_toolbar_input_select()}: Update a toolbar select input.

}}
\section{Updating toolbar select inputs}{


You can update the appearance and choices of a toolbar select input. This
function works similarly to \code{\link[shiny:updateSelectInput]{shiny::updateSelectInput()}}, but is specifically
designed for \code{\link[=toolbar_input_select]{toolbar_input_select()}}. It allows you to update the select's
label, icon, choices, selected value, and label visibility from the server.

Note that you cannot enable or disable the \code{tooltip} parameter after the
select has been created, only update the text of the tooltip.
When a tooltip is created for the select input, it will have an ID of
\code{"{id}_tooltip"} which can be used to update the tooltip text dynamically
via \code{\link[=update_tooltip]{update_tooltip()}}.

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(shiny)
library(bslib)

ui <- page_fluid(
  toolbar(
    align = "right",
    toolbar_input_select(
      "select",
      label = "Choose",
      icon = bsicons::bs_icon("filter"),
      choices = c("A", "B", "C")
    ),
    toolbar_input_button(
      "change_choices",
      label = "Change Choices",
      show_label = TRUE,
      icon = bsicons::bs_icon("arrow-repeat")
    )
  ),
  verbatimTextOutput("value")
)

server <- function(input, output, session) \{
  output$value <- renderPrint(\{
    input$select
  \})

  observeEvent(input$change_choices, \{
    update_toolbar_input_select(
      "select",
      label = "Pick one",
      choices = c("hi", "hello", "hey"),
      selected = "hello"
    )
    # Update the tooltip text
    update_tooltip("select_tooltip", "Choose your NEW option")
  \})
\}

shinyApp(ui, server)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
toolbar(
  align = "right",
  toolbar_input_select(
    id = "select",
    label = "Choose option",
    choices = c("Option 1", "Option 2", "Option 3"),
    selected = "Option 2"
  )
)
\dontshow{\}) # examplesIf}
}
\seealso{
Other toolbar components:
\code{\link[=toolbar]{toolbar()}},
\code{\link[=toolbar_divider]{toolbar_divider()}},
\code{\link[=toolbar_input_button]{toolbar_input_button()}}
}
\concept{toolbar components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme.R
\name{bs_theme}
\alias{bs_theme}
\alias{bs_theme_update}
\alias{is_bs_theme}
\title{Create a Bootstrap theme}
\usage{
bs_theme(
  version = version_default(),
  preset = NULL,
  ...,
  brand = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  font_scale = NULL,
  bootswatch = NULL
)

bs_theme_update(
  theme,
  ...,
  preset = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  font_scale = NULL,
  bootswatch = NULL
)

is_bs_theme(x)
}
\arguments{
\item{version}{The major version of Bootstrap to use (see \code{\link[=versions]{versions()}}
for possible values). Defaults to the currently recommended version
for new projects (currently Bootstrap 5).}

\item{preset}{The name of a theme preset, either a built-in theme provided by
bslib or a Bootswatch theme (see \code{\link[=builtin_themes]{builtin_themes()}} and
\code{\link[=bootswatch_themes]{bootswatch_themes()}} for possible values). This argument takes precedence
over the \code{bootswatch} argument and only one \code{preset} or \code{bootswatch} can be
provided. When no \code{bootswatch} theme is specified, and \code{version} is 5 or
higher, \code{preset} defaults to \code{"shiny"}. To remove the \code{"shiny"} preset,
provide a value of \code{"bootstrap"} (this value will also work in
\code{bs_theme_update()} to remove a \code{preset} or \code{bootswatch} theme).}

\item{...}{arguments passed along to \code{\link[=bs_add_variables]{bs_add_variables()}}.}

\item{brand}{Specifies how to apply branding to your theme using
\href{https://posit-dev.github.io/brand-yml/}{brand.yml}, a simple YAML file that
defines key brand elements like colors, fonts, and logos. Valid options:
\itemize{
\item \code{NULL} (default): Automatically looks for a \verb{_brand.yml} file in the
current directory or in \verb{_brand/} or \verb{brand/} in the current directory.
If not found, it searches parent project directories for a \verb{_brand.yml}
file (also possibly in \verb{_brand/} or \verb{brand/}). If a \verb{_brand.yml} file is
found, it is applied to the Bootstrap theme.
\item \code{TRUE} (default): Automatically looks for a \verb{_brand.yml} file in the
current or app directory as described above. If a \verb{_brand.yml} file
\emph{is not found}, \code{bs_theme()} will throw an error.
\item \code{FALSE}: Disables any brand.yml usage, even if a \verb{_brand.yml} file is
present.
\item A file path that directly points to a specific brand.yml file (with any
file name) that you want to use.
\item Use a list to directly provide brand settings directly in R, following
the brand.yml structure.
}

Learn more about creating and using brand.yml files at the
\href{https://posit-dev.github.io/brand-yml/}{brand.yml homepage} or run
\code{shiny::runExample("brand.yml", package = "bslib")} to try brand.yml in a
demo app.}

\item{bg}{A color string for the background.}

\item{fg}{A color string for the foreground.}

\item{primary}{A color to be used for hyperlinks, to indicate primary/default
actions, and to show active selection state in some Bootstrap components.
Generally a bold, saturated color that contrasts with the theme's base
colors.}

\item{secondary}{A color for components and messages that don't need to stand
out. (Not supported in Bootstrap 3.)}

\item{success}{A color for messages that indicate an operation has succeeded.
Typically green.}

\item{info}{A color for messages that are informative but not critical.
Typically a shade of blue-green.}

\item{warning}{A color for warning messages. Typically yellow.}

\item{danger}{A color for errors. Typically red.}

\item{base_font}{The default typeface.}

\item{code_font}{The typeface to be used for code. Be sure this is monospace!}

\item{heading_font}{The typeface to be used for heading elements.}

\item{font_scale}{A scalar multiplier to apply to the base font size. For
example, a value of \code{1.5} scales font sizes to 150\% and a value of \code{0.8}
scales to 80\%. Must be a positive number.}

\item{bootswatch}{The name of a bootswatch theme (see \code{\link[=bootswatch_themes]{bootswatch_themes()}}
for possible values). When provided to \code{bs_theme_update()}, any previous
Bootswatch theme is first removed before the new one is applied (use
\code{bootswatch = "bootstrap"} to effectively remove the Bootswatch theme).}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{x}{an object.}
}
\value{
Returns a \code{\link[sass:sass_layer]{sass::sass_bundle()}} (list-like) object.
}
\description{
Creates a Bootstrap theme object, where you can:
\itemize{
\item Choose a (major) Bootstrap \code{version}.
\item Choose a \href{https://bootswatch.com}{Bootswatch theme} (optional).
\item Customize main colors and fonts via explicitly named arguments (e.g.,
\code{bg}, \code{fg}, \code{primary}, etc).
\item Customize other, lower-level, Bootstrap Sass variable defaults via \code{...}.
}

To learn more about how to implement custom themes, as well as how to use
them inside Shiny and R Markdown, \href{https://rstudio.github.io/bslib/articles/theming.html}{see here}.
}
\section{Colors}{


Colors may be provided in any format that \code{\link[htmltools:parseCssColors]{htmltools::parseCssColors()}} can
understand. To control the vast majority of the ('grayscale') color
defaults, specify both the \code{fg} (foreground) and \code{bg} (background) colors.
The \code{primary} and \code{secondary} theme colors are also useful for accenting the
main grayscale colors in things like hyperlinks, tabset panels, and buttons.
}

\section{Fonts}{


Use \code{base_font}, \code{code_font}, and \code{heading_font} to control the main
typefaces. These arguments set new defaults for the relevant \code{font-family}
CSS properties, but don't necessarily import the relevant font files. To
both set CSS properties \emph{and} import font files, consider using the various
\code{\link[=font_face]{font_face()}} helpers.

Each \verb{*_font} argument may be a single font or a \code{font_collection()}.
A font can be created with \code{\link[=font_google]{font_google()}}, \code{\link[=font_link]{font_link()}}, or \code{\link[=font_face]{font_face()}},
or it can be a character vector of font names in the following format:
\itemize{
\item A single unquoted name (e.g., \code{"Source Sans Pro"}).
\item A single quoted name (e.g., \code{"'Source Sans Pro'"}).
\item A comma-separated list of names w/ individual names quoted as necessary.
(e.g. \code{c("Open Sans", "'Source Sans Pro'", "'Helvetica Neue', Helvetica, sans-serif")})
}

\code{font_google()} sets \code{local = TRUE} by default, which ensures that the font
files are downloaded from \href{https://fonts.google.com/}{Google Fonts} when
your document or app is rendered. This guarantees that the client has access
to the font family, making it relatively safe to specify just one font
family:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{bs_theme(base_font = font_google("Pacifico", local = TRUE))
}\if{html}{\out{</div>}}

That said, we recommend you specify multiple "fallback" font families,
especially when relying on remote and/or system fonts being available.
Fallback fonts are useful not only for handling missing fonts, but also
ensure that your users don't experience a Flash of Invisible Text (FOIT)
which can be quite noticeable with remote web fonts on a slow internet
connection.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{bs_theme(base_font = font_collection(font_google("Pacifico", local = FALSE), "Roboto", "sans-serif"))
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}

theme <- bs_theme(
  # Controls the default grayscale palette
  bg = "#202123", fg = "#B8BCC2",
  # Controls the accent (e.g., hyperlink, button, etc) colors
  primary = "#EA80FC", secondary = "#48DAC6",
  base_font = c("Grandstander", "sans-serif"),
  code_font = c("Courier", "monospace"),
  heading_font = "'Helvetica Neue', Helvetica, sans-serif",
  # Can also add lower-level customization
  "input-border-color" = "#EA80FC"
)

bs_theme_preview(theme)

# Lower-level bs_add_*() functions allow you to work more
# directly with the underlying Sass code
theme <- bs_add_variables(theme, "my-class-color" = "red")
theme <- bs_add_rules(theme, ".my-class { color: $my-class-color }")
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/theming/index.html}{Get Started: Theming}
introduces theming with bslib in Shiny apps and R Markdown documents.
\item \href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html}{Theming: Bootstrap 5 variables}
provides a searchable reference of all theming variables available in
Bootstrap 5.
\item \href{https://rstudio.github.io/bslib/articles/custom-components/index.html}{Theming: Custom components}
gives a tutorial on creating a dynamically themable custom component.
\item bslib's theming capabilities are powered by
\href{https://rstudio.github.io/sass/}{the {sass} package}.
\item \href{https://rstudio.github.io/bslib/articles/utility-classes/index.html}{Bootstrap's utility classes}
can be helpful when you want to change the appearance of an element
without writing CSS or customizing your \code{bs_theme()}.
}
}
\seealso{
Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme_preview}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-global.R, R/bs-theme.R
\name{bs_global_theme}
\alias{bs_global_theme}
\alias{bs_global_set}
\alias{bs_global_get}
\alias{bs_global_clear}
\alias{bs_global_add_variables}
\alias{bs_global_add_rules}
\alias{bs_global_bundle}
\alias{bs_global_theme_update}
\title{Global theming}
\usage{
bs_global_theme(
  version = version_default(),
  preset = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  ...,
  bootswatch = NULL
)

bs_global_set(theme = bs_theme())

bs_global_get()

bs_global_clear()

bs_global_add_variables(
  ...,
  .where = "defaults",
  .default_flag = identical(.where, "defaults")
)

bs_global_add_rules(...)

bs_global_bundle(...)

bs_global_theme_update(
  ...,
  preset = NULL,
  bg = NULL,
  fg = NULL,
  primary = NULL,
  secondary = NULL,
  success = NULL,
  info = NULL,
  warning = NULL,
  danger = NULL,
  base_font = NULL,
  code_font = NULL,
  heading_font = NULL,
  bootswatch = NULL
)
}
\arguments{
\item{version}{The major version of Bootstrap to use (see \code{\link[=versions]{versions()}}
for possible values). Defaults to the currently recommended version
for new projects (currently Bootstrap 5).}

\item{preset}{The name of a theme preset, either a built-in theme provided by
bslib or a Bootswatch theme (see \code{\link[=builtin_themes]{builtin_themes()}} and
\code{\link[=bootswatch_themes]{bootswatch_themes()}} for possible values). This argument takes precedence
over the \code{bootswatch} argument and only one \code{preset} or \code{bootswatch} can be
provided. When no \code{bootswatch} theme is specified, and \code{version} is 5 or
higher, \code{preset} defaults to \code{"shiny"}. To remove the \code{"shiny"} preset,
provide a value of \code{"bootstrap"} (this value will also work in
\code{bs_theme_update()} to remove a \code{preset} or \code{bootswatch} theme).}

\item{bg}{A color string for the background.}

\item{fg}{A color string for the foreground.}

\item{primary}{A color to be used for hyperlinks, to indicate primary/default
actions, and to show active selection state in some Bootstrap components.
Generally a bold, saturated color that contrasts with the theme's base
colors.}

\item{secondary}{A color for components and messages that don't need to stand
out. (Not supported in Bootstrap 3.)}

\item{success}{A color for messages that indicate an operation has succeeded.
Typically green.}

\item{info}{A color for messages that are informative but not critical.
Typically a shade of blue-green.}

\item{warning}{A color for warning messages. Typically yellow.}

\item{danger}{A color for errors. Typically red.}

\item{base_font}{The default typeface.}

\item{code_font}{The typeface to be used for code. Be sure this is monospace!}

\item{heading_font}{The typeface to be used for heading elements.}

\item{...}{arguments passed along to \code{\link[=bs_add_variables]{bs_add_variables()}}.}

\item{bootswatch}{The name of a bootswatch theme (see \code{\link[=bootswatch_themes]{bootswatch_themes()}}
for possible values). When provided to \code{bs_theme_update()}, any previous
Bootswatch theme is first removed before the new one is applied (use
\code{bootswatch = "bootstrap"} to effectively remove the Bootswatch theme).}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{.where}{Whether to place the variable definitions before other Sass
\code{"defaults"}, after other Sass \code{"declarations"}, or after other Sass
\code{"rules"}.}

\item{.default_flag}{Whether or not to add a \code{!default} flag (if missing) to
variable expressions. It's recommended to keep this as \code{TRUE} when \code{.where = "defaults"}.}
}
\value{
Functions that modify the global theme (e.g., \code{bs_global_set()})
invisibly return the previously set theme. \code{bs_global_get()} returns the
current global theme.
}
\description{
\code{bs_global_theme()} creates and sets the global Bootstrap Sass theme. This
theme is typically found by \code{\link[=bs_theme_dependencies]{bs_theme_dependencies()}} in the app or document
where the global theme is being used. You can obtain the current global theme
with \code{\link[=bs_global_get]{bs_global_get()}} or directly set the global theme with
\code{\link[=bs_global_set]{bs_global_set()}}.
}
\examples{
# Remember the global state now (so we can restore later)
theme <- bs_global_get()

# Use Bootstrap 3 (globally) with some theme customization
bs_global_theme(3, bg = "#444", fg = "#e4e4e4", primary = "#e39777")
if (rlang::is_interactive()) {
  bs_theme_preview(with_themer = FALSE)
}

# If no global theme is active, bs_global_get() returns NULL
bs_global_clear()
bs_global_get()

# Restore the original state
bs_global_set(theme)

}
\seealso{
Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme_preview}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layout.R
\name{layout_column_wrap}
\alias{layout_column_wrap}
\title{Column-first uniform grid layouts}
\usage{
layout_column_wrap(
  ...,
  width = "200px",
  fixed_width = FALSE,
  heights_equal = c("all", "row"),
  fill = TRUE,
  fillable = TRUE,
  height = NULL,
  height_mobile = NULL,
  min_height = NULL,
  max_height = NULL,
  gap = NULL,
  class = NULL
)
}
\arguments{
\item{...}{Unnamed arguments should be UI elements (e.g., \code{\link[=card]{card()}}). Named
arguments become attributes on the containing \link[htmltools:builder]{htmltools::tag} element.}

\item{width}{The desired width of each card, which can be any of the
following:
\itemize{
\item A (unit-less) number between 0 and 1.
\itemize{
\item This should be specified as \code{1/num}, where \code{num} represents the number
of desired columns.
}
\item A \link[htmltools:validateCssUnit]{CSS length unit}
\itemize{
\item Either the minimum (when \code{fixed_width=FALSE}) or fixed width
(\code{fixed_width=TRUE}).
}
\item \code{NULL}
\itemize{
\item Allows power users to set the \code{grid-template-columns} CSS property
manually, either via a \code{style} attribute or a CSS stylesheet.
}
}}

\item{fixed_width}{When \code{width} is greater than 1 or is a CSS length unit,
e.g. \code{"200px"}, \code{fixed_width} indicates whether that \code{width} value
represents the absolute size of each column (\code{fixed_width=TRUE}) or the
minimum size of a column (\code{fixed_width=FALSE}). When \code{fixed_width=FALSE},
new columns are added to a row when \code{width} space is available and columns
will never exceed the container or viewport size. When \code{fixed_width=TRUE},
all columns will be exactly \code{width} wide, which may result in columns
overflowing the parent container.}

\item{heights_equal}{If \code{"all"} (the default), every card in every row of the
grid will have the same height. If \code{"row"}, then every card in \emph{each} row
of the grid will have the same height, but heights may vary between rows.}

\item{fill}{Whether or not to allow the layout to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{fillable}{Whether or not each element is wrapped in a fillable container.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{height_mobile}{Any valid CSS unit to use for the height when on mobile
devices (or narrow windows).}

\item{min_height, max_height}{The maximum or minimum height of the layout container.
Can be any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{max_height="200px"}). Use these arguments in filling layouts to ensure that a
layout container doesn't shrink below \code{min_height} or grow beyond \code{max_height}.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}. This argument is only applicable when \code{fillable = TRUE}}

\item{class}{Additional CSS classes for the returned UI element.}
}
\description{
Wraps a 1d sequence of UI elements into a 2d grid. The number of columns (and
rows) in the grid dependent on the column \code{width} as well as the size of the
display. For more explanation and illustrative examples, see the \emph{References}
section below.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
x <- card("A simple card")

# Always has 2 columns (on non-mobile)
layout_column_wrap(width = 1/2, x, x, x)

# Automatically lays out three cards into columns
# such that each column is at least 200px wide:
layout_column_wrap(x, x, x)

# To use larger column widths by default, set `width`.
# This example has 3 columns when the screen is at least 900px wide:
layout_column_wrap(width = "300px", x, x, x)

# You can add a list of items, spliced with rlang's `!!!` operator
layout_column_wrap(!!!list(x, x, x))
\dontshow{\}) # examplesIf}
}
\references{
The bslib website features \code{layout_column_wrap()} in two places:
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/column-layout/index.html}{Column-based layouts}
\item \href{https://rstudio.github.io/bslib/articles/cards/index.html#multiple-cards}{Cards: Multiple cards}
}
}
\seealso{
Other Column layouts: 
\code{\link{layout_columns}()}
}
\concept{Column layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/value-box.R
\name{value_box}
\alias{value_box}
\alias{value_box_theme}
\alias{showcase_left_center}
\alias{showcase_top_right}
\alias{showcase_bottom}
\title{Value box}
\usage{
value_box(
  title,
  value,
  ...,
  showcase = NULL,
  showcase_layout = c("left center", "top right", "bottom"),
  full_screen = FALSE,
  theme = NULL,
  height = NULL,
  max_height = NULL,
  fill = TRUE,
  class = NULL,
  theme_color = deprecated()
)

value_box_theme(name = NULL, bg = NULL, fg = NULL)

showcase_left_center(
  width = 0.3,
  width_full_screen = "1fr",
  max_height = "100px",
  max_height_full_screen = 0.67
)

showcase_top_right(
  width = 0.4,
  width_full_screen = "1fr",
  max_height = "75px",
  max_height_full_screen = 0.67
)

showcase_bottom(
  width = "100\%",
  width_full_screen = NULL,
  height = "auto",
  height_full_screen = "2fr",
  max_height = "100px",
  max_height_full_screen = NULL
)
}
\arguments{
\item{title, value}{A string, number, or \code{\link[htmltools:builder]{htmltools::tag()}} child to display as
the title or value of the value box. The \code{title} appears above the \code{value}.}

\item{...}{Unnamed arguments may be any \code{\link[htmltools:builder]{htmltools::tag()}} children to
display below \code{value}. Named arguments become attributes on the containing
element.}

\item{showcase}{A \code{\link[htmltools:builder]{htmltools::tag()}} child to showcase (e.g., a
\code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, a \code{plotly::plotlyOutput()}, etc).}

\item{showcase_layout}{One of \code{"left center"} (default), \code{"top right"} or
\code{"bottom"}. Alternatively, you can customize the showcase layout options
with the \code{\link[=showcase_left_center]{showcase_left_center()}}, \code{\link[=showcase_top_right]{showcase_top_right()}}, or
\code{\link[=showcase_bottom]{showcase_bottom()}} functions. Use the options functions when you want to
control the height or width of the showcase area.}

\item{full_screen}{If \code{TRUE}, an icon will appear when hovering over the card
body. Clicking the icon expands the card to fit viewport size.}

\item{theme}{The name of a theme for the value box, or a theme constructed
with \code{value_box_theme()}. The theme names provide a convenient way to use
your app's Bootstrap theme colors as the foreground or background colors of
the value box. See below for more details on the provided themes. For more
control, you can create your own theme with \code{value_box_theme()} where you
can pass foreground and background colors directly. See the \strong{Themes}
section for more details.}

\item{max_height, max_height_full_screen}{A proportion (i.e., a number between
0 and 1) or any valid \link[htmltools:validateCssUnit]{CSS unit} defining the
showcase max_height.}

\item{fill}{Whether to allow the value box to grow/shrink to fit a fillable
container with an opinionated height (e.g., \code{page_fillable()}).}

\item{class}{Utility classes for customizing the appearance of the summary
card. Use \verb{bg-*} and \verb{text-*} classes (e.g, \code{"bg-danger"} and
\code{"text-light"}) to customize the background/foreground colors.}

\item{theme_color}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{theme} instead.}

\item{name}{The name of the theme, e.g. \code{"primary"}, \code{"danger"}, \code{"purple"}.}

\item{bg, fg}{The background and foreground colors for the theme. If only \code{bg}
is provided, then the foreground color is automatically chosen from
\verb{$black} or \verb{$white} to provide the best contrast with the background
color.}

\item{width, width_full_screen, height, height_full_screen}{one of the
following:
\itemize{
\item A proportion (i.e., a number between 0 and 1) of available width or
height to allocate to the showcase.
\item A valid \link[htmltools:validateCssUnit]{CSS unit} defining the width or
height of the showcase column, or a valid value accepted by the
\code{grid-template-columns} (width) or \code{grid-template-rows} (height) CSS
property to define the width or height of the showcase column or row.
Accepted values in the second category are \code{"auto"}, \code{"min-content"},
\code{"max-content"}, a fractional unit (e.g. \verb{2fr}), or a \code{minmax()} function
(e.g., \verb{minmax(100px, 1fr)}).
}}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

An opinionated (\code{\link[=card]{card()}}-powered) box, designed for displaying a \code{value} and
\code{title}. Optionally, a \code{showcase} can provide for context for what the
\code{value} represents (for example, it could hold a \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, or
even a \code{\link[shiny:plotOutput]{shiny::plotOutput()}}). Find examples and template code you can use to
create engaging \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html}{value boxes on the bslib website}.
}
\section{Build-a-Box App}{


Explore all of the \code{value_box()} options and layouts interactively with the
\href{https://bslib.shinyapps.io/build-a-box/}{Build-a-Box app}, available
online thanks to \href{https://www.shinyapps.io/}{shinyapps.io}. Or, you can
run the app locally with:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{shiny::runApp(system.file("examples", "build-a-box", package = "bslib"))
}\if{html}{\out{</div>}}
}

\section{Themes}{


The appearance of a \code{value_box()}
can be controlled via the \code{theme} argument
in one of two ways:
\enumerate{
\item a character value describing the theme,
such as \code{theme = "primary"} or \code{theme = "blue"}; or
\item \code{theme = value_box_theme()} to create a custom theme.
}

We recommend using named themes for most value boxes (the first approach),
because these themes will automatically match your Bootstrap theme.
\subsection{Named themes}{

Bootstrap provides a list of
\href{https://getbootstrap.com/docs/5.3/customize/color/#theme-colors}{theme colors},
with semantic names like \code{"primary"}, \code{"secondary"}, \code{"success"}, \code{"danger"}, etc.
You can set \code{theme} to one of these names to use the corresponding theme color as the background color of your value box.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "primary"
)
}\if{html}{\out{</div>}}

\figure{value-box-named-theme.png}{A value box using 'primary', a named Bootstrap theme color.}

Bootstrap's theme colors are drawn from
\href{https://getbootstrap.com/docs/5.3/customize/color/#all-colors}{a second color list}
that includes variations on several main colors,
named literally.
These colors include \code{"blue"}, \code{"purple"}, \code{"pink"}, \code{"red"}, \code{"orange"},
\code{"yellow"}, \code{"green"}, \code{"teal"}, and \code{"cyan"}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "teal"
)
}\if{html}{\out{</div>}}

\figure{value-box-named-color.png}{A value box using 'purple', a named Bootstrap main color.}
}

\subsection{Background colors}{

If the theme or color name is provided without any prefix, the color will be used for the background of the value box.
You can also explicitly prefix the theme or color name with \verb{bg-} to indicate that it should apply to the value box background.
When the theme sets the background color, either black or white is chosen automatically for the text color using Bootstrap's color contrast algorithm.

As before, you can reference semantic theme color names or literal color names.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-success"
)
}\if{html}{\out{</div>}}

\figure{value-box-background-theme.png}{A value box using 'bg-success', a named Bootstrap theme color.}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-purple"
)
}\if{html}{\out{</div>}}

\figure{value-box-background-color.png}{A value box using 'bg-orange', a named Bootstrap main color.}
}

\subsection{Foreground colors}{

To set only the foreground colors of the value box, you can prefix the theme or color name with \verb{text-}.
This changes the text color without affecting the background color.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "text-success"
)
}\if{html}{\out{</div>}}

\figure{value-box-text-theme.png}{A value box using 'text-success', a named Bootstrap theme color.}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "text-purple"
)
}\if{html}{\out{</div>}}

\figure{value-box-text-color.png}{A value box using 'text-orange', a named Bootstrap main color.}

Occasionally you may want to adjust use both background and foreground themes on your value box.
To achieve this, set \code{theme} to one of the theme names and use \code{class} for the complementary style.
The example below uses \code{theme = "purple"} (which could also be \code{"bg-purple"}) for a purple background, and \code{class = "text-light"} for light-colored text.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "purple",
  class = "text-light"
)
}\if{html}{\out{</div>}}

\figure{value-box-theme-class.png}{A value box using 'purple' for the background and 'text-light' for the text color.}
}

\subsection{Gradient backgrounds}{

For a vibrant and attention-grabbing effect, bslib provides an array of gradient
background options.
Provide \code{theme} with a theme name in the form \code{bg-gradient-{from}-{to}}, where \code{{from}} and \code{{to}} are named main colors, e.g. \code{bg-gradient-indigo-blue}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-gradient-indigo-blue"
)
}\if{html}{\out{</div>}}

\figure{value-box-gradient-1.png}{A value box using 'bg-gradient-blue-purple', a gradient background between two named Bootstrap colors.}
}

\subsection{Custom colors}{

Finally, for complete customization, you can use \code{value_box_theme()} to create a custom theme.
This function takes arguments \code{bg} and \code{fg} to set the background and foreground colors, respectively.
Like with the \verb{bg-} theme names, if only \code{bg} is provided, \code{value_box_theme()} will choose an appropriate light or dark color for the text color.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = value_box_theme(bg = "#e6f2fd", fg = "#0B538E"),
  class = "border"
)
}\if{html}{\out{</div>}}

\figure{value-box-custom.png}{A value box using 'value_box_theme()' for complete customization.}

Note that \code{value_box_theme()} optionally takes a \code{theme} name, which can be helpful if you want to use a named theme and modify the default \code{bg} or \code{fg} colors of that theme.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box_theme(name = "orange", bg = "#FFFFFF")
value_box_theme(name = "text-danger", fg = "#FFB6C1")
}\if{html}{\out{</div>}}

Also note that \code{bg}/\code{fg} \emph{must} be CSS colors, not Bootstrap theme or color names.
This means that \code{theme = "purple"} will use your Bootstrap theme's purple color, and \code{bg = "purple"} will use the CSS color for \emph{purple}, i.e. \code{"#800080"}.
}
}

\section{Showcase Layouts}{


Use the \code{showcase} argument to add a plot or icon to your \code{value_box()}.
There are three layouts available: \code{"left center"}, \code{"top right"}, and \code{"bottom"}.
Set \code{showcase} to the name of the layout you'd like, or use the \code{showcase_left_center()}, \code{showcase_top_right()}, or \code{showcase_bottom()} helper functions to customize the showcase area's size.

If you're using a plot as your showcase, you may also want to set \code{fullscreen = TRUE} so that your users can expand the value box into a full screen card.
See the \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html}{value box article} for more details.
\subsection{Left-center showcase}{

The \code{"left center"} showcase layout is the default, and is perfect for an icon or a small plot.
This layout works best for short value boxes.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = bsicons::bs_icon("ev-station-fill")
)
}\if{html}{\out{</div>}}

\figure{value-box-showcase-left-center.png}{A value box using the left-center showcase option.}
}

\subsection{Top-right showcase}{

The \code{"top right"} showcase layout places the icon or plot in the upper right corner of the value box.
This layout works best for medium-height to square value boxes.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = bsicons::bs_icon("ev-station-fill"),
  showcase_layout = "top right"
)
}\if{html}{\out{</div>}}

\figure{value-box-showcase-top-right.png}{A value box using the top-right showcase option.}
}

\subsection{Bottom showcase}{

Finally, the \code{"bottom"} showcase layout is perfect for full-bleed plots.
This layout places the plot below the title and value, with the plot taking up the full width of the bottom half.

Try this layout with sparkline-style plots.
These can be a little tricky to set up, so be sure to check out the \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html#expandable-sparklines}{Expandable sparklines} section of the \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html}{value boxes article} on the bslib website.
In this example, we've created a sparkline plot using base R graphics, which isn't generally recommended.
\if{latex}{View the bslib documentation online to see the source of \verb{sparkline_plot()}.}

\if{html}{\out{<details><summary>Code for a sparkline plot with base R</summary><br>  <pre><code>set.seed(4242)<br><br>random_sparkline_plot &lt;- function() {<br>  timeseries &lt;- cumsum(runif(100, -2, 2))<br>  x_axis &lt;- seq_along(timeseries)<br>  x_lim &lt;- c(1, length(timeseries))<br>  y_lim &lt;- range(timeseries) + c(-2, 0)<br><br>  par(mar = c(0, 0, 0, 0))<br><br>  # Set up the plot area<br>  plot(<br>    timeseries, type = "n",<br>    axes = FALSE, frame.plot = FALSE,<br>    ylim = y_lim, xlim = x_lim,<br>    ylab = "",    xlab = "",<br>    yaxs = "i",   xaxs = "i",<br>  )<br><br>  # Add the sparkline line<br>  lines(timeseries, type = "l", pch = NA, col = "#0B538E", lwd = 3)<br><br>  # Create polygon coordinates for shading<br>  polygon_x &lt;- c(1, x_axis, length(timeseries))<br>  polygon_y &lt;- c(min(y_lim), timeseries, min(y_lim))<br><br>  # Add shading under the line<br>  polygon(polygon_x, polygon_y, col = "#e6f2fd", border = NA)<br>}<br><br>sparkline_plot &lt;- function() {<br>  as_fill_item(<br>    htmltools::plotTag(<br>      random_sparkline_plot(),<br>      width = 500,<br>      height = 125,<br>      suppressSize = "xy",<br>      alt = paste(<br>        "A sparkline plot with a randomly-generated timeseries.",<br>        "The timeseries starts high and ends low, with lots of variation."<br>      )<br>    )<br>  )<br>}</code></pre></details>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = sparkline_plot(),
  showcase_layout = "bottom"
)
}\if{html}{\out{</div>}}

\figure{value-box-showcase-bottom.png}{A value box using the top-right showcase option.}
}
}

\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
library(htmltools)

value_box(
  "KPI Title",
  h1(HTML("$1 <i>Billion</i> Dollars")),
  span(
    bsicons::bs_icon("arrow-up"),
    " 30\% VS PREVIOUS 30 DAYS"
  ),
  showcase = bsicons::bs_icon("piggy-bank"),
  theme = "success"
)
\dontshow{\}) # examplesIf}
}
\references{
Value boxes are featured on the bslib website in a few articles:
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html}{Value boxes}
\item \href{https://bslib.shinyapps.io/build-a-box}{Build-a-Box App}
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html#value-boxes}{Get Started: Dashboards}
}
}
\seealso{
Value boxes are a specialized form of a \code{\link[=card]{card()}} component.

\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} help position multiple
value boxes into columns and rows.

Other Components: 
\code{\link{accordion}()},
\code{\link{card}()},
\code{\link{popover}()},
\code{\link{tooltip}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bslib-package.R
\docType{package}
\name{bslib-package}
\alias{bslib}
\alias{bslib-package}
\title{bslib: Custom 'Bootstrap' 'Sass' Themes for 'shiny' and 'rmarkdown'}
\description{
Simplifies custom 'CSS' styling of both 'shiny' and 'rmarkdown' via 'Bootstrap' 'Sass'. Supports 'Bootstrap' 3, 4 and 5 as well as their various 'Bootswatch' themes. An interactive widget is also provided for previewing themes in real time.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rstudio.github.io/bslib/}
  \item \url{https://github.com/rstudio/bslib}
  \item Report bugs at \url{https://github.com/rstudio/bslib/issues}
}

}
\author{
\strong{Maintainer}: Carson Sievert \email{carson@posit.co} (\href{https://orcid.org/0000-0002-4958-2844}{ORCID})

Authors:
\itemize{
  \item Joe Cheng \email{joe@posit.co}
  \item Garrick Aden-Buie \email{garrick@posit.co} (\href{https://orcid.org/0000-0002-7111-0077}{ORCID})
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
  \item  Bootstrap contributors (Bootstrap library) [contributor]
  \item  Twitter, Inc (Bootstrap library) [copyright holder]
  \item Javi Aguilar (Bootstrap colorpicker library) [contributor, copyright holder]
  \item Thomas Park (Bootswatch library) [contributor, copyright holder]
  \item  PayPal (Bootstrap accessibility plugin) [contributor, copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preset-bootswatch.R
\name{theme_version}
\alias{theme_version}
\title{Obtain a theme's Bootstrap version}
\usage{
theme_version(theme)
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}
}
\value{
Returns the major version of Bootstrap used in the \code{theme}.
}
\description{
Obtain a theme's Bootstrap version
}
\seealso{
Other Bootstrap theme utility functions: 
\code{\link{bootswatch_themes}()},
\code{\link{bs_get_variables}()},
\code{\link{builtin_themes}()},
\code{\link{theme_bootswatch}()},
\code{\link{versions}()}
}
\concept{Bootstrap theme utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nav-update.R
\name{nav_select}
\alias{nav_select}
\alias{nav_insert}
\alias{nav_remove}
\alias{nav_show}
\alias{nav_hide}
\title{Dynamically update nav containers}
\usage{
nav_select(id, selected = NULL, session = get_current_session())

nav_insert(
  id,
  nav,
  target = NULL,
  position = c("after", "before"),
  select = FALSE,
  session = get_current_session()
)

nav_remove(id, target, session = get_current_session())

nav_show(id, target, select = FALSE, session = get_current_session())

nav_hide(id, target, session = get_current_session())
}
\arguments{
\item{id}{a character string used to identify the nav container.}

\item{selected}{a character string used to identify a particular
\code{\link[=nav_panel]{nav_panel()}} item.}

\item{session}{a shiny session object (the default should almost always be
used).}

\item{nav}{a \code{\link[=nav_panel]{nav_panel()}} item.}

\item{target}{The \code{value} of an existing \code{nav_panel()} item, next to which tab will be added. If removing: the \code{value} of the \code{nav_panel()} item that you want to remove.}

\item{position}{Should \code{nav} be added before or after the target?}

\item{select}{Should \code{nav} be selected upon being inserted?}
}
\description{
Functions for dynamically updating nav containers (e.g., select, insert, and
remove nav items). These functions require an \code{id} on the nav container to be
specified and must be called within an active Shiny session.
}
\examples{

can_browse <- function() rlang::is_interactive() && require("shiny")

# Selecting a tab
if (can_browse()) {
  shinyApp(
    page_fluid(
      radioButtons("item", "Choose", c("A", "B")),
      navset_hidden(
        id = "container",
        nav_panel_hidden("A", "a"),
        nav_panel_hidden("B", "b")
      )
    ),
    function(input, output) {
      observe(nav_select("container", input$item))
    }
  )
}

# Inserting and removing
if (can_browse()) {
  ui <- page_fluid(
    actionButton("add", "Add 'Dynamic' tab"),
    actionButton("remove", "Remove 'Foo' tab"),
    navset_tab(
      id = "tabs",
      nav_panel("Hello", "hello"),
      nav_panel("Foo", "foo"),
      nav_panel("Bar", "bar tab")
    )
  )
  server <- function(input, output) {
    observeEvent(input$add, {
      nav_insert(
        "tabs", target = "Bar", select = TRUE,
        nav_panel("Dynamic", "Dynamically added content")
      )
    })
    observeEvent(input$remove, {
      nav_remove("tabs", target = "Foo")
    })
  }
  shinyApp(ui, server)
}

}
\seealso{
\link[=navset]{Navset functions} create the navigation container holding
the nav panels.

\code{\link[=nav_panel]{nav_panel()}}, \code{\link[=nav_panel_hidden]{nav_panel_hidden()}} create panels of content.

\code{\link[=nav_menu]{nav_menu()}}, \code{\link[=nav_item]{nav_item()}}, \code{\link[=nav_spacer]{nav_spacer()}} create menus, items, or
space in the navset control area.

Other Panel container functions: 
\code{\link{nav-items}},
\code{\link{navset}}
}
\concept{Panel container functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-submit.R
\name{input_submit_textarea}
\alias{input_submit_textarea}
\alias{update_submit_textarea}
\title{Create a textarea input control with explicit submission}
\usage{
input_submit_textarea(
  id,
  label = NULL,
  ...,
  placeholder = NULL,
  value = "",
  width = "min(680px, 100\%)",
  rows = 1,
  button = NULL,
  toolbar = NULL,
  submit_key = c("enter+modifier", "enter")
)

update_submit_textarea(
  id,
  ...,
  value = NULL,
  placeholder = NULL,
  label = NULL,
  submit = FALSE,
  focus = FALSE,
  session = get_current_session()
)
}
\arguments{
\item{id}{The input ID.}

\item{label}{The label to display above the input control. If \code{NULL}, no
label is displayed.}

\item{...}{Additional attributes to apply to the underlying \verb{<textarea>}
element (e.g., spellcheck, autocomplete, etc).}

\item{placeholder}{The placeholder text for the user input.}

\item{value}{The value to set the user input to.}

\item{width}{Any valid CSS unit (e.g., \code{width="100\%"}).}

\item{rows}{The number of rows (i.e., height) of the textarea. This essentially
sets the minimum height -- the textarea can grow taller as the user
enters more text.}

\item{button}{A \link[htmltools:builder]{htmltools::tags} element to use for the submit button. It's recommended
that this be a \code{\link[=input_task_button]{input_task_button()}} since it will automatically provide a
busy indicator (and disable) until the next flush occurs. Note also that if
the submit button launches a \link[shiny:ExtendedTask]{shiny::ExtendedTask}, this button can also be bound
to the task (\code{\link[=bind_task_button]{bind_task_button()}}) and/or manually updated for more
accurate progress reporting (\code{\link[=update_task_button]{update_task_button()}}).}

\item{toolbar}{A list of optional UI elements (e.g., links, icons) to
display next to the submit button.}

\item{submit_key}{A character string indicating what keyboard event should
trigger the submit button. The default is \code{enter+modifier}, which requires
the user to hold down Ctrl (or Cmd on Mac) before pressing Enter to
submit. This helps prevent accidental submissions. To allow submission with
just the Enter key, use \code{enter}. In this case, the user can still insert
new lines using Shift+Enter or Alt+Enter.}

\item{submit}{Whether to automatically submit the text for the user. Requires \code{value}.}

\item{focus}{Whether to move focus to the input element. Requires \code{value}.}

\item{session}{The \code{session} object; using the default is recommended.}
}
\value{
A textarea input control that can be added to a UI definition.
}
\description{
Creates a textarea input where users can enter multi-line text and submit
their input using a dedicated button or keyboard shortcut. This control is
ideal when you want to capture finalized input, rather than reacting to every
keystroke, making it useful for chat boxes, comments, or other scenarios
where users may compose and review their text before submitting.
}
\section{Server value}{

The server receives a character string containing the user's text input.

\strong{Important:} The initial server value is always \code{""} (empty string),
regardless of any \code{value} parameter provided to \code{input_submit_textarea()}.
The server value updates only when the user explicitly submits the input by
either pressing the Enter key (possibly with a modifier key) or clicking the
submit button.
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)

ui <- page_fluid(
  input_submit_textarea("text", placeholder = "Enter some input..."),
  verbatimTextOutput("value")
)
server <- function(input, output) {
  output$value <- renderText({
    req(input$text)
    Sys.sleep(2)
    paste("You entered:", input$text)
  })
}
shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=update_submit_textarea]{update_submit_textarea()}}, \code{\link[=input_task_button]{input_task_button()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preview.R
\name{bs_theme_preview}
\alias{bs_theme_preview}
\title{Preview a Bootstrap theme}
\usage{
bs_theme_preview(theme = bs_theme(), ..., with_themer = TRUE)
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{...}{passed along to \code{\link[shiny:runApp]{shiny::runApp()}}.}

\item{with_themer}{whether or not to run the app with \code{\link[=run_with_themer]{run_with_themer()}}.}
}
\value{
nothing, this function is called for its side-effects (launching an
application).
}
\description{
Launches an example shiny app that can be used to get a quick preview of a
\code{\link[=bs_theme]{bs_theme()}}, as well as an interactive GUI for tweaking some of the
main theme settings. Calling \code{bs_theme_preview()} with no arguments starts
the theme preview app with the default theme, which is a great way to see
the available theme presets or to start creating your own theme.
}
\details{
The app that this launches is subject to change as new features are
developed in \pkg{bslib} and \pkg{shiny}.
}
\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
theme <- bs_theme(bg = "#6c757d", fg = "white", primary = "orange")
bs_theme_preview(theme)
\dontshow{\}) # examplesIf}
}
\seealso{
Use \code{\link[=run_with_themer]{run_with_themer()}} or \code{\link[=bs_themer]{bs_themer()}} to add the theming UI to
an existing shiny app.

Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme_dependencies}()},
\code{\link{bs_theme}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page.R
\name{page}
\alias{page}
\alias{page_fluid}
\alias{page_fixed}
\title{Modern Bootstrap page layouts}
\usage{
page(..., title = NULL, theme = bs_theme(), lang = NULL)

page_fluid(..., title = NULL, theme = bs_theme(), lang = NULL)

page_fixed(..., title = NULL, theme = bs_theme(), lang = NULL)
}
\arguments{
\item{...}{UI elements for the page. Named arguments become HTML attributes.}

\item{title}{The browser window title (defaults to the host URL of the page)}

\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{lang}{ISO 639-1 language code for the HTML page, such as "en" or "ko".
This will be used as the lang in the \code{<html>} tag, as in \code{<html lang="en">}.
The default (NULL) results in an empty string.}
}
\description{
These functions are small wrappers around shiny's page constructors (i.e.,
\code{\link[shiny:fluidPage]{shiny::fluidPage()}}, \code{\link[shiny:navbarPage]{shiny::navbarPage()}}, etc) that differ in two ways:
\itemize{
\item The \code{theme} parameter defaults bslib's recommended version of Bootstrap (for new projects).
\item The return value is rendered as an static HTML page when printed interactively at the console.
}
}
\section{Functions}{
\itemize{
\item \code{page()}: A \pkg{bslib} wrapper for \code{\link[shiny:bootstrapPage]{shiny::bootstrapPage()}}, a basic
Boostrap page where the content is added directly to the page body.

\item \code{page_fluid()}: A \pkg{bslib} wrapper for \code{\link[shiny:fluidPage]{shiny::fluidPage()}}, a fluid
Bootstrap-based page layout that extends to the full viewport width.

\item \code{page_fixed()}: A \pkg{bslib} wrapper for \code{\link[shiny:fixedPage]{shiny::fixedPage()}}, a fixed
Bootstrap-based page layout where the page content container is centered
horizontally and its width is constrained.

}}
\seealso{
Dashboard-style pages: \code{\link[=page_sidebar]{page_sidebar()}}, \code{\link[=page_navbar]{page_navbar()}},
\code{\link[=page_fillable]{page_fillable()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card.R
\name{card}
\alias{card}
\title{A Bootstrap card component}
\usage{
card(
  ...,
  full_screen = FALSE,
  height = NULL,
  max_height = NULL,
  min_height = NULL,
  fill = TRUE,
  class = NULL,
  wrapper = card_body,
  id = NULL
)
}
\arguments{
\item{...}{Unnamed arguments can be any valid child of an \link[htmltools:builder]{htmltools tag} (which includes card items such as \code{\link[=card_body]{card_body()}}.
Named arguments become HTML attributes on returned UI element.}

\item{full_screen}{If \code{TRUE}, an icon will appear when hovering over the card
body. Clicking the icon expands the card to fit viewport size.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{max_height, min_height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{max_height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{max_height} in \code{\link[=card_body]{card_body()}}).}

\item{fill}{Whether or not to allow the card to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{class}{Additional CSS classes for the returned UI element.}

\item{wrapper}{A function (which returns a UI element) to call on unnamed
arguments in \code{...} which are not already card item(s) (like
\code{\link[=card_header]{card_header()}}, \code{\link[=card_body]{card_body()}}, etc.). Note that non-card items are grouped
together into one \code{wrapper} call (e.g. given \code{card("a", "b", card_body("c"), "d")}, \code{wrapper} would be called twice, once with \code{"a"} and
\code{"b"} and once with \code{"d"}).}

\item{id}{Provide a unique identifier for the \code{card()} or \code{value_box()} to
report its full screen state to Shiny. For example, using \code{id = "my_card"},
you can observe the card's full screen state with
\code{input$my_card_full_screen}.}
}
\value{
A \code{\link[htmltools:builder]{htmltools::div()}} tag.
}
\description{
A general purpose container for grouping related UI elements together with a
border and optional padding. To learn more about \code{\link[=card]{card()}}s, see \href{https://rstudio.github.io/bslib/articles/cards/index.html}{the Cards article} or the
other articles listed in the \emph{References} section below.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(htmltools)

card(
  full_screen = TRUE,
  card_header(
    "This is the header"
  ),
  card_body(
    p("This is the body."),
    p("This is still the body.")
  ),
  card_footer(
    "This is the footer"
  )
)
\dontshow{\}) # examplesIf}
}
\references{
Several articles on the bslib website feature the card component:
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/cards/index.html}{Cards}
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html}{Get Started: Dashboards}
\item \href{https://rstudio.github.io/bslib/articles/any-project/index.html}{Get Started: Any Project}
\item \href{https://rstudio.github.io/bslib/articles/column-layout/index.html}{Column-based layouts}
\item \href{https://rstudio.github.io/bslib/articles/filling/index.html#full-screen-cards}{Filling layouts: Full-screen cards}
}
}
\seealso{
\link[=card_body]{Card item functions} create the various parts of a
card.

\code{\link[=navset_card_tab]{navset_card_tab()}}, \code{\link[=navset_card_pill]{navset_card_pill()}} and
\code{\link[=navset_card_underline]{navset_card_underline()}} create cards with tabbed navigation.

\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} help position multiple
cards into columns and rows and can also be used inside a card.

\code{\link[=layout_sidebar]{layout_sidebar()}} adds a sidebar to a card when nested in \code{\link[=card]{card()}}
or \code{\link[=card_body]{card_body()}}.

\code{\link[=value_box]{value_box()}} uses \code{\link[=card]{card()}} to highlight a showcase a key piece of
information.

Other Components: 
\code{\link{accordion}()},
\code{\link{popover}()},
\code{\link{tooltip}()},
\code{\link{value_box}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layout.R
\name{layout_columns}
\alias{layout_columns}
\title{Responsive 12-column grid layouts}
\usage{
layout_columns(
  ...,
  col_widths = NA,
  row_heights = NULL,
  fill = TRUE,
  fillable = TRUE,
  gap = NULL,
  class = NULL,
  height = NULL,
  min_height = NULL,
  max_height = NULL
)
}
\arguments{
\item{...}{Unnamed arguments should be UI elements (e.g., \code{\link[=card]{card()}}). Named
arguments become attributes on the containing \link[htmltools:builder]{htmltools::tag} element.}

\item{col_widths}{One of the following:
\itemize{
\item \code{NA} (the default): Automatically determines a sensible number of columns
based on the number of children.
\item A numeric vector of integers between 1 and 12, where each element
represents the number of columns for the relevant UI element. Elements
that happen to go beyond 12 columns wrap onto the next row. For example,
\code{col_widths = c(4, 8, 12)} allocates 4 columns to the first element, 8
columns to the second element, and 12 columns to the third element (which
wraps to the next row). Negative values are also allowed, and are treated
as empty columns. For example, \code{col_widths = c(-2, 8, -2)} would allocate
8 columns to an element (with 2 empty columns on either side).
\item A \code{\link[=breakpoints]{breakpoints()}} object, where each breakpoint may be either of the
above.
}}

\item{row_heights}{One of the following:
\itemize{
\item A numeric vector, where each value represents the
\href{https://css-tricks.com/introduction-fr-css-unit/}{fractional unit}
(\code{fr}) height of the relevant row. If there are more rows than values
provided, the pattern will repeat. For example, \code{row_heights = c(1, 2)}
allows even rows to take up twice as much space as odd rows.
\item A list of numeric and \link[htmltools:validateCssUnit]{CSS length units},
where each value represents the height of the relevant row. If more rows
are needed than values provided, the pattern will repeat. For example,
\code{row_heights = list("auto", 1)} allows the height of odd rows to be
driven by its contents and even rows to be
\href{https://css-tricks.com/introduction-fr-css-unit/}{\verb{1fr}}.
\item A character vector/string of \link[htmltools:validateCssUnit]{CSS length units}.
In this case, the value is supplied directly to \code{grid-auto-rows}.
\item A \code{\link[=breakpoints]{breakpoints()}} object, where each breakpoint may be either of the above.
}}

\item{fill}{Whether or not to allow the layout to grow/shrink to fit a
fillable container with an opinionated height (e.g., \code{page_fillable()}).}

\item{fillable}{Whether or not each element is wrapped in a fillable container.}

\item{gap}{A \link[htmltools:validateCssUnit]{CSS length unit} defining the
\code{gap} (i.e., spacing) between elements provided to \code{...}. This argument is only applicable when \code{fillable = TRUE}}

\item{class}{Additional CSS classes for the returned UI element.}

\item{height}{Any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{height="200px"}). Doesn't apply when a card is made \code{full_screen}
(in this case, consider setting a \code{height} in \code{\link[=card_body]{card_body()}}).}

\item{min_height, max_height}{The maximum or minimum height of the layout container.
Can be any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g.,
\code{max_height="200px"}). Use these arguments in filling layouts to ensure that a
layout container doesn't shrink below \code{min_height} or grow beyond \code{max_height}.}
}
\description{
Create responsive, column-based grid layouts, based on a 12-column grid.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
x <- card("A simple card")

page_fillable(
  layout_columns(x, x, x, x)
)

# Or add a list of items, spliced with rlang's `!!!` operator
page_fillable(
 layout_columns(!!!list(x, x, x))
)

page_fillable(
  layout_columns(
    col_widths = c(6, 6, 12),
    x, x, x
  )
)

page_fillable(
  layout_columns(
    col_widths = c(6, 6, -2, 8),
    row_heights = c(1, 3),
    x, x, x
  )
)

page_fillable(
  fillable_mobile = TRUE,
  layout_columns(
    col_widths = breakpoints(
      sm = c(12, 12, 12),
      md = c(6, 6, 12),
      lg = c(4, 4, 4)
    ),
    x, x, x
  )
)
\dontshow{\}) # examplesIf}
}
\references{
\href{https://rstudio.github.io/bslib/articles/column-layout/index.html}{Column-based layouts}
on the bslib website.
}
\seealso{
\code{\link[=breakpoints]{breakpoints()}} for more information on specifying column widths at
responsive breakpoints.

Other Column layouts: 
\code{\link{layout_column_wrap}()}
}
\concept{Column layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-dependencies.R
\name{bs_theme_dependencies}
\alias{bs_theme_dependencies}
\title{Compile Bootstrap Sass with (optional) theming}
\usage{
bs_theme_dependencies(
  theme,
  sass_options = sass::sass_options_get(output_style = "compressed"),
  cache = sass::sass_cache_get(),
  jquery = jquerylib::jquery_core(3),
  precompiled = get_precompiled_option("bslib.precompiled", default = TRUE)
)
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{sass_options}{a \code{\link[sass:sass_options]{sass::sass_options()}} object.}

\item{cache}{This can be a directory to use for the cache, a \link[sass]{FileCache}
object created by \code{\link[sass:sass_file_cache]{sass_file_cache()}}, or \code{FALSE} or \code{NULL} for no caching.}

\item{jquery}{a \code{\link[jquerylib:jquery_core]{jquerylib::jquery_core()}} object.}

\item{precompiled}{Before compiling the theme object, first look for a
precompiled CSS file for the \code{\link[=theme_version]{theme_version()}}.  If \code{precompiled = TRUE}
and a precompiled CSS file exists for the theme object, it will be fetched
immediately and not compiled. At the moment, we only provide precompiled
CSS for "stock" builds of Bootstrap (i.e., no theming additions, Bootswatch
themes, or non-default \code{sass_options}).}
}
\value{
Returns a list of HTML dependencies containing Bootstrap CSS,
Bootstrap JavaScript, and \code{jquery}. This list may contain additional HTML
dependencies if bundled with the \code{theme}.
}
\description{
\code{bs_theme_dependencies()} compiles Bootstrap Sass into CSS and returns it,
along with other HTML dependencies, as a list of
\code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}}s. Most users won't need to call this function
directly as Shiny & R Markdown will perform this compilation automatically
when handed a \code{\link[=bs_theme]{bs_theme()}}. If you're here looking to create a themeable
component, see \code{\link[=bs_dependency]{bs_dependency()}}.
}
\section{Sass caching and precompilation}{


If Shiny Developer Mode is enabled (by setting \code{options(shiny.devmode = TRUE)} or calling \code{shiny::devmode(TRUE)}), both \pkg{sass} caching and
\pkg{bslib} precompilation are disabled by default; that is, a call to
\code{bs_theme_dependencies(theme)} expands to \code{bs_theme_dependencies(theme, cache = F, precompiled = F)}). This is useful for local development as enabling
caching/precompilation may produce incorrect results if local changes are
made to bslib's source files.
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}

# Function to preview the styling a (primary) Bootstrap button
library(htmltools)
button <- tags$a(class = "btn btn-primary", href = "#", role = "button", "Hello")
preview_button <- function(theme) {
  browsable(tags$body(bs_theme_dependencies(theme), button))
}

# Latest Bootstrap
preview_button(bs_theme())
# Bootstrap 3
preview_button(bs_theme(3))
# Bootswatch 4 minty theme
preview_button(bs_theme(4, bootswatch = "minty"))
# Bootswatch 4 sketchy theme
preview_button(bs_theme(4, bootswatch = "sketchy"))
\dontshow{\}) # examplesIf}
}
\seealso{
Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme}()},
\code{\link{bs_theme_preview}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{deprecated}
\alias{deprecated}
\alias{bs_theme_new}
\alias{bs_theme_clear}
\alias{bs_theme_get}
\alias{bs_theme_set}
\alias{bs_theme_base_colors}
\alias{bs_theme_accent_colors}
\alias{bs_theme_fonts}
\alias{bs_theme_add_variables}
\alias{bs_theme_add}
\alias{bs_theme_get_variables}
\alias{bootstrap}
\alias{bootstrap_sass}
\alias{bs_add_declarations}
\alias{card_body_fill}
\alias{page_fill}
\alias{sidebar_toggle}
\alias{nav}
\alias{nav_content}
\alias{navs_tab}
\alias{navs_pill}
\alias{navs_pill_list}
\alias{navs_hidden}
\alias{navs_bar}
\alias{navs_tab_card}
\alias{navs_pill_card}
\title{Deprecated functions}
\usage{
bs_theme_new(...)

bs_theme_clear(...)

bs_theme_get(...)

bs_theme_set(...)

bs_theme_base_colors(...)

bs_theme_accent_colors(...)

bs_theme_fonts(...)

bs_theme_add_variables(...)

bs_theme_add(...)

bs_theme_get_variables(...)

bootstrap(theme = bs_theme_get(), ...)

bootstrap_sass(rules = list(), theme = bs_theme_get(), ...)

bs_add_declarations(theme, declarations)

card_body_fill(...)

page_fill(...)

sidebar_toggle(id, open = NULL, session = get_current_session())

nav(...)

nav_content(...)

navs_tab(...)

navs_pill(...)

navs_pill_list(...)

navs_hidden(...)

navs_bar(...)

navs_tab_card(...)

navs_pill_card(...)
}
\value{
a \code{\link[=bs_theme]{bs_theme()}} object.
}
\description{
These functions have been deprecated but remain for backwards compatibility.
}
\section{v0.9.1}{

The function \code{sidebar_toggle()} is now deprecated in v0.9.1. Please use
\code{\link[=toggle_sidebar]{toggle_sidebar()}} instead.
}

\section{Navigation Containers}{


Several functions for navigation containers were renamed in version 0.5.0:
\itemize{
\item \code{nav()} was renamed \code{\link[=nav_panel]{nav_panel()}}
\item \code{nav_content()} was renamed \code{\link[=nav_panel_hidden]{nav_panel_hidden()}}
\item \code{navs_tab()} was renamed \code{\link[=navset_tab]{navset_tab()}}
\item \code{navs_pill()} was renamed \code{\link[=navset_pill]{navset_pill()}}
\item \code{navs_pill_list()} was renamed \code{\link[=navset_pill_list]{navset_pill_list()}}
\item \code{navs_hidden()} was renamed \code{\link[=navset_hidden]{navset_hidden()}}
\item \code{navs_bar()} was renamed \code{\link[=navset_bar]{navset_bar()}}
\item \code{navs_tab_card()} was renamed \code{\link[=navset_card_tab]{navset_card_tab()}}
\item \code{navs_pill_card()} was renamed \code{\link[=navset_card_pill]{navset_card_pill()}}
}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/value-box.R
\name{value_box}
\alias{value_box}
\alias{value_box_theme}
\alias{showcase_left_center}
\alias{showcase_top_right}
\alias{showcase_bottom}
\title{Value box}
\usage{
value_box(
  title,
  value,
  ...,
  showcase = NULL,
  showcase_layout = c("left center", "top right", "bottom"),
  full_screen = FALSE,
  theme = NULL,
  height = NULL,
  max_height = NULL,
  min_height = NULL,
  fill = TRUE,
  class = NULL,
  id = NULL,
  theme_color = deprecated()
)

value_box_theme(name = NULL, bg = NULL, fg = NULL)

showcase_left_center(
  width = 0.3,
  width_full_screen = "1fr",
  max_height = "100px",
  max_height_full_screen = 0.67
)

showcase_top_right(
  width = 0.4,
  width_full_screen = "1fr",
  max_height = "75px",
  max_height_full_screen = 0.67
)

showcase_bottom(
  width = "100\%",
  width_full_screen = NULL,
  height = "auto",
  height_full_screen = "2fr",
  max_height = "100px",
  max_height_full_screen = NULL
)
}
\arguments{
\item{title, value}{A string, number, or \code{\link[htmltools:builder]{htmltools::tag()}} child to display as
the title or value of the value box. The \code{title} appears above the \code{value}.}

\item{...}{Unnamed arguments may be any \code{\link[htmltools:builder]{htmltools::tag()}} children to
display below \code{value}. Named arguments become attributes on the containing
element.}

\item{showcase}{A \code{\link[htmltools:builder]{htmltools::tag()}} child to showcase (e.g., a
\code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, a \code{plotly::plotlyOutput()}, etc).}

\item{showcase_layout}{One of \code{"left center"} (default), \code{"top right"} or
\code{"bottom"}. Alternatively, you can customize the showcase layout options
with the \code{\link[=showcase_left_center]{showcase_left_center()}}, \code{\link[=showcase_top_right]{showcase_top_right()}}, or
\code{\link[=showcase_bottom]{showcase_bottom()}} functions. Use the options functions when you want to
control the height or width of the showcase area.}

\item{full_screen}{If \code{TRUE}, an icon will appear when hovering over the card
body. Clicking the icon expands the card to fit viewport size.}

\item{theme}{The name of a theme for the value box, or a theme constructed
with \code{value_box_theme()}. The theme names provide a convenient way to use
your app's Bootstrap theme colors as the foreground or background colors of
the value box. See below for more details on the provided themes. For more
control, you can create your own theme with \code{value_box_theme()} where you
can pass foreground and background colors directly. See the \strong{Themes}
section for more details.}

\item{max_height}{The maximum height of the \code{value_box()} or the showcase
area when used in a \verb{showcase_layout_*()} function. Can be any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g., \code{max_height="200px"}).}

\item{min_height}{The minimum height of the values box. Can be any valid \link[htmltools:validateCssUnit]{CSS unit} (e.g., \code{min_height="200px"}).}

\item{fill}{Whether to allow the value box to grow/shrink to fit a fillable
container with an opinionated height (e.g., \code{page_fillable()}).}

\item{class}{Utility classes for customizing the appearance of the summary
card. Use \verb{bg-*} and \verb{text-*} classes (e.g, \code{"bg-danger"} and
\code{"text-light"}) to customize the background/foreground colors.}

\item{id}{Provide a unique identifier for the \code{card()} or \code{value_box()} to
report its full screen state to Shiny. For example, using \code{id = "my_card"},
you can observe the card's full screen state with
\code{input$my_card_full_screen}.}

\item{theme_color}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{theme} instead.}

\item{name}{The name of the theme, e.g. \code{"primary"}, \code{"danger"}, \code{"purple"}.}

\item{bg, fg}{The background and foreground colors for the theme. If only \code{bg}
is provided, then the foreground color is automatically chosen from
\verb{$black} or \verb{$white} to provide the best contrast with the background
color.}

\item{width, width_full_screen, height, height_full_screen}{one of the
following:
\itemize{
\item A proportion (i.e., a number between 0 and 1) of available width or
height to allocate to the showcase.
\item A valid \link[htmltools:validateCssUnit]{CSS unit} defining the width or
height of the showcase column, or a valid value accepted by the
\code{grid-template-columns} (width) or \code{grid-template-rows} (height) CSS
property to define the width or height of the showcase column or row.
Accepted values in the second category are \code{"auto"}, \code{"min-content"},
\code{"max-content"}, a fractional unit (e.g. \verb{2fr}), or a \code{minmax()} function
(e.g., \verb{minmax(100px, 1fr)}).
}}

\item{max_height_full_screen}{A proportion (i.e., a number between
0 and 1) or any valid \link[htmltools:validateCssUnit]{CSS unit} defining the
showcase \code{max_height} in a full screen card.}
}
\description{
An opinionated (\code{\link[=card]{card()}}-powered) box, designed for displaying a \code{value} and
\code{title}. Optionally, a \code{showcase} can provide for context for what the
\code{value} represents (for example, it could hold a \code{\link[bsicons:bs_icon]{bsicons::bs_icon()}}, or
even a \code{\link[shiny:plotOutput]{shiny::plotOutput()}}). Find examples and template code you can use to
create engaging \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html}{value boxes on the bslib website}.
}
\section{Build-a-Box App}{


Explore all of the \code{value_box()} options and layouts interactively with the
\href{https://bslib.shinyapps.io/build-a-box/}{Build-a-Box app}, available
online thanks to \href{https://www.shinyapps.io/}{shinyapps.io}. Or, you can
run the app locally with:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# shiny >= 1.8.1
shiny::runExample("build-a-box", package = "bslib")

# shiny < 1.8.1
shiny::runApp(system.file("examples-shiny", "build-a-box", package = "bslib"))
}\if{html}{\out{</div>}}
}

\section{Themes}{


The appearance of a \code{value_box()}
can be controlled via the \code{theme} argument
in one of two ways:
\enumerate{
\item a character value describing the theme,
such as \code{theme = "primary"} or \code{theme = "blue"}; or
\item \code{theme = value_box_theme()} to create a custom theme.
}

We recommend using named themes for most value boxes (the first approach),
because these themes will automatically match your Bootstrap theme.
\subsection{Named themes}{

Bootstrap provides a list of
\href{https://getbootstrap.com/docs/5.3/customize/color/#theme-colors}{theme colors},
with semantic names like \code{"primary"}, \code{"secondary"}, \code{"success"}, \code{"danger"}, etc.
You can set \code{theme} to one of these names to use the corresponding theme color as the background color of your value box.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "primary"
)
}\if{html}{\out{</div>}}

\figure{value-box-named-theme.png}{A value box using 'primary', a named Bootstrap theme color.}

Bootstrap's theme colors are drawn from
\href{https://getbootstrap.com/docs/5.3/customize/color/#all-colors}{a second color list}
that includes variations on several main colors,
named literally.
These colors include \code{"blue"}, \code{"purple"}, \code{"pink"}, \code{"red"}, \code{"orange"},
\code{"yellow"}, \code{"green"}, \code{"teal"}, and \code{"cyan"}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "teal"
)
}\if{html}{\out{</div>}}

\figure{value-box-named-color.png}{A value box using 'purple', a named Bootstrap main color.}
}

\subsection{Background colors}{

If the theme or color name is provided without any prefix, the color will be used for the background of the value box.
You can also explicitly prefix the theme or color name with \verb{bg-} to indicate that it should apply to the value box background.
When the theme sets the background color, either black or white is chosen automatically for the text color using Bootstrap's color contrast algorithm.

As before, you can reference semantic theme color names or literal color names.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-success"
)
}\if{html}{\out{</div>}}

\figure{value-box-background-theme.png}{A value box using 'bg-success', a named Bootstrap theme color.}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-purple"
)
}\if{html}{\out{</div>}}

\figure{value-box-background-color.png}{A value box using 'bg-orange', a named Bootstrap main color.}
}

\subsection{Foreground colors}{

To set only the foreground colors of the value box, you can prefix the theme or color name with \verb{text-}.
This changes the text color without affecting the background color.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "text-success"
)
}\if{html}{\out{</div>}}

\figure{value-box-text-theme.png}{A value box using 'text-success', a named Bootstrap theme color.}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "text-purple"
)
}\if{html}{\out{</div>}}

\figure{value-box-text-color.png}{A value box using 'text-orange', a named Bootstrap main color.}

Occasionally you may want to adjust use both background and foreground themes on your value box.
To achieve this, set \code{theme} to one of the theme names and use \code{class} for the complementary style.
The example below uses \code{theme = "purple"} (which could also be \code{"bg-purple"}) for a purple background, and \code{class = "text-light"} for light-colored text.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "purple",
  class = "text-light"
)
}\if{html}{\out{</div>}}

\figure{value-box-theme-class.png}{A value box using 'purple' for the background and 'text-light' for the text color.}
}

\subsection{Gradient backgrounds}{

For a vibrant and attention-grabbing effect, bslib provides an array of gradient
background options.
Provide \code{theme} with a theme name in the form \code{bg-gradient-{from}-{to}}, where \code{{from}} and \code{{to}} are named main colors, e.g. \code{bg-gradient-indigo-blue}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = "bg-gradient-indigo-blue"
)
}\if{html}{\out{</div>}}

\figure{value-box-gradient-1.png}{A value box using 'bg-gradient-blue-purple', a gradient background between two named Bootstrap colors.}
}

\subsection{Custom colors}{

Finally, for complete customization, you can use \code{value_box_theme()} to create a custom theme.
This function takes arguments \code{bg} and \code{fg} to set the background and foreground colors, respectively.
Like with the \verb{bg-} theme names, if only \code{bg} is provided, \code{value_box_theme()} will choose an appropriate light or dark color for the text color.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Customer lifetime value",
  value = "$5,000",
  showcase = bsicons::bs_icon("bank2"),
  theme = value_box_theme(bg = "#e6f2fd", fg = "#0B538E"),
  class = "border"
)
}\if{html}{\out{</div>}}

\figure{value-box-custom.png}{A value box using 'value_box_theme()' for complete customization.}

Note that \code{value_box_theme()} optionally takes a \code{theme} name, which can be helpful if you want to use a named theme and modify the default \code{bg} or \code{fg} colors of that theme.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box_theme(name = "orange", bg = "#FFFFFF")
value_box_theme(name = "text-danger", fg = "#FFB6C1")
}\if{html}{\out{</div>}}

Also note that \code{bg}/\code{fg} \emph{must} be CSS colors, not Bootstrap theme or color names.
This means that \code{theme = "purple"} will use your Bootstrap theme's purple color, and \code{bg = "purple"} will use the CSS color for \emph{purple}, i.e. \code{"#800080"}.
}
}

\section{Showcase Layouts}{


Use the \code{showcase} argument to add a plot or icon to your \code{value_box()}.
There are three layouts available: \code{"left center"}, \code{"top right"}, and \code{"bottom"}.
Set \code{showcase} to the name of the layout you'd like, or use the \code{showcase_left_center()}, \code{showcase_top_right()}, or \code{showcase_bottom()} helper functions to customize the showcase area's size.

If you're using a plot as your showcase, you may also want to set \code{fullscreen = TRUE} so that your users can expand the value box into a full screen card.
See the \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html}{value box article} for more details.
\subsection{Left-center showcase}{

The \code{"left center"} showcase layout is the default, and is perfect for an icon or a small plot.
This layout works best for short value boxes.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = bsicons::bs_icon("ev-station-fill")
)
}\if{html}{\out{</div>}}

\figure{value-box-showcase-left-center.png}{A value box using the left-center showcase option.}
}

\subsection{Top-right showcase}{

The \code{"top right"} showcase layout places the icon or plot in the upper right corner of the value box.
This layout works best for medium-height to square value boxes.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = bsicons::bs_icon("ev-station-fill"),
  showcase_layout = "top right"
)
}\if{html}{\out{</div>}}

\figure{value-box-showcase-top-right.png}{A value box using the top-right showcase option.}
}

\subsection{Bottom showcase}{

Finally, the \code{"bottom"} showcase layout is perfect for full-bleed plots.
This layout places the plot below the title and value, with the plot taking up the full width of the bottom half.

Try this layout with sparkline-style plots.
These can be a little tricky to set up, so be sure to check out the \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html#expandable-sparklines}{Expandable sparklines} section of the \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html}{value boxes article} on the bslib website.
In this example, we've created a sparkline plot using base R graphics, which isn't generally recommended.
\if{latex}{View the bslib documentation online to see the source of \verb{sparkline_plot()}.}

\if{html}{\out{<details><summary>Code for a sparkline plot with base R</summary><br>  <pre><code>set.seed(4242)<br><br>random_sparkline_plot &lt;- function() {<br>  timeseries &lt;- cumsum(runif(100, -2, 2))<br>  x_axis &lt;- seq_along(timeseries)<br>  x_lim &lt;- c(1, length(timeseries))<br>  y_lim &lt;- range(timeseries) + c(-2, 0)<br><br>  par(mar = c(0, 0, 0, 0))<br><br>  # Set up the plot area<br>  plot(<br>    timeseries, type = "n",<br>    axes = FALSE, frame.plot = FALSE,<br>    ylim = y_lim, xlim = x_lim,<br>    ylab = "",    xlab = "",<br>    yaxs = "i",   xaxs = "i",<br>  )<br><br>  # Add the sparkline line<br>  lines(timeseries, type = "l", pch = NA, col = "#0B538E", lwd = 3)<br><br>  # Create polygon coordinates for shading<br>  polygon_x &lt;- c(1, x_axis, length(timeseries))<br>  polygon_y &lt;- c(min(y_lim), timeseries, min(y_lim))<br><br>  # Add shading under the line<br>  polygon(polygon_x, polygon_y, col = "#e6f2fd", border = NA)<br>}<br><br>sparkline_plot &lt;- function() {<br>  as_fill_item(<br>    htmltools::plotTag(<br>      random_sparkline_plot(),<br>      width = 500,<br>      height = 125,<br>      suppressSize = "xy",<br>      alt = paste(<br>        "A sparkline plot with a randomly-generated timeseries.",<br>        "The timeseries starts high and ends low, with lots of variation."<br>      )<br>    )<br>  )<br>}</code></pre></details>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{value_box(
  title = "Energy consumption",
  value = "345 kwh/month",
  showcase = sparkline_plot(),
  showcase_layout = "bottom"
)
}\if{html}{\out{</div>}}

\figure{value-box-showcase-bottom.png}{A value box using the top-right showcase option.}
}
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
library(htmltools)

value_box(
  "KPI Title",
  h1(HTML("$1 <i>Billion</i> Dollars")),
  span(
    bsicons::bs_icon("arrow-up"),
    " 30\% VS PREVIOUS 30 DAYS"
  ),
  showcase = bsicons::bs_icon("piggy-bank"),
  theme = "success"
)
\dontshow{\}) # examplesIf}
}
\references{
Value boxes are featured on the bslib website in a few articles:
\itemize{
\item \href{https://rstudio.github.io/bslib/articles/value-boxes/index.html}{Value boxes}
\item \href{https://bslib.shinyapps.io/build-a-box}{Build-a-Box App}
\item \href{https://rstudio.github.io/bslib/articles/dashboards/index.html#value-boxes}{Get Started: Dashboards}
}
}
\seealso{
Value boxes are a specialized form of a \code{\link[=card]{card()}} component.

\code{\link[=layout_columns]{layout_columns()}} and \code{\link[=layout_column_wrap]{layout_column_wrap()}} help position multiple
value boxes into columns and rows.

Other Components: 
\code{\link{accordion}()},
\code{\link{card}()},
\code{\link{popover}()},
\code{\link{tooltip}()}
}
\concept{Components}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bslib-package.R
\docType{package}
\name{bslib-package}
\alias{bslib}
\alias{bslib-package}
\title{bslib: Custom 'Bootstrap' 'Sass' Themes for 'shiny' and 'rmarkdown'}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Simplifies custom 'CSS' styling of both 'shiny' and 'rmarkdown' via 'Bootstrap' 'Sass'. Supports 'Bootstrap' 3, 4 and 5 as well as their various 'Bootswatch' themes. An interactive widget is also provided for previewing themes in real time.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rstudio.github.io/bslib/}
  \item \url{https://github.com/rstudio/bslib}
  \item Report bugs at \url{https://github.com/rstudio/bslib/issues}
}

}
\author{
\strong{Maintainer}: Carson Sievert \email{carson@posit.co} (\href{https://orcid.org/0000-0002-4958-2844}{ORCID})

Authors:
\itemize{
  \item Joe Cheng \email{joe@posit.co}
  \item Garrick Aden-Buie \email{garrick@posit.co} (\href{https://orcid.org/0000-0002-7111-0077}{ORCID})
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
  \item  Bootstrap contributors (Bootstrap library) [contributor]
  \item  Twitter, Inc (Bootstrap library) [copyright holder]
  \item Javi Aguilar (Bootstrap colorpicker library) [contributor, copyright holder]
  \item Thomas Park (Bootswatch library) [contributor, copyright holder]
  \item  PayPal (Bootstrap accessibility plugin) [contributor, copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preview.R
\name{bs_get_variables}
\alias{bs_get_variables}
\alias{bs_get_contrast}
\title{Retrieve Sass variable values from the current theme}
\usage{
bs_get_variables(theme, varnames)

bs_get_contrast(theme, varnames)
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{varnames}{A character string referencing a Sass variable in the current
theme.}
}
\value{
Returns a character string containing a CSS/Sass value. If the
variable(s) are not defined, their value is \code{NA}.
}
\description{
Useful for retrieving a variable from the current theme and using
the value to inform another R function.
}
\examples{
vars <- c("body-bg", "body-color", "primary", "border-radius")
bs_get_variables(bs_theme(), varnames = vars)
bs_get_variables(bs_theme(bootswatch = "darkly"), varnames = vars)

bs_get_contrast(bs_theme(), c("primary", "dark", "light"))

library(htmltools)
div(
  class = "bg-primary",
  style = css(
    color = bs_get_contrast(bs_theme(), "primary")
  )
)

}
\references{
\href{https://rstudio.github.io/bslib/articles/bs5-variables/index.html}{Theming: Bootstrap 5 variables}
provides a searchable reference of all theming variables available in
Bootstrap 5.
}
\seealso{
Other Bootstrap theme utility functions: 
\code{\link{bootswatch_themes}()},
\code{\link{builtin_themes}()},
\code{\link{theme_bootswatch}()},
\code{\link{theme_version}()},
\code{\link{versions}()}
}
\concept{Bootstrap theme utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bs-theme-preview.R
\name{bs_theme_preview}
\alias{bs_theme_preview}
\title{Preview a Bootstrap theme}
\usage{
bs_theme_preview(theme = bs_theme(), ..., with_themer = TRUE)
}
\arguments{
\item{theme}{A \code{\link[=bs_theme]{bs_theme()}} object.}

\item{...}{passed along to \code{\link[shiny:runApp]{shiny::runApp()}}.}

\item{with_themer}{whether or not to run the app with \code{\link[=run_with_themer]{run_with_themer()}}.}
}
\value{
nothing, this function is called for its side-effects (launching an
application).
}
\description{
Launches an example shiny app that can be used to get a quick preview of a
\code{\link[=bs_theme]{bs_theme()}}, as well as an interactive GUI for tweaking some of the
main theme settings. Calling \code{bs_theme_preview()} with no arguments starts
the theme preview app with the default theme, which is a great way to see
the available theme presets or to start creating your own theme.
}
\details{
The app that this launches is subject to change as new features are
developed in \pkg{bslib} and \pkg{shiny}.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
theme <- bs_theme(bg = "#6c757d", fg = "white", primary = "orange")
bs_theme_preview(theme)
\dontshow{\}) # examplesIf}
}
\seealso{
Use \code{\link[=run_with_themer]{run_with_themer()}} or \code{\link[=bs_themer]{bs_themer()}} to add the theming UI to
an existing shiny app.

Other Bootstrap theme functions: 
\code{\link{bs_add_variables}()},
\code{\link{bs_current_theme}()},
\code{\link{bs_dependency}()},
\code{\link{bs_global_theme}()},
\code{\link{bs_remove}()},
\code{\link{bs_theme}()},
\code{\link{bs_theme_dependencies}()}
}
\concept{Bootstrap theme functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nav-update.R
\name{nav_prepend}
\alias{nav_prepend}
\alias{nav_append}
\title{Append or prepend nav item to a dropdown menu}
\usage{
nav_prepend(
  id,
  nav,
  menu_title,
  select = FALSE,
  session = get_current_session()
)

nav_append(
  id,
  nav,
  menu_title,
  select = FALSE,
  session = get_current_session()
)
}
\arguments{
\item{id}{a character string used to identify the nav container.}

\item{nav}{a \code{\link[=nav_panel]{nav_panel()}} item.}

\item{menu_title}{The title of a \code{\link[=nav_menu]{nav_menu()}}.}

\item{select}{Should \code{nav} be selected upon being inserted?}

\item{session}{a shiny session object (the default should almost always be
used).}
}
\description{
Exported for use by \code{\link[shiny:insertTab]{shiny::prependTab()}}/\code{\link[shiny:insertTab]{shiny::appendTab()}}. These
functions have been superseded by \code{\link[=nav_insert]{nav_insert()}} (i.e.,
\code{\link[shiny:insertTab]{shiny::insertTab()}}), since it can do everything these functions do (i.e.,
add a \code{\link[=nav_panel]{nav_panel()}} to the start or end of a \code{\link[=nav_menu]{nav_menu()}}) and more (i.e., insert a
\code{\link[=nav_panel]{nav_panel()}} anywhere inside a nav container).
}
\keyword{internal}
